/***************************************************************************
                        dfg1394_device.c  -  description
                             -------------------
 Copyright (c) 2003 Norbert Wagner <nw@softwarekombinat.de>



    created              : 2003-03-12 Norbert Wagner <nw@softwarekombinat.de>
    modified             : 2003-10-07 Arne Caspari <arne_caspari@eu.theimagingsource.com>
                           - dfg1394_open: client__destroy was missing if no converter was found
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "dfg1394.h"

#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/videodev.h>


extern struct file_operations dfg1394_fops;

/* ************************************************************************ */
/** 
 *   @name Globals                                
 */
/* ************************************************************************ */
/* @{                                                                       */

/** @brief the name of the device */
extern char* dfg1394_name;

/**
 * @brief Helps to define pix_format in a readable way
 */
#define DEFINE_FORMAT(w,h,d,f) \
{	width: (w), height: (h), pixelformat: (f), \
        field: V4L2_FIELD_INTERLACED, bytesperline: 0, \
        sizeimage : (w)*(h)*(d)/8, colorspace: V4L2_COLORSPACE_SMPTE170M, \
        priv: 0 \
}



/**
   @brief All pixel formats of the DFG/1394-1 
 */
static 
struct v4l2_pix_format dfg1394_base_pix_formats[DFG1394_PIX_MODE__COUNT]=
{
	DEFINE_FORMAT( 320, 240,  8, V4L2_PIX_FMT_GREY ), // _MODE_320x240_400
	DEFINE_FORMAT( 320, 240, 12, V4L2_PIX_FMT_Y41P ), // _MODE_320x240_411,
	DEFINE_FORMAT( 320, 240, 16, V4L2_PIX_FMT_UYVY ), // _MODE_320x240_422,
	DEFINE_FORMAT( 640, 480,  8, V4L2_PIX_FMT_GREY ), // _MODE_640x480_400,
	DEFINE_FORMAT( 640, 480, 12, V4L2_PIX_FMT_Y41P ), // _MODE_640x480_411,
	DEFINE_FORMAT( 640, 480, 16, V4L2_PIX_FMT_UYVY ), // _MODE_640x480_422,
	DEFINE_FORMAT( 768, 576,  8, V4L2_PIX_FMT_GREY ), // _MODE_768x576_400,
	DEFINE_FORMAT( 768, 576, 12, V4L2_PIX_FMT_Y41P ), // _MODE_768x576_411,
	DEFINE_FORMAT( 768, 576, 16, V4L2_PIX_FMT_UYVY ), // _MODE_768x576_422,
};

/*@}*/

/* ************************************************************************ */
/** 
 *   @name dfg1394_buffer_t Methods
 */
/* ************************************************************************ */
/* @{                                                                       */

/**
   @brief Initializes the specified capture buffer
 */
inline 
void dfg1394_buffer__initialize( 
	dfg1394_buffer_t* buf, int idx, u32 buflen )
{
	memset( buf, 0, sizeof( dfg1394_buffer_t ));
	
	buf->is_requested = FALSE;
	buf->offset = 0;
	INIT_LIST_HEAD( &buf->list_entry );
	
	buf->v4l2buf.index = idx;
	buf->v4l2buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	buf->v4l2buf.memory = V4L2_MEMORY_MMAP;
	
	/* offset must be unique for each buffer, and a multiple */
	/* of PAGE_SIZE on 2.4.x  */
	buf->v4l2buf.m.offset = idx << PAGE_SHIFT;
	buf->v4l2buf.length = buflen;
	buf->frame_size = 0;
	buf->line_size = 0;
}

/*@}*/

/* ************************************************************************ */
/** 
 *   @name dfg1394_device Management                                
 */
/* ************************************************************************ */
/* @{                                                                       */


/**
 * @brief Creates a new dfg1394_device structure and initializes its members
 *
 * @param device the LDM device
 * @return a new dfg1394_device object
 */
dfg1394_device_t* dfg1394_device__create( struct device* device )
{
	int i;

	/* create and initialize device structure
	 */
	dfg1394_device_t* dev = kmalloc( sizeof(dfg1394_device_t), GFP_ATOMIC);
	memset( dev, 0, sizeof( dfg1394_device_t ));


	/* initialize node
	 */
	dfg1394_node__initialize( &dev->node );


	/* initialize misc stuff
	 */

	dev->client_count = 0;
	spin_lock_init( &dev->lock );
	sema_init( &dev->sema, 1 );
	dev->flags = 0;
	INIT_LIST_HEAD( &dev->list_entry );

	/* initialize LDM device field
         */
	dev->device = get_device( device );
	dev->device->driver_data = dev;
	
	/* initialize video device
	 */
	
	strcpy( dev->vdev.name, dfg1394_name );
	dev->vdev.fops  = &dfg1394_fops;
	dev->vdev.minor = -1;
	dev->vdev.priv  = dev;


	/* initialize device video settings
	 */
	dev->pix_format = dfg1394_base_pix_formats[ dev->node.pix_mode ];

	/* initialize capture related members 
	 */
	for ( i=0; i<DFG1394_STREAM_MAX_BUFFERS; ++i )
		dfg1394_buffer__initialize( &dev->cap_buffers[i], i, 0 );

	INIT_LIST_HEAD( &dev->cap_free_queue );
	INIT_LIST_HEAD( &dev->cap_done_queue );

	init_waitqueue_head( &dev->cap_done_wait );
	spin_lock_init( &dev->cap_lock );	

	return dev;
}

/**
 * @brief Destroys an existing dfg1394_dev structure
 *
 * @param dev       [in]  @c this pointer
 * @return @c SUCCESS or a negative error code
 */
int  dfg1394_device__destroy( dfg1394_device_t* dev )
{
	if (dev->device != NULL)
		put_device( dev->device );

	kfree( dev );
	return SUCCESS;
}


/**
 * @brief Registers a device with the kernel
 *
 * @param dev       [in]  @c this pointer
 * @param nr        [in]  the minor number to register with
 * @return @c SUCCESS or a negative error code
 */
int dfg1394_device__register( dfg1394_device_t* dev, int nr )
{
	int ret;
	/* register device
	 */

	ret = video_register_device( &dev->vdev, VFL_TYPE_GRABBER, nr );
	if ( IS_FAILURE( ret ) ) {
		PRINT_ERR( "failed to register the device: %d.\n", ret);
	} else 
		PRINT_INFO( "device registered (minor %d).\n", nr);

	return ret;
}

/**
 * @brief Unregisters a previously registered device
 *
 * @param dev       [in]  @c this pointer
 * @return @c SUCCESS or a negative error code
 *
 * Unregistering fails if the device still has clients.
 */
int dfg1394_device__unregister( dfg1394_device_t* dev )
{
	int cc;

	spin_lock(&dev->lock);
	cc = dev->client_count;
	spin_unlock(&dev->lock);

	/* bail out in case we still have clients */
	if (cc > 0) 
		return -EBUSY;
		
		
	video_unregister_device( &dev->vdev );
	
	PRINT_NOTICE( "device '%s' unregistered (minor %d).\n", 
		      dev->vdev.name, dev->vdev.minor );
	
	return SUCCESS;

}

/**
 * @brief Make a device a ghost
 *
 * @param dev       [in]  @c this pointer
 * @return @c SUCCESS or a negative error code
 *
 * Ghost devices have no corresponding hardware device,
 * they exist only because they were open when the hardware
 * was removed
 */
int dfg1394_device__ghostify( dfg1394_device_t* dev )
{

	set_bit(DFG1394_DEVICE_FLAG__GHOST, &dev->flags);

	/* kill transmission */
	if (dfg1394_node__is_transmitting(&dev->node))
		dfg1394_node__drop_dead(&dev->node);

	if (dev->device != NULL) {
		put_device(dev->device);
		dev->device = NULL;
	}

	PRINT_INFO("device disembodied - new ghost!\n");

	return SUCCESS;
}

/**
 * @brief Resurrects a ghost device
 *
 * @param dev       [in]  @c this pointer
 * @param device    [in]  The new LDM device object
 * @return @c SUCCESS or a negative error code
 */
int dfg1394_device__resurrect( dfg1394_device_t* dev, struct device* device  )
{
	int cc;

	spin_lock(&dev->lock);
	cc = dev->client_count;
	spin_unlock(&dev->lock);

	/* the device cannot be resurrected if the ghost has still
           something to do  */
	if (cc > 0) 
		return -EBUSY;


	/* initialize node
	 */
	dfg1394_node__initialize( &dev->node );

	/* initialize LDM device field
         */
	dev->device = get_device( device );
	dev->device->driver_data = dev;

	/* initialize capture related members 
	 */
	INIT_LIST_HEAD( &dev->cap_free_queue );
	INIT_LIST_HEAD( &dev->cap_done_queue );

	/* now we live again */
	clear_bit(DFG1394_DEVICE_FLAG__GHOST, &dev->flags);

	PRINT_INFO("ghost device resurrected.\n");
	
	return SUCCESS;
}

/*@}*/


/* ************************************************************************ */
/** 
 *   @name dfg1394_device IOCTL Handlers                                
 *      
 *   Each functions executes on V4L2 command.
 *                                                                    
 *   @warning They may be accessed concurrently on SMP machines,       
 *   so access must be synchronized!           
 */
/* ************************************************************************ */
/* @{                                                                       */


/**
 * @brief @c VIDIOC_QUERYCAP: Returns the device capabilities 
 */
int dfg1394_device__query_cap( 
	dfg1394_device_t* dev, struct v4l2_capability* caps )
{
	/* read-only access to static data, no sync required */

	static struct v4l2_capability dev_caps = 
	{
		driver:   "dfg1394-1",
		card:     "DFG1394 Video Converter",
		bus_info: "\0",
		version:  KERNEL_VERSION(0,5,1),
		capabilities: V4L2_CAP_VIDEO_CAPTURE | V4L2_CAP_STREAMING
		/*|  | V4L2_CAP_READWRITE | V4L2_CAP_ASYNCIO*/
	};

	*caps = dev_caps;
	return SUCCESS;
}

/**
 * @brief @c  VIDIOC_ENUM_STD: Get the device's supported video standards 
 *
 */
int dfg1394_device__enum_standard( 
	dfg1394_device_t* dev, struct v4l2_standard* std )
{
	/* no access to shared data, no sync required */

	switch ( std->index ) {
	case 0:
		v4l2_video_std_construct(std, V4L2_STD_PAL, "PAL" );
		break;
	case 1:
		v4l2_video_std_construct(std, V4L2_STD_NTSC, "NTSC" );
		break;
	default:
		return -EINVAL;
	}
	return SUCCESS;
}

/**
 * @brief @c  VIDIOC_GET_STD: Get currently used standards 
 */
int dfg1394_device__get_standard( dfg1394_device_t* dev, v4l2_std_id* std_id )
{
	dfg1394_freq_t freq;

	freq = dfg1394_node__get_frequency( &dev->node );

	switch ( freq ) {
	case DFG1394_FREQ_PAL_50:
		*std_id = V4L2_STD_PAL;
		break;
	case DFG1394_FREQ_NTSC_60:
		*std_id = V4L2_STD_NTSC;
		break;
	default:
		PRINT_ERR("Invalid frequency set in node: %d\n", freq );
		return -EINVAL;
	}

	return SUCCESS;
}

/**
 * @brief @c  VIDIOC_SET_STD: Set the video standard to be used
 */
int dfg1394_device__set_standard( dfg1394_device_t* dev, v4l2_std_id* std_id )
{
	dfg1394_freq_t freq;
	dfg1394_bool_t cap_is_active;
	int ret;

        /* -------- enter synchronized section -------- */
	if (down_interruptible(&dev->sema)) 
		return -ERESTARTSYS;

	cap_is_active = dfg1394_node__is_transmitting( &dev->node );
	
	if ( *std_id & V4L2_STD_PAL ) 
		freq = DFG1394_FREQ_PAL_50;
	else if ( *std_id & V4L2_STD_NTSC ) 
		freq = DFG1394_FREQ_NTSC_60;
	else
		return -EINVAL;

	if ( cap_is_active ) 
		dfg1394_node__stop_transmission( &dev->node );
	

	ret = dfg1394_node__set_frequency( &dev->node, freq );

	if ( cap_is_active ) {
		dfg1394_node__start_transmission( &dev->node, 0 );
	}

	
	up(&dev->sema); 
        /* -------- leave synchronized section -------- */

	return ret;
}

/**
 * @brief @c  VIDIOC_ENUM_INPUT: Allows to iterate over all inputs 
 */
int dfg1394_device__enum_input( 
	dfg1394_device_t* dev, struct v4l2_input* input )
{
	/* read-only access to static data, no sync required */

	unsigned int idx = input->index;
	static const char* dev_input_names[] = 
		{ "Composite-1", "Composite-2", "SVideo", "Auto Detect" };

	if( idx >= ( sizeof( dev_input_names ) / sizeof( char * ) ) )
	{
		return -EINVAL;
	}
	


	memset( input, 0, sizeof( *input ));
	input->index = idx;
	input->std = V4L2_STD_PAL | V4L2_STD_NTSC;
	input->type = V4L2_INPUT_TYPE_CAMERA;
	strncpy( input->name, dev_input_names[idx], sizeof(input->name) );
	
	return SUCCESS;
}

/**
 * @brief @c  VIDIOC_GET_INPUT: Gets the currently selected input
 */
int dfg1394_device__get_input( dfg1394_device_t* dev, int* input_index )
{
	*input_index = dfg1394_node__get_input(&dev->node);
	switch( *input_index )
	{
		case DFG1394_INPUT_COMPOSITE_4:
			*input_index = 0;
			break;
			
		case DFG1394_INPUT_COMPOSITE_5:
			*input_index = 1;
			break;
			
		case DFG1394_INPUT_SVIDEO_3:
			*input_index = 2;
			break;
			
		default:
			break;
	}

	return SUCCESS;
}

/**
 * @brief @c  VIDIOC_GET_INPUT: Selects a new input
 */
int dfg1394_device__set_input( dfg1394_device_t* dev, int* input_index )
{
	dfg1394_bool_t cap_is_active;
	int input;

	switch( *input_index )
	{
		case 0:
			input = DFG1394_INPUT_COMPOSITE_4;
			break;
			
		case 1:
			input = DFG1394_INPUT_COMPOSITE_5;
			break;
			
		case 2: 
			input = DFG1394_INPUT_SVIDEO_3;
			break;
			
		case 3:
			input = DFG1394_INPUT_DETECT;
			break;
			
		default:
			return -EINVAL;
	}

        /* -------- enter synchronized section -------- */
	if (down_interruptible(&dev->sema)) 
		return -ERESTARTSYS;

	cap_is_active = dfg1394_node__is_transmitting(&dev->node);

	/* stop transmission */
	if ( cap_is_active ) 
		dfg1394_node__stop_transmission( &dev->node );
	/* set input */
	if ( IS_FAILURE( dfg1394_node__set_input( &dev->node, input )))
		return -EAGAIN;
	/* restart transmission */
	if ( cap_is_active ) 
		dfg1394_node__start_transmission( &dev->node, 0 );
	
	up(&dev->sema); 
        /* -------- leave synchronized section -------- */

	dfg1394_device__get_input( dev, input_index );

	return SUCCESS;
}

/**
 * @brief @c  VIDIOC_ENUM_FMT: Allows to iterate over supported video formats
 */
int dfg1394_device__enum_formats( 
	dfg1394_device_t* dev, struct v4l2_fmtdesc* fmt )
{
	/* read-only access to static data, no sync required */

	unsigned int idx = fmt->index;

	if ( fmt->type != V4L2_BUF_TYPE_VIDEO_CAPTURE ) 
		return -EINVAL;

	memset( fmt, 0, sizeof(*fmt) );
	fmt->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	fmt->index = idx;

	switch ( idx ) {
	case 0:
		fmt->pixelformat = V4L2_PIX_FMT_GREY;
		strncpy(fmt->description, 
			"Grayscale", sizeof(fmt->description) );
		break;
	case 1:
		fmt->pixelformat = V4L2_PIX_FMT_Y41P;
		strncpy( fmt->description, 
			 "YUV 4:1:1", sizeof(fmt->description) );
		break;
	case 2:
		fmt->pixelformat = V4L2_PIX_FMT_UYVY;
		strncpy( fmt->description, 
			 "YUV 4:2:2", sizeof(fmt->description) );
		break;
	default:
		return -EINVAL;
	}
	
	return SUCCESS;
}

/**
 * @brief @c  VIDIOC_GET_FMT: Gets the active video format
 */
int dfg1394_device__get_format( 
	dfg1394_device_t* dev, struct v4l2_format* format )
{
	if ( format->type != V4L2_BUF_TYPE_VIDEO_CAPTURE ) 
		return -EINVAL;

	spin_lock(&dev->lock);
	format->fmt.pix = dev->pix_format;
	spin_unlock(&dev->lock);

	return SUCCESS;
}

/**
 * @brief @c  VIDIOC_SET_FMT: Tries to set a new video format
 */
int dfg1394_device__set_format( 
	dfg1394_device_t* dev, struct v4l2_format* format )
{
	u32 f = format->fmt.pix.pixelformat;
	u32 w = format->fmt.pix.width;
	u32 h = format->fmt.pix.height;
	struct v4l2_pix_format pix_format;
	dfg1394_pix_mode_t mode;
	int i = 0;
	dfg1394_bool_t cap_is_active;
	int ret = SUCCESS;

	if ( format->type != V4L2_BUF_TYPE_VIDEO_CAPTURE ) 
		return -EINVAL;

	PRINT_INFO(
		"set_format(IN ): w: %d h: %d field: %d, pix_fmt: %c%c%c%c\n",
		format->fmt.pix.width, format->fmt.pix.height,
		format->fmt.pix.field,
		(f >>  0) & 0xff, (f >>  8) & 0xff,
		(f >> 16) & 0xff, (f >> 24) & 0xff );

	/*
	 * (I) Try to find the best matching supported format
	 */

	/* find first base format with suitable size */
	mode = DFG1394_PIX_MODE__FIRST;
	for ( i=DFG1394_PIX_MODE__FIRST; i<DFG1394_PIX_MODE__COUNT; i+=3 ){
		if ( dfg1394_base_pix_formats[i].width  <= w 
		  && dfg1394_base_pix_formats[i].height <= h ) {
			mode = i;
		} else {
			break;
		}
	}

	/* find base format with suitable pixelformat */
	for ( i=0; i<4; i++ ) {
		if ( dfg1394_base_pix_formats[mode+i].pixelformat == f ) {
			mode += i;
			break;
		}
	}

	/* default to UYVY if we have no matching pixel format */
	if ( dfg1394_base_pix_formats[mode].pixelformat != f ) 
		mode += 2; /* move mode to UYVY */
	
	pix_format = dfg1394_base_pix_formats[mode];

	/*
	 * (II) Set the format
	 */

        /* -------- enter synchronized section -------- */
	if (down_interruptible(&dev->sema)) 
		return -ERESTARTSYS;

	cap_is_active = dfg1394_node__is_transmitting(&dev->node);

	if ( cap_is_active ) {
		/* just in case ... */
		PRINT_WARNING(
			"setting format with transmission enabled, "
			"forced to stop!\n" );
		dfg1394_node__stop_transmission( &dev->node );
	}


	/* change hardware video mode */
	if ( mode != dfg1394_node__get_mode(&dev->node) ) 
		ret = dfg1394_node__set_mode( &dev->node, mode );
	
	if (IS_SUCCESS(ret)) {
		/* update device structure while holding semaphore*/
		spin_lock(&dev->lock);
		dev->pix_format = pix_format;
		spin_unlock(&dev->lock);
	} else 
		PRINT_ERR("Failed to set mode to %d\n", mode);
	
	up(&dev->sema); 
        /* -------- leave synchronized section -------- */
	

	if (IS_SUCCESS(ret)) {

		/* return chosen format */
		format->fmt.pix = pix_format;

		f = format->fmt.pix.pixelformat;
		PRINT_DEBUG(
			"set_format(OUT): w: %d h: %d field: %d, "
			"pix_fmt: %c%c%c%c, imgsize: %d \n",
			format->fmt.pix.width, format->fmt.pix.height ,
			format->fmt.pix.field,
			(f >>  0) & 0xff, (f >>  8) & 0xff,
			(f >> 16) & 0xff, (f >> 24) & 0xff,
			format->fmt.pix.sizeimage );
	}

	return ret;
}

/**
 * @brief @c  VIDIOC_GET_PARAM: Gets capture parameters
 */
int dfg1394_device__get_param( 
	dfg1394_device_t* dev, struct v4l2_streamparm* parm )
{
	/* read-only access to static data, no sync required */

	struct v4l2_captureparm* cap_parm;

	if ( parm->type != V4L2_BUF_TYPE_VIDEO_CAPTURE ) 
		return -EINVAL;

	cap_parm = &parm->parm.capture;

	memset( cap_parm, 0, sizeof( struct v4l2_captureparm ));
	cap_parm->readbuffers = 2;
	
	return SUCCESS;
}

/**
 * @brief @c  VIDIOC_RQBUF: Request a specified number of capture buffers
 */
int dfg1394_device__request_buffers( 
	dfg1394_device_t* dev, struct v4l2_requestbuffers* rb )
{
        /* no need to synchronize client access:
         * only one client may be capturing 
	 */

	int i;
	int ret = SUCCESS;
	int buflen;
	int frame_size;
	unsigned long irq_flags;

	/* sanity checks */
	if ( rb->type != V4L2_BUF_TYPE_VIDEO_CAPTURE ) 
		return -EINVAL;
	if ( rb->memory != V4L2_MEMORY_MMAP ) 
		return -EINVAL;
	if ( rb->count < 6 ) 
		rb->count = 6;
	if ( rb->count > DFG1394_STREAM_MAX_BUFFERS ) 
		rb->count = DFG1394_STREAM_MAX_BUFFERS;
	if ( dfg1394_node__is_transmitting(&dev->node) )
		return -EPERM;

	/* -------- lock capture state */
	spin_lock_irqsave( &dev->cap_lock, irq_flags );

	/* check that no buffer is mapped already */
	for (i=0; i < DFG1394_STREAM_MAX_BUFFERS; ++i ) {
		if( dev->cap_buffers[i].v4l2buf.flags & V4L2_BUF_FLAG_MAPPED) {
			PRINT_WARNING( 
				"REQBUFS: buffer %d is already mapped.\n", i);
//			ret = -EPERM;
//			goto exit;
		}
	}

	/* reinitialize buffer queues */
	INIT_LIST_HEAD( &dev->cap_free_queue );
	INIT_LIST_HEAD( &dev->cap_done_queue );

	/* calculate buffer size according to current settings, 
	   ensure that buflen == n * PAGE_SIZE  */
	frame_size = dev->pix_format.sizeimage;
	buflen = (frame_size + PAGE_SIZE - 1) & ~(PAGE_SIZE-1);

	/* initialize all requested buffers */
	for ( i = 0; i < rb->count; ++i ) {
		dfg1394_buffer__initialize( 
			&dev->cap_buffers[i], i, buflen );
		dev->cap_buffers[i].frame_size = frame_size;
		dev->cap_buffers[i].line_size = 
			frame_size / dev->pix_format.height;
		dev->cap_buffers[i].is_requested = TRUE;
	}

	/* clear is_requested - flag on all other buffers */
	for ( i = rb->count; i < DFG1394_STREAM_MAX_BUFFERS; ++i )
		dev->cap_buffers[i].is_requested = FALSE;

	/* -------- unlock capture state */
	spin_unlock_irqrestore( &dev->cap_lock, irq_flags );
	
	PRINT_INFO("%d buffers were requested.\n", rb->count );

	return ret;
}

/**
 * @brief @c VIDIOC_QUERYBUF: Asks the driver how many buffers 
 *        are supported for mmap'ed capturing
 *  
 */
int dfg1394_device__query_buffer(
	dfg1394_device_t* dev, struct v4l2_buffer *buf )
{
	int ret = SUCCESS;
	int buf_idx = buf->index;
	unsigned long irq_flags;

	/* sanity checks */
	if ( buf_idx < 0 || buf_idx >= DFG1394_STREAM_MAX_BUFFERS )
		return -EINVAL;
	if ( buf->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
		return -EINVAL;


	/* -------- lock capture state */
	spin_lock_irqsave( &dev->cap_lock, irq_flags );

	/* check if queried buffer was requested  */
	if ( !dev->cap_buffers[buf_idx].is_requested ) {
		PRINT_WARNING( 
			"VIDEOC_QUERYBUF buffer %d not requested.\n", buf_idx);
		ret = -EINVAL; 
		goto exit;
	}

	/* the actual buffer parameters as set in
	   dfg1394_ioctl_request_buffers(...) */
	*buf = dev->cap_buffers[buf_idx].v4l2buf;

exit:
	/* -------- unlock capture state */
	spin_unlock_irqrestore( &dev->cap_lock, irq_flags );

	return ret;
}


/**
 * @brief @c VIDIOC_QBUF: Queues a previously requested buffer
 *
 * The buffer will be filled with a new frame during capturing
 */
int dfg1394_device__queue_buffer(
	dfg1394_device_t* dev, struct v4l2_buffer* qbuf )
{
        /* no need to synchronize client access:
         * only one client may be capturing 
	 */

	int ret = SUCCESS;
	unsigned long buf_idx = (unsigned long)qbuf->index;
	dfg1394_buffer_t* buf;
	unsigned long irq_flags;


	/* sanity checks */
	if ( qbuf->type != V4L2_BUF_TYPE_VIDEO_CAPTURE ) 
		return -EINVAL; 
	if ( buf_idx >= DFG1394_STREAM_MAX_BUFFERS )
		return -EINVAL; 

	/* -------- lock capture state */
	spin_lock_irqsave( &dev->cap_lock, irq_flags );

	buf = &dev->cap_buffers[buf_idx];
	if ( !buf->is_requested ) {
		PRINT_WARNING("VIDIOC_QBUF buffer %ld not requested.\n",
			      buf_idx);
		ret = -EINVAL; 
		goto exit;
	}

	/* check the addressed buffer's flags, don't trust what was
           given as parameter ... */
	qbuf = &buf->v4l2buf;

	if ( !(qbuf->flags & V4L2_BUF_FLAG_MAPPED) ) {
		PRINT_WARNING("VIDIOC_QBUF buffer %ld is not mapped\n", 
			      buf_idx);
		ret = -EINVAL; 
		goto exit;
	}
	if ((qbuf->flags & V4L2_BUF_FLAG_QUEUED)) {
		PRINT_WARNING("VIDIOC_QBUF buffer %ld is already queued\n", 
			      buf_idx);
		ret = -EINVAL; 
		goto exit;
	}

	/* update flags and queue buffer */
	qbuf->flags &= ~V4L2_BUF_FLAG_DONE;
	INIT_LIST_HEAD( &buf->list_entry );
	list_add_tail( &buf->list_entry, &dev->cap_free_queue );
	qbuf->flags |= V4L2_BUF_FLAG_QUEUED;

exit:
	/* -------- unlock capture state */
	spin_unlock_irqrestore( &dev->cap_lock, irq_flags );

	return ret;
}


/**
 * @brief @c VIDIOC_DQBUF: Dequeues a filled buffer.
 */
inline int dfg1394_device__dequeue_buffer(
	dfg1394_device_t* dev, struct v4l2_buffer* qbuf )
{
        /* no need to synchronize client access:
         * only one client may be capturing 
	 */

	int ret = SUCCESS;
	dfg1394_buffer_t* new_buf = NULL;
	unsigned long irq_flags;

	/* sanity checks */
	if ( qbuf->type != V4L2_BUF_TYPE_VIDEO_CAPTURE ) 
		return -EINVAL; 


	/* -------- lock capture state */
	spin_lock_irqsave( &dev->cap_lock, irq_flags );

	if ( list_empty( &dev->cap_done_queue )) {
		PRINT_WARNING("iotcl: VIDIOC_DQBUF done queue is empty.\n");
		ret = -EINVAL; 
		goto exit;
	}

	new_buf = list_entry( 
		dev->cap_done_queue.next, dfg1394_buffer_t, list_entry );
	list_del( &new_buf->list_entry );

 	new_buf->v4l2buf.flags &= ~V4L2_BUF_FLAG_QUEUED;
 	new_buf->v4l2buf.flags &= ~V4L2_BUF_FLAG_DONE;
	*qbuf = new_buf->v4l2buf;

exit:
	/* -------- unlock capture state */
	spin_unlock_irqrestore( &dev->cap_lock, irq_flags );

	return ret;
}

/**
 * @brief @c VIDIOC_STREAMON Starts the capturing
 */
int dfg1394_device__stream_on( dfg1394_device_t* dev, u32* type )
{

	int ret = SUCCESS;

	/* sanity checks */
	if ( *type != V4L2_BUF_TYPE_VIDEO_CAPTURE ) 
		return -EINVAL; 

        /* only one client may be capturing, but as we change the
	 * converter state here, we have to synchronize..
	 *
         * -------- enter synchronized section -------- */
	if (down_interruptible(&dev->sema)) 
		return -ERESTARTSYS;

	if (dfg1394_node__is_transmitting(&dev->node)) {
		PRINT_ERR("stream_on() called with is_transmitting == TRUE\n");
		dfg1394_node__stop_transmission(&dev->node);
	}

	/* start hardware */
	ret = dfg1394_node__start_transmission(&dev->node, 0 );

	up(&dev->sema); 
        /* -------- leave synchronized section -------- */
	
	return ret;
}

/**
 * @brief @c VIDIOC_STREAMOFF: Stops the capturing
 *
 * 
 */
int dfg1394_device__stream_off( dfg1394_device_t* dev, u32* type )
{
	int i;
	int ret = SUCCESS;
	unsigned long irq_flags;
	

	/* sanity checks */
	if ( *type != V4L2_BUF_TYPE_VIDEO_CAPTURE ) 
		return -EINVAL; 


        /* -------- enter synchronized section -------- */
	if (down_interruptible(&dev->sema)) 
		return -ERESTARTSYS;

	/* stop hardware */
	ret = dfg1394_node__stop_transmission(&dev->node);


	/* -------- lock capture state */
	spin_lock_irqsave(&dev->cap_lock, irq_flags);

	/* unqueue all buffers */
	for (i=0; i < DFG1394_STREAM_MAX_BUFFERS; ++i ) {
		dev->cap_buffers[i].v4l2buf.flags 
			&= ~(V4L2_BUF_FLAG_QUEUED | V4L2_BUF_FLAG_DONE); 
	}

	/* empty buffer queues */
	INIT_LIST_HEAD( &dev->cap_free_queue );
	INIT_LIST_HEAD( &dev->cap_done_queue );

	spin_unlock_irqrestore( &dev->cap_lock, irq_flags); 
	/* -------- unlock capture state */

	up(&dev->sema); 
        /* -------- leave synchronized section -------- */

	/* wake up any sleeping clients (they'll get an error) */
	wake_up_interruptible( &dev->cap_done_wait );

	return ret;
}



/**
 * @brief @c VIDIOC_QUERYCTRL: Query controls
 *
 * 
 */
int dfg1394_device__queryctrl( dfg1394_device_t *dev, struct v4l2_queryctrl *param)
{
	int ret = SUCCESS;

	switch( param->id )
	{
		case V4L2_CID_BRIGHTNESS:
			param->type = V4L2_CTRL_TYPE_INTEGER;
			strcpy( param->name, "brightness" );
			param->minimum = 0;
			param->maximum = 255;
			param->step = 1;
			param->default_value = 128;
			param->flags = 0;
			param->reserved[0] = param->reserved[1] = 0;
			break;

		case V4L2_CID_CONTRAST:
			param->type = V4L2_CTRL_TYPE_INTEGER;
			strcpy( param->name, "contrast" );
			param->minimum = 0;
			param->maximum = 127;
			param->step = 1;
			param->default_value = 64;
			param->flags = 0;
			param->reserved[0] = param->reserved[1] = 0;
			break;
			
		case V4L2_CID_SATURATION:
			param->type = V4L2_CTRL_TYPE_INTEGER;
			strcpy( param->name, "saturation" );
			param->minimum = 0;
			param->maximum = 127;
			param->step = 1;
			param->default_value = 64;
			param->flags = 0;
			param->reserved[0] = param->reserved[1] = 0;
			break;
			
		case V4L2_CID_HUE:
			param->type = V4L2_CTRL_TYPE_INTEGER;
			strcpy( param->name, "hue" );
			param->minimum = 0;
			param->maximum = 255;
			param->step = 1;
			param->default_value = 128;
			param->flags = 0;
			param->reserved[0] = param->reserved[1] = 0;
			break;
			
			
		default: 
			return -EINVAL;
	}
	
	return ret;
}


/**
 * @brief @c VIDIOC_S_CTRL: Set controls
 *
 * 
 */
int dfg1394_device__set_ctrl( dfg1394_device_t *dev, struct v4l2_control *param )
{
	int ret = SUCCESS;
	
	switch( param->id )
	{
		case V4L2_CID_BRIGHTNESS:
			if( param->value > 0 &&
			    param->value < 256 )
			{
				ret = dfg1394_node__set_brightness( &dev->node, param->value );
			}
			break;

		case V4L2_CID_CONTRAST:
			if( param->value > 0 &&
			    param->value < 127 )
			{
				ret = dfg1394_node__set_contrast( &dev->node, param->value );
			}
			break;
			
		case V4L2_CID_SATURATION:
			if( param->value > 0 &&
			    param->value < 127 )
			{
				ret = dfg1394_node__set_saturation( &dev->node, param->value );
			}
			break;

		case V4L2_CID_HUE:
			if( param->value > 0 &&
			    param->value < 255 )
			{
				ret = dfg1394_node__set_hue( &dev->node, param->value );
			}
			break;
			
		default:
			return -EINVAL;
	}
	
	return ret;
}
				
/**
 * @brief @c VIDIOC_G_CTRL: Get controls
 *
 * 
 */
int dfg1394_device__get_ctrl( dfg1394_device_t *dev, struct v4l2_control *param )
{
	int ret = SUCCESS;
	
	switch( param->id )
	{
		case V4L2_CID_BRIGHTNESS:
		{
			u8 value;
			ret = dfg1394_node__get_brightness( &dev->node, &value );
			param->value = value;
		}		
		break;
			
		case V4L2_CID_CONTRAST:
		{
			u8 value;
			ret = dfg1394_node__get_contrast( &dev->node, &value );
			param->value = value;
		}		
		break;
			
		case V4L2_CID_SATURATION:
		{
			u8 value;
			ret = dfg1394_node__get_saturation( &dev->node, &value );
			param->value = value;
		}		
		break;
			
		case V4L2_CID_HUE:
		{
			u8 value;
			ret = dfg1394_node__get_hue( &dev->node, &value );
			param->value = value;
		}		
		break;
			
			
		default: 
			return -EINVAL;
	}
	
	return ret;
}

/*@}*/

/* ************************************************************************ */
/** 
 *   @name dfg1394_device Capture related                                
 */
/* ************************************************************************ */
/* @{                                                                       */

/**
 * @brief Tries to pop an empty buffer from the incoming buffer queue
 *
 * @param dev       [in]  @c this pointer
 * @return An empty buffer or @c NULL if the queue was empty
 */
dfg1394_buffer_t* dfg1394_device__pop_empty_buffer( dfg1394_device_t* dev )
{
	dfg1394_buffer_t* buf = NULL;
	unsigned long irq_flags;

	/* -------- lock capture state */
	spin_lock_irqsave(&dev->cap_lock, irq_flags);

	/* try to get next buffer from free_queue */
	if ( !list_empty( &dev->cap_free_queue )) {

		buf = list_entry( 
			dev->cap_free_queue.next, 
			dfg1394_buffer_t, list_entry);

		list_del( &buf->list_entry );			

		buf->offset = 0;
	} 

	spin_unlock_irqrestore( &dev->cap_lock, irq_flags); 
	/* -------- unlock capture state */

	return buf;
}

/**
 * @brief Pushes a filled buffer into the outgoind buffer queue
 *
 * @param dev         [in]  @c this pointer
 * @param buf         [in]  the buffer 
 * @param frame_count [in]  the frame sequence number
 *
 */
void dfg1394_device__push_filled_buffer( 
	dfg1394_device_t* dev, 
	dfg1394_buffer_t* buf, u32 frame_count)
{
	unsigned long irq_flags;

	/* -------- lock capture state */
	spin_lock_irqsave(&dev->cap_lock, irq_flags);

	/* commit current buffer */
	if ( buf != NULL ) {

		/* set buffer attributes ... */
		buf->v4l2buf.flags |= V4L2_BUF_FLAG_DONE;
		buf->v4l2buf.flags &= ~V4L2_BUF_FLAG_QUEUED;
		buf->v4l2buf.sequence = frame_count;
		buf->v4l2buf.bytesused = buf->frame_size;
		do_gettimeofday( &buf->v4l2buf.timestamp );
		buf->v4l2buf.field = V4L2_FIELD_INTERLACED;

		/* ... and add it to cpa_done_queue */
		INIT_LIST_HEAD( &buf->list_entry );
		list_add_tail( &buf->list_entry, &dev->cap_done_queue );
	} 
	
	/* wake up sleeping client */
	wake_up_interruptible( &dev->cap_done_wait );
	
	spin_unlock_irqrestore( &dev->cap_lock, irq_flags); 
	/* -------- unlock capture state */
}

/**
 * @brief Declares the specified client as "capturing". 
 *
 * There may be only one capturing client connected to each device!
 */
inline int dfg1394_device__declare_capturing( 
	dfg1394_device_t* dev, dfg1394_client_t* client )
{
	int ret = SUCCESS;

	/* -------- lock capture state */
	spin_lock(&dev->cap_lock);

	if ( dev->cap_client == NULL ) {
		PRINT_INFO("client declared capturing\n");
		dev->cap_client = client;
	} else if ( dev->cap_client != client ) {
		PRINT_INFO("client is NOT capturing, access denied\n");
		ret = -EPERM;
	}

	spin_unlock(&dev->cap_lock);
	/* -------- unlocked capture state */

	return ret;
}

/**
 * @brief Checks wether the specified client is the capturing one.
 * 
 * @retval @c SUCCESS if the client is capturing, @c -EPERM otherwise
 */
inline int dfg1394_device__check_capturing( 
	dfg1394_device_t* dev, dfg1394_client_t* client )
{
	int ret;
	/* -------- lock capture state */
	spin_lock(&dev->cap_lock);

	ret = (dev->cap_client == client) ? SUCCESS : -EPERM;

	spin_unlock(&dev->cap_lock);
	/* -------- unlocked capture state */

	return ret;
}

/*@}*/

/* ************************************************************************ */
/** 
 *   @name dfg1394_client Methods                                        
 */
/* ************************************************************************ */
/* @{                                                                       */


/**
 * Allocates memory and initializes a new dfg1394_client_t in it
 */
dfg1394_client_t* dfg1394_client__create( 
	dfg1394_device_t* dev, dfg1394_bool_t is_blocking )
{
	dfg1394_client_t* client 
		= kmalloc( sizeof(dfg1394_client_t), GFP_ATOMIC );

	if ( client != NULL ) {
		memset( client, 0, sizeof( *client ));
		client->is_blocking  = is_blocking;
		client->is_open      = TRUE;
		client->dev = dev;

		try_module_get(THIS_MODULE);
	}
		
	return client;
}

/**
 * destroys the dfg1394_client_t object, frees memory
 */
void dfg1394_client__destroy( dfg1394_client_t* client )
{
	module_put(THIS_MODULE);
	kfree( client );
}



/*@}*/

/* ************************************************************************ */
/** 
 *   @name VMA Management
 */
/* ************************************************************************ */
/* @{                                                                       */


/**
 *  @brief Increments buffer usage count
 */
void dfg1394_vma_open( struct vm_area_struct* vma )
{
	dfg1394_device_t *dev = vma->vm_private_data;
	unsigned long buf_idx = vma->vm_pgoff;
	dfg1394_buffer_t* buf = NULL;
	unsigned long irq_flags;

	/* -------- lock capture state */
	spin_lock_irqsave(&dev->cap_lock, irq_flags);

	buf = &dev->cap_buffers[buf_idx];
	buf->vma_user_count++;

	spin_unlock_irqrestore( &dev->cap_lock, irq_flags );
	/* -------- unlocked capture state */
}


/**
 * @brief decrements buffer usage count. 
 *
 * Frees buffer memory if usage count drops to zero
 */
void dfg1394_vma_close( struct vm_area_struct* vma )
{
	dfg1394_device_t *dev = vma->vm_private_data;
	unsigned long buf_idx = vma->vm_pgoff;
	dfg1394_buffer_t* buf = NULL;
	unsigned long irq_flags;	

	/* -------- lock capture state */
	spin_lock_irqsave(&dev->cap_lock, irq_flags);

	buf = &dev->cap_buffers[buf_idx];

	if ( --buf->vma_user_count == 0 ) {
		buf->v4l2buf.flags &=  ~V4L2_BUF_FLAG_MAPPED;
		vfree( buf->data );
		buf->data = NULL;
	}

	spin_unlock_irqrestore( &dev->cap_lock, irq_flags );
	/* -------- unlocked capture state */

}

/**
 * @brief Mapps requested buffer page into userspace
 */
struct page* dfg1394_vma_nopage(
	struct vm_area_struct* vma, unsigned long address, int write_access )
{
	dfg1394_device_t *dev = vma->vm_private_data;
	dfg1394_buffer_t* buf = NULL;
	unsigned long buf_offs = address - vma->vm_start;
	unsigned long buf_idx = vma->vm_pgoff;
	struct page* page = NOPAGE_SIGBUS;
	unsigned long irq_flags;

	/* -------- lock capture state */
	spin_lock_irqsave(&dev->cap_lock, irq_flags);

	buf = &dev->cap_buffers[buf_idx];

	if ( buf_offs > buf->v4l2buf.length ) {
		PRINT_ERR("read past end of mmap'ed buffer.\n");
		goto exit;
	}

	page = vmalloc_to_page( buf->data + buf_offs );
	if (page)
		get_page(page);
	else
		page = NOPAGE_SIGBUS;
		
exit:
	spin_unlock_irqrestore( &dev->cap_lock, irq_flags );
	/* -------- unlocked capture state */

	return page;
}


/**
 * The dfg1394's vma operations. See also: dfg1394_mmap() function.
 */
static struct vm_operations_struct dfg1394_vma_ops =
{
	open :  dfg1394_vma_open,
	close:  dfg1394_vma_close,
	nopage: dfg1394_vma_nopage,
};


/*@}*/

/* ************************************************************************ */
/** 
 *   @name File Operations (system calls)
 */
/* ************************************************************************ */
/* @{                                                                       */

/**
 * @brief Opens a new connection to the device
 */
static int dfg1394_open(struct inode *inode, struct file *file)
{
        struct video_device *vdev = video_devdata(file);

	u32 cc = 0;
	int ret = SUCCESS;
	dfg1394_device_t *dev = (dfg1394_device_t*) vdev->priv ;
	dfg1394_client_t *client;

	if (dfg1394_device__is_ghost(dev)) 
		return -ENODEV;


	/* increment client count */
	spin_lock(&dev->lock);
	cc = ++ dev->client_count;
	spin_unlock(&dev->lock);

	/* create client object */
	client = dfg1394_client__create(
		dev, (file->f_flags & O_NONBLOCK) == 0 );
	if (client == NULL ) {
		ret = -ENOMEM;
		goto err1;
	}
	file->private_data = client;
	
	if ( cc == 1 ) {
		/*  open the first time
		   -> scan for device  and initialize converter hardware */

		if (IS_FAILURE( ret = dfg1394_node__start( &dev->node ))) {
			PRINT_ERR("Failed to initialize DFG/1394-1.\n");
			goto err1;
		}
		PRINT_NOTICE("Converter initialized.\n");
	}

	PRINT_INFO("device opened, now %d user(s)\n", cc );

	return ret;

 err1:
	dfg1394_client__destroy( client );
	spin_lock(&dev->lock);
	cc = -- dev->client_count;
	spin_unlock(&dev->lock);

	return ret;
}

/**
 * @brief Closes a device connection
 */
static int dfg1394_release(struct inode *inode, struct file *file)
{
	dfg1394_client_t *client = file->private_data;
	dfg1394_device_t *dev = client->dev;
	u32 cc = 0;
	dfg1394_bool_t still_active = FALSE;
	u32 type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	unsigned long irq_flags;

	/*  check synchronized, if this was the capturing client */
	spin_lock_irqsave(&dev->cap_lock, irq_flags);
	if (dev->cap_client == client) {
		dev->cap_client = NULL;
		still_active = dfg1394_node__is_transmitting(&dev->node);
	}
	spin_unlock_irqrestore( &dev->cap_lock, irq_flags );

	if ( still_active ) {
		PRINT_INFO("still streaming, I'll stop now!.\n");
		dfg1394_device__stream_off( dev, &type );
	}

	dfg1394_client__destroy( client );

	spin_lock(&dev->lock);
	cc = -- dev->client_count;
	spin_unlock(&dev->lock);
	
	if ( cc == 0 ) {
		PRINT_INFO("device closed, no more clients.\n");
	} else {
		PRINT_INFO("device closed, still %d clients\n", cc );
	}

	return SUCCESS;
}

/**
 * Unimplemented read system call
 */ 
static ssize_t dfg1394_read(
	struct file *file, char *data, size_t count, loff_t *ppos)
{
	PRINT_WARNING("read() not implemented.\n");
	return -EINVAL;
}


/**
 * @brief Waits for buffers in @c dev->cap_done_queue
 */ 
static unsigned int dfg1394_poll( struct file *file, poll_table *table )
{
	dfg1394_client_t *client = file->private_data;
	dfg1394_device_t *dev = client->dev;
	int ret = 0;
	unsigned long irq_flags;

	if (dfg1394_device__is_ghost(dev))
		return -ENODEV;

	poll_wait(file, &dev->cap_done_wait, table);

	/* -------- lock capture state */
	spin_lock_irqsave(&dev->cap_lock, irq_flags);

	if ( !list_empty( &dev->cap_done_queue )) {
		ret = (POLLIN | POLLRDNORM);/* data is ready now */
		goto exit;
	}

exit:
	/* -------- unlock capture state */
	spin_unlock_irqrestore( &dev->cap_lock, irq_flags );

	return ret;
}


/**
 * @brief Initializes mapping of one buffer into userspace
 */
static int dfg1394_mmap(struct file *file, struct vm_area_struct * vma)
{
	dfg1394_client_t *client = file->private_data;
	dfg1394_device_t *dev = client->dev;
	dfg1394_buffer_t* buf = NULL;
	/* this is valid, because request_buffers() set the offset to  
           "buf_idx << PAGE_SHIFT": */
	unsigned long buf_idx = vma->vm_pgoff;
	int ret = SUCCESS;

	/* sanity checks */
	if (dfg1394_device__is_ghost(dev))
		return -ENODEV;
	if ( buf_idx >= DFG1394_STREAM_MAX_BUFFERS ) {
		PRINT_ERR(
			"mmap() failed: buf_idx %ld out of range!\n", buf_idx);
		return -EINVAL;
	}

	/* only the capturing client may mmap */
	if (IS_FAILURE(ret = dfg1394_device__check_capturing(dev, client)))
		return ret;

        /* -------- enter synchronized section -------- */
	if (down_interruptible(&dev->sema)) 
		return -ERESTARTSYS;

	/* make sure we are not already transmitting */
	if ( dfg1394_node__is_transmitting( &dev->node )) {
		ret = -EPERM;
		goto exit;
	}

	/* we are not transmitting, so it's safe to access the 
	   capture buffers here without further synchronization
	*/
	buf = &dev->cap_buffers[buf_idx];

	if ( buf->v4l2buf.flags & V4L2_BUF_FLAG_MAPPED ) {
		PRINT_WARNING(
			"mmap() failed: buffer %ld already mapped.\n",buf_idx);
		ret = -EINVAL; 
		goto exit;
	}

	/* allocate memory for this buffer */
	buf->data = vmalloc( buf->v4l2buf.length );
	if ( buf->data == NULL) {
		PRINT_ERR(
			"mmap() failed: no space left for buffer %ld.\n", 
			buf_idx );
		ret = -ENOMEM;
		goto exit;
	}

	buf->v4l2buf.flags |=	V4L2_BUF_FLAG_MAPPED;

	/* set up VMA operatios */
	vma->vm_flags |= VM_RESERVED;
	vma->vm_ops = &dfg1394_vma_ops;
	vma->vm_private_data = dev;

	/* we have now exactly one user */
	buf->vma_user_count = 1;

	PRINT_INFO("mmap'ed buffer %ld.\n", buf_idx );

 exit:
	up(&dev->sema); 
        /* -------- leave synchronized section -------- */

	return SUCCESS;
}


/**
 * @brief Broker to call IOCTL functions
 */
static int dfg1394_do_ioctl(struct inode *inode, struct file *file,
			  unsigned int cmd, void* arg)
{
	dfg1394_client_t *client = file->private_data;
	dfg1394_device_t *dev    = client->dev;
	int ret = -EINVAL;

	if (dfg1394_device__is_ghost(dev))
		return -ENODEV;

/** @brief helper to report not implemented V4L2 ioctls  */
#define CASE_MISSING(c) \
        case c: PRINT_ERR("ioctl " #c " not implemented.\n"); \
                ret = -EINVAL; \
                break;

	 switch(cmd)
	 {
	 case VIDIOC_QUERYCAP: 
 //		PRINT_INFO("ioctl VIDIOC_QUERYCAP\n");
		 ret = dfg1394_device__query_cap( 
			 dev, (struct v4l2_capability*) arg );
		 break;
	 case VIDIOC_ENUM_FMT:
		 ret = dfg1394_device__enum_formats( 
			 dev, (struct v4l2_fmtdesc *) arg );
		 break;
	 case VIDIOC_G_FMT:
		ret = dfg1394_device__get_format( 
			dev, (struct v4l2_format*) arg);
		break;
	case VIDIOC_S_FMT:
		ret = dfg1394_device__set_format( 
			dev, (struct v4l2_format*) arg);
		break;

 	case VIDIOC_REQBUFS: 
		if (IS_FAILURE(ret = dfg1394_device__declare_capturing( 
					 dev, client ) ))
			break;
		ret = dfg1394_device__request_buffers( 
			dev, (struct v4l2_requestbuffers*) arg );
		break;
	case VIDIOC_QUERYBUF:
		ret = dfg1394_device__query_buffer( 
			dev, (struct v4l2_buffer *) arg );
		break;

 	case VIDIOC_QBUF:
		if (IS_FAILURE(ret = dfg1394_device__check_capturing(
				       dev, client )))
			break;
		ret = dfg1394_device__queue_buffer( 
			dev, (struct v4l2_buffer *) arg ); 
 		break; 
 	case VIDIOC_DQBUF:
		if (IS_FAILURE(ret = dfg1394_device__check_capturing( 
				       dev, client )))
			break;
		ret = dfg1394_device__dequeue_buffer( 
			dev, (struct v4l2_buffer *) arg ); 
 		break; 
	case VIDIOC_STREAMON:
		PRINT_INFO("ioctl VIDIOC_STREAMON\n");
		if (IS_FAILURE(ret = dfg1394_device__check_capturing( 
				       dev, client )))
			break;
		ret = dfg1394_device__stream_on( dev, (u32*) arg );
		break;
	case VIDIOC_STREAMOFF:
		PRINT_INFO("ioctl VIDIOC_STREAMOFF\n");
		if (IS_FAILURE(ret = dfg1394_device__check_capturing( 
				       dev, client )))
			break;
		ret = dfg1394_device__stream_off( dev, (u32*) arg );
		break;

	CASE_MISSING( VIDIOC_G_FBUF );
	CASE_MISSING( VIDIOC_S_FBUF );

	case VIDIOC_ENUMINPUT:
		ret = dfg1394_device__enum_input( 
			dev, (struct v4l2_input *) arg );
		break;
	case VIDIOC_G_INPUT:
		PRINT_INFO("ioctl VIDIOC_G_INPUT\n");
		ret = dfg1394_device__get_input( dev, (int*)arg );
		break;
	case VIDIOC_S_INPUT:
		PRINT_INFO("ioctl VIDIOC_S_INPUT\n");
		ret = dfg1394_device__set_input( dev, (int*)arg );
		break;
	case VIDIOC_G_PARM:
		PRINT_INFO("ioctl VIDIOC_G_INPUT\n");
		ret = dfg1394_device__get_param( 
			dev, (struct v4l2_streamparm*)arg );
		break;

	CASE_MISSING( VIDIOC_S_PARM );


	case VIDIOC_ENUMSTD:
		ret = dfg1394_device__enum_standard( 
			dev, (struct v4l2_standard*) arg);
		break;
 	case VIDIOC_G_STD: 
		PRINT_INFO("ioctl VIDIOC_G_STD\n");
		ret =  dfg1394_device__get_standard( dev, (v4l2_std_id*) arg );
		break;
 	case VIDIOC_S_STD: 
		PRINT_INFO("ioctl VIDIOC_S_STD\n");
		ret =  dfg1394_device__set_standard( dev, (v4l2_std_id*) arg );
		break;
	case VIDIOC_QUERYCTRL:
		PRINT_INFO("ioctl VIDIOC_QUERYCTRL\n");
		ret = dfg1394_device__queryctrl( dev, (struct v4l2_queryctrl *)arg );
		break;
	case VIDIOC_G_CTRL:
		PRINT_INFO("ioctl VIDIOC_G_CTRL\n");
		ret = dfg1394_device__get_ctrl( dev, (struct v4l2_control *)arg );
		break;
	case VIDIOC_S_CTRL:
		PRINT_INFO("ioctl VIDIOC_S_CTRL\n");
		ret = dfg1394_device__set_ctrl( dev, (struct v4l2_control *)arg );
		break;
	case VIDIOC_QUERYMENU: 
	case VIDIOC_G_TUNER:   /* no tuner */
	case VIDIOC_S_TUNER: 
	case VIDIOC_G_AUDIO:   /* no audio */
	case VIDIOC_S_AUDIO:
		ret = -EINVAL; 
		break;

	default:
		PRINT_TRACE1("do_ioctl: unknown\n");
		ret = v4l_compat_translate_ioctl( 
			inode, file, cmd, arg, dfg1394_do_ioctl);
		break;
	}

	return ret;
}

/**
 * @brief IOCTL system call
 *
 * This is were all the V4L2 commands are received from userspace
 */
static int dfg1394_ioctl(struct inode *inode, struct file *file,
			  unsigned int cmd, unsigned long arg)
{
        /*
	 * Just call the usercopy helper and pass a pointer to
	 * the function which actually does the real work.
	 */
	return video_usercopy( inode, file, cmd, arg, dfg1394_do_ioctl );
}

/**
 * @brief The drivers's file operations
 */
static struct file_operations dfg1394_fops =
{
	owner:	  THIS_MODULE,
	open:	  dfg1394_open,
	release:  dfg1394_release,
	read:	  dfg1394_read,
	poll:     dfg1394_poll,
	mmap:	  dfg1394_mmap,
	ioctl:	  dfg1394_ioctl,
	llseek:   no_llseek,
};


/*@}*/






