/*
 *
 * Copyright (c) 2004, Frederick Ros (sl33p3r@free.fr)
 *										    
 * Boot.c
 *
 *
 * This file is part of the eagle-usb driver package.
 *
 * eagle-usb driver package is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * "eagle-usb driver package" is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with "ADI USB ADSL Driver for Linux"; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: Boot.c,v 1.2 2004/06/04 18:56:19 sleeper Exp $
 */

#include "Adiutil.h"
#include "Boot.h"
#include "eu_utils.h"
#include "eu_sm.h"
#include "Me.h"
#include "macros.h"
#include "debug.h"

#define kMAILBOX_EMPTY          0
#define kMAILBOX_FULL           1

/* Structure required by the IDMA boot engine to load pages*/
typedef struct
{
    uint16_t  ADIHeader;
    uint16_t  Addr;
    uint16_t  Size;
    uint16_t  OvlOffset;
    uint16_t  Ovl;
    uint16_t  LastBlock;
} IDMALoadBlockInfo;

/*********************************************************************************/
/* Local only prototypes							 */
/*********************************************************************************/
static void IdmaUploadMainPage(eu_instance_t *ins);
static int IDMAWrite(eu_instance_t *ins, void *pData, UInt32 Size);
#ifdef LINUX_2_4
static void IDMACompletion(struct urb *urb);
#else
static void IDMACompletion(struct urb *urb, struct pt_regs *regs );
#endif

/*********************************************************************************/
/* BootModem									 */
/* Kinda like doing a ctrl-alt-delete on your PC. Reuploads (or does first	 */
/* upload) of the DSP code, and initializes the host's state machine. 		 */
/*********************************************************************************/
int BootTheModem (eu_instance_t *ins, Boolean Rebooting ) 
{
    int stat = 0;

    eu_enters (DBG_BOOT);

    /*********************************************************************************/
    /* The steps in the boot process are not documented, they were "gleaned"	     */
    /* from the NDIS code. A different, simpler process might be possible,	     */
    /* but we have no way to prove it at this point.				     */
    /*********************************************************************************/

    if (!Rebooting)
    {
	/*To kickoff the boot process, do a custom command, SET_MODE to BOOTMODE_IDMA*/
	stat = eu_cmd_to_modem(ins, EU_CMD_SET_MODE, MODE_BOOTMODE_IDMA, 0, NULL);
	if (stat != 0)
	{
	    eu_err( "eu_cmd_to_modem error %x on SET_MODE BOOTMODE_IDMA\n", stat);
	    goto end_boot;
	}
	
	/*LOOPBACK_ON correlates to ENUM_FPGALBK in the NDIS code*/
	stat = eu_cmd_to_modem(ins, EU_CMD_SET_MODE, MODE_LOOPBACK_ON, 0, NULL);
	if (stat != 0)
	{
	    eu_err("eu_cmd_to_modem error %x on SET_MODE LOOPBACK_ON\n", stat);
	    goto end_boot;
	}
    }

    /*I know we've already done this - but this is what the NDIS code does!*/
    stat = eu_cmd_to_modem(ins, EU_CMD_SET_MODE, MODE_BOOTMODE_IDMA, 0, NULL);
    if (stat != 0)
    {
	eu_err("eu_cmd_to_modem error %x on SET_MODE BOOTMODE_IDMA\n", stat);
	goto end_boot;
    }

    /*
      At this point, we must wait for 200 ms. Original code was doing a
      non-interruptible wait !!! This leads to freeze and possibly odd
      behaviours.
      The new way of doing it is adding steps in the state machine, and schedule
      the WatchDog to beep in 200 ms
    */
    
    ins->AdiModemSm.CurrentAdiState = STATE_BOOT_STAGE_1;
    ins->AdiModemSm.ModemReplyExpected = EVENT_TIMER_TICK;
    SetTimerInterval(ins, TRANS_TIME_2_BOOT_STAGE_1);
    
end_boot:
    eu_leaves (DBG_BOOT);
    return stat;
       
    
}

int BootTheModem_Stage_1 (eu_instance_t *ins) 
{
    int stat = 0;

    eu_enters (DBG_BOOT);
    
    stat = eu_cmd_to_modem(ins, EU_CMD_SET_MODE, MODE_START_RESET, 0, NULL);
    if (stat != 0)
    {
	eu_err ("eu_cmd_to_modem error %x on SET_MODE START_RESET\n", stat);
	goto end_boot;
    }

    /* Advance to stage 2 and wait for 200 ms */
    ins->AdiModemSm.CurrentAdiState = STATE_BOOT_STAGE_2;
    SetTimerInterval(ins, TRANS_TIME_2_BOOT_STAGE_2);

end_boot:
    eu_leaves (DBG_BOOT);
    return stat;
}

int BootTheModem_Stage_2 (eu_instance_t *ins) 
{
    int stat = 0;
    UInt16 myWord;

    eu_enters (DBG_BOOT);

    stat = eu_cmd_to_modem(ins, EU_CMD_SET_MODE, MODE_END_RESET, 0, NULL);
    if (stat != 0)
    {
	eu_err ("eu_cmd_to_modem error %x on SET_MODE END_RESET\n", stat);
	goto end_boot;
    }

    /*Clear all of the CMV mailboxes - this is a MUST!!!  [IIDos-Oct.18,01]*/
    myWord = cpu_to_le16(kMAILBOX_EMPTY);
    stat  = eu_cmd_to_modem(ins, EU_CMD_SET_2183_DATA, (DSP_MP_TX_MAILBOX | 0x4000),
                           2, (UInt8 *)&myWord);
    stat  = eu_cmd_to_modem(ins, EU_CMD_SET_2183_DATA, (DSP_MP_RX_MAILBOX | 0x4000),
                           2, (UInt8 *)&myWord);
    stat  = eu_cmd_to_modem(ins, EU_CMD_SET_2183_DATA, (DSP_SWAP_MAILBOX  | 0x4000),
                           2, (UInt8 *)&myWord);

    ins->AdiModemSm.CurrentAdiState = STATE_BOOT_STAGE_3;
    SetTimerInterval(ins, TRANS_TIME_2_BOOT_STAGE_3);

end_boot:
    eu_leaves (DBG_BOOT);
    return stat;
}

int BootTheModem_Stage_3 (eu_instance_t *ins) 
{
    int stat = 0;

    eu_enters (DBG_BOOT);
    
    /*Finally, we get to do something that seems substantial!*/
    IdmaUploadMainPage(ins);
    /*********************************************************************************/
    /*This is the last operation in our sequence of tasks for device		     */
    /*initialization. Everything else will be driven by the hardware.		     */
    /*Once the main page is loaded, the hardware will start requesting		     */
    /*swap pages.								     */
    /*Put our modem in the BOOT_WAIT state and start our watchdog!		     */
    /*********************************************************************************/
    ins->AdiModemSm.CurrentAdiState = STATE_BOOT_WAIT;
    SetTimerInterval(ins, TRANS_TIME_4_BOOTWAIT);

    eu_leaves (DBG_BOOT);
    return stat;
}
    

#if 0
/*********************************************************************************/
/* BootModem									 */
/* Kinda like doing a ctrl-alt-delete on your PC. Reuploads (or does first	 */
/* upload) of the DSP code, and initializes the host's state machine. 		 */
/*********************************************************************************/
int BootTheModem(eu_instance_t *ins, Boolean Rebooting)
{
    int stat = 0;
    UInt16 myWord;

    eu_enters (DBG_BOOT);

    /*********************************************************************************/
    /* The steps in the boot process are not documented, they were "gleaned"	     */
    /* from the NDIS code. A different, simpler process might be possible,	     */
    /* but we have no way to prove it at this point.				     */
    /*********************************************************************************/

    if (!Rebooting)
    {
	/*To kickoff the boot process, do a custom command, SET_MODE to BOOTMODE_IDMA*/
	stat = eu_cmd_to_modem(ins, EU_CMD_SET_MODE, MODE_BOOTMODE_IDMA, 0, NULL);
	if (stat != 0)
	{
	    eu_err ("eu_cmd_to_modem error %x on SET_MODE BOOTMODE_IDMA\n", stat);
	    goto end_boot;
	}
	
	/*LOOPBACK_ON correlates to ENUM_FPGALBK in the NDIS code*/
	stat = eu_cmd_to_modem(ins, EU_CMD_SET_MODE, MODE_LOOPBACK_ON, 0, NULL);
	if (stat != 0)
	{
	    eu_err ("eu_cmd_to_modem error %x on SET_MODE LOOPBACK_ON\n", stat);
	    goto end_boot;
	}
    }

    /*I know we've already done this - but this is what the NDIS code does!*/
    stat = eu_cmd_to_modem(ins, EU_CMD_SET_MODE, MODE_BOOTMODE_IDMA, 0, NULL);
    if (stat != 0)
    {
	eu_err ("eu_cmd_to_modem error %x on SET_MODE BOOTMODE_IDMA\n", stat);
	goto end_boot;
    }

    /*Analog folks say this delay is needed - but they couldn't tell us why!*/
    mdelay(200);

    stat = eu_cmd_to_modem(ins, EU_CMD_SET_MODE, MODE_START_RESET, 0, NULL);
    if (stat != 0)
    {
	eu_err ("eu_cmd_to_modem error %x on SET_MODE START_RESET\n", stat);
	goto end_boot;
    }

    /*Analog folks say this delay is needed - but they couldn't tell us why!*/
    mdelay(200);

    stat = eu_cmd_to_modem(ins, EU_CMD_SET_MODE, MODE_END_RESET, 0, NULL);
    if (stat != 0)
    {
	eu_err ("eu_cmd_to_modem error %x on SET_MODE END_RESET\n", stat);
	goto end_boot;
    }

    /*Clear all of the CMV mailboxes - this is a MUST!!!  [IIDos-Oct.18,01]*/
    myWord = cpu_to_le16(kMAILBOX_EMPTY);
    stat  = eu_cmd_to_modem(ins, EU_CMD_SET_2183_DATA, (DSP_MP_TX_MAILBOX | 0x4000),
                           2, (UInt8 *)&myWord);
    stat  = eu_cmd_to_modem(ins, EU_CMD_SET_2183_DATA, (DSP_MP_RX_MAILBOX | 0x4000),
                           2, (UInt8 *)&myWord);
    stat  = eu_cmd_to_modem(ins, EU_CMD_SET_2183_DATA, (DSP_SWAP_MAILBOX  | 0x4000),
                           2, (UInt8 *)&myWord);

    /*Analog folks say this delay is needed - but they couldn't tell us why!*/
    mdelay(1000);

    /*Finally, we get to do something that seems substantial!*/
    IdmaUploadMainPage(ins);
    /*********************************************************************************/
    /*This is the last operation in our sequence of tasks for device		     */
    /*initialization. Everything else will be driven by the hardware.		     */
    /*Once the main page is loaded, the hardware will start requesting		     */
    /*swap pages.								     */
    /*Put our modem in the BOOT_WAIT state and start our watchdog!		     */
    /*********************************************************************************/
    ins->AdiModemSm.CurrentAdiState = STATE_BOOT_WAIT;
    SetTimerInterval(ins, TRANS_TIME_4_BOOTWAIT);

end_boot:
    eu_leaves (DBG_BOOT);
    return stat;
}
#endif /* 0 */
 
/*********************************************************************************/
/* IdmaUploadMainPage								 */
/*										 */
/* Sends the Main page blocks of the DSP code through the IDMA interface.        */
/* Note that the bulk of this logic is straight from Buffer.c in the NDIS code,  */
/* there was virtually no documentation on how this is supposed to work.	 */
/*********************************************************************************/
static void IdmaUploadMainPage(eu_instance_t *ins)
{
    int stat;
    UInt32 i, myBlockSize;
    UInt8 *pLastBits = NULL;
    UInt8 *pBlockData;
    IDMALoadBlockInfo myBLI;

    eu_enters (DBG_BOOT);

    eu_dbg (DBG_BOOT,"Loading 0x%#x blocks from main page.\n",
             ins->MainPage.BlockCount);
    

    for (i=0; i<ins->MainPage.BlockCount; i++)
    {
        /*Init data that is not block-location-specific*/
        myBLI.ADIHeader = cpu_to_le16(0xABCD);
        myBLI.Ovl       = 0;
        myBLI.OvlOffset = cpu_to_le16(0x8000);
        myBLI.LastBlock = 0;

        /*Is this NOT the first block?*/
        if (ins->MainPage.Blocks[i].DSPAddr != 0)
        {
           myBLI.Size  = cpu_to_le16((UInt16)(ins->MainPage.Blocks[i].DSPSize));
           myBLI.Addr  = cpu_to_le16(ins->MainPage.Blocks[i].DSPAddr);
           myBlockSize = (UInt16)(ins->MainPage.Blocks[i].DSPSize);
           pBlockData  = ins->MainPage.Blocks[i].MemOffset;
        }
        /*This is block 0 - we want to send all but offset 0 - so we start at offset 1*/
        /*We should only get here once!*/
        else
        {
           myBLI.Size  = cpu_to_le16((UInt16)(ins->MainPage.Blocks[i].DSPSize - 3));
           myBLI.Addr  = cpu_to_le16(1);
           myBlockSize = (UInt16)(ins->MainPage.Blocks[i].DSPSize - 3);
           pBlockData  = ins->MainPage.Blocks[i].MemOffset+3;
           pLastBits   = ins->MainPage.Blocks[i].MemOffset;
        }

        /*Write the structure telling the DSP what to expect in the next block*/
        stat = IDMAWrite(ins, (void *)&myBLI, sizeof(IDMALoadBlockInfo));
        if (stat != 0)
	    eu_err ("Error %#X on IDMA write of BlockLoadInfo\n", stat);

        /*********************************************************************************/
        /* Occassionally the modem will freak out during idma boot and get in 		 */
        /* a weird mode where it requests wrong swap pages, etc.			 */
        /* Comparing the usb traces of linux and Windows reveals that linux		 */
        /* does things MUCH faster, so my theory is that we're overwhelming		 */
        /* the hardware. So, I'll inject a tiny delay here to keep things		 */
        /* under control								 */
        /*********************************************************************************/
        mdelay(1); 

        /*Now write the actual block data*/
        stat = IDMAWrite(ins, (void *)pBlockData, myBlockSize);
        if (stat != 0)
	    eu_err ("Error %#X on IDMA write of block data\n", stat);

        mdelay(1);
    }

    /*All blocks are loaded, now we can write out offset 0 of the first block,    */
    /*which will tell the hardware to start - resulting in requests for swap pages*/
    myBLI.ADIHeader = cpu_to_le16(0xABCD);
    myBLI.Ovl       = 0;
    myBLI.OvlOffset = cpu_to_le16(0x8000);
    myBLI.LastBlock = cpu_to_le16(1);
    myBLI.Size      = cpu_to_le16(3);
    myBLI.Addr      = 0;
    myBlockSize     = 3;
    pBlockData      = pLastBits;
    eu_dbg (DBG_BOOT,"sending last entry from main page, DSP will start soon!.\n");
    

    /*Write the structure telling the DSP what to expect in the next block*/
    stat = IDMAWrite(ins, (void *)&myBLI, sizeof(IDMALoadBlockInfo));
    if (stat != 0)
	eu_err ("Error %#X on IDMA write of BlockLoadInfo\n", stat);
    /*Now write the actual block data*/
    stat = IDMAWrite(ins, (void *)pBlockData, myBlockSize);
    if (stat != 0)
	eu_err ("Error %#X on IDMA write of block data\n", stat);

    eu_leaves (DBG_BOOT);
}

/*********************************************************************************/
/* IdmaUploadSwapPage								 */
/*										 */
/* Sends the blocks of the requested swap page through the IDMA interface.	 */
/* Note that the bulk of this logic is straight from Buffer.c in the NDIS code,  */
/* there was virtually no documentation on how this is supposed to work.	 */
/*********************************************************************************/
void IdmaUploadSwapPage(eu_instance_t *ins, UInt16 SwapInfo)
{
    int stat;
    UInt32 i, myBlockSize;
    UInt16 PageIndex, Ovl;
    IDMALoadBlockInfo myBLI;

    eu_enters (DBG_BOOT);

    PageIndex = (SwapInfo & 0x00FF) - 1;
    Ovl       = ((SwapInfo & 0xF000) >> 12) | ((SwapInfo & 0x0F00) >> 4);

    eu_dbg (DBG_BOOT,"Loading 0x%x blocks from swap page 0x%x, ovl=0x%x.\n", 
             ins->pSwapPages[PageIndex].BlockCount, PageIndex, Ovl);
    
    if (PageIndex > ins->SwapPageCount)
    {
       eu_err ("Request for swap page 0x%x when only 0x%x pages exist!\n",
               PageIndex, ins->SwapPageCount);
       
       return;
    }
    
    /*For swap pages, we actually know we won't have more than 2 blocks*/
    for (i=0; i<ins->pSwapPages[PageIndex].BlockCount; i++)
    {
        /*Init data that is not block-location-specific*/
        myBLI.ADIHeader = cpu_to_le16(0xABCD);
        myBLI.Ovl       = cpu_to_le16(Ovl);
        myBLI.OvlOffset = cpu_to_le16(0x8000 | Ovl);
        myBlockSize     = ins->pSwapPages[PageIndex].Blocks[i].DSPExtendedSize;
        myBLI.Size      = cpu_to_le16((UInt16)myBlockSize);
        myBLI.Addr      = cpu_to_le16(ins->pSwapPages[PageIndex].Blocks[i].DSPAddr);

        if ((ins->pSwapPages[PageIndex].BlockCount-1) == i)
           myBLI.LastBlock = cpu_to_le16(1);
        else
           myBLI.LastBlock = cpu_to_le16(0);

        /*Write the structure telling the DSP what to expect in the next block*/
        stat = IDMAWrite(ins, (void *)&myBLI, sizeof(IDMALoadBlockInfo));
        if (stat != 0)
	    eu_err ("Error %#X on IDMA write of BlockLoadInfo\n", stat);

        /*********************************************************************************/
        /* Occassionally the modem will freak out during idma boot and get in 		 */
        /* a weird mode where it requests wrong swap pages, etc.			 */
        /* Comparing the usb traces of linux and Windows reveals that linux		 */
        /* does things MUCH faster, so my theory is that we're overwhelming		 */
        /* the hardware. So, I'll inject a tiny delay here to keep things		 */
        /* under control								 */
        /*********************************************************************************/
        mdelay(1); 

        /*Now write the actual block data*/
        stat = IDMAWrite(ins, (void *)(ins->pSwapPages[PageIndex].Blocks[i].MemOffset), myBlockSize);
        if (stat != 0)
	    eu_err ("Error %#X on IDMA write of block data\n", stat);

        mdelay(1);
    }

    eu_leaves (DBG_BOOT);
}


/*********************************************************************************/
/* IDMAWrite									 */
/*										 */
/* Responsible for sending data to the IDMA interface via the BULK IDMA OUT pipe.*/
/*********************************************************************************/
static int IDMAWrite(eu_instance_t *ins, void *pData, UInt32 Size)
{
    int ret = 0;
    struct urb *urb;
    UInt8 *xfer_buf = GET_KBUFFER(Size);

    eu_enters (DBG_BOOT);

    if (!xfer_buf)
    {
       eu_err ("kmalloc in IDMAWrite for %#X bytes failed\n", Size);
       
       ret = -ENOMEM;
       goto IDMA_exit;
    }

    memcpy(xfer_buf, pData, Size);

    /* Get an URB and prepare it for submission*/
    urb = USB_ALLOC_URB ( 0 , GFP_ATOMIC );

    usb_fill_bulk_urb ( urb, ins->usbdev, ins->pipe_bulk_idma_out,
                        xfer_buf, Size, IDMACompletion, ins );
    
    urb->transfer_flags |= USB_QUEUE_BULK;

    eu_dbg (DBG_BOOT,"Submitting BULK URB %p\n", urb);
    
    ret = USB_SUBMIT_URB ( urb, GFP_ATOMIC );

IDMA_exit:
    eu_leaves (DBG_BOOT);
    return ret;
}


/*********************************************************************************/
/* IDMACompletion								 */
/*										 */
/* Called when an IDMAWrite urb has completed. All we have to do is free	 */
/* the memory allocated in IDMAWrite.						 */
/*********************************************************************************/
#ifdef LINUX_2_4
static void IDMACompletion(struct urb *urb)
#elif defined(LINUX_2_6)
static void IDMACompletion(struct urb *urb, struct pt_regs *regs )
#endif 
{
    eu_enters (DBG_BOOT);

    /* All we really need to do is free the memory we alloced and free the urb*/
    FREE_KBUFFER(urb->transfer_buffer);
    usb_free_urb(urb);

    eu_dbg (DBG_BOOT,"Freed BULK URB %p\n", urb);
    

    eu_leaves (DBG_BOOT);
}

