/******************************************************************************
  
  Copyright(c) 2003 - 2004 Intel Corporation. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of version 2 of the GNU General Public License as 
  published by the Free Software Foundation.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information:
  James P. Ketrenos <ipw2100-admin@linux.intel.com>
  Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497

******************************************************************************/

/*
 * This is the main type def for the ipw2200 driver.  It will be included
 * by all the modules in this driver.  When/if this driver makes it into the
 * kernel proper, I what the code structured in such a manner that one
 * can practically concatenate all the files starting with ipw2200.h and 
 * ending with ipw2200_main.c and have a single file driver acceptable for
 * inclusion into the kernel tree with only a few clean ups.
 * 
 * mgross@linux.intel.com
 */
#include <linux/module.h>
#include <linux/config.h>
#include <linux/init.h>

#include <linux/version.h>
#include <linux/proc_fs.h>
#include <linux/pci.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/delay.h>

#include <linux/firmware.h>
#include <linux/wireless.h>

#include <net/iw_handler.h>

#define DRV_NAME	"ipw2200"

struct alive_command_responce {
	u8 alive_command;
	u8 sequence_number;
	u16 software_revision;
	u8 device_identifier;
	u8 reserved1[5];
	u16 reserved2;
	u16 reserved3;
	u16 clock_settle_time;
	u16 powerup_settle_time;
	u16 reserved4;
	u8 time_stamp[5];	/* month, day, year, hours, minutes */
	u8 ucode_valid;
} __attribute__ ((packed));


struct command_block
{
	unsigned int control;
	void *source_addr;
	void *dest_addr;
	unsigned int status;
};

#define CB_NUMBER_OF_ELEMENTS_SMALL 64
struct fw_image_desc
{
	unsigned long last_cb_index;
	unsigned long current_cb_index;
	struct command_block cb_list[CB_NUMBER_OF_ELEMENTS_SMALL];
	void * v_addr;
	unsigned long p_addr;
	unsigned long len;
};


struct ipw2200_private {
	/* spinlock */
	spinlock_t lock;

	/* basic pci-network driver stuff */
	struct pci_dev *pci_dev;
	struct net_device *net_dev;

	/* pci harware address support */
	unsigned long hw_base;	/* (virt) */
	unsigned long hw_len;
	
	unsigned long interrupts_enabled;

	struct fw_image_desc sram_desc;

	/* result of ucode download */
	struct alive_command_responce dino_alive;

	struct pci_device_id hw_id;

	u8 eeprom[0x100];  /* 256 bytes of eeprom */  
	int eeprom_delay;

	struct iw_statistics wstats; 

	struct work_struct rx_transfer_work;
	struct work_struct tx_cmd_queue_work;
	struct work_struct tx_queue_1_work;
	struct work_struct tx_queue_2_work;
	struct work_struct tx_queue_3_work;
	struct work_struct tx_queue_4_work;
	struct work_struct status_change_work;
	struct work_struct beacon_period_expired_work;
	struct work_struct slave_mode_cmd_done_work;
	struct work_struct fw_init_done_work;
	struct work_struct phy_off_done_work;
	struct work_struct rf_kill_done_work;
	struct work_struct fatal_error_work;
	struct work_struct parity_error_work;

};				/*ipw2200_private */

/* external functions for driver */
void do_rx_transfer(void *data);
void do_tx_cmd_queue(void *data);
void do_tx_queue_1(void *data);
void do_tx_queue_2(void *data);
void do_tx_queue_3(void *data);
void do_tx_queue_4(void *data);
void do_status_change(void *data);
void do_beacon_period_expired(void *data);
void do_slave_mode_cmd_done(void *data);
void do_fw_init_done(void *data);
void do_phy_off_done(void *data);
void do_rf_kill_done(void *data);
void do_fatal_error(void *data);
void do_parity_error(void *data);

void ipw2200_write_reg32(struct ipw2200_private *priv, u32 reg,
				u32 value);

u32 ipw2200_read_reg32(struct ipw2200_private *priv, u32 reg);
void ipw2200_write(struct ipw2200_private *priv, u32 start, u8 * data,
			  int num);
void ipw2200_clear_bit(struct ipw2200_private *priv, u32 reg, u32 mask);
void ipw2200_set_bit(struct ipw2200_private *priv, u32 reg, u32 mask);

int ipw2200_load(struct ipw2200_private *priv);
int ipw2200_fw_dma_add_buffer(struct ipw2200_private *priv,
	u32 src_phys, u32 dest_address, u32 length);

int ipw2200_fw_dma_kick(struct ipw2200_private *priv);
int ipw2200_fw_dma_wait(struct ipw2200_private *priv);
int ipw2200_fw_dma_enable(struct ipw2200_private *priv);

void ipw2200_eeprom_read(struct ipw2200_private* priv);
void ipw2200_eeprom_parse_mac(struct ipw2200_private* priv, __u8* mac);
void ipw2200_eeprom_parse_version(struct ipw2200_private* priv,__u8* ver);
void ipw2200_eeprom_parse_country(struct ipw2200_private* priv,__u8* code);

/* debug macros */

#ifdef CONFIG_IPW_DEBUG
extern int ipw2200_debug_level;
#define IPW_DEBUG(level, fmt, args...) \
do { if (ipw2200_debug_level & level) \
  printk(KERN_DEBUG DRV_NAME": %c %s " fmt, \
         in_interrupt() ? 'I' : 'U', __FUNCTION__, ## args); } while (0)
extern unsigned long ipw2100_debug_level;
#else
#define IPW_DEBUG(level, fmt, args...) do {} while (0);
#endif				/* CONFIG_IPW_DEBUG */

/*
 * To use the debug system;
 *
 * If you are defining a new debug classification, simply add it to the #define
 * list here in the form of:
 *
 * #define IPW_DL_xxxx VALUE
 * 
 * shifting value to the left one bit from the previous entry.  xxxx should be
 * the name of the classification (for example, WEP)
 *
 * You then need to either add a IPW_xxxx_DEBUG() macro definition for your
 * classification, or use IPW_DEBUG(IPW_DL_xxxx, ...) whenever you want
 * to send output to that classification.
 *
 * To add your debug level to the list of levels seen when you perform
 *
 * % cat /proc/net/ipw2200/debug_level
 *
 * you simply need to add your entry to the ipw2100_debug_levels array.
 *
 * If you do not see debug_level in /proc/net/ipw2100 then you do not have 
 * CONFIG_IPW_DEBUG defined in your kernel configuration
 *
 */

#define IPW_DL_ERROR         0x00000001
#define IPW_DL_WARNING       0x00000002
#define IPW_DL_INFO          0x00000004
#define IPW_DL_WIRELESS_EXT  0x00000008
#define IPW_DL_ISR           0x00000010
#define IPW_DL_SCAN          0x00000020
#define IPW_DL_STATUS        0x00000040
#define IPW_DL_TRACE         0x00000080
#define IPW_DL_RX            0x00000100
#define IPW_DL_TX            0x00000200
#define IPW_DL_MANAGE        0x00000400
#define IPW_DL_WEP           0x00000800
#define IPW_DL_FW_COMMAND    0x00001000
#define IPW_DL_FRAG          0x00002000
#define IPW_DL_FW            0x00004000
#define IPW_DL_RF_KILL       0x00008000
#define IPW_DL_DROP          0x00010000
#define IPW_DL_IO            0x00020000

#define IPW_DEBUG_ERROR(f, a...)   IPW_DEBUG(IPW_DL_ERROR, f, ## a)
#define IPW_DEBUG_WARNING(f, a...) IPW_DEBUG(IPW_DL_WARNING, f, ## a)
#define IPW_DEBUG_INFO(f, a...)    IPW_DEBUG(IPW_DL_INFO, f, ## a)

#define IPW_DEBUG_WX(f, a...)     IPW_DEBUG(IPW_DL_WIRELESS_EXT, f, ## a)
#define IPW_DEBUG_SCAN(f, a...)   IPW_DEBUG(IPW_DL_SCAN, f, ## a)
#define IPW_DEBUG_STATUS(f, a...) IPW_DEBUG(IPW_DL_STATUS, f, ## a)
#define IPW_DEBUG_TRACE(f, a...)  IPW_DEBUG(IPW_DL_TRACE, f, ## a)
#define IPW_DEBUG_RX(f, a...)     IPW_DEBUG(IPW_DL_RX, f, ## a)
#define IPW_DEBUG_TX(f, a...)     IPW_DEBUG(IPW_DL_TX, f, ## a)
#define IPW_DEBUG_ISR(f, a...)    IPW_DEBUG(IPW_DL_ISR, f, ## a)
#define IPW_DEBUG_MANAGEMENT(f, a...) IPW_DEBUG(IPW_DL_MANAGE, f, ## a)
#define IPW_DEBUG_WEP(f, a...)    IPW_DEBUG(IPW_DL_WEP, f, ## a)
#define IPW_DEBUG_FW_COMMAND(f, a...) IPW_DEBUG(IPW_DL_FW_COMMAND, f, ## a)
#define IPW_DEBUG_FRAG(f, a...) IPW_DEBUG(IPW_DL_FRAG, f, ## a)
#define IPW_DEBUG_FW(f, a...) IPW_DEBUG(IPW_DL_FW, f, ## a)
#define IPW_DEBUG_RF_KILL(f, a...) IPW_DEBUG(IPW_DL_RF_KILL, f, ## a)
#define IPW_DEBUG_DROP(f, a...) IPW_DEBUG(IPW_DL_DROP, f, ## a)
#define IPW_DEBUG_IO(f, a...) IPW_DEBUG(IPW_DL_IO, f, ## a)

/*
* Register bit definitions
*/

/* Dino control registers bits */

#define DINO_ENABLE_SYSTEM 0x80
#define DINO_ENABLE_CS     0x40
#define DINO_RXFIFO_DATA   0x01	
#define DINO_CONTROL_REG   0x00200000

#define CX2_INTA_RW       0x00000008
#define CX2_INTA_MASK_R   0x0000000C
#define CX2_INDIRECT_ADDR 0x00000010
#define CX2_INDIRECT_DATA 0x00000014
#define CX2_AUTOINC_ADDR  0x00000018
#define CX2_AUTOINC_DATA  0x0000001C
#define CX2_RESET_REG     0x00000020
#define CX2_GP_CNTRL_RW   0x00000024

#define CX2_READ_INT_REGISTER 0xFF4

#define CX2_GP_CNTRL_BIT_INIT_DONE	0x00000004

#define CX2_SHARED_LOWER_BOUND          0x00000200
#define CLX2_REGISTER_DOMAIN1_END       0x00001000
#define CX2_INTERRUPT_AREA_LOWER_BOUND  0x00000f80
#define CX2_SRAM_READ_INT_REGISTER 		0x00000ff4
#define CX2_BIT_INT_HOST_SRAM_READ_INT_REGISTER (1 << 29)
#define CX2_GP_CNTRL_BIT_CLOCK_READY    0x00000001
#define CX2_GP_CNTRL_BIT_HOST_ALLOWS_STANDBY 0x00000002

/*
 * RESET Register Bit Indexes
 */
#define CBD_RESET_REG_PRINCETON_RESET 0x00000001  /* Bit 0 (LSB) */
#define CX2_RESET_REG_SW_RESET        0x00000080  /* Bit 7       */
#define CX2_RESET_REG_MASTER_DISABLED 0x00000100  /* Bit 8       */
#define CX2_RESET_REG_STOP_MASTER     0x00000200  /* Bit 9       */
#define CX2_ARC_KESHET_CONFIG         0x08000000  /* Bit 27      */
#define CX2_START_STANDBY             0x00000004  /* Bit 2       */

#define CX2_CSR_CIS_UPPER_BOUND	0x00000200
#define CX2_DOMAIN_0_END 0x1000

#define CX2_BASEBAND_CONTROL_STATUS	0X00200000
#define CX2_BASEBAND_TX_FIFO_WRITE	0X00200004
#define CX2_BASEBAND_RX_FIFO_READ	0X00200004
#define CX2_BASEBAND_CONTROL_STORE	0X00200010

#define CX2_INTERNAL_CMD_EVENT 	0X00300004
#define CX2_BASEBAND_POWER_DOWN 0x00000001

#define CX2_MEM_HALT_AND_RESET  0x003000e0

/* defgroup bits_halt_reset MEM_HALT_AND_RESET register bits */
#define CX2_BIT_HALT_RESET_ON	0x80000000
#define CX2_BIT_HALT_RESET_OFF 	0x00000000

#define CB_LAST_VALID     0x20000000
#define CB_INT_ENABLED    0x40000000
#define CB_VALID          0x80000000
#define CB_SRC_LE         0x08000000
#define CB_DEST_LE        0x04000000
#define CB_SRC_AUTOINC    0x00800000
#define CB_SRC_IO_GATED   0x00400000
#define CB_DEST_AUTOINC   0x00080000
#define CB_SRC_SIZE_LONG  0x00200000
#define CB_DEST_SIZE_LONG 0x00020000


/* DMA DEFINES */

#define DMA_CONTROL_SMALL_CB_CONST_VALUE 0x00540000
#define DMA_CB_STOP_AND_ABORT            0x00000C00
#define DMA_CB_START                     0x00000100 


#define CX2_SHARED_SRAM_SIZE               0x00030000
#define CX2_SHARED_SRAM_DMA_CNTRL          0x00027000
#define CX2_DMA_CONTROL_SHARED_LOWER_BOUND 0x00027000
#define CB_MAX_LENGTH                      0x1FFF

#define CX2_HOST_EEPROM_DATA_SRAM_SIZE 0xA18
#define CX2_EEPROM_IMAGE_SIZE          0x100


/* DMA defs */
#define CX2_DMA_I_CURRENT_CB  0x003000D0
#define CX2_DMA_O_CURRENT_CB  0x003000D4
#define CX2_DMA_I_DMA_CONTROL 0x003000A4
#define CX2_DMA_I_CB_BASE     0x003000A0

/*
 * EEPROM Related Definitions
 */

#define FW_MEM_REG_LOWER_BOUND          0x00300000
#define FW_MEM_REG_EEPROM_ACCESS        (FW_MEM_REG_LOWER_BOUND + 0x40)        

#define EEPROM_BIT_SK                   __BIT(0)
#define EEPROM_BIT_CS                   __BIT(1)
#define EEPROM_BIT_DI                   __BIT(2)
#define EEPROM_BIT_DO                   __BIT(4)

#define EEPROM_CMD_READ                 0x2

/* Interestings Offsets into the eeprom */
#define EEPROM_PME_CAPABILITY           0x13
#define EEPROM_MAC_ADDRESS              0x42    
#define EEPROM_VERSION                  0x49
#define EEPROM_NIC_TYPE                 0x4A     /* 1 byte   */
#define EEPROM_SKU_CAPABILITY           0x4B     /* 1 byte   */    
#define EEPROM_COUNTRY_CODE             0x4C     /* 3 bytes  */
#define EEPROM_IBSS_CHANNELS_BG         0x50     /* 2 bytes  */
#define EEPROM_IBSS_CHANNELS_A          0x53     /* 5 bytes  */
#define EEPROM_BSS_CHANNELS_BG          0x58     /* 2 bytes  */
#define EEPROM_BSS_CHANNELS_A           0x5B     /* 5 bytes  */
#define EEPROM_HW_VERSION               0xE4     /* 2 bytes  */

/* Defines a single bit in a by bit number (0-31) */
#define __BIT(x)                (1UL << (x))

/* OEM NIC types */
#define EEPROM_NIC_TYPE_STANDARD                0
#define EEPROM_NIC_TYPE_DELL                    1
#define EEPROM_NIC_TYPE_FUJITSU                 2
#define EEPROM_NIC_TYPE_IBM                     3
#define EEPROM_NIC_TYPE_HP                      4

/* SKU Capabilities */
#define EEPROM_SKU_CAP_BT_CHANNEL_SIGNALING     __BIT(0)
#define EEPROM_SKU_CAP_BT_PRIORITY              __BIT(1)
#define EEPROM_SKU_CAP_BIT_2                    __BIT(2)
#define EEPROM_SKU_CAP_BIT_3                    __BIT(3)
#define EEPROM_SKU_CAP_CPU_LOCK                 __BIT(4)
#define EEPROM_SKU_CAP_DUAL_BAND                __BIT(5)
#define EEPROM_SKU_CAP_RF_KILL                  __BIT(6)
#define EEPROM_SKU_CAP_REGULATORY_OVERRIDE      __BIT(7)

/* IBSS channels */
#define EEPROM_IBSS_CHANNELS_BG_ENABLE          __BIT(15)
#define EEPROM_IBSS_CHANNELS_A_ENABLE           __BIT(7)




/* Interrupts masks */
/* Interrupts masks */
/* Interrupts masks */
#define CX2_INTA_NONE   0x00000000

#define CX2_INTA_BIT_RX_TRANSFER                   0x00000002
#define CX2_INTA_BIT_STATUS_CHANGE                 0x00000010
#define CX2_INTA_BIT_BEACON_PERIOD_EXPIRED         0x00000020

//Inta Bits for CF
#define CX2_INTA_BIT_TX_CMD_QUEUE                  0x00000800
#define CX2_INTA_BIT_TX_QUEUE_1                    0x00001000
#define CX2_INTA_BIT_TX_QUEUE_2                    0x00002000
#define CX2_INTA_BIT_TX_QUEUE_3                    0x00004000
#define CX2_INTA_BIT_TX_QUEUE_4                    0x00008000

#define CX2_INTA_BIT_SLAVE_MODE_HOST_CMD_DONE      0x00010000

#define CX2_INTA_BIT_PREPARE_FOR_POWER_DOWN        0x00100000
#define CX2_INTA_BIT_POWER_DOWN                    0x00200000

#define CX2_INTA_BIT_FW_INITIALIZATION_DONE        0x01000000
#define CX2_INTA_BIT_FW_CARD_DISABLE_PHY_OFF_DONE  0x02000000
#define CX2_INTA_BIT_RF_KILL_DONE                  0x04000000
#define CX2_INTA_BIT_FATAL_ERROR                   0x40000000
#define CX2_INTA_BIT_PARITY_ERROR                  0x80000000

/* Interrupts enabled at init time. */
#define CX2_INTA_MASK_ALL                        \
        (CX2_INTA_BIT_TX_QUEUE_1               | \
	 CX2_INTA_BIT_TX_QUEUE_2               | \
	 CX2_INTA_BIT_TX_QUEUE_3               | \
	 CX2_INTA_BIT_TX_QUEUE_4               | \
	 CX2_INTA_BIT_TX_CMD_QUEUE             | \
	 CX2_INTA_BIT_RX_TRANSFER              | \
	 CX2_INTA_BIT_FATAL_ERROR              | \
	 CX2_INTA_BIT_PARITY_ERROR             | \
	 CX2_INTA_BIT_STATUS_CHANGE            | \
	 CX2_INTA_BIT_FW_INITIALIZATION_DONE   | \
	 CX2_INTA_BIT_BEACON_PERIOD_EXPIRED    | \
	 CX2_INTA_BIT_SLAVE_MODE_HOST_CMD_DONE | \
	 CX2_INTA_BIT_PREPARE_FOR_POWER_DOWN   | \
	 CX2_INTA_BIT_POWER_DOWN               | \
         CX2_INTA_BIT_RF_KILL_DONE )
