/******************************************************************************
  
  Copyright(c) 2003 - 2004 Intel Corporation. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of version 2 of the GNU General Public License as 
  published by the Free Software Foundation.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information:
  James P. Ketrenos <ipw2100-admin@linux.intel.com>
  Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497

******************************************************************************/

#include "ipw2200.h"

/*
 * The IPW2200 device contains a Microwire compatible EEPROM that stores
 * various data like the MAC address.  Usually the firmware has exclusive
 * access to the eeprom, but during device initialization (before the
 * device driver has sent the HostComplete command to the firmware) the
 * device driver has read access to the EEPROM by way of indirect addressing
 * through a couple of memory mapped registers.
 *
 * The following is a simplified implementation for pulling data out of the
 * the eeprom, along with some helper functions to find information in
 * the per device private data's copy of the eeprom.
 *
 * NOTE: To better understand how these functions work (i.e what is a chip
 *       select and why do have to keep driving the eeprom clock?), read
 *       just about any data sheet for a Microwire compatible EEPROM.
 */

/* write a 32 bit value into the indirect accessor register */
static inline void __eeprom_write_reg(struct ipw2200_private *p, u32 data)
{
	ipw2200_write_reg32(p,FW_MEM_REG_EEPROM_ACCESS,data);
	
	/* the eeprom requires some time to complete the operation */
	udelay(p->eeprom_delay);

	return;
}

/* read a 32 bit value from the indirect accessor register */
static inline u32 __eeprom_read_reg(struct ipw2200_private *p)
{
	u32 data = ipw2200_read_reg32(p,FW_MEM_REG_EEPROM_ACCESS);

	/* the eeprom requires some time to complete the operation */
	udelay(p->eeprom_delay);

	return data;
}

/* perform a chip select operation */
static inline void ipw2200_eeprom_cs(struct ipw2200_private* priv)
{
	__eeprom_write_reg(priv,0);
	__eeprom_write_reg(priv,EEPROM_BIT_CS);
	__eeprom_write_reg(priv,EEPROM_BIT_CS|EEPROM_BIT_SK);
	__eeprom_write_reg(priv,EEPROM_BIT_CS);
}

/* perform a chip select operation */
static inline void ipw2200_eeprom_disable_cs(struct ipw2200_private* priv)
{
	__eeprom_write_reg(priv,EEPROM_BIT_CS);
	__eeprom_write_reg(priv,0);
	__eeprom_write_reg(priv,EEPROM_BIT_SK);
}

/* push a single bit down to the eeprom */
static inline void ipw2200_eeprom_write_bit(struct ipw2200_private *p,__u8 bit)
{
	int d = ( bit ? EEPROM_BIT_DI : 0);
	__eeprom_write_reg(p,EEPROM_BIT_CS|d);
	__eeprom_write_reg(p,EEPROM_BIT_CS|d|EEPROM_BIT_SK);
}

/* push an opcode followed by an address down to the eeprom */
static void ipw2200_eeprom_op(struct ipw2200_private* priv, __u8 op, __u8 addr)
{
	int i;

	ipw2200_eeprom_cs(priv);
	ipw2200_eeprom_write_bit(priv,1);
	ipw2200_eeprom_write_bit(priv,op&2);
	ipw2200_eeprom_write_bit(priv,op&1);
	for ( i=7; i>=0; i-- ) {
		ipw2200_eeprom_write_bit(priv,addr&(1<<i));
	}
}

/* pull 16 bits off the eeprom, one bit at a time */
static __u16 ipw2200_eeprom_read_u16(struct ipw2200_private* priv, __u8 addr)
{
	int i;
	__u16 r=0;
	
	/* Send READ Opcode */
	ipw2200_eeprom_op(priv,EEPROM_CMD_READ,addr);

	/* Send dummy bit */
	__eeprom_write_reg(priv,EEPROM_BIT_CS);

	/* Read the byte off the eeprom one bit at a time */
	for ( i=0; i<16; i++ ) {
		u32 data = 0;
		__eeprom_write_reg(priv,EEPROM_BIT_CS|EEPROM_BIT_SK);
		__eeprom_write_reg(priv,EEPROM_BIT_CS);
		data = ipw2200_read_reg32(priv,FW_MEM_REG_EEPROM_ACCESS);
		r = (r<<1) | ((data & EEPROM_BIT_DO)?1:0);
	}
	
	/* Send another dummy bit */
	__eeprom_write_reg(priv,0);
	ipw2200_eeprom_disable_cs(priv);
	
	return r;
}

/* dump the entire contents of the eeprom into the device's private data */
void ipw2200_eeprom_read(struct ipw2200_private* priv)
{
	int i;
	u16 *eeprom = (u16 *)priv->eeprom;
  
	for ( i=0; i<128; i++ )
		eeprom[i] = ipw2200_eeprom_read_u16(priv,(__u8)i);
}

/* helper function for pulling the mac address out of the private */
/* data's copy of the eeprom data                                 */
void ipw2200_eeprom_parse_mac(struct ipw2200_private* priv, __u8* mac)
{
	__u8* ee=(__u8*)priv->eeprom;
	memcpy(mac,&ee[EEPROM_MAC_ADDRESS],6);
}

/* helper function for pulling the device version out of the private */
/* data's copy of the eeprom data                                    */
void ipw2200_eeprom_parse_version(struct ipw2200_private* priv,__u8* ver)
{
	__u8* ee=(__u8*)priv->eeprom;
	*ver=ee[EEPROM_VERSION];
}

/* helper function for pulling the country code out of the private */
/* data's copy of the eeprom data                                  */
void ipw2200_eeprom_parse_country(struct ipw2200_private* priv,__u8* code)
{
	__u8* ee=(__u8*)priv->eeprom;
	memcpy(code,&ee[EEPROM_COUNTRY_CODE],3);
}

