/******************************************************************************
  
  Copyright(c) 2003 - 2004 Intel Corporation. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of version 2 of the GNU General Public License as 
  published by the Free Software Foundation.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information:
  James P. Ketrenos <ipw2100-admin@linux.intel.com>
  Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497

******************************************************************************/

/* IPW2200 driver interface to the hardware provides the low level support for
 * talking to the BAR loading the firmware, and primitives for talking to the
 * firmware's interface.
 * 
 * mgross@linux.intel.com
 */

#include "ipw2200.h"

#if defined(CONFIG_SMP)
#define CHECK_HWLOCK(priv,reg) \
if (!spin_is_locked(&priv->lock)) { \
  IPW_DEBUG_INFO("Indirect access w/o lock [0x%08x]",reg); \
}
#else				/* CONFIG_SMP */
#define CHECK_HWLOCK(priv,lma)
#endif				/* CONFIG_SMP */

/**
 * Check that card is still alive. 
 * Reads debug register from domain0.
 * If card is present, pre-defined value should
 * be found there.
 * 
 * @param priv
 * @return 1 if card is present, 0 otherwise
 */
static inline int ipw2200_alive(struct ipw2200_private *priv)
{
	return readl(priv->hw_base + 0x90) == 0xd55555d5;
}

/* indirect write s */
void ipw2200_write_reg32(struct ipw2200_private *priv, u32 reg,
			 u32 value)
{
	IPW_DEBUG_IO(" %p : reg = 0x%8X : value = 0x%8X\n", 
			priv, reg, value);
	CHECK_HWLOCK(priv,reg);
	writel(reg, priv->hw_base + CX2_INDIRECT_ADDR);
	wmb();
	writel(value, priv->hw_base + CX2_INDIRECT_DATA );
}


static void ipw2200_write_reg8(struct ipw2200_private *priv, u32 reg, u8 value)
{
	IPW_DEBUG_IO(" reg = 0x%8X : value = 0x%8X\n", reg, value);
	CHECK_HWLOCK(priv,reg);
	writel(reg & ~0x3ul, priv->hw_base + CX2_INDIRECT_ADDR);
	wmb();
	IPW_DEBUG_IO(" reg = 0x%8X : value = 0x%8X\n", 
		       (int)priv->hw_base + CX2_INDIRECT_DATA + (reg&3), 
		       value);
	writeb(value, priv->hw_base + CX2_INDIRECT_DATA + (reg&3));
}

static void ipw2200_write_reg16(struct ipw2200_private *priv, u32 reg,
				u16 value)
{
	IPW_DEBUG_IO(" reg = 0x%8X : value = 0x%8X\n", reg, value);
	CHECK_HWLOCK(priv,reg);
	writel(reg & ~0x2ul, priv->hw_base + CX2_INDIRECT_ADDR);
	wmb();
	writew(value, priv->hw_base + CX2_INDIRECT_DATA + (reg&2));
}

/* indirect read s */

static u8 ipw2200_read_reg8(struct ipw2200_private *priv, u32 reg)
{
	CHECK_HWLOCK(priv,reg);
	writel(reg & ~0x3ul, priv->hw_base + CX2_INDIRECT_ADDR);
	wmb();
	
	IPW_DEBUG_IO(" reg = 0x%8X : \n", reg);

	return readb(priv->hw_base + CX2_INDIRECT_DATA + (reg&3) );
}

u32 ipw2200_read_reg32(struct ipw2200_private *priv, u32 reg)
{
	u32 value;
	CHECK_HWLOCK(priv,reg);

	IPW_DEBUG_IO("%p : reg = 0x%08x\n", priv, reg);

	writel(reg, priv->hw_base + CX2_INDIRECT_ADDR);
	wmb();
	value = readl(priv->hw_base + CX2_INDIRECT_DATA);
	IPW_DEBUG_IO(" reg = 0x%4X : value = 0x%4x \n", reg, value);
	return value;
}


#ifdef NOT_YET
static u16 ipw2200_read_reg16(struct ipw2200_private *priv, u32 reg)
{
	CHECK_HWLOCK(priv,reg);
	writel(reg & ~0x2ul, priv->hw_base + CX2_INDIRECT_ADDR);
	wmb();
	return readw(priv->hw_base + CX2_INDIRECT_DATA + (reg & 2));
}

/* iterative/auto-increment 32 bit reads and writes */
static void ipw2200_read_indirect(struct ipw2200_private *priv, u32 start, u32 * data,
				  int num)
{
	CHECK_HWLOCK(priv,start);
	writel(start, priv->hw_base + CX2_AUTOINC_ADDR);
	wmb();
	while (num-- > 0)
		*data++ = readl(priv->hw_base + CX2_AUTOINC_DATA);
}
#endif

static void ipw2200_read(struct ipw2200_private *priv, u32 start, u32 * data,
			 int num)
{
	int i = 0;

	CHECK_HWLOCK(priv,start);

	while (num-- > 0) {
		*data++ = readl(priv->hw_base + i );
		i += sizeof(u32); 
	};
}
void ipw2200_write(struct ipw2200_private *priv, u32 start, u8 * data,
		   int num)
{
	int i;
	int num_u32s = num/sizeof(u32);
	int num_u8s = num % sizeof(u32);
	u32 *buf = (u32 *) data;

	CHECK_HWLOCK(priv,start);

	if((start&~0x3) != start)
		IPW_DEBUG_INFO("start addrss not allined!!!!\n");

    	IPW_DEBUG_TRACE("start = 0x%4X length = 0x%4X\n", start, num);
	writel(start, priv->hw_base + CX2_INDIRECT_ADDR);
	writel(start, priv->hw_base + CX2_AUTOINC_ADDR);
	wmb();
	for(i = 0; i< num_u32s; i++) {
		IPW_DEBUG_TRACE(": value = 0x%8X\n", *buf);
		writel(*buf++, priv->hw_base + CX2_AUTOINC_DATA);
	}

	data += (num - num_u8s);
	for( i = 0; i < num_u8s; i++)
		ipw2200_write_reg8(priv, num_u32s*sizeof(u32) + i, data[i]);

}

static int ipw2200_poll_bit(struct ipw2200_private *priv, u32 addr, u32 mask,
			    int timeout)
{
	int i;

	CHECK_HWLOCK(priv,addr);

	for (i = 0; i < timeout; i++) {
		u32 reg = readl(priv->hw_base + addr);
		if ((reg & mask) == mask) {
    			IPW_DEBUG_INFO("0x%4X \n", i);
			return i;
		}
		mdelay(1);
	}
	return -ETIME;
}

void ipw2200_set_bit(struct ipw2200_private *priv, u32 reg, u32 mask)
{
	u32 temp;

	CHECK_HWLOCK(priv,reg);
	temp = readl(priv->hw_base + reg);
	temp |= mask;
	writel(temp, priv->hw_base +reg);
	 
	IPW_DEBUG_TRACE("<< reg= 0x%4x mask = 0x%4X value = 0x%4X\n", reg, mask, temp);

}

void ipw2200_clear_bit(struct ipw2200_private *priv, u32 reg, u32 mask)
{
	u32 temp;

	CHECK_HWLOCK(priv,reg);
	temp = readl(priv->hw_base + reg);
	temp &= ~mask;
	writel(temp, priv->hw_base +reg);
	 
	IPW_DEBUG_TRACE("<< reg= 0x%4x mask = 0x%4X value = 0x%4X\n", reg, mask, temp);

}

/* These functions load the firmware and micro code for the operation of 
 * the ipw2200 hardware.  It assumes the buffer has all the bits for the
 * image and the caller is handling the memory allocation and clean up.
 */


int ipw2200_stop_master(struct ipw2200_private * priv)
{
	int rc;
	
	IPW_DEBUG_TRACE(">> \n");
	/* stop master. typical delay - 0 */
	ipw2200_set_bit(priv, CX2_RESET_REG, CX2_RESET_REG_STOP_MASTER);

	rc = ipw2200_poll_bit(priv,CX2_RESET_REG,
			      CX2_RESET_REG_MASTER_DISABLED, 10);
	if (rc >= 0)
		IPW_DEBUG_WARNING("stop master %dms\n", rc);
	else {
		IPW_DEBUG_ERROR("stop master failed in 50ms\n");
		return -1;
	}
	return rc;
}


int ipw2200_arc_halt(struct ipw2200_private *priv)
{
	int rc = ipw2200_stop_master(priv);
	
	IPW_DEBUG_TRACE(">> \n");
	if (rc < 0)
		return rc;
	ipw2200_set_bit(priv, CX2_RESET_REG,
			CBD_RESET_REG_PRINCETON_RESET | CX2_RESET_REG_STOP_MASTER );
	/* no one know timing, for safety add some delay */
	mdelay(5);
	return rc;
}

void ipw2200_arc_release(struct ipw2200_private *priv)
{
	IPW_DEBUG_TRACE(">> \n");
	mdelay(5);

	ipw2200_clear_bit(priv, CX2_RESET_REG, CBD_RESET_REG_PRINCETON_RESET);

	/* no one know timing, for safety add some delay */
	mdelay(5);
	/* enable interrupts */
	
	writel(~0, priv->hw_base + CX2_INTA_MASK_R);
}

struct fw_header {
	u32 version;
	u32 mode;
};

struct fw_chunk {
	u32 address;
	u32 length;
};


/* TODO: UCODE DOWNLOAD SHOULD BE ON WORK QUEUE TO AVOID THE POLING BS 
 * also, the locking and busy waits in this function give me a rash 
 */
static int ipw2200_load_ucode(struct ipw2200_private *priv, u8 * data,
			      size_t len)
{
	int rc = 0;
	int i;
	u8 cr = 0;
	u16 *image;
	unsigned long flags;

	image = (u16 *)( (u32)data + sizeof(struct fw_header));
	
	IPW_DEBUG_TRACE(">> \n");

	rc = ipw2200_stop_master(priv);

	if (rc < 0)
		return rc;
	/* zero whole domain 1 */
	{
		int addr;
		for (addr = CX2_SHARED_LOWER_BOUND;
		     addr < CLX2_REGISTER_DOMAIN1_END; addr += 4) {
			writel(0, priv->hw_base + addr);
		}
	} 

	/* no ucode (yet) */
	memset(&priv->dino_alive, 0, sizeof(priv->dino_alive));
	/* destroy DMA queues */
	/* reset sequence */
	spin_lock_irqsave(&priv->lock, flags);

	ipw2200_write_reg32(priv, CX2_MEM_HALT_AND_RESET ,CX2_BIT_HALT_RESET_ON);
	ipw2200_arc_release(priv);
	ipw2200_write_reg32(priv, CX2_MEM_HALT_AND_RESET,
			    CX2_BIT_HALT_RESET_OFF);

	spin_unlock_irqrestore(&priv->lock, flags);
	mdelay(1);
	spin_lock_irqsave(&priv->lock, flags);

	/* reset PHY */
	ipw2200_write_reg32(priv, CX2_INTERNAL_CMD_EVENT, 
			    CX2_BASEBAND_POWER_DOWN);
	spin_unlock_irqrestore(&priv->lock, flags);
	mdelay(1);
	spin_lock_irqsave(&priv->lock, flags);

	ipw2200_write_reg32(priv, CX2_INTERNAL_CMD_EVENT, 0);
	
	spin_unlock_irqrestore(&priv->lock, flags);
	mdelay(1);
	spin_lock_irqsave(&priv->lock, flags);



	/* enable ucode store */
	ipw2200_write_reg8(priv, DINO_CONTROL_REG, 0x0);
	ipw2200_write_reg8(priv, DINO_CONTROL_REG, DINO_ENABLE_CS);
	spin_unlock_irqrestore(&priv->lock, flags);
	mdelay(1);
	spin_lock_irqsave(&priv->lock, flags);

	
	
	/* write ucode */
	/**
	 * @bug
	 * Do NOT set indirect address register once and then
	 * store data to indirect data register in the loop.
	 * It seems very reasonable, but in this case DINO do not
	 * accept ucode. It is essential to set address each time.
	 */
	/* load new ipw2200 uCode */
	for (i = 0; i < len / 2; i++)
		ipw2200_write_reg16(priv, CX2_BASEBAND_CONTROL_STORE,
				    image[i]);

	
	/* enable DINO */
	ipw2200_write_reg8(priv, CX2_BASEBAND_CONTROL_STATUS, 0);
	ipw2200_write_reg8(priv, CX2_BASEBAND_CONTROL_STATUS,
			   DINO_ENABLE_SYSTEM );

	/* this is where the igx / win driver deveates from the VAP driver.*/

	/* wait for alive response */
	for (i = 0; i < 100; i++) {
		/* poll for incoming data */
		cr = ipw2200_read_reg8(priv, CX2_BASEBAND_CONTROL_STATUS);
		if (cr & DINO_RXFIFO_DATA)
			break;
		spin_unlock_irqrestore(&priv->lock, flags);
		mdelay(1);
		spin_lock_irqsave(&priv->lock, flags);
	}

	if (cr & DINO_RXFIFO_DATA) {
		/* alive_command_responce size is NOT multiple of 4 */
		u32 responce_buffer[(sizeof(priv->dino_alive) + 3) / 4];

		for (i = 0; i < (sizeof(priv->dino_alive) + 3) / 4; i++) {
			responce_buffer[i] =
				ipw2200_read_reg32(priv, CX2_BASEBAND_RX_FIFO_READ);
		}
		memcpy(&priv->dino_alive, responce_buffer,
		       sizeof(priv->dino_alive));
		if (priv->dino_alive.alive_command == 1
		    && priv->dino_alive.ucode_valid == 1) {
			rc = 0;
			IPW_DEBUG_INFO
				("Microcode OK, rev. %d (0x%x) dev. %d (0x%x) of %02d/%02d/%02d %02d:%02d\n",
				 priv->dino_alive.software_revision,
				 priv->dino_alive.software_revision,
				 priv->dino_alive.device_identifier,
				 priv->dino_alive.device_identifier,
				 priv->dino_alive.time_stamp[0],
				 priv->dino_alive.time_stamp[1],
				 priv->dino_alive.time_stamp[2],
				 priv->dino_alive.time_stamp[3],
				 priv->dino_alive.time_stamp[4]);
		} else {
			IPW_DEBUG_INFO("Microcode is not alive\n");
			rc = -EINVAL;
		}
	} else {
		IPW_DEBUG_INFO("No alive response from DINO\n");
		rc = -ETIME;
	}
	/* disable DINO, otherwise for some reason
	   firmware have problem getting alive resp. */
	ipw2200_write_reg8(priv, CX2_BASEBAND_CONTROL_STATUS, 0);
	spin_unlock_irqrestore(&priv->lock, flags);
	IPW_DEBUG_TRACE(">>\n");

	return rc;
}

static int ipw2200_load_firmware(struct ipw2200_private *priv, u8 * data,
				 size_t len)
{
	int rc = -1;
	int offset = 0;
	struct fw_header *header;
	struct fw_chunk *chunk;
	u32 shared_phys;
	u8 *shared_virt;
	
	IPW_DEBUG_TRACE("<< : \n");
	shared_virt = pci_alloc_consistent(priv->pci_dev, len, &shared_phys);

	if(!shared_virt)
		return -1;

	memmove(shared_virt, data, len);

	/* Start the Dma */
	rc = ipw2200_fw_dma_enable( priv );

	header = (struct fw_header *) (data + offset);
	offset += sizeof(struct fw_header);
	
	if (priv->sram_desc.last_cb_index > 0)
	{
		/* the DMA is already ready this would be a bug. */
		BUG();
		goto out;
	}

	do {
		chunk = (struct fw_chunk *)(data + offset);
		offset += sizeof(struct fw_chunk);
		/* build DMA packet and queue up for sending */
		/* dma to chunk->address, the chunk->length bytes  from data + offeset*/
		/* Dma loading */
		rc = ipw2200_fw_dma_add_buffer(priv, shared_phys + offset,
					       chunk->address, chunk->length);
		if (rc)
		{
			IPW_DEBUG_INFO("imgLoadFw : dmaAddBuffer Failed\n");
			goto out;
		}
		
		offset += chunk->length;
	} while (offset < len);


	/* Run the DMA and wait for the answer*/
	rc = ipw2200_fw_dma_kick(priv);
	if (rc )
	{
		IPW_DEBUG_INFO(">>imgLoadFw : dmaKick Failed\n");
		goto out;
	}

	rc = ipw2200_fw_dma_wait(priv);
	if ( rc )
	{
		IPW_DEBUG_INFO("imgLoadFw : dmaWaitSync Failed\n");
		goto out;
	}
 out:
	pci_free_consistent( priv->pci_dev, len, shared_virt, shared_phys);
	return rc;
}

/* stop nic */
static int ipw2200_stop_nic(struct ipw2200_private *priv)
{
	int rc = 0;

	IPW_DEBUG_TRACE(">>\n");
	/* stop*/
	writel(CX2_RESET_REG_STOP_MASTER, priv->hw_base + CX2_RESET_REG);

	rc = ipw2200_poll_bit(priv, CX2_RESET_REG, 
			      CX2_RESET_REG_MASTER_DISABLED, 500); 
	if( 0 > rc)   
	{
		IPW_DEBUG_INFO("wait for reg master disabled failed\n");
		return rc;
	}   

	ipw2200_set_bit(priv, CX2_RESET_REG,CBD_RESET_REG_PRINCETON_RESET );
	
	IPW_DEBUG_TRACE("<<\n");

	return rc;
}

static void ipw2200_start_nic(struct ipw2200_private *priv)
{
	IPW_DEBUG_TRACE(">>\n");

	/* prvHwStartNic  release ARC*/
	ipw2200_clear_bit(priv, CX2_RESET_REG,
			  CX2_RESET_REG_MASTER_DISABLED | 
			  CX2_RESET_REG_STOP_MASTER | 
			  CBD_RESET_REG_PRINCETON_RESET);
	
	/* enable power management */
	ipw2200_set_bit(priv, CX2_GP_CNTRL_RW, CX2_GP_CNTRL_BIT_HOST_ALLOWS_STANDBY);
	IPW_DEBUG_TRACE("<<\n");
}


static int ipw2200_fw_start(struct ipw2200_private *priv)
{
	int rc;

	IPW_DEBUG_TRACE(">>\n");

	/* reset queues */
	/* TODO: */

	/*start nic*/
	ipw2200_start_nic(priv);

	/* start NIC prvHwStartNic */
	rc = ipw2200_poll_bit(priv,CX2_INTA_MASK_R,
			      CX2_INTA_BIT_FW_INITIALIZATION_DONE , 50000);
	if (rc >= 0)
		IPW_DEBUG_WARNING("fw initialized in %dms\n", rc);
	else {
		IPW_DEBUG_ERROR("fw not initialized in 50000ms\n");
		return rc;
	}

	priv->interrupts_enabled = 1;
	writel(CX2_INTA_BIT_FW_INITIALIZATION_DONE, priv->hw_base + CX2_INTA_RW);
	writel(CX2_INTA_MASK_ALL, priv->hw_base + CX2_INTA_MASK_R); /* ack them all */

	IPW_DEBUG_TRACE("<<\n");
	
	return 0;
}
	
static int ipw2200_init_nic(struct ipw2200_private *priv)
{
	int rc;

	IPW_DEBUG_TRACE(">>\n");
	/* reset */	
	/*prvHwInitNic */
	/* set "initialization complete" bit to move adapter to D0 state */
	ipw2200_set_bit(priv, CX2_GP_CNTRL_RW, CX2_GP_CNTRL_BIT_INIT_DONE);


	/* low-level PLL activation */
	writel( CX2_BIT_INT_HOST_SRAM_READ_INT_REGISTER,
		priv->hw_base + CX2_READ_INT_REGISTER );

	/* wait for clock stabilization */
	rc = ipw2200_poll_bit(priv, CX2_GP_CNTRL_RW, 
			      CX2_GP_CNTRL_BIT_CLOCK_READY, 25000); 
	if(rc < 0 )
		IPW_DEBUG_INFO("FAILED waite for clock stablization\n");

	/* assert SW reset */
	ipw2200_set_bit(priv, CX2_RESET_REG, CX2_RESET_REG_SW_RESET);

	udelay( 10);

	/* set "initialization complete" bit to move adapter to D0 state */
	ipw2200_set_bit(priv, CX2_GP_CNTRL_RW, CX2_GP_CNTRL_BIT_INIT_DONE);

	IPW_DEBUG_TRACE(">>\n");
	return 0;
}


/* Call this function from process context, it will sleep in request_firmware. 
 * Probe is an ok place to call this from.
 */
static int ipw2200_reset_nic(struct ipw2200_private *priv)
{
	int rc = 0;

	IPW_DEBUG_TRACE(">>\n");
	
	rc = ipw2200_init_nic(priv);

	IPW_DEBUG_TRACE("<<\n");
	return rc;
}

static int ipw2200_init_clocks(struct ipw2200_private *priv)
{
	int rc;

	IPW_DEBUG_TRACE(">>\n");
	/* disable interrupts */
	/* next do the loading of the images onto the hardware... */
	/*stop and ack ints drop them all*/
	writel(~CX2_INTA_MASK_ALL, priv->hw_base + CX2_INTA_MASK_R);
	priv->interrupts_enabled = 0;
	writel(CX2_INTA_MASK_ALL, priv->hw_base + CX2_INTA_RW);


	/* stop NIC */
	rc = ipw2200_stop_nic(priv);
	if( rc < 0)
		IPW_DEBUG_INFO("hw stop nic failed\n");

	/* reset NIC */
	rc = ipw2200_reset_nic(priv);

	IPW_DEBUG_TRACE("<<\n");
	return 0;
}
int ipw2200_load(struct ipw2200_private *priv)
{
	int rc = -1;
	const struct firmware *bootfw;
	const struct firmware *firmware;
	const struct firmware *ucode;
	u32 temp;

	/* first lest get the firmware from the file system. */

	rc = request_firmware(&bootfw, "ipw2200_boot.fw", &priv->pci_dev->dev);
	if (rc < 0) {
		IPW_DEBUG_ERROR("ipw2200_boot.fw load failed\n");
		return -EINVAL;
	} else {
		IPW_DEBUG_INFO(" boot fw data %p size %d \n", bootfw->data,
			       bootfw->size);
	}

	rc = request_firmware(&firmware, "ipw2200_bss.fw", &priv->pci_dev->dev);
	if (rc < 0) {
		IPW_DEBUG_ERROR("ipw2200_bss.fw load failed\n");
		release_firmware(bootfw);
		return -EINVAL;
	} else {
		IPW_DEBUG_INFO("fw data %p size %d \n", firmware->data,
			       firmware->size);
	}

	rc = request_firmware(&ucode, "ipw2200_ucode.fw", &priv->pci_dev->dev);
	if (rc < 0) {
		IPW_DEBUG_ERROR("ipw2200_ucode.fw load failed\n");
		release_firmware(bootfw);
		release_firmware(firmware);
		return -EINVAL;
	} else {
		IPW_DEBUG_INFO(" uCode data %p size %d \n", ucode->data,
			       ucode->size);
	}
	ipw2200_init_clocks(priv);

	rc = ipw2200_load_firmware(priv, bootfw->data, bootfw->size);
	if (rc >= 0) {

		ipw2200_start_nic(priv);
		/* waite for interrupt for init done */		
		rc = ipw2200_poll_bit(priv, CX2_INTA_RW, 
				      CX2_INTA_BIT_FW_INITIALIZATION_DONE, 5000); 
		if( 0 > rc)   
		{
			IPW_DEBUG_WARNING("didn't get CX2_INTA_BIT_FW_INITIALIZATION_DONE after 5000ms\n");
			return rc;
		} else
			IPW_DEBUG_WARNING("CX2_INTA_BIT_FW_INITIALIZATION_DONE!!! after %dms\n", rc);
		/* ack fw init done interrupt */	
		writel(CX2_INTA_BIT_FW_INITIALIZATION_DONE, priv->hw_base + CX2_INTA_RW);

		rc = ipw2200_load_ucode(priv, ucode->data, ucode->size);
		if (rc >= 0) {
			/* stop nic */
			ipw2200_stop_nic(priv);
			
			/* required by HW */
			temp = ipw2200_read_reg32(priv, CX2_INTERNAL_CMD_EVENT);
			temp = (temp | CX2_ARC_KESHET_CONFIG) & ~CX2_START_STANDBY;
			ipw2200_write_reg32(priv, CX2_INTERNAL_CMD_EVENT, temp);

			rc = ipw2200_load_firmware(priv, firmware->data,
						   firmware->size);
			if (rc < 0 )
			{
				IPW_DEBUG_ERROR("Opperational FW load failed.  you are screwed.\n");
				return -1;
			}

			/* enable interrupts */
			rc = ipw2200_fw_start(priv);
		}
	}

	if (rc < 0) {
		/* reset hardware or turn it off cuz it aint gonna work like this */
		IPW_DEBUG_ERROR("failure initializing the firmware \n");
		;
	}

	// BUGBUG just to keep the printk's from locking up the 
	// system while I get stuff working. The following 3 lines need to
	// be removed once the isr's are stablized.
//	writel(~CX2_INTA_MASK_ALL, priv->hw_base + CX2_INTA_MASK_R);
//	priv->interrupts_enabled = 0;
//	writel(CX2_INTA_MASK_ALL, priv->hw_base + CX2_INTA_RW);

	release_firmware(bootfw);
	release_firmware(firmware);
	release_firmware(ucode);

	return rc;
}
