/*
 * iSCSI driver for Linux
 * Copyright (C) 2002 Cisco Systems, Inc.
 * maintained by linux-iscsi-devel@lists.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi-attr.c,v 1.1.2.10 2004/01/09 09:29:02 surekhap Exp $
 *
 */

#include <linux/device.h>

#include "iscsi-protocol.h"
#include "iscsi-session.h"
#include "iscsi-attr.h"

volatile unsigned int iscsi_log_settings = LOG_SET(ISCSI_LOG_ERR) | LOG_SET(ISCSI_LOG_RETRY) | LOG_SET(ISCSI_LOG_TIMEOUT);

extern struct iscsi_hba *iscsi_hba;
extern int this_is_iscsi_boot;

#define is_digit(c)     (((c) >= '0') && ((c) <= '9'))
#define is_space(c)     ((c) == ' ' || (c) == '\t' || (c) == '\n' || (c) == '\0')

int
find_number(char *start, char *end, int *number)
{
	char *ptr = start;
	int found = 0;
	int acc = 0;

	/* skip leading whitespace */
	while ((ptr < end) && is_space(*ptr))
		ptr++;

	while (ptr < end) {
		if (is_space(*ptr)) {
			break;
		} else if (is_digit(*ptr)) {
			found = 1;
			acc = (acc * 10) + (*ptr - '0');
		} else {
			/* something bogus */
			return 0;
		}
		ptr++;
	}
	if (found) {
		if (number)
			*number = acc;
		return (ptr - start);
	} else
		return 0;
}

int
find_ip(char *start, char *end, char *addr)
{
	char *ptr = start;
	char *ptr1;
	int ip_length = 0;

	/* skip leading whitespace */
	while ((ptr < end) && is_space(*ptr))
		ptr++;

	ptr1 = ptr;
	while ((ptr1 < end) && !is_space(*ptr1)) {
		ptr1++;
		ip_length++;
	}
	if (ip_length) {
		memcpy(addr, ptr, ip_length);
		return (ptr1 - start);
	} else
		return 0;
}


struct iscsi_session *
get_session_by_target(const char *buf, char *bp, char *end)
{
	struct iscsi_session *session;
	unsigned int bus = 0, target = 0;
	int cmd_len;

	if ((cmd_len = find_number(bp, end, &bus)) == 0) {
		printk ("iSCSI: couldn't determine bus number of session\n");
		return NULL;
	}
	bp += cmd_len;

	if ((cmd_len = find_number(bp, end, &target)) == 0) {
		printk ("iSCSI: couldn't determine target id number of"
			" session\n");
		return NULL;
	}
	bp += cmd_len;

	session = find_session_by_bus(bus, target);
	if (session == NULL) {
		printk ("iSCSI: failed to find session for bus %u target %u\n",
		        bus, target);
	}
	return session;
}

static ssize_t
iscsi_log_options(struct class_device *class_dev, const char *buf, size_t count)
{
	struct Scsi_Host *shost = class_to_shost(class_dev);
	struct iscsi_hba *hba;
	char val[15], val1[10];
	unsigned int log_setting = 0;

	hba = (struct iscsi_hba *)shost->hostdata;
	if ( !hba ) {
		return -EINVAL;
	}

	sscanf(buf, "%s %s", val, val1); 

	if (strncmp(val, "all", 3) == 0) {
		iscsi_log_settings = 0xFFFFFFFF;
		printk("iSCSI: all logging enabled\n");
	}
	else if(strncmp(val, "none", 4) == 0) {
		iscsi_log_settings = 0;
		printk("iSCSI: all logging disabled\n");
	}
	else if(strncmp(val, "sense", 5) == 0) {
		log_setting = ISCSI_LOG_SENSE;
		if(!strcmp(val1, "always")) {
			iscsi_log_settings |= LOG_SET(log_setting);
			printk("iSCSI: log sense always\n");
		}
		else if(!strcmp(val1, "on")) {
			iscsi_log_settings |= LOG_SET(log_setting);
			printk("iSCSI: log sense yes\n");
		}
		else if(!strcmp(val1, "yes")) {
			iscsi_log_settings |= LOG_SET(log_setting);
			printk("iSCSI: log sense yes\n");
		}
		else if(!strcmp(val1, "l")) {
			iscsi_log_settings |= LOG_SET(log_setting);
			printk("iSCSI: log sense 1\n");
		}
		else if(!strcmp(val1, "minimal")) {
			iscsi_log_settings &= ~LOG_SET(log_setting);
			printk("iSCSI: log sense off\n");
		}
		else if(!strcmp(val1, "off")) {
			iscsi_log_settings &= ~LOG_SET(log_setting);
			printk("iSCSI: log sense off\n");
		}
		else if(!strcmp(val1, "no")) {
			iscsi_log_settings &= ~LOG_SET(log_setting);
			printk("iSCSI: log sense no\n");
		}
		else if(!strcmp(val1, "0")) {
			iscsi_log_settings &= ~LOG_SET(log_setting);
			printk("iSCSI: log sense 0\n");
 		}
	} else if (strncmp(val, "login", 5) == 0) {
		log_setting = ISCSI_LOG_LOGIN;
	}
	else if(strncmp(val, "init", 4) == 0) {
		log_setting = ISCSI_LOG_INIT;
	}
	else if(strncmp(val, "queue", 5) == 0) {
		log_setting = ISCSI_LOG_QUEUE;
	}
	else if(strncmp(val, "alloc", 5) == 0) {
		log_setting = ISCSI_LOG_ALLOC;
	}
	else if(strncmp(val, "flow", 4) == 0) {
		log_setting = ISCSI_LOG_FLOW;
	}
	else if(strncmp(val, "error", 5) == 0) {
		log_setting = ISCSI_LOG_ERR;
	}
	else if(strncmp(val, "eh", 2) == 0) {
		log_setting = ISCSI_LOG_EH;
	}
	else if(strncmp(val, "retry", 5) == 0) {
		log_setting = ISCSI_LOG_RETRY;	
	}
	else {
		printk(" invalid log setting option\n");
		return -EINVAL;
	}
			
	if (log_setting) {
		if(!strcmp(val1, "on")) {
			iscsi_log_settings |= LOG_SET(log_setting);
		}
		else if(!strcmp(val1, "yes")) {
			iscsi_log_settings |= LOG_SET(log_setting);
		}
		else if(!strcmp(val1, "1")) {
			iscsi_log_settings |= LOG_SET(log_setting);
		}
		else if(!strcmp(val1, "off")) {
			iscsi_log_settings &= ~LOG_SET(log_setting);
		}
		else if(!strcmp(val1, "no")) {
			iscsi_log_settings &= ~LOG_SET(log_setting);
		}
		else if(!strcmp(val1, "0")) {
			iscsi_log_settings &= ~LOG_SET(log_setting);
		}
		printk("iSCSI: log settings %8x\n", iscsi_log_settings);
	}
	return count;
}

static ssize_t
do_iscsi_shutdown(struct class_device *class_dev, const char *buf, size_t count)
{
	/* try to shutdown the driver */
	if (!iscsi_shutdown()) {
		printk("iSCSI: driver shutdown failed\n");
		return -EFAULT;
	}
	return count;
}

static ssize_t
iscsi_change_portal(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_portal_info *p = NULL;
	struct iscsi_session *session;
	unsigned int portal = 0;
	unsigned int bus, target;
	char *tmp, *bp, *end;
	int cmd_len;
	char ip[16];
	char address[17];
	uint32_t a1;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		bus = session->channel;
		target = session->target_id;
		if (session->portal_failover) {
			memset(address, 0, sizeof (address));
			if ((cmd_len = find_ip(bp, end, address)) == 0) {
				printk("iSCSI: session for bus %u target %u,"
				       " no ip address specified\n", bus, target);
			} else {
				p = session->portals;
				for (portal = 0; portal < session->num_portals;
					portal++) {
					a1 = iscsi_inet_aton(address);
					ip[0] = a1 & 0xFF;
					ip[1] = (a1 & 0xFF00) >> 8;
					ip[2] = (a1 & 0xFF0000) >> 16;
					ip[3] = (a1 & 0xFF000000) >> 24;

					if (memcmp (p[portal].ip_address, ip,
						 p[portal].ip_length) == 0) {
						break;
					}
				}

				spin_lock(&session->portal_lock);
				if (memcmp(session->ip_address, ip, p[portal].
					 ip_length) == 0) {
					printk("iSCSI: Requested address %s is"
					       " already the current portal,"
					       " failover not needed \n", address);
				} else
				    if (portal < session->num_portals) {
					session->requested_portal = portal;
					session->fallback_portal = session->
							    current_portal;

					printk("iSCSI:session for bus "
					       "%u target %u requesting switch "
					       "to ip %s at %lu\n", bus, target,
					       address, jiffies);
						
					/* request a logout for the current
					 * session
					 */

					spin_lock (&session-> task_lock);
					iscsi_request_logout (session, session->
						     active_timeout,
						     session-> active_timeout);
					spin_unlock (&session-> task_lock);
				} else {
					address[16] = '\0';
					printk("iSCSI: iscsi session for bus"
					       "%u target %u requesting switch "
					       "to ip %s \n", bus, target, address);
				}
				spin_unlock(&session->portal_lock);
			}
		} else
			printk("iSCSI: session for bus %u target %u can't"
			      " switch to requested ip, because portal failover"			      " is disabled\n", bus, target);
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}


static int
add_lun_line_info(struct iscsi_session *session, int lun, char **buffer)
{
	DEBUG_FLOW("iSCSI: add_lun_line iscsi bus %d target id %d, %d \n",
		session->iscsi_bus, session->target_id, lun);

	/* 2 spaces + (4 * (3-digit field + space)) + 4 spaces = 22 chars */
	if (lun >= 0)
		*buffer += sprintf(*buffer, " %3.1u %3.1u %3.1u    ",
			session->iscsi_bus, session->target_id, lun);
	else
		*buffer += sprintf(*buffer, " %3.1u %3.1u   ?    ",
			session->iscsi_bus, session->target_id);

	if (test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
		/* up to 15 char IP address + 2 spaces = 17 chars */
		*buffer += sprintf(*buffer, "   %u.%u.%u.%u  ",
			session->ip_address[0], session->ip_address[1],
			session->ip_address[2], session->ip_address[3]);
		/* right_justify(*buffer, 17); */

		/* 5-digit port + 2 spaces = 7 chars */
		*buffer += sprintf(*buffer, " %d  ", session->port);

	} else {
		/* fill in '?' for each empty field, so that
		 * it is easier for processing by tools such as awk and perl.
		 */
		*buffer += sprintf(*buffer, "              ?      ?  ");

	}
	*buffer += sprintf(*buffer, "%s ", session->TargetName);
	 /* and a newline */
	*buffer += sprintf(*buffer, "\n");
	return 1;
}

static int
show_session_lun_info(struct iscsi_session *session, char **buffer)
{
	/* if we've already found LUNs, show them all */
	int lfound = 0;
	int l;

	/* FIXME: IPv6 */
	if (session->ip_length != 4)
		return 1;

	for (l = 0; l < ISCSI_MAX_LUN; l++) {
		if (test_bit(l, session->luns_activated)) {
			lfound += 1;
			if (!add_lun_line_info(session, l, buffer)) {
				DEBUG_FLOW("iSCSI: show session luns"
					" returning 0 ");
				return 0;
			}
		}
	}

	/* if we haven't found any LUNs, use ? for a LUN number */
	if (!lfound) {
		if (!add_lun_line_info(session, -1, buffer)) {
			DEBUG_FLOW("iSCSI: show session luns returning 0 \n");
			return 0;
		}
	} else {
		DEBUG_FLOW("iSCSI: show session luns returning 1 ");
	}
	return 1;
}

static ssize_t
iscsi_print_info(struct class_device *class_dev, char *buf)
{
	struct Scsi_Host *host = class_to_shost(class_dev);
	struct iscsi_session *session;
	unsigned long noqueue_flags_;
	char *build_str = BUILD_STR;
	char *bp;

	bp = buf;

	/* comment header with version number */
	/* FIXME: we assume the buffer length is always large 
	 * enough for our header 
	 */
	if (build_str) {
		/* developer-built variant of a 4-digit internal 
		 * release 
		 */
		bp +=
		    sprintf(bp,
			    "# iSCSI driver version: %d.%d.%d.%d%s "
			    "variant (%s)\n#\n",
			    DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION,
			    DRIVER_PATCH_VERSION,
			    DRIVER_INTERNAL_VERSION,
			    DRIVER_EXTRAVERSION, ISCSI_DATE);
	} else if (DRIVER_INTERNAL_VERSION > 0) {
		/* 4-digit release */
		bp +=
		    sprintf(bp,
			    "# iSCSI driver version: %d.%d.%d.%d%s "
			    "(%s)\n#\n",
			    DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION,
			    DRIVER_PATCH_VERSION,
			    DRIVER_INTERNAL_VERSION,
			    DRIVER_EXTRAVERSION, ISCSI_DATE);
	} else {
		/* 3-digit release */
		bp +=
		    sprintf(bp,
			    "# iSCSI driver version: %d.%d.%d%s "
			    "(%s)\n#\n",
			    DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION,
			    DRIVER_PATCH_VERSION, DRIVER_EXTRAVERSION,
			    ISCSI_DATE);
	}
	bp += sprintf(bp, "# SCSI:               iSCSI:\n");
	bp += sprintf(bp, "# Bus Tgt LUN         IP address   Port  "
		    "TargetName\n");

	if (iscsi_hba) {
		spin_lock_irqsave(&iscsi_hba->session_lock, noqueue_flags_);
		list_for_each_entry(session, &iscsi_hba->session_head,
				    session_link) {
			if (!show_session_lun_info(session, &bp))
				break;
		}
		spin_unlock_irqrestore(&iscsi_hba->session_lock,
					noqueue_flags_);
	} else {
		printk("iSCSI: couldn't find HBA %p\n", host);
	}
	return (bp - buf);
}

static ssize_t 
iscsi_host_id(struct class_device *class_dev, char *buf)
{
	struct Scsi_Host *host = class_to_shost(class_dev);
	char *bp;

	bp = buf;

	bp += sprintf(bp, "%d\n", host->host_no);

	return (bp - buf);
}

static ssize_t 
iscsi_network_boot(struct class_device *class_dev, char *buf)
{
	char *bp;

	bp = buf;

	if (this_is_iscsi_boot)
		bp += sprintf(bp, "%s\n", "yes");
	else
		bp += sprintf(bp, "%s\n", "no");

	return (bp - buf);
}

struct class_device_attribute change_portal = {
	.attr = {
		.name =		"failover",
		.mode =		S_IWUSR,
	},
	.store = iscsi_change_portal,
};

struct class_device_attribute log_status = {
	.attr = {
		.name =		"log",
		.mode =		S_IWUSR,
	},
	.store = iscsi_log_options,
};

struct class_device_attribute shutdown = {
	.attr = {
		.name =		"shutdown",
		.mode =		S_IWUSR,
	},
	.store = do_iscsi_shutdown,
};

struct class_device_attribute print_info = {
	.attr = {
		.name =		"print_info",
		.mode =		S_IRUGO,
	},
	.show = iscsi_print_info,
};

struct class_device_attribute host_id = {
	.attr = {
		.name = "host_no",
		.mode = S_IRUGO,
	},
	.show = iscsi_host_id,
};

struct class_device_attribute network_boot = {
	.attr = {
		.name = "network_boot",
		.mode = S_IRUGO,
	},
	.show = iscsi_network_boot,
};
