/*
 * iSCSI connection daemon
 * Copyright (C) 2002 Cisco Systems, Inc.
 * maintained by linux-iscsi-devel@lists.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi-bindings.c,v 1.6.8.4 2003/12/16 05:59:52 smhatre Exp $
 *
 */

#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>

#include "iscsi-hooks.h"
#include "iscsi-bindings.h"

/* use a bitmap allocator for bus and target ids */
#define ULONG_BITS      (sizeof(unsigned long) * CHAR_BIT)
#define ID_BITMAP_SIZE \
((ISCSI_MAX_TARGET_IDS_PER_BUS + ULONG_BITS - 1) / ULONG_BITS)
#define BUS_BITMAP_SIZE ((ISCSI_MAX_BUS_IDS + ULONG_BITS - 1) / ULONG_BITS)
struct iscsi_id_info {
	unsigned long bitmap[ISCSI_MAX_BUS_IDS][ID_BITMAP_SIZE];
	unsigned long bus_bitmap[BUS_BITMAP_SIZE];
	int binding_index;
};

int
lookup_bindings(int fd, struct iscsi_target_list *targets,
		struct iscsi_id_info *id_info)
{
	struct iscsi_target *target;
	struct iscsi_session_config *session;
	int scan_fd;
	FILE *f = NULL;
	char buffer[4096];
	char *line;
	int unbound = 0;

	/* count the number of unbound sessions */
	for (target = targets->head; target; target = target->next) {
		if (target->new_config) {
			for (session = target->new_config->sessions; session;
			     session = session->next)
				if (session->target_id < 0)
					unbound++;
		}
	}

	if (unbound == 0) {
		/* don't need to do anything */
		debugmsg(1, "lookup_bindings called with 0 unbound sessions");
		return unbound;
	}

	scan_fd = dup(fd);
	if (scan_fd >= 0)
		f = fdopen(scan_fd, "r");

	if (f) {
		unsigned long line_nr = 0;

		id_info->binding_index = 0;

		debugmsg(1, "scanning bindings file for %d unbound sessions",
			 unbound);

		/* for every binding line in the file, see if it applies to one 
		 * of the sessions 
		 */
		while ((line = fgets(buffer, sizeof (buffer), f))) {
			unsigned long *word;
			unsigned int bit;
			unsigned long bus_id;
			unsigned long target_id;
			char *c = line;
			char *TargetName;

			/* strip comments */
			if ((c = strchr(line, '#'))) {
				*c = '\0';
			}

			line_nr++;

			/* each binding line should have at least: 
			 * bus_id target_id TargetName 
			 */
			c = strtok(line, "# \t\r\n");
			if (c == NULL)
				continue;	/* don't log, it's just 
						 * a blank line 
						 */
			if (sscanf(c, "%lu", &bus_id) != 1) {
				logmsg(AS_ERROR,
				       "ignoring malformed line %lu of "
				       "bindings file, no bus id",
				       line);
				continue;
			}

			c = strtok(NULL, "# \t\r\n");
			if ((c == NULL) || 
				(sscanf(c, "%lu", &target_id) != 1)) {
				logmsg(AS_ERROR,
				       "ignoring malformed line %lu of "
				       "bindings file, no target id",
				       line);
				continue;
			}

			TargetName = strtok(NULL, "# \t\r\n");
			if (TargetName == NULL) {
				logmsg(AS_ERROR,
				       "ignoring malformed line %lu of "
				       "bindings file, no iSCSI TargetName",
				       line);
				continue;
			}

			/* mark the bus and target id as allocated 
			 * in the bitmap 
			 */
			word =
			    &id_info->bitmap[bus_id][0] +
			    (target_id / ULONG_BITS);
			bit = target_id % ULONG_BITS;

			if (*word & (0x1 << bit)) {
				/* same bus & target id appears more than 
				 * once in the bindings file.
				 * somebody probably broke it when they 
				 * edited the file by hand.
				 */
				logmsg(AS_ERROR,
				       "bus %lu target %lu appears more "
				       "than once in bindings file",
				       bus_id, target_id);
			} else {
				*word |= (0x1 << bit);
			}

			/* see if we have a session that gets this binding */
			for (target = targets->head; target;
			     target = target->next) {
				if (target->new_config) {
					for (session =
					     target->new_config->sessions;
					     session; session = session->next) {
						if (session->target_id < 0) {
							if (strcmp
							    (TargetName,
							     session->target->
							     TargetName) == 0) {
								debugmsg(1,
									 "iSCSI"
" bus %d target %d bound to session #%d to %s",
									 bus_id,
									 target_id,
									 session->
									 path_number,
									 session->
									 target->
									 TargetName);
								session->
								    iscsi_bus =
								    bus_id;
								session->
								    target_id =
								    target_id;
								session->
								    binding_index
								    =
								    id_info->
								    binding_index;
								unbound--;
							/* only bind one session
							 * for each line in the 
							 * bindings file 
							 */
								goto next;
							}
						}
					}
				}
			}

		      next:
			id_info->binding_index++;

			/* if there are no unbound sessions, then we don't need 
			 * to keep scanning 
			 */
			if (unbound == 0)
				goto done;
		}

	      done:
		debugmsg(1, "done scanning bindings file at line %lu", line_nr);
		fclose(f);
		close(scan_fd);
	} else {
		logmsg(AS_ERROR,
		       "lookup_bindings failed to fdopen bindings file\n");
	}

	return unbound;
}

int
save_binding(int fd, unsigned long iscsi_bus, unsigned long target_id,
	     char *TargetName)
{
	char line[2048];
	char *l, *end;
	int length, result;
	off_t offset;

	/* build the binding line */
	length =
	    sprintf(line, "%lu\t%lu\t%s\n", iscsi_bus, target_id, TargetName);
	if (length > (sizeof (line) - 1)) {
		/* buffer too small */
		logmsg(AS_ERROR,
		       "couldn't save binding to the bindings file, "
		       "buffer too small");
		return 0;
	}
	line[sizeof (line) - 1] = '\0';

      reseek:
	/* seek to the end */
	if ((offset = lseek(fd, 0, SEEK_END)) < 0) {
		if (errno == EINTR)
			goto reseek;

		errormsg("failed to seek to the end of the bindings file");
		return 0;
	}

	/* write it to the file */
	l = line;
	end = line + length;
	while (l < end) {
		result = write(fd, l, end - l);
		if (result < 0) {
			if (errno != EINTR) {
				errormsg
				    ("failed to write to the bindings file");
				/* truncate the file to clear any partial 
				 * writes that have occured 
				 */
				ftruncate(fd, offset);
				return 0;
			}
		} else {
			l += result;
		}
	}

	/* make sure the data was written to disk */
      resync:
	result = fsync(fd);

	if (result < 0) {
		if (errno == EINTR)
			goto resync;

		errormsg("failed to sync the bindings file");
		return 0;
	}

	return 1;
}

/* try to allocate the next available, starting with bus_id,target_id */
int
allocate_next_available(int fd, struct iscsi_session_config *session,
			struct iscsi_id_info *id_info, int bus_id,
			int target_id)
{
	unsigned long starting_bus_id = bus_id, starting_target_id = target_id;
	unsigned long *word;
	unsigned int bit;

	for (;;) {
		/* FIXME: optimize this to skip words with all bits set */
		word = &id_info->bitmap[bus_id][0] + (target_id / ULONG_BITS);
		bit = target_id % ULONG_BITS;

		/* check if the selected id is free */
		if ((*word & (0x1UL << bit)) == 0) {
			if (save_binding
			    (fd, bus_id, target_id,
			     session->target->TargetName)) {
				/* got it */
				*word |= (0x1UL << bit);
				session->iscsi_bus = bus_id;
				session->target_id = target_id;
				session->binding_index = id_info->binding_index;
				id_info->binding_index++;

				debugmsg(1,
					 "iSCSI bus %d target %d allocated "
					 "for session #%d to %s",
					 bus_id, target_id,
					 session->path_number,
					 session->target->TargetName);
				return 1;
			} else {
				logmsg(AS_ERROR,
				       "failed to save binding for bus %lu, "
				       "target %lu, %s",
				       bus_id, target_id,
				       session->target->TargetName);
				return 0;
			}
		}

		/* circular scan through the bus&id space */
		if (++target_id >= ISCSI_MAX_TARGET_IDS_PER_BUS) {
			bus_id++;
			target_id = 0;
		}

		if (bus_id >= ISCSI_MAX_BUS_IDS) {
			bus_id = 0;
		}

		/* fail if we've looped back to the starting point */
		if ((bus_id == starting_bus_id)
		    && (target_id == starting_target_id)) {
			logmsg(AS_NOTICE,
			       "no unbound target ids, couldn't bind "
			       "session %p to %s",
			       session, session->target->TargetName);
			return 0;
		}
	}
}

/* try to allocate the specified target_id on buses other than bus_id */
int
allocate_on_next_bus(int fd, struct iscsi_session_config *session,
		     struct iscsi_id_info *id_info, int bus_id, int target_id)
{
	unsigned long starting_bus_id = bus_id;
	unsigned long *word;
	unsigned int bit;

	/* skip the specified bus id, but check all the rest */
	bus_id++;
	if (bus_id >= ISCSI_MAX_BUS_IDS)
		bus_id = 0;

	for (;;) {
		if (bus_id == starting_bus_id) {
			/* checked all the buses */
			return 0;
		}

		debugmsg(7, "checking bus %d target %d for session #%d to %s",
			 bus_id, target_id, session->path_number,
			 session->target->TargetName);

		word = &id_info->bitmap[bus_id][0] + (target_id / ULONG_BITS);
		bit = target_id % ULONG_BITS;

		/* check if the selected bus/target id is free */
		if ((*word & (0x1UL << bit)) == 0) {
			if (save_binding
			    (fd, bus_id, target_id,
			     session->target->TargetName)) {
				/* got it */
				*word |= (0x1UL << bit);
				session->iscsi_bus = bus_id;
				session->target_id = target_id;
				session->binding_index = id_info->binding_index;
				id_info->binding_index++;

				debugmsg(1,
					 "iSCSI bus %d target %d allocated "
					 "for session #%d to %s",
					 bus_id, target_id,
					 session->path_number,
					 session->target->TargetName);
				return 1;
			} else {
				logmsg(AS_ERROR,
				       "failed to save binding for bus %lu "
				       "target %lu, session #%d to %s",
				       bus_id, target_id, session->path_number,
				       session->target->TargetName);
				return 0;
			}
		} else {
			/* try the next bus */
			bus_id++;
			if (bus_id >= ISCSI_MAX_BUS_IDS) {
				bus_id = 0;
			}
		}
	}
}

int
allocate_on_different_bus(int fd, struct iscsi_session_config *session,
			  struct iscsi_id_info *id_info, int bus_id)
{
	unsigned long starting_bus_id = bus_id;
	unsigned long *word;
	unsigned int bit;
	int target_id;

	for (;;) {
		/* search for a different bus */
		for (;;) {
			if (++bus_id >= ISCSI_MAX_BUS_IDS)
				bus_id = 0;

			if (bus_id == starting_bus_id)
				return 0;	/* all buses already in use 
						 * by other sessions 
						 */

			word = id_info->bus_bitmap + (bus_id / ULONG_BITS);
			bit = bus_id % ULONG_BITS;

			if ((*word & (0x1UL << bit)) == 0) {
				/* we can use this bus */
				break;
			}
		}

		debugmsg(7, "allocate_on_different_buses checking bus %d",
			 bus_id);

		/* search for a free target id on this bus */
		for (target_id = 0; target_id < ISCSI_MAX_TARGET_IDS_PER_BUS;
		     target_id++) {
			/* FIXME: optimize this to skip words with all 
			 * bits set 
			 */
			word =
			    &id_info->bitmap[bus_id][0] +
			    (target_id / ULONG_BITS);
			bit = target_id % ULONG_BITS;

			/* check if the selected id is free */
			if ((*word & (0x1UL << bit)) == 0) {
				if (save_binding
				    (fd, bus_id, target_id,
				     session->target->TargetName)) {
					/* got it */
					*word |= (0x1UL << bit);
					session->iscsi_bus = bus_id;
					session->target_id = target_id;
					session->binding_index =
					    id_info->binding_index;
					id_info->binding_index++;

					/* update the bus_bitmap as well */
					word =
					    id_info->bus_bitmap +
					    (bus_id / ULONG_BITS);
					bit = bus_id % ULONG_BITS;
					*word |= (0x1UL << bit);

					debugmsg(1,
						 "iSCSI bus %d target %d "
						 "allocated for session #%d "
						 "to %s",
						 bus_id, target_id,
						 session->path_number,
						 session->target->TargetName);
					return 1;
				} else {
					logmsg(AS_ERROR,
					       "failed to save binding for "
					       "bus %lu, target %lu, %s",
					       bus_id, target_id,
					       session->target->TargetName);
					return 0;
				}
			}
		}
	}
}

/* try to allocate a binding for every session to this target that
 * isn't already bound, preferably with each session having a
 * different bus id, and ideally with each session having the same
 * target id on different buses.  For the first session to each target,
 * we start searching after the highest target id already in use on bus 0,
 * or the lowest id on buses above 0 if bus 0 is full.  
 *
 * There should only be gaps if people delete bindings by hand, so
 * looking for them will usually by a waste of time.  Even if there
 * are gaps, it's safer to re-use ids slowly by avoiding the gaps
 * until we're out of ids, and have no choice but to use an id that's
 * been used before, but is now free.  That way infrequently-used
 * software with an old config has a better chance of erroring out,
 * rather than using the wrong storage.
 */
int
allocate_target_bindings(int fd, struct iscsi_target *target,
			 struct iscsi_id_info *id_info)
{
	struct iscsi_session_config *session = NULL;
	struct iscsi_session_config *first_session;
	unsigned long bus_id = 0;
	unsigned long target_id = 0;
	unsigned long *word;
	unsigned int bit;

	if (target->new_config == NULL)
		return 0;

	first_session = target->new_config->sessions;

	if (first_session == NULL)
		return 0;

	/* log all the sessions that need bindings */
	for (session = target->new_config->sessions; session;
	     session = session->next) {
		if (session->target_id < 0)
			debugmsg(7,
				 "need to allocate a binding for session #%d "
				 "to %s",
				 session->path_number,
				 session->target->TargetName);
	}

	/* the first session to this target takes the first available bus & id.
	 * if this fails, we're completely out of ids.
	 */
	if (first_session->target_id < 0) {
		if (!allocate_next_available
		    (fd, first_session, id_info, bus_id, target_id))
			return 0;
	}

	if (first_session->next == NULL)
		return 1;

	debugmsg(7, "allocating same ids on different buses for target %s",
		 target->TargetName);

	/* any other sessions to the same target try to take the same 
	 * target id on the next higher bus 
	 */
	for (session = target->new_config->sessions; session;
	     session = session->next) {
		if (session->iscsi_bus >= 0) {
			bus_id = session->iscsi_bus;
			target_id = session->target_id;
		} else
		    if (allocate_on_next_bus
			(fd, session, id_info, bus_id, target_id)) {
			bus_id = session->iscsi_bus;
			target_id = session->target_id;
		} else {
			break;
		}
	}

	if (session == NULL)
		return 1;

	/* record what buses sessions to this target are already using, 
         * so that we can try to avoid them 
         */
	memset(id_info->bus_bitmap, 0, sizeof (id_info->bus_bitmap));
	for (session = target->new_config->sessions; session;
	     session = session->next) {
		if (session->iscsi_bus >= 0) {
			word =
			    id_info->bus_bitmap +
			    (session->iscsi_bus / ULONG_BITS);
			bit = session->iscsi_bus % ULONG_BITS;

			*word |= (0x1UL << bit);
		}
	}

	debugmsg(7, "allocating any ids on different buses for target %s",
		 target->TargetName);

	/* try to bind every unbound session for this target 
	 * on a different bus 
	 */
	for (session = target->new_config->sessions; session;
	     session = session->next) {
		if (session->iscsi_bus >= 0) {
			bus_id = session->iscsi_bus;
		} else
		    if (allocate_on_different_bus(fd, session, id_info, bus_id))
		{
			bus_id = session->iscsi_bus;
		} else {
			/* failed, try something else */
			break;
		}
	}

	if (session == NULL)
		return 1;

	debugmsg(7, "allocating any id on any bus for target %s",
		 target->TargetName);

	/* try to take any free bus/target id, regardless of 
	 * sharing buses with other sessions 
	 */
	for (session = target->new_config->sessions; session;
	     session = session->next) {
		if (session->target_id >= 0) {
			bus_id = session->iscsi_bus;
			target_id = session->target_id;
			debugmsg(7,
				 "final allocation found bus %d "
				 "target %d for %s",
				 bus_id, target_id, target->TargetName);
		} else
		    if (allocate_next_available
			(fd, session, id_info, bus_id, target_id)) {
			bus_id = session->iscsi_bus;
			target_id = session->target_id;
		} else {
			break;
		}
	}

	if (session == NULL)
		return 1;

	return 0;
}

void
allocate_bindings(int fd, struct iscsi_target_list *targets,
		  struct iscsi_id_info *id_info)
{
	struct iscsi_target *target;

	debugmsg(7, "need to allocate bindings");
	for (target = targets->head; target; target = target->next) {
		if (target->new_config && target->new_config->sessions)
			allocate_target_bindings(fd, target, id_info);
	}
}

struct iscsi_id_info *
alloc_id_info(void)
{
	struct iscsi_id_info *id_info = calloc(1, sizeof (*id_info));

	if (id_info) {
		/* FIXME: set the unused bits of the bitmap to 1, 
		 * to make it easier to scan it quickly 
		 */
	}

	return id_info;
}

static void
ensure_directories_exist(char *pathname, mode_t dir_mode)
{
	char path[100];
	char *end;

	strcpy(path, pathname);
	end = path;

	/* skip leading slashes */
	while (end && *end && (*end == '/'))
		end++;

	while ((end = strchr(end, '/'))) {
		/* if there is another slash, make the dir.
		 * FIXME: we ought to ignore errors when the directory exists,
		 * but report errors where the directory doesn't exist and
		 * we failed to create it.
		 */
		*end = '\0';
		debugmsg(7, "mkdir %s", path); 
		mkdir(path, dir_mode);
		*end = '/';
		end++;
	}
}

char bindings_file_header[] = "# iSCSI bindings, file format version 1.0.\n"
    "# NOTE: this file is automatically maintained by the iSCSI daemon.\n"
    "# You should not need to edit this file under most circumstances.\n"
    "# If iSCSI targets in this file have been permanently deleted, you \n"
    "# may wish to delete the bindings for the deleted targets.\n#\n"
    "# Format:\n" "# bus\ttarget\tiSCSI\t\n" "# id\tid\tTargetName\n#\n";

int
open_bindings_file(char *bindings_file)
{
	int fd = -1;
	int result;
	struct stat s;

	do {
		ensure_directories_exist(bindings_file, 0700);
		fd = open(bindings_file, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
	} while ((fd < 0) && (errno == EINTR));

	if (fd >= 0) {
		debugmsg(1, "opened bindings file %s", bindings_file);
	} else {
		logmsg(AS_ERROR, "failed to open bindings file %s",
		       bindings_file);
		return -1;
	}

	while (iscsi_lock_file(fd) < 0) {
		if (iscsi_process_should_exit())
			exit(0);

		if (errno == EINTR) {
			debugmsg(1, "waiting to lock bindings file");
			sleep(1);
		} else {
			logmsg(AS_ERROR, "failed to lock bindings file %s",
			       bindings_file);
			close(fd);
			return -1;
		}
	}

	debugmsg(1, "locked bindings file %s", bindings_file);
	memset(&s, 0, sizeof (s));

	do {
		result = fstat(fd, &s);
	} while ((result < 0) && (errno == EINTR));

	if (result < 0) {
		errormsg("failed to fstat the bindings file");
		return -1;
	}

	/* if the file is empty, add a comment header */
	if (s.st_size == 0) {
		const char *h = bindings_file_header;
		const char *end = h + strlen(h);

		while (h < end) {
			result = write(fd, h, end - h);
			if (result < 0) {
				if (errno != EINTR) {
					errormsg
					    ("failed to write to the "
					     "bindings file");
					/* truncate the file to clear 
					 * any partial writes that have 
					 * occured 
					 */
					ftruncate(fd, 0);
					close(fd);
					return -1;
				}
			} else {
				h += result;
			}
		}

		fsync(fd);
	}

	return fd;
}

void
iscsi_get_bindings(struct iscsi_target_list *targets, char *bindings_file)
{
	int fd;
	struct iscsi_target *target;
	struct iscsi_session_config *session;
	struct iscsi_id_info *id_info = NULL;

	/* initialize sessions to have no binding */
	for (target = targets->head; target; target = target->next) {
		if (target->new_config) {
			for (session = target->new_config->sessions; session;
			     session = session->next) {
				session->iscsi_bus = -1;
				session->target_id = -1;
			}
		}
	}

	if ((fd = open_bindings_file(bindings_file))) {

		/* allocate an id_info to use 
		 * while scanning the bindings file 
		 */
		id_info = alloc_id_info();

		if (id_info) {
			/* look for existing bindings */
			if (lookup_bindings(fd, targets, id_info)) {
				/* possibly allocate new ones if we 
				 * didn't find them all 
				 */
				allocate_bindings(fd, targets, id_info);
			}

			free(id_info);
		} else {
			logmsg(AS_ERROR,
			       "failed to allocate id_info for "
			       "processing bindings file %s",
			       bindings_file);
		}
		iscsi_unlock_file(fd);
		debugmsg(1, "unlocked bindings file %s", bindings_file);
		close(fd);
	} else {
		logmsg(AS_ERROR, "failed to open bindings file %s",
		       bindings_file);
	}
}
