/*
 * iSCSI driver for Linux
 * Copyright (C) 2002 Cisco Systems, Inc.
 * maintained by linux-iscsi-devel@lists.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi-debug-attr.c,v 1.1.2.7 2004/01/26 22:35:20 haapi Exp $
 *
 */

#include "iscsi-common.h"
#include "iscsi-protocol.h"
#include "iscsi-session.h"
#include "iscsi.h"

extern struct iscsi_hba *iscsi_hba;
extern struct class_device_attribute *iscsi_host_attrs[];
extern struct iscsi_session *
get_session_by_target(const char *buf, char *bp, char *end);

static struct iscsi_session *
get_session_by_lun(const char *buf, char *bp, char *end, int *lun_id)
{
	struct iscsi_session *session;
	unsigned int bus, target, lun;
	int cmd_len = 0;


	if ((cmd_len = find_number(bp, end, &bus)) == 0) {
		printk ("iSCSI: couldn't determine bus number of session\n");
		return NULL;
	}
	bp += cmd_len;

	if ((cmd_len = find_number(bp, end, &target)) == 0) {
		printk ("iSCSI: couldn't determine target id number of"
		" session\n");
		return NULL;
	}
	bp += cmd_len;

	if ((cmd_len = find_number(bp, end, &lun)) == 0) {
		printk("iSCSI: couldn't determine logical unit number\n");
		return NULL;
	}
	bp += cmd_len;
	*lun_id = lun;	
	session = find_session_by_bus(bus, target);
	if (session == NULL) {
		printk("iSCSI: failed to find session bus %u target %u LUN"
		       " %u\n", bus, target, lun);
	}
	return session;
}

static ssize_t
iscsi_lun_ignore(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int completions = 0;
	unsigned int aborts = 0;
	unsigned int abort_task_sets = 0;
	unsigned int lun_resets = 0;
	unsigned int warm_resets = 0;
	unsigned int cold_resets = 0;
	unsigned int bus, target, lun;
	char *end; 
	char *bp;
	char *tmp;
	int cmd_len;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
        bp = tmp;
        end = tmp + count;

	session = get_session_by_lun(buf, bp, end, &lun);
	if (session) {
		bus = session->channel;
		target = session->target_id;

		cmd_len = find_number(bp, end, &completions);
		bp += cmd_len;
		cmd_len = find_number(bp, end, &aborts);
		bp += cmd_len;
		cmd_len = find_number(bp, end, &abort_task_sets);
		bp += cmd_len;
		cmd_len = find_number(bp, end, &lun_resets);
		bp += cmd_len;
		cmd_len = find_number(bp, end, &warm_resets);
		bp += cmd_len;
		cmd_len = find_number(bp, end, &cold_resets);

		printk("iSCSI: session for bus %u target %u LUN %u at %lu,"
		       " ignore %u completions, %u aborts, %u abort task sets,"
		       " %u LUN resets, %u warm target resets, %u cold target"
		       " resets\n", bus, target, lun, jiffies, completions,
		       aborts, abort_task_sets, lun_resets, warm_resets,
		       cold_resets);
		spin_lock(&session->task_lock);
		session->ignore_lun = lun;
		session->ignore_completions = completions;
		session->ignore_aborts = aborts;
		session->ignore_abort_task_sets = abort_task_sets;
		session->ignore_lun_resets = lun_resets;
		session->ignore_warm_resets = warm_resets;
		session->ignore_cold_resets = cold_resets;
		spin_unlock(&session->task_lock);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_lun_reject(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int aborts = 0;
	unsigned int abort_task_sets = 0;
	unsigned int lun_resets = 0;
	unsigned int warm_resets = 0;
	unsigned int cold_resets = 0;
	unsigned int bus, target, lun;
	char *end; 
	char *bp;
	char *tmp;
	int cmd_len;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
        bp = tmp;
        end = tmp + count;

	session = get_session_by_lun(buf, bp, end, &lun);
	if (session) {
		bus = session->channel;
		target = session->target_id;

		cmd_len = find_number(bp, end, &aborts);
		bp += cmd_len;
		cmd_len = find_number(bp, end, &abort_task_sets);
		bp += cmd_len;
		cmd_len = find_number(bp, end, &lun_resets);
		bp += cmd_len;
		cmd_len = find_number(bp, end, &warm_resets);
		bp += cmd_len;
		cmd_len = find_number(bp, end, &cold_resets);
				
		printk("iSCSI:session for bus %u target %u LUN %u at %lu,"
		       " reject %u aborts, %u abort task sets, %u LUN resets,"
		       " %u warm target resets, %u cold target resets\n", bus,
		       target, lun, jiffies, aborts, abort_task_sets,
		       lun_resets, warm_resets, cold_resets);
		spin_lock(&session->task_lock);
		session->reject_lun = lun;
		session->reject_aborts = aborts;
		session->reject_abort_task_sets = abort_task_sets;
		session->reject_lun_resets = lun_resets;
		session->reject_warm_resets = warm_resets;
		session->reject_cold_resets = cold_resets;
		spin_unlock(&session->task_lock);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_lun_busy(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target, lun;
	char *end; 
	char *bp;
	char *tmp;
	int cmd_len;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
        bp = tmp;
        end = tmp + count;

	session = get_session_by_lun(buf, bp, end, &lun);
	if (session) {
		unsigned int count = 1;
		bus = session->channel;
		target = session->target_id;
		
		cmd_len = find_number(bp, end, &count);

		spin_lock(&session->task_lock);
		session->fake_status_lun = lun;
		session->fake_status_busy = count;
		printk("iSCSI: session iscsi bus %d target id %d will fake %u"
		       " SCSI status BUSY responses from LUN %u at %lu\n",
		session->iscsi_bus, session->target_id, count, lun, jiffies);
		spin_unlock(&session->task_lock);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_lun_queuefull(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target, lun;
	char *end; 
	char *bp;
	char *tmp;
	int cmd_len;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
        bp = tmp;
        end = tmp + count;

	session = get_session_by_lun(buf, bp, end, &lun);
	if (session) {
		unsigned int count = 1;
		bus = session->channel;
		target = session->target_id;

		cmd_len = find_number(bp, end, &count);
		spin_lock(&session->task_lock);
		session->fake_status_lun = lun;
		session->fake_status_queue_full = count;
		printk("iSCSI: session iscsi bus %d target id %d will fake %u"
		       " SCSI status QUEUE_FULL responses from LUN %u at %lu\n",
		       session->iscsi_bus, session->target_id, count, lun,
		       jiffies);
		spin_unlock(&session->task_lock);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_lun_unreachable(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target, lun;
	char *end; 
	char *bp;
	char *tmp;
	int cmd_len;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_lun(buf, bp, end, &lun);
	if (session) {
		unsigned int count = 1;
		bus = session->channel;
		target = session->target_id;

		cmd_len = find_number(bp, end, &count);

		spin_lock(&session->task_lock);
		session->fake_status_lun = lun;
		session->fake_status_unreachable = count;
		printk("iSCSI: session iscsi bus %d target id %d will fake %u"
		       " iSCSI transport errors from LUN %u at %lu\n",
		       session->iscsi_bus, session->target_id, count, lun,
		       jiffies);
		spin_unlock(&session->task_lock);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_lun_abort(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target, lun;
	char *end; 
	char *bp;
	char *tmp;
	int cmd_len;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
        bp = tmp;
        end = tmp + count;

	session = get_session_by_lun(buf, bp, end, &lun);

	if (session) {
		unsigned int count = 1;
		bus = session->channel;
		target = session->target_id;

		cmd_len = find_number(bp, end, &count);

		spin_lock(&session->task_lock);
		session->fake_status_lun = lun;
		session->fake_status_aborted = count;
		printk("iSCSI: session iscsi bus %d target id %d will fake %u"
		       " target command aborts from LUN %u at %lu\n",
		       session->iscsi_bus, session->target_id, count, lun,
		       jiffies);
		spin_unlock(&session->task_lock);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_lun_tasktimeout(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target, lun;
	struct iscsi_task *t;
	unsigned int cnt = 0xFFFFFFFF;
	char *end; 
	char *bp;
	char *tmp;
	int cmd_len;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
        bp = tmp;
        end = tmp + count;

	session = get_session_by_lun(buf, bp, end, &lun);

	if (session) {
		bus = session->channel;
		target = session->target_id;

		cmd_len = find_number(bp, end, &cnt);
		printk("iSCSI: session iscsi bus %d target id %d faking up to"
		       " %u task timeouts for LUN %u at %lu\n",
		       session->iscsi_bus, session->target_id, cnt, lun,
		       jiffies);

		spin_lock(&session->task_lock);
		/* fake task timeouts, to try to test
		 * the error recovery code
		 */
		list_for_each_entry(t, &session->task_arrival_order,
			task_order_link) {
			if (cnt == 0)
				break;

			if ((t->lun == lun) && !test_bit(0, &t-> timedout)) {
				printk("iSCSI: session iscsi bus %d target id"
				       " %d faking task timeout of itt %u,"
				       " LUN %u, sc 0x%x at %lu\n",
				       session->iscsi_bus, session->target_id,
				       t->itt, t->lun, t->scsi_cmnd-> cmnd[0],
				       jiffies);

				/* make the task look like it timedout */
				del_task_timer(t);
				set_bit(t->lun, session->
				luns_timing_out);
				smp_wmb();
				set_bit(0, &t->timedout);
				smp_mb();
				cnt--;
			}
		}
		wake_tx_thread(SESSION_TASK_TIMEDOUT, session);
		spin_unlock(&session->task_lock);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_lun_cmndtimeout(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target, lun;
	struct iscsi_task *t;
	Scsi_Cmnd *sc;
	unsigned int cnt = 0xFFFFFFFF;
	char *end; 
	char *bp;
	char *tmp;
	int cmd_len;
	unsigned long noqueue_flags_;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
        bp = tmp;
        end = tmp + count;

	session = get_session_by_lun(buf, bp, end, &lun);

	if (session) {
		bus = session->channel;
		target = session->target_id;

		cmd_len = find_number(bp, end, &cnt);
		printk("iSCSI: session iscsi bus %d target id %d faking up to"
		       " %u command timeouts for LUN %u at %lu\n",
		       session->iscsi_bus, session->target_id, cnt, lun,
		       jiffies);

		spin_lock(&session->task_lock);
		spin_lock_irqsave(&session->scsi_cmnd_lock, noqueue_flags_);
		/* fake command timeouts for all
		 * tasks and queued commands
		 */
		list_for_each_entry(t, &session->task_arrival_order,
			task_order_link) {
			if (cnt == 0)
				goto finished_lun;

			if ((t->lun == lun) && t->scsi_cmnd &&
				    !test_bit(COMMAND_TIMEDOUT,
					command_flags(t-> scsi_cmnd))) {
				printk("iSCSI: session iscsi bus %d target id"
				       " %d faking command timeout of itt %u,"
				       " LUN %u, cmnd 0x%x at %lu\n",
				       session->iscsi_bus, session->target_id,
				       t->itt, t->lun, t->scsi_cmnd-> cmnd[0],
				       jiffies);

				/* make the task look like it timedout */
				del_command_timer(t->scsi_cmnd);
				set_bit (COMMAND_TIMEDOUT,
				command_flags(t->scsi_cmnd));
				cnt--;
			}
		}

		for (sc = session->retry_cmnd_head; sc;
		    sc = (Scsi_Cmnd *) sc->host_scribble) {
			if (cnt == 0)
				goto finished_lun;

			if (sc->device->lun == lun) {
				printk("iSCSI: session iscsi bus %d target"
				       " id %d faking command timeout of retry"
				       " cmnd 0x%x, LUN %u, at %lu\n",
				       session->iscsi_bus, session->target_id,
				       sc->cmnd[0], sc->device->lun, jiffies);
				del_command_timer(sc);
				set_bit (COMMAND_TIMEDOUT, command_flags(sc));
				cnt--;
			}
		}
		for (sc = session->deferred_cmnd_head; sc;
		    sc = (Scsi_Cmnd *) sc->host_scribble) {
			if (cnt == 0)
				goto finished_lun;

			if (sc->device->lun == lun) {
				printk("iSCSI: session iscsi bus %d target id"
				       " %d faking command timeout of deferred"
				       " cmnd 0x%x, LUN %u, at %lu\n",
				       session->iscsi_bus, session->target_id,
				       sc->cmnd[0], sc->device->lun, jiffies);
				del_command_timer(sc);
				set_bit (COMMAND_TIMEDOUT, command_flags(sc));
				cnt--;
			}
		}
		for (sc = session->scsi_cmnd_head; sc;
			    sc = (Scsi_Cmnd *) sc->host_scribble) {
			if (cnt == 0)
				goto finished_lun;

			if (sc->device->lun == lun) {
				printk("iSCSI: session iscsi bus %d target id"
				       " %d faking command timeout of normal"
				       " cmnd 0x%x, LUN %u, at %lu\n",
				       session->iscsi_bus, session->target_id,
				       sc->cmnd[0], sc->device->lun, jiffies);
				del_command_timer(sc);
				set_bit (COMMAND_TIMEDOUT, command_flags(sc));
				cnt--;
			}
		}
		finished_lun:
		smp_wmb();
		set_bit(SESSION_COMMAND_TIMEDOUT, &session->control_bits);
		smp_wmb();

		/* wake up the tx thread to deal
		 * with the timeout
		 */
		set_bit(TX_WAKE, &session->control_bits);
		smp_mb();
		/* we can't know which wait_q the
		 * tx thread is in (if any), so wake
		 * them both
		 */
		wake_up(&session->tx_wait_q);
		wake_up(&session->login_wait_q);

		spin_unlock_irqrestore(&session->scsi_cmnd_lock, noqueue_flags_);
		spin_unlock(&session->task_lock);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_lun_status(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target, lun;
	char *end; 
	char *bp;
	char *tmp;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_lun(buf, bp, end, &lun);

	if (session) {
		bus = session->channel;
		target = session->target_id;

		spin_lock(&session->task_lock);
		printk("iSCSI: session iscsi bus %d target id %d LUN %u"
		       " detected=%s, activated=%s, timing out=%s, doing"
		       " recovery=%s, delaying commands=%s, unreachable=%s\n",
		       session->iscsi_bus, session->target_id, lun,
		       test_bit(lun, session->luns_detected)
		       ? "yes" : "no", test_bit(lun, session->luns_activated)
		       ? "yes" : "no", test_bit(lun, session->luns_timing_out)
		       ? "yes" : "no", test_bit(lun,
						session->luns_doing_recovery)
		       ? "yes" : "no", test_bit(lun,
					session-> luns_delaying_commands)
		       ? "yes" : "no", test_bit(lun,
						session-> luns_unreachable)
		       ? "yes" : "no");

		spin_unlock(&session->task_lock);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_nop_target(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target;

	char *end; 
	char *bp;
	char *tmp;
	int cmd_len;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		unsigned int data_length = 0;
		bus = session->channel;
		target = session->target_id;

		if ((cmd_len = find_number(bp, end, &data_length)) == 0) {
			printk("iSCSI: session iscsi bus %d target id %d Nop"
			       " test couldn't determine data length\n",
			       session->iscsi_bus, session->target_id);
			/* done with the session */
			drop_reference(session);
			kfree(tmp);
			return -EINVAL;
		}

		if (data_length) {
			printk ("iSCSI: session for bus %u target %u, %d byte"
			        " Nop data test requested at %lu\n", bus,
			        target, data_length, jiffies);
			iscsi_ping_test_session(session, data_length);
		} else {
			printk("iSCSI: session for bus %u target %u,"
			       " 0 byte Nop data test ignored at %lu\n", bus,
			       target, jiffies);
		}
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_target_portal(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target;
	char *end;
	char *bp;
	char *tmp;
	int cmd_len;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		bus = session->channel;
		target = session->target_id;

		if (session->portal_failover) {
			unsigned int portal = 0;

			if ((cmd_len = find_number(bp, end, &portal)) == 0) {
				printk("iSCSI: session for bus %u target %u,"
				       " no portal specified\n", bus, target);
				/* done with the session */
				drop_reference(session);
				kfree(tmp);
				return -EINVAL;
			}
			spin_lock(&session->portal_lock);
#ifndef DEBUG
			if (portal < session->num_portals) {
#endif
				session-> requested_portal = portal;
				session-> fallback_portal = session->
							    current_portal;
				printk("iSCSI: session for bus %u target %u "
				       "requesting switch to portal %u at"
				       "  %lu\n", bus, target, portal, jiffies);

				/* request a logout for the current session */
				spin_lock(&session-> task_lock);
				iscsi_request_logout(session,
					 session->active_timeout,
						 session->active_timeout);
				spin_unlock(&session->task_lock);
#ifndef DEBUG
			} else {
				printk("iSCSI:session for bus %u target %u "
				       "can't switch to portal %u, only %d "
				       "portals\n", bus, target, portal,
				       session->num_portals);
			}
#endif
			spin_unlock(&session-> portal_lock);
		} else
			printk("iSCSI: session for bus %u target %u can't"
			       " switch to requested portal, because portal"
			       " failover is disabled.\n", bus, target);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_logout_session(struct class_device *class_dev, const char *buf, size_t count)
{
        struct iscsi_session *session;
	unsigned int bus, target;
        char *bp, *end, *tmp;
        int cmd_len;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		unsigned int deadline = 0;
		unsigned int response_deadline = 0;

		bus = session->channel;
		target = session->target_id;

		if ((cmd_len = find_number(bp, end, &deadline)) == 0) {
			deadline = session->active_timeout;
		}
		bp += cmd_len;

		if ((cmd_len = find_number(bp, end, &response_deadline)) == 0) {
			response_deadline = session->active_timeout;
		}
		bp += cmd_len;

		printk("iSCSI: session for bus %u target %u requesting logout"
		       " at %lu, logout deadline %u seconds, response deadline"
		       " %u seconds\n", bus, target, jiffies, deadline,
		       response_deadline);
		/* request a logout for the current session */
		spin_lock(&session->task_lock);
		iscsi_request_logout(session, session->active_timeout,
				session-> active_timeout);
		spin_unlock(&session->task_lock);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_target_drop(struct class_device *class_dev, const char *buf, size_t count)
{
        struct iscsi_session *session;
        unsigned int bus, target;
        char *end;
        char *bp;
        char *tmp;
        int cmd_len;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		unsigned int time2wait = 0;
		bus = session->channel;
		target = session->target_id;

		if ((cmd_len = find_number(bp, end, &time2wait))) {
			session->time2wait = time2wait;
			smp_mb();
			printk("iSCSI: dropping session for bus %u target %u"
			       " at %lu, time2wait %u\n", bus, target, jiffies,
			       time2wait);
		} else {
			printk("iSCSI: dropping session for bus %u target %u"
			       " at %lu\n", bus, target, jiffies);
		}
		bp += cmd_len;

		iscsi_drop_session(session);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
terminate_session_info(struct class_device *class_dev, const char *buf, size_t count)
{
        struct iscsi_session *session;
        unsigned int bus, target;
        char *end;
        char *bp;
        char *tmp;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		bus = session->channel;
		target = session->target_id;
		printk("iSCSI: terminating session for bus %u target %u at"
		       " %lu\n", bus, target, jiffies);
		iscsi_terminate_session(session);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_target_queues(struct class_device *class_dev, const char *buf, size_t count)
{
        struct iscsi_session *session;
        unsigned int bus, target;
        char *end;
        char *bp;
        char *tmp;
	unsigned long noqueue_flags_;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		bus = session->channel;
		target = session->target_id;

		/* show all of the session's queues */
		spin_lock(&session->task_lock);
		spin_lock_irqsave(&session->scsi_cmnd_lock, noqueue_flags_);
		printk("iSCSI: session iscsi bus %d target id %d to %s, bits"
		       " 0x%08lx,  next itt %u at %lu\n", session->iscsi_bus,
		       session->target_id, session->log_name,
		       session->control_bits, session->itt, jiffies);
		printk("iSCSI: session iscsi bus %d target id %d ExpCmdSN"
		       " %08u, next CmdSN %08u, MaxCmdSN %08u, window closed"
		       " %lu times, full %lu times\n", session->iscsi_bus,
		       session->target_id, session->ExpCmdSn, session->CmdSn,
		       session->MaxCmdSn, session->window_closed,
		       session->window_full);
		print_session_tasks(session);
		print_session_cmnds(session);
		spin_unlock_irqrestore(&session->scsi_cmnd_lock, noqueue_flags_);
		spin_unlock(&session->task_lock);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_target_notready(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target;
	char *end;
	char *bp;
	char *tmp;
	int cmd_len;
	unsigned long noqueue_flags_;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		unsigned int notready = 1;
		bus = session->channel;
		target = session->target_id;
		cmd_len = find_number(bp, end, &notready);

		spin_lock_irqsave(&session->scsi_cmnd_lock, noqueue_flags_);
		session->fake_not_ready = notready;
		printk("iSCSI: session iscsi bus %d target id %d will fake"
		       " %u NOT_READY errors at %lu\n", session->iscsi_bus,
		       session->target_id, notready, jiffies);
		spin_unlock_irqrestore(&session->scsi_cmnd_lock, noqueue_flags_);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_target_printcmnd(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target;
	char *end;
	char *bp;
	char *tmp;
	int cmd_len;
	unsigned long noqueue_flags_;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		bus = session->channel;
		target = session->target_id;
		cmd_len = find_number(bp, end, &count);

		spin_lock_irqsave(&session->scsi_cmnd_lock, noqueue_flags_);
		session->print_cmnds = count;
		printk ("iSCSI: session iscsi bus %d target id %d will print"
		        " %u commands at %lu\n", session->iscsi_bus,
		        session->target_id, count, jiffies);
		spin_unlock_irqrestore(&session->scsi_cmnd_lock, noqueue_flags_);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_target_unreachable(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target;
	char *end;
	char *bp;
	char *tmp;
	int cmd_len;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		unsigned int count = 1;
		bus = session->channel;
		target = session->target_id;

		cmd_len = find_number(bp, end, &count);

		spin_lock(&session->task_lock);
		session->fake_status_lun = -1;
		session->fake_status_unreachable = count;
		printk("iSCSI: session iscsi bus %d target id %d will fake %u"
		       " iSCSI transport errors at %lu\n", session->iscsi_bus,
		       session->target_id, count, jiffies);
		spin_unlock(&session->task_lock);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_target_busy(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target;
	char *end;
	char *bp;
	char *tmp;
	int cmd_len;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		unsigned int count = 1;
		bus = session->channel;
		target = session->target_id;

		cmd_len = find_number(bp, end, &count);

		spin_lock(&session->task_lock);
		session->fake_status_lun = -1;
		session->fake_status_busy = count;
		printk("iSCSI: session iscsi bus %d target id %d will fake %u"
		       " SCSI status BUSY responses at %lu\n",
		       session->iscsi_bus, session->target_id, count, jiffies);
		spin_unlock(&session->task_lock);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_target_queuefull(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target;
	char *end;
	char *bp;
	char *tmp;
	int cmd_len;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		unsigned int count = 1;
		bus = session->channel;
		target = session->target_id;

		cmd_len = find_number(bp, end, &count);

		spin_lock(&session->task_lock);
		session->fake_status_lun = -1;
		session->fake_status_queue_full = count;
		printk("iSCSI: session iscsi bus %d target id %d will fake %u"
		       " SCSI status QUEUE_FULL responses at %lu\n",
		       session->iscsi_bus, session->target_id, count, jiffies);
		spin_unlock(&session->task_lock);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_target_abort(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target;
	char *end;
	char *bp;
	char *tmp;
	int cmd_len;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		unsigned int count = 1;
		bus = session->channel;
		target = session->target_id;

		cmd_len = find_number(bp, end, &count);

		spin_lock(&session->task_lock);
		session->fake_status_lun = -1;
		session->fake_status_aborted = count;
		printk("iSCSI: session iscsi bus %d target id %d will fake %u"
		       " target command aborted responses at %lu\n",
		       session->iscsi_bus, session->target_id, count, jiffies);
		spin_unlock(&session->task_lock);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}


static ssize_t
iscsi_target_tasktimeouts(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target;
	char *end;
	char *bp;
	char *tmp;
	int cmd_len;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		struct iscsi_task *t;
		unsigned int count = 0xFFFFFFFF;
		bus = session->channel;
		target = session->target_id;

		cmd_len = find_number(bp, end, &count);
		printk("iSCSI: session iscsi bus %d target id %d faking up to"
		       " %u task timeouts at %lu\n", session->iscsi_bus,
		       session->target_id, count, jiffies);

		spin_lock(&session->task_lock);
		/* fake task timeouts, to try to test 
		 * the error recovery code 
		 */
		list_for_each_entry(t, &session->task_arrival_order,
					    task_order_link) {
			if (count == 0)
				break;

			if (!test_bit(0, &t->timedout)) {
				printk("iSCSI: session iscsi bus %d target id"
				       " %d faking task timeout of itt %u,"
				       " LUN %u, sc 0x%x at %lu\n",
				       session->iscsi_bus, session->target_id,
				       t->itt, t->lun, t->scsi_cmnd->cmnd[0],
				       jiffies);

				/* make the task look like it timedout */
				del_task_timer(t);
				set_bit(t->lun, session-> luns_timing_out);
				smp_wmb();
				set_bit(0, &t->timedout);
				smp_mb();
				count--;
			}
		}
		wake_tx_thread(SESSION_TASK_TIMEDOUT, session);
		spin_unlock(&session->task_lock);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_target_cmndtimeouts(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target;
	char *end;
	char *bp;
	char *tmp;
	int cmd_len;
	unsigned long noqueue_flags_;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		unsigned int count = 0xFFFFFFFF;
		struct iscsi_task *t;
		Scsi_Cmnd *sc;

		bus = session->channel;
		target = session->target_id;

		cmd_len = find_number(bp, end, &count);
		printk("iSCSI: session iscsi bus %d target id %d faking up to"
		       " %u command timeouts at %lu\n", session->iscsi_bus,
		       session->target_id, count, jiffies);

		spin_lock(&session->task_lock);
		spin_lock_irqsave(&session->scsi_cmnd_lock, noqueue_flags_);

		/* fake command timeouts for all tasks 
		 * and queued commands 
		 */
		list_for_each_entry(t, &session->task_arrival_order,
						    task_order_link) {
			if (count == 0)
				goto finished;

			if (t->scsi_cmnd && !test_bit(COMMAND_TIMEDOUT,
					      command_flags(t-> scsi_cmnd))) {
				printk("iSCSI: session iscsi bus %d target id"
				       " %d faking command timeout of itt %u,"
				       " LUN %u, cmnd 0x%x at %lu\n",
				       session->iscsi_bus, session->target_id,
				       t->itt, t->lun, t->scsi_cmnd->cmnd[0],
				       jiffies);

				/* make the task look like it timedout */
				del_command_timer(t-> scsi_cmnd);
				set_bit (COMMAND_TIMEDOUT,
				    		 command_flags(t-> scsi_cmnd));
				count--;
			}
		}
		for (sc = session->retry_cmnd_head; sc;
		     sc = (Scsi_Cmnd *) sc->host_scribble) {
			if (count == 0)
				goto finished;

			printk("iSCSI: session iscsi bus %d target id %d"
			       " faking command timeout of retry cmnd 0x%x,"
			       " LUN %u, at %lu\n", session->iscsi_bus,
			       session->target_id, sc->cmnd[0],
			       sc->device->lun, jiffies);
			del_command_timer(sc);
			set_bit(COMMAND_TIMEDOUT, command_flags(sc));
			count--;
		}
		for (sc = session->deferred_cmnd_head; sc;
			sc = (Scsi_Cmnd *) sc->host_scribble) {
			if (count == 0)
				goto finished;

			printk("iSCSI: session iscsi bus %d target id %d"
			       " faking command timeout of deferred cmnd"
			       " 0x%x, LUN %u, at %lu\n", session->iscsi_bus,
			       session->target_id, sc->cmnd[0], sc->device->lun,
			       jiffies);
			del_command_timer(sc);
			set_bit(COMMAND_TIMEDOUT, command_flags(sc));
			count--;
		}
		for (sc = session->scsi_cmnd_head; sc;
			     sc = (Scsi_Cmnd *) sc->host_scribble) {
			if (count == 0)
				goto finished;

			printk("iSCSI: session iscsi bus %d target id %d"
			       " faking command timeout of normal cmnd 0x%x,"
			       " LUN %u, at %lu\n", session->iscsi_bus,
			       session->target_id, sc->cmnd[0],
			       sc->device->lun, jiffies);
			del_command_timer(sc);
			set_bit(COMMAND_TIMEDOUT, command_flags(sc));
			count--;
		}

		finished:
		smp_wmb();
		set_bit(SESSION_COMMAND_TIMEDOUT, &session->control_bits);
		smp_wmb();

		/* wake up the tx thread to deal with the timeout */
		set_bit(TX_WAKE, &session->control_bits);
		smp_mb();

		/* we can't know which wait_q the tx thread is in (if any), 
		 * so wake them both 
		 */
		wake_up(&session->tx_wait_q);
		wake_up(&session->login_wait_q);

		spin_unlock_irqrestore(&session->scsi_cmnd_lock, noqueue_flags_);
		spin_unlock(&session->task_lock);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_writeheadermismatch( struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target;
	char *end;
	char *bp;
	char *tmp;
	int cmd_len;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		unsigned int mismatch = 1;
		bus = session->channel;
		target = session->target_id;

		cmd_len = find_number(bp, end, &mismatch);

		spin_lock(&session->task_lock);
		session->fake_write_header_mismatch = mismatch;
		printk("iSCSI: session iscsi bus %d target id %d will fake %u"
		       " write HeaderDigest mismatches at %lu\n",
		       session->iscsi_bus, session->target_id, mismatch,
		       jiffies);
		spin_unlock(&session->task_lock);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_readdatamismatch(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target;
	char *end;
	char *bp;
	char *tmp;
	int cmd_len;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		unsigned int mismatch = 1;
		bus = session->channel;
		target = session->target_id;

		cmd_len = find_number(bp, end, &mismatch);

		spin_lock(&session->task_lock);
		session->fake_read_data_mismatch = mismatch;
		printk("iSCSI: session iscsi bus %d target id %d will fake %u"
		       " read DataDigest mismatches at %lu\n",
		       session->iscsi_bus, session->target_id, mismatch,
		       jiffies);
		spin_unlock(&session->task_lock);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_writedatamismatch(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target;
	char *end;
        char *bp;
        char *tmp;
        int cmd_len;

        tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
        memcpy(tmp, buf, count);
        bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		unsigned int mismatch = 1;
		bus = session->channel;
		target = session->target_id;

		cmd_len = find_number(bp, end, &mismatch);

		spin_lock(&session->task_lock);
		session->fake_write_data_mismatch = mismatch;
		printk("iSCSI: session iscsi bus %d target id %d will fake %u"
		       " write DataDigest mismatches at %lu\n",
		       session->iscsi_bus, session->target_id, mismatch,
		       jiffies);
		spin_unlock(&session->task_lock);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_target_ignore(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int completions = 0;
	unsigned int aborts = 0;
	unsigned int abort_task_sets = 0;
	unsigned int lun_resets = 0;
	unsigned int warm_resets = 0;
	unsigned int cold_resets = 0;
	unsigned int bus, target;
	char *end;
	char *bp;
	char *tmp;
	int cmd_len;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		bus = session->channel;
		target = session->target_id;

		cmd_len = find_number(bp, end, &completions);
		bp += cmd_len;
		cmd_len = find_number(bp, end, &aborts);
		bp += cmd_len;
		cmd_len = find_number(bp, end, &abort_task_sets);
		bp += cmd_len;
		cmd_len = find_number(bp, end, &lun_resets);
		bp += cmd_len;
		cmd_len = find_number(bp, end, &warm_resets);
		bp += cmd_len;
		cmd_len = find_number(bp, end, &cold_resets);

		printk("iSCSI: session for bus %u target %u at %lu, ignore %u"
		       " completions, %u aborts, %u abort task sets, %u"
		       " LUN resets, %u warm target resets, %u cold target"
		       " resets\n", bus, target, jiffies, completions, aborts,
		       abort_task_sets, lun_resets, warm_resets, cold_resets);

		spin_lock(&session->task_lock);
		session->ignore_lun = -1;
		session->ignore_completions = completions;
		session->ignore_aborts = aborts;
		session->ignore_abort_task_sets = abort_task_sets;
		session->ignore_lun_resets = lun_resets;
		session->ignore_warm_resets = warm_resets;
		session->ignore_cold_resets = cold_resets;
		spin_unlock(&session->task_lock);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_target_reject(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int aborts = 0;
	unsigned int abort_task_sets = 0;
	unsigned int lun_resets = 0;
	unsigned int warm_resets = 0;
	unsigned int cold_resets = 0;
	unsigned int bus, target;
	char *end;
	char *bp;
	char *tmp;
	int cmd_len;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		bus = session->channel;
		target = session->target_id;

		cmd_len = find_number(bp, end, &aborts);
		bp += cmd_len;
		cmd_len = find_number(bp, end, &abort_task_sets);
		bp += cmd_len;
		cmd_len = find_number(bp, end, &lun_resets);
		bp += cmd_len;
		cmd_len = find_number(bp, end, &warm_resets);
		bp += cmd_len;
		cmd_len = find_number(bp, end, &cold_resets);

		printk("iSCSI: session for bus %u target %u at %lu, reject %u"
		       " aborts, %u abort task sets, %u LUN resets, %u warm "
		       " target resets, %u cold target resets\n", bus, target,
		       jiffies, aborts, abort_task_sets, lun_resets,
		       warm_resets, cold_resets);

		spin_lock(&session->task_lock);
		session->reject_lun = -1;
		session->reject_aborts = aborts;
		session->reject_abort_task_sets = abort_task_sets;
		session->reject_lun_resets = lun_resets;
		session->reject_warm_resets = warm_resets;
		session->reject_cold_resets = cold_resets;
		spin_unlock(&session->task_lock);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_target_reset(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target;
	char *end;
	char *bp;
	char *tmp;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		bus = session->channel;
		target = session->target_id;

		printk("iSCSI: session iscsi bus %d target id %d warm target"
		       " reset requested at %lu\n", session->iscsi_bus,
		       session->target_id, jiffies);
		wake_tx_thread(SESSION_RESET_REQUESTED, session);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_target_cmdtimeout(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target;
	char *end;
	char *bp;
	char *tmp;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		bus = session->channel;
		target = session->target_id;

		printk("iSCSI: session iscsi bus %d target id %d waking "
		       "tx thread SESSION_COMMAND_TIMEDOUT at %lu\n",
		       session->iscsi_bus, session->target_id, jiffies);
		wake_tx_thread(SESSION_COMMAND_TIMEDOUT, session);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_target_tasktimeout(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target;
	char *end;
	char *bp;
	char *tmp;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		bus = session->channel;
		target = session->target_id;

		printk("iSCSI: session iscsi bus %d target id %d waking "
		       "tx thread SESSION_TASK_TIMEDOUT at %lu\n",
		       session->iscsi_bus, session->target_id, jiffies);
		wake_tx_thread(SESSION_TASK_TIMEDOUT, session);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_target_retry(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target;
	char *end;
	char *bp;
	char *tmp;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		bus = session->channel;
		target = session->target_id;

		printk("iSCSI: session iscsi bus %d target id %d waking "
		       "tx thread SESSION_RETRY_COMMANDS at %lu\n",
		       session->iscsi_bus, session->target_id, jiffies);
		wake_tx_thread(SESSION_RETRY_COMMANDS, session);
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

static ssize_t
iscsi_target_txcmnd(struct class_device *class_dev, const char *buf, size_t count)
{
	struct iscsi_session *session;
	unsigned int bus, target;
	char *end;
	char *bp;
	char *tmp;

	tmp = (char *)kmalloc(sizeof(char) * count, GFP_ATOMIC);
	if (tmp == NULL) {
		printk("iSCSI: could not allocate memory for attribute\n");
		return -ENOMEM;
	}
	memcpy(tmp, buf, count);
	bp = tmp;
	end = tmp + count;

	session = get_session_by_target(buf, bp, end);
	if (session) {
		bus = session->channel;
		target = session->target_id;

		printk("iSCSI: session iscsi bus %d target id %d waking "
		       "tx thread TX_SCSI_COMMAND at %lu\n", session->iscsi_bus,
		       session->target_id, jiffies);
		wake_tx_thread(TX_SCSI_COMMAND, session);
		/* done with the session */
		drop_reference(session);
	}
	kfree(tmp);
	return count;
}

/* Target specific attributes */

struct class_device_attribute target_cmdtimeout = {
	.attr = {
		.name =		"target_cmndtimeout",
		.mode =		S_IWUSR,
	},
	.store = iscsi_target_cmdtimeout,
};

struct class_device_attribute target_tasktimeout = {
	.attr = {
		.name =		"target_tasktimeout",
		.mode =		S_IWUSR,
	},
	.store = iscsi_target_tasktimeout,
};

struct class_device_attribute target_cmndtimeouts = {
	.attr = {
		.name =		"target_cmndtimeouts",
		.mode =		S_IWUSR,
	},
	.store = iscsi_target_cmndtimeouts,
};

struct class_device_attribute target_tasktimeouts = {
	.attr = {
		.name =		"target_tasktimeouts",
		.mode =		S_IWUSR,
	},
	.store = iscsi_target_tasktimeouts,
};

struct class_device_attribute target_unreachable = {
	.attr = {
		.name =		"target_unreachable",
		.mode =		S_IWUSR,
	},
	.store = iscsi_target_unreachable,
};

struct class_device_attribute target_queuefull = {
	.attr = {
		.name =		"target_queuefull",
		.mode =		S_IWUSR,
	},
	.store = iscsi_target_queuefull,
};

struct class_device_attribute target_busy = {
	.attr = {
		.name =		"target_busy",
		.mode =		S_IWUSR,
	},
	.store = iscsi_target_busy,
};

struct class_device_attribute target_ignore = {
	.attr = {
		.name =		"target_ignore",
		.mode =		S_IWUSR,
	},
	.store = iscsi_target_ignore,
};

struct class_device_attribute target_reject = {
	.attr = {
		.name =		"target_reject",
		.mode =		S_IWUSR,
	},
	.store = iscsi_target_reject,
};

struct class_device_attribute target_abort = {
	.attr = {
		.name =		"target_abort",
		.mode =		S_IWUSR,
	},
	.store = iscsi_target_abort,
};

struct class_device_attribute logout_session = {
	.attr = {
		.name =		"logout_session",
		.mode =		S_IWUSR,
	},
	.store = iscsi_logout_session,
};

struct class_device_attribute session_terminate = {
	.attr = {
		.name =		"session_terminate",
		.mode =		S_IWUSR,
	},
	.store = terminate_session_info,
};

struct class_device_attribute target_nop = {
	.attr = {
		.name =		"nop_target",
		.mode =		S_IWUSR,
	},
	.store = iscsi_nop_target,
};

struct class_device_attribute target_portal= {
	.attr = {
		.name =		"target_portal",
		.mode =		S_IWUSR,
	},
	.store = iscsi_target_portal,
};

struct class_device_attribute target_drop = {
	.attr = {
		.name =		"target_drop",
		.mode =		S_IWUSR,
	},
	.store = iscsi_target_drop,
};

struct class_device_attribute target_queues = {
	.attr = {
		.name =		"target_queues",
		.mode =		S_IWUSR,
	},
	.store = iscsi_target_queues,
};

struct class_device_attribute target_notready = {
	.attr = {
		.name =		"target_notready",
		.mode =		S_IWUSR,
	},
	.store = iscsi_target_notready,
};

struct class_device_attribute target_reset = {
	.attr = {
		.name =		"target_reset",
		.mode =		S_IWUSR,
	},
	.store = iscsi_target_reset,
};

struct class_device_attribute target_retry = {
	.attr = {
		.name =		"target_retry",
		.mode =		S_IWUSR,
	},
	.store = iscsi_target_retry,
};

struct class_device_attribute target_txcmnd = {
	.attr = {
		.name =		"target_txcmnd",
		.mode =		S_IWUSR,
	},
	.store = iscsi_target_txcmnd,
};

struct class_device_attribute target_printcmnd = {
	.attr = {
		.name =		"target_printcmnd",
		.mode =		S_IWUSR,
	},
	.store = iscsi_target_printcmnd,
};

struct class_device_attribute writeheadermismatch = {
	.attr = {
		.name =		"w_h_mismatch",
		.mode =		S_IWUSR,
	},
	.store = iscsi_writeheadermismatch,
};

struct class_device_attribute readdatamismatch = {
	.attr = {
		.name =		"r_d_mismatch",
		.mode =		S_IWUSR,
	},
	.store = iscsi_readdatamismatch,
};

struct class_device_attribute writedatamismatch = {
	.attr = {
		.name =		"w_d_mismatch",
		.mode =		S_IWUSR,
	},
	.store = iscsi_writedatamismatch,
};

/* Lun specific attributes are defined here */

struct class_device_attribute lun_cmndtimeout = {
	.attr = {
		.name =		"lun_cmndtimeout",
		.mode =		S_IWUSR,
	},
	.store = iscsi_lun_cmndtimeout,
};

struct class_device_attribute lun_tasktimeout = {
	.attr = {
		.name =		"lun_tasktimeout",
		.mode =		S_IWUSR,
	},
	.store = iscsi_lun_tasktimeout,
};

struct class_device_attribute lun_status = {
	.attr = {
		.name =		"lun_status",
		.mode =		S_IWUSR,
	},
	.store = iscsi_lun_status,
};

struct class_device_attribute lun_abort = {
	.attr = {
		.name =         "lun_abort",
		.mode =          S_IWUSR,
	},
	.store = iscsi_lun_abort,
};

struct class_device_attribute lun_unreachable = {
	.attr = {
		.name =		"lun_unreachable",
		.mode =		S_IWUSR,
	},
	.store = iscsi_lun_unreachable,
};

struct class_device_attribute lun_queuefull = {
	.attr = {
		.name =		"lun_queuefull",
		.mode =		S_IWUSR,
	},
	.store = iscsi_lun_queuefull,
};

struct class_device_attribute lun_busy = {
	.attr = {
		.name =		"lun_busy",
		.mode =		S_IWUSR,
	},
	.store = iscsi_lun_busy,
};

struct class_device_attribute lun_reject = {
	.attr = {
		.name =		"lun_reject",
		.mode =		S_IWUSR,
	},
	.store = iscsi_lun_reject,
};

struct class_device_attribute lun_ignore = {
	.attr = {
		.name =		"lun_ignore",
                .mode =		S_IWUSR,
	},
	.store = iscsi_lun_ignore,
};
