/*
 * iSCSI driver for Linux
 * Copyright (C) 2001 Cisco Systems, Inc.
 * maintained by linux-iscsi@cisco.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi-event-handler.c,v 1.1.4.1 2003/10/21 13:22:10 krishmnc Exp $
 *
 */

#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <syslog.h>
#include <scsi/sg.h>
#include <scsi/scsi.h>
#include <dirent.h>

#include "iscsi-kernel-event.h"
#define CONTROL_FILE "/dev/iscsictl"
#define AS_ERROR  LOG_ERR
#define MAX_SCSI_DISKS 128
#define MAX_SCSI_DISK_PARTITIONS 15
#define MAX_SCSI_GENERICS 256
#define MAX_SCSI_CDROMS 256
#define MAX_SCSI_TAPES 32
#define MAX_LUNS 256
#define BITS_PER_LONG 32

extern void logmsg(int priority, const char *fmt, ...);

struct scsi_idlun {
	int dev_id;
	int host_unique_id;
};

static int event_handler_state;
static volatile int stop_event = 0;

void
evt_sigterm_handler(int unused)
{
	stop_event = 0;
}

static void
ensure_directories_exist(char *pathname)
{
	char *end = pathname;

	/* skip leading slashes */
	while (end && *end && (*end == '/'))
		end++;

	while (end && (*end != '\0')) {
		/* if there is another slash, make the dir.
		 * FIXME: we ought to ignore errors when the directory exists,
		 * but report errors where the directory doesn't exist and
		 * we failed to create it.
		 */
		while ((*end != '/') && (*end != '\0'))
			end++;

		if (*end == '/') {
			*end = '\0';

			mkdir(pathname, 0755);

			*end = '/';
			end++;
		}
	}
}

static void
empty_directory(char *dir)
{
	struct dirent *dent;
	int rc;
	DIR *dirname;
	char name[100];

	dirname = opendir(dir);
	do {
		dent = readdir(dirname);
		if (dent)
			rc = 1;
		else {
			rc = 0;
			break;
		}
		strcpy(name, dir);
		strcat(name, "/");
		strcat(name, dent->d_name);
		if (strcmp(name, ".") && strcmp(name, "..")) {
			unlink(name);
		}
	} while (rc > 0);

	closedir(dirname);
}

static int
iscsi_remove_lun(unsigned int host_no, unsigned int channel, unsigned target_id,
		 int lun)
{
	char str[88];
	int fd, rc;

	sprintf(str, "scsi remove-single-device %d %d %d %d\n",
		host_no, channel, target_id, lun);
	str[sizeof (str) - 1] = '\0';

	if ((fd = open("/proc/scsi/scsi", O_WRONLY)) < 0) {
		logmsg(AS_ERROR, "failed to open /proc/scsi/scsi: %d\n", errno);
		return 0;
	}

	rc = write(fd, str, strlen(str));
	if (rc < 0) {
		logmsg(AS_ERROR,
		       "scsi remove-single-device %d %d %d %d failed\n",
		       host_no, channel, target_id, lun);
		return 0;
	}
	return rc;
}

static int
iscsi_add_lun(unsigned int host_no, unsigned int channel, unsigned target_id,
	      unsigned lun)
{
	char str[88];
	int fd, rc;

	sprintf(str, "scsi add-single-device %d %d %d %d\n",
		host_no, channel, target_id, lun);
	str[sizeof (str) - 1] = '\0';

	if ((fd = open("/proc/scsi/scsi", O_WRONLY)) < 0) {
		logmsg(AS_ERROR, "failed to open /proc/scsi/scsi: %d\n", errno);
		return 0;
	}

	rc = write(fd, str, strlen(str));
	if (rc < 0) {
		logmsg(AS_ERROR, "scsi add-single-device %d %d %d %d failed\n",
		       host_no, channel, target_id, lun);
		return 0;
	}
	return rc;
}

int
test_bit(unsigned long bit, unsigned long *map)
{
	int mod = bit % BITS_PER_LONG;
	int mask = 0;

	mask = (1U << mod);
	return (map[bit / BITS_PER_LONG] & mask);
}

void
clear_bit(unsigned long bit, unsigned long *map)
{
	int rem = bit % BITS_PER_LONG;
	int mask = 0;

	mask = (1U << rem);
	mask = ~mask;
	(map[bit / BITS_PER_LONG]) &= mask;
}

void
set_bit(unsigned long bit, unsigned long *map)
{
	int mod = bit % BITS_PER_LONG;
	int mask = 0;

	mask = (1U << mod);
	(map[bit / BITS_PER_LONG]) |= mask;
}

static void
get_device_scsi_quad(char *device_name, int *host, int *channel, int *target,
		     int *lun)
{
	int fd, rc;
	struct scsi_idlun tmp;

	fd = open(device_name, O_RDWR);
	rc = ioctl(fd, SCSI_IOCTL_GET_IDLUN, &tmp);
	*target = tmp.dev_id & 0xff;
	*lun = tmp.dev_id >> 8 & 0xff;
	*channel = tmp.dev_id >> 16 & 0xff;
	*host = tmp.dev_id >> 24 & 0xff;
}

static int
probe_luns(char *rd)
{
	char *c, *start;
	struct probe_devices *pd =
	    (struct probe_devices *) malloc(sizeof (struct probe_devices));
	int i;
	unsigned int lun_bitmap[8] = { 0xffffffff, 0, 0, 0, 0, 0, 0, 0 };
	start = rd;

	for (i = 0; i < ((ISCSI_MAX_LUNS + 31) / BITS_PER_LONG); i++) {
		c = start;
		c = strchr(c, '\n');
		*c = '\0';
		c++;
		pd->luns_found[i] = atoi(start);
		start = c;
	}

	for (i = 0; i < ((ISCSI_MAX_LUNS + 31) / BITS_PER_LONG); i++) {
		c = start;
		c = strchr(c, '\n');
		*c = '\0';
		c++;
		pd->luns_detected[i] = atoi(start);
		start = c;
	}

	for (i = 0; i < (ISCSI_MAX_LUNS + 31) / BITS_PER_LONG; i++) {
		c = start;
		c = strchr(c, '\n');
		*c = '\0';
		c++;
		pd->luns_activated[i] = atoi(start);
		start = c;
	}

	c = start;
	c = strchr(c, '\n');
	*c = '\0';
	c++;
	pd->host_no = atoi(start);

	start = c;
	c = strchr(c, '\n');
	*c = '\0';
	c++;
	pd->channel = atoi(start);

	start = c;
	c = strchr(c, '\n');
	*c = '\0';
	c++;
	pd->target_id = atoi(start);

	start = c;
	c = strchr(c, '\n');
	*c = '\0';
	c++;
	start = c;
	pd->lun_bitmap = &lun_bitmap[0];
//      pd->lun_bitmap[0] = atoi(start);

	start = c;
	c = strchr(c, '\n');
	*c = '\0';
	c++;
	pd->target_link_dir = start;

	start = c;
	c = strchr(c, '\n');
	*c = '\0';
	c++;
	pd->max_sd_devices = atoi(start);

	start = c;
	c = strchr(c, '\n');
	*c = '\0';
	c++;
	pd->max_sg_devices = atoi(start);

	start = c;
	c = strchr(c, '\n');
	*c = '\0';
	c++;
	pd->max_sr_devices = atoi(start);

	start = c;
	c = strchr(c, '\n');
	*c = '\0';
	c++;
	pd->max_st_devices = atoi(start);

	for (i = 0; i < ISCSI_MAX_LUNS; i++) {
		if (test_bit(i, pd->luns_detected)) {
			/* Check if lun has been removed */
			if (!test_bit(i, pd->luns_found)) {
				if (iscsi_remove_lun
				    (pd->host_no, pd->channel, pd->target_id,
				     i) != 0) {
					char c;
					char *lun_dir =
					    pd->target_link_dir +
					    strlen(pd->target_link_dir);
					char *bus_dir = lun_dir - 2;	
							/* before the slash */
					clear_bit(i, pd->luns_activated);
					clear_bit(i, pd->luns_detected);

					if (pd->target_link_dir[0] == '/') {
						sprintf(lun_dir, "lun%d/", i);
						rmdir(pd->target_link_dir);
					}
					/* If all luns on this target have 
					 * been deleted, remove the  target 
					 * entry.
					 */
					if (pd->target_link_dir[0] == '/') {
						/* and get rid of the target 
						 * dir itself 
						 */
						*lun_dir = '\0';
						rmdir(pd->target_link_dir);
						/* if the bus dir is empty now, 
						 * get rid of it too, but 
						 * don't corrupt the session's 
						 * target dir 
						 */
						while (*bus_dir != '/')
							bus_dir--;
						bus_dir++;
						c = *bus_dir;
						*bus_dir = '\0';
						rmdir(pd->target_link_dir);
						*bus_dir = c;
					}
				}
			} else {
				/* if allowed and not already activated 
				 * (successfully probed), probe it 
				 */
				if ((pd->lun_bitmap[i / 32] & (1 << (i % 32)))
				    && !test_bit(i, pd->luns_activated)) {
					iscsi_add_lun(pd->host_no, pd->channel,
						      pd->target_id, i);
				}
			}
		} else {
			if (test_bit(i, pd->luns_activated)) {
				char c;
				char *lun_dir =
				    pd->target_link_dir +
				    strlen(pd->target_link_dir);
				char *bus_dir = lun_dir - 2;	/* before the 
								 * slash 
								 */

				if (pd->target_link_dir[0] == '/') {
					sprintf(lun_dir, "lun%d/", i);
					empty_directory(pd->target_link_dir);
					rmdir(pd->target_link_dir);
					/* If  all luns on this target have 
					 * been deleted, remove the target 
					 * entry.
					 */
					if (pd->target_link_dir[0] == '/') {
						/* and get rid of the target 
						 * dir itself 
						 */
						*lun_dir = '\0';
						rmdir(pd->target_link_dir);
						/* if the bus dir is empty now, 
						 * get rid of it too, but 
						 * don't corrupt the session's 
						 * target dir 
						 */
						while (*bus_dir != '/')
							bus_dir--;
						bus_dir++;
						c = *bus_dir;
						*bus_dir = '\0';
						rmdir(pd->target_link_dir);
						*bus_dir = c;
					}
				}
			}
		}
	}

	if (pd->target_link_dir[0] == '/') {
		ensure_directories_exist(pd->target_link_dir);

		if (pd->max_sd_devices > 0) {
			char devname[20];
			char *lun_dir =
			    pd->target_link_dir + strlen(pd->target_link_dir);

			for (i = 0; i < MAX_SCSI_DISKS; i++) {
				int host = -1, channel = -1, id = -1, lun = -1;

				if (i < 26) {
					sprintf(devname, "/dev/sd%c", 'a' + i);
				} else {
					sprintf(devname, "/dev/sd%c%c",
						'a' + (i / 26) - 1,
						'a' + (i % 26));
				}

				get_device_scsi_quad(devname, &host, &channel,
						     &id, &lun);

				if ((host == pd->host_no)
				    && (channel == pd->channel)
				    && (id == pd->target_id)) {
					char *partition =
					    devname + strlen(devname);
					char *link;
					int p;

					sprintf(lun_dir, "lun%d/", lun);
					ensure_directories_exist(pd->
								 target_link_dir);
					link = lun_dir + strlen(lun_dir);

					strcpy(link, "disk");

					unlink(pd->target_link_dir);
					symlink(devname, pd->target_link_dir);
					for (p = 1;
					     p <= MAX_SCSI_DISK_PARTITIONS;
					     p++) {
						sprintf(partition, "%d", p);
						sprintf(link, "part%d", p);
						unlink(pd->target_link_dir);
						symlink(devname,
							pd->target_link_dir);
					}
				}
			}
			*lun_dir = '\0';
		}

		if (pd->max_sg_devices > 0) {
			char devname[20];
			char *lun_dir =
			    pd->target_link_dir + strlen(pd->target_link_dir);
			char *link;

			for (i = 0; i < MAX_SCSI_GENERICS; i++) {
				int host = -1, channel = -1, id = -1, lun = -1;

				sprintf(devname, "/dev/sg%d", i);

				get_device_scsi_quad(devname, &host, &channel,
						     &id, &lun);

				if ((host == pd->host_no)
				    && (channel == pd->channel)
				    && (id == pd->target_id)) {
					sprintf(lun_dir, "lun%d/", lun);
					ensure_directories_exist(pd->
								 target_link_dir);

					link = lun_dir + strlen(lun_dir);

					strcpy(link, "generic");

					unlink(pd->target_link_dir);
					symlink(devname, pd->target_link_dir);
					break;
				}
			}
			*lun_dir = '\0';
		}

		if (pd->max_sr_devices > 0) {
			char devname[20];
			char *lun_dir =
			    pd->target_link_dir + strlen(pd->target_link_dir);
			char *link;

			for (i = 0; i < MAX_SCSI_CDROMS; i++) {
				int host = -1, channel = -1, id = -1, lun = -1;

				sprintf(devname, "/dev/scd%d", i);

				get_device_scsi_quad(devname, &host, &channel,
						     &id, &lun);

				if ((host == pd->host_no)
				    && (channel == pd->channel)
				    && (id == pd->target_id)) {
					sprintf(lun_dir, "lun%d/", lun);
					ensure_directories_exist(pd->
								 target_link_dir);

					link = lun_dir + strlen(lun_dir);

					strcpy(link, "cd");

					unlink(pd->target_link_dir);
					symlink(devname, pd->target_link_dir);
					break;
				}
			}
			*lun_dir = '\0';
		}

		if (pd->max_st_devices > 0) {
			char devname[20];
			char *lun_dir =
			    pd->target_link_dir + strlen(pd->target_link_dir);
			char *link;

			for (i = 0; i < MAX_SCSI_TAPES; i++) {
				int host = -1, channel = -1, id = -1, lun = -1;

				sprintf(devname, "/dev/st%d", i);

				get_device_scsi_quad(devname, &host, &channel,
						     &id, &lun);

				if ((host == pd->host_no)
				    && (channel == pd->channel)
				    && (id == pd->target_id)) {
					sprintf(lun_dir, "lun%d/", lun);
					ensure_directories_exist(pd->
								 target_link_dir);

					link = lun_dir + strlen(lun_dir);

					strcpy(link, "mt");
					unlink(pd->target_link_dir);
					sprintf(devname, "/dev/st%d", i);
					symlink(devname, pd->target_link_dir);

					strcpy(link, "mtl");
					unlink(pd->target_link_dir);
					sprintf(devname, "/dev/st%dl", i);
					symlink(devname, pd->target_link_dir);

					strcpy(link, "mtm");
					unlink(pd->target_link_dir);
					sprintf(devname, "/dev/st%dm", i);
					symlink(devname, pd->target_link_dir);

					strcpy(link, "mta");
					unlink(pd->target_link_dir);
					sprintf(devname, "/dev/st%da", i);
					symlink(devname, pd->target_link_dir);

					strcpy(link, "mtn");
					unlink(pd->target_link_dir);
					sprintf(devname, "/dev/nst%d", i);
					symlink(devname, pd->target_link_dir);

					strcpy(link, "mtln");
					unlink(pd->target_link_dir);
					sprintf(devname, "/dev/nst%dl", i);
					symlink(devname, pd->target_link_dir);

					strcpy(link, "mtmn");
					unlink(pd->target_link_dir);
					sprintf(devname, "/dev/nst%dm", i);
					symlink(devname, pd->target_link_dir);

					strcpy(link, "mtan");
					unlink(pd->target_link_dir);
					sprintf(devname, "/dev/nst%da", i);
					symlink(devname, pd->target_link_dir);

					break;
				}
			}
			*lun_dir = '\0';
		}
	}

	return 0;
}

static int
remove_lun_req(char *req_d)
{
	char *c, *start;
	struct remove_lun *rd = (struct remove_lun *) 
				 malloc(sizeof (struct remove_lun));
	int i;
	char tmp;
	char *lun_dir = NULL;
	char *bus_dir = NULL;
	int lun_id;

	start = req_d;

	for (i = 0; i < (ISCSI_MAX_LUNS + 31) / BITS_PER_LONG; i++) {
		c = start;
		c = strchr(c, '\n');
		*c = '\0';
		c++;
		rd->luns_activated[i] = atoi(start);
		start = c;
	}

	start = c;
	c = strchr(c, '\n');
	*c = '\0';
	c++;
	rd->target_link_dir = start;

	c = start;
	c = strchr(c, '\n');
	*c = '\0';
	c++;
	rd->host_no = atoi(start);

	start = c;
	c = strchr(c, '\n');
	*c = '\0';
	c++;
	rd->channel = atoi(start);

	start = c;
	c = strchr(c, '\n');
	*c = '\0';
	c++;
	rd->target_id = atoi(start);

	start = c;
	c = strchr(c, '\n');
	*c = '\0';
	c++;
	rd->lun = atoi(start);

	lun_dir = rd->target_link_dir + strlen(rd->target_link_dir);
	bus_dir = lun_dir - 2;	/* before the slash */

	if (rd->lun == -1)
		lun_id = 0;
	else
		lun_id = rd->lun;

	for (; lun_id < ISCSI_MAX_LUNS; lun_id++) {
		if (rd->target_link_dir[0] == '/') {
			sprintf(lun_dir, "lun%d/", lun_id);
			empty_directory(rd->target_link_dir);

			rmdir(rd->target_link_dir);
		}

		if ((test_bit(lun_id, rd->luns_activated)) && (rd->lun != -1)) {
			/* tell Linux to release the Scsi_Devices */
			iscsi_remove_lun(rd->host_no, rd->channel,
					 rd->target_id, lun_id);
		}

		if (rd->target_link_dir[0] == '/') {
			/* and get rid of the target dir itself */
			*lun_dir = '\0';
			rmdir(rd->target_link_dir);

			/* if the bus dir is empty now, get rid of it too, 
			 * but don't corrupt the session's target dir 
			 */
			while (*bus_dir != '/')
				bus_dir--;
			bus_dir++;	/* leave the slash */
			tmp = *bus_dir;
			*bus_dir = '\0';

			rmdir(rd->target_link_dir);
			*bus_dir = tmp;
		}
		if (rd->lun == -1)
			continue;
	}

	return 0;
}

static int
clear_lun_req(char *rd)
{
	char *c, *start;
	struct clear_luns *r = (struct clear_luns *) 
				malloc(sizeof (struct clear_luns));
	char str[88];
	int fd;
	int rc = 0;
	int i = 0;
	int host, chan, target, lun;

	start = rd;
	c = start;
	for (i = 0; i < 2; i++) {
		start = c;
		c = strchr(c, '\n');
		*c = '\0';
		c++;
		r->host[i] = atoi(start);
	}
	for (i = 0; i < 2; i++) {
		start = c;
		c = strchr(c, '\n');
		*c = '\0';
		c++;
		r->chan[i] = atoi(start);
	}

	for (i = 0; i < 2; i++) {
		start = c;
		c = strchr(c, '\n');
		*c = '\0';
		c++;
		r->tid[i] = atoi(start);
	}
	for (i = 0; i < 2; i++) {
		start = c;
		c = strchr(c, '\n');
		*c = '\0';
		c++;
		r->lun[i] = atoi(start);
	}

	for (host = r->host[0]; host <= r->host[1]; host++) {
		for (chan = r->chan[0]; chan <= r->chan[1]; chan++) {
			for (target = r->tid[0]; target <= r->tid[1]; target++) 
			{
				for (lun = r->lun[0]; lun <= r->lun[1]; lun++) {
					sprintf(str,
						"scsi remove-single-device %d "
						"%d %d %d\n",
						host, chan, target, lun);

					str[sizeof (str) - 1] = '\0';

					if ((fd =
					     open("/proc/scsi/scsi",
						  O_WRONLY)) < 0) {
						printf
						    ("failed to open "
						     "/proc/scsi/scsi: %d\n",
						     errno);
						return -1;
					}
					rc = write(fd, str, strlen(str));
				}
			}
		}
	}
	return rc;
}

static int
process_kernel_request(struct kernel_request * req)
{

	switch (req->req_type) {
	case REMOVE_LUN:
		return remove_lun_req(req->data);
		break;

	case CLEAR_LUNS:
		return clear_lun_req(req->data);
		break;

	case PROBE_LUNS:
		return probe_luns(req->data);
		break;

	case STOP_EVENT_HANDLER:
		event_handler_state = STOP;
		stop_event = 1;
		break;

	default:
		logmsg(AS_ERROR, "illegal request\n");
		break;
	}

	return 1;
}

static void
send_response(struct kernel_request * req, int rc)
{
	switch (req->req_type) {
	case REMOVE_LUN:
		sprintf(req->data, "%d\n", rc);
		break;

	case PROBE_LUNS:
		sprintf(req->data, "%d\n", rc);
		break;

	case CLEAR_LUNS:
		sprintf(req->data, "%d\n", rc);
		break;

	default:
		break;
	}
}

static int
wait_for_kernel_event(void)
{
	struct kernel_request *req;
	int fd;
	int rc;
	struct sigaction action;

	fd = open(CONTROL_FILE, O_RDWR);
	if (fd < 0) {
		logmsg(AS_ERROR, "cannot open the control device");
		return -1;
	}
	memset(&action, 0, sizeof (struct sigaction));
	action.sa_sigaction = NULL;
	action.sa_flags = 0;
	action.sa_handler = evt_sigterm_handler;
	sigaction(SIGTERM, &action, NULL);

	req = (struct kernel_request *) malloc(sizeof (struct kernel_request));
	memset(req, 0, sizeof (*req));

	event_handler_state = START;
	req->req_type = START_EVENT_HANDLER;
	memset(req->data, 0, sizeof (req->data));
	while (!stop_event) {
		ioctl(fd, WAIT_FOR_EVENT, req);

		rc = process_kernel_request(req);
		if (event_handler_state == START)
			send_response(req, rc);
		else
			break;
	}

	free(req);
	close(fd);

	return 0;
}

void
start_event_handler()
{
	int pid;

      fork_again:
	pid = fork();
	if (pid == 0) {
		wait_for_kernel_event();
		exit(0);
	}
	if (pid < 0) {
		logmsg(AS_ERROR, "cannot fork kernel_event_handler\n");
		sleep(1);
		goto fork_again;
	}
}

int
stop_event_handler()
{
	int fd;
	int rc;

	fd = open(CONTROL_FILE, O_RDWR);
	if (fd < 0) {
		logmsg(AS_ERROR, "cannot open the control device");
		return -1;
	}

	rc = ioctl(fd, RELEASE_EVENT_HANDLER, STOP);
	if (rc != 0) {
		logmsg(AS_ERROR, "RELEASE_EVENT_HANDLER ioctl failed: %d\n",
		       errno);
		return -1;
	}

	close(fd);
	return 0;
}
