/*
 * iSCSI driver for Linux
 * Copyright (C) 2002 Cisco Systems, Inc.
 * maintained by linux-iscsi-devel@lists.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi-id.c,v 1.1.2.8 2004/01/30 20:42:06 haapi Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <syslog.h>
#include <ctype.h>
#include <stdarg.h>
#include <linux/limits.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <scsi/scsi.h>
#include <scsi/scsi_ioctl.h>
#include <sysfs/libsysfs.h>

/*
 * temporary names for mknod.
 */

#define TMP_DIR	"/tmp"
#define TMP_PREFIX "scsi"
#define MAX_NAME_LEN    80
#define ISCSI_SYS_DIR "/sys/class/scsi_host/"

void log_message (int level, const char *format, ...);

#define dprintf(format, arg...) \
	log_message(LOG_DEBUG, "%s: " format, __FUNCTION__, ## arg)

unsigned char sysfs_mnt_path[SYSFS_PATH_MAX];

static inline char *sysfs_get_attr(struct sysfs_class_device *dev,
                                    const unsigned char *attr)
{
	return sysfs_get_value_from_attributes(dev->directory->attributes, 
					       attr);
}


void 
log_message(int level, const char *format, ...)
{
	static int logging_init;
	va_list	args;
	int debug = 1;

	if (!debug && level == LOG_DEBUG)
		return;

	va_start (args, format);
	if (!logging_init) {
		openlog ("iscsi-id", LOG_PID, LOG_DAEMON);
		logging_init = 1;
	}
	vsyslog(level, format, args);
	va_end (args);
	return;
}

static int 
get_major_minor(const unsigned char *devpath, int *major, int *minor)
{
	struct sysfs_class_device *class_dev;
	char dev_value[SYSFS_NAME_LEN];
	char *dev;
	char *temp;

	dprintf("%s\n", devpath);
	class_dev = sysfs_open_class_device(devpath);
	if (!class_dev) {
		temp = strrchr(devpath, '/');		
		*temp = 0x00;
		class_dev = sysfs_open_class_device(devpath);
		if (!class_dev) {
			log_message(LOG_WARNING, "open class %s failed: %s\n",
				    devpath, strerror(errno));
			return -1;
		}
	}

	dev = sysfs_get_attr(class_dev, "dev");
	if (dev)
		strncpy(dev_value, dev, SYSFS_NAME_LEN);

	sysfs_close_class_device(class_dev);

	if (!dev) {
		log_message(LOG_DEBUG, "%s could not get dev attribute: %s\n",
			    devpath, strerror(errno));
		return -1;
	}
	dev = NULL;

	dprintf("dev %s", dev_value); /* Already has a trailing \n */
	if (sscanf(dev_value, "%u:%u", major, minor) != 2) {
		log_message(LOG_WARNING, "%s: invalid dev major/minor\n",
			    devpath);
		return -1;
	}
	return 0;
}

static int 
create_tmp_dev(const char *devpath, char *tmpdev, int dev_type)
{
	int major, minor;

	dprintf("(%s)\n", devpath);
	if (get_major_minor(devpath, &major, &minor)) {
		dprintf(" Failed to create device %s\n", devpath);
		return -1;
	}
	snprintf(tmpdev, MAX_NAME_LEN, "%s/%s-maj%d-min%d-%u",
		 TMP_DIR, TMP_PREFIX, major, minor, getpid());

	dprintf("tmpdev '%s'\n", tmpdev);

	if (mknod(tmpdev, 0600 | dev_type, makedev(major, minor))) {
		log_message(LOG_WARNING, "mknod failed: %s\n", strerror(errno));
		return -1;
	}
	return 0;
}

static int 
has_sysfs_prefix(const char *path, const char *prefix)
{
	char match[MAX_NAME_LEN];

	strncpy(match, sysfs_mnt_path, MAX_NAME_LEN);
	strncat(match, prefix, MAX_NAME_LEN);
	if (strncmp(path, match, strlen(match)) == 0)
		return 1;
	else
		return 0;
}

/*
 * iscsi_id: try to get the c:b:t:l quadruple, if one is found, printf it to
 * stdout. Returns a value passed to exit() - 0 if printed the value, else 1.
 */
static int 
iscsi_id(const char *target_path, char *maj_min_dev)
{
	FILE *sys_hba;
	int retval;
	int dev_type;
	char devname[PATH_MAX + 1];
	char buffer[MAX_NAME_LEN];
	int host, channel, id, lun;
	int iscsi_channel;
	int iscsi_tid;
	int iscsi_lun;
	char *line;
	int fd;
	u_long var[2];

	dprintf("target_path %s\n", target_path);

	/*
	 * depend on the sysfs path to tell us whether this is a
	 * block or char device.
	 */

	if (has_sysfs_prefix(target_path, "/block")) {
		dev_type = S_IFBLK;
	} else if (has_sysfs_prefix(target_path, "/class")) {
		dev_type = S_IFCHR;
	} else {
		dprintf("Non block or class device\n");
		return 1;
	}

	/*
	 * mknod a temp dev to communicate with the device.
	 */
	if (create_tmp_dev(target_path, maj_min_dev, dev_type)) {
		dprintf("create_tmp_dev failed\n");
		return 1;
	}

	if ((fd = open(maj_min_dev, O_RDONLY)) >= 0) {
		memset(&var, 0, sizeof (var));
		/* try to get the host, channel, id, port quadruple */
		if (ioctl(fd, SCSI_IOCTL_GET_IDLUN, &var) == 0) {
			id = var[0] & 0xff;
			lun = (var[0] >> 8) & 0xff;
			channel = (var[0] >> 16) & 0xff;
			host = (var[0] >> 24) & 0xff;
		} else {
			fprintf(stderr, "iscsi-id: ioctl SCSI_IOCTL_GET_IDLUN"
			" %s failed, %s\n", maj_min_dev, strerror(errno));
			exit(1);
		}
	} else {
		fprintf(stderr, "iscsi-id: couldn't open %s, %s\n", maj_min_dev,
			strerror(errno));
		exit(1);
	}

	/* see if the device node's host (HBA number) is an iSCSI HBA */
	sprintf(devname, "%s/iscsi/print_info", ISCSI_SYS_DIR);

	if ((sys_hba = fopen(devname, "r"))) {
		/* read every line, ignoring comments and blank lines */
		while ((line = fgets(buffer, sizeof (buffer), sys_hba))) {
			while (isspace(*line))
				line++;
			if (*line == '#' || *line == '\0')
				continue;

			if (sscanf(line, "%d %d %d", &iscsi_channel, &iscsi_tid,
				&iscsi_lun) == 3) {
				/* For now we assume there is only one host,
				 * so that the bus number maps directly to the
				 * channel number.  This breaks if we ever
				 * have multiple HBAs.
				 */
				if (channel == iscsi_channel &&
					id == iscsi_tid && lun == iscsi_lun) {
					/* found it. Print the device
					 * and /sysfs line
					 */
					printf("iscsib%dt%dl%d", channel, id, 
								 lun);
					break;
				}
			}
		}
		fclose(sys_hba);

		/* if /sys/class/scsi_host/ has a file for the device node's
		 * host, then we know it's an iSCSI device, even if we can't
		 * get any additional info about it.
		 */
		retval = 0;
	} else 
		retval = 1;

	fflush(stdout);
	unlink(maj_min_dev);
	return retval;
}

int 
main(int argc, char **argv)
{
	int retval;
	char *devpath;
	char target_path[MAX_NAME_LEN];
	char maj_min_dev[MAX_NAME_LEN];

	if (sysfs_get_mnt_path(sysfs_mnt_path, MAX_NAME_LEN)) {
		log_message(LOG_WARNING, "sysfs_get_mnt_path failed: %s\n",
			strerror(errno));
		exit(1);
	}

	devpath = getenv("DEVPATH");
	if (!devpath) {
		log_message(LOG_WARNING, "DEVPATH is not set\n");
		exit(1);
	}

	strncpy(target_path, sysfs_mnt_path, MAX_NAME_LEN);
	strncat(target_path, devpath, MAX_NAME_LEN);
	retval = iscsi_id(target_path, maj_min_dev);
	exit(retval);
}
