/*
 * iSCSI driver for Linux
 * Copyright (C) 2001 Cisco Systems, Inc.
 * maintained by linux-iscsi-devel@lists.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi-initiator.c,v 1.1.2.25 2004/01/14 10:22:22 surekhap Exp $
 *
 */

#include <linux/version.h>
#include <linux/moduleparam.h>
#include <linux/blkdev.h>
#include <net/sock.h>
#include <linux/inet.h>

/* if non-zero, do a TCP Abort when a session drops, 
 * instead of (attempting) a graceful TCP Close 
 */
#define TCP_ABORT_ON_DROP 0

#define RETRIES_BLOCK_DEVICES 1

/* some targets, such as the Intel Labs target on SourceForge, make
 * invalid assumptions about the relateive ordering of command and
 * data PDUs, but still advertise a CmdSN window greater than one
 * command.  When this is non-zero, we try to break ourselves in such
 * a way that the target's bogus assumptions are met.  No promises
 * though, since we may send Nops or task mgmt PDUs at any time, 
 * which the broken target may still choke on.
 */
#define INVALID_ORDERING_ASSUMPTIONS 0

/* periodically stall reading data to test data 
 * arriving after aborts have started 
 */
#define TEST_DELAYED_DATA 0

/* fake sense indicating ILLEGAL_REQUEST for all REPORT_LUNS commands */
#define FAKE_NO_REPORT_LUNS 0

/* fake check conditions on the first 2 attempts for each probe command */
#define FAKE_PROBE_CHECK_CONDITIONS 0

/* fake underflows on the first 4 attempts for each probe command */
#define FAKE_PROBE_UNDERFLOW 0

#include "iscsi-protocol.h"
#include "iscsi-session.h"
#include "iscsi-io.h"
#include "iscsi-login.h"
#include "iscsi.h"
#include "iscsi-version.h"
#include "iscsi-probe.h"
#include "iscsi-crc.h"

/*
 *  IMPORTANT NOTE: to prevent deadlock, when holding multiple locks,
 *  the following locking order must be followed at all times:
 *
 *  session->portal_lock      - access to a session's portal info
 *  session->task_lock        - access to a session's collections of tasks
 *  hba->session_lock         - access to an HBA's collection of sessions   
 *  session->scsi_cmnd_lock   - access to a session's list of Scsi_Cmnds
 * 								(IRQSAVE)
 *  io_request_lock/host_lock - mid-layer acquires before calling queuecommand,  * 				eh_*, we must acquire before done() callback  
 *				(IRQSAVE)
 *  iscsi_trace_lock          - for the tracing code     (IRQSAVE)
 *
 *  Note: callers not in interrupt context must locally disable/restore 
 *  interrupts when holding locks marked (IRQSAVE)
 */

MODULE_AUTHOR("Cisco Systems, Inc.");
MODULE_DESCRIPTION("iSCSI initiator");
MODULE_LICENSE("GPL");

int iscsi_system_is_rebooting;
int this_is_iscsi_boot;

/* Force tagged command queueing for all devices, regardless of whether 
 * they say they support it 
 */
static int force_tcq;
module_param(force_tcq, int, 0);
MODULE_PARM_DESC(force_tcq,
		"when non-zero, force tagged command queueing for all devices");

/* Queue depth for devices that don't support tagged command queueing.
 * The driver used to use ISCSI_CMDS_PER_LUN, which was probably a bug.
 * Default to 1 now, but let people who want to the old behavior set it higher.
 */
static int untagged_queue_depth = 1;
module_param(untagged_queue_depth, int, 0);
MODULE_PARM_DESC(untagged_queue_depth,
		 "queue depth to use for devices that don't support "
		 "tagged command queueing");

int translate_deferred_sense = 1;
module_param(translate_deferred_sense, int, 0);
MODULE_PARM_DESC(translate_deferred_sense,
		 "translate deferred sense data to current sense data "
		 "in disk command responses");

static int control_major;
static const char *control_name = "iscsictl";

static struct file_operations control_fops = {
      .owner = THIS_MODULE,
      .ioctl = ctl_ioctl,	/* ioctl */
};

struct iscsi_hba *iscsi_hba;
extern struct class_device_attribute *iscsi_host_attrs[];
volatile unsigned long init_module_complete;
volatile unsigned long iscsi_timer_running;
volatile pid_t iscsi_timer_pid;

#define is_hex_lower(c) (((c) >= 'a') && ((c) <= 'f'))
#define is_hex_upper(c) (((c) >= 'A') && ((c) <= 'F'))

#if DEBUG_TRACE
spinlock_t iscsi_trace_lock = SPIN_LOCK_UNLOCKED;
static struct iscsi_trace_entry trace_table[ISCSI_TRACE_COUNT];
static int trace_index;
#endif

/* Become a daemon kernel thread.  
 */
void
iscsi_daemonize(char *daemon_name)
{
	struct task_struct *this_task = current;

	/* use the kernel's daemonize */
	daemonize("%s",daemon_name);

	/* increase priority like the md driver does for it's kernel threads */
	this_task->policy = SCHED_NORMAL;
	set_user_nice(this_task, -20);
	smp_mb();
}

void
lun_reset_occured(struct iscsi_session *session, unsigned int lun)
{
	Scsi_Device *device; 
	struct Scsi_Host *shost = session->hba->host;

	if (shost) {
		device = scsi_device_lookup(shost, session->channel, 
					    session->target_id, lun);
		device->was_reset = 1;
		device->expecting_cc_ua = 1;
		scsi_device_put(device);
	}
}

/* determine whether a command is eligible to be retried internally. */
int
internally_retryable(Scsi_Cmnd * sc)
{
	if (sc->device->type == TYPE_DISK) {
		switch (sc->cmnd[0]) {
		case INQUIRY:
		case REPORT_LUNS:
		case TEST_UNIT_READY:
		case READ_CAPACITY:
		case START_STOP:
		case MODE_SENSE:
			return 0;
		default:
			return 1;
		}
	}

	return 0;
}

static inline unsigned long *
device_flags(Scsi_Device * sd)
{
	unsigned long *dev_flags;

	dev_flags  = &((struct iscsi_dev_data *)(sd->hostdata))->flags;
	return dev_flags;
}

/* device flags */
#define DEVICE_LOG_TERMINATING          0
#define DEVICE_LOG_REPLACEMENT_TIMEDOUT 1
#define DEVICE_LOG_NO_SESSION           2

/* newer kernels require long alignment for bitops.
 * we assume pointers have at least as much alignment as longs,
 * and use an unused pointer field to store bit flags.
 */
unsigned long *
command_flags(Scsi_Cmnd * sc)
{
	unsigned long *flags = (unsigned long *) &(sc->SCp.buffer);

	return flags;
}


/* wake up the tx_thread without ever losing the wakeup event */
void
wake_tx_thread(int control_bit, struct iscsi_session *session)
{
	/* tell the tx thread what to do when it wakes up. */
	set_bit(control_bit, &session->control_bits);
	smp_wmb();

	/* We make a condition variable out of a wait queue and atomic 
	 * test&clear.
	 * May get spurious wake-ups, but no wakeups will be lost.
	 * this is cv_signal().  wait_event_interruptible is cv_wait().
	 */
	set_bit(TX_WAKE, &session->control_bits);
	smp_mb();

	wake_up(&session->tx_wait_q);
}


/* caller must hold session->task_lock */
void
iscsi_request_logout(struct iscsi_session *session, int logout,
		     int logout_response)
{
	if (atomic_read(&session->num_active_tasks) == 0) {
		DEBUG_INIT
		    ("iSCSI: session iscsi bus %d target id %d " 
		     "currently has no active tasks, queueing logout at %lu\n",
		     session->iscsi_bus, session->target_id, jiffies);
		session->logout_response_deadline =
		    jiffies + (logout_response * HZ);
		if (session->logout_response_deadline == 0)
			session->logout_response_deadline = 1;
		smp_mb();
		set_bit(SESSION_LOGOUT_REQUESTED, &session->control_bits);
		smp_mb();
		wake_tx_thread(TX_LOGOUT, session);
	} else {
		session->logout_deadline = jiffies + (logout * HZ);
		if (session->logout_deadline == 0)
			session->logout_deadline = 1;
		session->logout_response_deadline =
		    session->logout_deadline + (logout_response * HZ);
		if (session->logout_response_deadline == 0)
			session->logout_response_deadline = 1;
		smp_mb();
		set_bit(SESSION_LOGOUT_REQUESTED, &session->control_bits);
		smp_mb();
	}
}

/* if a signal is pending, deal with it, and return 1.
 * Otherwise, return 0.
 */
int
iscsi_handle_signals(struct iscsi_session *session)
{
	pid_t pid;
	int ret = 0;

	/* if we got SIGHUP, try to establish a replacement session.
	 * if we got SIGKILL, terminate this session.
	 */
	if (signal_pending(current)) {
		spin_lock_irq(&current->sighand->siglock);

		/* iscsi_drop_session and iscsi_terminate_session signal both
		 * threads, but someone logged in as root may not.  So, we
		 * make sure whichever process gets signalled first propagates
		 * the signal when it looks like only one thread got
		 * signalled.
		 */

		/* on SIGKILL, terminate the session */
		if (sigismember(&current->pending.signal, SIGKILL)) {

			/*
			 * FIXME: We don't terminate the sessions if 
			 * "/" is iSCSI disk
			 * Need to fix this for other iSCSI targets.
			 */

			if (!this_is_iscsi_boot
			    || iscsi_system_is_rebooting) {
				if (!test_and_set_bit
				    (SESSION_TERMINATING,
				     &session->control_bits)) {
					if ((pid = session->tx_pid)
					    && (pid != current->pid)) {
						printk
						    ("iSCSI: rx thread %d "
						     "received SIGKILL, killing"
						     " tx thread %d\n",
						     current->pid, pid);
						kill_proc(pid, SIGKILL, 1);
					}
					if ((pid = session->rx_pid)
					    && (pid != current->pid)) {
						printk
						    ("iSCSI: tx thread %d "
						     "received SIGKILL, killing"
						     " rx thread %d\n",
						     current->pid, pid);
						kill_proc(pid, SIGKILL, 1);
					}
				}
				ret = 1;
			}
		}
		/* on SIGHUP, drop the session, and try to establish 
		 * a replacement session 
		 */
		if (sigismember(&current->pending.signal, SIGHUP)) {
			if (test_and_clear_bit
			    (SESSION_ESTABLISHED, &session->control_bits)) {
				if ((pid = session->tx_pid)
				    && (pid != current->pid)) {
					printk
					    ("iSCSI: rx thread %d received "
					     "SIGHUP, signaling tx thread %d\n",
					     current->pid, pid);
					kill_proc(pid, SIGHUP, 1);
				}
				if ((pid = session->rx_pid)
				    && (pid != current->pid)) {
					printk
					    ("iSCSI: tx thread %d received "
					     "SIGHUP, signaling rx thread %d\n",
					     current->pid, pid);
					kill_proc(pid, SIGHUP, 1);
				}
			}
			ret = 1;
		}
		/* we don't care about any other signals */
		spin_unlock_irq(&current->sighand->siglock);
		flush_signals(current);
	}
	if (ret) {
		session->session_drop_time = jiffies ? jiffies : 1;
		session->session_alive = 0;
	}

	return ret;
}

unsigned int
iscsi_expected_data_length(Scsi_Cmnd * sc)
{
	unsigned int length = 0;

	if (sc == NULL)
		return 0;

	switch (sc->cmnd[0]) {
	case INQUIRY:
	case REQUEST_SENSE:
		length = sc->cmnd[4];
		return length;
	case REPORT_LUNS:
		length |= sc->cmnd[6] << 24;
		length |= sc->cmnd[7] << 16;
		length |= sc->cmnd[8] << 8;
		length |= sc->cmnd[9];
		return length;
	default:
		return sc->request_bufflen;
	}
}

/* Serial Number Arithmetic, 32 bits, less than, RFC1982 */
int
sna_lt(uint32_t n1, uint32_t n2)
{
	return ((n1 != n2) && (((n1 < n2) && ((n2 - n1) < SNA32_CHECK))
			       || ((n1 > n2) && ((n2 - n1) < SNA32_CHECK))));
}

/* Serial Number Arithmetic, 32 bits, less than, RFC1982 */
int
sna_lte(uint32_t n1, uint32_t n2)
{
	return ((n1 == n2) || (((n1 < n2) && ((n2 - n1) < SNA32_CHECK))
			       || ((n1 > n2) && ((n2 - n1) < SNA32_CHECK))));
}

static inline void
set_not_ready(Scsi_Cmnd * sc)
{
	sc->sense_buffer[0] = 0x70;
	sc->sense_buffer[2] = NOT_READY;
	sc->sense_buffer[7] = 0x0;
}

/* mark a Scsi_Cmnd as having a LUN communication failure */
static inline void
set_lun_comm_failure(Scsi_Cmnd * sc)
{
	sc->sense_buffer[0] = 0x70;
	sc->sense_buffer[2] = NOT_READY;
	sc->sense_buffer[7] = 0x6;
	sc->sense_buffer[12] = 0x08;
	sc->sense_buffer[13] = 0x00;
}

/* decode common network errno values into more useful strings.
 * strerror would be nice right about now.
 */
char *
iscsi_strerror(int errno)
{
	switch (errno) {
	case EIO:
		return "I/O error";
	case EINTR:
		return "Interrupted system call";
	case ENXIO:
		return "No such device or address";
	case EFAULT:
		return "Bad address";
	case EBUSY:
		return "Device or resource busy";
	case EINVAL:
		return "Invalid argument";
	case EPIPE:
		return "Broken pipe";
	case ENONET:
		return "Machine is not on the network";
	case ECOMM:
		return "Communication error on send";
	case EPROTO:
		return "Protocol error";
	case ENOTUNIQ:
		return "Name not unique on network";
	case ENOTSOCK:
		return "Socket operation on non-socket";
	case ENETDOWN:
		return "Network is down";
	case ENETUNREACH:
		return "Network is unreachable";
	case ENETRESET:
		return "Network dropped connection because of reset";
	case ECONNABORTED:
		return "Software caused connection abort";
	case ECONNRESET:
		return "Connection reset by peer";
	case ESHUTDOWN:
		return "Cannot send after shutdown";
	case ETIMEDOUT:
		return "Connection timed out";
	case ECONNREFUSED:
		return "Connection refused";
	case EHOSTDOWN:
		return "Host is down";
	case EHOSTUNREACH:
		return "No route to host";
	default:
		return "";
	}
}

#if DEBUG_TRACE
void
iscsi_fill_trace(unsigned char type, Scsi_Cmnd * sc, struct iscsi_task *task,
		 unsigned long data1, unsigned long data2)
{
	struct iscsi_trace_entry *te;
	unsigned long flags;

	spin_lock_irqsave(&iscsi_trace_lock, flags);

	te = &trace_table[trace_index];
	trace_index++;
	if (trace_index >= ISCSI_TRACE_COUNT) {
		trace_index = 0;
	}
	memset(te, 0x0, sizeof (*te));

	te->type = type;
	if (task) {
		struct iscsi_session *session = task->session;

		te->host = session->host_no;
		te->channel = session->channel;
		te->target = session->target_id;
		te->lun = task->lun;
		te->itt = task->itt;
	}
	if (sc) {
		te->cmd = sc->cmnd[0];
		te->host = sc->device->host->host_no;
		te->channel = sc->device->channel;
		te->target = sc->device->id;
		te->lun = sc->device->lun;
	}
	te->data1 = data1;
	te->data2 = data2;
	te->jiffies = jiffies;

	spin_unlock_irqrestore(&iscsi_trace_lock, flags);
}
#endif

/* FIXME: update for 16 byte CDBs, such as:
   lock unlock cache 16
   pre-fetch 16
   read 16
   rebuild 16
   regenerate 16
   synchronize cache 16
   verify 16
   write 16
   write and verify 16
   write same 16
   xdwrite extended 16

   Then increase ISCSI_MAX_CMD_LEN to 16 in iscsi.h.
*/
/* FIXME: for that matter, check the existing list for correctness */
/* caller must either hold the task, or keep the task refcount non-zero while 
 * calling this 
 */

/* tagged queueing */
unsigned int
iscsi_command_attr(Scsi_Cmnd * cmd)
{
	if (cmd->device && (cmd->device->simple_tags || 
	    cmd->device->ordered_tags)) {
		switch (cmd->tag) {
		case HEAD_OF_QUEUE_TAG:
			return ISCSI_ATTR_HEAD_OF_QUEUE;
		case ORDERED_QUEUE_TAG:
			return ISCSI_ATTR_ORDERED;
		default:
			return ISCSI_ATTR_SIMPLE;
		}
	}

	return ISCSI_ATTR_UNTAGGED;
}

void
print_cmnd(Scsi_Cmnd * sc)
{
#ifdef HAS_CMND_REQUEST_STRUCT
	struct request *req = &sc->request;
	struct buffer_head *bh = NULL;
#endif

	printk("iSCSI: Scsi_Cmnd to (%u %u %u %u), Cmd 0x%x\n"
	       "iSCSI:   done %p, scsi_done %p, host_scribble %p\n"
	       "iSCSI:   reqbuf %p, req_len %u\n"
	       "iSCSI:   buffer %p, bufflen %u\n"
	       "iSCSI:   use_sg %u, old_use_sg %u, sglist_len %u\n"
	       "iSCSI:   owner 0x%x, state  0x%x, eh_state 0x%x\n"
	       "iSCSI:   cmd_len %u, old_cmd_len %u\n",
	       sc->device->host->host_no, sc->device->channel, 
	       sc->device->id, sc->device->lun, sc->cmnd[0], sc->done,
	       sc->scsi_done, sc->host_scribble, sc->request_buffer,
	       sc->request_bufflen, sc->buffer, sc->bufflen, sc->use_sg, 
               sc->old_use_sg, sc->sglist_len, sc->owner, sc->state,
	       sc->eh_state, sc->cmd_len, sc->old_cmd_len);

	if (sc->cmd_len >= 12)
		printk
		    ("iSCSI:   cdb %02x%02x%02x%02x %02x%02x%02x%02x "
		     "%02x%02x%02x%02x\n",
		     sc->cmnd[0], sc->cmnd[1], sc->cmnd[2], sc->cmnd[3],
		     sc->cmnd[4], sc->cmnd[5], sc->cmnd[6], sc->cmnd[7],
		     sc->cmnd[8], sc->cmnd[9], sc->cmnd[10], sc->cmnd[11]);
	else if (sc->cmd_len >= 10)
		printk
		    ("iSCSI:   cdb %02x%02x%02x%02x %02x%02x%02x%02x "
		     "%02x%02x\n",
		     sc->cmnd[0], sc->cmnd[1], sc->cmnd[2], sc->cmnd[3],
		     sc->cmnd[4], sc->cmnd[5], sc->cmnd[6], sc->cmnd[7],
		     sc->cmnd[8], sc->cmnd[9]);
	else if (sc->cmd_len >= 8)
		printk("iSCSI:   cdb %02x%02x%02x%02x %02x%02x%02x%02x\n",
		       sc->cmnd[0], sc->cmnd[1], sc->cmnd[2], sc->cmnd[3],
		       sc->cmnd[4], sc->cmnd[5], sc->cmnd[6], sc->cmnd[7]);
	else if (sc->cmd_len >= 6)
		printk("iSCSI:   cdb %02x%02x%02x%02x %02x%02x\n",
		       sc->cmnd[0], sc->cmnd[1], sc->cmnd[2], sc->cmnd[3],
		       sc->cmnd[4], sc->cmnd[5]);
	else if (sc->cmd_len >= 4)
		printk("iSCSI:   cdb %02x%02x%02x%02x\n",
		       sc->cmnd[0], sc->cmnd[1], sc->cmnd[2], sc->cmnd[3]);
	else if (sc->cmd_len >= 2)
		printk("iSCSI:   cdb %02x%02x\n", sc->cmnd[0], sc->cmnd[1]);

	if (sc->use_sg && sc->request_buffer) {
		struct scatterlist *sglist =
		    (struct scatterlist *) sc->request_buffer;
		int i;

		for (i = 0; i < sc->use_sg; i++) {
			printk
			    ("iSCSI:   sglist %p index %02d = page %p, "
			     "offset %u, len %u\n",
			     sglist, i, sglist->page, sglist->offset,
			     sglist->length);
			sglist++;
		}
	}
#ifdef HAS_CMND_REQUEST_STRUCT
	/* and log the struct request so we can check consistency */
	printk
	    ("iSCSI:   request status 0x%x, sector %lu, nr_sectors %lu, "
	     "hard_sector %lu, hard_nr_sectors %lu\n"
	     "iSCSI:   nr_segments %u, hard_nr_segments %u, "
	     "current_nr_sectors %lu\n"
	     "iSCSI:   special %p, buffer %p, bh %p, bhtail %p\n",
	     req->rq_status, req->sector, req->nr_sectors, req->hard_sector,
	     req->hard_nr_sectors, req->nr_segments, req->nr_hw_segments,
	     req->current_nr_sectors, req->special, req->buffer, req->bh,
	     req->bhtail);

	for (bh = req->bh; bh; bh = bh->b_reqnext) {
		printk
		    ("iSCSI:   bh %p = rsector %lu, blocknr %lu, size %u, "
		     "list %u, state 0x%lx, data %p, page %p\n",
		     bh, bh->b_rsector, bh->b_blocknr, bh->b_size, bh->b_list,
		     bh->b_state, bh->b_data, bh->b_page);
	}
#endif

	/* and log the scsi_request so we can check consistency */
	if (sc->sc_request) {
		printk
		    ("iSCSI:   Scsi_Request %p = sr_magic 0x%x, sr_bufflen %u,"
		     " sr_buffer %p, sr_allowed %u, sr_cmd_len %u\n"
		     "iSCSI:                     sr_use_sg %u, sr_sglist_len "
		     "%u, sr_underflow %u\n",
		     sc->sc_request, sc->sc_request->sr_magic,
		     sc->sc_request->sr_bufflen, sc->sc_request->sr_buffer,
		     sc->sc_request->sr_allowed, sc->sc_request->sr_cmd_len,
		     sc->sc_request->sr_use_sg, sc->sc_request->sr_sglist_len,
		     sc->sc_request->sr_underflow);

	}
}

int
add_cmnd(Scsi_Cmnd * sc, Scsi_Cmnd ** head, Scsi_Cmnd ** tail)
{
	sc->host_scribble = NULL;

	if (*head) {
		(*tail)->host_scribble = (void *) sc;
		*tail = sc;
	} else {
		*tail = *head = sc;
	}

	return 1;
}

void
request_command_retries(unsigned long arg)
{
	struct iscsi_session *session = (struct iscsi_session *) arg;

	DEBUG_RETRY
	    ("iSCSI: session iscsi bus %d target id %d retry timer "
	     "expired at %lu\n",
	     session->iscsi_bus, session->target_id, jiffies);
	session->retry_timer.expires = 0;
	smp_mb();
	wake_tx_thread(SESSION_RETRY_COMMANDS, session);
}

/* try to queue one command retry for each LUN that needs one */
void
iscsi_retry_commands(struct iscsi_session *session)
{
	Scsi_Cmnd *prior = NULL, *sc;
	Scsi_Cmnd *retry_head = NULL, *retry_tail = NULL;
	struct iscsi_task *task;
	int num_retries = 0;
	int l;
	unsigned long noqueue_flags_;

	spin_lock(&session->task_lock);
	spin_lock_irqsave(&session->scsi_cmnd_lock, noqueue_flags_);

	/* record which LUNs we're going to check for retries */
	memset(session->luns_checked, 0, sizeof (session->luns_checked));
	for (l = 0; l < ISCSI_MAX_LUN; l++) {
		if (test_bit(l, session->luns_delaying_commands)
		    && !test_bit(l, session->luns_timing_out))
			__set_bit(l, session->luns_checked);
	}

	/* skip LUNs that already have outstanding tasks */
	list_for_each_entry(task, &session->task_arrival_order,
			    task_order_link) {
		__clear_bit(task->lun, session->luns_checked);
	}

	/* skip LUNs that already have a retry queued */
	for (sc = session->retry_cmnd_head; sc;
	     sc = (Scsi_Cmnd *) sc->host_scribble) {
		__clear_bit(sc->device->lun, session->luns_checked);
	}

	/* find the oldest deferred command to each of the LUNs 
	 * we want to queue a retry to 
	 */
	while ((sc = session->deferred_cmnd_head)) {
		if (test_bit(sc->device->lun, session->luns_checked)) {
			/* pop this command off the head of 
			 * the deferred queue 
			 */
			session->deferred_cmnd_head =
			    (Scsi_Cmnd *) sc->host_scribble;
			if (session->deferred_cmnd_head == NULL)
				session->deferred_cmnd_tail = NULL;
			session->num_deferred_cmnds--;

			/* queue it for retry */
			if (retry_head) {
				retry_tail->host_scribble = (unsigned char *)sc;
				retry_tail = sc;
			}
			else {
				retry_head = retry_tail = sc;
			}

			sc->host_scribble = NULL;
			num_retries++;
			if (LOG_ENABLED(ISCSI_LOG_RETRY))
				printk
				    ("iSCSI: session iscsi bus %d queuing "
				     "command cdb 0x%02x to (%u %u %u %u) for "
				     "retry at %lu\n",
				     session->iscsi_bus, sc->cmnd[0],
				     session->host_no, session->channel,
				     session->target_id, sc->device->lun,
			  	     jiffies);

			/* and don't take any more commands for this LUN */
			__clear_bit(sc->device->lun, session->luns_checked);
		} else {
			prior = sc;
			break;
		}
	}
	while (prior && (sc = (Scsi_Cmnd *) prior->host_scribble)) {
		if (test_bit(sc->device->lun, session->luns_checked)) {
			/* remove this command from the deferred queue */
			prior->host_scribble = sc->host_scribble;
			if (session->deferred_cmnd_tail == sc)
				session->deferred_cmnd_tail = prior;
			session->num_deferred_cmnds--;

			/* queue it for retry */
			if (retry_head) {
				retry_tail->host_scribble = (unsigned char *)sc;
				retry_tail = sc;
			} else {
				retry_head = retry_tail = sc;
			}
			sc->host_scribble = NULL;
			num_retries++;
			if (LOG_ENABLED(ISCSI_LOG_RETRY))
				printk
				    ("iSCSI: session iscsi bus %d queuing "
				     "command cdb 0x%02x to (%u %u %u %u) for "
				     "retry at %lu\n",
				     session->iscsi_bus, sc->cmnd[0],
				     session->host_no, session->channel,
				     session->target_id, sc->device->lun, jiffies);

			/* and don't take any more commands for this LUN */
			__clear_bit(sc->device->lun, session->luns_checked);
		} else {
			prior = sc;
		}
	}

	if (num_retries) {
		/* append to the retry_cmnd queue */
		if (session->retry_cmnd_head)
			session->retry_cmnd_tail->host_scribble =
			    (void *) retry_head;
		else
			session->retry_cmnd_head = retry_head;

		session->retry_cmnd_tail = retry_tail;
		atomic_add(num_retries, &session->num_retry_cmnds);
		set_bit(TX_WAKE, &session->control_bits);
		set_bit(TX_SCSI_COMMAND, &session->control_bits);
	}

	spin_unlock_irqrestore(&session->scsi_cmnd_lock, noqueue_flags_);

	if (session->num_luns_delaying_commands
	    && (session->retry_timer.expires == 0)) {
		/* as long as at least one LUN is delaying commands, 
		 * we need to reset the timer 
		 */
		session->retry_timer.function = request_command_retries;
		session->retry_timer.data = (unsigned long) session;
		session->retry_timer.expires = jiffies + HZ;
		add_timer(&session->retry_timer);
	}

	spin_unlock(&session->task_lock);
}

void
requeue_deferred_commands(struct iscsi_session *session, unsigned int lun)
{
	Scsi_Cmnd *cmnd, *prior, *requeue_head = NULL, *requeue_tail = NULL;
	int num_requeued = 0;

	DEBUG_RETRY
	    ("iSCSI: session iscsi bus %d requeuing deferred commands for "
	     "(%u %u %u %u) at %lu\n",
	     session->iscsi_bus, session->host_no, session->channel,
	     session->target_id, lun, jiffies);

	prior = NULL;
	while ((cmnd = session->deferred_cmnd_head)) {
		if (cmnd->device->lun == lun) {
			/* remove it from the deferred queue */
			session->deferred_cmnd_head =
			    (Scsi_Cmnd *) cmnd->host_scribble;
			if (session->deferred_cmnd_head == NULL)
				session->deferred_cmnd_tail = NULL;
			session->num_deferred_cmnds--;
			cmnd->host_scribble = NULL;

			DEBUG_RETRY
			    ("iSCSI: session iscsi bus %d requeueing deferred "
			     "command cdb 0x%02x to (%u %u %u %u) at %lu\n",
			     session->iscsi_bus, cmnd->cmnd[0],
			     session->host_no, session->channel,
			     session->target_id, cmnd->device->lun, jiffies);
			add_cmnd(cmnd, &requeue_head, &requeue_tail);
			num_requeued++;
		} else {
			prior = cmnd;
			break;
		}
	}
	while (prior && (cmnd = (Scsi_Cmnd *) prior->host_scribble)) {
		if (cmnd->device->lun == lun) {
			/* remove it from the deferred queue */
			prior->host_scribble = cmnd->host_scribble;
			if (session->deferred_cmnd_tail == cmnd)
				session->deferred_cmnd_tail = prior;
			session->num_deferred_cmnds--;
			cmnd->host_scribble = NULL;

			DEBUG_RETRY
			    ("iSCSI: session iscsi bus %d requeueing deferred "
			     "command cdb 0x%02x to (%u %u %u %u) at %lu\n",
			     session->iscsi_bus, cmnd->cmnd[0],
			     session->host_no, session->channel,
			     session->target_id, cmnd->device->lun, jiffies);
			add_cmnd(cmnd, &requeue_head, &requeue_tail);
			num_requeued++;
		} else {
			prior = cmnd;
		}
	}

	if (requeue_head) {
		requeue_tail->host_scribble = (void *) session->scsi_cmnd_head;
		session->scsi_cmnd_head = requeue_head;
		if (session->scsi_cmnd_tail == NULL)
			session->scsi_cmnd_tail = requeue_tail;
		atomic_add(num_requeued, &session->num_cmnds);
		wake_tx_thread(TX_SCSI_COMMAND, session);
		DEBUG_RETRY
		    ("iSCSI: session iscsi bus %d target id %d requeued %d "
		     "deferred commands and woke tx thread at %lu\n",
		     session->iscsi_bus, session->target_id, num_requeued,
		     jiffies);
	}
}

static void
recheck_busy_commands(unsigned long arg)
{
	struct iscsi_session *session = (struct iscsi_session *) arg;

	session->busy_command_timer.expires = 0;
	smp_mb();
	wake_tx_thread(SESSION_COMMAND_TIMEDOUT, session);
}

int
process_timedout_commands(struct iscsi_session *session)
{
	struct iscsi_task *task;
	Scsi_Cmnd *fatal_head = NULL, *fatal_tail = NULL, *cmnd, *prior;
	int busy = 0;
	unsigned long noqueue_flags_;

	spin_lock(&session->task_lock);
	spin_lock_irqsave(&session->scsi_cmnd_lock, noqueue_flags_);

	DEBUG_TIMEOUT
	    ("iSCSI: session iscsi bus %d target id %d "
	     "processing timedout commands at %lu\n",
	     session->iscsi_bus, session->target_id, jiffies);
	clear_bit(SESSION_COMMAND_TIMEDOUT, &session->control_bits);

	/* by default, we can fail commands to any LUN */
	memset(session->luns_checked, 0xFF, sizeof (session->luns_checked));

	DEBUG_TIMEOUT
	    ("iSCSI: session iscsi bus %d target id %d checking %d "
	     "tasks for command timeouts at %lu\n",
	     session->iscsi_bus, session->target_id,
	     atomic_read(&session->num_active_tasks), jiffies);
	list_for_each_entry(task, &session->task_arrival_order,
			    task_order_link) {
		if (task->scsi_cmnd
		    && test_bit(COMMAND_TIMEDOUT,
				command_flags(task->scsi_cmnd))) {
			if (atomic_read(&task->refcount) == 0) {
				cmnd = task->scsi_cmnd;
				task->scsi_cmnd = NULL;

				if (LOG_ENABLED(ISCSI_LOG_TIMEOUT))
					printk
					    ("iSCSI: session iscsi bus %d "
					     "failing itt %u cdb 0x%02x to "
					     "(%u %u %u %u) at %lu, retries "
					     "%d, allowed %d\n",
					     session->iscsi_bus, task->itt,
					     cmnd->cmnd[0],
					     cmnd->device->host->host_no,
					     cmnd->device->channel,
					     cmnd->device->id,
					     cmnd->device->lun, jiffies,
					     cmnd->retries, cmnd->allowed);
				add_cmnd(cmnd, &fatal_head, &fatal_tail);
			} else {
				/* can't fail this command now, something may 
				 * be using it's buffers.
				 * Delay failing this command and all younger 
				 * commands to this LUN.
				 */
				DEBUG_TIMEOUT
				    ("iSCSI: session iscsi bus %d target id %d "
				     "itt %u cmnd 0x%x is timedout but busy "
				     "at %lu\n",
				     session->iscsi_bus, session->target_id,
				     task->itt, task->scsi_cmnd->cmnd[0],
				     jiffies);
				__clear_bit(task->lun, session->luns_checked);
				busy = 1;
				break;
			}
		}

	}

	if (busy) {
		/* schedule another scan in the near future */
		if ((session->busy_command_timer.expires == 0)
		    && !test_bit(SESSION_TERMINATING, &session->control_bits)) {
			session->busy_command_timer.expires =
			    jiffies + MSECS_TO_JIFFIES(40);
			session->busy_command_timer.data =
			    (unsigned long) session;
			session->busy_command_timer.function =
			    recheck_busy_commands;
			DEBUG_TIMEOUT
			    ("iSCSI: session iscsi bus %d target id %d "
			     "scheduling busy command scan for %lu at %lu\n",
			     session->iscsi_bus, session->target_id,
			     session->busy_command_timer.expires, jiffies);
			del_timer_sync(&session->busy_command_timer);	
					/* make sure it's not running now */
			add_timer(&session->busy_command_timer);
		}
	}

	/* if any commands in the retry queue have TIMEDOUT, 
	 * dequeue and fail them. 
	 */
	DEBUG_TIMEOUT
	    ("iSCSI: session iscsi bus %d target id %d checking %d "
	     "retry commands for timeouts at %lu\n",
	     session->iscsi_bus, session->target_id,
	     atomic_read(&session->num_retry_cmnds), jiffies);
	prior = NULL;
	while ((cmnd = session->retry_cmnd_head)) {
		if (test_bit(COMMAND_TIMEDOUT, command_flags(cmnd))) {
			/* remove it from the deferred queue */
			session->retry_cmnd_head =
			    (Scsi_Cmnd *) cmnd->host_scribble;
			if (session->retry_cmnd_head == NULL)
				session->retry_cmnd_tail = NULL;
			atomic_dec(&session->num_retry_cmnds);
			cmnd->host_scribble = NULL;

			if (LOG_ENABLED(ISCSI_LOG_TIMEOUT))
				printk
				    ("iSCSI: session iscsi bus %d failing "
				     "retryable command cdb 0x%02x to "
				     "(%u %u %u %u) at %lu, retries %d, "
				     "allowed %d\n",
				     session->iscsi_bus, cmnd->cmnd[0],
				     cmnd->device->host->host_no, 
				     cmnd->device->channel, cmnd->device->id, 
				     cmnd->device->lun, jiffies,
				     cmnd->retries, cmnd->allowed);
			add_cmnd(cmnd, &fatal_head, &fatal_tail);
		} else {
			prior = cmnd;
			break;
		}
	}
	while (prior && (cmnd = (Scsi_Cmnd *) prior->host_scribble)) {
		if (test_bit(COMMAND_TIMEDOUT, command_flags(cmnd))) {
			/* remove it from the deferred queue */
			prior->host_scribble = cmnd->host_scribble;
			if (session->retry_cmnd_tail == cmnd)
				session->retry_cmnd_tail = prior;
			atomic_dec(&session->num_retry_cmnds);
			cmnd->host_scribble = NULL;

			if (LOG_ENABLED(ISCSI_LOG_TIMEOUT))
				printk
				    ("iSCSI: session iscsi bus %d failing "
				     "retryable command cdb 0x%02x to "
				     "(%u %u %u %u) at %lu, retries %d, "
				     "allowed %d\n",
				     session->iscsi_bus, cmnd->cmnd[0],
				     cmnd->device->host->host_no, 
				     cmnd->device->channel, cmnd->device->id,
				     cmnd->device->lun, jiffies,
				     cmnd->retries, cmnd->allowed);
			add_cmnd(cmnd, &fatal_head, &fatal_tail);
		} else {
			prior = cmnd;
		}
	}

	/* if any commands in the deferred queue have TIMEDOUT, 
	 * dequeue and fail them 
	 */
	DEBUG_TIMEOUT
	    ("iSCSI: session iscsi bus %d target id %d checking %d "
	     "deferred commands for timeouts at %lu\n",
	     session->iscsi_bus, session->target_id,
	     session->num_deferred_cmnds, jiffies);
	prior = NULL;
	while ((cmnd = session->deferred_cmnd_head)) {
		if (test_bit(COMMAND_TIMEDOUT, command_flags(cmnd))) {
			/* remove it from the deferred queue */
			session->deferred_cmnd_head =
			    (Scsi_Cmnd *) cmnd->host_scribble;
			if (session->deferred_cmnd_head == NULL)
				session->deferred_cmnd_tail = NULL;
			session->num_deferred_cmnds--;
			cmnd->host_scribble = NULL;

			if (LOG_ENABLED(ISCSI_LOG_TIMEOUT))
				printk
				    ("iSCSI: session iscsi bus %d failing "
				     "deferred command cdb 0x%02x to "
				     "(%u %u %u %u) at %lu, retries %d, "
				     "allowed %d\n",
				     session->iscsi_bus, cmnd->cmnd[0],
				     cmnd->device->host->host_no,
				     cmnd->device->channel, cmnd->device->id,
				     cmnd->device->lun, jiffies,
				     cmnd->retries, cmnd->allowed);
			add_cmnd(cmnd, &fatal_head, &fatal_tail);
		} else {
			prior = cmnd;
			break;
		}
	}
	while (prior && (cmnd = (Scsi_Cmnd *) prior->host_scribble)) {
		if (test_bit(COMMAND_TIMEDOUT, command_flags(cmnd))) {
			/* remove it from the deferred queue */
			prior->host_scribble = cmnd->host_scribble;
			if (session->deferred_cmnd_tail == cmnd)
				session->deferred_cmnd_tail = prior;
			session->num_deferred_cmnds--;
			cmnd->host_scribble = NULL;

			if (LOG_ENABLED(ISCSI_LOG_TIMEOUT))
				printk
				    ("iSCSI: session iscsi bus %d failing "
				     "deferred command cdb 0x%02x to "
				     "(%u %u %u %u) at %lu, retries %d, "
				     "allowed %d\n",
				     session->iscsi_bus, cmnd->cmnd[0],
				     cmnd->device->host->host_no, 
				     cmnd->device->channel, cmnd->device->id,
				     cmnd->device->lun, jiffies,
				     cmnd->retries, cmnd->allowed);
			add_cmnd(cmnd, &fatal_head, &fatal_tail);
		} else {
			prior = cmnd;
		}
	}

	/* if any commands in the normal queue have TIMEDOUT, 
	 * dequeue and fail them 
	 */
	DEBUG_TIMEOUT
	    ("iSCSI: session iscsi bus %d target id %d checking %d "
	     "normal commands for timeouts at %lu\n",
	     session->iscsi_bus, session->target_id,
	     atomic_read(&session->num_cmnds), jiffies);
	prior = NULL;
	while ((cmnd = session->scsi_cmnd_head)) {
		if (test_bit(COMMAND_TIMEDOUT, command_flags(cmnd))) {
			/* remove it from the scsi_cmnd queue */
			session->scsi_cmnd_head =
			    (Scsi_Cmnd *) cmnd->host_scribble;
			if (session->scsi_cmnd_head == NULL)
				session->scsi_cmnd_tail = NULL;
			atomic_dec(&session->num_cmnds);
			cmnd->host_scribble = NULL;

			if (LOG_ENABLED(ISCSI_LOG_TIMEOUT))
				printk
				    ("iSCSI: session iscsi bus %d failing "
				     "normal command cdb 0x%02x to "
				     "(%u %u %u %u) at %lu, retries %d, "
				     "allowed %d\n",
				     session->iscsi_bus, cmnd->cmnd[0],
				     cmnd->device->host->host_no,
				     cmnd->device->channel,
				     cmnd->device->id, cmnd->device->lun,
				     jiffies, cmnd->retries, cmnd->allowed);

			/* and arrange for it to be completed with 
			 * a fatal error 
			 */
			add_cmnd(cmnd, &fatal_head, &fatal_tail);
		} else {
			prior = cmnd;
			break;
		}
	}
	while (prior && (cmnd = (Scsi_Cmnd *) prior->host_scribble)) {
		if (test_bit(COMMAND_TIMEDOUT, command_flags(cmnd)) == 0) {
			/* remove it from the scsi_cmnd queue */
			prior->host_scribble = cmnd->host_scribble;
			if (session->scsi_cmnd_tail == cmnd)
				session->scsi_cmnd_tail = prior;
			atomic_dec(&session->num_cmnds);
			cmnd->host_scribble = NULL;

			if (LOG_ENABLED(ISCSI_LOG_TIMEOUT))
				printk
				    ("iSCSI: session iscsi bus %d failing "
				     "normal command cdb 0x%02x to "
				     "(%u %u %u %u) at %lu, retries %d, "
				     "allowed %d\n",
				     session->iscsi_bus, cmnd->cmnd[0],
				     cmnd->device->host->host_no,
				     cmnd->device->channel,
				     cmnd->device->id, cmnd->device->lun,
				     jiffies, cmnd->retries, cmnd->allowed);

			/* and arrange for it to be completed 
			 * with a fatal error 
			 */
			add_cmnd(cmnd, &fatal_head, &fatal_tail);
		} else {
			prior = cmnd;
		}
	}

	spin_unlock_irqrestore(&session->scsi_cmnd_lock, noqueue_flags_);

	/* if we have commands to fail back to the high-level 
	 * driver with a fatal error, do so now 
	 */
	if (fatal_head) {
		unsigned long midlayer_flags_;

		DEBUG_TIMEOUT
		    ("iSCSI: session iscsi bus %d target id %d "
		     "completing timedout commands at %lu\n",
		     session->iscsi_bus, session->target_id, jiffies);

		spin_lock_irqsave(session->hba->host->host_lock, 
				  midlayer_flags_);
		while ((cmnd = fatal_head)) {
			fatal_head = (Scsi_Cmnd *) cmnd->host_scribble;

			cmnd->result = host_byte(DID_NO_CONNECT);
			cmnd->resid = iscsi_expected_data_length(cmnd);
			if (cmnd->allowed > 1)	/* we've exhausted all 
						 * retries 
						 */
				cmnd->retries = cmnd->allowed;

			set_not_ready(cmnd);	/* fail the whole command now, 
						 * rather than just 1 buffer 
						 * head 
						 */

			/* FIXME: if it's a disk write, take the device 
			 * offline?  We don't want the buffer cache data loss 
			 * to occur silently, but offlining the device will 
			 * break multipath drivers, and cause problems
			 * for future kernels that have the cache problem fixed.
			 */
			if (cmnd->scsi_done) {
				del_command_timer(cmnd);	
						/* must have already started 
						 * running, but may not have 
						 * finished yet 
						 */
				add_completion_timer(cmnd);
				cmnd->scsi_done(cmnd);
			}
		}
		spin_unlock_irqrestore(session->hba->host->host_lock,
				       midlayer_flags_);
	}

	spin_unlock(&session->task_lock);

	return 0;
}

void *
sg_virtual_address(struct scatterlist *sg)
{
	if (sg->page) {
		return page_address(sg->page) + sg->offset;
	}
	return NULL;
}

void *
kmap_sg(struct scatterlist *sg)
{
	if (sg->page) {
		return kmap(sg->page);
	}
	return NULL;
}

void
kunmap_sg(struct scatterlist *sg)
{
	if (sg->page)
		kunmap(sg->page);
	return;
}

/* called by the /proc code, so we can block */
void
iscsi_ping_test_session(struct iscsi_session *session, int total_data_length)
{
	unsigned char *rx_buffer;
	unsigned char *tx_buffer;

	/* assume that we can run the test, and allocate the memory we'll need.
	 * draft 8 only allows 4K of ping data per Nop.
	 */
	rx_buffer = kmalloc(4096, GFP_ATOMIC);
	tx_buffer = kmalloc(4096, GFP_ATOMIC);

	if (rx_buffer && tx_buffer) {
		unsigned char *data;
		unsigned int value = 0;

		/* put a simple pattern in the data */
		for (data = tx_buffer; data < tx_buffer + 4096; data++) {
			*data = value & 0xFF;
			value++;
		}

		spin_lock(&session->task_lock);
		if (session->ping_test_start == 0) {
			/* start a ping test */
			session->ping_test_start = jiffies;
			session->ping_test_data_length = total_data_length;
			session->ping_test_rx_length = total_data_length;
			session->ping_test_rx_start = 0;
			session->ping_test_tx_buffer = tx_buffer;

			smp_mb();
			wake_tx_thread(TX_PING_DATA, session);
			printk
			    ("iSCSI: session iscsi bus %d target id %d starting"
			     " Nop data test with total length %u at %lu\n",
			     session->iscsi_bus, session->target_id,
			     total_data_length, jiffies);
		} else {
			printk
			    ("iSCSI: session iscsi bus %d target id %d can't "
			     "start Nop data test, test started at %lu still "
			     "in progress at %lu\n",
			     session->iscsi_bus, session->target_id,
			     session->ping_test_start, jiffies);
		}
		spin_unlock(&session->task_lock);

		/* the tx and rx thread will free the buffers when they're 
		 * done with them, so that we can just return.
		 */
	} else {
		printk
		    ("iSCSI: session iscsi bus %d target id %d can't start "
		     "Nop data test, couldn't allocate buffers at %lu\n",
		     session->iscsi_bus, session->target_id, jiffies);
		kfree(rx_buffer);
		kfree(tx_buffer);
	}
}

/* the writer thread */
int
iscsi_tx_thread(void *vtaskp)
{
	struct iscsi_session *session;

	if (!vtaskp) {
		printk("iSCSI: tx thread task parameter NULL\n");
		return 0;
	}

	session = (struct iscsi_session *) vtaskp;
	/* whoever created the thread already incremented the 
	 * session's refcount for us 
	 */

	DEBUG_INIT
	    ("iSCSI: tx thread %d for session iscsi bus %d target id %d "
	     "about to daemonize on cpu%d\n",
	     current->pid, session->iscsi_bus, session->target_id,
	     smp_processor_id());

	/* become a daemon kernel thread, and abandon 
	 * any user space resources 
	 */
	iscsi_daemonize("iscsi-tx");
	session->tx_pid = current->pid;
	current->flags |= PF_MEMALLOC;
	smp_mb();

	/* check to see if iscsi_terminate_session was called before we
	 * started running, since we can't get a signal from it until
	 * until we set session->tx_pid.  
	 */
	if (test_bit(SESSION_TERMINATING, &session->control_bits))
		goto ThreadExit;

	/* Block all signals except SIGHUP and SIGKILL */
	spin_lock_irq(&current->sighand->siglock);
	siginitsetinv(&current->blocked, sigmask(SIGKILL) | sigmask(SIGHUP));
	recalc_sigpending();
	spin_unlock_irq(&current->sighand->siglock);

	while (!test_bit(SESSION_TERMINATING, &session->control_bits)) {
		wait_queue_t waitq;
		int timedout = 0;

		DEBUG_INIT
		    ("iSCSI: tx thread %d for session iscsi bus %d target id %d"
		     " waiting for new session to be established at %lu\n",
		     current->pid, session->iscsi_bus, session->target_id,
		     jiffies);

		/* add ourselves to the login wait q, so that the rx thread 
		 * can wake us up 
		 */
		init_waitqueue_entry(&waitq, current);
		add_wait_queue(&session->login_wait_q, &waitq);
		smp_mb();

		for (;;) {
			int replacement_timeout;
			unsigned long now;
			long sleep_jiffies = 0;

			/* tell the rx thread that we're blocked, and that 
			 * it can safely call iscsi_sendmsg now as part of 
			 * the Login phase, since we're guaranteed not to be 
			 * doing any IO until the session is up.  
			 */
			set_current_state(TASK_INTERRUPTIBLE);
			set_bit(TX_THREAD_BLOCKED, &session->control_bits);
			smp_mb();
			wake_up(&session->tx_blocked_wait_q);

			/* if the session is up, our wait is over */
			if (test_bit
			    (SESSION_ESTABLISHED, &session->control_bits))
				break;

			now = jiffies;
			replacement_timeout = session->replacement_timeout;

			/* check for a session replacement timeout */
			if (!timedout && replacement_timeout
			    && session->session_drop_time
			    && time_before_eq(session->session_drop_time +
					      (replacement_timeout * HZ),
					      now)) {
				Scsi_Cmnd *sc;
				unsigned long noqueue_flags_;
				unsigned long midlayer_flags_;

				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "replacement timed after %d seconds, drop "
				     "%lu, now %lu, failing all commands\n",
				     session->iscsi_bus, session->target_id,
				     replacement_timeout,
				     session->session_drop_time, jiffies);

				spin_lock_irqsave(&session->scsi_cmnd_lock,
						  noqueue_flags_);
				spin_lock_irqsave(session->hba->host->host_lock,
						  midlayer_flags_);

				/* make sure any future attempts to 
				 * queue a command fail immediately 
				 */
				set_bit(SESSION_REPLACEMENT_TIMEDOUT,
					&session->control_bits);

				/* don't need to do this again, since 
				 * we just put a barrier blocking any 
				 * more commands from being queued 
				 */
				timedout = 1;

				/* we're failing all commands, so any 
				 * outstanding command timeouts are also 
				 * handled 
				 */
				clear_bit(SESSION_COMMAND_TIMEDOUT,
					  &session->control_bits);

				/* complete all commands currently in 
				 * the driver.
				 * Note: this assumes that the completion 
				 * callback will not call iscsi_queuecommand,
				 * since we're holding the scsi_cmnd_lock, and 
				 * would deadlock with ourselves
				 * if queuecommand was called.
				 */
				while ((sc = session->scsi_cmnd_head)) {
					session->scsi_cmnd_head =
					    (Scsi_Cmnd *) sc->host_scribble;

					atomic_dec(&session->num_cmnds);
					sc->result = host_byte(DID_NO_CONNECT);
					sc->resid =
					    iscsi_expected_data_length(sc);

					set_lun_comm_failure(sc);

					/* FIXME: if this is the last retry of 
					 * a disk write, log a warning about 
					 * possible data loss from the buffer 
					 * cache?
					 */

					if (sc->scsi_done) {
						add_completion_timer(sc);
						DEBUG_EH
						    ("iSCSI: session iscsi bus "
						     "%d target id %d "
						     "replacement timeout "
						     "completing 0x%x at %lu\n",
						     session->iscsi_bus,
						     session->target_id,
						     sc->cmnd[0], jiffies);
						sc->scsi_done(sc);
					}
				}

				spin_unlock_irqrestore(session->
						       hba->host->host_lock, 
						       midlayer_flags_);
				spin_unlock_irqrestore(&session->scsi_cmnd_lock,
						       noqueue_flags_);
			}

			/* process any command timeouts */
			if (test_bit
			    (SESSION_COMMAND_TIMEDOUT,
			     &session->control_bits)) {
				DEBUG_INIT
				    ("iSCSI: session iscsi bus %d target id %d "
				     "processing command timeouts while not "
				     "established at %lu\n",
				     session->iscsi_bus, session->target_id,
				     jiffies);
				process_timedout_commands(session);
			}

			/* wait for either: 
			 *   the rx thread to tell us the session is up
			 *   the session replacement timeout to expire
			 *   a command timeout to expire for the last time
			 */
			if (!timedout && replacement_timeout
			    && session->session_drop_time) {
				unsigned long timeout = 0;

				/* calculate how long till the 
				 * replacement timer expires 
				 */
				now = jiffies;
				if (session->session_drop_time)
					timeout =
					    session->session_drop_time +
					    (HZ * replacement_timeout);
				else
					timeout =
					    now + (HZ * replacement_timeout);

				/* handle wrap-around */
				if (now <= timeout)
					sleep_jiffies = timeout - now;
				else
					sleep_jiffies =
					    ULONG_MAX - now + timeout;

				DEBUG_INIT
				    ("iSCSI: session iscsi bus %d target id %d "
				     "tx thread %d blocking at %lu, timeout "
				     "at %lu\n",
				     session->iscsi_bus, session->target_id,
				     current->pid, jiffies, timeout);
				schedule_timeout(sleep_jiffies);
			} else {
				DEBUG_INIT
				    ("iSCSI: session iscsi bus %d target id %d "
				     "tx thread %d blocking at %lu, timedout "
				     "%d, replacement %d, drop time %lu\n",
				     session->iscsi_bus, session->target_id,
				     current->pid, jiffies, timedout,
				     replacement_timeout,
				     session->session_drop_time);
				schedule();
			}

			if (iscsi_handle_signals(session)) {
				DEBUG_INIT
				    ("iSCSI: session iscsi bus %d target id %d "
				     "tx thread %d signalled at %lu while "
				     "waiting for session establishment\n",
				     session->iscsi_bus, session->target_id,
				     current->pid, jiffies);
			}

			if (test_bit
			    (SESSION_TERMINATING, &session->control_bits)) {
				/* we're all done */
				set_current_state(TASK_RUNNING);
				remove_wait_queue(&session->login_wait_q,
						  &waitq);
				goto ThreadExit;
			}
		}

		/* remove ourselves from the login wait q */
		set_current_state(TASK_RUNNING);
		remove_wait_queue(&session->login_wait_q, &waitq);

		/* we're up and running with a new session */
		clear_bit(TX_THREAD_BLOCKED, &session->control_bits);
		DEBUG_INIT
		    ("iSCSI: tx thread %d for session starting to process "
		     "new session iscsi bus %d target id %d at %lu\n",
		     current->pid, session->iscsi_bus, session->target_id,
		     jiffies);

		/* make sure we start sending commands again */
		set_bit(TX_PING, &session->control_bits);
		set_bit(TX_SCSI_COMMAND, &session->control_bits);
		set_bit(TX_WAKE, &session->control_bits);

		/* don't start any new commands if we're still 
		 * trying to do a reset 
		 */
		if (test_bit(SESSION_RESET_REQUESTED, &session->control_bits)) {
			DEBUG_INIT
			    ("iSCSI: session iscsi bus %d target id %d "
			     "still has a warm reset requested at %lu\n",
			     session->iscsi_bus, session->target_id, jiffies);
			set_bit(SESSION_RESETTING, &session->control_bits);
		}

		/* process tx requests for this session, until 
		 * the session drops 
		 */
		while (!signal_pending(current)) {

			DEBUG_FLOW
			    ("iSCSI: tx thread %d for session iscsi bus %d "
			     "target id %d waiting at %lu\n",
			     session->tx_pid, session->iscsi_bus,
			     session->target_id, jiffies);
			wait_event_interruptible(session->tx_wait_q,
						 test_and_clear_bit(TX_WAKE,
								    &session->
								    control_bits));

			DEBUG_FLOW
			    ("iSCSI: tx thread %d for session iscsi bus %d "
			     "target id %d is awake at %lu\n",
			     session->tx_pid, session->iscsi_bus,
			     session->target_id, jiffies);

			if (signal_pending(current))
				break;

			if (test_bit
			    (SESSION_TASK_TIMEDOUT, &session->control_bits)) {
				process_timedout_tasks(session);
			}

			if (signal_pending(current))
				break;

			if (test_bit
			    (SESSION_COMMAND_TIMEDOUT,
			     &session->control_bits)) {
				process_timedout_commands(session);
			}

			if (signal_pending(current))
				break;

			/* See if we should send a ping (Nop 
			 * with reply requested) 
			 */
			if (test_and_clear_bit(TX_PING, &session->control_bits)) {
				uint32_t itt;

				DEBUG_FLOW
				    ("iSCSI: sending Nop/poll on "
				     "session iscsi bus %d target id %d\n",
				     session->iscsi_bus, session->target_id);
				/* may need locking someday.  see 
				 * allocate_itt comment 
				 */
				itt = allocate_itt(session);
				iscsi_xmit_ping(session, itt, NULL, 0);
			}

			if (signal_pending(current))
				break;

			/* See if we should send a ping (Nop with 
			 * reply requested) containing test data 
			 */
			if (test_and_clear_bit
			    (TX_PING_DATA, &session->control_bits)) {
				int data_length, length;
				unsigned char *buffer;
				unsigned long tx_start, tx_stop;

				/* grab the total data length and buffer to 
				 * use 
				 */
				spin_lock(&session->task_lock);
				length = data_length =
				    session->ping_test_data_length;
				buffer = session->ping_test_tx_buffer;
				session->ping_test_tx_buffer = NULL;
				spin_unlock(&session->task_lock);

				tx_start = jiffies;
				while ((length > 0) && buffer) {
					/* may need locking someday.  
					 * see allocate_itt comment 
					 */
					uint32_t itt = allocate_itt(session);

					DEBUG_FLOW
					    ("iSCSI: sending Nop/poll with "
					     "data on session iscsi bus %d "
					     "target id %d\n",
					     session->iscsi_bus,
					     session->target_id);
					iscsi_xmit_ping(session, itt, buffer,
							MIN(4096, length));

					if (signal_pending(current)) {
						printk
						    ("iSCSI: session iscsi bus "
						     "%d target id %d Nop data "
						     "tx failed at %lu\n",
						     session->iscsi_bus,
						     session->target_id,
						     jiffies);
						break;
					}

					length -= MIN(4096, length);
				}
				tx_stop = jiffies;

				kfree(buffer);

				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "tx Nop data test - tx %d of %d bytes, "
				     "start %lu, stop %lu,jiffies %lu, HZ %u\n",
				     session->iscsi_bus, session->target_id,
				     data_length - length, data_length,
				     tx_start, tx_stop, tx_stop - tx_start, HZ);
			}

			if (signal_pending(current))
				break;

			/* See if we should send one or more Nops 
			 * (replies requested by the target) 
			 */
			if (test_and_clear_bit
			    (TX_NOP_REPLY, &session->control_bits)) {
				DEBUG_FLOW
				    ("iSCSI: sending Nop replies on session "
				     "iscsi bus %d target id %d\n",
				     session->iscsi_bus, session->target_id);
				iscsi_xmit_nop_replys(session);
			}

			if (signal_pending(current))
				break;

			/* See if we should warm reset the target */
			if (test_bit
			    (SESSION_RESET_REQUESTED, &session->control_bits)
			    && (session->warm_reset_itt == RSVD_TASK_TAG)) {
				if (test_bit
				    (SESSION_RESETTING,
				     &session->control_bits)) {
					/* error recovery is already doing a 
					 * reset, so we don't need to 
					 */
					printk
					    ("iSCSI: session iscsi bus %d "
					     "ignoring target reset request for"
					     " (%u %u %u *), reset already in "
					     "progress at %lu\n",
					     session->iscsi_bus,
					     session->host_no, session->channel,
					     session->target_id, jiffies);
					clear_bit(SESSION_RESET_REQUESTED,
						  &session->control_bits);
				} else {
					uint32_t itt;
					struct iscsi_task *task;

					spin_lock(&session->task_lock);

					session->warm_reset_itt = itt =
					    allocate_itt(session);
					session->reset_response_deadline =
					    jiffies +
					    (session->reset_timeout * HZ);
					if (session->reset_response_deadline ==
					    0)
						session->
						    reset_response_deadline = 1;

					printk
					    ("iSCSI: session iscsi bus %d "
					     "requested target reset for "
					     "(%u %u %u *), warm reset itt %u "
					     "at %lu\n",
					     session->iscsi_bus,
					     session->host_no, session->channel,
					     session->target_id, itt, jiffies);
					/* prevent any new tasks from starting 
					 * or existing tasks from completing 
					 */
					set_bit(SESSION_RESETTING,
						&session->control_bits);
					list_for_each_entry(task,
							    &session->task_arrival_order,
							    task_order_link) {
						DEBUG_EH
						    ("iSCSI: session iscsi bus "
						     "%d target id %d warm "
						     "target reset causing "
						     "problems for LUN %u\n",
						     session->iscsi_bus,
						     session->target_id,
						     task->lun);
						set_bit(task->lun,
							session->
							luns_timing_out);
						del_task_timer(task);
						set_bit(0, &task->timedout);
						/* the task mgmt response will 
						 * set SESSION_TASK_TIMEDOUT 
						 * and ensure these get 
						 * processed later 
						 */
					}
					spin_unlock(&session->task_lock);

					iscsi_xmit_task_mgmt(session,
							     ISCSI_TM_FUNC_TARGET_WARM_RESET,
							     NULL, itt);
				}
			}

			if (signal_pending(current))
				break;

			if (test_and_clear_bit
			    (SESSION_RETRY_COMMANDS, &session->control_bits)) {
				/* try to queue up delayed commands 
				 * for retries 
				 */
				iscsi_retry_commands(session);
			}

			if (signal_pending(current))
				break;

			/* New SCSI command received, or 
			 * MaxCmdSN incremented, or task freed 
			 */
			if (test_and_clear_bit
			    (TX_SCSI_COMMAND, &session->control_bits)) {
				/* if possible, issue new commands */
				iscsi_xmit_queued_cmnds(session);
			}

			if (signal_pending(current))
				break;

			/* See if we need to send more data */
			if (test_and_clear_bit(TX_DATA, &session->control_bits)) {
				/* NOTE: this may call iscsi_xmit_queued_cmnds 
				 * under some conditions 
				 */
				iscsi_xmit_r2t_data(session);
			}

			if (signal_pending(current))
				break;

			if (test_and_clear_bit
			    (TX_LOGOUT, &session->control_bits)) {
				uint32_t itt;

				DEBUG_INIT
				    ("iSCSI: session iscsi bus %d target id %d "
				     "sending Logout at %lu\n",
				     session->iscsi_bus, session->target_id,
				     jiffies);
				/* may need locking someday.  see 
				 * allocate_itt comment 
				 */
				itt = allocate_itt(session);
				session->logout_itt = itt;
				smp_mb();
				iscsi_xmit_logout(session, itt,
						  ISCSI_LOGOUT_REASON_CLOSE_SESSION);
			}
		}

		/* handle any signals that may have occured */
		iscsi_handle_signals(session);
	}

      ThreadExit:
	DEBUG_INIT
	    ("iSCSI: tx thread %d for session iscsi bus %d "
	     "target id %d exiting\n",
	     session->tx_pid, session->iscsi_bus, session->target_id);

	/* the rx thread may be waiting for the tx thread to block.  
	 * make it happy 
	 */
	set_bit(TX_THREAD_BLOCKED, &session->control_bits);
	wake_up(&session->tx_blocked_wait_q);

	/* we're done */
	set_current_state(TASK_RUNNING);
	session->tx_pid = 0;
	smp_mb();
	drop_reference(session);

	return 0;
}

/* update LUN info for sysfs read purpose.
 * Watch the inquiry data rather than TEST_UNIT_READY response,
 * since the TURs come from the target driver (sd, st, etc), and
 * that may not happen until after a SCSI device node is opened
 * and the target driver is loaded.  We get INQUIRY commands when
 * the HBA registers, and when our driver's LUN probing forces
 * them.  This is more reliable than waiting for TURs.
 */
void
process_inquiry_data(struct iscsi_session *session, Scsi_Cmnd * sc,
		     uint8_t * data)
{
	if (data && sc) {
		/* look at the peripheral qualifier (bits 5,6, and 7 of the 
		 * first byte), SPC-3 7.4.2 
		 */
		if (LOG_ENABLED(ISCSI_LOG_INIT))
			printk
			    ("iSCSI: session iscsi bus %d cmd 0x%x INQUIRY "
			     "peripheral 0x%02x from (%u %u %u %u), %s\n",
			     session->iscsi_bus, sc->cmnd[0], *data,
			     sc->device->host->host_no, sc->device->channel,
			     sc->device->id, sc->device->lun,
			     session->log_name);

		switch ((*data & 0xE0) >> 5) {
		case 0:	/* possibly a LUN */
			if (sc->device->lun < ISCSI_MAX_LUN) {
				/* detected a LUN */
				set_bit(sc->device->lun, session->luns_detected);
				set_bit(sc->device->lun, session->luns_found);

				/* unless it's one of our driver's LUN probes, 
				 * the SCSI layer will activate this LUN 
				 */
				if ((sc->scsi_done != iscsi_done) &&
				    !test_and_set_bit(sc->device->lun,
						      session->luns_activated))
			        {
					/* assume we found a useable LUN */
					session->num_luns++;
					smp_mb();
				}
			}
			break;
		case 1:	/* capable of supporting a physical device at this 
			 * LUN, but not currently connected 
			 */
			/* as of 2002may09, the available Linux kernels treat 
			 * qualifier 1 the same as qualifier 0,
			 * even though that's not really appropriate, and 
			 * fills the log with a bunch of messages
			 * about unknown device types.  Map it to qualifier 3, 
			 * which gets silently ignored.
			 */
			*data = 0x7F;	/* Linux only ignores a 0x7F */
			/* fall-through */
		case 2:	/* reserved */
		case 3:	/* not capable of supporting a physical 
			 * device at this LUN 
			 */
		default:
			if (sc->device->lun < ISCSI_MAX_LUN) {
				clear_bit(sc->device->lun,
					  session->luns_detected);

				if ((sc->scsi_done != iscsi_done) &&
				    test_and_clear_bit(sc->device->lun,
						       session->luns_activated))
			        {
					/* there's not really a useable LUN */
					session->num_luns--;
					smp_mb();
				}
			}
			break;
		}
	} else {
		printk
		    ("iSCSI: failed to process inquiry data for session "
		     "iscsi bus %d target id %d, sc 0x%x\n",
		     session->iscsi_bus, session->target_id, sc->cmnd[0]);
	}
}

int
iscsi_lun_thread(void *vtaskp)
{
	struct iscsi_session *session;
	int rc = -1;
	int lun;

	session = (struct iscsi_session *) vtaskp;

	printk
	    ("iSCSI: session iscsi bus %d target id %d lun thread %d "
	     "about to daemonize on cpu%d\n",
	     session->iscsi_bus, session->target_id, current->pid,
	     smp_processor_id());

	/* become a daemon kernel thread */
	iscsi_daemonize("iscsi-lun-thr");
	current->flags |= PF_MEMALLOC;
	smp_mb();

	/* Block all signals except SIGHUP and SIGKILL */
	spin_lock_irq(&current->sighand->siglock);
	siginitsetinv(&current->blocked, sigmask(SIGKILL) | sigmask(SIGHUP));
	recalc_sigpending();
	spin_unlock_irq(&current->sighand->siglock);

	if (test_and_set_bit(SESSION_PROBING_LUNS, &session->control_bits)) {
		printk
		    ("iSCSI: session already has a process probing or "
		     "waiting to probe LUNs for bus %d, target %d\n",
		     session->iscsi_bus, session->target_id);
		rc = -EBUSY;
		goto done;
	}
	iscsi_detect_luns(session);
	for (lun = 0; lun < ISCSI_MAX_LUN; lun++) {
		if (test_bit(lun, session->luns_detected)) {
			/* These are the original luns present */
			if (!test_bit(lun, session->luns_found)) {
				/* the lun seems to have changed */
				iscsi_remove_lun(session, lun);
			}
		}
	}

	if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
		printk
		    ("iSCSI: session iscsi bus %d target id %d terminating, "
		     "returning at %lu\n",
		     session->iscsi_bus, session->target_id, jiffies);
		clear_bit(SESSION_PROBING_LUNS, &session->control_bits);
		smp_mb();
		goto done;
	} else if (signal_pending(current)) {
		iscsi_terminate_session(session);
		printk
		    ("iSCSI: session iscsi bus %d target id %d "
		     "ioctl terminated, returning at %lu\n",
		     session->iscsi_bus, session->target_id, jiffies);
		clear_bit(SESSION_PROBING_LUNS, &session->control_bits);
		smp_mb();
		goto done;
	}

	iscsi_probe_luns(session, session->luns_allowed);

	/* and then we're done */
	clear_bit(SESSION_PROBING_LUNS, &session->control_bits);
	smp_mb();
	rc = 0;

	if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
		printk
		    ("iSCSI: session iscsi bus %d target id %d terminating, "
		     "ioctl returning at %lu\n",
		     session->iscsi_bus, session->target_id, jiffies);
	} else if (signal_pending(current)) {
		iscsi_terminate_session(session);
		printk
		    ("iSCSI: session iscsi bus %d target id %d "
		     "ioctl terminated, returning at %lu\n",
		     session->iscsi_bus, session->target_id, jiffies);
	}

      done:			/* lun change event finished */
	DEBUG_INIT("iSCSI: lun thread leaving kernel at %lu\n", jiffies);
	set_current_state(TASK_RUNNING);
	drop_reference(session);
	smp_mb();
	return rc;
}

/* wait for the tx thread to block or exit, ignoring signals.
 * the rx thread needs to know that the tx thread is not running before
 * it can safely close the socket and start a new login phase on a new socket,
 * Also, tasks still in use by the tx thread can't safely be completed on
 * a session drop.
 */
static int
wait_for_tx_blocked(struct iscsi_session *session)
{
	while (session->tx_pid) {
		DEBUG_INIT
		    ("iSCSI: session iscsi bus %d target id %d thread %d "
		     "waiting for tx thread %d to block\n",
		     session->iscsi_bus, session->target_id, current->pid,
		     session->tx_pid);

		wait_event_interruptible(session->tx_blocked_wait_q,
					 test_bit(TX_THREAD_BLOCKED,
						  &session->control_bits));

		if (iscsi_handle_signals(session)) {
			DEBUG_INIT
			    ("iSCSI: session iscsi bus %d target id %d "
			     "wait_for_tx_blocked signalled at %lu while "
			     "waiting for tx %d\n",
			     session->iscsi_bus, session->target_id, jiffies,
			     session->tx_pid);
		}
		/* if the session is terminating, the tx thread will exit, 
		 * waking us up in the process we don't want to return until 
		 * the tx thread is blocked, since there's not much
		 * the rx thread can do until the tx thread is guaranteed not 
		 * to be doing anything.
		 */
		if (test_bit(TX_THREAD_BLOCKED, &session->control_bits)) {
			DEBUG_INIT
			    ("iSCSI: session iscsi bus %d target id %d rx "
			     "thread %d found tx thread %d blocked\n",
			     session->iscsi_bus, session->target_id,
			     current->pid, session->tx_pid);
			return 1;
		}
	}

	/* dead and blocked are fairly similar, really */
	DEBUG_INIT
	    ("iSCSI: session iscsi bus %d target id %d rx thread %d found "
	     "tx thread %d exited\n",
	     session->iscsi_bus, session->target_id, current->pid,
	     session->tx_pid);
	return 1;
}

static inline void
append_queue(Scsi_Cmnd ** to_head, Scsi_Cmnd ** to_tail, Scsi_Cmnd ** from_head,
	     Scsi_Cmnd ** from_tail)
{
	if (*to_head && *from_head) {
		/* both non-empty, append 'from' to 'to' */
		(*to_tail)->host_scribble = (void *) *from_head;
		*to_tail = *from_tail;
		*from_head = NULL;
		*from_tail = NULL;
	} else if (*from_head) {
		/* 'from' becomes 'to' */
		*to_head = *from_head;
		*to_tail = *from_tail;
		*from_head = NULL;
		*from_tail = NULL;
	}
}

/* caller must hold the task_lock */
static void
requeue_or_fail_commands(struct iscsi_session *session)
{
	Scsi_Cmnd *fatal_head = NULL, *fatal_tail = NULL;
	Scsi_Cmnd *requeue_head = NULL, *requeue_tail = NULL;
	Scsi_Cmnd *sc;
	struct iscsi_task *task;
	int fail_all = 0;
	int num_failed = 0;
	int num_tasks = 0;
	unsigned long midlayer_flags_;
	unsigned long noqueue_flags_;

	if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
		/* no point in retrying anything */
		if (test_bit(ISCSI_HBA_SHUTTING_DOWN, &session->hba->flags))
			DEBUG_INIT
			    ("iSCSI: session iscsi bus %d target id %d "
			     "terminating, failing all SCSI commands\n",
			     session->iscsi_bus, session->target_id);
		else
			printk
			    ("iSCSI: session iscsi bus %d target id %d "
			     "terminating, failing all SCSI commands\n",
			     session->iscsi_bus, session->target_id);
		fail_all = 1;
	} else {
		DEBUG_INIT
		    ("iSCSI: session iscsi bus %d target id %d "
			     "requeue_or_fail_commands at %lu\n",
		     session->iscsi_bus, session->target_id, jiffies);
	}

	/* grab all the tasks for this connection */
        list_for_each_entry(task, &session->task_arrival_order,
			    task_order_link) {
		list_del_init(&task->task_group_link);
		list_del_init(&task->task_order_link);
		del_task_timer(task);

		if (atomic_read(&task->refcount) == 0) {
			sc = task->scsi_cmnd;
			ISCSI_TRACE(ISCSI_TRACE_TaskAborted, sc, task, 0, 0);
			task->scsi_cmnd = NULL;

			if (sc)
				add_cmnd(sc, &requeue_head, &requeue_tail);

			if (test_bit
			    (SESSION_TERMINATING, &session->control_bits))
				DEBUG_ALLOC
				    ("iSCSI: session iscsi bus %d target id %d "
				     "requeue_or_fail freeing task "
				     "itt %u at %lu\n",
				     session->iscsi_bus, session->target_id,
				     task->itt, jiffies);

			num_tasks++;
			free_task(session, task);
		} else {
			/* This should never happen, which is good, since 
			 * we don't really have any good options here.  Leak 
			 * the task memory, and fail to complete the cmnd, 
			 * which may leave apps blocked forever in the kernel.
			 */
			printk
			    ("iSCSI: bug - session iscsi bus %d target id %d "
			     "can't complete itt %u refcount %u, command 0x%x, "
			     "leaking task memory\n",
			     session->iscsi_bus, session->target_id, task->itt,
			     atomic_read(&task->refcount),
			     task->scsi_cmnd->cmnd[0]);
		}
		/* proceed with rest of elements.
		 * This is a bad hack. What if the first element was not
		 * task_arrival_order?
		 */
		task = (struct iscsi_task *) (&session->task_arrival_order);
	}
	INIT_LIST_HEAD(&(session->task_arrival_order));
	

	if (test_bit(SESSION_TERMINATING, &session->control_bits)
	    && LOG_ENABLED(ISCSI_LOG_ALLOC))
		printk
		    ("iSCSI: session iscsi bus %d for (%u %u %u *) "
		     "requeue_or_fail freed %d tasks at %lu, alloc %u "
		     "freed %u\n",
		     session->iscsi_bus, session->host_no, session->channel,
		     session->target_id, num_tasks, jiffies,
		     session->tasks_allocated, session->tasks_freed);

	atomic_set(&session->num_active_tasks, 0);
	/* clear out the task collections */
	INIT_LIST_HEAD(&(session->tx_task_head));
	session->warm_reset_itt = RSVD_TASK_TAG;
	session->cold_reset_itt = RSVD_TASK_TAG;

	/* grab the retry, deferred, and normal queues in that order */
	spin_lock_irqsave(&session->scsi_cmnd_lock, noqueue_flags_);
	append_queue(&requeue_head, &requeue_tail, &session->retry_cmnd_head,
		     &session->retry_cmnd_tail);
	atomic_set(&session->num_retry_cmnds, 0);
	append_queue(&requeue_head, &requeue_tail, &session->deferred_cmnd_head,
		     &session->deferred_cmnd_tail);
	session->num_deferred_cmnds = 0;
	append_queue(&requeue_head, &requeue_tail, &session->scsi_cmnd_head,
		     &session->scsi_cmnd_tail);
	atomic_set(&session->num_cmnds, 0);

	while ((sc = requeue_head)) {
		requeue_head = (Scsi_Cmnd *) sc->host_scribble;

		if (fail_all || (sc->allowed <= 1)) {
			/* fail it */
			add_cmnd(sc, &fatal_head, &fatal_tail);
			num_failed++;
		} else {
			/* requeue it */
			add_cmnd(sc, &session->scsi_cmnd_head,
				 &session->scsi_cmnd_tail);
			atomic_inc(&session->num_cmnds);
		}
	}

	spin_unlock_irqrestore(&session->scsi_cmnd_lock, noqueue_flags_);

	/* fail any commands that can't be retried */
	spin_lock_irqsave(session->hba->host->host_lock, midlayer_flags_);
	while ((sc = fatal_head)) {
		fatal_head = (Scsi_Cmnd *) sc->host_scribble;

		del_command_timer(sc);
		sc->host_scribble = NULL;
		sc->resid = iscsi_expected_data_length(sc);
		sc->result = host_byte(DID_NO_CONNECT);
		if (sc->allowed > 1)
			sc->retries = sc->allowed - 1;

		set_not_ready(sc);

		/* FIXME: always log these?  sometimes log these? */
		printk
		    ("iSCSI: session iscsi bus %d failing command "
		     "cdb 0x%02x to (%u %u %u %u) at %lu\n",
		     session->iscsi_bus, sc->cmnd[0], sc->device->host->host_no,
		     sc->device->channel, sc->device->id, sc->device->lun,
		     jiffies);

		if (sc->scsi_done) {
			add_completion_timer(sc);
			sc->scsi_done(sc);
		}
	}

	spin_unlock_irqrestore(session->hba->host->host_lock, midlayer_flags_);
}

int
iscsi_rx_thread(void *vtaskp)
{
	struct iscsi_session *session;
	int rc = -EPIPE, length, xlen;
	struct msghdr msg;
	struct iovec iov[2];
	struct IscsiHdr sth;
	uint32_t crc32c;
	unsigned char *rxbuf;
	long login_delay = 0;
	int pad;
	unsigned long session_failures = 0;

	if (vtaskp == NULL) {
		printk("iSCSI: rx thread task parameter NULL\n");
		return 0;
	}

	session = (struct iscsi_session *) vtaskp;
	/* whoever created the thread already incremented the session's 
	 * refcount for us 
	 */


	DEBUG_INIT
	    ("iSCSI: session iscsi bus %d target id %d rx thread %d about "
	     "to daemonize on cpu%d\n",
	     session->iscsi_bus, session->target_id, current->pid,
	     smp_processor_id());

	/* become a daemon kernel thread, and abandon any 
	 * user space resources 
	 */
	iscsi_daemonize("iscsi-rx");
	session->rx_pid = current->pid;
	current->flags |= PF_MEMALLOC;
	smp_mb();

	/* check to see if iscsi_terminate_session was called before we
	 * started running, since we can't get a signal from it until
	 * until we set session->rx_pid.
	 */
	if (test_bit(SESSION_TERMINATING, &session->control_bits))
		goto ThreadExit;

	/* Block all signals except SIGHUP and SIGKILL */
	spin_lock_irq(&current->sighand->siglock);
	siginitsetinv(&current->blocked, sigmask(SIGKILL) | sigmask(SIGHUP));
	recalc_sigpending();
	spin_unlock_irq(&current->sighand->siglock);

	DEBUG_INIT
	    ("iSCSI: session iscsi bus %d target id %d "
	     "rx thread %d starting on cpu%d\n",
	     session->iscsi_bus, session->target_id, current->pid,
	     smp_processor_id());

	while (!test_bit(SESSION_TERMINATING, &session->control_bits)) {
		unsigned long login_failures = 0;

		/* we need a session for the rx and tx threads to use */
		while (!test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
			if (login_delay) {
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "to %s waiting %ld seconds before "
				     "next login attempt\n",
				     session->iscsi_bus, session->target_id,
				     session->log_name, login_delay);
				set_current_state(TASK_INTERRUPTIBLE);
				schedule_timeout(login_delay * HZ);
			}

			/* ensure we can write to the socket 
			 * without interference 
			 */
			DEBUG_INIT
			    ("iSCSI: session iscsi bus %d target id %d "
			     "rx thread %d waiting for tx blocked for at %lu\n",
			     session->iscsi_bus, session->target_id,
			     current->pid, jiffies);
			wait_for_tx_blocked(session);
			if (test_bit
			    (SESSION_TERMINATING, &session->control_bits))
				goto ThreadExit;

			/* now that the tx thread is idle, it's safe to 
			 * clean up the old session, if there was one 
			 */
			iscsi_disconnect(session);
			/* FIXME: should clearing these bits move to 
			 * iscsi_establish_session? 
			 */
			clear_bit(SESSION_DROPPED, &session->control_bits);
			clear_bit(SESSION_TASK_ALLOC_FAILED,
				  &session->control_bits);
			clear_bit(SESSION_LOGOUT_REQUESTED,
				  &session->control_bits);
			clear_bit(SESSION_WINDOW_CLOSED,
				  &session->control_bits);
			clear_bit(SESSION_RESETTING, &session->control_bits);
			clear_bit(SESSION_RESET, &session->control_bits);
			clear_bit(SESSION_TASK_MGMT_TIMEDOUT,
				  &session->control_bits);
			smp_mb();

			/* try to get a new session */
			rc = iscsi_establish_session(session);
			if (rc > 0) {
				/* established or redirected */
				login_failures = 0;
				session->session_alive = 1;
			} else if (rc < 0) {
				/* failed, retry */
				login_failures++;
			} else {
				/* failed, give up */
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "giving up at %lu\n",
				     session->iscsi_bus, session->target_id,
				     jiffies);
				iscsi_terminate_session(session);
				goto ThreadExit;
			}

			/* slowly back off the frequency of login attempts */
			if (login_failures == 0)
				login_delay = 0;
			else if (login_failures < 30)
				login_delay = 1;	
					/* 30 seconds at 1 sec each */
			else if (login_failures < 48)
				login_delay = 5;	
					/* another 90 seconds at 5 sec each */
			else if (session->replacement_timeout &&
				 time_before_eq(session->session_drop_time +
						(HZ *
						 session->replacement_timeout),
						jiffies)) {
				login_delay = 10;	/* every 10 seconds */
			} else {
				/* we've already failed all commands out of the
				 * driver, but if we can bring the session back 
				 * up, we can stop failing new commands in 
				 * queuecommand.
				 */
				login_delay = 60;
			}
		}

		DEBUG_INIT
		    ("iSCSI: session iscsi bus %d target id %d established by "
		     "rx thread %d at %lu\n",
		     session->iscsi_bus, session->target_id, current->pid,
		     jiffies);

		/* handle rx for this session */
		while (!signal_pending(current)) {
			/* check for anything to read on socket */
			iov[0].iov_base = &sth;
			iov[0].iov_len = length = sizeof (sth);
			memset(&msg, 0, sizeof (msg));
			msg.msg_iov = iov;
			msg.msg_iovlen = 1;
			if (session->HeaderDigest == ISCSI_DIGEST_CRC32C) {
				iov[1].iov_base = &crc32c;
				iov[1].iov_len = sizeof (crc32c);
				msg.msg_iovlen = 2;
				length += sizeof (crc32c);
			}

			DEBUG_FLOW
			    ("iSCSI: session iscsi bus %d target id %d rx "
			     "thread %d waiting to receive %d header bytes\n",
			     session->iscsi_bus, session->target_id,
			     session->rx_pid, length);

			rc = iscsi_recvmsg(session, &msg, length);
			if (signal_pending(current)) {
				DEBUG_FLOW
				    ("iSCSI: session iscsi bus %d target id %d "
				     "rx thread %d received signal\n",
				     session->iscsi_bus, session->target_id,
				     session->rx_pid);
				goto EndSession;
			}
			if (rc == length) {
				DEBUG_FLOW
				    ("iSCSI: session iscsi bus %d target id %d "
				     "rx thread %d received %d header bytes, "
				     "opcode 0x%x\n",
				     session->iscsi_bus, session->target_id,
				     session->rx_pid, length, sth.opcode);
				/* HeaderDigests */
				if (session->HeaderDigest ==
				    ISCSI_DIGEST_CRC32C) {
					uint32_t calculated_crc32c =
					    iscsi_crc32c(&sth, sizeof (sth));

					if (session->fake_read_header_mismatch >
					    0) {
						session->
						    fake_read_header_mismatch--;
						smp_mb();
						printk
						    ("iSCSI: session iscsi bus "
						     "%d target id %d faking "
						     "HeaderDigest mismatch "
						     "for itt %u\n",
						     session->iscsi_bus,
						     session->target_id,
						     ntohl(sth.itt));
						calculated_crc32c = 0x01020304;
					}

					if (calculated_crc32c != crc32c) {
						printk
						    ("iSCSI: session iscsi bus "
						     "%d target id %d "
						     "HeaderDigest mismatch, "
						     "received 0x%08x, "
						     "calculated 0x%08x, "
						     "dropping session "
						     "at %lu\n",
						     session->iscsi_bus,
						     session->target_id, crc32c,
						     calculated_crc32c,
						     jiffies);
						iscsi_drop_session(session);
						goto EndSession;
					}
				}

				/* received something */
				xlen = ntoh24(sth.dlength);

				if (sth.hlength) {
					/* FIXME: read any additional header 
					 * segments. For now, drop the session 
					 * if one is received, since we can't 
					 * handle them.
					 */
					printk
					    ("iSCSI: session iscsi bus %d "
					     "target id %d received opcode %x, "
					     "ahs length %d, dlength %d, "
					     "itt %u at %lu\n",
					     session->iscsi_bus,
					     session->target_id, sth.opcode,
					     sth.hlength, xlen, ntohl(sth.itt),
					     jiffies);
					printk
					    ("iSCSI: session iscsi bus %d "
					     "target id %d dropping, "
					     "additional header segments not "
					     "supported by this driver "
					     "version.\n",
					     session->iscsi_bus,
					     session->target_id);
					iscsi_drop_session(session);
					goto EndSession;
				}

				/* If there are padding bytes, 
				 * read them as well 
				 */
				pad = xlen % PAD_WORD_LEN;
				if (pad) {
					pad = PAD_WORD_LEN - pad;
					xlen += pad;
				}

				DEBUG_FLOW
				    ("iSCSI: session iscsi bus %d target id %d "
				     "rx PDU, opcode 0x%x, dlength %d at %lu\n",
				     session->iscsi_bus, session->target_id,
				     sth.opcode, xlen, jiffies);

				if (xlen
				    && (sth.opcode != ISCSI_OP_SCSI_DATA_RSP)
				    && (sth.opcode != ISCSI_OP_NOOP_IN)) {
					/* unless it's got a (possibly large) 
					 * data payload, read the whole PDU 
					 * into memory beforehand 
					 */
					if (xlen > ISCSI_RXCTRL_SIZE) {
						printk
						    ("iSCSI: session iscsi bus "
						     "%d target id %d PDU data "
						     "length too large, opcode "
						     "%x, dlen %d\n",
						     session->iscsi_bus,
						     session->target_id,
						     sth.opcode, xlen);
						iscsi_drop_session(session);
						goto EndSession;
					}
					rxbuf = session->rx_buffer;
					iov[0].iov_base = rxbuf;
					iov[0].iov_len = xlen;
					memset(&msg, 0, sizeof (struct msghdr));
					msg.msg_iov = iov;
					msg.msg_iovlen = 1;
					length = xlen;

					if (session->DataDigest ==
					    ISCSI_DIGEST_CRC32C) {
						iov[1].iov_base = &crc32c;
						iov[1].iov_len =
						    sizeof (crc32c);
						msg.msg_iovlen = 2;
						length += sizeof (crc32c);
					}

					rc = iscsi_recvmsg(session, &msg,
							   length);
					if (rc != length) {
						printk
						    ("iSCSI: session iscsi bus "
						     "%d target id %d PDU "
						     "opcode 0x%x, recvmsg %d "
						     "failed, rc %d\n",
						     session->iscsi_bus,
						     session->target_id,
						     sth.opcode, length, rc);
						iscsi_drop_session(session);
						goto EndSession;
					}

					if (session->DataDigest ==
					    ISCSI_DIGEST_CRC32C) {
						uint32_t calculated_crc32c =
						    iscsi_crc32c(rxbuf, xlen);

						if (calculated_crc32c != crc32c) {
/* FIXME: if it's a command response, we MUST do a Logout and drop the session.
 * it's not a command response or data, we're allowed to just ignore the PDU.
 * It must have been Async with sense, or a Reject, or Nop-in with data, and
 * other timers should handle those.  For now, ignore the spec, and just
 * drop the session unconditionally.
 */
							printk
							    ("iSCSI: session "
							     "iscsi bus %d "
							     "target id %d "
							     "DataDigest "
							     "mismatch, opcode "
							     "0x%x, received "
							     "0x%08x, "
							     "calculated "
							     "0x%08x, dropping "
							     "session at %lu\n",
							     session->iscsi_bus,
							     session->target_id,
							     sth.opcode, crc32c,
							     calculated_crc32c,
							     jiffies);
							iscsi_drop_session
							    (session);
							goto EndSession;
						}
					}
				} else {
					rxbuf = NULL;
				}

				switch (sth.opcode) {
				case ISCSI_OP_NOOP_IN | 0xc0:	
			/* work-around a bug in the Intel Nov05 target */
				case ISCSI_OP_NOOP_IN:
					iscsi_recv_nop(session,
						       (struct IscsiNopInHdr *)
						       &sth);
					break;
				case ISCSI_OP_SCSI_RSP:
					iscsi_recv_cmd(session,
						       (struct IscsiScsiRspHdr
							*) &sth, rxbuf);
					break;
				case ISCSI_OP_SCSI_TASK_MGT_RSP:
					iscsi_recv_task_mgmt(session,
							     (struct
							      IscsiScsiTaskMgtRspHdr
							      *) &sth);
					break;
				case ISCSI_OP_RTT_RSP:
					iscsi_recv_r2t(session,
						       (struct IscsiRttHdr *)
						       &sth);
					break;
				case ISCSI_OP_SCSI_DATA_RSP:
					iscsi_recv_data(session,
							(struct IscsiDataRspHdr
							 *) &sth);
					break;
				case ISCSI_OP_ASYNC_EVENT:
					iscsi_recv_async_event(session,
							       (struct
								IscsiAsyncEvtHdr
								*) &sth, rxbuf);
					break;
				case ISCSI_OP_REJECT_MSG:
					iscsi_recv_reject(session,
							  (struct
							   IscsiRejectRspHdr *)
							  &sth, rxbuf);
					break;
				case ISCSI_OP_LOGOUT_RSP:
					iscsi_recv_logout(session,
							  (struct
							   IscsiLogoutRspHdr *)
							  &sth);
					break;
				default:
					printk
					    ("iSCSI: session iscsi bus %d "
					     "target id %d dropping after "
					     "receiving unexpected "
					     "opcode 0x%x\n",
					     session->iscsi_bus,
					     session->target_id, sth.opcode);
					session->time2wait = 2;	
						/* don't spin if the target 
						 * always sends illegal opcodes
						 */
					iscsi_drop_session(session);
					goto EndSession;
				}
			} else {
				if (rc != -EAGAIN) {
					if (rc == 0) {
						printk
						    ("iSCSI: session iscsi bus "
						     "%d target id %d closed by"
						     " target %s at %lu\n",
						     session->iscsi_bus,
						     session->target_id,
						     session->log_name,
						     jiffies);
					} else if (rc == -ECONNRESET) {
						printk
						    ("iSCSI: session iscsi bus "
						     "%d target id %d to %s "
						     "received connection "
						     "reset at %lu\n",
						     session->iscsi_bus,
						     session->target_id,
						     session->log_name,
						     jiffies);
					} else if (rc == -ERESTARTSYS) {
						printk
						    ("iSCSI: session iscsi bus "
						     "%d target id %d to %s "
						     "received signal at %lu\n",
						     session->iscsi_bus,
						     session->target_id,
						     session->log_name,
						     jiffies);
					} else {
						printk
						    ("iSCSI: session iscsi bus "
						     "%d target id %d to %s "
						     "short PDU header read, "
						     "%d of %d at %lu\n",
						     session->iscsi_bus,
						     session->target_id,
						     session->log_name, rc,
						     length, jiffies);
					}
					iscsi_drop_session(session);
					goto EndSession;
				}
			}
		}

	      EndSession:
		DEBUG_INIT
		    ("iSCSI: session iscsi bus %d target id %d "
		     "going down at %lu\n",
		     session->iscsi_bus, session->target_id, jiffies);

		/* calculate how long to wait before logging in again */
		if (session->time2wait >= 0) {
			/* the target gave us a specific Time2Wait */
			login_delay = session->time2wait;
			session->time2wait = -1;
			DEBUG_INIT
			    ("iSCSI: session iscsi bus %d target id %d "
			     "Time2Wait %ld\n",
			     session->iscsi_bus, session->target_id,
			     login_delay);
		} else {
			/* use the default */
			login_delay = session->DefaultTime2Wait;
			DEBUG_INIT
			    ("iSCSI: session iscsi bus %d target id %d "
			     "DefaultTime2Wait %ld\n",
			     session->iscsi_bus, session->target_id,
			     login_delay);
		}

		if (time_before_eq
		    (session->session_drop_time,
		     session->session_established_time + (2 * HZ))) {
			/* if the session dies really quicky after we reach
			 * full-feature phase, we ll not be interoperable due to
			 * bugs in the target (or this driver) that send illegal
			 * opcodes, or disagreements about how to do CRC
			 * calculations.  To avoid spinning, we track sessions
			 * with really short lifetimes, and decrease the login
			 * frequency if we keep getting session failures,like we
			 * do for login failures.
			 */
			session_failures++;

			if (session_failures < 30)
				login_delay = MAX(login_delay,(long) 1);	
						/* 30 seconds at 1 sec each */
			else if (session_failures < 48)
				login_delay = MAX(login_delay,((long) 5));	
						/* another 90 seconds at 
						 * 5 sec each 
						 */
			else if (session->replacement_timeout &&
				 time_before_eq(session->session_drop_time +
						(HZ *
						 session->replacement_timeout),
						jiffies)) {
				login_delay = MAX(login_delay,((long) 10));	
						/* every 10 seconds */
			} else {
				/* after the replacement timeout has expired, 
				 * the device will probably be offline, so we 
				 * probably don't need a session anymore, but 
				 * it's possible the device isn't offline yet 
				 * because of all the hard-coded sleeps in the 
				 * SCSI midlayer after resets occur, and in any 
				 * case it might be useful to know if we ever 
				 * get a session back for debugging
				 * purposes, so we'll keep trying occasionally.
				 */
				login_delay = MAX(login_delay,((long) 60));
			}

			printk
			    ("iSCSI: session iscsi bus %d target id %d has "
			     "ended quickly %lu times, login delay "
			     "%ld seconds\n",
			     session->iscsi_bus, session->target_id,
			     session_failures, login_delay);
		} else {
			/* session lived long enough that the 
			 * target is probably ok 
			 */
			session_failures = 0;
		}

		/* handle any signals that may have occured, 
		 * which may kill the tx thread 
		 */
		iscsi_handle_signals(session);

		/* we need to wait for the tx thread to block before 
		 * trying to complete commands, since it may be using a 
		 * task at the moment, which means we can't complete it yet.
		 * even if the session is terminating, we must wait for 
		 * the tx thread.
		 */
		wait_for_tx_blocked(session);

		spin_lock(&session->task_lock);

		if (session->warm_reset_itt != RSVD_TASK_TAG) {
			printk
			    ("iSCSI: session iscsi bus %d target id %d dropped "
			     "during warm target reset, assuming SCSI commands "
			     "completed by reset\n",
			     session->iscsi_bus, session->target_id);
			session->warm_reset_itt = RSVD_TASK_TAG;
			smp_mb();

			/* FIXME: complete everything with DID_RESET? */
			requeue_or_fail_commands(session);
		} else
		    if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
			requeue_or_fail_commands(session);
		} else if (test_bit(SESSION_LOGGED_OUT, &session->control_bits)) {
			/* the session has logged out, so there shouldn't 
			 * be any tasks, but check anyway 
			 */
			requeue_or_fail_commands(session);
		} else {
			/* session dropped unexpectedly, often due to 
			 * network problems 
			 */
			printk
			    ("iSCSI: session iscsi bus %d target id %d "
			     "to %s dropped\n",
			     session->iscsi_bus, session->target_id,
			     session->log_name);

			/* fail all commands that don't allow retries, 
			 * and requeue everything else 
			 */
			requeue_or_fail_commands(session);
		}

		/* can't send any nop replies now */
		session->nop_reply.ttt = RSVD_TASK_TAG;
		while (session->nop_reply_head) {
			struct iscsi_nop_info *nop_info =
			    session->nop_reply_head;
			session->nop_reply_head = nop_info->next;
			DEBUG_ALLOC("iSCSI: kfree nop_info ttt %u\n",
				    nop_info->ttt);
			kfree(nop_info);
		}
		session->nop_reply_tail = NULL;
		/* a ping test also fails on a connection drop */
		session->ping_test_start = 0;
		session->ping_test_rx_length = 0;
		session->ping_test_data_length = 0;
		kfree(session->ping_test_tx_buffer);

		/* no point trying to logout now */
		session->logout_deadline = 0;
		session->logout_response_deadline = 0;

		/* terminate error recovery and command retries */
		session->mgmt_itt = RSVD_TASK_TAG;
		session->task_mgmt_response_deadline = 0;
		del_timer_sync(&session->busy_task_timer);
		del_timer_sync(&session->busy_command_timer);
		del_timer_sync(&session->immediate_reject_timer);
		del_timer_sync(&session->retry_timer);
		memset(session->luns_timing_out, 0,
		       sizeof (session->luns_timing_out));
		memset(session->luns_doing_recovery, 0,
		       sizeof (session->luns_doing_recovery));
		memset(session->luns_delaying_commands, 0,
		       sizeof (session->luns_delaying_commands));
		session->num_luns_delaying_commands = 0;

		/* we'll never get a reset reply now */
		session->warm_reset_itt = RSVD_TASK_TAG;
		session->reset_response_deadline = 0;

		/* cancel any testing */
		session->ignore_lun = -2;
		session->ignore_completions = 0;
		session->ignore_aborts = 0;
		session->ignore_abort_task_sets = 0;
		session->ignore_lun_resets = 0;
		session->ignore_warm_resets = 0;
		session->ignore_cold_resets = 0;
		session->reject_lun = -2;
		session->reject_aborts = 0;
		session->reject_abort_task_sets = 0;
		session->reject_lun_resets = 0;
		session->reject_warm_resets = 0;
		session->fake_status_lun = -2;
		session->fake_status_unreachable = 0;
		session->fake_status_busy = 0;
		session->fake_status_queue_full = 0;

		spin_unlock(&session->task_lock);
	}

      ThreadExit:
	DEBUG_INIT
	    ("iSCSI: session iscsi bus %d for (%u %u %u *) "
	     "rx thread %d exiting\n",
	     session->iscsi_bus, session->host_no, session->channel,
	     session->target_id, session->rx_pid);
	/* indicate that we're already going down, 
	 * so that we don't get killed 
	 */
	session->rx_pid = 0;
	smp_mb();

	/* this will fail all commands, since the 
	 * SESSION_TERMINATING bit is set 
	 */
	requeue_or_fail_commands(session);

	spin_lock(&session->task_lock);
	/* no point trying to logout now */
	session->logout_deadline = 0;
	session->logout_response_deadline = 0;
	/* terminate error recovery */
	session->mgmt_itt = RSVD_TASK_TAG;
	session->task_mgmt_response_deadline = 0;
	session->reset_response_deadline = 0;
	/* ensure the timers have been deleted before we free their memory */
	del_timer_sync(&session->busy_task_timer);
	del_timer_sync(&session->busy_command_timer);
	del_timer_sync(&session->immediate_reject_timer);
	del_timer_sync(&session->retry_timer);
	memset(session->luns_timing_out, 0, sizeof (session->luns_timing_out));
	memset(session->luns_doing_recovery, 0,
	       sizeof (session->luns_doing_recovery));
	memset(session->luns_delaying_commands, 0,
	       sizeof (session->luns_delaying_commands));
	if (session->preallocated_task) {
		iscsi_task_ctor(session->preallocated_task, NULL, 0);
		kmem_cache_free(session->hba->task_cache,
				session->preallocated_task);
		session->preallocated_task = NULL;
	} else {
		printk
		    ("iSCSI: session iscsi bus %d for (%u %u %u *) terminating,"
		     " but has no preallocated task to free at %lu\n",
		     session->iscsi_bus, session->host_no, session->channel,
		     session->target_id, jiffies);
	}
	spin_unlock(&session->task_lock);

	/* cleanup the socket */
	if (session->socket) {
		/* wait for the tx thread to exit */
		while (session->tx_pid) {
			DEBUG_INIT
			    ("iSCSI: session iscsi bus %d target id %d rx "
			     "thread %d waiting for tx thread %d to exit\n",
			     session->iscsi_bus, session->target_id,
			     current->pid, session->tx_pid);
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(MSECS_TO_JIFFIES(10));
		}

		/* drop the connection */
		iscsi_disconnect(session);
	}

	set_bit(SESSION_TERMINATED, &session->control_bits);

	/* wake up any ioctls sleeping on the session */
	wake_up(&session->login_wait_q);
	up(&session->probe_sem);

	iscsi_remove_luns(session);

	drop_reference(session);

	return 0;
}

int
update_address(struct iscsi_session *session, char *address)
{
	char *tag;
	char *port;
	unsigned char ip[16];
	int ip_length = 4;
	uint32_t a1;

	memset(ip, 0, sizeof (ip));

	if ((tag = strrchr(address, ','))) {
		*tag = '\0';
		tag++;
	}
	if ((port = strrchr(address, ':'))) {
		*port = '\0';
		port++;
	}

	/* update the session's IP address and port, based on the
	 * TargetAddress passed to us.  For now, we can't resolve DNS
	 * names, since that would require us to pass the request up to a
	 * user-mode process, and use the NSS system.  We have our own
	 * equivalent of inet_aton, and fail to change the address if we
	 * get a DNS name instead of an IP address.
	 */
	a1 = iscsi_inet_aton(address);
	ip[0] = a1 & 0xFF;
	ip[1] = (a1 & 0xFF00) >> 8;
	ip[2] = (a1 & 0xFF0000) >> 16;
	ip[3] = (a1 & 0xFF000000) >> 24;

	memcpy(session->ip_address, ip, sizeof (session->ip_address));
	session->ip_length = ip_length;
	if (port)
		session->port = iscsi_strtoul(port, NULL, 0);
	else
		session->port = ISCSI_LISTEN_PORT;

	if (tag)
		session->portal_group_tag = iscsi_strtoul(tag, NULL, 0);


	/* restore the original strings */
	if (tag) {
		--tag;
		*tag = ',';
	}
	if (port) {
		--port;
		*port = ':';
	}

	smp_mb();
	return 1;
}

/* shutdown every session on the HBA */
int
iscsi_shutdown_hba(struct iscsi_hba *hba)
{
	int num_sessions;
	struct iscsi_session *session;
	unsigned long noqueue_flags_;

	/* FIXME: we lose info on LUNs probed when this happens.  After
	 * this, the kernel module must be reloaded in order for another
	 * LUN probe to work correctly.  Just restarting the daemon causes
	 * LUN probe attempts, but the kernel's scsi.c will detect that
	 * the device is already on the HBA's device list and error out 
	 * the add-single-device.  
	 */

	/* ensure no more sessions get added to the HBA while we're 
	 * trying to shut it down 
	 */
	set_bit(ISCSI_HBA_SHUTTING_DOWN, &hba->flags);

	do {
		num_sessions = 0;

		spin_lock_irqsave(&hba->session_lock, noqueue_flags_);
		list_for_each_entry(session, &hba->session_head,
				    session_link) {
			num_sessions++;
			set_bit(SESSION_TERMINATING, &session->control_bits);
			if ((session->last_kill == 0)
			    || time_before_eq(session->last_kill + (5 * HZ),
					      jiffies)) {
				session->last_kill = jiffies;
				iscsi_drop_session(session);
			}
		}
		spin_unlock_irqrestore(&hba->session_lock, noqueue_flags_);

		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(MSECS_TO_JIFFIES(20));

	} while (num_sessions);

	/* let sessions get added again */
	clear_bit(ISCSI_HBA_SHUTTING_DOWN, &hba->flags);
	smp_mb();

	return 1;
}

int
iscsi_shutdown(void)
{
	struct iscsi_session *session;
	int num_sessions;
	pid_t pid;
	unsigned long noqueue_flags_;

	/* terminate every session on every HBA */
	if (this_is_iscsi_boot && !iscsi_system_is_rebooting)
		return 1;

	do {
		num_sessions = 0;

		set_bit(ISCSI_HBA_SHUTTING_DOWN, &iscsi_hba->flags);
		spin_lock_irqsave(&iscsi_hba->session_lock, noqueue_flags_);
		list_for_each_entry(session, &iscsi_hba->session_head, 
				    session_link) {
			if (!this_is_iscsi_boot
			    || iscsi_system_is_rebooting) {
				num_sessions++;
				set_bit(SESSION_TERMINATING,
					&session->control_bits);
				if (session->last_kill == 0) {
					DEBUG_INIT("iSCSI: shutdown killing "
						   "session iscsi bus %d "
						   "target id %d with "
						   "refcount %u\n",
						   session->iscsi_bus,
						   session->target_id,
						   atomic_read(&session->
							        refcount));
					session->last_kill = jiffies;
					/* FIXME: should we try to cleanly 
					 * terminate the session the first time?
					 * May have locking issues with that 
					 */
					iscsi_drop_session(session);
				} else
				    if (time_before_eq
					(session->last_kill + (5 * HZ),
					 jiffies)) {
					printk("iSCSI: shutdown killing "
					       "session iscsi bus %d target "
					       "id %d with refcount %u\n",
					       session->iscsi_bus,
					       session->target_id,
					       atomic_read(&session->refcount));
					session->last_kill = jiffies;
					iscsi_drop_session(session);
				}
			}
		}
		spin_unlock_irqrestore(&iscsi_hba->session_lock, 
				       noqueue_flags_);

		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(MSECS_TO_JIFFIES(20));
	} while (num_sessions);

	/* kill the timer */
	if (!this_is_iscsi_boot || iscsi_system_is_rebooting) {
		if ((pid = iscsi_timer_pid)) {
			printk("iSCSI: driver shutdown killing timer %d\n",
			       pid);
			kill_proc(pid, SIGKILL, 1);
		}

		printk
		    ("iSCSI: driver shutdown waiting for timer to terminate\n");
		while (test_bit(0, &iscsi_timer_running)) {
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(MSECS_TO_JIFFIES(10));
			if (signal_pending(current))
				return 0;
		}
	}

	/* reset LUN probing */
	iscsi_reset_lun_probing();

	/* let sessions get added again later */
	clear_bit(ISCSI_HBA_SHUTTING_DOWN, &iscsi_hba->flags);

	if (!this_is_iscsi_boot || iscsi_system_is_rebooting)
		printk("iSCSI: driver shutdown complete at %lu\n", jiffies);

	return 1;
}

int
iscsi_slave_alloc(Scsi_Device * dev)
{
	struct iscsi_session * session;
	struct iscsi_dev_data *idd;

	session = find_session_by_hba((struct iscsi_hba *)dev->host->hostdata, 
				      dev->channel, dev->id);
	if (session) {
		idd = (struct iscsi_dev_data *) kmalloc(
						        sizeof(struct iscsi_dev_data),
							       GFP_KERNEL);
		if (idd) {
			idd->session = session;
			idd->flags = 0;
			dev->hostdata = idd;
		} else {
			drop_reference(session);
			return -ENOMEM;
		}
	}
				      
	return 0;
}

int
iscsi_slave_configure(Scsi_Device * dev)
{
	unsigned char depth;

	/* select queue depth and tcq for this device */
	if (dev->tagged_supported) {
		depth = ISCSI_CMDS_PER_LUN;
		scsi_adjust_queue_depth(dev, MSG_SIMPLE_TAG, depth);
		if (LOG_ENABLED(ISCSI_LOG_INIT))
			printk
			    ("iSCSI: enabled tagged command queueing for device"
			     " (%u %u %u %u), type 0x%x, depth %u\n",
			     dev->host->host_no, dev->channel, dev->id,
			     dev->lun, dev->type, depth);
	} else if (force_tcq) {
		depth = ISCSI_CMDS_PER_LUN;
		scsi_adjust_queue_depth(dev, MSG_SIMPLE_TAG, depth);
		if (LOG_ENABLED(ISCSI_LOG_INIT))
			printk
			    ("iSCSI: forced tagged command queueing for device "
			     "(%u %u %u %u), type 0x%x, depth %u\n",
			     dev->host->host_no, dev->channel, dev->id,
			     dev->lun, dev->type, depth);
	} else {
		depth = untagged_queue_depth;
		scsi_adjust_queue_depth(dev, 0, depth);
		if (LOG_ENABLED(ISCSI_LOG_INIT))
			printk
			    ("iSCSI: tagged command queueing not supported for "
			     "device (%u %u %u %u), type 0x%x, depth %u\n",
			     dev->host->host_no, dev->channel, dev->id,
			     dev->lun, dev->type, depth);
	}

	return 0;
}

void
iscsi_slave_destroy(Scsi_Device * dev)
{
	struct iscsi_dev_data *idd;

	if ((idd = dev->hostdata)) {
		drop_reference(idd->session);
		kfree(idd);
		dev->hostdata = NULL;
	}
}

/* remove a Scsi_Cmnd from a singly linked list joined by 
 * the host_scribble pointers. 
 */
static int
remove_cmnd(Scsi_Cmnd * sc, Scsi_Cmnd ** head, Scsi_Cmnd ** tail)
{
	if (!sc || !head || !tail) {
		printk("iSCSI: bug - remove_cmnd 0x%x, head %p, tail %p\n",
		       sc->cmnd[0], head, tail);
		return 0;
	}

	if (sc == *head) {
		/* it's the head, remove it */
		*head = (Scsi_Cmnd *) sc->host_scribble;	/* next */
		if (*head == NULL)
			*tail = NULL;
		sc->host_scribble = NULL;
		return 1;
	} else if (*head) {
		Scsi_Cmnd *prior, *next;

		/* try find the command prior to sc */
		prior = *head;
		next = (Scsi_Cmnd *) prior->host_scribble;
		while (next && (next != sc)) {
			prior = next;
			next = (Scsi_Cmnd *) prior->host_scribble;	
							/* next command */
		}
		if (prior && (next == sc)) {
			/* remove the command */
			prior->host_scribble = sc->host_scribble;
			if (*tail == sc)
				*tail = prior;
			sc->host_scribble = NULL;
			return 1;
		}
	}

	return 0;
}

/* unconditionally remove the cmnd from all driver data structures 
 * The probing code uses this when cmnds time out or the probe is killed.
 * It aborts the command on our side, but doesn't inform the target.
 * Since the cmnd is either INQUIRY or REPORT_LUNs, the target should
 * always complete the command, and we just discard the response if
 * it's already been removed from our data structures.
 */
int
iscsi_squash_cmnd(struct iscsi_session *session, Scsi_Cmnd * sc)
{
	struct iscsi_task *task;
	int ret = 0;
	unsigned long noqueue_flags_;

	spin_lock_irqsave(&session->scsi_cmnd_lock, noqueue_flags_);
	if (remove_cmnd
	    (sc, &session->retry_cmnd_head, &session->retry_cmnd_tail)) {
		del_command_timer(sc);
		ret = 1;
	} else
	    if (remove_cmnd
		(sc, &session->scsi_cmnd_head, &session->scsi_cmnd_tail)) {
		del_command_timer(sc);
		ret = 1;
	} else
	    if (remove_cmnd
		(sc, &session->deferred_cmnd_head,
		 &session->deferred_cmnd_tail)) {
		del_command_timer(sc);
		ret = 1;
	}
	spin_unlock_irqrestore(&session->scsi_cmnd_lock, noqueue_flags_);

	if (ret)
		return ret;

	/* remove any task for this cmnd */
	spin_lock(&session->task_lock);
	if ((task = remove_task_for_cmnd(&session->tx_task_head, sc))) {
		/* it's received an R2T, and is queued to have data sent */
		atomic_inc(&task->refcount);
	} else if ((task = find_task_for_cmnd(session, sc))) {
		atomic_inc(&task->refcount);
	}

	if (task) {
		DEBUG_EH
		    ("iSCSI: session iscsi bus %d target id %d "
		     "squashing task itt %u\n",
		     session->iscsi_bus, session->target_id, task->itt);
		remove_session_task(session, task);

		while (atomic_read(&task->refcount) > 1) {
			/* wait for the driver to quit using the task */
			set_current_state(TASK_UNINTERRUPTIBLE);
			schedule_timeout(MSECS_TO_JIFFIES(10));
		}

		/* delete the timers */
		del_task_timer(task);
		del_command_timer(sc);

		/* free the task */
		free_task(session, task);

		ret = 1;
		DEBUG_EH
		    ("iSCSI: session iscsi bus %d target id %d "
		     "squashed task itt %u\n",
		     session->iscsi_bus, session->target_id, task->itt);
	}
	spin_unlock(&session->task_lock);

	if (ret == 0) {
		printk
		    ("iSCSI: session iscsi bus %d target id %d "
		     "couldn't squash cmnd 0x%x\n",
		     session->iscsi_bus, session->target_id, sc->cmnd[0]);
	}
	return ret;
}

int
iscsi_eh_device_reset(Scsi_Cmnd * sc)
{
	struct Scsi_Host *host;
	struct iscsi_hba *hba;
	struct iscsi_session *session;
	int ret = FAILED;
	unsigned long noqueue_flags_;

	if (!sc) {
		printk("iSCSI: device reset, no SCSI command\n");
		return FAILED;
	}
	host = sc->device->host;
	if (!host) {
		printk("iSCSI: device reset, no host for SCSI command 0x%x\n",
		       sc->cmnd[0]);
		return FAILED;
	}
	hba = (struct iscsi_hba *) host->hostdata;
	if (!hba) {
		printk
		    ("iSCSI: device reset, no iSCSI HBA associated "
		     "with SCSI command 0x%x\n",
		     sc->cmnd[0]);
		return FAILED;
	}

	spin_unlock_irq(host->host_lock);

	/* find the appropriate session for the command */
	session = ((struct iscsi_dev_data *)(sc->device->hostdata))->session;

	spin_lock_irqsave(&hba->session_lock, noqueue_flags_);
	if (session)
		atomic_inc(&session->refcount);
	spin_unlock_irqrestore(&hba->session_lock, noqueue_flags_);

	if (session) {
		set_bit(SESSION_RESET_REQUESTED, &session->control_bits);
		printk
		    ("iSCSI: session iscsi bus %d eh_device_reset at %lu "
		     "for command to (%u %u %u %u), cdb 0x%x\n",
		     session->iscsi_bus, jiffies, sc->device->host->host_no,
		     sc->device->channel, sc->device->id, sc->device->lun,
		     sc->cmnd[0]);
		drop_reference(session);
		ret = SUCCESS;
	} else {
		printk
		    ("iSCSI: session iscsi bus %d eh_device_reset "
		     "failed at %lu, no session for command to "
		     "(%u %u %u %u), cdb 0x%x\n",
		     session->iscsi_bus, jiffies, sc->device->host->host_no,
		     sc->device->channel, sc->device->id, sc->device->lun,
		     sc->cmnd[0]);
		ret = FAILED;
	}

	spin_lock_irq(host->host_lock);
	return ret;
}


int
iscsi_eh_bus_reset(Scsi_Cmnd * sc)
{
	struct Scsi_Host *host;
	struct iscsi_hba *hba;
	struct iscsi_session *session;
	unsigned long noqueue_flags_;

	if (!sc) {
		return FAILED;
	}
	host = sc->device->host;
	if (!host) {
		printk("iSCSI: bus reset, no host for SCSI command 0x%x\n",
		       sc->cmnd[0]);
		return FAILED;
	}
	hba = (struct iscsi_hba *) host->hostdata;
	if (!hba) {
		printk
		    ("iSCSI: bus reset, no iSCSI HBA associated with "
		     "SCSI command 0x%x\n",
		     sc->cmnd[0]);
		return FAILED;
	}

	spin_unlock_irq(host->host_lock);

	spin_lock_irqsave(&hba->session_lock, noqueue_flags_);
	list_for_each_entry(session, &hba->session_head, session_link) {
		if (session->channel == sc->device->channel) {
			set_bit(SESSION_RESET_REQUESTED,
				&session->control_bits);
			printk
			     ("iSCSI: session iscsi bus %d eh_bus_reset "
			      "at %lu for command to (%u %u %u %u), cdb 0x%x\n"
			      ,session->iscsi_bus, jiffies,
			      sc->device->host->host_no, sc->device->channel,
			      sc->device->id, sc->device->lun, sc->cmnd[0]);
		}
	}
	spin_unlock_irqrestore(&hba->session_lock, noqueue_flags_);
	spin_lock_irq(host->host_lock);
	return SUCCESS;
}

int
iscsi_eh_host_reset(Scsi_Cmnd * sc)
{
	struct Scsi_Host *host;
	struct iscsi_hba *hba;
	struct iscsi_session *session;
	unsigned long noqueue_flags_;

	if (!sc) {
		return FAILED;
	}
	host = sc->device->host;
	if (!host) {
		printk("iSCSI: host reset, no host for SCSI command 0x%x\n",
		       sc->cmnd[0]);
		return FAILED;
	}
	hba = (struct iscsi_hba *) host->hostdata;
	if (!hba) {
		printk
		    ("iSCSI: host reset, no iSCSI HBA associated with "
		     "SCSI command 0x%x\n",
		     sc->cmnd[0]);
		return FAILED;
	}

	spin_unlock_irq(host->host_lock);

	spin_lock_irqsave(&hba->session_lock, noqueue_flags_);
	list_for_each_entry(session, &hba->session_head, session_link) {
		set_bit(SESSION_RESET_REQUESTED, &session->control_bits);
		printk
		    ("iSCSI: session iscsi bus %d eh_bus_reset at %lu "
		     "for command to (%u %u %u %u), cdb 0x%x\n",
		     session->iscsi_bus, jiffies, sc->device->host->host_no,
		     sc->device->channel, sc->device->id, sc->device->lun, 
		     sc->cmnd[0]);
	}
	spin_unlock_irqrestore(&hba->session_lock, noqueue_flags_);

	spin_lock_irq(host->host_lock);
	return SUCCESS;
}

/* try to queue a command to the session, returning a boolean 
 * indicating success or failure 
 */
int
iscsi_queue(struct iscsi_session *session, Scsi_Cmnd * sc,
	    void (*done) (Scsi_Cmnd *))
{
	unsigned long noqueue_flags_;

	if (session == NULL)
		return 0;

	/* make sure we can complete it properly later */
	sc->scsi_done = done;
	sc->result = 0;

	spin_lock_irqsave(&session->scsi_cmnd_lock, noqueue_flags_);

	if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
		BUG_ON(sc->device == NULL);
		if (LOG_ENABLED(ISCSI_LOG_QUEUE) ||
		    (test_bit(DEVICE_LOG_TERMINATING, device_flags(sc->device))
		     == 0)) {
			/* by default, log this only once per Scsi_Device, 
			 * to avoid flooding the log 
			 */
			printk
			    ("iSCSI: session iscsi bus %d terminating, failing "
			     "to queue cdb 0x%x and any following commands to "
			     "(%u %u %u %u), %s\n",
			     session->iscsi_bus, sc->cmnd[0], session->host_no,
   			     sc->device->channel, sc->device->id, 
			     sc->device->lun, session->log_name);
			set_bit(DEVICE_LOG_TERMINATING,
				 device_flags(sc->device));
		}
		spin_unlock_irqrestore(&session->scsi_cmnd_lock, 
				       noqueue_flags_);
		return 0;
	}

	if (test_bit(SESSION_REPLACEMENT_TIMEDOUT, &session->control_bits)) {
		BUG_ON(sc->device == NULL);
		if ( LOG_ENABLED(ISCSI_LOG_QUEUE) ||
		    (test_bit (DEVICE_LOG_REPLACEMENT_TIMEDOUT,
		      device_flags(sc->device)) == 0)) {
			/* by default, log this only once per Scsi_Device, 
			 * to avoid flooding the log 
			 */
			printk
			    ("iSCSI: session iscsi bus %d replacement "
			     "timed out, failing to queue cdb 0x%x and any "
			     "following commands to (%u %u %u %u), %s\n",
			     session->iscsi_bus, sc->cmnd[0], session->host_no,
			     sc->device->channel, sc->device->id, 
			     sc->device->lun, session->log_name);
			set_bit(DEVICE_LOG_REPLACEMENT_TIMEDOUT,
				device_flags(sc->device));
		}
		spin_unlock_irqrestore(&session->scsi_cmnd_lock, 
				       noqueue_flags_);
		return 0;
	}
#ifdef DEBUG
	/* make sure the command hasn't already been queued */
	{
		Scsi_Cmnd *search = session->scsi_cmnd_head;
		while (search) {
			if (search == sc) {
				printk
				    ("iSCSI: bug - cmnd 0x%x, state %x, "
				     "eh_state %x, scribble %p is already "
				     "queued to session iscsi bus %d "
				     "target id %d\n",
				     sc->cmnd[0], sc->state, sc->eh_state,
				     sc->host_scribble, session->iscsi_bus,
				     session->target_id);
				print_session_cmnds(session);
				spin_unlock_irqrestore(&session->scsi_cmnd_lock,
						       noqueue_flags_);
				ISCSI_TRACE(ISCSI_TRACE_QFailed, sc, NULL,
					    sc->retries, sc->allowed);
				return 0;
			}
			search = (Scsi_Cmnd *) search->host_scribble;
		}
	}
#endif

	/* initialize Scsi_Pointer fields that we might use later */
	memset(&sc->SCp, 0, sizeof (sc->SCp));
	sc->SCp.ptr = (char *) session;
	sc->host_scribble = NULL;

	if (session->print_cmnds > 0) {
		session->print_cmnds--;
		printk
		    ("iSCSI: session iscsi bus %d target id %d "
		     "iscsi_queue printing command at %lu\n",
		     session->iscsi_bus, session->target_id, jiffies);
		print_cmnd(sc);
	}

	/* add a command timer that tells us to fail the command back 
	 * to the OS 
	 */
	DEBUG_QUEUE
	    ("iSCSI: session iscsi bus %d target id %d adding timer to "
	     "command 0x%x at %lu\n",
	     session->iscsi_bus, session->target_id, sc->cmnd[0], jiffies);
	add_command_timer(session, sc, iscsi_command_times_out);

	/* add it to the session's command queue so the tx thread will send it 
	 */
	if (session->scsi_cmnd_head) {
		/* append at the tail */
		session->scsi_cmnd_tail->host_scribble = (unsigned char *) sc;
		session->scsi_cmnd_tail = sc;
	} else {
		/* make it the head */
		session->scsi_cmnd_head = session->scsi_cmnd_tail = sc;
	}
	atomic_inc(&session->num_cmnds);

	DEBUG_QUEUE
	    ("iSCSI: queued 0x%x to session iscsi bus %d target id %d at %lu, "
	     "%u cmnds, head %p, tail %p\n",
	     sc->cmnd[0], session->iscsi_bus, session->target_id, jiffies,
	     atomic_read(&session->num_cmnds), session->scsi_cmnd_head,
	     session->scsi_cmnd_tail);

	ISCSI_TRACE(ISCSI_TRACE_Qd, sc, NULL, sc->retries,
		    sc->timeout_per_command);
	wake_tx_thread(TX_SCSI_COMMAND, session);

	spin_unlock_irqrestore(&session->scsi_cmnd_lock, noqueue_flags_);

	return 1;
}

int
iscsi_queuecommand(Scsi_Cmnd * sc, void (*done) (Scsi_Cmnd *))
{
	struct iscsi_hba *hba;
	struct iscsi_session *session;
	struct Scsi_Host *host;
	int queued = 0;
	int fake_transport_error = 0;

	host = sc->device->host;
	BUG_ON(host == NULL);

	hba = (struct iscsi_hba *) sc->device->host->hostdata;
	if (!test_bit(ISCSI_HBA_ACTIVE, &hba->flags)) {
		ISCSI_TRACE(ISCSI_TRACE_QFailed, sc, NULL, sc->retries,
			    sc->allowed);
		printk("iSCSI: queuecommand but no HBA\n");
		sc->result = host_byte(DID_NO_CONNECT);
		set_lun_comm_failure(sc);
		done(sc);
		return 0;
	}

	if (!iscsi_timer_running) {
		/* iSCSI coming up or going down, fail the command */
		ISCSI_TRACE(ISCSI_TRACE_QFailed, sc, NULL, sc->retries,
			    sc->allowed);
		DEBUG_QUEUE
		    ("iSCSI: no timer, failing to queue cmnd to (%u %u %u %u), "
		     "cdb 0x%x\n",
		     hba->host->host_no, sc->device->channel, 
		     sc->device->id, sc->device->lun, sc->cmnd[0]);
		sc->result = host_byte(DID_NO_CONNECT);
		done(sc);
		return 0;
	}

#ifdef DEBUG
	if (sc->use_sg) {
		int index;
		struct scatterlist *sglist =
		    (struct scatterlist *) sc->request_buffer;
		unsigned int length = 0;
		int bogus = 0;

		/* sanity check the sglist, to make sure the segments 
		 * have at least bufflen 
		 */
		for (index = 0; index < sc->use_sg; index++) {
			length += sglist[index].length;
			if (sglist[index].length == 0)
				bogus = 1;
		}

		if (bogus || (length < sc->request_bufflen)) {
			printk
			    ("iSCSI: attempted to queue at %lu to "
			     "(%u %u %u %u), cdb 0x%x, corrupt sglist, "
			     "sg length %u, buflen %u\n",
			     jiffies, hba->host->host_no, sc->device->channel,
			     sc->device->id, sc->device->lun, sc->cmnd[0],
			     length, sc->request_bufflen);
			print_cmnd(sc);
			ISCSI_TRACE(ISCSI_TRACE_QFailed, sc, NULL, sc->retries,
				    sc->timeout_per_command);
			sc->result = host_byte(DID_NO_CONNECT);
			set_lun_comm_failure(sc);
			done(sc);
			return 0;
		}
	}
#endif

	spin_unlock_irq(host->host_lock);

	DEBUG_QUEUE
	    ("iSCSI: queueing to (%u %u %u %u) at %lu, cdb 0x%x, cpu%d\n",
	     hba->host->host_no, sc->device->channel, sc->device->id,
	     sc->device->lun, jiffies, sc->cmnd[0], smp_processor_id());

	if (!sc->device->hostdata) {

		session = find_session_for_cmnd(sc);
		if (session) {
			struct iscsi_dev_data *idd;
			idd = sc->device->hostdata = (struct iscsi_dev_data *)
						      kmalloc(sizeof(struct
							  	     iscsi_dev_data),
							      GFP_KERNEL);
			if (idd) {
				idd->session = session;
				idd->flags = 0;
			} else {
				drop_reference(session);	
				return -ENOMEM;
			}
		}
	} else {
		unsigned long noqueue_flags_;

		session = ((struct iscsi_dev_data *)
				(sc->device->hostdata))->session;
		spin_lock_irqsave(&hba->session_lock, noqueue_flags_);
		atomic_inc(&session->refcount);
		spin_unlock_irqrestore(&hba->session_lock, noqueue_flags_);
	}

	if (session) {
		DEBUG_QUEUE
		    ("iSCSI: session iscsi bus %d target id %d queuecommand %p"
		     " at %lu, retries %d, allowed %d, timeout %u\n",
		     session->iscsi_bus, session->target_id, sc, jiffies,
		     sc->retries, sc->allowed, sc->timeout_per_command);

		/* record whether I/O commands have been ever been sent 
		 * on this session, to help us decide when we need the 
		 * session and should retry logins regardless
		 * of the login status. Ignore all the commands sent 
		 * by default as part of the LUN being scanned or a 
		 * device being opened, so that sessions that have 
		 * always been idle can be dropped.  Of course, this 
		 * is always true for disks, since Linux will do reads 
		 * looking for a partition table.
		 */
		switch (sc->cmnd[0]) {
			case INQUIRY:
			case REPORT_LUNS:
			case TEST_UNIT_READY:
			case READ_CAPACITY:
			case START_STOP:
			case MODE_SENSE:
				break;
			default:
				session->commands_queued = 1;
				smp_mb();
				break;
		}

		/* For testing, possibly fake transport errors for 
		 * some commands 
		 */
		if (session->fake_not_ready > 0) {
			session->fake_not_ready--;	
			/* not atomic to avoid overhead, 
			 * and miscounts won't matter much 
			 */
			smp_mb();
			fake_transport_error = 1;
		} else {
			/* delete the existing command timer 
			 * before iscsi_queue adds ours 
			 */
			del_command_timer(sc);

			queued = iscsi_queue(session, sc, done);

			if (!queued)
				add_completion_timer(sc);	
					/* need a timer for the 
					 * midlayer to delete 
					 */
		}

		drop_reference(session);
	} else {
		/* couldn't find a session */
		BUG_ON(sc->device == NULL);
		if (LOG_ENABLED(ISCSI_LOG_QUEUE)
		    || (test_bit(DEVICE_LOG_NO_SESSION,
		      device_flags(sc->device)) == 0)) {
			printk
			    ("iSCSI: queuecommand 0x%x failed to find a session"
			     "for HBA (%u %u %u %u)\n", sc->cmnd[0],
			     hba->host->host_no, sc->device->channel,
			     sc->device->id, sc->device->lun);
			set_bit(DEVICE_LOG_NO_SESSION,device_flags(sc->device));
		}
	}
		

	spin_lock_irq(host->host_lock);

	if (fake_transport_error) {
		printk
		    ("iSCSI: session iscsi bus %d target id %d faking transport"
		     " failure for command 0x%x to (%u %u %u %u) at %lu\n",
		     session->iscsi_bus, session->target_id, sc->cmnd[0],
		     sc->device->host->host_no, sc->device->channel, 
		     sc->device->id, sc->device->lun, jiffies);
		/* act as if recv_cmd() received a non-zero iSCSI response */
		memset(sc->sense_buffer, 0, sizeof (sc->sense_buffer));
		set_lun_comm_failure(sc);
		sc->result = host_byte(DID_ERROR)
			     | status_byte(SAM_STAT_CHECK_CONDITION);
		sc->resid = iscsi_expected_data_length(sc);
		done(sc);
	} else if (!queued) {
		DEBUG_QUEUE
		    ("iSCSI: queuecommand completing 0x%x with DID_NO_CONNECT "
		     "at %lu\n",
		     sc->cmnd[0], jiffies);
		ISCSI_TRACE(ISCSI_TRACE_QFailed, sc, NULL, sc->retries,
			    sc->allowed);
		sc->result = host_byte(DID_NO_CONNECT);
		sc->resid = iscsi_expected_data_length(sc);
		set_lun_comm_failure(sc);
		done(sc);
		/* "queued" successfully, and already completed (with a fatal 
		 * error), so we still return 0 
		 */
	}

	return 0;
}

Scsi_Host_Template iscsi_driver_template = {
      .name = "Cisco iSCSI driver",
      .module = THIS_MODULE,
      .queuecommand = iscsi_queuecommand,
      .eh_device_reset_handler = iscsi_eh_device_reset,
      .eh_bus_reset_handler = iscsi_eh_bus_reset,
      .eh_host_reset_handler = iscsi_eh_host_reset,
      .slave_alloc = iscsi_slave_alloc,
      .slave_configure = iscsi_slave_configure,
      .slave_destroy = iscsi_slave_destroy,
      .this_id = -1,
      .can_queue = ISCSI_CANQUEUE,
      .sg_tablesize = ISCSI_MAX_SG,
      .cmd_per_lun = ISCSI_CMDS_PER_LUN,
      .present = 0,
      .unchecked_isa_dma = 0,
      .use_clustering = ENABLE_CLUSTERING,
      .emulated = 1,
};

static int __init
iscsi_init(void)
{
	char *build_str = BUILD_STR;
	int ret = -ENODEV;
	struct Scsi_Host *shost;
	struct iscsi_hba *hba;
	kmem_cache_t *task_cache;

	DEBUG_INIT("iSCSI: init module\n");

	if (build_str) {
		/* developer-built variant of a 4-digit internal release */
		printk
		    ("iSCSI: %d.%d.%d.%d%s variant (%s) built for "
		     "Linux %s\niSCSI: %s\n",
		     DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION,
		     DRIVER_PATCH_VERSION, DRIVER_INTERNAL_VERSION,
		     DRIVER_EXTRAVERSION, ISCSI_DATE, UTS_RELEASE, build_str);
	} else if (DRIVER_INTERNAL_VERSION > 0) {
		/* 4-digit release */
		printk("iSCSI: %d.%d.%d.%d%s (%s) built for Linux %s\n",
		       DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION,
		       DRIVER_PATCH_VERSION, DRIVER_INTERNAL_VERSION,
		       DRIVER_EXTRAVERSION, ISCSI_DATE, UTS_RELEASE);
	} else {
		/* 3-digit release */
		printk("iSCSI: %d.%d.%d%s (%s) built for Linux %s\n",
		       DRIVER_MAJOR_VERSION, DRIVER_MINOR_VERSION,
		       DRIVER_PATCH_VERSION, DRIVER_EXTRAVERSION, ISCSI_DATE,
		       UTS_RELEASE);
	}

	/* pool of iscsi tasks */
	task_cache = kmem_cache_create("iscsi_task_cache",
					sizeof(struct iscsi_task), 0, 
					SLAB_NO_REAP, iscsi_task_ctor, NULL);

	if (!task_cache) {
		printk("iSCSI: kmem_cache_create failed at %lu\n", jiffies);
		return -ENOMEM;
	}

	/* log any module param settings the user has changed */
	if (translate_deferred_sense)
		printk
		    ("iSCSI: will translate deferred sense to current "
		     "sense on disk command responses\n");

	if (force_tcq)
		printk
		    ("iSCSI: will force tagged command queueing "
		     "for all devices\n");

	if (untagged_queue_depth != 1)
		printk("iSCSI: untagged queue depth %d\n",
		       untagged_queue_depth);

	control_major = register_chrdev(0, control_name, &control_fops);
	if (control_major < 0) {
		printk("iSCSI: failed to register the control device\n");
		if (kmem_cache_destroy(task_cache)) {
			printk("iSCSI: kmem_cache_destroy failed for task "
			       "cache %p at %lu\n", task_cache, jiffies);
		}
		return control_major;
	}
	printk("iSCSI: control device major number %d\n", control_major);

	shost = scsi_host_alloc(&iscsi_driver_template,
				sizeof(struct iscsi_hba));
	if (!shost) {
		if (kmem_cache_destroy(task_cache)) {
			printk("iSCSI: kmem_cache_destroy failed for task "
			       "cache %p at %lu\n", task_cache, jiffies);
		}
		return -ENOMEM;
	}

	if (shost->hostt->shost_attrs == NULL)
		shost->hostt->shost_attrs = iscsi_host_attrs;

	snprintf(shost->shost_gendev.bus_id, BUS_ID_SIZE, "iscsi");
	snprintf(shost->shost_classdev.class_id, BUS_ID_SIZE, "iscsi");

	shost->max_id = ISCSI_MAX_TARGET_IDS_PER_BUS;
	shost->max_lun = ISCSI_MAX_LUNS_PER_TARGET;
	shost->max_channel = ISCSI_MAX_CHANNELS_PER_HBA - 1;
	shost->max_cmd_len = ISCSI_MAX_CMD_LEN;
	hba = (struct iscsi_hba *)shost->hostdata;
	memset((void *)hba, 0, sizeof(struct iscsi_hba));
	
	hba->host = shost;
	
	/* list of sessions on this HBA */
	spin_lock_init(&hba->session_lock);
	INIT_LIST_HEAD(&hba->session_head);
	atomic_set(&hba->num_sessions, 0);
	hba->task_cache = task_cache;
	
	set_bit(ISCSI_HBA_ACTIVE, &hba->flags);
	clear_bit(ISCSI_HBA_SHUTTING_DOWN, &hba->flags);
	hba->host_no = shost->host_no;

	/* for now,there's just one iSCSI HBA */
	smp_mb();
	iscsi_hba = hba;
	INIT_LIST_HEAD(&hba->session_head);
	
	ret = scsi_add_host(shost,NULL);
	if (ret) {
		if (kmem_cache_destroy(task_cache)) {
			printk("iSCSI: kmem_cache_destroy failed for task "
			       "cache %p at %lu\n", task_cache, jiffies);
		}
		scsi_host_put(shost);
		printk("iSCSI: failed to register SCSI HBA driver\n");
	} else {
		printk("iSCSI:detected HBA host #%d\n",shost->host_no);
	}
	set_bit(0, &init_module_complete);
	return ret;
}

static void __exit
iscsi_cleanup(void)
{
	pid_t pid = 0;
	struct Scsi_Host *shost = iscsi_hba->host;
	int rc;

	DEBUG_INIT("iSCSI: cleanup module\n");
	if (control_major > 0) {
		rc = unregister_chrdev(control_major, control_name);
		if (rc) {
			printk
			    ("iSCSI: error trying to unregister "
			     "control device\n");
		} else {
			control_major = 0;
		}
	}
	if (shost) {
		struct iscsi_hba *hba;
		
		hba = (struct iscsi_hba *) shost->hostdata;
		BUG_ON(hba == NULL);
		printk("iSCSI: releasing HBA host #%d\n", hba->host->host_no);
		set_bit(ISCSI_HBA_RELEASING, &hba->flags);
		smp_mb();
		
		/* remove all sessions on this HBA, and prevent any from
		 * being added.
		 */
		
		if (!iscsi_shutdown_hba(hba)) {
			printk("iSCSI:can't release HBA host #%u failed to"
			       "shutdown\n",shost->host_no);
			return;
		}

		/* free this HBA's tasks */
		if (hba->task_cache) {
			DEBUG_INIT("iSCSI: HBA host #%u destroying task cache"
				   "%p at %lu\n",hba->host->host_no,
				   hba->task_cache, jiffies);
			if (kmem_cache_destroy(hba->task_cache)) {
				printk("iSCSI: HBA host #%u failed to destroy"
				       " task cache %p at %lu\n",
				       hba->host->host_no, hba->task_cache, 
				       jiffies);
			}
			hba->task_cache = NULL;
		}
		scsi_remove_host(shost);
		scsi_host_put(shost);
		BUG_ON(hba != iscsi_hba);
		iscsi_hba = NULL;
	}

	/* kill the timer */
	if ((pid = iscsi_timer_pid))
		kill_proc(pid, SIGKILL, 1);

	/* wait for the timer to exit */
	while (test_bit(0, &iscsi_timer_running)) {
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(MSECS_TO_JIFFIES(10));
	}

	DEBUG_INIT("iSCSI: cleanup module complete\n");
	return;
}
module_init(iscsi_init);
module_exit(iscsi_cleanup);
