/*
 * iSCSI driver for Linux
 * Copyright (C) 2001 Cisco Systems, Inc.
 * maintained by linux-iscsi-devel@lists.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi-ioctl.c,v 1.1.2.13 2004/01/09 09:29:02 surekhap Exp $
 */

#include <linux/reboot.h>
#include <linux/tcp.h>

#include "iscsi-protocol.h"
#include "iscsi-session.h"
#include "iscsi-probe.h"
#include "iscsi.h"

extern volatile unsigned long iscsi_timer_running;
extern int this_is_iscsi_boot;
extern struct sapiNBP iscsi_inbp_info;
extern struct iscsi_hba *iscsi_hba;
extern char inbp_interface_name[IFNAMSIZ];

static struct notifier_block iscsi_reboot_notifier = {
      notifier_call:iscsi_reboot_notifier_function,
      next:NULL,
      priority:255   /* priority, might need to have a relook at the value */
};


/*
 * This is called as part of ISCSI_SET_NETWORK_BOOT ioctl which gets called
 * only from iscsi-network-boot.c from initrd. iscsid will never call
 * ISCSI_SET_NETWORK_BOOT.
 */
/* For now always returns 0 */
static int 
set_iscsi_network_boot(void)
{
	if (this_is_iscsi_boot != 1) {
		this_is_iscsi_boot = 1;
	
		/* register reboot notifier  */
		if ( register_reboot_notifier (&iscsi_reboot_notifier)) {
			/* FIXME: return error */
			DEBUG_INIT("\niSCSI: register_reboot_notifier"
				   " failed\n");
		} else {
			DEBUG_INIT("\niSCSI: register_reboot_notifier succeeded\n");
		}
	}
	return 0;
}


static int
check_inbp_info(unsigned long *arg)
{
	if (copy_to_user((int *) arg, &this_is_iscsi_boot,
			 sizeof (this_is_iscsi_boot))) {
		printk("iSCSI: Cannot copy out this_is_iscsi_boot variable\n");
		return -EFAULT;
	}
	return 0;
}

static int 
handle_iscsi_probe_luns(unsigned long *arg)
{
	int rc=0;
	struct iscsi_session *session = NULL;
	struct iscsi_probe_luns_ioctl *ioctld = kmalloc(sizeof (*ioctld), 
							GFP_KERNEL);

	if (!ioctld) {
		printk("iSCSI: couldn't allocate space for probe ioctl data\n");
		return -ENOMEM;
	}

	if (copy_from_user(ioctld, (void *) arg, sizeof (*ioctld))) {
		printk("iSCSI: Cannot copy session ioctl data\n");
		kfree(ioctld);
		return -EFAULT;
	}

	if (ioctld->ioctl_size != sizeof (*ioctld)) {
		printk("iSCSI: ioctl size %u incorrect, expecting %u\n",
		       ioctld->ioctl_size, sizeof (*ioctld));
		kfree(ioctld);
		return -EINVAL;
	}

	if (ioctld->ioctl_version != ISCSI_PROBE_LUNS_IOCTL_VERSION) {
		printk("iSCSI: ioctl version %u incorrect, expecting %u\n", 
			ioctld->ioctl_version, ISCSI_PROBE_LUNS_IOCTL_VERSION);
		kfree(ioctld);
		return -EINVAL;
	}

	/* find the session */
	session = find_session_by_bus(ioctld->iscsi_bus, ioctld->target_id);
	if (session == NULL) {
		printk("iSCSI: ioctl probe LUNs (bus %d, target %d) failed, "
		       "no session\n", ioctld->iscsi_bus, ioctld->target_id);
		goto done_probing;
	}

	/* if another ioctl is already trying to probe LUNs, we don't need a 2nd
	 */
	if (test_and_set_bit(SESSION_PROBING_LUNS, &session->control_bits)) {
		DEBUG_INIT("iSCSI: session already has an ioctl probing or "
			   "waiting to probe LUNs for bus %d, target %d\n",
			   ioctld->iscsi_bus, ioctld->target_id);
		rc = 1;
		goto done_probing;
	}

	if (signal_pending(current))
		goto done_probing;
	if (test_bit(SESSION_TERMINATING, &session->control_bits))
		goto done_probing;

	/* if the session has been established before, but isn't established 
	 * now, try to wait for it, at least until we get signalled or the 
	 * session replacement timeout expires.
	 */
	if (!test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
		DEBUG_INIT("iSCSI: session LUN probe ioctl for bus %d, target "
			   "%d waiting for session to be established at %lu\n", 
			   ioctld->iscsi_bus, ioctld->target_id, jiffies);
	}
	if (wait_for_session(session, TRUE)) {
		DEBUG_INIT("iSCSI: session ioctl triggering LUN probe for bus "
			   "%d, target %d at %lu\n",
			   ioctld->iscsi_bus, ioctld->target_id, jiffies);

		iscsi_detect_luns(session);

		if (signal_pending(current))
			goto done_probing;
		if (test_bit
		    (SESSION_TERMINATING,
		     &session->control_bits))
			goto done_probing;

		iscsi_probe_luns(session, session->luns_allowed);
		if (signal_pending(current))
			goto done_probing;
		if (test_bit(SESSION_TERMINATING, &session->control_bits))
			goto done_probing;
		rc = 1;
	} else {
		/* we got signalled or the session replacement timer expired 
		 */
		DEBUG_INIT("iSCSI: session LUN probe ioctl for bus %d, target "
			   "%d failed\n",ioctld->iscsi_bus, ioctld->target_id);
		rc = 0;
	}

      done_probing:
	if (session) {
		clear_bit(SESSION_PROBING_LUNS, &session->control_bits);
		smp_mb();
		drop_reference(session);
	}
	kfree(ioctld);
	return rc;
}

static int 
handle_iscsi_terminate_session(unsigned long *arg)
{
	struct iscsi_session *session = NULL;
	struct iscsi_terminate_session_ioctl ioctld;

	if (copy_from_user(&ioctld, (void *) arg, sizeof (ioctld))) {
		printk("iSCSI: Cannot copy session ioctl data\n");
		return -EFAULT;
	}

	if (ioctld.ioctl_size != sizeof (ioctld)) {
		printk("iSCSI: terminate session ioctl size %u incorrect, "
		       "expecting %Zu\n", ioctld.ioctl_size, sizeof (ioctld));
		return -EINVAL;
	}


	if (ioctld.ioctl_version != ISCSI_TERMINATE_SESSION_IOCTL_VERSION) {
		printk("iSCSI: terminate session ioctl version %u incorrect, "
		       "expecting %u\n", ioctld.ioctl_version,
		       ISCSI_TERMINATE_SESSION_IOCTL_VERSION);
		return -EINVAL;
	}

	/* find the session */
	session = find_session_by_bus(ioctld.iscsi_bus, ioctld.target_id);
	if (session) {
		printk("iSCSI: session bus %d target %d terminated by "
		       "ioctl\n", ioctld.iscsi_bus, ioctld.target_id);
		iscsi_terminate_session(session);
		drop_reference(session);
		return 1;
	} else {
		printk
		    ("iSCSI: terminate session ioctl for bus %d target %d "
		     "failed, no session\n",
		     ioctld.iscsi_bus, ioctld.target_id);
		return 0;
	}
}

static int 
handle_iscsi_ls_target_info(unsigned long * arg)
{
	int target_index;
	int bus_index;
	struct iscsi_session *session;
	struct target_info tmp_buf;
	
	if (copy_from_user(&tmp_buf, (void *) arg, 
			   sizeof (struct target_info))) {
		printk("iSCSI: Cannot copy user-level data\n");
		return -EFAULT;
	}

	target_index = tmp_buf.target_id;
	bus_index = tmp_buf.channel;
	spin_lock(&iscsi_hba->session_lock);
	list_for_each_entry(session, &iscsi_hba->session_head, session_link) {
		if ((session->target_id == target_index) && 
		    (session->channel == bus_index)) {
			tmp_buf.host_no = session->host_no;
			tmp_buf.channel = session->channel;
			strcpy(tmp_buf.target_name, session->TargetName);
			strcpy(tmp_buf.target_alias, session->TargetAlias);
			tmp_buf.num_portals = session->num_portals;
			memcpy(tmp_buf.session_data.isid, session->isid,
			       sizeof (session->isid));
			tmp_buf.session_data.tsid = session->tsid;
			tmp_buf.session_data.addr[0] = session->ip_address[0];
			tmp_buf.session_data.addr[1] = session->ip_address[1];
			tmp_buf.session_data.addr[2] = session->ip_address[2];
			tmp_buf.session_data.addr[3] = session->ip_address[3];
			tmp_buf.session_data.port = session->port;

			tmp_buf.session_data.establishment_time =
						(jiffies - session->
						 session_established_time)/ HZ;
			tmp_buf.session_data.session_drop_time =
						(jiffies - session->
						 session_drop_time) / HZ;
			tmp_buf.session_data.ever_established =
						session->ever_established;
			tmp_buf.session_data.session_alive =
						session->session_alive;
			tmp_buf.session_data.InitialR2T = session->InitialR2T;
			tmp_buf.session_data.ImmediateData =
						session->ImmediateData;
			tmp_buf.session_data.HeaderDigest =
						session->HeaderDigest;
			tmp_buf.session_data.DataDigest = session->DataDigest;
			tmp_buf.session_data.FirstBurstLength =
						session->FirstBurstLength;
			tmp_buf.session_data.MaxBurstLength =
						session->MaxBurstLength;
			tmp_buf.session_data.MaxRecvDataSegmentLength =
						session->
						MaxRecvDataSegmentLength;
			tmp_buf.session_data.MaxXmitDataSegmentLength =
						session->
						MaxXmitDataSegmentLength;
			tmp_buf.session_data.login_timeout =
						session->login_timeout;
			tmp_buf.session_data.auth_timeout =
						session->auth_timeout;
			tmp_buf.session_data.active_timeout =
						session->active_timeout;
			tmp_buf.session_data.idle_timeout =
						session->idle_timeout;
			tmp_buf.session_data.ping_timeout =
						session->ping_timeout;
		}
	}
	spin_unlock(&iscsi_hba->session_lock);

	if (copy_to_user((void *) arg, &tmp_buf, sizeof (struct target_info))) {
		printk("iSCSI: failed to copy target-specific data back to "
		       "user mode\n");
		return -EFAULT;
	}
	return 0;
}

static int 
handle_iscsi_ls_portal_info(unsigned long *arg)
{
	struct iscsi_session *session;
	int flag = 0, num_portals = 0;
	struct iscsi_portal_list portalp;
	struct iscsi_portal_info *portal_copy = NULL;

	if (copy_from_user(&portalp, (void *) arg, sizeof (portalp))) {
		printk("iSCSI: Cannot copy user-level data\n");
		return -EFAULT;
	}

	spin_lock(&iscsi_hba->session_lock);
	list_for_each_entry(session, &iscsi_hba->session_head, session_link) {
		if ((session->target_id == portalp.target_id) &&
		    (session->channel == portalp.bus_id)) {
			portal_copy = (struct iscsi_portal_info *)
				      kmalloc(session->num_portals *
					      sizeof(struct iscsi_portal_info),
						     GFP_ATOMIC);
			if (!portal_copy) {
				printk("\niSCSI: Could not allocate "
				       "memory for portal copy\n");
				return -ENOMEM;
			}

			memcpy(portal_copy, session->portals,
			       session->num_portals * sizeof (struct
						iscsi_portal_info));
			num_portals = session->num_portals;
			flag = 1;
			break;
		}
	}
	spin_unlock(&iscsi_hba->session_lock);

	if (flag == 1) {
		if (copy_to_user(portalp.portals, portal_copy,
				 num_portals * sizeof (*portal_copy))) {
			printk("iSCSI: failed to copy target-specific data back"
			       " to user mode\n");
			kfree(portal_copy);
			return -EFAULT;
		}
		kfree(portal_copy);
		return 0;
	}
	return 0;
}

#if DEBUG_TRACE
static int
handle_iscsi_gettrace(unsigned long *arg)
{
	int rc;
	DECLARE_NOQUEUE_FLAGS;
	struct iscsi_trace_dump dump;
	struct iscsi_trace_dump *user_dump;
	int rc;

	user_dump = (struct iscsi_trace_dump *) arg;

	if (copy_from_user(dump, user_dump, sizeof (*dump))) {
		printk("iSCSI: trace copy_from_user %p, %p, %Zu failed\n",
		       dump, user_dump, sizeof (*dump));
		return -EFAULT;
	}

	if (dump->dump_ioctl_size != sizeof (struct iscsi_trace_dump)) {
		printk("iSCSI: trace dump ioctl size is %Zu, but caller uses "
		       "%u\n", sizeof (struct iscsi_trace_dump),
		       dump->dump_ioctl_size);
		return -EINVAL;
	}

	if (dump->dump_version != TRACE_DUMP_VERSION) {
		printk("iSCSI: trace dump version is %u, but caller uses %u\n",
		       TRACE_DUMP_VERSION, dump->dump_version);
		return -EINVAL;
	}

	if (dump->trace_entry_size != sizeof (struct iscsi_trace_entry)) {
		printk("iSCSI: trace dump ioctl size is %Zu, but caller uses "
		       "%u\n", sizeof (struct iscsi_trace_dump),
		       dump->dump_ioctl_size);
		return -EINVAL;
	}

	if (dump->num_entries < ISCSI_TRACE_COUNT) {
		/* tell the caller to use a bigger buffer */
		dump->num_entries = ISCSI_TRACE_COUNT;
		if (copy_to_user(user_dump, dump, sizeof (*dump)))
			return -EFAULT;
		else
			return -E2BIG;
	}

	/* the caller is responsible for zeroing the buffer 
	 * before the ioctl, so if the caller asks for too many 
	 * entries, it should be able to tell which ones 
	 * actually have data.
	 */

	/* only send what we've got */
	dump->num_entries = ISCSI_TRACE_COUNT;
	if (copy_to_user(user_dump, dump, sizeof (*dump))) {
		printk("iSCSI: trace copy_to_user %p, %p, %Zu failed\n",
		       user_dump, dump, sizeof (*dump));
		return -EFAULT;
	}

	spin_lock_irqsave(&iscsi_trace_lock, noqueue_flags_);
	/* FIXME: copy_to_user may sleep, but we're holding 
	 * a spin_lock with interrupts off 
	 */
	if (copy_to_user(user_dump->trace, &trace_table[0], dump->num_entries *
			 sizeof (struct iscsi_trace_entry))) {
		printk("iSCSI: trace copy_to_user %p, %p, %u failed\n",
		       user_dump->trace, &trace_table[0], dump->num_entries);
		spin_unlock_irqrestore(&iscsi_trace_lock, noqueue_flags_);
		return -EFAULT;
	}
	rc = trace_index;
	printk("iSCSI: copied %d trace entries to %p at %lu\n",
	       dump->num_entries, user_dump->trace, jiffies);
	spin_unlock_irqrestore(&iscsi_trace_lock, noqueue_flags_);

	return rc;
}
#endif

static int
handle_iscsi_establish_session(unsigned long *arg)
{
	struct iscsi_session_ioctl *ioctld = kmalloc(sizeof (*ioctld), 
						     GFP_KERNEL);
	struct iscsi_session *session = NULL;
	struct iscsi_portal_info *portals = NULL;
	int probe_luns, rc = 0;

	if (!ioctld) {
		printk("iSCSI: couldn't allocate space for session "
		       "ioctl data\n");
		return -ENOMEM;
	}
	if (copy_from_user(ioctld, (void *) arg, sizeof (*ioctld))) {
		printk("iSCSI: Cannot copy session ioctl data\n");
		kfree(ioctld);
		return -EFAULT;
	}

	DEBUG_INIT("iSCSI: ioctl establish session to %s at %lu\n",
		   ioctld->TargetName, jiffies);

	if (ioctld->ioctl_size != sizeof (struct iscsi_session_ioctl)) {
		printk("iSCSI: ioctl size %u incorrect, expecting %Zu\n",
		       ioctld->ioctl_size, sizeof (*ioctld));
		kfree(ioctld);
		return -EINVAL;
	}
	if (ioctld->ioctl_version != ISCSI_SESSION_IOCTL_VERSION) {
		printk("iSCSI: ioctl version %u incorrect, expecting %u\n", 
		       ioctld->ioctl_version, ISCSI_SESSION_IOCTL_VERSION);
		kfree(ioctld);
		return -EINVAL;
	}
	if (ioctld->portal_info_size != sizeof (struct iscsi_portal_info)) {
		printk ("iSCSI: ioctl portal info size %u incorrect, "
			"expecting %Zu\n",
			ioctld->portal_info_size, sizeof (*portals));
		kfree(ioctld);
		return -EINVAL;
	}
	if (ioctld->num_portals == 0) {
		printk("iSCSI: ioctl has no portals\n");
		kfree(ioctld);
		return -EINVAL;
	}

	/* allocate the portals */
	if (ioctld->num_portals <= 0) {
		printk("iSCSI: bus %d target %d has no portals in "
		       "session ioctl\n", ioctld->iscsi_bus, ioctld->target_id);
		kfree(ioctld);
		return -EINVAL;
	}
	portals = (struct iscsi_portal_info *) kmalloc(ioctld->num_portals *
						       sizeof(*portals),
						       GFP_KERNEL);
	if (portals == NULL) {
		printk("iSCSI: bus %d target %d cannot allocate %d portals "
		       "for session ioctl\n", ioctld->iscsi_bus, 
			ioctld->target_id, ioctld->num_portals);
		kfree(ioctld);
		return -ENOMEM;
	}
	DEBUG_INIT("iSCSI: bus %d target %d allocated portals %p (size %u) "
		   "at %lu\n", ioctld->iscsi_bus, ioctld->target_id, portals,
		   ioctld->num_portals * sizeof (*portals), jiffies);
	memset(portals, 0, ioctld->num_portals * sizeof (*portals));

	/* copy the portal info from the user ioctl structure */
	if (copy_from_user(portals,
			   (void *) arg + offsetof(struct iscsi_session_ioctl,
			   portals), ioctld->num_portals * sizeof (*portals))) {
		printk ("iSCSI: bus %d target %d cannot copy portal info, "
			"ioctl size %Zu\n", ioctld->iscsi_bus, 
			ioctld->target_id, sizeof (*portals));
		kfree(ioctld);
		kfree(portals);
		return -EFAULT;
	} else {
		DEBUG_ALLOC("iSCSI: copied %u bytes of portal info from "
			    "%p to %p\n",
			    ioctld->num_portals * sizeof (*portals),
			    (void *) arg + offsetof(struct iscsi_session_ioctl,
			    portals), portals);
	}

	/* if this is the daemon's ioctl for a new session
	 * process, then we need to wait for the session to be
	 * established and probe LUNs, regardless of whether 
	 * or not the session already exists, and regardless 
	 * of the config number of any existing session.  This 
	 * is because the config is guaranteed to be the 
	 * newest, regardless of the config number, and 
	 * because the daemon needs to know if the session
	 * failed to start in order for the session process to 
	 * exit with the appropriate exit value.
	 * if this is an update and there is no existing 
	 * session, then we need to create a session, wait for 
	 * it to be established, and probe LUNs for the new 
	 * session.  if this is an update and there is an 
	 * existing session, then we need to update the config 
	 * if the ioctl config number is
	 * greater than the existing session's config number.
	 * Regardless of whether or not the config update was
	 * accepted, LUN probing must occur if requested. If LUN
	 * probing is already in progress, return -EBUSY so 
	 * that the daemon tries again later.  
	 */
	probe_luns = ioctld->probe_luns;

	/* create or update a session */
	do {
		if ((session = find_session_by_bus(ioctld->iscsi_bus,
						   ioctld->target_id))) {
			if (strcmp(ioctld->TargetName,
				   session->TargetName) == 0) {
				rc = update_session(session, ioctld, portals);
				if (rc < 0) {
					/* error out */
					goto done;
				}
			} else {
				/* otherwise error out */
				printk ("iSCSI: bus %d target %d already bound "
					"to %s\n", ioctld->iscsi_bus,
					ioctld->target_id, session->TargetName);
				drop_reference(session);
				kfree(ioctld);
				kfree(portals);
				return 0;
			}
		} else
		    if ((session =
			 allocate_session(ioctld, portals))) {
			/* the config_mutex is initialized to 
			 * locked, any calls to update_session 
			 * block if they see the session we're 
			 * about to add before we've had a 
			 * chance to start the session threads.
			 */
			if (add_session(session)) {
				/* preallocate a task for this 
				 * session to use in case the 
				 * HBA's task_cache ever 
				 * becomes empty, since we can 
				 * use SLAB_KERNEL now, but 
				 * would have to use SLAB_ATOMIC
				 * later.  We have to do this 
				 * after adding the session to 
				 * the HBA, so that a driver 
				 * shutdown can see this 
				 * session and wait for it to 
				 * terminate. Otherwise we'd 
				 * fail to detroy the task cache
				 * because this session still 
				 * had a task allocated, but 
				 * wasn't visible to the 
				 * shutdown code.
				 */
				if ((session->
				     preallocated_task = kmem_cache_alloc(
								session->hba->
								task_cache,
								SLAB_KERNEL))) {
					DEBUG_ALLOC("iSCSI: session iscsi bus "
						    "%d target id %d " 
						    "preallocated task itt %u "
						    "at %lu\n",
						    session->iscsi_bus,
						    session->target_id,
						    session->preallocated_task->itt, 
						    jiffies);
					__set_bit(TASK_PREALLOCATED,
						  &session->preallocated_task->
						  flags);
				} else {
					printk("iSCSI: session iscsi bus %d "
					       "target id %d couldn't "
					       "preallocate task at %lu\n",
					       session->iscsi_bus,
					       session->target_id, jiffies);
					drop_reference(session);
					kfree(ioctld);
					return -ENOMEM;
				}

				/* give the caller the host 
				 * and channel numbers we just 
				 * claimed 
				 */
				ioctld->host_number = session->host_no;
				ioctld->channel = session->channel;

				/* unless we already have one, 
				 * start a timer thread 
				 */
				if (!test_and_set_bit(0, &iscsi_timer_running)) {
					printk("iSCSI: starting timer "
						   "thread at %lu\n", jiffies);
					if (kernel_thread(iscsi_timer_thread,
							  NULL, 0) < 0) {
						printk("iSCSI: failed to "
						       "start timer thread at "
						       "%lu\n", jiffies);
						drop_reference(session);
						kfree(ioctld);
						return -ENOMEM;
					}
				}
				/* try to start the threads 
				 * for this session 
				 */
				rc = start_session_threads
						    (session);
				/* unlock the mutex so that any 
				 * waiting or future 
				 * update_session() calls 
				 * can proceed 
				 */
				up(&session->config_mutex);

				if (rc < 0)
					goto done;
				/* we failed to start 
				 * the session threads 
				 */

				/* always probe LUNs when we 
				 * create a new session 
				 */
				probe_luns = 1;
			} else {
				/* some session claimed this 
				 * bus/target id while we were 
				 * allocating a session. Loop, 
				 * so that we either update the 
				 * existing session or error 
				 * out.
				 */
				drop_reference(session);
				kfree(ioctld);
				session = NULL;
			}
		} else {
			/* couldn't allocate a new session. 
			 * error out and let the daemon retry 
			 * the ioctl 
			 */
			kfree(ioctld);
			kfree(portals);
			return -ENOMEM;
		}
	} while (session == NULL);
	if (probe_luns || !ioctld->update) {
		/* wait for the session login to complete */
		DEBUG_INIT("iSCSI: ioctl waiting for session iscsi bus %d "
			   "target id %d at %lu\n",session->iscsi_bus, 
			   session->target_id, jiffies);
		wait_for_session(session, FALSE);
		if (test_bit(SESSION_TERMINATING,&session->control_bits)) {
			printk("iSCSI: session iscsi bus %d target id %d "
			       "terminating, ioctl returning at %lu\n",
			       session->iscsi_bus, session->target_id, jiffies);
			goto done;
		} else if (signal_pending(current)) {
			iscsi_terminate_session(session);
			printk
			    ("iSCSI: session iscsi bus %d target id %d "
			     "ioctl terminated, returning at %lu\n",
			     session->iscsi_bus, session->target_id, jiffies);
			goto done;
		}
	}
	if (probe_luns) {
		/* if another ioctl is already trying to probe 
		 * LUNs, must wait for it to finish 
		 */
		if (test_and_set_bit(SESSION_PROBING_LUNS,
				     &session->control_bits)) {
			DEBUG_INIT("iSCSI: session already has an ioctl "
				   "probing or waiting to probe LUNs for "
				   "bus %d, target %d\n", ioctld->iscsi_bus,
				   ioctld->target_id);
			rc = -EBUSY;
			goto done;
		}
		/* first figure out what LUNs actually exist */
		iscsi_detect_luns(session);
		if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
			printk("iSCSI: session iscsi bus %d target id %d "
			       "terminating, ioctl returning at %lu\n",
			       session->iscsi_bus, session->target_id, jiffies);
			clear_bit(SESSION_PROBING_LUNS, &session->control_bits);
			smp_mb();
			goto done;
		} else if (signal_pending(current)) {
			iscsi_terminate_session(session);
			printk("iSCSI: session iscsi bus %d target id %d "
			       "ioctl terminated, returning at %lu\n",
			       session->iscsi_bus, session->target_id, jiffies);
			clear_bit(SESSION_PROBING_LUNS, &session->control_bits);
			smp_mb();
			goto done;
		}

		/* and then try to probe the intersection of the
		 * allowed and detected LUNs 
		 */
		iscsi_probe_luns(session, session->luns_allowed);

		/* and then we're done */
		clear_bit(SESSION_PROBING_LUNS,
			  &session->control_bits);
		smp_mb();

		if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
			printk("iSCSI: session iscsi bus %d target id %d "
			       "terminating, ioctl returning at %lu\n",
			       session->iscsi_bus, session->target_id, jiffies);
			goto done;
		} else if (signal_pending(current)) {
			iscsi_terminate_session(session);
			printk ("iSCSI: session iscsi bus %d target id %d "
				"ioctl terminated, returning at %lu\n",
				session->iscsi_bus,session->target_id, jiffies);
			goto done;
		}
	}

	rc = 1;

      done:
	/* pass back the TargetAlias to the caller */
	memcpy(ioctld->TargetAlias, session->TargetAlias,
	       MIN(sizeof (ioctld->TargetAlias),
	       sizeof (session->TargetAlias)));
	ioctld->TargetAlias[sizeof (ioctld->TargetAlias) - 1] = '\0';
	if (copy_to_user ((void *) arg, ioctld, sizeof (*ioctld))) {
		printk("iSCSI: failed to copy ioctl data back to user mode for "
		       "session iscsi bus %d target id %d\n",
		       session->iscsi_bus, session->target_id);
		kfree(ioctld);
		drop_reference(session);
		return -EFAULT;
	}
	kfree(ioctld);

	drop_reference(session);

	if (signal_pending(current))
		return -EINTR;
	else
		return rc;
}

int
ctl_ioctl(struct inode *inode,
	  struct file *file, unsigned int cmd, unsigned long arg)
{

	switch (cmd) {
	case ISCSI_ESTABLISH_SESSION:
			return (handle_iscsi_establish_session((void *)arg));
	case ISCSI_SET_NETWORK_BOOT:
			return (set_iscsi_network_boot());
	case ISCSI_CHECK_INBP_BOOT:
		return(check_inbp_info((void *)arg));
	case ISCSI_SHUTDOWN:
			return(iscsi_shutdown());
	case ISCSI_RESET_PROBING:
			return(iscsi_reset_lun_probing());
	case ISCSI_PROBE_LUNS:
			return(handle_iscsi_probe_luns((void*) arg));
	case ISCSI_TERMINATE_SESSION:
			return(handle_iscsi_terminate_session((void *)arg));
	case ISCSI_GETTRACE:
#if DEBUG_TRACE
			return(handle_iscsi_gettrace((void *) arg));
#else
			printk("iSCSI: iSCSI kernel module does not implement "
			       "tracing\n");
			return -ENXIO;
#endif
	case ISCSI_LS_TARGET_INFO:
			return (handle_iscsi_ls_target_info((void *) arg));
	case ISCSI_LS_PORTAL_INFO:
			return (handle_iscsi_ls_portal_info((void *) arg));
	}

	printk("iSCSI: Requested ioctl not found\n");
	return -EINVAL;
}
