/*
 * iSCSI driver for Linux
 * Copyright (C) 2001 Cisco Systems, Inc.
 * maintained by linux-iscsi-devel@lists.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi-portal.c,v 1.1.2.5 2003/12/16 05:59:53 smhatre Exp $
 */


#include <linux/kernel.h>
#include <linux/inet.h>

#include "iscsi-protocol.h"
#include "iscsi-session.h"
#include "iscsi.h"

/* caller must hold the session's portal_lock */
unsigned int
find_portal(struct iscsi_session *session, unsigned char *ip_address,
	    int ip_length, int port)
{
	struct iscsi_portal_info *portals = session->portals;
	unsigned int p;

	for (p = 0; p < session->num_portals; p++) {
		if (portals[p].ip_length == 0)
			continue;

		if (portals[p].ip_length != ip_length)
			continue;

		if (portals[p].port != port)
			continue;

		if (memcmp(portals[p].ip_address, ip_address, ip_length))
			continue;

		DEBUG_INIT ("iSCSI: session iscsi bus %d target id %d "
			    "found portal %u\n",
			    session->iscsi_bus, session->target_id, p);
		break;
	}

	if (p < session->num_portals)
		return p;

	return UINT_MAX;
}

void
set_portal_config(struct iscsi_session *session, unsigned int p)
{
	/* Set the session timeouts and iSCSI op params based on the portal's 
	 * settings. Don't change the address, since a termporary redirect may 
	 * have already changed the address, and we want to use the redirected 
	 * address rather than the portal's address.
	 */
	session->login_timeout = session->portals[p].login_timeout;
	session->auth_timeout = session->portals[p].auth_timeout;
	session->active_timeout = session->portals[p].active_timeout;
	session->idle_timeout = session->portals[p].idle_timeout;
	session->ping_timeout = session->portals[p].ping_timeout;
	session->abort_timeout = session->portals[p].abort_timeout;
	session->reset_timeout = session->portals[p].reset_timeout;
	session->replacement_timeout = session->portals[p].replacement_timeout;

	/* FIXME: get the scsi_cmnd_lock when setting these? */
	session->min_disk_command_timeout =
	    session->portals[p].min_disk_command_timeout;
	session->max_disk_command_timeout =
	    session->portals[p].max_disk_command_timeout;

	session->InitialR2T = session->portals[p].InitialR2T;
	session->ImmediateData = session->portals[p].ImmediateData;
	session->MaxRecvDataSegmentLength =
	    session->portals[p].MaxRecvDataSegmentLength;
	session->FirstBurstLength = session->portals[p].FirstBurstLength;
	session->MaxBurstLength = session->portals[p].MaxBurstLength;
	session->DefaultTime2Wait = session->portals[p].DefaultTime2Wait;
	session->DefaultTime2Retain = session->portals[p].DefaultTime2Retain;

	session->HeaderDigest = session->portals[p].HeaderDigest;
	session->DataDigest = session->portals[p].DataDigest;

	session->portal_group_tag = session->portals[p].tag;

	/* TCP options */
	session->tcp_window_size = session->portals[p].tcp_window_size;
	/* FIXME: type_of_service */
}

/* caller must hold the session's portal_lock */
int
set_portal(struct iscsi_session *session, unsigned int p)
{
	struct iscsi_portal_info *portals = session->portals;

	if (portals == NULL) {
		printk ("iSCSI: session iscsi bus %d target id %d has no "
			"portal info, can't set portal %d\n",
			session->iscsi_bus, session->target_id, p);
		return 0;
	}

	if (p >= session->num_portals) {
		printk ("iSCSI: session iscsi bus %d target id %d has "
			"only %d portals, can't set portal %d\n",
			session->iscsi_bus, session->target_id,
			session->num_portals, p);
		return 0;
	}

	session->current_portal = p;

	/* address */
	session->ip_length = portals[p].ip_length;
	memcpy(session->ip_address, portals[p].ip_address,
	       portals[p].ip_length);
	session->port = portals[p].port;

	/* timeouts, operational params, other settings */
	set_portal_config(session, p);

	DEBUG_INIT ("iSCSI: session iscsi bus %d target id %d set to "
		    "portal %d, group %d\n",
		    session->iscsi_bus, session->target_id, 
		    session->current_portal, session->portal_group_tag);

	return 1;
}

void
set_preferred_subnet_bitmap(struct iscsi_session *session)
{
	unsigned int bitmap = 0;
	struct iscsi_portal_info *portals = session->portals;
	unsigned int p;
	uint32_t a1, a2;

	if (portals == NULL) {
		printk ("iSCSI: session iscsi bus %d target id %d has no "
			"portal info, therefore no preferred subnet bitmap\n",
			session->iscsi_bus, session->target_id);
		return;
	}

	a1 = iscsi_inet_aton(session->preferred_subnet);
	a1 &= session->preferred_subnet_mask;

	for (p = 0; p < session->num_portals; p++) {
		a2 = portals[p].ip_address[3] << 24;
		a2 |= portals[p].ip_address[2] << 16;
		a2 |= portals[p].ip_address[1] << 8;
		a2 |= portals[p].ip_address[0];
		a2 &= session->preferred_subnet_mask;

		if (a1 == a2)
			bitmap = bitmap | (1 << (p % MAX_PORTALS));
	}
	session->preferred_subnet_bitmap = bitmap;
}

void
set_preferred_portal_bitmap(struct iscsi_session *session)
{
	unsigned int bitmap = 0;
	struct iscsi_portal_info *portals = session->portals;
	unsigned char ip[16];
	unsigned int p;
	uint32_t a1;

	if (portals == NULL) {
		printk ("iSCSI: session iscsi bus %d target id %d has no "
			"portal info, therefore no preferred portal bitmap\n",
			session->iscsi_bus, session->target_id);
		return;
	}
	a1 = iscsi_inet_aton(session->preferred_portal);

	ip[0] = a1 & 0xFF;
	ip[1] = (a1 & 0xFF00) >> 8;
	ip[2] = (a1 & 0xFF0000) >> 16;
	ip[3] = (a1 & 0xFF000000) >> 24;

	for (p = 0; p < session->num_portals; p++) {
		if (memcmp(ip, portals[p].ip_address, portals[p].ip_length) ==
		    0) {
			bitmap = bitmap | (1 << (p % MAX_PORTALS));
			break;
		}
	}
	session->preferred_portal_bitmap = bitmap;
}

int
get_appropriate_portal(struct iscsi_session *session)
{
	unsigned int p;
	int pp = -1;
	unsigned int portal_bitmap = session->preferred_portal_bitmap;
	unsigned int subnet_bitmap = session->preferred_subnet_bitmap;

	if (!portal_bitmap && !subnet_bitmap)
		return -1;

	for (p = 0; p < session->num_portals; p++) {
		if (portal_bitmap & (1 << (p % MAX_PORTALS))) {
			pp = p;
			break;
		}
	}

	if (pp < 0) {
		for (p = 0; p < session->num_portals; p++) {
			if (subnet_bitmap & (1 << (p % MAX_PORTALS))) {
				pp = p;
				break;
			}
		}
	}
	return pp;
}

/* caller must hold the session's portal_lock */
void
next_portal(struct iscsi_session *session)
{
	unsigned int desired_portal = UINT_MAX;
	int allow_any_tag = 1;
	int current_tag = session->portal_group_tag;

	if (!allow_any_tag && (session->portal_group_tag < 0)) {
		printk
		    ("iSCSI: session iscsi bus %d target id %d "
		     "current portal %u group tag unknown, "
		     "can't switch portals\n",
		     session->iscsi_bus, session->target_id,
		     session->current_portal);
		set_portal(session, session->current_portal);
		return;
	}

	/* requested portals and fallbacks after requested portals 
	 * are handled similarly 
	 */
	if (session->requested_portal != UINT_MAX) {
		DEBUG_INIT
		    ("iSCSI: session iscsi bus %d target id %d requested to "
		     "switch to portal %u\n",
		     session->iscsi_bus, session->target_id,
		     session->requested_portal);
		desired_portal = session->requested_portal;
		session->requested_portal = UINT_MAX;
	} else if (session->fallback_portal != UINT_MAX) {
		DEBUG_INIT
		    ("iSCSI: session iscsi bus %d target id %d falling back to "
		     "portal %u\n",
		     session->iscsi_bus, session->target_id,
		     session->fallback_portal);
		desired_portal = session->fallback_portal;
		session->fallback_portal = UINT_MAX;
	}

	if (desired_portal != UINT_MAX) {
		/* a particular portal has been requested */
		if (desired_portal >= session->num_portals) {
			/* the portal doesn't exist */
			printk ("iSCSI: session iscsi bus %d target id %d "
				"desired portal %u does not exist, staying "
				"with portal %u\n",
				session->iscsi_bus, session->target_id,
				desired_portal, session->current_portal);
			/* don't reset the address, so that we stay wherever 
			 * we are if we can't switch portals 
			 */
			set_portal_config(session, session->current_portal);
		} else if (session->portals[desired_portal].ip_length == 0) {
			/* the requested portal is dead (probably killed 
			 * by a permanent redirect) 
			 */
			printk ("iSCSI: session iscsi bus %d target id %d "
				"desired portal %u is dead, staying with "
				"portal %u\n",
				session->iscsi_bus, session->target_id,
				desired_portal, session->current_portal);
			/* don't reset the address, so that we stay 
			 * wherever we are if we can't switch portals 
			 */
			set_portal_config(session, session->current_portal);
		} else if (!allow_any_tag
			   && (session->portals[desired_portal].tag !=
			       session->portal_group_tag)) {
			/* the requested portal is in the wrong portal group */
			printk ("iSCSI: session iscsi bus %d target id %d "
				"desired portal %u is in portal group %u, "
				"but portal group %u is required, staying with "
				"portal %u\n",
				session->iscsi_bus, session->target_id,
				desired_portal,
				session->portals[desired_portal].tag,
				session->portal_group_tag,
				session->current_portal);
			/* don't reset the address, so that we stay wherever 
			 * we are if we can't switch portals 
			 */
			set_portal_config(session, session->current_portal);
		} else {
			/* try the requested portal */
			session->current_portal = desired_portal;
			set_portal(session, session->current_portal);
		}
	} else if (session->portal_failover) {
		unsigned int p;
		int failed = 1;
		unsigned int bitmap = 0;
		unsigned int num_portals = session->num_portals;

		/* Look for the preferred portal */
		bitmap = session->preferred_portal_bitmap;
		if (bitmap) {
			for (p = 0; p < num_portals; p++) {
				if (bitmap & (1 << (p % MAX_PORTALS))) {
					if (!
					    (session->
					     tried_portal_bitmap & (1 <<
								    (p %
								     MAX_PORTALS))))
					{
						if (session->portals[p].
						    ip_length == 0) {
							/* this portal is dead 
							 * (probably killed by 
							 * a permanent redirect)
							 */
							DEBUG_INIT
							    ("iSCSI: session "
							     "iscsi bus %d "
							     "target id %d "
							     "skipping dead "
							     "portal %u\n",
							     session->iscsi_bus,
							     session->target_id,
							     p);
						} else if (allow_any_tag) {
							/* we can use any portal
							 * group, so a tag 
							 * mismatch isn't a 
							 * problem 
							 */
							session->
							    current_portal = p;
							session->
							    tried_portal_bitmap
							    |=
							    (1 <<
							     (p % MAX_PORTALS));
							failed = 0;
							break;
						} else if (session->portals[p].
							   tag < 0) {
							DEBUG_INIT
							    ("iSCSI: session "
							     "iscsi bus %d "
							     "target id %d "
							     "skipping portal "
							     "%u group unknown,"
							     " must login to "
							     "group %u\n",
							     session->iscsi_bus,
							     session->target_id,
							     p, current_tag);
						} else if (session->portals[p].
							   tag == current_tag) {
							/* tag allowed, 
							 * go ahead and try it 
							 */
							session->
							    current_portal = p;
							session->
							    tried_portal_bitmap
							    |=
							    (1 <<
							     (p % MAX_PORTALS));
							failed = 0;
							break;
						}
					}
				}
			}
		}

		if (failed) {
			/* Look for the portal in the preferred subnet */
			bitmap = session->preferred_subnet_bitmap;
			if (bitmap) {
				for (p = 0; p < num_portals; p++) {
					if (bitmap & (1 << (p % MAX_PORTALS))) {
						if (!
						    (session->
						     tried_portal_bitmap & (1 <<
									    (p %
									     MAX_PORTALS))))
						{
							if (session->portals[p].
							    ip_length == 0) {
								/* this portal 
								 * is dead 
								 * (probably 
								 * killed by a 
								 * permanent 
								 * redirect) 
								 */
								DEBUG_INIT
								    ("iSCSI: "
								     "session "
								     "iscsi bus"
								     " %d "
								     "target id"
								     " %d "
								     "skipping "
								     "dead "
								     "portal "
								     "%u\n",
								     session->
								     iscsi_bus,
								     session->
								     target_id,
								     p);
							} else
							    if (allow_any_tag) {
								/* we can use 
								 * any portal 
								 * group, so a 
								 * tag mismatch
								 * isn't a 
								 * problem 
								 */
								session->
								    current_portal
								    = p;
								session->
								    tried_portal_bitmap
								    |=
								    (1 <<
								     (p %
								      MAX_PORTALS));
								failed = 0;
								break;
							} else if (session->
								   portals[p].
								   tag < 0) {
								DEBUG_INIT
								    ("iSCSI: "
								     "session "
								     "iscsi bus"
								     " %d "
								     "target id"
								     " %d "
								     "skipping "
								     "portal %u"
								     " group "
								     "unknown, "
								     "must "
								     "login to "
								     "group "
								     "%u\n",
								     session->
								     iscsi_bus,
								     session->
								     target_id,
								     p,
								     current_tag);
							} else if (session->
								   portals[p].
								   tag ==
								   current_tag)
							{
								/* tag allowed, 
								 * go ahead and
								 * try it 
								 */
								session->
								    current_portal
								    = p;
								session->
								    tried_portal_bitmap
								    |=
								    (1 <<
								     (p %
								      MAX_PORTALS));
								failed = 0;
								break;
							}
						}
					}
				}
			}
		}

		if (failed) {
			/* Now, look for portal in the rest of the available 
			 * portals 
			 */
      retry:		for (p = 0; p < num_portals; p++) {
				if (!
				    (session->
				     tried_portal_bitmap & (1 <<
							    (p %
							     MAX_PORTALS)))) {
					if (session->portals[p].ip_length == 0) {
						/* this portal is dead 
						 * (probably killed by a 
						 * permanent redirect) 
						 */
						DEBUG_INIT
						    ("iSCSI: session iscsi bus "
						     "%d target id %d skipping "
						     "dead portal %u\n",
						     session->iscsi_bus,
						     session->target_id, p);
					} else if (allow_any_tag) {
						/* we can use any portal group,
						 * so a tag mismatch isn't a 
						 * problem 
						 */
						session->current_portal = p;
						session->tried_portal_bitmap |=
						    (1 << (p % MAX_PORTALS));
						failed = 0;
						break;
					} else if (session->portals[p].tag < 0) {
						DEBUG_INIT
						    ("iSCSI: session iscsi bus "
						     "%d target id %d skipping "
						     "portal %u group unknown, "
						     "must login to group %u\n",
						     session->iscsi_bus,
						     session->target_id, p,
						     current_tag);
					} else if (session->portals[p].tag ==
						   current_tag) {
						/* tag allowed, go ahead 
						 * and try it 
						 */
						session->current_portal = p;
						session->tried_portal_bitmap |=
						    (1 << (p % MAX_PORTALS));
						failed = 0;
						break;
					}
				}
			}

			if (failed) {
				/* We have exhausted all portals, will traverse
				 * again, and try if we can get any active
				 * portal to connect.
				 */
				session->tried_portal_bitmap = 0;
				logmsg(AS_NOTICE,
				       "iSCSI: session %p retrying all the "
				       "portals again, since the portal "
				       "list got exhausted\n", session);
				goto retry;
			}
			
		}

		/* set the portal, even if it hasn't changed, so that we
		 * replace the session's address and undo any temporary
		 * redirects.
		 */
		set_portal(session, session->current_portal);
	}
}

int
same_network_portal(struct iscsi_portal_info *p1, struct iscsi_portal_info *p2)
{
	if (p1->port != p2->port)
		return 0;

	if (p1->tag != p2->tag)
		return 0;

	if (p1->ip_length != p2->ip_length)
		return 0;

	if (memcmp(p1->ip_address, p2->ip_address, p1->ip_length))
		return 0;

	return 1;
}
