/*
 * iSCSI driver for Linux
 * Copyright (C) 2001 Cisco Systems, Inc.
 * maintained by linux-iscsi-devel@lists.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi-recv-pdu.c,v 1.1.2.11 2004/01/30 20:42:06 haapi Exp $
 */

#include "iscsi-login.h"
#include "iscsi-crc.h"
#include "iscsi.h"

#ifndef USE_LOCAL_CRC32C
#include "linux/crypto.h"
#endif

#ifndef UINT32_MAX
# define UINT32_MAX 0xFFFFFFFFU
#endif

/* difference isn't really a defined operation in SNA, but we'd like it so that
 * we can determine how many commands can be queued to a session.
 */
static int
cmdsn_window_size(uint32_t expected, uint32_t max)
{
	if ((expected <= max) && ((max - expected) < SNA32_CHECK)) {
		return (max - expected + 1);
	} else if ((expected > max) && ((expected - max) < SNA32_CHECK)) {
		/* window wraps around */
		return ((UINT32_MAX - expected) + 1 + max + 1);
	} else {
		/* window closed, or numbers bogus */
		return 0;
	}
}

/* possibly update the ExpCmdSN and MaxCmdSN, and peak window sizes */
static void
updateSN(struct iscsi_session *session, uint32_t expcmdsn, uint32_t maxcmdsn)
{
	int window_size;

	/* standard specifies this check for when to update expected and 
	 * max sequence numbers 
	 */
	if (!sna_lt(maxcmdsn, expcmdsn - 1)) {
		if ((expcmdsn != session->ExpCmdSn)
		    && !sna_lt(expcmdsn, session->ExpCmdSn)) {
			session->ExpCmdSn = expcmdsn;
		}
		if ((maxcmdsn != session->MaxCmdSn)
		    && !sna_lt(maxcmdsn, session->MaxCmdSn)) {

			session->MaxCmdSn = maxcmdsn;

			/* look for the peak window size */
			window_size = cmdsn_window_size(expcmdsn, maxcmdsn);
			if (window_size > session->current_peak_window_size)
				session->current_peak_window_size = window_size;

			/* age peak window size info */
			if (time_before
			    (session->window_peak_check + (15 * HZ), jiffies)) {
				session->last_peak_window_size =
				    session->current_peak_window_size;
				session->current_peak_window_size = window_size;
				session->window_peak_check = jiffies;
			}

			/* memory barrier for all of that */
			smp_mb();

			/* wake the tx thread to try sending more commands */
			wake_tx_thread(TX_SCSI_COMMAND, session);
		}

		/* record whether or not the command window for this session 
		 * has closed, so that we can ping the target periodically to 
		 * ensure we eventually find out that the window has re-opened.
		 */
		if (maxcmdsn == expcmdsn - 1) {
			/* record how many times this happens, to see 
			 * how often we're getting throttled 
			 */
			session->window_closed++;
			/* prepare to poll the target to see if 
			 * the window has reopened 
			 */
			session->current_peak_window_size = 0;
			session->last_window_check = jiffies;
			smp_wmb();
			set_bit(SESSION_WINDOW_CLOSED, &session->control_bits);
			smp_mb();
			DEBUG_QUEUE
			    ("iSCSI: session iscsi bus %d target id %d command "
			     "window closed, ExpCmdSN %u, MaxCmdSN %u at %lu\n",
			     session->iscsi_bus, session->target_id,
			     session->ExpCmdSn, session->MaxCmdSn, jiffies);
		} else
		    if (test_bit(SESSION_WINDOW_CLOSED, &session->control_bits))
		{
			DEBUG_QUEUE
			    ("iSCSI: session iscsi bus %d target id %d command "
			     "window opened, ExpCmdSN %u, MaxCmdSN %u at %lu\n",
			     session->iscsi_bus, session->target_id,
			     session->ExpCmdSn, session->MaxCmdSn, jiffies);
			clear_bit(SESSION_WINDOW_CLOSED,
				  &session->control_bits);
			smp_mb();
		} else {
			DEBUG_FLOW
			    ("iSCSI: session iscsi bus %d target id %d"
			     " - ExpCmdSN %u, MaxCmdSN %u at %lu\n",
			     session->iscsi_bus, session->target_id,
			     session->ExpCmdSn, session->MaxCmdSn, jiffies);
		}
	}
}

/* caller must hold the session's task_lock */
static void
trigger_error_recovery(struct iscsi_session *session, unsigned int lun)
{
	struct iscsi_task *t;

	/* stop new tasks from being sent to this LUN */
	set_bit(lun, session->luns_timing_out);
	smp_wmb();

	/* fake timeouts for all tasks to the specified LUN in order to 
	 * trigger error recovery. 
	 */
	DEBUG_EH ("iSCSI: session iscsi bus %d target id %d faking task "
		  "timeouts to trigger error recovery for LUN %u at %lu\n",
		  session->iscsi_bus, session->target_id, lun, jiffies);

	list_for_each_entry(t, &session->task_arrival_order, task_order_link) {
		if ((t->lun == lun) && t->scsi_cmnd
		    && !test_bit(0, &t->timedout)) {
			DEBUG_EH ("iSCSI: session iscsi bus %d target id %d "
				  "faking timeout of task %p itt %u, LUN %u, "
				  "command 0x%x at %lu\n", session->iscsi_bus, 
				  session->target_id, t, t->itt, t->lun, 
				  t->scsi_cmnd->cmnd[0], jiffies);

			/* make the command look like it timedout */
			del_task_timer(t);
			set_bit(0, &t->timedout);
			/* ensure nothing will be completed until 
			 * error recovery finishes 
			 */
			set_bit(lun, session->luns_doing_recovery);
		}
	}
	smp_mb();
	wake_tx_thread(SESSION_TASK_TIMEDOUT, session);
}

static void
retry_immediate_mgmt_pdus(unsigned long arg)
{
	struct iscsi_session *session = (struct iscsi_session *) arg;

	session->immediate_reject_timer.expires = 0;
	smp_mb();
	wake_tx_thread(SESSION_TASK_TIMEDOUT, session);
}

/* try to read an entire login PDU into the buffer, 
 * timing out after timeout seconds 
 */
int
iscsi_recv_pdu(struct iscsi_session *session, struct IscsiHdr *header,
	       int max_header_length, char *data, int max_data_length,
	       int timeout)
{
	struct msghdr msg;
	struct iovec iov[2];
	char padding[PAD_WORD_LEN];
	int rc = 0;
	int data_length;
	int ret = 0;

	if (header == NULL) {
		printk ("iSCSI: session iscsi bus %d target id %d, "
			"can't receive PDU header into NULL\n",
			session->iscsi_bus, session->target_id);
		return 0;
	}

	if (max_header_length < sizeof (*header)) {
		printk ("iSCSI: session iscsi bus %d target id %d, can't "
			"receive %Zu PDU header bytes into %d byte buffer\n",
			session->iscsi_bus, session->target_id, 
			sizeof (*header), max_header_length);
		return 0;
	}

	/* set the timer to implement the timeout requested */
	if (timeout)
		session->login_phase_timer = jiffies + (timeout * HZ);
	else
		session->login_phase_timer = 0;
	smp_mb();
	if (LOG_ENABLED(ISCSI_LOG_LOGIN)) {
		printk ("iSCSI: session iscsi bus %d target id %d trying to "
			"recv login pdu at %lu, timeout at %lu (%d seconds)\n",
			session->iscsi_bus, session->target_id, jiffies,
			session->login_phase_timer, timeout);
	}

	/* read the PDU header */
	memset(iov, 0, sizeof (iov));
	iov[0].iov_base = (void *) header;
	iov[0].iov_len = sizeof (*header);
	memset(&msg, 0, sizeof (struct msghdr));
	msg.msg_iov = iov;
	msg.msg_iovlen = 1;

	rc = iscsi_recvmsg(session, &msg, sizeof (*header));

	/* FIXME: check for additional header segments */

	if (signal_pending(current)) {
		printk ("iSCSI: session iscsi bus %d target id %d "
			"recv_login_pdu timed out at %lu\n",
			session->iscsi_bus, session->target_id, jiffies);
		goto done;
	}

	if (rc != sizeof (*header)) {
		if (rc < 0) {
			char *error = iscsi_strerror(-rc);
			if (error && error[0] != '\0') {
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "recv_login_pdu failed to recv %d login "
				     "PDU bytes, rc %d, %s\n",
				     session->iscsi_bus, session->target_id,
				     iov[0].iov_len, rc, iscsi_strerror(-rc));
			} else {
				printk ("iSCSI: session iscsi bus %d target id "
					"%d recv_login_pdu failed to recv %d "
					"login PDU bytes, rc %d\n",
					session->iscsi_bus, session->target_id,
					iov[0].iov_len, rc);
			}
		} else if (rc == 0) {
			printk ("iSCSI: session iscsi bus %d target id %d "
				"recv_login_pdu: connection closed\n",
				session->iscsi_bus, session->target_id);
		} else {
			/* short reads should be impossible unless a signal
			 * occured, which we already checked for.
			 */
			printk ("iSCSI: bug - session iscsi bus %d target id "
				"%d recv_login_pdu, short read %d of %Zu\n",
				session->iscsi_bus, session->target_id, rc,
				sizeof (*header));
		}
		goto done;
	}
	/* assume a PDU round-trip, connection is ok */
	session->last_rx = jiffies;
	smp_mb();

	/* possibly read PDU data */
	data_length = ntoh24(header->dlength);
	if (data_length) {
		/* check for buffer overflow */
		if (data_length > max_data_length) {
			printk ("iSCSI: session iscsi bus %d target id %d "
				"recv_login_pdu can't read %d bytes of login "
				"PDU data, only %d bytes of buffer available\n",
				session->iscsi_bus, session->target_id,
				data_length, max_data_length);
			goto done;
		}

		/* read the PDU's text data payload */
		memset(&msg, 0, sizeof (struct msghdr));
		msg.msg_iov = iov;
		msg.msg_iovlen = 1;

		memset(iov, 0, sizeof (iov));
		iov[0].iov_base = data;
		iov[0].iov_len = data_length;

		/* handle PDU padding */
		if (data_length % PAD_WORD_LEN) {
			int pad = PAD_WORD_LEN - (data_length % PAD_WORD_LEN);

			iov[1].iov_base = padding;
			iov[1].iov_len = pad;
			msg.msg_iovlen = 2;
			data_length += pad;
		}

		rc = iscsi_recvmsg(session, &msg, data_length);

		if (signal_pending(current)) {
			printk ("iSCSI: session iscsi bus %d target id %d "
				"recv_login_pdu timed out at %lu\n",
				session->iscsi_bus, session->target_id, 
				jiffies);
			goto done;
		}

		if (rc != data_length) {
			if (rc < 0) {
				char *error = iscsi_strerror(-rc);
				if (error && error[0] != '\0') {
					printk ("iSCSI: session iscsi bus %d "
						"target id %d recv_login_pdu "
						"failed to recv %d login data "
						"PDU bytes, rc %d, %s\n",
						session->iscsi_bus,
						session->target_id, data_length,
						rc, iscsi_strerror(-rc));
				} else {
					printk ("iSCSI: session iscsi bus %d "
						"target id %d recv_login_pdu "
						"failed to recv %d login data "
						"PDU bytes, rc %d\n",
						session->iscsi_bus,
						session->target_id, data_length,
						rc);
				}
				ret = rc;
			} else if (rc == 0) {
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "recv_login_pdu: connection closed\n",
				     session->iscsi_bus, session->target_id);
			} else {
				/* short reads should be impossible unless a 
				 * signal occured,
				 * which we already checked for.
				 */
				printk ("iSCSI: bug - session iscsi bus %d "
					"target id %d recv_login_pdu, "
					"short read %d of %d\n",
					session->iscsi_bus, session->target_id, 
					rc, data_length);
			}
			goto done;
		}

		/* assume a PDU round-trip, connection is ok */
		session->last_rx = jiffies;
		smp_mb();
	}

	if (LOG_ENABLED(ISCSI_LOG_LOGIN)) {
		char *text = data;
		char *end = text + ntoh24(header->dlength);
		int show_text = 0;

		if (header->opcode == ISCSI_OP_LOGIN_RSP) {
			struct IscsiLoginRspHdr *login_pdu =
			    (struct IscsiLoginRspHdr *) header;
			/* show the login phases and transit bit */
			printk
			    ("iSCSI: session iscsi bus %d target id %d "
			     "received login pdu response at %lu with current "
			     "stage %d, next %d, transit 0x%x, dlength %d\n",
			     session->iscsi_bus, session->target_id, jiffies,
			     ISCSI_LOGIN_CURRENT_STAGE(login_pdu->flags),
			     ISCSI_LOGIN_NEXT_STAGE(login_pdu->flags),
			     login_pdu->flags & ISCSI_FLAG_LOGIN_TRANSIT,
			     ntoh24(header->dlength));
			show_text = 1;
		} else if (header->opcode == ISCSI_OP_TEXT_RSP) {
			printk ("iSCSI: session iscsi bus %d target id %d "
				"received text pdu response with dlength %d "
				"at %lu\n",
				session->iscsi_bus, session->target_id,
				ntoh24(header->dlength), jiffies);
			show_text = 1;
		} else {
			printk ("iSCSI: session iscsi bus %d target id %d "
				"received pdu with opcode 0x%x, dlength %d at "
				"%lu\n", session->iscsi_bus, session->target_id,
				header->opcode, ntoh24(header->dlength), 
				jiffies);
		}

		/* show all the text that we're sending */
		while (show_text && (text < end)) {
			printk ("iSCSI: session iscsi bus %d target id %d "
				"login resp text: %s\n",
				session->iscsi_bus, session->target_id, text);
			text += strlen(text);
			while ((text < end) && (*text == '\0'))
				text++;
		}
	}

	ret = 1;

      done:
	/* clear the timer */
	session->login_phase_timer = 0;
	smp_mb();
	iscsi_handle_signals(session);

	return ret;
}

void
iscsi_recv_logout(struct iscsi_session *session, struct IscsiLogoutRspHdr *stlh)
{
	updateSN(session, ntohl(stlh->expcmdsn), ntohl(stlh->maxcmdsn));

	/* assume a PDU round-trip, connection is ok */
	session->last_rx = jiffies;
	session->logout_itt = RSVD_TASK_TAG;
	session->logout_response_deadline = 0;
	smp_mb();

	if (test_bit(SESSION_LOGOUT_REQUESTED, &session->control_bits)) {
		switch (stlh->response) {
		case ISCSI_LOGOUT_SUCCESS:
			/* set session's time2wait to zero?  
			 * use DefaultTime2Wait? 
			 */
			session->time2wait = 0;
			printk ("iSCSI: session iscsi bus %d target id %d to "
				"%s logged out at %lu\n",
				session->iscsi_bus, session->target_id,
				session->log_name, jiffies);
			set_bit(SESSION_LOGGED_OUT, &session->control_bits);
			smp_mb();
			iscsi_drop_session(session);
			break;
		case ISCSI_LOGOUT_CID_NOT_FOUND:
			printk ("iSCSI: session iscsi bus %d target id %d "
				"logout failed, cid not found\n",
				session->iscsi_bus, session->target_id);
			iscsi_drop_session(session);
			break;
		case ISCSI_LOGOUT_RECOVERY_UNSUPPORTED:
			printk ("iSCSI: session iscsi bus %d target id %d "
				"logout failed, connection recovery not "
				"supported\n",
				session->iscsi_bus, session->target_id);
			iscsi_drop_session(session);
			break;
		case ISCSI_LOGOUT_CLEANUP_FAILED:
			printk ("iSCSI: session iscsi bus %d target id %d "
				"logout failed, cleanup failed\n",
				session->iscsi_bus, session->target_id);
			iscsi_drop_session(session);
			break;
		default:
			printk ("iSCSI: session iscsi bus %d target id %d "
				"logout failed, response 0x%x\n",
				session->iscsi_bus, session->target_id,
				stlh->response);
			iscsi_drop_session(session);
			break;
		}
	} else {
		printk
		    ("iSCSI: session iscsi bus %d target id %d received logout "
		     "response at %lu, but never sent a login request\n",
		     session->iscsi_bus, session->target_id, jiffies);
		iscsi_drop_session(session);
	}
}

int
iscsi_recv_nop_data(struct iscsi_session *session, unsigned char *buffer,
		    int data_length)
{
	/* read the nop data into the nop_info struct, 
	 * and throw any pad bytes away 
	 */
	struct msghdr msg;
	int bytes_read = 0, rc; 
	int num_bytes = data_length;
	int iovn = 1;
	int pad = (data_length % PAD_WORD_LEN) ? (PAD_WORD_LEN -
					    (data_length % PAD_WORD_LEN)) : 0;
	uint32_t received_crc32c, calculated_crc32c;

	while (bytes_read < num_bytes) {

		/* data */
		session->rx_iov[0].iov_base = buffer + bytes_read;
		session->rx_iov[0].iov_len = data_length - bytes_read;
		num_bytes = data_length - bytes_read;
		iovn = 1;

		if (pad) {
			session->rx_iov[1].iov_base = session->rx_buffer;
			session->rx_iov[1].iov_len = pad;
			num_bytes += pad;
			iovn++;
		}

		if (session->DataDigest == ISCSI_DIGEST_CRC32C) {
			session->rx_iov[1].iov_base = &received_crc32c;
			session->rx_iov[1].iov_len = sizeof (received_crc32c);
			num_bytes += sizeof (received_crc32c);
			iovn++;
		}

		memset(&msg, 0, sizeof (struct msghdr));
		msg.msg_iov = session->rx_iov;
		msg.msg_iovlen = iovn;

		rc = iscsi_recvmsg(session, &msg, num_bytes);
		if (rc <= 0) {
			printk
			    ("iSCSI: session iscsi bus %d target id %d "
			     "recv_nop_data failed to recv %d bytes, rc %d\n",
			     session->iscsi_bus, session->target_id, num_bytes,
			     rc);
			iscsi_drop_session(session);
			return bytes_read;
		}
		if (signal_pending(current)) {
			return bytes_read;
		}

		bytes_read += rc;
	}

	DEBUG_FLOW ("iSCSI: session iscsi bus %d target id %d recv_nop_data "
		    "read %d bytes at %lu\n",
		    session->iscsi_bus, session->target_id, num_bytes, jiffies);

	if (session->DataDigest == ISCSI_DIGEST_CRC32C) {
		calculated_crc32c = iscsi_crc32c(buffer, data_length + pad);
		if (calculated_crc32c != received_crc32c) {
			printk ("iSCSI: session iscsi bus %d target id %d "
				"recv_nop_data DataDigest mismatch, "
				"received 0x%08x, calculated 0x%08x\n",
				session->iscsi_bus, session->target_id,
				received_crc32c, calculated_crc32c);
			/* we're not required to do anything if Nop 
			 * data has a digest error 
			 */
		}
	}

	return data_length;
}

void
iscsi_recv_nop(struct iscsi_session *session, struct IscsiNopInHdr *stnih)
{
	int dlength = ntoh24(stnih->dlength);

	DEBUG_FLOW
	    ("iSCSI: recv_nop for session iscsi bus %d target id %d from %s\n",
	     session->iscsi_bus, session->target_id, session->log_name);

	if (stnih->itt != RSVD_TASK_TAG) {
		/* FIXME: check StatSN */
		session->ExpStatSn = ntohl(stnih->statsn) + 1;
		updateSN(session, ntohl(stnih->expcmdsn),
			 ntohl(stnih->maxcmdsn));
		/* it's a reply to one of our Nop-outs, so there was a 
		 * PDU round-trip, and the connection is ok 
		 */
		session->last_rx = jiffies;
		smp_mb();

		ISCSI_TRACE(ISCSI_TRACE_RxPingReply, NULL, NULL,
			    ntohl(stnih->itt), dlength);

		/* if there is ping data in the reply, check to see 
		 * if it matches what we expect 
		 */
		if (dlength) {
			unsigned long rx_start, rx_stop;

			/* FIXME: make sure the dlength won't 
			 * overflow the buffer 
			 */
			rx_start = jiffies;
			if (iscsi_recv_nop_data
			    (session, session->rx_buffer, dlength) != dlength) {
				return;
			}
			rx_stop = jiffies;

			spin_lock(&session->task_lock);
			if (session->ping_test_rx_start == 0)
				session->ping_test_rx_start = rx_start;

			session->ping_test_rx_length -= dlength;

			if (session->ping_test_rx_length <= 0) {
				/* all done */
				printk ("iSCSI: session iscsi bus %d target id "
					"%d rx Nop data test - rx %d of %d "
					"bytes, start %lu, stop %lu, jiffies "
					"%lu, HZ %u\n",
					session->iscsi_bus, session->target_id,
					session->ping_test_data_length -
					session->ping_test_rx_length,
					session->ping_test_data_length,
					session->ping_test_rx_start, rx_stop,
					rx_stop - session->ping_test_rx_start, 
					HZ);
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "Nop data test %d bytes, start %lu, "
				     "stop %lu, jiffies %lu, HZ %u\n",
				     session->iscsi_bus, session->target_id,
				     session->ping_test_data_length,
				     session->ping_test_start, rx_stop,
				     rx_stop - session->ping_test_start, HZ);
				session->ping_test_start = 0;
				session->ping_test_rx_start = 0;
			}
			spin_unlock(&session->task_lock);
		}
	} else {
		/* FIXME: check StatSN, but don't advance it */
		updateSN(session, ntohl(stnih->expcmdsn),
			 ntohl(stnih->maxcmdsn));
	}

	/* check the ttt to decide whether to reply with a Nop-out */
	if (stnih->ttt != RSVD_TASK_TAG) {
		struct iscsi_nop_info *nop_info;

		ISCSI_TRACE(ISCSI_TRACE_RxNop, NULL, NULL, ntohl(stnih->itt),
			    stnih->ttt);

		if (dlength == 0) {
			/* we preallocate space for one data-less nop reply in 
			 * session structure, to avoid having to invoke kernel
			 * memory allocator in the common case where the target
			 * has at most one outstanding data-less nop reply
			 * requested at any given time.
			 */
			spin_lock(&session->task_lock);
			if ((session->nop_reply.ttt == RSVD_TASK_TAG)
			    && (session->nop_reply_head == NULL)) {
				session->nop_reply.ttt = stnih->ttt;
				memcpy(session->nop_reply.lun, stnih->lun,
				       sizeof (session->nop_reply.lun));
				spin_unlock(&session->task_lock);
				DEBUG_FLOW
				    ("iSCSI: preallocated nop reply for "
				     "ttt %u, dlength %d\n",
				     ntohl(stnih->ttt), dlength);
				wake_tx_thread(TX_NOP_REPLY, session);
				return;
			}
			spin_unlock(&session->task_lock);
		}

		/* otherwise, try to allocate a nop_info struct and 
		 * queue it up 
		 */
		nop_info = kmalloc(sizeof (struct iscsi_nop_info) + dlength,
				   GFP_ATOMIC);
		if (nop_info) {
			nop_info->next = NULL;
			nop_info->ttt = stnih->ttt;
			memcpy(nop_info->lun, stnih->lun,
			       sizeof (nop_info->lun));
			nop_info->dlength = dlength;

			/* try to save any data from the nop for the reply */
			if (dlength) {
				if (iscsi_recv_nop_data
				    (session, nop_info->data,
				     dlength) != dlength) {
					kfree(nop_info);
					return;
				}
			}

			/* queue it up */
			spin_lock(&session->task_lock);
			if (session->nop_reply_head) {
				session->nop_reply_tail->next = nop_info;
				session->nop_reply_tail = nop_info;
			} else {
				session->nop_reply_head =
				    session->nop_reply_tail = nop_info;
			}
			spin_unlock(&session->task_lock);

			DEBUG_FLOW
			    ("iSCSI: queued nop reply for ttt %u, dlength %d\n",
			     ntohl(stnih->ttt), dlength);
			wake_tx_thread(TX_NOP_REPLY, session);
		} else {
			printk ("iSCSI: session iscsi bus %d target id %d "
				"couldn't queue nop reply for ttt %u\n",
				session->iscsi_bus, session->target_id,
				ntohl(stnih->ttt));
		}
	}
}

void
iscsi_recv_cmd(struct iscsi_session *session, struct IscsiScsiRspHdr *stsrh,
	       unsigned char *sense_data)
{
	struct iscsi_task *task;
	Scsi_Cmnd *sc;
	unsigned int senselen = 0;
	uint32_t itt = ntohl(stsrh->itt);

	/* FIXME: check StatSN */
	session->ExpStatSn = ntohl(stsrh->statsn) + 1;
	updateSN(session, ntohl(stsrh->expcmdsn), ntohl(stsrh->maxcmdsn));
	/* assume a PDU round-trip, connection is ok */
	session->last_rx = jiffies;
	smp_mb();

	/* find the task for the itt we received */
	spin_lock(&session->task_lock);
	if ((task = find_session_task(session, itt))) {
		/* task was waiting for this command response */
		__set_bit(TASK_COMPLETED, &task->flags);
		sc = task->scsi_cmnd;

		/* for testing, we may want to ignore this command completion */
		if (session->ignore_completions && ((session->ignore_lun == -1)
						    || (session->ignore_lun ==
							task->lun))) {
			/* for testing, the driver can be told to 
			 * ignore command completion 
			 */
			printk ("iSCSI: session iscsi bus %d recv_cmd ignoring "
				"completion of itt %u, LUN %u, sc cdb 0x%x to "
				"(%u %u %u %u) at %lu\n",
				session->iscsi_bus, itt, task->lun, sc->cmnd[0],
				sc->device->host->host_no, sc->device->channel,
				sc->device->id, sc->device->lun, jiffies);
			session->ignore_completions--;
			spin_unlock(&session->task_lock);
			return;
		}

		del_task_timer(task);

		if (sc == NULL) {
			printk ("iSCSI: session iscsi bus %d target id %d "
				"recv_cmd itt %u, refcount %d, no SCSI "
				"command at %lu\n",
				session->iscsi_bus, session->target_id, itt,
				atomic_read(&task->refcount), jiffies);
			/* this will just wait for the refcount to drop and 
			 * then free the task 
			 */
			complete_task(session, itt);
			return;
		}

		DEBUG_QUEUE ("iSCSI: session iscsi bus %d target id %d "
			     "recv_cmd 0x%x, itt %u, refcount %d\n",
			     session->iscsi_bus, session->target_id, 
			     sc->cmnd[0], itt, atomic_read(&task->refcount));
	} else {
		DEBUG_INIT
		    ("iSCSI: session iscsi bus %d target id %d "
		     "recv_cmd - response for itt %u, but no such task\n",
		     session->iscsi_bus, session->target_id, itt);
		spin_unlock(&session->task_lock);
		return;
	}

	/* check for sense data */
	if ((ntoh24(stsrh->dlength) > 1) && sense_data) {
		/* Sense data format per draft-08, 3.4.6.  2-byte sense length,
		 * then sense data, then iSCSI response data 
		 */
		senselen = (sense_data[0] << 8) | sense_data[1];
		if (senselen > (ntoh24(stsrh->dlength) - 2))
			senselen = (ntoh24(stsrh->dlength) - 2);
		sense_data += 2;
	}

	/* determine command result based on the iSCSI response, 
	 * status, and sense 
	 */
	process_task_response(session, task, stsrh, sense_data, senselen);

#if TEST_PROBE_RECOVERY
	if (test_bit(SESSION_ESTABLISHED, &session->control_bits) && sc &&
	    ((sc->cmnd[0] == REPORT_LUNS) || (sc->cmnd[0] == INQUIRY)) &&
	    (stsrh->cmd_status == 0) &&
	    (task->cmdsn >= ABORT_FREQUENCY) &&
	    ((task->cmdsn % ABORT_FREQUENCY) >= 0)
	    && ((task->cmdsn % ABORT_FREQUENCY) < ABORT_COUNT)) {
		/* don't complete this command, so that 
		 * we can test the probe error handling code. 
		 */
		if ((task = remove_task(&session->completing_tasks, itt))) {
			add_task(&session->rx_tasks, task);
			printk
			    ("iSCSI: ignoring completion of itt %u, CmdSN %u, "
			     "cdb 0x%x to (%u %u %u %u)\n",
			     itt, task->cmdsn, sc->cmnd[0],
			     sc->host->host_no, sc->channel, sc->target,
			     sc->lun);
		}
		atomic_dec(&task->refcount);
		spin_unlock(&session->task_lock);
		return;
	}
#endif

	/* now that we're done with it, try to complete it.  */
	DEBUG_FLOW ("iSCSI: session iscsi bus %d target id %d recv_cmd "
		    "attempting to complete itt %u\n",
		    session->iscsi_bus, session->target_id, itt);
	complete_task(session, itt);
	/* Note: the task_lock will be unlocked by complete_task */
}

void
iscsi_recv_r2t(struct iscsi_session *session, struct IscsiRttHdr *strh)
{
	struct iscsi_task *task = NULL;
	uint32_t itt = ntohl(strh->itt);

	updateSN(session, ntohl(strh->expcmdsn), ntohl(strh->maxcmdsn));
	/* assume a PDU round-trip, connection is ok */
	session->last_rx = jiffies;
	smp_mb();

	spin_lock(&session->task_lock);
	if ((task = find_session_task(session, itt))) {
		if (!test_bit(TASK_WRITE, &task->flags)) {
			/* bug in the target.  the command isn't a write, 
			 * so we have no data to send 
			 */
			printk
			    ("iSCSI: session iscsi bus %d target id %d ignoring"
			     " unexpected R2T for task itt %u, %u bytes @ "
			     "offset %u, ttt %u, not a write command\n",
			     session->iscsi_bus, session->target_id,
			     ntohl(strh->itt), ntohl(strh->data_length),
			     ntohl(strh->data_offset), ntohl(strh->ttt));
			iscsi_drop_session(session);
		} else if (task->scsi_cmnd == NULL) {
			printk
			    ("iSCSI: session iscsi bus %d target id %d ignoring"
			     " R2T for task itt %u, %u bytes @ offset %u, "
			     "ttt %u, no SCSI command\n",
			     session->iscsi_bus, session->target_id,
			     ntohl(strh->itt), ntohl(strh->data_length),
			     ntohl(strh->data_offset), ntohl(strh->ttt));
		} else if (task->ttt != RSVD_TASK_TAG) {
			/* bug in the target.  MaxOutsandingR2T == 1 should 
			 * have prevented this from occuring 
			 */
			printk
			    ("iSCSI: session iscsi bus %d target id %d ignoring"
			     " R2T for task itt %u, %u bytes @ offset %u, "
			     "ttt %u, "
			     "already have R2T for %u @ %u, ttt %u\n",
			     session->iscsi_bus, session->target_id,
			     ntohl(strh->itt), ntohl(strh->data_length),
			     ntohl(strh->data_offset), ntohl(strh->ttt),
			     task->data_length, task->data_offset,
			     ntohl(task->ttt));
		} else {
			/* record the R2T */
			task->ttt = strh->ttt;
			task->data_length = ntohl(strh->data_length);
			task->data_offset = ntohl(strh->data_offset);
			ISCSI_TRACE(ISCSI_TRACE_R2T, task->scsi_cmnd, task,
				    task->data_offset, task->data_length);
			DEBUG_FLOW
			    ("iSCSI: session iscsi bus %d target id %d "
			     "R2T for task itt %u, %u bytes @ offset %u\n",
			     session->iscsi_bus, session->target_id,
			     ntohl(strh->itt), ntohl(strh->data_length),
			     ntohl(strh->data_offset));

			/* even if we've issued an abort task set, we need
			 * to respond to R2Ts for this task, though we can
			 * apparently set the F-bit and terminate the data burst
			 * early.  Rather than hope targets handle that 
			 * correctly, we just send the data requested as usual.
			 */
			add_task(&session->tx_task_head, task);
			wake_tx_thread(TX_DATA, session);
		}
	} else {
		/* the task no longer exists */
		DEBUG_FLOW
		    ("iSCSI: session iscsi bus %d target id %d ignoring R2T "
		     "for itt %u, %u bytes @ offset %u\n",
		     session->iscsi_bus, session->target_id, ntohl(strh->itt),
		     ntohl(strh->data_length), ntohl(strh->data_offset));
	}
	spin_unlock(&session->task_lock);
}

#ifndef USE_LOCAL_CRC32C

void
iscsi_recv_data(struct iscsi_session *session, struct IscsiDataRspHdr *stdrh)
{
	struct iscsi_task *task; 
	Scsi_Cmnd *sc = NULL; 
	struct scatterlist *sglist, *sg, *first_sg = NULL, *last_sg = NULL;
	int length, dlength, remaining, rc, i;
	int bytes_read;
	uint32_t offset, expected_offset = 0;
	unsigned int iovn = 0, pad;
	unsigned int segment_offset;
	struct msghdr msg;
	uint32_t itt = ntohl(stdrh->itt);
	uint8_t *peripheral = NULL;
	uint32_t received_crc32c;
	int fake_data_mismatch = 0;
	int ignore_completion = 0;

	if (stdrh->flags & ISCSI_FLAG_DATA_STATUS) {
		/* FIXME: check StatSN */
		session->ExpStatSn = ntohl(stdrh->statsn) + 1;
	}
	updateSN(session, ntohl(stdrh->expcmdsn), ntohl(stdrh->maxcmdsn));
	/* assume a PDU round-trip, connection is ok */
	session->last_rx = jiffies;
	smp_mb();

	length = dlength = ntoh24(stdrh->dlength);
	offset = ntohl(stdrh->offset);

	/* Compute padding bytes that follow the data */
	pad = dlength % PAD_WORD_LEN;
	if (pad) {
		pad = PAD_WORD_LEN - pad;
	}

	spin_lock(&session->task_lock);

	task = find_session_task(session, itt);
	if (task == NULL) {
		printk ("iSCSI: session iscsi bus %d target id %d recv_data, "
			"no task for itt %u (next itt %u), discarding received "
			"data, offset %u len %u\n",
			session->iscsi_bus, session->target_id, 
			ntohl(stdrh->itt), session->itt, offset, dlength);
	} else if (!test_bit(TASK_READ, &task->flags)) {
		/* we shouldn't be getting Data-in unless it's a read */
		if (task->scsi_cmnd)
			printk
			    ("iSCSI: session iscsi bus %d recv_data itt %u, "
			     "command cdb 0x%02x, dropping session due to "
			     "unexpected Data-in from (%u %u %u %u)\n",
			     session->iscsi_bus, itt,
			     task->scsi_cmnd->cmnd[0], session->host_no,
			     session->channel, session->target_id, task->lun);
		else
			printk
			    ("iSCSI: session iscsi bus %d recv_data itt %u, "
			     "command NULL, dropping session due to unexpected "
			     "Data-in from (%u %u %u %u)\n",
			     session->iscsi_bus, itt, session->host_no,
			     session->channel, session->target_id, task->lun);

		/* print the entire PDU header */
		printk ("iSCSI: bogus Data-in PDU header: itt 0x%0x ttt 0x%0x, "
			"hlength %u dlength %u lun %u, statsn 0x%0x "
			"expcmdsn 0x%0x maxcmdsn 0x%0x datasn 0x%0x, "
			"offset %u residual %u\n",
			ntohl(stdrh->itt), ntohl(stdrh->ttt), stdrh->hlength,
			ntoh24(stdrh->dlength), stdrh->lun[1],
			ntohl(stdrh->statsn), ntohl(stdrh->expcmdsn),
			ntohl(stdrh->maxcmdsn), ntohl(stdrh->datasn),
			ntohl(stdrh->offset), ntohl(stdrh->residual_count));

		iscsi_drop_session(session);
		task = NULL;
		spin_unlock(&session->task_lock);
		return;
	} else {
		/* accept all of the data for this task */
		sc = task->scsi_cmnd;
		expected_offset = task->rxdata;

		if (sc) {
			/* either we'll read it all, or we'll drop the session 
			 * and requeue the command, so it's safe to increment 
			 * the received data count before we actually read 
			 * the data, while we still have the task_lock.
			 */
			task->rxdata += dlength;

			/* ensure the task's command won't be completed 
			 * while we're using it 
			 */
			atomic_inc(&task->refcount);

			DEBUG_FLOW
			    ("iSCSI: session iscsi bus %d target id %d "
			     "recv_data itt %u, sc 0x%x, datasn %u, offset %u, "
			     "dlength %u\n",
			     session->iscsi_bus, session->target_id, itt,
			     sc->cmnd[0], ntohl(stdrh->datasn), offset,
			     dlength);
		} else {
			/* command has already been completed (by a timeout) */
			printk ("iSCSI: session iscsi bus %d target id %d "
				"recv_data itt %u, no SCSI command at %lu\n",
				session->iscsi_bus, session->target_id, itt,
				jiffies);
		}

		/* if there is piggybacked status, ensure that we're not 
		 * delaying commands to this LUN 
		 */
		if (stdrh->flags & ISCSI_FLAG_DATA_STATUS) {
			/* mark the task completed */
			__set_bit(TASK_COMPLETED, &task->flags);

			if (sc && session->ignore_completions
			    && ((session->ignore_lun < 0)
				|| (session->ignore_lun == task->lun))) {
				/* for testing, the driver can be told 
				 * to ignore command completion 
				 */
				printk ("iSCSI: session iscsi bus %d ignoring "
					"completion of itt %u, cmnd cdb 0x%x "
					"to (%u %u %u %u) at %lu\n",
					session->iscsi_bus, itt, sc->cmnd[0],
					session->host_no, session->channel,
					session->target_id, task->lun, jiffies);

				session->ignore_completions--;
				ignore_completion = 1;
			} else {
				del_task_timer(task);

				/* piggybacked status is always good */
				if (test_bit
				    (task->lun,
				     session->luns_delaying_commands)) {
					__clear_bit(task->lun,
						    session->
						    luns_delaying_commands);
					session->num_luns_delaying_commands--;
					DEBUG_RETRY
					    ("iSCSI: session iscsi bus %d no "
					     "longer delaying commands to "
					     "(%u %u %u %u) at %lu\n",
					     session->iscsi_bus,
					     session->host_no, session->channel,
					     session->target_id, task->lun,
					     jiffies);
					if (session->
					    num_luns_delaying_commands == 0) {
						del_timer_sync(&session->
							       retry_timer);
						clear_bit
						    (SESSION_RETRY_COMMANDS,
						     &session->control_bits);
						DEBUG_RETRY
						    ("iSCSI: session iscsi bus "
						     "%d target id %d stopping "
						     "retry timer at %lu\n",
						     session->iscsi_bus,
						     session->target_id,
						     jiffies);
					}
					if (!test_bit
					    (task->lun,
					     session->luns_timing_out)) {
						unsigned long noqueue_flags_;

						spin_lock_irqsave(&session->
								  scsi_cmnd_lock,
								  noqueue_flags_);
						requeue_deferred_commands
						    (session, task->lun);
						spin_unlock_irqrestore(&session->
								       scsi_cmnd_lock,
								       noqueue_flags_);
					}
					smp_mb();
				}
			}
		}

		/* if there is no command, we don't increment the 
		 * task refcount, so we can't keep using it 
		 */
		if (sc == NULL)
			task = NULL;

		/* for testing, possibly fake a digest mismatch */
		if (session->fake_read_data_mismatch > 0) {
			session->fake_read_data_mismatch--;
			fake_data_mismatch = 1;
		}
	}
#if TEST_DELAYED_DATA
	if (task && dlength && ((task->cmdsn % 500) == 0)) {
		printk ("iSCSI: testing delayed data for task itt %u, "
			"cmdsn %u, dlength %u at %lu\n",
			task->itt, task->cmdsn, dlength, jiffies);
		atomic_dec(&task->refcount);
		task = NULL;
		sc = NULL;
		spin_unlock(&session->task_lock);
		session->last_rx = jiffies + (45 * HZ);
		smp_mb();
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(37 * HZ);
		spin_lock(&session->task_lock);
		printk("iSCSI: test of delayed data continuing at %lu\n",
		       jiffies);
	}
#endif
	spin_unlock(&session->task_lock);

	if (sc == NULL)
		goto toss_data;

	/* sanity check the PDU against the command */
	if ((offset + dlength) > sc->request_bufflen) {
		/* buffer overflow, often because of a corrupt PDU header */
		printk
		    ("iSCSI: session iscsi bus %d target id %d recv_data for "
		     "itt %u, cmnd 0x%x, bufflen %u, Data PDU with offset %u "
		     "len %u overflows command buffer, dropping session\n",
		     session->iscsi_bus, session->target_id, itt, sc->cmnd[0],
		     sc->request_bufflen, offset, dlength);

		if (task)
			atomic_dec(&task->refcount);
		iscsi_drop_session(session);
		return;
	} else if (expected_offset != offset) {
		/* if the data arrives out-of-order, it becomes much harder
		 * for us to correctly calculate the residual if we don't get
		 * enough data and also don't get an underflow from the
		 * target.  This can happen if we discard Data PDUs due to
		 * bogus offsets/lengths.  Since we always negotiate for 
		 * Data PDUs in-order, this should never happen, but check
		 * for it anyway.
		 */
		/* buffer overflow, often because of a corrupt PDU header */
		printk
		    ("iSCSI: session iscsi bus %d target id %d recv_data for "
		     "itt %u, cmnd 0x%x, bufflen %u, offset %u does not match "
		     "expected offset %u, dropping session\n",
		     session->iscsi_bus, session->target_id, itt, sc->cmnd[0],
		     sc->request_bufflen, offset, expected_offset);

		if (task)
			atomic_dec(&task->refcount);
		iscsi_drop_session(session);
		return;
	}

	/* configure for receiving the data */
	if (sc->use_sg) {
		int index;

		/* scatter-gather */
		sg = sglist = (struct scatterlist *) sc->request_buffer;
		segment_offset = offset;

		for (index = 0; index < sc->use_sg; index++) {
			if (segment_offset < sglist[index].length)
				break;
			else
				segment_offset -= sglist[index].length;
		}

		if (index >= sc->use_sg) {
			/* didn't find the offset, toss the data and 
			 * let the command underflow 
			 */
			printk ("iSCSI: session iscsi bus %d target id %d "
				"recv_data for itt %u couldn't find offset %u "
				"in sglist %p, sc 0x%x, bufflen %u, use_sg %u, "
				"dropping session\n",
				session->iscsi_bus, session->target_id, 
				task->itt, offset, sglist, sc->cmnd[0], 
				sc->request_bufflen, sc->use_sg);
			print_cmnd(sc);
			ISCSI_TRACE(ISCSI_TRACE_BadOffset, sc, task, offset,
				    sc->request_bufflen);
			/* FIXME: discard the data, or drop the session? */
			atomic_dec(&task->refcount);
			iscsi_drop_session(session);
			return;
		} else {
			remaining = dlength;

			/* setup all the data buffers */
			while (sc && (remaining > 0) && (index < sc->use_sg)) {
				sg = &sglist[index];

				if (!kmap_sg(sg)) {
					printk ("iSCSI: session iscsi bus %d "
						"target id %d recv_data itt %u "
						"failed to map sg %p\n",
						session->iscsi_bus, session->
						target_id, task->itt, sg);
					print_cmnd(sc);
					/* FIXME: discard the data, or 
					 * drop the session? 
					 */
					atomic_dec(&task->refcount);
					iscsi_drop_session(session);
					return;
				} else if (first_sg == NULL) {
					first_sg = sg;
				}

				last_sg = sg;

				/* sanity check the sglist segment length */
				if (sg->length <= segment_offset) {
					/* the sglist is corrupt */
					printk
					    ("iSCSI: session iscsi bus %d "
					     "target id %d recv_data index %d, "
					     "length %u too small for offset "
					     "%u, remaining %d, sglist has "
					     "been corrupted\n",
					     session->iscsi_bus,
					     session->target_id, index,
					     sg->length, segment_offset,
					     remaining);
					print_cmnd(sc);
					ISCSI_TRACE(ISCSI_TRACE_BadRxSeg, sc,
						    task, sg->length,
						    segment_offset);
					/* FIXME: discard the data, 
					 * or drop the session? 
					 */
					atomic_dec(&task->refcount);
					iscsi_drop_session(session);
					return;
				}

				session->rx_iov[iovn].iov_base =
				    sg_virtual_address(sg) + segment_offset;
				session->rx_iov[iovn].iov_len =
				    MIN(remaining,
					(int) (sg->length - segment_offset));
				remaining -= session->rx_iov[iovn].iov_len;

				DEBUG_FLOW
				    ("iSCSI: recv_data itt %u, iov[%2d] = "
				     "sg[%2d] = %p, %u of %u bytes, "
				     "remaining %u\n",
				     itt, iovn, sg - sglist,
				     session->rx_iov[iovn].iov_base,
				     session->rx_iov[iovn].iov_len, sg->length,
				     remaining);
				index++;
				iovn++;
				segment_offset = 0;
			}

			if (remaining != 0) {
				/* we ran out of buffer space with more 
				 * data remaining. This should never happen 
				 * if the Scsi_Cmnd's bufflen matches the 
				 * combined length of the sglist segments.
				 */
				printk ("iSCSI: session iscsi bus %d target id "
					"%d recv_data for cmnd 0x%x, bufflen "
					"%u, offset %u len %u, remaining data "
					"%u, dropping session\n",
					session->iscsi_bus, session->target_id,
					sc->cmnd[0], sc->request_bufflen, 
					offset, dlength, remaining);
				print_cmnd(sc);
				/* FIXME: discard the data, 
				 * or drop the session? 
				 */
				atomic_dec(&task->refcount);
				iscsi_drop_session(session);
				return;
			}
		}
	} else {
		/* no scatter-gather, just read it into the buffer */
		session->rx_iov[0].iov_base = sc->request_buffer + offset;
		session->rx_iov[0].iov_len = dlength;
		iovn = 1;
	}

	if (pad) {
		session->rx_iov[iovn].iov_base = session->rx_buffer;
		session->rx_iov[iovn].iov_len = pad;
		iovn++;
		length += pad;
	}

	if (session->DataDigest == ISCSI_DIGEST_CRC32C) {
		/* If we're calculating a data digest, we need to save the 
		 * pointer and length values in the iovecs before the recvmsg 
		 * modifies them (or walk through the sglist again and 
		 * recalculate them later, which seems inefficient).
		 */
		for (i = 0; i < iovn; i++) {
			session->crc_rx_iov[i].iov_base =
			    session->rx_iov[i].iov_base;
			session->crc_rx_iov[i].iov_len =
			    session->rx_iov[i].iov_len;
		}

		/* and we need to receive the target's digest */
		session->rx_iov[iovn].iov_base = &received_crc32c;
		session->rx_iov[iovn].iov_len = sizeof (received_crc32c);
		iovn++;
		length += sizeof (received_crc32c);
	}

	/* save the address of the first byte of INQUIRY data */
	if ((sc->cmnd[0] == INQUIRY) && (offset == 0) && (dlength > 0))
		peripheral = session->rx_iov[0].iov_base;

	/* accept the data */
	memset(&msg, 0, sizeof (struct msghdr));
	msg.msg_iov = session->rx_iov;
	msg.msg_iovlen = iovn;

	DEBUG_FLOW
	    ("iSCSI: recv_data itt %u calling recvmsg %d bytes, iovn %u, "
	     "rx_iov[0].base = %p\n",
	     itt, dlength + pad, iovn, session->rx_iov[0].iov_base);
	rc = iscsi_recvmsg(session, &msg, length);

	if (rc == length) {
		/* assume a PDU round-trip, connection is ok */
		session->last_rx = jiffies;
		smp_mb();

		if (session->DataDigest == ISCSI_DIGEST_CRC32C) {
			uint32_t calculated_crc32c;
			/*
			 * It is required that the memory not be dually-mapped, 
			 * and we don't need anything mapped after 
			 * iscsi_recvmsg(), anyway.
			 * Now let's inefficiently step back over the sglist 
			 * again so we can use the crypto chksum routines on 
			 * the scatterlist.
			 * All the error checking is omitted because it was done
			 * above.
			 */
			if (first_sg) {
				int index, used_sglen;
				struct crypto_tfm *tfm;
				struct scatterlist tmpsg[2];
				
				tfm = crypto_alloc_tfm("crc32c", 0);
				crypto_chksum_init(tfm);
				
				/* undo any temporary mappings */
				for (sg = first_sg; sg <= last_sg; sg++) {
					kunmap_sg(sg);
				}
				first_sg = NULL;	/* done with sg list */

				/* scatter-gather */
				sg = sglist = (struct scatterlist *) 
							sc->request_buffer;
				segment_offset = offset;

				for (index = 0; index < sc->use_sg; index++) {
					if (segment_offset <
					    sglist[index].length)
						break;
					else
						segment_offset -=
						    sglist[index].length;
				}
				remaining = dlength;
				while (sc && (remaining > 0)
				       && (index < sc->use_sg)) {
					sg = &sglist[index];

					used_sglen = MIN(remaining,
							 (int) (sg->length -
								segment_offset));

					memcpy(tmpsg, sg,
					       sizeof (struct scatterlist));
					/* override fields if we have an 
					 * offset 
					 */
					if (segment_offset
					    || used_sglen != sg->length) {
						tmpsg[0].page =
						    virt_to_page
						    (sg_virtual_address(sg) +
						     segment_offset);
						tmpsg[0].offset =
						    offset_in_page
						    (sg_virtual_address(sg) +
						     segment_offset);
						tmpsg[0].length = used_sglen;
					}
					crypto_chksum_update(tfm, tmpsg, 1);

					remaining -= used_sglen;
					index++;
					segment_offset = 0;
				}
			
				/*
				 * Figure pad
				 */
				if (pad) {
					tmpsg[0].page = virt_to_page(session->
								     rx_buffer);
					tmpsg[0].offset =
					    offset_in_page(session->rx_buffer);
					tmpsg[0].length = pad;
					crypto_chksum_update(tfm, tmpsg, 1);
				}

				crypto_chksum_final(tfm, &calculated_crc32c);
				crypto_free_tfm(tfm);
				tfm = NULL;
			} else {
				calculated_crc32c =
					iscsi_crc32c(session->
						     crc_rx_iov[0].iov_base,
						     session->
						     crc_rx_iov[0].iov_len);

				/* add in all other segments, except 
				 * for the digest itself 
				 */
				for (i = 1; i < iovn - 1; i++) {
					calculated_crc32c =
					    iscsi_crc32c_continued(session->
								   crc_rx_iov
								   [i].iov_base,
								   session->
								   crc_rx_iov
								   [i].iov_len,
								   calculated_crc32c);
				}
			}
			
			if (fake_data_mismatch) {
				printk ("iSCSI: session iscsi bus %d target id "
					"%d faking read DataDigest mismatch "
					"for itt %u\n",
					session->iscsi_bus, session->target_id,
					task->itt);
				calculated_crc32c = 0x01020304;
			}

			if (calculated_crc32c != received_crc32c) {
				unsigned int lun = task->lun;
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "recv_data for itt %u, cmnd 0x%x "
				     "DataDigest mismatch, received 0x%08x, "
				     "calculated 0x%08x, offset %u, pad %u, "
				     "triggering error recovery for LUN %u\n",
				     session->iscsi_bus, session->target_id,
				     itt, sc->cmnd[0], received_crc32c,
				     calculated_crc32c, offset, pad, lun);
				if (first_sg) {
					/* undo any temporary mappings */
					for (sg = first_sg; sg <= last_sg; sg++) {
						kunmap_sg(sg);
					}
					first_sg = NULL;
				}
				/* we MUST abort this task.  To avoid 
				 * reordering, we trigger recovery for all 
				 * tasks to this LUN.  
				 */
				spin_lock(&session->task_lock);
				task->rxdata = 0;
				atomic_dec(&task->refcount);
				trigger_error_recovery(session, lun);
				spin_unlock(&session->task_lock);
				return;
			}
		}
	} else {
		printk ("iSCSI: session iscsi bus %d target id %d recv_data "
			"for itt %u, cmnd 0x%x failed to recv %d data PDU "
			"bytes, rc %d\n",
			session->iscsi_bus, session->target_id, task->itt,
			sc->cmnd[0], length, rc);
		atomic_dec(&task->refcount);
		iscsi_drop_session(session);
		return;
	}

	/* update LUN info based on the INQUIRY data, since we've got it 
	 * mapped now 
	 */
	if (peripheral)
		process_inquiry_data(session, task->scsi_cmnd, peripheral);

	/* done with the data buffers */
	if (first_sg) {
		/* undo any temporary mappings */
		for (sg = first_sg; sg <= last_sg; sg++) {
			kunmap_sg(sg);
		}
	}

	ISCSI_TRACE(ISCSI_TRACE_RxData, sc, task, offset, dlength);

	if ((stdrh->flags & ISCSI_FLAG_DATA_STATUS) && !ignore_completion) {
		unsigned int expected = iscsi_expected_data_length(sc);

		/* we got status, meaning the command completed in a way that
		 * doesn't give us any sense data, and the command must be
		 * completed now, since we won't get a command response PDU.
		 */
		DEBUG_FLOW
		    ("iSCSI: Data-in with status 0x%x for itt %u, sc 0x%x\n",
		     stdrh->cmd_status, ntohl(stdrh->itt),
		     task->scsi_cmnd->cmnd[0]);
		ISCSI_TRACE(ISCSI_TRACE_RxDataCmdStatus, sc, task,
			    stdrh->cmd_status, 0);
		sc->result = host_byte(DID_OK) | status_byte(stdrh->cmd_status);

		spin_lock(&session->task_lock);

		if ((stdrh->flags & ISCSI_FLAG_DATA_OVERFLOW)
		    || (stdrh->flags & ISCSI_FLAG_DATA_UNDERFLOW)
		    || ((test_bit(TASK_READ, &task->flags))
			&& (task->rxdata != expected))) {
			if (LOG_ENABLED(ISCSI_LOG_QUEUE)
			    || LOG_ENABLED(ISCSI_LOG_FLOW)) {
				printk
				    ("iSCSI: session iscsi bus %d task itt %u "
				     "to (%u %u %u %u), cdb 0x%x, %c%c %s, "
				     "received %u, residual %u, expected %u\n",
				     session->iscsi_bus, task->itt,
				     sc->device->host->host_no,
				     sc->device->channel, sc->device->id,
				     sc->device->lun, sc->cmnd[0],
				     (stdrh->flags & ISCSI_FLAG_DATA_OVERFLOW) ?
				     'O' : ' ',
				     (stdrh->flags & ISCSI_FLAG_DATA_UNDERFLOW)
				     ? 'U' : ' ', (stdrh-> flags & ISCSI_FLAG_DATA_OVERFLOW) ?
				     "overflow" : "underflow", task->rxdata,
				     ntohl(stdrh->residual_count), expected);
			}

			if (stdrh->flags & ISCSI_FLAG_DATA_UNDERFLOW) {
				ISCSI_TRACE(ISCSI_TRACE_RxUnderflow, sc, task,
					    ntohl(stdrh->residual_count),
					    expected);
				sc->resid = ntohl(stdrh->residual_count);
			} else if (stdrh->flags & ISCSI_FLAG_DATA_OVERFLOW) {
				/* FIXME: not sure how to tell the SCSI layer 
				 * of an overflow, so just give it an error 
				 */
				ISCSI_TRACE(ISCSI_TRACE_RxOverflow, sc, task,
					    ntohl(stdrh->residual_count),
					    expected);
				sc->result =
				    host_byte(DID_ERROR) | status_byte(stdrh->
								       cmd_status);
			} else {
				/* All the read data did not arrive */
				ISCSI_TRACE(ISCSI_TRACE_HostUnderflow, sc, task,
					    task->rxdata, expected);
				/* we don't know which parts of the buffer 
				 * didn't get data, so report the whole buffer 
				 * missing 
				 */
				sc->resid = expected;
			}
		}

		/* done using the command's data buffers and structure fields */
		atomic_dec(&task->refcount);

		/* try to complete the task.  complete_task expects the 
		 * task_lock held, but returns with it unlocked 
		 */
		complete_task(session, itt);
	} else {
		/* done modifying the command and task */
		atomic_dec(&task->refcount);
	}

	return;

      toss_data:
	/* just throw away the PDU */
	if (first_sg) {
		/* undo any temporary mappings */
		for (sg = first_sg; sg <= last_sg; sg++) {
			kunmap_sg(sg);
		}
	}

	bytes_read = 0;
	length = dlength + pad;
	if (session->DataDigest == ISCSI_DIGEST_CRC32C) {
		printk ("iSCSI: session iscsi bus %d target id %d recv_data "
			"discarding %d data PDU bytes, %d pad bytes, %Zu "
			"digest bytes\n",
			session->iscsi_bus, session->target_id, dlength, pad,
			sizeof (received_crc32c));
		length += sizeof (received_crc32c);
	} else {
		printk ("iSCSI: session iscsi bus %d target id %d recv_data "
			"discarding %d data PDU bytes, %d pad bytes\n",
			session->iscsi_bus, session->target_id, dlength, pad);
	}

	while (!signal_pending(current) && (bytes_read < length)) {
		int num_bytes =
		    MIN(length - bytes_read,(int) sizeof (session->rx_buffer));

		/* FIXME: can we use the same rx_buffer in all the iovecs, 
		 * since we're discarding the data anyway? 
		 * That would reduce the number of recvmsg calls we have 
		 * to make.
		 */
		session->rx_iov[0].iov_base = session->rx_buffer;
		session->rx_iov[0].iov_len = sizeof (session->rx_buffer);
		memset(&msg, 0, sizeof (struct msghdr));
		msg.msg_iov = &session->rx_iov[0];
		msg.msg_iovlen = 1;
		rc = iscsi_recvmsg(session, &msg, num_bytes);
		if (rc <= 0) {
			printk ("iSCSI: session iscsi bus %d target id %d "
				"recv_data failed to recv and discard %d data "
				"PDU bytes, rc %d, bytes_read %d\n",
				session->iscsi_bus, session->target_id, 
				length, rc, bytes_read);
			iscsi_drop_session(session);
		} else {
			/* assume a PDU round-trip, connection is ok */
			bytes_read += rc;
			DEBUG_FLOW ("iSCSI: session iscsi bus %d target id %d "
				    "recv_data discarded %d bytes, tossed %d "
				    "of %d bytes at %lu\n",
				    session->iscsi_bus, session->target_id, rc,
				    bytes_read, length, jiffies);
			session->last_rx = jiffies;
			smp_mb();
		}
	}

	/* We don't bother checking the CRC, since we couldn't 
	 * retry the command anyway 
	 */
	if (task) {
		atomic_dec(&task->refcount);
		task = NULL;
	}

	if (stdrh->flags & ISCSI_FLAG_DATA_STATUS) {
		spin_lock(&session->task_lock);
		complete_task(session, itt);
		/* complete_task will release the lock */
	}
}

#else

void
iscsi_recv_data(struct iscsi_session *session, struct IscsiDataRspHdr *stdrh)
{
	struct iscsi_task *task; 
	Scsi_Cmnd *sc = NULL; 
	struct scatterlist *sglist, *sg, *first_sg = NULL, *last_sg = NULL;
	int length, dlength, remaining, rc, i;
	int bytes_read;
	uint32_t offset, expected_offset = 0;
	unsigned int iovn = 0, pad;
	unsigned int segment_offset;
	struct msghdr msg;
	uint32_t itt = ntohl(stdrh->itt);
	uint8_t *peripheral = NULL;
	uint32_t received_crc32c;
	int fake_data_mismatch = 0;
	int ignore_completion = 0;

	if (stdrh->flags & ISCSI_FLAG_DATA_STATUS) {
		/* FIXME: check StatSN */
		session->ExpStatSn = ntohl(stdrh->statsn) + 1;
	}
	updateSN(session, ntohl(stdrh->expcmdsn), ntohl(stdrh->maxcmdsn));
	/* assume a PDU round-trip, connection is ok */
	session->last_rx = jiffies;
	smp_mb();

	length = dlength = ntoh24(stdrh->dlength);
	offset = ntohl(stdrh->offset);

	/* Compute padding bytes that follow the data */
	pad = dlength % PAD_WORD_LEN;
	if (pad) {
		pad = PAD_WORD_LEN - pad;
	}

	spin_lock(&session->task_lock);

	task = find_session_task(session, itt);
	if (task == NULL) {
		printk ("iSCSI: session iscsi bus %d target id %d recv_data, "
			"no task for itt %u (next itt %u), discarding received "
			"data, offset %u len %u\n",
			session->iscsi_bus, session->target_id, 
			ntohl(stdrh->itt), session->itt, offset, dlength);
	} else if (!test_bit(TASK_READ, &task->flags)) {
		/* we shouldn't be getting Data-in unless it's a read */
		if (task->scsi_cmnd)
			printk
			    ("iSCSI: session iscsi bus %d recv_data itt %u, "
			     "command cdb 0x%02x, dropping session due to "
			     "unexpected Data-in from (%u %u %u %u)\n",
			     session->iscsi_bus, itt,
			     task->scsi_cmnd->cmnd[0], session->host_no,
			     session->channel, session->target_id, task->lun);
		else
			printk
			    ("iSCSI: session iscsi bus %d recv_data itt %u, "
			     "command NULL, dropping session due to unexpected "
			     "Data-in from (%u %u %u %u)\n",
			     session->iscsi_bus, itt, session->host_no,
			     session->channel, session->target_id, task->lun);

		/* print the entire PDU header */
		printk ("iSCSI: bogus Data-in PDU header: itt 0x%0x ttt 0x%0x, "
			"hlength %u dlength %u lun %u, statsn 0x%0x "
			"expcmdsn 0x%0x maxcmdsn 0x%0x datasn 0x%0x, "
			"offset %u residual %u\n",
			ntohl(stdrh->itt), ntohl(stdrh->ttt), stdrh->hlength,
			ntoh24(stdrh->dlength), stdrh->lun[1],
			ntohl(stdrh->statsn), ntohl(stdrh->expcmdsn),
			ntohl(stdrh->maxcmdsn), ntohl(stdrh->datasn),
			ntohl(stdrh->offset), ntohl(stdrh->residual_count));

		iscsi_drop_session(session);
		task = NULL;
		spin_unlock(&session->task_lock);
		return;
	} else {
		/* accept all of the data for this task */
		sc = task->scsi_cmnd;
		expected_offset = task->rxdata;

		if (sc) {
			/* either we'll read it all, or we'll drop the session 
			 * and requeue the command, so it's safe to increment 
			 * the received data count before we actually read 
			 * the data, while we still have the task_lock.
			 */
			task->rxdata += dlength;

			/* ensure the task's command won't be completed 
			 * while we're using it 
			 */
			atomic_inc(&task->refcount);

			DEBUG_FLOW
			    ("iSCSI: session iscsi bus %d target id %d "
			     "recv_data itt %u, sc 0x%x, datasn %u, offset %u, "
			     "dlength %u\n",
			     session->iscsi_bus, session->target_id, itt,
			     sc->cmnd[0], ntohl(stdrh->datasn), offset,
			     dlength);
		} else {
			/* command has already been completed (by a timeout) */
			printk ("iSCSI: session iscsi bus %d target id %d "
				"recv_data itt %u, no SCSI command at %lu\n",
				session->iscsi_bus, session->target_id, itt,
				jiffies);
		}

		/* if there is piggybacked status, ensure that we're not 
		 * delaying commands to this LUN 
		 */
		if (stdrh->flags & ISCSI_FLAG_DATA_STATUS) {
			/* mark the task completed */
			__set_bit(TASK_COMPLETED, &task->flags);

			if (sc && session->ignore_completions
			    && ((session->ignore_lun < 0)
				|| (session->ignore_lun == task->lun))) {
				/* for testing, the driver can be told 
				 * to ignore command completion 
				 */
				printk ("iSCSI: session iscsi bus %d ignoring "
					"completion of itt %u, cmnd cdb 0x%x "
					"to (%u %u %u %u) at %lu\n",
					session->iscsi_bus, itt, sc->cmnd[0],
					session->host_no, session->channel,
					session->target_id, task->lun, jiffies);

				session->ignore_completions--;
				ignore_completion = 1;
			} else {
				del_task_timer(task);

				/* piggybacked status is always good */
				if (test_bit
				    (task->lun,
				     session->luns_delaying_commands)) {
					__clear_bit(task->lun,
						    session->
						    luns_delaying_commands);
					session->num_luns_delaying_commands--;
					DEBUG_RETRY
					    ("iSCSI: session iscsi bus %d no "
					     "longer delaying commands to "
					     "(%u %u %u %u) at %lu\n",
					     session->iscsi_bus,
					     session->host_no, session->channel,
					     session->target_id, task->lun,
					     jiffies);
					if (session->
					    num_luns_delaying_commands == 0) {
						del_timer_sync(&session->
							       retry_timer);
						clear_bit
						    (SESSION_RETRY_COMMANDS,
						     &session->control_bits);
						DEBUG_RETRY
						    ("iSCSI: session iscsi bus "
						     "%d target id %d stopping "
						     "retry timer at %lu\n",
						     session->iscsi_bus,
						     session->target_id,
						     jiffies);
					}
					if (!test_bit
					    (task->lun,
					     session->luns_timing_out)) {
						unsigned long noqueue_flags_;

						spin_lock_irqsave(&session->
								  scsi_cmnd_lock,
								  noqueue_flags_);
						requeue_deferred_commands
						    (session, task->lun);
						spin_unlock_irqrestore(&session->
								       scsi_cmnd_lock,
								       noqueue_flags_);
					}
					smp_mb();
				}
			}
		}

		/* if there is no command, we don't increment the 
		 * task refcount, so we can't keep using it 
		 */
		if (sc == NULL)
			task = NULL;

		/* for testing, possibly fake a digest mismatch */
		if (session->fake_read_data_mismatch > 0) {
			session->fake_read_data_mismatch--;
			fake_data_mismatch = 1;
		}
	}
#if TEST_DELAYED_DATA
	if (task && dlength && ((task->cmdsn % 500) == 0)) {
		printk ("iSCSI: testing delayed data for task itt %u, "
			"cmdsn %u, dlength %u at %lu\n",
			task->itt, task->cmdsn, dlength, jiffies);
		atomic_dec(&task->refcount);
		task = NULL;
		sc = NULL;
		spin_unlock(&session->task_lock);
		session->last_rx = jiffies + (45 * HZ);
		smp_mb();
		set_current_state(TASK_UNINTERRUPTIBLE);
		schedule_timeout(37 * HZ);
		spin_lock(&session->task_lock);
		printk("iSCSI: test of delayed data continuing at %lu\n",
		       jiffies);
	}
#endif
	spin_unlock(&session->task_lock);

	if (sc == NULL)
		goto toss_data;

	/* sanity check the PDU against the command */
	if ((offset + dlength) > sc->request_bufflen) {
		/* buffer overflow, often because of a corrupt PDU header */
		printk
		    ("iSCSI: session iscsi bus %d target id %d recv_data for "
		     "itt %u, cmnd 0x%x, bufflen %u, Data PDU with offset %u "
		     "len %u overflows command buffer, dropping session\n",
		     session->iscsi_bus, session->target_id, itt, sc->cmnd[0],
		     sc->request_bufflen, offset, dlength);

		if (task)
			atomic_dec(&task->refcount);
		iscsi_drop_session(session);
		return;
	} else if (expected_offset != offset) {
		/* if the data arrives out-of-order, it becomes much harder
		 * for us to correctly calculate the residual if we don't get
		 * enough data and also don't get an underflow from the
		 * target.  This can happen if we discard Data PDUs due to
		 * bogus offsets/lengths.  Since we always negotiate for 
		 * Data PDUs in-order, this should never happen, but check
		 * for it anyway.
		 */
		/* buffer overflow, often because of a corrupt PDU header */
		printk
		    ("iSCSI: session iscsi bus %d target id %d recv_data for "
		     "itt %u, cmnd 0x%x, bufflen %u, offset %u does not match "
		     "expected offset %u, dropping session\n",
		     session->iscsi_bus, session->target_id, itt, sc->cmnd[0],
		     sc->request_bufflen, offset, expected_offset);

		if (task)
			atomic_dec(&task->refcount);
		iscsi_drop_session(session);
		return;
	}

	/* configure for receiving the data */
	if (sc->use_sg) {
		int index;

		/* scatter-gather */
		sg = sglist = (struct scatterlist *) sc->request_buffer;
		segment_offset = offset;

		for (index = 0; index < sc->use_sg; index++) {
			if (segment_offset < sglist[index].length)
				break;
			else
				segment_offset -= sglist[index].length;
		}

		if (index >= sc->use_sg) {
			/* didn't find the offset, toss the data and 
			 * let the command underflow 
			 */
			printk ("iSCSI: session iscsi bus %d target id %d "
				"recv_data for itt %u couldn't find offset %u "
				"in sglist %p, sc 0x%x, bufflen %u, use_sg %u, "
				"dropping session\n",
				session->iscsi_bus, session->target_id, 
				task->itt, offset, sglist, sc->cmnd[0], 
				sc->request_bufflen, sc->use_sg);
			print_cmnd(sc);
			ISCSI_TRACE(ISCSI_TRACE_BadOffset, sc, task, offset,
				    sc->request_bufflen);
			/* FIXME: discard the data, or drop the session? */
			atomic_dec(&task->refcount);
			iscsi_drop_session(session);
			return;
		} else {
			remaining = dlength;

			/* setup all the data buffers */
			while (sc && (remaining > 0) && (index < sc->use_sg)) {
				sg = &sglist[index];

				if (!kmap_sg(sg)) {
					printk ("iSCSI: session iscsi bus %d "
						"target id %d recv_data itt %u "
						"failed to map sg %p\n",
						session->iscsi_bus, session->
						target_id, task->itt, sg);
					print_cmnd(sc);
					/* FIXME: discard the data, or 
					 * drop the session? 
					 */
					atomic_dec(&task->refcount);
					iscsi_drop_session(session);
					return;
				} else if (first_sg == NULL) {
					first_sg = sg;
				}

				last_sg = sg;

				/* sanity check the sglist segment length */
				if (sg->length <= segment_offset) {
					/* the sglist is corrupt */
					printk
					    ("iSCSI: session iscsi bus %d "
					     "target id %d recv_data index %d, "
					     "length %u too small for offset "
					     "%u, remaining %d, sglist has "
					     "been corrupted\n",
					     session->iscsi_bus,
					     session->target_id, index,
					     sg->length, segment_offset,
					     remaining);
					print_cmnd(sc);
					ISCSI_TRACE(ISCSI_TRACE_BadRxSeg, sc,
						    task, sg->length,
						    segment_offset);
					/* FIXME: discard the data, 
					 * or drop the session? 
					 */
					atomic_dec(&task->refcount);
					iscsi_drop_session(session);
					return;
				}

				session->rx_iov[iovn].iov_base =
				    sg_virtual_address(sg) + segment_offset;
				session->rx_iov[iovn].iov_len =
				    MIN(remaining,
					(int) (sg->length - segment_offset));
				remaining -= session->rx_iov[iovn].iov_len;

				DEBUG_FLOW
				    ("iSCSI: recv_data itt %u, iov[%2d] = "
				     "sg[%2d] = %p, %u of %u bytes, "
				     "remaining %u\n",
				     itt, iovn, sg - sglist,
				     session->rx_iov[iovn].iov_base,
				     session->rx_iov[iovn].iov_len, sg->length,
				     remaining);
				index++;
				iovn++;
				segment_offset = 0;
			}

			if (remaining != 0) {
				/* we ran out of buffer space with more 
				 * data remaining. This should never happen 
				 * if the Scsi_Cmnd's bufflen matches the 
				 * combined length of the sglist segments.
				 */
				printk ("iSCSI: session iscsi bus %d target id "
					"%d recv_data for cmnd 0x%x, bufflen "
					"%u, offset %u len %u, remaining data "
					"%u, dropping session\n",
					session->iscsi_bus, session->target_id,
					sc->cmnd[0], sc->request_bufflen, 
					offset, dlength, remaining);
				print_cmnd(sc);
				/* FIXME: discard the data, 
				 * or drop the session? 
				 */
				atomic_dec(&task->refcount);
				iscsi_drop_session(session);
				return;
			}
		}
	} else {
		/* no scatter-gather, just read it into the buffer */
		session->rx_iov[0].iov_base = sc->request_buffer + offset;
		session->rx_iov[0].iov_len = dlength;
		iovn = 1;
	}

	if (pad) {
		session->rx_iov[iovn].iov_base = session->rx_buffer;
		session->rx_iov[iovn].iov_len = pad;
		iovn++;
		length += pad;
	}

	if (session->DataDigest == ISCSI_DIGEST_CRC32C) {
		/* If we're calculating a data digest, we need to save the 
		 * pointer and length values in the iovecs before the recvmsg 
		 * modifies them (or walk through the sglist again and 
		 * recalculate them later, which seems inefficient).
		 */
		for (i = 0; i < iovn; i++) {
			session->crc_rx_iov[i].iov_base =
			    session->rx_iov[i].iov_base;
			session->crc_rx_iov[i].iov_len =
			    session->rx_iov[i].iov_len;
		}

		/* and we need to receive the target's digest */
		session->rx_iov[iovn].iov_base = &received_crc32c;
		session->rx_iov[iovn].iov_len = sizeof (received_crc32c);
		iovn++;
		length += sizeof (received_crc32c);
	}

	/* save the address of the first byte of INQUIRY data */
	if ((sc->cmnd[0] == INQUIRY) && (offset == 0) && (dlength > 0))
		peripheral = session->rx_iov[0].iov_base;

	/* accept the data */
	memset(&msg, 0, sizeof (struct msghdr));
	msg.msg_iov = session->rx_iov;
	msg.msg_iovlen = iovn;

	DEBUG_FLOW
	    ("iSCSI: recv_data itt %u calling recvmsg %d bytes, iovn %u, "
	     "rx_iov[0].base = %p\n",
	     itt, dlength + pad, iovn, session->rx_iov[0].iov_base);
	rc = iscsi_recvmsg(session, &msg, length);

	if (rc == length) {
		/* assume a PDU round-trip, connection is ok */
		session->last_rx = jiffies;
		smp_mb();

		if (session->DataDigest == ISCSI_DIGEST_CRC32C) {
			uint32_t calculated_crc32c =
			    iscsi_crc32c(session->crc_rx_iov[0].iov_base,
					 session->crc_rx_iov[0].iov_len);

			/* add in all other segments, except 
			 * for the digest itself 
			 */
			for (i = 1; i < iovn - 1; i++) {
				calculated_crc32c =
				    iscsi_crc32c_continued(session->
							   crc_rx_iov[i].
							   iov_base,
							   session->
							   crc_rx_iov[i].
							   iov_len,
							   calculated_crc32c);
			}

			if (fake_data_mismatch) {
				printk ("iSCSI: session iscsi bus %d target id "
					"%d faking read DataDigest mismatch "
					"for itt %u\n",
					session->iscsi_bus, session->target_id,
					task->itt);
				calculated_crc32c = 0x01020304;
			}

			if (calculated_crc32c != received_crc32c) {
				unsigned int lun = task->lun;
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "recv_data for itt %u, cmnd 0x%x "
				     "DataDigest mismatch, received 0x%08x, "
				     "calculated 0x%08x, triggering "
				     "error recovery for LUN %u\n",
				     session->iscsi_bus, session->target_id,
				     itt, sc->cmnd[0], received_crc32c,
				     calculated_crc32c, lun);
				if (first_sg) {
					/* undo any temporary mappings */
					for (sg = first_sg; sg <= last_sg; sg++) {
						kunmap_sg(sg);
					}
					first_sg = NULL;
				}
				/* we MUST abort this task.  To avoid 
				 * reordering, we trigger recovery for all 
				 * tasks to this LUN.  
				 */
				spin_lock(&session->task_lock);
				task->rxdata = 0;
				atomic_dec(&task->refcount);
				trigger_error_recovery(session, lun);
				spin_unlock(&session->task_lock);
				return;
			}
		}
	} else {
		printk ("iSCSI: session iscsi bus %d target id %d recv_data "
			"for itt %u, cmnd 0x%x failed to recv %d data PDU "
			"bytes, rc %d\n",
			session->iscsi_bus, session->target_id, task->itt,
			sc->cmnd[0], length, rc);
		atomic_dec(&task->refcount);
		iscsi_drop_session(session);
		return;
	}

	/* update LUN info based on the INQUIRY data, since we've got it 
	 * mapped now 
	 */
	if (peripheral)
		process_inquiry_data(session, task->scsi_cmnd, peripheral);

	/* done with the data buffers */
	if (first_sg) {
		/* undo any temporary mappings */
		for (sg = first_sg; sg <= last_sg; sg++) {
			kunmap_sg(sg);
		}
	}

	ISCSI_TRACE(ISCSI_TRACE_RxData, sc, task, offset, dlength);

	if ((stdrh->flags & ISCSI_FLAG_DATA_STATUS) && !ignore_completion) {
		unsigned int expected = iscsi_expected_data_length(sc);

		/* we got status, meaning the command completed in a way that
		 * doesn't give us any sense data, and the command must be
		 * completed now, since we won't get a command response PDU.
		 */
		DEBUG_FLOW
		    ("iSCSI: Data-in with status 0x%x for itt %u, sc 0x%x\n",
		     stdrh->cmd_status, ntohl(stdrh->itt),
		     task->scsi_cmnd->cmnd[0]);
		ISCSI_TRACE(ISCSI_TRACE_RxDataCmdStatus, sc, task,
			    stdrh->cmd_status, 0);
		sc->result = host_byte(DID_OK) | status_byte(stdrh->cmd_status);

		spin_lock(&session->task_lock);

		if ((stdrh->flags & ISCSI_FLAG_DATA_OVERFLOW)
		    || (stdrh->flags & ISCSI_FLAG_DATA_UNDERFLOW)
		    || ((test_bit(TASK_READ, &task->flags))
			&& (task->rxdata != expected))) {
			if (LOG_ENABLED(ISCSI_LOG_QUEUE)
			    || LOG_ENABLED(ISCSI_LOG_FLOW)) {
				printk
				    ("iSCSI: session iscsi bus %d task itt %u "
				     "to (%u %u %u %u), cdb 0x%x, %c%c %s, "
				     "received %u, residual %u, expected %u\n",
				     session->iscsi_bus, task->itt,
				     sc->device->host->host_no,
				     sc->device->channel, sc->device->id,
				     sc->device->lun, sc->cmnd[0],
				     (stdrh->flags & ISCSI_FLAG_DATA_OVERFLOW) ?
				     'O' : ' ',
				     (stdrh->flags & ISCSI_FLAG_DATA_UNDERFLOW)
				     ? 'U' : ' ', (stdrh-> flags & ISCSI_FLAG_DATA_OVERFLOW) ?
				     "overflow" : "underflow", task->rxdata,
				     ntohl(stdrh->residual_count), expected);
			}

			if (stdrh->flags & ISCSI_FLAG_DATA_UNDERFLOW) {
				ISCSI_TRACE(ISCSI_TRACE_RxUnderflow, sc, task,
					    ntohl(stdrh->residual_count),
					    expected);
				sc->resid = ntohl(stdrh->residual_count);
			} else if (stdrh->flags & ISCSI_FLAG_DATA_OVERFLOW) {
				/* FIXME: not sure how to tell the SCSI layer 
				 * of an overflow, so just give it an error 
				 */
				ISCSI_TRACE(ISCSI_TRACE_RxOverflow, sc, task,
					    ntohl(stdrh->residual_count),
					    expected);
				sc->result =
				    host_byte(DID_ERROR) | status_byte(stdrh->
								       cmd_status);
			} else {
				/* All the read data did not arrive */
				ISCSI_TRACE(ISCSI_TRACE_HostUnderflow, sc, task,
					    task->rxdata, expected);
				/* we don't know which parts of the buffer 
				 * didn't get data, so report the whole buffer 
				 * missing 
				 */
				sc->resid = expected;
			}
		}

		/* done using the command's data buffers and structure fields */
		atomic_dec(&task->refcount);

		/* try to complete the task.  complete_task expects the 
		 * task_lock held, but returns with it unlocked 
		 */
		complete_task(session, itt);
	} else {
		/* done modifying the command and task */
		atomic_dec(&task->refcount);
	}

	return;

      toss_data:
	/* just throw away the PDU */
	if (first_sg) {
		/* undo any temporary mappings */
		for (sg = first_sg; sg <= last_sg; sg++) {
			kunmap_sg(sg);
		}
	}

	bytes_read = 0;
	length = dlength + pad;
	if (session->DataDigest == ISCSI_DIGEST_CRC32C) {
		printk ("iSCSI: session iscsi bus %d target id %d recv_data "
			"discarding %d data PDU bytes, %d pad bytes, %Zu "
			"digest bytes\n",
			session->iscsi_bus, session->target_id, dlength, pad,
			sizeof (received_crc32c));
		length += sizeof (received_crc32c);
	} else {
		printk ("iSCSI: session iscsi bus %d target id %d recv_data "
			"discarding %d data PDU bytes, %d pad bytes\n",
			session->iscsi_bus, session->target_id, dlength, pad);
	}

	while (!signal_pending(current) && (bytes_read < length)) {
		int num_bytes =
		    MIN(length - bytes_read,(int) sizeof (session->rx_buffer));

		/* FIXME: can we use the same rx_buffer in all the iovecs, 
		 * since we're discarding the data anyway? 
		 * That would reduce the number of recvmsg calls we have 
		 * to make.
		 */
		session->rx_iov[0].iov_base = session->rx_buffer;
		session->rx_iov[0].iov_len = sizeof (session->rx_buffer);
		memset(&msg, 0, sizeof (struct msghdr));
		msg.msg_iov = &session->rx_iov[0];
		msg.msg_iovlen = 1;
		rc = iscsi_recvmsg(session, &msg, num_bytes);
		if (rc <= 0) {
			printk ("iSCSI: session iscsi bus %d target id %d "
				"recv_data failed to recv and discard %d data "
				"PDU bytes, rc %d, bytes_read %d\n",
				session->iscsi_bus, session->target_id, 
				length, rc, bytes_read);
			iscsi_drop_session(session);
		} else {
			/* assume a PDU round-trip, connection is ok */
			bytes_read += rc;
			DEBUG_FLOW ("iSCSI: session iscsi bus %d target id %d "
				    "recv_data discarded %d bytes, tossed %d "
				    "of %d bytes at %lu\n",
				    session->iscsi_bus, session->target_id, rc,
				    bytes_read, length, jiffies);
			session->last_rx = jiffies;
			smp_mb();
		}
	}

	/* We don't bother checking the CRC, since we couldn't 
	 * retry the command anyway 
	 */
	if (task) {
		atomic_dec(&task->refcount);
		task = NULL;
	}

	if (stdrh->flags & ISCSI_FLAG_DATA_STATUS) {
		spin_lock(&session->task_lock);
		complete_task(session, itt);
		/* complete_task will release the lock */
	}
}

#endif

void
iscsi_recv_task_mgmt(struct iscsi_session *session,
		     struct IscsiScsiTaskMgtRspHdr *ststmrh)
{
	struct iscsi_task *task;
	uint32_t mgmt_itt = ntohl(ststmrh->itt);
	int ignored = 0;

	/* FIXME: check StatSN */
	session->ExpStatSn = ntohl(ststmrh->statsn) + 1;
	updateSN(session, ntohl(ststmrh->expcmdsn), ntohl(ststmrh->maxcmdsn));
	/* assume a PDU round-trip, connection is ok */
	session->last_rx = jiffies;
	smp_mb();

	spin_lock(&session->task_lock);

	/* we should always find the task, since we don't allow them to leave
	 * the driver once we've started error recovery, and we shouldn't
	 * receive a task mgmt response until we've started error recovery.
	 */
	if ((task = find_session_mgmt_task(session, mgmt_itt))) {
		/* we save the recovery state in the session when we send task 
		 * mgmt PDUs, since a command completion that arrives after 
		 * we start recovery may change the task's state after we send 
		 * the task mgmt PDU.  We want to remember what we sent and 
		 * act accordingly.
		 */
		if (test_bit(TASK_TRY_ABORT, &task->flags)) {
			ISCSI_TRACE(ISCSI_TRACE_RxAbort, task->scsi_cmnd, task,
				    mgmt_itt, ststmrh->response);
			if (session->ignore_aborts && ((session->ignore_lun < 0)
						       || (session->
							   ignore_lun ==
							   task->lun))) {
				session->ignore_aborts--;
				ignored = 1;
				if (task->scsi_cmnd)
					printk
					    ("iSCSI: session iscsi bus %d "
					     "target id %d ignoring abort "
					     "response 0x%x for mgmt %u, "
					     "itt %u, cmnd  cdb 0x%x at %lu\n",
					     session->iscsi_bus,
					     session->target_id,
					     ststmrh->response,
					     ntohl(ststmrh->itt), task->itt,
					     task->scsi_cmnd->cmnd[0], jiffies);
				else
					printk
					    ("iSCSI: session iscsi bus %d "
					     "target id %d ignoring abort "
					     "response 0x%x for mgmt %u, "
					     "itt %u, cmnd NULL at %lu\n",
					     session->iscsi_bus,
					     session->target_id,
					     ststmrh->response,
					     ntohl(ststmrh->itt), task->itt,
					     jiffies);
			} else if (session->reject_aborts
				   && ((session->reject_lun < 0)
				       || (session->reject_lun == task->lun))) {
				session->reject_aborts--;
				if (task->scsi_cmnd)
					printk
					    ("iSCSI: session iscsi bus %d "
					     "target id %d treating abort "
					     "response 0x%x as reject for "
					     "mgmt %u, itt %u, cmnd cdb 0x%x\n",
					     session->iscsi_bus,
					     session->target_id,
					     ststmrh->response,
					     ntohl(ststmrh->itt), task->itt,
					     task->scsi_cmnd->cmnd[0]);
				else
					printk
					    ("iSCSI: session iscsi bus %d "
					     "target id %d treating abort "
					     "response 0x%x as reject for "
					     "mgmt %u, itt %u, cmnd NULL\n",
					     session->iscsi_bus,
					     session->target_id,
					     ststmrh->response,
					     ntohl(ststmrh->itt), task->itt);

				task->flags &= ~TASK_RECOVERY_MASK;
				__set_bit(TASK_TRY_ABORT_TASK_SET,
					  &task->flags);
				wake_tx_thread(SESSION_TASK_TIMEDOUT, session);
			} else if (ststmrh->response == 0) {
				if (task->scsi_cmnd)
					printk
					    ("iSCSI: session iscsi bus %d "
					     "target id %d abort success for "
					     "mgmt %u, itt %u, cdb 0x%x\n",
					     session->iscsi_bus,
					     session->target_id,
					     ntohl(ststmrh->itt), task->itt,
					     task->scsi_cmnd->cmnd[0]);
				else
					printk
					    ("iSCSI: session iscsi bus %d "
					     "target id %d abort success for "
					     "mgmt %u, itt %u, cmnd NULL\n",
					     session->iscsi_bus,
					     session->target_id,
					     ntohl(ststmrh->itt), task->itt);
				task->flags &= ~TASK_RECOVERY_MASK;
				wake_tx_thread(SESSION_TASK_TIMEDOUT, session);
			} else if (test_bit(TASK_COMPLETED, &task->flags)) {
				/* we received a command completion before the 
				 * abort response, so the task mgmt abort 
				 * doesn't need to succeed.
				 */
				if (task->scsi_cmnd)
					printk
					    ("iSCSI: session iscsi bus %d "
					     "target id %d abort success for "
					     "mgmt %u due to completion of "
					     "itt %u, cmnd cdb 0x%x\n",
					     session->iscsi_bus,
					     session->target_id,
					     ntohl(ststmrh->itt), task->itt,
					     task->scsi_cmnd->cmnd[0]);
				else
					printk
					    ("iSCSI: session iscsi bus %d "
					     "target id %d abort success for "
					     "mgmt %u due to completion of "
					     "itt %u, cmnd NULL\n",
					     session->iscsi_bus,
					     session->target_id,
					     ntohl(ststmrh->itt), task->itt);

				task->flags &= ~TASK_RECOVERY_MASK;
				wake_tx_thread(SESSION_TASK_TIMEDOUT, session);
			} else {
				if (task->scsi_cmnd)
					printk ("iSCSI: session iscsi bus %d "
						"target id %d abort rejected "
						"(0x%x) for mgmt %u, itt %u, "
						"cmnd cdb 0x%x\n",
						session->iscsi_bus,
						session->target_id,
						ststmrh->response,
						ntohl(ststmrh->itt), task->itt,
						task->scsi_cmnd->cmnd[0]);
				else
					printk ("iSCSI: session iscsi bus %d "
						"target id %d abort rejected "
						"(0x%x) for mgmt %u, itt %u, "
						"cmnd NULL\n",
						session->iscsi_bus,
						session->target_id,
						ststmrh->response,
						ntohl(ststmrh->itt), task->itt);

				task->flags &= ~TASK_RECOVERY_MASK;
				__set_bit(TASK_TRY_ABORT_TASK_SET,
					  &task->flags);
				wake_tx_thread(SESSION_TASK_TIMEDOUT, session);
			}
		} else if (test_bit(TASK_TRY_ABORT_TASK_SET, &task->flags)) {
			ISCSI_TRACE(ISCSI_TRACE_RxAbortTaskSet,
				    task ? task->scsi_cmnd : NULL, task,
				    mgmt_itt, ststmrh->response);
			if (session->ignore_abort_task_sets
			    && ((session->ignore_lun < 0)
				|| (session->ignore_lun == task->lun))) {
				session->ignore_abort_task_sets--;
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "ignoring abort task set response "
				     "0x%x for mgmt %u, itt %u, "
				     "cmnd 0x%x, at %lu\n",
				     session->iscsi_bus, session->target_id,
				     ststmrh->response, ntohl(ststmrh->itt),
				     task->itt, task->scsi_cmnd->cmnd[0],
				     jiffies);
				ignored = 1;
			} else if (session->reject_abort_task_sets
				   && ((session->reject_lun < 0)
				       || (session->reject_lun == task->lun))) {
				session->reject_abort_task_sets--;
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "treating abort task set response "
				     "0x%x as reject for mgmt %u, itt "
				     "%u, cmnd 0x%x\n",
				     session->iscsi_bus, session->target_id,
				     ststmrh->response, ntohl(ststmrh->itt),
				     task->itt, task->scsi_cmnd->cmnd[0]);
				task->flags &= ~TASK_RECOVERY_MASK;
				__set_bit(TASK_TRY_LUN_RESET, &task->flags);
				wake_tx_thread(SESSION_TASK_TIMEDOUT, session);
			} else if (ststmrh->response == 0) {
				struct iscsi_task *t;
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "abort task set success for mgmt %u, "
				     "itt %u, cmnd 0x%x\n",
				     session->iscsi_bus, session->target_id,
				     ntohl(ststmrh->itt), task->itt,
				     task->scsi_cmnd->cmnd[0]);
				/* all tasks to this LUN have been recovered */
				list_for_each_entry(t,
						    &session->task_arrival_order,
						    task_order_link) {
					if (task->lun == t->lun)
						t->flags &= ~TASK_RECOVERY_MASK;
				}
				task->flags &= ~TASK_RECOVERY_MASK;
				wake_tx_thread(SESSION_TASK_TIMEDOUT, session);
			} else {
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "abort task set rejected (0x%x) for "
				     "mgmt %u, itt %u, cmnd 0x%x\n",
				     session->iscsi_bus, session->target_id,
				     ststmrh->response, ntohl(ststmrh->itt),
				     task->itt, task->scsi_cmnd->cmnd[0]);
				task->flags &= ~TASK_RECOVERY_MASK;
				__set_bit(TASK_TRY_LUN_RESET, &task->flags);
				wake_tx_thread(SESSION_TASK_TIMEDOUT, session);
			}
		} else if (test_bit(TASK_TRY_LUN_RESET, &task->flags)) {
			ISCSI_TRACE(ISCSI_TRACE_RxLunReset,
				    task ? task->scsi_cmnd : NULL, task,
				    mgmt_itt, ststmrh->response);
			if (session->ignore_lun_resets
			    && ((session->ignore_lun < 0)
				|| (session->ignore_lun == task->lun))) {
				session->ignore_lun_resets--;
				ignored = 1;
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "ignoring LUN reset response 0x%x for "
				     "mgmt %u, itt %u, cmnd 0x%x at %lu\n",
				     session->iscsi_bus, session->target_id,
				     ststmrh->response, ntohl(ststmrh->itt),
				     task->itt, task->scsi_cmnd->cmnd[0],
				     jiffies);
			} else if (session->reject_lun_resets
				   && ((session->reject_lun < 0)
				       || (session->reject_lun == task->lun))) {
				session->reject_lun_resets--;
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "treating LUN reset response 0x%x as "
				     "reject for mgmt %u, itt %u, cmnd 0x%x\n",
				     session->iscsi_bus, session->target_id,
				     ststmrh->response, ntohl(ststmrh->itt),
				     task->itt, task->scsi_cmnd->cmnd[0]);
				task->flags &= ~TASK_RECOVERY_MASK;
				__set_bit(TASK_TRY_WARM_RESET, &task->flags);
				wake_tx_thread(SESSION_TASK_TIMEDOUT, session);
			} else if (ststmrh->response == 0) {
				struct iscsi_task *t;
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "LUN reset success for mgmt %u, itt %u, "
				     "cmnd 0x%x\n",
				     session->iscsi_bus, session->target_id,
				     ntohl(ststmrh->itt), task->itt,
				     task->scsi_cmnd->cmnd[0]);

				/* tell all devices attached to this LUN 
				 * that a reset occured 
				 */
				lun_reset_occured(session, task->lun);

				/* all tasks to this LUN have been recovered */
				list_for_each_entry(t,
						    &session->task_arrival_order,
						    task_order_link) {
					if (task->lun == t->lun) {
						printk
						    ("iSCSI: session iscsi bus "
						     "%d target id %d LUN reset"
						     " success recovering itt "
						     "%u, cmnd 0x%x\n",
						     session->iscsi_bus,
						     session->target_id, t->itt,
						     t->scsi_cmnd->cmnd[0]);
						t->flags &= ~TASK_RECOVERY_MASK;
					}
				}
				wake_tx_thread(SESSION_TASK_TIMEDOUT, session);
			} else {
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "LUN reset rejected (0x%x) for mgmt %u, "
				     "itt %u, cmnd 0x%x\n",
				     session->iscsi_bus, session->target_id,
				     ststmrh->response, ntohl(ststmrh->itt),
				     task->itt, task->scsi_cmnd->cmnd[0]);
				task->flags &= ~TASK_RECOVERY_MASK;
				__set_bit(TASK_TRY_WARM_RESET, &task->flags);
				wake_tx_thread(SESSION_TASK_TIMEDOUT, session);
			}
		} else if (test_bit(TASK_TRY_WARM_RESET, &task->flags)) {
			ISCSI_TRACE(ISCSI_TRACE_RxWarmReset,
				    task ? task->scsi_cmnd : NULL, task,
				    mgmt_itt, ststmrh->response);
			if (session->ignore_warm_resets
			    && ((session->ignore_lun < 0)
				|| (session->ignore_lun == task->lun))) {
				session->ignore_warm_resets--;
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "ignoring warm reset response 0x%x for "
				     "mgmt %u, itt %u, cmnd 0x%x at %lu\n",
				     session->iscsi_bus, session->target_id,
				     ststmrh->response, ntohl(ststmrh->itt),
				     task->itt, task->scsi_cmnd->cmnd[0],
				     jiffies);
				ignored = 1;
			} else if (session->reject_warm_resets
				   && ((session->reject_lun < 0)
				       || (session->reject_lun == task->lun))) {
				session->reject_warm_resets--;
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "treating warm reset response 0x%x as "
				     "reject for mgmt %u, itt %u, cmnd 0x%x\n",
				     session->iscsi_bus, session->target_id,
				     ststmrh->response, ntohl(ststmrh->itt),
				     task->itt, task->scsi_cmnd->cmnd[0]);
				task->flags &= ~TASK_RECOVERY_MASK;
				__set_bit(TASK_TRY_COLD_RESET, &task->flags);
				wake_tx_thread(SESSION_TASK_TIMEDOUT, session);
			} else if (ststmrh->response == 0) {
				struct iscsi_task *t;
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "warm target reset success for mgmt %u, "
				     "itt %u, cmnd 0x%x\n",
				     session->iscsi_bus, session->target_id,
				     ntohl(ststmrh->itt), task->itt,
				     task->scsi_cmnd->cmnd[0]);

				/* tell all devices attached to this target 
				 * that a reset occured 
				 */
				scsi_report_device_reset(session->hba->host,
							 session->channel,
							 session->target_id);

				/* mark all tasks recovered */
				list_for_each_entry(t,
						    &session->task_arrival_order,
						    task_order_link) {
					printk
					    ("iSCSI: session iscsi bus %d "
					     "target id %d warm target reset "
					     "success recovering itt %u, "
					     "cmnd 0x%x\n",
					     session->iscsi_bus,
					     session->target_id, t->itt,
					     t->scsi_cmnd->cmnd[0]);
					t->flags &= ~TASK_RECOVERY_MASK;
				}

				/* and recover them */
				set_bit(SESSION_RESET, &session->control_bits);
				smp_mb();
				wake_tx_thread(SESSION_TASK_TIMEDOUT, session);
			} else {
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "warm target reset rejected (0x%x) for "
				     "mgmt %u, itt %u, cmnd 0x%x\n",
				     session->iscsi_bus, session->target_id,
				     ststmrh->response, ntohl(ststmrh->itt),
				     task->itt, task->scsi_cmnd->cmnd[0]);
				task->flags &= ~TASK_RECOVERY_MASK;
				__set_bit(TASK_TRY_COLD_RESET, &task->flags);
				wake_tx_thread(SESSION_TASK_TIMEDOUT, session);
			}
		} else if (test_bit(TASK_TRY_COLD_RESET, &task->flags)) {
			/* we probably won't ever get a task mgmt response 
			 * for a cold reset that works, since the target should 
			 * drop the session as part of the reset.
			 */
			ISCSI_TRACE(ISCSI_TRACE_RxColdReset,
				    task ? task->scsi_cmnd : NULL, task,
				    mgmt_itt, ststmrh->response);
			task->flags &= ~TASK_RECOVERY_MASK;
			if (session->ignore_cold_resets
			    && ((session->ignore_lun < 0)
				|| (session->ignore_lun == task->lun))) {
				session->ignore_cold_resets--;
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "ignoring cold reset response 0x%x for "
				     "mgmt %u, itt %u, cmnd 0x%x at %lu\n",
				     session->iscsi_bus, session->target_id,
				     ststmrh->response, ntohl(ststmrh->itt),
				     task->itt, task->scsi_cmnd->cmnd[0],
				     jiffies);
				ignored = 1;
			} else if (session->reject_cold_resets
				   && ((session->reject_lun < 0)
				       || (session->reject_lun == task->lun))) {
				session->reject_cold_resets--;
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "treating cold reset response 0x%x as "
				     "reject for mgmt %u, itt %u, cmnd 0x%x\n",
				     session->iscsi_bus, session->target_id,
				     ststmrh->response, ntohl(ststmrh->itt),
				     task->itt, task->scsi_cmnd->cmnd[0]);
				task->flags &= ~TASK_RECOVERY_MASK;
				__set_bit(TASK_TRY_COLD_RESET, &task->flags);
				wake_tx_thread(SESSION_TASK_TIMEDOUT, session);
			} else if (ststmrh->response == 0) {
				struct iscsi_task *t;

				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "cold target reset success for mgmt %u, "
				     "itt %u, cmnd 0x%x\n",
				     session->iscsi_bus, session->target_id,
				     ntohl(ststmrh->itt), task->itt,
				     task->scsi_cmnd->cmnd[0]);

				/* mark all tasks recovered */
				list_for_each_entry(t,
						    &session->task_arrival_order,
						    task_order_link) {
					printk ("iSCSI: session iscsi bus %d "
						"target id %d cold target "
						"reset success recovering itt "
						"%u, cmnd 0x%x\n",
						session->iscsi_bus,
						session->target_id, t->itt,
						t->scsi_cmnd->cmnd[0]);
					t->flags &= ~TASK_RECOVERY_MASK;
				}

				/* clear any requested reset, since we just 
				 * did one 
				 */
				session->warm_reset_itt = RSVD_TASK_TAG;
				clear_bit(SESSION_RESET_REQUESTED,
					  &session->control_bits);
				/* and recover all the tasks */
				set_bit(SESSION_RESET, &session->control_bits);
				smp_mb();
				wake_tx_thread(SESSION_TASK_TIMEDOUT, session);
			} else {
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "cold target reset rejected (0x%x) for "
				     "mgmt %u, itt %u, cmnd 0x%x\n",
				     session->iscsi_bus, session->target_id,
				     ststmrh->response, ntohl(ststmrh->itt),
				     task->itt, task->scsi_cmnd->cmnd[0]);
				/* nothing left to try, just drop the session 
				 * and hope the target clears the problem 
				 */
				iscsi_drop_session(session);
			}
		}
	} else if (mgmt_itt == session->warm_reset_itt) {
		/* response to a requested reset */
		if (session->ignore_warm_resets && ((session->ignore_lun < 0)
						    || (session->ignore_lun ==
							task->lun))) {
			session->ignore_warm_resets--;
			printk ("iSCSI: session iscsi bus %d target id %d "
				"ignoring warm reset response 0x%x for mgmt %u "
				"at %lu\n",
				session->iscsi_bus, session->target_id,
				ststmrh->response, mgmt_itt, jiffies);
			ignored = 1;
		} else if (session->reject_warm_resets
			   && ((session->reject_lun < 0)
			       || (session->reject_lun == task->lun))) {
			session->reject_warm_resets--;
			printk ("iSCSI: session iscsi bus %d target id %d "
				"ignoring warm reset response 0x%x for mgmt %u "
				"at %lu\n",
				session->iscsi_bus, session->target_id,
				ststmrh->response, mgmt_itt, jiffies);

			session->warm_reset_itt = RSVD_TASK_TAG;
			clear_bit(SESSION_RESET_REQUESTED,
				  &session->control_bits);
			smp_mb();
		} else if (ststmrh->response == 0) {
			struct iscsi_task *t;
			printk ("iSCSI: session iscsi bus %d target id %d warm "
				"target reset success for mgmt %u at %lu\n",
				session->iscsi_bus, session->target_id, 
				mgmt_itt, jiffies);

			session->warm_reset_itt = RSVD_TASK_TAG;
			clear_bit(SESSION_RESET_REQUESTED,
				  &session->control_bits);
			smp_mb();

			/* tell all devices attached to this target that a 
			 * reset occured 
			 */
			scsi_report_device_reset(session->hba->host,
						 session->channel,
						 session->target_id);

			/* mark all tasks recovered */
			list_for_each_entry(t,&session->task_arrival_order,
					    task_order_link) {
					    
				printk ("iSCSI: session iscsi bus %d target id "
					"%d warm target reset killed itt %u, "
					"cmnd 0x%x\n",
					session->iscsi_bus, session->target_id,
					t->itt, t->scsi_cmnd->cmnd[0]);
				t->flags &= ~TASK_RECOVERY_MASK;
			}

			/* and recovery them */
			set_bit(SESSION_RESET, &session->control_bits);
			smp_mb();
			wake_tx_thread(SESSION_TASK_TIMEDOUT, session);
		} else {
			/* didn't work.  just give up */
			session->warm_reset_itt = RSVD_TASK_TAG;
			clear_bit(SESSION_RESET_REQUESTED,
				  &session->control_bits);
			smp_mb();

			printk ("iSCSI: session iscsi bus %d target id %d warm "
				"target reset rejected (0x%x) for mgmt %u "
				"at %lu\n",
				session->iscsi_bus, session->target_id,
				ststmrh->response, mgmt_itt, jiffies);
		}
	} else {
		printk ("iSCSI: session iscsi bus %d target id %d "
			"mgmt response 0x%x for unknown itt %u, rtt %u\n",
			session->iscsi_bus, session->target_id, 
			ststmrh->response, ntohl(ststmrh->itt), 
			ntohl(ststmrh->rtt));
	}

	if (!ignored && (session->mgmt_itt == mgmt_itt)) {
		/* we got the expected response, allow the tx thread to send 
		 * another task mgmt PDU whenever it wants to 
		 */
		session->mgmt_itt = RSVD_TASK_TAG;
		session->task_mgmt_response_deadline = 0;
		smp_mb();
	}

	spin_unlock(&session->task_lock);
}

void
iscsi_recv_reject(struct iscsi_session *session,
		  struct IscsiRejectRspHdr *reject, unsigned char *xbuf)
{
	int dlength = ntoh24(reject->dlength);
	uint32_t itt;
	struct iscsi_task *task = NULL;
	struct IscsiHdr pdu;

	/* FIXME: check StatSN */
	session->ExpStatSn = ntohl(reject->statsn) + 1;
	updateSN(session, ntohl(reject->expcmdsn), ntohl(reject->maxcmdsn));
	/* assume a PDU round-trip, connection is ok */
	session->last_rx = jiffies;
	smp_mb();

	if (reject->reason == DATA_DIGEST_ERROR) {
		/* we don't need to do anything about these, 
		 * timers or other PDUs will handle the problem 
		 */
		if (dlength >= sizeof (pdu)) {
			memcpy(&pdu, xbuf, sizeof (pdu));
			itt = ntohl(pdu.itt);
			printk
			    ("iSCSI: session iscsi bus %d target id %d itt %u "
			     "(opcode 0x%x) rejected because of a "
			     "DataDigest error at %lu\n",
			     session->iscsi_bus, session->target_id, itt,
			     pdu.opcode, jiffies);
		} else {
			printk
			    ("iSCSI: session iscsi bus %d target id %d "
			     "target rejected a PDU because of a "
			     "DataDigest error at %lu\n",
			     session->iscsi_bus, session->target_id, jiffies);
		}
	} else if (reject->reason == IMM_CMD_REJECT) {
		if (dlength >= sizeof (pdu)) {
			/* look at the rejected PDU */
			memcpy(&pdu, xbuf, sizeof (pdu));
			itt = ntohl(pdu.itt);

			/* try to find the task corresponding to this itt, 
			 * and wake up any process waiting on it 
			 */
			spin_lock(&session->task_lock);

			if (session->mgmt_itt == itt)
				session->mgmt_itt = RSVD_TASK_TAG;

			if ((task = find_session_mgmt_task(session, itt))) {
				if (task->scsi_cmnd)
					DEBUG_EH
					    ("iSCSI: session iscsi bus %d "
					     "target id %d task mgmt PDU "
					     "rejected, mgmt %u, itt %u, "
					     "cmnd cdb 0x%x\n",
					     session->iscsi_bus,
					     session->target_id, itt, task->itt,
					     task->scsi_cmnd->cmnd[0]);
				else
					DEBUG_EH
					    ("iSCSI: session iscsi bus %d "
					     "target id %d task mgmt PDU "
					     "rejected, mgmt %u, itt %u, "
					     "cmnd NULL\n",
					     session->iscsi_bus,
					     session->target_id, itt,
					     task->itt);

				if (session->immediate_reject_timer.expires ==
				    0) {
					session->immediate_reject_timer.
					    expires =
					    jiffies + MSECS_TO_JIFFIES(40);
					session->immediate_reject_timer.data =
					    (unsigned long) session;
					session->immediate_reject_timer.
					    function =
					    retry_immediate_mgmt_pdus;
					DEBUG_EH
					    ("iSCSI: session iscsi bus %d "
					     "target id %d scheduling task "
					     "mgmt %u retry for %lu at %lu\n",
					     session->iscsi_bus,
					     session->target_id, itt,
					     session->busy_task_timer.expires,
					     jiffies);
					del_timer_sync
					     (&session->busy_task_timer);	
						/* make sure it's not 
						 * running now 
						 */
					add_timer(&session->
						  immediate_reject_timer);
				}
			} else if ((pdu.opcode & ISCSI_OPCODE_MASK) ==
				   ISCSI_OP_LOGOUT_CMD) {
				/* our Logout was rejected.  just let the 
				 * logout response timer drop the session 
				 */
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "logout PDU rejected, itt %u\n",
				     session->iscsi_bus, session->target_id,
				     itt);
				session->logout_itt = RSVD_TASK_TAG;
				smp_mb();
			} else {
				printk
				    ("iSCSI: session iscsi bus %d target id %d,"
				     " itt %u immediate command rejected "
				     "at %lu\n",
				     session->iscsi_bus, session->target_id,
				     itt, jiffies);
			}
			spin_unlock(&session->task_lock);
		} else {
			printk
			    ("iSCSI: session iscsi bus %d target id %d, "
			     "immediate command rejected at %lu, dlength %u\n",
			     session->iscsi_bus, session->target_id, jiffies,
			     dlength);
		}
	} else {
		if (dlength >= sizeof (pdu)) {
			/* look at the rejected PDU */
			memcpy(&pdu, xbuf, sizeof (pdu));
			itt = ntohl(pdu.itt);
			printk
			    ("iSCSI: dropping session iscsi bus %d target id %d"
			     " because target rejected a PDU, reason 0x%x, "
			     "dlength %d, rejected itt %u, opcode 0x%x\n",
			     session->iscsi_bus, session->target_id,
			     reject->reason, dlength, itt, pdu.opcode);
		} else {
			printk
			    ("iSCSI: dropping session iscsi bus %d target id %d"
			     " because target rejected a PDU, reason 0x%x, "
			     "dlength %u\n",
			     session->iscsi_bus, session->target_id,
			     reject->reason, dlength);
		}
		iscsi_drop_session(session);
	}
}

void
iscsi_recv_async_event(struct iscsi_session *session,
		       struct IscsiAsyncEvtHdr *staeh, unsigned char *xbuf)
{
	unsigned int senselen = ntoh24(staeh->dlength);

	/* FIXME: check StatSN */
	session->ExpStatSn = ntohl(staeh->statsn) + 1;
	updateSN(session, ntohl(staeh->expcmdsn), ntohl(staeh->maxcmdsn));

	ISCSI_TRACE(ISCSI_TRACE_RxAsyncEvent, NULL, NULL, staeh->async_event,
		    staeh->async_vcode);

	switch (staeh->async_event) {
	case ASYNC_EVENT_SCSI_EVENT:
		printk
		    (" iSCSI: SCSI Async event ASC=%0x2x, ASCQ=%0x2x "
		     "received on session iscsi bus %d target id %d "
		     "for target %s\n",
		     xbuf[14], xbuf[15], session->iscsi_bus, session->target_id,
		     session->log_name);

		if (xbuf[14] == 0x3f && xbuf[15] == 0x0e) {
			atomic_inc(&session->refcount);
			/* Lun change event has occured for a target */
			if (kernel_thread(iscsi_lun_thread, (void *) session, 0)
			    < 0) {
				printk("iSCSI: failed to start the thread \n");
			}
		}

		/* no way to pass this up to the SCSI layer, 
		 * since there is no command associated with it 
		 */
		if (LOG_ENABLED(ISCSI_LOG_SENSE)) {
			if (senselen >= 26) {
				printk
				    ("iSCSI: SCSI Async event, senselen %d, "
				     "key %02x, ASC/ASCQ %02X/%02X, session "
				     "iscsi bus %d target id %d to %s\n"
				     "iSCSI: Sense %02x%02x%02x%02x "
				     "%02x%02x%02x%02x %02x%02x%02x%02x "
				     "%02x%02x%02x%02x "
				     "%02x%02x%02x%02x %02x%02x%02x%02x "
				     "%02x%02x\n",
				     senselen, SENSE_KEY(xbuf), ASC(xbuf),
				     ASCQ(xbuf), session->iscsi_bus,
				     session->target_id, session->log_name,
				     xbuf[0], xbuf[1], xbuf[2], xbuf[3],
				     xbuf[4], xbuf[5], xbuf[6], xbuf[7],
				     xbuf[8], xbuf[9], xbuf[10], xbuf[11],
				     xbuf[12], xbuf[13], xbuf[14], xbuf[15],
				     xbuf[16], xbuf[17], xbuf[18], xbuf[19],
				     xbuf[20], xbuf[21], xbuf[22], xbuf[23],
				     xbuf[24], xbuf[25]);
			} else if (senselen >= 18) {
				printk
				    ("iSCSI: SCSI Async event, senselen %d, "
				     "key %02x, ASC/ASCQ %02X/%02X, session "
				     "iscsi bus %d target id %d to %s\n"
				     "iSCSI: Sense %02x%02x%02x%02x "
				     "%02x%02x%02x%02x %02x%02x%02x%02x "
				     "%02x%02x%02x%02x %02x%02x\n",
				     senselen, SENSE_KEY(xbuf), ASC(xbuf),
				     ASCQ(xbuf), session->iscsi_bus,
				     session->target_id, session->log_name,
				     xbuf[0], xbuf[1], xbuf[2], xbuf[3],
				     xbuf[4], xbuf[5], xbuf[6], xbuf[7],
				     xbuf[8], xbuf[9], xbuf[10], xbuf[11],
				     xbuf[12], xbuf[13], xbuf[14], xbuf[15],
				     xbuf[16], xbuf[17]);
			} else if (senselen >= 14) {
				printk
				    ("iSCSI: SCSI Async event, senselen %d, "
				     "key %02x, ASC/ASCQ %02X/%02X, session "
				     "iscsi bus %d target id %d to %s\n"
				     "iSCSI: Sense %02x%02x%02x%02x "
				     "%02x%02x%02x%02x %02x%02x%02x%02x "
				     "%02x%02x\n",
				     senselen, SENSE_KEY(xbuf), ASC(xbuf),
				     ASCQ(xbuf), session->iscsi_bus,
				     session->target_id, session->log_name,
				     xbuf[0], xbuf[1], xbuf[2], xbuf[3],
				     xbuf[4], xbuf[5], xbuf[6], xbuf[7],
				     xbuf[8], xbuf[9], xbuf[10], xbuf[11],
				     xbuf[12], xbuf[13]);
			} else {
				printk
				    ("iSCSI: SCSI Async event, senselen %d, "
				     "key %02x, session iscsi bus %d "
				     "target id %d to %s\n"
				     "iSCSI: Sense %02x%02x%02x%02x "
				     "%02x%02x%02x%02x\n",
				     senselen, SENSE_KEY(xbuf),
				     session->iscsi_bus, session->target_id,
				     session->log_name, xbuf[0], xbuf[1],
				     xbuf[2], xbuf[3], xbuf[4], xbuf[5],
				     xbuf[6], xbuf[7]);
			}
		}
		break;
	case ASYNC_EVENT_REQUEST_LOGOUT:
		printk
		    ("iSCSI: target requests logout within %u seconds "
		     "for session to %s\n",
		     ntohs(staeh->param3), session->log_name);
		/* FIXME: this is really a request to drop a connection, 
		 * not the whole session, but we currently only have one 
		 * connection per session, so there's no difference
		 * at the moment.
		 */

		/* we need to get the task lock to make sure the TX thread 
		 * isn't in the middle of adding another task to the session.
		 */
		spin_lock(&session->task_lock);
		iscsi_request_logout(session, ntohs(staeh->param3) - (HZ / 10),
				     session->active_timeout);
		spin_unlock(&session->task_lock);
		break;
	case ASYNC_EVENT_DROPPING_CONNECTION:
		printk
		    ("iSCSI: session iscsi bus %d target id %d target dropping "
		     "connection %u, reconnect min %u max %u\n",
		     session->iscsi_bus, session->target_id,
		     ntohs(staeh->param1), ntohs(staeh->param2),
		     ntohs(staeh->param3));
		session->time2wait = (long) ntohs(staeh->param2) & 0x0000FFFFFL;
		smp_mb();
		break;
	case ASYNC_EVENT_DROPPING_ALL_CONNECTIONS:
		printk
		    ("iSCSI: session iscsi bus %d target id %d target dropping "
		     "all connections, reconnect min %u max %u\n",
		     session->iscsi_bus, session->target_id,
		     ntohs(staeh->param2), ntohs(staeh->param3));
		session->time2wait = (long) ntohs(staeh->param2) & 0x0000FFFFFL;
		smp_mb();
		break;
	case ASYNC_EVENT_VENDOR_SPECIFIC:
		printk
		    ("iSCSI: session iscsi bus %d target id %d ignoring "
		     "vendor-specific async event, vcode 0x%x\n",
		     session->iscsi_bus, session->target_id,
		     staeh->async_vcode);
		break;
	case ASYNC_EVENT_PARAM_NEGOTIATION:
		printk
		    ("iSCSI: session iscsi bus %d target id %d received async "
		     "event param negotiation, dropping session\n",
		     session->iscsi_bus, session->target_id);
		iscsi_drop_session(session);
		break;
	default:
		printk
		    ("iSCSI: session iscsi bus %d target id %d received "
		     "unknown async event 0x%x at %lu\n",
		     session->iscsi_bus, session->target_id, staeh->async_event,
		     jiffies);
		break;
	}
	if (staeh->async_event == ASYNC_EVENT_DROPPING_CONNECTION ||
	    staeh->async_event == ASYNC_EVENT_DROPPING_ALL_CONNECTIONS ||
	    staeh->async_event == ASYNC_EVENT_REQUEST_LOGOUT) {
		spin_lock(&session->portal_lock);
		session->ip_length =
		    session->portals[session->current_portal].ip_length;
		memcpy(session->ip_address,
		       session->portals[session->current_portal].ip_address,
		       session->portals[session->current_portal].ip_length);

		spin_unlock(&session->portal_lock);
	}
}
