/*
 * iSCSI driver for Linux
 * Copyright (C) 2001 Cisco Systems, Inc.
 * maintained by linux-iscsi-devel@lists.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi-session.c,v 1.1.2.17 2004/01/07 12:33:57 smhatre Exp $
 */

#include "iscsi-login.h"
#include "iscsi-io.h"
#include "iscsi.h"

/*
 *  IMPORTANT NOTE: to prevent deadlock, when holding multiple locks,
 *  the following locking order must be followed at all times:
 *
 *  session->portal_lock      - access to a session's portal info
 *  session->task_lock        - access to a session's collections of tasks
 *  hba->session_lock         - access to an HBA's collection of sessions   
 *  session->scsi_cmnd_lock   - access to a session's list of Scsi_Cmnds
 * 								(IRQSAVE)
 *  io_request_lock/host_lock - mid-layer acquires before calling queuecommand,  * 				eh_*, we must acquire before done() callback  
 *				(IRQSAVE)
 *
 *  Note: callers not in interrupt context must locally disable/restore 
 *  interrupts when holding locks marked (IRQSAVE)
 */

extern struct iscsi_hba *iscsi_hba;
extern int this_is_iscsi_boot;
extern struct sapiNBP iscsi_inbp_info;


/* drop an iscsi session */
void
iscsi_drop_session(struct iscsi_session *session)
{
	pid_t pid;

	DEBUG_INIT
	    ("iSCSI: iscsi_drop_session iscsi bus %d target id %d,"
	     " rx %d, tx %d at %lu\n",
	     session->iscsi_bus, session->target_id, session->rx_pid,
	     session->tx_pid, jiffies);

	set_bit(SESSION_DROPPED, &session->control_bits);	
				/* so we know whether to abort the connection */
	session->session_drop_time = jiffies ? jiffies : 1;	
				/* for replacement timeouts */
	smp_wmb();
	clear_bit(SESSION_ESTABLISHED, &session->control_bits);
	smp_mb();

	if ((pid = session->tx_pid))
		kill_proc(pid, SIGHUP, 1);
	if ((pid = session->rx_pid))
		kill_proc(pid, SIGHUP, 1);
	session->session_alive = 0;
}


/* Note: may acquire the task_lock */
void
iscsi_terminate_session(struct iscsi_session *session)
{
	pid_t pid;

	if ((test_and_set_bit(SESSION_TERMINATING, &session->control_bits) == 0)
	    && test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
		DEBUG_INIT
		    ("iSCSI: iscsi_terminate_session iscsi bus %d "
		     "target id %d, requesting logout at %lu\n",
		     session->iscsi_bus, session->target_id, jiffies);

		/* on the first terminate request while the session is up, 
		 * request a logout in the next 3 seconds 
		 */
		spin_lock(&session->task_lock);
		iscsi_request_logout(session, 3, session->active_timeout);
		spin_unlock(&session->task_lock);
	} else {
		/* either we've already tried to terminate once, or the 
		 * session is down.  just kill everything. 
		 */
		clear_bit(SESSION_ESTABLISHED, &session->control_bits);
		session->session_drop_time = jiffies ? jiffies : 1;
		smp_mb();

		DEBUG_INIT ("iSCSI: iscsi_terminate_session iscsi bus %d "
			    "target id %d, killing rx %d, tx %d at %lu\n",
			    session->iscsi_bus, session->target_id, 
			    session->rx_pid, session->tx_pid, jiffies);

		/* kill the session's threads */
		if ((pid = session->tx_pid))
			kill_proc(pid, SIGKILL, 1);
		if ((pid = session->rx_pid))
			kill_proc(pid, SIGKILL, 1);
	}
	session->session_alive = 0;
}


/* remember old peak cmdsn window size, and report the largest */
int
max_tasks_for_session(struct iscsi_session *session)
{
	if (session->ExpCmdSn == session->MaxCmdSn + 1)
		/* if the window is closed, report nothing, 
		 * regardless of what it was in the past 
		 */
		return 0;
	else if (session->last_peak_window_size <
		 session->current_peak_window_size)
		/* window increasing, so report the current peak size */
		return MIN(session->current_peak_window_size,
			   ISCSI_CMDS_PER_LUN * session->num_luns);
	else
		/* window decreasing.  report the previous peak size, in case 
		 * it's temporary decrease caused by the commands we're sending.
		 * we want to keep the right number of commands queued in the 
		 * driver, ready to go as soon as they can.
		 */
		return MIN(session->last_peak_window_size,
			   ISCSI_CMDS_PER_LUN * session->num_luns);
}


/* add a session to some iSCSI HBA's collection of sessions. */
int
add_session(struct iscsi_session *session)
{
	struct iscsi_session *session_next;
	int hba_number;
	int channel_number;
	int ret, p;
	unsigned long noqueue_flags_;

	hba_number = session->iscsi_bus / ISCSI_MAX_CHANNELS_PER_HBA;
	channel_number = session->iscsi_bus % ISCSI_MAX_CHANNELS_PER_HBA;

	if (!iscsi_hba) {
		printk("iSCSI: couldn't find HBA with iSCSI bus %d\n",
		       session->iscsi_bus);
		return 0;
	}
	if (!test_bit(ISCSI_HBA_ACTIVE, &iscsi_hba->flags)) {
		printk ("iSCSI: HBA host no %d is not active, "
			"can't add session with target id %d\n",
			iscsi_hba->host_no, session->target_id);
		return 0;
	}
	if (!iscsi_hba->host) {
		printk ("iSCSI: HBA has no host, can't add session with "
			"target id %d\n", session->target_id);
		return 0;
	}
	if (test_bit(ISCSI_HBA_RELEASING, &iscsi_hba->flags)) {
		printk ("iSCSI: releasing HBA host no %d, can't add session "
			"with target id %d\n",
			iscsi_hba->host_no, session->target_id);
		return 0;
	}
	if (test_bit(ISCSI_HBA_SHUTTING_DOWN, &iscsi_hba->flags)) {
		printk ("iSCSI: HBA host no %d is shutting down, can't add "
			"session with target id %d\n",
			iscsi_hba->host_no, session->target_id);
		return 0;
	}

	spin_lock_irqsave(&iscsi_hba->session_lock, noqueue_flags_);

	/* skip earlier channels and earlier targets on the same channel */
        list_for_each_entry(session_next, &iscsi_hba->session_head, 
			    session_link) {
		if (session_next->channel < session->channel)
			continue;
		if ((session_next->channel != session->channel) 
		 ||(session_next->target_id >= session->target_id)) 
			break;
        }

	/* same linux SCSI address? */
        if (((&session_next->session_link) != (&iscsi_hba->session_head)) 
	    && (session_next->channel == session->channel)
	    && (session_next->target_id == session->target_id)) {
		if (strcmp(session_next->TargetName,session->TargetName)) {
			/* warn that some other target has it */
			printk
			    ("iSCSI: bus %d target %d is already claimed "
			     "for %s, can't claim for %s\n",
			     session->iscsi_bus, session_next->target_id,
			     session->TargetName, session_next->TargetName);
		}
        	ret = 0;
	} else {
        	list_add(&session->session_link,
			 session_next->session_link.prev);
		session->hba = iscsi_hba;
		session->host_no = iscsi_hba->host->host_no;
		atomic_inc(&iscsi_hba->num_sessions);

		/* log the session's bus, target id, TargetName, and all of
		 * the portals, so that the user has a record of what targets
		 * the kernel module was given.  We do this with locks held so
		 * that no other session's info will get interleaved while
		 * we're printing this one's.
		 */

		printk("iSCSI: bus %d target %d = %s\n", session->iscsi_bus,
		       session->target_id, session->TargetName);
		for (p = 0; p < session->num_portals; p++) {
			/* FIXME: IPv6 */
			printk
			    ("iSCSI: bus %d target %d portal %u = address "
			     "%u.%u.%u.%u port %d group %d\n",
			     session->iscsi_bus, session->target_id, p,
			     session->portals[p].ip_address[0],
			     session->portals[p].ip_address[1],
			     session->portals[p].ip_address[2],
			     session->portals[p].ip_address[3],
			     session->portals[p].port, session->portals[p].tag);
		}
		ret = 1;
	}
 
	spin_unlock_irqrestore(&iscsi_hba->session_lock, noqueue_flags_);

	return ret;
}

/* remove a session from an HBA's collection of sessions.
 * caller must hold the HBA's session lock.
 */
int
remove_session(struct iscsi_hba *hba, struct iscsi_session *session)
{
	if (session->hba && (hba != session->hba)) {
		printk
		    ("iSCSI: tried to remove session iscsi bus %d target id %d"
		     " from hba host no %d, but session->hba host no %d\n",
		     session->iscsi_bus, session->target_id, hba->host_no,
		     session->hba->host_no);
		return 0;
	}

	/* remove the session from the HBA */
	list_del(&session->session_link);
	return 1;
}


struct iscsi_session *
find_session_for_cmnd(Scsi_Cmnd * sc)
{
	struct iscsi_session *session; 
	struct iscsi_hba *hba;
	unsigned long noqueue_flags_;

	if (!sc->device->host)
		return NULL;

	if (!sc->device->host->hostdata)
		return NULL;

	hba = (struct iscsi_hba *) sc->device->host->hostdata;

	/* find the session for this command.
	 * FIXME: may want to cache the last session we looked for, since 
	 * we'll often get burst of requests for the same session when multiple 
	 * commands are queued. Would need to invalidate the cache when a 
	 * session is removed from the HBA.
	 */
	spin_lock_irqsave(&hba->session_lock, noqueue_flags_);
	list_for_each_entry(session, &hba->session_head, session_link){
		if ((session->channel == sc->device->channel) 
		    && (session->target_id == sc->device->id)) {
			break;
		}
	}
	if (&(session->session_link)== &(hba->session_head))
		session = NULL;
	if (session)
		atomic_inc(&session->refcount);	/* caller must use 
						 * drop_reference when it's 
						 * done with the session 
						 */
	spin_unlock_irqrestore(&hba->session_lock, noqueue_flags_);

	return session;
}

struct iscsi_session *
find_session_by_bus(int iscsi_bus, int target_id)
{
	struct iscsi_session *session = NULL;
	unsigned int hba_index;
	unsigned int channel;
	unsigned long noqueue_flags_;

	/* compute the appropriate HBA and channel numbers */
	hba_index = iscsi_bus / ISCSI_MAX_CHANNELS_PER_HBA;
	channel = iscsi_bus % ISCSI_MAX_CHANNELS_PER_HBA;

	/* find the session for this command */
	if (iscsi_hba) {
		spin_lock_irqsave(&iscsi_hba->session_lock, noqueue_flags_);
		list_for_each_entry(session, &iscsi_hba->session_head, 
				    session_link) {
			if ((session->channel == channel) 
			    && (session->target_id == target_id)) {
				break;
			}
		}
		if (&(session->session_link) != &(iscsi_hba->session_head))
			atomic_inc(&session->refcount);	/* caller must use 
							 * drop_reference when 
							 * it's done with the 
							 * session */
                else 
			session = NULL;
		spin_unlock_irqrestore(&iscsi_hba->session_lock, 
				       noqueue_flags_);
	}

	return session;
}

struct iscsi_session *
find_session_by_hba(struct iscsi_hba *hba, int channel, int target_id)
{
	struct iscsi_session *session; 
	unsigned long noqueue_flags_;

	spin_lock_irqsave(&hba->session_lock, noqueue_flags_);
	list_for_each_entry(session,&hba->session_head,session_link) {
		if ((session->channel == channel) 
		    && (session->target_id == target_id)) {
			break;
		}
	}
	if (&(session->session_link) != &(hba->session_head))
		atomic_inc(&session->refcount); /* caller must use 
						 * drop_reference when it's
						 * done with the session. 
						 */
	else
		session = NULL;
	spin_unlock_irqrestore(&hba->session_lock, noqueue_flags_);
	return session;
}

/* caller must hold the session's scsi_cmnd_lock */
void
print_session_cmnds(struct iscsi_session *session)
{
	Scsi_Cmnd *search = session->retry_cmnd_head;
	printk ("iSCSI: session iscsi bus %d target id %d , num %u at %lu\n",
		session->iscsi_bus, session->target_id,
		atomic_read(&session->num_retry_cmnds), jiffies);
	while (search) {
		printk
		    ("iSCSI: session retry cmnd : cdb 0x%x to (%u %u %u %u) "
		     "flags 0x%01lx expires %lu\n",
		     search->cmnd[0], search->device->host->host_no,
		     search->device->channel, search->device->id, 
		     search->device->lun,(unsigned long) *command_flags(search),
		     search->eh_timeout.expires);
		search = (Scsi_Cmnd *) search->host_scribble;
	}

	printk ("iSCSI: session iscsi bus %d target id %d , deferred cmnd "
		"queue: num %u at %lu\n", session->iscsi_bus, 
		session->target_id, session->num_deferred_cmnds, jiffies);
	search = session->deferred_cmnd_head;
	while (search) {
		printk
		    ("iSCSI: session iscsi bus %d deferred cmnd : cdb 0x%x to "
		     "(%u %u %u %u) flags 0x%01lx expires %lu\n",
		     session->iscsi_bus, search->cmnd[0],
		     search->device->host->host_no, search->device->channel, 
		     search->device->id, search->device->lun,
		     (unsigned long) *command_flags(search),
		     search->eh_timeout.expires);
		search = (Scsi_Cmnd *) search->host_scribble;
	}

	printk ("iSCSI: session iscsi bus %d target id %d normal cmnd queue:"
		" num %u at %lu\n", session->iscsi_bus, session->target_id,
		atomic_read(&session->num_cmnds), jiffies);
	search = session->scsi_cmnd_head;
	while (search) {
		printk ("iSCSI: session iscsi bus %d normal cmnd : cdb 0x%x to "
			"(%u %u %u %u) flags 0x%01lx expires %lu\n",
			session->iscsi_bus, search->cmnd[0],
			search->device->host->host_no,
			search->device->channel, search->device->id,
			search->device->lun,
			(unsigned long) *command_flags(search),
			search->eh_timeout.expires);
		search = (Scsi_Cmnd *) search->host_scribble;
	}
}

/* caller must hold the session's task_lock */
void
print_session_tasks(struct iscsi_session *session)
{
	struct iscsi_task *task; 
	Scsi_Cmnd *cmnd;

	printk ("iSCSI: session iscsi bus %d target id %d task queue: "
		"num %u at %lu\n", session->iscsi_bus, session->target_id,
		atomic_read(&session->num_active_tasks), jiffies);

	list_for_each_entry(task, &session->task_arrival_order,
			    task_order_link) {
		if ((cmnd = task->scsi_cmnd))
			printk
			    ("iSCSI: session iscsi bus %d target id %d task :"
			     " itt %u flags 0x%04lx expires %lu %c, cdb 0x%x "
			     "to (%u %u %u %u) flags 0x%01lx expires %lu\n",
			     session->iscsi_bus, session->target_id, task->itt,
			     task->flags, task->timer.expires, test_bit(0,
									&task->
									timedout)
			     ? 'T' : ' ', cmnd->cmnd[0],
			     cmnd->device->host->host_no,
			     cmnd->device->channel, cmnd->device->id, 
			     cmnd->device->lun,
			     (unsigned long) *command_flags(cmnd),
			     cmnd->eh_timeout.expires);
		else
			printk
			    ("iSCSI: session iscsi bus %d target id %d task :"
			     " itt %u flags 0x%04lx expires %lu timedout %u, "
			     "cmnd NULL, LUN %u\n",
			     session->iscsi_bus, session->target_id, task->itt,
			     task->flags, task->timer.expires, test_bit(0,
									&task->
									timedout)
			     ? 1 : 0, task->lun);
	}
}

void
delete_session(struct iscsi_session *session)
{
	unsigned int host, channel, target;

	host = session->host_no;
	channel = session->channel;
	target = session->target_id;

	if (session->preallocated_task) {
		DEBUG_ALLOC
		    ("iSCSI: session for (%u %u %u *) freeing preallocated "
		     "task itt = %u \n",
		     host, channel, target, session->preallocated_task->itt);
		iscsi_task_ctor(session->preallocated_task, NULL, 0);
		kmem_cache_free(session->hba->task_cache,
				session->preallocated_task);
		session->preallocated_task = NULL;
	}

	/* free the auth structures */
	kfree(session->auth_client_block);
	kfree(session->auth_recv_string_block);
	kfree(session->auth_send_string_block);
	kfree(session->auth_recv_binary_block);
	kfree(session->auth_send_binary_block);
	kfree(session->username);
	kfree(session->password);
	kfree(session->username_in);
	kfree(session->password_in);
	kfree(session->portals);
	kfree(session->InitiatorName);
	kfree(session->InitiatorAlias);

	kfree(session);
}

/* decrement the session refcount, and remove it and free it if the 
 * refcount hit zero 
 */
void
drop_reference(struct iscsi_session *session)
{
	struct iscsi_hba *hba;
	unsigned long noqueue_flags_;

	if (!session) {
		printk("iSCSI: bug - drop_reference(NULL)\n");
		return;
	}

	if ((hba = session->hba)) {
		/* may need to remove it from the HBA's session list */
		spin_lock_irqsave(&hba->session_lock, noqueue_flags_);
		if (atomic_dec_and_test(&session->refcount)) {
			if (remove_session(hba, session)) {
				delete_session(session);
				atomic_dec(&hba->num_sessions);
				DEBUG_INIT
				    ("iSCSI: terminated and deleted session "
				     "for (%u %u %u *)\n",
				     session->host_no,
				     session->channel, session->target_id);
			} else {
				printk
				    ("iSCSI: bug - failed to remove "
				     "unreferenced session \n");
			}
		}
		spin_unlock_irqrestore(&hba->session_lock, noqueue_flags_);
	} else {
		/* session isn't in an HBA's list at the moment, so just check
		 * the refcount, and possibly free it.  
		 */
		if (atomic_dec_and_test(&session->refcount)) {
			delete_session(session);
			DEBUG_INIT
			    ("iSCSI: terminated and deleted session for "
			     "(%u %u %u *)\n",
			     session->host_no, session->channel,
			     session->target_id);
		}
	}
}


/* Wait for a session to be established.  
 * Returns 1 if the session is established, zero if the timeout expires
 * or the session is terminating/has already terminated.
 */
int
wait_for_session(struct iscsi_session *session, int use_timeout)
{
	int ret; 
	wait_queue_t waitq;

	if (test_bit(SESSION_ESTABLISHED, &session->control_bits))
		return 1;

	if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
		printk
		    ("iSCSI: session iscsi bus %d target id %d terminating, "
		     "wait_for_session failed\n",
		     session->iscsi_bus, session->target_id);
		return 0;
	}

	init_waitqueue_entry(&waitq, current);
	add_wait_queue(&session->login_wait_q, &waitq);
	smp_mb();

	DEBUG_INIT ("iSCSI: pid %d waiting for session iscsi bus %d "
		    "target id %d at %lu\n", current->pid, session->iscsi_bus, 
		    session->target_id, jiffies);

	for (;;) {
		set_current_state(TASK_INTERRUPTIBLE);

		if (test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
			ret = 1;
			goto done;
		}

		if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
			ret = 0;
			goto done;
		}

		if (signal_pending(current)) {
			ret = 0;
			goto done;
		}

		if (use_timeout && session->replacement_timeout) {
			unsigned long timeout, now;
			long sleep_jiffies = 0;

			if (test_bit
			    (SESSION_REPLACEMENT_TIMEDOUT,
			     &session->control_bits)) {
				ret = 0;
				goto done;
			}

			if (session->session_drop_time)
				timeout = session->session_drop_time +
					  (HZ * session->replacement_timeout);
			else
				timeout = jiffies +
					  (HZ * session->replacement_timeout);

			if (time_before_eq(timeout, jiffies)) {
				printk ("iSCSI: pid %d timed out in "
					"wait_for_session iscsi bus %d "
					"target id %d\n",
					current->pid, session->iscsi_bus,
					session->target_id);
				ret = 0;
				goto done;
			}

			/* handle wrap-around */
			now = jiffies;
			if (now < timeout)
				sleep_jiffies = timeout - now;
			else
				sleep_jiffies = ULONG_MAX - now + timeout;

			schedule_timeout(sleep_jiffies);
		} else
			schedule();
	}

      done:
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&session->login_wait_q, &waitq);

	if (ret == 0)
		printk ("iSCSI: wait_for_session iscsi bus %d target id %d "
			"failed\n", session->iscsi_bus, session->target_id);

	return ret;
}


int
iscsi_establish_session(struct iscsi_session *session)
{
	int ret = -1;
	uint8_t status_class;
	uint8_t status_detail;
	enum iscsi_login_status login_status = 0;

	spin_lock(&session->portal_lock);
	if (session->requested_portal != UINT_MAX) {
		/* request to change to a specific portal */
		next_portal(session);
	} else {
		/* Set almost everything based on the portal's settings.
		 * Don't change the address, since a temporary redirect 
		 * may have already changed the address,
		 * and we want to use the redirected address rather than 
		 * the portal's address.
		 */
		set_portal_config(session, session->current_portal);
	}
	spin_unlock(&session->portal_lock);

	if (LOG_ENABLED(ISCSI_LOG_LOGIN) || LOG_ENABLED(ISCSI_LOG_INIT))
		printk
		    ("iSCSI: bus %d target %d trying to establish session to "
		     "portal %u, address %u.%u.%u.%u port %d group %d, "
		     "rx %d, tx %d at %lu\n",
		     session->iscsi_bus, session->target_id,
		     session->current_portal, session->ip_address[0],
		     session->ip_address[1], session->ip_address[2],
		     session->ip_address[3], session->port,
		     session->portal_group_tag, session->rx_pid,
		     session->tx_pid, jiffies);
	else
		printk
		    ("iSCSI: bus %d target %d trying to establish session to "
		     "portal %u, address %u.%u.%u.%u port %d group %d\n",
		     session->iscsi_bus, session->target_id,
		     session->current_portal, session->ip_address[0],
		     session->ip_address[1], session->ip_address[2],
		     session->ip_address[3], session->port,
		     session->portal_group_tag);

	/* set a timer on the connect */
	if (session->login_timeout) {
		session->login_phase_timer =
		    jiffies + (session->login_timeout * HZ);
		smp_mb();
	}
	if (LOG_ENABLED(ISCSI_LOG_LOGIN))
		printk
		    ("iSCSI: session iscsi bus %d target id %d attempting to "
		     "connect at %lu, timeout at %lu (%d seconds)\n",
		     session->iscsi_bus, session->target_id, jiffies,
		     session->login_phase_timer, session->login_timeout);

	if (!iscsi_connect(session)) {
		if (signal_pending(current))
			printk
			    ("iSCSI: session iscsi bus %d target id %d connect "
			     "timed out at %lu\n",
			     session->iscsi_bus, session->target_id, jiffies);
		else
			printk
			    ("iSCSI: session iscsi bus %d target id %d connect "
			     "failed at %lu\n",
			     session->iscsi_bus, session->target_id, jiffies);
		/* switch to the next portal */
		spin_lock(&session->portal_lock);
		next_portal(session);
		spin_unlock(&session->portal_lock);
		goto done;
	}

	/* We need to grab the config_mutex before we start trying to
	 * login, to ensure update_session doesn't try to change the
	 * per-session settings while the login code is using them.  Any
	 * config updates will be deferred until after the login
	 * completes.  We grab the mutex now, so that the connect timeout
	 * will break us out if we can't get the mutex for some reason.
	 */
	if (down_interruptible(&session->config_mutex)) {
		printk
		    ("iSCSI: session iscsi bus %d target id %d failed to "
		     "acquire mutex before login at %lu\n",
		     session->iscsi_bus, session->target_id, jiffies);
		goto done;
	}

	/* make sure we have auth buffers for the login library to use */
	if (session->bidirectional_auth || session->username
	    || session->password) {
		/* make sure we've allocated everything we need */
		if (session->auth_client_block == NULL) {
			session->auth_client_block =
			    kmalloc(sizeof (*session->auth_client_block),
				    GFP_KERNEL);
			if (session->auth_client_block)
				DEBUG_INIT
				    ("iSCSI: session iscsi bus %d target id %d "
				     "allocated auth_client_block (size %Zu) "
				     "while establishing session\n",
				     session->iscsi_bus, session->target_id,
				     sizeof (*session->auth_client_block));
		}
		if (session->auth_recv_string_block == NULL) {
			session->auth_recv_string_block =
			    kmalloc(sizeof (*session->auth_recv_string_block),
				    GFP_KERNEL);
			if (session->auth_recv_string_block)
				DEBUG_INIT
				    ("iSCSI: session iscsi bus %d target id %d "
				     "allocated auth_recv_string_block "
				     "(size %Zu) while establishing session\n",
				     session->iscsi_bus, session->target_id,
				     sizeof (*session->auth_recv_string_block));
		}
		if (session->auth_send_string_block == NULL) {
			session->auth_send_string_block =
			    kmalloc(sizeof (*session->auth_send_string_block),
				    GFP_KERNEL);
			if (session->auth_send_string_block)
				DEBUG_INIT
				    ("iSCSI: session iscsi bus %d target id %d "
				     "allocated auth_send_string_block "
				     "(size %Zu) while establishing session\n",
				     session->iscsi_bus, session->target_id,
				     sizeof (*session->auth_send_string_block));
		}
		if (session->auth_recv_binary_block == NULL) {
			session->auth_recv_binary_block =
			    kmalloc(sizeof (*session->auth_recv_binary_block),
				    GFP_KERNEL);
			if (session->auth_recv_binary_block)
				DEBUG_INIT
				    ("iSCSI: session iscsi bus %d target id %d "
				     "allocated auth_recv_binary_block "
				     "(size %Zu) while establishing session\n",
				     session->iscsi_bus, session->target_id,
				     sizeof (*session->auth_recv_binary_block));
		}
		if (session->auth_send_binary_block == NULL) {
			session->auth_send_binary_block =
			    kmalloc(sizeof (*session->auth_send_binary_block),
				    GFP_KERNEL);
			if (session->auth_send_binary_block)
				DEBUG_INIT
				    ("iSCSI: session iscsi bus %d target id %d "
				     "allocated auth_send_binary_block "
				     "(size %Zu) while establishing session\n",
				     session->iscsi_bus, session->target_id,
				     sizeof (*session->auth_send_binary_block));
		}

		/* if we have everything we need, setup the auth buffer 
		 * descriptors for the login library 
		 */
		session->num_auth_buffers = 0;
		memset(&session->auth_buffers, 0,
		       sizeof (session->auth_buffers));
		if (session->auth_client_block
		    && session->auth_recv_string_block
		    && session->auth_send_string_block
		    && session->auth_recv_binary_block
		    && session->auth_send_binary_block) {
			session->auth_buffers[0].address =
			    session->auth_client_block;
			session->auth_buffers[0].length =
			    sizeof (*session->auth_client_block);

			session->auth_buffers[1].address =
			    session->auth_recv_string_block;
			session->auth_buffers[1].length =
			    sizeof (*session->auth_recv_string_block);

			session->auth_buffers[2].address =
			    session->auth_send_string_block;
			session->auth_buffers[2].length =
			    sizeof (*session->auth_send_string_block);

			session->auth_buffers[3].address =
			    session->auth_recv_binary_block;
			session->auth_buffers[3].length =
			    sizeof (*session->auth_recv_binary_block);

			session->auth_buffers[4].address =
			    session->auth_send_binary_block;
			session->auth_buffers[4].length =
			    sizeof (*session->auth_send_binary_block);

			session->num_auth_buffers = 5;
		} else if (session->bidirectional_auth) {
			/* we must authenticate, but can't. error out */
			printk
			    ("iSCSI: session iscsi bus %d target id %d requires"
			     " birectional authentication, but couldn't "
			     "allocate authentication stuctures\n",
			     session->iscsi_bus, session->target_id);
			ret = -1;	/* retry */
			up(&session->config_mutex);
			goto done;
		} else {
			/* try to login without auth structures, and see 
			 * if the target will let us in anyway.  If we get 
			 * rejected, retry, and hope we can allocate auth 
			 * structures next time.
			 */
			DEBUG_INIT
			    ("iSCSI: session iscsi bus %d target id %d "
			     "authentication configured, but couldn't allocate "
			     "authentication structures\n",
			     session->iscsi_bus, session->target_id);
		}
	}

	/* clear the connect timer */
	session->login_phase_timer = 0;
	smp_mb();
	iscsi_handle_signals(session);

	/* try to make sure other timeouts don't go off as soon as the 
	 * session is established 
	 */
	session->last_rx = jiffies;
	session->last_ping = jiffies - 1;

	/* initialize session fields for the iscsi-login code */
	session->type = ISCSI_SESSION_TYPE_NORMAL;
	/* iSCSI default, unless declared otherwise by the 
	 * target during login 
	 */
	session->MaxXmitDataSegmentLength =
	    DEFAULT_MAX_RECV_DATA_SEGMENT_LENGTH;
	session->vendor_specific_keys = 1;
	smp_mb();

	/* use the session's rx_buffer for a login PDU buffer, since it is
	 * currently unused.  We can't afford to dynamically allocate
	 * memory right now, since it's possible we're reconnecting, and
	 * the VM system is already blocked trying to write dirty pages to
	 * the iSCSI device we're trying to reconnect.  The session's
	 * rx_buffer was sized to have enough space for us to handle the login
	 * phase.  
	 */
	login_status = iscsi_login(session, session->rx_buffer,
				   sizeof (session->rx_buffer), &status_class,
				   &status_detail);

	/* release the lock on the per-session settings used by the login code 
	 */
	up(&session->config_mutex);

	switch (login_status) {
	case LOGIN_OK:
		/* check the status class and detail */
		break;

	case LOGIN_IO_ERROR:
	case LOGIN_WRONG_PORTAL_GROUP:
	case LOGIN_REDIRECTION_FAILED:
		/* these may indicate problems with just the current portal.  
		 * Try a different one 
		 */
		iscsi_disconnect(session);
		spin_lock(&session->portal_lock);
		next_portal(session);
		printk ("iSCSI: session iscsi bus %d target id %d retrying "
			"login to portal %u at %lu\n",
			session->iscsi_bus, session->target_id,
			session->current_portal, jiffies);
		spin_unlock(&session->portal_lock);
		ret = -1;
		goto done;

	default:
	case LOGIN_FAILED:
	case LOGIN_NEGOTIATION_FAILED:
	case LOGIN_AUTHENTICATION_FAILED:
	case LOGIN_VERSION_MISMATCH:
	case LOGIN_INVALID_PDU:
		/* these are problems that will probably occur with any portal 
		 * of this target. 
		 */
		if (session->ever_established && session->num_luns
		    && session->commands_queued) {
			/* the session has found LUNs and been used before, so
			 * applications or the buffer cache may be expecting
			 * it to continue working.  Keep trying to login even
			 * though clearing the error may require
			 * reconfiguration on the target.
			 */
			iscsi_disconnect(session);
			spin_lock(&session->portal_lock);
			next_portal(session);
			printk
			    ("iSCSI: session iscsi bus %d target id %d may be "
			     "in use, retrying login to portal %u at %lu\n",
			     session->iscsi_bus, session->target_id,
			     session->current_portal, jiffies);
			spin_unlock(&session->portal_lock);
			ret = -1;
		} else {
			printk ("iSCSI: session iscsi bus %d target id %d "
				"giving up on login attempts at %lu\n",
				session->iscsi_bus, session->target_id, 
				jiffies);
			iscsi_disconnect(session);
			ret = 0;
		}
		goto done;
	}

	/* check the login status */
	switch (status_class) {
	case STATUS_CLASS_SUCCESS:
		session->auth_failures = 0;
		ret = 1;
		break;
	case STATUS_CLASS_REDIRECT:
		switch (status_detail) {
		case ISCSI_LOGIN_STATUS_TGT_MOVED_TEMP:{
				unsigned int portal;

				/* the session IP address was changed by the 
				 * login library, sp just try again with this 
				 * portal config but the new address.
				 */
				session->auth_failures = 0;
				smp_mb();
				ret = 1;	/* not really success, but we 
						 * want to retry immediately, 
						 * with no delay 
						 */

				spin_lock(&session->portal_lock);
				portal =
				    find_portal(session, session->ip_address,
						session->ip_length,
						session->port);
				if (portal != UINT_MAX) {
					/* FIXME: IPv6 */
					printk
					    ("iSCSI: session iscsi bus %d "
					     "target id %d login to portal %u "
					     "temporarily redirected to portal "
					     "%u = %u.%u.%u.%u port %d\n",
					     session->iscsi_bus,
					     session->target_id,
					     session->current_portal, portal,
					     session->ip_address[0],
					     session->ip_address[1],
					     session->ip_address[2],
					     session->ip_address[3],
					     session->port);

					/* try to switch to the portal we've 
					 * been redirected to. if that fails, 
					 * try to come back to the portal we 
					 * were redirected away from.
					 * if that fails, try any other portals.
					 */
					session->requested_portal = portal;
					session->fallback_portal =
					    session->current_portal;
				} else {
					/* FIXME: IPv6 */
					printk
					    ("iSCSI: session iscsi bus %d "
					     "target id %d login to portal %u "
					     "temporarily redirected to "
					     "%u.%u.%u.%u port %d\n",
					     session->iscsi_bus,
					     session->target_id,
					     session->current_portal,
					     session->ip_address[0],
					     session->ip_address[1],
					     session->ip_address[2],
					     session->ip_address[3],
					     session->port);

					/* we'll connect to the session's 
					 * address next time.  If that fails,
					 * we'll fallback to the current 
					 * portal automatically.
					 */
				}
				spin_unlock(&session->portal_lock);
				goto done;
			}
		case ISCSI_LOGIN_STATUS_TGT_MOVED_PERM:{
				unsigned int portal;

				/* for a permanent redirect, we need to update 
				 * the portal address, and then try again. 
				 */
				session->auth_failures = 0;
				smp_mb();
				ret = 1;	/* not really success, but we 
						 * want to retry immediately, 
						 * with no delay 
						 */

				spin_lock(&session->portal_lock);
				portal =
				    find_portal(session, session->ip_address,
						session->ip_length,
						session->port);
				if (portal != UINT_MAX) {
					/* FIXME: IPv6 */
					printk
					    ("iSCSI: session iscsi bus %d "
					     "target id %d login to portal %u "
					     "permanently redirected to "
					     "portal %u = %u.%u.%u.%u "
					     "port %d\n",
					     session->iscsi_bus,
					     session->target_id,
					     session->current_portal, portal,
					     session->ip_address[0],
					     session->ip_address[1],
					     session->ip_address[2],
					     session->ip_address[3],
					     session->port);

					/* We want to forget about the current 
					 * portal. Mark this portal dead, and 
					 * switch to the new portal.
					 */
					session->portals[session->
							 current_portal].
					    ip_length = 0;

					/* and switch to the other portal */
					set_portal(session, portal);
				} else {
					printk
					    ("iSCSI: session iscsi bus %d "
					     "target id %d login to portal %u "
					     "permanently redirected to "
					     "%u.%u.%u.%u port %d\n",
					     session->iscsi_bus,
					     session->target_id,
					     session->current_portal,
					     session->ip_address[0],
					     session->ip_address[1],
					     session->ip_address[2],
					     session->ip_address[3],
					     session->port);

					/* reset the address in the 
					 * current portal info 
					 */
					session->portals[session->
							 current_portal].
					    ip_length = session->ip_length;
					memcpy(session->
					       portals[session->current_portal].
					       ip_address, session->ip_address,
					       session->ip_length);
					session->portals[session->
							 current_portal].port =
					    session->port;

					/* and just try logging in again with 
					 * the current portal's config.
					 * It'd be nice for Subnet entries in 
					 * the iscsi.conf file to take effect,
					 * but arranging for that means 
					 * exporting them all into the kernel 
					 * module.
					 */
				}

				spin_unlock(&session->portal_lock);
				goto done;
			}
		default:
			ret = -1;
			session->auth_failures = 0;
			smp_mb();
			printk
			    ("iSCSI: session iscsi bus %d target id %d login "
			     "rejected: redirection type 0x%x not supported\n",
			     session->iscsi_bus, session->target_id,
			     status_detail);
			break;
		}
		iscsi_disconnect(session);
		goto done;
	case STATUS_CLASS_INITIATOR_ERR:
		switch (status_detail) {
		case ISCSI_LOGIN_STATUS_AUTH_FAILED:
			printk
			    ("iSCSI: session iscsi bus %d target id %d login "
			     "rejected: initiator failed authentication with "
			     "target %s\n",
			     session->iscsi_bus, session->target_id,
			     session->TargetName);
			iscsi_disconnect(session);
			spin_lock(&session->portal_lock);
			if ((session->num_auth_buffers < 5) &&
			    (session->username || session->password_length
			     || session->bidirectional_auth)) {
				/* retry, and hope we can allocate the 
				 * auth structures next time 
				 */
				DEBUG_INIT
				    ("iSCSI: session iscsi bus %d target id %d "
				     "retrying the same portal, no "
				     "authentication structures allocated\n",
				     session->iscsi_bus, session->target_id);
				ret = -1;
			} else if ((!session->ever_established)
				   && (session->auth_failures >=
				       session->num_portals)) {
				/* give up, since we've tried every portal, 
				 * and have never established a session 
				 */
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "terminating login attempts, %d of %d "
				     "portals failed authentication or "
				     "authorization\n",
				     session->iscsi_bus, session->target_id,
				     session->auth_failures,
				     session->num_portals);
				ret = 0;
			} else if (session->portal_failover) {
				/* try a different portal */
				session->auth_failures++;
				next_portal(session);
				ret = -1;
			} else {
				session->auth_failures = 0;
				ret = 0;
			}
			spin_unlock(&session->portal_lock);
			goto done;
		case ISCSI_LOGIN_STATUS_TGT_FORBIDDEN:
			printk
			    ("iSCSI: session iscsi bus %d target id %d login "
			     "rejected: initiator failed authorization "
			     "with target %s\n",
			     session->iscsi_bus, session->target_id,
			     session->TargetName);
			iscsi_disconnect(session);
			spin_lock(&session->portal_lock);
			session->auth_failures++;
			if ((!session->ever_established)
			    && (session->auth_failures >= session->num_portals)) {
				/* give up, since we've tried every portal, 
				 * and have never established a session 
				 */
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "terminating login attempts, %d of %d "
				     "portals failed authentication or "
				     "authorization\n",
				     session->iscsi_bus, session->target_id,
				     session->auth_failures,
				     session->num_portals);
				ret = 0;
			} else if (session->portal_failover) {
				/* try a different portal */
				next_portal(session);
				ret = -1;
			} else {
				session->auth_failures = 0;
				ret = 0;
			}
			spin_unlock(&session->portal_lock);
			goto done;
		case ISCSI_LOGIN_STATUS_TGT_NOT_FOUND:
			printk
			    ("iSCSI: session iscsi bus %d target id %d login "
			     "rejected: initiator error - target not found "
			     "(%02x/%02x)\n",
			     session->iscsi_bus, session->target_id,
			     status_class, status_detail);
			session->auth_failures = 0;
			iscsi_disconnect(session);
			ret = 0;
			goto done;
		case ISCSI_LOGIN_STATUS_NO_VERSION:
			/* FIXME: if we handle multiple protocol versions, 
			 * before we log an error, try the other supported 
			 * versions. 
			 */
			printk
			    ("iSCSI: session iscsi bus %d target id %d login "
			     "rejected: incompatible version (%02x/%02x), "
			     "non-retryable, giving up\n",
			     session->iscsi_bus, session->target_id,
			     status_class, status_detail);
			session->auth_failures = 0;
			iscsi_disconnect(session);
			ret = 0;
			goto done;
		default:
			printk
			    ("iSCSI: session iscsi bus %d target id %d login "
			     "rejected: initiator error (%02x/%02x), "
			     "non-retryable, giving up\n",
			     session->iscsi_bus, session->target_id,
			     status_class, status_detail);
			session->auth_failures = 0;
			iscsi_disconnect(session);
			ret = 0;
			goto done;
		}
	case STATUS_CLASS_TARGET_ERR:
		printk
		    ("iSCSI: session iscsi bus %d target id %d login rejected: "
		     "target error (%02x/%02x)\n",
		     session->iscsi_bus, session->target_id, status_class,
		     status_detail);
		session->auth_failures = 0;
		iscsi_disconnect(session);
		/* Try a different portal for the retry.  We have no idea
		 * what the problem is, but maybe a different portal will
		 * work better.
		 */
		spin_lock(&session->portal_lock);
		if (session->portal_failover) {
			next_portal(session);
			ret = -1;
		} else
			ret = 0;
		spin_unlock(&session->portal_lock);
		goto done;
	default:
		printk
		    ("iSCSI: session iscsi bus %d target id %d login response "
		     "with unknown status class 0x%x, detail 0x%x\n",
		     session->iscsi_bus, session->target_id, status_class,
		     status_detail);
		session->auth_failures = 0;
		iscsi_disconnect(session);
		ret = 0;
		goto done;
	}

	/* logged in, get the new session ready */
	clear_bit(SESSION_LOGGED_OUT, &session->control_bits);
	session->fallback_portal = UINT_MAX;
	session->tried_portal_bitmap = 0;
	session->ever_established = 1;
	session->session_alive = 1;
	session->generation++;
	session->auth_failures = 0;
	session->last_rx = jiffies;
	session->last_ping = jiffies - 1;
	session->last_window_check = jiffies;
	session->last_peak_window_size = 0;
	session->last_kill = 0;
	session->window_closed = 0;
	session->window_full = 0;
	session->current_peak_window_size = max_tasks_for_session(session);
	session->window_peak_check = jiffies;
	session->warm_reset_itt = RSVD_TASK_TAG;
	session->cold_reset_itt = RSVD_TASK_TAG;
	session->nop_reply.ttt = RSVD_TASK_TAG;
	session->nop_reply_head = session->nop_reply_tail = NULL;
	session->session_established_time = jiffies;	
		/* used to detect sessions that die as soon as we hit FFP */
	session->session_drop_time = 0;	
		/* used to detect sessions that aren't coming back up */
	session->login_phase_timer = 0;
	if (session->TargetAlias[0])
		session->log_name = session->TargetAlias;
	smp_mb();

	/* announce it */
	if (session->TargetAlias[0] != '\0')
		printk
		    ("iSCSI: bus %d target %d established session #%lu to "
		     "portal %u, address %u.%u.%u.%u port %d group %d, "
		     "alias %s\n",
		     session->iscsi_bus, session->target_id,
		     session->generation, session->current_portal,
		     session->ip_address[0], session->ip_address[1],
		     session->ip_address[2], session->ip_address[3],
		     session->port, session->portal_group_tag,
		     session->TargetAlias);
	else
		printk
		    ("iSCSI: bus %d target %d established session #%lu, "
		     "portal %u, address %u.%u.%u.%u port %d group %d\n",
		     session->iscsi_bus, session->target_id,
		     session->generation, session->current_portal,
		     session->ip_address[0], session->ip_address[1],
		     session->ip_address[2], session->ip_address[3],
		     session->port, session->portal_group_tag);

	if (LOG_ENABLED(ISCSI_LOG_INIT) || LOG_ENABLED(ISCSI_LOG_EH)) {
		printk
		    ("iSCSI: session iscsi bus %d target id %d #%lu established"
		     " at %lu, isid 0x%02x%02x%02x%02x%02x%02x, tsih %u, %u "
		     "normal cmnds, %u deferred cmnds, %u tasks, "
		     "bits 0x%08lx\n",
		     session->iscsi_bus, session->target_id,
		     session->generation, jiffies, session->isid[0],
		     session->isid[1], session->isid[2], session->isid[3],
		     session->isid[4], session->isid[5], session->tsid,
		     atomic_read(&session->num_cmnds),
		     session->num_deferred_cmnds,
		     atomic_read(&session->num_active_tasks),
		     session->control_bits);
	}

	/* mark the session as up and accepting commands again */
	clear_bit(SESSION_REPLACEMENT_TIMEDOUT, &session->control_bits);
	smp_wmb();
	set_bit(SESSION_ESTABLISHED, &session->control_bits);
	smp_mb();

	/* wake up everyone waiting for the session to be established */
	wake_up(&session->login_wait_q);

	/* make sure we start sending commands again */
	wake_tx_thread(TX_SCSI_COMMAND, session);

      done:
	/* clear any timer that may have been left running */
	session->login_phase_timer = 0;
	smp_mb();
	/* cleanup after a possible timeout expiration */
	if (iscsi_handle_signals(session)) {
		if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
			DEBUG_INIT
			    ("iSCSI: session iscsi bus %d target id %d "
			     "terminating, giving up on login attempts\n",
			     session->iscsi_bus, session->target_id);
			return 0;
		} else {
			DEBUG_INIT
			    ("iSCSI: session iscsi bus %d target id %d "
			     "received signal during login, retrying\n",
			     session->iscsi_bus, session->target_id);
			return -1;
		}
	}

	return ret;
}

int
update_session(struct iscsi_session *session,
	       struct iscsi_session_ioctl *ioctld,
	       struct iscsi_portal_info *portals)
{
	struct iscsi_portal_info *old_portals;
	struct iscsi_portal_info *old_portal;
	struct iscsi_portal_info *new_portal = NULL;
	struct iscsi_portal_info *q = NULL;
	char *username = NULL;
	unsigned char *password = NULL;
	int password_length = 0;
	char *username_in = NULL;
	unsigned char *password_in = NULL;
	int password_length_in = 0;
	int bidirectional = 0;
	int auth_update_failed = 0;
	int p;
	int relogin = 0;
	size_t length;
	char *str;
	unsigned int requested_portal = UINT_MAX;
	unsigned int portal; 
	int ret, found = 0;

	if (down_interruptible(&session->config_mutex)) {
		/* signalled before we got the mutex */
		printk
		    ("iSCSI: session iscsi bus %d target id %d configuration "
		     "update aborted by signal at %lu\n",
		     session->iscsi_bus, session->target_id, jiffies);
		return 0;
	}

	if (ioctld->update && (ioctld->config_number < session->config_number)) {
		/* this update is obsolete, ignore it */
		DEBUG_INIT
		    ("iSCSI: session iscsi bus %d target id %d ignoring "
		     "obsolete update #%u, currently on config #%u\n",
		     session->iscsi_bus, session->target_id,
		     ioctld->config_number, session->config_number);
		return 0;
	}

	session->config_number = ioctld->config_number;

	printk
	    ("iSCSI: bus %d target %d updating configuration of session iscsi "
	     "bus %d target id %d to %s\n",
	     ioctld->iscsi_bus, ioctld->target_id, session->iscsi_bus,
	     session->target_id, session->log_name);

	/* once we have the mutex, we're guaranteed that the session is
	 * initialized and not logging in at the moment, so we can safely
	 * change the per-session settings stored in the session itself:
	 * isid, InitiatorName, InitiatorAlias, username, password
	 */

	if (memcmp(session->isid, ioctld->isid, sizeof (session->isid))) {
		/* FIXME: the explicit logout better work, since there won't 
		 * be an implicit logout 
		 */
		memcpy(session->isid, ioctld->isid, sizeof (session->isid));
		relogin = 1;
	}

	if ((session->InitiatorName == NULL)
	    || strcmp(ioctld->InitiatorName, session->InitiatorName)) {
		length = strlen(ioctld->InitiatorName);
		if ((str = kmalloc(length + 1, GFP_ATOMIC))) {
			kfree(session->InitiatorName);
			session->InitiatorName = str;
			strcpy(session->InitiatorName, ioctld->InitiatorName);
			relogin = 1;
			DEBUG_INIT
			    ("iSCSI: session iscsi bus %d target id %d updated "
			     "InitiatorName at %lu\n",
			     session->iscsi_bus, session->target_id, jiffies);
		} else {
			printk
			    ("iSCSI: session iscsi bus %d target id %d failed "
			     "to change InitiatorName from %s to %s\n",
			     session->iscsi_bus, session->target_id,
			     session->InitiatorName, ioctld->InitiatorName);
			up(&session->config_mutex);
			return 0;
		}
	}

	if ((session->InitiatorAlias == NULL)
	    || strcmp(ioctld->InitiatorAlias, session->InitiatorAlias)) {
		length = strlen(ioctld->InitiatorAlias);
		if ((str = kmalloc(length + 1, GFP_ATOMIC))) {
			kfree(session->InitiatorAlias);
			session->InitiatorAlias = str;
			strcpy(session->InitiatorAlias, ioctld->InitiatorAlias);
			relogin = 1;
			DEBUG_INIT
			    ("iSCSI: session iscsi bus %d target id %d updated "
			     "InitiatorAlias at %lu\n",
			     session->iscsi_bus, session->target_id, jiffies);
		} else {
			printk
			    ("iSCSI: session iscsi bus %d target id %d failed "
			     "to change InitiatorAlias from %s to %s\n",
			     session->iscsi_bus, session->target_id,
			     session->InitiatorAlias, ioctld->InitiatorAlias);
			up(&session->config_mutex);
			return 0;
		}
	}

	/* transactional (all-or-nothing) update of the auth config */
	if (ioctld->username_in[0] || ioctld->password_length_in)
		bidirectional = 1;

	/* cases:
	 * 1) no new or current value (unchanged), NULL, NULL
	 * 2) new but no current value (transactional update)  NULL, *
	 * 3) no new but current value (transactional delete) *, NULL
	 * 4) new and current value are different (transactional update) *,*
	 * 5) new and current value are the same (unchanged) * == *
	 */
	if (ioctld->username[0]) {
		if ((session->username == NULL)
		    || strcmp(ioctld->username, session->username)) {
			/* update the username */
			length = strlen(ioctld->username);
			if ((username = kmalloc(length + 1, GFP_ATOMIC))) {
				strncpy(username, ioctld->username, length);
				username[length] = '\0';
			} else {
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "failed to change outgoing username\n",
				     session->iscsi_bus, session->target_id);
				ret = -ENOMEM;
				auth_update_failed = 1;
			}
		} else {
			/* they're the same, just keep the current one */
			username = session->username;
		}
	}

	if (ioctld->password_length) {
		if ((session->password == NULL)
		    || (session->password_length != ioctld->password_length)
		    || memcmp(ioctld->password, session->password,
			      session->password_length)) {
			/* update the existing password */
			if ((password =
			     kmalloc(ioctld->password_length + 1,
				     GFP_ATOMIC))) {
				password_length = ioctld->password_length;
				memcpy(password, ioctld->password,
				       password_length);
				password[password_length] = '\0';
			} else {
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "failed to change outgoing password\n",
				     session->iscsi_bus, session->target_id);
				password_length = 0;
				ret = -ENOMEM;
				auth_update_failed = 1;
			}
		} else {
			/* they're the same, just keep the current one */
			password = session->password;
		}
	}

	if (ioctld->username_in[0]) {
		if ((session->username_in == NULL)
		    || strcmp(ioctld->username_in, session->username_in)) {
			/* update the username */
			length = strlen(ioctld->username_in);
			if ((username_in = kmalloc(length + 1, GFP_ATOMIC))) {
				strncpy(username_in, ioctld->username_in,
					length);
				username_in[length] = '\0';
			} else {
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "failed to change incoming username\n",
				     session->iscsi_bus, session->target_id);
				ret = -ENOMEM;
				auth_update_failed = 1;
			}
		} else {
			/* they're the same, just keep the current one */
			username_in = session->username_in;
		}
	}

	if (ioctld->password_length_in) {
		if ((session->password_in == NULL)
		    || (session->password_length_in !=
			ioctld->password_length_in)
		    || memcmp(ioctld->password_in, session->password_in,
			      session->password_length_in)) {
			/* update the existing password */
			if ((password_in =
			     kmalloc(ioctld->password_length_in + 1,
				     GFP_ATOMIC))) {
				password_length = ioctld->password_length_in;
				memcpy(password_in, ioctld->password_in,
				       password_length_in);
				password[password_length_in] = '\0';
			} else {
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "failed to change incoming password\n",
				     session->iscsi_bus, session->target_id);
				password_length_in = 0;
				ret = -ENOMEM;
				auth_update_failed = 1;
			}
		} else {
			/* they're the same, just keep the current one */
			password_in = session->password_in;
		}
	}

	if (!auth_update_failed) {
		/* update to the new auth config */
		session->bidirectional_auth = bidirectional;

		if (username != session->username) {
			/* update current */
			kfree(session->username);
			session->username = username;
			if (username)
				DEBUG_INIT
				    ("iSCSI: session iscsi bus %d target id %d "
				     "updated outgoing username to %s at %lu\n",
				     session->iscsi_bus, session->target_id,
				     session->username, jiffies);
		}

		if (password != session->password) {
			/* update current */
			kfree(session->password);
			session->password = password;
			session->password_length = password_length;
			if (password)
				DEBUG_INIT
				    ("iSCSI: session iscsi bus %d target id %d "
				     "updated outgoing password at %lu\n",
				     session->iscsi_bus, session->target_id,
				     jiffies);
		}

		if (username_in != session->username_in) {
			/* update current */
			kfree(session->username_in);
			session->username_in = username_in;
			if (username_in)
				DEBUG_INIT
				    ("iSCSI: session iscsi bus %d target id %d "
				     "updated incoming username to %s at %lu\n",
				     session->iscsi_bus, session->target_id,
				     session->username_in, jiffies);
		}

		if (password_in != session->password_in) {
			/* update current */
			kfree(session->password);
			session->password_in = password_in;
			session->password_length_in = password_length_in;
			if (password_in)
				DEBUG_INIT
				    ("iSCSI: session iscsi bus %d target id %d "
				     "updated incoming password at %lu\n",
				     session->iscsi_bus, session->target_id,
				     jiffies);
		}
	} else {
		/* update failed, free anything we allocated */
		kfree(username);
		kfree(password);
		kfree(username_in);
		kfree(password_in);
	}

	/* iscsi_establish_session will ensure we have auth structures,
	 * or error out if bidi auth is required and we can't do authentication.
	 */

	up(&session->config_mutex);

	/* the portals are guarded by a spinlock instead of the config
	 * mutex, so that we can request portal changes while a login is
	 * occuring.
	 */
	spin_lock(&session->portal_lock);

	/* replace the portals */
	old_portals = session->portals;
	session->portals = portals;
	session->num_portals = ioctld->num_portals;
	session->requested_portal = UINT_MAX;	/* cancel any request, since 
						 * the portals may have changed 
						 */
	session->fallback_portal = UINT_MAX;	/* cancel any fallback, since 
						 * the portals may have changed 
						 */
	session->portal_failover = ioctld->portal_failover;
	memset(session->preferred_portal, 0,
	       sizeof (session->preferred_portal));
	memset(session->preferred_subnet, 0,
	       sizeof (session->preferred_subnet));
	session->preferred_portal_bitmap = 0;
	session->preferred_subnet_bitmap = 0;
	session->tried_portal_bitmap = 0;

	if (ioctld->preferred_portal && strlen(ioctld->preferred_portal)) {
		memcpy(session->preferred_portal, ioctld->preferred_portal,
		       strlen(ioctld->preferred_portal));
		set_preferred_portal_bitmap(session);
	}

	if (ioctld->preferred_subnet && strlen(ioctld->preferred_subnet)) {
		memcpy(session->preferred_subnet, ioctld->preferred_subnet,
		       strlen(ioctld->preferred_subnet));
		session->preferred_subnet_mask = ioctld->preferred_subnet_mask;
		set_preferred_subnet_bitmap(session);
	}

	printk("iSCSI: bus %d target %d = %s\n", ioctld->iscsi_bus,
	       ioctld->target_id, ioctld->TargetName);
	for (p = 0; p < session->num_portals; p++) {
		/* FIXME: IPv6 */
		printk
		    ("iSCSI: bus %d target %d portal %u = address "
		     "%u.%u.%u.%u port %d group %d\n",
		     ioctld->iscsi_bus, ioctld->target_id, p,
		     portals[p].ip_address[0], portals[p].ip_address[1],
		     portals[p].ip_address[2], portals[p].ip_address[3],
		     portals[p].port, portals[p].tag);
	}

	old_portal = &old_portals[session->current_portal];

	/* figure out which new portal (if any) we're currently 
	 * connected/connecting to 
	 */
	for (p = 0; p < ioctld->num_portals; p++) {
		if (same_network_portal(&portals[p], old_portal)) {
			new_portal = &portals[p];

			if (session->current_portal == p) {
				DEBUG_INIT
				    ("iSCSI: bus %d target %d staying with "
				     "portal %u\n",
				     session->iscsi_bus, session->target_id, p);
			} else if (session->portal_failover) {
				printk
				    ("iSCSI: bus %d target %d portals have "
				     "changed, old portal %u is "
				     "new portal %u\n",
				     session->iscsi_bus, session->target_id,
				     session->current_portal, p);
				/* request the new portal if we decide we need 
				 * to relogin 
				 */
				requested_portal = p;
				/* but reset the current portal in case we 
				 * don't need to relogin 
				 */
				session->current_portal = p;
			}
		}
	}

	if ((new_portal == NULL) && session->portal_failover) {
		/* no matching new portal. try to find a portal in 
		 * the same portal group 
		 */
		for (p = 0; p < ioctld->num_portals; p++) {
			if (portals[p].tag == old_portal->tag) {
				printk
				    ("iSCSI: bus %d target %d portals have "
				     "changed, session switching to portal %u "
				     "in group %u\n",
				     session->iscsi_bus, session->target_id, p,
				     portals[p].tag);
				requested_portal = p;
				new_portal = &portals[p];
				relogin = 1;
			}
		}
	}

	if ((new_portal == NULL) && session->portal_failover) {
		/* we couldn't find a portal in the same portal group.
		 * if we can do a clean logout, we can login to any portal.
		 * if we can't logout, we risk command reordering if we 
		 * login to a different group.
		 */
		new_portal = &portals[0];
		requested_portal = 0;
		printk
		    ("iSCSI: bus %d target %d portals have changed, failed to "
		     "find a new portal in portal group %u, session "
		     "trying portal 0 group %u\n",
		     session->iscsi_bus, session->target_id, old_portal->tag,
		     new_portal->tag);
		/* FIXME: if the logout fails, we'll need to 
		 * error out somehow. 
		 */
		relogin = 1;
	}

	if (session->portal_failover) {
		/* the driver timeouts can change on the fly, with no relogin */
		if (new_portal->login_timeout != old_portal->login_timeout)
			session->login_timeout = new_portal->login_timeout;

		if (new_portal->auth_timeout != old_portal->auth_timeout)
			session->auth_timeout = new_portal->auth_timeout;

		if (new_portal->active_timeout != old_portal->active_timeout)
			session->active_timeout = new_portal->active_timeout;

		if (new_portal->idle_timeout != old_portal->idle_timeout)
			session->idle_timeout = new_portal->idle_timeout;

		if (new_portal->ping_timeout != old_portal->ping_timeout) {
			session->ping_timeout = new_portal->ping_timeout;
			relogin = 1;	/* because we ask the target to use it 
					 * as well with com.cisco.PingTimeout 
					 */
		}

		if (new_portal->abort_timeout != old_portal->abort_timeout)
			session->abort_timeout = new_portal->abort_timeout;

		if (new_portal->reset_timeout != old_portal->reset_timeout)
			session->reset_timeout = new_portal->reset_timeout;

		/* FIXME: this should probably be per-session rather 
		 * than per-portal 
		 */
		if (new_portal->replacement_timeout !=
		    old_portal->replacement_timeout)
			session->replacement_timeout =
			    new_portal->replacement_timeout;

		/* FIXME: get the scsi_cmnd_lock when setting these? */
		if (new_portal->min_disk_command_timeout !=
		    old_portal->min_disk_command_timeout)
			session->min_disk_command_timeout =
			    new_portal->min_disk_command_timeout;

		if (new_portal->max_disk_command_timeout !=
		    old_portal->max_disk_command_timeout)
			session->max_disk_command_timeout =
			    new_portal->max_disk_command_timeout;

		/* the iSCSI op params need a relogin to change */
		if (new_portal->InitialR2T != old_portal->InitialR2T)
			relogin = 1;

		if (new_portal->ImmediateData != old_portal->ImmediateData)
			relogin = 1;

		if (new_portal->MaxRecvDataSegmentLength !=
		    old_portal->MaxRecvDataSegmentLength)
			relogin = 1;

		if (new_portal->FirstBurstLength !=
		    old_portal->FirstBurstLength)
			relogin = 1;

		if (new_portal->MaxBurstLength != old_portal->MaxBurstLength)
			relogin = 1;

		if (new_portal->DefaultTime2Wait !=
		    old_portal->DefaultTime2Wait)
			relogin = 1;

		if (new_portal->DefaultTime2Retain !=
		    old_portal->DefaultTime2Retain)
			relogin = 1;

		if (new_portal->HeaderDigest != old_portal->HeaderDigest)
			relogin = 1;

		if (new_portal->DataDigest != old_portal->DataDigest)
			relogin = 1;

		/* the TCP connection settings need a relogin */
		if (new_portal->tcp_window_size != old_portal->tcp_window_size)
			relogin = 1;

		/* FIXME: TCP type_of_service */

	} else {
		relogin = 0;
		q = session->portals;
		for (portal = 0; portal < session->num_portals; portal++) {
			if (memcmp
			    (q[portal].ip_address, session->ip_address,
			     session->ip_length) == 0) {
				found = 1;
				break;
			}
		}
		if (!found) {
			iscsi_terminate_session(session);
			drop_reference(session);
			return 1;
		}
	}

	if (relogin) {
		/* if we have to relogin, place any portal request 
		 * decided on earlier 
		 */
		session->requested_portal = requested_portal;
		session->fallback_portal = UINT_MAX;
	}

	spin_unlock(&session->portal_lock);

	kfree(old_portals);

	smp_mb();

	if (relogin) {
		if (test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
			spin_lock(&session->task_lock);
			printk
			    ("iSCSI: bus %d target %d configuration updated "
			     "at %lu, session to %s must logout\n",
			     session->iscsi_bus, session->target_id, jiffies,
			     session->log_name);
			iscsi_request_logout(session, 3,
					     session->active_timeout);
			spin_unlock(&session->task_lock);
		} else {
			printk
			    ("iSCSI: bus %d target %d configuration updated "
			     "at %lu while session to %s is not established\n",
			     session->iscsi_bus, session->target_id, jiffies,
			     session->log_name);
		}
	} else {
		printk
		    ("iSCSI: bus %d target %d configuration updated at %lu, "
		     "session to %s does not need to logout\n",
		     session->iscsi_bus, session->target_id, jiffies,
		     session->log_name);
	}

	return 1;
}

struct iscsi_session *
allocate_session(struct iscsi_session_ioctl *ioctld,
		 struct iscsi_portal_info *portals)
{
	struct iscsi_session *session =
	    (struct iscsi_session *) kmalloc(sizeof (*session), GFP_KERNEL);
	size_t length;
	int pp;

	if (session == NULL) {
		printk
		    ("iSCSI: bus %d target %d cannot allocate new session "
		     "(size %Zu) at %lu\n",
		     ioctld->iscsi_bus, ioctld->target_id, sizeof (*session),
		     jiffies);
		return NULL;
	}

	memset(session, 0, sizeof (*session));
	atomic_set(&session->refcount, 1);
	DEBUG_INIT
	    ("iSCSI: bus %d target %d allocated session (size %Zu) at %lu\n",
	     ioctld->iscsi_bus, ioctld->target_id, sizeof (*session), jiffies);

	/* an InitiatorName is required */
	length = strlen(ioctld->InitiatorName);
	if (length) {
		if ((session->InitiatorName = kmalloc(length + 1, GFP_KERNEL))) {
			strncpy(session->InitiatorName, ioctld->InitiatorName,
				length);
			session->InitiatorName[length] = '\0';
		} else {
			printk
			    ("iSCSI: bus %d target %d cannot allocate "
			     "InitiatorName at %lu\n",
			     ioctld->iscsi_bus, ioctld->target_id, jiffies);
			delete_session(session);
			return NULL;
		}
	} else {
		printk("iSCSI: bus %d target %d has no InitiatorName at %lu\n",
		       ioctld->iscsi_bus, ioctld->target_id, jiffies);
		delete_session(session);
		return NULL;
	}

	/* an InitiatorAlias is optional */
	length = strlen(ioctld->InitiatorAlias);
	if (length
	    && (session->InitiatorAlias = kmalloc(length + 1, GFP_KERNEL))) {
		strncpy(session->InitiatorAlias, ioctld->InitiatorAlias,
			length);
		session->InitiatorAlias[length] = '\0';
	} else {
		printk
		    ("iSCSI: bus %d target %d cannot allocate "
		     "InitiatorAlias at %lu\n",
		     ioctld->iscsi_bus, ioctld->target_id, jiffies);
		delete_session(session);
		return NULL;
	}

	memcpy(session->isid, ioctld->isid, sizeof (session->isid));

	strncpy(session->TargetName, ioctld->TargetName,
		sizeof (session->TargetName));
	session->TargetName[sizeof (session->TargetName) - 1] = '\0';
	session->log_name = session->TargetName;
	session->TargetAlias[0] = '\0';	/* none unless declared by the target */

	session->num_auth_buffers = 0;
	session->auth_client_block = NULL;
	session->auth_recv_string_block = NULL;
	session->auth_send_string_block = NULL;
	session->auth_recv_binary_block = NULL;
	session->auth_send_binary_block = NULL;

	/* allocate authentication info */
	if (ioctld->username_in[0] || ioctld->password_length_in) {
		/* we must authenticate the target or refuse to login */
		session->bidirectional_auth = 1;
	} else {
		session->bidirectional_auth = 0;	/* authentication is 
							 * optional 
							 */
	}

	/* FIXME: should we fail the ioctl if the allocation fails? */
	if ((length = strlen(ioctld->username))) {
		if ((session->username = kmalloc(length + 1, GFP_KERNEL))) {
			strncpy(session->username, ioctld->username, length);
			session->username[length] = '\0';
		} else {
			printk
			    ("iSCSI: bus %d target %d failed to allocate "
			     "outgoing username at %lu\n",
			     ioctld->iscsi_bus, ioctld->target_id, jiffies);
			delete_session(session);
			return NULL;
		}
	}

	if (ioctld->password_length) {
		if ((session->password =
		     kmalloc(ioctld->password_length + 1, GFP_KERNEL))) {
			memcpy(session->password, ioctld->password,
			       ioctld->password_length);
			session->password_length = ioctld->password_length;
			session->password[session->password_length] = '\0';
		} else {
			printk
			    ("iSCSI: bus %d target %d failed to allocate "
			     "outgoing password at %lu\n",
			     ioctld->iscsi_bus, ioctld->target_id, jiffies);
			delete_session(session);
			return NULL;
		}
	}

	if ((length = strlen(ioctld->username_in))) {
		if ((session->username_in = kmalloc(length + 1, GFP_KERNEL))) {
			strncpy(session->username_in, ioctld->username_in,
				length);
			session->username_in[length] = '\0';
		} else {
			printk
			    ("iSCSI: bus %d target %d failed to allocate "
			     "incoming username at %lu\n",
			     ioctld->iscsi_bus, ioctld->target_id, jiffies);
			delete_session(session);
			return NULL;
		}
	}

	if (ioctld->password_length_in) {
		if ((session->password_in =
		     kmalloc(ioctld->password_length_in + 1, GFP_KERNEL))) {
			memcpy(session->password_in, ioctld->password_in,
			       ioctld->password_length_in);
			session->password_length_in =
			    ioctld->password_length_in;
			session->password_in[session->password_length_in] =
			    '\0';
		} else {
			printk
			    ("iSCSI: bus %d target %d failed to allocate "
			     "incoming password at %lu\n",
			     ioctld->iscsi_bus, ioctld->target_id, jiffies);
			delete_session(session);
			return NULL;
		}
	}

	/* the auth structures are allocated in iscsi_establish_session, so that
	 * any allocation failures are retried automatically.
	 */

	/* initialize the session structure */
	session->socket = NULL;

	session->config_number = ioctld->config_number;

	spin_lock_init(&session->portal_lock);
	session->num_portals = 0;
	session->portals = NULL;
	session->auth_failures = 0;
	session->portal_failover = 1;
	session->current_portal = 0;
	session->requested_portal = UINT_MAX;
	session->fallback_portal = UINT_MAX;
	session->portal_group_tag = -1;
	memset(session->preferred_portal, 0,
	       sizeof (session->preferred_portal));
	memset(session->preferred_subnet, 0,
	       sizeof (session->preferred_subnet));
	session->preferred_portal_bitmap = 0;
	session->preferred_subnet_bitmap = 0;
	session->tried_portal_bitmap = 0;

	spin_lock_init(&session->scsi_cmnd_lock);
	session->retry_cmnd_head = session->retry_cmnd_tail = NULL;
	atomic_set(&session->num_retry_cmnds, 0);
	session->scsi_cmnd_head = session->scsi_cmnd_tail = NULL;
	atomic_set(&session->num_cmnds, 0);
	session->deferred_cmnd_head = session->deferred_cmnd_tail = NULL;
	session->num_deferred_cmnds = 0;

	sema_init(&session->probe_sem, 0);	/* the first down should block 
						 */
	session->probe_order = ioctld->probe_order;
	memcpy(session->luns_allowed, ioctld->lun_bitmap,
		sizeof(session->luns_allowed));

	sema_init(&session->config_mutex, 0);	/* the first down should block 
						 */

	spin_lock_init(&session->task_lock);
	INIT_LIST_HEAD(&(session->task_arrival_order));
	INIT_LIST_HEAD (&(session->tx_task_head));
	INIT_LIST_HEAD(&(session->probe_order_link));
	atomic_set(&session->num_active_tasks, 0);
	session->preallocated_task = NULL;

	init_waitqueue_head(&session->tx_wait_q);
	init_waitqueue_head(&session->tx_blocked_wait_q);
	init_waitqueue_head(&session->login_wait_q);

	init_timer(&session->busy_task_timer);
	init_timer(&session->busy_command_timer);
	init_timer(&session->immediate_reject_timer);
	init_timer(&session->retry_timer);

	/* save the portal info, and pick which portal to start with */
	session->update_address = &update_address;
	session->portals = portals;
	session->num_portals = ioctld->num_portals;
	session->portal_failover = ioctld->portal_failover;
	session->channel = ioctld->iscsi_bus % ISCSI_MAX_CHANNELS_PER_HBA;
	session->iscsi_bus = ioctld->iscsi_bus;
	session->target_id = ioctld->target_id;

	if (ioctld->preferred_portal && strlen(ioctld->preferred_portal)) {
		memcpy(session->preferred_portal, ioctld->preferred_portal,
		       strlen(ioctld->preferred_portal));
		set_preferred_portal_bitmap(session);
	}

	if (ioctld->preferred_subnet && strlen(ioctld->preferred_subnet)) {
		memcpy(session->preferred_subnet, ioctld->preferred_subnet,
		       strlen(ioctld->preferred_subnet));
		session->preferred_subnet_mask = ioctld->preferred_subnet_mask;
		set_preferred_subnet_bitmap(session);
	}

	pp = get_appropriate_portal(session);
	if (pp < 0)
		set_portal(session, 0U);
	else
		set_portal(session, pp);

	session->itt = ioctld->target_id;	/* to make reading login 
						 * traces easier 
						 */
	session->generation = 0;
	session->ever_established = 0;
	session->session_alive = 0;
	session->time2wait = -1;
	session->logout_itt = RSVD_TASK_TAG;
	session->mgmt_itt = RSVD_TASK_TAG;

	session->ignore_lun = -2;
	session->reject_lun = -2;
	session->fake_status_lun = -2;

	/* in case the session never comes up */
	session->session_drop_time = jiffies;

	smp_mb();

	return session;
}

int
start_session_threads(struct iscsi_session *session)
{
	pid_t rx_pid, tx_pid;

	/* start a tx thread */
	DEBUG_INIT ("iSCSI: session iscsi bus %d target id %d about to start "
		    "tx and rx threads at %lu\n",
		    session->iscsi_bus, session->target_id, jiffies);
	atomic_inc(&session->refcount);
	tx_pid = kernel_thread(iscsi_tx_thread, (void *) session,
			  CLONE_VM | CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
	if (tx_pid > 0) {
		DEBUG_INIT
		    ("iSCSI: session iscsi bus %d target id %d started "
		     "tx thread %u at %lu\n",
		     session->iscsi_bus, session->target_id, tx_pid, jiffies);
	} else {
		printk
		    ("iSCSI: session iscsi bus %d target id %d failed to "
		     "start tx thread, terminating session\n",
		     session->iscsi_bus, session->target_id);
		atomic_dec(&session->refcount);	/* the thread isn't actually 
						 * using it 
						 */
		iscsi_terminate_session(session);
		/* No need to drop reference of this session as this job will
		 * be done in the calling function
		 * handle_iscsi_establish_session()
		 */
		return -EAGAIN;
	}

	/* start an rx thread */
	atomic_inc(&session->refcount);
	rx_pid =
	    kernel_thread(iscsi_rx_thread, (void *) session,
			  CLONE_VM | CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
	if (rx_pid > 0) {
		DEBUG_INIT
		    ("iSCSI: session iscsi bus %d target id %d started "
		     "rx thread %u at %lu\n",
		     session->iscsi_bus, session->target_id, rx_pid, jiffies);
	} else {
		printk
		    ("iSCSI: session iscsi bus %d target id %d failed to "
		     "start rx thread, terminating session\n",
		     session->iscsi_bus, session->target_id);
		atomic_dec(&session->refcount);	/* the thread isn't actually 
						 * using it 
						 */
		iscsi_terminate_session(session);
		/* No need to drop reference of this session as this job will
		 * be done in the calling function
		 * handle_iscsi_establish_session()
		 */
		return -EAGAIN;
	}
	DEBUG_INIT
	    ("iSCSI: session iscsi bus %d target id %d waiting for "
	     "rx %d and tx %d at %lu\n",
	     session->iscsi_bus, session->target_id, rx_pid, tx_pid, jiffies);
	/* wait for the threads to start */
	while ((session->tx_pid == 0) || (session->rx_pid == 0)) {
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(MSECS_TO_JIFFIES(10));
		if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
			printk
			    ("iSCSI: session iscsi bus %d target id %d "
			     "terminating, failed to start threads at %lu\n",
			     session->iscsi_bus, session->target_id, jiffies);
			return -EAGAIN;
		} else if (signal_pending(current)) {
			iscsi_terminate_session(session);
			printk
			    ("iSCSI: session iscsi bus %d target id %d thread "
			     "start terminated, returning at %lu\n",
			     session->iscsi_bus, session->target_id, jiffies);
			return -EAGAIN;
		}
		smp_mb();
	}

	DEBUG_INIT
	    ("iSCSI: session iscsi bus %d target id %d started tx_pid %d, "
	     "rx_pid %d\n",
	     session->iscsi_bus, session->target_id, session->tx_pid,
	     session->rx_pid);

	if (signal_pending(current))
		return -EINTR;
	else
		return 0;
}

/* do timer processing for one session, and return the length of time
 * (in jiffies) til this session needs to be checked again.
 */
unsigned long
check_session_timeouts(struct iscsi_session *session)
{
	unsigned long timeout;
	unsigned long session_timeout = 0;

	if (!test_bit(SESSION_ESTABLISHED, &session->control_bits)) {
		/* check login phase timeouts */
		if (session->login_phase_timer) {
			session_timeout = session->login_phase_timer;
			if (time_before_eq(session_timeout, jiffies)) {
				printk
				    ("iSCSI: login phase for session iscsi bus "
				     "%d target id %d (rx %d, tx %d) timed out "
				     "at %lu, timeout was set for %lu\n",
				     session->iscsi_bus, session->target_id,
				     session->rx_pid, session->tx_pid, jiffies,
				     session_timeout);
				session->login_phase_timer = 0;
				smp_mb();
				session_timeout = 0;
				iscsi_drop_session(session);
			}
		}
	} else {
		/* check full-feature phase timeouts. */
		if (atomic_read(&session->num_active_tasks))
			timeout = session->active_timeout;
		else
			timeout = session->idle_timeout;

		if (timeout) {
			if (session->ping_timeout &&
			    time_before_eq(session->last_rx + (timeout * HZ) +
					   (session->ping_timeout * HZ),
					   jiffies)) {
				/* should have received something by now, 
				 * kill the connection 
				 */
				if ((session->last_kill == 0)
				    || time_before_eq(session->last_kill + HZ,
						      jiffies)) {

					session->last_kill = jiffies;

					printk
					    ("iSCSI: %lu second timeout expired"
					     " for session iscsi bus %d target "
					     "id %d, rx %lu, ping %lu, "
					     "now %lu\n",
					     timeout + session->ping_timeout,
					     session->iscsi_bus,
					     session->target_id,
					     session->last_rx,
					     session->last_ping, jiffies);

					iscsi_drop_session(session);

					session_timeout = jiffies + HZ;
				} else
					session_timeout = 0;
			} else
			    if (time_before_eq
				(session->last_rx + (timeout * HZ), jiffies)) {

				if (time_before_eq
				    (session->last_ping, session->last_rx)) {
					/* send a ping to try to provoke 
					 * some traffic 
					 */
					DEBUG_FLOW
					    ("iSCSI: timer queuing ping for "
					     "session iscsi bus %d target id "
					     "%d, rx %lu, ping %lu, now %lu\n",
					     session->iscsi_bus,
					     session->target_id,
					     session->last_rx,
					     session->last_ping, jiffies);
					session->last_ping = jiffies - 1;

					wake_tx_thread(TX_PING, session);
				}
				session_timeout =
				    session->last_rx + (timeout * HZ) +
				    (session->ping_timeout * HZ);
			} else {
				if (atomic_read(&session->num_active_tasks)) {
					session_timeout =
					    session->last_rx +
					    (session->active_timeout * HZ);
				} else {
					unsigned long active_timeout,
					    idle_timeout;

					/* session is idle, but may become 
					 * active without the timer being 
					 * notified, so use smaller of 
					 * (now + active_timeout, last_rx + 
					 * idle_timeout)
					 */
					idle_timeout =
					    session->last_rx +
					    (session->idle_timeout * HZ);
					active_timeout =
					    jiffies +
					    (session->active_timeout * HZ);
					if (time_before_eq
					    (idle_timeout, active_timeout)) {
						session_timeout = idle_timeout;
					} else {
						session_timeout =
						    active_timeout;
					}
				}
			}
		}

		/* we limit how long we'll wait for a task mgmt 
		 * response, to avoid blocking forever in error recovery.
		 */
		if (session->task_mgmt_response_deadline &&
		    time_before_eq(session->task_mgmt_response_deadline,
				   jiffies)) {
			printk
			    ("iSCSI: session iscsi bus %d target id %d task "
			     "mgmt %u response timeout at %lu\n",
			     session->iscsi_bus, session->target_id,
			     session->mgmt_itt, jiffies);
			session->task_mgmt_response_deadline = 0;
			smp_mb();

			/* tell the tx thread that the task mgmt PDU timed out, 
			 * and have it escalate error recovery 
			 */
			set_bit(SESSION_TASK_MGMT_TIMEDOUT,
				&session->control_bits);
			wake_tx_thread(SESSION_TASK_TIMEDOUT, session);

			if ((session_timeout == 0)
			    || time_before(session->task_mgmt_response_deadline,
					   session_timeout))
				session_timeout =
				    session->task_mgmt_response_deadline;
		}

		/* there's a separate deadline for responses to requested 
		 * resets, so that error recovery and reset requests don't 
		 * get in each other's way.
		 */
		if (session->reset_response_deadline &&
		    time_before_eq(session->reset_response_deadline, jiffies)) {
			session->reset_response_deadline = 0;
			smp_mb();

			if (test_and_clear_bit
			    (SESSION_RESET_REQUESTED, &session->control_bits)) {
				/* FIXME: what should we do when a requested 
				 * reset times out? For now, just give up on 
				 * the reset and drop the session.  The
				 * target should always reply, so we probably 
				 * have a network problem.
				 */
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "timed out waiting for mgmt %u reset "
				     "response, dropping session at %lu\n",
				     session->iscsi_bus, session->target_id,
				     session->warm_reset_itt, jiffies);
				iscsi_drop_session(session);
				return 0;
			}

			if ((session_timeout == 0)
			    || time_before(session->task_mgmt_response_deadline,
					   session_timeout))
				session_timeout =
				    session->task_mgmt_response_deadline;
		}

		if (test_bit(SESSION_LOGOUT_REQUESTED, &session->control_bits)) {
			/* we're waiting for tasks to complete before logging 
			 * out. No need to check the CmdSN window, since 
			 * we won't be starting any more tasks.
			 */
			if (time_before_eq
			    (session->logout_response_deadline, jiffies)) {
				/* passed the deadline for a logout response, 
				 * just drop the session 
				 */
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "logout response timeout at %lu, "
				     "dropping session\n",
				     session->iscsi_bus, session->target_id,
				     jiffies);
				session->logout_response_deadline = 0;
				session->logout_deadline = 0;
				smp_mb();
				iscsi_drop_session(session);
			} else
			    if (time_before_eq
				(session->logout_deadline, jiffies)) {
				/* send a logout */
				DEBUG_INIT
				    ("iSCSI: session iscsi bus %d target id %d "
				     "logout deadline reached at %lu\n",
				     session->iscsi_bus, session->target_id,
				     jiffies);
				session->logout_deadline = 0;
				smp_mb();
				wake_tx_thread(TX_LOGOUT, session);
				if ((session_timeout == 0)
				    || time_before(session->
						   logout_response_deadline,
						   session_timeout))
					session_timeout =
					    session->logout_response_deadline;
			} else {
				if ((session_timeout == 0)
				    || time_before(session->logout_deadline,
						   session_timeout))
					session_timeout =
					    session->logout_deadline;
			}
		} else
		    if (test_bit(SESSION_WINDOW_CLOSED, &session->control_bits)
			&& time_before_eq(session->last_window_check + HZ,
					  jiffies)) {
			/* command window closed, ping once a second to ensure 
			 * we find out when it re-opens.  Target should send 
			 * us an update when it does, but we're not very 
			 * trusting of target correctness.
			 */
			session->last_window_check = jiffies;
			DEBUG_QUEUE
			    ("iSCSI: session iscsi bus %d target id %d command "
			     "window closed, ExpCmdSN %u, MaxCmdSN %u, "
			     "polling target at %lu\n",
			     session->iscsi_bus, session->target_id,
			     session->ExpCmdSn, session->MaxCmdSn, jiffies);

			/* request a window update from the target with Nops */
			wake_tx_thread(TX_PING, session);

			if ((session_timeout == 0)
			    || time_before(session->last_window_check + HZ,
					   session_timeout))
				session_timeout =
				    session->last_window_check + HZ;
		}
	}

	return session_timeout;
}
