#include <unistd.h>
#include <slp.h>
#include "iscsi-slp-discovery.h"

#define ISCSI_SERVICE "service:iscsi:target"
#define MAX_IFACES 10
#define IP_ADDR_LEN 16
#define RECORD_LEN 2048
#define PREDICATE_LEN 2048
#define CRED_PREDICATE_LEN 512
#define NAME_PREDICATE_LEN 512
#define INITIATOR_NAME "/etc/initiatorname.iscsi"

static int record_begin;

struct InterfaceInfo {
	int iface_count;
	struct sockaddr_in iface_addr[MAX_IFACES];
};

struct slp_portal {
	struct slp_portal *next;
	char *ip;
	char *port;
	char *tag;
};

struct slp_target {
	struct slp_target *next;
	char *target_name;
	struct slp_portal *portals;
	struct slp_portal *tail;
};

struct slp_discovery_process {
	struct slp_discovery_process *next;
	struct slp_target *targets;
	struct slp_target *tail;
	struct iscsi_discovery_process *discovery;
};

struct slp_discovery_process_list {
	struct slp_discovery_process *head;
	struct slp_discovery_process *tail;
};

struct slp_discovery_pg_cookie {
	struct iscsi_discovery_process *discovery;
	char *target_name;
	char *ip;
	char *port;
};

struct slp_discovery_process_list *list = NULL;

void
add_slp_portal(struct slp_target *target, char *ip, char *port, char *tag)
{
	struct slp_portal *portal =
	    (struct slp_portal *) malloc(sizeof (struct slp_portal));

	portal->ip = ip;
	portal->port = port;
	portal->tag = tag;
	portal->next = NULL;

	target->tail->next = portal;
	target->tail = portal;

	debugmsg(7, "slp portal %p added\n", portal);
}

void
add_slp_target(struct slp_discovery_process *process,
	       char *name, char *ip, char *port, char *tag)
{
	struct slp_target *target = NULL;
	struct slp_portal *portal = NULL;

	target = (struct slp_target *) malloc(sizeof (struct slp_target));
	portal = (struct slp_portal *) malloc(sizeof (struct slp_portal));

	target->target_name = name;
	target->portals = target->tail = portal;
	target->next = NULL;

	portal->ip = ip;
	portal->port = port;
	portal->tag = tag;
	portal->next = NULL;

	process->tail->next = target;
	process->tail = target;

	debugmsg(7, "slp target %s added\n", target->target_name);
}

void
add_slp_discovery_process(struct iscsi_discovery_process *discovery,
			  char *name, char *ip, char *port, char *tag)
{
	struct slp_discovery_process *process;
	struct slp_target *target;
	struct slp_portal *portal;

	process = (struct slp_discovery_process *)
	    malloc(sizeof (struct slp_discovery_process));
	target = (struct slp_target *) malloc(sizeof (struct slp_target));
	portal = (struct slp_portal *) malloc(sizeof (struct slp_portal));

	process->discovery = discovery;
	process->targets = process->tail = target;
	process->next = NULL;

	target->target_name = name;
	target->portals = target->tail = portal;
	target->next = NULL;

	portal->ip = ip;
	portal->port = port;
	portal->tag = tag;
	portal->next = NULL;

	if (list) {
		list->tail->next = process;
		list->tail = process;
	} else {
		list = (struct slp_discovery_process_list *)
		    malloc(sizeof (struct slp_discovery_process_list));
		list->head = list->tail = process;
	}

	debugmsg(7, "slp discovery data buffered for discovery process: %d\n",
		 discovery->pid);
}

void
save_slp_discovery_data(char *name, char *ip,
			char *port, char *tag,
			struct iscsi_discovery_process *discovery)
{
	struct slp_discovery_process *process; 
	struct slp_target *target = NULL;
	int slp_discovery_process_exit = 0;
	int slp_target_exit = 0;

	if (list) {
		for (process = list->head; process; process = process->next) {
			if (process->discovery == discovery) {
				slp_discovery_process_exit = 1;
				for (target = process->targets; target;
				     target = target->next) {
					if (strcmp(target->target_name, 
					     name) == 0) {
						slp_target_exit = 1;
						break;
					}
				}
				break;
			}
		}

		if (slp_discovery_process_exit) {
			if (slp_target_exit)
				add_slp_portal(target, ip, port, tag);
			else
				add_slp_target(process, name, ip, port, tag);
		} else
			add_slp_discovery_process(discovery, name, ip, port,
						  tag);
	} else
		add_slp_discovery_process(discovery, name, ip, port, tag);
}

void
send_target_record(struct slp_target *target,
		   struct iscsi_discovery_process *discovery)
{
	struct slp_portal *portal;
	char record[RECORD_LEN];
	char *c = record;
	int len = 0;
	int result;

	memset(record, 0, RECORD_LEN);
	if (!record_begin) {
		sprintf(c, "DTN=%s\n", target->target_name);
		record_begin = 1;
	} else {
		sprintf(c, "TN=%s\n", target->target_name);
	}
	len += strlen(c);
	c += strlen(c);

	for (portal = target->portals; portal; portal = portal->next) {
		sprintf(c, "TT=%s\n", portal->tag);
		len += strlen(c);
		c += strlen(c);

		sprintf(c, "TP=%s\n", portal->port);
		len += strlen(c);
		c += strlen(c);

		sprintf(c, "TA=%s\n", portal->ip);
		len += strlen(c);
		c += strlen(c);
	}
	sprintf(c, ";\n");
	len += strlen(c);

	debugmsg(7, "sending target record %s\n", record);

	result = write(discovery->pipe_fd, record, len);
	if (result < 0)
		logmsg(AS_ERROR,
		       "Error in sending discovery data for target %s "
		       "by discovery process %d\n",
		       target->target_name, discovery->pid);

}

void
free_slp_portals(struct slp_portal *portals)
{
	struct slp_portal *portal; 

	while (portals) {
		portal = portals;

		if (portal->ip)
			free(portal->ip);
		if (portal->port)
			free(portal->port);
		if (portal->tag)
			free(portal->tag);

		portals = portals->next;
		free(portal);
	}
}

void
free_slp_targets(struct slp_target *targets)
{
	struct slp_target *target;

	while (targets) {
		target = targets;
		if (target->target_name)
			free(target->target_name);
		if (target->tail)
			target->tail = NULL;
		if (target->portals)
			free_slp_portals(target->portals);

		targets = targets->next;
		free(target);
	}
}

void
free_slp_discovery_process(struct slp_discovery_process *process)
{
	struct slp_discovery_process *p;
	struct slp_discovery_process *n;

	if (process) {
		if (process->targets) {
			process->tail = NULL;
			free_slp_targets(process->targets);
		}

		for (n = list->head, p = list->head; n; n = n->next) {
			if (n == process) {
				if (n == p)	/* head */
					list->head = process->next;
				else
					p->next = n->next;

				if (n == list->tail) {
					if (n == p)	/* this is the only 
							 * process in list 
							 */
						list->tail = list->head;
					else
						list->tail = p;
				}
				process->discovery = NULL;
				free(process);
				break;
			}
			p = n;
		}
	}
}

void
send_slp_discovery_data(struct iscsi_discovery_process *discovery)
{
	struct slp_discovery_process *process;
	struct slp_target *target; 

	if (list) {
		for (process = list->head; process; process = process->next) {
			if (process->discovery == discovery) {
				for (target = process->targets; target;
				     target = target->next) {
					send_target_record(target, discovery);
				}
				write(discovery->pipe_fd, "!\n",
				      sizeof ("!\n") - 1);
				record_begin = 0;
				break;
			}
		}
	}
}

void
free_slp_discovery_data(struct iscsi_discovery_process *discovery)
{
	struct slp_discovery_process *process;

	if (list) {
		for (process = list->head; process; process = process->next) {
			if (process->discovery == discovery) {
				free_slp_discovery_process(process);
				break;
			}
		}
		if ((list->head == NULL) && (list->tail == NULL)) {
			free(list);
			list = NULL;
		}
	}
}

int
IfaceGetInfo(struct InterfaceInfo *ifaceinfo,
	     struct iscsi_discovery_process *discovery)
{
	struct sockaddr *sa;
	struct sockaddr_in *sin;
	struct ifreq ifrlist[MAX_IFACES];
	struct ifreq ifrflags;
	struct ifconf ifc;
	int fd;
	int i;

	ifc.ifc_len = sizeof (struct ifreq) * MAX_IFACES;
	ifc.ifc_req = ifrlist;

	fd = socket(AF_INET, SOCK_STREAM, 0);
	if (fd == -1) {
		logmsg(AS_ERROR,
		       "Failed to create socket, for getting all IP interfaces "
		       "through discovery process pid = %d\n",
		       discovery->pid);
		return -1;
	}
	if (ioctl(fd, SIOCGIFCONF, &ifc) == -1) {
		logmsg(AS_ERROR,
		       "ioctl failure, in getting all IP interfaces through "
		       "discovery process pid = %d\n",
		       discovery->pid);
		return -1;
	}
	memset(ifaceinfo, 0, sizeof (struct InterfaceInfo));
	for (i = 0; i < ifc.ifc_len / sizeof (struct ifreq); i++) {
		sa = (struct sockaddr *) &(ifrlist[i].ifr_addr);
		if (sa->sa_family == AF_INET) {
			/* Get interface flags */
			memcpy(&ifrflags, &(ifrlist[i]), sizeof (struct ifreq));
			if (ioctl(fd, SIOCGIFFLAGS, &ifrflags) == 0) {
				/* skip the loopback interfaces */
				if ((ifrflags.ifr_flags & IFF_LOOPBACK) == 0) {
					sin = (struct sockaddr_in *) sa;
					memcpy(&
					       (ifaceinfo->
						iface_addr[ifaceinfo->
							   iface_count]), sin,
					       sizeof (struct sockaddr_in));
					ifaceinfo->iface_count++;
				}
			}
		}
	}
	return 0;
}

char *
get_all_ip(struct iscsi_discovery_process *discovery)
{
	int i;
	struct InterfaceInfo ifaceinfo;
	char *addrstr = NULL;

	if (IfaceGetInfo(&ifaceinfo, discovery) == 0) {
		addrstr = (char *) malloc(ifaceinfo.iface_count * IP_ADDR_LEN);
		memset(addrstr, 0, ifaceinfo.iface_count * IP_ADDR_LEN);

		for (i = 0; i < ifaceinfo.iface_count; i++) {
			strcat(addrstr,
			       inet_ntoa(ifaceinfo.iface_addr[i].sin_addr));
			if (i + 1 < ifaceinfo.iface_count) {
				strcat(addrstr, ",");
			}
		}
	}
	return addrstr;
}

SLPBoolean
SrvUrlCallback(SLPHandle hslp,
	       const char *srvurl,
	       unsigned short lifetime, int errcode, void *cookie)
{
	char *target_name, *targetname;
	char *ip;
	char *port, *port_id;
	char *tmp_url, *start_url= NULL, *attr_url;
	struct iscsi_discovery_process *discovery = NULL;
	char *addr;
	char *tag;

	if (errcode == SLP_OK) {
		debugmsg(7, "SLP url :%s,%i\n", srvurl, lifetime);
	}

	if (srvurl)
		start_url = strdup(srvurl);

	tmp_url = start_url;

	if (cookie)
		discovery = (struct iscsi_discovery_process *) cookie;

	if (tmp_url) {
		attr_url = strdup(tmp_url);
		if (strncmp(tmp_url, "service:iscsi:target://", 23) == 0) {
			tmp_url += 22;
			*tmp_url = '\0';
			tmp_url++;
			ip = tmp_url;

			port = strchr(tmp_url, ':');
			*port = '\0';
			port++;
			tmp_url = port;

			target_name = strchr(tmp_url, '/');
			*target_name = '\0';
			target_name++;

			tmp_url = target_name;

			debugmsg(7,
				 "iSCSI Target: ip %s port %s target name %s, "
				 "discovered by discovery process %d\n",
				 ip, port, target_name, discovery->pid);

			/* slp query do not return any information related to
			 * target tag. For time being setting it to 1.
			 */
			addr = (char *) malloc(strlen(ip) + 1);
			memset(addr, 0, strlen(ip) + 1);
			memcpy(addr, ip, strlen(ip));

			port_id = (char *) malloc(strlen(port) + 1);
			memset(port_id, 0, strlen(port) + 1);
			memcpy(port_id, port, strlen(port));

			targetname = (char *) malloc(strlen(target_name) + 1);
			memset(targetname, 0, strlen(target_name) + 1);
			memcpy(targetname, target_name, strlen(target_name));

			tag = (char *) malloc(strlen("0") + 1);
			memset(tag, 0, strlen("0") + 1);
			strcpy(tag, "0");
			save_slp_discovery_data(targetname, addr, port_id, tag,
						discovery);
			if ( discovery->flag == SLP_UNICAST_ENABLED )
				slp_unicast_attr_query(discovery, attr_url, 
						       "portal-group", 
						       targetname, addr, 
						       port_id);
			if ( discovery->flag == SLP_MULTICAST_ENABLED )
				slp_multicast_attr_query(discovery, attr_url, 
						       "portal-group", 
						       targetname, addr, 
						       port_id);
		}
		free(start_url);
	} else {
		if (discovery) {
			send_slp_discovery_data(discovery);
			free_slp_discovery_data(discovery);
		}
	}
	return SLP_TRUE;
}

SLPBoolean 
AttrQueryCallback(SLPHandle hslp,
		  const char* attrlist,
		  int errcode,
		  void* cookie)
{
	struct slp_discovery_process *process;
	struct slp_target *target;
	struct slp_portal *portal;
	struct iscsi_discovery_process *discovery = NULL;
	char *tmp_url, *start_url = NULL;
	char *tag, *tmp;
	char *ip, *port, *target_name = NULL;

	if (errcode == SLP_OK) {
		debugmsg(7, "SLP attribute list :%s\n",attrlist);
	}

	if (attrlist) {
		start_url = strdup(attrlist);
	}

	tmp_url = start_url;

	if (cookie) {
		discovery = (struct iscsi_discovery_process *)
			     ((struct slp_discovery_pg_cookie *)cookie)->
				discovery;
		ip = (char *)((struct slp_discovery_pg_cookie *)cookie)->ip;
		port = (char *)((struct slp_discovery_pg_cookie *)cookie)->port;
		target_name = (char *)((struct slp_discovery_pg_cookie *)
					cookie)->target_name;
	}

	if (tmp_url) {
		tmp = strstr(tmp_url, "portal-group=");
		tmp = strchr(tmp, '=');
		tmp++;
		tmp = strtok(tmp,")");
		tag = strdup(tmp);
		if (list) {
			for (process = list->head; process; 
			     process = process->next) {
				if (process->discovery == discovery) {
					for (target = process->targets; target;
					     target = target->next) {
						if (strcmp(target->target_name, 
							   target_name) == 0) {
							for (portal = target->
							     portals; portal; 
							     portal = portal->
							     next) {
								portal->tag = 
								tag;
							}
						}
					}
				}
			}
		}
		free(start_url);
	}
	else {
		return SLP_TRUE;
	}
	return SLP_TRUE;
}

void
slp_multicast_srv_query(struct iscsi_discovery_process *discovery, char *pl,
			unsigned short flag)
{
	SLPError result;
	SLPHandle hslp;
	struct iscsi_slp_config *config = discovery->entry->config.slp;

	if (SLPOpen(NULL, SLP_FALSE, &hslp) == SLP_OK) {

		if ((result =
		     SLPAssociateIFList(hslp, config->interfaces)) != SLP_OK) {
			logmsg(AS_ERROR,
			       "SLP query error on discovery pid = %d "
			       "with SLP errorcode: %i\n",
			       discovery->pid, result);
			SLPClose(hslp);
			return;
		}
	}

	switch (flag) {

	case GENERIC_QUERY:
		result = SLPFindSrvs(hslp,
				     ISCSI_SERVICE,
				     NULL,
				     pl,
				     SrvUrlCallback,
				     discovery);
		break;

	default:
		result = SLP_FALSE;
	
	}

	if (result != SLP_OK) {
		logmsg(AS_ERROR,
		       "SLP query error on discovery pid = %d "
		       "with SLP errorcode: %i\n",
		       discovery->pid, result);
	}
	SLPClose(hslp);
}

void 
slp_multicast_attr_query(struct iscsi_discovery_process *discovery, char* url, 
			 char *pl, char *target_name, char *ip, char *port)
{
	SLPError    result;
	SLPHandle   hslp;
	struct iscsi_slp_config *config = discovery->entry->config.slp;
	struct slp_discovery_pg_cookie cookie;

	if (SLPOpen(NULL,SLP_FALSE,&hslp) == SLP_OK) {
		if ((result = SLPAssociateIFList(hslp, config->interfaces)) \
				!= SLP_OK) {
			logmsg(AS_ERROR, "SLP query error on discovery %p with "
				       	 "SLP errorcode: %i\n",discovery, 
							       result);
			SLPClose(hslp);
			return;
		}       
	}

	cookie.discovery = (struct iscsi_discovery_process *)discovery;
	cookie.target_name = (char *) malloc(strlen(target_name) + 1);
	memset(cookie.target_name, 0, strlen(target_name) + 1);
	strcpy(cookie.target_name, target_name);

	cookie.ip = (char *) malloc(strlen(ip) + 1);
	memset(cookie.ip, 0, strlen(ip) + 1);
	strcpy(cookie.ip, ip);

	cookie.port = (char *) malloc(strlen(port) + 1);
	memset(cookie.port, 0, strlen(port) + 1);
	strcpy(cookie.port, port);

	result = SLPFindAttrs(hslp,
			      ISCSI_SERVICE,
			      NULL,
			      pl,
			      AttrQueryCallback,
			      &cookie);
                       
	if (result != SLP_OK) {
		logmsg(AS_ERROR, "SLP query error on discovery %p "
				 "with SLP errorcode: %i\n",discovery, result);
	}   
	SLPClose(hslp);
}

void
slp_unicast_srv_query(struct iscsi_discovery_process *discovery, char *pl,
			unsigned short flag)
{
	SLPError result;
	SLPHandle hslp;
	struct iscsi_slp_config *config = discovery->entry->config.slp;

	if (SLPOpen(NULL, SLP_FALSE, &hslp) == SLP_OK) {

		if ((result =
		     SLPAssociateIP(hslp, config->address)) != SLP_OK) {
			logmsg(AS_ERROR,
			       "SLP query error on discovery pid = %d "
			       "with SLP errorcode: %i\n",
			       discovery->pid, result);
			SLPClose(hslp);
			return;
		}
	}

	switch (flag) {

	case GENERIC_QUERY:
		result = SLPFindSrvs(hslp,
				     ISCSI_SERVICE,
				     NULL,
				     pl,
				     SrvUrlCallback,
				     discovery);
		break;

	default:
		result = SLP_FALSE;

	}

	if (result != SLP_OK) {
		logmsg(AS_ERROR,
		       "SLP query error on discovery pid = %d "
		       "with SLP errorcode: %i\n",
		       discovery->pid, result);
	}
	SLPClose(hslp);
}

void 
slp_unicast_attr_query(struct iscsi_discovery_process *discovery, char* url, 
			 char *pl, char *target_name, char *ip, char *port)
{
	SLPError    result;
	SLPHandle   hslp;
	struct iscsi_slp_config *config = discovery->entry->config.slp;
	struct slp_discovery_pg_cookie cookie;

	if (SLPOpen(NULL,SLP_FALSE,&hslp) == SLP_OK) {
		if ((result = SLPAssociateIP(hslp, config->address)) \
				!= SLP_OK) {
			logmsg(AS_ERROR, "SLP query error on discovery %p with "
				       	 "SLP errorcode: %i\n",discovery, 
							       result);
			SLPClose(hslp);
			return;
		}       
	}   

	cookie.discovery = (struct iscsi_discovery_process *)discovery;
	cookie.target_name = (char *) malloc(strlen(target_name) + 1);
	memset(cookie.target_name, 0, strlen(target_name) + 1);
	strcpy(cookie.target_name, target_name);

	cookie.ip = (char *) malloc(strlen(ip) + 1);
	memset(cookie.ip, 0, strlen(ip) + 1);
	strcpy(cookie.ip, ip);

	cookie.port = (char *) malloc(strlen(port) + 1);
	memset(cookie.port, 0, strlen(port) + 1);
	strcpy(cookie.port, port);

	result = SLPFindAttrs(hslp,
			      url,
			      NULL,
			      pl,
			      AttrQueryCallback,
			      &cookie);
                       
	if (result != SLP_OK) {
		logmsg(AS_ERROR, "SLP query error on discovery %p "
				 "with SLP errorcode: %i\n",discovery, result);
	}   
	SLPClose(hslp);
}

char *
generate_pl(char *interfaces)
{
	char slp_predicate_list[PREDICATE_LEN];
	char *pl = slp_predicate_list;
	int SLPpl;
	char *temp, *startp = strdup(interfaces);
	char *endp;
	char *c;

	temp = startp;

	if (strlen(temp)) {
		sprintf(pl, "(auth-addr=any)");

		endp = strrchr(temp, ',');	/* more than one IP address */
		if (endp) {
			while (endp > temp) {
				c = strchr(temp, ',');
				*c = '\0';
				c += 1;
				sprintf(pl + strlen(pl), "(auth-addr=%s)",
					temp);
				temp = c;
			}
		}
		sprintf(pl + strlen(pl), "(auth-addr=%s)", temp);
		free(startp);
		SLPpl = 1;
	} else
		SLPpl = 0;

	if (SLPpl) {
		pl = strdup(slp_predicate_list);
		return pl;
	}
	return NULL;
}

char *
multicast_predicate_list(struct iscsi_discovery_process *discovery)
{
	struct iscsi_slp_config *config = discovery->entry->config.slp;
	char *mpl = NULL;

	if (config->interfaces && strlen(config->interfaces)) {
		if (strcmp(config->interfaces, "none") == 0) {
			debugmsg(1,
				 "SLP Multicast is disabled for "
				 "discovery process pid = %d\n",
				 discovery->pid);
		} else if (strcmp(config->interfaces, "all") == 0) {
			free(config->interfaces);
			config->interfaces = get_all_ip(discovery);
			if (config->interfaces)
				mpl = generate_pl(config->interfaces);
		} else {	/* comma separated list of IP's */
			mpl = generate_pl(config->interfaces);
		}
	}
	return mpl;
}

char *
unicast_predicate_list(struct iscsi_discovery_process *discovery)
{
	struct iscsi_slp_config *config = discovery->entry->config.slp;
	char *interfaces; 
	char *upl = NULL;

	if (config->address && strlen(config->address)) {
		interfaces = get_all_ip(discovery);
		if (interfaces)
			upl = generate_pl(interfaces);
	}

	return upl;
}

char *
cred_predicate_list(struct iscsi_discovery_process *discovery)
{
	struct iscsi_slp_config *config = discovery->entry->config.slp;
	char cred_predicate_list[CRED_PREDICATE_LEN];
	char *slpc = cred_predicate_list;

	sprintf(slpc, "(auth-cred=any)");

	if (strlen(config->auth_options.username)) {
		/* We ll have one more case when we add SRP support
		 */
		switch(config->auth_options.authmethod) {

		case CHAP_AUTHENTICATION:
		default:
			sprintf(slpc + strlen(slpc), "(auth-cred=chap/%s)",
				config->auth_options.username);
		}
	}

	slpc = strdup(cred_predicate_list);

	return slpc;
}

char *
name_predicate_list(void)
{
	char name_predicate[NAME_PREDICATE_LEN];
	char *slpn = name_predicate;
	char *initiator_name;

	sprintf(slpn, "(auth-name=any)");
	initiator_name = get_iscsi_initiatorname(INITIATOR_NAME);

	if (initiator_name && strlen(initiator_name)) {
		sprintf(slpn + strlen(slpn), "(auth-name=%s)", initiator_name);
	}

	slpn = strdup(name_predicate);

	return slpn;
}

char *
generate_predicate_list(struct iscsi_discovery_process *discovery,
			unsigned short *flag)
{
	char *mpl;
	char *upl;
	char *cpl;
	char *npl;
	char *pl;
	int plen = 0;

	mpl = multicast_predicate_list(discovery);
	upl = unicast_predicate_list(discovery);
	cpl = cred_predicate_list(discovery);
	npl = name_predicate_list();

	if (mpl) {
		debugmsg(7,
			 "multicast predicate list for "
			 "discovery process pid = %d is %s\n",
			 discovery->pid, mpl);
		plen += strlen(mpl);
		*flag = SLP_MULTICAST_ENABLED;
	}

	if (upl) {
		debugmsg(7,
			 "unicast predicate list for discovery "
			 "process pid = %d is %s\n",
			 discovery->pid, upl);
		plen += strlen(upl);
		*flag = SLP_UNICAST_ENABLED;
	}

	if (plen) {
		if (cpl) {
			debugmsg(7,
				 "credential predicate list for "
				 "discovery process pid = %d is %s\n",
				 discovery->pid, cpl);
			plen += strlen(cpl);
		}
		if (npl) {
			debugmsg(7,
				 "name predicate list for discovery "
				 "process pid = %d is %s\n",
				 discovery->pid, npl);
			logmsg(AS_ERROR,
			       "name predicate list for "
			       "discovery process pid = %d is %s\n",
			       discovery->pid, npl);
			plen += strlen(npl);
		}
	}

	if (plen) {
		char *c;

		plen += 16;	/* This extra space is required for 
				 * "(&(|%s)(|%s)(|%s)(|%s))"
				 * whearas, 1st %s is for mpl,
				 *          2nd %s is for upl,
				 *          3rd %s is for cpl,
				 *          4th %s is for npl.
				 */
		pl = (char *) malloc(sizeof (char) * plen);
		memset(pl, 0, plen);
		c = pl;

		sprintf(c, "(&");
		c += strlen(c);

		if (mpl) {
			sprintf(c, "(|%s)", mpl);
			c += strlen(c);
			free(mpl);
		}
		if (upl) {
			sprintf(c, "(|%s)", upl);
			c += strlen(c);
			free(upl);
		}

		sprintf(c, "(|%s)", cpl);
		c += strlen(c);
		free(cpl);

		sprintf(c, "(|%s)", npl);
		c += strlen(c);
		free(npl);

		sprintf(c, ")");	/* close brace correspond to '(&' */

		debugmsg(3,
			 "SLP predicate list for discover process %d is %s\n",
			 discovery->pid, pl);

		return pl;
	}

	return NULL;
}
