/*
 * iSCSI driver for Linux
 * Copyright (C) 2001 Cisco Systems, Inc.
 * maintained by linux-iscsi-devel@lists.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi-task.c,v 1.1.2.10 2003/12/23 10:53:24 krishmnc Exp $
 */

#include "iscsi-login.h"
#include "iscsi-probe.h"
#include "iscsi-crc.h"
#include "iscsi.h"

extern int translate_deferred_sense; 


void
iscsi_task_ctor(void *obj, kmem_cache_t * cache, unsigned long flags)
{
	struct iscsi_task *task = (struct iscsi_task *) obj;

	memset(task, 0, sizeof (*task));
	task->flags = 0;
	task->itt = RSVD_TASK_TAG;
	task->ttt = RSVD_TASK_TAG;
	task->mgmt_itt = RSVD_TASK_TAG;
	INIT_LIST_HEAD(&task->task_group_link);
	INIT_LIST_HEAD(&task->task_order_link);
	init_timer(&task->timer);
	atomic_set(&task->refcount, 0);
}

/* must hold the task_lock to call this */
struct iscsi_task *
find_session_task(struct iscsi_session *session, uint32_t itt)
{
	struct iscsi_task *task = NULL;

	list_for_each_entry(task, &session->task_arrival_order, 
			    task_order_link) {
		if (task->itt == itt) {
			DEBUG_FLOW
			    ("iSCSI: found itt %u, refcount %d\n", itt,
			     atomic_read(&task->refcount));
			return task;
		}
	}

	return NULL;
}

/* must hold the task_lock to call this */
struct iscsi_task *
find_session_mgmt_task(struct iscsi_session *session, uint32_t mgmt_itt)
{
	struct iscsi_task *task; 

	list_for_each_entry(task, &session->task_arrival_order,
 			    task_order_link) {
		if (task->mgmt_itt == mgmt_itt) {
			DEBUG_FLOW
			    ("iSCSI: found mgmt_itt %u, refcount %d\n",
			     mgmt_itt, atomic_read(&task->refcount));
			return task;
		}
	}

	return NULL;
}

/* don't actually need this at the moment */
/* must hold the task_lock to call this */
struct iscsi_task *
find_task_for_cmnd(struct iscsi_session *session, Scsi_Cmnd * sc)
{
	struct iscsi_task *task;

	list_for_each_entry(task, &session->task_arrival_order,
			    task_order_link) {
		if (task->scsi_cmnd == sc) {
			DEBUG_FLOW("iSCSI: found itt %u, for cmnd 0x%x\n",
				   task->itt, sc->cmnd[0]);
			return task;
		}
	}

	return NULL;
}

/* add a task to the collection.  Must hold the task_lock to do this. */
void
add_task(struct list_head *collection, struct iscsi_task *task)
{
	list_add_tail(&task->task_group_link,collection);
}

/* must hold the task_lock when calling this */
struct iscsi_task *
pop_task(struct list_head *collection)
{
	struct iscsi_task *task; 
	struct list_head * task_link;

	if (list_empty(collection)) {
		return NULL;
	} else {
		task_link = collection->next;
		list_del_init(task_link);
		task = list_entry(task_link, struct iscsi_task, 
				  task_group_link);
		return task;
	}	
}

void
unlink_task(struct list_head *collection, struct iscsi_task *task)
{
	/* unlink the task from the collection */
	
	list_del_init(&task->task_group_link);
}

/* if the task for the itt is found in the collection, remove it, and return it.
 * otherwise, return NULL.  Must hold the task_lock to call this.
 */
struct iscsi_task *
remove_task(struct list_head *collection, uint32_t itt)
{
	struct iscsi_task *task; 
	struct iscsi_task *search;

	list_for_each_entry(search, collection, task_group_link) {
		if (search->itt == itt) {
			task = search;
			unlink_task(collection, task);
			return task;
		}
	}

	return NULL;
}

/* if the task for the itt is found in the collection, remove it, and return it.
 * otherwise, return NULL.  Must hold the task_lock to call this.
 */
struct iscsi_task *
remove_task_for_cmnd(struct list_head *collection, Scsi_Cmnd * sc)
{
	struct iscsi_task *task; 
	struct iscsi_task *search;

	list_for_each_entry(search, collection, task_group_link) {
		if (search->scsi_cmnd == sc) {
			task = search;
			unlink_task(collection, task);
			return task;
		}
	}

	return NULL;
}

/* caller must hold the session's task lock */
struct iscsi_task *
alloc_task(struct iscsi_session *session)
{
	struct iscsi_task *task;
	struct iscsi_hba *hba = session->hba;

	if (hba == NULL) {
		printk
		    ("iSCSI: session iscsi bus %d target id %d "
		     "alloc_task failed - NULL HBA\n",
		     session->iscsi_bus, session->target_id);
		return NULL;
	} else if (hba->task_cache == NULL) {
		printk
		    ("iSCSI: session iscsi bus %d target id %d "
		     "alloc_task failed - NULL HBA task cache\n",
		     session->iscsi_bus, session->target_id);
		return NULL;
	}

	if ((task = kmem_cache_alloc(hba->task_cache, SLAB_ATOMIC))) {
		session->tasks_allocated++;
		task->session = session;
		DEBUG_ALLOC
		    ("iSCSI: session iscsi bus %d target id %d "
		     "allocated task itt %u from cache at %lu\n",
		     session->iscsi_bus, session->target_id, task->itt,
		     jiffies);
	} else if (session->preallocated_task) {
		/* if the task cache is empty, we fall back 
		 * to the session's preallocated task, which
		 * guarantees us at least some forward progress 
		 * on every session.
		 */
		task = session->preallocated_task;
		session->preallocated_task = NULL;
		task->session = session;
		/* don't log by default.  We're more concerned with when a
		 * task alloc fails than when we use the preallocated task. 
		 */
		if (LOG_ENABLED(ISCSI_LOG_ALLOC))
			printk
			    ("iSCSI: session iscsi bus %d to (%u %u %u *) "
			     "task cache empty, using preallocated task itt %u "
			     "at %lu\n",
			     session->iscsi_bus, session->host_no,
			     session->channel, session->target_id, task->itt,
			     jiffies);
	} else {
		/* better luck later */
		task = NULL;
	}
	if (task) {
		 INIT_LIST_HEAD((&(task->task_order_link)));
		 INIT_LIST_HEAD((&(task->task_group_link)));
	}
	return task;
}

/* caller must hold the session's task lock */
void
free_task(struct iscsi_session *session, struct iscsi_task *task)
{
	struct iscsi_hba *hba;

	if (!task) {
		printk("iSCSI: free_task couldn't free NULL task\n");
		return;
	}
	if (!session) {
		printk
		    ("iSCSI: free_task couldn't find session for task itt %u\n",
		     task->itt);
		return;
	}
	hba = session->hba;
	if (!hba) {
		printk("iSCSI: free_task couldn't find HBA for task %u\n",
		       task->itt);
		return;
	}

	if ((!list_empty(&task->task_group_link))
	    ||(!list_empty(&task->task_order_link)) ) {
		/* this is a memory leak, which is better than memory corruption
		 */
		printk
		    ("iSCSI: bug - tried to free task itt %u but is currently"
                     " in a task list\n ", task->itt);

		return;
	}

	DEBUG_QUEUE
	    ("iSCSI: session iscsi bus %d target id %d  free_task itt %u\n",
	     task->session->iscsi_bus, task->session->target_id, task->itt);

	if (test_bit(TASK_PREALLOCATED, &task->flags)) {
		if (session->preallocated_task) {
			printk
			    ("iSCSI: bug - session iscsi bus %d target id %d "
			     "has preallocated task, really freeing itt %u "
			     "flags 0x%0lx at %lu\n",
			     session->iscsi_bus, session->target_id,
			     task->itt, task->flags, jiffies);

			/* reinitialize the task for later use */
			iscsi_task_ctor(task, NULL, 0);

			kmem_cache_free(hba->task_cache, task);
		} else {
			/* reinitialize the task for later use */
			iscsi_task_ctor(task, NULL, 0);
			__set_bit(TASK_PREALLOCATED, &task->flags);

			/* save it for the next memory emergency */
			session->preallocated_task = task;

			/* wake up the tx thread, since it may have been 
			 * forced to stop sending tasks once the prealloacted 
			 * task was in use. Now that the preallocated task is 
			 * back, we can guarantee this session can allocate at 
			 * least one more task.  Too many wakeups is better 
			 * than too few.  
			 */
			printk
			    ("iSCSI: session iscsi bus %d to (%u %u %u *) "
			     "done using preallocated task itt %u at %lu\n",
			     session->iscsi_bus, session->host_no,
			     session->channel, session->target_id, task->itt,
			     jiffies);
			wake_tx_thread(TX_SCSI_COMMAND, session);
		}
	} else {
		session->tasks_freed++;

		/* reinitialize the task for later use */
		iscsi_task_ctor(task, NULL, 0);

		/* return it to the cache */
		kmem_cache_free(hba->task_cache, task);
	}
}

/* As long as the tx thread is the only caller, no locking
 * is required.  If any other thread also needs to call this,
 * then all callers must be changed to agree on some locking
 * protocol.  Currently, some but not all caller's are holding
 * the session->task_lock.
 */
uint32_t
allocate_itt(struct iscsi_session *session)
{
	uint32_t itt = 0;

	if (session) {
		itt = session->itt++;
		/* iSCSI reserves 0xFFFFFFFF, this driver reserves 0 */
		if (session->itt == RSVD_TASK_TAG)
			session->itt = 1;
	}
	return itt;
}

/* Caller must hold the session's task_lock.  Associating a task with
 * a session causes it to be completed on a session drop or target
 * reset, along with all other session tasks, in the order they were
 * added to the session.  Preserving the ordering is required by the
 * Linux SCSI architecture.  Tasks that should not be completed to the
 * Linux SCSI layer (because the eh_abort_handler has or will return
 * SUCCESS for it) get removed from the session, though they may still
 * be in various task collections so that PDUs relating to them can be
 * sent or received.
 */
void
add_session_task(struct iscsi_session *session, struct iscsi_task *task)
{
	if (atomic_read(&session->num_active_tasks) == 0) {
		/* session going from idle to active, pretend we just
		 * received something, so that the idle period before 
		 * this doesn't cause an immediate timeout.
		 */
		session->last_rx = jiffies;
		smp_mb();
	}
	atomic_inc(&session->num_active_tasks);

	/* set task info */
	task->session = session;
	task->itt = allocate_itt(session);

	DEBUG_QUEUE
	    ("iSCSI: task allocated itt %u for command 0x%x, session "
	     "iscsi bus %d target id %d to %s\n",
	     task->itt, task->scsi_cmnd->cmnd[0], session->iscsi_bus,
	     session->target_id, session->log_name);

	/* add it to the session task ordering list */
	list_add_tail(&task->task_order_link, &session->task_arrival_order);

	DEBUG_FLOW
	    ("iSCSI: task %p itt %u, added to session iscsi bus %d "
	     "target id %d to %s\n",
	     task, task->itt, session->iscsi_bus, session->target_id,
	     session->log_name);
}

int
remove_session_task(struct iscsi_session *session, struct iscsi_task *task)
{
	/* remove the task from the session's arrival_order collection */

	list_del_init(&task->task_order_link);

	if (atomic_dec_and_test(&session->num_active_tasks)) {
		/* no active tasks, ready to logout */
		if (test_bit(SESSION_LOGOUT_REQUESTED, &session->control_bits)) {
			DEBUG_INIT
			    ("iSCSI: session iscsi bus %d target id %d now has "
			     "no active tasks, queueing logout at %lu\n",
			     session->iscsi_bus, session->target_id, jiffies);
			wake_tx_thread(TX_LOGOUT, session);
		}
	}

	return 1;
}

void
process_task_response(struct iscsi_session *session, struct iscsi_task *task,
		      struct IscsiScsiRspHdr *stsrh, unsigned char *sense_data,
		      unsigned int senselen)
{
	Scsi_Cmnd *sc = task->scsi_cmnd;
	int needs_retry = 0;
	int slow_retry = 0;
	unsigned int expected;

	DEBUG_FLOW
	    ("iSCSI: session iscsi bus %d target id %d recv_cmd - itt %u, "
	     "cdb 0x%x, cmd_len %d, rsp dlength %d, senselen %d\n",
	     session->iscsi_bus, session->target_id, task->itt, sc->cmnd[0],
	     sc->cmd_len, ntoh24(stsrh->dlength), senselen);

	/* default to just passing along the SCSI status.  
	 * We may change this later 
	 */
	sc->result = host_byte(DID_OK) | status_byte(stsrh->cmd_status);

	/* grab any sense data that came with the command.  It could be
	 * argued that we should only do this if the SCSI status is check
	 * condition.  It could also be argued that the target should only
	 * send sense if the SCSI status is check condition.  If the
	 * target bothered to send sense, we pass it along, since it
	 * may indicate a problem, and it's safer to report a possible
	 * problem than it is to assume everything is fine.
	 */
	if (senselen) {
		/* fill in the Scsi_Cmnd's sense data */
		memset(sc->sense_buffer, 0, sizeof (sc->sense_buffer));
		memcpy(sc->sense_buffer, sense_data,
		       MIN(senselen, sizeof (sc->sense_buffer)));

		/* if sense data logging is enabled, or it's deferred
		 * sense that we're going to do something special with, 
		 * or if it's an unexpected unit attention, which Linux doesn't
		 * handle well, log the sense data.
		 */
		if ((LOG_ENABLED(ISCSI_LOG_SENSE)) ||
		    (((sense_data[0] == 0x71) || (sense_data[0] == 0xF1))
		     && translate_deferred_sense)
		    || ((SENSE_KEY(sense_data) == UNIT_ATTENTION)
			&& (test_bit(SESSION_RESETTING, &session->control_bits)
			    == 0))) {
			if (senselen >= 26) {
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "recv_cmd cdb 0x%x, status 0x%x, response "
				     "0x%x, senselen %d, "
				     "key %02x, ASC/ASCQ %02X/%02X, itt %u to "
				     "(%u %u %u %u), %s\n"
				     "iSCSI: Sense %02x%02x%02x%02x "
				     "%02x%02x%02x%02x %02x%02x%02x%02x "
				     "%02x%02x%02x%02x "
				     "%02x%02x%02x%02x %02x%02x%02x%02x "
				     "%02x%02x\n",
				     session->iscsi_bus, session->target_id,
				     sc->cmnd[0], stsrh->cmd_status,
				     stsrh->response, senselen,
				     SENSE_KEY(sense_data), ASC(sense_data),
				     ASCQ(sense_data), task->itt,
				     sc->device->host->host_no,
				     sc->device->channel, sc->device->id,
				     sc->device->lun, session->log_name,
				     sense_data[0], sense_data[1], 
				     sense_data[2], sense_data[3], 
				     sense_data[4], sense_data[5],
				     sense_data[6], sense_data[7], 
				     sense_data[8], sense_data[9],
				     sense_data[10], sense_data[11],
				     sense_data[12], sense_data[13],
				     sense_data[14], sense_data[15], 
				     sense_data[16], sense_data[17],
  				     sense_data[18], sense_data[19],
				     sense_data[20], sense_data[21],
				     sense_data[22], sense_data[23], 
				     sense_data[24], sense_data[25]);
			} else if (senselen >= 18) {
				printk
				    ("iSCSI: session iscsi_bus %d target id %d "
				     "recv_cmd cdb 0x%x, status 0x%x, response "
				     "0x%x, senselen %d, "
				     "key %02x, ASC/ASCQ %02X/%02X, itt %u to "
				     "(%u %u %u %u), %s\n"
				     "iSCSI: Sense %02x%02x%02x%02x "
				     "%02x%02x%02x%02x %02x%02x%02x%02x "
				     "%02x%02x%02x%02x %02x%02x\n",
				     session->iscsi_bus, session->target_id,
				     sc->cmnd[0], stsrh->cmd_status,
				     stsrh->response, senselen,
				     SENSE_KEY(sense_data), ASC(sense_data),
				     ASCQ(sense_data), task->itt,
				     sc->device->host->host_no, 
				     sc->device->channel, sc->device->id,
				     sc->device->lun, session->log_name,
				     sense_data[0], sense_data[1], 
				     sense_data[2], sense_data[3],
				     sense_data[4], sense_data[5],
				     sense_data[6], sense_data[7],
				     sense_data[8], sense_data[9], 
				     sense_data[10], sense_data[11], 
				     sense_data[12], sense_data[13],
				     sense_data[14], sense_data[15],
				     sense_data[16], sense_data[17]);
			} else {
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "recv_cmd cdb 0x%x, status 0x%x, response "
				     "0x%x, senselen %d, key %02x, "
				     "itt %u to (%u %u %u %u), %s\n"
				     "iSCSI: Sense %02x%02x%02x%02x "
				     "%02x%02x%02x%02x\n",
				     session->iscsi_bus, session->target_id,
				     sc->cmnd[0], stsrh->cmd_status,
				     stsrh->response, senselen,
				     SENSE_KEY(sense_data), task->itt,
				     sc->device->host->host_no,
				     sc->device->channel, sc->device->id,
				     sc->device->lun, session->log_name,
				     sense_data[0], sense_data[1],
				     sense_data[2], sense_data[3], 
				     sense_data[4], sense_data[5], 
				     sense_data[6], sense_data[7]);
			}
		}
	} else if ((stsrh->cmd_status == SAM_STAT_CHECK_CONDITION)
		   && (senselen == 0)) {
		/* check condition with no sense.  We need to avoid this, 
		 * since the Linux SCSI code could put the command in 
		 * SCSI_STATE_FAILED, which it's error recovery doesn't appear 
		 * to handle correctly, and even if it does, we're trying to 
		 * bypass all of the Linux error recovery code
		 * to avoid blocking all I/O to the HBA.  Fake some sense 
		 * data that gets a retry from Linux.
		 */
		printk
		    ("iSCSI: session iscsi bus %d target id %d recv_cmd "
		     "generating sense for itt %u, status 0x%x, response 0x%x,"
		     " senselen %d, cdb 0x%x to (%u %u %u %u) at %lu\n",
		     session->iscsi_bus, session->target_id, task->itt,
		     stsrh->cmd_status, stsrh->response, senselen, sc->cmnd[0],
		     sc->device->host->host_no, sc->device->channel,
		     sc->device->id, sc->device->lun, jiffies);

		/* report a complete underflow */
		stsrh->residual_count = htonl(iscsi_expected_data_length(sc));
		stsrh->flags |= ISCSI_FLAG_CMD_UNDERFLOW;

		memset(sc->sense_buffer, 0, sizeof (sc->sense_buffer));
		sc->sense_buffer[0] = 0x70;
		sc->sense_buffer[2] = ABORTED_COMMAND;	
			/* so that scsi_check_sense always returns 
			 * NEEDS_RETRY to scsi_decide_dispostion 
			 */
		sc->sense_buffer[7] = 0x6;
		sc->sense_buffer[12] = 0x04;	
			/* ASC/ASCQ 04/01 appears to always get a retry 
			 * from scsi_io_completion, so we use that 
			 */
		sc->sense_buffer[13] = 0x01;
	}
#if FAKE_NO_REPORT_LUNS
	else if (test_bit(SESSION_ESTABLISHED, &session->control_bits) && sc
		 && (senselen == 0) && (sc->cmnd[0] == REPORT_LUNS)
		 && (stsrh->cmd_status == 0) && (stsrh->response == 0)) {
		printk
		    ("iSCSI: session iscsi bus %d target id %d faking failed "
		     "REPORT_LUNS itt %u, CmdSN %u, cdb 0x%x to "
		     "(%u %u %u %u)\n",
		     session->iscsi_bus, session->target_id, itt, task->cmdsn,
		     sc->cmnd[0], sc->device->host->host_no, sc->device->channel
		     , sc->device->id, sc->device->lun);

		/* fake an illegal request check condition for this command.
		 */
		sc->sense_buffer[0] = 0x70;
		sc->sense_buffer[2] = ILLEGAL_REQUEST;
		sc->sense_buffer[7] = 0x6;
		sc->sense_buffer[12] = 0x20;	/* INVALID COMMAND OPCODE */
		sc->sense_buffer[13] = 0x00;
		sc->result = host_byte(DID_OK) | status_byte(0x02);
		stsrh->cmd_status = 0x2;
		stsrh->residual_count = htonl(iscsi_expected_data_length(sc));
		stsrh->flags |= ISCSI_FLAG_CMD_UNDERFLOW;
	}
#endif
#if FAKE_PROBE_CHECK_CONDITIONS
	else if (test_bit(SESSION_ESTABLISHED, &session->control_bits) && sc
		 && (senselen == 0) && (sc->scsi_done == iscsi_done)
		 && (sc->retries <= 1) && (stsrh->cmd_status == 0)
		 && (stsrh->response == 0)) {
		printk
		    ("iSCSI: session iscsi bus %d target id %d faking failed "
		     "probe itt %u, CmdSN %u, cdb 0x%x to (%u %u %u %u)\n",
		     session->iscsi_bus, session->target_id, itt, task->cmdsn,
		     sc->cmnd[0], sc->device->host->host_no, 
		     sc->device->channel, sc->device->id, sc->device->lun);

		/* fake an command aborted check condition to test the 
		 * recovery of probe commands 
		 */
		sc->sense_buffer[0] = 0x70;
		sc->sense_buffer[2] = NOT_READY;
		sc->sense_buffer[7] = 0x6;
		sc->sense_buffer[12] = 0x08;
		sc->sense_buffer[13] = 0x00;
		stsrh->cmd_status = 0x2;
		stsrh->residual_count = htonl(iscsi_expected_data_length(sc));
		stsrh->flags |= ISCSI_FLAG_CMD_UNDERFLOW;
		sc->result = host_byte(DID_OK) | status_byte(0x02);
	}
#endif
#if FAKE_PROBE_UNDERFLOW
	else if (test_bit(SESSION_ESTABLISHED, &session->control_bits) && sc
		 && (senselen == 0) && (sc->scsi_done == iscsi_done)
		 && (sc->retries <= 3) && (stsrh->cmd_status == 0)
		 && (stsrh->response == 0)) {
		printk
		    ("iSCSI: session iscsi bus %d target id %d faking probe "
		     "underflow for itt %u, CmdSN %u, cdb 0x%x to "
		     "(%u %u %u %u)\n",
		     session->iscsi_bus, session->target_id, itt, task->cmdsn,
		     sc->cmnd[0], sc->device->host->host_no,
		     sc->device->channel, sc->device->id, sc->device->lun);

		stsrh->residual_count = htonl(iscsi_expected_data_length(sc));
		stsrh->flags |= ISCSI_FLAG_CMD_UNDERFLOW;
		sc->resid = iscsi_expected_data_length(sc);
		sc->result = host_byte(DID_OK) | status_byte(0x0);
	}
#endif

	/* record the (possibly fake) status in the trace */
	ISCSI_TRACE(ISCSI_TRACE_RxCmdStatus, sc, task, stsrh->cmd_status,
		    stsrh->response);

	if (senselen && ((sense_data[0] == 0x71) || (sense_data[0] == 0xF1)) &&
	    (sc->device->type == TYPE_DISK) && translate_deferred_sense) {
		printk ("iSCSI: session iscsi bus %d target id %d recv_cmd cdb "
		     "0x%x translating deferred sense to current sense for "
		     "itt %u\n",
		     session->iscsi_bus, session->target_id, sc->cmnd[0],
		     task->itt);
		sc->sense_buffer[0] &= 0xFE;
	}

	/* check for underflow and overflow */
	expected = iscsi_expected_data_length(sc);
	if ((stsrh->flags & ISCSI_FLAG_CMD_OVERFLOW)
	    || (stsrh->flags & ISCSI_FLAG_CMD_UNDERFLOW)
	    || ((test_bit(TASK_READ, &task->flags))
		&& (task->rxdata < expected))) {
		if (LOG_ENABLED(ISCSI_LOG_QUEUE) || LOG_ENABLED(ISCSI_LOG_FLOW)
		    || (senselen && (SENSE_KEY(sense_data) == UNIT_ATTENTION))) {
			/* for debugging, always log this for UNIT ATTENTION */
			/* FIXME: bidi flags as well someday */
			printk
			    ("iSCSI: session iscsi bus %d recv_cmd itt %u, to "
			     "(%u %u %u %u), cdb 0x%x, %c%c %s, received %u, "
			     "residual %u, expected %u\n",
			     session->iscsi_bus, task->itt,
			     sc->device->host->host_no, sc->device->channel,
			     sc->device->id, sc->device->lun, sc->cmnd[0],
			     (stsrh->flags & ISCSI_FLAG_CMD_OVERFLOW) ? 'O' :
			     ' ', (stsrh-> flags & ISCSI_FLAG_CMD_UNDERFLOW) ?
			     'U' : ' ',(stsrh-> flags & ISCSI_FLAG_CMD_OVERFLOW)
			      ? "overflow" : "underflow", task->rxdata,
			     ntohl(stsrh->residual_count), expected);
		}
#ifdef DEBUG
		/* FIXME: fake a bad driver or SCSI status if there is a
		 * residual for certain commands?  The Linux high-level
		 * drivers appear to ignore the resid field.  This may cause
		 * data corruption if a device returns a residual for a read
		 * or write command, but a good SCSI status and iSCSI
		 * response.  The problem is that for some commands an
		 * underflow is normal, such as INQUIRY.  We have to check 
		 * the cdb to determine if an underflow should be 
		 * translated to an error.
		 * For now, just log about it, so we can see if the problem 
		 * is ever occuring.
		 */
		switch (sc->cmnd[0]) {
		case READ_6:
		case READ_10:
		case READ_12:
		case WRITE_6:
		case WRITE_10:
		case WRITE_12:
			if ((sc->device->type == TYPE_DISK)
			    && stsrh->residual_count && (stsrh->response == 0)
			    && (stsrh->cmd_status == 0)) {
				/* log if we get an underflow with a good status
				 * and response for data transfer commands, 
				 * since Linux appears to ignore the residual 
				 * field of Scsi_Cmnds, and only consider the 
				 * data invalid if the driver of SCSI status 
				 * was bad.  
				 */
				printk
				    ("iSCSI: session iscsi bus %d itt %u to "
				     "(%u %u %u %u), cdb 0x%x, received %u, "
				     "residual %u, expected %u, but normal "
				     "status\n",
				     session->iscsi_bus, task->itt,
				     sc->device->host->host_no, 
				     sc->device->channel, sc->device->id,
				     sc->device->lun, sc->cmnd[0], task->rxdata,
				     ntohl(stsrh->residual_count), expected);
			}
			break;
		}
#endif

		if (stsrh->flags & ISCSI_FLAG_CMD_UNDERFLOW) {
			ISCSI_TRACE(ISCSI_TRACE_RxUnderflow, sc, task,
				    ntohl(stsrh->residual_count), expected);
			sc->resid = ntohl(stsrh->residual_count);
		} else if (stsrh->flags & ISCSI_FLAG_CMD_OVERFLOW) {
			ISCSI_TRACE(ISCSI_TRACE_RxOverflow, sc, task,
				    ntohl(stsrh->residual_count), expected);
			sc->result =
			    host_byte(DID_ERROR) | status_byte(stsrh->
							       cmd_status);
			sc->resid = expected;
		} else if (task->rxdata < expected) {
			/* All the read data did not arrive. This can happen 
			 * without an underflow indication from the target if 
			 * the data is discarded by the driver, due to failed 
			 * sanity checks on the PDU or digest errors.
			 */
			ISCSI_TRACE(ISCSI_TRACE_HostUnderflow, sc, task,
				    task->rxdata, expected);
			sc->resid = expected - task->rxdata;
		}
	}

	if (stsrh->response) {
		needs_retry = 1;
		slow_retry = 1;

		/* log when we transition from no transport errors to 
		 * transport errors 
		 */
		if (__test_and_set_bit(sc->device->lun, 
		    session->luns_unreachable) == 0) {
			printk
			    ("iSCSI: session iscsi_bus %d recv_cmd status 0x%x,"
			     " iSCSI transport response 0x%x, itt %u, to "
			     "(%u %u %u %u) at %lu\n",
			     session->iscsi_bus, stsrh->cmd_status,
			     stsrh->response, task->itt, 
			     sc->device->host->host_no, sc->device->channel,
			     sc->device->id, sc->device->lun, jiffies);
		}
	} else {
		/* log when we transition from transport errors to 
		 * no transport errors 
		 */
		if (__test_and_clear_bit(sc->device->lun,
		    session->luns_unreachable)) {
			printk
			    ("iSCSI: session iscsi bus %d recv_cmd status 0x%x,"
			     "iSCSI transport response 0x%x, itt %u, to "
			     "(%u %u %u %u) at %lu\n",
			     session->iscsi_bus, stsrh->cmd_status,
			     stsrh->response, task->itt,
			     sc->device->host->host_no, sc->device->channel,
			     sc->device->id, sc->device->lun, jiffies);
		}

		/* now we basically duplicate what scsi_decide_disposition and 
		 * scsi_check_sense would have done if we completed the 
		 * command, but we do it ourselves so that we can requeue 
		 * internally.
		 */
		if ((stsrh->cmd_status == SAM_STAT_BUSY)
		    || (stsrh->cmd_status == SAM_STAT_TASK_SET_FULL)) {
			/* slow retries, at least until a command completes */
			needs_retry = 1;
			slow_retry = 1;
		} else if (stsrh->cmd_status == SAM_STAT_CHECK_CONDITION) {
			/* check conditions can only be retried if the command 
			 * allows retries. Tapes for example, can't retry, 
			 * since the tape head may have moved.
			 */
			/* FIXME: possible interactions with ACA. Do we need 
			 * to complete the command back to the SCSI layer when 
			 * ACA is enabled?
			 */
			if (sc->allowed > 1) {
				if (senselen == 0) {
					/* for check conditions with no sense, 
					 * fast retry when possible 
					 */
					needs_retry = 1;
				} else if ((sc->sense_buffer[0] & 0x70) == 0) {
					/* check conditions with invalid sense 
					 */
					needs_retry = 1;
				} else if (sc->sense_buffer[2] & 0xe0) {
					/* can't retry internally */
					/* FIXME: why not? what are these bits? 
					 */
				} else if ((SENSE_KEY(sc->sense_buffer) ==
					    ABORTED_COMMAND)) {
					needs_retry = 1;
				}
				else if ((SENSE_KEY(sc->sense_buffer) ==
					  MEDIUM_ERROR)) {
					needs_retry = 1;
				} else
				    if ((SENSE_KEY(sc->sense_buffer) ==
					 NOT_READY)) {
					if ((ASC(sc->sense_buffer) == 0x04)
					    && (ASCQ(sc->sense_buffer) == 0x01)) {
						/* LUN in the process of 
						 * becoming ready 
						 */
						needs_retry = 1;
						slow_retry = 1;
					}
				}

				/* switch to slow retries if the fast 
				 * retries don't seem to be working 
				 */
				if (needs_retry && (sc->SCp.sent_command > 10))
					slow_retry = 1;
			}
		}
	}

	if (needs_retry && internally_retryable(sc)) {
		/* need to requeue this command for a retry later.
		 * Philsophically we ought to complete the command and let the
		 * midlayer or high-level driver deal with retries.  Since the
		 * way the midlayer does retries is undesirable, we instead
		 * keep the command in the driver, but requeue it for the same
		 * cases the midlayer checks for retries.  This lets us ignore
		 * the command's retry count, and do retries until the command
		 * timer expires.  
		 */
		sc->result = 0;
		sc->resid = 0;
		sc->SCp.sent_command++;	/* count how many internal retries 
					 * we've done 
					 */
		memset(sc->sense_buffer, 0, sizeof (sc->sense_buffer));
		__set_bit(TASK_NEEDS_RETRY, &task->flags);

		if (slow_retry) {
			/* delay commands for slower retries */
			if (__test_and_set_bit
			    (task->lun, session->luns_delaying_commands) == 0) {
				DEBUG_RETRY
				    ("iSCSI: session iscsi bus %d starting to "
				     "delay commands to (%u %u %u %u) at %lu\n",
				     session->iscsi_bus, session->host_no,
				     session->channel, session->target_id,
				     sc->device->lun, jiffies);
				if (session->num_luns_delaying_commands == 0) {
					session->retry_timer.data =
					    (unsigned long) session;
					session->retry_timer.expires =
					    jiffies + HZ;
					session->retry_timer.function =
					    request_command_retries;
					add_timer(&session->retry_timer);
					DEBUG_RETRY
					    ("iSCSI: session iscsi bus %d "
					     "starting retry timer at %lu\n",
					     session->iscsi_bus, jiffies);
				}
				session->num_luns_delaying_commands++;
			}
		}
#if RETRIES_BLOCK_DEVICES
		/* try to stop the mid-layer from queueing any more commands 
		 * to this LUN until a command completes, by setting 
		 * sc->device->device_blocked.
		 */
		/* FIXME: locking? */
# ifdef SCSI_DEFAULT_DEVICE_BLOCKED
		sc->device->device_blocked = sc->device->max_device_blocked;
# else
		sc->device->device_blocked = TRUE;
# endif
			smp_mb();
#endif

		/* FIXME: warn if the command's tag is ORDERED or 
		 * HEAD_OF_QUEUE, since we're reordering commands by requeuing 
		 * to the tail of the scsi_cmnd queue, rather than retrying
		 * this task and all younger tasks to this LUN.  We emulate 
		 * what the SCSI midlayer would do, even though what it does 
		 * is probably broken if the command is ORDERED or 
		 * HEAD_OF_QUEUE. We probably need something like ACA to make 
		 * this work right, and it doesn't look like the midlayer uses 
		 * ACA, but rather it just assumes everything is untagged or 
		 * simple, so command reordering doesn't matter.  If the 
		 * midlayer ever changes, we'll need to make similar changes, 
		 * or go back to actually completing the command back to the 
		 * midlayer and letting it figure out how to retry.
		 */
		if (sc->tag == ORDERED_QUEUE_TAG)
			printk
			    ("iSCSI: session iscsi bus %d target id %d "
			     "retrying ORDERED command 0x%x, possible "
			     "reordering hazard at %lu\n",
			     session->iscsi_bus, session->target_id,
			     sc->cmnd[0], jiffies);
		else if (sc->tag == HEAD_OF_QUEUE_TAG)
			printk
			    ("iSCSI: session iscsi bus %d target id %d retrying"
			     " HEAD_OF_QUEUE command 0x%x, possible reordering "
			     "hazard at %lu\n",
			     session->iscsi_bus, session->target_id,
			     sc->cmnd[0], jiffies);

		smp_mb();
	} else {
		/* if we're not retrying this command, we go back to full
		 * speed unless command timeouts have triggered or will
		 * trigger error recovery.  
		 */
		if (test_bit(task->lun, session->luns_delaying_commands)) {
			__clear_bit(task->lun, session->luns_delaying_commands);
			session->num_luns_delaying_commands--;
			DEBUG_RETRY
			    ("iSCSI: session iscsi bus %d no longer delaying "
			     "commands to (%u %u %u %u) at %lu\n",
			     session->iscsi_bus, session->host_no,
			     session->channel, session->target_id, 
			     sc->device->lun, jiffies);
			if (session->num_luns_delaying_commands == 0) {
				del_timer_sync(&session->retry_timer);
				clear_bit(SESSION_RETRY_COMMANDS,
					  &session->control_bits);
				DEBUG_RETRY
				    ("iSCSI: session iscsi bus %d target id %d "
				     "stopping retry timer at %lu\n",
				     session->iscsi_bus, session->target_id,
				     jiffies);
			}
			if (!test_bit(task->lun, session->luns_timing_out)) {
				unsigned long noqueue_flags_;

				spin_lock_irqsave(&session->scsi_cmnd_lock,
						  noqueue_flags_);
				requeue_deferred_commands(session, task->lun);
				spin_unlock_irqrestore(&session->scsi_cmnd_lock,
						       noqueue_flags_);
			}
			smp_mb();
		}
	}

	ISCSI_TRACE(ISCSI_TRACE_RxCmd, sc, task, task->rxdata, expected);
}

/* 
 * complete a task in the session's completing queue, and return a pointer to 
 * it, or NULL if the task could not be completed.  Caller must hold the 
 * task_lock, but the lock is always released before returning.
 */
void
complete_task(struct iscsi_session *session, uint32_t itt)
{
	struct iscsi_task *task;
	unsigned long last_log = 0;
	int refcount;
	unsigned long midlayer_flags_;

	while (!signal_pending(current)) {
		DEBUG_QUEUE
		    ("iSCSI: session iscsi bus %d target id %d attempting to "
		     "complete itt %u\n",
		     session->iscsi_bus, session->target_id, itt);

		if ((task = find_session_task(session, itt))) {
			Scsi_Cmnd *sc = task->scsi_cmnd;

			if (test_bit(SESSION_RESETTING, &session->control_bits)) {
				/* we don't trust the target to give us 
				 * correct responses once we've issued a reset.
				 * Ensure that none of the outstanding tasks 
				 * complete.
				 */
				spin_unlock(&session->task_lock);
				DEBUG_EH
				    ("iSCSI: session iscsi bus %d target id %d "				     "can't complete itt %u, cmnd 0x%x, "
				     "reset in progress at %lu\n",
				     session->iscsi_bus, session->target_id,
				     itt, sc->cmnd[0], jiffies);
				return;
			} else
			    if (test_bit
				(task->lun, session->luns_doing_recovery)) {
				/* don't complete any tasks once a LUN has 
				 * started doing error recovery.
				 * Leave the recovery state as it is, since 
				 * we may have an outstanding
				 * task mgmt PDU for this task.
				 */
				spin_unlock(&session->task_lock);
				DEBUG_EH
				    ("iSCSI: session iscsi bus %d target id %d "
				     "can't complete itt %u, cmnd 0x%x, LUN %u "
				     "doing error recovery at %lu\n",
				     session->iscsi_bus, session->target_id,
				     itt, sc->cmnd[0], task->lun, jiffies);
				return;
			}

			/* no need to do error recovery for this task */
			task->flags &= ~TASK_RECOVERY_MASK;

			/* it's possible the tx thread is using the task right 
			 * now. The task's refcount can't increase while it's 
			 * in the completing collection, so wait for the 
			 * refcount to hit zero, or the task to leave the 
			 * completing collection, whichever happens first.
			 */
			if ((refcount = atomic_read(&task->refcount)) == 0) {
				/* this is the expected case */
#if INCLUDE_DEBUG_EH
				if (LOG_ENABLED(ISCSI_LOG_EH) && sc
				    && (sc->cmnd[0] == TEST_UNIT_READY)) {
					printk
					    ("iSCSI: completing TUR at %lu, "
					     "itt %u, (%u %u %u %u), cdb 0x%x,"
					     " result 0x%x\n",
					     jiffies, itt,
					     sc->device->host->host_no,
					     sc->device->channel,
					     sc->device->id, sc->device->lun,
					     sc->cmnd[0], sc->result);
				} else
#endif
				{
#if INCLUDE_DEBUG_QUEUE
					if (LOG_ENABLED(ISCSI_LOG_QUEUE)) {
						if (sc)
							printk
							    ("iSCSI: completing"
							     " itt %u, (%u %u "
							     "%u %u), cdb 0x%x,"							     " result 0x%x\n",
							     itt,
							     sc->device->host->
								host_no,
							     sc->device->
								channel,
							     sc->device->id,
							     sc->device->lun,
							     sc->cmnd[0],
							     sc->result);
						else
							printk
							    ("iSCSI: completing"
							     " itt %u, command "
							     "NULL, (%u %u %u "
							     "%u)\n",
							     itt,
							     session->host_no,
							     session->channel,
							     session->target_id,
							     task->lun);
					}
#endif
				}

				/* remove the task from the session, to ensure a
				 * session drop won't try to complete the task 
				 * again.
				 */
				if (remove_session_task(session, task)) {
					DEBUG_QUEUE
					    ("iSCSI: removed itt %u, from "
					     "session iscsi bus %d target id %d"
					     " to %s\n",
					     task->itt, session->iscsi_bus,
					     session->target_id,
					     session->log_name);
				}

				if (test_bit
				    (task->lun, session->luns_timing_out)) {
					/* this task may be the last thing 
					 * delaying error recovery.
					 * Make sure the tx thread scans tasks 
					 * again.
					 */
					DEBUG_EH
					    ("iSCSI: session iscsi bus %d "
					     "target id %d completing itt %u, "
					     "while LUN %u is timing out at "
					     "%lu\n",
					     session->iscsi_bus,
					     session->target_id, itt, task->lun,
					     jiffies);
					set_bit(SESSION_TASK_TIMEDOUT,
						&session->control_bits);
					smp_mb();
				}

				/* this task no longer has a Scsi_Cmnd 
				 * associated with it 
				 */
				task->scsi_cmnd = NULL;
				if (sc)
					sc->host_scribble = NULL;

				if (sc == NULL) {
					/* already completed, nothing to do */
					printk
					    ("iSCSI: session iscsi bus %d "
					     "already completed itt %u, "
					     "(%u %u %u %u)\n",
					     session->iscsi_bus, itt,
					     session->host_no, session->channel,
					     session->target_id, task->lun);

					free_task(session, task);
				} else
				    if (test_bit
					(TASK_NEEDS_RETRY, &task->flags)) {
					unsigned long noqueue_flags_;

					/* done with this task */
					free_task(session, task);

					/* just requeue the task back to the 
					 * scsi_cmnd queue so that it gets 
					 * retried 
					 */
					DEBUG_RETRY
					    ("iSCSI: session iscsi bus %d "
					     "requeueing itt %u command cdb "
					     "0x%x for retry to (%u %u %u %u)"
					     "\n",
					     session->iscsi_bus, itt,
					     sc->cmnd[0], sc->device->host->
								host_no,
					     sc->device->channel, 
					     sc->device->id, sc->device->lun);

					spin_lock_irqsave(&session->
							  scsi_cmnd_lock,
							  noqueue_flags_);
					add_cmnd(sc, &session->scsi_cmnd_head,
						 &session->scsi_cmnd_tail);
					atomic_inc(&session->num_cmnds);

#if RETRIES_BLOCK_DEVICES
/* try to prevent the midlayer from issuing more commands to this device
 * until we complete a command for this device back to the midlayer.  
 * This hopefully keeps the midlayer queueing commands to other LUNs,
 * rather than filling up the driver's limit of 64 with commands that 
 * we can't complete, which would effectively block other LUNs that 
 * are still working from getting any commands.
 */
					/* FIXME: locking? */
# ifdef SCSI_DEFAULT_DEVICE_BLOCKED
					sc->device->device_blocked =
					    sc->device->max_device_blocked;
# else
					sc->device->device_blocked = TRUE;
# endif
					smp_mb();
#endif
					spin_unlock_irqrestore(&session->
							       scsi_cmnd_lock,
							       noqueue_flags_);
					wake_tx_thread(TX_SCSI_COMMAND,
						       session);
				} else {
					/* delete our command timer */
					del_command_timer(sc);

					/* we're completing it out of the 
					 * driver 
					 */
					ISCSI_TRACE(ISCSI_TRACE_CmdDone, sc,
						    task, sc->result, 0);

					/* done with this task */
					free_task(session, task);

					/* FIXME: if we want to get lots of 
					 * retries for cases we don't retry 
					 * internally, we'll need to 
					 * conditionally alter sc->retries 
					 * before completing the command.
					 */

					if (sc->scsi_done == NULL) {
						printk
						    ("iSCSI: no completion "
						     "callback for command "
						     "0x%x\n",
						     sc->cmnd[0]);
					} else if (sc->scsi_done == iscsi_done) {
						/* it came from iscsi-probe.c, 
						 * and doesn't need a timer 
						 * added or lock held 
						 */
						sc->scsi_done(sc);
					} else {
						/* add a useless timer for the 
						 * midlayer to delete 
						 */
						add_completion_timer(sc);

						/* tell the SCSI midlayer that 
						 * the command is done 
						 */
						spin_lock_irqsave(
							session->hba->host->
								host_lock,
							midlayer_flags_);
						sc->scsi_done(sc);
						spin_unlock_irqrestore(
							session->hba->host->
								host_lock, 
							midlayer_flags_);
					}

					DEBUG_QUEUE
					    ("iSCSI: session iscsi bus %d "
					     "completed itt %u, (%u %u %u %u), "					     "cdb 0x%x, result 0x%x\n",
					     session->iscsi_bus, itt,
					     sc->device->host->host_no,
					     sc->device->channel,
					     sc->device->id, sc->device->lun,
					     sc->cmnd[0], sc->result);
				}

				spin_unlock(&session->task_lock);

				return;
			} else {
				/* task is still in use, can't complete it yet. 
				 * Since this only happens when a command is 
				 * aborted by the target unexpectedly, this 
				 * error case can be slow. Just keep polling 
				 * for the refcount to hit zero.  If the tx 
				 * thread is blocked while using a task, the
				 * timer thread will eventually send a signal 
				 * to both the rx thread and tx thread, so 
				 * this loop will terminate one way or another. 
				 */
				if ((last_log == 0)
				    || time_before_eq(last_log + HZ, jiffies)) {
					DEBUG_QUEUE
					    ("iSCSI: waiting to complete itt "
					     "%u, cmnd 0x%x, refcount %d\n",
					     itt, sc->cmnd[0], refcount);
				}

				spin_unlock(&session->task_lock);

				set_current_state(TASK_INTERRUPTIBLE);
				schedule_timeout(MSECS_TO_JIFFIES(10));

				spin_lock(&session->task_lock);
			}
		} else {
			/* not a valid task */
			DEBUG_QUEUE
			    ("iSCSI: can't complete itt %u, task not found\n",
			     itt);
			spin_unlock(&session->task_lock);
			return;
		}
	}

	printk
	    ("iSCSI: session iscsi bus %d target id %d complete_task %u failed "
	     "at %lu\n",
	     session->iscsi_bus, session->target_id, itt, jiffies);
	spin_unlock(&session->task_lock);
}

static void
recheck_busy_tasks(unsigned long arg)
{
	struct iscsi_session *session = (struct iscsi_session *) arg;

	session->busy_task_timer.expires = 0;
	smp_mb();
	wake_tx_thread(SESSION_TASK_TIMEDOUT, session);
}

int
process_timedout_tasks(struct iscsi_session *session)
{
	struct iscsi_task *task, *t,*task_tmp;
	Scsi_Cmnd *requeue_head, *requeue_tail, *defer_head, *defer_tail;
	Scsi_Cmnd *cmnd, *prior;
	int luns_checked, luns_recovering, tasks_recovering = 0;
	int num_requeue_cmnds = 0, num_deferred_cmnds = 0;
	int l, busy = 0;
	unsigned long noqueue_flags_;

	spin_lock(&session->task_lock);
	spin_lock_irqsave(&session->scsi_cmnd_lock, noqueue_flags_);

	if (LOG_ENABLED(ISCSI_LOG_EH))
		printk
		    ("iSCSI: session iscsi bus %d target id %d "
		     "processing timedout tasks at %lu\n",
		     session->iscsi_bus, session->target_id, jiffies);

	do {
		if (signal_pending(current))
			break;	/* the session drop will take care 
				 * of everything 
				 */

		if (test_bit(SESSION_TERMINATING, &session->control_bits))
			break;	/* the session termination will 
				 * take care of everything 
				 */

		/* calculate the state of each LUN based on the tasks, so that
		 * we know how to deal with the tasks and commands
		 * later on.
		 */

		clear_bit(SESSION_TASK_TIMEDOUT, &session->control_bits);
		/* we could use per-LUN data structures instead 
		 * of bitmaps for these 
		 */
		memset(session->luns_checked, 0,
		       sizeof (session->luns_checked));
		memset(session->luns_needing_recovery, 0,
		       sizeof (session->luns_needing_recovery));
		memset(session->luns_delaying_recovery, 0,
		       sizeof (session->luns_delaying_recovery));
		luns_checked = 0;
		luns_recovering = 0;
		tasks_recovering = 0;

		if (test_bit(SESSION_RESETTING, &session->control_bits)) {
			DEBUG_EH
			    ("iSCSI: session iscsi bus %d target id %d "
			     "resetting, task timeout processing checking "
			     "all LUNs\n",
			     session->iscsi_bus, session->target_id);

			memset(session->luns_checked, 0xFF,
			       sizeof (session->luns_checked));
			luns_checked += ISCSI_MAX_LUN;
		} else {
			/* record which LUNS currently are timing out,
			 * so that we know which ones we've checked for 
			 * recovery.
			 */
			for (l = 0; l < ISCSI_MAX_LUN; l++) {
				if (test_bit(l, session->luns_timing_out)) {
					DEBUG_EH
					    ("iSCSI: session iscsi bus %d "
					     "target id %d task timeout "
					     "processing checking LUN %u\n",
					     session->iscsi_bus,
					     session->target_id, l);
					__set_bit(l, session->luns_checked);
					luns_checked++;
				}
			}
		}

		/* scan all outstanding tasks to determine which LUNs need 
		 * error recovery, and whether recovery must be delayed.
		 */
		list_for_each_entry(task, &session->task_arrival_order,
				    task_order_link){
			if (test_bit(task->lun, session->luns_checked)) {
				if (TASK_NEEDS_RECOVERY(task)) {
					/* we must do error recovery 
					 * for this LUN 
					 */
					tasks_recovering++;
					if (__test_and_set_bit
					    (task->lun,
					     session->luns_needing_recovery))
						luns_recovering++;

					DEBUG_EH
					    ("iSCSI: session iscsi bus %d "
					     "target id %d itt %u sc 0x%x "
					     "LUN %u needs error recovery\n",
					     session->iscsi_bus,
					     session->target_id, task->itt,
					     task->scsi_cmnd->cmnd[0],
					     task->lun);
				}

				if (!test_bit(0, &task->timedout)) {
					/* don't do error recovery for this 
					 * LUN while outstanding tasks
					 * have not yet completed or timed out.
					 */
					__set_bit(task->lun,
						  session->luns_delaying_recovery);
					DEBUG_EH
					    ("iSCSI: session iscsi bus %d "
					     "target id %d itt %u sc 0x%x has "
					     "not timed out, delaying "
					     "recovery for LUN %u\n",
					     session->iscsi_bus,
					     session->target_id, task->itt,
					     task->scsi_cmnd->cmnd[0],
					     task->lun);
				} else if (atomic_read(&task->refcount)) {
					/* the task refcount may be non-zero 
					 * if we're in the middle of sending 
					 * or receiving data for this task.  
					 * Make sure that we don't try to
					 * finish recovery and complete the 
					 * task when it's in use.
					 */
					/* FIXME: we only want to delay 
					 * finishing recovery for this LUN.
					 * we don't have to delay sending 
					 * task mgmt PDUs for this task,
					 * though we currently do.
					 */
					__set_bit(task->lun,
						  session->
						  luns_needing_recovery);
					__set_bit(task->lun,
						  session->
						  luns_delaying_recovery);
					DEBUG_EH
					    ("iSCSI: session iscsi bus %d "
					     "target id %d itt %u sc 0x%x has "
					     "timed out but is busy, delaying "
					     "recovery for LUN %u\n",
					     session->iscsi_bus,
					     session->target_id, task->itt,
					     task->scsi_cmnd->cmnd[0],
					     task->lun);
					busy = 1;
				} else {
					DEBUG_EH
					    ("iSCSI: session iscsi bus %d "
					     "target id %d itt %u sc 0x%x has "
					     "timed out\n",
					     session->iscsi_bus,
					     session->target_id, task->itt,
					     task->scsi_cmnd->cmnd[0]);
				}

				/* Note: draft 16 - 9.5.1 says we MUST keep
				 * responding to valid target transfer tags, 
				 * though we can terminate them early with 
				 * the F-bit, and that the target must wait 
				 * for all outstanding target transfer tags to 
				 * complete before doing an abort task set.  
				 * For simplicity's sake, we currently
				 * always continue responding to ttts, and send
				 * the actual data if we still have the command,
				 * or empty data PDUs if the command has already
				 * been completed out of the driver.
				 */
			}
		}

		smp_mb();

	} while (test_bit(SESSION_TASK_TIMEDOUT, &session->control_bits));

	if (busy) {
		/* either xmit_data invoked us with a task refcount held high,
		 * or the rx thread is in the middle of receiving data for
		 * a task.
		 */
		if ((session->busy_task_timer.expires == 0)
		    && !test_bit(SESSION_TERMINATING, &session->control_bits)) {
			session->busy_task_timer.expires =
			    jiffies + MSECS_TO_JIFFIES(40);
			session->busy_task_timer.data = (unsigned long) session;
			session->busy_task_timer.function = recheck_busy_tasks;
			DEBUG_EH
			    ("iSCSI: session iscsi bus %d target id %d "
			     "scheduling busy task scan for %lu at %lu\n",
			     session->iscsi_bus, session->target_id,
			     session->busy_task_timer.expires, jiffies);
			del_timer_sync(&session->busy_task_timer);	
					/* make sure it's not running now */
			add_timer(&session->busy_task_timer);
		}
	}

	if (test_bit(SESSION_RESETTING, &session->control_bits)) {
		if (!test_bit(SESSION_RESET, &session->control_bits)) {
			/* don't complete anything if a reset is in 
			 * progress but has not yet occured 
			 */
			DEBUG_EH
			    ("iSCSI: session iscsi bus %d target id %d reset "
			     "in progress at %lu, deferring recovery for "
			     "all LUNs\n",
			     session->iscsi_bus, session->target_id, jiffies);
			spin_unlock_irqrestore(&session->scsi_cmnd_lock,
					       noqueue_flags_);
			/* we may need to escalate a timedout reset though */
			goto error_recovery;
		} else if (busy) {
			/* reset has finished, but a task is busy, 
			 * complete everything later 
			 */
			DEBUG_EH
			    ("iSCSI: session iscsi bus %d target id %d reset "
			     "complete but tasks busy at %lu, deferring "
			     "recovery for all LUNs\n",
			     session->iscsi_bus, session->target_id, jiffies);
			spin_unlock_irqrestore(&session->scsi_cmnd_lock,
					       noqueue_flags_);
			/* we may need to escalate a timedout reset though */
			goto error_recovery;
		} else {
			/* go ahead and recovery everything */
			DEBUG_EH
			    ("iSCSI: session iscsi bus %d target id %d reset "
			     "complete at %lu, recovering tasks for all LUNs\n",
			     session->iscsi_bus, session->target_id, jiffies);
		}
	}

	/* if we've cleared a LUN's problems, we need to requeue tasks 
	 * and commands to that LUN 
	 */

	/* process the tasks */
	DEBUG_EH
	    ("iSCSI: session iscsi bus %d target id %d checking %d tasks " 
	     "for recovery at %lu\n",
	     session->iscsi_bus, session->target_id,
	     atomic_read(&session->num_active_tasks), jiffies);
	list_for_each_entry(task, &session->task_arrival_order,
			    task_order_link) {
		if (test_bit(task->lun, session->luns_checked)
		    && !test_bit(task->lun, session->luns_needing_recovery)) {
			task_tmp = list_entry(task->task_order_link.prev,
					      struct iscsi_task,task_order_link);
			/* we're done with this task */
			if (remove_task(&session->tx_task_head, task->itt)) {
				DEBUG_EH
				    ("iSCSI: session iscsi bus %d target id %d "
				     "task data transmit cancelled to LUN %u\n",
				     session->iscsi_bus, session->target_id,
				     task->lun);
				task->ttt = RSVD_TASK_TAG;
			}
			remove_session_task(session, task);
			del_task_timer(task);

			/* and this task's command */
			if ((cmnd = task->scsi_cmnd)) {
				/* clear any Scsi_Cmnd fields that may have 
				 * been modified 
				 */
				memset(cmnd->sense_buffer, 0,
				       sizeof (cmnd->sense_buffer));
				cmnd->result = 0;
				cmnd->resid = 0;
				cmnd->host_scribble = NULL;
				/* prepare to requeue it 
				 * someplace appropriate 
				 */
				if (test_bit
				    (task->lun,
				     session->luns_delaying_commands)) {
					/* we need to defer this task's command 
					 * and any commands for this LUN in 
					 * the retry queue (since the tasks 
					 * should be retried first).
					 */
					if (LOG_ENABLED(ISCSI_LOG_EH))
						printk
						    ("iSCSI: session iscsi bus "
						     "%d deferring itt %u cmnd "
						     "cdb 0x%02x to (%u %u %u "
						     "%u) at %lu\n",
						     session->iscsi_bus,
						     task->itt, cmnd->cmnd[0],
						     session->host_no,
						     session->channel,
						     session->target_id,
						     cmnd->device->lun, jiffies);
					add_cmnd(cmnd, &defer_head,
						 &defer_tail);
					num_deferred_cmnds++;
				} else {
					/* requeue all tasks and retry commands 
					 * back to the scsi_cmd queue.
					 * There may be a command retry queued 
					 * even when the LUN isn't failing
					 * delivery, in cases where a command 
					 * completion arrived during error
					 * recovery and cleared the 
					 * failing_delivery bit.
					 */
					if (LOG_ENABLED(ISCSI_LOG_EH))
						printk
						    ("iSCSI: session iscsi bus "
						     "%d requeueing itt %u cmnd"
						     " cdb 0x%02x to (%u %u %u "
						     "%u) at %lu\n",
						     session->iscsi_bus,
						     task->itt, cmnd->cmnd[0],
						     session->host_no,
						     session->channel,
						     session->target_id,
						     cmnd->device->lun, jiffies);
					add_cmnd(cmnd, &requeue_head,
						 &requeue_tail);
					num_requeue_cmnds++;
				}
			}

			/* the task has a refcount of zero and has already been
			 * removed from the session, so we can safely free it
			 * now.
			 */
			free_task(session, task);
			/* proceed with rest of tasks */
			task = task_tmp;
		}

	}

	/* anything in the retry queue needs to get requeued along with the 
	 * tasks, to avoid reordering commands.
	 */
	DEBUG_EH
	    ("iSCSI: session iscsi bus %d target id %d checking %d retry "
	     "queue commands following task timeouts at %lu\n",
	     session->iscsi_bus, session->target_id,
	     atomic_read(&session->num_retry_cmnds), jiffies);
	prior = NULL;
	while ((cmnd = session->retry_cmnd_head)) {
		if (test_bit(cmnd->device->lun, session->luns_checked)
		    && !test_bit(cmnd->device->lun,
				 session->luns_needing_recovery)) {
			/* remove it from the retry_cmnd queue */
			session->retry_cmnd_head =
			    (Scsi_Cmnd *) cmnd->host_scribble;
			if (session->retry_cmnd_head == NULL)
				session->retry_cmnd_tail = NULL;
			atomic_dec(&session->num_retry_cmnds);
			cmnd->host_scribble = NULL;

			if (test_bit
			    (task->lun, session->luns_delaying_commands)) {
				if (LOG_ENABLED(ISCSI_LOG_EH))
					printk
					    ("iSCSI: session iscsi bus %d "
					     "deferring retryable cmnd cdb "
					     "0x%02x to (%u %u %u %u) at %lu\n",
					     session->iscsi_bus, cmnd->cmnd[0],
					     session->host_no, session->channel,
					     session->target_id,
					     cmnd->device->lun, jiffies);
				add_cmnd(cmnd, &defer_head, &defer_tail);
				num_deferred_cmnds++;
			} else {
				if (LOG_ENABLED(ISCSI_LOG_EH))
					printk
					    ("iSCSI: session iscsi bus %d "
					     "requeueing retryable cmnd cdb "
					     "0x%02x to (%u %u %u %u) at %lu\n",
					     session->iscsi_bus, cmnd->cmnd[0],
					     session->host_no, session->channel,
					     session->target_id, 
					     cmnd->device->lun, jiffies);
				add_cmnd(cmnd, &requeue_head, &requeue_tail);
				num_requeue_cmnds++;
			}
		} else {
			prior = cmnd;
			break;
		}
	}
	while (prior && (cmnd = (Scsi_Cmnd *) prior->host_scribble)) {
		if (test_bit(cmnd->device->lun, session->luns_checked)
		    && !test_bit(cmnd->device->lun, 
				 session->luns_needing_recovery)) {
			/* remove it from the retry_cmnd queue */
			prior->host_scribble = cmnd->host_scribble;
			if (session->retry_cmnd_tail == cmnd)
				session->retry_cmnd_tail = prior;
			atomic_dec(&session->num_retry_cmnds);
			cmnd->host_scribble = NULL;

			if (test_bit
			    (task->lun, session->luns_delaying_commands)) {
				if (LOG_ENABLED(ISCSI_LOG_EH))
					printk
					    ("iSCSI: session iscsi bus %d "
					     "deferring retryable cmnd cdb "
					     "0x%02x to (%u %u %u %u) at %lu\n",
					     session->iscsi_bus, cmnd->cmnd[0],
					     session->host_no, session->channel,
					     session->target_id,
					     cmnd->device->lun, jiffies);
				add_cmnd(cmnd, &defer_head, &defer_tail);
				num_deferred_cmnds++;
			} else {
				if (LOG_ENABLED(ISCSI_LOG_EH))
					printk
					    ("iSCSI: session iscsi bus %d "
					     "requeueing retryable cmnd cdb "
					     "0x%02x to (%u %u %u %u) at %lu\n",
					     session->iscsi_bus, cmnd->cmnd[0],
					     session->host_no, session->channel,
					     session->target_id, 
					     cmnd->device->lun, jiffies);
				add_cmnd(cmnd, &requeue_head, &requeue_tail);
				num_requeue_cmnds++;
			}
		} else {
			prior = cmnd;
		}
	}

	/* scan the deferred queue, moving commands to the requeue list unless
	 * the LUN is currently delaying commands.
	 */
	DEBUG_EH
	    ("iSCSI: session iscsi bus %d target id %d checking %d deferred "
	     "queue commands following task timeouts at %lu\n",
	     session->iscsi_bus, session->target_id,
	     session->num_deferred_cmnds, jiffies);
	prior = NULL;
	while ((cmnd = session->deferred_cmnd_head)) {
		if (test_bit(cmnd->device->lun, session->luns_checked) &&
		    !test_bit(cmnd->device->lun, session->luns_needing_recovery) &&
		    !test_bit(cmnd->device->lun, 
			      session->luns_delaying_commands)) {
			/* remove it from the deferred_cmnd queue */
			session->deferred_cmnd_head =
			    (Scsi_Cmnd *) cmnd->host_scribble;
			if (session->deferred_cmnd_head == NULL)
				session->deferred_cmnd_tail = NULL;
			session->num_deferred_cmnds--;
			cmnd->host_scribble = NULL;

			/* and requeue it to be sent */
			if (LOG_ENABLED(ISCSI_LOG_EH))
				printk
				    ("iSCSI: session iscsi bus %d requeueing "
				     "deferred cmnd cdb 0x%02x to (%u %u %u %u)"
				     " at %lu\n",
				     session->iscsi_bus, cmnd->cmnd[0],
				     session->host_no, session->channel,
				     session->target_id, cmnd->device->lun,
				     jiffies);
			add_cmnd(cmnd, &requeue_head, &requeue_tail);
			num_requeue_cmnds++;
		} else {
			prior = cmnd;
			break;
		}
	}
	while (prior && (cmnd = (Scsi_Cmnd *) prior->host_scribble)) {
		if (test_bit(cmnd->device->lun, session->luns_checked) &&
		    !test_bit(cmnd->device->lun, session->luns_needing_recovery)
		    && !test_bit(cmnd->device->lun,
				 session->luns_delaying_commands)) {
			/* remove it from the deferred_cmnd queue */
			prior->host_scribble = cmnd->host_scribble;
			if (session->deferred_cmnd_tail == cmnd)
				session->deferred_cmnd_tail = prior;
			session->num_deferred_cmnds--;
			cmnd->host_scribble = NULL;

			if (LOG_ENABLED(ISCSI_LOG_EH))
				printk
				    ("iSCSI: session iscsi bus %d requeueing "
				     "deferred cmnd cdb 0x%02x to "
				     "(%u %u %u %u) at %lu\n",
				     session->iscsi_bus, cmnd->cmnd[0],
				     session->host_no, session->channel,
				     session->target_id, cmnd->device->lun,
				     jiffies);
			add_cmnd(cmnd, &requeue_head, &requeue_tail);
			num_requeue_cmnds++;
		} else {
			prior = cmnd;
		}
	}

	if (requeue_head) {
		/* requeue to the head of the scsi_cmnd queue */
		DEBUG_EH
		    ("iSCSI: session iscsi bus %d target id %d requeueing "
		     "%d commands at %lu\n",
		     session->iscsi_bus, session->target_id, num_requeue_cmnds,
		     jiffies);
		requeue_tail->host_scribble = (void *) session->scsi_cmnd_head;
		session->scsi_cmnd_head = requeue_head;
		if (session->scsi_cmnd_tail == NULL)
			session->scsi_cmnd_tail = requeue_tail;
		atomic_add(num_requeue_cmnds, &session->num_cmnds);
	}

	if (defer_head) {
		/* requeue to the head of the deferred_cmnd queue */
		DEBUG_EH
		    ("iSCSI: session iscsi bus %d target id %d deferring "
		     "%d commands at %lu\n",
		     session->iscsi_bus, session->target_id, num_deferred_cmnds,
		     jiffies);
		defer_tail->host_scribble =
		    (void *) session->deferred_cmnd_head;
		session->deferred_cmnd_head = defer_head;
		if (session->deferred_cmnd_tail == NULL)
			session->deferred_cmnd_tail = defer_tail;
		session->num_deferred_cmnds += num_deferred_cmnds;
	}

	/* we no longer need the scsi_cmnd lock */
	spin_unlock_irqrestore(&session->scsi_cmnd_lock, noqueue_flags_);

	/* clear bits and let I/O to these LUNs restart */
	if (test_and_clear_bit(SESSION_RESET, &session->control_bits)) {
		printk
		    ("iSCSI: session iscsi bus %d (%u %u %u *) "
		     "finished reset at %lu\n",
		     session->iscsi_bus, session->host_no, session->channel,
		     session->target_id, jiffies);
		for (l = 0; l < ISCSI_MAX_LUN; l++) {
			clear_bit(l, session->luns_doing_recovery);	
						/* allow completion again */
			clear_bit(l, session->luns_timing_out);	
						/* allow new tasks again */
		}
		clear_bit(SESSION_RESETTING, &session->control_bits);
		set_bit(TX_SCSI_COMMAND, &session->control_bits);
		set_bit(TX_WAKE, &session->control_bits);
	} else {
		for (l = 0; l < ISCSI_MAX_LUN; l++) {
			if (test_bit(l, session->luns_checked)
			    && !test_bit(l, session->luns_needing_recovery)) {
				printk
				    ("iSCSI: session iscsi bus %d (%u %u %u %u)"
				     " finished error recovery at %lu\n",
				     session->iscsi_bus, session->host_no,
				     session->channel, session->target_id, l,
				     jiffies);
				clear_bit(l, session->luns_doing_recovery);	
						/* allow completion again */
				clear_bit(l, session->luns_timing_out);	
						/* allow new tasks again */
				set_bit(TX_SCSI_COMMAND,
					&session->control_bits);
				set_bit(TX_WAKE, &session->control_bits);
			}
		}
	}
	smp_mb();

      error_recovery:
	if (signal_pending(current)) {
		DEBUG_EH
		    ("iSCSI: session iscsi bus %d target id %d signalled "
		     "during timeout processing, skipping error recovery\n",
		     session->iscsi_bus, session->target_id);
		spin_unlock(&session->task_lock);
		return 0;	/* the session drop will take 
				 * care of everything 
				 */
	}

	if (test_bit(SESSION_TERMINATING, &session->control_bits)) {
		DEBUG_EH
		    ("iSCSI: session iscsi bus %d target id %d "
		     "terminating, skipping error recovery\n",
		     session->iscsi_bus, session->target_id);
		spin_unlock(&session->task_lock);
		return 0;	/* the session termination will take 
				 * care of everything 
				 */
	}

	if (test_and_clear_bit
	    (SESSION_TASK_MGMT_TIMEDOUT, &session->control_bits)
	    && ((task = find_session_mgmt_task(session, session->mgmt_itt)))) {
		/* a timeout has occured, escalate the task's recovery 
		 * method, and quit waiting for a response 
		 */
		if (__test_and_clear_bit(TASK_TRY_ABORT, &task->flags))
			__set_bit(TASK_TRY_ABORT_TASK_SET, &task->flags);
		else if (__test_and_clear_bit
			 (TASK_TRY_ABORT_TASK_SET, &task->flags))
			__set_bit(TASK_TRY_LUN_RESET, &task->flags);
		else if (__test_and_clear_bit(TASK_TRY_LUN_RESET, &task->flags))
			__set_bit(TASK_TRY_WARM_RESET, &task->flags);
		else if (__test_and_clear_bit
			 (TASK_TRY_WARM_RESET, &task->flags))
			__set_bit(TASK_TRY_COLD_RESET, &task->flags);
		else {
			printk
			    ("iSCSI: session iscsi bus %d target id %d "
			     "cold reset timed out, dropping session at %lu\n",
			     session->iscsi_bus, session->target_id, jiffies);
			spin_unlock(&session->task_lock);
			iscsi_drop_session(session);
			return 0;
		}

		session->mgmt_itt = task->mgmt_itt = RSVD_TASK_TAG;
	}

	/* if tasks need recovery and we don't have an 
	 * oustanding task mgmt PDU, send one 
	 */
	if (tasks_recovering && (session->mgmt_itt == RSVD_TASK_TAG)) {

		DEBUG_EH
		    ("iSCSI: session iscsi bus %d target id %d doing "
		     "error recovery at %lu, %d tasks need recovery\n",
		     session->iscsi_bus, session->target_id, jiffies,
		     tasks_recovering);

		/* send a PDU for the oldest TIMEDOUT task needing recovery 
		 * for a LUN that needs error recovery and isn't delaying it.
		 */
		list_for_each_entry(task, &session->task_arrival_order,
				    task_order_link) {
			DEBUG_EH
			    ("iSCSI: session iscsi bus %d target id %d error "
			     "recovery checking itt %u LUN %u flags 0x%04lx\n",
			     session->iscsi_bus, session->target_id, task->itt,
			     task->lun, task->flags);

			if (TASK_NEEDS_RECOVERY(task) &&
			    test_bit(task->lun, session->luns_needing_recovery)
			    && !test_bit(task->lun,
					 session->luns_delaying_recovery)) {
				break;
			}
		}

		if (&(task->task_order_link) != &(session->task_arrival_order))
		{
			/* prevent any command completions once we start error
			 * recovery for a LUN.  We want to hang on to all of the
			 * tasks, so that we can complete them in order once 
			 * error recovery finishes.
			 */
			set_bit(task->lun, session->luns_doing_recovery);

			if (test_bit(TASK_TRY_ABORT, &task->flags)) {
				session->mgmt_itt = task->mgmt_itt =
				    allocate_itt(session);
				if (session->abort_timeout) {
					session->task_mgmt_response_deadline =
					    jiffies +
					    (session->abort_timeout * HZ);
					if (session->
					    task_mgmt_response_deadline == 0)
						session->
						    task_mgmt_response_deadline
						    = 1;
				}
				atomic_inc(&task->refcount);

				if (task->scsi_cmnd)
					printk
					    ("iSCSI: session iscsi bus %d "
					     "sending mgmt %u abort for itt %u "
					     "cdb 0x%02x to (%u %u %u %u) "
					     "at %lu\n",
					     session->iscsi_bus, task->mgmt_itt,
					     task->itt,
					     task->scsi_cmnd->cmnd[0],
					     session->host_no, session->channel,
					     session->target_id, task->lun,
					     jiffies);
				else
					printk
					    ("iSCSI: session iscsi bus %d "
					     "sending mgmt %u abort for itt %u "
					     "to (%u %u %u %u) at %lu\n",
					     session->iscsi_bus, task->mgmt_itt,
					     task->itt, session->host_no,
					     session->channel,
					     session->target_id, task->lun,
					     jiffies);

				spin_unlock(&session->task_lock);

				iscsi_xmit_task_mgmt(session,
						     ISCSI_TM_FUNC_ABORT_TASK,
						     task, task->mgmt_itt);
				atomic_dec(&task->refcount);
			} else
			    if (test_bit(TASK_TRY_ABORT_TASK_SET, &task->flags))
			{
				session->mgmt_itt = task->mgmt_itt =
				    allocate_itt(session);
				if (session->abort_timeout) {
					session->task_mgmt_response_deadline =
					    jiffies +
					    (session->abort_timeout * HZ);
					if (session->
					    task_mgmt_response_deadline == 0)
						session->
						    task_mgmt_response_deadline
						    = 1;
				}
				atomic_inc(&task->refcount);
				spin_unlock(&session->task_lock);

				printk
				    ("iSCSI: session iscsi bus %d sending mgmt "
				     "%u abort task set to (%u %u %u %u) "
				     "at %lu\n",
				     session->iscsi_bus, task->mgmt_itt,
				     session->host_no, session->channel,
				     session->target_id, task->lun, jiffies);
				iscsi_xmit_task_mgmt(session,
						   ISCSI_TM_FUNC_ABORT_TASK_SET,
						     task, task->mgmt_itt);
				atomic_dec(&task->refcount);
			} else if (test_bit(TASK_TRY_LUN_RESET, &task->flags)) {
				session->mgmt_itt = task->mgmt_itt =
				    allocate_itt(session);
				if (session->reset_timeout) {
					session->task_mgmt_response_deadline =
					    jiffies +
					    (session->reset_timeout * HZ);
					if (session->
					    task_mgmt_response_deadline == 0)
						session->
						    task_mgmt_response_deadline
						    = 1;
				}
				atomic_inc(&task->refcount);
				spin_unlock(&session->task_lock);

				printk
				    ("iSCSI: session iscsi bus %d sending mgmt "
				     "%u LUN reset to (%u %u %u %u) at %lu\n",
				     session->iscsi_bus, task->mgmt_itt,
				     session->host_no, session->channel,
				     session->target_id, task->lun, jiffies);
				iscsi_xmit_task_mgmt(session,
						     ISCSI_TM_FUNC_LOGICAL_UNIT_RESET,
						     task, task->mgmt_itt);
				atomic_dec(&task->refcount);
			} else if (test_bit(TASK_TRY_WARM_RESET, &task->flags)) {
				/* block any new tasks from starting 
				 * and existing tasks from completing 
				 */
				set_bit(SESSION_RESETTING,
					&session->control_bits);

				list_for_each_entry(t,
						    &session->task_arrival_order,
						    task_order_link) {
					DEBUG_EH
					    ("iSCSI: session iscsi bus %d "
					     "target id %d warm target reset "
					     "causing problems for LUN %u\n",
					     session->iscsi_bus,
					     session->target_id, t->lun);
					set_bit(t->lun,
						session->luns_timing_out);
					/* the task scans above assume that all 
					 * tasks TIMEDOUT before error recovery
					 * could have killed the tasks.  Make it
					 * look like all tasks have TIMEDOUT,
					 * so that the LUNs affected by the 
					 * target reset can be recovered 
					 * in the same way as usual.
					 */
					del_task_timer(t);
					set_bit(0, &t->timedout);
					/* the task mgmt response will set 
					 * SESSION_TASK_TIMEDOUT and ensure 
					 * these get processed later 
					 */
				}

				session->mgmt_itt = task->mgmt_itt =
				    allocate_itt(session);
				if (session->reset_timeout) {
					session->task_mgmt_response_deadline =
					    jiffies +
					    (session->reset_timeout * HZ);
					if (session->
					    task_mgmt_response_deadline == 0)
						session->
						    task_mgmt_response_deadline
						    = 1;
				}
				atomic_inc(&task->refcount);
				spin_unlock(&session->task_lock);

				printk
				    ("iSCSI: session iscsi bus %d sending mgmt "
				     "%u warm target reset to (%u %u %u *) "
				     "at %lu\n",
				     session->iscsi_bus, task->mgmt_itt,
				     session->host_no, session->channel,
				     session->target_id, jiffies);
				iscsi_xmit_task_mgmt(session,
						     ISCSI_TM_FUNC_TARGET_WARM_RESET,
						     task, task->mgmt_itt);
				atomic_dec(&task->refcount);
			} else if (test_bit(TASK_TRY_COLD_RESET, &task->flags)) {

				/* block any new tasks from starting and 
				 * existing tasks from completing 
				 */
				set_bit(SESSION_RESETTING,
					&session->control_bits);

				list_for_each_entry(t,
						    &session->task_arrival_order,
						    task_order_link) {
					DEBUG_EH
					    ("iSCSI: session iscsi bus %d "
					     "target id %d cold target reset "
					     "causing problems for LUN %u\n",
					     session->iscsi_bus,
					     session->target_id, t->lun);
					set_bit(t->lun,
						session->luns_timing_out);
					/* the task scans above assume that 
					 * all tasks TIMEDOUT before error 
					 * recovery could have killed the tasks.					 * Make it look like all tasks have 
					 * TIMEDOUT, so that the LUNs affected 
					 * by the target reset can be recovered 					 * in the same way as usual.
					 */
					del_task_timer(t);
					set_bit(0, &t->timedout);
					/* the task mgmt response will set 
					 * SESSION_TASK_TIMEDOUT and ensure 
					 * these get processed later 
					 */
				}

				/* tell all devices attached to this target 
				 * that a reset occured we do this now, since 
				 * a cold reset should cause the target to drop
				 * the session, and we probably won't get a 
				 * task mgmt response for a cold reset.  
				 * FIXME: better to do this when the session 
				 * actually drops?
				 */
				scsi_report_device_reset(session->hba->host,
							 session->channel,
							 session->target_id);

				/* this is our last resort, so force 
				 * a 10 second deadline 
				 */
				session->task_mgmt_response_deadline =
				    jiffies + (10 * HZ);
				if (session->task_mgmt_response_deadline == 0)
					session->task_mgmt_response_deadline =
					    1;
				atomic_inc(&task->refcount);
				spin_unlock(&session->task_lock);

				printk
				    ("iSCSI: session iscsi bus %d sending mgmt "
				     "%u cold target reset to (%u %u %u *) "
				     "at %lu\n",
				     session->iscsi_bus, task->mgmt_itt,
				     session->host_no, session->channel,
				     session->target_id, jiffies);
				iscsi_xmit_task_mgmt(session,
						     ISCSI_TM_FUNC_TARGET_COLD_RESET,
						     task, task->mgmt_itt);
				atomic_dec(&task->refcount);
			}
		} else {
			spin_unlock(&session->task_lock);
			DEBUG_EH
			    ("iSCSI: session iscsi bus %d target id %d "
			     "couldn't find a task ready for error recovery "
			     "at %lu\n",
			     session->iscsi_bus, session->target_id, jiffies);
		}
	} else {
		/* Either don't need or can't do any recovery right now. */
		spin_unlock(&session->task_lock);
	}

	DEBUG_EH
	    ("iSCSI: session iscsi bus %d target id %d finished processing "
	     "timedout commands at %lu\n",
	     session->iscsi_bus, session->target_id, jiffies);
	return 0;
}

int
iscsi_xmit_task(struct iscsi_task *task)
{
	struct msghdr msg;
	struct iovec iov[(ISCSI_MAX_SG + 1 + 1)];
	struct IscsiScsiCmdHdr stsch;
	int rc, wlen;
	int remain;
	struct iscsi_session *session = task->session;
	Scsi_Cmnd *sc = task->scsi_cmnd;
	unsigned int segment_offset = 0, index = 0;
	uint32_t data_offset = 0;
	int xfrlen = 0;
	struct scatterlist *sglist = NULL, *sg, *first_sg = NULL, *last_sg =
	    NULL;
	int iovn, first_data_iovn = 0;
	int bytes_to_fill, bytes_from_segment;
	int pad_bytes = 0;
	char padding[4];
	uint32_t header_crc32c, data_crc32c;

	if (!task) {
		printk("iSCSI: xmit_task NULL\n");
		return 0;
	}

	if (!sc) {
		printk("iSCSI: xmit_task %u, cmnd NULL\n", task->itt);
		return 0;
	}

	DEBUG_FLOW
	    ("iSCSI: xmit_task itt %u to (%u %u %u %u), "
	     "cdb 0x%x, cmd_len %u, bufflen %u\n",
	     task->itt, sc->device->host->host_no, sc->device->channel,
	     sc->device->id, sc->device->lun, sc->cmnd[0], sc->cmd_len,
	     sc->request_bufflen);

	wlen = sizeof (stsch);
	memset(&stsch, 0, sizeof (stsch));

	if (test_bit(TASK_READ, &task->flags)) {
		/* read */
		stsch.flags |= ISCSI_FLAG_CMD_READ;
		stsch.data_length = htonl(iscsi_expected_data_length(sc));
	}
	if (test_bit(TASK_WRITE, &task->flags)) {
		/* write */
		stsch.flags |= ISCSI_FLAG_CMD_WRITE;
		stsch.data_length = htonl(iscsi_expected_data_length(sc));
	}

	/* tagged command queueing */
	stsch.flags |= (iscsi_command_attr(sc) & ISCSI_FLAG_CMD_ATTR_MASK);

	/* FIXME: if it's an untagged command, and we've already sent
	 * an untagged command to the LUN, don't send a 2nd untagged command.
	 * Leave it queued up and send it after the other command completes.
	 * We also don't want to block commands for other LUNs.  Basically,
	 * we need a per-LUN command queue.  For now, deal with it by
	 * setting the Scsi_Device queue_depth to 1 without TCQ.  We can
	 * reduce latency by keeping multiple commands per LUN queued to
	 * the HBA, but only sending one.  That takes a more code though.
	 */

	stsch.opcode = ISCSI_OP_SCSI_CMD;
	stsch.itt = htonl(task->itt);
	task->cmdsn = session->CmdSn;
	stsch.cmdsn = htonl(session->CmdSn);
	stsch.expstatsn = htonl(session->ExpStatSn);

	/* set the final bit when there are no unsolicited Data-out 
	 * PDUs following the command PDU 
	 */
	if (!test_bit(TASK_INITIAL_R2T, &task->flags))
		stsch.flags |= ISCSI_FLAG_FINAL;
	/* FIXME: why does clearing the flags crash the kernel? */

	/* single level LUN format puts LUN in byte 1, 0 everywhere else */
	stsch.lun[1] = sc->device->lun;

	memcpy(stsch.scb, sc->cmnd, MIN(sizeof (stsch.scb),
					(unsigned int) sc->cmd_len));

	ISCSI_TRACE(ISCSI_TRACE_TxCmd, sc, task, session->CmdSn,
		    ntohl(stsch.data_length));

	/* FIXME: Sending ImmediateData along with the cmd PDU */

	/* PDU header */
	iov[0].iov_base = &stsch;
	iov[0].iov_len = sizeof (stsch);
	iovn = 1;
	wlen = sizeof (stsch);

	/* HeaderDigests */
	if (session->HeaderDigest == ISCSI_DIGEST_CRC32C) {
		iov[1].iov_base = &header_crc32c;
		iov[1].iov_len = sizeof (header_crc32c);
		iovn = 2;
		wlen += sizeof (header_crc32c);
	}

	/* For ImmediateData, we need to compute the DataDigest also
	 */
	if (session->ImmediateData
	    && (sc->sc_data_direction == SCSI_DATA_WRITE)) {
		/* make sure we have data to send when we expect to */
		if (sc && (iscsi_expected_data_length(sc) == 0)
		    && ((sc->request_bufflen == 0)
			|| (sc->request_buffer == NULL))) {
			printk
			    ("iSCSI: xmit_task for itt %u, sc 0x%x, "
			     "expected %u, no data in buffer\n"
			     "             request_buffer %p len %u, "
			     "buffer %p len %u\n",
			     task->itt, sc->cmnd[0],
			     iscsi_expected_data_length(sc), sc->request_buffer,
			     sc->request_bufflen, sc->buffer, sc->bufflen);
			print_cmnd(sc);
			return 0;
		}
		remain = 0;
		/* Find the segment and offset within the segment 
		 * to start writing from. 
		 */
		if (sc && sc->use_sg) {
			sg = sglist = (struct scatterlist *) sc->request_buffer;
			segment_offset = data_offset;
			for (index = 0; index < sc->use_sg; index++) {
				if (segment_offset < sglist[index].length)
					break;
				else
					segment_offset -= sglist[index].length;
			}
			if (index >= sc->use_sg) {
				/* didn't find the offset, command 
				 * will eventually timeout 
				 */
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "xmit_data for itt %u couldn't find "
				     "offset %u in sglist %p, sc 0x%x, "
				     "bufflen %u, use_sg %u\n",
				     session->iscsi_bus, session->target_id,
				     task->itt, data_offset, sglist,
				     sc->cmnd[0], sc->request_bufflen,
				     sc->use_sg);
				print_cmnd(sc);
				ISCSI_TRACE(ISCSI_TRACE_OutOfData, sc, task,
					    index, sc->use_sg);
				return 0;
			}
		}

		first_data_iovn = iovn;
		if (session->FirstBurstLength)
			bytes_to_fill =
			    MIN(session->FirstBurstLength,
				session->MaxXmitDataSegmentLength);
		else
			bytes_to_fill = session->MaxXmitDataSegmentLength;
		bytes_to_fill = MIN(bytes_to_fill, (int)sc->request_bufflen);

		/* check if we need to pad the PDU */

		if (bytes_to_fill % PAD_WORD_LEN) {
			pad_bytes =
			    PAD_WORD_LEN - (bytes_to_fill % PAD_WORD_LEN);
			memset(padding, 0x0, sizeof (padding));
		} else {
			pad_bytes = 0;
		}

		if (sc) {
			/* find all the PDU data */
			if (sc->use_sg) {
				/* while there is more data and we want 
				 * to send more data 
				 */
				while (bytes_to_fill > 0) {
					if (index >= sc->use_sg) {
						printk
						    ("iSCSI: session iscsi bus "
						     "%d target id %d xmit_data"
						     " index %d exceeds "
						     "sc->use_sg %d, "
						     "bytes_to_fill %d, out of "
						     "buffers\n",
						     session->iscsi_bus,
						     session->target_id, index,
						     sc->use_sg, bytes_to_fill);
						/* the command will 
						 * eventually timeout 
						 */
						print_cmnd(sc);
						ISCSI_TRACE
						    (ISCSI_TRACE_OutOfData, sc,
						     task, index, sc->use_sg);
						goto done;
					}
					sg = &sglist[index];
					/* make sure the segment is mapped */
					if (!kmap_sg(sg)) {
						printk
						    ("iSCSI: session iscsi bus "
						     "%d target id %d xmit_data"
						     " couldn't map segment "
						     "%p\n",
						     session->iscsi_bus,
						     session->target_id, sg);
						goto done;
					} else if (first_sg == NULL) {
						first_sg = sg;
					}
					last_sg = sg;
					/* sanity check the sglist 
					 * segment length 
					 */
					if (sg->length <= segment_offset) {
						/* the sglist is corrupt */
						printk
						    ("iSCSI: session iscsi bus "
						     "%d target id %d xmit_data"
						     " index %d, length %u too "
						     "small for offset %u, "
						     "bytes_to_fill %d, sglist "
						     "has been corrupted\n",
						     session->iscsi_bus,
						     session->target_id, index,
						     sg->length, segment_offset,
						     bytes_to_fill);
						/* the command will 
						 * eventually timeout 
						 */
						print_cmnd(sc);
						ISCSI_TRACE
						    (ISCSI_TRACE_BadTxSeg, sc,
						     task, sg->length,
						     segment_offset);
						goto done;
					}
					bytes_from_segment =
					    sg->length - segment_offset;
					if (bytes_from_segment > bytes_to_fill) {
					/* only need part of this segment */
						iov[iovn].iov_base =
						page_address(sg->page) 
						+ sg->offset + segment_offset;
						iov[iovn].iov_len =
						    bytes_to_fill;
						xfrlen += bytes_to_fill;
						DEBUG_FLOW
						    ("iSCSI: session iscsi bus "
						     "%d target id %d xmit_data"
						     " xfrlen %d, to_fill %d, "
						     "from_segment %d, iov[%2d]"
						     " = partial sg[%2d]\n",
						     session->iscsi_bus,
						     session->target_id, xfrlen,
						     bytes_to_fill,
						     bytes_from_segment, iovn,
						     index);
						iovn++;
						segment_offset += bytes_to_fill;
						break;
					} else {
						/* need all of this segment, 
						 * and possibly more from the 
						 * next 
						 */
						iov[iovn].iov_base =
						    sg_virtual_address(sg) +
						    segment_offset;
						iov[iovn].iov_len =
						    bytes_from_segment;
						xfrlen += bytes_from_segment;
						DEBUG_FLOW
						    ("iSCSI: session iscsi bus "
						     "%d target id %d xmit_data"
						     " xfrlen %d, to_fill %d, "
						     "from_segment %d, iov[%2d]"
						     " = sg[%2d]\n",
						     session->iscsi_bus,
						     session->target_id, xfrlen,
						     bytes_to_fill,
						     bytes_from_segment, iovn,
						     index);
						bytes_to_fill -=
						    bytes_from_segment;
						iovn++;
						/* any remaining data starts at
						 * offset 0 of the next segment 
						 */
						index++;
						segment_offset = 0;
					}
				}
			} else {
				/* no scatter-gather */
				if ((sc->request_buffer + data_offset +
				     bytes_to_fill) <=
				    (sc->request_buffer +
				     sc->request_bufflen)) {
					/* send all the data */
					iov[iovn].iov_base =
					    sc->request_buffer + data_offset;
					iov[iovn].iov_len = xfrlen =
					    bytes_to_fill;
					iovn++;
				} else if ((sc->request_buffer + data_offset) <
					   (sc->request_buffer +
					    sc->request_bufflen)) {
					/* send some data, but can't send 
					 * all requested 
					 */
					xfrlen =
					    sc->request_bufflen - data_offset;
					printk
					    ("iSCSI: xmit_data ran out of data,"
					     " buffer %p len %u but offset %d "
					     "length %d, sending final %d "
					     "bytes\n",
					     sc->request_buffer,
					     sc->request_bufflen, data_offset,
					     bytes_to_fill, xfrlen);
					iov[iovn].iov_base =
					    sc->request_buffer + data_offset;
					iov[iovn].iov_len = xfrlen;
					iovn++;
					stsch.flags = ISCSI_FLAG_FINAL;
					remain = xfrlen;
				} else {
					/* can't send any data */
					printk
					    ("iSCSI: xmit_data ran out of data,"
					     " buffer %p len %u but offset %d "
					     "length %d, sending no more "
					     "data\n",
					     sc->request_buffer,
					     sc->request_bufflen, data_offset,
					     bytes_to_fill);
					goto done;
				}
			}

			if (pad_bytes) {
				iov[iovn].iov_base = padding;
				iov[iovn].iov_len = pad_bytes;
				iovn++;
				wlen += pad_bytes;
			}
		}

		/* put the data length in the PDU header */
		hton24(stsch.dlength, xfrlen);
		stsch.data_length = htonl(sc->request_bufflen);
		wlen += xfrlen;
	}

	/* header complete, we can finally calculate the HeaderDigest */
	if (session->HeaderDigest == ISCSI_DIGEST_CRC32C) {
		header_crc32c = iscsi_crc32c(&stsch, sizeof (stsch));
		/* FIXME: this may not be SMP safe, but it's only for 
		 * testing anyway, so it probably doesn't need to be 
		 */
		if (session->fake_write_header_mismatch > 0) {
			session->fake_write_header_mismatch--;
			smp_mb();
			printk
			    ("iSCSI: session iscsi bus %d target id %d faking "
			     "HeaderDigest mismatch for itt %u, \n",
			     session->iscsi_bus, session->target_id, task->itt);
			header_crc32c = 0x01020304;
		}
	}
	/* DataDigest */
	if (xfrlen && (session->DataDigest == ISCSI_DIGEST_CRC32C)) {
		int i;

		data_crc32c =
		    iscsi_crc32c(iov[first_data_iovn].iov_base,
				 iov[first_data_iovn].iov_len);
		for (i = first_data_iovn + 1; i < iovn; i++) {
			data_crc32c =
			    iscsi_crc32c_continued(iov[i].iov_base,
						   iov[i].iov_len, data_crc32c);
		}

		/* FIXME: this may not be SMP safe, but it's only for testing 
		 * anyway, so it probably doesn't need to be 
		 */
		if (session->fake_write_data_mismatch > 0) {
			session->fake_write_data_mismatch--;
			smp_mb();
			printk
			    ("iSCSI: session iscsi bus %d target id %d faking "
			     "DataDigest mismatch for itt %u\n",
			     session->iscsi_bus, session->target_id, task->itt);
			data_crc32c = 0x01020304;
		}
		iov[iovn].iov_base = &data_crc32c;
		iov[iovn].iov_len = sizeof (data_crc32c);
		iovn++;
		wlen += sizeof (data_crc32c);
	}

	memset(&msg, 0, sizeof (msg));
	msg.msg_iov = &iov[0];
	msg.msg_iovlen = iovn;
	
	ISCSI_TRACE(ISCSI_TRACE_TxDataPDU, sc, task, data_offset,
		    xfrlen);
	
	rc = iscsi_sendmsg(session, &msg, wlen);
	if (rc != wlen) {
		printk("iSCSI: session iscsi bus %d target id %d "
		       "xmit_data failed to send %d bytes, rc %d\n",
		       session->iscsi_bus, session->target_id, wlen, rc);
		iscsi_drop_session(session);
		goto done;
	}
	session->CmdSn++;
	
	return 1;
	
 done:
	if (first_sg) {
		/* undo any temporary mappings */
		for (sg = first_sg; sg <= last_sg; sg++) {
			kunmap_sg(sg);
		}
	}
	return 0;
}

int
fake_task_completion(struct iscsi_session *session, struct iscsi_task *task)
{
	struct IscsiScsiRspHdr stsrh;
	Scsi_Cmnd *sc = task->scsi_cmnd;
	unsigned char sense_buffer[32];
	unsigned int senselen = 0;
	uint32_t itt = task->itt;

	/* For testing, fake a completion with various status
	 * codes when requested, without ever sending the task or
	 * any data to the target, so that data corruption
	 * problems will occur if the retry isn't handled
	 * correctly.  
	 */

	memset(&stsrh, 0, sizeof (stsrh));
	stsrh.itt = htonl(itt);

	if (session->fake_status_unreachable) {
		session->fake_status_unreachable--;
		stsrh.response = 0x82;
		printk
		    ("iSCSI: session iscsi bus %d target id %d faking iSCSI "
		     "response 0x82 for itt %u command 0x%x to LUN %u at %lu\n",
		     session->iscsi_bus, session->target_id, task->itt,
		     sc->cmnd[0], task->lun, jiffies);
	} else if (session->fake_status_busy) {
		session->fake_status_busy--;
		stsrh.cmd_status = SAM_STAT_BUSY;
		printk
		    ("iSCSI: session iscsi bus %d target id %d faking SCSI "
		     "status BUSY for itt %u command 0x%x to LUN %u at %lu\n",
		     session->iscsi_bus, session->target_id, task->itt,
		     sc->cmnd[0], task->lun, jiffies);
	} else if (session->fake_status_queue_full) {
		session->fake_status_queue_full--;
		stsrh.cmd_status = SAM_STAT_TASK_SET_FULL;
		printk
		    ("iSCSI: session iscsi bus %d target id %d faking SCSI "
		     "status QUEUE_FULL for itt %u command 0x%x to "
		     "LUN %u at %lu\n",
		     session->iscsi_bus, session->target_id, task->itt,
		     sc->cmnd[0], task->lun, jiffies);
	} else if (session->fake_status_aborted) {
		session->fake_status_aborted--;
		stsrh.cmd_status = SAM_STAT_CHECK_CONDITION;
		stsrh.residual_count = htonl(iscsi_expected_data_length(sc));
		stsrh.flags |= ISCSI_FLAG_CMD_UNDERFLOW;
		sense_buffer[0] = 0x70;
		sense_buffer[2] = ABORTED_COMMAND;
		senselen = 8;
		printk
		    ("iSCSI: session iscsi bus %d target id %d faking SCSI "
		     "status CHECK_CONDITION key ABORTED_COMMAND for itt %u "
		     "command 0x%x to LUN %u at %lu\n",
		     session->iscsi_bus, session->target_id, task->itt,
		     sc->cmnd[0], task->lun, jiffies);
	} else {
		/* nothing left to fake */
		session->fake_status_lun = -2;
		return 0;
	}

	/* determine command result based on the iSCSI response, 
	 * status, and sense 
	 */
	process_task_response(session, task, &stsrh, sense_buffer, senselen);

	/* try to complete the command */
	complete_task(session, itt);
	/* Note: we lose the task_lock by calling complete_task */

	return 1;
}
