/*
 * iSCSI driver for Linux
 * Copyright (C) 2001 Cisco Systems, Inc.
 * maintained by linux-iscsi-devel@lists.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi-timer.c,v 1.1.2.7 2003/12/16 05:59:53 smhatre Exp $
 */

#include <linux/kernel.h>

#include "iscsi-protocol.h"
#include "iscsi-session.h"
#include "iscsi-probe.h"
#include "iscsi.h"

extern struct iscsi_hba *iscsi_hba;
extern int this_is_iscsi_boot;
extern int iscsi_system_is_rebooting;
extern volatile pid_t iscsi_timer_pid;
extern volatile unsigned long iscsi_timer_running;
extern volatile unsigned long init_module_complete;

static void
useless_timeout_function(unsigned long arg)
{

}

void
add_completion_timer(Scsi_Cmnd * sc)
{
	if (sc->scsi_done != iscsi_done) {
		sc->eh_timeout.data = (unsigned long) sc;
		sc->eh_timeout.expires = jiffies + sc->timeout_per_command;
		sc->eh_timeout.function = useless_timeout_function;
		add_timer(&sc->eh_timeout);
	}
}

void
add_command_timer(struct iscsi_session *session, Scsi_Cmnd * sc,
		  void (*fn) (unsigned long))
{
	unsigned long now = jiffies;
	unsigned long expires;

	if (sc->eh_timeout.function != NULL) {
		DEBUG_QUEUE
		    ("iSCSI: add_command_timer cdb 0x%x host %u channel %u "
		     "target %u lun %u deleting existing timer at %lu\n",
		     sc->cmnd[0], sc->device->host->host_no, 
		     sc->device->channel, sc->device->id, sc->device->lun,
		     jiffies);
		del_timer_sync(&sc->eh_timeout);
	}

	/* default is based on the number of retries remaining and 
	 * the timeout for each 
	 */
	if ((sc->allowed > 1) && (sc->retries < sc->allowed))
		sc->eh_timeout.expires =
		    now +
		    ((sc->allowed - sc->retries) * sc->timeout_per_command);
	else
		sc->eh_timeout.expires = now + sc->timeout_per_command;

	if (sc->eh_timeout.expires == 0)
		sc->eh_timeout.expires = 1;

	/* but each session may override that timeout value for 
	 * certain disk commands 
	 */
	if (sc->device->type == TYPE_DISK) {
		if (internally_retryable(sc)) {
			/* we only increase timeouts on commands 
			 * that are internally retryable,
			 * since some commands must be completed in a 
			 * reasonable amount of time in order for the 
			 * upper layers to behave properly.
			 */
			if ((session->min_disk_command_timeout < 0)
			    && (session->max_disk_command_timeout <= 0)) {
				/* no command timeout (infinite retries) 
				 * when the min is infinite and there is no max 
				 */
				sc->eh_timeout.expires = 0;
			} else if (session->min_disk_command_timeout > 0) {
				/* clamp the SCSI layer's timeout 
				 * to be at least the minimum 
				 */
				expires =
				    now +
				    (session->min_disk_command_timeout * HZ);
				if (expires > sc->eh_timeout.expires) {
					/* use a longer timeout */
					if (expires)
						sc->eh_timeout.expires =
						    expires;
					else
						sc->eh_timeout.expires = 1;
				}
			}
		}

		/* we always allow the timeout to be shortened, to make 
		 * multipath drivers happy.
		 */
		if (session->max_disk_command_timeout > 0) {
			/* clamp the SCSI layer's timeout to be 
			 * at most the maximum 
			 */
			expires =
			    now + (session->max_disk_command_timeout * HZ);
			if (expires < sc->eh_timeout.expires) {
				/* use a shorter timeout */
				if (expires)
					sc->eh_timeout.expires = expires;
				else
					sc->eh_timeout.expires = 1;
			}
		}
	}

	if (sc->eh_timeout.expires) {
		DEBUG_QUEUE
		    ("iSCSI: add_command_timer cdb 0x%x host %u channel %u "
		     "target %u lun %u adding timer at %lu, expires %lu, "
		     "timeout %u, retries %u, allowed %u\n",
		     sc->cmnd[0], sc->device->host->host_no,
		     sc->device->channel, sc->device->id, sc->device->lun,
		     jiffies, sc->eh_timeout.expires, sc->timeout_per_command, 
		     sc->retries, sc->allowed);
		sc->eh_timeout.data = (unsigned long) sc;
		sc->eh_timeout.function = fn;
		add_timer(&sc->eh_timeout);
	}
}

void
add_task_timer(struct iscsi_task *task, void (*fn) (unsigned long))
{
	if (task->timer.function != NULL) {
		DEBUG_QUEUE
		    ("iSCSI: add_task_timer task of itt %u "
		     "deleting existing timer at %lu\n",
		     task->itt, jiffies);

		del_timer_sync(&task->timer);
	}
	task->timer.data = (unsigned long) task;
	task->timer.expires = jiffies + task->scsi_cmnd->timeout_per_command;
	task->timer.function = fn;

	DEBUG_QUEUE
	    ("iSCSI: add_task_timer task of itt %u adding timer at %lu, "
	     "expires %lu, timeout %u\n",
	     task->itt, jiffies, task->timer.expires,
	     task->scsi_cmnd->timeout_per_command);
	add_timer(&task->timer);
}

int
del_command_timer(Scsi_Cmnd * sc)
{
	int ret;

	DEBUG_QUEUE
	    ("iSCSI: del_command_timer cdb 0x%x host %u channel %u "
	     "target %u lun %u deleting timer at %lu\n",
		     sc->cmnd[0], sc->device->host->host_no, 
		     sc->device->channel, sc->device->id, sc->device->lun,
		     jiffies);

	ret = del_timer_sync(&sc->eh_timeout);

	sc->eh_timeout.expires = 0;
	sc->eh_timeout.data = (unsigned long) NULL;
	sc->eh_timeout.function = NULL;
	return ret;
}

int
del_task_timer(struct iscsi_task *task)
{
	int ret;

	DEBUG_QUEUE
	    ("iSCSI: del_task_timer task of itt %u deleting timer at %lu\n",
	     task->itt, jiffies);

	ret = del_timer_sync(&task->timer);

	task->timer.expires = 0;
	task->timer.data = (unsigned long) NULL;
	task->timer.function = NULL;
	return ret;
}

void
iscsi_command_times_out(unsigned long arg)
{
	Scsi_Cmnd *sc = (Scsi_Cmnd *) arg;
	struct iscsi_session *session = (struct iscsi_session *) sc->SCp.ptr;

	/* we can safely use the session pointer, since during a session 
	 * termination the rx thread will make sure all commands have been 
	 * completed before it drops the session refcount.
	 */

	if (session == NULL)
		return;

	DEBUG_EH
	    ("iSCSI: session iscsi bus %d target id %d timer for command "
	     "0x%x host %u channel %u target %u lun %u expired at %lu, "
	     "retries %d, allowed %d\n",
	     session->iscsi_bus, session->target_id, sc->cmnd[0],
	     sc->device->host->host_no, sc->device->channel, sc->device->id,
	     sc->device->lun, jiffies, sc->retries, sc->allowed);

	/* tell the tx thread that a command has timed out */
	set_bit(COMMAND_TIMEDOUT, command_flags(sc));
	smp_wmb();

	set_bit(SESSION_COMMAND_TIMEDOUT, &session->control_bits);
	smp_wmb();

	/* wake up the tx thread to deal with the timeout */
	set_bit(TX_WAKE, &session->control_bits);
	smp_mb();
	/* we can't know which wait_q the tx thread is in (if any), 
	 * so wake them both 
	 */
	wake_up(&session->tx_wait_q);
	wake_up(&session->login_wait_q);
}

void
iscsi_task_times_out(unsigned long arg)
{
	struct iscsi_task *task = (struct iscsi_task *) arg;
	struct iscsi_session *session = task->session;

	/* we can safely use the session pointer, since during a session 
	 * termination the rx thread will make sure all tasks have been 
	 * completed before it drops the session refcount.
	 */
	if (session == NULL)
		return;

	DEBUG_TIMEOUT
	    ("iSCSI: session iscsi bus %d target id %d timer for task "
	     "of itt %u expired at %lu\n",
	     session->iscsi_bus, session->target_id, task->itt, jiffies);

	/* stop new tasks from being sent to this LUN (force error recovery) */
	set_bit(task->lun, session->luns_timing_out);
	smp_wmb();

	/* tell the tx thread that a task has timed out */
	set_bit(0, &task->timedout);
	smp_wmb();

	set_bit(SESSION_TASK_TIMEDOUT, &session->control_bits);
	smp_wmb();

	/* wake up the tx thread to deal with the timeout 
	 * and possible error recovery 
	 */
	set_bit(TX_WAKE, &session->control_bits);
	smp_mb();

	/* we can't know which wait_q the tx thread is in (if any), 
	 * so wake them both 
	 */
	wake_up(&session->tx_wait_q);
	wake_up(&session->login_wait_q);
}


/*
 *  FIXME: it'd probably be cleaner to move the timeout logic to the rx thread.
 *         The only danger is if the rx thread somehow blocks indefinately.
 *         Doing timeouts here makes sure the timeouts get checked, at the
 *         cost of having this code constantly loop.
 */
int
iscsi_timer_thread(void *vtaskp)
{
	struct iscsi_session *session;
	unsigned long noqueue_flags_;

	/* become a child of init, and abandon any user space resources */
	iscsi_daemonize("iscsi-timer");

	iscsi_timer_pid = current->pid;
	smp_mb();
	DEBUG_INIT("iSCSI: timer pid %d starting at %lu\n", iscsi_timer_pid,
		   jiffies);

	spin_lock_irq(&current->sighand->siglock);
	/* Block all signals except SIGKILL */
	siginitsetinv(&current->blocked, sigmask(SIGKILL));
	recalc_sigpending();
	spin_unlock_irq(&current->sighand->siglock);

	/* wait for the module to initialize */
	while (test_bit(0, &init_module_complete) == 0) {
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(MSECS_TO_JIFFIES(10));
		if (signal_pending(current)) {
			iscsi_timer_running = 0;
			smp_mb();
			return 0;
		}
	}

	DEBUG_INIT("iSCSI: timer waiting for HBA at %lu\n", jiffies);
	while (!signal_pending(current)) {
		if (iscsi_hba)
			break;

		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(MSECS_TO_JIFFIES(10));
	}

	DEBUG_INIT("iSCSI: timer looping over HBAs at %lu\n", jiffies);

      continue_timer_thread:

	while (!signal_pending(current)) {
		unsigned long next_timeout = jiffies + (5 * HZ);
		spin_lock_irqsave(&iscsi_hba->session_lock, noqueue_flags_);
		list_for_each_entry(session, &iscsi_hba->session_head, 
				    session_link) {
			unsigned long session_timeout = 0;

			if (LOG_ENABLED(ISCSI_LOG_ALLOC))
				printk("iSCSI: session iscsi bus %d "
				       "target id %d, rx %5u, tx %5u, %u "
				       "luns, %3u r, %3u d, %3u n, %3u t,"
				       " bits 0x%08lx at %lu\n",
					session->iscsi_bus, session->target_id,
					session->rx_pid, session->tx_pid,
					session->num_luns,
					atomic_read(&session->num_retry_cmnds),
					session->num_deferred_cmnds,
					atomic_read(&session->num_cmnds),
					atomic_read(&session->num_active_tasks),
					session->control_bits, jiffies);

			session_timeout = check_session_timeouts(session);

			/* find the earliest timeout that might occur, 
			 * so that we know how long to sleep 
			 */
			if (session_timeout && time_before_eq(session_timeout, 
							      jiffies))
				printk("iSCSI: ignoring session timeout "
				       "%lu at %lu, last rx %lu, for "
				       "session iscsi bus %d target id %d\n",
					session_timeout, jiffies,
					session->last_rx, session->iscsi_bus,
					session->target_id);
			else if (session_timeout && time_before(session_timeout,
								next_timeout))
				next_timeout = session_timeout;

		}
		spin_unlock_irqrestore(&iscsi_hba->session_lock,
				       noqueue_flags_);

		if (LOG_ENABLED(ISCSI_LOG_ALLOC))
			printk("iSCSI: timer - host %d can_queue %d "
			       "at %lu\n", iscsi_hba->host->host_no, 
				iscsi_hba->host->can_queue, jiffies);

		/* possibly start LUN probing */
		if (iscsi_lun_probe_start) {
			if (time_before_eq(iscsi_lun_probe_start, jiffies)) {
				iscsi_possibly_start_lun_probing();
			} else
			    if (time_before_eq
				(iscsi_lun_probe_start, next_timeout)) {
				next_timeout = iscsi_lun_probe_start;
			}
		}

		/* sleep for a while */
		if (time_before(jiffies, next_timeout)) {
			unsigned long sleep;

			/* sleep til the next time a timeout might occur, 
			 * and handle jiffies wrapping 
			 */
			if (next_timeout < jiffies)
				sleep = (ULONG_MAX - jiffies + next_timeout);
			else
				sleep = (next_timeout - jiffies);
			DEBUG_FLOW
			    ("iSCSI: timer sleeping for %lu jiffies, "
			     "now %lu, next %lu, HZ %u\n",
			     sleep, jiffies, next_timeout, HZ);

			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(sleep);
			if (signal_pending(current))
				goto finished;
		} else {
			/* this should never happen, but make sure we block 
			 * for at least a little while if it does somehow, 
			 * otherwise it'll lock up the machine and be impossible
			 * to debug what went wrong.
			 */
			DEBUG_FLOW
			    ("iSCSI: timer forced to sleep, now %lu, "
			     "next %lu, HZ %u\n",
			     jiffies, next_timeout, HZ);
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(1);
			if (signal_pending(current))
				goto finished;
		}
	}

      finished:
	/* timer finished */

	if ((this_is_iscsi_boot) && (!iscsi_system_is_rebooting)) {
		printk("\niSCSI: timer_thread got signalled\n");
		flush_signals(current);
		goto continue_timer_thread;
	}

	DEBUG_INIT("iSCSI: timer leaving kernel at %lu\n", jiffies);

	set_current_state(TASK_RUNNING);

	iscsi_timer_running = 0;
	iscsi_timer_pid = 0;
	smp_mb();

	return 0;
}

