/*
 * iSCSI driver for Linux
 * Copyright (C) 2001 Cisco Systems, Inc.
 * maintained by linux-iscsi-devel@lists.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsi-xmit-pdu.c,v 1.1.2.10 2004/01/30 20:42:06 haapi Exp $
 */

#include <linux/kernel.h>

#include "iscsi-protocol.h"
#include "iscsi-login.h"
#include "iscsi-crc.h"
#include "iscsi.h"

#ifndef USE_LOCAL_CRC32C
#include <linux/crypto.h>
#endif

int
iscsi_send_pdu(struct iscsi_session *session, struct IscsiHdr *pdu, char *data,
	       int timeout)
{
	struct msghdr msg;
	struct iovec iov[3];
	char padding[4];
	int pad = 0;
	int rc;
	int pdu_length = 0;
	int data_length;

	if (pdu == NULL) {
		printk
		    ("iSCSI: session iscsi bus %d target id %d, "
		     "pdu NULL, can't send PDU header\n",
		     session->iscsi_bus, session->target_id);
		return 0;
	}

	memset(iov, 0, sizeof (iov));
	memset(&msg, 0, sizeof (msg));
	msg.msg_iov = iov;
	msg.msg_iovlen = 1;

	/* pdu header */
	iov[0].iov_base = pdu;
	iov[0].iov_len = sizeof (*pdu);
	pdu_length = sizeof (*pdu);

	/* pdu data */
	data_length = ntoh24(pdu->dlength);
	if (data) {
		iov[msg.msg_iovlen].iov_base = data;
		iov[msg.msg_iovlen].iov_len = data_length;
		msg.msg_iovlen++;
		pdu_length += ntoh24(pdu->dlength);
	} else if (data_length) {
		printk
		    ("iSCSI: session iscsi bus %d target id %d pdu 0x%x with "
		     "dlength %d, but data NULL\n",
		     session->iscsi_bus, session->target_id, pdu->opcode,
		     data_length);
		return 0;
	}

	/* add any padding needed */
	if (pdu_length % PAD_WORD_LEN) {
		memset(padding, 0x0, sizeof (padding));
		pad = PAD_WORD_LEN - (pdu_length % PAD_WORD_LEN);
	}
	if (pad) {
		iov[msg.msg_iovlen].iov_base = padding;
		iov[msg.msg_iovlen].iov_len = pad;
		msg.msg_iovlen++;
		pdu_length += pad;
	}

	/* set a timer, though we shouldn't really need one */
	if (timeout) {
		session->login_phase_timer = jiffies + (timeout * HZ);
		smp_mb();
	}

	if (LOG_ENABLED(ISCSI_LOG_LOGIN)) {
		char *text = data;
		char *end = text + ntoh24(pdu->dlength);
		int show_text = 0;

		if ((pdu->opcode & ISCSI_OPCODE_MASK) == ISCSI_OP_LOGIN_CMD) {
			struct IscsiLoginHdr *login_pdu =
			    (struct IscsiLoginHdr *) pdu;
			/* show the login phases and tbit */
			printk
			    ("iSCSI: session iscsi bus %d target id %d sending "
			     "login pdu with current phase %d, next %d, transit"
			     " 0x%x, dlength %d at %lu, timeout at %lu "
			     "(%d seconds)\n",
			     session->iscsi_bus, session->target_id,
			     ISCSI_LOGIN_CURRENT_STAGE(login_pdu->flags),
			     ISCSI_LOGIN_NEXT_STAGE(login_pdu->flags),
			     login_pdu->flags & ISCSI_FLAG_LOGIN_TRANSIT,
			     ntoh24(pdu->dlength), jiffies,
			     session->login_phase_timer,
			     session->login_timeout);
			show_text = 1;
		} else if ((pdu->opcode & ISCSI_OPCODE_MASK) ==
			   ISCSI_OP_TEXT_CMD) {
			printk
			    ("iSCSI: session iscsi bus %d target id %d sending "
			     "text pdu, dlength %d at %lu, timeout at %lu "
			     "(%d seconds)\n",
			     session->iscsi_bus, session->target_id,
			     ntoh24(pdu->dlength), jiffies,
			     session->login_phase_timer,
			     session->login_timeout);
			show_text = 1;
		} else {
			printk
			    ("iSCSI: session iscsi bus %d target id %d sending "
			     "pdu with opcode 0x%x, dlength %d at %lu, timeout "
			     "at %lu (%d seconds)\n",
			     session->iscsi_bus, session->target_id,
			     pdu->opcode, ntoh24(pdu->dlength), jiffies,
			     session->login_phase_timer,
			     session->login_timeout);
		}

		/* show all the text that we're sending */
		while (show_text && (text < end)) {
			printk
			    ("iSCSI: session iscsi bus %d target id %d "
			     "login text: %s\n",
			     session->iscsi_bus, session->target_id, text);
			text += strlen(text);
			while ((text < end) && (*text == '\0'))
				text++;
		}
	}

	rc = iscsi_sendmsg(session, &msg, pdu_length);

	/* clear the timer */
	session->login_phase_timer = 0;
	smp_mb();

	if (rc != pdu_length) {
		char *error;
		if ((rc < 0) && (error = iscsi_strerror(-rc))
		    && (error[0] != '\0'))
			printk
			    ("iSCSI: session iscsi bus %d target id %d "
			     "failed to send login PDU, rc %d, %s\n",
			     session->iscsi_bus, session->target_id, rc,
			     iscsi_strerror(-rc));
		else
			printk
			    ("iSCSI: session iscsi bus %d target id %d "
			     "failed to send login PDU, rc %d\n",
			     session->iscsi_bus, session->target_id, rc);

		return 0;
	}

	DEBUG_FLOW
	    ("iSCSI: session iscsi bus %d target id %d sent login pdu 0x%x"
	     " at %lu, length %d, dlength %d\n",
	     session->iscsi_bus, session->target_id, pdu->opcode, jiffies,
	     pdu_length, ntoh24(pdu->dlength));

	return 1;
}

int
iscsi_xmit_task_mgmt(struct iscsi_session *session, uint8_t func_type,
		     struct iscsi_task *task, uint32_t mgmt_itt)
{
	struct msghdr msg;
	struct iovec iov[2];
	int rc, wlen;
	struct IscsiScsiTaskMgtHdr ststmh;
	uint32_t crc32c;

	memset(&ststmh, 0, sizeof (ststmh));
	ststmh.opcode = ISCSI_OP_SCSI_TASK_MGT_MSG | ISCSI_OP_IMMEDIATE;
	ststmh.flags =
	    ISCSI_FLAG_FINAL | (func_type & ISCSI_FLAG_TASK_MGMT_FUNCTION_MASK);
	ststmh.rtt = RSVD_TASK_TAG;
	ststmh.itt = htonl(mgmt_itt);
	ststmh.cmdsn = htonl(session->CmdSn);	/* CmdSN not incremented 
						 * after imm cmd 
						 */
	ststmh.expstatsn = htonl(session->ExpStatSn);

	switch (func_type) {
	case ISCSI_TM_FUNC_ABORT_TASK:
		/* need a task for this */
		if (task) {
			ststmh.refcmdsn = htonl(task->cmdsn);
			ststmh.rtt = htonl(task->itt);
			ststmh.lun[1] = task->lun;
			ISCSI_TRACE(ISCSI_TRACE_TxAbort, task->scsi_cmnd, task,
				    task->mgmt_itt, 0);
		} else {
			printk
			    ("iSCSI: session iscsi bus %d target id %d "
			     "failed to send abort, task unknown\n",
			     session->iscsi_bus, session->target_id);
			return 0;
		}
		break;
	case ISCSI_TM_FUNC_ABORT_TASK_SET:
		/* need a LUN for this */
		if (task) {
			ststmh.lun[1] = task->lun;
			ISCSI_TRACE(ISCSI_TRACE_TxAbortTaskSet, task->scsi_cmnd,
				    task, task->mgmt_itt, 0);
		} else {
			printk
			    ("iSCSI: session iscsi bus %d target id %d "
			     "failed to send abort task set, LUN unknown\n",
			     session->iscsi_bus, session->target_id);
			return 0;
		}
		break;
	case ISCSI_TM_FUNC_LOGICAL_UNIT_RESET:
		/* need a LUN for this */
		if (task) {
			ststmh.lun[1] = task->lun;
			ISCSI_TRACE(ISCSI_TRACE_TxLunReset, task->scsi_cmnd,
				    task, task->mgmt_itt, 0);
		} else {
			printk
			    ("iSCSI: session iscsi bus %d target id %d "
			     "failed to send logical unit reset, no task\n",
			     session->iscsi_bus, session->target_id);
			return 0;
		}
		break;
	case ISCSI_TM_FUNC_TARGET_WARM_RESET:
		ISCSI_TRACE(ISCSI_TRACE_TxWarmReset,
			    task ? task->scsi_cmnd : NULL, task, mgmt_itt, 0);
		break;
	case ISCSI_TM_FUNC_TARGET_COLD_RESET:
		ISCSI_TRACE(ISCSI_TRACE_TxColdReset,
			    task ? task->scsi_cmnd : NULL, task, mgmt_itt, 0);
		break;
	default:
		printk
		    ("iSCSI: unknown task mgmt function type %u for session "
		     "iscsi bus %d target id %d to %s\n",
		     func_type, session->iscsi_bus, session->target_id,
		     session->log_name);
		return 0;
		break;
	}

	iov[0].iov_base = &ststmh;
	iov[0].iov_len = sizeof (ststmh);
	memset(&msg, 0, sizeof (msg));
	msg.msg_iov = iov;
	msg.msg_iovlen = 1;
	wlen = sizeof (ststmh);

	/* HeaderDigests */
	if (session->HeaderDigest == ISCSI_DIGEST_CRC32C) {
		crc32c = iscsi_crc32c(&ststmh, sizeof (ststmh));
		iov[msg.msg_iovlen].iov_base = &crc32c;
		iov[msg.msg_iovlen].iov_len = sizeof (crc32c);
		msg.msg_iovlen++;
		wlen += sizeof (crc32c);
	}

	rc = iscsi_sendmsg(session, &msg, wlen);
	if (rc != wlen) {
		printk
		    ("iSCSI: session iscsi bus %d target id %d "
		     "xmit_task_mgmt failed, rc %d\n",
		     session->iscsi_bus, session->target_id, rc);
		iscsi_drop_session(session);
		return 0;
	}

	return 1;
}

static int
iscsi_set_direction(struct iscsi_task *task)
{
	if (task && task->scsi_cmnd) {
		switch (task->scsi_cmnd->cmnd[0]) {
		case TEST_UNIT_READY:
		case START_STOP:
		case REZERO_UNIT:
		case WRITE_FILEMARKS:
		case SPACE:
		case ERASE:
		case ALLOW_MEDIUM_REMOVAL:
			/* just control commands */
			__set_bit(TASK_CONTROL, &task->flags);
			return TASK_CONTROL;
		case WRITE_6:
		case WRITE_10:
		case WRITE_12:
		case 0x8a:	/* WRITE_16 */
		case 0x8e:	/* write and verify 16 */
		case 0x93:	/* write same 16 */
		case WRITE_LONG:
		case WRITE_SAME:
		case WRITE_BUFFER:
		case WRITE_VERIFY:
		case WRITE_VERIFY_12:
		case COMPARE:
		case COPY:
		case COPY_VERIFY:
		case SEARCH_EQUAL:
		case SEARCH_HIGH:
		case SEARCH_LOW:
		case SEARCH_EQUAL_12:
		case SEARCH_HIGH_12:
		case SEARCH_LOW_12:
		case FORMAT_UNIT:
		case REASSIGN_BLOCKS:
		case RESERVE:
		case MODE_SELECT:
		case MODE_SELECT_10:
		case LOG_SELECT:
		case SEND_DIAGNOSTIC:
		case CHANGE_DEFINITION:
		case UPDATE_BLOCK:
		case SET_WINDOW:
		case MEDIUM_SCAN:
		case SEND_VOLUME_TAG:
		case PERSISTENT_RESERVE_OUT:
		case WRITE_LONG_2:
			__set_bit(TASK_WRITE, &task->flags);
			return TASK_WRITE;
		default:
			__set_bit(TASK_READ, &task->flags);
			return TASK_READ;
		}
	}

	return -1;
}

void
iscsi_xmit_queued_cmnds(struct iscsi_session *session)
{
	Scsi_Cmnd *sc;
	struct iscsi_task *task = NULL;
	unsigned long noqueue_flags_;
	uint32_t imm_data_length = 0;

	if (!session) {
		printk("iSCSI: can't xmit queued commands, no session\n");
		return;
	}

	for (;;) {

		if (signal_pending(current)) {
			DEBUG_QUEUE
			    ("iSCSI: session iscsi bus %d target id %d can't "
			     "start tasks now, signal pending\n",
			     session->iscsi_bus, session->target_id);
			break;
		}

		if ((atomic_read(&session->num_cmnds) == 0)
		    && (atomic_read(&session->num_retry_cmnds) == 0)) {
			DEBUG_QUEUE
			    ("iSCSI: no SCSI cmnds queued for session "
			     "iscsi bus %d target id %d to %s\n",
			     session->iscsi_bus, session->target_id,
			     session->log_name);
			break;
		}

		if (!sna_lte(session->CmdSn, session->MaxCmdSn)) {
			DEBUG_QUEUE
			    ("iSCSI: session iscsi bus %d target id %d can't "
			     "start %u tasks now, ExpCmdSN %u, CmdSn %u, "
			     "MaxCmdSN %u\n",
			     session->iscsi_bus, session->target_id,
			     atomic_read(&session->num_cmnds),
			     session->ExpCmdSn, session->CmdSn,
			     session->MaxCmdSn);
			if (test_bit
			    (SESSION_WINDOW_CLOSED,
			     &session->control_bits) == 0) {
				/* window is open, but not large enough for us 
				 * to send everything we have queued.
				 * record how many times we hit this situation, 
				 * to see how often we're getting throttled.
				 */
				session->window_full++;
				smp_mb();
			}
			break;
		}

		if (test_bit(SESSION_RESETTING, &session->control_bits)) {
			DEBUG_EH
			    ("iSCSI: session iscsi bus %d target id %d "
			     "resetting, can't start tasks at %lu\n",
			     session->iscsi_bus, session->target_id, jiffies);
			break;
		}

		DEBUG_QUEUE
		    ("iSCSI: session iscsi bus %d target id %d "
		     "xmit_queued_cmnds, CmdSN %u, MaxCmdSN %u\n",
		     session->iscsi_bus, session->target_id, session->CmdSn,
		     session->MaxCmdSn);

		spin_lock(&session->task_lock);

		if (task == NULL) {
			/* allocate a task */
			task = alloc_task(session);
			if (task == NULL) {
				printk
				    ("iSCSI: session iscsi bus %d to "
				     "(%u %u %u *) couldn't allocate task "
				     "at %lu\n",
				     session->iscsi_bus, session->host_no,
				     session->channel, session->target_id,
				     jiffies);
				spin_unlock(&session->task_lock);
				/* to prevent a stall of the driver, 
				 * free_task must wakeup the tx thread later.
				 */
				return;
			}
		}

		/* Don't start any new tasks if a Logout has been requested.  */
		if (test_bit(SESSION_LOGOUT_REQUESTED, &session->control_bits)) {
			spin_unlock(&session->task_lock);
			DEBUG_QUEUE
			    ("iSCSI: session iscsi bus %d target id %d "
			     "logout requested, can't start tasks now\n",
			     session->iscsi_bus, session->target_id);
			break;
		}

		spin_lock_irqsave(&session->scsi_cmnd_lock, noqueue_flags_);

		if ((sc = session->retry_cmnd_head)) {
			/* remove the command from the retry_cmnd queue */
			session->retry_cmnd_head =
			    (Scsi_Cmnd *) sc->host_scribble;
			sc->host_scribble = NULL;
			if (session->retry_cmnd_head == NULL)
				session->retry_cmnd_tail = NULL;

			/* FIXME: we could stop using an atomic counter, if 
			 * we're willing to acquire the session's 
			 * scsi_cmnd_lock every time the TX_SCSI_COMMAND
			 * bit is set.  For now, we use atomic counters so 
			 * that we can skip the lock acquisition if there 
			 * are no commands queued.
			 */
			atomic_dec(&session->num_retry_cmnds);

			/* commands in the retry queue are sent even when 
			 * the LUN is delaying commands, since this is how 
			 * we detect that they no longer need to be delayed.
			 */

			/* if error recovery has started or will start, 
			 * don't start any new tasks 
			 */
			if (test_bit(sc->device->lun, session->luns_timing_out)) {
				/* defer the command until later */
				DEBUG_EH
				    ("iSCSI: session iscsi bus %d deferring "
				     "command 0x%x retry to (%u %u %u %u) "
				     "at %lu\n",
				     session->iscsi_bus, sc->cmnd[0],
				     sc->device->host->host_no, 
				     sc->device->channel, sc->device->id,
				     sc->device->lun, jiffies);

				/* these go back on the head of the deferred 
				 * queue, not the tail, to preserve ordering 
				 * of commands to each LUN.
				 */
				sc->host_scribble =
				    (void *) session->deferred_cmnd_head;
				if (session->deferred_cmnd_head == NULL)
					session->deferred_cmnd_tail = sc;
				session->deferred_cmnd_head = sc;
				session->num_deferred_cmnds++;

				spin_unlock_irqrestore(&session->scsi_cmnd_lock,
						       noqueue_flags_);
				spin_unlock(&session->task_lock);

				/* there may be commands for other LUNs 
				 * that we can send 
				 */
				continue;
			}
		} else if ((sc = session->scsi_cmnd_head)) {
			/* remove the command from the scsi_cmnd queue */
			session->scsi_cmnd_head =
			    (Scsi_Cmnd *) sc->host_scribble;
			sc->host_scribble = NULL;
			if (session->scsi_cmnd_head == NULL)
				session->scsi_cmnd_tail = NULL;

			/* FIXME: we could stop using an atomic counter, 
			 * if we're willing to acquire the session's 
			 * scsi_cmnd_lock every time the TX_SCSI_COMMAND
			 * bit is set.  For now, we use atomic counters so 
			 * that we can skip the lock acquisition if there 
			 * are no commands queued.
			 */
			atomic_dec(&session->num_cmnds);

			/* FIXME: should we check delaying_commands first, 
			 * or timing_out first?  Does it matter? 
			 */
			if (test_bit(sc->device->lun,
				     session->luns_delaying_commands)) {
				/* defer the command until later */
				DEBUG_RETRY
				    ("iSCSI: session iscsi bus %d deferring "
				     "command 0x%x to (%u %u %u %u) at %lu\n",
				     session->iscsi_bus, sc->cmnd[0],
				     sc->device->host->host_no, 
				     sc->device->channel, sc->device->id,
				     sc->device->lun, jiffies);

				/* append it to the tail of the deferred queue 
				 */
				add_cmnd(sc, &session->deferred_cmnd_head,
					 &session->deferred_cmnd_tail);
				session->num_deferred_cmnds++;

				spin_unlock_irqrestore(&session->scsi_cmnd_lock,
						       noqueue_flags_);
				spin_unlock(&session->task_lock);

				/* there may be commands for other LUNs 
				 * that we can send 
				 */
				continue;
			}

			if (test_bit(sc->device->lun,
				     session->luns_timing_out)) {
				/* defer the command until later */
				DEBUG_EH
				    ("iSCSI: session iscsi bus %d deferring "
				     "command 0x%x to (%u %u %u %u) at %lu\n",
				     session->iscsi_bus, sc->cmnd[0],
				     sc->device->host->host_no, 
				     sc->device->channel, sc->device->id,
				     sc->device->lun, jiffies);

				/* append it to the tail of the deferred queue 
				 */
				add_cmnd(sc, &session->deferred_cmnd_head,
					 &session->deferred_cmnd_tail);
				session->num_deferred_cmnds++;

				spin_unlock_irqrestore(&session->scsi_cmnd_lock,
						       noqueue_flags_);
				spin_unlock(&session->task_lock);

				/* there may be commands for other LUNs 
				 * that we can send 
				 */
				continue;
			}
		} else {
			/* this should never happen if the command counts 
			 * are accurate 
			 */
			printk
			    ("iSCSI: bug - no SCSI cmnds queued at %lu for "
			     "session iscsi bus %d target id %d, num_cmnds %u, "
			     "head %p, tail %p, num_retry %u, head %p, "
			     "tail %p\n",
			     jiffies, session->iscsi_bus, session->target_id,
			     atomic_read(&session->num_cmnds),
			     session->scsi_cmnd_head, session->scsi_cmnd_tail,
			     atomic_read(&session->num_retry_cmnds),
			     session->retry_cmnd_head,
			     session->retry_cmnd_tail);

			atomic_set(&session->num_cmnds, 0);
			atomic_set(&session->num_retry_cmnds, 0);

			spin_unlock_irqrestore(&session->scsi_cmnd_lock,
					       noqueue_flags_);
			spin_unlock(&session->task_lock);
			break;
		}

		/* prepare to start a new task */
		__set_bit(TASK_TRY_ABORT, &task->flags);
		task->lun = sc->device->lun;
		task->scsi_cmnd = sc;
		iscsi_set_direction(task);
		add_session_task(session, task);

		DEBUG_QUEUE
		    ("iSCSI: cmnd 0x%x became task itt %u at %lu for session "
		     "iscsi bus %d target id %d, num_cmnds %u, head %p, "
		     "tail %p\n",
		     sc->cmnd[0], task->itt, jiffies, session->iscsi_bus,
		     session->target_id, atomic_read(&session->num_cmnds),
		     session->scsi_cmnd_head, session->scsi_cmnd_tail);

		spin_unlock_irqrestore(&session->scsi_cmnd_lock, 
				       noqueue_flags_);

		if ((session->fake_status_lun >= -1) &&
		    ((session->fake_status_lun == -1)
		     || (session->fake_status_lun == sc->device->lun))) {
			if (fake_task_completion(session, task)) {
				/* the task has been completed, and we've 
				 * lost the task_lock 
				 */
				task = NULL;
				continue;
			} else {
				/* nothing left to fake, still have the 
				 * task_lock 
				 */
				session->fake_status_lun = -2;
			}
		}

		/* start a timer, queue up any unsolicited data, 
		 * and send the task 
		 */
		add_task_timer(task, iscsi_task_times_out);
		atomic_inc(&task->refcount);
		smp_mb();

		/* possibly queue up unsolicited data PDUs.
		 * With ImmediateData, we may or may not have to send
		 * additional Data PDUs, depending on the amount of data, and
		 * the Max PDU Length.  For now, we never use immediate data.
		 */
		if (test_bit(TASK_WRITE, &task->flags) && sc->request_buffer
		    && sc->request_bufflen && iscsi_expected_data_length(sc)) {

			/* queue up unsolicited data PDUs.  the implied 
			 * initial R2T doesn't count against the 
			 * MaxOutstandingR2T, so we can't use the normal R2T 
			 * fields of the task for the implied initial R2T.  
			 * Use a special flag for the implied initial R2T, and 
			 * let the rx thread update tasks in the tx_tasks 
			 * collection if an R2T comes in before the implied 
			 * initial R2T has been processed.
			 */
			if (session->ImmediateData) {
				imm_data_length =
				    session->MaxXmitDataSegmentLength;
				imm_data_length =
				    MIN(imm_data_length, sc->request_bufflen);
				if (sc->request_bufflen > imm_data_length) {
					if (!session->InitialR2T) {
						__set_bit(TASK_INITIAL_R2T,
							  &task->flags);
						/* queue up an implied R2T data 
						 * transfer for later 
						 */
						add_task(&session->tx_task_head,
							 task);
						set_bit(TX_DATA,
							&session->control_bits);
						set_bit(TX_WAKE,
							&session->control_bits);
					}
				}
			} else {
				if (!session->InitialR2T) {
					__set_bit(TASK_INITIAL_R2T,
						  &task->flags);
					/* queue up an implied R2T data 
					 * transfer for later 
					 */
					add_task(&session->tx_task_head, task);
					set_bit(TX_DATA,
						&session->control_bits);
					set_bit(TX_WAKE,
						&session->control_bits);
				}
			}
		}

		spin_unlock(&session->task_lock);

		DEBUG_FLOW
		    ("iSCSI: sending itt %u on session iscsi bus %d "
		     "target id %d as CmdSN %u, MaxCmdSn %u\n",
		     task->itt, session->iscsi_bus, session->target_id,
		     session->CmdSn, session->MaxCmdSn);

		/* we don't bother to check if the xmit works, since if it
		 * fails, the session will drop, and all tasks and cmnds
		 * will be completed by the drop.
		 */
		iscsi_xmit_task(task);

#if INVALID_ORDERING_ASSUMPTIONS
		/* some broken targets choke if a command PDU is followed 
		 * by anything other than the data for that command, but 
		 * still advertise a CmdSN window of more than 1 command.  
		 * For such broken targets, we stop the loop after the first 
		 * write, to try and give them the data they want.  The 
		 * target may still choke if it doesn't get us an R2T before 
		 * we send the next command.  Nothing we can do about that, 
		 * other than log bugs against the broken targets.
		 */
		if (test_bit(TASK_WRITE, &task->flags)) {
			atomic_dec(&task->refcount);
			set_bit(TX_SCSI_COMMAND, &session->control_bits);
			set_bit(TX_WAKE, &session->control_bits);
			return;
		}
#endif

		atomic_dec(&task->refcount);
		DEBUG_FLOW
		    ("iSCSI: after sending itt %u, now has refcount %d\n",
		     task->itt, atomic_read(&task->refcount));
		task = NULL;
	}

	/* we still have a task we never used.  free it before returning */
	if (task) {
		spin_lock(&session->task_lock);
		free_task(session, task);
		spin_unlock(&session->task_lock);
	}
}

#ifndef USE_LOCAL_CRC32C

void
iscsi_xmit_data(struct iscsi_task *task, uint32_t ttt, uint32_t data_offset,
		uint32_t data_length)
{
	struct msghdr msg;
	struct IscsiDataHdr stdh;
	Scsi_Cmnd *sc = NULL;
	struct iscsi_session *session = task->session;
	struct scatterlist *sglist = NULL, *sg, *first_sg = NULL, *last_sg =
	    NULL;
	int wlen, rc, iovn = 0, first_data_iovn = 0;
	unsigned int segment_offset = 0, index = 0;
	int remain, xfrlen;
	uint32_t data_sn = 0;
	int bytes_to_fill, bytes_from_segment;
	char padding[4];
	int pad_bytes;
	uint32_t header_crc32c;
	uint32_t data_crc32c;
	struct crypto_tfm *tfm = NULL;
	struct scatterlist tmpsg[2];

	sc = task->scsi_cmnd;
	/* make sure we have data to send when we expect to */
	if (sc && (iscsi_expected_data_length(sc) == 0)
	    && ((sc->request_bufflen == 0) || (sc->request_buffer == NULL))) {
		printk
		    ("iSCSI: xmit_data for itt %u, sc 0x%x, dlength %u, "
		     "expected %u, no data in buffer\n"
		     "       request_buffer %p len %u, buffer %p len %u\n",
		     task->itt, sc->cmnd[0], data_length,
		     iscsi_expected_data_length(sc), sc->request_buffer,
		     sc->request_bufflen, sc->buffer, sc->bufflen);
		print_cmnd(sc);
		return;
	}

	remain = data_length;
	if (sc == NULL)
		remain = 0;

	memset(&stdh, 0, sizeof (stdh));
	stdh.opcode = ISCSI_OP_SCSI_DATA;
	stdh.itt = htonl(task->itt);
	stdh.ttt = ttt;
	stdh.offset = htonl(data_offset);

	/* PDU header */
	session->tx_iov[0].iov_base = &stdh;
	session->tx_iov[0].iov_len = sizeof (stdh);

	DEBUG_FLOW("iSCSI: xmit_data for itt %u, credit %d @ %u\n"
		   "       request_buffer %p len %u, buffer %p len %u\n",
		   task->itt, remain, data_offset,
		   sc->request_buffer, sc->request_bufflen, sc->buffer,
		   sc->bufflen);

	/* Find the segment and offset within the segment to 
	 * start writing from.  
	 */
	if (sc && sc->use_sg) {
		sg = sglist = (struct scatterlist *) sc->request_buffer;

		segment_offset = data_offset;

		for (index = 0; index < sc->use_sg; index++) {
			if (segment_offset < sglist[index].length)
				break;
			else
				segment_offset -= sglist[index].length;
		}

		if (index >= sc->use_sg) {
			/* didn't find the offset, command will 
			 * eventually timeout 
			 */
			printk
			    ("iSCSI: session iscsi bus %d target id %d "
			     "xmit_data for itt %u couldn't find offset %u "
			     "in sglist %p, sc %p, bufflen %u, use_sg %u\n",
			     session->iscsi_bus, session->target_id, task->itt,
			     data_offset, sglist, sc, sc->request_bufflen,
			     sc->use_sg);
			print_cmnd(sc);
			ISCSI_TRACE(ISCSI_TRACE_OutOfData, sc, task, index,
				    sc->use_sg);
			return;
		}
	}

	ISCSI_TRACE(ISCSI_TRACE_TxData, sc, task, data_offset, data_length);

	do {
		if (signal_pending(current))
			break;

#if (INVALID_ORDERING_ASSUMPTIONS == 0)
		/* since this loop may take a while, check 
		 * for TIMEDOUT tasks and commands 
		 */
		/* Note: this means a task may have a non-zero 
		 * refcount during timeout processing 
		 */
		if (test_bit(SESSION_TASK_TIMEDOUT, &session->control_bits)) {
			process_timedout_tasks(session);
		}
		if (test_bit(SESSION_COMMAND_TIMEDOUT, &session->control_bits)) {
			process_timedout_commands(session);
		}

		/* also queue up command retries */
		if (test_and_clear_bit
		    (SESSION_RETRY_COMMANDS, &session->control_bits)) {
			/* try to queue up delayed commands for retries */
			iscsi_retry_commands(session);
		}

		/* if command PDUs are small (no immediate data),
		 * start commands as soon as possible, so that we can
		 * overlap the R2T latency with the time it takes to
		 * send data for commands already issued.  This increases 
		 * throughput without significantly increasing the completion 
		 * time of commands already issued.  Some broken targets
		 * such as the one by Intel Labs will choke if they receive
		 * another command before they get all of the data for preceding
		 * commands, so this can be conditionally compiled out.
		 */
		if (!session->ImmediateData) {
			DEBUG_FLOW
			    ("iSCSI: checking for new commands before "
			     "sending data to %s\n",
			     session->log_name);
			iscsi_xmit_queued_cmnds(session);
		}
#endif

		iovn = 1;
		wlen = sizeof (stdh);
		if (session->HeaderDigest == ISCSI_DIGEST_CRC32C) {
			/* we'll need to send a digest, 
			 * but can't compute it yet 
			 */
			session->tx_iov[1].iov_base = &header_crc32c;
			session->tx_iov[1].iov_len = sizeof (header_crc32c);
			iovn = 2;
			wlen += sizeof (header_crc32c);
		}

		first_data_iovn = iovn;

		stdh.datasn = htonl(data_sn++);
		stdh.offset = htonl(data_offset);
		stdh.expstatsn = htonl(session->ExpStatSn);

		if (session->MaxXmitDataSegmentLength
		    && (remain > session->MaxXmitDataSegmentLength)) {
			/* enforce the target's data segment limit */
			bytes_to_fill = session->MaxXmitDataSegmentLength;
		} else {
			/* final PDU of a data burst */
			bytes_to_fill = remain;
			stdh.flags = ISCSI_FLAG_FINAL;
		}

		/* check if we need to pad the PDU */
		if (bytes_to_fill % PAD_WORD_LEN) {
			pad_bytes =
			    PAD_WORD_LEN - (bytes_to_fill % PAD_WORD_LEN);
			memset(padding, 0x0, sizeof (padding));
		} else {
			pad_bytes = 0;
		}

		DEBUG_FLOW
		    ("iSCSI: remain %d, bytes_to_fill %d, "
		     "sc->use_sg %u, MaxRecvDataSegmentLength %d\n",
		     remain, bytes_to_fill, sc->use_sg,
		     session->MaxRecvDataSegmentLength);

		xfrlen = 0;

		if (sc) {
			/* find all the PDU data */
			if (sc->use_sg) {
				/*
				 * Calculate the CRC right here before 
				 * any kmapping happens.
				 */
				if (session->DataDigest == 
				    ISCSI_DIGEST_CRC32C) {
					tfm = crypto_alloc_tfm("crc32c", 0);
					crypto_chksum_init(tfm);
				}

				/* while there is more data and 
				 * we want to send more data 
				 */
				while (bytes_to_fill > 0) {

					if (index >= sc->use_sg) {
						printk
						    ("iSCSI: session iscsi bus "
						     "%d target id %d xmit_data"
						     " index %d exceeds "
						     "sc->use_sg %d, "
						     "bytes_to_fill %d, "
						     "out of buffers\n",
						     session->iscsi_bus,
						     session->target_id, index,
						     sc->use_sg, bytes_to_fill);
						/* the command will eventually 
						 * timeout 
						 */
						print_cmnd(sc);
						ISCSI_TRACE
						    (ISCSI_TRACE_OutOfData, sc,
						     task, index, sc->use_sg);
						goto done;
					}
					if (signal_pending(current)) {
						DEBUG_FLOW
						    ("iSCSI: session iscsi bus "
						     "%d target id %d signal "
						     "pending, returning from "
						     "xmit_data\n",
						     session->iscsi_bus,
						     session->target_id);
						goto done;
					}

					sg = &sglist[index];

					bytes_from_segment =
					    sg->length - segment_offset;

					if (tfm) {
						/*
						 * Update chksum here
						 */
						tmpsg[0].page =
						    virt_to_page
						    (sg_virtual_address(sg) +
						     segment_offset);
						tmpsg[0].offset =
						    offset_in_page
						    (sg_virtual_address(sg) +
						     segment_offset);
						tmpsg[0].length =
						    (bytes_from_segment >
						     bytes_to_fill ?
						     bytes_to_fill :
						     bytes_from_segment);
						crypto_chksum_update(tfm, tmpsg,
								     1);
					}
					
					/* make sure the segment is mapped */
					if (!kmap_sg(sg)) {
						printk
						    ("iSCSI: session iscsi bus "
						     "%d target id %d xmit_data"
						     " couldn't map segment "
						     "%p\n",
						     session->iscsi_bus,
						     session->target_id, sg);
						goto done;
					} else if (first_sg == NULL) {
						first_sg = sg;
					}
					last_sg = sg;

					/* sanity check the sglist 
					 * segment length 
					 */
					if (sg->length <= segment_offset) {
						/* the sglist is corrupt */
						printk
						    ("iSCSI: session iscsi bus "
						     "%d target id %d xmit_data"
						     " index %d, length %u too "
						     "small for offset %u, "
						     "bytes_to_fill %d, sglist "
						     "has been corrupted\n",
						     session->iscsi_bus,
						     session->target_id, index,
						     sg->length, segment_offset,
						     bytes_to_fill);
						/* the command will eventually 
						 * timeout 
						 */
						print_cmnd(sc);
						ISCSI_TRACE
						    (ISCSI_TRACE_BadTxSeg, sc,
						     task, sg->length,
						     segment_offset);
						goto done;
					}

					if (bytes_from_segment > bytes_to_fill) {
						/* only need part of 
						 * this segment 
						 */
						session->tx_iov[iovn].iov_base =
						    sg_virtual_address(sg) +
						    segment_offset;
						session->tx_iov[iovn].iov_len =
						    bytes_to_fill;
						xfrlen += bytes_to_fill;
						DEBUG_FLOW
						    ("iSCSI: session iscsi bus "
						     "%d target id %d xmit_data"
						     " xfrlen %d, to_fill %d, "
						     "from_segment %d, iov[%2d]"
						     " = partial sg[%2d]\n",
						     session->iscsi_bus,
						     session->target_id, xfrlen,
						     bytes_to_fill,
						     bytes_from_segment, iovn,
						     index);
						iovn++;
						segment_offset += bytes_to_fill;
						break;
					} else {
						/* need all of this segment, and
						 * possibly more from the next 
						 */
						session->tx_iov[iovn].iov_base =
						    sg_virtual_address(sg) +
						    segment_offset;
						session->tx_iov[iovn].iov_len =
						    bytes_from_segment;
						xfrlen += bytes_from_segment;
						DEBUG_FLOW
						    ("iSCSI: session iscsi bus "
						     "%d target id %d xmit_data"
						     " xfrlen %d, to_fill %d, "
						     "from_segment %d, iov[%2d]"
						     " = sg[%2d]\n",
						     session->iscsi_bus,
						     session->target_id, xfrlen,
						     bytes_to_fill,
						     bytes_from_segment, iovn,
						     index);
						bytes_to_fill -=
						    bytes_from_segment;
						iovn++;
						/* any remaining data starts at
 						 * offset 0 of the next segment
						 */
						index++;
						segment_offset = 0;
					}
				}

				if (xfrlen <= 0) {
					printk
					    ("iSCSI: session iscsi bus %d "
					     "target id %d xmit_data picked "
					     "xfrlen of 0, sc->use_sg %d, "
					     "bytes_to_fill %d\n",
					     session->iscsi_bus,
					     session->target_id, sc->use_sg,
					     bytes_to_fill);
					iscsi_drop_session(session);
					goto done;
				}
				if (sc->use_sg && tfm) {
					crypto_chksum_final(tfm, &data_crc32c);
					crypto_free_tfm(tfm);
					tfm = NULL;
				}
			} else {
				/* no scatter-gather */
				if ((sc->request_buffer + data_offset +
				     bytes_to_fill) <=
				    (sc->request_buffer +
				     sc->request_bufflen)) {
					/* send all the data */
					session->tx_iov[iovn].iov_base =
					    sc->request_buffer + data_offset;
					session->tx_iov[iovn].iov_len = xfrlen =
					    bytes_to_fill;
					iovn++;
				} else if ((sc->request_buffer + data_offset) <
					   (sc->request_buffer +
					    sc->request_bufflen)) {
					/* send some data, but can't send all 
					 * requested 
					 */
					xfrlen =
					    sc->request_bufflen - data_offset;
					printk
					    ("iSCSI: xmit_data ran out of data,"
					     " buffer %p len %u but offset %d "
					     "length %d, sending final %d "
					     "bytes\n",
					     sc->request_buffer,
					     sc->request_bufflen, data_offset,
					     bytes_to_fill, xfrlen);
					session->tx_iov[iovn].iov_base =
					    sc->request_buffer + data_offset;
					session->tx_iov[iovn].iov_len = xfrlen;
					iovn++;
					stdh.flags = ISCSI_FLAG_FINAL;
					remain = xfrlen;
				} else {
					/* can't send any data */
					printk
					    ("iSCSI: xmit_data ran out of data,"
					     " buffer %p len %u but offset %d "
					     "length %d, sending no more "
					     "data\n",
					     sc->request_buffer,
					     sc->request_bufflen, data_offset,
					     bytes_to_fill);
					goto done;
				}
			}

			if (pad_bytes) {
				session->tx_iov[iovn].iov_base = padding;
				session->tx_iov[iovn].iov_len = pad_bytes;
				iovn++;
				wlen += pad_bytes;

				if (first_sg) {
					/* add padding bytes to CRC computed 
					 * above
					 */
					data_crc32c =
					    iscsi_crc32c_continued(&padding,
								   pad_bytes,
								   data_crc32c);
				}
			}
		}

		/* put the data length in the PDU header */
		hton24(stdh.dlength, xfrlen);
		wlen += xfrlen;

		/* header complete, we can finally calculate the HeaderDigest */
		if (session->HeaderDigest == ISCSI_DIGEST_CRC32C)
			header_crc32c = iscsi_crc32c(&stdh, sizeof (stdh));

		/* DataDigest */
		if (xfrlen && (session->DataDigest == ISCSI_DIGEST_CRC32C)) {
			int i;


			if (!first_sg) {
				/*
				 * No scatter-gather, so compute CRC right here.
				 */
				data_crc32c =
				    iscsi_crc32c(session->
						 tx_iov[first_data_iovn].
						 iov_base,
						 session->
						 tx_iov[first_data_iovn].
						 iov_len);
				for (i = first_data_iovn + 1; i < iovn; i++) {
					data_crc32c =
					    iscsi_crc32c_continued(session->
								   tx_iov[i].
								   iov_base,
								   session->
								   tx_iov[i].
								   iov_len,
								   data_crc32c);
				}
			}
			
			/* FIXME: this may not be SMP safe, but it's only for 
			 * testing anyway, so it probably doesn't need to be 
			 */
			if (session->fake_write_data_mismatch > 0) {
				session->fake_write_data_mismatch--;
				smp_mb();
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "faking DataDigest mismatch for itt %u\n",
				     session->iscsi_bus, session->target_id,
				     task->itt);
				data_crc32c = 0x01020304;
			}

			session->tx_iov[iovn].iov_base = &data_crc32c;
			session->tx_iov[iovn].iov_len = sizeof (data_crc32c);
			iovn++;
			wlen += sizeof (data_crc32c);
		}

		memset(&msg, 0, sizeof (msg));
		msg.msg_iov = &session->tx_iov[0];
		msg.msg_iovlen = iovn;
		
		ISCSI_TRACE(ISCSI_TRACE_TxDataPDU, sc, task,
			    data_offset, xfrlen);
		
		rc = iscsi_sendmsg(session, &msg, wlen);
		if (rc != wlen) {
			printk("iSCSI: session iscsi_bus %d target id %d "
			       "xmit_data failed to send "
			       "%d bytes, rc %d\n",
			       session->iscsi_bus, session->target_id,
			       wlen, rc);
			iscsi_drop_session(session);
			goto done;
		}
		
		remain -= xfrlen;

		DEBUG_FLOW
		    ("iSCSI: xmit_data sent %d @ %u for itt %u, "
		     "remaining %d, final %d\n",
		     xfrlen, data_offset, task->itt, remain,
		     stdh.flags & ISCSI_FLAG_FINAL);

		data_offset += xfrlen;
		
		if (first_sg) {
			/* undo any temporary mappings */
			for (sg = first_sg; sg <= last_sg; sg++) {
				kunmap_sg(sg);
			}
			first_sg = last_sg = NULL;
		}
		
	} while (remain);
	
      done:
	if (first_sg) {
		/* undo any temporary mappings */
		for (sg = first_sg; sg <= last_sg; sg++) {
			kunmap_sg(sg);
		}
	}
	if (tfm)
		crypto_free_tfm(tfm);
}

#else

void
iscsi_xmit_data(struct iscsi_task *task, uint32_t ttt, uint32_t data_offset,
		uint32_t data_length)
{
	struct msghdr msg;
	struct IscsiDataHdr stdh;
	Scsi_Cmnd *sc = NULL;
	struct iscsi_session *session = task->session;
	struct scatterlist *sglist = NULL, *sg, *first_sg = NULL, *last_sg =
	    NULL;
	int wlen, rc, iovn = 0, first_data_iovn = 0;
	unsigned int segment_offset = 0, index = 0;
	int remain, xfrlen;
	uint32_t data_sn = 0;
	int bytes_to_fill, bytes_from_segment;
	char padding[4];
	int pad_bytes;
	uint32_t header_crc32c;
	uint32_t data_crc32c;

	sc = task->scsi_cmnd;
	/* make sure we have data to send when we expect to */
	if (sc && (iscsi_expected_data_length(sc) == 0)
	    && ((sc->request_bufflen == 0) || (sc->request_buffer == NULL))) {
		printk
		    ("iSCSI: xmit_data for itt %u, sc 0x%x, dlength %u, "
		     "expected %u, no data in buffer\n"
		     "       request_buffer %p len %u, buffer %p len %u\n",
		     task->itt, sc->cmnd[0], data_length,
		     iscsi_expected_data_length(sc), sc->request_buffer,
		     sc->request_bufflen, sc->buffer, sc->bufflen);
		print_cmnd(sc);
		return;
	}

	remain = data_length;
	if (sc == NULL)
		remain = 0;

	memset(&stdh, 0, sizeof (stdh));
	stdh.opcode = ISCSI_OP_SCSI_DATA;
	stdh.itt = htonl(task->itt);
	stdh.ttt = ttt;
	stdh.offset = htonl(data_offset);

	/* PDU header */
	session->tx_iov[0].iov_base = &stdh;
	session->tx_iov[0].iov_len = sizeof (stdh);

	DEBUG_FLOW("iSCSI: xmit_data for itt %u, credit %d @ %u\n"
		   "       request_buffer %p len %u, buffer %p len %u\n",
		   task->itt, remain, data_offset,
		   sc->request_buffer, sc->request_bufflen, sc->buffer,
		   sc->bufflen);

	/* Find the segment and offset within the segment to 
	 * start writing from.  
	 */
	if (sc && sc->use_sg) {
		sg = sglist = (struct scatterlist *) sc->request_buffer;

		segment_offset = data_offset;

		for (index = 0; index < sc->use_sg; index++) {
			if (segment_offset < sglist[index].length)
				break;
			else
				segment_offset -= sglist[index].length;
		}

		if (index >= sc->use_sg) {
			/* didn't find the offset, command will 
			 * eventually timeout 
			 */
			printk
			    ("iSCSI: session iscsi bus %d target id %d "
			     "xmit_data for itt %u couldn't find offset %u "
			     "in sglist %p, sc %p, bufflen %u, use_sg %u\n",
			     session->iscsi_bus, session->target_id, task->itt,
			     data_offset, sglist, sc, sc->request_bufflen,
			     sc->use_sg);
			print_cmnd(sc);
			ISCSI_TRACE(ISCSI_TRACE_OutOfData, sc, task, index,
				    sc->use_sg);
			return;
		}
	}

	ISCSI_TRACE(ISCSI_TRACE_TxData, sc, task, data_offset, data_length);

	do {
		if (signal_pending(current))
			break;

#if (INVALID_ORDERING_ASSUMPTIONS == 0)
		/* since this loop may take a while, check 
		 * for TIMEDOUT tasks and commands 
		 */
		/* Note: this means a task may have a non-zero 
		 * refcount during timeout processing 
		 */
		if (test_bit(SESSION_TASK_TIMEDOUT, &session->control_bits)) {
			process_timedout_tasks(session);
		}
		if (test_bit(SESSION_COMMAND_TIMEDOUT, &session->control_bits)) {
			process_timedout_commands(session);
		}

		/* also queue up command retries */
		if (test_and_clear_bit
		    (SESSION_RETRY_COMMANDS, &session->control_bits)) {
			/* try to queue up delayed commands for retries */
			iscsi_retry_commands(session);
		}

		/* if command PDUs are small (no immediate data),
		 * start commands as soon as possible, so that we can
		 * overlap the R2T latency with the time it takes to
		 * send data for commands already issued.  This increases 
		 * throughput without significantly increasing the completion 
		 * time of commands already issued.  Some broken targets
		 * such as the one by Intel Labs will choke if they receive
		 * another command before they get all of the data for preceding
		 * commands, so this can be conditionally compiled out.
		 */
		if (!session->ImmediateData) {
			DEBUG_FLOW
			    ("iSCSI: checking for new commands before "
			     "sending data to %s\n",
			     session->log_name);
			iscsi_xmit_queued_cmnds(session);
		}
#endif

		iovn = 1;
		wlen = sizeof (stdh);
		if (session->HeaderDigest == ISCSI_DIGEST_CRC32C) {
			/* we'll need to send a digest, 
			 * but can't compute it yet 
			 */
			session->tx_iov[1].iov_base = &header_crc32c;
			session->tx_iov[1].iov_len = sizeof (header_crc32c);
			iovn = 2;
			wlen += sizeof (header_crc32c);
		}

		first_data_iovn = iovn;

		stdh.datasn = htonl(data_sn++);
		stdh.offset = htonl(data_offset);
		stdh.expstatsn = htonl(session->ExpStatSn);

		if (session->MaxXmitDataSegmentLength
		    && (remain > session->MaxXmitDataSegmentLength)) {
			/* enforce the target's data segment limit */
			bytes_to_fill = session->MaxXmitDataSegmentLength;
		} else {
			/* final PDU of a data burst */
			bytes_to_fill = remain;
			stdh.flags = ISCSI_FLAG_FINAL;
		}

		/* check if we need to pad the PDU */
		if (bytes_to_fill % PAD_WORD_LEN) {
			pad_bytes =
			    PAD_WORD_LEN - (bytes_to_fill % PAD_WORD_LEN);
			memset(padding, 0x0, sizeof (padding));
		} else {
			pad_bytes = 0;
		}

		DEBUG_FLOW
		    ("iSCSI: remain %d, bytes_to_fill %d, "
		     "sc->use_sg %u, MaxRecvDataSegmentLength %d\n",
		     remain, bytes_to_fill, sc->use_sg,
		     session->MaxRecvDataSegmentLength);

		xfrlen = 0;

		if (sc) {
			/* find all the PDU data */
			if (sc->use_sg) {
				/* while there is more data and 
				 * we want to send more data 
				 */
				while (bytes_to_fill > 0) {

					if (index >= sc->use_sg) {
						printk
						    ("iSCSI: session iscsi bus "
						     "%d target id %d xmit_data"
						     " index %d exceeds "
						     "sc->use_sg %d, "
						     "bytes_to_fill %d, "
						     "out of buffers\n",
						     session->iscsi_bus,
						     session->target_id, index,
						     sc->use_sg, bytes_to_fill);
						/* the command will eventually 
						 * timeout 
						 */
						print_cmnd(sc);
						ISCSI_TRACE
						    (ISCSI_TRACE_OutOfData, sc,
						     task, index, sc->use_sg);
						goto done;
					}
					if (signal_pending(current)) {
						DEBUG_FLOW
						    ("iSCSI: session iscsi bus "
						     "%d target id %d signal "
						     "pending, returning from "
						     "xmit_data\n",
						     session->iscsi_bus,
						     session->target_id);
						goto done;
					}

					sg = &sglist[index];

					/* make sure the segment is mapped */
					if (!kmap_sg(sg)) {
						printk
						    ("iSCSI: session iscsi bus "
						     "%d target id %d xmit_data"
						     " couldn't map segment "
						     "%p\n",
						     session->iscsi_bus,
						     session->target_id, sg);
						goto done;
					} else if (first_sg == NULL) {
						first_sg = sg;
					}
					last_sg = sg;

					/* sanity check the sglist 
					 * segment length 
					 */
					if (sg->length <= segment_offset) {
						/* the sglist is corrupt */
						printk
						    ("iSCSI: session iscsi bus "
						     "%d target id %d xmit_data"
						     " index %d, length %u too "
						     "small for offset %u, "
						     "bytes_to_fill %d, sglist "
						     "has been corrupted\n",
						     session->iscsi_bus,
						     session->target_id, index,
						     sg->length, segment_offset,
						     bytes_to_fill);
						/* the command will eventually 
						 * timeout 
						 */
						print_cmnd(sc);
						ISCSI_TRACE
						    (ISCSI_TRACE_BadTxSeg, sc,
						     task, sg->length,
						     segment_offset);
						goto done;
					}

					bytes_from_segment =
					    sg->length - segment_offset;
					if (bytes_from_segment > bytes_to_fill) {
						/* only need part of 
						 * this segment 
						 */
						session->tx_iov[iovn].iov_base =
						    sg_virtual_address(sg) +
						    segment_offset;
						session->tx_iov[iovn].iov_len =
						    bytes_to_fill;
						xfrlen += bytes_to_fill;
						DEBUG_FLOW
						    ("iSCSI: session iscsi bus "
						     "%d target id %d xmit_data"
						     " xfrlen %d, to_fill %d, "
						     "from_segment %d, iov[%2d]"
						     " = partial sg[%2d]\n",
						     session->iscsi_bus,
						     session->target_id, xfrlen,
						     bytes_to_fill,
						     bytes_from_segment, iovn,
						     index);
						iovn++;
						segment_offset += bytes_to_fill;
						break;
					} else {
						/* need all of this segment, and
						 * possibly more from the next 
						 */
						session->tx_iov[iovn].iov_base =
						    sg_virtual_address(sg) +
						    segment_offset;
						session->tx_iov[iovn].iov_len =
						    bytes_from_segment;
						xfrlen += bytes_from_segment;
						DEBUG_FLOW
						    ("iSCSI: session iscsi bus "
						     "%d target id %d xmit_data"
						     " xfrlen %d, to_fill %d, "
						     "from_segment %d, iov[%2d]"
						     " = sg[%2d]\n",
						     session->iscsi_bus,
						     session->target_id, xfrlen,
						     bytes_to_fill,
						     bytes_from_segment, iovn,
						     index);
						bytes_to_fill -=
						    bytes_from_segment;
						iovn++;
						/* any remaining data starts at 						 * offset 0 of the next segment
						 */
						index++;
						segment_offset = 0;
					}
				}

				if (xfrlen <= 0) {
					printk
					    ("iSCSI: session iscsi bus %d "
					     "target id %d xmit_data picked "
					     "xfrlen of 0, sc->use_sg %d, "
					     "bytes_to_fill %d\n",
					     session->iscsi_bus,
					     session->target_id, sc->use_sg,
					     bytes_to_fill);
					iscsi_drop_session(session);
					goto done;
				}
			} else {
				/* no scatter-gather */
				if ((sc->request_buffer + data_offset +
				     bytes_to_fill) <=
				    (sc->request_buffer +
				     sc->request_bufflen)) {
					/* send all the data */
					session->tx_iov[iovn].iov_base =
					    sc->request_buffer + data_offset;
					session->tx_iov[iovn].iov_len = xfrlen =
					    bytes_to_fill;
					iovn++;
				} else if ((sc->request_buffer + data_offset) <
					   (sc->request_buffer +
					    sc->request_bufflen)) {
					/* send some data, but can't send all 
					 * requested 
					 */
					xfrlen =
					    sc->request_bufflen - data_offset;
					printk
					    ("iSCSI: xmit_data ran out of data,"
					     " buffer %p len %u but offset %d "
					     "length %d, sending final %d "
					     "bytes\n",
					     sc->request_buffer,
					     sc->request_bufflen, data_offset,
					     bytes_to_fill, xfrlen);
					session->tx_iov[iovn].iov_base =
					    sc->request_buffer + data_offset;
					session->tx_iov[iovn].iov_len = xfrlen;
					iovn++;
					stdh.flags = ISCSI_FLAG_FINAL;
					remain = xfrlen;
				} else {
					/* can't send any data */
					printk
					    ("iSCSI: xmit_data ran out of data,"
					     " buffer %p len %u but offset %d "
					     "length %d, sending no more "
					     "data\n",
					     sc->request_buffer,
					     sc->request_bufflen, data_offset,
					     bytes_to_fill);
					goto done;
				}
			}

			if (pad_bytes) {
				session->tx_iov[iovn].iov_base = padding;
				session->tx_iov[iovn].iov_len = pad_bytes;
				iovn++;
				wlen += pad_bytes;
			}
		}

		/* put the data length in the PDU header */
		hton24(stdh.dlength, xfrlen);
		wlen += xfrlen;

		/* header complete, we can finally calculate the HeaderDigest */
		if (session->HeaderDigest == ISCSI_DIGEST_CRC32C)
			header_crc32c = iscsi_crc32c(&stdh, sizeof (stdh));

		/* DataDigest */
		if (xfrlen && (session->DataDigest == ISCSI_DIGEST_CRC32C)) {
			int i;

			data_crc32c =
			    iscsi_crc32c(session->tx_iov[first_data_iovn].
					 iov_base,
					 session->tx_iov[first_data_iovn].
					 iov_len);
			for (i = first_data_iovn + 1; i < iovn; i++) {
				data_crc32c =
				    iscsi_crc32c_continued(session->tx_iov[i].
							   iov_base,
							   session->tx_iov[i].
							   iov_len,
							   data_crc32c);
			}

			/* FIXME: this may not be SMP safe, but it's only for 
			 * testing anyway, so it probably doesn't need to be 
			 */
			if (session->fake_write_data_mismatch > 0) {
				session->fake_write_data_mismatch--;
				smp_mb();
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "faking DataDigest mismatch for itt %u\n",
				     session->iscsi_bus, session->target_id,
				     task->itt);
				data_crc32c = 0x01020304;
			}

			session->tx_iov[iovn].iov_base = &data_crc32c;
			session->tx_iov[iovn].iov_len = sizeof (data_crc32c);
			iovn++;
			wlen += sizeof (data_crc32c);
		}

		memset(&msg, 0, sizeof (msg));
		msg.msg_iov = &session->tx_iov[0];
		msg.msg_iovlen = iovn;
		
		ISCSI_TRACE(ISCSI_TRACE_TxDataPDU, sc, task,
			    data_offset, xfrlen);
		
		rc = iscsi_sendmsg(session, &msg, wlen);
		if (rc != wlen) {
			printk("iSCSI: session iscsi_bus %d target id %d "
			       "xmit_data failed to send "
			       "%d bytes, rc %d\n",
			       session->iscsi_bus, session->target_id,
			       wlen, rc);
			iscsi_drop_session(session);
			goto done;
		}
		
		remain -= xfrlen;

		DEBUG_FLOW
		    ("iSCSI: xmit_data sent %d @ %u for itt %u, "
		     "remaining %d, final %d\n",
		     xfrlen, data_offset, task->itt, remain,
		     stdh.flags & ISCSI_FLAG_FINAL);

		data_offset += xfrlen;
		
		if (first_sg) {
			/* undo any temporary mappings */
			for (sg = first_sg; sg <= last_sg; sg++) {
				kunmap_sg(sg);
			}
			first_sg = last_sg = NULL;
		}
		
	} while (remain);
	
      done:
	if (first_sg) {
		/* undo any temporary mappings */
		for (sg = first_sg; sg <= last_sg; sg++) {
			kunmap_sg(sg);
		}
	}
}

#endif

void
iscsi_xmit_r2t_data(struct iscsi_session *session)
{
	struct iscsi_task *task;
	uint32_t itt;
	uint32_t ttt;
	uint32_t offset;
	uint32_t length;
	int initial_r2t = 0;
	uint32_t implied_length = 0;
	uint32_t imm_data_length = 0;

	spin_lock(&session->task_lock);
	while ((task = pop_task(&session->tx_task_head))) {
		itt = task->itt;

		if ((initial_r2t =
		     __test_and_clear_bit(TASK_INITIAL_R2T, &task->flags))) {
			if (session->FirstBurstLength)
				implied_length =
				    MIN(session->FirstBurstLength,
					(int)iscsi_expected_data_length(task->
								   scsi_cmnd));
			else
				implied_length = 
				    iscsi_expected_data_length(task->scsi_cmnd);
					/* FirstBurstLength 0 means no limit */

			/* For ImmediateData, we'll have to subtract it 
			 * off as well 
			 */
			if (session->ImmediateData) {
				imm_data_length =
				    session->MaxXmitDataSegmentLength;
				imm_data_length =
				    MIN((int)imm_data_length,
					session->FirstBurstLength);
				imm_data_length =
				    MIN(imm_data_length,
					iscsi_expected_data_length(task->
								   scsi_cmnd));
				implied_length -= imm_data_length;
			}

			if (implied_length == 0)
				printk
				    ("iSCSI: session iscsi bus %d target id %d "
				     "sending empty Data PDU for implied R2T "
				     "of itt %u, cmnd NULL at %lu\n",
				     session->iscsi_bus, session->target_id,
				     task->itt, jiffies);
		}

		/* save the values that get set when we receive an R2T from 
		 * the target, so that we can receive another one while 
		 * we're sending data.
		 */
		ttt = task->ttt;
		offset = task->data_offset;
		length = task->data_length;
		task->ttt = RSVD_TASK_TAG;
		if (task->scsi_cmnd == NULL) {
			printk
			    ("iSCSI: session iscsi bus %d target id %d sending "
			     "empty Data PDU for R2T (%u @ %u), itt %u, "
			     "ttt %u, cmnd NULL at %lu\n",
			     session->iscsi_bus, session->target_id, offset,
			     length, task->itt, ntohl(ttt), jiffies);
			length = 0;
		}

		atomic_inc(&task->refcount);
		spin_unlock(&session->task_lock);

		/* implied initial R2T */
		if (initial_r2t) {
			DEBUG_FLOW
			    ("iSCSI: session iscsi bus %d target id %d sending "
			     "implied initial R2T data (%u @ 0) for itt %u, "
			     "to %s\n",
			     session->iscsi_bus, session->target_id,
			     implied_length, itt, session->log_name);

			/* we now send an empty PDU if the implied length is 0,
			 * to handle cases where a task's command is removed and
			 * completed while the task is still queued to have data
			 * sent.  We could trigger error recovery at this point,
			 * or send an ABORT_TASK to try to quiet error message 
			 * on the target about 0 length data PDUs. If we end up
			 * trying ABORT_TASK_SET, we're required to continue
			 * responding to all outstanding ttts, though we can 
			 * send empty Data PDUs with the F-bit set (like we do 
			 * here).
			 */
			iscsi_xmit_data(task, RSVD_TASK_TAG, imm_data_length,
					implied_length);
		}

		if (signal_pending(current)) {
			atomic_dec(&task->refcount);
			return;
		}

		/* normal R2T from the target */
		if (ttt != RSVD_TASK_TAG) {
			DEBUG_FLOW
			    ("iSCSI: session iscsi bus %d target id %d sending "
			     "R2T data (%u @ %u) for itt %u, ttt %u to %s\n",
			     session->iscsi_bus, session->target_id, length,
			     offset, itt, ntohl(ttt), session->log_name);

			iscsi_xmit_data(task, ttt, offset, length);
		}

		atomic_dec(&task->refcount);

		if (signal_pending(current))
			return;

		/* relock before checking loop condition */
		spin_lock(&session->task_lock);
	}
	spin_unlock(&session->task_lock);
}

/* send a reply to a nop that requested one */
void
iscsi_xmit_nop_reply(struct iscsi_session *session,
		     struct iscsi_nop_info *nop_info)
{
	struct IscsiNopOutHdr stnoh;
	struct msghdr msg;
	struct iovec iov[5];
	int rc;
	int pad[4];
	uint32_t header_crc32c, data_crc32c;
	int length, iovn, first_data_iovn, i;

	memset(&stnoh, 0, sizeof (stnoh));
	stnoh.opcode = ISCSI_OP_NOOP_OUT | ISCSI_OP_IMMEDIATE;
	stnoh.itt = RSVD_TASK_TAG;
	stnoh.ttt = nop_info->ttt;
	stnoh.flags = ISCSI_FLAG_FINAL;
	memcpy(stnoh.lun, nop_info->lun, sizeof (stnoh.lun));
	hton24(stnoh.dlength, nop_info->dlength);
	stnoh.cmdsn = htonl(session->CmdSn);	/* don't increment after 
						 * immediate cmds 
						 */
	stnoh.expstatsn = htonl(session->ExpStatSn);

	/* PDU header */
	iov[0].iov_base = &stnoh;
	iov[0].iov_len = sizeof (stnoh);
	length = sizeof (stnoh);
	iovn = 1;

	/* HeaderDigest */
	if (session->HeaderDigest == ISCSI_DIGEST_CRC32C) {
		iov[iovn].iov_base = &header_crc32c;
		iov[iovn].iov_len = sizeof (header_crc32c);
		iovn++;
		length += sizeof (header_crc32c);
	}

	first_data_iovn = iovn;

	if (nop_info->dlength) {
		/* data */
		iov[iovn].iov_base = nop_info->data;
		iov[iovn].iov_len = nop_info->dlength;
		length += nop_info->dlength;
		iovn++;

		/* pad */
		if (nop_info->dlength % PAD_WORD_LEN) {
			memset(pad, 0, sizeof (pad));
			iov[iovn].iov_base = pad;
			iov[iovn].iov_len =
			    PAD_WORD_LEN - (nop_info->dlength % PAD_WORD_LEN);
			length += iov[iovn].iov_len;
			iovn++;
		}

		/* DataDigest */
		if (session->DataDigest == ISCSI_DIGEST_CRC32C) {
			data_crc32c =
			    iscsi_crc32c(iov[first_data_iovn].iov_base,
					 iov[first_data_iovn].iov_len);

			for (i = first_data_iovn + 1; i < iovn; i++) {
				data_crc32c =
				    iscsi_crc32c_continued(iov[i].iov_base,
							   iov[i].iov_len,
							   data_crc32c);
			}

			iov[iovn].iov_base = &data_crc32c;
			iov[iovn].iov_len = sizeof (data_crc32c);
			length += sizeof (data_crc32c);
			iovn++;
		}
	}

	/* HeaderDigest */
	if (session->HeaderDigest == ISCSI_DIGEST_CRC32C)
		header_crc32c = iscsi_crc32c(&stnoh, sizeof (stnoh));

	memset(&msg, 0, sizeof (msg));
	msg.msg_iov = iov;
	msg.msg_iovlen = iovn;

	rc = iscsi_sendmsg(session, &msg, length);
	if (rc != length) {
		printk("iSCSI: xmit_nop %d failed, rc %d\n", length, rc);
		iscsi_drop_session(session);
	}

	ISCSI_TRACE(ISCSI_TRACE_TxNopReply, NULL, NULL, nop_info->ttt,
		    nop_info->dlength);
}

/* send replies for NopIns that requested them */
void
iscsi_xmit_nop_replys(struct iscsi_session *session)
{
	struct iscsi_nop_info *nop_info;

	/* these aren't really tasks, but it's not worth having 
	 * a separate lock for them 
	 */
	spin_lock(&session->task_lock);

	/* space for one data-less reply is preallocated in 
	 * the session itself 
	 */
	if (session->nop_reply.ttt != RSVD_TASK_TAG) {
		spin_unlock(&session->task_lock);

		iscsi_xmit_nop_reply(session, &session->nop_reply);
		session->nop_reply.ttt = RSVD_TASK_TAG;

		spin_lock(&session->task_lock);
	}

	/* if we get multiple reply requests, or they have data, 
	 * they'll get queued up 
	 */
	while ((nop_info = session->nop_reply_head)) {
		session->nop_reply_head = nop_info->next;
		if (!session->nop_reply_head)
			session->nop_reply_tail = NULL;
		spin_unlock(&session->task_lock);

		iscsi_xmit_nop_reply(session, nop_info);
		kfree(nop_info);

		if (signal_pending(current))
			return;

		/* relock before checking loop condition */
		spin_lock(&session->task_lock);
	}
	spin_unlock(&session->task_lock);
}

void
iscsi_xmit_logout(struct iscsi_session *session, uint32_t itt, int reason)
{
	struct IscsiLogoutHdr stlh;
	struct msghdr msg;
	struct iovec iov[2];
	uint32_t crc32c;
	int rc, wlen;

	memset(&stlh, 0, sizeof (stlh));
	stlh.opcode = ISCSI_OP_LOGOUT_CMD | ISCSI_OP_IMMEDIATE;
	stlh.flags =
	    ISCSI_FLAG_FINAL | (reason & ISCSI_FLAG_LOGOUT_REASON_MASK);
	stlh.itt = htonl(itt);
	stlh.cmdsn = htonl(session->CmdSn);
	stlh.expstatsn = htonl(session->ExpStatSn);

	memset(iov, 0, sizeof (iov));
	iov[0].iov_base = &stlh;
	iov[0].iov_len = sizeof (stlh);
	memset(&msg, 0, sizeof (msg));
	msg.msg_iov = iov;
	msg.msg_iovlen = 1;
	wlen = sizeof (stlh);

	/* HeaderDigests */
	if (session->HeaderDigest == ISCSI_DIGEST_CRC32C) {
		crc32c = iscsi_crc32c(&stlh, sizeof (stlh));
		iov[msg.msg_iovlen].iov_base = &crc32c;
		iov[msg.msg_iovlen].iov_len = sizeof (crc32c);
		msg.msg_iovlen++;
		wlen += sizeof (crc32c);
	}

	rc = iscsi_sendmsg(session, &msg, wlen);
	if (rc != wlen) {
		printk
		    ("iSCSI: session iscsi bus %d target id %d "
		     "xmit_logout error, rc %d, wlen %d\n",
		     session->iscsi_bus, session->target_id, rc, wlen);
		iscsi_drop_session(session);
	}
}

void
iscsi_xmit_ping(struct iscsi_session *session, uint32_t itt,
		unsigned char *data, int length)
{
	struct IscsiNopOutHdr stph;
	struct msghdr msg;
	struct iovec iov[5];
	unsigned char pad[4];
	uint32_t header_crc32c, data_crc32c;
	int rc, wlen, iovn = 0, first_data_iovn, i;

	memset(&stph, 0, sizeof (stph));
	stph.opcode = ISCSI_OP_NOOP_OUT | ISCSI_OP_IMMEDIATE;
	stph.flags = ISCSI_FLAG_FINAL;
	stph.itt = htonl(itt);	/* reply request */
	stph.ttt = RSVD_TASK_TAG;
	stph.cmdsn = htonl(session->CmdSn);
	stph.expstatsn = htonl(session->ExpStatSn);

	memset(iov, 0, sizeof (iov));
	iov[0].iov_base = &stph;
	iov[0].iov_len = sizeof (stph);
	iovn = 1;
	wlen = sizeof (stph);

	/* HeaderDigests */
	if (session->HeaderDigest == ISCSI_DIGEST_CRC32C) {
		iov[iovn].iov_base = &header_crc32c;
		iov[iovn].iov_len = sizeof (header_crc32c);
		iovn++;
		wlen += sizeof (header_crc32c);
	}

	first_data_iovn = iovn;

	if (data && length) {
		hton24(stph.dlength, length);

		/* add the data */
		iov[iovn].iov_base = data;
		iov[iovn].iov_len = length;
		iovn++;
		wlen += length;

		/* may need to pad as well */
		if (length % PAD_WORD_LEN) {
			memset(pad, 0, sizeof (pad));
			iov[iovn].iov_base = pad;
			iov[iovn].iov_len =
			    PAD_WORD_LEN - (length % PAD_WORD_LEN);
			wlen += iov[iovn].iov_len;
			iovn++;
		}

		/* DataDigest */
		if (session->DataDigest == ISCSI_DIGEST_CRC32C) {
			data_crc32c =
			    iscsi_crc32c(iov[first_data_iovn].iov_base,
					 iov[first_data_iovn].iov_len);

			for (i = first_data_iovn + 1; i < iovn; i++) {
				data_crc32c =
				    iscsi_crc32c_continued(iov[i].iov_base,
							   iov[i].iov_len,
							   data_crc32c);
			}

			iov[iovn].iov_base = &data_crc32c;
			iov[iovn].iov_len = sizeof (data_crc32c);
			wlen += sizeof (data_crc32c);
			iovn++;
		}

		DEBUG_FLOW
		    ("iSCSI: session iscsi bus %d target id %d tx Nop/data "
		     "itt %u, lengths %d, %d, %d\n",
		     session->iscsi_bus, session->target_id, itt,
		     iov[0].iov_len, iov[1].iov_len, iov[2].iov_len);
	} else {
		DEBUG_FLOW
		    ("iSCSI: session iscsi bus %d target id %d tx Nop/data "
		     "itt %u, lengths %d, %d, %d\n",
		     session->iscsi_bus, session->target_id, itt,
		     iov[0].iov_len, iov[1].iov_len, iov[2].iov_len);
	}

	/* can't calculate the HeaderDigest until after 
	 * we've filled in the dlength 
	 */
	if (session->HeaderDigest == ISCSI_DIGEST_CRC32C)
		header_crc32c = iscsi_crc32c(&stph, sizeof (stph));

	ISCSI_TRACE(ISCSI_TRACE_TxPing, NULL, NULL, itt, length);

	memset(&msg, 0, sizeof (msg));
	msg.msg_iov = iov;
	msg.msg_iovlen = iovn;

	rc = iscsi_sendmsg(session, &msg, wlen);

	if (rc != wlen) {
		printk
		    ("iSCSI: session iscsi bus %d target id %d "
		     "xmit_ping error, rc %d, wlen %d\n",
		     session->iscsi_bus, session->target_id, rc, wlen);
		iscsi_drop_session(session);
	}
}

