/*
 * iSCSI connection daemon
 * Copyright (C) 2002 Cisco Systems, Inc.
 * maintained by linux-iscsi-devel@lists.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsid.c,v 1.74.6.10 2004/01/16 14:04:55 surekhap Exp $
 *
 */

#include <unistd.h>
#include <errno.h>
#include <poll.h>
#include <sys/wait.h>

#include "iscsi-protocol.h"
#include "iscsi-discovery.h"
#include "iscsi-bindings.h"
#include "iscsi-hooks.h"

/* global config info */
struct iscsi_daemon_config daemon_config;

/* global so that the signal handlers can reset the pids */
static struct iscsi_discovery_process_list discovery_processes;
static struct iscsi_session_process_list session_processes;
static struct iscsi_target_list targets;

static void
sigpipe_handler(int unused)
{
	debugmsg(1, "received sigpipe");
}

void
sigchld_handler(int unused)
{
	pid_t pid = 0;
	int status = 0;
	struct iscsi_discovery_process *discovery;
	struct iscsi_session_process *session;

      wait_pid:
	while ((pid = waitpid(-1, &status, WNOHANG)) > 0) {
		if (WIFEXITED(status))
			debugmsg(1, "child %d exited with status %d", pid,
				 WEXITSTATUS(status));
		if (WIFSIGNALED(status))
			debugmsg(1, "child %d exited on signal %d", pid,
				 WTERMSIG(status));
		if (WIFSTOPPED(status))
			debugmsg(1, "child %d stopped on signal %d", pid,
				 WSTOPSIG(status));

		/* reset pids if the child was a discovery or target process */
		for (discovery = discovery_processes.head; discovery;
		     discovery = discovery->next) {
			if (discovery->pid == pid) {
				discovery->pid = 0;

				if (WIFSIGNALED(status)) {
					/* if the process doesn't want a 
					 * particular signal to restart it,
					 * it should catch the signal and exit
					 * (0);
					 */
					discovery->restart = 1;
					discovery_processes.changed = 1;
					debugmsg(1,
						 "discovery process %p needs "
						 "restart because pid %d "
						 "killed by signal %d",
						 discovery, pid,
						 WTERMSIG(status));
				} else if (WIFEXITED(status)
					   && WEXITSTATUS(status)) {
					/* abnormal exit, try again */
					discovery->restart = 1;
					discovery_processes.changed = 1;
					debugmsg(1,
						 "discovery process %p needs "
						 "restart because pid %d "
						 "exited with status 0x%x",
						 discovery, pid,
						 WEXITSTATUS(status));
				} else {
					/* normal exit. */
					discovery->remove = 1;
					discovery_processes.changed = 1;
					debugmsg(1,
						 "discovery process %p pid %d "
						 "fd %d exited normally, "
						 "marking for removal",
						 discovery, pid,
						 discovery->pipe_fd);
				}
				goto wait_pid;
			}
		}
		for (session = session_processes.head; session;
		     session = session->next) {
			if (session->pid == pid) {
				session->pid = 0;
				if (WIFSIGNALED(status)) {
					/* if the process doesn't want a 
					 * particular signal to restart it,
					 * it should catch the signal and exit
					 * (0);
					 */
					session_processes.changed = 1;
					session->restart = 1;
					debugmsg(1,
						 "session process %p needs "
						 "restart because pid %d "
						 "killed by signal %d",
						 session, pid,
						 WTERMSIG(status));
				} else if (WIFEXITED(status)) {
					switch (WEXITSTATUS(status)) {
					case 0:
						/* normal exit.  we don't 
						 * remove session processes 
						 * that exit with status 0,
						 * since that always happens 
						 * on Linux because the session 
						 * process exits as soon as 
						 * the kernel module ioctl
						 * returns, and the kernel 
						 * module does all the real work
						 */
						debugmsg(1,
							 "session process %p "
							 "with pid %d exited "
							 "normally",
							 session, pid);
						session->pid = 0;
						break;
					case ISCSI_SESSION_FAILED_NO_RETRY:
						/* the session process died 
						 * after the session failed, 
						 * but we don't want
						 * to restart it unless the 
						 * user asks for a config 
						 * reload.  This can occur if 
						 * the session fails due to 
						 * authentication or 
						 * authorization failures to 
						 * all portals.
						 * We don't want to retry 
						 * because we might have 
						 * discovered a target that we
						 * aren't authorized to use, 
						 * and retrying forever would 
						 * fill the log.
						 * We do want the user to be 
						 * able to request a retry, in 
						 * case they've changed
						 * the access controls on the 
						 * target and we're now 
						 * authorized.
						 */
						debugmsg(1,
							 "session process %p "
							 "with pid %d exited, "
							 "no retry",
							 session, pid);
						session->pid = 0;
						session->failed = 1;
						break;
					default:
						session_processes.changed = 1;
						session->restart = 1;
						debugmsg(1,
							 "session process %p "
							 "needs restart "
							 "becaused pid %d "
							 "exited with status "
							 "0x%x",
							 session, pid,
							 WEXITSTATUS(status));
						break;
					}
				}
				goto wait_pid;
			}
		}
	}
}

void
fork_discovery_process(struct iscsi_discovery_process *discovery)
{
	int pipes[2];

	discovery->in_progress = 1;

	/* make pipes */
	if (pipe(pipes) < 0) {
		discovery->restart = 1;
		return;
	}

      fork_again:
	if (iscsi_process_should_exit())
		return;

	discovery->pid = fork();
	if (discovery->pid == 0) {
		/* close the read side */
		close(pipes[0]);
		/* save the write side */
		discovery->pipe_fd = pipes[1];

		discovery->pid = getpid();
		debugmsg(3, "started discovery process %p for entry %p",
			 discovery, discovery->entry);
		discovery_process(discovery);
		exit(0);
	} else if (discovery->pid < 0) {
		errormsg("cannot fork discovery %p", discovery);
		sleep(1);
		goto fork_again;
	} else {
		/* close the write side */
		close(pipes[1]);
		/* save the read side */
		discovery->pipe_fd = pipes[0];

		debugmsg(3, "forked child %d for discovery %p, pipe fd %d",
			 discovery->pid, discovery, discovery->pipe_fd);
	}
}

void
kill_discovery(struct iscsi_discovery_process *discovery)
{
	pid_t pid = discovery->pid;

	if (pid) {
		debugmsg(3, "kill_discovery %p sending SIGTERM to pid %d",
			 discovery, pid);
		kill(pid, SIGTERM);
	}
}

void
trigger_rediscovery(struct iscsi_discovery_process *discovery)
{
	pid_t pid = discovery->pid;

	debugmsg(1, "triggering rediscovery from process %p", discovery);
	if (pid) {
		kill(pid, SIGHUP);
	}
}

void
add_discovery(struct iscsi_discovery_process_list *list,
	      struct iscsi_discovery_process *discovery)
{
	sigset_t chld;

	debugmsg(7, "adding discovery %p, head %p, tail %p", discovery,
		 list->head, list->tail);

	/* block SIGCHLD while we're doing this */
	sigemptyset(&chld);
	sigaddset(&chld, SIGCHLD);
	sigprocmask(SIG_BLOCK, &chld, NULL);

	if (list->head) {
		/* append at the tail */
		discovery->next = NULL;
		discovery->prev = list->tail;
		list->tail->next = discovery;
		list->tail = discovery;
	} else {
		/* become the only list element */
		discovery->next = discovery->prev = NULL;
		list->head = list->tail = discovery;
	}
	list->count++;

	sigprocmask(SIG_UNBLOCK, &chld, NULL);
}

void
remove_discovery(struct iscsi_discovery_process_list *list,
		 struct iscsi_discovery_process *discovery)
{
	sigset_t chld;

	/* block SIGCHLD while we're doing this */
	sigemptyset(&chld);
	sigaddset(&chld, SIGCHLD);
	sigprocmask(SIG_BLOCK, &chld, NULL);

	if (discovery == list->head) {
		debugmsg(7, "removing discovery head %p, tail %p", discovery,
			 list->tail);
		list->head = discovery->next;
		if (list->head == NULL)
			list->tail = NULL;
		discovery->next = discovery->prev = NULL;
	} else if (discovery == list->tail) {
		debugmsg(7, "removing discovery tail %p, head %p", discovery,
			 list->head);
		list->tail = discovery->prev;
		list->tail->next = NULL;
		discovery->next = discovery->prev = NULL;
	} else if (discovery->next && discovery->prev) {
		debugmsg(7, "removing interior discovery %p, head %p, tail %p",
			 discovery, list->head, list->tail);
		discovery->next->prev = discovery->prev;
		discovery->prev->next = discovery->next;
		discovery->next = discovery->prev = NULL;
	} else {
		logmsg(AS_ERROR,
		       "couldn't remove discovery process %p from list %p",
		       discovery, list);
	}
	list->count--;

	sigprocmask(SIG_UNBLOCK, &chld, NULL);
}

/* recursively free dynamically allocated memory */
void
free_discovery(struct iscsi_discovery_process *discovery)
{
	if (discovery->entry) {
		debugmsg(6, "freeing discovery %p with config %p", discovery,
			 discovery->entry);
		free_config_entry(discovery->entry);
		discovery->entry = NULL;
	} else {
		debugmsg(6, "freeing discovery %p with no config", discovery);
	}

	free(discovery);
}

void
show_target(struct iscsi_target *target)
{
	int p = 0;
	struct iscsi_portal_descriptor *portal;

	debugmsg(1, "target %p name %s", target, target->TargetName);
	for (portal = target->current_portals; portal; portal = portal->next) {
		debugmsg(1,
			 "target %p portal[%d]=%p, address %s, "
			 "ip %u.%u.%u.%u, port %d, tag %d",
			 target, p++, portal, portal->address, portal->ip[0],
			 portal->ip[1], portal->ip[2], portal->ip[3],
			 portal->port, portal->tag);
	}
}

void
fork_session_process(struct iscsi_session_process *process)
{
	struct iscsi_session_config *config = process->config;

      fork_again:
	if (iscsi_process_should_exit())
		return;

	process->pid = fork();
	if (process->pid == 0) {
		process->pid = getpid();
		debugmsg(2,
			 "starting iSCSI session process %p config %p "
			 "for bus %d target %d, TargetName %s",
			 process, config, config->iscsi_bus, config->target_id,
			 config->target->TargetName);
		iscsi_session_process(process, process->config);
		exit(0);
	} else if (process->pid < 0) {
		errormsg("cannot fork session process %p", process);
		sleep(1);
		goto fork_again;
	} else {
		debugmsg(2, "forked pid %d for session process %p",
			 process->pid, process);
	}
}

void
add_session_process(struct iscsi_session_process_list *list,
		    struct iscsi_session_process *process)
{
	sigset_t chld;

	/* block SIGCHLD while we're doing this */
	sigemptyset(&chld);
	sigaddset(&chld, SIGCHLD);
	sigprocmask(SIG_BLOCK, &chld, NULL);

	debugmsg(7, "adding session process %p, head %p, tail %p", process,
		 list->head, list->tail);
	if (list->head) {
		/* append at the tail */
		process->next = NULL;
		process->prev = list->tail;
		list->tail->next = process;
		list->tail = process;
	} else {
		/* become the only list element */
		process->next = process->prev = NULL;
		list->head = list->tail = process;
	}
	list->count++;

	sigprocmask(SIG_UNBLOCK, &chld, NULL);
}

void
remove_session_process(struct iscsi_session_process_list *list,
		       struct iscsi_session_process *process)
{
	sigset_t chld;

	/* block SIGCHLD while we're doing this */
	sigemptyset(&chld);
	sigaddset(&chld, SIGCHLD);
	sigprocmask(SIG_BLOCK, &chld, NULL);

	if (process == list->head) {
		debugmsg(7, "removing session process head %p, tail %p",
			 process, list->tail);
		list->head = process->next;
		if (list->head == NULL)
			list->tail = NULL;
		process->next = process->prev = NULL;
	} else if (process == list->tail) {
		debugmsg(7, "removing session process tail %p, head %p",
			 process, list->head);
		list->tail = process->prev;
		list->tail->next = NULL;
		process->next = process->prev = NULL;
	} else if (process->next && process->prev) {
		debugmsg(7, "removing interior process %p, head %p, tail %p",
			 process, list->head, list->tail);
		process->next->prev = process->prev;
		process->prev->next = process->next;
		process->next = process->prev = NULL;
	} else {
		logmsg(AS_ERROR,
		       "couldn't remove session process %p from list %p",
		       process, list);
	}
	list->count--;

	sigprocmask(SIG_UNBLOCK, &chld, NULL);
}

void
free_session_process(struct iscsi_session_process *process)
{
	if (process->config) {
		debugmsg(6, "freeing session process %p with config %p",
			 process, process->config);
		/* Note: caller must free the config if appropriate to do so */
		process->config->process = NULL;
		process->config = NULL;
	} else {
		debugmsg(6, "freeing process %p with no config", process);
	}

	free(process);
}

void
stop_session_process(struct iscsi_session_process *process)
{
	debugmsg(1, "stopping session process %p pid %d", process,
		 process->pid);
	remove_session_process(&session_processes, process);
	iscsi_kill_session(process);
	if (process->config)
		process->config->process = NULL;
	free_session_process(process);
}

void
stop_session_processes(struct iscsi_session_process_list *session_processes)
{
	struct iscsi_session_process *process;

	debugmsg(1, "stopping all session processes");
	while ((process = session_processes->head)) {
		stop_session_process(process);
	}
}

void
start_discovery_processes(struct iscsi_config *config,
			  struct iscsi_discovery_process_list
			  *discovery_processes)
{
	struct iscsi_config_entry *entry, *next;
	struct iscsi_discovery_process *discovery = NULL;
	struct iscsi_discovery_process *chk_discovery;
	int order = 1;
	int dup_addr = 0;

	if (config == NULL || config->head == NULL) {
		if (config)
			debugmsg(1,
				 "no discovery processes have been configured, "
				 "config %p, head %p, tail %p",
				 config, config->head, config->tail);
		else
			debugmsg(1,
				 "no discovery processes have been configured");

		logmsg(AS_ERROR,
		       " Configuration file is empty, exiting the "
		       "discovery process\n");

		return;
	}

	debugmsg(1, "starting discovery processes");

	/* scan the config looking for SendTargets or SLP configs */
	entry = config->head;

	while (entry) {
		/* save the next config to process, since we may 
		 * splice out the current one. 
		 */
		next = entry->next;

		debugmsg(7, "start discovery checking config entry %p, type %d",
			 entry, entry->type);
		switch (entry->type) {
		case CONFIG_TYPE_SENDTARGETS:
		case CONFIG_TYPE_DISCOVERY_FILE:
		case CONFIG_TYPE_SLP:
			/* grab the discovery entries out of the config, 
			 * since nothing else needs them 
			 */
			discovery = calloc(1, sizeof (*discovery));
			if (discovery) {
				discovery->order = order++;
				discovery->in_progress = 1;

				/* splice out the config entry, since nothing 
				 * else will need it 
				 */
				remove_config_entry(config, entry);

				/* save the config for the process */
				discovery->entry = entry;
				dup_addr = 0;
				for (chk_discovery = discovery_processes->head;
				     chk_discovery;
				     chk_discovery = discovery->next) {
					if (strcmp
					    (chk_discovery->entry->config.
					     sendtargets->address,
					     discovery->entry->config.
					     sendtargets->address) == 0) {
						logmsg(AS_ERROR,
						       " ignoring duplicate "
						       "discovery address %s in"
						       " the config file\n",
						       discovery->entry->config.
						       sendtargets->address);
						dup_addr = 1;
						break;
					}
				}

				if (!dup_addr) {
					/* start the process */
					/* add the process to the 
					 * discovery processes 
					 */
					add_discovery(discovery_processes,
						      discovery);
					fork_discovery_process(discovery);
					/* give it a chance to start before 
					 * we go on to the next 
					 */
					usleep(10 * 1000);
				}
			} else {
				logmsg(AS_ERROR,
				       "couldn't allocate discovery process");
			}
			break;
		default:
			break;
		}

		entry = next;
	}

	debugmsg(1, "done starting discovery processes");
}

void
stop_discovery_process(struct iscsi_discovery_process *process)
{
	debugmsg(1, "stopping discovery process %p", process);
	remove_discovery(&discovery_processes, process);
	discovery_processes.changed = 1;
	kill_discovery(process);
	free_discovery(process);
}

void
stop_discovery_processes(struct iscsi_discovery_process_list
			 *discovery_processes)
{
	struct iscsi_discovery_process *process;

	while ((process = discovery_processes->head)) {
		stop_discovery_process(process);
		discovery_processes->changed = 1;
	}
}

/* dynamic arrays for poll fds and process pointers */
struct discovery_poll_info {
	size_t allocated_length;
	size_t valid_length;
	struct iscsi_discovery_process **processes;
	struct pollfd *poll_fds;
};

int
setup_polling(struct discovery_poll_info *poll_info,
	      struct iscsi_discovery_process_list *discovery_processes)
{
	struct iscsi_discovery_process *discovery;
	struct iscsi_discovery_process **new_processes = NULL;
	struct pollfd *new_fds = NULL;
	int i;
	int needed = 0;

	/* determine how many fds are needed (pipe_fd >= 0) */
	for (discovery = discovery_processes->head; discovery;
	     discovery = discovery->next) {
		if (discovery->pipe_fd >= 0)
			needed++;
	}
	debugmsg(7, "need %d poll fds", needed);

	/* see if we need to allocate more memory */
	if (poll_info->allocated_length < needed) {
		/* we need more */
		new_fds = realloc(poll_info->poll_fds, needed * sizeof (*new_fds));	
			/* the actual pollfds to give to poll */
		new_processes = realloc(poll_info->processes, needed * sizeof (*new_processes));	
			/* pointers to processes */

		if (new_fds && new_processes) {
			/* update the poll info */
			poll_info->allocated_length = needed;
			poll_info->poll_fds = new_fds;
			poll_info->processes = new_processes;

			/* initialize the added memory */
			memset(poll_info->poll_fds + poll_info->valid_length, 0,
			       (poll_info->allocated_length -
				poll_info->valid_length) *
			       sizeof (struct pollfd));
			for (i = poll_info->valid_length;
			     i < poll_info->allocated_length; i++) {
				poll_info->processes[i] = NULL;
			}
		} else {
			/* failed */
			logmsg(AS_ERROR,
			       "couldn't realloc poll info, have %d, needed %d",
			       poll_info->allocated_length, needed);
			/* mark the process list changed, 
			 * so that we try again later 
			 */
			discovery_processes->changed = 1;
			return 0;
		}
	}

	/* setup the poll info based on the current discovery processes */
	i = 0;
	for (discovery = discovery_processes->head; discovery;
	     discovery = discovery->next) {
		if (discovery->pipe_fd >= 0) {
			if (i >= poll_info->allocated_length) {
				/* should be impossible for the process list to 
				 * grow while we're running, but we're paranoid 
				 */
				logmsg(AS_ERROR,
				       "couldn't set poll info[%d] for "
				       "discovery %p, allocated length only %d",
				       i, discovery,
				       poll_info->allocated_length);
				break;
			} else {
				debugmsg(7,
					 "poll_info[%d] = fd %d for "
					 "discovery %p",
					 i, discovery->pipe_fd, discovery);
				poll_info->processes[i] = discovery;
				poll_info->poll_fds[i].fd = discovery->pipe_fd;
				poll_info->poll_fds[i].events =
				    POLLIN | POLLPRI;
				poll_info->poll_fds[i].revents = 0;
				i++;
			}
		}
	}

	/* record how many entries we're using */
	poll_info->valid_length = i;

	/* success */
	return 1;
}

struct iscsi_target *
find_target(char *name, struct iscsi_target_list *targets)
{
	struct iscsi_target *target;

	if ((name == NULL) || (*name == '\0'))
		return NULL;

	for (target = targets->head; target; target = target->next) {
		if (strcmp(target->TargetName, name) == 0)
			return target;
	}

	return NULL;
}

int
update_target_list(struct iscsi_target_list *targets,
		   struct iscsi_target *target)
{
	struct iscsi_target *prev = NULL;
	struct iscsi_target *current = targets->head;

	if (targets == NULL) {
		return (-1);
	}

	while (current != NULL
	       && strcmp(current->TargetName, target->TargetName)) {
		prev = current;
		current = current->next;
	}
	if (current == NULL) {
		return (-1);
	} else {
		if (prev == NULL) {
			targets->head = current->next;
		} else if (current->next == NULL) {
			targets->tail = prev;
			prev->next = current->next;
		} else {
			prev->next = current->next;
		}
		current->next = NULL;
		free(current);
		return (0);
	}
}

int
update_session_portal(struct iscsi_session_config *config,
		      struct iscsi_portal_descriptor *portal)
{
	struct iscsi_portal_config *prev = NULL;
	struct iscsi_portal_config *current = config->portals.head;

	if (config == NULL) {
		return (-1);
	}

	while (current != NULL
	       && !same_portal_descriptor(current->descriptor, portal)) {
		prev = current;
		current = current->next;
	}
	if (current == NULL) {
		return (-1);
	} else {
		if (prev == NULL) {
			config->portals.head = current->next;
		} else if (current->next == NULL) {
			config->portals.tail = prev;
			prev->next = current->next;
		} else {
			prev->next = current->next;
		}
		current->next = NULL;
		current->descriptor = NULL;
		free(current);
		return (0);
	}
}

void
delete_target(struct iscsi_target *existing_target,
	      struct iscsi_target_list *targets)
{
	struct iscsi_session_config *existing_session = NULL, *new_session = NULL;
	struct iscsi_target_config *current_config = NULL, *new_config = NULL;
	struct iscsi_target *target = NULL;

	target = existing_target;
	target->discovery = NULL;
	current_config = target->current_config;
	new_config = target->new_config;
	existing_session = current_config->sessions;

	while (existing_session) {
		iscsi_config_changed(existing_session->process, new_session,
				     existing_session);
		remove_session_process(&session_processes,
				       existing_session->process);
		free_session_process(existing_session->process);
		existing_session->process = NULL;
		existing_session = existing_session->next;
	}
	current_config->sessions = NULL;
	if (current_config != NULL)
		free_target_config(current_config);
	target->current_config = NULL;
	if (new_config != NULL)
		free_target_config(new_config);
	target->new_config = NULL;
	if (target->current_portals)
		free_portal_descriptors(target->current_portals);
	target->current_portals = NULL;
	if (target->new_portals)
		free_portal_descriptors(target->new_portals);
	target->new_portals = NULL;
	free(target->TargetName);
	update_target_list(targets, target);
}

void
process_targets_for_portal(char *address, struct iscsi_target_list *targets,
			   struct iscsi_config *config)
{
	/* Check for targets using this particular portal so that we can
	 * appropriately update the specific target and session config
	 * information.
	 1) If the portal is used by this target and if this is the only portal,
	 then delete the target and session config and ask the kernel to
	 terminate the target session in the kernel.
	 2) If there are more than one portal for the target, update the session
	 config and target portal list with new number of portals.Send an update
         ioctl to update the kernel session structure.
	 3) If multipath is enabled and more than one portal is available to the
	 same target, remove the session config for the portal removed,
	 and terminate the session. If only one path is now active and that
	 portal is made offline, need to delete the target config and
	 session config and terminate the session.
	 *
	 */
	struct iscsi_session_config *session = NULL, *prev_session = NULL;
	struct iscsi_session_config *old_session = NULL, *existing_session;
	struct iscsi_session_process *process;
	struct iscsi_target *target;
	struct iscsi_portal_descriptor *portal, *tmpportal, *new_portals, *desc;
	struct iscsi_auth_config *auth_options = NULL;
	struct iscsi_target_config *target_config;
	int num_portals = 0;
	int multipath_type;
	int ret = 0;
	int matched;

	for (target = targets->head; target; target = target->next) {
		tmpportal = portal = target->current_portals;
		multipath_type = target->current_config->multipath_options.type;
		existing_session = target->current_config->sessions;
		process = existing_session->process;
		new_portals = NULL;
		matched = 0;

		/* Get the number of portals for this target */
		num_portals = 0;
		for (portal = target->current_portals; portal;
		     portal = portal->next)
			num_portals++;

		for (portal = target->current_portals; portal;
		     portal = portal->next) {
			if (!strcmp(address, portal->address)) {
				matched = 1;
				break;
			}
		}

		/* Create a new target config structure to save the 
		 * old portal info 
		 */
		if (matched && num_portals > 1
		    && multipath_type == MULTIPATH_NO) {
			target_config =
			    create_target_config(target->TargetName,
						 target->current_portals,
						 config, auth_options);
			if (target_config) {
				if (target->new_config)
					free_target_config(target->new_config);
				target->new_config = target_config;
				if (target->new_portals)
					free_portal_descriptors(target->
								new_portals);
				target->new_portals = NULL;
			}
			old_session = target->new_config->sessions;
		}

		while (tmpportal != NULL && strcmp(address, tmpportal->address)) {
			new_portals = tmpportal;
			tmpportal = tmpportal->next;
		}

		if (tmpportal == NULL) {
			logmsg(AS_ERROR,
			       "Portal %s to be deleted is not in the "
			       "portal list of target %s\n",
			       address, target->TargetName);
			return;
		}
		if (num_portals == 1) {
			/* Delete the target and session config completely */
			delete_target(target, targets);
		} else if (num_portals > 1 && multipath_type == MULTIPATH_NO) {
			/* Multipath is not configured, so just need to update 
			 * portal structure of the target session,with an update
			 * ioctl
			 */
			process->config_number++;
			ret =
			    update_session_portal(existing_session, tmpportal);
			if (ret == -1) {
				logmsg(AS_ERROR,
				       "specified portal does not exist\n");
				return;
			}
			iscsi_config_changed(process, existing_session,
					     old_session);
		} else {
			/* Delete the session config for this portal */
			session = existing_session;
			desc = session->portals.head->descriptor;
			prev_session = NULL;
			while (session != NULL
			       && !same_portal_descriptor(desc, tmpportal)) {
				prev_session = session;
				session = session->next;
				desc = session->portals.head->descriptor;
			}
			if (prev_session == NULL)
				target->current_config->sessions =
				    session->next;
			else
				prev_session->next = session->next;
			session->next = NULL;
			stop_session_process(session->process);
			session->process = NULL;
			session->target = NULL;
			if (target->current_portals) {
				session->portals.head->descriptor = NULL;
				session->portals.head = session->portals.tail =
				    NULL;
			}
			free_session_config(session);
		}

		if (num_portals > 1) {
			if (target->new_config)
				free_target_config(target->new_config);
			target->new_config = NULL;

			if (tmpportal == NULL) {
				logmsg(AS_ERROR, " no portals are present\n");
			}
			if (new_portals == NULL) {
				target->current_portals = tmpportal->next;
			} else {
				new_portals->next = tmpportal->next;
			}
			tmpportal->next = NULL;
			free(tmpportal->address);
			free(tmpportal);
		}
	}
}

void
add_target(struct iscsi_target_list *targets, struct iscsi_target *target)
{
	target->next = NULL;
	if (targets->head) {
		targets->tail->next = target;
		targets->tail = target;
	} else {
		targets->head = targets->tail = target;
	}
}

struct iscsi_portal_descriptor *
alloc_portal_descriptor(void)
{
	struct iscsi_portal_descriptor *portal = calloc(1, sizeof (*portal));

	if (portal) {
		/* defaults */
		portal->address = NULL;
		portal->tag = PORTAL_GROUP_TAG_UNKNOWN;
		portal->port = ISCSI_LISTEN_PORT;
	}

	return portal;
}

static char discovery_buffer[256 * 1024];

void
update_async_data(struct iscsi_discovery_process *discovery,
		  struct iscsi_config *config,
		  struct iscsi_target_list *targets)
{
	char *input, *line, *eol;
	char *address = NULL;
	int count;
	int tgtoffline = 0;
	int portoffline = 0;
	struct iscsi_target *target = NULL, *existing_target = NULL;

	input = discovery_buffer;

	for (;;) {
		/* read discovery info into the buffer, always leaving the 
		 * last byte for a NUL terminator 
		 */
		count =
		    read(discovery->pipe_fd, input,
			 sizeof (discovery_buffer) - 1 - (input -
							  discovery_buffer));

		if (count > 0) {
			char *data = input;

			input += count;
			*input = '\0';
			debugmsg(7, "discovery %p pipe %d, %d bytes input '%s'",
				 discovery, discovery->pipe_fd, count, data);
			discovery->in_progress = 1;
		} else if (count == 0) {
			/* end of file, guess we're done */
			debugmsg(7, "discovery %p pipe at end of file",
				 discovery, discovery->pipe_fd);
			close(discovery->pipe_fd);
			discovery->pipe_fd = -1;
			discovery->in_progress = 0;
			discovery_processes.changed = 1;
			/* so that we remove the discovery process */
			return;
		} else if (errno != EINTR) {
			/* some sort of non-retryable error */
			errormsg("error reading from discovery process %p",
				 discovery);
			/* FIXME: maybe we should kill or restart 
			 * the discovery process. 
			 */
			return;
		}

		/* scan the input for line boundaries, and process each line */
		line = discovery_buffer;
		while (line < input) {
			if ((eol = strchr(line, '\n'))) {
				*eol = '\0';
				debugmsg(7, "processing pipe line %p: %s", line,
					 line);

				if (eol == line) {
					/* ignore blank lines */
				} else if (strcmp(line, ";") == 0) {
					/* The target record has been terminated
					 * we are done processing the record */
				} else if (strcmp(line, "!") == 0) {
					/* This means the end of data in pipe
					 * we are done with reading data
					 */
					if (tgtoffline) {
						/* target offline occured ... 
						 * need to orphan the target
						 */
						delete_target(existing_target,
							      targets);
					}
					if (portoffline)
						process_targets_for_portal
						    (address, targets, config);
					return;
				} else if (strncmp(line, "PF=", 3) == 0) {
					char *port;
					address = line + 3;

					if ((port = strrchr(line, ':'))) {
						*port = '\0';
						port++;
					}
					debugmsg(1,
						 " Failing over all targets "
						 "using portal %s\n",
						 address);
					portoffline = 1;
				} else if (strncmp(line, "TF=", 3) == 0) {
					char *name = line + 3;
					debugmsg(1, " Deleting target %s\n",
						 name);
					tgtoffline = 1;
					/* start of a target description.
					 * if the target has already been 
					 * discovered, we will need to delete it
					 */
					if ((existing_target =
					     find_target(name, targets))) {
						target = existing_target;

						/* track which discovery 
						 * process owns this target 
						 */
						if (target->discovery == NULL) {
							target->discovery =
							    discovery;
							target->discovered = 1;
						} else if (target->discovery ==
							   discovery) {
							target->discovered = 1;	
								/* we found it 
								 * again 
								 */
						}
					}
				} else {
					debugmsg(1, " Could not read %s\n",
						 line);
				}
				/* on to the next line */
				line = eol + 1;
			} else {
				/* move any remaining input to the beginning of 				 * the buffer, and wait for more input 				 		 */
				size_t bytes = input - line;

				debugmsg(7,
					 "deferring partial pipe line %p: %s",
					 line, line);

				memmove(discovery_buffer, line, bytes);
				line = input = discovery_buffer + bytes;
				break;
			}
		}

		if (line >= input)
			input = discovery_buffer;
	}
}

/* process discovery info and update targets */
void
update_targets(struct iscsi_discovery_process *discovery,
	       struct iscsi_config *config, struct iscsi_target_list *targets)
{
	char *input, *line, *eol;
	char ip[10];
	int count;
	int flag = 0, no_data = 0;
	struct iscsi_target *target = NULL, *existing_target =
	    NULL, *new_target = NULL;
	struct iscsi_target_config *target_config;
	struct iscsi_portal_descriptor *portal = NULL, *port = NULL;
	struct iscsi_portal_descriptor *first_portal = NULL, *last_portal =
	    NULL;
	struct iscsi_auth_config *discovery_auth_options = NULL;
	struct iscsi_auth_config *auth_options = NULL;

	/* FIXME: for now, for backwards compatibility with the old config file 	 * semantics, pass on authentication settings from the discovery 
	 * process to the targets it discovers.
	 * This will behave oddly if the same target is discovered from 
	 * multiple discovery processes, and they have different 
	 * authentication settings.
	 */
	if (discovery->entry) {
		switch (discovery->entry->type) {
		case CONFIG_TYPE_SENDTARGETS:{
				struct iscsi_sendtargets_config *sendtargets =
				    discovery->entry->config.sendtargets;

				debugmsg(7,
					 "discovery %p will propagate "
					 "authentication options to "
					 "discovered targets",
					 discovery);
				discovery_auth_options =
				    &sendtargets->auth_options;
				break;
			}
		case CONFIG_TYPE_SLP:{
				struct iscsi_slp_config *slp =
				    discovery->entry->config.slp;

				debugmsg(7,
					 "discovery %p will propagate "
					 "authentication options to discovered "
					 "targets",
					 discovery);
				discovery_auth_options = &slp->auth_options;
				break;
			}
		case CONFIG_TYPE_DISCOVERY_FILE:{
				struct iscsi_discovery_file_config *file =
				    discovery->entry->config.file;

				debugmsg(7,
					 "discovery %p will propagate "
					 "authentication options to discovered "
					 "targets",
					 discovery);
				discovery_auth_options = &file->auth_options;
				break;
			}
		}
	}

	/* we're about to start a round of discovery for this process.
	 * Mark all targets formerly discovered by the process as not
	 * discovered, so that any target that we don't rediscover in
	 * this round can be orphaned, and possibly get taken over by
	 * a different discovery process that can still find it.
	 */
	if (!discovery->in_progress) {
		for (target = targets->head; target; target = target->next) {
			if (target->discovery == discovery) {
				debugmsg(3,
					 "discovery %p expecting to rediscover "
					 "target %p name %s",
					 discovery, target, target->TargetName);
				target->discovered = 0;
			}
		}
	}

	/* We determine the data coming in the pipe,
	 * whether it is sendtargets data or async data */

	count = read(discovery->pipe_fd, &ip[0], 1);
	ip[2] = '\0';
	if (count > 0 && ip[0] == 'A') {
		debugmsg(1, "Received Async data from the pipe %c\n", ip[0]);
		update_async_data(discovery, config, targets);
	} else {
		if (count > 0 && ip[0] == 'D') {
			flag = 0;
			debugmsg(1,
				 "Received Sendtargets data from the pipe %c\n",
				 ip[0]);
		} else if (count > 0 && ip[0] == '!') {
			/* Sendtargets returns with no data */
			no_data = 1;
		}

		input = discovery_buffer;

		for (;;) {
			/* read discovery info into the buffer, always leaving 
			 * the last byte for a NUL terminator 
			 */
			count =
			    read(discovery->pipe_fd, input,
				 sizeof (discovery_buffer) - 1 - (input -
								  discovery_buffer));

			if (count > 0) {
				char *data = input;

				input += count;
				*input = '\0';
				debugmsg(7,
					 "discovery %p pipe %d, %d bytes "
					 "input '%s'",
					 discovery, discovery->pipe_fd, count,
					 data);
				discovery->in_progress = 1;
			} else if (count == 0) {
				/* end of file, guess we're done */
				debugmsg(7, "discovery %p pipe at end of file",
					 discovery, discovery->pipe_fd);
				close(discovery->pipe_fd);
				discovery->pipe_fd = -1;
				discovery->in_progress = 0;
				discovery_processes.changed = 1;	
					/* so that we remove the 
					 * discovery process 
					 */
				return;
			} else if (errno != EINTR) {
				/* some sort of non-retryable error */
				errormsg
				    ("error reading from discovery process %p",
				     discovery);
				/* FIXME: maybe we should kill or 
				 * restart the discovery process. 
				 */
				return;
			}

			/* scan the input for line boundaries, 
			 * and process each line 
			 */
			line = discovery_buffer;

			while (line < input) {
				if ((eol = strchr(line, '\n'))) {
					*eol = '\0';

					debugmsg(7,
						 "processing pipe line %p: %s",
						 line, line);

					if (flag) {
						/* We have already scanned for 
						 * 'D' data in pipe, so 
						 * advance the line pointer 
						 */
						line++;
						flag = 0;
					}
					if (no_data) {
						/* No Sendtargets data received
						 * in the pipe 
						 */
						no_data = 0;
						return;
					}

					if (eol == line) {
						/* ignore blank lines */
					} else if (strcmp(line, ";") == 0) {
						/* a target record has been 
						 * terminated 
						 */
						if (new_target) {
							new_target->
							    new_portals =
							    first_portal;

							/* create the per-target
							 * configuration 
							 */
							target_config =
							    create_target_config
							    (new_target->
							     TargetName,
							     first_portal,
							     config,
							     discovery_auth_options);
							if (target_config) {
					/* replace any unprocessed new config 
					 * with the latest one 
					 */
								if (new_target->
								    new_config)
								{
									debugmsg
									    (6,
									     "freeing unused new target config %p for new target %p",
									     new_target->
									     new_config,
									     new_target);
									free_target_config
									    (new_target->
									     new_config);
								}

								new_target->
								    new_config =
								    target_config;

								targets->
								    check_configs
								    = 1;

								debugmsg(6,
									 "new target %p has new target config %p and portals %p",
									 new_target,
									 target_config,
									 first_portal);
							} else {
								logmsg(AS_ERROR,
								       "failed to allocate new target config for new target %p",
								       new_target);
							}

							/* add the target to 
							 * the list 
							 */
							add_target(targets,
								   new_target);

							target = new_target =
							    NULL;
							first_portal =
							    last_portal = NULL;
						} else if (existing_target) {
							int make_config = 0;

							/* this target already 
							 * exists, possibly 
							 * update it 
							 */
							if (!same_portal_descriptors(first_portal, existing_target->current_portals)) {
								debugmsg(1,
									 "target %p portal descriptors have changed\n",
									 existing_target);
								make_config = 1;
								/* stay with 
								 * any existing 
								 * auth 
								 * credentials 
								 * unless the 
								 * discovery 
								 * overrides 
								 */
								if (existing_target->new_config)
									auth_options
									    =
									    &existing_target->
									    new_config->
									    auth_options;
								else if
								    (existing_target->
								     current_config)
									auth_options
									    =
									    &existing_target->
									    current_config->
									    auth_options;
								else
									auth_options
									    =
									    NULL;
							} else {

								for (port =
								     existing_target->
								     current_portals;
								     port;
								     port =
								     port->next)
									debugmsg
									    (7,
									     "target %p portal descriptors have not changed\n",
									     existing_target);
							}
/* make a new config if we need to propagate different auth credentials.
 * to prevent flapping if the same target is found by multiple discovery 
 * processes, we only let the first process to discover the target pass 
 * along auth credentials.
 * A config reload resets this, so that new credentials can be obtained 
 * after a reload. Also, when the owning process stops discovering that 
 * target, it will orphan it so that any other discovery processes that 
 * still find it can propagate discovery credentials to it.
 */
							if (discovery
							    &&
							    ((target->
							      discovery == NULL)
							     || (target->
								 discovery ==
								 discovery))) {
								/* check if the credentials need to change */
								if (existing_target->current_config && discovery_auth_options) {
									if (memcmp(&existing_target->current_config->auth_options, discovery_auth_options, sizeof (existing_target->current_config->auth_options))) {
										debugmsg
										    (1,
										     "target %p discovered by %p needs to update to new auth credentials from "
     "discovery %p\n",
										     existing_target,
										     existing_target->
										     discovery,
										     discovery);
										auth_options
										    =
										    discovery_auth_options;
										make_config
										    =
										    1;
									} else
										debugmsg
										    (7,
										     "target %p discovered by %p does not need to update to auth credentials "
     "from discovery %p\n",
										     existing_target,
										     existing_target->
										     discovery,
										     discovery);
								}

/* ensure that only this discovery process gets to propagate 
 * credentials to this target 
 */
								target->
								    discovery =
								    discovery;
							} else
							    if
							    (discovery_auth_options)
							{
								debugmsg(3,
									 "target %p discovered by %p ignoring auth credentials from discovery %p",
									 target,
									 target->
									 discovery,
									 discovery);
							}

							if (make_config) {
								target_config =
								    create_target_config
								    (existing_target->
								     TargetName,
								     first_portal,
								     config,
								     auth_options);
								if (target_config) {
						/* update the new_config */
									if (existing_target->new_config) {
										debugmsg
										    (6,
										     "freeing unused new target config %p for target %p",
										     existing_target->
										     new_config,
										     existing_target);
										free_target_config
										    (existing_target->
										     new_config);
									}
									existing_target->
									    new_config
									    =
									    target_config;

						/* update the new portals */
									if (existing_target->new_portals) {
										debugmsg
										    (6,
										     "freeing unused new portal descriptors %p for target %p",
										     existing_target->
										     new_portals,
										     existing_target);
										free_portal_descriptors
										    (existing_target->
										     new_portals);
									}
									existing_target->
									    new_portals
									    =
									    first_portal;

									targets->
									    check_configs
									    = 1;

									debugmsg
									    (1,
									     "existing target %p has new target config %p and new portals %p",
									     existing_target,
									     target_config,
									     first_portal);
								} else {
									logmsg
									    (AS_ERROR,
									     "failed to allocate new target config for existing target %p",
									     existing_target);

									/* free the unused new portals */
									free_portal_descriptors
									    (first_portal);
								}
							} else {
								/* free the unused new portals */
								free_portal_descriptors
								    (first_portal);
							}

							target =
							    existing_target =
							    NULL;
							first_portal =
							    last_portal = NULL;
						}
					} else if (strcmp(line, ".") == 0) {
				/* we may want to stop reading now.
				 * the discovery processes use these lines to 
				 * mark the end of a group of target records.  
				 * More records may be coming, but it may take 
				 * a while for them to arrive, so we may want 
				 * to stop reading now in order to avoid 
				 * blocking.  For SendTargets, this
				 * corresponds to a non-final PDU boundary.
				 */
						if (eol + 1 < input) {
							debugmsg(7,
								 "discovery %p "
								 "reached "
								 "stopping "
								 "point with "
								 "more input "
								 "remaining, "
								 "continuing",
								 discovery);
						} else {
							debugmsg(7,
								 "discovery %p "
								 "reached "
								 "stopping "
								 "point with "
								 "no more "
								 "input, "
								 "stopping",
								 discovery);
							return;
						}
					} else if (strcmp(line, "!") == 0) {
				/* a round of discovery is complete, mark the
				 * discovery as no longer in progress. For
				 * SendTargets, this means the PDU exchange is
				 * done.
				 */
						if (eol + 1 < input) {
							debugmsg(7,
								 "discovery %p "
								 "round "
								 "completed "
								 "with more "
								 "input "
								 "remaining, "
								 "continuing",
								 discovery);
				/* keep reading, and start a new round of 
				 * discovery.  We stay in progress, 
				 * but we need to mark all of the targets 
				 * we own undiscovered again.
				 */
							for (target =
							     targets->head;
							     target;
							     target =
							     target->next) {
								if (target->
								    discovery ==
								    discovery) {
									debugmsg
									    (3,
									     "discovery %p expecting to rediscover target %p name %s",
									     discovery,
									     target,
									     target->
									     TargetName);
									target->
									    discovered
									    = 0;
								}
							}
						} else {
							int orphans = 0;

							debugmsg(7,
								 "discovery %p "
								 "round "
								 "completed "
								 "with no more "
								 "input "
								 "remaining, "
								 "stopping",
								 discovery);
							discovery->in_progress =
							    0;

				/* orphan any of our targets that we didn't 
				 * rediscover this round, so that some other 
				 * discovery process can claim them.
				 */
							for (target =
							     targets->head;
							     target;
							     target =
							     target->next) {
								if ((target->
								     discovery
								     ==
								     discovery)
								    && !target->
								    discovered)
								{
									debugmsg
									    (3,
									     "discovery process %p orphaning undiscovered target %p name %s",
									     discovery,
									     target,
									     target->
									     TargetName);
									target->
									    discovery
									    =
									    NULL;
									orphans
									    = 1;
								}
							}

							if (orphans) {
								struct iscsi_discovery_process
								*other;
				/* trigger rediscover for all other discovery 
				 * processes, in the hopes that one of them 
				 * will claim the targets we orphaned.
				 */
								for (other =
								     discovery_processes.
								     head;
								     other;
								     other =
								     other->
								     next) {
									if (other != discovery)
										trigger_rediscovery
										    (other);
								}
							}
							return;
						}
					} else
					    if ((strncmp(line, "TN=", 3) == 0)
						|| (strncmp(line, "LC=", 3) ==
						    0)) {
						char *name = line + 3;

				/* start of a target description.
				 * if the target has already been discovered, 
				 * we may need to update it.
				 * if not, we need to add it.
				 */
						if ((existing_target =
						     find_target(name,
								 targets))) {
							target =
							    existing_target;

						/* track which discovery 
						 * process owns this target 
						 */
							if (target->discovery ==
							    NULL) {
								target->
								    discovery =
								    discovery;
								target->
								    discovered =
								    1;
							} else if (target->
								   discovery ==
								   discovery) {
								target->discovered = 1;	/* we found it again */
							}

							if (*line == 'L') {
								struct iscsi_session_config
								*s;

								debugmsg(1,
									 "target %p name %s rediscovered by process %p",
									 target,
									 name,
									 discovery);

								if (target->
								    current_config)
								{
									for (s =
									     target->
									     current_config->
									     sessions;
									     s;
									     s =
									     s->
									     next)
									{
										if (s->process) {
											s->process->lun_inventory_changed;
											session_processes.
											    lun_inventory_changed
											    =
											    1;
											debugmsg
											    (1,
											     "target %p name %s session process %p LUN inventory "
	     "may have changed",
											     target,
											     name,
											     s->
											     process);
										}
									}
								}
							} else if (target->
								   current_portals)
							{
								debugmsg(1,
									 "target %p name %s rediscovered by process %p",
									 target,
									 name,
									 discovery);
							} else {
					/* can this ever happen now? why would 
					 * we have a target with no portals? 
					 */
								debugmsg(1,
									 "target %p name %s discovered by process %p",
									 target,
									 name,
									 discovery);
							}
						} else {
							/* create a new target */
							target =
							    calloc(1,
								   sizeof
								   (*target));
							if (target) {
								new_target =
								    target;
								target->
								    TargetName =
								    strdup
								    (name);
								target->discovery = discovery;	/* record which process found it first */
								target->discovered = 1;	/* make sure it doesn't get orphaned immediately */
								debugmsg(1,
									 "target %p name %s discovered by process %p",
									 new_target,
									 name,
									 discovery);
							} else {
								logmsg(AS_ERROR,
								       "failed to allocate a new target %s",
								       name);
							}
						}
					} else if (strncmp(line, "TT=", 3) == 0) {
						char *tag = line + 3;

						if (portal == NULL)
							portal =
							    alloc_portal_descriptor
							    ();

						if (portal) {
							portal->tag =
							    iscsi_strtoul(tag,
									  NULL,
									  0);
						} else {
							logmsg(AS_ERROR,
							       "failed to allocate a new portal for target %s",
							       target->
							       TargetName);
						}
					} else if (strncmp(line, "TP=", 3) == 0) {
						/* target TCP port number */
						char *port = line + 3;

						if (portal == NULL)
							portal =
							    alloc_portal_descriptor
							    ();

						if (portal) {
							portal->port =
							    iscsi_strtoul(port,
									  NULL,
									  0);
						}
					} else if (strncmp(line, "RA=", 3) == 0) {
						char *address = line + 3;
						struct in_addr addr;

						/* resolved target ip address */
						if (portal == NULL)
							portal =
							    alloc_portal_descriptor
							    ();

						if (portal) {
							/* FIXME: IPv6 */
							if (iscsi_inet_aton
							    (address, &addr)) {
								memcpy(&portal->
								       ip,
								       &addr,
								       4);
								portal->
								    ip_length =
								    4;
							} else {
								logmsg(AS_ERROR,
								       "couldn't get IP address for %s",
								       address);
							}
						}
					} else if (strncmp(line, "TA=", 3) == 0) {
						char *address = line + 3;
						struct in_addr addr;

						/* ip address or DNS name 
						 * of target portal 
						 */
						if (portal == NULL)
							portal =
							    alloc_portal_descriptor
							    ();

						if (portal) {
							portal->address =
							    strdup(address);

							/* if we didn't get an 
							 * RA line, the IP is 
							 * on the TA line 
							 */
							if (portal->ip_length ==
							    0) {
								/* FIXME: IPv6 */
								if (iscsi_inet_aton(address, &addr)) {
									memcpy
									    (&portal->
									     ip,
									     &addr,
									     4);
									portal->
									    ip_length
									    = 4;
								} else {
									logmsg
									    (AS_ERROR,
									     "couldn't get IP address for %s",
									     address);
								}
							}

					/* we now have all the info for this 
					 * portal, so add it to the list and 
					 * get ready for the next portal.
					 */
							if (first_portal) {
								last_portal->
								    next =
								    portal;
								last_portal =
								    portal;
							} else {
								first_portal =
								    last_portal
								    = portal;
							}

							debugmsg(1,
								 "target %p "
								 "portal "
								 "descriptor %p"
								 " = address "
								 "%s, ip "
								 "%u.%u.%u.%u "
								 "length %u, "
								 "port %d, "
								 "tag %d",
								 target, portal,
								 portal->
								 address,
								 portal->ip[0],
								 portal->ip[1],
								 portal->ip[2],
								 portal->ip[3],
								 portal->
								 ip_length,
								 portal->port,
								 portal->tag);

							portal = NULL;
						}
					}

					/* on to the next line */
					line = eol + 1;
				} else {
					/* move any remaining input to the 
					 * beginning of the buffer, and wait 
					 * for more input 
					 */
					size_t bytes = input - line;

					debugmsg(7,
						 "deferring partial pipe line "
						 "%p: %s",
						 line, line);

					memmove(discovery_buffer, line, bytes);
					line = input = discovery_buffer + bytes;
					break;
				}
			}

			if (line >= input)
				input = discovery_buffer;
		}
	}
}

/* make the target's new_config the current_config */
static void
use_new_target_config(struct iscsi_target *target)
{
	struct iscsi_target_config *old_target_config = NULL;
	struct iscsi_session_config *new_sessions, *existing_sessions, *tail;
	struct iscsi_session_config *new_session, *existing_session;
	struct iscsi_session_process *process;
	int new_target_config = 0;

	if (target->new_config == NULL)
		return;

	/* if the target config has changed, we consider all session
	 * configs as changed.  we want to propagate the fact that the
	 * per-target settings have changed to all session processes.
	 */
	if (!same_target_config(target->new_config, target->current_config)) {
		debugmsg(1, "target config %p for target %p to %s has changed",
			 target->new_config, target, target->TargetName);

		new_target_config = 1;
	}

	/* we always switch to the new data structures, even when they're
	 * the same, to simplify the memory management.  This way we
	 * can always free the old stuff.
	 */
	new_sessions = target->new_config->sessions;
	target->new_config->sessions = NULL;

	if (target->current_config) {
		old_target_config = target->current_config;
		existing_sessions = target->current_config->sessions;
		target->current_config->sessions = NULL;
	} else
		existing_sessions = NULL;

	target->current_config = target->new_config;
	target->new_config = NULL;

	/* For simplicity, we use a sequential mapping from the new session 
	 * configs to the existing session configs.  If portals are reported 
	 * in a different order after a portal is added or removed, this may 
	 * cause us to change the config of every existing session just to 
	 * shuffle the portals around. Since portals don't change often, this 
	 * shouldn't cause any real problems.
	 * We iterate over both the new session configs and the existing 
	 * session configs at the same time.  If both exist and have the same 
	 * config, we do nothing.
	 * If both exist with different configs, we call the config_changed() 
	 * hook.
	 * If a new session exists and there is no existing session, we add it.
	 * If an existing session exists and there is no new session, 
	 * we delete it.
	 */
	new_session = new_sessions;
	existing_session = existing_sessions;
	tail = NULL;

	while ((new_session = new_sessions)) {
		/* pop off a new session */
		new_sessions = new_session->next;
		new_session->next = NULL;

		/* we always ignore any unbound new sessions */
		if (new_session->target_id < 0) {
			logmsg(AS_NOTICE, "discarding unbound session %p to %s",
			       new_session, new_session->target->TargetName);

			free_session_config(new_session);
			continue;
		}

		/* append the new session to the existing sessions */
		if (tail) {
			tail->next = new_session;
			tail = new_session;
		} else {
			target->current_config->sessions = tail = new_session;
		}

		/* get a process using this session config */
		if (existing_session && (process = existing_session->process)) {
			/* connect the new_config and the existing process */
			new_session->process = process;
			process->config = new_session;

			if (new_target_config
			    || !same_session_config(new_session,
						    existing_session)) {
				process->config_number++;

				debugmsg(2,
					 "new session config %p is different "
					 "than existing session config %p, "
					 "notifying process %p\n",
					 new_session, existing_session,
					 process);

				/* notify platform-specific code of the 
				 * config change 
				 */
				iscsi_config_changed(process, new_session,
						     existing_session);
			} else {
				debugmsg(7,
					 "new session config %p is the same "
					 "as existing session config %p\n",
					 new_session, existing_session);
			}
		} else {
			/* try to start a new process */
			process = calloc(1, sizeof (*process));

			if (process == NULL) {
				logmsg(AS_ERROR,
				       "failed to allocate session process for "
				       "session config %p",
				       new_session);
				free_session_config(new_session);
				continue;
			}

			/* cross-link them */
			process->config = new_session;
			process->config_number = 1;
			new_session->process = process;

			debugmsg(1,
				 "allocated session process %p for "
				 "session config %p",
				 process, new_session);

			/* do we really need these in a separate list? */
			add_session_process(&session_processes, process);

			/* start the session */
			fork_session_process(process);

			/* and give it a chance to get going */
			usleep(10 * 1000);
		}
		if (existing_session) {
			/* pop the existing session config, and 
			 * go on to the next 
			 */
			debugmsg(6,
				 "popping and freeing existing session "
				 "config %p",
				 existing_session);
			existing_sessions = existing_session->next;
			free_session_config(existing_session);
			existing_session = existing_sessions;
		} else {
			debugmsg(7, "no existing session config");
		}
	}

	/* any other existing sessions need to be killed off */
	while ((existing_session = existing_sessions)) {
		/* pop an existing session */
		existing_sessions = existing_session->next;

		if (existing_session->process) {
			debugmsg(1,
				 "killing existing session process %p "
				 "for bus %d target %d TargetName %s",
				 existing_session, existing_session->iscsi_bus,
				 existing_session->target_id,
				 existing_session->target->TargetName);

			stop_session_process(existing_session->process);
			existing_session->process = NULL;
		}

		/* and free the obsolete session config */
		free_session_config(existing_session);
		existing_session = NULL;
	}

	/* we've already freed every old session, so the old target 
	 * config is no longer needed either 
	 */
	if (old_target_config)
		free_target_config(old_target_config);

	if (target->new_portals) {
		/* we're now done with the target's old portals.  Everything has
		 * been updated to use the new portals, even if they're
		 * equivalent, so that we can free all of the old ones.  
		 */
		if (target->current_portals)
			free_portal_descriptors(target->current_portals);
		target->current_portals = target->new_portals;
		target->new_portals = NULL;
	}
}

int
main(int argc, char *argv[])
{
	struct sigaction action;
	int rc = 0;
	struct iscsi_discovery_process *discovery = NULL;
	struct iscsi_config config;
	struct discovery_poll_info poll_info;
	int notified_session_count = 0;

	memset(&session_processes, 0, sizeof (session_processes));
	memset(&discovery_processes, 0, sizeof (discovery_processes));
	memset(&targets, 0, sizeof (targets));
	memset(&poll_info, 0, sizeof (poll_info));

	/* initialize configuration defaults */
	memset(&config, 0, sizeof (config));
	iscsi_init_config_defaults(&config.defaults);

	/* signal handling */
	memset(&action, 0x0, sizeof (action));
	action.sa_sigaction = NULL;
	action.sa_flags = 0;

	action.sa_handler = sigchld_handler;
	sigaction(SIGCHLD, &action, 0);

	action.sa_handler = sigpipe_handler;
	sigaction(SIGPIPE, &action, 0);

	/* let platform-specific code know that we're starting, 
	 * and let it process arguments 
	 */
	memset(&daemon_config, 0, sizeof (daemon_config));
	iscsi_daemon_starting(argc, argv);

	if (!daemon_config.foreground) {
		pid_t pid;

		/*
		 * Turn into a daemon
		 */
		close(STDIN_FILENO);
		close(STDOUT_FILENO);
		/* leave stderr open for debug messages */
		chdir("/");

	      again:
		if (iscsi_process_should_exit())
			goto done;

		if ((pid = fork())) {
			/* parent exits */
			exit(0);
		} else if (pid < 0) {
			errormsg("fork failed");
			goto again;
		} else {
			/* child starts a new session */
			setsid();
		}
	}

	/* make sure we're the only instance running */
	if (!lock_pid_file()) {
		exit(1);
	}

	/* load the configuration for the first time */
	if (!update_iscsi_config(daemon_config.config_file, &config)) {
		logmsg(AS_ERROR, "failed to load configuration from %s",
		       daemon_config.config_file);
		exit(2);
	}


	/* start discovery processes based on the current config */
	start_discovery_processes(&config, &discovery_processes);

	/* build an array of pollfds containing every discovery process pipe */
	setup_polling(&poll_info, &discovery_processes);

	while (!iscsi_process_should_exit()) {
		if (iscsi_should_reload_config()) {
			if (update_iscsi_config
			    (daemon_config.config_file, &config)) {
				struct iscsi_session_process *process;
				struct iscsi_target *target;

				/* kill all the existing discovery processes */
				stop_discovery_processes(&discovery_processes);

				for (target = targets.head; target;
				     target = target->next) {
					struct iscsi_portal_descriptor *portals;
					struct iscsi_target_config
					    *target_config;

					/* always use the most recent 
					 * portal descriptors 
					 */
					portals =
					    target->new_portals ? target->
					    new_portals : target->
					    current_portals;

					/* let the first discovery process to 
					 * rediscover this target pass on new 
					 * auth credentials 
					 */
					target->discovery = NULL;
					target->discovered = 0;

					/* Propagate the current target 
					 * config's auth settings for now.
					 * These may change when the target 
					 * gets rediscovered, but for now we
					 * want to keep using what we already 
					 * have, since it may have been
					 * propagated from a previous discovery,
					 * rather than a global default.
					 * If it was a global default that 
					 * included the discovery entry, and 
					 * that default has changed, we'll 
					 * switch to the new default when the 
					 * target gets rediscovered. If a 
					 * global default applies to a 
					 * TargetName entry but not the 
					 * discovery entry, we'll probably 
					 * lose it.  Propagation from 
					 * discovery and dynamic config reloads 
					 * largely make it impossible to 
					 * configure auth credentials by 
					 * TargetName, but nobody has really 
					 * asked for that yet either, so as 
					 * long as we train everyone put to 
					 * put global defaults before the 
					 * discovery entries, there shouldn't 
					 * be any problems.
					 */
					target_config =
					    create_target_config(target->
								 TargetName,
								 portals,
								 &config,
								 &target->
								 current_config->
								 auth_options);

					if (target_config) {
						/* replace any unused new config						 */
						if (target->new_config) {
							debugmsg(6,
								 "freeing "
								 "unused new "
								 "target config"
								 " %p for "
								 "target %p "
								 "after config "
								 "reload",
								 target->
								 new_config,
								 target);
							free_target_config
							    (target->
							     new_config);
						}

						target->new_config =
						    target_config;
						targets.check_configs = 1;
					} else {
						logmsg(AS_ERROR,
						       "failed to allocate "
						       "target config for "
						       "target %p after config "
						       "reload\n",
						       target);
					}
				}

				for (process = session_processes.head; process;
				     process = process->next) {
					if (process->failed) {
						/* restart any process that 
						 * failed with no retries.  
						 * We treat a config reload as 
						 * an explicit user request to 
						 * retry the failed session.  
						 */
						debugmsg(1,
							 "restarting failed "
							 "session process %p",
							 process);
						process->restart = 1;
						session_processes.changed = 1;
					} else {
						/* treat a config reload as a 
						 * request for a new LUN probe 
						 */
						debugmsg(1,
							 "assuming LUN "
							 "inventory may have "
							 "changed for session "
							 "process %p",
							 process);
						process->lun_inventory_changed =
						    1;
						session_processes.
						    lun_inventory_changed = 1;
					}
				}

				/* start new discovery processes based on 
				 * the current config 
				 */
				debugmsg(1,
					 "restarting discovery after "
					 "config reload");
				start_discovery_processes(&config,
							  &discovery_processes);
				setup_polling(&poll_info, &discovery_processes);

				/* give discovery a chance to produce some 
				 * results before we start LUN probing.
				 */
				usleep(100 * 1000);
			} else {
				logmsg(AS_ERROR,
				       "failed to reload configuration file %s",
				       daemon_config.config_file);
				/* keep running with what we have */
			}
		}

		if (discovery_processes.changed) {
			int setup_poll = 0;

			discovery_processes.changed = 0;

			/* scan the discovery list and make any changes needed 
			 */
			discovery = discovery_processes.head;
			while (discovery) {
				debugmsg(7,
					 "checking discovery process %p, pid %d"
					 " pipe_fd %d, restart %d, remove %d, "
					 "next %p",
					 discovery, discovery->pid,
					 discovery->pipe_fd, discovery->restart,
					 discovery->remove, discovery->next);

				if (discovery->restart) {
					discovery->restart = 0;
					if (discovery->pid == 0) {
						debugmsg(1,
							 "restarting discovery "
							 "%p",
							 discovery);
						if (discovery->pipe_fd >= 0) {
							close(discovery->
							      pipe_fd);
							discovery->pipe_fd = -1;
						}
						fork_discovery_process
						    (discovery);
						setup_poll = 1;
					}

					discovery = discovery->next;
				} else if (discovery->pipe_fd == -1) {
					struct iscsi_discovery_process *next =
					    discovery->next;
					struct iscsi_target *target;

					/* no point keeping it around if we 
					 * can't read from it.
					 * we don't remove it right after 
					 * SIGCHLD, since we may still
					 * need to drain it's pipe.
					 */
					debugmsg(1,
						 "removing discovery %p, "
						 "pipe has closed",
						 discovery);

					/* any target first discovered by this 
					 * process needs it's discovery pointer
					 * cleared, so that there are no stale 
					 * pointers, and so that the next 
					 * process to discover this target can 
					 * pass along auth credentials.
					 */
					for (target = targets.head; target;
					     target = target->next) {
						if (target->discovery ==
						    discovery) {
							target->discovery =
							    NULL;
						}
					}

					stop_discovery_process(discovery);

					discovery = next;
					setup_poll = 1;
				} else {
					discovery = discovery->next;
				}
			}

			if (setup_poll) {
				/* rebuild the pollfd array */
				debugmsg(7,
					 "rebuilding pollfds after "
					 "discovery changes");
				setup_polling(&poll_info, &discovery_processes);
			}
		}

		if (session_processes.changed) {
			struct iscsi_session_process *process;

			session_processes.changed = 0;

			/* scan the session processes and make any 
			 * changes needed 
			 */
			process = session_processes.head;
			while (process) {
				if (process->remove) {
					struct iscsi_session_process *next =
					    process->next;

					stop_session_process(process);

					process = next;
				} else if (process->restart) {
					debugmsg(1,
						 "restarting session "
						 "process %p",
						 process);
					process->restart = 0;
					process->failed = 0;
					if (process->pid == 0)
						fork_session_process(process);

					process = process->next;
				} else {
					process = process->next;
				}
			}
		}

		if (iscsi_process_should_exit())
			goto done;

		/* wait forever a discovery process pipe to have 
		 * input for us, or a signal 
		 */
		debugmsg(7, "main daemon polling %d pipes",
			 poll_info.valid_length);
		rc = poll(poll_info.poll_fds, poll_info.valid_length, -1);

		if (rc < 0)
			debugmsg(7,
				 "daemon returned from poll, rc %d, "
				 "errno %d, %s",
				 rc, errno, strerror(errno));
		else
			debugmsg(7, "daemon returned from poll, rc %d", rc);

		while (rc > 0) {
			int n;
			int setup_poll = 0;

			if (iscsi_process_should_exit())
				goto done;

			/* if one or more pipes has discovery info, 
			 * process them all 
			 */
			for (n = 0; n < poll_info.valid_length; n++) {
				if (poll_info.poll_fds[n].
				    revents & (POLLIN | POLLPRI)) {
					/* process discovery info from this 
					 * pipe, and update sessions based 
					 * on it 
					 */
					debugmsg(7,
						 "input on fd %d from discovery"
						 " process %p",
						 poll_info.poll_fds[n].fd,
						 poll_info.processes[n]);

					update_targets(poll_info.processes[n],
						       &config, &targets);
				}

				if (poll_info.poll_fds[n].revents & POLLHUP) {
					debugmsg(1,
						 "poll hangup on index %d, fd "
						 "%d, discovery %p",
						 n, poll_info.poll_fds[n].fd,
						 poll_info.processes[n]);

					if (poll_info.processes[n]->pipe_fd >=
					    0) {
						close(poll_info.processes[n]->
						      pipe_fd);
						poll_info.processes[n]->
						    pipe_fd = -1;
						discovery_processes.changed = 1;
					/* make sure we remove it later */
					}
					usleep(20 * 1000);
					setup_poll = 1;
				}
				if (poll_info.poll_fds[n].revents & POLLNVAL) {
					logmsg(AS_ERROR,
					       "invalid poll on index %d, fd "
					       "%d, discovery %p",
					       n, poll_info.poll_fds[n].fd,
					       poll_info.processes[n]);
					usleep(20 * 1000);
					setup_poll = 1;
				}
				if (poll_info.poll_fds[n].revents & POLLERR) {
					logmsg(AS_ERROR,
					       "poll error on index %d, fd "
					       "%d, discovery %p",
					       n, poll_info.poll_fds[n].fd,
					       poll_info.processes[n]);
					usleep(20 * 1000);
					setup_poll = 1;
				}

				poll_info.poll_fds[n].revents = 0;
			}

			if (setup_poll) {
				/* rebuild the pollfd array */
				debugmsg(7,
					 "rebuilding poll info after poll "
					 "revents");
				setup_polling(&poll_info, &discovery_processes);
			}

			/* non-blocking poll to check for more input before 
			 * processing any config changes or new targets, so 
			 * that we tend to do that all at once.
			 */
			rc = poll(poll_info.poll_fds, poll_info.valid_length,
				  0);
			if (rc < 0)
				debugmsg(7,
					 "daemon returned from non-blocking "
					 "poll, rc %d, errno %d, %s",
					 rc, errno, strerror(errno));
			else
				debugmsg(7,
					 "daemon returned from non-blocking "
					 "poll, rc %d",
					 rc);
		}

		if (iscsi_process_should_exit())
			goto done;

		if (targets.check_configs) {
			struct iscsi_target *target;

			targets.check_configs = 0;

			/* get bindings for every new session config */
			debugmsg(1, "getting bindings for new session configs");
			iscsi_get_bindings(&targets,
					   daemon_config.bindings_file);
			/* and try to switch to the new configs */
			for (target = targets.head; target;
			     target = target->next) {
				if (target->new_config)
					use_new_target_config(target);
			}
		}

		if (iscsi_process_should_exit())
			goto done;

		/* check for LUNs changing after starting any new sessions, 
		 * so that they're more likely to get probed in the proper 
		 * order.
		 */
		if (session_processes.lun_inventory_changed) {
			struct iscsi_session_process *process;

			session_processes.lun_inventory_changed = 0;

			for (process = session_processes.head; process;
			     process = process->next) {
				if (process->lun_inventory_changed
				    && !process->failed) {
					debugmsg(2,
						 "LUNs changed for session "
						 "process %p target %p to %s",
						 process,
						 process->config->target,
						 process->config->target->
						 TargetName);
					iscsi_lun_inventory_changed(process);
				}
			}
		}

		/* if the session count has changed, notify the 
		 * platform-specific code 
		 */
		if (notified_session_count != session_processes.count) {
			int in_progress = 0;
			int not_in_progress = 0;

			for (discovery = discovery_processes.head; discovery;
			     discovery = discovery->next) {
				if (discovery->in_progress)
					in_progress++;
				else
					not_in_progress++;
			}

			debugmsg(2,
				 "session count changed from %d to %d, "
				 "discovery in progress %d, not in "
				 "progress %d\n",
				 notified_session_count,
				 session_processes.count, in_progress,
				 not_in_progress);
			iscsi_session_count_changed(session_processes.count,
						    in_progress,
						    not_in_progress);
			notified_session_count = session_processes.count;
		}
	}

      done:
	/* we're done */
	debugmsg(1, "daemon stopping");
	iscsi_daemon_stopping();

	/* stop discovery */
	stop_discovery_processes(&discovery_processes);

	/* stop any sessions */
	stop_session_processes(&session_processes);


	/* kill any other processes */
	debugmsg(1, "daemon killing process group");
	kill(0, SIGTERM);
	
	return 0;
}
