/*
 * iSCSI driver trace for Linux
 * Copyright (C) 2001 Cisco Systems, Inc.
 * maintained by linux-iscsi-devel@lists.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsigt.c,v 1.13.10.3 2003/12/16 05:59:53 smhatre Exp $ 
 *
 * iscsigt.c
 *
 *     User program to retrieve the iSCSI driver's traces
 *
 */

#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#ifdef _POSIX_MEMLOCK
#  include <sys/mman.h>
#endif

#define DEBUG_TRACE         1
#include "iscsi-ioctl.h"
#include "iscsi-trace.h"

void
dump_trace(struct iscsi_trace_entry *te)
{
	switch (te->type) {
	case 0:
		/* entry not valid */
		return;
	case ISCSI_TRACE_Qd:
		printf("Cmd Queued     ");
		break;
	case ISCSI_TRACE_QFailed:
		printf("Cmd Q failed   ");
		break;
	case ISCSI_TRACE_QSessionLockFailed:
		printf("Q sess failed  ");
		break;
	case ISCSI_TRACE_QCmndLockFailed:
		printf("Q cmnd failed  ");
		break;
	case ISCSI_TRACE_TxCmd:
		printf("Tx Cmd         ");
		break;
	case ISCSI_TRACE_RxCmd:
		printf("Rx Cmd         ");
		break;
	case ISCSI_TRACE_RxCmdStatus:
		printf("Rx Cmd/Status  ");
		break;
	case ISCSI_TRACE_RxUnderflow:
		printf("Rx Underflow   ");
		break;
	case ISCSI_TRACE_RxOverflow:
		printf("Rx Overflow    ");
		break;
	case ISCSI_TRACE_HostUnderflow:
		printf("Host Underflow ");
		break;
	case ISCSI_TRACE_RxAsyncEvent:
		printf("Rx Async Event ");
		break;
	case ISCSI_TRACE_TxData:
		printf("Tx Data Burst  ");
		break;
	case ISCSI_TRACE_TxDataPDU:
		printf("Tx Data PDU    ");
		break;
	case ISCSI_TRACE_RxData:
		printf("Rx Data        ");
		break;
	case ISCSI_TRACE_RxDataCmdStatus:
		printf("Rx Data/Status ");
		break;
	case ISCSI_TRACE_TxAbort:
		printf("Tx Abort       ");
		break;
	case ISCSI_TRACE_RxAbort:
		printf("Rx Abort       ");
		break;
	case ISCSI_TRACE_TxAbortTaskSet:
		printf("Tx Abort Tasks ");
		break;
	case ISCSI_TRACE_RxAbortTaskSet:
		printf("Rx Abort Tasks ");
		break;
	case ISCSI_TRACE_TxLunReset:
		printf("Tx LUN Reset   ");
		break;
	case ISCSI_TRACE_RxLunReset:
		printf("Rx LUN Reset   ");
		break;
	case ISCSI_TRACE_TxWarmReset:
		printf("Tx Warm Reset  ");
		break;
	case ISCSI_TRACE_RxWarmReset:
		printf("Rx Warm Reset  ");
		break;
	case ISCSI_TRACE_TxColdReset:
		printf("Tx Cold Reset  ");
		break;
	case ISCSI_TRACE_RxColdReset:
		printf("Rx Cold Reset  ");
		break;
	case ISCSI_TRACE_CmdDone:
		printf("Cmd Done       ");
		break;
	case ISCSI_TRACE_CmndAborted:
		printf("Cmnd Aborted   ");
		break;
	case ISCSI_TRACE_TaskAborted:
		printf("Task Aborted   ");
		break;
	case ISCSI_TRACE_R2T:
		printf("R2T            ");
		break;
	case ISCSI_TRACE_TxNopReply:
		printf("Tx Nop reply   ");
		break;
	case ISCSI_TRACE_TxPing:
		printf("Tx Nop/Ping    ");
		break;
	case ISCSI_TRACE_RxNop:
		printf("Rx Nop         ");
		break;
	case ISCSI_TRACE_RxPingReply:
		printf("Rx Ping Reply  ");
		break;
	case ISCSI_TRACE_BadOffset:
		printf("Rx Bad Offset  ");
		break;
	case ISCSI_TRACE_BadRxSeg:
		printf("Rx Bad Segment ");
		break;
	case ISCSI_TRACE_OutOfData:
		printf("Tx Out of Data ");
		break;
	case ISCSI_TRACE_BadTxSeg:
		printf("Tx Bad Segment ");
		break;
	default:
		printf("Unknown 0x%3x  ", te->type);
		break;
	}

	printf("%3u %3u %3u %3u  0x%02x %10u %10lu %10lu %10lu\n",
	       te->host, te->channel, te->target, te->lun, te->cmd,
	       te->itt, te->data1, te->data2, te->jiffies);

	return;
}

int
main(int argc, char **argv)
{
	FILE *f;
	int rc, devn, ctlfd, ti, i;
	char buf[80], devname[80];
	struct iscsi_trace_dump *dump = NULL;
	uint32_t num_entries = ISCSI_TRACE_COUNT;

	/*
	 * Establish the control path to the driver
	 */
	f = fopen("/proc/devices", "r");
	if (!f) {
		fprintf(stderr, "Cannot open control path to the driver.\n");
		return 2;
	}

	devn = 0;
	while (!feof(f)) {
		if (!fgets(buf, sizeof (buf), f)) {
			break;
		}
		if (sscanf(buf, "%d %s", &devn, devname) != 2) {
			continue;
		}
		if (!strcmp(devname, "iscsictl")) {
			break;
		}
		devn = 0;
	}

	fclose(f);
	if (!devn) {
		fprintf(stderr, "Cannot find control file to the driver.\n");
		return 3;
	}

	ctlfd = open("/dev/iscsictl", O_RDWR);
	if (ctlfd < 0) {
		fprintf(stderr, "Cannot open control node.\n");
		return 6;
	}

	do {
		/*
		 * allocate a trace dump for the ioctl.
		 */
		dump =
		    malloc(sizeof (dump) +
			   (num_entries * sizeof (struct iscsi_trace_entry)));
		if (!dump) {
			fprintf(stderr,
				"Couldn't allocate trace dump with "
				"%u entries\n",
				num_entries);
			return 7;
		}
		memset(dump, 0x0,
		       sizeof (dump) +
		       (num_entries * sizeof (struct iscsi_trace_entry)));
		dump->dump_ioctl_size = sizeof (*dump);
		dump->dump_version = TRACE_DUMP_VERSION;
		dump->trace_entry_size = sizeof (struct iscsi_trace_entry);
		dump->num_entries = num_entries;

		/* try to lock all pages in memory, so that no page faults can
		 * occur when the ioctl is copying data into our buffer.  We
		 * don't want to block while the ioctl is holding a lock,
		 * since it can stall the driver.
		 */
#ifdef _POSIX_MEMLOCK
		if (mlockall(MCL_CURRENT)) {
			fprintf(stderr, "failed to mlockall, %d, %s\n", errno,
				strerror(errno));
		}
#endif

		/* try to get the trace info */
		rc = ioctl(ctlfd, ISCSI_GETTRACE, (void *) dump);
		if (rc < 0) {
			if (errno == E2BIG) {
				/* need a bigger buffer */
				fprintf(stderr,
					"Cannot get %u trace entries, "
					"kernel has %u\n",
					num_entries, dump->num_entries);
				num_entries = dump->num_entries;
				free(dump);
				dump = NULL;
			} else {
				perror
				    ("Cannot get trace entries from "
				     "iSCSI driver ");
			}
		}
	} while (!dump);

	if (dump && (rc >= 0)) {
		ti = rc;
		printf("received %d trace entries, starting index %d\n",
		       dump->num_entries, rc);
		printf
		    ("Trace tag      Hst Chn Tgt Lun   Cmd      ITT      "
		     "Data1      Data2     Jiffies\n");
		printf
		    ("=============================================="
		     "==================================\n");
		for (i = 0; i < dump->num_entries; i++) {
			dump_trace(&dump->trace[ti]);
			ti++;
			if (ti >= dump->num_entries)
				ti = 0;
		}
	}

	close(ctlfd);

	return 0;
}
