/*
 * iSCSI LUN activation
 * Copyright (C) 2001 Cisco Systems, Inc.
 * maintained by linux-iscsi@cisco.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: iscsilun.c,v 1.14.16.1 2003/10/07 13:03:11 krishmnc Exp $ 
 *
 * iscsilun.c
 *
 *     User program to dynamically activate iSCSI LUNs.
 * 
 */

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <ctype.h>

#include "iscsi-protocol.h"
#include "iscsi-common.h"

/* what we find in /proc/scsi/iscsi */
struct proc_entry {
	struct proc_entry *next;
	int hostno;
	int channel;
	int target;
	int lun;
	int remtarget;
	char ipaddr[32];
	char targetname[TARGET_NAME_MAXLEN];
};

/* Target,Lun= entries in the config file /etc/iscsi.conf */
struct target_lun_entry {
	struct target_lun_entry *next;
	int target;
	int lun;
};

const char *config_file = "/etc/iscsi.conf";

void
ActivateLun(struct proc_entry *pe, int lun)
{
	FILE *fdout;
	char outbuf[128];

	fdout = fopen("/proc/scsi/scsi", "w");
	if (!fdout) {
		perror("Cannot open file, /proc/scsi/scsi\n");
		return;
	}

	/* announce what we're activating */
	printf("activating:  %3d %3d %3d %3d             %s %3d %s\n",
	       pe->hostno, pe->channel, pe->target, lun, pe->ipaddr,
	       pe->remtarget, pe->targetname);

	/* tell SCSI to add the LUN */
	sprintf(outbuf, "scsi add-single-device %d %d %d %d\n",
		pe->hostno, pe->channel, pe->target, lun);

	fputs(outbuf, fdout);

	fclose(fdout);

	/* give the LUN time to respond, since Linux SCSI device nodes
	 * are allocated in discovery order, and we'd like to keep the order
	 * relatively consistent.
	 */
	usleep(30 * 1000);
}

struct target_lun_entry *
GetConfig(int *numremp)
{
	FILE *f;
	int numrem;
	unsigned int target, lun;
	struct target_lun_entry *tle, *ftle, *ltle;
	char *rdbufp, rdbuf[1024];

	f = fopen(config_file, "r");
	if (!f) {
		return NULL;
	}

	numrem = 0;
	ftle = ltle = NULL;
	do {
		rdbufp = fgets(rdbuf, sizeof (rdbuf), f);
		if (rdbufp && (*rdbufp != '#')) {
			while (isspace(*rdbufp)) {
				rdbufp++;
			}
			/* if it looks like Target,Lun=x,y */
			if (sscanf(rdbufp, "Target,Lun=%u,%u", &target, &lun) ==
			    2) {
				tle = (struct target_lun_entry *)
				    malloc(sizeof (struct target_lun_entry));
				if (tle) {
					tle->target = target;
					tle->lun = lun;
					tle->next = NULL;
					if ((tle->target >= 0) &&
					    (tle->target < ISCSI_MAX_TARGETS) &&
					    (tle->lun >= 0) &&
					    (tle->lun < ISCSI_MAX_LUN)) {
						if (ftle) {
							ltle->next = tle;
						} else {
							ftle = tle;
						}
						ltle = tle;
					}
				}
			} else {
				/* count the number of remotes that 
				 * have appeared 
				 */
				if ((strncmp(rdbufp, "TargetIpAddr=", 13) == 0)
				    || (strncmp(rdbufp, "DiscoveryAddress=", 17)
					== 0)) {
					numrem += 1;
				}
			}
		}
	} while (rdbufp);

	fclose(f);

	if (numremp) {
		*numremp = numrem;
	}

	/* return the first target_lun_entry */
	return ftle;
}

int
main(int argc, char **argv)
{
	int rc, hostno, hitarget, bootflg, foundrem, finished = 0;
	DIR *dir;
	struct dirent *dirv;
	FILE *fdin;
	struct proc_entry *pe, *pehead, *petail;
	struct target_lun_entry *tle_list = NULL, *tle, justonetle;
	int port;
	char *cp, lunch[4], procflnm[80], filebuf[1024];

	hitarget = 0;
	if (argc == 2) {
		/* 1 argument provided */
		if (strcmp(argv[1], "boot") == 0) {
			/* become a daemon later */
			tle_list = GetConfig(&bootflg);
			if (bootflg == 0) {
				fprintf(stderr,
					"No targets in configuration file.\n");
				exit(0);
			}
		} else {
			/* activate the one LUN specified on the command line */
			bootflg = 0;
			tle_list = tle = &justonetle;
			cp = strchr(argv[1], ',');
			if (cp) {
				*cp = '\0';
				cp += 1;
				tle->lun = atoi(cp);
				if ((tle->lun < 0)
				    || (tle->lun >= ISCSI_MAX_LUN)) {
					fprintf(stderr,
						"Invalid argument (%s,%s)\n",
						argv[1], cp);
					return 1;
				}
			}
			tle->target = atoi(argv[1]);
			if ((tle->target < 0)
			    || (tle->target >= ISCSI_MAX_TARGETS)) {
				if (cp) {
					fprintf(stderr,
						"Invalid argument (%s,%s)\n",
						argv[1], cp);
				} else {
					fprintf(stderr,
						"Invalid argument (%s)\n",
						argv[1]);
				}
				return 1;
			}
			tle->next = NULL;
		}
	} else {
		/* no args provided.  Stay in the forground, but 
		 * behave like the daemon 
		 */
		bootflg = 0;
		tle_list = GetConfig(NULL);
	}

	if (bootflg) {
		/*
		 * Turn into a daemon
		 */
		chdir("/");
		if (fork()) {
			exit(0);
		}
		setsid();
	}

	dir = opendir("/proc/scsi/iscsi");
	if (!dir) {
		perror("Cannot open directory, /proc/scsi/iscsi\n");
		return 1;
	}

	do {
		dirv = readdir(dir);
		if (!dirv) {
			perror("Cannot read directory, /proc/scsi/iscsi\n");
			return 1;
		}

		/* find the /proc file for the HBA, which is the host number
		 * assigned by Linux.  Ignore any non-numeric filenames.
		 */
		if ((dirv->d_name[0] >= '0') && (dirv->d_name[0] <= '9')) {
			break;
		}
	} while (dirv);

	closedir(dir);

	hostno = atoi(dirv->d_name);
	sprintf(procflnm, "/proc/scsi/iscsi/%d", hostno);

      ScanAgain:

	fdin = fopen(procflnm, "r");
	if (!fdin) {
		perror("Cannot open file, %s\n", procflnm);
		return 1;
	}

	foundrem = 0;
	pehead = petail = NULL;
	do {
		/* scan /proc/scsi/iscsi for iSCSI info */
		cp = fgets(filebuf, sizeof (filebuf), fdin);
		if (cp) {
			if (*cp != '#') {
				pe = (struct proc_entry *)
				    malloc(sizeof (struct proc_entry));
				if (!pe) {
					fprintf(stderr, "Out of memory\n");
					return 1;
				}
				rc = sscanf(cp, "%d %d %d %s %s %d %s",
					    &pe->hostno, &pe->channel,
					    &pe->target, lunch, pe->ipaddr,
					    &port, pe->targetname);
				if (rc == 7) {
					if (lunch[0] == '?') {
						/* LUNs unknown */
						pe->lun = -1;
					} else {
						/* LUN known */
						pe->lun = atoi(lunch);
					}
					/* record highest target number */
					if (pe->target > hitarget) {
						hitarget = pe->target;
					}

					/* record number of remotes */
					if (!petail
					    || strcmp(petail->ipaddr,
						      pe->ipaddr)) {
						foundrem += 1;
					}

					/* add entry to the list */
					if (pehead) {
						petail->next = pe;
					} else {
						pehead = pe;
					}
					petail = pe;
#ifdef DEBUG_DAEMON
					fprintf(stderr,
						"\niscsilun found proc entry "
					        "for: %d %d %d %s %s %d %s\n",
						pe->hostno, pe->channel,
						pe->target, lunch, pe->ipaddr,
						port, pe->targetname);
#endif
				} else {
					fprintf(stderr, "Bad parse, %d\n", rc);
				}
			}
		}
	} while (cp);

	fclose(fdin);

	/* no remotes found, or more remotes configured than found */
	if ((pehead == NULL) || (bootflg > foundrem)) {
		/* wait for all configured remotes */
		/* FIXME: activate them as soon as we find 
		 * them for better availability 
		 */
		usleep(100 * 1000);
		goto ScanAgain;
	}

	/* FIXME: now that we use one session per target, rather than one 
	 * session per remote, this check really isn't sufficient.  We may 
	 * probe for all targets on a remote as soon as one target is up, 
	 * missing all but one target.  This is just fundamentally broken.  
	 * For now, just sleep for 2 seconds after all remotes have started 
	 * bringing up targets, to give all the targets a chance to come up
	 * before we probe for them.  A better solution is to have each target 
	 * request LUN activation when it is ready.
	 */
	if (!finished) {
		finished = 1;
		sleep(3);
		/* we need to scan /proc one more time to get all of the targets
		 */
		goto ScanAgain;
	}

	if (tle_list) {
		/* Target,Lun entries exist, and limit what we'll probe for */
		tle = tle_list;
		while (tle) {
			pe = pehead;
			while (pe) {
#ifdef DEBUG_DAEMON
				fprintf(stderr,
					"\niscsilun processing proc entry for: "
				        "%d %d %d %d %s %d %s\n",
					pe->hostno, pe->channel, pe->target,
					pe->lun, pe->ipaddr, port,
					pe->targetname);
#endif
				/* compare each line of /proc/scsi/iscsi 
				 * against the Target,Lun entries 
				 */
				if (pe->lun >= 0) {
					/* proc entry with a LUN */
					/* FIXME: so why would we need to tell 
					 * SCSI about it? 
					 * It's the probes we do for ??? lines 
					 * that cause the lines with LUNs to 
					 * exist, as a side-effect of a TUR
					 * command coming back as part of the 
					 * scsi add-single-device.
					 */
					if ((tle->target >= 0) &&
					    (tle->target == pe->target) &&
					    (tle->lun >= 0)) {

						/* Target,Lun exists and 
						 * matches the proc entry 
						 */
						ActivateLun(pe, tle->lun);

						/* make sure we don't activate 
						 * it again later 
						 */
						tle->target = hitarget + 1;
					}
				} else {
					/* proc entry with ??? */
					if (tle->target == pe->target) {
						ActivateLun(pe, tle->lun);
					}
				}
				pe = pe->next;
			}
			tle = tle->next;
		}
	} else {
		/* no Target,Lun entries in config file. */
		pe = pehead;
		while (pe) {
			/* if the proc line has a LUN of ???, probe for all LUNs
			 */
			if (pe->lun < 0) {
				int i;

				for (i = 0; i < ISCSI_DEFAULT_LUNS_TO_PROBE;
				     i++) {
					ActivateLun(pe, i);
				}
			}
			pe = pe->next;
		}
	}

	return 0;
}
