/*
 * iSCSI connection daemon
 * Copyright (C) 2001 Cisco Systems, Inc.
 * maintained by linux-iscsi@cisco.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * See the file COPYING included with this distribution for more details.
 *
 * $Id: textkey.cpp,v 1.2 2001/08/08 15:13:59 smferris Exp $ 
 *
 * textkey.cpp
 *
 *      iSCSI text command block
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "textkey.hpp"


// Text key:value pair
struct TextKeyPair {
    TextKeyPair *next_m;
    TextKeyPair *prev_m;
    bool freeMem_m;
    char *key_m;
    char *value_m;
};


// Construct a text command block from a rcvd packet
TextKey::TextKey(char *zb, int textBlockLen) :
  count_m(0),
  textBlock_m(NULL),
  first_m(NULL),
  last_m(NULL)
{
    if (textBlockLen == 0)
	return;

    // Allocate space to retain a copy of the text block
    int len = textBlockLen + 1;
    if ((textBlock_m = new char[len]) == NULL) {
        fprintf(stderr, "insufficient memory to create %d byte text block\n",
		len);
	return;
    }

    // Copy the text block from the supplied zBuf
    memcpy(textBlock_m, zb, textBlockLen);
    textBlock_m[textBlockLen] = '\0';

    // Create the set of key:val pairs from the text block
    char *key = NULL;
    int j = 0;
    for (int i=0; i < len; i++) {
	if (!key) {
	    if (textBlock_m[i] == '\0')
		j = i + 1;
	    else if (textBlock_m[i] == '=') {
		key = &textBlock_m[j];
		textBlock_m[i] = '\0';
		j = i + 1;
	    }
	} else if (textBlock_m[i] == '\0') {
	    struct TextKeyPair *tp = new struct TextKeyPair;
	    if (tp == NULL) {
		fprintf(stderr, "Insufficient memory to create text pair\n");
		return;
	    }
	    tp->next_m = NULL;
	    tp->prev_m = last_m;
	    tp->freeMem_m = false;
	    tp->key_m = key;
	    tp->value_m = &textBlock_m[j];
	    if (first_m == NULL)
		first_m = tp;
	    if (last_m)
		last_m->next_m = tp;
	    last_m = tp;
	    count_m++;
	    key = NULL;
	    j = i + 1;
	}
    }
}


// Construct an empty text holder
TextKey::TextKey() :
  count_m(0),
  textBlock_m(NULL),
  first_m(NULL),
  last_m(NULL)
{
}

// Destructor
TextKey::~TextKey()
{
    if (textBlock_m)
	delete [] textBlock_m;

    // Free the text pairs
    while (first_m) {
	struct TextKeyPair *next = first_m->next_m;
	if (first_m->freeMem_m) {
	    delete [] first_m->key_m;
	    delete [] first_m->value_m;
	}
	delete first_m;
	first_m = next;
    }
}


// Search for the specified key and return the value if found (NULL if not)
char *
TextKey::find(char *key)
{
    char *val = NULL;
    struct TextKeyPair *tp = first_m;

    while (tp) {
	if (!cmp(key, tp->key_m)) {
	    val = tp->value_m;
	    break;
	}
	tp = tp->next_m;
    }

    return val;
}


// Add an entry
void
TextKey::add(char *key, char *val)
{
    int len = strlen(key);
    struct TextKeyPair *tp = new struct TextKeyPair;
    if (!tp)
	return;
    tp->next_m = NULL;
    tp->prev_m = last_m;
    tp->freeMem_m = true;
    if ((tp->key_m = new char[len+1]) != NULL)
	memcpy(tp->key_m, key, len+1);
    else {
	delete tp;
	return;
    }
    len = strlen(val);
    if ((tp->value_m = new char[len+1]) != NULL)
	memcpy(tp->value_m, val, len+1);
    else {
	delete tp->key_m;
	delete tp;
	return;
    }
    if (last_m)
	last_m->next_m = tp;
    else
	first_m = tp;
    last_m = tp;
}


// Get a copy of the text command block.  The length of the text block
// is returned in len.  If len == 0, NULL is returned.  If len > 0,
// the caller is responsible for freeing the memory returned.
char *
TextKey::get(int &len)
{
    len = 0;

    // Compute the length of buffer we will need
    struct TextKeyPair *tp = first_m;
    while (tp) {
	len += strlen(tp->key_m) + 1;
	len += strlen(tp->value_m) + 1;
	tp = tp->next_m;
    }

    if (!len)
	return NULL;

    // Get a buffer
    char *buff = new char[len];
    if (buff == NULL) {
	len = 0;
	return NULL;
    }
    
    // Fill the buffer
    char *bp = buff;
    tp = first_m;
    while (tp) {
	sprintf(bp, "%s=%s", tp->key_m, tp->value_m);
	bp += strlen(bp) + 1;
	tp = tp->next_m;
    }

    return buff;
}


// Show the key:val pairs stored in this text block
void
TextKey::show(int indent)
{
    struct TextKeyPair *tp = first_m;
    while (tp) {
	for (int i=0; i < indent; i++)
	    printf(" ");
	fprintf(stderr, "%s=%s\n", tp->key_m, tp->value_m);
	tp = tp->next_m;
    }
}


// Compare two strings.  Returns 0 if the same, -1 if not.
int
TextKey::cmp(char *t1, char *t2)
{
    while (*t1 != '\0' && *t2 != '\0') {
	if (tolower(*t1) != tolower(*t2))
	    return -1;
	t1++;
	t2++;
    }

    return (*t1 == '\0' && *t2 == '\0') ? 0 : -1;
}
