/*
 *
 * This module is an interface to the ZR36060 module to mgacodec.c
 *
 * Copyright (C) 1999  Mike Bernson <mike@mlb.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *****************************************************************************
 *
 * Modified and extended by
 *	Leon van Stuivenberg <leonvs@iae.nl>
 */

#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/signal.h>
#include <linux/interrupt.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <linux/sched.h>
#include <asm/segment.h>
#include <linux/types.h>
#include <linux/proc_fs.h>

#include <linux/spinlock.h>
#include <linux/videodev.h>
#include <linux/version.h>
#include <asm/uaccess.h>

#include "aiee.h"
#include "mgacore.h"
#include "mgacodec.h"
#include "zr36060.h"
#include "mgavideo.h"
#include "ks0127.h"
#include "media/audiochip.h"


#if LINUX_VERSION_CODE >= 0x020100
MODULE_PARM(debug,"i");
#endif

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif


struct	i33_dev_s {
	int			last_input;
	int			last_norm;

	struct	mga_dev_s	*mga;
	char			device_name[32];

	enum buz_codec_mode	mode;
	struct	work_struct	compress;
	long			compress_active;

	struct	work_struct	decompress;
	long			decompress_active;

	struct	timeval		timestamp[32];
	u32			lcode[32];
	u8			field[32];
	u32			hardptr[32];
	u32			snapshot_head;
	u32			snapshot_tail;
	
	u32			field_time;
	u32			compress_started;

	u32			buffer_len[256];
	int			buffer_status[256];
	u32			buffer_head;
	u32			buffer_tail;
	u32			last_stream;
	u32			last_hardptr;
	u32			last_lcode;
	
	int			empty_field_count;
};

static struct tvnorm zrnorms[] = {
   /* PAL-BDGHI */
   { 864, 720, 625, 576, HStart:64, VStart:18, Hjpeg:14, Vjpeg:14 },

   /* NTSC */
     { 858, 720, 525, 480, HStart:68, VStart:14, Hjpeg:9, Vjpeg:12 },

   /* SECAM: like PAL-BDGHI */
   { 864, 720, 625, 576, HStart:64, VStart:18, Hjpeg:14, Vjpeg:14 },
};

static void i33_fill_stream(void *cte_private);

static void i33_vsync_compress(struct mga_dev_s *mga);


static int debug = 0;
#define dprintk     if (debug) printk



/*	miscctl, for (at least) marvel G200 & marvel G400:
	bit 0:	?
	bit 1:	?
	bit 2:	1-> ZR36060.CS asserted		0->ZR36060.RESET asserted
	bit 3:	1-> ZR36060.SLEEP asserted
	bit 4:	1-> ZR36060.START asserted
	bit 5:					0-> ZR36700.RESET asserted
	bit 6:					0-> ZR36700.STNDBY asserted
	bit 7:	1-> ZR36700.HCS asserted
*/

static void
i33_powerup( struct i33_dev_s *i33, int flag )
{
	struct	mga_codec_s	*codec = i33->mga->codec;
	/* note: all delays are (still) over-dimensioned. */

	/* pulse the reset line */
	mgacodec_and_or_miscctl(codec, ~0x1c, 0x04);	/* put lines where they should have been already */
	mdelay(1);					/* wait a bit... */
	mgacodec_and_or_miscctl(codec, ~0x1c, 0x00);	/* assert RESET (SLEEP must be deasserted)*/
	udelay(5);					/* wait at least 8 Vclkx2, or 8/27M sec (.3u) */
	mgacodec_and_or_miscctl(codec, ~0x1c, 0x04);	/* deassert RESET, assert CS */
	udelay(10);					/* wait for stable Vclkx2;
							arbitrary delay, since Vclkx2 is always stable? */
	if (flag == 0) 	
		return;

	/* pulse the sleep line */
	mgacodec_and_or_miscctl(codec, ~0x1c, 4|8);	/* assert SLEEP */
	udelay(5);					/* keep asserted for at least 8 Vclkx2 (0.3usec)*/
	mgacodec_and_or_miscctl(codec, ~0x1c, 0x04);	/* deassert SLEEP */
	mdelay(1);					/* wait for lock: 5000 Vclkx2 cycles! (185usec)*/
}


static void
i33_powerdown( struct i33_dev_s *i33 )
{
	struct	mga_codec_s	*codec = i33->mga->codec;
	
	/* CS=0, RESET=1, SLEEP=0, START=0 */
	mgacodec_and_or_miscctl(codec, ~0x1c, 0);
	udelay(1); /* wait at least 8 Vclkx2, or 8/27M sec (.3u) */
}


#define COMP_AHEAD_THRES 4096

static void
i33_callback(struct mga_dev_s *mga, int state)
{
	dprintk( "i33_callback %i\n", state );
	
	switch(state) {
	case MGACODEC_START:
		mgacodec_and_or_miscctl(mga->codec, ~0, 0x10);
		break;

	case MGACODEC_STOP:
		/* codec (near) under/overrun; stall transfer */
		printk("i33 STOP, hard %6x\n", mgacodec_hardptr(mga->codec) );
		mgacodec_stall_transfer(mga->codec);
		break;

	default:
		printk(KERN_ERR "%s: unknown callback\n", mga->i33->device_name);
	}
}


static void
i33_enable(struct i33_dev_s *i33)
{
	struct	mga_codec_s	*codec = i33->mga->codec;

	/* note: KS0127 needs to be on first; it gives the i33 its clock */
	/* reset codec interface to correct protocol */
	mgacodec_init(codec, MGACODEC_MODE_I33, 0,
		(i33->mga->options & MGAVIDEO_HAS_RRG) != 0 ? 0x03 : 0 );
	i33_powerup(i33, 1);
	mgacodec_callback(codec, i33_callback);
}


static void
i33_disable(struct i33_dev_s *i33)
{
	struct	mga_codec_s	*codec = i33->mga->codec;

	mgacodec_callback(codec, NULL);
	i33_powerdown(i33);
	mgacodec_reset(codec);
}


static void
i33_reset(void *cte_private, int flag)
{
        struct  i33_dev_s       *i33;
        struct  mga_codec_s     *codec;

        i33 = cte_private;
        codec = i33->mga->codec;

	i33_enable(i33);
        mgacodec_reset( codec );
        i33_powerup( i33, flag );
}


static int
i33_read_memory(void *cte_private, void *memory, int start_reg, int size)
{
	int			i;
	struct	mga_codec_s	*codec;
	struct  i33_dev_s	*i33;

	i33 = cte_private;
	codec = i33->mga->codec;

	for(i=0; i < size; i++, start_reg++) {
		mgacodec_push_writeb(codec, 1, start_reg >> 8 );
		mgacodec_push_writeb(codec, 2, start_reg & 0xff);
		mgacodec_push_readb(codec, 3);

		if (mgacodec_exec_commands(codec, ~0, 0) != 0)
			return 1;

		*(u8 *)memory++  = mgacodec_get_byte(codec);
	}
	return 0;
}


static int
i33_write_memory(void *cte_private, void *memory, int start_reg, int size)
{
	int			i;
	struct	mga_codec_s	*codec;
	struct  i33_dev_s	*i33;

	i33 = cte_private;
	codec = i33->mga->codec;

	for(i=0; i < size; i++, start_reg++) {
		if (i == 0 || (start_reg&0xff) == 0)
			mgacodec_push_writeb(codec, 1, start_reg >> 8 );
		mgacodec_push_writeb(codec, 2,  start_reg & 0xff);
		mgacodec_push_writeb(codec, 3 ,*(u8 *)memory++);

		/* cmdbuf can hold 32 writes; per cycle we write 2 x reg msb (max),
		   N x reg lsb, and N x data. 2N+2<=32 -> N=15 */
		if (((size-i-1)%15) == 0)
			if (mgacodec_exec_commands(codec, ~0, 0) != 0)
				return 1;
	}
	return 0;
}




static void
i33_start_transfer(void *cte_private, enum buz_codec_mode mode)
{
	struct	mga_codec_s	*codec;
	struct  i33_dev_s	*i33;


	i33 = cte_private;
	codec = i33->mga->codec;

	if (mgacodec_get_transfer(codec) != CODEC_NO_TRANSFER)
		return;

	switch(mode) {
	case BUZ_MODE_MOTION_COMPRESS:
	case BUZ_MODE_STILL_COMPRESS:
		dprintk("%s: starting transfer\n", i33->device_name);
		mgacodec_set_transfer_limit(codec, -COMP_AHEAD_THRES );
		mgacodec_start_transfer(codec, CODEC_RECV_TRANSFER);
		break;

	case BUZ_MODE_MOTION_DECOMPRESS:
	case BUZ_MODE_STILL_DECOMPRESS:
		dprintk("%s: starting transfer\n", i33->device_name);
		i33_fill_stream(i33);
		mgacodec_start_transfer(codec, CODEC_SEND_TRANSFER);
	        break;

	case BUZ_MODE_IDLE:
		printk(MGA_WARNING "%s: start_transfer called will codec is idle\n",
		       i33->device_name);
		break;
	}
}


static void
i33_vsync_decompress(struct mga_dev_s *mga)
{
	struct zr36060_dev *zr;
	struct i33_dev_s *i33;
	struct timeval	stamp;
	u32	hardptr;
	u8	field;

	zr = mga->zr;
	i33 = mga->i33;

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,4,17)
	get_fast_time(&stamp);
#else 	
	do_gettimeofday(&stamp);
#endif	
	hardptr = mgacodec_hardptr(mga->codec);
	
	if ((i33->mga->status & VS_FIELD) == VS_TOPFIELD)
		field = 1;
	else
		field = 2;

	i33->timestamp[i33->snapshot_head] = stamp;
	i33->hardptr[i33->snapshot_head] = hardptr;
	i33->field[i33->snapshot_head] = field;
	
	i33->snapshot_head = (i33->snapshot_head+1) & 31; /* update head atomic */

	dprintk("%s: decompress vsync hardptr = %x, field %i\n", 
		i33->device_name, mgacodec_hardptr(mga->codec), 
		(mga->status>>8)&1);

	if (mgacodec_get_transfer(mga->codec) == CODEC_NO_TRANSFER)
		return;

        schedule_work(&mga->i33->decompress);
}


static void
i33_vsync_compress(struct mga_dev_s *mga)
{
	struct zr36060_dev *zr;

	struct i33_dev_s *i33;
	int	diff;
	u32	hardptr;
	u32	lcode;
	u8	field;
	struct timeval	stamp;
	struct timeval *last_stamp;
	int timely;
	int dif1, dif2;
	zr = mga->zr;
	i33 = mga->i33;
	
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,4,17)
	get_fast_time(&stamp);
#else 	
	do_gettimeofday(&stamp);
#endif

	last_stamp = &i33->timestamp[(i33->snapshot_head -1) & 31];

	diff = (stamp.tv_sec - last_stamp->tv_sec) * 1000000 +
		(stamp.tv_usec - last_stamp->tv_usec);

	lcode = mgacodec_lcode(mga->codec);
	hardptr = mgacodec_hardptr(mga->codec);

	if ((i33->mga->status & VS_FIELD) == VS_TOPFIELD)
		field = 1;
	else
		field = 2;

	/* make field correction if this isr runs before the lcode of last field
	   makes it through the codec interface */
	dif1 = hardptr - lcode;                         /* hardptr always comes after lcode */
	if (dif1 < 0) dif1 += CODEC_BUFFER_SIZE;
	dif2 = lcode - i33->last_hardptr;               /* lcode always comes after old hardptr */
	if (dif2 < 0) dif2 += CODEC_BUFFER_SIZE;

	if ( dif1 > dif2 ) {
		/* new lcode and last_hardptr (and therefor previous vstatus) belong together,
		reverse detected field. */
		field ^= 3;
	}
		
	timely = ((diff > 16000 && diff < 17000)  || 
              (diff > 19500 && diff < 20500));

	i33->timestamp[i33->snapshot_head] = stamp;
	i33->lcode[i33->snapshot_head] = lcode;
	i33->field[i33->snapshot_head] = field;

	i33->snapshot_head = (i33->snapshot_head+1) & 31; /* update head atomic */
	
	i33->last_hardptr = hardptr;

	if (!timely && i33->timestamp[(i33->snapshot_head-1) & 31].tv_sec) {
		dprintk("%s: interrupt missed %d\n", i33->device_name, diff);
	}

	dprintk("%s: vsync comp hptr = %5x, lcode %5x\n", 
		i33->device_name, mgacodec_hardptr(mga->codec),
		mgacodec_lcode(mga->codec) );

	if (mgacodec_get_transfer(mga->codec) == CODEC_NO_TRANSFER)
		return;

        schedule_work(&mga->i33->compress);
}


static void
i33_get_stream(void *data)
{
        struct i33_dev_s *i33 = data;
	int		count;
	int		left;
	int		size;
	int		cur_field;
	u8		*mem;
	struct mga_codec_s *codec;
	struct mga_dev_s *mga;
	u32		cur_lcode;
	int		copied;
	struct timeval	cur_stamp;
	u32	limit;


	mga = i33->mga;
	codec = i33->mga->codec;

	if (test_and_set_bit(0, &i33->compress_active)) {
		printk(MGA_ERR "%s: get_stream already active\n", i33->device_name);
		return;
	}
	
	while(i33->snapshot_tail != i33->snapshot_head) {
		cur_lcode = i33->lcode[i33->snapshot_tail];
		cur_field = i33->field[i33->snapshot_tail];
		cur_stamp = i33->timestamp[i33->snapshot_tail];

		size = cur_lcode - i33->last_lcode;
		if (size < 0)
			size += CODEC_BUFFER_SIZE;

		if (size == 0) {
			i33->empty_field_count++;
			
			if (i33->compress_started) {
				dprintk("%s: %d: No data for field\n", 
					i33->device_name,
					i33->mga->zr->jpg_seq_num);
			}
					
			if (i33->empty_field_count == 8) {
				dprintk("i33: Input stream stalled, signalling restart\n");
				zr36060_ioctl(i33->mga->zr, ZR36060_RESTART, 0);
				i33->empty_field_count = 0;
			}
				
			goto bottom;
		} else {
			i33->compress_started = 1;
		}
		
		while ( i33->last_lcode != cur_lcode ) {

			if ((mem=zr36060_current_buffer(mga->zr,
				&left, cur_field)) == NULL) {

				dprintk("%s: %d: no buffer\n", 
				i33->device_name,
					i33->mga->zr->jpg_seq_num );
				goto bottom;
			}

			copied = mgacodec_recv_bitstream(codec, mem, 
				i33->last_lcode, cur_lcode, left);
			
			if ( copied > COMP_AHEAD_THRES*2 )
				limit = i33->last_lcode + copied - COMP_AHEAD_THRES;
			else
				limit = i33->last_lcode + copied/2;
			limit &= (CODEC_BUFFER_SIZE-1);
			mgacodec_set_transfer_limit(codec, limit );
			
			if ( mgacodec_is_stalled(codec) ) {
				mgacodec_start_transfer(codec, 0);
			}

			count = zr36060_current_put(mga->zr, copied, &cur_field,
				&cur_stamp);

			if (count == 0) {
			
				size = cur_lcode - i33->last_lcode;
				if (size < 0)
					size += CODEC_BUFFER_SIZE;
			
				if (copied != size) {
					/* copy was limited by buffer size, and no field found */
					printk(MGA_WARNING "%s: %d: buffer overflow\n",
					i33->device_name,
					i33->mga->zr->jpg_seq_num);
					
					zr36060_current_flush(mga->zr);
				}

				goto bottom;
			}

			i33->last_lcode = (i33->last_lcode + count) % CODEC_BUFFER_SIZE;
		}

bottom:
		i33->last_lcode = cur_lcode;
		i33->snapshot_tail = (i33->snapshot_tail + 1) & 31; /* update tail atomic */
	}

	i33->compress_active = 0;
}
 

static void
i33_fill_stream(void *cte_private)
{
	int		 size;
	int		 i;
	u32		 left;
	u8		 *mem;
	struct mga_dev_s *mga;
	struct i33_dev_s *i33;
	u32		stream;
	int		streamed;
	u32		buffer;
	int		donecode;
	int		cur_field;

	i33 = cte_private;
	mga = (struct mga_dev_s *)i33->mga;

	if (test_and_set_bit(0, &i33->decompress_active)) {
		dprintk("%s: fill_stream: called while active\n",
		       i33->device_name);
		return;
	}

	cur_field = i33->field[i33->snapshot_tail];
	
	while(i33->snapshot_tail != i33->snapshot_head) {
		stream = i33->hardptr[i33->snapshot_tail];
		cur_field = i33->field[i33->snapshot_tail];

		streamed = stream - i33->last_stream;
		if (streamed < 0)
			streamed += CODEC_BUFFER_SIZE;

		if (streamed==0 && stream) {
			dprintk("i33: Output stream stalled, signalling restart\n");
			zr36060_ioctl(i33->mga->zr, ZR36060_RESTART, 0);

			while(i33->buffer_tail != i33->buffer_head) {
				if (i33->buffer_status[i33->buffer_tail & 0xff] == 1)
					zr36060_done(i33->mga->zr, &i33->timestamp[i33->snapshot_tail]);
				i33->buffer_status[i33->buffer_tail & 0xff] = -1;
				i33->buffer_tail++;
			}
			zr36060_current_flush(mga->zr);
			i33->snapshot_tail = i33->snapshot_head = 0;
			i33->decompress_active = 0;
			return;
		}

		/*
		* Loop removing buffer that have been displayed. 
		*/
		while(i33->buffer_tail != i33->buffer_head) {
			buffer = i33->buffer_len[i33->buffer_tail & 0xff];
			/*
			* buffer larger then amount streamed this buffer
			* has not been displayed
			*/
			if (buffer > streamed) {
				i33->buffer_len[i33->buffer_tail & 0xff] -= streamed;
				i33->last_stream = (i33->last_stream + streamed) % CODEC_BUFFER_SIZE;
				break;
			}
			/* only done none replay buffers */
			if (i33->buffer_status[i33->buffer_tail & 0xff] == 1)
				zr36060_done(i33->mga->zr, &i33->timestamp[i33->snapshot_tail]);

			/* Mark buffer as free */
			i33->buffer_status[i33->buffer_tail & 0xff] = -1;

			/* Move to next buffer */
			i33->buffer_tail++;

			/* remove buffer from amount streamed */
			streamed -= buffer;

			/* move last stream up to show buffer removed */
			i33->last_stream = (i33->last_stream + buffer) % CODEC_BUFFER_SIZE;
		}
		
		i33->snapshot_tail = (i33->snapshot_tail + 1) & 31; /* update tail atomic */
	}

	if ( (i33->buffer_tail > 2) && (i33->buffer_head != i33->buffer_tail) ) {
	
		/* cur_field corresponds to the last field of buffer
                   at buffer_tail-1, so include it in the count. */
		cur_field = (( (cur_field-1) + (i33->buffer_head - i33->buffer_tail + 1) )&1) + 1;
	}
	else {
		/* only just started playing, don't check fields yet */
		cur_field = 0;
	}


	for(i=0;i < 64; i++) {
		/* amount of codecbuffer to keep unfilled.
		   note: hardptr behaves inconsistently sometimes; steps backwards a
		     few bytes (around 64 seems typical) */
		const int headroom = 2048;
		
		/*
		 * If we are playing the last buffer then only
		 * allow a fixed 4 fields/frames to be played.
		 */
		if (zr36060_current_last(mga->zr)) {
			if ((abs(i33->buffer_head - i33->buffer_tail)) >= 4) {
				break;
			}
		}
		
		size = mgacodec_buffer_free(mga->codec);
		if (size <= headroom)
			break;

		if ((mem = zr36060_current_buffer(mga->zr, &left, cur_field)) == NULL) {
			break;
		}


		if (left > size-headroom) /* dont fill complete buffer */
			left = size-headroom;
		if (left < size)
			size = left;

		mgacodec_send_bitstream_irq(mga->codec, mem, size);
		
		i33->buffer_len[i33->buffer_head & 0xff] += size;

		donecode = zr36060_current_get(mga->zr, size);
		if (donecode) {
			i33->buffer_status[i33->buffer_head & 0xff] = donecode;
			if ( cur_field ) {
				cur_field ^= 3;
			}
			i33->buffer_head++;
			i33->buffer_len[i33->buffer_head & 0xff] = 0;
		}
	}
	i33->decompress_active = 0;
}


static int
i33_set_video(void *cte_private, enum buz_codec_mode mode)
{
	int	temp;
	int	norm;
	int	input;
	struct  mga_dev_s *mga;
	struct  i33_dev_s *i33;

	i33 = cte_private;
	mga = i33->mga;

	switch (mga->zr->params.norm) {
	case VIDEO_MODE_PAL:
		i33->field_time = 1000000/50;
		norm = KS_STD_PAL;
		break;
	case VIDEO_MODE_NTSC:
	default:
		i33->field_time = 1001000/60;
		norm = KS_STD_NTSC;
		break;
	}

	switch(mode) {
	case BUZ_MODE_MOTION_COMPRESS:
	case BUZ_MODE_STILL_COMPRESS:
		mgavideo_get_active(mga->video, &i33->last_input, 
			&i33->last_norm);

		switch (mga->zr->params.input) {
		case 0:
			input = KS_INPUT_COMPOSITE;
			break;
		default:
		case 1:
			input = KS_INPUT_SVIDEO;
			break;
		case 2:
			input = KS_INPUT_TUNER;
			break;
		}
		mgavideo_set_active(mga->video, input, norm);

		temp = KS_OUTPUT_EXV;
		mgavideo_decoder(mga->video, KS0127_SET_OUTPUT, &temp);
		break;

	case BUZ_MODE_MOTION_DECOMPRESS:
	case BUZ_MODE_STILL_DECOMPRESS:
		mgavideo_get_active(mga->video, &i33->last_input, 
			&i33->last_norm);

		temp = KS_OUTPUT_YUV656E;
		mgavideo_decoder(mga->video, KS0127_SET_OUTPUT, &temp);

		mgavideo_set_active(mga->video, KS_INPUT_YUV656, norm);
		break;

	case BUZ_MODE_IDLE:
		if (i33->last_input != -1 && i33->last_norm != -1)
			mgavideo_set_active(mga->video, i33->last_input,
				i33->last_norm);
		i33->last_input = -1;
		i33->last_norm = -1;

		temp = KS_OUTPUT_YUV656E;
		mgavideo_decoder(mga->video, KS0127_SET_OUTPUT, &temp);
		break;
	}
	return 0;
}


static int
i33_set_jpg(void *cte_private, enum buz_codec_mode mode)
{
	struct i33_dev_s *i33;

	i33 = cte_private;

	switch(mode) {
	case BUZ_MODE_MOTION_COMPRESS:
	case BUZ_MODE_STILL_COMPRESS:
		i33->last_hardptr = 0;
		i33->last_lcode = 0;
		i33->buffer_tail = 0;
		i33->buffer_head = 0;
		i33->snapshot_head = 0;
		i33->snapshot_tail = 0;
		i33->compress_started = 0;
		i33->empty_field_count = 0;
		mga_enable_irq(i33->mga, MGA_IRQ_VINVSYNC, (mga_interrupt_handler_t)i33_vsync_compress);
		break;

	case BUZ_MODE_MOTION_DECOMPRESS:
	case BUZ_MODE_STILL_DECOMPRESS:
		i33->buffer_tail = 0;
		i33->buffer_head = 0;
		i33->last_stream = 0;
		i33->snapshot_head = 0;
		i33->snapshot_tail = 0;
		i33->buffer_len[i33->buffer_tail & 0xff] = 0;
		mga_enable_irq(i33->mga, MGA_IRQ_VINVSYNC, (mga_interrupt_handler_t)i33_vsync_decompress);
	        break;

	case BUZ_MODE_IDLE:
		switch(i33->mode) {
		case BUZ_MODE_MOTION_DECOMPRESS:
		case BUZ_MODE_STILL_DECOMPRESS:
			mga_disable_irq(i33->mga, MGA_IRQ_VINVSYNC, (mga_interrupt_handler_t)i33_vsync_decompress);
			break;

		case BUZ_MODE_MOTION_COMPRESS:
		case BUZ_MODE_STILL_COMPRESS:
			mga_disable_irq(i33->mga, MGA_IRQ_VINVSYNC, (mga_interrupt_handler_t)i33_vsync_compress);

			// deassert START line
			mgacodec_and_or_miscctl(i33->mga->codec, ~0x10, 0x00);
			break;

		case BUZ_MODE_IDLE:
			break;
		}
		break;
	}
	i33->mode = mode;
	return 0;
}


static int
i33_attach(struct mga_dev_s *mga)
{
	char	chip_id[2];
	struct  i33_dev_s	*i33;
	struct	zr36060_dev	*zr;
	int	inv_vclk;

	mga->zr = zr = kmalloc(sizeof(*zr), GFP_KERNEL);
	if (zr == NULL)
		return -ENOMEM;
	mga->i33 = i33 = kmalloc(sizeof(*i33), GFP_KERNEL);
	if (i33 == NULL) {
		kfree(zr);
		mga->zr = NULL;
		return -ENOMEM;
	}

	memset(i33, 0, sizeof(*i33));
	memset(zr, 0, sizeof(*zr));

	sprintf(i33->device_name, "mga%d-i33", mga->card_num);
	sprintf(zr->device_name, "mga%d-zr36060", mga->card_num);

	i33->mga = mga;

	zr->params.input = KS_INPUT_TUNER;
	zr->params.norm = KS_STD_NTSC;

	zr->memory_size = 8;

	zr->input_count = 3;
	zr->video_bus_size = 8;
	zr->code_master = 0;

	zr->reset = i33_reset;
	zr->write_memory = i33_write_memory;
	zr->read_memory = i33_read_memory;
	zr->start_transfer = i33_start_transfer;

	zr->codec_set_video = i33_set_video;
	zr->codec_set_jpg   = i33_set_jpg;

	zr->norms = zrnorms;

	i33_enable(i33);

	if ( i33_read_memory(i33, chip_id, 0x22, 2) ||
		(chip_id[0] != 0x33) ) {
		printk(MGA_NOTICE "%s: zr36060 not found (0x%x 0x%x)\n",
		       i33->device_name, chip_id[0], chip_id[1]);
		
		kfree(i33);
		kfree(zr);

		mga->i33 = NULL;
		mga->zr = NULL;
		return -1;
	}
	
	i33_disable(i33);

	dprintk("%s: zr36060 found revision %x\n", 
	       i33->device_name, chip_id[1]);

	/* currently known device(s) that needs reversed vclk: marvel g400 */
	inv_vclk = (i33->mga->options & MGAVIDEO_HAS_MARVEL) && 
		(mga->mga_type == MGA_G400);
	zr36060_attach(zr, i33, inv_vclk );

	/*HACK*/
	if ( inv_vclk ) {
		zrnorms[0].HStart = ((zrnorms[0].HStart+1)&-4)-1;
		zrnorms[1].HStart = ((zrnorms[1].HStart+1)&-4)-1;
		zrnorms[2].HStart = ((zrnorms[2].HStart+1)&-4)-1;
	}

	i33->mga->options |= MGAVIDEO_HAS_ZR36060;

        INIT_WORK(&i33->compress, i33_get_stream, mga->i33);
        INIT_WORK(&i33->decompress, i33_fill_stream, mga->i33);

	i33->last_input = -1;
	i33->last_norm = -1;

	return 0;
}


static void
i33_deattach(struct i33_dev_s	*i33)
{
	if ( i33 != NULL ) {
		kfree(i33->mga->zr);
		kfree(i33);

		i33->mga->i33 = NULL;
		i33->mga->zr = NULL;
	}
}


////////////////////////////////


int
i33_lock(struct mga_dev_s *mga)
{
	if (mga->i33 != NULL) {
		if (mga_card_lock(mga)==0) {
			i33_enable(mga->i33);
			MOD_INC_USE_COUNT;
			return 0;
		}
		return -EBUSY;
	}
	return 0;
}
EXPORT_SYMBOL(i33_lock);

void
i33_unlock(struct mga_dev_s *mga)
{
	if (mga->i33 != NULL) {
		i33_disable(mga->i33);
		mga_card_unlock(mga);
		MOD_DEC_USE_COUNT;
	}
}
EXPORT_SYMBOL(i33_unlock);

int init_module(void)
{
	struct mga_dev_s *mga;
	int		 i;

	printk(KERN_INFO "Matrox MJPEG driver\n");

	for(i=0; (mga = mga_get_card(i)) != NULL ; i++) {
		/* detect the zr36060 for mjpeg */
		if ( i33_attach(mga) ) {
			if ( mga->i33 != NULL ) {
				i33_deattach(mga->i33);
			}
			continue;
		}
	}
	if(i == 0) {
//		return -ENXIO;	allow insmod on zoran-less cards
	}
	return 0;
}


void cleanup_module(void)
{
	int i = 0;
	struct mga_dev_s *mga;

	for(i=0; (mga = mga_get_card(i)) != NULL; i++) {
		if ( mga->i33 != NULL ) {
			i33_deattach(mga->i33);
		}
		/* need to release it 2 times */
		mga_release_card(mga);
		mga_release_card(mga);
	}
}

