/*
 * for the Matrox Marvel G200 and Rainbow Runner-G series
 *
 * This module is an interface to the ZR36060 jpeg codec chip.
 *
 * Copyright (C) 1999  Mike Bernson <mike@mlb.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *
 * This code is based on Rainer Johanni buz driver. <Rainer@Johanni.de>
 *
 *****************************************************************************
 *
 * Modified and extended by
 *	Leon van Stuivenberg <leonvs@iae.nl>
 */

#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/signal.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <linux/sched.h>
#include <asm/segment.h>
#include <linux/types.h>

#include <linux/videodev.h>
#include <linux/version.h>
#include <asm/uaccess.h>

#include <linux/proc_fs.h>

#include "aiee.h"
#include "mgacore.h"
#include "zr36060.h"


#define dprintk     if (debug) printk

static int debug = 0; /* insmod parameter */

#if LINUX_VERSION_CODE >= 0x020100
MODULE_PARM(debug,"i");
#endif

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif


#define	MAX_KMALLOC_MEM	(32 * 1024)

static void jpg_fbuffer_free(struct zr36060_dev *zr);

#ifndef	MAP_NR
#define	MAP_NR(x)	virt_to_page(x)
#endif

static void
add_time(struct timeval *v, const long time_us)
{
	v->tv_usec += time_us;
	if (time_us > 0) {
		while(v->tv_usec >= 1000000) {
			v->tv_usec -= 1000000;
			v->tv_sec++;
		}
	} else {
		while(v->tv_usec < 0) {
			v->tv_usec += 1000000;
			v->tv_sec--;
		}
	}
}



/*
 *   Allocate the MJPEG grab buffers.
 *
 */
static int jpg_fbuffer_alloc(struct zr36060_dev * zr)
{
	u8	*mem;
	int	i;

	if ((mem = mga_rvmalloc(zr->jpg_bufsize * zr->jpg_nbufs)) == NULL) {
		return -ENOBUFS;
	}

	memset(mem, 0, zr->jpg_bufsize * zr->jpg_nbufs);

	if(zr->jpg_buffers_allocated) {
		printk(KERN_WARNING "%s: jpg_fbuffer_alloc: buffers allready "
		       "allocated ???\n", zr->device_name);
	}

	for (i = 0; i < zr->jpg_nbufs; i++) {
		zr->jpg_gbuf[i].virt = mem;
		mem += zr->jpg_bufsize;
	}

	dprintk("%s: jpg_fbuffer_alloc: %d KB allocated\n",
		zr->device_name, (zr->jpg_nbufs*zr->jpg_bufsize)>>10);

	zr->jpg_buffers_allocated = 1;

	return 0;
}

/* free the MJPEG grab buffers */
static void jpg_fbuffer_free(struct zr36060_dev * zr)
{
	if (zr->jpg_buffers_allocated == 0)
		return;

	mga_rvfree(zr->jpg_gbuf[0].virt, zr->jpg_bufsize * zr->jpg_nbufs);

	zr->jpg_buffers_allocated = 0;
}

int
zr36060_mmap_bufs(struct vm_area_struct *vma, struct zr36060_dev *zr
                  , unsigned long adr, unsigned long size)
{
	unsigned long start;
	unsigned long pos;

	start = adr;

	/* Map the MJPEG buffers */

	if(!zr->jpg_buffers_allocated) {
		dprintk("%s: zr36060_mmap_bufs(MJPEG): buffers not yet allocated\n",
			zr->device_name);
		return -ENOMEM;
	}

	if(size > zr->jpg_nbufs * zr->jpg_bufsize) {
		dprintk("%s: Max size is %u - you wanted %lu\n",
			zr->device_name, zr->jpg_nbufs * zr->jpg_bufsize, size);
		return -EINVAL;
	}

	if(size != zr->jpg_nbufs * zr->jpg_bufsize) {
		dprintk("%s: Expected %u - you wanted %lu\n",
			zr->device_name, zr->jpg_nbufs * zr->jpg_bufsize, size);
	}

	pos = (unsigned long)zr->jpg_gbuf[0].virt;
	while(size > 0) {
                unsigned long page = virt_to_phys((void*)pos);
		if (remap_page_range(vma, start, page, PAGE_SIZE, PAGE_SHARED))
			return -EAGAIN;
		start += PAGE_SIZE;
		pos += PAGE_SIZE;
		size -= PAGE_SIZE;
	}
	return 0;
}
EXPORT_SYMBOL(zr36060_mmap_bufs);

/* Check a buz_params struct for correctness, insert default params */
static int zr36060_check_params(struct zr36060_dev *zr, struct buz_params * params)
{
	int err = 0, err0 = 0;
	int errcode = 0;

	/* insert constant params */
	params->major_version = 0;
	params->minor_version = 0xdeadbeef;

	/* Check input and norm */

	if(params->input < 0 || params->input > zr->input_count) {
		dprintk("%s CHECK PARAMS: input = %d, must be 0-%d\n",
			zr->device_name, params->input, zr->input_count);
		err++;
	}

	/* Deactivated in the moment */
	/*     if(params->norm != VIDEO_MODE_PAL && params->norm != VIDEO_MODE_NTSC) */
	/*       { */
	/*       printk("%s: CHECK PARAMS: norm = %d, must be 0 (PAL) or 1 (NTSC)\n",zr->device_name, params->norm); */
	/*       err++; */
	/*       } */

	/* Check decimation, set default values for decimation = 1, 2, 4 */
	switch (params->decimation) {
	case 1:
		params->HorDcm         = 1;
		params->VerDcm         = 1;
		params->TmpDcm         = 1;
		params->field_per_buff = 2;

		params->img_x          = 0;
		params->img_y          = 0;
		params->img_width      = 720;
		params->img_height     = zr->norms[params->norm].Ha/2;
		break;

	case 2:
		params->HorDcm         = 2;
		params->VerDcm         = 1;
		params->TmpDcm         = 2;
		params->field_per_buff = 1;

		params->img_x          = 8;
		params->img_y          = 0;
		params->img_width      = 704;
		params->img_height     = zr->norms[params->norm].Ha/2;
		break;

	case 4:
		params->HorDcm         = 4;
		params->VerDcm         = 2;
		params->TmpDcm         = 2;
		params->field_per_buff = 1;

		params->img_x          = 8;
		params->img_y          = 0;
		params->img_width      = 704;
		params->img_height     = zr->norms[params->norm].Ha/2;
		break;

	case 0:
	/* We have to check the data the user has set */
	  if(params->HorDcm!=1 && params->HorDcm!=2 && params->HorDcm!=4) { err0++; errcode = 0x1; }
	  if(params->VerDcm!=1 && params->VerDcm!=2) { err0++; errcode |= 0x2; }
	  if(params->TmpDcm!=1 && params->TmpDcm!=2) { err0++; errcode |= 0x4; }
	  if(params->field_per_buff!=1 && params->field_per_buff!=2) { err0++; errcode |= 0x8; }

	  if(params->img_x     <0) { err0++; errcode |= 0x10; }
	  if(params->img_y     <0) { err0++; errcode |= 0x20; }
	  if(params->img_width <0) { err0++; errcode |= 0x40; }
	  if(params->img_height<0) { err0++; errcode |= 0x80; }
	  
	  if(params->img_x + params->img_width  > 720) { err0++; errcode |= 0x100; }
	  if(params->img_y + params->img_height > zr->norms[params->norm].Ha/2) { err0++; errcode |= 0x200; }
	  if(params->img_width  % (16*params->HorDcm) != 0 ) { err0++; errcode |= 0x400; }
	  if(params->img_height % ( 8*params->VerDcm) != 0 ) { err0++; errcode |= 0x800; }
	  
		if(err0)
		{
		dprintk("%s: SET PARAMS: error in params for decimation = 0, error code 0x%x\n",
			zr->device_name, errcode);
		err++;
		}
		break;

	default:
		dprintk("%s:  SET PARAMS: decimation = %d, must be 0, 1, 2 or 4\n",
			zr->device_name, params->decimation);
		err++;
		break;
	}

	if(params->quality > 100) params->quality = 100;
	if(params->quality <   5) params->quality =   5;

	if(params->APPn <  0) params->APPn =  0;
	if(params->APPn > 15) params->APPn = 15;
	if(params->APP_len <  0) params->APP_len =  0;
	if(params->APP_len > 60) params->APP_len = 60;
	if(params->COM_len <  0) params->COM_len =  0;
	if(params->COM_len > 60) params->COM_len = 60;

	if(err) return -EINVAL;

	return 0;

}

static int zr36060_default_params(struct zr36060_dev *zr)
{
	int i;

	/* re-initialize DMA ring stuff */  
	/* see set_jpg 
	zr->jpg_que_head = 0;
	zr->jpg_dma_head = 0;
	zr->jpg_dma_tail = 0;
	zr->jpg_que_tail = 0;
	zr->jpg_seq_num  = 0;

	for(i=0; i < zr->jpg_nbufs; i++)
		zr->jpg_gbuf[i].state = BUZ_STATE_USER;
	*/

	/* Set necessary params and call buz_check_params to set the defaults */
	zr->params.decimation = 1;

	zr->params.quality    = 50; /* default compression factor 8 */
	zr->params.odd_even   = 1;

	zr->params.APPn    = 0;
	zr->params.APP_len = 0; /* No APPn marker */
	for(i=0;i<60;i++) 
		zr->params.APP_data[i] = 0;

	zr->params.COM_len = 0; /* No COM marker */
	for(i=0;i<60;i++) 
		zr->params.COM_data[i] = 0;

	zr->params.VFIFO_FB = 1;

	memset(zr->params.reserved,0,sizeof(zr->params.reserved));

	/* only enable DRI when some driver/app actually makes use of it. */
	zr->params.jpeg_markers = JPEG_MARKER_DHT | JPEG_MARKER_DQT; /* | JPEG_MARKER_DRI;*/

	i = zr36060_check_params(zr, &zr->params);
	if (i) printk(KERN_ERR "%s: buz_open_init_params internal error\n", zr->device_name);

	return i;
}

/****************************************************************************
* linux kernel module api
****************************************************************************/
#ifdef MODULE
int init_module(void)
#else
int zr36060_init(void)
#endif
{

	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{
}
#endif

static int
zr36060_load_memory(struct zr36060_dev *zr) 
{
	int	i;
	char	status;
	char	data[1];

	data[0] = 0x80;
	zr->write_memory(zr->cte_private, data, 0, 1);
	
	/* memload takes ~100usec ; wait a bit before checking if it's done */
	udelay(75);

	for(i=0; i < 200; i++) {
		status = 0;
		zr->read_memory(zr->cte_private, &status, 1, 1);

		if ((status & 0x80) == 0) {
			return 0;
		}
		udelay(10);
	}

	printk(KERN_ERR "%s: load from memory did not complete status=0x%x\n",
		zr->device_name, status);
	return -1;
 }



static u8 SOF[19] = {
	0xff, 0xc0,		/* Start of frame marker */
	0x00, 0x11,		/* Length of this segment */
	0x08,			/* Precision */
	0x00, 0x78,		/* number of lines */
	0x00, 0xb0,		/* line length */
	3,			/* number of components */
	0, 0x21, 0,		/* Y component */
	1, 0x11, 1,		/* U component */
	2, 0x11, 1		/* V component */
};

static u8 SOS[14] = {
	0xff, 0xda,		/* start of Scan marker */
	0x00, 0x0c,		/* lenght of this segment */
	0x03,			/* number of components */
	0x00, 0x00,		/* Y component */
	0x01, 0x11,		/* U component */
	0x02, 0x11,		/* V component */
	0x00, 0x3f, 0x00	/* constant */
};

static u8 DRI[6] = {
	0xff, 0xdd,		/* define Restart Interval */
	0x00, 0x04,		/* length of this segment */
	0x00, 0x08,		/* lenght of restart interval in mcu units */
};

static u8 DQT[134] = {
	0xff, 0xdb,		/* DQT marker */
	0x00, 0x84,		/* DQT length */
	0x00,			/* table ID 0 */
	0x10, 0x0b, 0x0c, 0x0e, 0x0c, 0x0a, 0x10, 0x0e,
	0x0d, 0x0e, 0x12, 0x11, 0x10, 0x13, 0x18, 0x28,
	0x1a, 0x18, 0x16, 0x16, 0x18, 0x31, 0x23, 0x25,
	0x1d, 0x28, 0x3a, 0x33, 0x3d, 0x3c, 0x39, 0x33,
	0x38, 0x37, 0x40, 0x48, 0x5c, 0x4e, 0x40, 0x44,
	0x57, 0x45, 0x37, 0x38, 0x50, 0x6d, 0x51, 0x57,
	0x5f, 0x62, 0x67, 0x68, 0x67, 0x3e, 0x4d, 0x71,
	0x79, 0x70, 0x64, 0x78, 0x5c, 0x65, 0x67, 0x63,
	0x01,			/* table ID 1 */
	0x11, 0x12, 0x12, 0x18, 0x15, 0x18, 0x2f, 0x1a,
	0x1a, 0x2f, 0x63, 0x42, 0x38, 0x42, 0x63, 0x63,
	0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
	0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
	0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
	0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
	0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
	0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63
};

static u8 DHT[420] = {
	0xff, 0xc4,   /* DHT marker */
	0x01, 0xa2,   /* DHT length */
	0x00,      /* table class 0, ID 0 */
	0x00, 0x01, 0x05, 0x01, 0x01, 0x01, 0x01, 0x01, /* # codes of length 1..8 */
	0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* # codes of length 8..16 */
	0x00,                  /* values for codes of length 2 */
	0x01, 0x02, 0x03, 0x04, 0x05,         /* values for codes of length 3 */
	0x06,                  /* values for codes of length 4 */
	0x07,                  /* values for codes of length 5 */
	0x08,                  /* values for codes of length 6 */
	0x09,                  /* values for codes of length 7 */
	0x0a,                  /* values for codes of length 8 */
	0x0b,                  /* values for codes of length 9 */
	0x01,      /* table class 0, ID 1 */
	0x00, 0x03, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,   /* # codes of length 1..8 */
	0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00,   /* # codes of length 9..16 */
	0x00, 0x01, 0x02,            /* values for codes of length 2 */
	0x03,                  /* values for codes of length 3 */
	0x04,                  /* values for codes of length 4 */
	0x05,                  /* values for codes of length 5 */
	0x06,                  /* values for codes of length 6 */
	0x07,                  /* values for codes of length 7 */
	0x08,                  /* values for codes of length 8 */
	0x09,                  /* values for codes of length 9 */
	0x0a,                  /* values for codes of length 10 */
	0x0b,                  /* values for codes of length 11 */
	0x10,
	0x00, 0x02, 0x01, 0x03, 0x03, 0x02, 0x04, 0x03,
	0x05, 0x05, 0x04, 0x04, 0x00, 0x00, 0x01, 0x7d,
	0x01, 0x02, 0x03, 0x00, 0x04, 0x11, 0x05, 0x12,
	0x21, 0x31, 0x41, 0x06, 0x13, 0x51, 0x61, 0x07,
	0x22, 0x71, 0x14, 0x32, 0x81, 0x91, 0xa1, 0x08,
	0x23, 0x42, 0xb1, 0xc1, 0x15, 0x52, 0xd1, 0xf0,
	0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0a, 0x16,
	0x17, 0x18, 0x19, 0x1a, 0x25, 0x26, 0x27, 0x28,
	0x29, 0x2a, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39,
	0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49,
	0x4a, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59,
	0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69,
	0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79,
	0x7a, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
	0x8a, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98,
	0x99, 0x9a, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,
	0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6,
	0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3, 0xc4, 0xc5,
	0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2, 0xd3, 0xd4,
	0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xe1, 0xe2,
	0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea,
	0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8,
	0xf9, 0xfa, 0x11, 0x00, 0x02, 0x01, 0x02, 0x04,
	0x04, 0x03, 0x04, 0x07, 0x05, 0x04, 0x04, 0x00,
	0x01, 0x02, 0x77, 0x00, 0x01, 0x02, 0x03, 0x11,
	0x04, 0x05, 0x21, 0x31, 0x06, 0x12, 0x41, 0x51,
	0x07, 0x61, 0x71, 0x13, 0x22, 0x32, 0x81, 0x08,
	0x14, 0x42, 0x91, 0xa1, 0xb1, 0xc1, 0x09, 0x23,
	0x33, 0x52, 0xf0, 0x15, 0x62, 0x72, 0xd1, 0x0a,
	0x16, 0x24, 0x34, 0xe1, 0x25, 0xf1, 0x17, 0x18,
	0x19, 0x1a, 0x26, 0x27, 0x28, 0x29, 0x2a, 0x35,
	0x36, 0x37, 0x38, 0x39, 0x3a, 0x43, 0x44, 0x45,
	0x46, 0x47, 0x48, 0x49, 0x4a, 0x53, 0x54, 0x55,
	0x56, 0x57, 0x58, 0x59, 0x5a, 0x63, 0x64, 0x65,
	0x66, 0x67, 0x68, 0x69, 0x6a, 0x73, 0x74, 0x75,
	0x76, 0x77, 0x78, 0x79, 0x7a, 0x82, 0x83, 0x84,
	0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x92, 0x93,
	0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0xa2,
	0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa,
	0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9,
	0xba, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8,
	0xc9, 0xca, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7,
	0xd8, 0xd9, 0xda, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6,
	0xe7, 0xe8, 0xe9, 0xea, 0xf2, 0xf3, 0xf4, 0xf5,
	0xf6, 0xf7, 0xf8, 0xf9, 0xfa
};


static int
set_sync_generator( struct zr36060_dev *zr, u8 *data, enum buz_codec_mode mode,
	unsigned v_total, unsigned h_total,
	unsigned v_sync_size, unsigned h_sync_size,
	unsigned bv_start, unsigned bv_end,
	unsigned bh_start, unsigned bh_end )
{
	data[0x36] = v_total >> 8;
	data[0x37] = v_total;
	data[0x38] = h_total >> 8;
	data[0x39] = h_total;
	data[0x3a] = v_sync_size;
	data[0x3b] = h_sync_size;

	data[0x3c] = bv_start;
	data[0x3d] = bh_start;

	data[0x3e] = bv_end >> 8;
	data[0x3f] = bv_end;

	data[0x40] = bh_end >> 8;
	data[0x41] = bh_end;

	return 0;
}



static int
set_active_area( struct zr36060_dev *zr, u8 *data, enum buz_codec_mode mode,
	unsigned v_start, unsigned v_end,
	unsigned h_start, unsigned h_end )
{
	unsigned Ntf;

	if ( v_start < 2 ) {
		dprintk("%s: v_start too small (%i)\n", 
			zr->device_name, v_start );
	}
	if ( (v_end - v_start) > 32768 ) {
		dprintk("%s: active area too tall (%i)\n", 
		       zr->device_name, v_end - v_start);
	}
	if ( (h_end - h_start) > 768 ) {
		dprintk("%s: active area too wide (%i)\n", 
		       zr->device_name, h_end - h_start );
	}
	if ( (h_start >= 1024) || (h_end >= 1024) ) {
		dprintk("%s: active invalid horz (%i,%i)\n", 
		       zr->device_name, h_start, h_end );
	}

	if ( zr->params.norm == VIDEO_MODE_NTSC )
		Ntf = 262;
	else
		Ntf = 312;
	if ( v_end > (Ntf-10) ) {
		dprintk("%s: v_end too large (%i, max %i)\n", 
		       zr->device_name, v_end, Ntf-10 );
	}

	if (	((mode == BUZ_MODE_MOTION_COMPRESS || mode == BUZ_MODE_STILL_COMPRESS) &&
			(h_start < 8)) ||
		((mode == BUZ_MODE_MOTION_DECOMPRESS || mode == BUZ_MODE_STILL_DECOMPRESS) &&
			(h_start < 12)) ) {
		dprintk("%s: h_start too small (%i)\n", 
			zr->device_name, h_start );
	}

	data[0x42] = v_start >> 8;
	data[0x43] = v_start;

	data[0x44] = v_end >> 8;
	data[0x45] = v_end;

	data[0x46] = h_start >> 8;
	data[0x47] = h_start;

	data[0x48] = h_end >> 8;
	data[0x49] = h_end;

	return 0;
}


static int
set_subimg_area( struct zr36060_dev *zr, u8 *data, enum buz_codec_mode mode,
	unsigned sv_start, unsigned sv_end,
	unsigned sh_start, unsigned sh_end )
{
	if ( (sv_end - sv_start) > 32768 ) {
		dprintk("%s: subimg area too tall (%i)\n", 
			zr->device_name, sv_end - sv_start );
	}
	if ( (sh_end - sh_start) > 1024 ) {
		dprintk( "%s: subimg area too wide (%i)\n", 
		       zr->device_name, sh_end - sh_start );
	}
	if ( (sh_start >= 1024) || (sh_end >= 1024) ) {
		dprintk( "%s: subimg invalid horz (%i,%i)\n", 
		       zr->device_name, sh_start, sh_end );
	}

	data[0x4a] = sv_start >> 8;
	data[0x4b] = sv_start;

	data[0x4c] = sv_end >> 8;
	data[0x4d] = sv_end;

	data[0x4e] = sh_start >> 8;
	data[0x4f] = sh_start;

	data[0x50] = sh_end >> 8;
	data[0x51] = sh_end;

	return 0;
}




int zr36060_set_cap(struct zr36060_dev *zr, enum buz_codec_mode mode)
{
	u8	data[1024];
	u32	size;
	int	lines;
	int	i;
	int	len;
	int	Vstart;
	int	Hstart;
	struct	tvnorm *tvn;

	static struct {
		short addr;
		short len;
	} mem_region[] = {
		{ 0x02, 6 },	{ 0x09, 4 },	{ 0x11, 5 },
		{ 0x24, 2 },	{ 0x30, 34 },	{ 0x60, 19 },
		{ 0x7a, 14 },	{ 0xc0, 6 },	{ 0xcc, 134 },
		{ 0x1d4, 420 },	{ 0x380, 64 },	{ 0x3c0, 64 }
	};

	for(i=10; i; i--) {
		/* reset zr36060 on wake it up */
		zr->reset(zr->cte_private, 1);

		if (zr->read_memory(zr->cte_private, data, 0x22, 2) == 0) {
			if (data[0] != 0x33) {
				printk(KERN_WARNING "%s: Bad chip id %x\n", 
				       zr->device_name, data[0]);
				continue;
			}
			break;
		}
		printk(KERN_ERR "%s: error reading chip id. Reset again\n",
		       zr->device_name);
	}
	if (i == 0) {
		return -EINVAL;
	}

  
	data[0] = 0x01;

	zr->write_memory(zr->cte_private, data, 0, 1);

	memset(data, 0, sizeof(data));

	data[0x02] = 0x00;	/* from matrox */

	if (mode == BUZ_MODE_MOTION_COMPRESS || mode == BUZ_MODE_STILL_COMPRESS) {
		data[0x03] = 0xa4; /* compress pass, variable scaling */
	} else {
		data[0x03] = 0x00; /* decompress pass */
	}
	data[0x04] = 0x00; /* reserved */
	data[0x05] = 0xff; /* max block code volume in bits (510) */
	data[0x06] = zr->params.jpeg_markers & 0x38;
	if (zr->params.COM_len) 
		data[0x06] |= JPEG_MARKER_COM;
	if (zr->params.APP_len)
		data[0x06] |= JPEG_MARKER_APP;

#ifdef	NEVER
	data[0x07] = 0x01; /* from matrox */
#else
	data[0x07] = 0x00;
#endif

	/* Target- field size in pixels: */
	tvn = &zr->norms[zr->params.norm];

	size = (tvn->Ha/2)*(tvn->Wa)/(zr->params.HorDcm)/(zr->params.VerDcm);
         
	/* Target compressed field size in bits: */
	size = size*16; /* uncompressed size in bits */
	size = size*zr->params.quality/400; /* quality = 100 is a compression ratio 1:4 */

	if (size < 8192)
		size = 8192;

	/* Upper limit: 7/8 of the code buffers */
	if(size*zr->params.field_per_buff > zr->jpg_bufsize*7)
		size = zr->jpg_bufsize*7 / zr->params.field_per_buff;

	dprintk("%s: max image size = %x %d\n",  
		zr->device_name, size/8, size/8);

	zr->max_image = size;

	data[0x09] = (size >> 24) & 0xff;
	data[0x0a] = (size >> 16) & 0xff;
	data[0x0b] = (size >> 8) & 0xff;
	data[0x0c] = size & 0xff;

	data[0x11] = 0x01;	/* Scale Factor 1:1 */
	data[0x12] = 0x00;

	data[0x13] = 0xff;	/* allocation factor Must be 0xffffff */
	data[0x14] = 0xff;
	data[0x15] = 0xff;

	data[0x24] = 0x00;
	data[0x25] = 0x00;

	data[0x30] = 0x06; 	/* sync on end of VSYNC pulse; use external FI indicator */


	if (zr->video_bus_size == 8)
		data[0x30] |= 0x80;

	if (zr->params.odd_even)
		data[0x30] |= (1 << 3);

	/* limit pixels to range 16..235 as per CICIR-601 */
	data[0x30] |= (1 << 6);

	// polarity; zr and ks appear to have different opinions when to clock data in/out
	if (mode == BUZ_MODE_MOTION_COMPRESS || mode == BUZ_MODE_STILL_COMPRESS) {
		data[0x31] = 0x06;
	} else {
		data[0x31] = 0x86;
	}

	if ( zr->inv_vclk )
		data[0x31] ^= 0x80;

	switch(zr->params.HorDcm) {
	default:
	case 1:
		data[0x32] = 0;
		break;
	case 2:
		data[0x32] = 1;
		break;

	case 4:
		data[0x32] = 2;
		break;
	}

	if (zr->params.VerDcm == 2)
		data[0x32] |= 0x04;

	data[0x33] = 0x10;
	data[0x34] = 0x80;
	data[0x35] = 0x80;
	
	switch(mode) {
	case BUZ_MODE_MOTION_DECOMPRESS:
	case BUZ_MODE_STILL_DECOMPRESS:
		Vstart = tvn->Vjpeg;
		Hstart = tvn->Hjpeg;
		break;

	case BUZ_MODE_MOTION_COMPRESS:
	case BUZ_MODE_STILL_COMPRESS:
	default:
		Vstart = tvn->VStart;
		Hstart = tvn->HStart;
		break;
	}
	set_sync_generator( zr, data, mode,
		tvn->Ht-1, tvn->Wt-1,
		5, 99,
		Vstart-1, (Vstart-1) + tvn->Ha/2,
		Hstart-1, (Hstart-1) + tvn->Wa );

	set_active_area( zr, data, mode,
		zr->params.img_y + Vstart, (zr->params.img_y + Vstart) + zr->params.img_height,
		zr->params.img_x + Hstart, (zr->params.img_x + Hstart) + zr->params.img_width );
		
	set_subimg_area( zr, data, mode,
		zr->params.img_y + Vstart, (zr->params.img_y + Vstart) + zr->params.img_height,
		zr->params.img_x + Hstart, (zr->params.img_x + Hstart) + zr->params.img_width );

	memcpy(&data[0x060], SOF, sizeof(SOF));
	memcpy(&data[0x07a], SOS, sizeof(SOS));
	memcpy(&data[0x0c0], DRI, sizeof(DRI));
	memcpy(&data[0x0cc], DQT, sizeof(DQT));
	memcpy(&data[0x1d4], DHT, sizeof(DHT));

	lines = zr->params.img_height / zr->params.VerDcm;
	data[0x65] = lines >> 8;
	data[0x66] = lines;

	lines = zr->params.img_width / zr->params.HorDcm;
	data[0x67] = lines >> 8;
	data[0x68] = lines;

	len = zr->params.APP_len;
	if (len < 0) 
		len = 0;
	if (len > 60)
		len = 60;

	i = zr->params.APPn;
	if (i < 0) i = 0;
	if (i > 15) i = 15;

	data[0x380] = 0xff;
	data[0x381] = 0xe0 + i;
	data[0x382] = 0x00;
	data[0x383] = len + 2;
	for(i=0; i < len; i++)
		data[0x384+i] = zr->params.APP_data[i];


	len = zr->params.COM_len;
	if (len < 0) 
		len = 0;
	if (len > 60)
		len = 60;

	data[0x3c0] = 0xff;
	data[0x3c1] = 0xfe;
	data[0x3c2] = 0x00;
	data[0x3c3] = len + 2;
	for(i=0; i < len; i++)
		data[0x3c4+i] = zr->params.COM_data[i];

	for ( i=0; i<sizeof(mem_region)/sizeof(mem_region[0]); i++ ) {
		if (zr->write_memory(zr->cte_private, &data[ mem_region[i].addr ],
			mem_region[i].addr, mem_region[i].len))
			return -1;
	}
	
	return zr36060_load_memory(zr);
}


static void zr36060_set_jpg(struct zr36060_dev *zr, enum  buz_codec_mode mode)
{
	int		i;
	unsigned long	flags;

	spin_lock_irqsave(&zr->lock, flags);

	zr->jpg_que_head = 0;
	zr->jpg_que_tail = 0;

	zr->jpg_dma_head = 0;
	zr->jpg_dma_tail = 0;

	zr->jpg_seq_num = 0;

	for(i=0; i < zr->jpg_nbufs; i++) {
		zr->jpg_gbuf[i].state = BUZ_STATE_USER;
	}

	spin_unlock_irqrestore(&zr->lock, flags);
}


static void jpg_enable(struct zr36060_dev *zr, enum buz_codec_mode mode)
{
	int mode2 = (int)mode;
	enum   buz_codec_mode old_mode = zr->codec_mode;
	
	dprintk("%s: Setting mode 0x%x\n", zr->device_name, mode);
  
	if ( zr->codec_mode == mode ) {
  		dprintk( "zr: already in mode %i, ignoring jpg_enable\n", mode );
		return;
	}
	zr->codec_mode = mode;

	if ((zr->params.norm == VIDEO_MODE_PAL) || (zr->params.norm == VIDEO_MODE_SECAM)) {
		zr->field_time = 1000000/50;
	} else if (zr->params.norm == VIDEO_MODE_NTSC) {
		zr->field_time = 1001000/60;
	} else {
		zr->field_time = 1001000/60;
	}

	switch (mode2) {
	case BUZ_MODE_MOTION_COMPRESS:
		zr->field_delete = 2;
		zr36060_set_cap(zr, mode);
		zr36060_set_jpg(zr, mode);
		zr->codec_set_jpg(zr->cte_private, mode);
		zr->codec_set_video(zr->cte_private, mode);
		zr->start_transfer(zr->cte_private, zr->codec_mode);
		break;

	case BUZ_MODE_MOTION_DECOMPRESS:
		zr->codec_set_video(zr->cte_private, mode);
		zr36060_set_cap(zr, mode);
		zr36060_set_jpg(zr, mode);
		zr->codec_set_jpg(zr->cte_private, mode);
		break;

	case BUZ_MODE_IDLE:
	default:
		/* shut down processing */
		if ( old_mode ==  BUZ_MODE_MOTION_COMPRESS ) {
			zr->codec_set_jpg(zr->cte_private, mode);
			zr->codec_set_video(zr->cte_private, mode);
			zr->reset(zr->cte_private, 0);
		} else {
			zr->codec_set_jpg(zr->cte_private, mode);
			zr->reset(zr->cte_private, 0);
			zr->codec_set_video(zr->cte_private, mode);
		}
		break;
	}
}

#define M_SOF0  0xC0
#define M_SOF1  0xC1
#define M_DHT   0xC4
#define M_SOI   0xD8		/* Start Of Image (beginning of datastream) */
#define M_EOI   0xD9		/* End Of Image (end of datastream) */
#define M_SOS   0xDA		/* Start Of Scan (begins compressed data) */
#define M_DQT   0xDB
#define M_DRI	0xDD
#define M_APP0  0xE0
#define M_APP1  0xE1

#define OM_NOP	0
#define OM_SKIP	1
#define OM_NEXT	2

static int
parse_jpeg_markers(const u8 *jpegdata, const int jpeglen, int *soi, int* eoi, const int stop_at_eoi)
{
	register int p;	/* index of next byte to be read from jpegdata */
	int s;		/* states: scan-for-soi, scan-for-markers, process-padding */
	unsigned len;
	u8 m;

	static const u8 on_marker[] = {
		0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,		// 0x00
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,		// 0x10
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,		// 0x20
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,		// 0x30
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,		// 0x40
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,		// 0x50
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,		// 0x60
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,		// 0x70
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,		// 0x80
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,		// 0x90
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,		// 0xa0
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,		// 0xb0
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,		// 0xc0
		0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 1, 1, 1, 1, 1, 1,		// 0xd0
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,		// 0xe0
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1		// 0xf0
	};

	p = 0;
	s = 0;

	for(;;) {
		do {
			if (p == jpeglen)
				return p;
		} while (jpegdata[p++] != 0xFF);

		do {
			if (p == jpeglen)
				return p;
		} while ((m=jpegdata[p++]) == 0xFF);

		switch (s) {
		case 1:
			if (m == M_EOI) {
				*eoi = p;
				if (stop_at_eoi)
					return p;
			}
			break;
		case 0:
			if (m == M_SOI)
				*soi = p-2;
			else
				continue;
			break;
		case 2:
		default:
			return p-2;
		}
			
		switch (on_marker[m]) {
		case OM_SKIP:
			if (p+2 > jpeglen)
				return p;
			len = ((unsigned)jpegdata[p] << 8) + jpegdata[p + 1];
			if ( len < 2 )
				return p;
			p += 2;
			len -= 2;
			if ( (p+len) > jpeglen )
				return p;
			p += len; /* skip marker payload */
			break;

		case OM_NEXT:
			s++;
			break;
		case OM_NOP:
		default:
			break;
		}
	}
}


/*
 *   Queue a MJPEG buffer for capture/playback
 */
static int jpg_qbuf(struct zr36060_dev *zr, int frame, enum buz_codec_mode mode)
{
	int			offset;
	int			count;
	struct buz_gbuffer	*gbuf;
	u8			*mem;
	int			i;
	int			soi, eoi;

	/* Check if buffers are allocated */
	if(!zr->jpg_buffers_allocated) {
		dprintk("%s: jpg_qbuf: buffers not yet allocated\n", zr->device_name);
		return -ENOMEM;
	}

	/* Does the user want to stop streaming? */
	if (frame < 0) {
		if (zr->codec_mode == mode) {
			dprintk("%s: (REMOVEME) Stop the streaming mode.\n", zr->device_name);
			jpg_enable(zr, BUZ_MODE_IDLE);
			return 0;
		}
		else {
			dprintk("%s: jpg_qbuf - stop streaming but not in streaming mode\n", 
				zr->device_name);
			return -EINVAL;
		}
	}

	/* No grabbing outside the buffer range! */
	if (frame >= zr->jpg_nbufs) {
		dprintk("%s: jpg_qbuf: buffer %d out of range\n", zr->device_name, frame);
		return -EINVAL;
	}
   
	/* what is the codec mode right now? */
	if ( zr->codec_mode == BUZ_MODE_IDLE ) {
		/* Ok load up the zr36060 and go */
		dprintk("%s: Starting up zr36060\n", zr->device_name);
		jpg_enable(zr, mode);
	}
	else if ( zr->codec_mode != mode ) {
		/* wrong codec mode active - invalid */
		dprintk("%s: jpg_qbuf - codec in wrong mode\n", zr->device_name);
		return -EINVAL;
	}

	gbuf = &zr->jpg_gbuf[frame];
   
	/* make sure a grab isn't going on currently with this buffer */
	switch (gbuf->state) {
		case BUZ_STATE_USER:
			break;

		default:
		case BUZ_STATE_DMA:
		case BUZ_STATE_PEND:
		case BUZ_STATE_DONE:
			dprintk("%s: jbuf - buffer %i in bad state\n", zr->device_name, frame );
			return -EBUSY;         /* what are you doing? */
	}

	gbuf->times_done = 0;
	gbuf->fields_seen = 0;
	gbuf->offset = 0;

	if (mode == BUZ_MODE_MOTION_DECOMPRESS) {
		mem = gbuf->virt;

		offset = 0;
		for(i=0; i < zr->params.field_per_buff; i++) {
			soi = eoi = -9;
			count = parse_jpeg_markers(mem+offset, 
					zr->jpg_bufsize - offset, &soi, &eoi,
					(i+1 == zr->params.field_per_buff) );
			if ((soi < 0) || (eoi < 0)) {
				dprintk("%s: bad jpg data (%i %i)\n", zr->device_name, soi, eoi);
				return -EINVAL;
			}
			if (soi) {
				/* fill garbage before field with escape code */
				memset(mem+offset, 0xff, soi);
			}
			offset += count;
			gbuf->end[ i ] = offset;
		}
		gbuf->bs.length = offset;
	}
	else if (mode == BUZ_MODE_MOTION_COMPRESS) {
		gbuf->bs.length = 0;
	}
	
	/* since there is at least one unused buffer there's room for at least one more pend[] entry */
	zr->jpg_pend[zr->jpg_que_head & BUZ_MASK_FRAME] = frame;
	gbuf->state = BUZ_STATE_PEND;
	zr->jpg_que_head++;

	return 0;
}

/*
 *   Sync on a MJPEG buffer
 */
static int jpg_sync(struct zr36060_dev *zr, struct buz_sync * bs)
{
	unsigned long flags;
	int frame;

	if( zr->codec_mode != BUZ_MODE_MOTION_DECOMPRESS &&
		zr->codec_mode != BUZ_MODE_MOTION_COMPRESS ) {
		dprintk("%s: BUZIOCSYNC: - codec not in streaming mode\n", 
			zr->device_name);
		return -EINVAL;
	}
	
	if (zr->restart) {
		zr->restart = 0;
		dprintk("%s: restart zr36060 stream\n", zr->device_name);
		zr36060_set_cap(zr, zr->codec_mode);
	}
	zr->start_transfer(zr->cte_private, zr->codec_mode);

	spin_lock_irqsave(&zr->lock, flags);

	while(zr->jpg_que_tail == zr->jpg_dma_tail) {
		spin_unlock_irqrestore(&zr->lock, flags);
		
		if (zr->restart) {
			zr->restart = 0;
			dprintk("%s: restart zr36060 stream\n", zr->device_name);
			zr36060_set_cap(zr, zr->codec_mode);
		}
		zr->start_transfer(zr->cte_private, zr->codec_mode);
		
		interruptible_sleep_on(&zr->jpg_capq);
		if (signal_pending(current))
			return -ERESTARTSYS;

		spin_lock_irqsave(&zr->lock, flags);
	}

	frame = zr->jpg_pend[zr->jpg_que_tail & BUZ_MASK_FRAME];

	/* buffer should now be in BUZ_STATE_DONE */

	if(zr->jpg_gbuf[frame].state != BUZ_STATE_DONE)
		printk(KERN_WARNING "%s: jpg_sync - internal error\n", zr->device_name);

	memcpy(bs, &(zr->jpg_gbuf[frame].bs), sizeof(struct buz_sync));
	zr->jpg_gbuf[frame].state = BUZ_STATE_USER;

	zr->jpg_que_tail++;

	spin_unlock_irqrestore(&zr->lock, flags);

	return 0; /* needs to return the processed frame (fixme?) */
}


int
zr36060_done(struct zr36060_dev *zr, struct timeval *timestamp)
{
	int	frame;
	struct	buz_gbuffer *gbuf;

	if (zr->codec_mode != BUZ_MODE_STILL_COMPRESS && 
	    zr->codec_mode != BUZ_MODE_MOTION_COMPRESS &&
	    zr->codec_mode != BUZ_MODE_STILL_DECOMPRESS &&
	    zr->codec_mode != BUZ_MODE_MOTION_DECOMPRESS) {
		return 0;
	}

	if (zr->jpg_dma_tail == zr->jpg_dma_head) {
		return 0;
        }

	frame = zr->jpg_pend[zr->jpg_dma_tail & BUZ_MASK_FRAME];
	gbuf = &zr->jpg_gbuf[frame];
		
	if (zr->codec_mode == BUZ_MODE_STILL_COMPRESS ||
	    zr->codec_mode == BUZ_MODE_MOTION_COMPRESS) {
		
		gbuf->bs.length = gbuf->offset;
		gbuf->bs.seq = zr->jpg_seq_num++ - zr->field_delete;

	} else {
		gbuf->bs.length = 0;
	}

	gbuf->bs.timestamp = *timestamp;
	gbuf->state = BUZ_STATE_DONE;
	zr->jpg_dma_tail++;

	wake_up_interruptible(&zr->jpg_capq);
	return 0;
}
EXPORT_SYMBOL(zr36060_done);


int
zr36060_current_last(struct zr36060_dev *zr)
{
	if (zr->jpg_dma_head+1 == zr->jpg_que_head) {
		return 1;
	}
	return 0;
}
EXPORT_SYMBOL(zr36060_current_last);


u8 *
zr36060_current_buffer(struct zr36060_dev *zr, u32 *left, int field)
{
	int			frame;
	struct	buz_gbuffer	*gbuf;

	if (zr->codec_mode == BUZ_MODE_IDLE) {
		return 0;
	}

	if (zr->jpg_dma_head == zr->jpg_que_head) {
		return 0;
	}
	frame = zr->jpg_pend[zr->jpg_dma_head & BUZ_MASK_FRAME];
	gbuf = &zr->jpg_gbuf[frame];

	if (zr->codec_mode == BUZ_MODE_STILL_COMPRESS ||
	    zr->codec_mode == BUZ_MODE_MOTION_COMPRESS) {

                *left = zr->jpg_bufsize - gbuf->offset;

	} else {
		gbuf->state = BUZ_STATE_DMA;

		if (gbuf->offset == 0) {

			/* at start of buffer */
			
			if ((field!=0) && (zr->params.TmpDcm == 1)) {
				/* got field info, and no tmpdcm is active
					(field is dontcare otherwise (always same type)) */
			
				if (zr->params.field_per_buff == 1) {
			
					/* fixme: need better test for which type of field is in gbuf */
					int oddeven = zr->params.odd_even;
					if (zr->jpg_dma_head & 1)
						oddeven ^= 1; 
					if (	((oddeven == 1) && (field != 1)) ||
						((oddeven != 1) && (field != 2)) ) {
						/* field mismatch */
					
						if (zr->jpg_dma_head+1 != zr->jpg_que_head) {
							/* it's not the replay buffer */

							/* do this field twice */
							gbuf->times_done = -1;
							dprintk("%s: fieldmis replay\n", zr->device_name);
						}
					}
				}
				else if (((zr->params.odd_even == 1) && (field != 1)) ||
					((zr->params.odd_even != 1) && (field != 2)) ) {
					/* field mismatch in multi-field-per-buff config */
			
					/* skip first field of buffer */

					gbuf->offset = gbuf->end[0];
					gbuf->fields_seen = 1;
					gbuf->times_done = -1;  /* play whole buffer again later */
					dprintk("%s: skip first %i\n", zr->device_name, gbuf->end[0] );
				}
			}
		}
		*left = gbuf->end[ gbuf->fields_seen ] - gbuf->offset;
	}
	return &gbuf->virt[gbuf->offset];
}
EXPORT_SYMBOL(zr36060_current_buffer);


/*	@return 0: buffer partially emptied
		1: buffer empty, advancing to next
		2: buffer empty, this was a replay
*/
int
zr36060_current_get(struct zr36060_dev *zr, u32 size)
{
	int			frame;
	struct	buz_gbuffer	*gbuf;
	int rv = 0;

	frame = zr->jpg_pend[zr->jpg_dma_head & BUZ_MASK_FRAME];
	gbuf = &zr->jpg_gbuf[frame];
	
        gbuf->offset += size;

	if (gbuf->offset > gbuf->bs.length) {
		printk(KERN_ERR "%s: current get: got too much (%d bs.length %lu\n",
			zr->device_name, size, gbuf->bs.length);
	}

	if (gbuf->offset >= gbuf->end[gbuf->fields_seen]) {
		rv = 2;
		if (++gbuf->fields_seen >= zr->params.field_per_buff) {
			gbuf->fields_seen = 0;
			
			if (zr->params.TmpDcm == ++gbuf->times_done) {
				if (zr->jpg_dma_head+1 == zr->jpg_que_head) {
					/* re-use buffer */
					gbuf->times_done = 0;
				} else {
					zr->jpg_dma_head++;
					rv = 1;
				}
			}
			gbuf->offset = 0;
		}
	}
	return rv;
}
EXPORT_SYMBOL(zr36060_current_get);


int
zr36060_current_flush(struct zr36060_dev *zr)
{
	int			frame;
	struct	buz_gbuffer	*gbuf;
	
	if (zr->codec_mode == BUZ_MODE_IDLE) {
		return 1;
	}

	if (zr->jpg_dma_head == zr->jpg_que_head) {
		return 1;
	}

	frame = zr->jpg_pend[zr->jpg_dma_head & BUZ_MASK_FRAME];
	gbuf = &zr->jpg_gbuf[frame];
	
	gbuf->offset = 0;
	gbuf->fields_seen = 0;
	gbuf->times_done = 0;
	
	if (zr->codec_mode == BUZ_MODE_STILL_DECOMPRESS ||
	    zr->codec_mode == BUZ_MODE_MOTION_DECOMPRESS) {

		/* buffer might have caused the stall; don't replay but give back to user */
		struct timeval tv;
		
	    	if (zr36060_current_last(zr)) {
			/* can't give last buffer back */
			printk( "%s: choking on bad buffer\n", zr->device_name);
			return 0;
		}
		do_gettimeofday( &tv );
		zr->jpg_dma_head++;
		zr36060_done(zr, &tv);
	}
	return 0;
}
EXPORT_SYMBOL(zr36060_current_flush);


int
zr36060_current_put(struct zr36060_dev *zr, u32 size, int *pfield, struct timeval *stamp)
{
	int			frame;
	int			count;
	struct	buz_gbuffer	*gbuf;
	u8			*mem;
	u8			*buffer;
	u8			*mem_soi[10];
	u8			*mem_eoi[10];
	u8			do_buf_done[10];
	struct timeval		buf_stamp[10];
	int			i;
	int			field_count;
	int			buffer_used;
	int			soi;
	int			eoi;
	u32			used;
	struct	timeval		active_stamp;
	int			field = *pfield;
	
	used = 0;

	if (zr->codec_mode != BUZ_MODE_STILL_COMPRESS && 
	    zr->codec_mode != BUZ_MODE_MOTION_COMPRESS) {
		return used;
	}
	
	if (zr->jpg_seq_num < zr->field_delete) {
		zr->jpg_seq_num++;
		return size;
	}

	if (zr->jpg_dma_head == zr->jpg_que_head) {
		return used;
	}
	
	frame = zr->jpg_pend[zr->jpg_dma_head & BUZ_MASK_FRAME];
	gbuf = &zr->jpg_gbuf[frame];
	
	mem = &gbuf->virt[gbuf->offset];

	field_count = 0;

	/*
	 * Count number of fields in buffer. Also keep track of
	 * where the soi and eoi marker are.
	 */
	for(i=0; i < 10; i++) {
		soi = eoi = -9;
		count = parse_jpeg_markers(mem, size-used, &soi, &eoi, 0);

		if (soi < 0 || eoi < 0) 
			break;

		field_count++;
		mem_soi[i] = &mem[soi];
		mem_eoi[i] = &mem[eoi]; /* using mem[count] would be nice for dw alignment but
						appears to break some apps */
		mem += count;
		used += count;
	}

	/* If no field in buffer then do nothing */
	if (field_count == 0) {
		return used;
	}

	mem = &gbuf->virt[gbuf->offset];

	/*
	 * adjust timestamp and field if more then 1 field
	 * in buffer.
	 */
	active_stamp = *stamp;
	if (field_count > 1)
		add_time(&active_stamp, - ((field_count-1) * zr->field_time));
	if ((field_count & 1) == 0)
		field ^= 3;

	buffer_used = 0;
	for(i=0; i < field_count; i++, field ^= 3) {
		do_buf_done[ i ] = 0;
		if ((zr->jpg_dma_head + buffer_used) == zr->jpg_que_head) {
			field_count = i;
			break;
		}
		frame = zr->jpg_pend[(zr->jpg_dma_head + buffer_used) & BUZ_MASK_FRAME];
		gbuf = &zr->jpg_gbuf[frame];
		
		/* drop odd or even fields if tmpdcm is on */
		if (zr->params.TmpDcm == 2) {
			if (((zr->params.odd_even == 1) && (field != 1)) ||
			    ((zr->params.odd_even != 1) && (field != 2)) ) {
				gbuf->offset = 0;
				continue;
			}
		}

		/* check if correct field is coming into first position of buffer. */
		if ((gbuf->fields_seen==0) && (zr->params.field_per_buff>1)) {
			if (((zr->params.odd_even == 1) && (field != 1)) ||
			    ((zr->params.odd_even != 1) && (field != 2)) ) {
				gbuf->offset = 0;
				dprintk("%s: dump, wrong field\n", 
				       zr->device_name);
				continue;
			}
		}
		
		buffer = &gbuf->virt[gbuf->offset];
		
		do_buf_done[ i ] = 0;
		buf_stamp[ i ] = active_stamp;
		add_time(&active_stamp, zr->field_time);

		count = (mem_eoi[i] - mem_soi[i]);
		if (mem_soi[i] != buffer) {
			memmove(buffer,  mem_soi[i], count);
		}
		gbuf->offset += count;

		if (++gbuf->fields_seen < zr->params.field_per_buff) {
			continue;
		}

		do_buf_done[ i ] = 1;
		buffer_used++;
	}
	
	for(i = 0; i < field_count; i++) {
		if ( do_buf_done[ i ] ) {
			zr->jpg_dma_head++;
			zr36060_done(zr, &buf_stamp[ i ]);
		}
	}
	
	*pfield = field;
	*stamp = active_stamp;
	
	return used;	
}
EXPORT_SYMBOL(zr36060_current_put);



int zr36060_ioctl(struct zr36060_dev *zr, int command, void *arg)
{
	int rv;
	
	if (zr->initialized == 0) {
		dprintk("%s: not configure\n", zr->device_name);
		return -1;
	}

	switch(command) {


	case ZR36060_DEF_PARAMS:
		dprintk("%s: ioctl ZR36060_DEF_PARAMS\n", zr->device_name);
		return zr36060_default_params(zr);

	case ZR36060_GET_PARAMS:
		zr->params.major_version = 0;
		zr->params.minor_version = 10;
		dprintk("%s: ioctl ZR36060_GET_PARAMS\n", zr->device_name);
		memcpy(arg, &zr->params, sizeof(zr->params));
		return 0;

	case ZR36060_SET_PARAMS:
		dprintk("zr36060: ioctl ZR36060_SET_PARAMS\n");
		if (zr->codec_mode != BUZ_MODE_IDLE) {
			dprintk("%s: SET_PARAMS called but zr36060 in capture/play mode\n", 
			       zr->device_name);
			return -EINVAL;
		}

		rv = zr36060_check_params(zr, (struct buz_params *)arg);
		if (rv)
			return rv;

		zr->params = *(struct buz_params *)arg;
		return 0;

	case ZR36060_SET_INPUT:
		dprintk("zr36060: ioctl ZR36060_SET_INPUT\n");
		if (zr->codec_mode != BUZ_MODE_IDLE) {
			dprintk("%s: SET_INPUT called but zr36060 in capture/play mode\n", 
			       zr->device_name);
			return -EINVAL;
		}
		if(*(int *)arg < 0 || *(int *)arg > zr->input_count) {
			dprintk("%s SET INPUT: input = %d, must be 0-%d\n",
			zr->device_name, *(int *)arg, zr->input_count);
			return -EINVAL;
		}

		zr->params.input = *(int *)arg;
		return 0;


	case ZR36060_SET_NORM:
		dprintk("zr36060: ioctl ZR36060_SET_NORM\n");
		if (zr->codec_mode != BUZ_MODE_IDLE) {
			dprintk("%s: SET_NORM called but zr36060 in capture/play mode\n", 
			       zr->device_name);
			return -EINVAL;
		}


		zr->params.norm = *(int *)arg;
		return 0;

	case ZR36060_GET_MODE:
		dprintk("%s: ioctl ZR36060_GET_MODE\n", zr->device_name);
		*(enum buz_codec_mode *)arg = zr->codec_mode;
		return 0;

	case ZR36060_SET_MODE:
		dprintk("%s: ioctl ZR36060_SET_MODE\n", zr->device_name);
		jpg_enable(zr, *(enum buz_codec_mode *)arg);
		return 0;

	case ZR36060_REQBUFS:
	{
		struct buz_requestbuffers *br;

		dprintk("%s: ioctl ZR36060_REQBUFS\n", zr->device_name);
		br = (struct buz_requestbuffers *)arg;

		if(zr->jpg_buffers_allocated) {
			dprintk("%s: buffers already allocated\n", zr->device_name);
			return -EINVAL;
		}

		/* Enforce reasonable lower and upper limits */
		if(br->count<4)
			br->count = 4; /* Could be chosen smaller */

		if(br->count>BUZ_MAX_FRAME) 
			br->count = BUZ_MAX_FRAME;

		br->size = PAGE_ALIGN(br->size);

		if(br->size < 8192)
			br->size = 8192; /* Arbitrary */

		/* br->size is limited by 1 page for the stat_com 
		   tables to a Maximum of 2 MB */
		if(br->size > (512*1024)) 
			br->size = (512*1024); /* 512 K should be enough */

			
		if(zr->need_contiguous && br->size>MAX_KMALLOC_MEM) 
			br->size = MAX_KMALLOC_MEM;

		zr->jpg_nbufs   = br->count;
		zr->jpg_bufsize = br->size;

		if(jpg_fbuffer_alloc(zr)) 
			return -ENOMEM;

		return 0;
	}

	case ZR36060_CANBUFS:
		dprintk("%s: ioctl ZR36060_CANBUFS\n", zr->device_name);
		jpg_fbuffer_free(zr);
		zr->jpg_nbufs = 0;
		return 0;

	case ZR36060_SYNC:
		dprintk("%s: ioctl ZR36060_SYNC\n", zr->device_name);
		return jpg_sync(zr, (struct buz_sync *)arg);

	case ZR36060_QBUF_CAPT:
		dprintk("%s: ioctl ZR36060_QBUF_CAPT\n", zr->device_name);
		return jpg_qbuf(zr, *(int *)arg, BUZ_MODE_MOTION_COMPRESS);

	case ZR36060_QBUF_PLAY:
		dprintk("%s: ioctl ZR36060_QBUF_PLAY\n", zr->device_name);
		return jpg_qbuf(zr, *(int *)arg, BUZ_MODE_MOTION_DECOMPRESS);

	case ZR36060_RESTART:
		dprintk("%s: ioctl ZR36060_RESTART\n", zr->device_name);
		zr->restart = 1;
		wake_up_interruptible(&zr->jpg_capq);
		return 0;

	default:
		return -1;
	}
	return -1;
}
EXPORT_SYMBOL(zr36060_ioctl);


int
zr36060_read_proc(char *page, char **start, off_t offset, int len, int *eof, void *data)
{
	struct zr36060_dev *zr = (struct zr36060_dev *)data;
	char	buffer[1024];
	char	*s = page;

	int	i;

	len = 0;

	zr->read_memory(zr->cte_private, buffer, 0, 1023);

	for(i=0; i < 1024; i++) {
		if (i > 0 && (i % 16) == 0) {
			s = s + sprintf(s, "\n");
		}

		if ((i % 16) == 0) {
			s = s + sprintf(s, "%03x:", i);
		}

		s = s + sprintf(s, " %02x", buffer[i] & 0xff);
	}
		
	len = s - page;

	return len;
}

int
zr36060_write_proc(struct file *file, const char *buf, unsigned long count, 
void *data)
{
	return count;
}



/* This is run from mgavideo.c - it attaches this module to the "public" 
   communication paths */
int
zr36060_attach(struct zr36060_dev *zr, void *cte_private, int inv_vclk)
{
	int	j;

	zr->cte_private = cte_private;

	init_waitqueue_head(&zr->jpg_capq);

	dprintk("%s: attaching zr36060\n", zr->device_name);

	zr->initialized = 0;
	zr->user = 0;
	zr->inv_vclk = inv_vclk;


	spin_lock_init(&zr->lock);

	zr->jpg_nbufs   = 0;
	zr->jpg_bufsize = 0;
	zr->jpg_buffers_allocated = 0;

#ifdef	NEVER
	zr->params.norm     = default_norm  ? 1 : 0; /* Avoid nonsense settings from user */
	zr->params.input    = default_input ? 1 : 0; /* Avoid nonsense settings from user */
#else
	zr->params.norm     = 1;
	zr->params.input    = 0;
#endif

	for (j = 0; j < BUZ_MAX_FRAME; j++) {
		zr->jpg_gbuf[j].virt       = 0;
		zr->jpg_gbuf[j].state          = BUZ_STATE_USER;
		zr->jpg_gbuf[j].bs.frame       = j;
	}

	dprintk("%s: reset codec interface\n", zr->device_name);
	zr->reset(zr->cte_private, 1);

	zr36060_default_params(zr);

	zr->initialized = 1;

#ifdef	NEVER
	mga_create_proc_entry(mga, "zr36060", zr36060_read_proc, 
		zr36060_write_proc, zr);
#endif

#ifdef	NEVER
	MOD_INC_USE_COUNT;
#endif
	return 0;
}
EXPORT_SYMBOL(zr36060_attach);


int
zr36060_detach(struct zr36060_dev *zr36060)
{
#ifdef	NEVER
	MOD_DEC_USE_COUNT;
#endif
	return 0;
}



