/*
 *
 *    Copyright 1995-2000 Regents of the University of Minnesota
 *    Portions Copyright 1999-2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 */

#include "ogfs_debug.h"
#include "ogfs.h"
#include "blklist.h"
//#include "bits.h"
#include "dio.h"
#include "glock.h"
#include "glops.h"
#include "file.h"
#include "fs.h"
#include "log.h"
#include "trans.h"

#define AVL_MIN_RG_COUNT 10
#define rd_avl_empty ((struct ogfs_rgrpd *)NULL)
#define BLKALLOC_INTERNAL_NOENT ((uint32)0xffffffff)

#include "rg_avl.c"

/**
 * ogfs_rgrpd_hold - increment the useage count on a resource group
 * @rgd: the resource group
 *
 * Returns: the rgrp that is passed in
 */

static __inline__ ogfs_rgrpd_t *
ogfs_rgrpd_hold(ogfs_rgrpd_t * rgd)
{
	ENTER(GFN_RGRPD_HOLD);
	if (rgd)
		atomic_inc(&rgd->rd_count);
	EXIT(GFN_RGRPD_HOLD);
	return rgd;
}

/**
 * __ogfs_blk2rgrpd_list - Find the RGRP for a given block number (using list)
 * @sdp: the filesystem
 * @blk: the block number
 *
 * Find the Resource group for a given block number.
 *
 * Returns: The resource group, or NULL if not found
 */

ogfs_rgrpd_t *
__ogfs_blk2rgrpd_list(ogfs_sbd_t * sdp, uint64 blk)
{
	struct list_head *tmp, *head;
	ogfs_rgrpd_t *rgd;
	ogfs_rindex_t *ri;

	ENTER(GFN_BLK2RGRPD_LIST);

	head = tmp = &sdp->sd_rglist;
	for (;;) {
		rgd = NULL;
		tmp = tmp->next;
		if (tmp == head)
			break;

		rgd = list_entry(tmp, ogfs_rgrpd_t, rd_list);
		ri = &rgd->rd_ri;

		if (blk >= (ri->ri_data1 + ri->ri_data))
			continue;
		if (blk >= ri->ri_data1)
			break;

		rgd = NULL;
		break;
	}

	EXIT(GFN_BLK2RGRPD_LIST);

	return rgd;
}

/**
 * __ogfs_blk2rgrpd_tree - Find RGRP for a given data block (using tree)
 * @sdp: The OGFS superblock
 * @n: The data block number
 *
 * Returns: Ths resource group, or NULL if not found
 */

static ogfs_rgrpd_t *
__ogfs_blk2rgrpd_tree(ogfs_sbd_t * sdp, uint64 n)
{
	ogfs_rgrpd_t *tree = sdp->sd_rgavl;
	ogfs_rgrpd_t *ret;

	ENTER(GFN_BLK2RGRPD_TREE);

	for (;;) {
		ret = NULL;
		if (tree == rd_avl_empty)
			break;
		if (tree->rd_avl_key > n) {
			tree = tree->rd_avl_left;
		} else {
			ret = tree;
			if ((tree->rd_ri.ri_data1 <= n) &&
			    (tree->rd_ri.ri_data1 + tree->rd_ri.ri_data) > n)
				break;
			tree = tree->rd_avl_right;
		}
	}

	EXIT(GFN_BLK2RGRPD_TREE);
	return ret;
}

/**
 * ogfs_blk2rgrpd - Find resource group for a given data block number
 * @sdp: The OGFS superblock
 * @n: The data block number
 *
 * Returns: Ths resource group, or NULL if not found
 */

ogfs_rgrpd_t *
ogfs_blk2rgrpd(ogfs_sbd_t * sdp, uint64 n)
{
	ogfs_rgrpd_t *rgd;
	int set_cache = 0;

	ENTER(GFN_BLK2RGRPD);

	read_lock(&sdp->sd_rglock);
	rgd = sdp->sd_rgcache;

	if (rgd && (n >= rgd->rd_ri.ri_data1)
	    && (n < (rgd->rd_ri.ri_data1 + rgd->rd_ri.ri_data)))
		goto out;

	if (sdp->sd_rgavl)
		rgd = __ogfs_blk2rgrpd_tree(sdp, n);
	else
		rgd = __ogfs_blk2rgrpd_list(sdp, n);

	if (rgd && (sdp->sd_rgcache != rgd))
		set_cache = 1;

out:
	ogfs_rgrpd_hold(rgd);
	read_unlock(&sdp->sd_rglock);
	if (set_cache) {
		write_lock(&sdp->sd_rglock);
		sdp->sd_rgcache = rgd;
		write_unlock(&sdp->sd_rglock);
	}

	EXIT(GFN_BLK2RGRPD);
	return rgd;
}

/**
 * ogfs_rgrpd_get_by_num - get the @num-th RG
 * @sdp: The OGFS superblock
 * @num: The RG number to get
 *
 * Returns: The rgrp, or NULL if no rgrp with this number
 */

ogfs_rgrpd_t *
ogfs_rgrpd_get_by_num(ogfs_sbd_t * sdp, unsigned int num)
{
	struct list_head *tmp, *head;
	ogfs_rgrpd_t *rgd = NULL;
	unsigned int x = 0;

	ENTER(GFN_RGRPD_GET_BY_NUM);

	read_lock(&sdp->sd_rglock);
	head = &sdp->sd_rglist;
	for (tmp = head->next; tmp != head; tmp = tmp->next) {
		rgd = list_entry(tmp, ogfs_rgrpd_t, rd_list);
		if (x == num) {
			ogfs_rgrpd_hold(rgd);
			break;
		}
		rgd = NULL;
		x++;
	}
	read_unlock(&sdp->sd_rglock);

	EXIT(GFN_RGRPD_GET_BY_NUM);

	return rgd;
}

/**
 * __ogfs_find_rgrpd_list - Find an rgrp by disk block (using the list)
 * @sdp: The OGFS superblock
 * @n: The block number
 *
 */

static ogfs_rgrpd_t *
__ogfs_find_rgrpd_list(ogfs_sbd_t * sdp, uint64 n)
{
	struct list_head *tmp, *head;
	ogfs_rgrpd_t *rgrp;

	ENTER(GFN_FIND_RGRPD_LIST);

	tmp = head = &sdp->sd_rglist;
	for (;;) {
		rgrp = NULL;
		tmp = tmp->next;
		if (tmp == head)
			break;
		rgrp = list_entry(tmp, ogfs_rgrpd_t, rd_list);
		if (rgrp->rd_ri.ri_addr < n)
			continue;
		if (rgrp->rd_ri.ri_addr > n)
			break;
		EXIT(GFN_FIND_RGRPD_LIST);
		return rgrp;
	}

	EXIT(GFN_FIND_RGRPD_LIST);

	return NULL;
}

/**
 * __ogfs_find_rgrpd_tree - Find an rgrp by disk block (using the AVL tree)
 * @sdp: The OGFS superblock
 * @n: The diskblock number
 *
 */

static ogfs_rgrpd_t *
__ogfs_find_rgrpd_tree(ogfs_sbd_t * sdp, uint64 n)
{
	ogfs_rgrpd_t *tree = sdp->sd_rgavl;
	ogfs_rgrpd_t *ret;

	ENTER(GFN_FIND_RGRPD_TREE);

	for (;;) {
		ret = NULL;
		if (tree == rd_avl_empty)
			break;
		if (tree->rd_avl_key >= n) {
			ret = tree;
			if (tree->rd_avl_key == n)
				break;
			tree = tree->rd_avl_left;
		} else {
			tree = tree->rd_avl_right;
		}
	}

	EXIT(GFN_FIND_RGRPD_TREE);
	return ret;
}

/**
 * __ogfs_find_rgrpd - Find an rgrp by disk block
 * @sdp: The OGFS superblock
 * @n: The disk block
 *
 * Lookup a resource group by disk block number. First we look in the
 * cache. We often look up the same resource group several times in
 * a row, so this saves time. If we have an AVL tree, then we used that
 * method to perform a lookup, otherwise we fall back to scanning the
 * ordered list.
 */

static __inline__ ogfs_rgrpd_t *
__ogfs_find_rgrpd(ogfs_sbd_t * sdp, uint64 n)
{
	ogfs_rgrpd_t *rgd = sdp->sd_rgcache;

	ENTER(GFN_FIND_RGRPD);

	if (rgd && (rgd->rd_avl_key == n))
		goto out;
	if (sdp->sd_rgavl)
		rgd = __ogfs_find_rgrpd_tree(sdp, n);
	else
		rgd = __ogfs_find_rgrpd_list(sdp, n);

out:
	EXIT(GFN_FIND_RGRPD);
	return ogfs_rgrpd_hold(rgd);
}

/**
 * build_rgrp_avl - Build an AVL tree of resource groups
 * @sdp: The OGFS superblock
 *
 */

static void
build_rgrp_avl(ogfs_sbd_t * sdp)
{
	struct ogfs_rgrpd *rd;
	struct list_head *tmp, *head;

	ENTER(GFN_BUILD_RGRP_AVL);

	sdp->sd_rgavl = NULL;
	tmp = head = &sdp->sd_rglist;
	for (;;) {
		tmp = tmp->next;
		if (tmp == head)
			break;
		rd = list_entry(tmp, struct ogfs_rgrpd, rd_list);
		avl_insert(rd, &sdp->sd_rgavl);
	}

	EXIT(GFN_BUILD_RGRP_AVL);
}

/**
 * __ogfs_rgrpd_insert - Add a new rgrp into the list, ensure its unique
 * @rgd: The new rgrp
 *
 * Returns: NULL on success, the duplicate rgrpd if not unique
 */

static void
__ogfs_rgrpd_insert(ogfs_rgrpd_t * nrgd)
{
	ogfs_sbd_t *sdp = nrgd->rd_sbd;
	ogfs_rgrpd_t *rgd;
	struct list_head *tmp, *head;

	ENTER(GFN_RGRPD_INSERT);

	tmp = head = &sdp->sd_rglist;
	for (;;) {
		tmp = tmp->next;
		if (tmp == head)
			break;
		rgd = list_entry(tmp, ogfs_rgrpd_t, rd_list);

		if (rgd->rd_ri.ri_addr > nrgd->rd_ri.ri_addr)
			break;
	}

	list_add_tail(&nrgd->rd_list, tmp);
	if (sdp->sd_rgavl)	/* Add to AVL tree as well */
		avl_insert(nrgd, &sdp->sd_rgavl);
	sdp->sd_rgcount++;
	if (sdp->sd_rgcount >= AVL_MIN_RG_COUNT && !sdp->sd_rgavl)
		build_rgrp_avl(sdp);

	EXIT(GFN_RGRPD_INSERT);
}

/**
 * ogfs_rgrpd_get - Get an rgrp by disk block number
 * @sdp: The OGFS superblock
 * @n: The block number
 * @create: May create a new rgrp structure
 *
 */

ogfs_rgrpd_t *
ogfs_rgrpd_get(ogfs_sbd_t * sdp, uint64 n, int create)
{
	ogfs_rgrpd_t *rgrp, *nrgrp;

	ENTER(GFN_RGRPD_GET);

	read_lock(&sdp->sd_rglock);
	rgrp = __ogfs_find_rgrpd(sdp, n);
	read_unlock(&sdp->sd_rglock);
	if (rgrp) {
		write_lock(&sdp->sd_rglock);
		sdp->sd_rgcache = rgrp;
		 write_unlock(&sdp->sd_rglock);
		if (!create)
			goto out;
	}
	nrgrp = ogfs_alloc_rgrpd();
	if (!nrgrp)
		goto out;
	memset(nrgrp, 0, sizeof(ogfs_rgrpd_t));
	nrgrp->rd_ri.ri_addr = n;
	nrgrp->rd_sbd = sdp;
	ogfs_rgrpd_hold(nrgrp);
	ogfs_rgrpd_hold(nrgrp);
	/*
	 * Just because we created a new resource group doesn't mean that we are
	 * going to use it right away, so don't bother caching it.
	 */
	write_lock(&sdp->sd_rglock);
	if ((rgrp = __ogfs_find_rgrpd(sdp, n)) == NULL)
		__ogfs_rgrpd_insert(nrgrp);
	write_unlock(&sdp->sd_rglock);
	if (rgrp)
		ogfs_free_rgrpd(nrgrp);
	else
		rgrp = nrgrp;

out:
	EXIT(GFN_RGRPD_GET);
	return rgrp;
}

/**
 * ogfs_rgrpd_get_with_hint_tree - get with hint using AVL tree
 * @sdp: The OGFS superblock
 * @hint: The minimum disk address of the rgrp
 *
 * Returns: The rgrp, or NULL if no rgrp is below the hinted disk address
 */
static __inline__ ogfs_rgrpd_t *
__ogfs_rgrpd_get_with_hint_tree(ogfs_sbd_t * sdp, uint64 hint)
{
        ogfs_rgrpd_t *tree = sdp->sd_rgavl;
	ogfs_rgrpd_t *min = NULL;
		
	while (tree != rd_avl_empty) {
	  if (tree->rd_ri.ri_addr > hint) {
	    min = tree;
	    tree = tree->rd_avl_left;
	  } else if (tree->rd_ri.ri_addr < hint) {
	    tree = tree->rd_avl_right;
	  } else { 
	    min = tree;
	    break;
	  }
	}
	if (min)
	  ogfs_rgrpd_hold(min);
	return min;
}

/**
 * ogfs_rgrpd_get_with_hint_list - get with hint using linear list
 * @sdp: The OGFS superblock
 * @hint: The minimum disk address of the rgrp
 *
 * Returns: The rgrp, or NULL if no rgrp is below the hinted disk address
 */
static __inline__ ogfs_rgrpd_t *
__ogfs_rgrpd_get_with_hint_list(ogfs_sbd_t * sdp, uint64 hint)
{
	struct list_head *tmp, *head;
	ogfs_rgrpd_t *rgd = NULL;
	
	tmp = head = &sdp->sd_rglist;
	for (;;) {
		tmp = tmp->next;
		rgd = NULL;
		if (tmp == head)
			break;
		rgd = list_entry(tmp, ogfs_rgrpd_t, rd_list);
		if (rgd->rd_ri.ri_addr < hint)
			continue;
		ogfs_rgrpd_hold(rgd);
		break;
	}
	return rgd;
}

/**
 * ogfs_rgrpd_get_with_hint - get minimum disk address >= hint
 * @sdp: The OGFS superblock
 * @hint: The minimum disk address of the rgrp
 *
 * uses AVL tree of resource groups if present
 *
 * Returns: The rgrp, or NULL if no rgrp is below the hinted disk address
 */
ogfs_rgrpd_t *
ogfs_rgrpd_get_with_hint(ogfs_sbd_t * sdp, uint64 hint)
{
	ogfs_rgrpd_t *rgd = NULL;

	ENTER(GFN_RGRPD_GET_WITH_HINT);

	read_lock(&sdp->sd_rglock);
	
	if (sdp->sd_rgavl)
	  rgd = __ogfs_rgrpd_get_with_hint_tree(sdp, hint);
	else
	  rgd = __ogfs_rgrpd_get_with_hint_list(sdp, hint);
	
	read_unlock(&sdp->sd_rglock);
	
	EXIT(GFN_RGRPD_GET_WITH_HINT);
	return rgd;
}

/**
 * __ogfs_rgrpd_put - Free an rgrpd structure
 * @sdp: The OGFS superblock
 * @rgd: The rgrpd to put
 *
 * Called only from ogfs_rgrpd_put()
 */

void
__ogfs_rgrpd_put(ogfs_rgrpd_t * rgd)
{
	ENTER(GFN___RGRPD_PUT);

	OGFS_ASSERT(list_empty(&rgd->rd_list),);
	OGFS_ASSERT(rgd->rd_avl_left == NULL,);
	OGFS_ASSERT(rgd->rd_avl_right == NULL,);
	if (rgd->rd_gl) {
		ogfs_force_gl_drop(rgd->rd_sbd, rgd->rd_gl);
		if (rgd->rd_gl->gl_lvb)
			ogfs_unhold_lvb(rgd->rd_sbd, rgd->rd_gl);
		rgd->rd_gl->gl_object = NULL;
		ogfs_put_glstruct(rgd->rd_gl);
	}
	if (rgd->rd_bh)
		gfree(
			rgd->rd_bh,
			rgd->rd_ri.ri_length * sizeof(struct buffer_head *));
	if (rgd->rd_bits)
		gfree(rgd->rd_bits,
		      rgd->rd_ri.ri_length * sizeof(ogfs_bitmap_t));
	ogfs_free_rgrpd(rgd);

	EXIT(GFN___RGRPD_PUT);
}

/**
 * ogfs_clear_rgrpd - Clear up rgrps
 * @sdp: The OGFS superblock
 *
 */

void
ogfs_clear_rgrpd(ogfs_sbd_t * sdp)
{
	ogfs_rgrpd_t *rgd;

	ENTER(GFN_CLEAR_RGRPD);

	for (;;) {
		rgd = NULL;
		write_lock(&sdp->sd_rglock);
		if (!list_empty(&sdp->sd_rglist)) {
			rgd = list_entry(
				sdp->sd_rglist.next, ogfs_rgrpd_t, rd_list);
			list_del(&rgd->rd_list);
			INIT_LIST_HEAD(&rgd->rd_list);
			avl_remove(rgd, &sdp->sd_rgavl);
			rgd->rd_avl_left = rgd->rd_avl_right = NULL;
			rgd->rd_avl_height = 0;
			sdp->sd_rgcount--;
			if (sdp->sd_rgcache == rgd)
				sdp->sd_rgcache = NULL;
		}
		write_unlock(&sdp->sd_rglock);
		if (rgd == NULL)
			break;
		OGFS_ASSERT(atomic_read(&rgd->rd_count) == 1,);
		ogfs_rgrpd_put(rgd);
	}

	EXIT(GFN_CLEAR_RGRPD);
}

/**
 * ogfs_compute_bitstructs - Compute the bitmap sizes
 * @rgd: The resource group descriptor
 *
 * Returns: 0, or negative on error
 */

int
compute_bitstructs(ogfs_rgrpd_t * rgd)
{
	ogfs_sbd_t *sdp = rgd->rd_sbd;
	ogfs_bitmap_t *bits;
	uint32 length = rgd->rd_ri.ri_length;
	uint32 bytes_left, bytes;
	unsigned int x;
	int error = 0;

	ENTER(GFN_COMPUTE_BITSTRUCTS);

	rgd->rd_bh = (struct buffer_head **) gmalloc(
		length * sizeof(struct buffer_head *));
	if (!rgd->rd_bh) {
		error = -ENOMEM;
		goto out;
	}
	memset(rgd->rd_bh, 0, length * sizeof(struct buffer_head *));
	rgd->rd_bits = (ogfs_bitmap_t *) gmalloc(
		length * sizeof(ogfs_bitmap_t));
	if (!rgd->rd_bits) {
		gfree(rgd->rd_bh, length * sizeof(struct buffer_head *));
		rgd->rd_bh = NULL;
		error = -ENOMEM;
		goto out;
	}
	memset(rgd->rd_bits, 0, length * sizeof(ogfs_bitmap_t));
	bytes_left = rgd->rd_ri.ri_bitbytes;
	for (x = 0; x < length; x++) {
		bits = &rgd->rd_bits[x];
		if (length == 1) {
			bytes = bytes_left;
			bits->bi_offset = sizeof(ogfs_rgrp_t);
			bits->bi_start = 0;
			bits->bi_len = bytes;
		} else if (x == 0) {
			bytes = sdp->sd_sb.sb_bsize - sizeof(ogfs_rgrp_t);
			bits->bi_offset = sizeof(ogfs_rgrp_t);
			bits->bi_start = 0;
			bits->bi_len = bytes;
		} else if (x + 1 == length) {
			bytes = bytes_left;
			bits->bi_offset = sizeof(ogfs_rgrpbits_t);
			bits->bi_start = rgd->rd_ri.ri_bitbytes - bytes_left;
			bits->bi_len = bytes;
		} else {
			bytes = sdp->sd_sb.sb_bsize - sizeof(ogfs_rgrpbits_t);
			bits->bi_offset = sizeof(ogfs_rgrpbits_t);
			bits->bi_start = rgd->rd_ri.ri_bitbytes - bytes_left;
			bits->bi_len = bytes;
		}

		bytes_left -= bytes;
	}
	OGFS_ASSERT(!bytes_left,);
	OGFS_ASSERT((rgd->rd_bits[length - 1].bi_start +
		    rgd->rd_bits[length - 1].bi_len) * OGFS_NBBY ==
		   rgd->rd_ri.ri_data, con_printf("start=%u len=%u offset=%u\n",
						  rgd->rd_bits[length -
							       1].bi_start,
						  rgd->rd_bits[length -
							       1].bi_len,
						  rgd->rd_bits[length -
							       1].bi_offset);
		   ogfs_rindex_print(&rgd->rd_ri, TRUE);
	    );

out:
	EXIT(GFN_COMPUTE_BITSTRUCTS);

	return error;
}

/**
 * ogfs_rgrpd_update - Update the rgrpd structures
 * @gl: The glock covering the rindex inode
 *
 * This is rather inefficient as it stands. I'm not too worried about
 * it for two reasons: firstly, the only time this is called is when
 * someone updates the resource index, which will be a very infrequent
 * operation, and secondly because when we move to per-block locking the
 * problem will go away anyway. For now though, this will do.
 *
 * Returns: 0 on successful update, error code otherwise
 */

static int
ogfs_rgrp_update(ogfs_inode_t * ip)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	ogfs_rgrpd_t *rgd;
	ogfs_rindex_t buf;
	unsigned int rg;
	int error;

	ENTER(GFN_RGRP_UPDATE);

	OGFS_ASSERT((ip->i_di.di_size % sizeof(ogfs_rindex_t)) == 0,);
	ogfs_clear_rgrpd(sdp);
	for (rg = 0;; rg++) {
		error =
		    ogfs_internal_read(
			    ip, (char *) &buf, rg * sizeof(ogfs_rindex_t),
			    sizeof(ogfs_rindex_t));
		if (!error)
			break;
		if (error != sizeof(ogfs_rindex_t)) {
			if (error > 0)
				error = -EIO;
			goto fail;
		}
		rgd = ogfs_rgrpd_get(sdp, ogfs64_to_cpu(buf.ri_addr), CREATE);
		if (!rgd)
			goto fail;
		ogfs_rindex_in(&rgd->rd_ri, (char *) &buf);
		error = compute_bitstructs(rgd);
		if (error)
			goto fail;
		rgd->rd_gl = ogfs_get_glstruct(
			sdp, rgd->rd_ri.ri_addr, &ogfs_rgrp_glops, ip->i_gl,
			CREATE);
		if (!rgd->rd_gl) {
			error = -ENOMEM;
			goto fail;
		}
		rgd->rd_gl->gl_object = rgd;
		error = ogfs_hold_lvb(sdp, rgd->rd_gl);
		if (error)
			goto fail;
		error = ogfs_rgrp_lvb_init(rgd, FALSE);
		if (error)
			goto fail;
		ogfs_rgrpd_put(rgd);
	}
	sdp->sd_riinode_vn = ip->i_gl->gl_vn;

	EXIT(GFN_RGRP_UPDATE);
	return 0;

fail:
	ogfs_clear_rgrpd(sdp);
	EXIT(GFN_RGRP_UPDATE);

	return error;
}

/**
 * ogfs_rindex_hold - Grab a lock on the rindex
 * @sdp: The OGFS superblock
 *
 * We grab a lock in the rindex inode to make sure that it doesn't
 * change whilst we are performing an operation. We keep this lock
 * for quite long periods of time compared to other locks. This
 * doesn't matter, since its shared and it is very, very rarely
 * accessed in the exclusive mode.
 *
 * Returns: 0 on success, error code otherwise
 */

int
ogfs_rindex_hold(ogfs_sbd_t * sdp)
{
	ogfs_inode_t *ip = sdp->sd_riinode;
	ogfs_glock_t *gl = ip->i_gl;
	int error;

	ENTER(GFN_RINDEX_HOLD);

	error = ogfs_glock_i(ip, GL_SHARED);
	if (error)
		goto out;
	if (sdp->sd_riinode_vn != gl->gl_vn) {
		error = ogfs_rgrp_update(ip);
		if (error)
			ogfs_gunlock_i(ip, 0);
	}

out:
	EXIT(GFN_RINDEX_HOLD);
	return error;
}

/**
 * ogfs_rindex_release - Release hold over rindex
 * @sdp: The OGFS superblock
 *
 * Drop the lock on the resource index, allows filesystem grow and
 * shrink to take place once more.
 *
 */

void
ogfs_rindex_release(ogfs_sbd_t * sdp)
{
	ENTER(GFN_RINDEX_RELEASE);
	ogfs_gunlock_i(sdp->sd_riinode, 0);
	EXIT(GFN_RINDEX_RELEASE);
}

/**
 * ogfs_rgrp_read - Read in a RG's bitmaps
 * @rgd: the ogfs_rgrpd_t describing the RG to read in
 *
 * Read in RG bitmaps.  Must call ogfs_rgrp_relse() it free the bitmaps.
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_rgrp_read(ogfs_rgrpd_t * rgd)
{
	ogfs_sbd_t *sdp = rgd->rd_sbd;
	ogfs_glock_t *gl = rgd->rd_gl;
	unsigned int x;
	unsigned int length = rgd->rd_ri.ri_length;
	unsigned int y;
	int error;

	ENTER(GFN_RGRP_READ);

	for (x = 0; x < length; x++) {
		OGFS_ASSERT(!rgd->rd_bh[x],);
		rgd->rd_bh[x] = ogfs_dgetblk(sdp, rgd->rd_ri.ri_addr + x, gl);
		if (!rgd->rd_bh[x]) {
			error = -EIO;
			goto fail;
		}
	}
	for (y = 0; y < length; y++)
		ogfs_dreread(sdp, rgd->rd_bh[y], DR_START);
	for (y = length; y--;) {
		error = ogfs_dreread(sdp, rgd->rd_bh[y], DR_WAIT);
		if (error)
			goto fail;
		ogfs_metatype_check(
			rgd->rd_bh[y],
			(y) ? OGFS_METATYPE_RB : OGFS_METATYPE_RG);
	}
	ogfs_rgrp_in(&rgd->rd_rg, BH_DATA(rgd->rd_bh[0]));

	EXIT(GFN_RGRP_READ);

	return 0;

fail:
	while (x--) {
		brelse(rgd->rd_bh[x]);
		rgd->rd_bh[x] = NULL;
	}

	EXIT(GFN_RGRP_READ);
	return error;
}

/**
 * ogfs_rgrp_relse - Release RG bitmaps read in with ogfs_rgrp_read()
 * @rgd: the ogfs_rgrpd_t describing the RG to read in
 *
 */

void
ogfs_rgrp_relse(ogfs_rgrpd_t * rgd)
{
	int x, length = rgd->rd_ri.ri_length;

	ENTER(GFN_RGRP_RELSE);

	for (x = 0; x < length; x++) {
		brelse(rgd->rd_bh[x]);
		rgd->rd_bh[x] = NULL;
	}

	EXIT(GFN_RGRP_RELSE);
}

/**
 * ogfs_rgrp_lvb_fill - copy RG usage data out of the ogfs_rgrp_t into the
 * ogfs_rgrp_save_t
 * @rgd: the resource group data structure
 *
 */

void
ogfs_rgrp_lvb_fill(ogfs_rgrpd_t * rgd)
{
	ogfs_rgrp_t *rg = &rgd->rd_rg;
	ogfs_rgrp_save_t *rs = &rgd->rd_rs;

	ENTER(GFN_RGRP_LVB_FILL);

	rs->rs_magic = OGFS_MAGIC;
	rs->rs_free = rg->rg_free;
	rs->rs_useddi = rg->rg_useddi;
	rs->rs_freedi = rg->rg_freedi;
	rs->rs_usedmeta = rg->rg_usedmeta;
	rs->rs_freemeta = rg->rg_freemeta;
	rgd->rd_rs_dirty = TRUE;

	EXIT(GFN_RGRP_LVB_FILL);
}

/**
 * ogfs_rgrp_lvb_init - Init the data of a RG LVB
 * @rgd: the resource group data structure
 * @force: Fill in the data even if it seems that it's correct
 *
 * Returns:  0 on success, -EXXX on failure
 */

int
ogfs_rgrp_lvb_init(ogfs_rgrpd_t * rgd, int force)
{
	int error = 0;

	ENTER(GFN_RGRP_LVB_INIT);

	if (!force) {
		error = ogfs_glock_rg(rgd, GL_SHARED | GL_SKIP);
		if (error)
			goto out;
		if (rgd->rd_rs.rs_magic == OGFS_MAGIC) {
			ogfs_gunlock_rg(rgd, GL_SKIP);
			error = 0;
			goto out;
		}
		ogfs_gunlock_rg(rgd, GL_SKIP);
	}
	error = ogfs_glock_rg(rgd, 0);
	if (error)
		goto out;
	IFDB(GFN_RGRP_LVB_INIT,
	     ogfs_rindex_print(&rgd->rd_ri, FALSE);
	     ogfs_rgrp_print(&rgd->rd_rg, FALSE); ogfs_rgrp_verify(rgd);
	    );
	ogfs_rgrp_lvb_fill(rgd);
	rgd->rd_rs.rs_owner = 0;
	ogfs_rgrp_save_out(&rgd->rd_rs, rgd->rd_gl->gl_lvb);
	ogfs_sync_lvb(rgd->rd_sbd, rgd->rd_gl);
	rgd->rd_rs_dirty = FALSE;
	ogfs_gunlock_rg(rgd, 0);

out:
	EXIT(GFN_RGRP_LVB_INIT);

	return error;
}

/**
 * rgrp_lvb_own - Mark a RG LVB as being owned by a machine
 * @rgd: the resource group data structure
 *
 * If a machine ever sees a RG LVB that's owned by another
 * machine, it knows that the data in that LVB is stale because
 * the owning machine crashed.
 */

static void
rgrp_lvb_own(ogfs_rgrpd_t * rgd)
{
	ogfs_sbd_t *sdp = rgd->rd_sbd;

	ENTER(GFN_RGRP_LVB_OWN);

	if (rgd->rd_rs.rs_owner != sdp->sd_lockstruct.ls_jid + 1) {
		rgd->rd_rs.rs_owner = sdp->sd_lockstruct.ls_jid + 1;
		ogfs_rgrp_save_out(&rgd->rd_rs, rgd->rd_gl->gl_lvb);
		ogfs_sync_lvb(rgd->rd_sbd, rgd->rd_gl);
	}

	EXIT(GFN_RGRP_LVB_OWN);
}

/**
 * ogfs_rgrp_verify - Verify that a resource group is consistent
 * @sdp: the filesystem
 * @rgd: the rgrp
 *
 * Somebody should have already called ogfs_glock_rg() on this RG.
 */

void
ogfs_rgrp_verify(ogfs_rgrpd_t * rgd)
{
	ogfs_bitmap_t *bits = NULL;
	uint32 length = rgd->rd_ri.ri_length;
	uint32 count[OGFS_NBBY], tmp;
	unsigned int buf;
	int x;

	ENTER(GFN_RGRP_VERIFY);

	for (x = 0; x < OGFS_NBBY; x++) {
		count[x] = 0;

		for (buf = 0; buf < length; buf++) {
			bits = &rgd->rd_bits[buf];
			count[x] +=
			    ogfs_travel_bitmap(BH_DATA(rgd->rd_bh[buf]) +
					 bits->bi_offset, bits->bi_len, x, OGFS_BLKST_INVALID);
		}

		db_printf("count[%d] = %u\n", x, count[x]);
	}
	if (count[OGFS_BLKST_FREE] != rgd->rd_rg.rg_free)
		con_printf(
			"free data mismatch:  %u != %u\n",
			count[OGFS_BLKST_FREE], rgd->rd_rg.rg_free);
	tmp = rgd->rd_ri.ri_data -
	    (rgd->rd_rg.rg_usedmeta + rgd->rd_rg.rg_freemeta) -
	    (rgd->rd_rg.rg_useddi + rgd->rd_rg.rg_freedi) - rgd->rd_rg.rg_free;
	if (count[OGFS_BLKST_USED] != tmp)
		con_printf(
			"used data mismatch:  %u != %u\n",
			count[OGFS_BLKST_USED], tmp);
	if (count[OGFS_BLKST_FREEMETA] != rgd->rd_rg.rg_freemeta)
		con_printf(
			"free metadata mismatch:  %u != %u\n",
			count[OGFS_BLKST_FREEMETA], rgd->rd_rg.rg_freemeta);
	tmp = rgd->rd_rg.rg_usedmeta +
		(rgd->rd_rg.rg_useddi + rgd->rd_rg.rg_freedi);
	if (count[OGFS_BLKST_USEDMETA] != tmp)
		con_printf(
			"used metadata mismatch:  %u != %u\n",
			count[OGFS_BLKST_USEDMETA], tmp);

	EXIT(GFN_RGRP_VERIFY);
}

/**
 * try_rgrp_fit - See if a given reservation will fit in a given RG
 * @rgd: the RG data
 * @res: the ogfs_ipreserv_t structure describing the reservation
 * @accept_partial_meta_fit: if set, return as many meta blocks as possible if
 * the reservation does not fit completely
 *
 * Sets the $ir_datares field in @res.
 * Sets the $ir_metares field in @res.
 *
 * Returns: 1 on success, 0 on failure, -1 on partial success
 */

static int
try_rgrp_fit(
	ogfs_rgrpd_t * rgd, ogfs_ipreserv_t * res, int accept_partial_meta_fit)
{
	uint32 freeblks = rgd->rd_rg.rg_free;
	uint32 freemeta = rgd->rd_rg.rg_freemeta;
	uint32 metares = res->ir_metablks;
	uint32 datares = res->ir_datablks;
	int rc;

	ENTER(GFN_TRY_RGRP_FIT);

	/*  First take care of the data blocks required  */
	if (freeblks < datares)
		goto fail;
	freeblks -= datares;
	/*  Then take care of the dinodes  */
	if (rgd->rd_rg.rg_freedi < res->ir_diblks)
		metares += res->ir_diblks - rgd->rd_rg.rg_freedi;
	/*  Then take care of the metadata blocks  */
	while (freemeta < metares) {
		if (freeblks < OGFS_META_CLUMP) {
			datares += freeblks;
			freemeta += freeblks;
			freeblks = 0;
			break;
		}
		freeblks -= OGFS_META_CLUMP;
		datares += OGFS_META_CLUMP;
		freemeta += OGFS_META_CLUMP;
	}
	rc = 1;
	if (accept_partial_meta_fit) {
		if (rgd->rd_rg.rg_freedi < res->ir_diblks &&
		    freemeta < res->ir_diblks - rgd->rd_rg.rg_freedi) {
			goto fail;
		}
		if (freemeta < metares) {
			metares = freemeta;
			rc = -1;
		}
	} else {
		if (freemeta < metares) {
			goto fail;
		}
	}
	res->ir_rgd = rgd;
	res->ir_metares = metares;
	res->ir_datares = datares;

	EXIT(GFN_TRY_RGRP_FIT);
	return rc;

fail:
	EXIT(GFN_TRY_RGRP_FIT);
	return 0;
}

/**
 * __get_pure_metares - Calculate the number of meta blocks reserved in a
 * reservation by try_rgrp_fit()
 * @rgd:  the RG data
 * @res:  reservation filled by try_rgrp_fit
 *
 * Calculate the number of meta blocks reserved in a reservation by
 * try_rgrp_fit().  Assumes that th RG locks are held.
 *
 * Returns: the number of reserved non-dinode meta blocks
 */

static __inline__ uint32
__get_pure_metares(ogfs_rgrpd_t *rgd, ogfs_ipreserv_t *res)
{
	uint32 metares;

	ENTER(GFN___GET_PURE_METARES);

	metares = res->ir_metares;
	if (rgd->rd_rg.rg_freedi < res->ir_diblks) {
		metares -= (res->ir_diblks - rgd->rd_rg.rg_freedi);
	}

	EXIT(GFN___GET_PURE_METARES);
	return metares;
}

/**
 * __get_best_rg_fit - Find and lock the rg that best fits a reservation
 * @ret_rgd:  Pointer to the returned resource group
 * @res:  The target reservation
 * @sdp:  The superblock structure.
 *
 * Acquire resource group locks to allow for the maximum allocation
 * described by "res".  Looks for the resource group that best fits the
 * reservation.  Locks all resource groups to do this.
 *
 * Must be called with the rindex lock held.  Only to be called by
 * ogfs_inplace_reserv().
 *
 * Returns: 0 on success, -EXXX on failure
 */

static __inline__ int
__get_best_rg_fit(ogfs_rgrpd_t **ret_rgd, ogfs_ipreserv_t *res, ogfs_sbd_t *sdp)
{
	ogfs_ipreserv_t tres;
	ogfs_rgrpd_t *rgd;
	uint64 hint;
	uint64 best_hint = 0;
	uint64 max_hint;
	uint32 best_metares = 0;
	int have_best_match = 0;
	int error = 0;
	int rc;

	ENTER(GFN___GET_BEST_RG_FIT);

	*ret_rgd = NULL;
	res->ir_rgd = NULL;
	res->ir_metares = 0;
	res->ir_datares = 0;
	tres = *res;
	/* lock all rgs in ascending order and look for the best match */
	for (hint = 0;
	     (rgd = ogfs_rgrpd_get_with_hint(sdp, hint)) != NULL;
	     hint = rgd->rd_ri.ri_addr + 1) {
		error = ogfs_glock_rg(rgd, 0);
		if (error) {
			ogfs_rgrpd_put(rgd);
			*ret_rgd = NULL;
			break;
		}
		rc = try_rgrp_fit(rgd, &tres, 1);
		if (rc == 1) {
			/* perfect match */
			*ret_rgd = rgd;
			*res = tres;
			have_best_match = 1;
			best_hint = hint;
			break;
		} else if (rc == -1) {
			uint32 metares;

			metares = __get_pure_metares(rgd, res);
			if (!have_best_match || metares > best_metares) {
				/* this is the best match yet */
				*ret_rgd = rgd;
				*res = tres;
				best_metares = metares;
				have_best_match = 1;
				best_hint = hint;
			}
		}
	}
	max_hint = hint;
	/* unlock everything but the selected rg */
	for (hint = 0; hint < max_hint &&
		     (rgd = ogfs_rgrpd_get_with_hint(sdp, hint)) != NULL;
	     hint = rgd->rd_ri.ri_addr + 1) {
		if (!have_best_match || best_hint != hint || error != 0) {
			ogfs_gunlock_rg(rgd, 0);
			ogfs_rgrpd_put(rgd);
		}
	}
	if (error == 0 && !have_best_match) {
		error = -ENOSPC;
	}

	EXIT(GFN___GET_BEST_RG_FIT);
	return error;
}

/**
 * ogfs_inplace_reserv - Reserve space in the filesystem
 * @ip: the inode to reserve space for
 * @accept_partial_meta_fit: if the reservation fits in no resource group, try
 * to fit all the data blocks and as many meta data blocks as possible in the
 * resource group with the most free space instead.
 *
 * Acquire resource group locks to allow for the maximum allocation
 * described by "res".
 *
 * This should probably become more complex again, but for now, let's go
 * for simple (one resource group) reservations.
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_inplace_reserv(ogfs_inode_t * ip, int accept_partial_meta_fit)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	ogfs_rgrpd_t *rgd;
	ogfs_ipreserv_t *res = &ip->i_res;
	ogfs_ipreserv_t best_res;
	uint64 hint;
	uint64 best_hint = 0;
	uint64 start_blk;
	uint32 best_metares = 0;
	int error = 0;
	int have_best_match = 0;

	ENTER(GFN_INPLACE_RESERV);
	IFDB(GFN_INPLACE_RESERV,
	     db_printf(
		     "\nogfs_inplace_reserv: req: %u data, %u inodes, %u meta"
		     " @ %" FMT_64 "u, type %d\n",
		     res->ir_datablks, res->ir_diblks, res->ir_metablks,
		     res->ir_target, res->ir_trans);
		);
	OGFS_ASSERT(res->ir_diblks || res->ir_datablks || res->ir_metablks,);
	res->ir_rgd = NULL;
	res->ir_dialloced = 0;
	res->ir_metaalloced = 0;
	res->ir_dataalloced = 0;
	error = ogfs_rindex_hold(sdp);
	if (error)
		return error;
	/* Plan A: see if reservation fits completely in any resource group. */
	hint = res->ir_target;
	rgd = ogfs_rgrpd_get_with_hint(sdp, hint);
	if (rgd == NULL) {
		hint = 0;
		rgd = ogfs_rgrpd_get_with_hint(sdp, hint);
	}
	start_blk = rgd->rd_ri.ri_addr;
	do {
		error = ogfs_glock_rg(rgd, 0);
		if (error) {
			ogfs_rgrpd_put(rgd);
			ogfs_rindex_release(sdp);
			goto out;
		}
		/* Check if we may alloc the requested types of blocks here */
		switch (try_rgrp_fit(rgd, res, accept_partial_meta_fit)) {
		case 1:
			goto success;
		case -1:
			if (!have_best_match ||
			    res->ir_metares > best_res.ir_metares) {
				/* this is the best match yet */
				best_hint = hint;
				best_res = *res;
				best_metares = __get_pure_metares(rgd, res);
				have_best_match = 1;
			}
			break;
		default:
		case 0:
			break;
		}
		hint = rgd->rd_ri.ri_addr + 1;
		ogfs_gunlock_rg(rgd, 0);
		ogfs_rgrpd_put(rgd);
		rgd = ogfs_rgrpd_get_with_hint(sdp, hint);
		if (!rgd) {
			hint = 0;
			rgd = ogfs_rgrpd_get_with_hint(sdp, hint);
		}
	} while (rgd->rd_ri.ri_addr != start_blk);
	ogfs_rgrpd_put(rgd);
	if (!accept_partial_meta_fit || !have_best_match) {
		ogfs_rindex_release(sdp);
		error = -ENOSPC;
		goto out;
	}
	/* Plan B: the reservation does not fit in any resource group.  We
	 * already recorded the maximum free space in the last scan.  If it did
	 * not change in the mean time, use that one. */
	rgd = ogfs_rgrpd_get_with_hint(sdp, best_hint);
	error = ogfs_glock_rg(rgd, 0);
	if (error) {
		ogfs_rgrpd_put(rgd);
		ogfs_rindex_release(sdp);
		goto out;
	}
	switch (try_rgrp_fit(rgd, res, 1)) {
	case 1:
		goto success;
	case -1:
		if (__get_pure_metares(rgd, res) >= best_metares) {
			goto success;
		}
		break;
	default:
	case 0:
		break;
	}
	ogfs_gunlock_rg(rgd, 0);
	ogfs_rgrpd_put(rgd);
	/* Plan C: lock all resource groups and take the one with the most
	 * available space. */
	error = __get_best_rg_fit(&rgd, res, sdp);
	if (error) {
		ogfs_rindex_release(sdp);
		goto out;
	}

success:
	IFDB(GFN_INPLACE_RESERV,
	     db_printf(
		     "\nogfs_inplace_reserv: got: %u data, %u meta"
		     " in rg %" FMT_64 "u\n",
		     res->ir_datares, res->ir_metares,
		     res->ir_rgd->rd_ri.ri_addr);
		);
	IFDB(GFN_INPLACE_RESERV,
	     db_printf(
		     "\nogfs_inplace_reserv: rgd: %u data, %u (%u) inodes,"
		     " %u (%u) meta",
		     rgd->rd_rs.rs_free,
		     rgd->rd_rs.rs_freedi, rgd->rd_rs.rs_useddi,
		     rgd->rd_rs.rs_freemeta, rgd->rd_rs.rs_usedmeta);
		);

out:
	EXIT(GFN_INPLACE_RESERV);

	return error;
}

/**
 * ogfs_inplace_relse - release an inplace reservation
 * @ip: the inode the reservation was taken out on
 *
 * Release a reservation made by ogfs_inplace_reserv().
 */

void
ogfs_inplace_relse(ogfs_inode_t * ip)
{
	ogfs_ipreserv_t *res = &ip->i_res;
	ogfs_rgrpd_t *rgd = res->ir_rgd;

	ENTER(GFN_INPLACE_RELSE);

	IFDB(GFN_INPLACE_RELSE,
	     db_printf("  di alloced %d of %d\n", res->ir_dialloced,
		       res->ir_diblks);
	     db_printf("  meta alloced %d of %d\n", res->ir_metaalloced,
		       res->ir_metares);
	     db_printf("  data alloced %d of %d\n", res->ir_dataalloced,
		       res->ir_datares);
	    );
	OGFS_ASSERT(res->ir_dialloced <= res->ir_diblks,
		   con_printf("ir_dialloced = %u, ir_diblks = %u, trans = %d\n",
			      res->ir_dialloced, res->ir_diblks, res->ir_trans);
	    );
	OGFS_ASSERT(res->ir_metaalloced <= res->ir_metares,
		   con_printf
		   ("ir_metaalloced = %u, ir_metablks = %u, trans = %d\n",
		    res->ir_metaalloced, res->ir_metablks, res->ir_trans);
	    );
	OGFS_ASSERT(res->ir_dataalloced <= res->ir_datares,
		   con_printf
		   ("ir_dataalloced = %u, ir_datares = %u, trans = %d\n",
		    res->ir_dataalloced, res->ir_datares, res->ir_trans);
	    );
	res->ir_rgd = NULL;
	ogfs_gunlock_rg(rgd, 0);
	ogfs_rgrpd_put(rgd);
	ogfs_rindex_release(ip->i_sbd);

	EXIT(GFN_INPLACE_RELSE);
}

/**
 * blkalloc_internal - allocate a single block
 * @rgd: the resource group descriptor
 * @goal: the goal block in the RG
 * @old_state: the type of block to find
 * @new_state: the resulting block type
 * @do_it: if FALSE, we just find the block we would allocate
 *
 * Returns:  returns the block allocated or 0 if no block could be allocated
 */

static uint32
blkalloc_internal(
	ogfs_rgrpd_t * rgd, uint32 goal, unsigned char old_state,
	unsigned char new_state, int do_it)
{
	ogfs_sbd_t *sdp = rgd->rd_sbd;
	ogfs_bitmap_t *bits = NULL;
	uint32 length = rgd->rd_ri.ri_length;
	uint32 block = 0;
	uint32 buf, x;

	ENTER(GFN_BLKALLOC_INTERNAL);

	for (buf = 0; buf < length; buf++) {
		bits = &rgd->rd_bits[buf];
		if (goal < (bits->bi_start + bits->bi_len) * OGFS_NBBY)
			break;
	}
	OGFS_ASSERT(buf < length,);
	goal -= bits->bi_start * OGFS_NBBY;
	/*  "x <= length" because we're skipping over some of the first
	   buffer when the goal is non-zero.  */
	for (x = 0; x <= length; x++) {
		block = ogfs_bitfit(
			BH_DATA(rgd->rd_bh[buf]) + bits->bi_offset,
			bits->bi_len, goal, old_state);
		if (block != BFITNOENT)
			break;
		buf = (buf + 1) % length;
		bits = &rgd->rd_bits[buf];
		goal = 0;
	}
	if (block == BFITNOENT) {
		return BLKALLOC_INTERNAL_NOENT;
	}
	OGFS_ASSERT(x <= length,);
	if (do_it) {
		ogfs_trans_add_bh(sdp, rgd->rd_bh[buf]);
		ogfs_setbit(
			BH_DATA(rgd->rd_bh[buf]) + bits->bi_offset,
			bits->bi_len, block, new_state);
	}

	EXIT(GFN_BLKALLOC_INTERNAL);

	return bits->bi_start * OGFS_NBBY + block;
}

/**
 * blkfree_internal - Free a block
 * @sdp: the filesystem
 * @bstart: the start of a run of blocks to free
 * @blen: the length of the block run
 * @new_state: the new state of the block
 *
 */

static ogfs_rgrpd_t *
blkfree_internal(ogfs_sbd_t * sdp, uint64 bstart, uint32 blen,
		 unsigned char new_state)
{
	ogfs_rgrpd_t *rgd;
	ogfs_bitmap_t *bits = NULL;
	uint32 length, rgrp_block, buf_block;
	unsigned int buf;

	ENTER(GFN_BLKFREE_INTERNAL);

	rgd = ogfs_blk2rgrpd(sdp, bstart);
	OGFS_ASSERT(rgd, con_printf("block = %" FMT_64 "u\n", bstart);
	    );
	length = rgd->rd_ri.ri_length;
	rgrp_block = bstart - rgd->rd_ri.ri_data1;
	while (blen--) {
		for (buf = 0; buf < length; buf++) {
			bits = &rgd->rd_bits[buf];
			if (rgrp_block <
			    (bits->bi_start + bits->bi_len) * OGFS_NBBY)
				break;
		}
		OGFS_ASSERT(buf < length,);
		buf_block = rgrp_block - bits->bi_start * OGFS_NBBY;
		rgrp_block++;
		ogfs_trans_add_bh(sdp, rgd->rd_bh[buf]);
		ogfs_setbit(BH_DATA(rgd->rd_bh[buf]) + bits->bi_offset,
			   bits->bi_len, buf_block, new_state);
	}

	EXIT(GFN_BLKFREE_INTERNAL);

	return rgd;
}

/**
 * clump_alloc - Allocate a clump of metadata
 * @sdp: the filesystem
 * @rgd: the resource group descriptor
 *
 * If this routine's performance sucks, try allocating
 * more than block at a time.
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
clump_alloc(ogfs_rgrpd_t * rgd)
{
	ogfs_sbd_t *sdp = rgd->rd_sbd;
	ogfs_meta_header_t mh;
	struct buffer_head *bh[OGFS_META_CLUMP];
	uint32 block, goal = 0;
	int x, y;
	int error;

	ENTER(GFN_CLUMP_ALLOC);

	error = 0;
	if (test_and_clear_bit(RDF_SYNC_B4_ALLOC, &rgd->rd_flags))
		ogfs_log_flush(sdp, rgd->rd_gl);
	memset(&mh, 0, sizeof(ogfs_meta_header_t));
	mh.mh_magic = OGFS_MAGIC;
	mh.mh_type = OGFS_METATYPE_NONE;
	for (x = 0; x < OGFS_META_CLUMP; x++) {
		block = blkalloc_internal(
			rgd, goal, OGFS_BLKST_FREE, OGFS_BLKST_FREEMETA, TRUE);
		if (block == BLKALLOC_INTERNAL_NOENT) {
			if (x == 0) {
				error = -EIO;
				goto out;
			} else {
				/* use whatever we got */
				break;
			}
		}
		bh[x] = ogfs_dgetblk(
			sdp, rgd->rd_ri.ri_data1 + block, rgd->rd_gl);
		if (!bh[x]) {
			error = -EIO;
			goto out;
		}
		ogfs_prep_new_buffer(bh[x]);
		ogfs_meta_header_out(&mh, BH_DATA(bh[x]));
		/*  Zero out the generation number.  */
		((ogfs_meta_header_t *) BH_DATA(bh[x]))->mh_generation = 0;
		goal = block;
	}
	for (y = 0; y < x; y++)
		ogfs_dwrite(sdp, bh[y], DW_DIRTY | DW_START);
	for (y = 0; y < x; y++) {
		error = ogfs_dwrite(sdp, bh[y], DW_WAIT);
		if (error) {
			goto out;
		}
	}
	OGFS_ASSERT(rgd->rd_rg.rg_free >= x,);
	rgd->rd_rg.rg_free -= x;
	rgd->rd_rg.rg_freemeta += x;

out:
	while (x--)
		brelse(bh[x]);

	EXIT(GFN_CLUMP_ALLOC);
	return error;
}

/**
 * ogfs_blkalloc - Allocate a data block
 * @ip: the inode to allocate the data block for
 * @block: the block allocated
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_blkalloc(ogfs_inode_t * ip, uint64 * block)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	ogfs_ipreserv_t *res = &ip->i_res;
	ogfs_rgrpd_t *rgd = res->ir_rgd;
	uint32 goal;
	int same;

	ENTER(GFN_BLKALLOC);

	OGFS_ASSERT(rgd,);
	if (test_and_clear_bit(RDF_SYNC_B4_ALLOC, &rgd->rd_flags))
		ogfs_log_flush(sdp, rgd->rd_gl);
	/*  Any necessary quota checking goes here  */
	same = (rgd->rd_ri.ri_addr == ip->i_di.di_goal_rgrp);
	goal = (same) ? ip->i_di.di_goal_dblk : 0;
	*block = blkalloc_internal(
		rgd, goal, OGFS_BLKST_FREE, OGFS_BLKST_USED, TRUE);
	OGFS_ASSERT(*block != BLKALLOC_INTERNAL_NOENT,);
	if (!same) {
		ip->i_di.di_goal_rgrp = rgd->rd_ri.ri_addr;
		ip->i_di.di_goal_mblk = 0;
	}
	ip->i_di.di_goal_dblk = *block;
	*block += rgd->rd_ri.ri_data1;
	OGFS_ASSERT(rgd->rd_rg.rg_free,);
	rgd->rd_rg.rg_free--;
	ogfs_trans_add_bh(sdp, rgd->rd_bh[0]);
	ogfs_rgrp_out(&rgd->rd_rg, BH_DATA(rgd->rd_bh[0]));
	rgrp_lvb_own(rgd);
	res->ir_dataalloced++;

	EXIT(GFN_BLKALLOC);

	return 0;
}

/**
 * ogfs_metaalloc - Allocate a metadata block to a file
 * @ip:  the file
 * @block: the block allocated
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_metaalloc(ogfs_inode_t * ip, uint64 * block)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	ogfs_ipreserv_t *res = &ip->i_res;
	ogfs_rgrpd_t *rgd = res->ir_rgd;
	uint32 goal;
	int error;
	int same;

	ENTER(GFN_METAALLOC);

	OGFS_ASSERT(rgd,);
	/*  Any necessary quota checking goes here  */
	if (!rgd->rd_rg.rg_freemeta) {
		error = clump_alloc(rgd);
		if (error) {
			EXIT(GFN_METAALLOC);
			return error;
		}
		res->ir_dataalloced += OGFS_META_CLUMP;
	}
	same = (rgd->rd_ri.ri_addr == ip->i_di.di_goal_rgrp);
	goal = (same) ? ip->i_di.di_goal_mblk : 0;
	*block = blkalloc_internal(
		rgd, goal, OGFS_BLKST_FREEMETA, OGFS_BLKST_USEDMETA, TRUE);
	OGFS_ASSERT(*block != BLKALLOC_INTERNAL_NOENT,);
	if (!same) {
		ip->i_di.di_goal_rgrp = rgd->rd_ri.ri_addr;
		ip->i_di.di_goal_dblk = 0;
	}
	ip->i_di.di_goal_mblk = *block;
	*block += rgd->rd_ri.ri_data1;
	OGFS_ASSERT(rgd->rd_rg.rg_freemeta,);
	rgd->rd_rg.rg_freemeta--;
	rgd->rd_rg.rg_usedmeta++;
	ogfs_trans_add_bh(sdp, rgd->rd_bh[0]);
	ogfs_rgrp_out(&rgd->rd_rg, BH_DATA(rgd->rd_bh[0]));
	rgrp_lvb_own(rgd);
	res->ir_metaalloced++;

	EXIT(GFN_METAALLOC);

	return 0;
}

/**
 * ogfs_dipick - Look for a free dinode
 * @sdp: the filesystem
 * @res: the reservation
 * @inum: the inode that the allocator will pick when you call ogfs_dialloc()
 *
 */

void
ogfs_dipick(ogfs_sbd_t * sdp, ogfs_ipreserv_t * res, ogfs_inum_t * inum)
{
	ogfs_rgrpd_t *rgd = res->ir_rgd;
	unsigned char from, to;

	ENTER(GFN_DIPICK);

	OGFS_ASSERT(rgd,);
	/*  If there are free dinodes use those first  */
	if (rgd->rd_rg.rg_freedi) {
		*inum = rgd->rd_rg.rg_freedi_list;
		OGFS_ASSERT(inum->no_formal_ino && inum->no_addr,);
		goto out;
	}
	/*  Else, pick a metadata block  */
	if (rgd->rd_rg.rg_freemeta) {
		from = OGFS_BLKST_FREEMETA;
		to = OGFS_BLKST_USEDMETA;
	} else {
		from = OGFS_BLKST_FREE;
		to = OGFS_BLKST_FREEMETA;
	}
	inum->no_formal_ino = 0;
	inum->no_addr = blkalloc_internal(rgd, 0, from, to, FALSE);
	OGFS_ASSERT(inum->no_addr != BLKALLOC_INTERNAL_NOENT,);
	inum->no_addr += rgd->rd_ri.ri_data1;

out:
	EXIT(GFN_DIPICK);
}

/**
 * dialloc_preexist - Allocate a preexisting unused dinode
 * @sdp: the filesystem
 * @res: the reservation
 * @expected_inum: the number of the dinode to be allocated
 * @expected_gl: the glock of the dinode to be allocated
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
dialloc_preexist(ogfs_sbd_t * sdp, ogfs_ipreserv_t * res,
		 ogfs_inum_t * expected_inum, ogfs_glock_t * expected_gl)
{
	ogfs_rgrpd_t *rgd = res->ir_rgd;
	ogfs_dinode_t *di;
	struct buffer_head *bh;
	ogfs_inum_t next;
	uint32 flags;

	ENTER(GFN_DIALLOC_PREEXIST);

	OGFS_ASSERT(OGFS_INUM_EQUAL(
			    &rgd->rd_rg.rg_freedi_list, expected_inum),);
	bh = ogfs_dread(sdp, expected_inum->no_addr, expected_gl,
		       DR_START | DR_WAIT);
	if (!bh) {
		EXIT(GFN_DIALLOC_PREEXIST);
		return -EIO;
	}
	di = (ogfs_dinode_t *) BH_DATA(bh);
	flags = di->di_flags;
	flags = ogfs32_to_cpu(flags);
	OGFS_ASSERT(flags & OGFS_DIF_UNUSED,);
	ogfs_inum_in(&next, (char *) &di->di_next_unused);
	brelse(bh);
	rgd->rd_rg.rg_freedi--;
	rgd->rd_rg.rg_useddi++;
	rgd->rd_rg.rg_freedi_list = next;
	ogfs_trans_add_bh(sdp, rgd->rd_bh[0]);
	ogfs_rgrp_out(&rgd->rd_rg, BH_DATA(rgd->rd_bh[0]));
	rgrp_lvb_own(rgd);
	res->ir_dialloced++;

	EXIT(GFN_DIALLOC_PREEXIST);

	return 0;
}

/**
 * ogfs_dialloc - Allocate a dinode
 * @sdp: the filesystem
 * @res: the reservation
 * @expected_inum: The expected inode number
 * @expected_gl: the glock of the dinode to be allocated
 *
 * Assumes expected_gl is held exclusively by the calling process.
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_dialloc(
	ogfs_sbd_t * sdp, ogfs_ipreserv_t * res, ogfs_inum_t * expected_inum,
	ogfs_glock_t * expected_gl)
{
	ogfs_rgrpd_t *rgd = res->ir_rgd;
	uint64 block;
	uint32 goal;
	int error;

	ENTER(GFN_DIALLOC);

	OGFS_ASSERT(rgd,);
	/*  Any necessary quota checking goes here  */
	if (rgd->rd_rg.rg_freedi) {
		error = dialloc_preexist(sdp, res, expected_inum, expected_gl);
		EXIT(GFN_DIALLOC);
		return error;
	}
	if (!rgd->rd_rg.rg_freemeta) {
		error = clump_alloc(rgd);
		if (error) {
			EXIT(GFN_DIALLOC);
			return error;
		}
		res->ir_dataalloced += OGFS_META_CLUMP;
	}
	goal = expected_inum->no_addr - rgd->rd_ri.ri_data1;
	block = blkalloc_internal(
		rgd, goal, OGFS_BLKST_FREEMETA, OGFS_BLKST_USEDMETA, TRUE);
	block += rgd->rd_ri.ri_data1;
	OGFS_ASSERT(block == expected_inum->no_addr,);
	OGFS_ASSERT(rgd->rd_rg.rg_freemeta,);
	rgd->rd_rg.rg_freemeta--;
	rgd->rd_rg.rg_useddi++;
	ogfs_trans_add_bh(sdp, rgd->rd_bh[0]);
	ogfs_rgrp_out(&rgd->rd_rg, BH_DATA(rgd->rd_bh[0]));
	rgrp_lvb_own(rgd);
	res->ir_dialloced++;
	res->ir_metaalloced++;

	EXIT(GFN_DIALLOC);

	return 0;
}

/**
 * ogfs_blkfree - free a piece of data
 * @ip: the inode these blocks are being free from
 * @bstart: the start of a run of blocks to free
 * @blen: the length of the block run
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_blkfree(ogfs_inode_t * ip, uint64 bstart, uint32 blen)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	ogfs_rgrpd_t *rgd;
	int error;

	ENTER(GFN_BLKFREE);

	rgd = blkfree_internal(sdp, bstart, blen, OGFS_BLKST_FREE);
	rgd->rd_rg.rg_free += blen;
	set_bit(RDF_SYNC_B4_ALLOC, &rgd->rd_flags);
	ogfs_trans_add_bh(sdp, rgd->rd_bh[0]);
	ogfs_rgrp_out(&rgd->rd_rg, BH_DATA(rgd->rd_bh[0]));
	rgrp_lvb_own(rgd);
	error = ogfs_add_gl_dependency(sdp, ip->i_gl, rgd->rd_gl);
	ogfs_rgrpd_put(rgd);

	EXIT(GFN_BLKFREE);

	return error;
}

/**
 * ogfs_metafree - free a piece of metadata
 * @ip: the inode these blocks are being free from
 * @bstart: the start of a run of blocks to free
 * @blen: the length of the block run
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_metafree(ogfs_inode_t * ip, uint64 bstart, uint32 blen)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	ogfs_rgrpd_t *rgd;
	int error;

	ENTER(GFN_METAFREE);

	rgd = blkfree_internal(sdp, bstart, blen, OGFS_BLKST_FREEMETA);

	OGFS_ASSERT(rgd->rd_rg.rg_usedmeta >= blen,);
	rgd->rd_rg.rg_usedmeta -= blen;
	rgd->rd_rg.rg_freemeta += blen;
	ogfs_trans_add_bh(sdp, rgd->rd_bh[0]);
	ogfs_rgrp_out(&rgd->rd_rg, BH_DATA(rgd->rd_bh[0]));
	rgrp_lvb_own(rgd);
	error = ogfs_add_gl_dependency(sdp, ip->i_gl, rgd->rd_gl);
	ogfs_rgrpd_put(rgd);

	EXIT(GFN_METAFREE);

	return error;
}

/**
 * ogfs_difree - free a piece of metadata
 * @ip: the inode representing the dinode to free
 * @rgd: the resource group that contains the dinode
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_difree(ogfs_inode_t * ip, ogfs_rgrpd_t * rgd)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	struct buffer_head *dibh;
	ogfs_dinode_t *di;
	struct timespec ctime;
	uint32 flags;
	int error;

	ENTER(GFN_DIFREE);

	dibh = ogfs_dread(sdp, ip->i_num.no_addr, ip->i_gl, DR_START | DR_WAIT);
	if (!dibh) {
		error = -EIO;
		goto out;
	}
	di = (ogfs_dinode_t *) BH_DATA(dibh);
	ogfs_trans_add_bh(sdp, dibh);
	ctime = current_kernel_time();
	di->di_ctime = cpu_to_ogfs64((int64)ctime.tv_sec);
	flags = ogfs32_to_cpu(di->di_flags);
	flags |= OGFS_DIF_UNUSED;
	di->di_flags = cpu_to_ogfs32(flags);
	ogfs_inum_out(&rgd->rd_rg.rg_freedi_list, (char *) &di->di_next_unused);
	brelse(dibh);
	OGFS_ASSERT(rgd->rd_rg.rg_useddi,);
	rgd->rd_rg.rg_useddi--;
	rgd->rd_rg.rg_freedi++;
	rgd->rd_rg.rg_freedi_list = ip->i_num;
	ogfs_trans_add_bh(sdp, rgd->rd_bh[0]);
	ogfs_rgrp_out(&rgd->rd_rg, BH_DATA(rgd->rd_bh[0]));
	rgrp_lvb_own(rgd);
	error = ogfs_add_gl_dependency(sdp, ip->i_gl, rgd->rd_gl);

out:
	EXIT(GFN_DIFREE);

	return error;
}

/**
 * ogfs_rlist_add - do stuff
 * @sdp: the filesystem
 * @rlist: the list of resource groups
 * @rgrp: the resource group to add
 *
 * Figure out what RG a block belongs to and add that RG to the list
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_rlist_add(ogfs_sbd_t * sdp, ogfs_rgrp_list_t * rlist, uint64 block)
{
	ogfs_rgrpd_t *rgd, **tmp;
	unsigned int new_space;
	unsigned int x;
	int error = 0;
	int do_put = TRUE;

	ENTER(GFN_RLIST_ADD);

	OGFS_ASSERT(rlist->rl_rgrps <= rlist->rl_space,);
	rgd = ogfs_blk2rgrpd(sdp, block);
	OGFS_ASSERT(rgd, con_printf("block = %" FMT_64 "u\n", block);
	    );
	for (x = 0; x < rlist->rl_rgrps; x++)
		if (rlist->rl_list[x] == rgd)
			goto out;
	if (rlist->rl_rgrps == rlist->rl_space) {
		new_space = rlist->rl_space + 10;
		tmp = (ogfs_rgrpd_t **) gmalloc(
			new_space * sizeof(ogfs_rgrpd_t *));
		if (!tmp) {
			error = -EIO;
			goto out;
		}
		if (rlist->rl_list) {
			memcpy(tmp, rlist->rl_list,
				   rlist->rl_space * sizeof(ogfs_rgrpd_t *));
			gfree(rlist->rl_list,
			      rlist->rl_space * sizeof(ogfs_rgrpd_t *));
		}
		rlist->rl_list = tmp;
		rlist->rl_space = new_space;
	}
	do_put = FALSE;
	rlist->rl_list[rlist->rl_rgrps++] = rgd;

out:
	if (do_put)
		ogfs_rgrpd_put(rgd);

	EXIT(GFN_RLIST_ADD);

	return error;
}

/**
 * ogfs_rlist_compar - compare address. Used by ogfs_rlist_sort.
 * @a: resource group a
 * @b: resource group b
 * Return : 1 a > b  -1 a < b 
 */
int ogfs_rlist_compar(const void *a, const void *b)
{
	return ((*(ogfs_rgrpd_t **)a)->rd_ri.ri_addr > (*(ogfs_rgrpd_t **)b)->rd_ri.ri_addr) ? 1:-1;
}


/**
 * ogfs_rlist_sort - sort a list of resource groups
 * @list: the list of resource groups
 *
 */
void
ogfs_rlist_sort(ogfs_rgrp_list_t * rlist)
{
	ogfs_rgrpd_t **list = rlist->rl_list;
	uint32 rgrps = rlist->rl_rgrps;

	ENTER(GFN_RLIST_SORT);

	if (!rgrps) {
		EXIT(GFN_RLIST_SORT);
		return;
	}
	ogfs_sort((void *)list, (rgrps - 1), sizeof(*list), ogfs_rlist_compar);
	EXIT(GFN_RLIST_SORT);
}

/**
 * ogfs_rlist_free - free a resource group list
 * @list: the list of resource groups
 *
 */

void
ogfs_rlist_free(ogfs_rgrp_list_t * rlist)
{
	unsigned int x;

	ENTER(GFN_RLIST_FREE);

	if (rlist->rl_list) {
		for (x = 0; x < rlist->rl_rgrps; x++)
			ogfs_rgrpd_put(rlist->rl_list[x]);
		gfree(rlist->rl_list, rlist->rl_space * sizeof(ogfs_rgrpd_t *));
	}

	EXIT(GFN_RLIST_FREE);
}


/**
 * ogfs_reclaim_one - reclaim all free metadata block from one rgrp
 * @sdp: the filesystem
 * @rnum: rgrp number
 * */

int ogfs_reclaim_one(ogfs_sbd_t *sdp, unsigned int rnum)
{
	uint32 freemeta;
	unsigned int buf;
	uint32 length;
        ogfs_bitmap_t *bits = NULL;
	ogfs_rgrpd_t *rgd;
	int error = 0;
		
	rgd = ogfs_rgrpd_get_by_num(sdp, rnum);
	if (rgd == NULL)
		return error;

	error = ogfs_glock_rg(rgd, 0);
	if (error) {
		ogfs_rgrpd_put(rgd);
		return error;
	}
	
	length = rgd->rd_ri.ri_length;
	freemeta = rgd->rd_rg.rg_freemeta;
	if (freemeta > 0) {
		error = ogfs_trans_begin(sdp, TRANS_BMAP, length + 1, 0);
		if (error) {
			ogfs_gunlock_rg(rgd, 0);
			ogfs_rgrpd_put(rgd);
			return error;
		}
	} else {
		ogfs_gunlock_rg(rgd, 0);
		ogfs_rgrpd_put(rgd);
		return error;
	}
	
	for (buf = 0; buf < length; buf++) {
		bits = &rgd->rd_bits[buf];
		freemeta -= ogfs_travel_bitmap(BH_DATA(rgd->rd_bh[buf]) + bits->bi_offset,
						bits->bi_len, OGFS_BLKST_FREEMETA,
						OGFS_BLKST_FREE);
		ogfs_trans_add_bh(sdp, rgd->rd_bh[buf]);
	}
	rgd->rd_rg.rg_free += (rgd->rd_rg.rg_freemeta - freemeta);
	rgd->rd_rg.rg_freemeta = freemeta;
	ogfs_trans_add_bh(sdp, rgd->rd_bh[0]);
	ogfs_rgrp_out(&rgd->rd_rg, BH_DATA(rgd->rd_bh[0]));
	rgrp_lvb_own(rgd);
	set_bit(RDF_SYNC_B4_ALLOC, &rgd->rd_flags);
	ogfs_trans_end(sdp);
	ogfs_gunlock_rg(rgd, 0);
	ogfs_rgrpd_put(rgd);
	return error;
}

/*
 * These routines are used by the block-list routines (blklist.c)
 * to keep track of block allocation.  Each block is represented by two
 * bits.  One bit indicates whether or not the block is used.  (1=used,
 * 0=free)  The other bit indicates whether or not the block contains a
 * dinode or not.  (1=dinode, 0=data block)
 * So, each byte represents OGFS_NBBY (i.e. 4) blocks.
 */

static const char valid_change[16] = {
	/* current */

	/* n */ 0, 1, 1, 0,
	/* e */ 1, 0, 0, 0,
	/* w */ 1, 0, 0, 1,
	0, 0, 1, 0
};

/**
 * ogfs_setbit - Set a bit in the bitmaps
 * @buffer: the buffer that holds the bitmaps
 * @buflen: the length (in bytes) of the buffer
 * @block: the block to set
 * @new_state: the new state of the block
 *
 */

void
ogfs_setbit(unsigned char *buffer, unsigned int buflen,
	   uint32 block, unsigned char new_state)
{
	unsigned char *byte, *end, cur_state;
	unsigned int bit;

	ENTER(GFN_SET_BITS);

	byte = buffer + (block / OGFS_NBBY);
	bit = (block % OGFS_NBBY) * OGFS_BIT_SIZE;
	end = buffer + buflen;

	OGFS_ASSERT(byte < end,);

	cur_state = (*byte >> bit) & OGFS_BIT_MASK;
	OGFS_ASSERT(valid_change[new_state * 4 + cur_state],
		   con_printf("cur_state = %u, new_state = %u\n", cur_state,
			      new_state);
	    );

	*byte ^= cur_state << bit;
	*byte |= new_state << bit;

	EXIT(GFN_SET_BITS);
}

/**
 * ogfs_bitfit - Find a free block in the bitmaps
 * @buffer: the buffer that holds the bitmaps
 * @buflen: the length (in bytes) of the buffer
 * @goal: the block to try to allocate
 * @old_state: the state of the block we're looking for
 *
 * Return: the block number that was allocated
 */

uint32
ogfs_bitfit(unsigned char *buffer, unsigned int buflen,
	   uint32 goal, unsigned char old_state)
{
	unsigned char *byte, *end, alloc;
	uint32 blk = goal;
	unsigned int bit;

	ENTER(GFN_BITFIT);

	byte = buffer + (goal / OGFS_NBBY);
	bit = (goal % OGFS_NBBY) * OGFS_BIT_SIZE;
	end = buffer + buflen;
	alloc = (old_state & 1) ? 0 : 0x55;

	while (byte < end) {
		if ((*byte & 0x55) == alloc) {
			blk += (8 - bit) >> 1;

			bit = 0;
			byte++;

			continue;
		}

		if (((*byte >> bit) & OGFS_BIT_MASK) == old_state) {
			EXIT(GFN_BITFIT);
			return blk;
		}

		bit += OGFS_BIT_SIZE;
		if (bit >= 8) {
			bit = 0;
			byte++;
		}

		blk++;
	}

	EXIT(GFN_BITFIT);

	return BFITNOENT;
}

/**
 * ogfs_travel_bitmap - travel the bitmap and change blocks state
 * @buffer: the buffer that holds the bitmaps
 * @buflen: the length (in bytes) of the buffer
 * @old_state: the old_state of the block we're looking for
 * @new_state: the new_state of the block we're looking for
 *
 * Returns: The number of blocks we changed
 *
 * This function would travel the bitmap and change all "old_state" blocks to
 * "new_state" blocks when "new_state" is not equal to OGFS_BLKST_INVALID. Or,
 * the number of "old_state" blocks is return and nothing is changed.
 */

uint32
ogfs_travel_bitmap(unsigned char *buffer, unsigned int buflen, unsigned char old_state,
		unsigned char new_state)
{
	unsigned char *byte, *end, cur_state;
	unsigned int bit;
	uint32 count = 0;

	ENTER(GFN_BITCOUNT);

	if (new_state & ~OGFS_BIT_MASK) /*INVALID new_state*/
		new_state = OGFS_BLKST_INVALID;
	
	byte = buffer;
	bit = 0;
	end = buffer + buflen;

	while (byte < end) {
		cur_state = (*byte >> bit) & OGFS_BIT_MASK;
		if (cur_state == old_state) {
		 	if ( new_state != OGFS_BLKST_INVALID ) {
				OGFS_ASSERT(valid_change[new_state * 4 + cur_state],
			   		con_printf("cur_state = %u, new_state = %u\n", 
						cur_state, new_state);
				);
				*byte ^= cur_state << bit;
				*byte |= new_state << bit;
			}
			count++;
		}
		bit += OGFS_BIT_SIZE;
		if (bit >= 8) {
			bit = 0;
			byte++;
		}
	}

	EXIT(GFN_BITCOUNT);

	return count;
}
