/*
 *
 *    Copyright 1995-1999 Regents of the University of Minnesota
 *    Portions Copyright 1999-2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 */

/*
 * These routines translate from extents in a file to block numbers.
 */

#include "ogfs_debug.h"
#include "ogfs.h"

#include "blklist.h"
#include "bmap.h"
#include "dio.h"
#include "file.h"
#include "glock.h"
#include "inode.h"
#include "log.h"
#include "trans.h"

struct metapath {
	uint64 mp_list[OGFS_MAX_META_HEIGHT];
};
typedef struct metapath metapath_t;

typedef int (*block_call_t) (ogfs_inode_t * ip, struct buffer_head * bh, uint64 * top,
			     uint64 * bottom, unsigned int height, void *data);

struct strip_mine {
	int sm_first;
	unsigned int sm_height;
};
typedef struct strip_mine strip_mine_t;

/**
 * ogfs_unstuff_dinode - Unstuff a dinode when the data has grown too big
 * @ip: The OGFS inode to unstuff
 *
 * This routine unstuffs a dinode and returns it to a "normal" state such
 * that the height can be grown in the traditional way.
 *
 * Returns: 0 on success, -EXXXX on failure
 */

int
ogfs_unstuff_dinode(ogfs_inode_t * ip)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	struct buffer_head *bh, *dibh;
	uint64 block = 0;
	int error;
	int journaled = ogfs_is_jdata(ip);

	ENTER(GFN_UNSTUFF_DINODE);

	OGFS_ASSERT(ogfs_is_stuffed(ip),);

	dibh = ogfs_get_inode_buffer(ip);
	if (!dibh) {
		error = -EIO;
		goto fail;
	}

	if (ip->i_di.di_size) {
		/*
		 * Get a free block, fill it with the stuffed data,
		 * and write it out to disk
		 */

		if (journaled)
			error = ogfs_metaalloc(ip, &block);
		else
			error = ogfs_blkalloc(ip, &block);

		if (error)
			goto fail_di_relse;

		bh = ogfs_get_data_buffer(ip, block, TRUE);
		if (!bh) {
			error = -EIO;
			goto fail_di_relse;
		}

		ogfs_buffer_copy_tail(bh, (journaled) ? sizeof(ogfs_jdata_t) : 0,
				     dibh, sizeof(ogfs_dinode_t));

		if (journaled)
			brelse(bh);
		else {
			error =
			    ogfs_dwrite(sdp, bh, DW_DIRTY | DW_START | DW_WAIT);
			if (error) {
				brelse(bh);
				goto fail_di_relse;
			}

			ogfs_dforget(bh);
		}
	}

	/*  Set up the pointer to the new block  */

	ogfs_trans_add_bh(sdp, dibh);

	ogfs_buffer_clear_tail(dibh, sizeof(ogfs_dinode_t));

	if (ip->i_di.di_size) {
		((uint64 *) (BH_DATA(dibh) + sizeof(ogfs_dinode_t)))[0] =
		    cpu_to_ogfs64(block);
		ip->i_di.di_blocks++;
	}

	ip->i_di.di_height = 1;

	ogfs_dinode_out(&ip->i_di, BH_DATA(dibh));
	brelse(dibh);

	EXIT(GFN_UNSTUFF_DINODE);

	return 0;

fail_di_relse:
	brelse(dibh);

fail:
	EXIT(GFN_UNSTUFF_DINODE);
	return error;
}

/**
 * ogfs_calc_tree_height - Calculate the height of a metadata tree
 * @ip: The OGFS inode
 * @size: The proposed size of the file
 *
 * Work out how tall a metadata tree needs to be in order to accommodate a
 * file of a particular size. If size is less than the current size of
 * the inode, then the current size of the inode is used instead of the
 * supplied one.
 *
 * Returns: the height the tree should be
 */

unsigned int
ogfs_calc_tree_height(ogfs_inode_t * ip, uint64 size)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	uint64 sz;
	unsigned int height;
	uint64 space, old_space;
	unsigned int bsize;

	ENTER(GFN_CALC_TREE_HEIGHT);

	sz = (size > ip->i_di.di_size) ? size : ip->i_di.di_size;

	if (sz <= (sdp->sd_sb.sb_bsize - sizeof(ogfs_dinode_t))) {
		EXIT(GFN_CALC_TREE_HEIGHT);
		return 0;
	}

	bsize = (ogfs_is_jdata(ip)) ? sdp->sd_jbsize : sdp->sd_sb.sb_bsize;

	height = 1;
	space = sdp->sd_diptrs * bsize;

	while (sz > space) {
		old_space = space;

		height++;
		space *= sdp->sd_inptrs;

		if (space / sdp->sd_inptrs != old_space ||
		    space % sdp->sd_inptrs != 0)
			break;
	}

	EXIT(GFN_CALC_TREE_HEIGHT);

	return height;
}

/**
 * ogfs_build_height - Build a metadata tree of the requested height
 * @ip: The OGFS inode
 * @height: The height to build to
 *
 * This routine makes sure that the metadata tree is tall enough to hold
 * "size" bytes of data.
 *
 * Returns: 0 on success, -EXXXX on failure
 */

int
build_height(ogfs_inode_t * ip, int height)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	struct buffer_head *bh, *dibh;
	uint64 block;
	int error;

	ENTER(GFN_BUILD_HEIGHT);

	while (ip->i_di.di_height < height) {
		dibh = ogfs_get_inode_buffer(ip);
		if (!dibh) {
			error = -EIO;
			goto fail;
		}

		/*
		 * Get a new block, fill it with the old direct pointers
		 * and write it out.
		 */
		error = ogfs_metaalloc(ip, &block);
		if (error)
			goto fail_drelse;

		bh = ogfs_dread(sdp, block, ip->i_gl,
			       DR_FORCE | DR_START | DR_WAIT);
		if (!bh) {
			error = -EIO;
			goto fail_drelse;
		}

		ogfs_trans_add_bh(sdp, bh);
		ogfs_metatype_set(bh, OGFS_METATYPE_IN, OGFS_FORMAT_IN);
		ogfs_buffer_copy_tail(bh, sizeof(ogfs_indirect_t), dibh,
				     sizeof(ogfs_dinode_t));

		brelse(bh);

		/*
		 * Set up the new direct pointer and write it out to disk.
		 */
		ogfs_trans_add_bh(sdp, dibh);

		ogfs_buffer_clear_tail(dibh, sizeof(ogfs_dinode_t));
		((uint64 *) (BH_DATA(dibh) + sizeof(ogfs_dinode_t)))[0] =
		    cpu_to_ogfs64(block);

		ip->i_di.di_height++;
		ip->i_di.di_blocks++;

		ogfs_dinode_out(&ip->i_di, BH_DATA(dibh));
		brelse(dibh);
	}

	EXIT(GFN_BUILD_HEIGHT);

	return 0;

fail_drelse:
	brelse(dibh);

fail:
	EXIT(GFN_BUILD_HEIGHT);
	return error;
}

/**
 * find_metapath - Find path through the metadata tree
 * @ip: The inode pointer
 * @mp: The metapath to return the result in
 * @block: The disk block to look up
 *
 *   This routine returns a metapath_t structure that defines a path through
 *   the metadata of inode "ip" to get to block "block".
 *
 *   Example:
 *   Given:  "ip" is a height 3 file, "offset" is 101342453, and this is a
 *   filesystem with a blocksize of 4096.
 *
 *   find_metapath() would return a metapath_t structure set to:
 *   mp_offset = 101342453, mp_height = 3, mp_list[0] = 0, mp_list[1] = 48,
 *   and mp_list[2] = 165.
 *
 *   That means that in order to get to the block containing the byte at
 *   offset 101342453, we would load the indirect block pointed to by pointer
 *   0 in the dinode.  We would then load the indirect block pointed to by
 *   pointer 48 in that indirect block.  We would then load the data block
 *   pointed to by pointer 165 in that indirect block.
 *
 *             ----------------------------------------
 *             | Dinode |                             |
 *             |        |                            4|
 *             |        |0 1 2 3 4 5                 9|
 *             |        |                            6|
 *             ----------------------------------------
 *                       |
 *                       |
 *                       V
 *             ----------------------------------------
 *             | Indirect Block                       |
 *             |                                     5|
 *             |            4 4 4 4 4 5 5            1|
 *             |0           5 6 7 8 9 0 1            2|
 *             ----------------------------------------
 *                                |
 *                                |
 *                                V
 *             ----------------------------------------
 *             | Indirect Block                       |
 *             |                         1 1 1 1 1   5|
 *             |                         6 6 6 6 6   1|
 *             |0                        3 4 5 6 7   2|
 *             ----------------------------------------
 *                                           |
 *                                           |
 *                                           V
 *             ----------------------------------------
 *             | Data block containing offset         |
 *             |            101342453                 |
 *             |                                      |
 *             |                                      |
 *             ----------------------------------------
 *
 */

static void
find_metapath(ogfs_inode_t * ip, metapath_t * mp, uint64 block)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	unsigned int i;

	ENTER(GFN_FIND_METAPATH);

	for (i = ip->i_di.di_height; i--;) {
		mp->mp_list[i] = block % sdp->sd_inptrs;
		block /= sdp->sd_inptrs;
	}

	EXIT(GFN_FIND_METAPATH);
}

/**
 * metapointer - Return pointer to start of metadata in a buffer
 * @bh: The buffer
 * @level: The metadata level (0 = dinode)
 * @mp: The metapath
 *
 * Return a pointer to the block number of the next level of the metadata
 * tree given a buffer containing the pointer to the current level of the
 * metadata tree.
 */

static __inline__ uint64 *
metapointer(struct buffer_head * bh, int level, metapath_t * mp)
{
	int head_size =
	    (level > 0) ? sizeof(ogfs_indirect_t) : sizeof(ogfs_dinode_t);
	return ((uint64 *) (BH_DATA(bh) + head_size)) + mp->mp_list[level];
}

/**
 * get_metablock - Get the next metadata block in metadata tree
 * @ip: The OGFS inode
 * @bh: Buffer containing the pointers to metadata blocks
 * @level: The level of the tree (0 = dinode)
 * @mp: The metapath
 * @create: Non-zero if we may create a new meatdata block
 * @new: Used to indicate if we did create a new metadata block
 * @block: the returned disk block number
 *
 * Given a metatree, complete to a particular level, checks to see if the next
 * level of the tree exists. If not the next level of the tree is created.
 * The block number of the next level of the metadata tree is returned.
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
get_metablock(ogfs_inode_t * ip,
	      struct buffer_head * bh, int level, metapath_t * mp,
	      int create, int *new, uint64 * block)
{
	uint64 *ptr = metapointer(bh, level, mp);
	int error = 0;

	ENTER(GFN_GET_METABLOCK);

	*new = 0;
	*block = 0;

	if (*ptr) {
		*block = ogfs64_to_cpu(*ptr);
		goto out;
	}

	if (!create)
		goto out;

	error = ogfs_metaalloc(ip, block);
	if (error)
		goto out;

	ogfs_trans_add_bh(ip->i_sbd, bh);

	*ptr = cpu_to_ogfs64(*block);
	ip->i_di.di_blocks++;

	*new = 1;

out:
	EXIT(GFN_GET_METABLOCK);
	return error;
}

/**
 * get_datablock - Get datablock number from metadata block
 * @ip: The OGFS inode
 * @bh: The buffer containing pointers to datablocks
 * @mp: The metapath
 * @create: Non-zero if we may create a new data block
 * @new: Used to indicate if we created a new data block
 * @block: the returned disk block number
 *
 * Given a fully built metadata tree, checks to see if a particular data
 * block exists. It is created if it does not exist and the block number
 * on disk is returned.
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
get_datablock(ogfs_inode_t * ip,
	      struct buffer_head * bh, metapath_t * mp,
	      int create, int *new, uint64 * block)
{
	int level = ip->i_di.di_height - 1;
	uint64 *ptr = metapointer(bh, level, mp);
	int error = 0;

	ENTER(GFN_GET_DATABLOCK);

	*new = 0;
	*block = 0;

	if (*ptr) {
		*block = ogfs64_to_cpu(*ptr);
		goto out;
	}

	if (!create)
		goto out;

	if (ogfs_is_jdata(ip))
		error = ogfs_metaalloc(ip, block);
	else
		error = ogfs_blkalloc(ip, block);

	if (error)
		goto out;

	ogfs_trans_add_bh(ip->i_sbd, bh);

	*ptr = cpu_to_ogfs64(*block);
	ip->i_di.di_blocks++;

	*new = 1;

out:
	EXIT(GFN_GET_DATABLOCK);
	return error;
}

/**
 * ogfs_inode_get_block - Map a block from an inode to a disk block
 * @ip: The OGFS inode
 * @blknum: The block number (inode relative)
 * @new: Value/Result argument (1 = may create/did create new blocks)
 *
 * Find the block number on the current device which corresponds to an
 * inode's block. If the block had to be created, "new" will be set.
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_inode_get_block(ogfs_inode_t * ip, uint64 blknum, int *new, uint64 * block)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	struct buffer_head *bh;
	metapath_t mp;
	int create = *new;
	unsigned int bsize;
	unsigned int height;
	unsigned int i, end_of_metadata;
	int error = 0;

	ENTER(GFN_INODE_GET_BLOCK);

	*new = 0;
	*block = 0;

	if (ogfs_is_stuffed(ip)) {
		*block = ip->i_num.no_addr;
		goto out;
	}

	bh = ogfs_get_inode_buffer(ip);
	if (!bh) {
		error = -EIO;
		goto fail;
	}

	if (create)
		ogfs_trans_add_bh(sdp, bh);

	bsize = (ogfs_is_jdata(ip)) ? sdp->sd_jbsize : sdp->sd_sb.sb_bsize;

	height = ogfs_calc_tree_height(ip, (blknum + 1) * bsize);
	if (ip->i_di.di_height < height) {
		if (!create) {
			brelse(bh);
			goto out;
		}

		error = build_height(ip, height);
		if (error) {
			brelse(bh);
			goto fail;
		}
	}

	find_metapath(ip, &mp, blknum);
	end_of_metadata = ip->i_di.di_height - 1;

	for (i = 0; i < end_of_metadata; i++) {
		error = get_metablock(ip, bh, i, &mp, create, new, block);
		brelse(bh);
		if (error)
			goto fail;
		if (!*block)
			goto out;

		bh = ogfs_get_meta_buffer(ip, i + 1, *block, *new);
		if (!bh) {
			error = -EIO;
			goto fail;
		}
	}

	error = get_datablock(ip, bh, &mp, create, new, block);
	brelse(bh);
	if (error)
		goto fail;

out:
	if (*new) {
		bh = ogfs_get_inode_buffer(ip);
		OGFS_ASSERT(bh,); /* This buffer should be already pinned */
		ogfs_dinode_out(&ip->i_di, BH_DATA(bh));
		brelse(bh);
	}

	EXIT(GFN_INODE_GET_BLOCK);

	return 0;

fail:
	EXIT(GFN_INODE_GET_BLOCK);
	return error;
}

/**
 * do_grow - Make a file look bigger than it is
 * @ip: the inode
 * @size: the size to set the file to
 *
 * Called with an exclusive lock on @ip.
 *
 * Returns: 0 on succes, -EXXX on failure
 */

static int
do_grow(ogfs_inode_t * ip, uint64 size)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	ogfs_ipreserv_t *res = &ip->i_res;
	struct buffer_head *dibh;
	struct timespec ctime;
	int h;
	int error;

	ENTER(GFN_DO_GROW);

	res->ir_diblks = 0;
	res->ir_metablks = sdp->sd_max_height;
	res->ir_datablks = 1;
	res->ir_target = ip->i_di.di_goal_rgrp;
	res->ir_trans = TRANS_TRUNC_GROW;

	error = ogfs_inplace_reserv(ip, 0);
	if (error)
		goto fail;

	/*  Trans may require:
	   Full extention of the metadata tree, block allocation,
	   and a dinode modification  */

	error =
	    ogfs_trans_begin(sdp, TRANS_TRUNC_GROW,
			    sdp->sd_max_height + res->ir_rgd->rd_ri.ri_length +
			    1, 0);
	if (error)
		goto fail_ipres;

	if (size > sdp->sd_sb.sb_bsize - sizeof(ogfs_dinode_t)) {
		if (ogfs_is_stuffed(ip)) {
			error = ogfs_unstuff_dinode(ip);
			if (error)
				goto fail_end_trans;
		}

		h = ogfs_calc_tree_height(ip, size);
		if (ip->i_di.di_height < h) {
			error = build_height(ip, h);
			if (error)
				goto fail_end_trans;
		}
	}

	ip->i_di.di_size = size;
	ctime = current_kernel_time();
	ip->i_di.di_mtime = ip->i_di.di_ctime = (int64)ctime.tv_sec;

	dibh = ogfs_get_inode_buffer(ip);
	if (!dibh) {
		error = -EIO;
		goto fail_end_trans;
	}

	ogfs_trans_add_bh(sdp, dibh);
	ogfs_dinode_out(&ip->i_di, BH_DATA(dibh));
	brelse(dibh);

	ogfs_trans_end(sdp);

	ogfs_inplace_relse(ip);

	EXIT(GFN_DO_GROW);

	return 0;

fail_end_trans:
	ogfs_trans_mark(sdp, TRF_POISONED);
	ogfs_trans_end(sdp);

fail_ipres:
	ogfs_inplace_relse(ip);

fail:
	EXIT(GFN_DO_GROW);
	return error;
}

/**
 * recursive_scan - recursively scan through the end of a file
 * @ip: the inode
 * @mp: the path through the metadata to the point to start
 * @height: the height the recursion is at
 * @block: the indirect block to look at
 * @first: TRUE if this is the first block
 * @bc: the call to make for each piece of metadata
 * @data: data opaque to this function to pass to @bc
 *
 * When this is first called @height and @block should be zero and
 * @first should be TRUE.
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
recursive_scan(ogfs_inode_t * ip, metapath_t * mp,
	       unsigned int height, uint64 block, int first,
	       block_call_t bc, void *data)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	struct buffer_head *bh = NULL;
	uint64 *top, *bottom;
	uint64 bn;
	int error;

	ENTER(GFN_RECURSIVE_SCAN);

	IFDB(GFN_RECURSIVE_SCAN,
	     db_printf("  height = %d\n", height);
	     db_printf("  block = %" FMT_64 "u\n", block);
	    );

	if (!height) {
		bh = ogfs_get_inode_buffer(ip);
		if (!bh) {
			error = -EIO;
			goto fail;
		}

		top =
		    (uint64 *) (BH_DATA(bh) + sizeof(ogfs_dinode_t)) +
		    mp->mp_list[0];
		bottom =
		    (uint64 *) (BH_DATA(bh) + sizeof(ogfs_dinode_t)) +
		    sdp->sd_diptrs;
	} else {
		bh = ogfs_get_meta_buffer(ip, height, block, FALSE);
		if (!bh) {
			error = -EIO;
			goto fail;
		}

		top =
		    (uint64 *) (BH_DATA(bh) + sizeof(ogfs_indirect_t)) +
		    ((first) ? mp->mp_list[height] : 0);
		bottom =
		    (uint64 *) (BH_DATA(bh) + sizeof(ogfs_indirect_t)) +
		    sdp->sd_inptrs;
	}

	error = bc(ip, bh, top, bottom, height, data);
	if (error)
		goto fail;

	if (height+1 < ip->i_di.di_height )
		for (; top < bottom; top++, first = FALSE) {
			if (!*top)
				continue;

			bn = ogfs64_to_cpu(*top);

			IFDB(GFN_RECURSIVE_SCAN,
			     db_printf("  Recursing into indirect block %"
				       FMT_64 "u\n", bn);
			    );

			error = recursive_scan(ip, mp,
					       height + 1, bn, first, bc, data);
			if (error)
				goto fail;
		}

	brelse(bh);

	EXIT(GFN_RECURSIVE_SCAN);

	return 0;

fail:
	if (bh)
		brelse(bh);

	EXIT(GFN_RECURSIVE_SCAN);
	return error;
}

/**
 * do_strip - Look for a particular layer of the file and strip it off
 * @ip: the inode
 * @bh: A buffer of pointers
 * @top: The first pointer in the buffer
 * @bottom: One more than the last pointer
 * @height: the height this buffer is at
 * @data: a pointer to a strip_mine_t
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
do_strip(ogfs_inode_t * ip, struct buffer_head * bh, uint64 * top, uint64 * bottom,
	 unsigned int height, void *data)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	strip_mine_t *sm = (strip_mine_t *) data;
	ogfs_rgrp_list_t rlist;
	ogfs_rgrpd_t *rgd;
	struct buffer_head *dibh;
	struct timespec ctime;
	uint64 bn, bstart;
	uint32 blen;
	uint64 *p;
	unsigned int blocks = 0;
	unsigned int x;
	int error;

	ENTER(GFN_DO_STRIP);

	if (!*top)
		sm->sm_first = FALSE;

	if (height != sm->sm_height) {
		error = 0;
		goto fail;
	}

	memset(&rlist, 0, sizeof(ogfs_rgrp_list_t));

	if (sm->sm_first) {
		top++;
		sm->sm_first = FALSE;
	}

	error = ogfs_rindex_hold(sdp);
	if (error)
		goto fail;

	bstart = 0;
	blen = 0;

	for (p = top; p < bottom; p++) {
		if (!*p)
			continue;

		bn = ogfs64_to_cpu(*p);

		if (bstart + blen == bn)
			blen++;
		else {
			if (bstart) {
				error = ogfs_rlist_add(sdp, &rlist, bstart);
				if (error)
					goto fail_rlist_free;
			}

			bstart = bn;
			blen = 1;
		}
	}

	if (bstart) {
		error = ogfs_rlist_add(sdp, &rlist, bstart);
		if (error)
			goto fail_rlist_free;
	} else {
		/*  Nothing to do  */
		error = 0;
		goto fail_rlist_free;
	}

	ogfs_rlist_sort(&rlist);

	for (x = 0; x < rlist.rl_rgrps; x++) {
		rgd = rlist.rl_list[x];

		error = ogfs_glock_rg(rgd, 0);
		if (error)
			goto fail_rg_gunlock;

		blocks += rgd->rd_ri.ri_length;
	}

	/*  Trans may require:
	   All the bitmaps that were reserved.
	   One block for the dinode.
	   One block for the indirect block being cleared.
	 */

	error = ogfs_trans_begin(sdp, TRANS_FILE_DEALLOC, blocks + 2, 0);
	if (error)
		goto fail_rg_gunlock;

	dibh = ogfs_get_inode_buffer(ip);
	if (!dibh) {
		error = -EIO;
		goto fail_end_trans;
	}

	ogfs_trans_add_bh(sdp, dibh);
	ogfs_trans_add_bh(sdp, bh);

	bstart = 0;
	blen = 0;

	for (p = top; p < bottom; p++) {
		if (!*p)
			continue;

		bn = ogfs64_to_cpu(*p);

		if (bstart + blen == bn)
			blen++;
		else {
			if (bstart) {
				if ((height+1 == ip->i_di.di_height )
				    && !ogfs_is_jdata(ip))
					error = ogfs_blkfree(ip, bstart, blen);
				else
					error = ogfs_metafree(ip, bstart, blen);

				if (error)
					goto fail_drelse;
			}

			bstart = bn;
			blen = 1;
		}

		*p = 0;
		ip->i_di.di_blocks--;
	}

	if (bstart) {
		if ((height+1 == ip->i_di.di_height ) && !ogfs_is_jdata(ip))
			error = ogfs_blkfree(ip, bstart, blen);
		else
			error = ogfs_metafree(ip, bstart, blen);

		if (error)
			goto fail_drelse;
	}

	ctime = current_kernel_time();
	ip->i_di.di_mtime = ip->i_di.di_ctime = (int64)ctime.tv_sec;

	ogfs_dinode_out(&ip->i_di, BH_DATA(dibh));
	brelse(dibh);

	ogfs_trans_end(sdp);

	for (x = 0; x < rlist.rl_rgrps; x++) {
		rgd = rlist.rl_list[x];
		ogfs_gunlock_rg(rgd, 0);
	}
	ogfs_rlist_free(&rlist);
	ogfs_rindex_release(sdp);

	EXIT(GFN_DO_STRIP);

	return 0;

fail_drelse:
	brelse(dibh);

fail_end_trans:
	ogfs_trans_mark(sdp, TRF_POISONED);
	ogfs_trans_end(sdp);

fail_rg_gunlock:
	while (x--) {
		rgd = rlist.rl_list[x];
		ogfs_gunlock_rg(rgd, 0);
	}

fail_rlist_free:
	ogfs_rlist_free(&rlist);
	ogfs_rindex_release(sdp);

fail:
	EXIT(GFN_DO_STRIP);
	return error;
}

/**
 * ogfs_shrink - make a file smaller
 * @ip: the inode
 * @size: the size to make the file
 *
 * Called with an exclusive lock on @ip.
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_shrink(ogfs_inode_t * ip, uint64 size)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	ogfs_glock_t *gl = ip->i_gl;
	metapath_t mp;
	strip_mine_t sm;
	struct buffer_head *bh;
	struct timespec ctime;
	unsigned int height;
	uint64 bn;
	int not_new = 0;
	int error = 0;

	ENTER(GFN_SHRINK);

	OGFS_ASSERT(!(ogfs_is_jdata(ip) && size),);

	find_metapath(ip, &mp,
		      (size) ? ((size - 1) >> sdp->sd_sb.sb_bsize_shift) : 0);

	/*  Get rid of all the data/metadata blocks  */

	for (height = ip->i_di.di_height; height--;) {
		sm.sm_first = (size) ? TRUE : FALSE;
		sm.sm_height = height;

		error = recursive_scan(ip, &mp, 0, 0, TRUE, do_strip, &sm);
		if (error)
			goto out;
	}

	/*
	 * If we truncated in the middle of a block, zero out the leftovers.
	 */
	if (ip->i_di.di_height && (size % sdp->sd_sb.sb_bsize)) {
		error =
		    ogfs_inode_get_block(ip, size >> sdp->sd_sb.sb_bsize_shift,
					&not_new, &bn);
		if (error)
			goto out;

		if (bn) {
			bh = ogfs_dread(sdp, bn, gl, DR_START | DR_WAIT);
			if (!bh) {
				error = -EIO;
				goto out;
			}

			ogfs_buffer_clear_tail(bh, size % sdp->sd_sb.sb_bsize);

			ogfs_dwrite(sdp, bh, DW_DIRTY);
			brelse(bh);
		}
	}

	/*
	 * Set the new size (and possibly the height).
	 */
	error = ogfs_trans_begin(sdp, TRANS_SET_SIZE, 1, 0);
	if (error)
		goto out;

	bh = ogfs_get_inode_buffer(ip);
	if (!bh) {
		ogfs_trans_end(sdp);
		error = -EIO;
		goto out;
	}

	ogfs_trans_add_bh(sdp, bh);

	ip->i_di.di_size = size;
	if (!size)
		ip->i_di.di_height = 0;
	ctime = current_kernel_time();
	ip->i_di.di_mtime = ip->i_di.di_ctime = (int64)ctime.tv_sec;

	if (!ip->i_di.di_height
	    && (size < sdp->sd_sb.sb_bsize - sizeof(ogfs_dinode_t)))
		ogfs_buffer_clear_tail(bh, sizeof(ogfs_dinode_t) + size);

	ogfs_dinode_out(&ip->i_di, BH_DATA(bh));
	brelse(bh);

	ogfs_trans_end(sdp);

out:
	EXIT(GFN_SHRINK);

	return error;
}

/**
 * ogfs_truncate - make a file a give size
 * @ip: the inode
 * @size: the size to make the file
 * @cred: the caller's credentials
 *
 * The file size can grow, shrink, or stay the same size.
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_truncate(ogfs_inode_t * ip, uint64 size)
{
	int error;

	ENTER(GFN_TRUNCATE);

	OGFS_ASSERT(ip->i_di.di_type == FILE_REG,);
	OGFS_ASSERT(!ogfs_is_jdata(ip),);

	error = ogfs_glock_i(ip, 0);
	if (error)
		goto out;

	error = ogfs_permission(ip, MAY_WRITE);
	if (error) {
		ogfs_gunlock_i(ip, 0);
		goto out;
	}

	if (size >= ip->i_di.di_size)
		/* Make bigger (or just modify the timestamps) */
		error = do_grow(ip, size);
	else
		/* Make smaller */
		error = ogfs_shrink(ip, size);

	ogfs_gunlock_i(ip, 0);

out:
	EXIT(GFN_TRUNCATE);

	return error;
}
