/*
 *
 *    Copyright 1995-2000 Regents of the University of Minnesota
 *    Portions Copyright 1999-2001 Sistina Software, Inc.
 *    Portions Copyright 2001 The OpenGFS Project
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 */

#include <linux/dcache.h>

#include "ogfs_debug.h"
#include "ogfs.h"

#include <linux/smp_lock.h>

#include "glock.h"
#include "inode.h"
#include "recovery.h"

/**
 * ogfs_glockd - Glock-harvesting kernel thread
 * @data: Pointer to OGFS superblock
 *
 * A kernel thread which harvests ogfs_bhdata_t structures laying
 * around. It also demotes glocks which are no longer being used.
 *
 * Returns: Process exit status
 */

int
ogfs_glockd(void *data)
{
	ogfs_sbd_t *sdp = (ogfs_sbd_t *) data;

	ENTER(GFN_GLOCKD);

	OGFS_ASSERT(sdp,);

	/*  Set up the thread  */

	lock_kernel();

	daemonize("glockd");

	sprintf(current->comm, "ogfs_glockd");

	sdp->sd_glockd_process = current;

	up(&sdp->sd_glockd_sema);

	/*
	 * sdp->sd_glockd_ctrl is the number of seconds between each
	 * run. When set to zero it causes the thread to exit.
	 */

	while (atomic_read(&sdp->sd_glockd_ctrl)) {
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(atomic_read(&sdp->sd_glockd_ctrl)*HZ); /* FIXME: race */

		IFDB(GFN_GLOCKD,
		     db_printf("\n\n\n  ogfs_glockd wakes up\n");
		     db_printf("Entering ogfs_glockd() core\n"););

		/*
		 * Run the arch independent part of ogfs_glockd
		 */

		ogfs_glockd_scan(sdp, sdp->sd_gl_heldtime);

		IFDB(GFN_GLOCKD, db_printf("Exiting ogfs_glockd() core\n"););
	}

	/*
	 * Signal to umount that we have finished
	 */

	up(&sdp->sd_glockd_sema);

	unlock_kernel();

	EXIT(GFN_GLOCKD);

	return 0;
}

/**
 * ogfs_inoded - Inode-harvesting kernel thread
 * @data: Pointer to OGFS superblock
 *
 * A kernel thread that wakes up every so often and gets rid of all
 * ogfs_bhdata_t structures laying around.
 *
 * Returns: Process exit status
 */

int
ogfs_inoded(void *data)
{
	ogfs_sbd_t *sdp = (ogfs_sbd_t *) data;
	int error;

	ENTER(GFN_INODED);

	OGFS_ASSERT(sdp,);

	/*  Set up the thread  */

	lock_kernel();

	daemonize("inoded");

	sprintf(current->comm, "ogfs_inoded");

	sdp->sd_inoded_process = current;

	up(&sdp->sd_inoded_sema);

	/*
	 * sdp->sd_inoded_ctrl is the number of seconds between each
	 * run. When set to zero it causes the thread to exit.
	 */

	while (atomic_read(&sdp->sd_inoded_ctrl)) {
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(atomic_read(&sdp->sd_inoded_ctrl)*HZ); /* FIXME race */

		IFDB(GFN_INODED,
		     db_printf("\n\n\n  ogfs_inoded wakes up\n");
		     db_printf("Entering ogfs_inoded() core\n"););

		/*
		 * Run the arch independent part of ogfs_inoded
		 */

		error = ogfs_dealloc_inodes(sdp);
		if (error)
			con_printf("ogfs_inoded:  error = %d\n", error);

		IFDB(GFN_INODED, db_printf("Exiting ogfs_inoded() core\n"););
	}

	/*
	 * Signal to umount that we have finished
	 */

	up(&sdp->sd_inoded_sema);

	unlock_kernel();

	EXIT(GFN_INODED);

	return 0;
}

/**
 * ogfs_recoverd - Recovery kernel thread
 * @data: Pointer to OGFS superblock
 *
 * A kernel thread that frees any inodes that have a zero usage count and
 * no VFS inode pointer.
 *
 * Returns: Process exit status
 */

int
ogfs_recoverd(void *data)
{
	ogfs_sbd_t *sdp = (ogfs_sbd_t *) data;
	int error;

	ENTER(GFN_RECOVERD);

	OGFS_ASSERT(sdp,);

	/*  Set up the thread  */

	lock_kernel();

	daemonize("recoverd");

	sprintf(current->comm, "ogfs_recoverd");

	sdp->sd_recoverd_process = current;

	up(&sdp->sd_recoverd_sema);

	/*
	 * sdp->sd_recoverd_ctrl is the number of seconds between each
	 * run. When set to zero it causes the thread to exit.
	 */

	while (atomic_read(&sdp->sd_recoverd_ctrl)) {
		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(atomic_read(&sdp->sd_recoverd_ctrl)*HZ); /* FIXME race */

		IFDB(GFN_RECOVERD,
		     db_printf("\n\n\n  ogfs_recoverd wakes up\n");
		     db_printf("Entering ogfs_recoverd() core\n"););

		/*
		 * Run the arch independent part of ogfs_recoverd
		 */

		error = ogfs_check_journals(sdp);
		if (error)
			con_printf("ogfs_recoverd:  error = %d\n", error);

		IFDB(GFN_RECOVERD, db_printf("Exiting ogfs_recoverd() core\n"););
	}

	up(&sdp->sd_recoverd_sema);

	unlock_kernel();

	EXIT(GFN_RECOVERD);

	return 0;
}
