/*
 *
 *    Copyright 1995-1999 Regents of the University of Minnesota
 *    Portion Copyright 2000-2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 */

#include <linux/dcache.h>
#include <linux/fs.h>

#include "ogfs_debug.h"
#include "ogfs.h"

#include <linux/smp_lock.h>

#include "blklist.h"
#include "dcache.h"
#include "dir.h"
#include "ogfs_linux.h"
#include "glock.h"

/**
 * ogfs_drevalidate - Check directory lookup consistency
 * @dentry: the mapping to check
 * @flags: Who knows?
 *
 * Check to make sure the lookup necessary to arrive at this inode from its
 * parent is still good.
 *
 * Returns: 1 if the dentry is ok, 0 if it isn't
 */
int
ogfs_drevalidate(struct dentry *dentry, struct nameidata *nidata)
{
	ogfs_inode_t *dip = vi2ino(dentry->d_parent->d_inode);
	ogfs_dcached_t *dcd;
	ogfs_inode_t *ip;
	ogfs_inum_t inum;
	int ret = 0;
	int error;

	ENTER(GFN_DREVALIDATE);

	lock_kernel();

	OGFS_ASSERT(dip,);

	if (!dentry->d_inode || is_bad_inode(dentry->d_inode))
		goto out;

	ip = vi2ino(dentry->d_inode);
	OGFS_ASSERT(ip,);

	error = ogfs_glock_i(dip, GL_SHARED);
	if (error)
		goto out_print;

	dcd = vdc2dcd(dentry);

	if (dcd && dcd->dc_vn == dip->i_gl->gl_vn) {
		ret = 1;
		goto out_gunlock;
	}

	error = ogfs_dir_search(dip, &dentry->d_name, &inum, NULL);
	if (error)
		goto out_gunlock;

	if (inum.no_formal_ino == ip->i_num.no_formal_ino) {
		if (dcd)
			dcd->dc_vn = dip->i_gl->gl_vn;
		ret = 1;
	}

out_gunlock:
	ogfs_gunlock_i(dip, 0);

out_print:
	IFDB(GFN_DREVALIDATE,
	     db_printf("(%" FMT_64 "u, %s --> %" FMT_64 "u) is %s\n",
		       dip->i_num.no_formal_ino,
		       dentry->d_name.name,
		       ip->i_num.no_formal_ino, (ret) ? "valid" : "invalid"););

	if (ret == 0) {
		shrink_dcache_parent(dentry);
		/* If we have submounts, don't unhash ! */
		if (have_submounts(dentry)) {
			ret = 1;
			goto out;
		}
		d_drop(dentry);
	}

out:
	unlock_kernel();

	EXIT(GFN_DREVALIDATE);

	return ret;
}

/**
 * ogfs_drelease - Free the OGFS specific data from a dentry about to be freed
 * @dentry: the dentry being freed
 *
 */

void
ogfs_drelease(struct dentry *dentry)
{
	ogfs_dcached_t *dcd = vdc2dcd(dentry);

	ENTER(GFN_DRELEASE);

	if (dcd)
		gfree(dcd, sizeof(ogfs_dcached_t));

	vdc2dcd(dentry) = NULL;

	EXIT(GFN_DRELEASE);
}

struct dentry_operations ogfs_dops = {
	d_revalidate:	ogfs_drevalidate,
	d_release:	ogfs_drelease
};
