/*
 *
 *    Copyright 2000-2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 */

/*
 * disk I/O (buffer cache) accounting wrappers
 */

#include "ogfs_debug.h"
#include "ogfs.h"
#include "dio.h"
#include "fs.h"

#ifdef DEBUG_BUF

struct buf_call {
	struct list_head list;

	struct buffer_head *bh;
	unsigned int count;

	char *file;
	unsigned int line;
};
typedef struct buf_call buf_call_t;

/**
 * buf_hash - compute a hash on a buffer pointer
 * @bh: the buffer pointer
 *
 * Returns: the hash
 */

static __inline__ uint32
buf_hash(struct buffer_head * bh)
{
	uint32 val;

	val = ogfs_hash((char *) &bh, sizeof(struct buffer_head *));
	val &= (BUF_HASH_SIZE - 1);

	return val;
}

/**
 * buf_alloc_brute - mark a buffer as allocated
 * @bh: the buffer
 * @file: the file containing the getblk
 * @line: the line it's on
 * @new: Set to TRUE if this is a new buffer
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
buf_alloc_brute(struct buffer_head * bh, char *file, unsigned int line, int *new)
{
	struct list_head *tmp;
	buf_call_t *oc, *oc2;
	unsigned int index = buf_hash(bh);
	int found = FALSE;

	ENTER(GFN_BUF_ALLOC_BRUTE);

	*new = TRUE;

	oc = (buf_call_t *) gmalloc(sizeof(buf_call_t));
	if (!oc) {
		EXIT(GFN_BUF_ALLOC_BRUTE);
		return -ENOMEM;
	}

	oc->bh = bh;
	oc->count = 1;

	oc->file = file;
	oc->line = line;

	spin_lock(&buf_lock);

	for (tmp = buf_head[index].next; tmp != &buf_head[index];
	     tmp = tmp->next) {
		oc2 = list_entry(tmp, buf_call_t, list);

		if (oc2->bh != bh)
			continue;

		*new = FALSE;

		oc2->count++;
		oc->count = oc2->count;

		if (oc2->line == line && !strcmp(oc2->file, file))
			found = TRUE;
	}

	if (!found)
		list_add(&oc->list, &buf_head[index]);

	spin_unlock(&buf_lock);

	if (found)
		gfree(oc, sizeof(buf_call_t));

	EXIT(GFN_BUF_ALLOC_BRUTE);

	return 0;
}

/**
 * buf_free_brute - mark a buffer as deallocated
 * @bh: the buffer
 * @file: the file containing the relse
 * @line: the line it's on
 * @last_one: Set to TRUE if this is the last allocation
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
buf_free_brute(struct buffer_head * bh, char *file, unsigned int line, int *last_one)
{
	unsigned int index = buf_hash(bh);
	struct list_head *tmp, *head, *last;
	buf_call_t *oc;
	int found = FALSE;

	ENTER(GFN_BUF_FREE_BRUTE);

	*last_one = FALSE;

	spin_lock(&buf_lock);

	last = head = &buf_head[index];
	for (tmp = head->next; tmp != head; tmp = last->next) {
		oc = list_entry(tmp, buf_call_t, list);

		if (oc->bh != bh) {
			last = tmp;
			continue;
		}

		found = TRUE;

		OGFS_ASSERT(oc->count,);
		oc->count--;

		if (!oc->count) {
			list_del(&oc->list);
			gfree(oc, sizeof(buf_call_t));
			*last_one = TRUE;
		} else
			last = tmp;
	}

	spin_unlock(&buf_lock);

	if (!found) {
		con_printf("OGFS: freeing non allocated blk <%lX> (%s, %u)\n",
			   (unsigned long) bh, file, line);
		EXIT(GFN_BUF_FREE_BRUTE);
		return -EINVAL;
	}

	EXIT(GFN_BUF_FREE_BRUTE);

	return 0;
}

/**
 * buf_inc_stats - Increment the stats
 * @new:  TRUE if this is a new allocation
 *
 */

static void
buf_inc_stats(int new)
{
	ENTER(GFN_BUF_INC_STATS);

	spin_lock(&buf_lock);

	if (new) {
		buf_stats.ct++;
		buf_stats.ct_total++;

		if (buf_stats.ct > buf_stats.ct_max)
			buf_stats.ct_max = buf_stats.ct;
	}

	buf_stats.refs++;
	buf_stats.refs_total++;

	if (buf_stats.refs > buf_stats.refs_max)
		buf_stats.refs_max = buf_stats.refs;

	spin_unlock(&buf_lock);

	EXIT(GFN_BUF_INC_STATS);
}

/**
 * buf_dec_stats - Decrement the stats
 * @last_one:  TRUE if this was the last deallocation
 *
 */

static void
buf_dec_stats(int last_one)
{
	ENTER(GFN_BUF_DEC_STATS);

	spin_lock(&buf_lock);

	if (last_one)
		buf_stats.ct--;

	buf_stats.refs--;

	spin_unlock(&buf_lock);

	EXIT(GFN_BUF_DEC_STATS);
}

/**
 * buf_alloc_main - Handle a buffer allocation
 * @bh: the buffer
 * @file: the file the getblk occurs on
 * @line: the line
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
buf_alloc_main(struct buffer_head * bh, char *file, unsigned int line)
{
	int error;
	int new;

	ENTER(GFN_BUF_ALLOC_MAIN);

	error = buf_alloc_brute(bh, file, line, &new);
	if (error) {
		EXIT(GFN_BUF_ALLOC_MAIN);
		return error;
	}

	buf_inc_stats(new);

	EXIT(GFN_BUF_ALLOC_MAIN);

	return 0;
}

/**
 * ogfs_dgetblk_wrapper - the wrapper around ogfs_dgetblk()
 * @sdp: the filesystem
 * @blkno: the block number
 * @gl: the glock
 * @file: the file the getblk occurs on
 * @line: the line
 *
 * Returns: a new buffer
 */

struct buffer_head *
ogfs_dgetblk_wrapper(ogfs_sbd_t * sdp, uint64 blkno, ogfs_glock_t * gl,
		    char *file, unsigned int line)
{
	struct buffer_head *bh = NULL;
	int error;

	ENTER(GFN_DGETBLK_WRAPPER);

	bh = ogfs_dgetblk_arch(sdp, blkno, gl);
	if (!bh)
		goto out;

	error = buf_alloc_main(bh, file, line);
	if (error) {
		brelse(bh);
		bh = NULL;
	}

out:
	EXIT(GFN_DGETBLK_WRAPPER);

	return bh;
}

/**
 * ogfs_dread_wrapper - the wrapper around ogfs_dread()
 * @sdp: the filesystem
 * @blkno: the block number
 * @gl: the glock
 * @flags: flags
 * @file: the file the dread occurs on
 * @line: the line
 *
 * Returns: a new buffer
 */

struct buffer_head *
ogfs_dread_wrapper(ogfs_sbd_t * sdp, uint64 blkno, ogfs_glock_t * gl, int flags,
		  char *file, unsigned int line)
{
	struct buffer_head *bh = NULL;
	int error;

	ENTER(GFN_DREAD_WRAPPER);

	bh = ogfs_dread_arch(sdp, blkno, gl, flags);
	if (!bh)
		goto out;

	error = buf_alloc_main(bh, file, line);
	if (error) {
		brelse(bh);
		bh = NULL;
	}

out:
	EXIT(GFN_DREAD_WRAPPER);

	return bh;
}

/**
 * buf_free_main - Handle a buffer deallocation
 * @bh: the buffer
 * @file: the file the getblk occurs on
 * @line: the line
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
buf_free_main(struct buffer_head * bh, char *file, unsigned int line)
{
	int last_one;
	int error;

	ENTER(GFN_BUF_FREE_MAIN);

	if (!bh) {
		con_printf("OGFS:  releasing NULL buffer (%s, %u)\n", file,
			   line);
		EXIT(GFN_BUF_FREE_MAIN);
		return -EINVAL;
	}

	error = buf_free_brute(bh, file, line, &last_one);
	if (error) {
		EXIT(GFN_BUF_FREE_MAIN);
		return error;
	}

	buf_dec_stats(last_one);

	EXIT(GFN_BUF_FREE_MAIN);

	return 0;
}

/**
 * brelse_wrapper - the wrapper around brelse()
 * @bh: the buffer
 * @file: the file the brelse() occurs on
 * @line: the line
 *
 * Function description
 *
 * Returns: what is returned
 */

void
brelse_wrapper(struct buffer_head * bh, char *file, unsigned int line)
{
	int error;

	ENTER(GFN_DRELSE_WRAPPER);

	error = buf_free_main(bh, file, line);
	if (error) {
		EXIT(GFN_DRELSE_WRAPPER);
		return;
	}

	brelse_arch(bh);

	EXIT(GFN_DRELSE_WRAPPER);
}

/**
 * ogfs_dforget_wrapper - the wrapper around ogfs_dforget()
 * @bh: the buffer
 * @file: the file the ogfs_dforget() occurs on
 * @line: the line
 *
 * Function description
 *
 * Returns: what is returned
 */

void
ogfs_dforget_wrapper(struct buffer_head * bh, char *file, unsigned int line)
{
	int error;

	ENTER(GFN_DFORGET_WRAPPER);

	error = buf_free_main(bh, file, line);
	if (error) {
		EXIT(GFN_DFORGET_WRAPPER);
		return;
	}

	ogfs_dforget_arch(bh);

	EXIT(GFN_DFORGET_WRAPPER);
}

/**
 * buf_check - Called at unload to print stats
 *
 */

void
buf_check(void)
{
	buf_call_t *oc;
	unsigned int index;

	ENTER(GFN_BUF_CHECK);

	spin_lock(&buf_lock);

	for (index = 0; index < BUF_HASH_SIZE; index++)
		while (!list_empty(&buf_head[index])) {
			oc = list_entry(buf_head[index].next, buf_call_t,
					    list);
			list_del(&oc->list);

			spin_unlock(&buf_lock);

			con_printf("OGFS:  unfreed buffer <%lX>, %" FMT_64
				   "u, (%s, %u), %u\n", (unsigned long) oc->bh,
				   BH_BLKNO(oc->bh), oc->file, oc->line,
				   oc->count);

			gfree(oc, sizeof(buf_call_t));

			spin_lock(&buf_lock);
		}

	spin_unlock(&buf_lock);

	con_printf("OGFS:  buf_ct:          %u\n", buf_stats.ct);
	con_printf("OGFS:  buf_ct_max:      %u\n", buf_stats.ct_max);
	con_printf("OGFS:  buf_ct_total:    %u\n", buf_stats.ct_total);
	con_printf("OGFS:  refs:            %u\n", buf_stats.refs);
	con_printf("OGFS:  refs_max:        %u\n", buf_stats.refs_max);
	con_printf("OGFS:  refs_total:      %u\n", buf_stats.refs_total);

	EXIT(GFN_BUF_CHECK);
}

#endif				/*  DEBUG_BUF  */
