/*
 *
 *    Copyright 1995-2000 Regents of the University of Minnesota
 *    Portions Copyright 1999-2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 */
#ifndef _DIO_H
#define _DIO_H

#include "dio_arch.h"
#include <linux/buffer_head.h>
/*  Arguments to ogfs_dread()  */

#define DR_FORCE                (0x01000000)
#define DR_START                (0x02000000)
#define DR_WAIT                 (0x04000000)

/*  Arguments to ogfs_dwrite()  */

#define DW_DIRTY                (0x00100000)
#define DW_START                (0x00200000)
#define DW_WAIT                 (0x00400000)

/*  Buffer accounting stuff

    the ct variables are changed each time we ask for
    a pointer to a buffer that we don't already have
    somewhere else. For

    bh = ogfs_getblk(1024);
    bh1 = ogfs_getblk(1024);

    ct = 1 and refs = 2
    */

#ifdef DEBUG_BUF

#define BUF_HASH_SIZE 4096
extern struct list_head buf_head[BUF_HASH_SIZE];

struct buf_stats {
	unsigned int ct;
	unsigned int ct_max;
	unsigned int ct_total;
	unsigned int refs;
	unsigned int refs_max;
	unsigned int refs_total;
};
typedef struct buf_stats buf_stats_t;

extern buf_stats_t buf_stats;
extern spinlock_t buf_lock;

struct buffer_head *ogfs_dgetblk_wrapper(ogfs_sbd_t * sdp, uint64 blkno,
			       ogfs_glock_t * gl, char *file, unsigned int line);
struct buffer_head *ogfs_dread_wrapper(ogfs_sbd_t * sdp, uint64 blkno,
			     ogfs_glock_t * gl, int flags,
			     char *file, unsigned int line);
struct buffer_head *ogfs_dread_wrapper(ogfs_sbd_t * sdp, uint64 blkno, int flags,
					char *file, unsigned int line);
void ogfs_drelse_wrapper(struct buffer_head * bh, char *file, unsigned int line);
void ogfs_dforget_wrapper(struct buffer_head * bh, char *file, unsigned int line);
int buf_alloc_main(struct buffer_head * bh, char *file, unsigned int line);
void buf_check(void);

#define ogfs_dgetblk(sdp, blkno, gl) ogfs_dgetblk_wrapper(sdp, blkno, gl, __FILE__, __LINE__)
#define ogfs_dread(sdp, blkno, gl, flags) ogfs_dread_wrapper(sdp, blkno, gl, flags, __FILE__, __LINE__)
#define ogfs_drelse(bh) brelse(bh)
#define ogfs_dforget(bh) ogfs_dforget_wrapper(bh, __FILE__, __LINE__)

#else				/*  DEBUG_BUF  */

#define buf_alloc_main(bh, file, line)
#define buf_check()

#define ogfs_dgetblk(sdp, blkno, gl) ogfs_dgetblk_arch(sdp, blkno, gl)
#define ogfs_dread(sdp, blkno, gl, flags) ogfs_dread_arch(sdp, blkno, gl, flags)
#define ogfs_drelse(bh) brelse(bh)
#define ogfs_drelse_arch(bh) brelse(bh)
#define ogfs_dforget(bh) ogfs_dforget_arch(bh)

#endif				/*  DEBUG_BUF  */

/*  Asynchronous I/O Routines  */

struct buffer_head *ogfs_dgetblk_arch(ogfs_sbd_t * sdp, uint64 blkno, ogfs_glock_t * gl);
struct buffer_head *ogfs_dread_arch(ogfs_sbd_t * sdp, uint64 blkno, ogfs_glock_t * gl,
			  int flags);
void brelse(struct buffer_head * bh);
void ogfs_dforget_arch(struct buffer_head * bh);

void ogfs_prep_new_buffer(struct buffer_head * bh);
int ogfs_dreread(ogfs_sbd_t * sdp, struct buffer_head * bh, int flags);
int ogfs_dwrite(ogfs_sbd_t * sdp, struct buffer_head * bh, int flags);

void ogfs_dpin(struct buffer_head * bh);
void ogfs_dunpin(struct buffer_head * bh, ogfs_trans_t * tr);
int ogfs_is_pinned(struct buffer_head * bh);

int ogfs_put_jdev_info(ogfs_sbd_t * sdp, int jid);
int ogfs_get_jdev_info(ogfs_sbd_t * sdp, int jid);

#define ogfs_lock_buffer(bh) down(&(BH_BUFDATA((bh)))->bd_lock)
#define ogfs_unlock_buffer(bh) up(&(BH_BUFDATA((bh)))->bd_lock)

struct buffer_head *ogfs_init_temp_bh(ogfs_sbd_t * sdp, uint64 blocknr, char *data,
			    int write);
void ogfs_dispose_temp_bh(struct buffer_head * bh);

/*  Buffer Caching routines  */

struct buffer_head *ogfs_get_meta_buffer(ogfs_inode_t * ip, int height, uint64 num,
			       int new);
struct buffer_head *ogfs_get_data_buffer(ogfs_inode_t * ip, uint64 block, int new);

static __inline__ void
ogfs_flush_meta_cache(ogfs_inode_t * ip, int start)
{
	int i;

	for (i = start; i < OGFS_MAX_META_HEIGHT; i++) {
		if (!ip->i_cache[i])
			break;
		brelse(ip->i_cache[i]);
		ip->i_cache[i] = NULL;
	}
}

static __inline__ struct buffer_head *
ogfs_get_inode_buffer(ogfs_inode_t * ip)
{
	return ogfs_get_meta_buffer(ip, 0, ip->i_num.no_addr, 0);
}

/*  Buffer Content Functions  */

/**
 * ogfs_buffer_clear - Zeros out a buffer
 * @ip: The OGFS inode
 * @bh: The buffer to zero
 *
 */

static __inline__ void
ogfs_buffer_clear(struct buffer_head * bh)
{
	ogfs_memset(BH_DATA(bh), 0, BH_SIZE(bh));
}

/**
 * ogfs_buffer_clear_tail - Clear buffer beyond the dinode
 * @bh: The buffer containing the on-disk inode
 * @head: the size of the head of the buffer
 *
 * Clears the remaining part of an on-disk inode that is not a dinode.
 * i.e. The data part of a stuffed inode, or the top level of metadata
 * of a non-stuffed inode.
 */

static __inline__ void
ogfs_buffer_clear_tail(struct buffer_head * bh, int head)
{
	ogfs_memset(BH_DATA(bh) + head, 0, BH_SIZE(bh) - head);
}

/**
 * ogfs_buffer_clear_ends - Zero out any bits of a buffer which are not being written
 * @bh: The buffer
 * @offset: Offset in buffer where write starts
 * @amount: Amount of data being written
 * @journaled: TRUE if this is a journaled buffer
 *
 */

static __inline__ void
ogfs_buffer_clear_ends(struct buffer_head * bh, int offset, int amount, int journaled)
{
	int z_off1 = (journaled) ? sizeof(ogfs_jdata_t) : 0;
	int z_len1 = offset - z_off1;
	int z_off2 = offset + amount;
	int z_len2 = BH_SIZE(bh) - z_off2;

	if (z_len1)
		ogfs_memset(BH_DATA(bh) + z_off1, 0, z_len1);

	if (z_len2)
		ogfs_memset(BH_DATA(bh) + z_off2, 0, z_len2);
}

/**
 * ogfs_buffer_copy_tail - copies the tail of one buffer to another
 * @to_bh: the buffer to copy to
 * @to_head: the size of the head of to_bh
 * @from_bh: the buffer to copy from
 * @from_head: the size of the head of from_bh
 *
 * from_head is guaranteed to bigger than to_head 
 */

static __inline__ void
ogfs_buffer_copy_tail(struct buffer_head * to_bh, int to_head,
		     struct buffer_head * from_bh, int from_head)
{
	ogfs_memcpy(BH_DATA(to_bh) + to_head,
		   BH_DATA(from_bh) + from_head, BH_SIZE(from_bh) - from_head);
	ogfs_memset(BH_DATA(to_bh) + BH_SIZE(to_bh) + to_head - from_head,
		   0, from_head - to_head);
}

/**
 * ogfs_buffer_print - print a buffer to the debug console
 * @bh: the buffer
 * @string:  what to print before the contents of the buffer
 *
 */

static __inline__ void
ogfs_buffer_print(struct buffer_head * bh, char *string)
{
#ifdef DEBUG_PRINT
	unsigned int x, size = BH_SIZE(bh);
	unsigned char *c = BH_DATA(bh);

	db_printf("%s\n", string);

	for (x = 0; x < size; x++) {
		db_printf("%.2X ", c[x]);
		if (x % 16 == 15)
			db_printf("\n");
	}
	if (x % 16 != 0)
		db_printf("\n");
#endif				/*  DEBUG_PRINT  */
}

#endif /* _DIO_H */
