/*
 *
 *    Copyright 1995-2000 Regents of the University of Minnesota
 *    Portions Copyright 1999-2001 Sistina Software, Inc.
 *    Portions Copyright 2001-2003 The OpenGFS Project
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 */

/*
 * This module implements a bunch of routines that do I/O through
 * the buffer cache.
 */


#include "ogfs_debug.h"
#include "ogfs.h"
//#include "yield.h"

#include <linux/buffer_head.h>
#include <linux/blkdev.h>
#include <linux/namei.h>
#include <linux/mount.h>
#include <linux/fs.h>
//extern int bd_acquire(struct inode *inode);

#include "dio.h"
#include "ogfs_linux.h"
#include "glock.h"
#include "inode.h"
#include "log.h"
#include "trans.h"


/*
 * Protects the OGFS-Internal buffer lists:
 *
 *  o ogfs_bufdata->bd_arch.bd_list
 *  o ogfs_glock->gl_arch.gl_bufs;
 */
spinlock_t ogfs_bufferlist_lock = SPIN_LOCK_UNLOCKED;


/**
 * rm_bh_internal - Remove a buffer from the list maintained for invalidation
 * @bh: The buffer head to remove
 *
 */

static void
rm_bh_internal(struct buffer_head * bh)
{
	ogfs_bufdata_t *bd = BH_BUFDATA(bh);
	ogfs_glock_t *gl;

	OGFS_ASSERT(bd,);
	gl = bd->bd_gl;

	if (bd->bd_arch.bd_ail_tr)
		list_del(&bd->bd_arch.bd_ail_list);

	//bh->b_rm_fn = NULL;
	bh->b_private = NULL;
	//bh->b_journal_head = NULL;
	list_del(&bd->bd_arch.bd_list);

	if (list_empty(&gl->gl_arch.gl_bufs))
		set_bit(GLF_DEMOTEME, &gl->gl_flags);

	ogfs_put_glstruct(gl);
	ogfs_free_bufdata(bd);
}

/**
 * ogfs_remove_bh - Remove a buffer from the list maintainted for invalidation
 * @bh: The buffer head to be removed
 *
 * This function is the one the buffer cache sees.  It checks to make sure
 * the buffer cache isn't throwing away an active buffer and then calls
 * rm_bh_internal().
 *
 */

void
ogfs_remove_bh(struct buffer_head * bh)
{
	OGFS_ASSERT(!atomic_read(&bh->b_count),);
	rm_bh_internal(bh);
}

/**
 * ogfs_inval_bh - Invalidate all buffers associated with a glock
 * @sdp: the filesystem
 * @gl: the glock
 *
 */

void
ogfs_inval_bh(ogfs_sbd_t * sdp, ogfs_glock_t * gl)
{
	struct list_head *tmp, *head;
	ogfs_bufdata_t *bd;
	struct buffer_head *bh;

	ENTER(GFN_INVAL_BH);

 restart:
	spin_lock(&ogfs_bufferlist_lock);

	tmp = head = &gl->gl_arch.gl_bufs;
	for (;;) {
		tmp = tmp->next;
		if (tmp == head)
			break;
		bd = list_entry(tmp, ogfs_bufdata_t, bd_arch.bd_list);
		bh = bd->bd_bh;

		OGFS_ASSERT(!buffer_dirty(bh),);

		if (buffer_locked(bh)) {
			spin_unlock(&ogfs_bufferlist_lock);
			wait_on_buffer(bh);
			goto restart;
		}

		clear_buffer_uptodate(bh);
	}

	spin_unlock(&ogfs_bufferlist_lock);

	EXIT(GFN_INVAL_BH);
}

/**
 * ogfs_sync_bh - Sync all buffers associated with a glock
 * @sdp: the filesystem
 * @gl: The glock
 * @wait: Non-zero means wait for completion of I/O
 *
 */

void
ogfs_sync_bh(ogfs_sbd_t * sdp, ogfs_glock_t * gl, int wait)
{
	struct list_head *tmp, *head;
	ogfs_bufdata_t *bd;
	struct buffer_head *bh;
	int pass = 0;

	ENTER(GFN_SYNC_BH);

restart:
	spin_lock(&ogfs_bufferlist_lock);

	tmp = head = &gl->gl_arch.gl_bufs;
	for (;;) {
		tmp = tmp->next;
		if (tmp == head)
			break;
		bd = list_entry(tmp, ogfs_bufdata_t, bd_arch.bd_list);
		bh = bd->bd_bh;

		if (buffer_dirty(bh) && !buffer_locked(bh)) {
			spin_unlock(&ogfs_bufferlist_lock);
			ll_rw_block(WRITE, 1, &bh);
			goto restart;
		}

		if (pass && buffer_locked(bh)) {
			spin_unlock(&ogfs_bufferlist_lock);
			wait_on_buffer(bh);
			goto restart;
		}
	}

	spin_unlock(&ogfs_bufferlist_lock);

	if (wait && !(pass++))
		goto restart;

	EXIT(GFN_SYNC_BH);
}

/**
 * ogfs_free_bh - Free all buffers associated with a glock
 * @sdp: the filesystem
 * @gl: The glock
 *
 * Remove all ogfs_linux_bhdata_t structures on the list.  Be careful to avoid
 * races with other processes removeing them naturally through
 * ogfs_remove_bh().  We do this by doing a getblk() on the buffer to
 * increment the count on the buffer.  (bh->b_count++ isn't good enough
 * to stop the race)
 *
 * Only called on unmount.
 *
 */

void
ogfs_free_bh(ogfs_sbd_t * sdp, ogfs_glock_t * gl)
{
	ogfs_bufdata_t *bd;
	struct buffer_head *bh;

	ENTER(GFN_FREE_BH);

	spin_lock(&ogfs_bufferlist_lock);
	while (!list_empty(&gl->gl_arch.gl_bufs)) {
		bd = list_entry(gl->gl_arch.gl_bufs.next, ogfs_bufdata_t,
				    bd_arch.bd_list);
		OGFS_ASSERT(bd,);

		bh = bd->bd_bh;
		OGFS_ASSERT(bh,);

		//bh->b_rm_fn = NULL;
		bh->b_private = NULL;
		//bh->b_journal_head = NULL;
		list_del(&bd->bd_arch.bd_list);

		ogfs_put_glstruct(gl);
		ogfs_free_bufdata(bd);
	}
	spin_unlock(&ogfs_bufferlist_lock);

	EXIT(GFN_FREE_BH);
}

/**
 * clear_from_ail - make sure a buffer isn't in the AIL
 * @sdp: the filesystem
 * @bh: the buffer
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
clear_from_ail(ogfs_sbd_t * sdp, struct buffer_head * bh)
{
	ogfs_bufdata_t *bd = BH_BUFDATA(bh);
	int error = 0;

	if (!bd)
		return 0;

	if (ogfs_is_pinned(bh))
		ogfs_log_flush(sdp, bd->bd_gl);

	spin_lock(&ogfs_bufferlist_lock);
	while (bd->bd_arch.bd_ail_tr) {
		if (!buffer_dirty(bh) && !buffer_locked(bh)) {
			bd->bd_arch.bd_ail_tr = NULL;
			list_del(&bd->bd_arch.bd_ail_list);
			break;
		}

		spin_unlock(&ogfs_bufferlist_lock);
		error = ogfs_dwrite(sdp, bh, DW_START | DW_WAIT);
		spin_lock(&ogfs_bufferlist_lock);

		if (error)
			break;
	}
	spin_unlock(&ogfs_bufferlist_lock);

	return error;
}

/**
 * ogfs_put_jdev_info - release block device struct for external journal
 * @sdp: The OGFS superblock
 * @jid: jid of journal we want to release
 *
 * Returns: 0 on success, error code on failure
 */

int
ogfs_put_jdev_info(ogfs_sbd_t * sdp, int jid)
{
	struct block_device *bdev;
	int error = 0;

	bdev = sdp->sd_jdevs[jid]; 
	blkdev_put(bdev, BDEV_FS);

	return error;
}

/**
 * ogfs_get_jdev_info - Get block device struct for external journal
 * @sdp: The OGFS superblock
 * @jid: jid of journal we want
 *
 * Get block device struct, using device path/name (e.g. /dev/sdb3)
 * Based on code in kernel's fs/super.c get_sb_bdev()
 *
 * Returns: 0 on success, error code on failure
 */

int
ogfs_get_jdev_info(ogfs_sbd_t * sdp, int jid)
{
	struct inode *inode;
	struct block_device *bdev;
	struct nameidata nd;
	dev_t dev;
	int error = 0;
	mode_t mode;
	char * dev_name = (sdp->sd_jindex + jid)->ji_device;
	char b[BDEVNAME_SIZE];

#if 0 /* devfs stuff */
	struct block_device_operations *bdops;
	devfs_handle_t de;
#endif 

	/* What device is it? */
	if (!dev_name || !*dev_name)
		return -EINVAL;
	error = path_lookup(dev_name, LOOKUP_FOLLOW, &nd);
	// LOOKUP_CONTINUE was above, I just dropped it, I think that's correct
	if (error)
		return error;
	inode = nd.dentry->d_inode;
	error = -ENOTBLK;
	if (!S_ISBLK(inode->i_mode))
		goto out;
	error = -EACCES;
	if (nd.mnt->mnt_flags & MNT_NODEV)
		goto out;
	//bd_acquire(inode);
	//bdev = inode->i_bdev;
	#warning MAYBE?
	bdev = bdget(inode->i_rdev);

	sdp->sd_jdevs[jid] = bdev; 

#if 0 /* devfs stuff */
	de = devfs_get_handle_from_inode (inode);
	bdops = devfs_get_ops (de);         /*  Increments module use count  */
	if (bdops) bdev->bd_op = bdops;
#endif 

	dev = bdev->bd_dev;
	printk("external journal ID %d, kdev = 0x%x, name = %s\n", jid, dev,
			__bdevname(dev, b));
	mode = FMODE_READ | FMODE_WRITE;
	error = blkdev_get(bdev, mode, 0, BDEV_FS);

#if 0 /* devfs stuff */
	devfs_put_ops (de);   /*  Decrement module use count now we're safe  */
#endif
	if (error) {
		printk("blkdev_get failed\n");
		goto out;
	}
#if 0 /* disk change, RO device? */
	check_disk_change(dev);
	error = -EACCES;
	if (!(flags & MS_RDONLY) && is_read_only(dev))
		goto out1;
#endif

	path_release(&nd);
	return 0;  /* success */

out:
	path_release(&nd);
	return error;
}


/**
 * translate_block - translate block address for external journal
 * @sdp: the filesystem
 *
 * Tries *this* machine's jindex first, then those for other machines.
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
translate_block(ogfs_sbd_t * sdp, dev_t * dev, uint64 * blkno)
{
	uint32 numblocks;
	uint64 base;
	uint32 j;

	/* try this machine's journal first */
	base = sdp->sd_my_jdesc.ji_addr;
	numblocks = sdp->sd_my_jdesc.ji_nsegment * sdp->sd_sb.sb_seg_size;
	if ((*blkno >= base) && ((*blkno - base) < numblocks)) {
		*blkno -= base;
		*dev = sdp->sd_my_jdev->bd_dev;
	}
	/* else search other journals' block ranges */
	else {
		for (j = 0;j < sdp->sd_journals; j++) {
			base = sdp->sd_jindex[j].ji_addr;
			numblocks = sdp->sd_jindex[j].ji_nsegment 
					* sdp->sd_sb.sb_seg_size;
			if ((*blkno >= base) && ((*blkno - base) < numblocks)) {
				*blkno -= base;
				*dev = sdp->sd_jdevs[j]->bd_dev;
				break;
			}
		}
		if (j == sdp->sd_journals) {
			printk("couldn't find a jindex for this block\n");
			return -1;
		}
	}
	if (*dev == 0)
		return -1;
	return 0;
}

/**
 * ogfs_dgetblk_arch - Get a block
 * @sdp: The OGFS superblock
 * @blkno: The block number
 * @gl: The glock associated with this block
 *
 * Returns: The buffer on success, or NULL
 */

struct buffer_head *
ogfs_dgetblk_arch(ogfs_sbd_t * sdp, uint64 blkno, ogfs_glock_t * gl)
{
	struct buffer_head *bh;
	dev_t dev;
	ogfs_bufdata_t *bd;

	ENTER(GFN_DGETBLK_ARCH);

	dev = sdp->sd_vsb->s_dev;  /* assume filesystem device */
	if (blkno >= sdp->sd_jext_base) {
		if (translate_block(sdp, &dev, &blkno)) {
			printk(KERN_ERR "translate_block failed\n");
			goto fail;
		}
	}
	bh = sb_getblk(sdp->sd_vsb, blkno);
	if (!bh) {
		printk(KERN_ERR "getblk failed\n");
		goto fail;
	}

	if (gl) {
		spin_lock(&ogfs_bufferlist_lock);

		/* buffer is from a deleted file */
#warning FIXME
		//if ((bh->b_rm_fn == ogfs_remove_bh)
		    //&& (gl != BH_BUFDATA(bh)->bd_gl)) {
		if(gl != BH_BUFDATA(bh)->bd_gl) {
			spin_unlock(&ogfs_bufferlist_lock);

			if (clear_from_ail(sdp, bh) != 0)
				goto fail_brelse;

			spin_lock(&ogfs_bufferlist_lock);

			rm_bh_internal(bh);
		}

		/* insert new bh into this lock's bufs list */

#warning FIXME TOO!
		//if (bh->b_rm_fn != ogfs_remove_bh) {
		if (0) {
			//OGFS_ASSERT(!bh->b_rm_fn,);

			spin_unlock(&ogfs_bufferlist_lock);

			bd = ogfs_alloc_bufdata();
			if (!bd)
				goto fail_brelse;

			memset(bd, 0, sizeof(ogfs_bufdata_t));

			bd->bd_bh = bh;
			bd->bd_gl = gl;

			INIT_LIST_HEAD(&bd->bd_new_list);
			INIT_LIST_HEAD(&bd->bd_incore_list);

			init_MUTEX(&bd->bd_lock);

			spin_lock(&ogfs_bufferlist_lock);

			ogfs_hold_glstruct(gl);
			list_add(&bd->bd_arch.bd_list,
				     &gl->gl_arch.gl_bufs);
			clear_bit(GLF_DEMOTEME, &gl->gl_flags);

			bh->b_private = bd;
			//bh->b_rm_fn = ogfs_remove_bh;

			if (!sdp->sd_args.ar_localcaching)
				clear_buffer_uptodate(bh);
		}

		spin_unlock(&ogfs_bufferlist_lock);
	}

	EXIT(GFN_DGETBLK_ARCH);

	return bh;

fail_brelse:
	brelse(bh);

fail:
	EXIT(GFN_DGETBLK_ARCH);
	return NULL;
}

#if 0

/**
 * ogfs_dread_arch - Read a block from disk
 * @sdp: The OGFS superblock
 * @blkno: The block number
 * @gl: The glock covering the block
 * @flags: flags to ogfs_dreread()
 *
 * Returns: The buffer on success, NULL on failure
 */

struct buffer_head *
ogfs_dread_arch(ogfs_sbd_t * sdp, uint64 blkno, ogfs_glock_t * gl, int flags)
{
	struct buffer_head *bh;

	ENTER(GFN_DREAD_ARCH);

	bh = ogfs_dgetblk_arch(sdp, blkno, gl);
	if (!bh) {
		printk("ogfs_degetblk_arch() failed\n");
		EXIT(GFN_DREAD_ARCH);
		return bh;
	}

	if (ogfs_dreread(sdp, bh, flags) != 0) {
		printk("ogfs_dreread() failed\n");
		brelse(bh);
		bh = NULL;
	}

	EXIT(GFN_DREAD_ARCH);

	return bh;
}
#endif /* 0 */

/**
 * ogfs_dread_nolock - Read a block from disk
 * @sdp: The OGFS superblock
 * @blkno: The block number
 * Removed for 2.6: @gl: The glock covering the block
 * @flags: flags to ogfs_dreread()
 *
 * Returns: The buffer on success, NULL on failure
 *
 * FIXME this is a hacked version of ogfs_dread_arch, it only gets called once at most
 */

#warning FIXME

struct buffer_head *
ogfs_dread_arch(ogfs_sbd_t * sdp, uint64 blkno, ogfs_glock_t * gl, int flags)
{
	struct buffer_head *bh;

	ENTER(GFN_DREAD_ARCH);

	bh = ogfs_dgetblk_arch(sdp, blkno, gl);
	if (!bh) {
		printk("ogfs_degetblk_arch() failed\n");
		EXIT(GFN_DREAD_ARCH);
		return bh;
	}

	if (ogfs_dreread(sdp, bh, flags) != 0) {
		printk("ogfs_dreread() failed\n");
		brelse(bh);
		bh = NULL;
	}

	EXIT(GFN_DREAD_ARCH);

	return bh;
}

/**
 * ogfs_prep_new_buffer - Mark a new buffer we just ogfs_dgetblk()ed uptodate
 * @bh: the buffer
 *
 */

void
ogfs_prep_new_buffer(struct buffer_head * bh)
{
	ENTER(GFN_PREP_NEW_BUFFER);

	lock_buffer(bh);

/* fixme	mark_buffer_clean(bh); */
	set_buffer_uptodate(bh);

	unlock_buffer(bh);

	EXIT(GFN_PREP_NEW_BUFFER);
}

/**
 * try_readahead - Do readahead on the device
 * @sdp: the filesystem
 * @bh: the buffer we're about to sleep on
 *
 */

static void
try_readahead(ogfs_sbd_t * sdp, struct buffer_head * bh)
{
	struct buffer_head *tmp_bh;
	//unsigned int major, minor;
	uint64 blk;
	/* calculating the readahead size is all broken due to block layer
		changes, so we'll hardcode the readahead size to 256k. ext3 uses
		128k, we can change this later if it turns out to be bad */
	uint64 size = (256 * 1024);
	unsigned int x;
	dev_t dev;

	ENTER(GFN_TRY_READAHEAD);

	/*  All buffers should be on the same device, right?
	   The device shouldn't shrink much while this function is running, right?  */

/* TODO at some point we should fix the readahead value calculation */
#if 0
	major = MAJOR(sdp->sd_vsb->s_dev);
	minor = MINOR(sdp->sd_vsb->s_dev);

	if (!blk_size[major]) {
		EXIT(GFN_TRY_READAHEAD);
		return;
	}

	if (sdp->sd_sb.sb_bsize_shift >= 10)
		size =
		    blk_size[major][minor] >> (sdp->sd_sb.sb_bsize_shift - 10);
	else
		size =
		    blk_size[major][minor] << (10 - sdp->sd_sb.sb_bsize_shift);
#endif

	for (x = 0; buffer_locked(bh) && x < sdp->sd_args.ar_buf_readahead; x++) {
		blk = BH_BLKNO(bh) + x + 1;

		if (blk >= size)
			break;

		dev = sdp->sd_vsb->s_dev;
		if (blk >= sdp->sd_jext_base) {
			if (translate_block(sdp, &dev, &blk))
				break;
		}
/* fix me		tmp_bh = getblk(dev, blk, sdp->sd_sb.sb_bsize); */
		if (!tmp_bh)
			break;

		if (!buffer_uptodate(tmp_bh) && !buffer_locked(tmp_bh))
			ll_rw_block(READA, 1, &tmp_bh);

		brelse(tmp_bh);
	}

	EXIT(GFN_TRY_READAHEAD);
}

/**
 * ogfs_dreread - Reread a block from disk
 * @sdp: the filesystem
 * @bh: The block to read
 * @flags: Flags that control the read
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_dreread(ogfs_sbd_t * sdp, struct buffer_head * bh, int flags)
{
	int error = 0;

	ENTER(GFN_DREREAD);

	if (flags & DR_FORCE) {
		error = clear_from_ail(sdp, bh);
		if (error)
			goto out;
		if (!sdp->sd_args.ar_localcaching)
			//mark_buffer_uptodate(bh, FALSE);
			clear_bit(BH_Uptodate, &bh->b_state);
	}

	if ((flags & DR_START) && !buffer_uptodate(bh) && !buffer_locked(bh))
		ll_rw_block(READ, 1, &bh);

	if (flags & DR_WAIT) {
		/*  Do readahead if it's convenient and allowed  */

		if (buffer_locked(bh) &&
		    sdp->sd_args.ar_localcaching &&
		    sdp->sd_args.ar_buf_readahead)
			try_readahead(sdp, bh);

		wait_on_buffer(bh);

		if (!buffer_uptodate(bh)) {
			con_printf("OGFS:  read error on block %" FMT_64 "u\n",
				   BH_BLKNO(bh));
			error = -EIO;

			OGFS_ASSERT(0,
				   con_printf
				   ("Panicking because of read error on block %"
				    FMT_64 "u\n", BH_BLKNO(bh)););
		}
	}

out:
	EXIT(GFN_DREREAD);

	return error;
}

/**
 * ogfs_dwrite - Write a buffer
 * @sdp: the filesystem
 * @bh: The buffer to write
 * @flags: The type of write operation to do
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_dwrite(ogfs_sbd_t * sdp, struct buffer_head * bh, int flags)
{
	ogfs_bufdata_t *bd;
	ogfs_glock_t *gl;
	int error = 0;

	ENTER(GFN_DWRITE);

	OGFS_ASSERT(buffer_uptodate(bh),);

#warning FIXME
	//if ((flags & DW_DIRTY) && !test_bit(BH_Temp, &bh->b_state)) {
	if (flags & DW_DIRTY) {
		bd = BH_BUFDATA(bh);
		if (bd) {
			gl = bd->bd_gl;
			OGFS_ASSERT(atomic_read(&gl->gl_locked) &&
				   (gl->gl_pid == ((long)(current->pid))
				    || gl->gl_pid == -1)
				   && gl->gl_state == LM_ST_EXCLUSIVE,);
			set_bit(GLF_DIRTY, &gl->gl_flags);
		}

		wait_on_buffer(bh);  /* FIXME: AUDIT this is a race */
		mark_buffer_dirty(bh);
	}

	if ((flags & DW_START) && buffer_dirty(bh) && !buffer_locked(bh))
		ll_rw_block(WRITE, 1, &bh);

	if (flags & DW_WAIT) {
		wait_on_buffer(bh);

		if (!buffer_uptodate(bh) || buffer_dirty(bh)) {
			con_printf("OGFS:  write error on block %" FMT_64 "u\n",
				   BH_BLKNO(bh));
			error = -EIO;

			OGFS_ASSERT(0,
				   con_printf
				   ("Panicking because of write error on block %"
				    FMT_64 "u\n", BH_BLKNO(bh)););
		}
	}

	EXIT(GFN_DWRITE);

	return error;
}

/**
 * ogfs_dforget_arch - Forget a data buffer
 * @bh: The buffer to release
 *
 */

void
ogfs_dforget_arch(struct buffer_head * bh)
{
	ENTER(GFN_DFORGET_ARCH);
	bforget(bh);
	EXIT(GFN_DFORGET_ARCH);
}

/**
 * ogfs_dpin - Pin a metadata buffer in memory
 * @bh: The buffer to be pinned
 *
 */

void
ogfs_dpin(struct buffer_head * bh)
{
	ogfs_bufdata_t *bd;

	ENTER(GFN_DPIN);

	OGFS_ASSERT(buffer_uptodate(bh),);

	bd = BH_BUFDATA(bh);
	OGFS_ASSERT(bd,);

	ogfs_lock_buffer(bh);

	if (!bd->bd_arch.bd_pinned) {
		wait_on_buffer(bh);  /* FIXME: AUDIT this is a race */

		/*  If this buffer is in the AIL and it has already been written,
		   remove it from the AIL.  */

		spin_lock(&ogfs_bufferlist_lock);
		if (bd->bd_arch.bd_ail_tr &&
		    !buffer_dirty(bh) && !buffer_locked(bh)) {
			bd->bd_arch.bd_ail_tr = NULL;
			list_del(&bd->bd_arch.bd_ail_list);
		}
		spin_unlock(&ogfs_bufferlist_lock);

		wait_on_buffer(bh);  /* FIXME: AUDIT this is a race */
/* fixme		mark_buffer_clean(bh); */
		atomic_inc(&bh->b_count);
	}

	bd->bd_arch.bd_pinned++;

	ogfs_unlock_buffer(bh);

	EXIT(GFN_DPIN);
}

/**
 * ogfs_dunpin - Unpin a buffer
 * @bh: The buffer to unpin
 * @tr: The transaction in the AIL that contains this buffer
 *
 */

void
ogfs_dunpin(struct buffer_head * bh, ogfs_trans_t * tr)
{
	ogfs_bufdata_t *bd;
	int dec = FALSE;

	ENTER(GFN_DUNPIN);

	bd = BH_BUFDATA(bh);
	OGFS_ASSERT(bd,);

	/*  Add the buffer to the AIL
	   and get rid of an old reference if there is one  */

	if (tr) {
		spin_lock(&ogfs_bufferlist_lock);
		if (bd->bd_arch.bd_ail_tr)
			list_del(&bd->bd_arch.bd_ail_list);
		bd->bd_arch.bd_ail_tr = tr;
		list_add(&bd->bd_arch.bd_ail_list, &tr->tr_arch.tr_bufs);
		spin_unlock(&ogfs_bufferlist_lock);
	}

	ogfs_lock_buffer(bh);

	OGFS_ASSERT(bd->bd_arch.bd_pinned,);
	bd->bd_arch.bd_pinned--;

	if (!bd->bd_arch.bd_pinned) {
		mark_buffer_dirty(bh);
		dec = TRUE;
	}

	ogfs_unlock_buffer(bh);

	if (dec)
		atomic_dec(&bh->b_count);

	EXIT(GFN_DUNPIN);
}

/**
 * ogfs_is_pinned - check to see if a buffer is pinned
 * @bh: The buffer
 *
 * Returns:  TRUE if the buffer is pinned, FALSE if it isn't.
 */

int
ogfs_is_pinned(struct buffer_head * bh)
{
	ogfs_bufdata_t *bd;
	int ret = FALSE;

	ENTER(GFN_IS_PINNED);

	bd = BH_BUFDATA(bh);
	if (bd) {
		ogfs_lock_buffer(bh);
		ret = bd->bd_arch.bd_pinned;
		ogfs_unlock_buffer(bh);
	}

	EXIT(GFN_IS_PINNED);

	return ret;
}

/**
 * ogfs_init_temp_bh - Initialize a temporary buffer head
 * @sdp: The OGFS superblock
 * @blkno: The block number
 * @buf: The data area
 * @write: TRUE if this buffer will be writen, FALSE if it will be read
 *
 * Returns:  the temp buffer head
 */

struct buffer_head *
ogfs_init_temp_bh(ogfs_sbd_t * sdp, uint64 blkno, char *buf, int write)
{
	struct buffer_head *bh;
	dev_t dev;

	ENTER(GFN_INIT_TEMP_BH);

	dev = sdp->sd_vsb->s_dev;  /* assume filesystem device */
	if (blkno >= sdp->sd_jext_base) {
		if (translate_block(sdp, &dev, &blkno)) {
			printk(KERN_ERR "translate_block failed\n");
			return NULL;
		}
	}

	/*  We don't need to worry about aliases  */

	//bh = get_unused_buffer_head(0);
	bh = alloc_buffer_head(0);

	if (!bh)
		goto out;

	init_buffer(bh, NULL, NULL);

	//bh->b_bdev = dev;
	bh->b_data = buf;
	bh->b_page = virt_to_page(buf);
	bh->b_this_page = NULL;
	//bh->b_prev_free = NULL;
	//bh->b_next_free = NULL;
	//bh->b_next = NULL;
	//bh->b_pprev = NULL;

	//bh->b_state = (1 << BH_Mapped) | (1 << BH_Temp);
	bh->b_state = (1 << BH_Mapped);
	if (write) {
		set_bit(BH_Dirty, &bh->b_state);
		set_bit(BH_Uptodate, &bh->b_state);
	}
	atomic_set(&bh->b_count, 1);
	bh->b_size = sdp->sd_sb.sb_bsize;
	bh->b_blocknr = blkno;

	//refile_buffer(bh);
	clear_buffer_dirty(bh);

out:
	EXIT(GFN_INIT_TEMP_BH);

	return bh;
}

/**
 * ogfs_dispose_temp_bh - Get rid of a temporary buffer_head
 * @bh: The buffer head we no longer want
 *
 */

void
ogfs_dispose_temp_bh(struct buffer_head * bh)
{
	ENTER(GFN_DISPOSE_TEMP_BH);

	OGFS_ASSERT(!buffer_dirty(bh) && !buffer_locked(bh),);
	//OGFS_ASSERT(test_bit(BH_Temp, &bh->b_state),);
	//OGFS_ASSERT(!bh->b_next,);
	while (atomic_read(&bh->b_count) != 1) {
		//ogfs_yield();
		schedule_timeout(1);
	}

	bforget(bh);

	EXIT(GFN_DISPOSE_TEMP_BH);
}

/**
 * ogfs_get_meta_buffer - Get a metadata buffer
 * @ip: The OGFS inode
 * @depth: The depth in the metadata tree
 * @num: The block number (device relative) of the buffer
 * @new: Non-zero if we may create a new buffer
 *
 * Returns: The metadata buffer or NULL
 */

struct buffer_head *
ogfs_get_meta_buffer(ogfs_inode_t * ip, int height, uint64 num, int new)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	struct buffer_head *bh = ip->i_cache[height];
	int flags = ((new) ? DR_FORCE : 0) | DR_START | DR_WAIT;

	ENTER(GFN_GET_META_BUFFER);

	if (bh && (BH_BLKNO(bh) == num)) {
		if (ogfs_dreread(sdp, bh, flags) != 0) {
			bh = NULL;
			goto out;
		}
		goto check_it;
	}

	ogfs_flush_meta_cache(ip, height);

	bh = ogfs_dread(sdp, num, ip->i_gl, flags);
	if (!bh)
		goto out;
	ip->i_cache[height] = bh;

check_it:
	if (new) {
		ogfs_trans_add_bh(sdp, bh);
		ogfs_metatype_set(bh, OGFS_METATYPE_IN, OGFS_FORMAT_IN);
		ogfs_buffer_clear_tail(bh, sizeof(ogfs_indirect_t));
	} else
		ogfs_metatype_check(bh,
				   (height) ? OGFS_METATYPE_IN :
				   OGFS_METATYPE_DI);

	atomic_inc(&bh->b_count);
	buf_alloc_main(bh, __FILE__, __LINE__);

out:
	EXIT(GFN_GET_META_BUFFER);

	return bh;
}

/**
 * ogfs_get_data_buffer - Get a data buffer
 * @ip: The OGFS inode
 * @num: The block number (device relative) of the data block
 * @new: Non-zero if this is a new allocation
 *
 * Returns: The data buffer or NULL
 */

struct buffer_head *
ogfs_get_data_buffer(ogfs_inode_t * ip, uint64 block, int new)
{
	struct buffer_head *bh;

	ENTER(GFN_GET_DATA_BUFFER);

	if (block == ip->i_num.no_addr) {
		OGFS_ASSERT(!new,);

		bh = ogfs_dread(ip->i_sbd, block, ip->i_gl, DR_START | DR_WAIT);
		if (bh)
			ogfs_metatype_check(bh, OGFS_METATYPE_DI);
	} else if (ogfs_is_jdata(ip)) {
		if (new) {
			bh = ogfs_dread(ip->i_sbd, block, ip->i_gl,
				       DR_FORCE | DR_START | DR_WAIT);
			if (bh) {
				ogfs_trans_add_bh(ip->i_sbd, bh);
				ogfs_metatype_set(bh, OGFS_METATYPE_JD,
						 OGFS_FORMAT_JD);
				ogfs_buffer_clear_tail(bh, sizeof(ogfs_jdata_t));
			}
		} else {
			bh = ogfs_dread(ip->i_sbd, block, ip->i_gl,
				       DR_START | DR_WAIT);
			if (bh)
				ogfs_metatype_check(bh, OGFS_METATYPE_JD);
		}
	} else {
		if (new) {
			bh = ogfs_dgetblk(ip->i_sbd, block, ip->i_gl);
			if (bh)
				ogfs_prep_new_buffer(bh);
		} else
			bh = ogfs_dread(ip->i_sbd, block, ip->i_gl,
				       DR_START | DR_WAIT);
	}

	EXIT(GFN_GET_DATA_BUFFER);

	return bh;
}
