/*
 *
 *    Copyright 1995-1999 Regents of the University of Minnesota
 *    Portions Copyright 1999-2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 */

/*
 * This module defines all OGFS routines that have to do with directories.
 *
 *      Implements Extendible Hashing as described in:
 *        "Extendible Hashing" by Fagin, et al in
 *           __ACM Trans. on Database Systems__, Sept 1979.
 */

#include <linux/dcache.h>

#include "ogfs_debug.h"
#include "ogfs.h"
#include "blklist.h"
#include "bmap.h"
#include "dir.h"
#include "dio.h"
#include "file.h"
#include "fs.h"
#include "glock.h"
#include "inode.h"
#include "log.h"
#include "trans.h"

#define IS_LEAF     (1)
#define IS_DINODE   (2)

/*
 * Here's the layout of dirents which is essentially the same as that of
 * ext2 within a single block.
 * The field de_name_len is the number of bytes actually required for the
 * name (no null terminator). The field de_rec_len is the number of bytes
 * allocated to the dirent. The offset of the next dirent in the block is
 * (dirent + dirent->de_rec_len).
 * When a dirent is deleted, the preceding dirent inherits its allocated
 * space, ie prev->de_rec_len += deleted->de_rec_len. Since the next dirent
 * is obtained by adding de_rec_len to the current dirent, this essentially
 * causes the deleted dirent to get jumped over when iterating through all
 * the dirents.  When deleting the first dirent in a block, there is no
 * previous dirent so the field de_ino is set to zero to designate it as
 * deleted. When allocating a dirent, ogfs_dirent_alloc iterates through the
 * dirents in a block. If the first dirent has (de_ino == 0) and de_rec_len
 * is large enough, this first dirent is allocated. Otherwise it must go
 * through all the 'used' dirents searching for one in which the amount of
 * total space minus the amount of used space will provide enough space for
 * the new dirent.
 * There are two types of blocks in which dirents reside. In a stuffed dinode,
 * the dirents begin at offset sizeof(ogfs_dinode_t) from the beginning of the
 * block.  In leaves, they begin at offset sizeof (ogfs_leaf_t) from the
 * beginning of the leaf block. The dirents reside in leaves when
 *
 *	(dip->i_di.di_flags & OGFS_DIF_EXHASH) != 0
 *
 * The dirents are in the stuffed dinode when
 *
 *	(dip->i_di.di_flags & OGFS_DIF_EXHASH) == 0.
 *
 * When the dirents are in leaves, the actual contents of the directory
 * file are used as an array of 64-bit block pointers pointing to the leaf
 * blocks. The dirents are NOT in the directory file itself. There can be
 * more than one block pointer in the array that points to the same leaf.
 * In fact, when a directory is first converted from linear to exhash,
 * all of the pointers point to the same leaf. When a leaf is completely
 * full, the size of the hash table can be doubled unless it is already at
 * the maximum size which is hard coded into OGFS_DIR_MAX_DEPTH.
 * After that, leaves are chained together in a linked list but never before
 * the maximum hash table size has been reached.
 */

/**
 * int ogfs_filecmp - Compare two filenames
 * @file1: The first filename
 * @file2: The second filename
 * @len_of_file2: The length of the second file
 *
 * This routine compares two filenames and returns TRUE if they are equal.
 *
 * Returns: TRUE (!=0) if the files are the same, otherwise FALSE (0).
 */

int
ogfs_filecmp(struct qstr * file1, char *file2, loff_t len_of_file2)
{
	ENTER(GFN_FILECMP);

	if (file1->len != len_of_file2) {
		EXIT(GFN_FILECMP);
		return FALSE;
	}

	if (memcmp(file1->name, file2, file1->len)) {
		EXIT(GFN_FILECMP);
		return FALSE;
	}

	EXIT(GFN_FILECMP);

	return TRUE;
}

/**
 * dir_hash - Compute hash on a filename
 * @filename: The name to compute the hash upon
 *
 * Return: The hash value
 */

static __inline__ uint32
dir_hash(struct qstr * filename)
{
	return ogfs_hash(filename->name, filename->len);
}

/**
 * ogfs_dirent_first - Return the first dirent
 * @bh: The buffer
 * @dent: Pointer to list of dirents
 *
 * return first dirent whether bh points to leaf or stuffed dinode
 *
 * Returns: IS_LEAF or IS_DINODE
 */

static int
ogfs_dirent_first(struct buffer_head * bh, ogfs_dirent_t ** dent)
{
	ogfs_leaf_t *leaf;
	ogfs_dinode_t *dinode;

	ENTER(GFN_DIRENT_FIRST);

	leaf = (ogfs_leaf_t *) BH_DATA(bh);

	if (ogfs32_to_cpu(leaf->lf_header.mh_type) == OGFS_METATYPE_LF) {
		*dent = (ogfs_dirent_t *) (BH_DATA(bh) + sizeof(ogfs_leaf_t));

		EXIT(GFN_DIRENT_FIRST);
		return IS_LEAF;
	} else {
		dinode = (ogfs_dinode_t *) BH_DATA(bh);
		OGFS_ASSERT(ogfs32_to_cpu(dinode->di_header.mh_type) ==
			   OGFS_METATYPE_DI,);

		*dent = (ogfs_dirent_t *) (BH_DATA(bh) + sizeof(ogfs_dinode_t));

		EXIT(GFN_DIRENT_FIRST);
		return IS_DINODE;
	}
}

/**
 * ogfs_dirent_next - Next dirent
 * @bh: The buffer
 * @dent: Pointer to list of dirents
 *
 * Returns: 0 on success, error code otherwise
 */

static int
ogfs_dirent_next(struct buffer_head * bh, ogfs_dirent_t ** dent)
{
	ogfs_dirent_t *tmp, *cur;
	char *bh_end;
	uint32 cur_rec_len;

	ENTER(GFN_DIRENT_NEXT);

	cur = *dent;
	bh_end = BH_DATA(bh) + BH_SIZE(bh);

	cur_rec_len = ogfs16_to_cpu(cur->de_rec_len);

	if ((char *) cur + cur_rec_len >= bh_end) {
		OGFS_ASSERT((char *) cur + cur_rec_len == bh_end,);

		EXIT(GFN_DIRENT_NEXT);
		return -ENOENT;
	}

	tmp = (ogfs_dirent_t *) ((char *) cur + cur_rec_len);

	OGFS_ASSERT((char *) tmp + ogfs16_to_cpu(tmp->de_rec_len) <= bh_end,);
	/* only the first dent could ever have de_ino == 0 */
	OGFS_ASSERT(tmp->de_inum.no_formal_ino,);

	*dent = tmp;

	EXIT(GFN_DIRENT_NEXT);

	return 0;
}

/**
 * ogfs_dirent_del - Delete a dirent
 * @dip: The OGFS inode
 * @bh: The buffer
 * @prev: The previous dirent
 * @cur: The current dirent
 *
 * Returns: 0 on success, error code otherwise
 */

static int
ogfs_dirent_del(ogfs_inode_t * dip, struct buffer_head * bh, ogfs_dirent_t * prev,
	       ogfs_dirent_t * cur)
{
	uint32 cur_rec_len, prev_rec_len;

	ENTER(GFN_DIRENT_DEL);

	OGFS_ASSERT(cur->de_inum.no_formal_ino,);

	ogfs_trans_add_bh(dip->i_sbd, bh);

	/*  If there is no prev entry, this is the first entry in the block.
	   The de_rec_len is already as big as it needs to be.  Just zero
	   out the inode number and return.  */

	if (!prev) {
		cur->de_inum.no_formal_ino = 0;	/*  No endianess worries  */
		EXIT(GFN_DIRENT_DEL);
		return 0;
	}

	/*  Combine this dentry with the previous one.  */

	prev_rec_len = ogfs16_to_cpu(prev->de_rec_len);
	cur_rec_len = ogfs16_to_cpu(cur->de_rec_len);

	OGFS_ASSERT((char *) prev + prev_rec_len == (char *) cur,);
	OGFS_ASSERT((char *) cur + cur_rec_len <= BH_DATA(bh) + BH_SIZE(bh),);

	prev_rec_len += cur_rec_len;
	prev->de_rec_len = cpu_to_ogfs16(prev_rec_len);

	EXIT(GFN_DIRENT_DEL);

	return 0;
}

/**
 * ogfs_dirent_alloc - Allocate a directory entry
 * @dip: The OGFS inode
 * @bh: The buffer
 * @name_len: The length of the name
 * @dent_out: Pointer to list of dirents
 *
 * Returns: 0 on success, error code otherwise
 */

int
ogfs_dirent_alloc(ogfs_inode_t * dip, struct buffer_head * bh, int name_len,
		 ogfs_dirent_t ** dent_out)
{
	ogfs_sbd_t *sdp = dip->i_sbd;
	ogfs_dirent_t *dent, *new;
	ogfs_leaf_t *leaf = NULL;
	ogfs_dinode_t *dinode;
	unsigned int rec_len;
	int type;
	int entries = 0, offset = 0, x = 0;

	ENTER(GFN_DIRENT_ALLOC);

	type = ogfs_dirent_first(bh, &dent);

	if (type == IS_LEAF) {
		leaf = (ogfs_leaf_t *) BH_DATA(bh);
		entries = ogfs16_to_cpu(leaf->lf_entries);
		offset = sizeof(ogfs_leaf_t);
	} else if (type == IS_DINODE) {
		dinode = (ogfs_dinode_t *) BH_DATA(bh);
		entries = ogfs32_to_cpu(dinode->di_entries);
		offset = sizeof(ogfs_dinode_t);
	} else {
		OGFS_ASSERT(0,);
	}

	rec_len = OGFS_DIR_REC_LEN(name_len);

	if (!entries) {
		ogfs_trans_add_bh(sdp, bh);

		dent->de_rec_len = BH_SIZE(bh) - offset;
		dent->de_rec_len = cpu_to_ogfs16(dent->de_rec_len);
		dent->de_name_len = cpu_to_ogfs16(name_len);

		*dent_out = dent;
		goto success;
	}

	do {
		uint32 cur_rec_len, cur_name_len;

		cur_rec_len = ogfs16_to_cpu(dent->de_rec_len);
		cur_name_len = ogfs16_to_cpu(dent->de_name_len);

		if ((!dent->de_inum.no_formal_ino && cur_rec_len >= rec_len) ||
		    (cur_rec_len >= OGFS_DIR_REC_LEN(cur_name_len) + rec_len)) {
			ogfs_trans_add_bh(sdp, bh);

			if (dent->de_inum.no_formal_ino) {
				new =
				    (ogfs_dirent_t *) ((char *) dent +
						      OGFS_DIR_REC_LEN
						      (cur_name_len));
				new->de_rec_len =
				    cpu_to_ogfs16(cur_rec_len -
						 OGFS_DIR_REC_LEN(cur_name_len));
				new->de_name_len = cpu_to_ogfs16(name_len);

				dent->de_rec_len =
				    cur_rec_len - ogfs16_to_cpu(new->de_rec_len);
				dent->de_rec_len =
				    cpu_to_ogfs16(dent->de_rec_len);

				*dent_out = new;
				goto success;
			}

			dent->de_name_len = cpu_to_ogfs16(name_len);

			*dent_out = dent;
			goto success;
		}

		OGFS_ASSERT(x < entries,);

		if (dent->de_inum.no_formal_ino)
			x++;
	} while (ogfs_dirent_next(bh, &dent) == 0);

	EXIT(GFN_DIRENT_ALLOC);

	return -ENOSPC;

success:
	EXIT(GFN_DIRENT_ALLOC);
	return 0;
}

/**
 * leaf_search
 * @bh:
 * @filename:
 * @dent_out:
 * @dent_prev:
 *
 * Returns:
 */

static int
leaf_search(struct buffer_head * bh, struct qstr * filename,
	    ogfs_dirent_t ** dent_out, ogfs_dirent_t ** dent_prev)
{
	uint32 hash;
	ogfs_dirent_t *dent, *prev = NULL;
	int type, entries = 0, x = 0;

	ENTER(GFN_LEAF_SEARCH);

	type = ogfs_dirent_first(bh, &dent);

	if (type == IS_LEAF) {
		ogfs_leaf_t *leaf = (ogfs_leaf_t *) BH_DATA(bh);
		entries = ogfs16_to_cpu(leaf->lf_entries);
	} else if (type == IS_DINODE) {
		ogfs_dinode_t *dinode = (ogfs_dinode_t *) (BH_DATA(bh));
		entries = ogfs32_to_cpu(dinode->di_entries);
	} else {
		OGFS_ASSERT(0,);
	}

	hash = dir_hash(filename);

	do {
		if (!dent->de_inum.no_formal_ino) {
			prev = dent;
			continue;
		}

		if (ogfs32_to_cpu(dent->de_hash) == hash &&
		    ogfs_filecmp(filename, (char *) (dent + 1),
				ogfs16_to_cpu(dent->de_name_len))) {
			*dent_out = dent;
			if (dent_prev)
				*dent_prev = prev;

			EXIT(GFN_LEAF_SEARCH);
			return 0;
		}

		OGFS_ASSERT(x < entries,);
		x++;
		prev = dent;
	} while (ogfs_dirent_next(bh, &dent) == 0);

	EXIT(GFN_LEAF_SEARCH);

	return -ENOENT;
}

/**
 * get_leaf - Get leaf
 * @dip:
 * @leaf_no:
 * @bh_out:
 *
 * Returns: 0 on success, error code otherwise
 */

static int
get_leaf(ogfs_inode_t * dip, uint64 leaf_no, struct buffer_head ** bh_out)
{
	ogfs_sbd_t *sdp = dip->i_sbd;
	struct buffer_head *bh;

	ENTER(GFN_GET_LEAF);

	bh = ogfs_dread(sdp, leaf_no, dip->i_gl, DR_START | DR_WAIT);
	if (!bh) {
		EXIT(GFN_GET_LEAF);
		return -EIO;
	}
	ogfs_metatype_check(bh, OGFS_METATYPE_LF);

	*bh_out = bh;

	EXIT(GFN_GET_LEAF);

	return 0;
}

/**
 * get_leaf_nr - Get a leaf number associated with the index
 * @dip: The OGFS inode
 * @index:
 * @leaf_out:
 *
 * Returns: 0 on success, error code otherwise
 */

static int
get_leaf_nr(ogfs_inode_t * dip, uint32 index, uint64 * leaf_out)
{
	uint64 leaf_no;
	int64 error;

	ENTER(GFN_GET_LEAF_NR);

	error =
	    ogfs_internal_read(dip, (char *) &leaf_no, index * sizeof(uint64),
			      sizeof(uint64));
	if (error != sizeof(uint64)) {
		EXIT(GFN_GET_LEAF_NR);
		return (error < 0) ? error : -EIO;
	}

	*leaf_out = ogfs64_to_cpu(leaf_no);

	EXIT(GFN_GET_LEAF_NR);

	return 0;
}

/**
 * get_first_leaf - Get first leaf
 * @dip: The OGFS inode
 * @index:
 * @bh_out:
 *
 * Returns: 0 on success, error code otherwise
 */

static int
get_first_leaf(ogfs_inode_t * dip, uint32 index, struct buffer_head ** bh_out)
{
	uint64 leaf_no;
	int64 error;

	ENTER(GFN_GET_FIRST_LEAF);

	error = get_leaf_nr(dip, index, &leaf_no);
	if (!error)
		error = get_leaf(dip, leaf_no, bh_out);

	EXIT(GFN_GET_FIRST_LEAF);

	return error;
}

/**
 * get_next_leaf - Get next leaf
 * @dip: The OGFS inode
 * @bh_in: The buffer
 * @bh_out:
 *
 * Returns: 0 on success, error code otherwise
 */

static int
get_next_leaf(ogfs_inode_t * dip, struct buffer_head * bh_in, struct buffer_head ** bh_out)
{
	ogfs_leaf_t *leaf;
	int error;

	ENTER(GFN_GET_NEXT_LEAF);

	leaf = (ogfs_leaf_t *) BH_DATA(bh_in);

	if (!leaf->lf_next)
		error = -ENOENT;
	else
		error = get_leaf(dip, ogfs64_to_cpu(leaf->lf_next), bh_out);

	EXIT(GFN_GET_NEXT_LEAF);

	return error;
}

/**
 * linked_leaf_search - Linked leaf search
 * @dip: The OGFS inode
 * @filename: The filename to search for
 * @dent_out:
 * @dent_prev:
 * @bh_out:
 *
 * Returns: 0 on sucess, error code otherwise
 */

static int
linked_leaf_search(ogfs_inode_t * dip, struct qstr * filename,
		   ogfs_dirent_t ** dent_out, ogfs_dirent_t ** dent_prev,
		   struct buffer_head ** bh_out)
{
	struct buffer_head *bh = NULL, *bh_next;
	uint32 hsize, index;
	uint32 hash;
	int64 error;

	ENTER(GFN_LINKED_LEAF_SEARCH);

	hsize = 1 << dip->i_di.di_depth;
	OGFS_ASSERT(hsize * sizeof(uint64) == dip->i_di.di_size,);

	/*  Figure out the address of the leaf node.  */

	hash = dir_hash(filename);
	index = hash >> (32 - dip->i_di.di_depth);
	OGFS_ASSERT(index < hsize,);

	error = get_first_leaf(dip, index, &bh_next);
	if (error) {
		EXIT(GFN_LINKED_LEAF_SEARCH);
		return error;
	}

	/*  Find the entry  */

	do {
		if (bh)
			brelse(bh);

		bh = bh_next;

		error = leaf_search(bh, filename, dent_out, dent_prev);
		if (error) {
			if (error != -ENOENT) {
				brelse(bh);
				EXIT(GFN_LINKED_LEAF_SEARCH);
				return error;
			}
		} else {
			*bh_out = bh;
			EXIT(GFN_LINKED_LEAF_SEARCH);
			return 0;
		}

		error = get_next_leaf(dip, bh, &bh_next);
	} while (!error);

	brelse(bh);

	EXIT(GFN_LINKED_LEAF_SEARCH);

	return error;
}

/**
 * dir_make_exhash - Convet a stuffed directory into an ExHash directory
 * @dip: The OGFS inode
 *
 * Returns: 0 on success, error code otherwise
 */

static int
dir_make_exhash(ogfs_inode_t * dip)
{
	ogfs_sbd_t *sdp = dip->i_sbd;
	ogfs_dirent_t *dent;
	struct buffer_head *bh, *dibh;
	ogfs_leaf_t *leaf;
	int y;
	uint32 x;
	uint64 *lp, bn;
	int error;

	ENTER(GFN_DIR_MAKE_EXHASH);

	/*  Sanity checks  */

	OGFS_ASSERT(sizeof(ogfs_leaf_t) <= sizeof(ogfs_dinode_t),);
	OGFS_ASSERT(!(dip->i_di.di_flags & OGFS_DIF_EXHASH),);
	OGFS_ASSERT(ogfs_is_stuffed(dip),);

	dibh = ogfs_get_inode_buffer(dip);
	if (!dibh) {
		error = -EIO;
		goto fail;
	}

	/*  Allocate a new block for the first leaf node  */

	error = ogfs_metaalloc(dip, &bn);
	if (error)
		goto fail_drelse;

	/*  Turn over a new leaf  */

	bh = ogfs_dread(sdp, bn, dip->i_gl, DR_FORCE | DR_START | DR_WAIT);
	if (!bh) {
		error = -EIO;
		goto fail_drelse;
	}

	ogfs_trans_add_bh(sdp, bh);
	ogfs_metatype_set(bh, OGFS_METATYPE_LF, OGFS_FORMAT_LF);
	ogfs_buffer_clear_tail(bh, sizeof(ogfs_meta_header_t));

	/*  Fill in the leaf structure  */

	leaf = (ogfs_leaf_t *) BH_DATA(bh);

	OGFS_ASSERT(dip->i_di.di_entries < (1 << 16),);

	leaf->lf_dirent_format = cpu_to_ogfs32(OGFS_FORMAT_DE);
	leaf->lf_entries = cpu_to_ogfs16(dip->i_di.di_entries);

	/*  Copy dirents  */

	ogfs_buffer_copy_tail(bh, sizeof(ogfs_leaf_t), dibh,
			     sizeof(ogfs_dinode_t));

	/*  Find last entry  */

	x = 0;
	ogfs_dirent_first(bh, &dent);

	do {
		if (!dent->de_inum.no_formal_ino)
			continue;

		if (++x == dip->i_di.di_entries)
			break;
	} while (ogfs_dirent_next(bh, &dent) == 0);

	/*  Adjust the last dirent's record length
	   (Remember that dent still points to the last entry.)  */

	dent->de_rec_len =
	    ogfs16_to_cpu(dent->de_rec_len) + sizeof(ogfs_dinode_t) -
	    sizeof(ogfs_leaf_t);
	dent->de_rec_len = cpu_to_ogfs16(dent->de_rec_len);

	brelse(bh);

	/*  We're done with the new leaf block, now setup the new
	   hash table.  */

	ogfs_trans_add_bh(sdp, dibh);
	ogfs_buffer_clear_tail(dibh, sizeof(ogfs_dinode_t));

	lp = (uint64 *) (BH_DATA(dibh) + sizeof(ogfs_dinode_t));

	for (x = sdp->sd_hash_ptrs; x--; lp++)
		*lp = cpu_to_ogfs64(bn);

	dip->i_di.di_size = sdp->sd_sb.sb_bsize / 2;
	dip->i_di.di_blocks++;
	dip->i_di.di_flags |= OGFS_DIF_EXHASH;
	dip->i_di.di_payload_format = 0;

	for (x = sdp->sd_hash_ptrs, y = -1; x; x >>= 1, y++) ;
	dip->i_di.di_depth = y;

	ogfs_dinode_out(&dip->i_di, BH_DATA(dibh));

	brelse(dibh);

	EXIT(GFN_DIR_MAKE_EXHASH);

	return 0;

fail_drelse:
	brelse(dibh);

fail:
	EXIT(GFN_DIR_MAKE_EXHASH);
	return error;
}

/**
 * dir_split_leaf - Split a leaf block into two
 * @dip: The OGFS inode
 * @index:
 * @leaf_no:
 *
 * Returns: 0 on success, error code on failure
 */

static int
dir_split_leaf(ogfs_inode_t * dip, uint32 index, uint64 leaf_no)
{
	ogfs_sbd_t *sdp = dip->i_sbd;
	struct buffer_head *nbh, *obh, *dibh;
	ogfs_leaf_t *nleaf, *oleaf;
	ogfs_dirent_t *dent, *prev = NULL, *next = NULL, *new;
	uint32 hsize, start, len, half_len, divider;
	uint64 bn, *lp;
	uint32 name_len;
	int moved = FALSE;
	unsigned int x;
	int64 error;

	ENTER(GFN_DIR_SPLIT_LEAF);

	/*  Sanity Checks  */

	OGFS_ASSERT(dip->i_di.di_flags & OGFS_DIF_EXHASH,);

	hsize = 1 << dip->i_di.di_depth;
	OGFS_ASSERT(hsize * sizeof(uint64) == dip->i_di.di_size,);
	OGFS_ASSERT(index < hsize,);

	/*  Allocate the new leaf block  */

	error = ogfs_metaalloc(dip, &bn);
	if (error)
		goto fail;

	/*  Get the new leaf block  */

	nbh = ogfs_dread(sdp, bn, dip->i_gl, DR_FORCE | DR_START | DR_WAIT);
	if (!nbh) {
		error = -EIO;
		goto fail;
	}

	ogfs_trans_add_bh(sdp, nbh);
	ogfs_metatype_set(nbh, OGFS_METATYPE_LF, OGFS_FORMAT_LF);
	ogfs_buffer_clear_tail(nbh, sizeof(ogfs_meta_header_t));

	nleaf = (ogfs_leaf_t *) BH_DATA(nbh);

	nleaf->lf_dirent_format = cpu_to_ogfs32(OGFS_FORMAT_DE);

	/*  Get the old leaf block  */

	error = get_leaf(dip, leaf_no, &obh);
	if (error)
		goto fail_nrelse;

	ogfs_trans_add_bh(sdp, obh);

	oleaf = (ogfs_leaf_t *) BH_DATA(obh);

	/*  Compute the start and len of leaf pointers in the hash table.  */

	len = 1 << (dip->i_di.di_depth - ogfs16_to_cpu(oleaf->lf_depth));
	OGFS_ASSERT(len != 1,);
	half_len = len >> 1;

	start = (index & ~(len - 1));

	/*  Change the pointers.
	   Don't bother distinguishing stuffed from non-stuffed.
	   This code is complicated enough already.  */

	lp = (uint64 *) gmalloc(half_len * sizeof(uint64));
	if (!lp) {
		error = -ENOMEM;
		goto fail_orelse;
	}

	error =
	    ogfs_internal_read(dip, (char *) lp, start * sizeof(uint64),
			      half_len * sizeof(uint64));
	if (error != half_len * sizeof(uint64)) {
		if (error >= 0)
			error = -EIO;
		goto fail_lpfree;
	}

	/*  Change the pointers  */

	for (x = 0; x < half_len; x++)
		lp[x] = cpu_to_ogfs64(bn);

	error =
	    ogfs_internal_write(dip, (char *) lp, start * sizeof(uint64),
			       half_len * sizeof(uint64));
	if (error != half_len * sizeof(uint64)) {
		if (error >= 0)
			error = -EIO;
		goto fail_lpfree;
	}

	gfree(lp, half_len * sizeof(uint64));

	/*  Compute the divider  */

	divider = (start + half_len) << (32 - dip->i_di.di_depth);

	IFDB(GFN_DIR_SPLIT_LEAF, db_printf("  Divider = 0x%.8X\n", divider);
	    );

	/*  Copy the entries  */

	ogfs_dirent_first(obh, &dent);

	do {
		next = dent;
		if (ogfs_dirent_next(obh, &next))
			next = NULL;

		if (dent->de_inum.no_formal_ino &&
		    (ogfs32_to_cpu(dent->de_hash) < divider)) {
			name_len = ogfs16_to_cpu(dent->de_name_len);

			error = ogfs_dirent_alloc(dip, nbh, name_len, &new);
			OGFS_ASSERT(!error,);

			/*
			 * No endianness worries.
			 */
			new->de_inum = dent->de_inum;
			new->de_hash = dent->de_hash;
			new->de_type = dent->de_type;
			memcpy((char *) (new + 1), (char *) (dent + 1),
				   name_len);

			nleaf->lf_entries = ogfs16_to_cpu(nleaf->lf_entries) + 1;
			nleaf->lf_entries = cpu_to_ogfs16(nleaf->lf_entries);

			ogfs_dirent_del(dip, obh, prev, dent);

			OGFS_ASSERT(ogfs16_to_cpu(oleaf->lf_entries),);
			oleaf->lf_entries = ogfs16_to_cpu(oleaf->lf_entries) - 1;
			oleaf->lf_entries = cpu_to_ogfs16(oleaf->lf_entries);

			if (!prev)
				prev = dent;

			moved = TRUE;
		} else
			prev = dent;

		dent = next;
	} while (dent);

	/*  If none of the entries got moved into the new leaf,
	   artificially fill in the first entry.  */

	if (!moved) {
		error = ogfs_dirent_alloc(dip, nbh, 0, &new);
		OGFS_ASSERT(!error,);

		new->de_inum.no_formal_ino = 0;
	}

	oleaf->lf_depth = ogfs16_to_cpu(oleaf->lf_depth) + 1;
	oleaf->lf_depth = cpu_to_ogfs16(oleaf->lf_depth);
	nleaf->lf_depth = oleaf->lf_depth;

	dibh = ogfs_get_inode_buffer(dip);
	OGFS_ASSERT(dibh,);	/*  Pinned in ogfs_internal_write()  */

	dip->i_di.di_blocks++;

	ogfs_trans_add_bh(sdp, dibh);
	ogfs_dinode_out(&dip->i_di, BH_DATA(dibh));
	brelse(dibh);

	brelse(obh);
	brelse(nbh);

	EXIT(GFN_DIR_SPLIT_LEAF);

	return 0;

fail_lpfree:
	gfree(lp, half_len * sizeof(uint64));

fail_orelse:
	brelse(obh);

fail_nrelse:
	brelse(nbh);

fail:
	EXIT(GFN_DIR_SPLIT_LEAF);
	return error;
}

/**
 * dir_double_exhash - Double size of ExHash table
 * @dip: The OGFS dinode
 *
 * Returns: 0 on success, error code on failure
 */

static int
dir_double_exhash(ogfs_inode_t * dip)
{
	ogfs_sbd_t *sdp = dip->i_sbd;
	struct buffer_head *dibh;

	uint32 hsize;
	uint64 *buf;
	uint64 *from, *to;
	uint64 block;

	int x;
	int64 error = 0;

	ENTER(GFN_DIR_DOUBLE_EXHASH);

	/*  Sanity Checks  */

	OGFS_ASSERT(dip->i_di.di_flags & OGFS_DIF_EXHASH,);
	hsize = 1 << dip->i_di.di_depth;
	OGFS_ASSERT(hsize * sizeof(uint64) == dip->i_di.di_size,);

	/*  Allocate both the "from" and "to" buffers in one big chunk  */

	buf = (uint64 *) gmalloc(3 * sdp->sd_hash_bsize);
	if (!buf) {
		error = -ENOMEM;
		goto out;
	}

	for (block = dip->i_di.di_size / sdp->sd_hash_bsize; block--;) {
		error =
		    ogfs_internal_read(dip, (char *) buf,
				      block * sdp->sd_hash_bsize,
				      sdp->sd_hash_bsize);
		if (error != sdp->sd_hash_bsize) {
			if (error >= 0)
				error = -EIO;
			goto out_gfree;
		}

		from = buf;
		to = (uint64 *) ((char *) buf + sdp->sd_hash_bsize);

		for (x = sdp->sd_hash_ptrs; x--; from++) {
			*to++ = *from;	/*  No endianess worries  */
			*to++ = *from;
		}

		error =
		    ogfs_internal_write(dip, (char *) buf + sdp->sd_hash_bsize,
				       block * sdp->sd_sb.sb_bsize,
				       sdp->sd_sb.sb_bsize);
		if (error != sdp->sd_sb.sb_bsize) {
			if (error >= 0)
				error = -EIO;
			goto out_gfree;
		}
	}

	gfree(buf, 3 * sdp->sd_hash_bsize);

	dibh = ogfs_get_inode_buffer(dip);
	OGFS_ASSERT(dibh,);	/*  Pinned in ogfs_internal_write()  */

	dip->i_di.di_depth++;

	ogfs_trans_add_bh(sdp, dibh);
	ogfs_dinode_out(&dip->i_di, BH_DATA(dibh));
	brelse(dibh);

	EXIT(GFN_DIR_DOUBLE_EXHASH);

	return 0;

out_gfree:
	gfree(buf, 3 * sdp->sd_hash_bsize);

out:
	EXIT(GFN_DIR_DOUBLE_EXHASH);
	return error;
}

/**
 * compare_dents - compare directory entries by hash value
 * @dent1: first dent
 * @dent2: second dent
 *
 * When comparing the hash entries of dent1 to dent2:
 *   gt: returns 1
 *   lt: returns -1
 *   eq: returns 0
 */

static int
compare_dents(const void *dent1, const void *dent2)
{
	ogfs_dirent_t **foo1 = (ogfs_dirent_t **) dent1;
	ogfs_dirent_t **foo2 = (ogfs_dirent_t **) dent2;

	int ret = 0;

	ENTER(GFN_COMPARE_DENTS);
	if (ogfs32_to_cpu((*foo1)->de_hash) > ogfs32_to_cpu((*foo2)->de_hash)) {
		ret = 1;
	} else if (ogfs32_to_cpu((*foo1)->de_hash) <
		   ogfs32_to_cpu((*foo2)->de_hash)) {
		ret = -1;
	}

	EXIT(GFN_COMPARE_DENTS);
	return ret;
}

/**
 * find_dent_index - Finds a dent in a array of sorted dents
 * @sorted_dents: pre-sorted array of dents
 * @fpos: hash of the dent we're looking for
 * @dent_no: contains the total number of entries stored in the array when it
 * comes into the function, variable to store the index of the dent in on exit
 *
 * Finds a dent in the pre-sorted array based on the hash value passed in.
 * Stores the index in the index pointer.  Currently is a simple linear
 * search.  This should maybe be converted to at least a binary search,
 * but I don't know how much of a difference it would really make.
 * I guess it might make a big difference in *really* large directories
 * that have linked leaves.
 *
 */

static int
find_dent_index(ogfs_dirent_t ** sorted_dents, int64 * f_pos, uint16 * dent_no)
{
	uint16 cur_index;

	ENTER(GFN_FIND_DENT_INDEX);

	for (cur_index = 0; cur_index < *dent_no; cur_index++) {
		if ((ogfs32_to_cpu((sorted_dents[cur_index])->de_hash) >> 1) >=
		    (uint32) * f_pos) {
			*dent_no = cur_index;
			EXIT(GFN_FIND_DENT_INDEX);
			return 0;
		}
	}

	if (*f_pos == 0) {
		*dent_no = 0;
		EXIT(GFN_FIND_DENT_INDEX);
		return 0;
	}

	EXIT(GFN_FIND_DENT_INDEX);

	return -ENOENT;
}

/* CLEAN_UP_BH - helper macro for do_filldir
 * @type - Leaf or stuffed dinode?
 * @orig_leaf_no - The leaf number to start cleaning at
 * @dip - the dinode pointer we're working with
 *
 * Since we can't release the buffer heads til we're done with the sorted
 * sorted list of dentries, but we need to release them at every exit point
 * in do_filldir, I created this macro to do so.  It goes through all the
 * linked leaves and releases the buffer heads twice - once for the current
 * loop, and once for the loop where the dentries were stuffed into the
 * sorted list in the first place
 */

#define CLEAN_UP_BH(type, orig_leaf_no, dip)                                 \
  if (type == IS_LEAF)                                                       \
  {                                                                          \
    /* grab the leaf no that was passed into the fxn; */                     \
    /* this is the head leaf */                                              \
    uint64 lf_no = orig_leaf_no;                                             \
    do                                                                       \
    {                                                                        \
      int err;                                                               \
      ogfs_leaf_t *lf = NULL;                                                 \
      struct buffer_head *lfbh = NULL;                                                \
      err = get_leaf(dip, lf_no, &lfbh);                                     \
      /* if we get an error, it means we found an invalid leaf */            \
      if (err)                                                               \
      {                                                                      \
	/* We can exit 'cause noone else got this far either */              \
	break;                                                               \
      }                                                                      \
      lf = (ogfs_leaf_t *)BH_DATA(lfbh);                                      \
      lf_no = ogfs64_to_cpu(lf->lf_next);                                     \
      /* release the buffer twice; once for this get_leaf, */                \
      /* and once for the one above */                                       \
      brelse(lfbh);                                                      \
      brelse(lfbh);                                                      \
    }                                                                        \
    while (lf_no);                                                           \
  }

/**
 * do_filldir - Fill directory
 * @dip: The OGFS inode
 * @f_pos:
 * @dirent:
 * @filldir: the buffer to fill with entries
 * @bh:  The disk order buffer
 * @index: Used to verify current dirent belongs in current leaf
 * @leaf_no:
 *
 * Does the filldir given a leaf or stuffed dinode
 * seven arguments stink. Ideas appreciated. (put in structure ?)
 *
 * New strategy:
 *  o get the leaf or stuffed dinode (or leaves if there are linked leaves)
 *  o sort the entries by their hash
 *  o find the last entry read if there is one, if not start at the first
 *    in sorted order
 *  o send them to filldir in order starting with the entry found in step 3
 *
 * Returns: 0 on success, error code on failure
 */

static int
do_filldir(ogfs_inode_t * dip, int64 * f_pos,
	   void *dirent, ogfs_filldir_t filldir,
	   struct buffer_head * bh, int index, uint64 leaf_no)
{
	ogfs_dirent_t *dent;
	ogfs_leaf_t *leaf = NULL;
	int type, len, error;
	ogfs_dirent_t **sorted_dents; /* Entries from current dent are sorted
				        and run throught the filldir fxn */
	uint16 dent_no, cur_dent_no;
	uint32 entries = 0;
	/* variables used to calculate things for the linked leaves */
	struct buffer_head *leafbh;
	uint64 orig_leaf_no = leaf_no;

	ENTER(GFN_DO_FILLDIR);

	type = ogfs_dirent_first(bh, &dent);

	/*
	 * We have to find all entries in the leaf and any leaves attached to
	 * the first leaf in a linked list.
	 */
	if (type == IS_LEAF) {
		do {
			error = get_leaf(dip, leaf_no, &leafbh);
			if (error) {
				EXIT(GFN_DO_FILLDIR);
				return error;
			}

			leaf = (ogfs_leaf_t *) BH_DATA(leafbh);
			entries += ogfs16_to_cpu(leaf->lf_entries);
			leaf_no = ogfs64_to_cpu(leaf->lf_next);
			brelse(leafbh);
		} while (leaf_no);
	} else {
		/* stuffed dinodes are easy */
		entries = dip->i_di.di_entries;
	}

	/*
	 * if there aren't any entries, don't bother doing anything
	 * with the dent.
	 */
	if (!entries) {
		EXIT(GFN_DO_FILLDIR);
		return 0;
	}

	sorted_dents =
	    (ogfs_dirent_t **) gmalloc(sizeof(ogfs_dirent_t *) * entries);
	if (!sorted_dents) {
		EXIT(GFN_DO_FILLDIR);
		return -ENOMEM;
	}
	/* Make sure we're starting from the beginning */
	dent_no = 0;

	/* Get all dent hash entries and stuff them somewhere for later
	 * sorting and processing */
	if (type == IS_LEAF) {
		/*
		 * grab the leaf no that was passed into the fxn,
		 * this is the head leaf
		 */
		leaf_no = orig_leaf_no;

		do {
			error = get_leaf(dip, leaf_no, &leafbh);

			/*
			 * if we get an error, it means we didn't find
			 * the right leaf
			 */
			if (error) {
				gfree(sorted_dents,
				      sizeof(ogfs_dirent_t *) * entries);
				CLEAN_UP_BH(type, orig_leaf_no, dip);
				EXIT(GFN_DO_FILLDIR);
				return error;
			}
			type = ogfs_dirent_first(leafbh, &dent);
			leaf = (ogfs_leaf_t *) BH_DATA(leafbh);

			/*
			 * if there aren't any dents, don't bother
			 * with this leaf
			 */
			if (!ogfs16_to_cpu(leaf->lf_entries)) {
				leaf_no = ogfs64_to_cpu(leaf->lf_next);

				/*
				 * don't need to release this now,
				 * a'cause I'll do it later
				 */
				/*brelse(leafbh); */
				continue;
			}

			/* put the dents into the list */
			do {
				if (!dent->de_inum.no_formal_ino)
					continue;

				/* sanity checking */
				len =
				    1 << (dip->i_di.di_depth -
					  ogfs16_to_cpu(leaf->lf_depth));
				OGFS_ASSERT((ogfs32_to_cpu(dent->de_hash) >>
					    (32 - dip->i_di.di_depth)) <
					   (index + len),);

				sorted_dents[dent_no] = dent;
				dent_no++;
			} while (ogfs_dirent_next(leafbh, &dent) == 0);

			leaf_no = ogfs64_to_cpu(leaf->lf_next);
			/*
			 * can't release the buffer now, 'cause I'm still
			 * referencing the  dentry inside it
			 *- See CLEAN_UP_BH macro for more info on this
			 */
		} while (leaf_no);
	} else {
		/*
		 * put the stuffed dinode entries into an array
		 */
		do {
			if (!dent->de_inum.no_formal_ino)
				continue;

			sorted_dents[dent_no] = dent;
			dent_no++;
		} while (ogfs_dirent_next(bh, &dent) == 0);
	}

	/* Sort hash entries */
	ogfs_sort(sorted_dents, (dent_no - 1), sizeof(ogfs_dirent_t *),
		 compare_dents);

	cur_dent_no = dent_no;
	/* Grab the current dent_no based on fpos */
	error = find_dent_index(sorted_dents, f_pos, &cur_dent_no);
	if (error) {
		/*
		 * If there is an error returned by find_dent_index,
		 * no matching entry was found */
		gfree(sorted_dents, sizeof(ogfs_dirent_t *) * entries);
		CLEAN_UP_BH(type, orig_leaf_no, dip);
		EXIT(GFN_DO_FILLDIR);
		return 0;
	} else {
		/* Run the dents through the filldir fxn */
		for (; cur_dent_no < dent_no; cur_dent_no++) {
			/* make sure the latest dent hash value is stored */
			*f_pos =
			    (uint64) (ogfs32_to_cpu
				      ((sorted_dents[cur_dent_no])->
				       de_hash) >> 1);

			/* get the current dent */
			dent = sorted_dents[cur_dent_no];

			error =
			    filldir(dirent, (char *) (dent + 1),
				    ogfs16_to_cpu(dent->de_name_len), *f_pos,
				    ogfs64_to_cpu(dent->de_inum.no_formal_ino),
				    ogfs16_to_cpu(dent->de_type));

			/*
			 * If there is an error, the filldir buffer is full,
			 * and we need to go back to userspace to let the
			 * calling program handle the buffer
			 */
			if (error < 0) {
				error = -EINVAL;
				gfree(sorted_dents,
				      sizeof(ogfs_dirent_t *) * entries);
				CLEAN_UP_BH(type, orig_leaf_no, dip);
				EXIT(GFN_DO_FILLDIR);
				return error;
			}
		}
	}

	/*
	 * Increment the *f_pos by one, so the next time we come into the
	 * do_filldir fxn, we get the next entry instead of the last one
	 * in the current leaf.
	 */
	(*f_pos)++;
	gfree(sorted_dents, sizeof(ogfs_dirent_t *) * entries);
	CLEAN_UP_BH(type, orig_leaf_no, dip);

	EXIT(GFN_DO_FILLDIR);

	return 0;
}

/**
 * dir_e_search -
 * @dip: The OGFS inode
 * @filename:
 * @inode:
 *
 * Returns:
 */

static int
dir_e_search(ogfs_inode_t * dip, struct qstr * filename,
	     ogfs_inum_t * inum, unsigned int *type)
{
	struct buffer_head *bh;
	ogfs_dirent_t *dent;
	int64 error;

	ENTER(GFN_DIR_E_SEARCH);

	/*  Sanity Checks  */

	OGFS_ASSERT(dip->i_di.di_flags & OGFS_DIF_EXHASH,);

	error = linked_leaf_search(dip, filename, &dent, NULL, &bh);
	if (error) {
		EXIT(GFN_DIR_E_SEARCH);
		return error;
	}

	if (inum)
		ogfs_inum_in(inum, (char *) &dent->de_inum);
	if (type)
		*type = ogfs16_to_cpu(dent->de_type);

	brelse(bh);

	EXIT(GFN_DIR_E_SEARCH);

	return 0;
}

/**
 * dir_e_add -
 * @dip: The OGFS inode
 * @filename:
 * @inode:
 * @type:
 *
 */

static int
dir_e_add(ogfs_inode_t * dip, struct qstr * filename,
	  ogfs_inum_t * inum, unsigned int type)
{
	ogfs_sbd_t *sdp = dip->i_sbd;
	struct buffer_head *bh, *nbh, *dibh;
	struct timespec ctime;
	ogfs_leaf_t *leaf, *nleaf;
	ogfs_dirent_t *dent;
	uint32 hsize, index;
	uint32 hash;
	uint64 leaf_no, bn;
	int64 error;

	ENTER(GFN_DIR_E_ADD);

restart:
	/*  Sanity Checks  */

	OGFS_ASSERT(dip->i_di.di_flags & OGFS_DIF_EXHASH,);
	hsize = 1 << dip->i_di.di_depth;
	OGFS_ASSERT(hsize * sizeof(uint64) == dip->i_di.di_size,);

	/*  Figure out the address of the leaf node.  */

	hash = dir_hash(filename);
	index = hash >> (32 - dip->i_di.di_depth);
	OGFS_ASSERT(index < hsize,);

	error = get_leaf_nr(dip, index, &leaf_no);
	if (error) {
		EXIT(GFN_DIR_E_ADD);
		return error;
	}

	/*  Add entry to the leaf  */

	while (TRUE) {
		error = get_leaf(dip, leaf_no, &bh);
		if (error) {
			EXIT(GFN_DIR_E_ADD);
			return error;
		}

		leaf = (ogfs_leaf_t *) BH_DATA(bh);

		if (ogfs_dirent_alloc(dip, bh, filename->len, &dent)) {
			if (ogfs16_to_cpu(leaf->lf_depth) < dip->i_di.di_depth) {
				/*  Can we split the leaf?  */
				brelse(bh);

				error = dir_split_leaf(dip, index, leaf_no);
				if (error) {
					EXIT(GFN_DIR_E_ADD);
					return error;
				}

				goto restart;
			} else if (dip->i_di.di_depth < OGFS_DIR_MAX_DEPTH) {
				/*  Can we double the hash table?  */
				brelse(bh);

				error = dir_double_exhash(dip);
				if (error) {
					EXIT(GFN_DIR_E_ADD);
					return error;
				}

				goto restart;
			} else if (leaf->lf_next) {
				/* Can we try the next leaf in the list? */
				leaf_no = ogfs64_to_cpu(leaf->lf_next);
				brelse(bh);
				continue;
			} else {
				/* Create a new leaf and add it to the list. */
				error = ogfs_metaalloc(dip, &bn);
				if (error) {
					brelse(bh);
					EXIT(GFN_DIR_E_ADD);
					return error;
				}

				nbh =
				    ogfs_dread(sdp, bn, dip->i_gl,
					      DR_FORCE | DR_START | DR_WAIT);
				if (!nbh) {
					brelse(bh);
					EXIT(GFN_DIR_E_ADD);
					return -EIO;
				}

				ogfs_trans_add_bh(sdp, nbh);
				ogfs_metatype_set(nbh, OGFS_METATYPE_LF,
						 OGFS_FORMAT_LF);
				ogfs_buffer_clear_tail(nbh,
						      sizeof
						      (ogfs_meta_header_t));

				ogfs_trans_add_bh(sdp, bh);
				leaf->lf_next = cpu_to_ogfs64(bn);

				nleaf = (ogfs_leaf_t *) BH_DATA(nbh);
				nleaf->lf_depth = leaf->lf_depth;
				nleaf->lf_dirent_format =
				    cpu_to_ogfs32(OGFS_FORMAT_DE);

				if (ogfs_dirent_alloc
				    (dip, nbh, filename->len, &dent))
					OGFS_ASSERT(0,);

				dip->i_di.di_blocks++;

				brelse(bh);

				bh = nbh;
				leaf = nleaf;
			}
		}

		/*  If the ogfs_dirent_alloc() succeeded, it pinned the "bh".  */

		ogfs_inum_out(inum, (char *) &dent->de_inum);
		dent->de_hash = cpu_to_ogfs32(hash);
		dent->de_type = cpu_to_ogfs16(type);
		memcpy((char *) (dent + 1), filename->name, filename->len);

		leaf->lf_entries = ogfs16_to_cpu(leaf->lf_entries) + 1;
		leaf->lf_entries = cpu_to_ogfs16(leaf->lf_entries);

		brelse(bh);

		dibh = ogfs_get_inode_buffer(dip);
		if (!dibh) {
			EXIT(GFN_DIR_E_ADD);
			return -EIO;
		}

		dip->i_di.di_entries++;
		ctime = current_kernel_time();
		dip->i_di.di_mtime = dip->i_di.di_ctime = (int64)ctime.tv_sec;

		ogfs_trans_add_bh(sdp, dibh);
		ogfs_dinode_out(&dip->i_di, BH_DATA(dibh));
		brelse(dibh);

		EXIT(GFN_DIR_E_ADD);

		return 0;
	}

	EXIT(GFN_DIR_E_ADD);

	return -ENOENT;
}

/**
 * dir_e_del -
 * @dip: The OGFS inode
 * @filename:
 *
 * Returns:
 */

static int
dir_e_del(ogfs_inode_t * dip, struct qstr * filename)
{
	struct buffer_head *bh, *dibh;
	struct timespec ctime;
	ogfs_dirent_t *dent, *prev;
	ogfs_leaf_t *leaf;
	unsigned int entries;
	int64 error;

	ENTER(GFN_DIR_E_DEL);

	/*  Sanity Checks  */

	OGFS_ASSERT(dip->i_di.di_flags & OGFS_DIF_EXHASH,);

	error = linked_leaf_search(dip, filename, &dent, &prev, &bh);
	if (error) {
		EXIT(GFN_DIR_E_DEL);
		return error;
	}

	ogfs_dirent_del(dip, bh, prev, dent);	/*  Pins bh  */

	leaf = (ogfs_leaf_t *) BH_DATA(bh);
	entries = ogfs16_to_cpu(leaf->lf_entries);
	OGFS_ASSERT(entries,);
	entries--;
	leaf->lf_entries = cpu_to_ogfs16(entries);

	brelse(bh);

	OGFS_ASSERT(dip->i_di.di_entries,);
	dip->i_di.di_entries--;

	ctime = current_kernel_time();
	dip->i_di.di_mtime = dip->i_di.di_ctime = (int64)ctime.tv_sec;

	dibh = ogfs_get_inode_buffer(dip);
	if (!dibh) {
		EXIT(GFN_DIR_E_DEL);
		return -EIO;
	}

	ogfs_trans_add_bh(dip->i_sbd, dibh);
	ogfs_dinode_out(&dip->i_di, BH_DATA(dibh));
	brelse(dibh);

	EXIT(GFN_DIR_E_DEL);

	return 0;
}

/**
 * dir_e_read - Reads the entries from a directory into a filldir buffer
 * @dip: dinode pointer
 * @f_pos: the hash of the last entry read shifted to the right once
 * @dirent: buffer for the filldir function to fill
 * @filldir: points to the filldir function to use
 *
 */

static int
dir_e_read(ogfs_inode_t * dip, int64 * f_pos, void *dirent,
	   ogfs_filldir_t filldir)
{
	ogfs_sbd_t *sdp = dip->i_sbd;	/* */
	struct buffer_head *bh;		/* */
	ogfs_leaf_t leaf;	/* */
	uint16 reread_lp;	/* 1 if the leaf pointer needs to be reread */
	uint32 hsize, len,	/* */
	 ht_offset,		/* a multiple of sdp->sd_hash_ptrs.
				   used to find the offset in the
				   inode of the current leaf block. */
	 lp_offset,		/* offset in the current leaf */
	 index,			/* The current index based on the hash value
				   stored in f_pos */
	 hash;			/* */
	uint64 *lp;		/* the leaf pointer */
	uint64 leaf_no;		/* */
	int64 error;		/* */

	ENTER(GFN_DIR_E_READ);

	/* we must re-read the lp the first time we come into this
	 * it used to be that we could use the value of lp_offset
	 * to determine this, but since we calculate lp_offset now
	 * we can't be sure that it will be 0 the first time.  */
	reread_lp = 1;

	/*  Sanity Checks  */
	OGFS_ASSERT(dip->i_di.di_flags & OGFS_DIF_EXHASH,);
	hsize = 1 << dip->i_di.di_depth;
	OGFS_ASSERT(hsize * sizeof(uint64) == dip->i_di.di_size,);

	lp = (uint64 *) gmalloc(sdp->sd_hash_bsize);
	if (!lp) {
		EXIT(GFN_DIR_E_READ);
		return -ENOMEM;
	}

	/* we need to shift the hash back to the left one so that the
	 * correct leaf is selected */
	hash = (uint32) ((*f_pos) << 1);

	/* calculate the index based on the hash */
	index = hash >> (32 - dip->i_di.di_depth);

	OGFS_ASSERT(index < hsize,);

	/* ht_offset is index rounded down to nearest half block size */
	ht_offset = index - (index % (sdp->sd_hash_ptrs));
	OGFS_ASSERT(ht_offset < hsize,);
	/* lp_offset is difference between the index and the ht_offset */
	lp_offset = index - ht_offset;
	if (lp_offset >= sdp->sd_hash_ptrs)
		lp_offset = 0;

	while (index < hsize) {
		if (reread_lp) {
			error =
			    ogfs_internal_read(dip, (char *) lp,
					      ht_offset * sizeof(uint64),
					      sdp->sd_hash_bsize);
			if (error != sdp->sd_hash_bsize) {
				gfree(lp, sdp->sd_hash_bsize);
				EXIT(GFN_DIR_E_READ);
				return (error < 0) ? error : -EIO;
			}
			reread_lp = 0;
		}

		leaf_no = ogfs64_to_cpu(lp[lp_offset]);

		error = get_leaf(dip, leaf_no, &bh);
		if (error) {
			gfree(lp, sdp->sd_hash_bsize);
			EXIT(GFN_DIR_E_READ);
			return error;
		}

		ogfs_leaf_in(&leaf, BH_DATA(bh));

		len = 1 << (dip->i_di.di_depth - leaf.lf_depth);

		/*
		 * resets leaf pointer offset to the beginning of the leaf.
		 * Without this, we tend to end up missing several entries
		 * every time the filldir buffer fills  make sure we aren't
		 * trying to mod by 0
		 */
		if (dip->i_di.di_depth - leaf.lf_depth)
			lp_offset =
			    lp_offset -
			    (lp_offset %
			     (2 << (dip->i_di.di_depth - leaf.lf_depth - 1)));

		error =
		    do_filldir(dip, f_pos, dirent, filldir, bh, index, leaf_no);
		if (error) {
			if (error == -EINVAL)
				error = 0;
			gfree(lp, sdp->sd_hash_bsize);

			brelse(bh);

			EXIT(GFN_DIR_E_READ);
			return error;
		}
		brelse(bh);

		lp_offset += len;
		index += len;

		if (lp_offset >= sdp->sd_hash_ptrs) {
			/*
			 * the ht_offset should always be a multiple of
			 * the sd_hash_ptrs
			 */
			ht_offset += sdp->sd_hash_ptrs;
			lp_offset = 0;
			reread_lp = 1;
		}
	}

	gfree(lp, sdp->sd_hash_bsize);

	EXIT(GFN_DIR_E_READ);

	return 0;
}

/**
 * dir_e_mvino -
 * @dip: The OGFS inode
 * @filename:
 * @new_inode:
 *
 * Returns:
 */

static int
dir_e_mvino(ogfs_inode_t * dip, struct qstr * filename,
	    ogfs_inum_t * inum, unsigned int new_type)
{
	ogfs_sbd_t *sdp = dip->i_sbd;
	struct buffer_head *bh, *dibh;
	struct timespec ctime;
	ogfs_dirent_t *dent;
	int64 error;

	ENTER(GFN_DIR_E_MVINO);

	/*  Sanity Checks  */

	OGFS_ASSERT(dip->i_di.di_flags & OGFS_DIF_EXHASH,);

	error = linked_leaf_search(dip, filename, &dent, NULL, &bh);
	if (error) {
		EXIT(GFN_DIR_E_MVINO);
		return error;
	}

	ogfs_trans_add_bh(sdp, bh);

	ogfs_inum_out(inum, (char *) &dent->de_inum);
	dent->de_type = cpu_to_ogfs16(new_type);

	brelse(bh);

	dibh = ogfs_get_inode_buffer(dip);
	if (!dibh) {
		EXIT(GFN_DIR_E_MVINO);
		return -EIO;
	}

	ctime = current_kernel_time();
	dip->i_di.di_mtime = dip->i_di.di_ctime = (int64)ctime.tv_sec;

	ogfs_trans_add_bh(sdp, dibh);
	ogfs_dinode_out(&dip->i_di, BH_DATA(dibh));
	brelse(dibh);

	EXIT(GFN_DIR_E_MVINO);

	return 0;
}

/**
 * dir_l_search -
 * @dip: The OGFS inode
 * @filename:
 * @inode:
 *
 * Returns:
 */

static int
dir_l_search(ogfs_inode_t * dip, struct qstr * filename,
	     ogfs_inum_t * inum, unsigned int *type)
{
	struct buffer_head *dibh;
	ogfs_dirent_t *dent;
	int64 error;

	ENTER(GFN_DIR_L_SEARCH);

	/*  Sanity checks  */

	OGFS_ASSERT(!(dip->i_di.di_flags & OGFS_DIF_EXHASH),);
	OGFS_ASSERT(ogfs_is_stuffed(dip),);

	dibh = ogfs_get_inode_buffer(dip);
	if (!dibh) {
		EXIT(GFN_DIR_L_SEARCH);
		return -EIO;
	}

	error = leaf_search(dibh, filename, &dent, NULL);
	if (error) {
		brelse(dibh);
		EXIT(GFN_DIR_L_SEARCH);
		return error;
	}

	if (inum)
		ogfs_inum_in(inum, (char *) &dent->de_inum);
	if (type)
		*type = ogfs16_to_cpu(dent->de_type);

	brelse(dibh);

	EXIT(GFN_DIR_L_SEARCH);

	return 0;
}

/**
 * dir_l_add -
 * @dip: The OGFS inode
 * @filename:
 * @inode:
 * @type:
 *
 * Returns:
 */

static int
dir_l_add(ogfs_inode_t * dip, struct qstr * filename,
	  ogfs_inum_t * inum, unsigned int type)
{
	struct buffer_head *dibh;
	struct timespec ctime;
	ogfs_dirent_t *dent;
	int error = 0;

	ENTER(GFN_DIR_L_ADD);

	/*  Sanity checks  */

	OGFS_ASSERT(!(dip->i_di.di_flags & OGFS_DIF_EXHASH),);
	OGFS_ASSERT(ogfs_is_stuffed(dip),);

	dibh = ogfs_get_inode_buffer(dip);
	if (!dibh) {
		EXIT(GFN_DIR_L_ADD);
		return -EIO;
	}

	if (ogfs_dirent_alloc(dip, dibh, filename->len, &dent)) {
		brelse(dibh);

		error = dir_make_exhash(dip);

		if (!error)
			error = dir_e_add(dip, filename, inum, type);

		EXIT(GFN_DIR_L_ADD);
		return error;
	}

	/*  ogfs_dirent_alloc() pins  */

	ogfs_inum_out(inum, (char *) &dent->de_inum);
	dent->de_hash = dir_hash(filename);
	dent->de_hash = cpu_to_ogfs32(dent->de_hash);
	dent->de_type = cpu_to_ogfs16(type);
	memcpy((char *) (dent + 1), filename->name, filename->len);

	dip->i_di.di_entries++;
	ctime = current_kernel_time();
	dip->i_di.di_mtime = dip->i_di.di_ctime = (int64)ctime.tv_sec;

	ogfs_dinode_out(&dip->i_di, BH_DATA(dibh));
	brelse(dibh);

	EXIT(GFN_DIR_L_ADD);

	return 0;
}

/**
 * dir_l_del -
 * @dip: The OGFS inode
 * @filename:
 *
 * Returns:
 */

static int
dir_l_del(ogfs_inode_t * dip, struct qstr * filename)
{
	struct buffer_head *dibh;
	struct timespec ctime;
	ogfs_dirent_t *dent, *prev;
	int64 error;

	ENTER(GFN_DIR_L_DEL);

	/*  Sanity checks  */

	OGFS_ASSERT(!(dip->i_di.di_flags & OGFS_DIF_EXHASH),);
	OGFS_ASSERT(ogfs_is_stuffed(dip),);

	dibh = ogfs_get_inode_buffer(dip);
	if (!dibh) {
		EXIT(GFN_DIR_L_DEL);
		return -EIO;
	}

	error = leaf_search(dibh, filename, &dent, &prev);
	if (error) {
		brelse(dibh);

		EXIT(GFN_DIR_L_DEL);
		return error;
	}

	ogfs_dirent_del(dip, dibh, prev, dent);

	/*  ogfs_dirent_del() pins  */

	OGFS_ASSERT(dip->i_di.di_entries,);
	dip->i_di.di_entries--;

	ctime = current_kernel_time();
	dip->i_di.di_mtime = dip->i_di.di_ctime = (int64)ctime.tv_sec;

	ogfs_dinode_out(&dip->i_di, BH_DATA(dibh));
	brelse(dibh);

	EXIT(GFN_DIR_L_DEL);

	return 0;
}

/**
 * dir_l_read -
 * @dip:
 * @f_pos:
 * @dirent:
 * @filldir:
 *
 * Returns:
 */

static int
dir_l_read(ogfs_inode_t * dip, int64 * f_pos, void *dirent,
	   ogfs_filldir_t filldir)
{
	struct buffer_head *dibh;
	int error;

	ENTER(GFN_DIR_L_READ);

	/*  Sanity checks  */

	OGFS_ASSERT(!(dip->i_di.di_flags & OGFS_DIF_EXHASH),);
	OGFS_ASSERT(ogfs_is_stuffed(dip),);

	dibh = ogfs_get_inode_buffer(dip);
	if (!dibh) {
		EXIT(GFN_DIR_L_READ);
		return -EIO;
	}

	error = do_filldir(dip, f_pos, dirent, filldir, dibh, 0, 0);
	if (error == -EINVAL)
		error = 0;

	brelse(dibh);

	EXIT(GFN_DIR_L_READ);

	return error;
}

/**
 * dir_l_mvino -
 * @dip:
 * @filename:
 * @new_inode:
 *
 * Returns:
 */

static int
dir_l_mvino(ogfs_inode_t * dip, struct qstr * filename,
	    ogfs_inum_t * inum, unsigned int new_type)
{
	struct buffer_head *dibh;
	struct timespec ctime;
	ogfs_dirent_t *dent;
	int64 error;

	ENTER(GFN_DIR_L_MVINO);

	/*  Sanity checks  */

	OGFS_ASSERT(!(dip->i_di.di_flags & OGFS_DIF_EXHASH),);
	OGFS_ASSERT(ogfs_is_stuffed(dip),);

	dibh = ogfs_get_inode_buffer(dip);
	if (!dibh) {
		EXIT(GFN_DIR_L_MVINO);
		return -EIO;
	}

	error = leaf_search(dibh, filename, &dent, NULL);
	if (error) {
		brelse(dibh);

		EXIT(GFN_DIR_L_MVINO);
		return error;
	}

	ogfs_trans_add_bh(dip->i_sbd, dibh);

	ogfs_inum_out(inum, (char *) &dent->de_inum);
	dent->de_type = cpu_to_ogfs16(new_type);

	ctime = current_kernel_time();
	dip->i_di.di_mtime = dip->i_di.di_ctime = (int64)ctime.tv_sec;

	ogfs_dinode_out(&dip->i_di, BH_DATA(dibh));
	brelse(dibh);

	EXIT(GFN_DIR_L_MVINO);

	return 0;
}

/**
 * ogfs_dir_search - Search a directory
 * @dip: The OGFS inode
 * @filename:
 * @inode:
 *
 * This routine searches a directory for a file or another directory.
 * Assumes a glock is held on dip.
 *
 * Returns: Inode number if found, -EXXXX on failure.
 */

int
ogfs_dir_search(ogfs_inode_t * dip, struct qstr * filename,
	       ogfs_inum_t * inum, unsigned int *type)
{
	int error;

	ENTER(GFN_DIR_SEARCH);

	if (dip->i_di.di_flags & OGFS_DIF_EXHASH)
		error = dir_e_search(dip, filename, inum, type);
	else
		error = dir_l_search(dip, filename, inum, type);

	EXIT(GFN_DIR_SEARCH);

	return error;
}

/**
 * ogfs_dir_add - Add new filename into directory
 * @dip: The OGFS inode
 * @filename: The new name
 * @inode: The inode number of the entry
 * @type: The type of the entry
 *
 * Returns: 0 on success, error code on failure
 */

int
ogfs_dir_add(ogfs_inode_t * dip, struct qstr * filename,
	    ogfs_inum_t * inum, unsigned int type)
{
	int error;

	ENTER(GFN_DIR_ADD);

	if (dip->i_di.di_flags & OGFS_DIF_EXHASH)
		error = dir_e_add(dip, filename, inum, type);
	else
		error = dir_l_add(dip, filename, inum, type);

	EXIT(GFN_DIR_ADD);

	return error;
}

/**
 * ogfs_dir_del - Delete a directory entry
 * @dip: The OGFS inode
 * @filename: The filename
 *
 * Returns: 0 on success, error code on failure
 */

int
ogfs_dir_del(ogfs_inode_t * dip, struct qstr * filename)
{
	int error;

	ENTER(GFN_DIR_DEL);

	if (dip->i_di.di_flags & OGFS_DIF_EXHASH)
		error = dir_e_del(dip, filename);
	else
		error = dir_l_del(dip, filename);

	EXIT(GFN_DIR_DEL);

	return error;
}

/**
 * ogfs_dir_read - Translate an OGFS filename
 * @dip: The OGFS inode
 * @f_pos:
 * @dirent:
 * @filldir:
 *
 * Returns: 0 on success, error code otherwise
 */

int
ogfs_dir_read(ogfs_inode_t * dip, int64 * f_pos, void *dirent,
	     ogfs_filldir_t filldir)
{
	int error;

	ENTER(GFN_DIR_READ);

	if (dip->i_di.di_flags & OGFS_DIF_EXHASH)
		error = dir_e_read(dip, f_pos, dirent, filldir);
	else
		error = dir_l_read(dip, f_pos, dirent, filldir);

	EXIT(GFN_DIR_READ);

	return error;
}

/**
 * ogfs_dir_mvino - Change inode number of directory entry
 * @dip: The OGFS inode
 * @filename:
 * @new_inode:
 *
 * This routine changes the inode number of a directory entry.  It's used
 * by rename to change ".." when a directory is moved.
 * Assumes a glock is held on dvp.
 *
 * Returns: 0 on success, -EXXXX on failure
 */

int
ogfs_dir_mvino(ogfs_inode_t * dip, struct qstr * filename,
	      ogfs_inum_t * inum, unsigned int new_type)
{
	int error;

	ENTER(GFN_DIR_MVINO);

	if (dip->i_di.di_flags & OGFS_DIF_EXHASH)
		error = dir_e_mvino(dip, filename, inum, new_type);
	else
		error = dir_l_mvino(dip, filename, inum, new_type);

	EXIT(GFN_DIR_MVINO);

	return error;
}

/**
 * leaf_free - Deallocate a directory leaf
 * @dip: the directory
 * @ht_offset: the hash table offset in the directory
 * @len: the number of pointers to this leaf
 * @leaf_no: the leaf number
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
leaf_free(ogfs_inode_t * dip, uint32 ht_offset, uint32 len, uint64 leaf_no)
{
	ogfs_sbd_t *sdp = dip->i_sbd;
	ogfs_leaf_t tmp_leaf;
	ogfs_rgrp_list_t rlist;
	ogfs_rgrpd_t *rgd;
	struct buffer_head *bh;
	uint64 blk;
	unsigned int blocks = 0;
	char *ht;
	unsigned int x, size = len * sizeof(uint64);
	int error;

	ENTER(GFN_LEAF_FREE);

	memset(&rlist, 0, sizeof(ogfs_rgrp_list_t));

	ht = (char *) gmalloc(size);
	if (!ht) {
		error = -ENOMEM;
		goto fail;
	}
	memset(ht, 0, size);

	error = ogfs_rindex_hold(sdp);
	if (error)
		goto fail_gfree;

	/*  Count the number of leaves  */

	for (blk = leaf_no; blk; blk = tmp_leaf.lf_next) {
		error = get_leaf(dip, blk, &bh);
		if (error)
			goto fail_free_rlist;
		ogfs_leaf_in(&tmp_leaf, BH_DATA(bh));
		brelse(bh);

		error = ogfs_rlist_add(sdp, &rlist, blk);
		if (error)
			goto fail_free_rlist;
	}

	ogfs_rlist_sort(&rlist);

	for (x = 0; x < rlist.rl_rgrps; x++) {
		rgd = rlist.rl_list[x];

		error = ogfs_glock_rg(rgd, 0);
		if (error)
			goto fail_rg_gunlock;

		blocks += rgd->rd_ri.ri_length;
	}

	/*  Trans may require:
	   All the bitmaps that were reserved.
	   One block for the dinode.
	   All the hash blocks that will be changed.
	 */

	error =
	    ogfs_trans_begin(sdp, TRANS_LEAF_DEALLOC,
			    blocks + 1 + (DIV_RU(size, sdp->sd_jbsize) + 1), 0);
	if (error)
		goto fail_rg_gunlock;

	for (blk = leaf_no; blk; blk = tmp_leaf.lf_next) {
		error = get_leaf(dip, blk, &bh);
		if (error)
			goto fail_end_trans;
		ogfs_leaf_in(&tmp_leaf, BH_DATA(bh));
		brelse(bh);

		error = ogfs_metafree(dip, blk, 1);
		if (error)
			goto fail_end_trans;
	}

	error = ogfs_internal_write(dip, ht, ht_offset * sizeof(uint64), size);
	if (error != (signed int)size) {  /* FIXME: AUDIT signed/unsigned bug */
		if (error >= 0)
			error = -EIO;
		goto fail_end_trans;
	}

	ogfs_trans_end(sdp);

	for (x = 0; x < rlist.rl_rgrps; x++) {
		rgd = rlist.rl_list[x];
		ogfs_gunlock_rg(rgd, 0);
	}
	ogfs_rlist_free(&rlist);
	ogfs_rindex_release(sdp);
	gfree(ht, size);

	EXIT(GFN_LEAF_FREE);

	return 0;

fail_end_trans:
	ogfs_trans_mark(sdp, TRF_POISONED);
	ogfs_trans_end(sdp);

fail_rg_gunlock:
	while (x--) {
		rgd = rlist.rl_list[x];
		ogfs_gunlock_rg(rgd, 0);
	}

fail_free_rlist:
	ogfs_rlist_free(&rlist);
	ogfs_rindex_release(sdp);

fail_gfree:
	gfree(ht, size);

fail:
	EXIT(GFN_LEAF_FREE);
	return error;
}

/**
 * ogfs_dir_exhash_free - free all the leaf block in a directory
 * @dip: the directory
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_dir_exhash_free(ogfs_inode_t * dip)
{
	ogfs_sbd_t *sdp = dip->i_sbd;
	struct buffer_head *bh;
	ogfs_leaf_t leaf;
	uint32 hsize, len, ht_offset, lp_offset;
	uint64 *lp;
	uint64 leaf_no;
	int64 error;

	ENTER(GFN_DIR_EXHASH_FREE);

	/*  Sanity Checks  */

	OGFS_ASSERT(dip->i_di.di_flags & OGFS_DIF_EXHASH,);
	hsize = 1 << dip->i_di.di_depth;
	OGFS_ASSERT(hsize * sizeof(uint64) == dip->i_di.di_size,);

	lp = (uint64 *) gmalloc(sdp->sd_hash_bsize);
	if (!lp) {
		error = -ENOMEM;
		goto fail;
	}

	ht_offset = 0;
	lp_offset = 0;

	while (ht_offset < hsize) {
		if (!lp_offset) {
			error =
			    ogfs_internal_read(dip, (char *) lp,
					      ht_offset * sizeof(uint64),
					      sdp->sd_hash_bsize);
			if (error != sdp->sd_hash_bsize) {
				if (error >= 0)
					error = -EIO;
				goto fail_gfree;
			}
		}

		leaf_no = ogfs64_to_cpu(lp[lp_offset]);

		if (leaf_no) {
			error = get_leaf(dip, leaf_no, &bh);
			if (error)
				goto fail_gfree;
			ogfs_leaf_in(&leaf, BH_DATA(bh));
			brelse(bh);

			len = 1 << (dip->i_di.di_depth - leaf.lf_depth);

			error = leaf_free(dip, ht_offset, len, leaf_no);
			if (error)
				goto fail_gfree;

			ht_offset += len;
			lp_offset += len;
		} else {
			ht_offset++;
			lp_offset++;
		}

		if (lp_offset >= sdp->sd_hash_ptrs)
			lp_offset = 0;
	}

	/*  Make this a regular file in case we crash.
	   (We don't want to free these blocks a second time.)  */

	error = ogfs_trans_begin(sdp, TRANS_DIR2FILE, 1, 0);
	if (error)
		goto fail_gfree;

	bh = ogfs_get_inode_buffer(dip);
	if (!bh) {
		error = -EIO;
		goto fail_end_trans;
	}

	ogfs_trans_add_bh(sdp, bh);
	((ogfs_dinode_t *) (BH_DATA(bh)))->di_type = cpu_to_ogfs16(FILE_REG);

	brelse(bh);

	ogfs_trans_end(sdp);

	gfree(lp, sdp->sd_hash_bsize);

	EXIT(GFN_DIR_EXHASH_FREE);

	return 0;

fail_end_trans:
	ogfs_trans_mark(sdp, TRF_POISONED);
	ogfs_trans_end(sdp);

fail_gfree:
	gfree(lp, sdp->sd_hash_bsize);

fail:
	EXIT(GFN_DIR_EXHASH_FREE);
	return error;
}
