/*
 *
 *    Copyright 1995-1999 Regents of the University of Minnesota
 *    Portions Copyright 1999-2001 Sistina Software, Inc.
 *    Portions Copyright 2003 The OpenGFS Project
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 */

/*
 * FIXME!  Use the page cache for metadata in 2.4/2.5.
 */

#include "ogfs_debug.h"
#include "ogfs.h"

#include <linux/pagemap.h>
#include <linux/file.h>
#include <linux/smp_lock.h>
#include <linux/mount.h>
#include <linux/uio.h>

#include "blklist.h"
#include "bmap.h"
#include "dcache.h"
#include "dio.h"
#include "dir.h"
#include "file.h"
#include "flock.h"
#include "ogfs_linux.h"
#include "glock.h"
#include "glops.h"
#include "inode.h"
#include "inode_linux.h"
//#include "file_linux.h"
#include "ioctl.h"
#include "log.h"
#include "plock.h"
#include "trans.h"

struct ogfs_filldir_struct {
	filldir_t vfs_filldir;
	void *vfs_opaque;
};
typedef struct ogfs_filldir_struct ogfs_filldir_struct_t;

/**
 * ogfs_copy2mem - Trivial copy function for ogfs_readi()
 * @bh: The buffer to copy from, or NULL meaning zero the buffer
 * @buf: The buffer to copy/zero
 * @offset: The offset in the buffer to copy from
 * @size: The amount of data to copy/zero
 *
 * This function is used as an argument to ogfs_readi() when we want
 * to read data as part of another OGFS internal function.
 *
 * Returns: The amount of data copied/cleared
 */

int
ogfs_copy2mem(struct buffer_head * bh, char *buf, int offset, int size)
{
	ENTER(GFN_COPY2MEM);

	if (bh)
		memcpy(buf, BH_DATA(bh) + offset, size);
	else
		memset(buf, 0, size);

	EXIT(GFN_COPY2MEM);

	return size;
}

/**
 * ogfs_copy2user - Used only for the ioctl() read function
 * @bh: The buffer
 * @buf: The destination of the data
 * @offset: The offset into the buffer
 * @size: The amount of data to copy
 *
 * Returns: Amount of data copied
 */

int
ogfs_copy2user(struct buffer_head * bh, char *buf, int offset, int size)
{
	int remainder;

	ENTER(GFN_COPY2USER);

	if (bh)
		remainder = copy_to_user(buf, BH_DATA(bh) + offset, size);
	else
		remainder = clear_user(buf, size);

	EXIT(GFN_COPY2USER);

	return (remainder) ? -EFAULT : size;
}

/**
 * ogfs_readi - Read a file
 * @ip: The OGFS Inode
 * @buf: The buffer to place result into
 * @offset: File offset to begin reading from
 * @size: Amount of data to transfer
 * @copy_fn: Function to actually perform the copy
 *
 * The @copy_fn only copies a maximum of a single block at once so
 * we are safe calling it with int arguments. It is done so that
 * we don't needlessly put 64bit arguments on the stack and it
 * also makes the code in the @copy_fn nicer too.
 *
 * Returns: The amount of data actually copied or the error
 */

int64
ogfs_readi(ogfs_inode_t * ip, char *buf, uint64 offset, uint64 size,
	  read_copy_fn_t copy_fn)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	struct buffer_head *bh;
	int64 copied = 0;
	uint64 block, nr;
	unsigned int amount;
	unsigned int bsize;
	int not_new = 0;
	int journaled = ogfs_is_jdata(ip);
	int64 error = 0;

	ENTER(GFN_READI);

	if (offset >= ip->i_di.di_size)
		goto out;

	if ((offset + size) > ip->i_di.di_size)
		size = ip->i_di.di_size - offset;

	if (!size)
		goto out;

	bsize = (journaled) ? sdp->sd_jbsize : sdp->sd_sb.sb_bsize;

	block = offset / bsize;
	offset %= bsize;

	if (ogfs_is_stuffed(ip))
		offset += sizeof(ogfs_dinode_t);
	else if (journaled)
		offset += sizeof(ogfs_jdata_t);

	while (copied < size) {
		amount = size - copied;
		if (amount > sdp->sd_sb.sb_bsize - offset)
			amount = sdp->sd_sb.sb_bsize - offset;

		error = ogfs_inode_get_block(ip, block, &not_new, &nr);
		if (error)
			goto out;

		if (nr) {
			bh = ogfs_get_data_buffer(ip, nr, not_new);
			if (!bh) {
				error = -EIO;
				goto out;
			}
		} else
			bh = NULL;

		error = copy_fn(bh, buf, (int) offset, amount);
		if (bh)
			brelse(bh);
		if (error < 0)
			goto out;

		copied += amount;
		buf += amount;
		block++;

		offset = (journaled) ? sizeof(ogfs_jdata_t) : 0;
	}

out:
	EXIT(GFN_READI);
	return (error < 0) ? error : copied;
}

/**
 * ogfs_read - Read bytes from a file
 * @filp: The file to read from
 * @buf: The buffer to copy into
 * @size: The amount of data requested
 * @offset: The current file offset
 *
 * Outputs: Offset - updated according to number of bytes read
 *
 * Returns: The number of bytes read, -EXXX on failure
 */

static ssize_t
ogfs_read(struct file * filp, char *buf, size_t size, loff_t * offset)
{
	ogfs_inode_t *ip = vi2ino(filp->f_dentry->d_inode);
	int error, count;

	ENTER(GFN_READ);

	lock_kernel();

	error = ogfs_glock_i(ip, GL_SHARED);
	if (error)
		goto fail;

	error = OGFS_UPDATE_ATIME(ip);
	if (error)
		goto fail_gunlock;

	count = generic_file_read(filp, buf, size, offset);

	ogfs_gunlock_i(ip, 0);

	unlock_kernel();

	EXIT(GFN_READ);

	return count;

fail_gunlock:
	ogfs_gunlock_i(ip, 0);

fail:
	unlock_kernel();
	EXIT(GFN_READ);
	return error;
}

/**
 * ogfs_copy_from_mem - Trivial copy function for ogfs_writei()
 * @ip: The file to write to
 * @bh: The buffer to copy to or clear
 * @buf: The buffer to copy from
 * @offset: The offset in the buffer to write to
 * @size: The amount of data to write
 * @new: Flag indicating that remaining space in the buffer should be zeroed
 *
 * Whilst this function does deal with stuffed dinodes, they are never new
 * buffers here, which is why a call to zero_buffer_ends is safe.
 *
 * Returns: The amount of data copied
 */

int
ogfs_copy_from_mem(ogfs_inode_t * ip, struct buffer_head * bh, const char *buf,
		  int offset, int size, int new)
{
	ENTER(GFN_COPY_FROM_MEM);

	if (BH_BLKNO(bh) == ip->i_num.no_addr) {
		OGFS_ASSERT(!new,);
		ogfs_trans_add_bh(ip->i_sbd, bh);
		memcpy(BH_DATA(bh) + offset, buf, size);
	} else if (ogfs_is_jdata(ip)) {
		ogfs_trans_add_bh(ip->i_sbd, bh);
		memcpy(BH_DATA(bh) + offset, buf, size);
		if (new)
			ogfs_buffer_clear_ends(bh, offset, size, TRUE);
	} else {
		memcpy(BH_DATA(bh) + offset, buf, size);
		if (new)
			ogfs_buffer_clear_ends(bh, offset, size, FALSE);
		ogfs_dwrite(ip->i_sbd, bh, DW_DIRTY);
	}

	EXIT(GFN_COPY_FROM_MEM);

	return size;
}

/**
 * ogfs_copy_from_user - Copy bytes from user space for ogfs_writei()
 * @ip: The file to write to
 * @bh: The buffer to copy to or clear
 * @buf: The buffer to copy from
 * @offset: The offset in the buffer to write to
 * @size: The amount of data to write
 * @new: Flag indicating that remaining space in the buffer should be zeroed
 *
 * If the buffer is a new one, and the copy from user space fails,
 * then we must zero out the space that would have been occupied by
 * data.
 *
 * Returns: The amount of data copied
 */

int
ogfs_copy_from_user(ogfs_inode_t * ip, struct buffer_head * bh, const char *buf,
		   int offset, int size, int new)
{
	int remainder;

	ENTER(GFN_COPY_FROM_USER);

	if (BH_BLKNO(bh) == ip->i_num.no_addr) {
		OGFS_ASSERT(!new,);
		ogfs_trans_add_bh(ip->i_sbd, bh);
		remainder = copy_from_user(BH_DATA(bh) + offset, buf, size);
	} else if (ogfs_is_jdata(ip)) {
		ogfs_trans_add_bh(ip->i_sbd, bh);
		remainder = copy_from_user(BH_DATA(bh) + offset, buf, size);
		if (!remainder && new)
			ogfs_buffer_clear_ends(bh, offset, size, TRUE);
	} else {
		remainder = copy_from_user(BH_DATA(bh) + offset, buf, size);
		if (!remainder) {
			if (new)
				ogfs_buffer_clear_ends(bh, offset, size, FALSE);
			ogfs_dwrite(ip->i_sbd, bh, DW_DIRTY);
		}
	}

	EXIT(GFN_COPY_FROM_USER);

	return (remainder) ? -EFAULT : size;
}

/**
 * ogfs_writei - Write bytes to a file
 * @ip: The OGFS inode
 * @buf: The buffer containing information to be written
 * @offset: The file offset to start writing at
 * @size: The amount of data to write
 * @copy_fn: Function to do the actual copying
 *
 * Returns: The number of bytes correctly written or error code
 */

int64
ogfs_writei(ogfs_inode_t * ip, const char *buf, uint64 offset, uint64 size,
	   write_copy_fn_t copy_fn)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	struct buffer_head *dibh, *bh;
	struct timespec ctime;
	int64 copied = 0;
	uint64 block, nr;
	unsigned int amount;
	unsigned int bsize;
	int new;
	int journaled = ogfs_is_jdata(ip);
	const uint64 start = offset;
	int64 error = 0;

	ENTER(GFN_WRITEI);

	/*  Bomb out on writing nothing.
	   Posix says we can't change the time here.  */

	if (!size)
		goto fail;	/*  Not really an error  */

	dibh = ogfs_get_inode_buffer(ip);
	if (!dibh) {
		error = -EIO;
		goto fail_di_relse;
	}

	ogfs_trans_add_bh(sdp, dibh);

	if (ogfs_is_stuffed(ip) &&
	    ((start + size) > (sdp->sd_sb.sb_bsize - sizeof(ogfs_dinode_t)))) {
		error = ogfs_unstuff_dinode(ip);
		if (error)
			goto fail_di_relse;
	}

	bsize = (journaled) ? sdp->sd_jbsize : sdp->sd_sb.sb_bsize;

	block = offset / bsize;
	offset %= bsize;

	if (ogfs_is_stuffed(ip))
		offset += sizeof(ogfs_dinode_t);
	else if (journaled)
		offset += sizeof(ogfs_jdata_t);

	while (copied < size) {
		amount = size - copied;
		if (amount > sdp->sd_sb.sb_bsize - offset)
			amount = sdp->sd_sb.sb_bsize - offset;

		new = TRUE;
		error = ogfs_inode_get_block(ip, block, &new, &nr);
		if (error)
			goto fail_di_relse;

		bh = ogfs_get_data_buffer(ip, nr,
					 (amount ==
					  sdp->sd_sb.sb_bsize) ? TRUE : new);
		if (!bh) {
			error = -EIO;
			goto fail_di_relse;
		}

		error = copy_fn(ip, bh, buf, (int) offset, amount, new);
		brelse(bh);
		if (error < 0)
			goto fail_di_relse;

		copied += amount;
		buf += amount;
		block++;

		offset = (journaled) ? sizeof(ogfs_jdata_t) : 0;
	}

	if (ip->i_di.di_size < start + copied)
		ip->i_di.di_size = start + copied;
	ctime = current_kernel_time();
	ip->i_di.di_mtime = ip->i_di.di_ctime = (int64)ctime.tv_sec;

	ogfs_dinode_out(&ip->i_di, BH_DATA(dibh));
	brelse(dibh);

	EXIT(GFN_WRITEI);

	return copied;

fail_di_relse:
	brelse(dibh);

fail:
	EXIT(GFN_WRITEI);
	return error;
}

/**
 * ogfs_write - Write bytes to a file
 * @filp: The file to write to
 * @buf: The buffer to copy from
 * @size: The amount of data requested
 * @offset: The current file offset
 *
 * Outputs: Offset - updated according to number of bytes written
 *
 * Returns: The number of bytes written, -EXXX on failure
 */

static ssize_t
ogfs_write(struct file * filp, const char *buf, size_t size, loff_t * offset)
{
	struct inode *inode = filp->f_dentry->d_inode;
	ogfs_inode_t *ip = vi2ino(inode);
	ogfs_sbd_t *sdp = ip->i_sbd;
	ogfs_ipreserv_t *res = &ip->i_res;
	struct buffer_head *bh;
	//struct iovec local_iov;
	struct timespec ctime;
	int64 count;
	uint32 ind = 0, tmp, data_blocks;
	int error;

	ENTER(GFN_WRITE);

	lock_kernel();

	/*  Compute maximum number of data blocks we need.
	   (We might need a block to unstuff the dinode)  */

	data_blocks = (size >> sdp->sd_sb.sb_bsize_shift) + 3;

	/*  Compute maximum number of indirect blocks we need  */

	for (tmp = data_blocks; tmp > sdp->sd_diptrs;) {
		tmp = DIV_RU(tmp, sdp->sd_inptrs) + 1;
		ind += tmp;
	}

	ind += 2 * (sdp->sd_max_height - 1);

	down(&inode->i_sem);

	error = ogfs_glock_i(ip, 0);
	if (error)
		goto fail;

	res->ir_diblks = 0;
	res->ir_metablks = ind;
	res->ir_datablks = data_blocks;
	res->ir_target = ip->i_di.di_goal_rgrp;
	res->ir_trans = TRANS_WRITE;

	error = ogfs_inplace_reserv(ip, 0);
	if (error)
		goto fail_gunlock;

	/*  Trans may require:
	   All blocks for a RG bitmap, whatever indirect blocks we
	   need, and a modified dinode  */

	error =
	    ogfs_trans_begin(sdp, TRANS_WRITE,
			    1 + res->ir_rgd->rd_ri.ri_length + ind, 0);
	if (error)
		goto fail_ipres;

	struct iovec local_iov = { (void *) buf, size };

	count = generic_file_write_nolock(filp, &local_iov, 1, offset);
	if (count < 0) {
		error = count;
		goto fail_end_trans;
	}

	bh = ogfs_get_inode_buffer(ip);
	if (!bh) {
		error = -EIO;
		goto fail_end_trans;
	}

	if (ip->i_di.di_size < *offset)
		ip->i_di.di_size = *offset;
	if (!capable(CAP_FSETID))
		ip->i_di.di_mode &=
		    (ip->i_di.
		     di_mode & S_IXGRP) ? (~(S_ISUID | S_ISGID)) : (~S_ISUID);
	ctime = current_kernel_time();
	ip->i_di.di_mtime = ip->i_di.di_ctime = (int64)ctime.tv_sec;

	ogfs_trans_add_bh(sdp, bh);
	ogfs_dinode_out(&ip->i_di, BH_DATA(bh));
	brelse(bh);

	ogfs_trans_end(sdp);

	ogfs_inplace_relse(ip);

	ogfs_gunlock_i(ip, 0);

	up(&inode->i_sem);

	unlock_kernel();

	EXIT(GFN_WRITE);

	return count;

fail_end_trans:
	ogfs_trans_mark(sdp, TRF_POISONED);
	ogfs_trans_end(sdp);

fail_ipres:
	ogfs_inplace_relse(ip);

fail_gunlock:
	ogfs_gunlock_i(ip, 0);

fail:
	up(&inode->i_sem);
	unlock_kernel();
	EXIT(GFN_WRITE);
	return error;
}

/**
 * ogfs_filldir_default - copy a directory entry into a user space buffer
 * @opaque: opaque type
 * @name: the name of the file
 * @length: the length of the directory name
 * @offset: this entry's offset in the directory
 * @inode: the entry's inode number
 * @type: the type of the inode
 *
 * Returns: 0 on success, -EXXX on full or failure
 */

static int
ogfs_filldir_default(void *opaque,
		    const char *name, unsigned int length,
		    uint64 offset, uint64 inode, unsigned type)
{
	ogfs_filldir_struct_t *fs = (ogfs_filldir_struct_t *) opaque;
	unsigned int vfs_type;
	int ret;

	ENTER(GFN_FILLDIR_DEFAULT);

	switch (type) {
	case FILE_NON:
		vfs_type = DT_UNKNOWN;
		break;
	case FILE_REG:
		vfs_type = DT_REG;
		break;
	case FILE_DIR:
		vfs_type = DT_DIR;
		break;
	case FILE_LNK:
		vfs_type = DT_LNK;
		break;
	case FILE_BLK:
		vfs_type = DT_BLK;
		break;
	case FILE_CHR:
		vfs_type = DT_CHR;
		break;
	case FILE_FIFO:
		vfs_type = DT_FIFO;
		break;
	case FILE_SOCK:
		vfs_type = DT_SOCK;
		break;
	default:
		OGFS_ASSERT(0, con_printf("type = %u\n", type););
	}

	ret =
	    fs->vfs_filldir(fs->vfs_opaque, name, length, offset, inode,
			    vfs_type);

	EXIT(GFN_FILLDIR_DEFAULT);

	return ret;
}

/**
 * ogfs_readdir - Read directory entries from a directory
 * @filp: The directory to read from
 * @dirent: Buffer for dirents
 * @filldir: Function used to do the copying
 *
 * Returns: 0 on success, -EXXXX on failure
 */

static int
ogfs_readdir(struct file *filp, void *dirent, filldir_t filldir)
{
	ogfs_inode_t *dip = vi2ino(filp->f_dentry->d_inode);
	ogfs_filldir_struct_t fs;
	int error;

	ENTER(GFN_READDIR);

	fs.vfs_filldir = filldir;
	fs.vfs_opaque = dirent;

	error = ogfs_glock_i(dip, GL_SHARED);
	if (error)
		goto fail;

	error = OGFS_UPDATE_ATIME(dip);
	if (error)
		goto fail_gunlock;

	error =
	    ogfs_dir_read(dip, (int64 *) & (filp->f_pos), &fs,
			 ogfs_filldir_default);

	ogfs_gunlock_i(dip, 0);

	EXIT(GFN_READDIR);

	return error;

fail_gunlock:
	ogfs_gunlock_i(dip, 0);

fail:
	EXIT(GFN_READDIR);
	return error;
}

/**
 * ogfs_open_by_number - Open a file by inode number
 * @sb: The superblock
 * @ptr: The inode number to open
 * @vfsmnt: The mount point through which the inode is to be accessed
 *
 * The inode is always opened in read/write mode. If the inode does not
 * already exist, then an error will be returned. This function is for
 * use only by sysadmin through the OGFS ioctl() function. It is intended
 * as an access point for special files such as the rindex which may
 * not be read by other means.
 *
 * Returns: The newly opened fd, or an error
 */

static int
ogfs_open_by_number(struct super_block *sb, char *ptr, struct vfsmount *vfsmnt)
{
	int fd;
	struct qstr this;
	char name[32];
	struct inode *inode;
	ogfs_sbd_t *sdp = sb->s_fs_info;
	ogfs_glock_t *gl;
	ogfs_inode_t *ip;
	ogfs_inum_t inum;

	ENTER(GFN_OPEN_BY_NUMBER);

	if (copy_from_user(&inum, ptr, sizeof(ogfs_inum_t))) {
		EXIT(GFN_OPEN_BY_NUMBER);
		return -EFAULT;
	}

	fd = ogfs_glock_num(sdp, inum.no_formal_ino, &ogfs_inode_glops, NULL,
			   GL_SHARED, &gl);
	if (fd) {
		EXIT(GFN_OPEN_BY_NUMBER);
		return fd;
	}

	ip = ogfs_get_istruct(sdp, gl, &inum, CREATE);
	if (!ip) {
		EXIT(GFN_OPEN_BY_NUMBER);
		return -ENOMEM;
	}

	ogfs_gunlock(sdp, gl, 0);
	ogfs_put_glstruct(gl);

	inode = ogfs_iget(sb, ip);
	ogfs_put_istruct(ip);

	if (!inode) {
		EXIT(GFN_OPEN_BY_NUMBER);
		return -ENOENT;
	}

	fd = get_unused_fd();
	if (fd >= 0) {
		struct file *file = get_empty_filp();

		if (!file) {
			put_unused_fd(fd);
			fd = -ENFILE;
			goto out;
		}

		sprintf(name, "inode:[%016" FMT_64 "x]", inum.no_formal_ino);
		this.name = name;
		this.len = strlen(name);
		this.hash = inum.no_formal_ino;

		file->f_dentry = d_alloc(vfsmnt->mnt_sb->s_root, &this);
		if (!file->f_dentry) {
			put_filp(file);
			put_unused_fd(fd);
			fd = -ENOMEM;
			goto out;
		}

		file->f_dentry->d_op = &ogfs_dops;
		d_add(file->f_dentry, inode);
		file->f_vfsmnt = mntget(vfsmnt);
		file->f_op = inode->i_fop;
		file->f_mode = 3;
		file->f_flags = O_RDWR;
		file->f_pos = 0;
		fd_install(fd, file);
	}

out:
	if (fd < 0)
		iput(inode);

	EXIT(GFN_OPEN_BY_NUMBER);
	return fd;
}

/**
 * ogfs_ioctl - do an ioctl on a file
 * @inode: the inode
 * @filp: the file pointer
 * @cmd: the ioctl command
 * @arg: the argument
 *
 * Returns: 0 on success, -EXXXX on failure
 */

int
ogfs_ioctl(struct inode *inode, struct file *filp, unsigned int cmd,
	  unsigned long arg)
{
	ogfs_sbd_t *sdp = NULL;
	ogfs_inode_t *ip = NULL;
	int error = 0;

	ENTER(GFN_IOCTL);

	switch (cmd) {
	case OGFS_DEBUG_DUMP:
	case OGFS_PRINT_FRAG:
	case OGFS_SHRINK:
	case OGFS_PROF_DUMP:
	case OGFS_STATGFS:
		break;

	default:
		if (!capable(CAP_SYS_ADMIN)) {
			EXIT(GFN_IOCTL);
			return -EPERM;
		}
		break;
	}

	if (inode) {
		ip = vi2ino(inode);
		OGFS_ASSERT(ip,);
		sdp = ip->i_sbd;
		OGFS_ASSERT(sdp,);
	}

	switch (cmd) {
	case OGFS_DEBUG_DUMP:
#ifdef DEBUG_PRINT
		error = ogfs_debug_dump((void *) arg);
		break;
#else
		con_printf("OGFS:  incore printing not compiled in\n");
		error = -ENOSYS;
		break;
#endif				/*  DEBUG_PRINT  */

#ifdef DEBUG_TRACE
	case OGFS_DEBUG_ON:
	case OGFS_DEBUG_OFF:
		error = ogfs_debug_change(cmd, (void *) arg);
		break;
#else
	case OGFS_DEBUG_ON:
	case OGFS_DEBUG_OFF:
		con_printf("OGFS:  enter/exit tracing not compiled in\n");
		error = -ENOSYS;
		break;
#endif				/*  DEBUG_TRACE  */

	case OGFS_PRINT_FRAG:
#ifdef DEBUG_PRINT
		ogfs_print_frag(ip);
#else
		con_printf("OGFS:  incore printing not compiled in\n");
		error = -ENOSYS;
#endif				/*  DEBUG_PRINT  */
		break;

	case OGFS_SHRINK:
		ogfs_drop_excess_inodes(sdp);
		break;

	case OGFS_PROF_DUMP:
#ifdef DEBUG_PROFILE
		error = ogfs_prof_dump((void *) arg);
		break;
#else
		con_printf("OGFS:  profiling not compiled in\n");
		error = -ENOSYS;
		break;
#endif				/*  DEBUG_PROFILE  */

	case OGFS_OBYNUM:
		error =
		    ogfs_open_by_number(inode->i_sb, (char *) arg,
				       filp->f_vfsmnt);
		break;

	case OGFS_GET_SUPER:
		error = ogfs_get_super(sdp, (void *) arg);
		break;

	case OGFS_JREAD:
		error = ogfs_jread_ioctl(ip, (void *) arg);
		break;

	case OGFS_JWRITE:
		error = ogfs_jwrite_ioctl(ip, (void *) arg);
		break;

	case OGFS_LOCK_DUMP:
		ogfs_dump_lockstate(sdp);
		break;

	case OGFS_STATGFS:
		error = ogfs_stat_ogfs_ioctl(sdp, (void *) arg);
		break;

	case OGFS_STATRGRP:
		error = ogfs_stat_rgrp_ioctl(sdp, (void *) arg);
		break;

	case OGFS_RECLAIM_ONE:
		lock_kernel();
		error = ogfs_reclaim_one_ioctl(sdp, (void*) arg);
		unlock_kernel();
		break;

	case OGFS_RECLAIM_ALL:
		lock_kernel();
		error = ogfs_reclaim_all_ioctl(sdp);
		unlock_kernel();
		break;

	default:
		error = -ENOTTY;
		break;
	}

	EXIT(GFN_IOCTL);

	return error;
}

/**
 * ogfs_open - open a file
 * @inode: the inode to open
 * @filp: the struct file for this opening
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_open(struct inode *inode, struct file *filp)
{
	ogfs_inode_t *ip = vi2ino(inode);
	int error = 0;

	ENTER(GFN_OPEN);

	/*  Conform to the LFS spec on non-O_LARGEFILE files.  */

	if (!(filp->f_flags & O_LARGEFILE) && inode->i_size > 0x7FFFFFFFull) {
		error = -EFBIG;
		goto out;
	}

	/*  If this is an exclusive create, make sure our ogfs_create()
	   says we created the file.  The O_EXCL flag isn't passed
	   to ogfs_create(), so we have to check it here.  */

	if ((filp->f_flags & O_EXCL)
	    && !(test_bit(GIF_CREATED, &ip->i_flags)))
		error = -EEXIST;

out:
	EXIT(GFN_OPEN);

	return error;
}

/**
 * ogfs_sync_file - sync the dirty data for a file (across the cluster)
 * @filp: the filp that points to the dentry (Huh?)
 * @dentry: the dentry that points to the inode to sync
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_sync_file(struct file *filp, struct dentry *dentry, int datasync)
{
	ogfs_inode_t *ip = vi2ino(dentry->d_inode);
	int error = 0;

	ENTER(GFN_SYNC_FILE);

	lock_kernel();

	error = ogfs_glock_i(ip, 0);
	if (!error)
		ogfs_gunlock_i(ip, GL_SYNC);

	unlock_kernel();

	EXIT(GFN_SYNC_FILE);

	return error;
}

/**
 * ogfs_lock_file - acquire/release a flock or posix lock on a file
 * @filp: The file pointer, we just want the inode # from this.
 * @cmd: modifying a lock? or just peeking at it?
 * @fl: THe file Lock structure with all of the gory details.
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_lock_file(struct file *filp, int cmd, struct file_lock *fl)
{
	ogfs_inode_t *ip = vi2ino(filp->f_dentry->d_inode);
	ogfs_sbd_t *sdp = ip->i_sbd;
	int err = 0;
	char lckcmd = 0;

	ENTER(GFN_LOCK_FILE);

	lock_kernel();

	if (sdp->sd_args.ar_localflocks) {
		unlock_kernel();
		EXIT(GFN_LOCK_FILE);
		return LOCK_USE_CLNT;
	}

	/* ok, no mandatory locks. */

	if ((ip->i_di.di_mode & (S_ISGID | S_IXGRP)) == S_ISGID) {
		unlock_kernel();
		EXIT(GFN_LOCK_FILE);
		return -ENOLCK;
	}

	if (fl->fl_flags & FL_FLOCK) {
		if (fl->fl_type == F_UNLCK)
			ogfs_funlock(ip, (long) filp);	/* flip is the unique id for flocks. */
		else {
			if (fl->fl_type == F_RDLCK)
				lckcmd = FLOCK_SH;
			if (fl->fl_type == F_WRLCK)
				lckcmd = FLOCK_EX;

			if (cmd == F_GETLK)	/* query lock status, doesn't apply to flocks */
				err = -EINVAL;
			else if (cmd == F_SETLK)	/* attempt to set a lock, NonBlocking */
				err = ogfs_flock(ip, lckcmd, FALSE, (long) filp);
			else if (cmd == F_SETLKW)	/* attempt to set a lock, Blocking */
				err = ogfs_flock(ip, lckcmd, TRUE, (long) filp);
			else
				err = -EINVAL;
		}
	} else if (fl->fl_flags & FL_POSIX) {
		if (fl->fl_type == F_UNLCK)
			ogfs_plock_act(ip, fl->fl_start, fl->fl_end, PLOCK_UN,
				      FALSE);
		else {
			if (fl->fl_type == F_RDLCK)
				lckcmd = PLOCK_SH;
			if (fl->fl_type == F_WRLCK)
				lckcmd = PLOCK_EX;

			if (cmd == F_GETLK) {	/* query lock status */
				err =
				    ogfs_ptestlock(ip, (u64 *) & fl->fl_start,
						  (u64 *) & fl->fl_end,
						  &lckcmd, &fl->fl_pid);
				if (lckcmd == PLOCK_SH)
					fl->fl_type = F_RDLCK;
				else
					fl->fl_type = F_WRLCK;
			} else if (cmd == F_SETLK)	/* attempt to set a lock, NonBlocking */
				err =
				    ogfs_plock_act(ip, fl->fl_start, fl->fl_end,
						  lckcmd, FALSE);
			else if (cmd == F_SETLKW)	/* attempt to set a lock, Blocking */
				err =
				    ogfs_plock_act(ip, fl->fl_start, fl->fl_end,
						  lckcmd, TRUE);
			else
				err = -EINVAL;
		}
	} else			/* Not supported. */
		err = -ENOLCK;

	unlock_kernel();

	EXIT(GFN_LOCK_FILE);

	return err;
}

#ifdef OGFS_SHARED_WRITEABLE_MMAP
/**
 * ogfs_shared_nopage - support shared writable mappings
 * @area - vm_area to pass through to filemap_nopage
 * @address - pass through to filemap_nopage
 * @no_share - pass through to filemap_nopage
 *
 * Currently we don't support the shared state. Thus, if somebody owns a page
 * in the file, they also have an exclusive lock on the whole file.
 *
 * Returns: the new page
 */

static struct page *
ogfs_shared_nopage(struct vm_area_struct *area,
		  unsigned long address, int no_share)
{
	ogfs_inode_t *ip = vi2ino(area->vm_file->f_dentry->d_inode);
	struct page *result = NULL;
	int error;

	ENTER(GFN_SHARED_NOPAGE);

	lock_kernel();

	error = ogfs_glock_i(ip, 0);
	if (error)
		goto out;

	result = filemap_nopage(area, address, no_share);

	ogfs_gunlock_i(ip, 0);

out:
	unlock_kernel();

	EXIT(GFN_SHARED_NOPAGE);

	return result;
}

static struct vm_operations_struct ogfs_shared_mmap = {
	nopage:		ogfs_shared_nopage,
};
#endif				/*  OGFS_SHARED_WRITEABLE_MMAP  */

/**
 * ogfs_private_nopage - do safe locking on private mappings
 * @area - vm_area to pass through to filemap_nopage
 * @address - pass through to filemap_nopage
 * @no_share - pass through to filemap_nopage
 *
 * Returns: the new page
 */

static struct page *
ogfs_private_nopage(struct vm_area_struct *area,
		   unsigned long address, int no_share)
{
	ogfs_inode_t *ip = vi2ino(area->vm_file->f_dentry->d_inode);
	struct page *result = NULL;
	int error;

	ENTER(GFN_PRIVATE_NOPAGE);

	lock_kernel();

	error = ogfs_glock_i(ip, GL_SHARED);
	if (error)
		goto out;

	result = filemap_nopage(area, address, no_share);

	ogfs_gunlock_i(ip, 0);

out:
	unlock_kernel();

	EXIT(GFN_PRIVATE_NOPAGE);

	return result;
}

static struct vm_operations_struct ogfs_private_mmap = {
	nopage:		ogfs_private_nopage,
};

/**
 * ogfs_file_mmap - We don't support shared writable mappings right now
 * @file: The file to map
 * @vma: The VMA which described the mapping
 *
 * Returns: 0 or error code
 */

int
ogfs_file_mmap(struct file *file, struct vm_area_struct *vma)
{
	ogfs_inode_t *ip = vi2ino(file->f_dentry->d_inode);
	ogfs_sbd_t *sdp = ip->i_sbd;
	int error = 0;

	ENTER(GFN_FILE_MMAP);

	if ((vma->vm_flags & (VM_MAYWRITE | VM_SHARED)) ==
	    (VM_MAYWRITE | VM_SHARED))
#ifdef OGFS_SHARED_WRITEABLE_MMAP
		vma->vm_ops = &ogfs_shared_mmap;
#else
		error = -ENOSYS;
#endif				/*  OGFS_SHARED_WRITEABLE_MMAP  */
	else
		vma->vm_ops = &ogfs_private_mmap;

	if (!error && !sdp->sd_noatime) {
		error = ogfs_glock_i(ip, GL_SHARED);
		if (error)
			goto fail;

		error = ogfs_update_atime(ip);
		if (error)
			goto fail_gunlock;

		ogfs_gunlock_i(ip, 0);
	}

	EXIT(GFN_FILE_MMAP);

	return error;

fail_gunlock:
	ogfs_gunlock_i(ip, 0);

fail:
	EXIT(GFN_FILE_MMAP);
	return error;
}

struct file_operations ogfs_file_fops = {
	read:		ogfs_read,
	write:		ogfs_write,
	ioctl:		ogfs_ioctl,
	mmap:		ogfs_file_mmap,
	open:		ogfs_open,
	fsync:		ogfs_sync_file,
	lock:		ogfs_lock_file
};

struct file_operations ogfs_dir_fops = {
	readdir:	ogfs_readdir,
	ioctl:		ogfs_ioctl,
	fsync:		ogfs_sync_file
};
