/*
 *
 *    Copyright 1995-1999 Regents of the University of Minnesota
 *    Portions Copyright 2000-2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 */

#include "ogfs_debug.h"
#include "ogfs.h"
#include "glock.h"
#include "glops.h"
#include "flock.h"

struct pids_lst {
	struct list_head pid_list;
	long pid;
};
typedef struct pids_lst pid_lst_t;

/**
 * ogfs_flock - Acquire a flock on a file
 * @ip: <> the file
 * @type: < the type of lock (FLOCK_EX or FLOCK_SH)
 * @wait: < TRUE if the call should block until successful
 * @owner: < An identifier of who owns this lock.
 *
 * A list of inode numbers that have been flock'ed is kept. This is because
 * the upper layers let someone go and unlock a lock that wasn't locked.
 * The glock assumes this will never happen.  So we need to check and make
 * sure that the inode they are unflocking was previously flocked.
 *
 *
 * Returns: 0 on success, -EXXX on failure
 */
int
ogfs_flock(ogfs_inode_t * ip, int type, int wait, long owner)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	int flags;
	int error = 0;
	pid_lst_t *pid = NULL;

	ENTER(GFN_FLOCK);

	IFDB(GFN_FLOCK,
	     db_printf("flocking %" FMT_64
		       "u, flags = %d, wait = %d owner = %lu\n",
		       ip->i_num.no_formal_ino, type, wait, owner);
	    );

	/*
	 * this sema forces one pid at a time through the lock calls.
	 *
	 * Why is this here? Well, one pid gets list lock, scans, leaves goes
	 * for glock. Another pid scans list sees its clean goes for glock.
	 * First pid adds info to list that the second pid should have
	 * conflicted on.
	 *
	 * So this should force some ordering.  probably not the best solution,
	 * but I am getting so sick of this.
	 *
	 * mmmmm, temphack.  made it interuptible. and if we got a signal, we
	 * just leave right away.  So this gets one case out of the D state...
	 *
	 * I wonder if I shouldn't look into redesigning this to not be needed.
	 * (if it can be done ofcourse.)
	 *
	 * I feel like I should respond differently if I get the lmx and see
	 * that the pending type is EXL
	 */
	down_interruptible(&ip->i_flock_lmx);
	if (osi_pending_signals())
		goto exit_ogfs_flock;

	/* ACK!  glock works at the machine level, so if I want multiple
	 * processes sharing a flock on one host, I need to esscape that out!
	 * (eeww, bet this will have a bug or two before I get it working.)
	 */
	{
		struct list_head *pidtmp, *pidhead;
		int found = FALSE;
		pid = NULL;
		down(&ip->i_flock_mux);
		pidtmp = pidhead = &ip->i_flock_pids;
		for (;;) {
			pidtmp = pidtmp->next;
			if (pidtmp == pidhead)
				break;
			pid = list_entry(pidtmp, pid_lst_t, pid_list);

			if (pid->pid == owner) {
				/* we have this lock already. */
				found = TRUE;
				break;
			}
		}		/* pid loop */
		up(&ip->i_flock_mux);
		if (found)
			goto exit_ogfs_flock;

		if (type == FLOCK_SH && ip->i_flock_st == FLOCK_SH) {
			pid = NULL;
			/*
			 * someone here has it shared already, lets just add
			 * ourselves to the list of shared holders since we
			 * also want it shared.
			 *
			 * with flocks it is ok to starve EXL requests by
			 * other shares comming along. its mean and evil,
			 * but its the way flocks work.
			 */
			pid = (pid_lst_t *) gmalloc(sizeof(pid_lst_t));
			if (!pid) {
				error = -ENOMEM;
				goto exit_ogfs_flock;
			}
			memset(pid, 0, sizeof(pid_lst_t));
			INIT_LIST_HEAD(&pid->pid_list);
			pid->pid = owner;

			down(&ip->i_flock_mux);
			list_add(&pid->pid_list, &ip->i_flock_pids);
			up(&ip->i_flock_mux);

			error = 0;
			goto exit_ogfs_flock;
		}		/* if want shared, and is shared */
	}

	/* at this point;
	 *   either noone on this node holds the lock.
	 *   or someone holds it in conflist type of what we're requesting.
	 *  The other holder could be either on this node or on a different one.
	 */

	/* grab some memory before we do the glock.  this way if this fails, the
	 * glock won't be held.
	 */
	pid = (pid_lst_t *) gmalloc(sizeof(pid_lst_t));
	if (!pid) {
		error = -ENOMEM;
		goto exit_ogfs_flock;
	}
	memset(pid, 0, sizeof(pid_lst_t));

	/* ok, now map it to a glock */
	flags = GL_PERM | GL_DISOWN | ((wait) ? 0 : LM_FLAG_TRY);
	switch (type) {
	case FLOCK_EX:
		break;
	case FLOCK_SH:
		flags |= GL_SHARED;
		break;
	default:
		OGFS_ASSERT(0,);
	}

	error = ogfs_glock_num(sdp, ip->i_num.no_formal_ino, &ogfs_flock_glops,
			      NULL, flags, NULL);
	if (error == GLR_TRYFAILED)
		error = -EAGAIN;
	OGFS_ASSERT(error <= 0,);

	/* check signals.  If any, drop lock and leave. */
	if (osi_pending_signals()) {
		/* drop glock if we have it. */
		if (error == 0) {
			ogfs_gunlock_num(
				sdp, ip->i_num.no_formal_ino, &ogfs_flock_glops,
				NULL, GL_NOCACHE);
		}
		/* set new error, and go. */
		error = -ERESTARTSYS;
	}

	/* if we got it, put it onto the list. */
	if (error == 0) {
		down(&ip->i_flock_mux);
		/* the next line could race with unlock.... */
		ip->i_flock_st = type;
		INIT_LIST_HEAD(&pid->pid_list);
		pid->pid = owner;

		list_add(&pid->pid_list, &ip->i_flock_pids);
		up(&ip->i_flock_mux);
	} else {
		gfree(pid, sizeof(pid_lst_t));	/*didn't get it, so free the mem */
	}

exit_ogfs_flock:
	up(&ip->i_flock_lmx);
	EXIT(GFN_FLOCK);
	return error;
}

/**
 * ogfs_funlock - Release a flock on a file
 * @ip: <> the file
 * @owner: < An identifier of who owns this lock.
 *
 */
void
ogfs_funlock(ogfs_inode_t * ip, long owner)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	struct list_head *pidtmp, *pidhead;
	pid_lst_t *pid;

	ENTER(GFN_FUNLOCK);

	IFDB(GFN_FUNLOCK,
	     db_printf("unflocking %" FMT_64 "u owner = %lu\n",
		       ip->i_num.no_formal_ino, owner);
	    );

	/* if in list, unlock */
	down(&ip->i_flock_mux);
	if (!list_empty(&ip->i_flock_pids)) {
		/* someone here holds this lock. see if it is us. */
		pidtmp = pidhead = &ip->i_flock_pids;
		for (;;) {
			pidtmp = pidtmp->next;
			if (pidtmp == pidhead)
				break;	/* we don't hold it */
			pid = list_entry(pidtmp, pid_lst_t, pid_list);

			if (pid->pid == owner) {
				/* yep we hold it. drop our hold */
				list_del(pidtmp);
				gfree(pid, sizeof(pid_lst_t));
				/* anyone else left? */
				if (list_empty(&ip->i_flock_pids)) {
					/* nope, means we gotta clear the glock. */
					ogfs_gunlock_num(sdp,
							ip->i_num.no_formal_ino,
							&ogfs_flock_glops, NULL,
							GL_NOCACHE);
					ip->i_flock_st = 0;
				}
				break;
			}
		}		/* pid loop */
	}
	up(&ip->i_flock_mux);

	EXIT(GFN_FUNLOCK);
}
