/*
 *
 *    Copyright 1995-1999 Regents of the University of Minnesota
 *    Portions Copyright 2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 */

#include "ogfs_debug.h"
#include "ogfs.h"
#include "fs.h"

/**
 * ogfs_random - Generate a random 32-bit number
 *
 * Generate a semi-crappy 32-bit pseudo-random number without using
 * floating point.
 *
 * The PRNG is from "Numerical Recipes in C", page 284.
 *
 * Returns: a 32-bit random number
 */

uint32
ogfs_random(void)
{
	ENTER(GFN_RANDOM);

	ogfs_random_number =
	    (uint32) 1664525 *ogfs_random_number + (uint32) 1013904223;

	IFDB(GFN_RANDOM, db_printf("  %u\n", ogfs_random_number);
	    );

	EXIT(GFN_RANDOM);

	return ogfs_random_number;
}

/**
 * ogfs_randrange - Create a random number in a give range
 * @low: the smallest number that can be returned
 * @high: one more than the highest number that can be returned
 *
 * Use ogfs_random() to create a random number in a given range.
 *
 * Returns: the number
 */

uint32
ogfs_randrange(uint32 low, uint32 high)
{
	uint32 rand;
	uint32 scale;

	ENTER(GFN_RANDRANGE);

	scale = 0x80000000 / (high - low);

	rand = ogfs_random() >> 1;
	rand = (rand / scale) + low;

	IFDB(GFN_RANDRANGE,
	     db_printf("  low = %u, high = %u, scale = %u, rand = %u\n",
		       low, high, scale, rand);
	    );

	EXIT(GFN_RANDRANGE);

	return rand;
}

//#include <crc.h>
#include "crc.h"

/**
 * ogfs_hash - hash an array of data
 * @data: the data to be hashed
 * @len: the length of data to be hashed
 *
 * Take some data and convert it to a 32-bit hash.
 *
 * The hash function is a 32-bit CRC of the data.  The algorithm uses
 * the crc_32_tab table above.
 *
 * This may not be the fastest hash function, but it does a fair bit better
 * at providing uniform results than the others I've looked at.  That's
 * really important for efficient directories.
 *
 * Returns: the hash
 */

uint32
ogfs_hash(const char *data, int len)
{
	uint32 hash = 0xFFFFFFFF;

	ENTER(GFN_HASH);

	for (; len--; data++)
		hash = crc_32_tab[(hash ^ *data) & 0xFF] ^ (hash >> 8);

	hash = ~hash;

	EXIT(GFN_HASH);

	return (hash);
}

/* Byte-wise swap two items of size SIZE. */

#define SWAP(a, b, size) \
do \
{ \
  register size_t __size = (size); \
  register char *__a = (a), *__b = (b); \
  do \
  { \
    char __tmp = *__a; \
    *__a++ = *__b; \
    *__b++ = __tmp; \
  } \
  while (__size-- > 1); \
} \
while (0)

/**
 * ogfs_sort - Sort base array using shell sort algorithm
 * @base: the input array
 * @num_elem: number of elements in array
 * @size: size of each element in array
 * @compar: fxn to compare array elements (returns negative for lt, 0 for eq, and positive for gt
 *
 * Sorts the array passed in using the compar fxn to compare elements using
 * the shell sort algorithm
 */

int
ogfs_sort(void *base, unsigned int num_elem, unsigned int size,
	 int (*compar) (const void *, const void *))
{
	register char *pbase = (char *) base;
	int i, j, k, h;
	int cols[16] = {1391376, 463792, 198768, 86961, 33936, 13776, 4592,
			1968, 861, 336, 112, 48, 21, 7, 3, 1};

	ENTER(GFN_SORT);

	for (k=0; k<16; k++) {
		h=cols[k];
		for (i=h; i<=num_elem; i++) {
			j=i;
			while (j>=h && (*compar)((void *)(pbase+size*(j-h)), (void *)(pbase+size*j))>0) {
				SWAP(pbase+size*j,pbase+size*(j-h),size);
				j=j-h;
			}
		}
	}
	EXIT(GFN_SORT);

	return 0;
}
