/*
 *
 *    Copyright 1995-2000 Regents of the University of Minnesota
 *    Portions Copyright 1999-2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 */

/*
 * This module implements the OGFS interface to the global locks.
 */
#ifndef _OGFS_GLOCK_H
#define _OGFS_GLOCK_H

/*  Flags passed in to ogfs_glock().  These are passed down to the
    lock module in one way or another.

    GL_SHARED means that OGFS is giving permission for the lock layers to acquire
    the lock in a shared state.

                                  GL_SHARED
      Lock actually
      acquired as     |      Set      |      Not set
    ------------------+---------------+--------------------
         Shared       |      GOOD     |       BAD!!!!
    ------------------+---------------+--------------------
         Exclusive    |       OK      |       GOOD 

    GL_DEFERRED asks for the lock in the deferred state.  It's mutual exclusive
    with GL_SHARED.  If a lock is asked for in the deferred state, it must
    be acquired in the deferred state.

    GL_PERM lets the lock layer know that this lock will be held for a long time.

    GL_DISOWN tells the lock layer to disallow recursive locking and allows a
    different process to unlock the lock.

    GL_SKIP tells the go_lock() and go_unlock() calls to do nothing

    LM_FLAG_NOEXP means that if this lock is expired, ignore that fact -- Treat the 
    lock exactly as it would if it wasn't expired.  (And the lock
    module shouldn't send OGFS callbacks about finding this lock expired.)
    (This flag is probably ignored for a DLM-style lock module.)

    LM_FLAG_TRY means that the locking layers should only try to acquire the lock
    once.  If the lock can't be acquired immediately give up and return
    GLR_TRYFAILED.

    */

#define GL_SHARED         (0x00001000)
#define GL_DEFERRED       (0x00002000)
#define GL_PERM           (0x00004000)
#define GL_DISOWN         (0x00008000)
#define GL_SKIP           (0x00010000)
/*  Also LM_FLAG_TRY and LM_FLAG_NOEXP.  */

/*  Potential errors returned by ogfs_glock() when used with some particular flags.
    GLR_CANCELED is never seen outside of glock.c  */

#define GLR_TRYFAILED (13)
#define GLR_CANCELED  (14)

/*  Flags passed in to ogfs_gunlock()

    GL_SYNC causes all the data and metadata for this lock to be synced to
    disk before its released.  It doesn't do anything with transactions
    that have been commited incore.  In fact it's really dangerous to use
    this flag.  If you don't know why don't use it.

    GL_NOCACHE causes the lock layers to drop the lock immediately after
    it is unlocked.  It isn't cache (held) by the lock layers.

    */

#define GL_SYNC       (0x00100000)
#define GL_NOCACHE    (0x00200000)

struct ogfs_lockop {
	ogfs_glock_t *lo_lock;
	ogfs_inode_t *lo_ip;
	int lo_flags;
};
typedef struct ogfs_lockop ogfs_lockop_t;

/*  Lockname Functions  */

void ogfs_blk2lockname(ogfs_sbd_t * sdp, uint64 block, unsigned int type,
		      lm_lockname_t * name);
uint64 ogfs_lockname2blk(ogfs_sbd_t * sdp, lm_lockname_t * name);

/*  Glock structure functions  */

ogfs_glock_t *ogfs_get_glstruct(ogfs_sbd_t * sdp,
			      uint64 number, ogfs_glock_operations_t * glops,
			      ogfs_glock_t * parent, int create);
static __inline__ void
ogfs_hold_glstruct(ogfs_glock_t * gl)
{
	atomic_inc(&gl->gl_count);
}

void ogfs_put_glstruct(ogfs_glock_t * gl);

/*  Lock acquisition and release  */

int ogfs_glock(ogfs_sbd_t * sdp, ogfs_glock_t * gl, int flags);
void ogfs_gunlock(ogfs_sbd_t * sdp, ogfs_glock_t * gl, int flags);
void ogfs_force_gl_drop(ogfs_sbd_t * sdp, ogfs_glock_t * gl);

#define ogfs_glock_i(ip, flags) ogfs_glock((ip)->i_sbd, (ip)->i_gl, (flags))
#define ogfs_gunlock_i(ip, flags) ogfs_gunlock((ip)->i_sbd, (ip)->i_gl, (flags))

#define ogfs_glock_rg(rgd, flags) ogfs_glock((rgd)->rd_sbd, (rgd)->rd_gl, (flags))
#define ogfs_gunlock_rg(rgd, flags) ogfs_gunlock((rgd)->rd_sbd, (rgd)->rd_gl, (flags))

int ogfs_glock_num(ogfs_sbd_t * sdp,
		  uint64 number, ogfs_glock_operations_t * glops,
		  ogfs_glock_t * parent, int flags, ogfs_glock_t ** gl);
void ogfs_gunlock_num(ogfs_sbd_t * sdp, uint64 number,
		     ogfs_glock_operations_t * glops, ogfs_glock_t * parent,
		     int flags);

int ogfs_glock_m(ogfs_sbd_t * sdp, int num_lock, ogfs_lockop_t * ops);
void ogfs_gunlock_m(ogfs_sbd_t * sdp, int num_lock, ogfs_lockop_t * ops);

/*  Lock Value Block functions  */

int ogfs_hold_lvb(ogfs_sbd_t * sdp, ogfs_glock_t * gl);
void ogfs_unhold_lvb(ogfs_sbd_t * sdp, ogfs_glock_t * gl);
void ogfs_sync_lvb(ogfs_sbd_t * sdp, ogfs_glock_t * gl);

/*  Inter-glock dependencies  */

int ogfs_add_gl_dependency(ogfs_sbd_t * sdp, ogfs_glock_t * sync_this_gl,
			  ogfs_glock_t * before_releasing_this_gl);

/*  Only used in ogfs_mount_lockproto() (and the locking modules)  */

void ogfs_glock_cb(lm_fsdata_t * fsdata, unsigned int type, void *data);

/*  Crap  */

void ogfs_pitch_inodes(ogfs_sbd_t * sdp);
void ogfs_glockd_scan(ogfs_sbd_t * sdp, unsigned int timeout);
void ogfs_clear_gla(ogfs_sbd_t * sdp);

void ogfs_dump_lockstate(ogfs_sbd_t * sdp);

/*  These are architecture specific functions
    Either implement them or stub them out in your arch_* directory.  */

void ogfs_glock_arch_init(ogfs_sbd_t * sdp, ogfs_glock_t * gl);

#endif /* _OGFS_GLOCK_H */
