/*
 *
 *    Copyright 2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 */

#include "ogfs_debug.h"
#include "ogfs.h"
#include "blklist.h"
#include "dio.h"
#include "glock.h"
#include "inode.h"
#include "log.h"
#include "page.h"
#include "super.h"

/**
 * sync_meta - Sync the dirty data for a regular metadata glock
 * @sdp: the filesystem
 * @gl: the glock
 * @flags: OGFS_SYNC_WAIT or OGFS_SYNC_INVISIBLE
 *
 * Flush all incore committed transactions for the glock
 * Sync to disk all dirty buffers associated with glock
 * Clear glock's GLF_DIRTY and GLF_SYNC bits
 */

static void
sync_meta(ogfs_sbd_t * sdp, ogfs_glock_t * gl, int flags)
{
	ENTER(GFN_SYNC_META);

	if (test_bit(GLF_DIRTY, &gl->gl_flags)) {
		ogfs_log_flush(sdp, gl);
		ogfs_sync_bh(sdp, gl, (flags & OGFS_SYNC_WAIT));
	}
	if (!(flags & OGFS_SYNC_INVISIBLE)) {
		clear_bit(GLF_DIRTY, &gl->gl_flags);
		clear_bit(GLF_SYNC, &gl->gl_flags);
	}

	EXIT(GFN_SYNC_META);
}

/**
 * release_meta - prepare a regular metadata glock to be released
 *     to another node
 * @sdp: the filesystem
 * @gl: the glock
 *
 * Sync committed transactions and dirty data, clear GLF_DIRTY, GLF_SYNC.
 * Invalidate buffers so we don't read them while another node writes them
 */

static void
release_meta(ogfs_sbd_t * sdp, ogfs_glock_t * gl)
{
	ENTER(GFN_RELEASE_META);

	sync_meta(sdp, gl, OGFS_SYNC_WAIT);

	if (!sdp->sd_args.ar_localcaching) {
		ogfs_inval_bh(sdp, gl);
		gl->gl_vn++;
	}

	EXIT(GFN_RELEASE_META);
}

/**
 * sync_inode - Sync the dirty data for a inode glock
 * @sdp: the filesystem
 * @gl: the glock
 * @flags: OGFS_SYNC_WAIT or OGFS_SYNC_INVISIBLE
 *
 * Flush all incore committed transactions for the glock
 * Sync to disk all dirty buffers associated with glock
 * Sync to disk all dirty pages associated with glock
 * Clear glock's GLF_DIRTY and GLF_SYNC bits
 */

static void
sync_inode(ogfs_sbd_t * sdp, ogfs_glock_t * gl, int flags)
{
	ENTER(GFN_SYNC_INODE);

	if (test_bit(GLF_DIRTY, &gl->gl_flags)) {
		ogfs_log_flush(sdp, gl);
		ogfs_sync_pg(sdp, gl, (flags & OGFS_SYNC_WAIT));
		ogfs_sync_bh(sdp, gl, (flags & OGFS_SYNC_WAIT));
	}
	if (!(flags & OGFS_SYNC_INVISIBLE)) {
		clear_bit(GLF_DIRTY, &gl->gl_flags);
		clear_bit(GLF_SYNC, &gl->gl_flags);
	}

	EXIT(GFN_SYNC_INODE);
}

/**
 * release_inode - prepare a inode glock to be released
 * @sdp: the filesystem
 * @gl: the glock
 *
 * Sync committed transactions and dirty data, clear GLF_DIRTY, GLF_SYNC.
 * Invalidate buffers so we don't read them while another node writes them
 */

static void
release_inode(ogfs_sbd_t * sdp, ogfs_glock_t * gl)
{
	ogfs_inode_t *ip = (ogfs_inode_t *) gl->gl_object;

	ENTER(GFN_RELEASE_INODE);

	if (ip)
		ogfs_flush_meta_cache(ip, 0);

	sync_inode(sdp, gl, OGFS_SYNC_WAIT);

	if (!sdp->sd_args.ar_localcaching) {
		ogfs_inval_pg(sdp, gl);
		ogfs_inval_bh(sdp, gl);
		gl->gl_vn++;
	}

	EXIT(GFN_RELEASE_INODE);
}

/**
 * lock_inode - operation done after an inode lock is locked by a process
 * @sdp: the filesystem
 * @gl: the glock
 * @flags: the flags passed into ogfs_glock()
 *
 * Returns: 0 on success, -EXXX on failure
 *
 * If needed, read fresh copy of dinode structure from disk
 *
 */

static int
lock_inode(ogfs_sbd_t * sdp, ogfs_glock_t * gl, int flags)
{
	ogfs_inode_t *ip = (ogfs_inode_t *) gl->gl_object;
	int lockone = !(flags & GL_SKIP)
	    && (atomic_read(&gl->gl_locked) == 1);
	int error = 0;

	ENTER(GFN_LOCK_INODE);

	if (lockone && ip && ip->i_vn != gl->gl_vn)
		error = ogfs_copyin_dinode(ip);

	EXIT(GFN_LOCK_INODE);

	return error;
}

/**
 * unlock_inode - operation done before an inode lock is unlocked by a process
 * @sdp: the filesystem
 * @gl: the glock
 * @flags: the flags passed into ogfs_gunlock()
 *
 * Returns: 0 on success, -EXXX on failure
 *
 * If attributes have changed, copy them from dinode to VFS inode structure
 */

static void
unlock_inode(ogfs_sbd_t * sdp, ogfs_glock_t * gl, int flags)
{
	ogfs_inode_t *ip = (ogfs_inode_t *) gl->gl_object;
	int lockone = !(flags & GL_SKIP)
	    && (atomic_read(&gl->gl_locked) == 1);

	ENTER(GFN_UNLOCK_INODE);

	if (test_and_clear_bit(GLF_POISONED, &gl->gl_flags))
		gl->gl_vn++;
	else if (lockone && ip && test_bit(GLF_DIRTY, &gl->gl_flags))
		ogfs_inode_attr_in(ip);

	EXIT(GFN_UNLOCK_INODE);
}

/**
 * acquire_rgrp - operation done after a rgrp lock is acquired
 * @sdp: the filesystem
 * @gl: the glock
 *
 * Copy rgrp usage statistics from lock's Lock Value Block
 *    (passed from other node).
 *
 */

static void
acquire_rgrp(ogfs_sbd_t * sdp, ogfs_glock_t * gl)
{
	ogfs_rgrpd_t *rgd = (ogfs_rgrpd_t *) gl->gl_object;

	ENTER(GFN_ACQUIRE_RGRP);

	OGFS_ASSERT(rgd && gl->gl_lvb,);

	ogfs_rgrp_save_in(&rgd->rd_rs, gl->gl_lvb);

	EXIT(GFN_ACQUIRE_RGRP);
}

/**
 * release_rgrp - prepare a rgrp glock to be released
 * @sdp: the filesystem
 * @gl: the glock
 *
 * Sync committed transactions and dirty data, clear GLF_DIRTY, GLF_SYNC.
 * Invalidate buffers so we don't read them while another node writes them.
 * Copy rgrp usage statistics to lock's Lock Value Block
 *    (to pass to other node).
 */

static void
release_rgrp(ogfs_sbd_t * sdp, ogfs_glock_t * gl)
{
	ogfs_rgrpd_t *rgd = (ogfs_rgrpd_t *) gl->gl_object;

	ENTER(GFN_RELEASE_RGRP);

	OGFS_ASSERT(rgd && gl->gl_lvb,);

	sync_meta(sdp, gl, OGFS_SYNC_WAIT);

	if (!sdp->sd_args.ar_localcaching) {
		ogfs_inval_bh(sdp, gl);
		gl->gl_vn++;
	}

	if (rgd->rd_rs_dirty) {
		rgd->rd_rs.rs_owner = 0;
		ogfs_rgrp_save_out(&rgd->rd_rs, gl->gl_lvb);

		rgd->rd_rs_dirty = FALSE;
	}

	EXIT(GFN_RELEASE_RGRP);
}

/**
 * lock_rgrp - operation done after an rgrp lock is locked by a process
 * @sdp: the filesystem
 * @gl: the glock
 * @flags: the flags passed into ogfs_glock()
 *
 * Returns: 0 on success, -EXXX on failure
 *
 * If needed, read fresh copy of rgrp's block usage bitmap from disk
 *
 */

static int
lock_rgrp(ogfs_sbd_t * sdp, ogfs_glock_t * gl, int flags)
{
	ogfs_rgrpd_t *rgd = (ogfs_rgrpd_t *) gl->gl_object;
	int lockone = !(flags & GL_SKIP)
	    && (atomic_read(&gl->gl_locked) == 1);
	int error = 0;

	ENTER(GFN_LOCK_RGRP);

	OGFS_ASSERT(rgd && gl->gl_lvb,);

	if (lockone)
		error = ogfs_rgrp_read(rgd);

	EXIT(GFN_LOCK_RGRP);

	return error;
}

/**
 * unlock_rgrp - operation done before an rgrp lock is unlocked by a process
 * @sdp: the filesystem
 * @gl: the glock
 * @flags: the flags passed into ogfs_gunlock()
 *
 * Returns: 0 on success, -EXXX on failure
 *
 * brelse() rgrp's block usage bitmap, so kernel can write to disk.
 * Copy rgrp's usage statistics to Lock Value Block structure.
 *
 */

static void
unlock_rgrp(ogfs_sbd_t * sdp, ogfs_glock_t * gl, int flags)
{
	ogfs_rgrpd_t *rgd = (ogfs_rgrpd_t *) gl->gl_object;
	int lockone = !(flags & GL_SKIP)
	    && (atomic_read(&gl->gl_locked) == 1);

	ENTER(GFN_UNLOCK_RGRP);

	OGFS_ASSERT(rgd && gl->gl_lvb,);

	if (lockone)
		ogfs_rgrp_relse(rgd);

	if (test_and_clear_bit(GLF_POISONED, &gl->gl_flags))
		gl->gl_vn++;
	else if (lockone && test_bit(GLF_DIRTY, &gl->gl_flags))
		ogfs_rgrp_lvb_fill(rgd);

	EXIT(GFN_UNLOCK_RGRP);
}

/**
 * release_trans - prepare the transaction glock to be released
 * @sdp: the filesystem
 * @gl: the glock
 *
 * Flush all incore committed transactions for the glock.
 * Sync everything (not just this glock) to disk.
 *
 * We want to sync the device even with localcaching.  Remember
 * that localcaching journal replay only marks buffers dirty.
 */

static void
release_trans(ogfs_sbd_t *sdp, ogfs_glock_t *gl)
{
	ENTER(GFN_RELEASE_TRANS);

	ogfs_log_flush(sdp, NULL);
	sync_blockdev(sdp->sd_vsb->s_bdev);
/*	if (fsync_no_super(sdp->sd_vsb->s_dev))
		OGFS_ASSERT(0,
			   con_printf("Panicking because of write error\n"););
*/
	EXIT(GFN_RELEASE_TRANS);
}

ogfs_glock_operations_t ogfs_meta_glops = {
	go_sync:	sync_meta,
	go_release:	release_meta,
	go_free:	ogfs_free_bh,
	go_type:	LM_TYPE_META,
	go_name:	"Misc On-Disk Metadata"
};

ogfs_glock_operations_t ogfs_inode_glops = {
	go_sync:	sync_inode,
	go_release:	release_inode,
	go_lock:	lock_inode,
	go_unlock:	unlock_inode,
	go_free:	ogfs_free_bh,
	go_type:	LM_TYPE_INODE,
	go_name:	"Inode"
};

ogfs_glock_operations_t ogfs_rgrp_glops = {
	go_sync:	sync_meta,
	go_acquire:	acquire_rgrp,
	go_release:	release_rgrp,
	go_lock:	lock_rgrp,
	go_unlock:	unlock_rgrp,
	go_free:	ogfs_free_bh,
	go_type:	LM_TYPE_RGRP,
	go_name:	"Inode"
};

ogfs_glock_operations_t ogfs_nopen_glops = {
	go_type:	LM_TYPE_NOPEN,
	go_name:	"N-Open"
};

ogfs_glock_operations_t ogfs_flock_glops = {
	go_type:	LM_TYPE_FLOCK,
	go_name:	"Flock"
};

ogfs_glock_operations_t ogfs_plock_head_glops = {
	go_type:	LM_TYPE_PLOCK_HEAD,
	go_name:	"Plock head"
};

ogfs_glock_operations_t ogfs_plock_glops = {
	go_type:	LM_TYPE_PLOCK,
	go_name:	"Plock"
};

ogfs_glock_operations_t ogfs_nondisk_glops = {
	go_type:	LM_TYPE_NONDISK,
	go_name:	"Non-Disk"
};

ogfs_glock_operations_t ogfs_trans_glops = {
	go_release:	release_trans,
	go_type:	LM_TYPE_NONDISK,
	go_name:	"Trans-Lock"
};
