/*
 *
 *    Copyright 1995-1999 Regents of the University of Minnesota
 *    Portion Copyright 2000-2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 */

/*
 * gmalloc accounting code
 *
 * a normal malloc call from ogfs goes like
 * osi_gmalloc_wrapper -> gmalloc_update_stats -> gmalloc_record_brute_force
 *
 * a malloc call allocating a gmalloc_call_t goes like
 * osi_gmalloc_wrapper -> gmalloc_update_stats
 */

#include "ogfs_debug.h"
#include "ogfs.h"
#include "fs.h"
#include "gmalloc.h"

#ifdef DEBUG_MEM

struct gmalloc_call {
	struct list_head list;

	void *p;
	unsigned int bytes;

	char *file;
	unsigned int line;
};
typedef struct gmalloc_call gmalloc_call_t;

/**
 * gmalloc_hash - Gmalloc hash function
 * @ptr: Pointer to hash
 *
 * Returns: hash of pointer
 */

static __inline__ uint32
gmalloc_hash(void *ptr)
{
	uint32 val;

	val = ogfs_hash((char *) &ptr, sizeof(void *));
	val &= (GMALLOC_HASH_SIZE - 1);

	return val;
}

/**
 * gmalloc_record_brute - Record stats
 * @data: Pointer to data
 * @size:
 * @file:
 * @line:
 *
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
gmalloc_record_brute(void *data, unsigned int size, char *file,
		     unsigned int line)
{
	gmalloc_call_t *gm;
	unsigned int index = gmalloc_hash(data);

	ENTER(GFN_GMALLOC_RECORD_BRUTE);

	gm = (gmalloc_call_t *) osi_malloc(sizeof(gmalloc_call_t));
	if (!gm) {
		EXIT(GFN_GMALLOC_RECORD_BRUTE);
		return -ENOMEM;
	}

	gm->p = data;
	gm->bytes = size;
	gm->file = file;
	gm->line = line;

	spin_lock(&gmalloc_lock);
	list_add(&gm->list, &gmalloc_head[index]);
	spin_unlock(&gmalloc_lock);

	EXIT(GFN_GMALLOC_RECORD_BRUTE);

	return 0;
}

/**
 * gfree_record_brute - Record stats
 * @data: Pointer to data
 * @size:
 * @file:
 * @line:
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
gfree_record_brute(void *data, unsigned int size, char *file, unsigned int line)
{
	unsigned int index = gmalloc_hash(data);
	struct list_head *tmp;
	gmalloc_call_t *gm = NULL;

	ENTER(GFN_GFREE_RECORD_BRUTE);

	spin_lock(&gmalloc_lock);

	for (tmp = gmalloc_head[index].next; tmp != &gmalloc_head[index];
	     tmp = tmp->next) {
		gm = list_entry(tmp, gmalloc_call_t, list);

		if (gm->p == data) {
			if (gm->bytes == size)
				list_del(&gm->list);
			break;
		} else
			gm = NULL;
	}

	spin_unlock(&gmalloc_lock);

	if (!gm) {
		con_printf("OGFS:  freeing non-malloced (%u, %s, %u)\n", size,
			   file, line);
		EXIT(GFN_GFREE_RECORD_BRUTE);
		return -EINVAL;
	}

	if (gm->bytes != size) {
		con_printf
		    ("OGFS:  freeing wrong size:  malloced = (%u, %s, %u), freed = (%u, %s, %u)\n",
		     gm->bytes, gm->file, gm->line, size, file, line);
		EXIT(GFN_GFREE_RECORD_BRUTE);
		return -EINVAL;
	}

	osi_free(gm, sizeof(gmalloc_call_t));

	EXIT(GFN_GFREE_RECORD_BRUTE);

	return 0;
}

/**
 * gmalloc_set_garbage - Write a pattern into a region of memory
 * @data: The region
 * @size: The size of the region
 *
 */

static __inline__ void
gmalloc_set_garbage(void *data, unsigned int size)
{
	memset(data, 0x69, size);
}

/**
 * gfree_set_garbage - Write a pattern into a region of memory
 * @data: The region to write to
 * @size: The size to write
 *
 */

static __inline__ void
gfree_set_garbage(void *data, unsigned int size)
{
	memset(data, 0x96, size);
}

/**
 * gmalloc_update_stats - Update statistics
 * @size: Size of allocated region
 *
 */

static void
gmalloc_update_stats(unsigned int size)
{
	ENTER(GFN_GMALLOC_UPDATE_STATS);

	spin_lock(&gmalloc_lock);

	gm_stats.ct++;
	gm_stats.ct_total++;

	if (gm_stats.ct > gm_stats.ct_max)
		gm_stats.ct_max = gm_stats.ct;

	gm_stats.bytes += size;
	gm_stats.bytes_total += size;

	if (gm_stats.bytes > gm_stats.bytes_max)
		gm_stats.bytes_max = gm_stats.bytes;

	spin_unlock(&gmalloc_lock);

	EXIT(GFN_GMALLOC_UPDATE_STATS);
}

/**
 * gfree_update_stats - Update stats
 * @size: Size of the region being freed
 *
 */

static void
gfree_update_stats(unsigned int size)
{
	ENTER(GFN_GFREE_UPDATE_STATS);

	spin_lock(&gmalloc_lock);

	gm_stats.ct--;
	gm_stats.bytes -= size;

	spin_unlock(&gmalloc_lock);

	EXIT(GFN_GFREE_UPDATE_STATS);
}

/**
 * gmalloc_wrapper - Wrapper for malloc
 * @size: Size of region
 * @file:
 * @line:
 *
 * Returns: The requested memory or NULL
 */

void *
gmalloc_wrapper(unsigned int size, char *file, unsigned int line)
{
	void *data;
	int error;

	ENTER(GFN_GMALLOC_WRAPPER);

	data = osi_malloc(size);
	if (!data) {
		EXIT(GFN_GMALLOC_WRAPPER);
		return NULL;
	}

	error = gmalloc_record_brute(data, size, file, line);
	if (error) {
		osi_free(data, size);
		EXIT(GFN_GMALLOC_WRAPPER);
		return NULL;
	}

	gmalloc_update_stats(size);
	gmalloc_set_garbage(data, size);	/*  flush out initialization bugs  */

	EXIT(GFN_GMALLOC_WRAPPER);

	return data;
}

/**
 * gfree_wrapper - Wrapper for free
 * @data: The memory to free
 * @size: The size of the area
 * @file:
 * @line:
 *
 */

void
gfree_wrapper(void *data, unsigned int size, char *file, unsigned int line)
{
	int error;

	ENTER(GFN_GFREE_WRAPPER);

	if (!data) {
		con_printf("OGFS:  freeing NULL pointer (%u, %s, %u)\n", size,
			   file, line);
		EXIT(GFN_GFREE_WRAPPER);
		return;
	}

	error = gfree_record_brute(data, size, file, line);
	if (error) {
		EXIT(GFN_GFREE_WRAPPER);
		return;
	}

	gfree_set_garbage(data, size);
	gfree_update_stats(size);

	osi_free(data, size);

	EXIT(GFN_GFREE_WRAPPER);
}

/**
 * gmalloc_check - Checks memory allocations at module unload
 *
 */

void
gmalloc_check()
{
	gmalloc_call_t *gm;
	unsigned int index;
	unsigned int ct = 0;

	ENTER(GFN_GMALLOC_CHECK);

	spin_lock(&gmalloc_lock);

	for (index = 0; index < GMALLOC_HASH_SIZE; index++)
		while (!list_empty(&gmalloc_head[index])) {
			gm = list_entry(gmalloc_head[index].next,
					    gmalloc_call_t, list);
			list_del(&gm->list);

			spin_unlock(&gmalloc_lock);

			con_printf("OGFS:  unfreed (%u, %s, %u)\n", gm->bytes,
				   gm->file, gm->line);

			OGFS_ASSERT(gm->p,);
			osi_free(gm->p, gm->bytes);

			osi_free(gm, sizeof(gmalloc_call_t));

			ct++;

			spin_lock(&gmalloc_lock);
		}

	spin_unlock(&gmalloc_lock);

	con_printf("OGFS:  # of unfreed mallocs: %u\n", ct);

	con_printf("OGFS:  gmalloc_ct:          %u\n", gm_stats.ct);
	con_printf("OGFS:  gmalloc_ct_max:      %u\n", gm_stats.ct_max);
	con_printf("OGFS:  gmalloc_ct_total:    %u\n", gm_stats.ct_total);
	con_printf("OGFS:  gmalloc_bytes:       %u\n", gm_stats.bytes);
	con_printf("OGFS:  gmalloc_bytes_max:   %u\n", gm_stats.bytes_max);
	con_printf("OGFS:  gmalloc_bytes_total: %" FMT_64 "u\n",
		   gm_stats.bytes_total);

	EXIT(GFN_GMALLOC_CHECK);
}

#endif				/*  DEBUG_MEM  */
