/*
 *
 *    Copyright 1995-2000 Regents of the University of Minnesota
 *    Portions Copyright 1999-2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 */

/*
 * The OGFS in-core structures.
 */
#ifndef _INCORE_H
#define _INCORE_H

/*  Generic Constants  */

#define OGFS_CLIENTS             (1024)
#define OGFS_CLIENT_BYTES        (OGFS_CLIENTS >> 3)

/*  Structure prototypes  */

struct task_struct;

struct ogfs_bitmap;
struct ogfs_rgrpd;
struct ogfs_bufdata;
struct ogfs_gl_depend;
struct ogfs_glock_operations;
struct ogfs_glock;
struct ogfs_ipreserv;
struct ogfs_inode;
struct ogfs_tune_pr;
struct ogfs_tr_nop;
struct ogfs_trans;
struct ogfs_nopen_block;
struct ogfs_bhlist;
struct ogfs_ghash_bucket;
struct ogfs_sbd;

typedef struct ogfs_bitmap ogfs_bitmap_t;
typedef struct ogfs_rgrpd ogfs_rgrpd_t;
typedef struct ogfs_bufdata ogfs_bufdata_t;
typedef struct ogfs_gl_depend ogfs_gl_depend_t;
typedef struct ogfs_glock_operations ogfs_glock_operations_t;
typedef struct ogfs_glock ogfs_glock_t;
typedef struct ogfs_ipreserv ogfs_ipreserv_t;
typedef struct ogfs_inode ogfs_inode_t;
typedef struct ogfs_tune_pr ogfs_tune_pr_t;
typedef struct ogfs_tr_nop ogfs_tr_nop_t;
typedef struct ogfs_trans ogfs_trans_t;
typedef struct ogfs_nopen_block ogfs_nopen_block_t;
typedef struct ogfs_bhlist ogfs_bhlist_t;
typedef struct ogfs_ghash_bucket ogfs_ghash_bucket_t;
typedef struct ogfs_sbd ogfs_sbd_t;

#include "incore_arch.h"

#if !defined(__KERNEL__)
typedef void *wait_queue_head_t;
#endif

/*
 *  Structure containing information about the allocation bitmaps.
 *  There are one of these for each fs block that the bitmap for
 *  the resource group header covers.
 */

struct ogfs_bitmap {
	uint32 bi_offset;	/* The offset in the buffer of the first byte */
	uint32 bi_start;	/* The position of the first byte in this block */
	uint32 bi_len;		/* The number of bytes in this block */
};

/*
 *  Structure containing information Resource Groups
 */

#define RDF_SYNC_B4_ALLOC       (0)

struct ogfs_rgrpd {
	struct list_head rd_list;	/* Link with superblock */

	struct ogfs_rgrpd *rd_avl_left;
	struct ogfs_rgrpd *rd_avl_right;
	unsigned rd_avl_height;

	atomic_t rd_count;	/* Reference count */
	ogfs_glock_t *rd_gl;	/* Glock for rgrp */

	unsigned long rd_flags;

	ogfs_rindex_t rd_ri;	/* Resource Index structure */
	ogfs_rgrp_t rd_rg;	/* Resource Group structure */

	int rd_rs_dirty;	/* TRUE if the rd_rs needs to be synced out */
	ogfs_rgrp_save_t rd_rs;	/* Resource Group LVB structure */

	ogfs_bitmap_t *rd_bits;
	struct buffer_head **rd_bh;

	ogfs_sbd_t *rd_sbd;
};

/*
 *  Per-buffer data
 */

struct ogfs_bufdata {
	struct buffer_head *bd_bh;	/* struct buffer_head which this struct belongs to */
	ogfs_glock_t *bd_gl;	/* Pointer to Glock struct for this bh */

	struct list_head bd_new_list;	/* List of buffers in this new transaction */
	struct list_head bd_incore_list;	/* List of buffers in this incore transaction */

	char *bd_frozen;
	struct semaphore bd_lock;

	ogfs_bufdata_arch_t bd_arch;
};

/*
 *  Glock dependencies
 */

struct ogfs_gl_depend {
	struct list_head gd_list;

	lm_lockname_t gd_name;
	ogfs_glock_t *gd_parent;

	osi_clock_ticks_t gd_time;
};

/*
 *  Glock operations
 */

#define OGFS_SYNC_WAIT           (1)
#define OGFS_SYNC_INVISIBLE      (2)

struct ogfs_glock_operations {
	void (*go_sync) (ogfs_sbd_t * sdp, ogfs_glock_t * gl, int flags);

	void (*go_acquire) (ogfs_sbd_t * sdp, ogfs_glock_t * gl);
	void (*go_release) (ogfs_sbd_t * sdp, ogfs_glock_t * gl);

	int (*go_lock) (ogfs_sbd_t * sdp, ogfs_glock_t * gl, int flags);
	void (*go_unlock) (ogfs_sbd_t * sdp, ogfs_glock_t * gl, int flags);

	void (*go_free) (ogfs_sbd_t * sdp, ogfs_glock_t * gl);

	int go_type;
	char *go_name;
};

/*
 *  Glock Structure  (One per active Glock)
 */

/*
 * Glock flags. Can be changed atomically with bit setting functions, but
 * _only_ one at a time.
 */

#define GLF_HELD                (0)
#define GLF_PERM                (1)

#define GLF_LOCK                (2)

#define GLF_SYNC                (3)
#define GLF_DIRTY               (4)
#define GLF_POISONED            (5)

#define GLF_RELEXCL             (6)
#define GLF_RELDFRD		(7)
#define GLF_RELSHRD		(8)

#define GLF_STICKY              (9)
#define GLF_DEMOTEME            (10)

#define GLF_DEPCHECK            (11)

#define GLF_RECOVERY            (12)

/*
 * Explanations of the ogfs_glock fields:
 *
 * gl_list   - Hash table hook
 * gl_name   - Unique name for lock
 * gl_count  - Reference count of ogfs_glock
 * gl_flags  - Flags (bit definitions as above)
 * gl_bucket - Hash bucket that we inhabit
 * gl_head_lock - Spinlock which covers fields marked: (*)
 * gl_wait_lock - Wait for exclusive access to glock (see GLF_LOCK)
 * gl_wait_unlock - Wait for glock state to be LM_ST_UNLOCKED
 * gl_pid - PID of locking process (*)
 * gl_locked - Count of processes locking this glock (recursively) (*)
 * gl_ops - Operations which get called at various times
 * gl_state - The state of the lock
 *            LM_ST_UNLOCKED - Unlocked
 *            LM_ST_EXCLUSIVE - Enclusive lock
 *            LM_ST_SHARED - Shared lock
 *            LM_ST_DEFERRED - Another shared lock mode which doesn't share with
 *                             the LM_ST_SHARED mode.
 *
 * gl_lock - Private data for the lock module
 * gl_parent - The locks parent lock (if one exists)
 * gl_children - List of children of this lock
 * gl_lvb_count - Number of LVB references held on this glock
 * gl_lvb - LVB data attached by lock module
 * gl_vn - Version number (incremented when cache is not valid any more)
 * gl_stamp - Time of create or last unlock
 * gl_object - The object the glock is protecting
 *
 * gl_new_list - List of glocks in current trans
 * gl_new_bufs - List of buffers in current trans
 * gl_trans_t - The current transaction
 * 
 * gl_incore_list - List of glocks in incore transaction
 * gl_incore_bufs - List of buffers in incore transaction
 * gl_incode_tr - The incore transaction
 *
 * gl_num_dep - The number of glocks that need to be synced before this one
 *              can be released
 * gl_depend - The list of glocks that need to be synced before this one
 *              can be released
 *
 * gl_arch - Arch-specific (kernel vs. user-space) data.  Kernel code uses
 *           this for list of buffers covered by this glock.
 */

struct ogfs_glock {
	struct list_head gl_list;	/* Pointers to other ogfs_glock_t */
	lm_lockname_t gl_name;	/* Name of the lock */
	atomic_t gl_count;	/* Usage Counter */
	unsigned long gl_flags;	/* Flags */
	unsigned int gl_bucket;	/* Hash bucket we are in */

	spinlock_t gl_head_lock;	/* Spinlock covering various fields */
	wait_queue_head_t gl_wait_lock;	/* Wait to lock glock */
	wait_queue_head_t gl_wait_unlock;/* Wait for glock to be unlocked */

	long gl_pid;		/* Process ID of the locking process */
	atomic_t gl_locked;	/* Glock recursive lock count */
	ogfs_glock_operations_t *gl_ops;	/* Glock operations */

	unsigned int gl_state;	/* The state in which the lock is held */
	lm_lock_t *gl_lock;	/* Private data for lock module */

	struct ogfs_glock *gl_parent;	/* This lock's parent (or NULL) */
	struct list_head gl_parlist;	/* Parent's list of children */
	struct list_head gl_children;	/* This lock's children */

	unsigned int gl_lvb_count;	/* LVB reference count */
	lm_lvb_t *gl_lvb;	/* LVB data */

	uint64 gl_vn;		/* Version number */
	osi_clock_ticks_t gl_stamp;	/* Time of create or last unlock */
	void *gl_object;	/* The object the glock is protecting */

	/*  Modified under the glock  (i.e. gl_locked > 0)  */

	struct list_head gl_new_list;	/* List of glocks in trans currently being built */
	struct list_head gl_new_bufs;	/* List of buffers for this lock in new trans */
	ogfs_trans_t *gl_new_tr;	/* The new transaction */

	/*  Modified under the log lock  */

	struct list_head gl_incore_list;	/* List of glocks in incore trans */
	struct list_head gl_incore_bufs;	/* List of buffers for this lock in incore trans */
	ogfs_trans_t *gl_incore_tr;	/* The incore transaction */

	atomic_t gl_num_dep;	/* Number of dependent Glocks */
	struct list_head gl_depend;	/* List of dependent Glocks */

	ogfs_glock_arch_t gl_arch;
};

/*
 *  In-Place Reservation structure
 */

struct ogfs_ipreserv {
	/*  Filled in by the caller to ogfs_inplace_reserv()  */

	uint32 ir_diblks;	/* Number of dinodes to reserve */
	uint32 ir_metablks;	/* Number of metadata blocks to reserve */
	uint32 ir_datablks;	/* Number of datadata blocks to reserve */
	uint64 ir_target;	/* this is the target RG  */
	int ir_trans;		/* transaction type (for debugging) */

	/*  Filled in by ogfs_inplace_reserv()  */

	ogfs_rgrpd_t *ir_rgd;
	uint32 ir_metares;
	uint32 ir_datares;

	/*  Filled in by ogfs_blkalloc()  */

	uint32 ir_dialloced;
	uint32 ir_metaalloced;
	uint32 ir_dataalloced;
};

/*
 *  Incore inode structure
 */

#define GIF_CREATED             (0)

struct ogfs_inode {
	ogfs_inum_t i_num;

	atomic_t i_count;	/* Usage count */
	unsigned long i_flags;

	uint64 i_vn;		/* Clock associated with the glock */

	ogfs_dinode_t i_di;	/* Dinode Structure */
	ogfs_glock_t *i_gl;	/* Glock pointer */

	ogfs_glock_t *i_plock_gl;	/* plock head lock */
	struct semaphore i_plock_mux;	/* one pid at a time */
	atomic_t i_plock_cnt;	/* plock usage */

	struct semaphore i_flock_mux;	/* one pid at a time in list */
	struct semaphore i_flock_lmx;	/* one pid at a time in lock */
	int i_flock_st;		/* current flock state */
	struct list_head i_flock_pids;	/* list of pids that have this flock */

	ogfs_sbd_t *i_sbd;	/* OGFS superblock pointer */
	struct inode *i_vino;	/* FS's device independent ino */

	ogfs_ipreserv_t i_res;
	struct buffer_head *i_cache[OGFS_MAX_META_HEIGHT];
};

/*
 *  Transaction structures
 */

struct ogfs_tr_nop {
	struct list_head nop_list;

	ogfs_inum_t nop_inum;	/* The inode number */
	unsigned int nop_type;	/* OGFS_LOG_DESC_NOPOPEN, or OGFS_LOG_DESC_NOPCLOSE */
};

#define TRF_POISONED            (0x00000001)
#define TRF_NOPEN_DUMP          (0x00000002)

struct ogfs_trans {
	struct list_head tr_list;	/* List of transactions */

	/* Initial creation stuff */

	unsigned int tr_type;	/* The type of the transaction TRANS_CREATE, ...  */

	unsigned int tr_blks_asked;	/* Number of log blocks asked to be reserved */
	unsigned int tr_seg_reserved;	/* Number of segments reserved */

	/* Stuff filled in during creation */

	unsigned int tr_flags;
	unsigned int tr_num_gl;	/* Number of Glocks in the trans */
	unsigned int tr_num_nop_o;	/* Number of Nopen opens in the trans */
	unsigned int tr_num_nop_c;	/* Number of Nopen closes in the trans */
	unsigned int tr_num_buf;	/* Number of buffers in the trans */

	struct list_head tr_gl;	/* List of glocks in this trans */
	struct list_head tr_nop_change;	/* List of nopen changes in this trans */
	struct list_head tr_nop_change_ic;

	/* Stuff modified during the commit */

	uint64 tr_log_head;	/* The current log head */
	uint64 tr_first_head;	/* First header block */

	struct list_head tr_bhlist;	/* List of buffers going to the log */

	ogfs_trans_arch_t tr_arch;
};

#define BLF_FAKE                (0x00000001)

struct ogfs_bhlist {
	struct list_head bl_list;

	struct buffer_head *bl_bh;
	struct buffer_head *bl_unlock;

	int bl_flags;
};

struct ogfs_ghash_bucket {
	atomic_t hb_notheld_count;
	atomic_t hb_held_count;
	atomic_t hb_perm_count;

	struct list_head hb_notheld;	/* !GLF_HELD */
	struct list_head hb_held;	/* GLF_HELD */
	struct list_head hb_perm;	/* (gl_locked > 0) for a long time */

	rwlock_t hb_lock;
}
#if defined(__KERNEL__)
__attribute__ ((__aligned__(SMP_CACHE_BYTES)))
#endif
    ;

/*
 *  Super Block Data Structure  (One per filesystem)
 */

#define OGFS_GHASH_SHIFT         (9)
#define OGFS_GHASH_SIZE          (1 << OGFS_GHASH_SHIFT)
#define OGFS_GHASH_MASK          (OGFS_GHASH_SIZE - 1)

struct ogfs_sbd {
	ogfs_sb_t sd_sb;		/* Super Block */
	struct super_block *sd_vsb;	/* FS's device independent sb */

	ogfs_args_t sd_args;

	/* Resource group stuff */

	ogfs_inode_t *sd_riinode;	/* rindex inode */
	uint64 sd_riinode_vn;	/* Version number of the resource index inode */

	struct list_head sd_rglist;	/* List of resource groups */
	ogfs_rgrpd_t *sd_rgavl;	/* AVL tree of resource groups */
	ogfs_rgrpd_t *sd_rgcache;	/* Cache of previous lookup */
	unsigned int sd_rgcount;	/* Count of resource groups */
	rwlock_t sd_rglock;	/* Lock upon list & AVL tree, etc */

	/*  Constants computed on mount  */

	uint32 sd_fsb2bb_shift;	/* Shift FS Block numbers to the left by
				   this to get buffer cache blocks  */
	uint32 sd_diptrs;	/* Number of pointers in a dinode */
	uint32 sd_inptrs;	/* Number of pointers in a indirect block */
	uint32 sd_jbsize;	/* Size of a journaled data block */
	uint32 sd_hash_bsize;	/* sizeof(exhash block) */
	uint32 sd_hash_ptrs;	/* Number of points in a hash block */
	uint32 sd_max_dirres;	/* Maximum space needed to add a directory entry */
	uint32 sd_max_height;	/* Maximum height of a file's metadata tree */

	/*  Mount flags  */

	int sd_noatime;
	int sd_sync_mount;

	/*  Lock Stuff  */

	ogfs_ghash_bucket_t sd_ghash[OGFS_GHASH_SIZE];

	unsigned int sd_gl_heldtime;	/* Time to cache locks for */

	struct lm_lockstruct sd_lockstruct;

	/*  Inode Stuff  */

	ogfs_inode_t *sd_rooti;	/* FS's root inode */

	atomic_t sd_free_icount;	/* Approx free inodes hanging around */
	uint64 sd_dir_alloc_target;	/* The address of the RG to allocate dirs from */
	osi_clock_ticks_t sd_shrink_time;

	ogfs_glock_t *sd_rename_gl;	/* rename glock */

	/*  Daemon stuff  */

	struct task_struct *sd_glockd_process;
	atomic_t sd_glockd_ctrl;
	struct semaphore sd_glockd_sema;

	struct task_struct *sd_inoded_process;
	atomic_t sd_inoded_ctrl;
	struct semaphore sd_inoded_sema;

	struct task_struct *sd_recoverd_process;
	atomic_t sd_recoverd_ctrl;
	struct semaphore sd_recoverd_sema;

	/*  Log stuff  */

	ogfs_glock_t *sd_trans_gl;	/* transaction glock */

	ogfs_inode_t *sd_jiinode;	/* jindex inode */
	uint64 sd_jiinode_vn;	/* Version number of the journal index inode */

	unsigned int sd_journals;	/* Number of journals in the FS */
	ogfs_jindex_t *sd_jindex;	/* Array of Jindex structures describing FS's journals */
	uint64 sd_jext_base;		/* base block # for external journals */
	struct block_device **sd_jdevs;	/* Array of ptrs to FS' external journal devs (if any) */

	ogfs_jindex_t sd_my_jdesc;	/* Jindex structure describing this machine's journal */
	struct block_device * sd_my_jdev;	/* block dev for this machine's external journal (if any) */
	ogfs_glock_t *sd_my_jnl_gl;	/* the glock for this machine's journal */

	uint64 sd_sequence;	/* Assigned to xactions in order they commit */
	uint64 sd_log_head;	/* Block number of next journal write */

	unsigned int sd_log_seg_free;	/* Free segments in the log */
	struct list_head sd_log_ail;	/* ogfs_trans_t structures that form the Active Items List 
					   "next" is the head, "prev" is the tail  */

	struct list_head sd_log_incore;	/* transactions that have been commited incore (but not ondisk)
					   "next" is the newest, "prev" is the oldest  */
	unsigned int sd_log_buffers;	/* Number of buffers in the incore log */

	struct semaphore sd_log_lock;	/* Lock for access to log values */

	struct list_head sd_log_bhlist_cache;

	/*  Nopen crap  */

	struct list_head sd_nopen_od_list;	/* List of inodes waiting to be deallocated */
	atomic_t sd_nopen_od_count;	/* Count of inodes waiting to be deallocated */

	struct list_head sd_nopen_ic_list;
	atomic_t sd_nopen_ic_count;	/* Like sd_nopen_count, but not
					   including deallocations already commited incore */

	uint64 sd_last_nopd;
	int sd_in_nopen_dump;

	/* Recovery stuff */

	unsigned long sd_dirty_j[OGFS_CLIENT_BYTES / sizeof(unsigned long)];
};

#define ogfs_meta_check(bh) \
do \
{ \
  uint32 meta_check_magic = ((ogfs_meta_header_t *)BH_DATA((bh)))->mh_magic; \
  meta_check_magic = ogfs32_to_cpu(meta_check_magic); \
  OGFS_ASSERT(meta_check_magic == OGFS_MAGIC, \
	     ogfs_meta_header_t meta_check_mh; \
	     con_printf("Bad metadata at %"FMT_64"u\n", BH_BLKNO((bh))); \
	     ogfs_meta_header_in(&meta_check_mh, BH_DATA((bh))); \
	     ogfs_meta_header_print(&meta_check_mh, TRUE);); \
} \
while (0)

#define ogfs_metatype_check(bh, type) \
do \
{ \
  uint32 metatype_check_magic = ((ogfs_meta_header_t *)BH_DATA((bh)))->mh_magic; \
  uint32 metatype_check_type = ((ogfs_meta_header_t *)BH_DATA((bh)))->mh_type; \
  metatype_check_magic = ogfs32_to_cpu(metatype_check_magic); \
  metatype_check_type = ogfs32_to_cpu(metatype_check_type); \
  OGFS_ASSERT(metatype_check_magic == OGFS_MAGIC && \
	     metatype_check_type == (type), \
	     ogfs_meta_header_t metatype_check_mh; \
	     con_printf("Bad metadata at %"FMT_64"u, should be %u\n", BH_BLKNO((bh)), (type)); \
	     ogfs_meta_header_in(&metatype_check_mh, BH_DATA((bh))); \
	     ogfs_meta_header_print(&metatype_check_mh, TRUE);); \
} \
while (0)

#define ogfs_metatype_set(bh, type, format) \
do \
{ \
  ogfs_meta_check((bh)); \
  ((ogfs_meta_header_t *)BH_DATA((bh)))->mh_type = cpu_to_ogfs32((type)); \
  ((ogfs_meta_header_t *)BH_DATA((bh)))->mh_format = cpu_to_ogfs32((format)); \
} \
while (0)

#endif /* _INCORE_H */
