/*
 *
 *    Copyright 1995-1999 Regents of the University of Minnesota
 *    Portions Copyright 1999-2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 */

#include <linux/dcache.h>

#include "ogfs_debug.h"
#include "ogfs.h"
#include "blklist.h"
#include "bmap.h"
#include "dio.h"
#include "dir.h"
#include "fs.h"
#include "glock.h"
#include "glops.h"
#include "inode.h"
#include "log.h"
#include "super.h"
#include "trans.h"

/**
 * make_dinode - Fill in a new dinode structure
 * @sdp: The OGFS superblock
 * @gl: The glock covering the new inode
 * @inum: the inode number
 * @new: TRUE if this is a new dinode
 *
 */

static int
make_dinode(ogfs_sbd_t * sdp, ogfs_glock_t * gl, ogfs_inum_t * inum, int new)
{
	ogfs_dinode_t di, *tmp_di;
	struct buffer_head *dibh;
	struct timespec ctime;
	ogfs_rgrpd_t *rgd;
	ogfs_inum_t tmp_inum;
	uint32 incarn = 0;

	ENTER(GFN_MAKE_DINODE);

	dibh =
	    ogfs_dread(sdp, inum->no_addr, gl,
		      ((new) ? DR_FORCE : 0) | DR_START | DR_WAIT);
	if (!dibh) {
		EXIT(GFN_MAKE_DINODE);
		return -EIO;
	}

	if (!new) {
		tmp_di = (ogfs_dinode_t *) BH_DATA(dibh);
		ogfs_inum_in(&tmp_inum, (char *) &tmp_di->di_num);
		OGFS_ASSERT(OGFS_INUM_EQUAL(&tmp_inum, inum),);
		incarn = ogfs32_to_cpu(tmp_di->di_incarn) + 1;
	}

	ogfs_trans_add_bh(sdp, dibh);
	ogfs_metatype_set(dibh, OGFS_METATYPE_DI, OGFS_FORMAT_DI);
	ogfs_buffer_clear_tail(dibh, sizeof(ogfs_dinode_t));

	memset(&di, 0, sizeof(ogfs_dinode_t));

	ogfs_meta_header_in(&di.di_header, BH_DATA(dibh));

	di.di_num = *inum;

	di.di_nlink = 1;
	di.di_blocks = 1;
	ctime = current_kernel_time();
	di.di_atime = di.di_mtime = di.di_ctime = (int64)ctime.tv_sec;

	rgd = ogfs_blk2rgrpd(sdp, inum->no_addr);
	OGFS_ASSERT(rgd, con_printf("block = %" FMT_64 "u\n", inum->no_addr);
	    );

	di.di_rgrp = rgd->rd_ri.ri_addr;
	di.di_goal_rgrp = di.di_rgrp;
	di.di_goal_dblk = inum->no_addr - rgd->rd_ri.ri_data1;
	di.di_goal_mblk = di.di_goal_dblk;
	ogfs_rgrpd_put(rgd);

	di.di_type = FILE_REG;
	di.di_incarn = incarn;

	ogfs_dinode_out(&di, BH_DATA(dibh));
	brelse(dibh);

	EXIT(GFN_MAKE_DINODE);

	return 0;
}

/**
 * ogfs_copyin_dinode - Refresh the incore copy of the dinode
 * @ip: The OGFS inode
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_copyin_dinode(ogfs_inode_t * ip)
{
	struct buffer_head *dibh;

	ENTER(GFN_COPYIN_DINODE);

	dibh = ogfs_get_inode_buffer(ip);
	if (!dibh) {
		EXIT(GFN_COPYIN_DINODE);
		return -EIO;
	}

	ogfs_metatype_check(dibh, OGFS_METATYPE_DI);
	ogfs_dinode_in(&ip->i_di, BH_DATA(dibh));

	brelse(dibh);

	OGFS_ASSERT(ip->i_num.no_formal_ino == ip->i_di.di_num.no_formal_ino,
		   con_printf("inode = %" FMT_64 "u/%" FMT_64 "u\n",
			      ip->i_num.no_formal_ino, ip->i_num.no_addr);
		   ogfs_dinode_print(&ip->i_di, TRUE);
	    );

	/*  Handle a moved inode  */

	if (ip->i_num.no_addr != ip->i_di.di_num.no_addr) {
		OGFS_ASSERT(0,);	/*  Not implemented yet  */
	}

	ip->i_vn = ip->i_gl->gl_vn;

	ogfs_inode_attr_in(ip);

	EXIT(GFN_COPYIN_DINODE);

	return 0;
}

/**
 * read_dinode - Read an inode into the OGFS inode cache
 * @ip: The OGFS inode
 * @skip_write: Don't change nopen count
 *
 * Returns: 0 on success, -EXXXX on failure
 */

static int
read_dinode(ogfs_inode_t * ip)
{
	ogfs_glock_t *gl;
	int error;

	ENTER(GFN_READ_DINODE);

	error =
	    ogfs_glock_num(ip->i_sbd, ip->i_num.no_addr, &ogfs_nopen_glops, NULL,
			  GL_SHARED | GL_PERM | GL_DISOWN, &gl);
	if (error) {
		EXIT(GFN_READ_DINODE);
		return error;
	}

	/*  Read in dinode  */

	error = ogfs_copyin_dinode(ip);
	if (error) {
		ogfs_gunlock(ip->i_sbd, gl, GL_NOCACHE);
		ogfs_put_glstruct(gl);
		EXIT(GFN_READ_DINODE);
		return error;
	}

	ogfs_put_glstruct(gl);

	EXIT(GFN_READ_DINODE);

	return 0;
}

/**
 * ogfs_get_istruct - Get an inode given its number
 * @sdp: The OGFS superblock
 * @gl: The glock covering the inode
 * @inum: The inode number
 * @create: Flag to say if we are allowed to create a new ogfs_inode_t
 *
 * Returns: the inode, or NULL if we need to create one but can't.
 */

ogfs_inode_t *
ogfs_get_istruct(ogfs_sbd_t * sdp, ogfs_glock_t * gl, ogfs_inum_t * inum,
		int create)
{
	ogfs_inode_t *ip;
	int error;

	ENTER(GFN_GET_ISTRUCT);

	ip = (ogfs_inode_t *) gl->gl_object;
	if (ip) {
		/*  Don't care about no_addr unless we're creating the inode  */
		OGFS_ASSERT(ip->i_num.no_formal_ino == inum->no_formal_ino,);

		error = ogfs_copyin_dinode(ip);
		if (error) {
			ip = NULL;
			goto out;
		}

		atomic_inc(&ip->i_count);
		goto out;
	}

	if (!create)
		goto out;

	ip = ogfs_alloc_inode();
	if (!ip)
		goto out;

	memset(ip, 0, sizeof(ogfs_inode_t));

	ip->i_num = *inum;
	atomic_set(&ip->i_count, 1);
	ip->i_gl = gl;

	init_MUTEX(&ip->i_plock_mux);
	atomic_set(&ip->i_plock_cnt, 0);

	init_MUTEX(&ip->i_flock_mux);
	init_MUTEX(&ip->i_flock_lmx);
	ip->i_flock_st = 0;	/*unflocked */
	INIT_LIST_HEAD(&ip->i_flock_pids);

	ip->i_sbd = sdp;

	error = read_dinode(ip);
	if (error) {
		ogfs_free_inode(ip);
		ip = NULL;
		goto out;
	}

	ogfs_hold_glstruct(gl);
	gl->gl_object = ip;

out:
	EXIT(GFN_GET_ISTRUCT);

	return ip;
}

/**
 * ogfs_put_istruct - Put ogfs_inode_t structure
 * @ip: The OGFS inode
 *
 */

void
ogfs_put_istruct(ogfs_inode_t * ip)
{
	ENTER(GFN_PUT_ISTRUCT);

	OGFS_ASSERT(atomic_read(&ip->i_count) != 0,);
	atomic_dec(&ip->i_count);

	EXIT(GFN_PUT_ISTRUCT);
}

/**
 * ogfs_put_istruct - Destroy an inode structure with no references on it
 * @ip: The OGFS inode
 *
 * This function must be called with a glock held on the inode.
 *
 */

void
ogfs_destroy_istruct(ogfs_inode_t * ip)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	ogfs_glock_t *gl = ip->i_gl;

	ENTER(GFN_DESTROY_ISTRUCT);

	OGFS_ASSERT(!atomic_read(&ip->i_count),);

	gl->gl_object = NULL;
	ogfs_gunlock_num(sdp, ip->i_num.no_addr, &ogfs_nopen_glops, NULL,
			GL_NOCACHE);
	ogfs_flush_meta_cache(ip, 0);
	ogfs_free_inode(ip);
	ogfs_put_glstruct(gl);

	EXIT(GFN_DESTROY_ISTRUCT);
}

/**
 * dinode_dealloc - Put deallocate a dinode
 * @ip: The OGFS inode
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
dinode_dealloc(ogfs_inode_t * ip)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	ogfs_rgrpd_t *rgd;
	int error;

	ENTER(GFN_DINODE_DEALLOC);

	error = ogfs_rindex_hold(sdp);
	if (error)
		goto fail;

	rgd = ogfs_blk2rgrpd(sdp, ip->i_num.no_addr);
	OGFS_ASSERT(rgd, con_printf("block = %" FMT_64 "u\n", ip->i_num.no_addr);
	    );

	error = ogfs_glock_rg(rgd, 0);
	if (error)
		goto fail_rgrpd_put;

	/*  Trans may require:
	   One block for the RG header.  One block for the dinode bit.
	   One block for the dinode.
	   We also need a block for the nopen change.  */

	error = ogfs_trans_begin(sdp, TRANS_I_DEALLOC, 3, 1);
	if (error)
		goto fail_rg_gunlock;

	error = ogfs_difree(ip, rgd);
	if (error)
		goto fail_end_trans;

	error = ogfs_trans_nopen_change(sdp, OGFS_LOG_DESC_NOPCLOSE, &ip->i_num);
	if (error)
		goto fail_end_trans;

	ogfs_trans_add_gl(sdp, ip->i_gl);

	ogfs_trans_end(sdp);

	ogfs_gunlock_rg(rgd, 0);
	ogfs_rgrpd_put(rgd);
	ogfs_rindex_release(sdp);

	EXIT(GFN_DINODE_DEALLOC);

	return 0;

fail_end_trans:
	ogfs_trans_mark(sdp, TRF_POISONED);
	ogfs_trans_end(sdp);

fail_rg_gunlock:
	ogfs_gunlock_rg(rgd, 0);

fail_rgrpd_put:
	ogfs_rgrpd_put(rgd);
	ogfs_rindex_release(sdp);

fail:
	EXIT(GFN_DINODE_DEALLOC);
	return error;
}

/**
 * inode_dealloc - Deallocate an inode
 * @sdp: the filesystem
 * @inum: the inode number to deallocate
 * @nopen_gl: the nopen glock structure for this inode (It's not held)
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
inode_dealloc(ogfs_sbd_t * sdp, ogfs_inum_t * inum, ogfs_glock_t * nopen_gl)
{
	ogfs_inode_t *ip;
	ogfs_glock_t *gl;
	int error;

	ENTER(GFN_INODE_DEALLOC);

	error =
	    ogfs_glock_num(sdp, inum->no_formal_ino, &ogfs_inode_glops, NULL, 0,
			  &gl);
	if (error)
		goto fail;

	/*  We reacquire the nopen lock here to avoid a race with the NFS server
	   calling ogfs_read_inode() with the inode number of a inode we're in the
	   process of deallocating.  And we can't keep our hold on the lock
	   from ogfs_dealloc_inodes() for deadlock reasons.  */

	error = ogfs_glock(sdp, nopen_gl, LM_FLAG_TRY);
	switch (error) {
	case 0:
		break;
	case GLR_TRYFAILED:
		error = 0;
		goto fail_gunlock;
	default:
		OGFS_ASSERT(error < 0,);
		goto fail_gunlock;
	}

	ip = ogfs_get_istruct(sdp, gl, inum, CREATE);

	ogfs_gunlock(sdp, nopen_gl, 0);

	if (!ip) {
		error = -ENOMEM;
		goto fail_gunlock;
	}

	OGFS_ASSERT(!ip->i_di.di_nlink, ogfs_dinode_print(&ip->i_di, TRUE);
	    );
	OGFS_ASSERT(atomic_read(&ip->i_count) == 1,);
	OGFS_ASSERT(!ip->i_vino,);

	if (ip->i_di.di_type == FILE_DIR
	    && (ip->i_di.di_flags & OGFS_DIF_EXHASH)) {
		error = ogfs_dir_exhash_free(ip);
		if (error)
			goto fail_iput;
	}

	error = ogfs_shrink(ip, 0);
	if (error)
		goto fail_iput;

	error = dinode_dealloc(ip);
	if (error)
		goto fail_iput;

	ogfs_put_istruct(ip);
	ogfs_destroy_istruct(ip);

	ogfs_gunlock(sdp, gl, 0);
	ogfs_put_glstruct(gl);

	EXIT(GFN_INODE_DEALLOC);

	return 0;

fail_iput:
	ogfs_put_istruct(ip);
	ogfs_destroy_istruct(ip);

fail_gunlock:
	ogfs_gunlock(sdp, gl, 0);
	ogfs_put_glstruct(gl);

fail:
	EXIT(GFN_INODE_DEALLOC);
	return error;
}

/**
 * ogfs_dealloc_inodes - Go through the list of inodes to be deallocated
 * @sdp: the filesystem
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_dealloc_inodes(ogfs_sbd_t * sdp)
{
	ogfs_glock_t *gl;
	ogfs_inum_t inum, first_stuck;
	unsigned int freed;
	int error;

	ENTER(GFN_DEALLOC_INODES);

	do {
		ogfs_pitch_inodes(sdp);

		freed = 0;
		first_stuck.no_formal_ino = 0;

		for (;;) {
			ogfs_nopen_find(sdp, &inum);
			if (!inum.no_formal_ino
			    || OGFS_INUM_EQUAL(&inum, &first_stuck))
				break;

			error =
			    ogfs_glock_num(sdp, inum.no_addr, &ogfs_nopen_glops,
					  NULL, LM_FLAG_TRY, &gl);
			switch (error) {
			case 0:
				break;
			case GLR_TRYFAILED:
				if (!first_stuck.no_formal_ino)
					first_stuck = inum;
				continue;
			default:
				OGFS_ASSERT(error < 0,);
				goto fail;
			}

			ogfs_gunlock(sdp, gl, 0);
			error = inode_dealloc(sdp, &inum, gl);
			ogfs_put_glstruct(gl);

			if (error)
				goto fail;

			freed++;
		}
	} while (freed);

	EXIT(GFN_DEALLOC_INODES);

	return 0;

fail:
	EXIT(GFN_DEALLOC_INODES);
	return error;
}

/**
 * ogfs_limit_unlinked - limit the number of inodes waiting to be deallocated
 * @sdp: the filesystem
 *
 * Returns: 0 on success, -EXXX on failure;
 */

int
ogfs_limit_unlinked(ogfs_sbd_t * sdp)
{
	int error = 0;

	ENTER(GFN_LIMIT_UNLINKED);

	while (atomic_read(&sdp->sd_nopen_ic_count) > 500) {
		wake_up_process(sdp->sd_inoded_process);

		IFDB(GFN_LIMIT_UNLINKED,
		     db_printf("ondisk = %d, incore = %d\n",
			       atomic_read(&sdp->sd_nopen_od_count),
			       atomic_read(&sdp->sd_nopen_ic_count));
		    );

		osi_sleep(1);

		if (osi_pending_signals()) {
			error = -ERESTARTSYS;
			break;
		}
	}

	EXIT(GFN_LIMIT_UNLINKED);

	return error;
}

/**
 * ogfs_change_nlink - Change nlink count on inode
 * @ip: The OGFS inode
 * @diff: The change in the nlink count required
 *
 * Change a dinode's nlink field by diff.
 * Assumes the lock for dvp is already held.
 *
 * Returns: 0 on success, -EXXXX on failure.
 */

int
ogfs_change_nlink(ogfs_inode_t * ip, int diff)
{
	struct buffer_head *dibh;
	struct timespec ctime;
	uint32 udiff;
	int error = 0;

	ENTER(GFN_CHANGE_NLINK);

	dibh = ogfs_get_inode_buffer(ip);
	if (!dibh) {
		error = -EIO;
		goto fail;
	}

	if (diff < 0) {
		udiff = (uint32) (-diff);

		OGFS_ASSERT(ip->i_di.di_nlink >= udiff,
			   ogfs_dinode_print(&ip->i_di, TRUE);
		    );

		ip->i_di.di_nlink -= udiff;
	} else {
		udiff = (uint32) diff;

		if (ip->i_di.di_nlink + udiff < ip->i_di.di_nlink) {
			error = -EMLINK;
			goto fail_drelse;
		}

		ip->i_di.di_nlink += udiff;
	}

	ctime = current_kernel_time();
	ip->i_di.di_ctime = (int64)ctime.tv_sec;

	ogfs_trans_add_bh(ip->i_sbd, dibh);
	ogfs_dinode_out(&ip->i_di, BH_DATA(dibh));
	brelse(dibh);

	EXIT(GFN_CHANGE_NLINK);

	return 0;

fail_drelse:
	brelse(dibh);

fail:
	EXIT(GFN_CHANGE_NLINK);
	return error;
}

/**
 * ogfs_link_size - Return expanded size required to store a symlink
 * @str: The symlink
 * @ops: The symlink substitution operations table
 *
 * Returns: The size of the expanded symlink.
 */

unsigned int
ogfs_link_size(struct ogfs_symlink_ops *ops, char *str)
{
	unsigned int len = 0;
	unsigned int i;

	ENTER(GFN_LINK_SIZE);

	while (*str) {
		len++;
		if (*str++ != '{')
			continue;

		for (i = 0; ops[i].name; i++) {
			if (memcmp(str, ops[i].name, ops[i].len) == 0) {
				len--;
				len += ops[i].subst_fn(NULL);
				str += ops[i].len;
				break;
			}
		}
	}

	EXIT(GFN_LINK_SIZE);

	return len + 1;
}

/**
 * ogfs_link_expand - Expand a context sensitive symlink
 * @ops: The symlink substitution operations table
 * @out: Buffer to place result in
 * @in: Buffer to get symlink from
 *
 * Returns: 0 or error code
 */

void
ogfs_link_expand(struct ogfs_symlink_ops *ops, char *out, char *in)
{
	unsigned int i;

	ENTER(GFN_LINK_EXPAND);

	while (*in) {
		*out++ = *in;
		if (*in++ != '{')
			continue;

		for (i = 0; ops[i].name; i++) {
			if (memcmp(in, ops[i].name, ops[i].len) == 0) {
				out--;
				out += ops[i].subst_fn(out);
				in += ops[i].len;
			}
		}
	}

	*out = 0;

	EXIT(GFN_LINK_EXPAND);
}

/**
 * ogfs_lookupi - Look up a filename in a directory and return its inode
 * @dip: The directory to search
 * @name: The name of the inode to look for
 * @unlock: Non-zero if we should unlock the two inodes before returning
 * @ipp: Used to return the found inode if any
 *
 * Returns: 0 on success, -EXXXX on failure
 */

int
ogfs_lookupi(ogfs_inode_t * dip, struct qstr * name,
	    int unlock, ogfs_inode_t ** ipp)
{
	ogfs_sbd_t *sdp = dip->i_sbd;
	ogfs_glock_t *gl;
	int error;
	ogfs_inum_t inum, inum2;

	ENTER(GFN_LOOKUPI);

	*ipp = NULL;

	if (name->len > OGFS_FNAMESIZE) {
		EXIT(GFN_LOOKUPI);
		return -ENAMETOOLONG;
	}

	if (ogfs_filecmp(name, (char *) ".", 1)) {
		*ipp = dip;
		EXIT(GFN_LOOKUPI);
		return 0;
	}

	error = ogfs_glock_i(dip, GL_SHARED);
	if (error)
		goto fail;

	if ((error = ogfs_permission(dip, MAY_EXEC)) ||
	    (error = ogfs_dir_search(dip, name, &inum, NULL))) {
		ogfs_gunlock_i(dip, 0);
		goto fail;
	}

restart:
	gl = ogfs_get_glstruct(sdp, inum.no_formal_ino, &ogfs_inode_glops, NULL,
			      CREATE);
	if (!gl) {
		ogfs_gunlock_i(dip, 0);
		error = -ENOMEM;
		goto fail;
	}

	/*  Acquire the second lock  */

	if (gl->gl_name.ln_number < dip->i_gl->gl_name.ln_number) {
		ogfs_gunlock_i(dip, 0);

		error = ogfs_glock(sdp, gl, GL_SHARED);
		if (error)
			goto fail_glput;

		error = ogfs_glock_i(dip, GL_SHARED);
		if (error) {
			ogfs_gunlock(sdp, gl, 0);
			goto fail_glput;
		}

		if ((error = ogfs_permission(dip, MAY_EXEC)) ||
		    (error = ogfs_dir_search(dip, name, &inum2, NULL))) {
			ogfs_gunlock_i(dip, 0);
			ogfs_gunlock(sdp, gl, 0);
			goto fail_glput;
		}

		if (!OGFS_INUM_EQUAL(&inum, &inum2)) {
			ogfs_gunlock(sdp, gl, 0);
			ogfs_put_glstruct(gl);
			inum = inum2;
			goto restart;
		}
	} else {
		error = ogfs_glock(sdp, gl, GL_SHARED);
		if (error) {
			ogfs_gunlock_i(dip, 0);
			goto fail_glput;
		}
	}

	*ipp = ogfs_get_istruct(sdp, gl, &inum, CREATE);
	if (!*ipp) {
		ogfs_gunlock(sdp, gl, 0);
		ogfs_gunlock_i(dip, 0);
		error = -EIO;
		goto fail_glput;
	}

	if (unlock) {
		ogfs_gunlock_i(*ipp, 0);
		ogfs_gunlock_i(dip, 0);
	}

	ogfs_put_glstruct(gl);

	EXIT(GFN_LOOKUPI);

	return 0;

fail_glput:
	ogfs_put_glstruct(gl);

fail:
	EXIT(GFN_LOOKUPI);

	return error;
}

/**
 * pick_formal_ino - Pick a formal inode number for a given inode
 * @sdp: the filesystem
 * @inum: the inode number structure
 *
 */

static void
pick_formal_ino(ogfs_sbd_t * sdp, ogfs_inum_t * inum)
{
	ENTER(GFN_PICK_FORMAL_INO);

	/*  This won't always be true  */
	inum->no_formal_ino = inum->no_addr;

	EXIT(GFN_PICK_FORMAL_INO);
}

/**
 * ogfs_createi - Create a new inode
 * @dip: The directory for the new inode
 * @name: The name of the new file
 * @type: The type of dinode (FILE_REG, FILE_DIR, FILE_LNK, ...)
 * @exclusive: if TRUE, return an error if the inode already exists
 * @new: If non-NULL, set to one if the inode is newly created
 * @ipp: Returns the inode that was created
 *
 * If the return value is 0, the glocks on both the directory and the new
 * file are held.  A transaction has been started and an inplace reservation
 * is held, as well.
 *
 * Returns: 0 on success, -EXXXX on failure
 */

int
ogfs_createi(ogfs_inode_t * dip, struct qstr * name,
	    unsigned int type, int exclusive,
	    int *new, ogfs_inode_t ** ipp)
{
	ogfs_sbd_t *sdp = dip->i_sbd;
	ogfs_glock_t *gl = NULL, *nopen_gl = NULL;
	ogfs_ipreserv_t *res = &dip->i_res;
	ogfs_inum_t inum;
	int new_inode = FALSE;
	int error;

	ENTER(GFN_CREATEI);

	*ipp = NULL;
	if (new)
		*new = TRUE;
	if (name->len > OGFS_FNAMESIZE) {
		error = -ENAMETOOLONG;
		goto fail;
	}

restart:
	error = ogfs_glock_i(dip, 0);
	if (error)
		goto fail;
	error = ogfs_permission(dip, MAY_WRITE|MAY_EXEC);
	if (error)
		goto fail_gunlock;
	/*  Don't create entries in an unlinked directory  */
	if (!dip->i_di.di_nlink) {
		error = -EPERM;
		goto fail_gunlock;
	}
	error = ogfs_dir_search(dip, name, NULL, NULL);
	switch (error) {
	case -ENOENT:
		break;
	case 0:
		if (exclusive) {
			error = -EEXIST;
			goto fail_gunlock;
		} else {
			ogfs_gunlock_i(dip, 0);

			error = ogfs_lookupi(dip, name, FALSE, ipp);
			switch (error) {
			case 0:
				*new = FALSE;
				EXIT(GFN_CREATEI);
				return 0;
			case -ENOENT:
				goto restart;
			default:
				goto fail;
			}
		}

		break;
	default:
		goto fail_gunlock;
	}
	res->ir_diblks = 1;
	res->ir_metablks = sdp->sd_max_dirres;
	res->ir_datablks = 0;
	res->ir_target = dip->i_di.di_rgrp;
	res->ir_trans = TRANS_CREATE;
	if (type == FILE_DIR) {
		res->ir_target = sdp->sd_dir_alloc_target;

		error = ogfs_inplace_reserv(dip, 1);
		if (error)
			goto fail_gunlock;

		ogfs_update_datarget(sdp, res);
	} else {
		error = ogfs_inplace_reserv(dip, 1);
		if (error)
			goto fail_gunlock;
	}
	ogfs_dipick(sdp, res, &inum);
	if (!inum.no_formal_ino) {
		new_inode = TRUE;
		pick_formal_ino(sdp, &inum);
	}
	error = ogfs_glock_num(
		sdp, inum.no_formal_ino, &ogfs_inode_glops, NULL, 0, &gl);
	if (error)
		goto fail_ipres;
	error = ogfs_glock_num(
		sdp, inum.no_addr, &ogfs_nopen_glops, NULL, GL_SHARED,
		&nopen_gl);
	if (error)
		goto fail_gunlock2;

	/*  Trans may require:
	   blocks for two dinodes, the directory blocks necessary for
	   a new entry, RG bitmap blocks for an allocation.  */
	error = ogfs_trans_begin(
		sdp, TRANS_CREATE,
		2 + sdp->sd_max_dirres + res->ir_rgd->rd_ri.ri_length, 0);
	if (error)
		goto fail_gunlock3;
	error = ogfs_dialloc(sdp, res, &inum, gl);
	if (error)
		goto fail_end_trans;
	error = ogfs_dir_add(dip, name, &inum, type);
	if (error)
		goto fail_end_trans;
	error = make_dinode(sdp, gl, &inum, new_inode);
	if (error)
		goto fail_end_trans;
	*ipp = ogfs_get_istruct(sdp, gl, &inum, CREATE);
	if (!*ipp) {
		error = -ENOMEM;
		goto fail_end_trans;
	}
	ogfs_gunlock(sdp, nopen_gl, 0);
	ogfs_put_glstruct(nopen_gl);
	ogfs_put_glstruct(gl);

	EXIT(GFN_CREATEI);

	return 0;

fail_end_trans:
	ogfs_trans_mark(sdp, TRF_POISONED);
	ogfs_trans_end(sdp);

fail_gunlock3:
	ogfs_gunlock(sdp, nopen_gl, 0);
	ogfs_put_glstruct(nopen_gl);

fail_gunlock2:
	ogfs_gunlock(sdp, gl, 0);
	ogfs_put_glstruct(gl);

fail_ipres:
	ogfs_inplace_relse(dip);

fail_gunlock:
	ogfs_gunlock_i(dip, 0);

fail:
	EXIT(GFN_CREATEI);
	return error;
}

/**
 * ogfs_unlinki - Unlink a file
 * @dip: The inode of the directory
 * @dname: The name of the file to be unlinked
 * @ip: The inode of the file to be removed
 *
 * Assumes Glocks on both dip and ip are held.
 *
 * Returns: 0 on success, -EXXXX on failure
 */

int
ogfs_unlinki(ogfs_inode_t * dip, struct qstr * dname, ogfs_inode_t * ip)
{
	ogfs_sbd_t *sdp = dip->i_sbd;
	int error = 0;

	ENTER(GFN_UNLINKI);

	error = ogfs_dir_del(dip, dname);
	if (error)
		goto out;

	ogfs_change_nlink(ip, -1);

	/*  If this inode is being unlinked from the directory structure,
	   we need to mark that in the log so that it isn't lost during
	   a crash.  */

	if (!ip->i_di.di_nlink)
		error =
		    ogfs_trans_nopen_change(sdp, OGFS_LOG_DESC_NOPOPEN,
					   &ip->i_num);

out:
	EXIT(GFN_UNLINKI);

	return error;
}

/**
 * ogfs_rmdiri - Remove a directory
 * @dip: The parent directory of the directory to be removed
 * @dname: The name of the directory to be removed
 * @ip: The OGFS inode of the directory to be removed
 *
 * Assumes Glocks on dip and ip are held
 *
 * Returns: 0 on success, -EXXXX on failure
 */

int
ogfs_rmdiri(ogfs_inode_t * dip, struct qstr * dname, ogfs_inode_t * ip)
{
	ogfs_sbd_t *sdp = dip->i_sbd;
	struct qstr name;
	int error = 0;

	ENTER(GFN_RMDIRI);

	/*  Check to make sure the directory is empty!!!!  */

	if (ip->i_di.di_entries > 2) {
		error = -ENOTEMPTY;
		goto out;
	}

	OGFS_ASSERT(ip->i_di.di_entries == 2, ogfs_dinode_print(&ip->i_di, TRUE);
	    );

	error = ogfs_dir_del(dip, dname);
	if (error)
		goto out;

	ogfs_change_nlink(dip, -1);

	name.len = 1;
	name.name = ".";
	ogfs_dir_del(ip, &name);

	name.len = 2;
	name.name = "..";
	ogfs_dir_del(ip, &name);

	OGFS_ASSERT(ip->i_di.di_nlink == 2, ogfs_dinode_print(&ip->i_di, TRUE);
	    );
	ogfs_change_nlink(ip, -2);

	/*  This inode is being unlinked from the directory structure and
	   we need to mark that in the log so that it isn't lost during
	   a crash.  */

	error = ogfs_trans_nopen_change(sdp, OGFS_LOG_DESC_NOPOPEN, &ip->i_num);

out:
	EXIT(GFN_RMDIRI);

	return error;
}

/**
 * ogfs_update_atime - Conditionally update the atime on an inode
 * @ip: ogfs_inode_t to test and possibly update
 *
 * Tests atime for ogfs_read, ogfs_readdir and ogfs_test_mmap
 * Update if the difference between the current time and the current atime
 * is greater than a interval specfied at mount.  default quantum is 1 second.
 *
 * Returns: 0 on success, -Exxx on error
 */

int
ogfs_update_atime(ogfs_inode_t * ip)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	uint64 curtime, quantum = sdp->sd_args.ar_atime_quantum;
	struct buffer_head *dibh;
	struct timespec ctime;
	int error = 0;

	ENTER(GFN_UPDATE_ATIME);

	ctime = current_kernel_time();
	curtime = (int64)ctime.tv_sec;
	if (curtime - ip->i_di.di_atime >= quantum) {
		ogfs_gunlock_i(ip, 0);
		error = ogfs_glock_i(ip, 0);
		if (error)
			goto fail;

		/* verify this hasn't been updated while we were trying to get exclusive lock */

		if (curtime - ip->i_di.di_atime >= quantum) {
			error = ogfs_trans_begin(sdp, TRANS_UPDATE_ATIME, 1, 0);
			if (error)
				goto fail_gunlock;

			dibh = ogfs_get_inode_buffer(ip);
			if (!dibh) {
				error = -EIO;
				goto fail_end_trans;
			}

			ip->i_di.di_atime = curtime;

			ogfs_trans_add_bh(sdp, dibh);
			ogfs_dinode_out(&ip->i_di, BH_DATA(dibh));
			brelse(dibh);

			ogfs_trans_end(sdp);
		}
	}

	EXIT(GFN_UPDATE_ATIME);

	return error;

fail_end_trans:
	ogfs_trans_mark(sdp, TRF_POISONED);
	ogfs_trans_end(sdp);

fail_gunlock:
	ogfs_gunlock_i(ip, 0);

fail:
	EXIT(GFN_UPDATE_ATIME);
	return error;
}
