/*
 *
 *    Copyright 1995-1999 Regents of the University of Minnesota
 *    Portions Copyright 1999-2001 Sistina Software, Inc.
 *    Portions Copyright 2001 The OpenGFS Project
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 */

#include "ogfs_debug.h"
#include "ogfs.h"

#include <linux/utsname.h>
#include <linux/version.h>
#include <linux/pagemap.h>
#include <linux/smp_lock.h>
#include <linux/page-flags.h>

#include "blklist.h"
#include "bmap.h"
#include "dcache.h"
#include "dio.h"
#include "dir.h"
#include "ogfs_linux.h"
#include "glock.h"
#include "glops.h"
#include "fs.h"
#include "inode.h"
#include "inode_linux.h"
#include "trans.h"

//static LIST_HEAD(inode_unused);
//spinlock_t *inode_lock;

/**
 * ogfs_imatch - The match function used in ogfs_iget()
 * @inode: The inode
 * @ino: The inode number (32 bits)
 * @ptr: The ogfs_inode_t pointer we're looking for
 *
 * Returns: TRUE on a match, FALSE on a failure
 */

static int
ogfs_imatch(struct inode *inode, unsigned long ino, void *ptr)
{
	ogfs_inode_t *ip = vi2ino(inode);
	int ret = FALSE;

	ENTER(GFN_IMATCH);

	if (ip == (ogfs_inode_t *) ptr)
		ret = TRUE;

	EXIT(GFN_IMATCH);

	return ret;
}

/**
 * ogfs_iget - Gets an inode, does a 64 bit compare
 * @sb: The superblock
 * @ip: the ogfs_inode_t to get the inode for
 *
 * Returns: An inode
 */

struct inode *
ogfs_iget(struct super_block *sb, ogfs_inode_t * ip)
{
	struct inode *inode;
	ENTER(GFN_IGET);

	inode =
	    iget_locked(sb, (unsigned long)ip->i_num.no_formal_ino);
	    //iget5_locked(sb, (unsigned long) ip->i_num.no_formal_ino, ogfs_imatch, ip);

	EXIT(GFN_IGET);
	return inode;
}

/**
 * ogfs_inode_attr_in - Copy attributes from the dinode into the VFS inode
 * @ip: The OGFS inode
 *
 */

void
ogfs_inode_attr_in(ogfs_inode_t * ip)
{
	struct inode *ino = ip->i_vino;

	ENTER(GFN_INODE_ATTR_IN);

	if (!ino)
		goto out;

	ino->i_mode = ip->i_di.di_mode & S_IALLUGO;

	switch (ip->i_di.di_type) {
	case FILE_REG:
		ino->i_mode |= S_IFREG;
		ino->i_rdev = 0;
		break;
	case FILE_DIR:
		ino->i_mode |= S_IFDIR;
		ino->i_rdev = 0;
		break;
	case FILE_LNK:
		ino->i_mode |= S_IFLNK;
		ino->i_rdev = 0;
		break;
	case FILE_BLK:
		ino->i_mode |= S_IFBLK;
		ino->i_rdev = MKDEV(ip->i_di.di_major, ip->i_di.di_minor);
		break;
	case FILE_CHR:
		ino->i_mode |= S_IFCHR;
		ino->i_rdev = MKDEV(ip->i_di.di_major, ip->i_di.di_minor);
		break;
	case FILE_FIFO:
		ino->i_mode |= S_IFIFO;
		ino->i_rdev = 0;
		break;
	case FILE_SOCK:
		ino->i_mode |= S_IFSOCK;
		ino->i_rdev = 0;
		break;
	default:
		OGFS_ASSERT(0,
			   con_printf("fino/addr = %" FMT_64 "u/%" FMT_64
				      "u, type = %u\n",
				      ip->i_num.no_formal_ino,
				      ip->i_num.no_addr, ip->i_di.di_type););
		break;
	};

	ino->i_nlink = ip->i_di.di_nlink;
	ino->i_uid = ip->i_di.di_uid;
	ino->i_gid = ip->i_di.di_gid;
	ino->i_size = ip->i_di.di_size;
	ino->i_atime.tv_sec = ip->i_di.di_atime;
	ino->i_mtime.tv_sec = ip->i_di.di_mtime;
	ino->i_ctime.tv_sec = ip->i_di.di_ctime;
	ino->i_blksize = PAGE_SIZE;
	ino->i_blocks = ip->i_di.di_blocks * ip->i_sbd->sd_sb.sb_bsize / 512;
	ino->i_generation = ip->i_di.di_incarn;
	//ino->i_version = ++event;

out:
	EXIT(GFN_INODE_ATTR_IN);
}

/**
 * ogfs_inode_attr_out - Copy attributes from VFS inode into the dinode
 * @ip: The OGFS inode
 *
 * Only copy out the attributes that we want the VFS layer
 * to be able to modify.
 */

void
ogfs_inode_attr_out(ogfs_inode_t * ip)
{
	struct inode *ino = ip->i_vino;

	ENTER(GFN_INODE_ATTR_OUT);

	if (!ino)
		goto out;

	ip->i_di.di_mode = ino->i_mode & S_IALLUGO;
	ip->i_di.di_uid = ino->i_uid;
	ip->i_di.di_gid = ino->i_gid;
	ip->i_di.di_atime = ino->i_atime.tv_sec;
	ip->i_di.di_mtime = ino->i_mtime.tv_sec;
	ip->i_di.di_ctime = ino->i_ctime.tv_sec;

out:
	EXIT(GFN_INODE_ATTR_OUT);
}

/**
 * ogfs_setattr - Change attributes on an inode
 * @dentry: The dentry which is changing
 * @attr: The structure describing the change
 *
 * The VFS layer wants to change one or more of an inodes attributes.  Write
 * that change out to disk.
 *
 * Returns: 0 on success, -EXXXX on failure
 */

static int
ogfs_setattr(struct dentry *dentry, struct iattr *attr)
{
	ogfs_inode_t *ip = vi2ino(dentry->d_inode);
	ogfs_sbd_t *sdp = ip->i_sbd;
	struct buffer_head *dibh;
	int error = 0;

	ENTER(GFN_SETATTR);

	error = ogfs_glock_i(ip, 0);
	if (error)
		goto fail;

	error = inode_change_ok(dentry->d_inode, attr);
	if (error)
		goto fail_gunlock;

	if (attr->ia_valid & ATTR_SIZE) {
		vmtruncate(dentry->d_inode, attr->ia_size);
		error = ogfs_truncate(ip, attr->ia_size);
	} else {
		/*  Trans may require:
		   one dinode block  */

		error = ogfs_trans_begin(sdp, TRANS_SETATTR, 1, 0);
		if (error)
			goto fail_gunlock;

		dibh = ogfs_get_inode_buffer(ip);
		if (!dibh) {
			error = -EIO;
			goto fail_end_trans;
		}

		inode_setattr(dentry->d_inode, attr);
		ogfs_inode_attr_out(ip);

		ogfs_trans_add_bh(sdp, dibh);
		ogfs_dinode_out(&ip->i_di, BH_DATA(dibh));
		brelse(dibh);

		ogfs_trans_end(sdp);
	}

	ogfs_gunlock_i(ip, 0);

	EXIT(GFN_SETATTR);

	return 0;

fail_end_trans:
	ogfs_trans_mark(sdp, TRF_POISONED);
	ogfs_trans_end(sdp);

fail_gunlock:
	ogfs_gunlock_i(ip, 0);

fail:
	EXIT(GFN_SETATTR);
	return error;
}

/**
 * ogfs_lookup - Look up a filename in a directory and return its inode
 * @dir: The directory inode
 * @dentry: The dentry of the new inode
 *
 * Called by the VFS layer. Lock dir and call ogfs_lookupi()
 *
 * Returns: 0 on success, -EXXXX on failure
 */

struct dentry *
ogfs_lookup(struct inode *dir, struct dentry *dentry, struct nameidata *nd)
{
	ogfs_inode_t *dip = vi2ino(dir), *ip;
	ogfs_sbd_t *sdp = dip->i_sbd;
	struct inode *inode = NULL;
	ogfs_dcached_t *dcd;
	int error;

	ENTER(GFN_LOOKUP);

	IFDB(GFN_LOOKUP,
	     db_printf("  %" FMT_64 "u, %s\n", dip->i_num.no_formal_ino,
		       dentry->d_name.name););

	dcd = (ogfs_dcached_t *) gmalloc(sizeof(ogfs_dcached_t));
	if (!dcd) {
		error = -ENOMEM;
		goto fail;
	}
	memset(dcd, 0, sizeof(ogfs_dcached_t));

	error = ogfs_lookupi(dip, &dentry->d_name, TRUE, &ip);
	if (error != 0 && error != -ENOENT)
		goto fail_gfree;

	if (ip) {
		inode = ogfs_iget(sdp->sd_vsb, ip);
		ogfs_put_istruct(ip);
	}

	/*  Set up the dcache data and operations  */

	if (!vdc2dcd(dentry)) {
		dcd->dc_vn = dip->i_gl->gl_vn;

		vdc2dcd(dentry) = dcd;
		dentry->d_op = &ogfs_dops;
	} else
		gfree(dcd, sizeof(ogfs_dcached_t));

	d_add(dentry, inode);

	IFDB(GFN_LOOKUP, db_printf("  error = %d\n", error););

	EXIT(GFN_LOOKUP);

	return NULL;

fail_gfree:
	gfree(dcd, sizeof(ogfs_dcached_t));

fail:
	EXIT(GFN_LOOKUP);
	return ERR_PTR(error);
}

/**
 * ogfs_create - Create a file
 * @dir: The directory in which to create the file
 * @dentry: The dentry of the new file
 * @mode: The permissions of the new file
 *
 * Returns: 0 on success, -EXXXX on failure
 */

static int
ogfs_create(struct inode *dir, struct dentry *dentry, int mode,
		struct nameidata *nd)
{
	ogfs_inode_t *dip = vi2ino(dir), *ip;
	ogfs_sbd_t *sdp = dip->i_sbd;
	struct inode *inode;
	struct buffer_head *dibh;
	int new;
	int error;

	ENTER(GFN_CREATE);

	error =
	    ogfs_createi(dip, &dentry->d_name, FILE_REG, FALSE, &new, &ip);
	if (error) {
		EXIT(GFN_CREATE);
		return error;
	}

	OGFS_ASSERT(ip,);

	if (new) {
		ip->i_di.di_mode = mode;
		ip->i_di.di_uid = current->fsuid;
		if (dip->i_di.di_mode & S_ISGID)
			ip->i_di.di_gid = dip->i_di.di_gid;
		else
			ip->i_di.di_gid = current->fsgid;
		dibh = ogfs_get_inode_buffer(ip);
		OGFS_ASSERT(dibh,);

		ogfs_dinode_out(&ip->i_di, BH_DATA(dibh));

		brelse(dibh);

		set_bit(GIF_CREATED, &ip->i_flags);

		ogfs_trans_end(sdp);
		ogfs_inplace_relse(dip);
	}

	ogfs_gunlock_i(ip, 0);
	ogfs_gunlock_i(dip, 0);

	inode = ogfs_iget(sdp->sd_vsb, ip);
	ogfs_put_istruct(ip);

	d_instantiate(dentry, inode);

	EXIT(GFN_CREATE);

	return 0;
}

/**
 * ogfs_link - Link to a file
 * @old_dentry: The inode to link
 * @dir: Add link to this directory
 * @dentry: The name of the link
 *
 * Link the inode in "old_dentry" into the directory "dir" with the
 * name in "dentry".
 *
 * Returns: 0 on success, -EXXXX on failure
 */

static int
ogfs_link(struct dentry *old_dentry, struct inode *dir, struct dentry *dentry)
{
	ogfs_inode_t *dip = vi2ino(dir);
	ogfs_sbd_t *sdp = dip->i_sbd;
	ogfs_inode_t *ip = vi2ino(old_dentry->d_inode);
	ogfs_ipreserv_t *res = &dip->i_res;
	ogfs_lockop_t lock[2];
	int error;

	ENTER(GFN_LINK);

	if (ip->i_di.di_type == FILE_DIR) {
		error = -EPERM;
		goto fail;
	}

	memset(lock, 0, 2 * sizeof(ogfs_lockop_t));

	lock[0].lo_ip = dip;
	lock[1].lo_ip = ip;

	error = ogfs_glock_m(sdp, 2, lock);
	if (error)
		goto fail;

	error = permission(dir, MAY_WRITE | MAY_EXEC, NULL);
	if (error)
		goto fail_gunlock;

	error = ogfs_dir_search(dip, &dentry->d_name, NULL, NULL);
	switch (error) {
	case -ENOENT:
		break;
	case 0:
		error = -EEXIST;
		goto fail_gunlock;
	default:
		goto fail_gunlock;
	}

	res->ir_diblks = 0;
	res->ir_metablks = sdp->sd_max_dirres;
	res->ir_datablks = 0;
	res->ir_target = dip->i_di.di_goal_rgrp;
	res->ir_trans = TRANS_LINK;

	error = ogfs_inplace_reserv(dip, 1);
	if (error)
		goto fail_gunlock;

	/*  Trans may require:
	   two dinode blocks, directory modifications to add an entry,
	   and RG bitmap blocks to allocate from  */

	error =
	    ogfs_trans_begin(sdp, TRANS_LINK,
			    2 + sdp->sd_max_dirres +
			    res->ir_rgd->rd_ri.ri_length, 0);
	if (error)
		goto fail_ipres;

	error = ogfs_dir_add(dip, &dentry->d_name, &ip->i_num, ip->i_di.di_type);
	if (error)
		goto fail_end_trans;

	error = ogfs_change_nlink(ip, +1);
	if (error)
		goto fail_end_trans;

	ogfs_trans_end(sdp);

	ogfs_inplace_relse(dip);
	ogfs_gunlock_m(sdp, 2, lock);

	atomic_inc(&ip->i_vino->i_count);

	d_instantiate(dentry, ip->i_vino);

	EXIT(GFN_LINK);

	return 0;

fail_end_trans:
	ogfs_trans_mark(sdp, TRF_POISONED);
	ogfs_trans_end(sdp);

fail_ipres:
	ogfs_inplace_relse(dip);

fail_gunlock:
	ogfs_gunlock_m(sdp, 2, lock);

fail:
	EXIT(GFN_LINK);
	return error;
}

static int
ogfs_dir_permission(struct inode *dir, int mask)
{
	struct task_struct *tsk = current;
	struct ogfs_inode *dip;
	int error, i;

	error = permission(dir, mask, NULL);
	if (error)
		return error;

	dip = vi2ino(dir);
	if (!dip)
		return -ENOENT;  /* kludge */

	if (dip->i_di.di_mode & S_ISVTX)
		return 0;
	if (dip->i_di.di_uid == tsk->fsuid)
		return 0;
	if (dip->i_di.di_gid == tsk->fsgid)
		return 0;

	for (i = 0; i < tsk->ngroups; i++)
		if (dip->i_di.di_gid == tsk->groups[i])
			return 0;
	if (capable(CAP_FOWNER))
		return 0;
	return -EPERM;
}


/**
 * ogfs_unlink - Unlink a file
 * @dir: The inode of the directory containing the file to unlink
 * @dentry: The file itself
 *
 * Unlink a file.  Call ogfs_unlinki()
 *
 * Returns: 0 on success, -EXXXX on failure
 */

static int
ogfs_unlink(struct inode *dir, struct dentry *dentry)
{
	ogfs_inode_t *dip = vi2ino(dir);
	ogfs_sbd_t *sdp = dip->i_sbd;
	ogfs_inode_t *ip = vi2ino(dentry->d_inode);
	ogfs_lockop_t lock[2];
	int error;

	ENTER(GFN_UNLINK);

	error = ogfs_limit_unlinked(sdp);
	if (error)
		goto fail;

	memset(lock, 0, 2 * sizeof(ogfs_lockop_t));

	lock[0].lo_ip = dip;
	lock[1].lo_ip = ip;

	error = ogfs_glock_m(sdp, 2, lock);
	if (error)
		goto fail;

	error = ogfs_dir_permission(dir, MAY_WRITE | MAY_EXEC);
	if (error)
		goto fail_gunlock;

	error = -ENOENT;
	if (!ogfs_drevalidate(dentry, 0))
		goto fail_gunlock;

	/*
	 * Trans may require:  Two dinode blocks and one modified directory
	 * leaf block and one nopen tag.
	 */
	error = ogfs_trans_begin(sdp, TRANS_UNLINK, 3, 1);
	if (error)
		goto fail_gunlock;

	error = ogfs_unlinki(dip, &dentry->d_name, ip);
	if (error)
		goto fail_end_trans;

	ogfs_trans_end(sdp);

	ogfs_gunlock_m(sdp, 2, lock);

	if (!error)
		d_delete(dentry);

	EXIT(GFN_UNLINK);

	return 0;

fail_end_trans:
	ogfs_trans_mark(sdp, TRF_POISONED);
	ogfs_trans_end(sdp);

fail_gunlock:
	ogfs_gunlock_m(sdp, 2, lock);

fail:
	EXIT(GFN_UNLINK);
	return error;
}

/**
 * ogfs_symlink - Create a symlink
 * @dir: The directory to create the symlink in
 * @dentry: The dentry to put the symlink in
 * @symname: The thing which the link points to
 *
 * Returns: 0 on success, -EXXXX on failure
 */

static int
ogfs_symlink(struct inode *dir, struct dentry *dentry, const char *symname)
{
	ogfs_inode_t *dip = vi2ino(dir), *ip;
	ogfs_sbd_t *sdp = dip->i_sbd;
	struct inode *inode;
	struct buffer_head *dibh;
	int size;
	int error;

	ENTER(GFN_SYMLINK);

	/*  Must be stuffed with a null terminator for ogfs_follow_link()  */

	size = strlen(symname);
	if (size > sdp->sd_sb.sb_bsize - sizeof(ogfs_dinode_t) - 1) {
		error = -ENAMETOOLONG;
		goto fail;
	}

	error =
	    ogfs_createi(dip, &dentry->d_name, FILE_LNK, TRUE, NULL, &ip);
	if (error)
		goto fail;

	OGFS_ASSERT(ip,);

	ip->i_di.di_mode = 0777;
	ip->i_di.di_uid = current->fsuid;
	if (dip->i_di.di_mode & S_ISGID)
		ip->i_di.di_gid = dip->i_di.di_gid;
	else
		ip->i_di.di_gid = current->fsgid;
	ip->i_di.di_size = size;
	ip->i_di.di_type = FILE_LNK;

	dibh = ogfs_get_inode_buffer(ip);
	OGFS_ASSERT(dibh,);

	ogfs_dinode_out(&ip->i_di, BH_DATA(dibh));
	memcpy(BH_DATA(dibh) + sizeof(ogfs_dinode_t), symname, size);

	brelse(dibh);

	ogfs_trans_end(sdp);

	ogfs_inplace_relse(dip);

	ogfs_gunlock_i(ip, 0);
	ogfs_gunlock_i(dip, 0);

	inode = ogfs_iget(sdp->sd_vsb, ip);
	ogfs_put_istruct(ip);

	d_instantiate(dentry, inode);

	EXIT(GFN_SYMLINK);

	return 0;

fail:
	EXIT(GFN_SYMLINK);
	return error;
}

/**
 * ogfs_mkdir - Make a directory
 * @dir: The parent directory of the new one
 * @dentry: The dentry of the new directory
 * @mode: The mode of the new directory
 *
 * Returns: 0 on success, -EXXXX on failure
 */

static int
ogfs_mkdir(struct inode *dir, struct dentry *dentry, int mode)
{
	ogfs_inode_t *dip = vi2ino(dir), *ip;
	ogfs_sbd_t *sdp = dip->i_sbd;
	struct inode *inode;
	struct buffer_head *dibh;
	ogfs_dinode_t *di;
	ogfs_dirent_t *dent;
	int error;

	ENTER(GFN_MKDIR);

	error =
	    ogfs_createi(dip, &dentry->d_name, FILE_DIR, TRUE, NULL, &ip);
	if (error) {
		EXIT(GFN_MKDIR);
		return error;
	}

	OGFS_ASSERT(ip,);

	dibh = ogfs_get_inode_buffer(ip);
	OGFS_ASSERT(dibh,);

	di = (ogfs_dinode_t *) BH_DATA(dibh);

	di->di_uid = cpu_to_ogfs32(current->fsuid);
	if (dip->i_di.di_mode & S_ISGID) {
		di->di_gid = cpu_to_ogfs32(dip->i_di.di_gid);
		mode |= S_ISGID;
	} else
		di->di_gid = cpu_to_ogfs32(current->fsgid);
	di->di_mode = cpu_to_ogfs32(mode);
	di->di_nlink = cpu_to_ogfs32(2);
	di->di_size = cpu_to_ogfs64(sdp->sd_sb.sb_bsize - sizeof(ogfs_dinode_t));

	di->di_flags = cpu_to_ogfs32(OGFS_DIF_JDATA);
	di->di_payload_format = cpu_to_ogfs32(OGFS_FORMAT_DE);
	di->di_type = cpu_to_ogfs16(FILE_DIR);

	error = ogfs_dirent_alloc(ip, dibh, 1, &dent);
	OGFS_ASSERT(!error,);	/*  This should never fail  */

	dent->de_inum = di->di_num;	/*  already OGFS endian  */
	dent->de_hash = ogfs_hash(".", 1);
	dent->de_hash = cpu_to_ogfs32(dent->de_hash);
	dent->de_type = cpu_to_ogfs16(FILE_DIR);
	memcpy((char *) (dent + 1), ".", 1);
	di->di_entries = cpu_to_ogfs32(1);

	error = ogfs_dirent_alloc(ip, dibh, 2, &dent);
	OGFS_ASSERT(!error,);	/*  This should never fail  */

	ogfs_inum_out(&dip->i_num, (char *) &dent->de_inum);
	dent->de_hash = ogfs_hash("..", 2);
	dent->de_hash = cpu_to_ogfs32(dent->de_hash);
	dent->de_type = cpu_to_ogfs16(FILE_DIR);
	memcpy((char *) (dent + 1), "..", 2);
	di->di_entries = cpu_to_ogfs32(2);

	ogfs_dinode_in(&ip->i_di, (char *) di);

	brelse(dibh);

	error = ogfs_change_nlink(dip, +1);
	OGFS_ASSERT(!error,);

	ogfs_trans_end(sdp);

	ogfs_inplace_relse(dip);

	ogfs_gunlock_i(ip, 0);
	ogfs_gunlock_i(dip, 0);

	inode = ogfs_iget(sdp->sd_vsb, ip);
	ogfs_put_istruct(ip);

	d_instantiate(dentry, inode);

	EXIT(GFN_MKDIR);

	return 0;
}

/**
 * ogfs_rmdir - Remove a directory
 * @dir: The parent directory of the directory to be removed
 * @dentry: The dentry of the directory to remove
 *
 * Remove a directory. Call ogfs_rmdiri()
 *
 * Returns: 0 on success, -EXXXX on failure
 */

static int
ogfs_rmdir(struct inode *dir, struct dentry *dentry)
{
	ogfs_inode_t *dip = vi2ino(dir);
	ogfs_sbd_t *sdp = dip->i_sbd;
	ogfs_inode_t *ip = vi2ino(dentry->d_inode);
	ogfs_lockop_t lock[2];
	int error;

	ENTER(GFN_RMDIR);

	error = ogfs_limit_unlinked(sdp);
	if (error)
		goto fail;

	memset(lock, 0, 2 * sizeof(ogfs_lockop_t));

	lock[0].lo_ip = dip;
	lock[1].lo_ip = ip;

	error = ogfs_glock_m(sdp, 2, lock);
	if (error)
		goto fail;

	error = ogfs_dir_permission(dir, MAY_WRITE | MAY_EXEC);
	if (error)
		goto fail_gunlock;

	error = -ENOENT;
	if (!ogfs_drevalidate(dentry, 0))
		goto fail_gunlock;

	/*
	 * Trans may require:  Two dinode blocks, one directory leaf block
	 * containing the entry to be rmdired, two leaf blocks containing
	 * . and .. of the directory being rmdired, and one nopen tag.
	 */
	error = ogfs_trans_begin(sdp, TRANS_RMDIR, 5, 1);
	if (error)
		goto fail_gunlock;

	error = ogfs_rmdiri(dip, &dentry->d_name, ip);
	if (error)
		goto fail_end_trans;

	ogfs_trans_end(sdp);

	ogfs_gunlock_m(sdp, 2, lock);

	if (!error)
		d_delete(dentry);

	EXIT(GFN_RMDIR);

	return 0;

fail_end_trans:
	ogfs_trans_mark(sdp, TRF_POISONED);
	ogfs_trans_end(sdp);

fail_gunlock:
	ogfs_gunlock_m(sdp, 2, lock);

fail:
	EXIT(GFN_RMDIR);
	return error;
}

/**
 * ogfs_mknod - Make a special file
 * @dir: The directory in which the special file will reside
 * @dentry: The dentry of the special file
 * @mode: The mode of the special file
 * @rdev: The device specification of the special file
 *
 */

int
ogfs_mknod(struct inode *dir, struct dentry *dentry, int mode, dev_t rdev)
{
	ogfs_inode_t *dip = vi2ino(dir), *ip;
	ogfs_sbd_t *sdp = dip->i_sbd;
	struct inode *inode;
	struct buffer_head *dibh;
	uint16 type = 0;
	uint32 major = 0, minor = 0;
	int error;

	ENTER(GFN_MKNOD);

	switch (mode & S_IFMT) {
	case S_IFBLK:
		type = FILE_BLK;
		major = imajor(dir);
		minor = iminor(dir);
		break;
	case S_IFCHR:
		type = FILE_CHR;
		major = imajor(dir);
		minor = iminor(dir);
		break;
	case S_IFIFO:
		type = FILE_FIFO;
		break;
	case S_IFSOCK:
		type = FILE_SOCK;
		break;
	default:
		OGFS_ASSERT(0, con_printf("mode = %d\n", mode););
		break;
	};

	error = ogfs_createi(dip, &dentry->d_name, type, TRUE, NULL, &ip);
	if (error) {
		EXIT(GFN_MKNOD);
		return error;
	}

	OGFS_ASSERT(ip,);

	ip->i_di.di_mode = mode & S_IALLUGO;
	ip->i_di.di_uid = current->fsuid;
	if (dip->i_di.di_mode & S_ISGID)
		ip->i_di.di_gid = dip->i_di.di_gid;
	else
		ip->i_di.di_gid = current->fsgid;
	ip->i_di.di_major = major;
	ip->i_di.di_minor = minor;

	ip->i_di.di_type = type;

	dibh = ogfs_get_inode_buffer(ip);
	OGFS_ASSERT(dibh,);

	ogfs_dinode_out(&ip->i_di, BH_DATA(dibh));

	brelse(dibh);

	ogfs_trans_end(sdp);

	ogfs_inplace_relse(dip);

	ogfs_gunlock_i(ip, 0);
	ogfs_gunlock_i(dip, 0);

	inode = ogfs_iget(sdp->sd_vsb, ip);
	ogfs_put_istruct(ip);

	d_instantiate(dentry, inode);

	EXIT(GFN_MKNOD);

	return 0;
}

/**
 * ogfs_is_subdir - Return TRUE if @new_dentry is a subdir of @old_dentry
 * @new_dentry:
 * @old_dentry:
 *
 * Do the same thing as Linux's is_subdir(), but do drevalidates to make
 * sure the dcache is consistent.
 *
 * Returns: TRUE if new_dentry is a subdir of old_dentry, FALSE otherwise
 */

static int
ogfs_is_subdir(struct dentry *new_dentry, struct dentry *old_dentry)
{
	int ret = FALSE;
	int error;

	ENTER(GFN_IS_SUBDIR);

	while (TRUE) {
		if (new_dentry != old_dentry) {
			if (new_dentry->d_parent == new_dentry)
				break;

			if (!ogfs_drevalidate(new_dentry, 0)) {
				error = -ENOENT;
				EXIT(GFN_IS_SUBDIR);
				return error;
			}

			new_dentry = new_dentry->d_parent;
			continue;
		}

		ret = TRUE;
		break;
	}

	EXIT(GFN_IS_SUBDIR);

	return ret;
}

/**
 * ogfs_rename - Rename a file
 * @odir: Parent directory of old file name
 * @odentry: The old dentry of the file
 * @ndir: Parent directory of new file name
 * @ndentry: The new dentry of the file
 *
 * Returns: 0 on success, -EXXXX on failure
 */

int
ogfs_rename(struct inode *odir, struct dentry *odentry,
	    struct inode *ndir, struct dentry *ndentry)
{
	ogfs_inode_t *odip = vi2ino(odir), *ndip = vi2ino(ndir);
	ogfs_sbd_t *sdp = odip->i_sbd;
	ogfs_inode_t *ip = vi2ino(odentry->d_inode), *nip = NULL;
	struct qstr name;
	ogfs_ipreserv_t *res = &ndip->i_res;
	ogfs_lockop_t lock[4];
	int dir_rename = FALSE;
	int error;
	int num_lock;

	ENTER(GFN_RENAME);

	error = ogfs_limit_unlinked(sdp);
	if (error)
		goto fail;

	if (ndentry->d_inode) {
		nip = vi2ino(ndentry->d_inode);

		if (ip == nip) {
			error = 0;
			goto fail;
		}
	}

	/*  Make sure we aren't trying to move a directory into its subdir  */

	if (ip->i_di.di_type == FILE_DIR && odip != ndip) {
		error = ogfs_glock(sdp, sdp->sd_rename_gl, 0);
		if (error)
			goto fail;

		error = ogfs_is_subdir(ndentry->d_parent, odentry);
		switch (error) {
		case TRUE:
			error = -EINVAL;
			goto fail_renlock;
			break;
		case FALSE:
			break;
		default:
			goto fail_renlock;
			break;
		}

		dir_rename = TRUE;
	}

	memset(lock, 0, 4 * sizeof(ogfs_lockop_t));

	lock[0].lo_ip = odip;
	lock[1].lo_ip = ndip;

	num_lock=2;

	if (nip) {
		lock[num_lock].lo_ip = nip;
		num_lock++;
	}

	if (dir_rename) {
		lock[num_lock].lo_ip = ip;
		num_lock++;
	}

	error = ogfs_glock_m(sdp, 4, lock);
	if (error)
		goto fail_renlock;

	/*  Check out the old directory  */

	error = ogfs_dir_permission(odir, MAY_WRITE | MAY_EXEC);
	if (error)
		goto fail_gunlock;

	if (!ogfs_drevalidate(odentry, 0)) {
		error = -ENOENT;
		goto fail_gunlock;
	}

	/*
	 * Check out the new directory.
	 */
	error = ogfs_dir_permission(ndir, MAY_WRITE | MAY_EXEC);
	if (error == -ENOENT) {
		error = ogfs_dir_search(ndip, &ndentry->d_name, NULL, NULL);
		switch (error) {
		case -ENOENT:
			error = 0;
			break;
		case 0:
			error = -EEXIST;
		default:
			goto fail_gunlock;
		}
	} else if (error)
		goto fail_gunlock;

	if (nip && !ogfs_drevalidate(ndentry, 0)) {
		error = -ENOENT;
		goto fail_gunlock;
	}

	res->ir_diblks = 0;
	res->ir_metablks = sdp->sd_max_dirres;
	res->ir_datablks = 0;
	res->ir_target = ndip->i_di.di_goal_rgrp;
	res->ir_trans = TRANS_RENAME;

	error = ogfs_inplace_reserv(ndip, 1);
	if (error)
		goto fail_gunlock;

	/*  Trans may require:
	   Dinodes for the srcdir, srcino, dstdir, dstino.  Blocks for
	   adding the entry to dstdir.  RG bitmaps for that allocation.
	   One leaf block in the srcdir for removal of the entry.
	   One leaf block for changing .. in srcino (if it's a directory).
	   Two leaf blocks for removing . and .. from dstino (if it exists
	   and it's a directory) and one nopen tag.  */

	error =
	    ogfs_trans_begin(sdp, TRANS_RENAME,
			    8 + sdp->sd_max_dirres +
			    res->ir_rgd->rd_ri.ri_length, 1);
	if (error)
		goto fail_ipres;

	/*  Remove the target file, if it exists  */

	if (nip) {
		if (nip->i_di.di_type == FILE_DIR)
			error = ogfs_rmdiri(ndip, &ndentry->d_name, nip);
		else
			error = ogfs_unlinki(ndip, &ndentry->d_name, nip);

		if (error)
			goto fail_end_trans;
	}

	if (dir_rename) {
		ogfs_change_nlink(ndip, +1);
		ogfs_change_nlink(odip, -1);
		name.len = 2;
		name.name = "..";

		error = ogfs_dir_mvino(ip, &name, &ndip->i_num, FILE_DIR);
		if (error)
			goto fail_end_trans;
	}

	error =
	    ogfs_dir_add(ndip, &ndentry->d_name, &ip->i_num, ip->i_di.di_type);
	if (error)
		goto fail_end_trans;

	error = ogfs_dir_del(odip, &odentry->d_name);
	if (error)
		goto fail_end_trans;

	if (dir_rename)
		ogfs_trans_add_gl(sdp, sdp->sd_rename_gl);

	ogfs_trans_end(sdp);

	ogfs_inplace_relse(ndip);

	ogfs_gunlock_m(sdp, num_lock, lock);

	if (dir_rename)
		ogfs_gunlock(sdp, sdp->sd_rename_gl, 0);

	EXIT(GFN_RENAME);

	return 0;

fail_end_trans:
	ogfs_trans_mark(sdp, TRF_POISONED);
	ogfs_trans_end(sdp);

fail_ipres:
	ogfs_inplace_relse(ndip);

fail_gunlock:
	ogfs_gunlock_m(sdp, num_lock, lock);

fail_renlock:
	if (dir_rename)
		ogfs_gunlock(sdp, sdp->sd_rename_gl, 0);

fail:
	EXIT(GFN_RENAME);
	return error;
}

/**
 * ogfs_irevalidate - Check inode hasn't changed
 * @dentry: The dentry
 *
 * Revalidate the inode. i.e. grab the Glock and sync the incore dinode
 * with the ondisk inode
 *
 * Returns: 0 on success, or error
 */

int
ogfs_irevalidate(struct dentry *dentry)
{
	ogfs_inode_t *ip = vi2ino(dentry->d_inode);
	int error;

	ENTER(GFN_IREVALIDATE);

	lock_kernel();

	error = ogfs_glock_i(ip, GL_SHARED);
	if (!error)
		ogfs_gunlock_i(ip, 0);

	unlock_kernel();

	EXIT(GFN_IREVALIDATE);

	return error;
}

/**
 * ogfs_readlink - Read the value of a symlink
 * @dentry: the symlink
 * @buf: the buffer to read the symlink data into
 * @size: the size of the buffer
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
ogfs_readlink(struct dentry *dentry, char *buf, int size)
{
	ogfs_inode_t *ip = vi2ino(dentry->d_inode);
	struct buffer_head *dibh;
	int error;

	ENTER(GFN_READLINK);

	lock_kernel();

	error = ogfs_glock_i(ip, GL_SHARED);
	if (error)
		goto fail;

	error = OGFS_UPDATE_ATIME(ip);
	if (error)
		goto fail_gunlock;

	dibh = ogfs_get_inode_buffer(ip);
	if (!dibh) {
		error = -EIO;
		goto fail_gunlock;
	}

	if (size > ip->i_di.di_size)
		size = ip->i_di.di_size;

	if (copy_to_user(buf, BH_DATA(dibh) + sizeof(ogfs_dinode_t), size)) {
		error = -EFAULT;
		goto fail_drelse;
	}

	brelse(dibh);

	ogfs_gunlock_i(ip, 0);

	unlock_kernel();

	EXIT(GFN_READLINK);

	return size;

fail_drelse:
	brelse(dibh);

fail_gunlock:
	ogfs_gunlock_i(ip, 0);

fail:
	unlock_kernel();
	EXIT(GFN_READLINK);
	return error;
}

/**
 * sym_hostname - Substitute system host name
 * @str: String for result
 * @len: Length of result buffer
 *
 * Returns: Length of hostname
 */

static unsigned int
sym_hostname(char *str)
{
	unsigned int l = strlen(system_utsname.nodename);

	if (str)
		memcpy(str, system_utsname.nodename, l);

	return l;
}

/**
 * sym_machine - Substitute machine type
 * @str: String for result
 * @len: Length of result buffer
 *
 * Returns: Length of machine type
 */

static unsigned int
sym_machine(char *str)
{
	unsigned int l = strlen(system_utsname.machine);

	if (str)
		memcpy(str, system_utsname.machine, l);

	return l;
}

/**
 * sym_os - Substitute OS name
 * @str: String for result
 * @len: Length of result buffer
 *
 * Returns: Length of OS name
 */

static unsigned int
sym_os(char *str)
{
	unsigned int l = strlen(system_utsname.sysname);

	if (str)
		memcpy(str, system_utsname.sysname, l);

	return l;
}

static struct ogfs_symlink_ops symlink_ops[] = {
	{"hostname}", 9, sym_hostname},
	{"mach}", 5, sym_machine},
	{"os}", 3, sym_os},
	{NULL, 0, NULL}
};

/**
 * ogfs_follow_link - Follow a symbolic link
 * @dentry: The dentry of the link
 * @nd: Data that we pass to vfs_follow_link()
 *
 * We do CDSL expansion for the following strings:
 * {hostname}, {mach}, and {os}.  (The values substituted are the outputs
 * of "uname -n", "uname -m", and "uname -s" respectively.)
 *
 * This can handle symlinks of any size. It is optimised for symlinks
 * under OGFS_FAST_SYMLINK_SIZE.
 *
 * Returns: 0 on success or error code
 */

int
ogfs_follow_link(struct dentry *dentry, struct nameidata *nd)
{
	ogfs_inode_t *ip = vi2ino(dentry->d_inode);
	struct buffer_head *dibh;
	char array[OGFS_FAST_SYMLINK_SIZE];
	char *buf = array;
	unsigned int len = 0;
	int error;

	ENTER(GFN_FOLLOW_LINK);

	lock_kernel();

	error = ogfs_glock_i(ip, GL_SHARED);
	if (error)
		goto fail;

	error = OGFS_UPDATE_ATIME(ip);
	if (error)
		goto fail_gunlock;

	dibh = ogfs_get_inode_buffer(ip);
	if (!dibh) {
		error = -EIO;
		goto fail_gunlock;
	}

	/*  Don't forget the trailing zero  */

	len =
	    ogfs_link_size(symlink_ops,
			  BH_DATA(dibh) + sizeof(ogfs_dinode_t)) + 1;

	if (len > OGFS_FAST_SYMLINK_SIZE) {
		buf = gmalloc(len);
		if (!buf) {
			error = -ENOMEM;
			goto fail_drelse;
		}
	}

	ogfs_link_expand(symlink_ops, buf, BH_DATA(dibh) + sizeof(ogfs_dinode_t));

	IFDB(GFN_FOLLOW_LINK,
	     db_printf("  symlink substituted %s -> %s\n",
		       BH_DATA(dibh) + sizeof(ogfs_dinode_t), buf););

	brelse(dibh);

	ogfs_gunlock_i(ip, 0);

	error = vfs_follow_link(nd, buf);

	if (buf != array)
		gfree(buf, len);

	unlock_kernel();

	EXIT(GFN_FOLLOW_LINK);

	return error;

fail_drelse:
	brelse(dibh);

fail_gunlock:
	ogfs_gunlock_i(ip, 0);

fail:
	unlock_kernel();
	EXIT(GFN_FOLLOW_LINK);
	return error;
}

/**
 * ogfs_get_block - Fills in a buffer head with details about a block
 * @inode: The inode
 * @iblock: The block number to look up
 * @bh_result: The buffer head to return the result in
 * @create: Non-zero if we may add block to the file
 *
 */

static int
ogfs_get_block(struct inode *inode, sector_t iblock, struct buffer_head *bh_result,
	      int create)
{
	ogfs_inode_t *ip = vi2ino(inode);
	int new = create;
	uint64 block;
	int error;

	ENTER(GFN_GET_BLOCK);

	lock_kernel();

	OGFS_ASSERT(atomic_read(&ip->i_gl->gl_locked),);
	OGFS_ASSERT(ip->i_gl->gl_pid == current->pid,);

	if (iblock < 0) {
		con_printf("OGFS:  ogfs_get_block() called with iblock < 0\n");
		error = -EINVAL;
		goto out;
	}

	error = ogfs_inode_get_block(ip, (uint64) iblock, &new, &block);
	if (error)
		goto out;

	OGFS_ASSERT(block || !create,);

	if (!block)
		goto out;

#if (BITS_PER_LONG < 64)
	if (block & 0xFFFFFFFF80000000ll) {
		con_printf
		    ("OGFS:  ogfs_get_block():  block maps out of \"unsigned long\" range\n");
		error = -EINVAL;
		goto out;
	}
#endif

	bh_result->b_bdev = inode->i_bdev;
	bh_result->b_blocknr = (unsigned long) block;
	bh_result->b_state |= (1UL << BH_Mapped);
	if (new)
		bh_result->b_state |= (1UL << BH_New);

out:
	unlock_kernel();

	EXIT(GFN_GET_BLOCK);

	return error;
}

/**
 * stuffed_readpage - Fill in a Linux page with stuffed file data
 * @ip: the file
 * @page: the page
 *
 * Lots of stuff lifted out of generic_readpage() and
 *   brw_page().
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
stuffed_readpage(ogfs_inode_t * ip, struct page *page)
{
	struct buffer_head *bh;
	void *kaddr;

	ENTER(GFN_STUFFED_READPAGE);

	OGFS_ASSERT(PageLocked(page),);

	bh = ogfs_get_inode_buffer(ip);
	if (!bh) {
		EXIT(GFN_STUFFED_READPAGE);
		return -EIO;
	}

	kaddr = kmap(page);
	memcpy((char *) kaddr, BH_DATA(bh) + sizeof(ogfs_dinode_t),
		   ip->i_di.di_size);
	memset((char *) kaddr + ip->i_di.di_size, 0,
		   PAGE_CACHE_SIZE - ip->i_di.di_size);
	kunmap(page);

	brelse(bh);

	SetPageUptodate(page);

	EXIT(GFN_STUFFED_READPAGE);

	return 0;
}

/**
 * ogfs_readpage - Read a page of data for a file
 * @d: The dentry for the file
 * @page: The page to write data into
 *
 * Returns:  0 on success, -EXXX on failure
 */

static int
ogfs_readpage(struct file *filp, struct page *page)
{
	struct inode *inode = page->mapping->host;
	ogfs_inode_t *ip = vi2ino(inode);
	int error;

	ENTER(GFN_READPAGE);

	lock_kernel();

	error = ogfs_glock_i(ip, GL_SHARED);
	if (error) {
		unlock_page(page);
		goto out;
	}

	if (ogfs_is_stuffed(ip) && !page->index) {
		error = stuffed_readpage(ip, page);
		unlock_page(page);
	} else
		error = block_read_full_page(page, ogfs_get_block);

	ogfs_gunlock_i(ip, 0);

      out:
	unlock_kernel();

	EXIT(GFN_READPAGE);

	return error;
}

/**
 * stuffed_writepage - Write to a stuffed inode
 * @inode: The inode to write to
 * @page: The page containing the data
 * @from: From (byte range within page)
 * @to: To (byte range within page)
 *
 * This helper routine is used to write data into a stuffed dinode.
 */

static int
stuffed_writepage(struct inode *inode, struct page *page, unsigned long from,
		  unsigned long to)
{
	ogfs_inode_t *ip = vi2ino(inode);
	ogfs_sbd_t *sdp = ip->i_sbd;
	struct buffer_head *bh;
	void *kaddr;

	ENTER(GFN_STUFFED_WRITEPAGE);

	OGFS_ASSERT(PageLocked(page),);

	bh = ogfs_get_inode_buffer(ip);
	if (!bh) {
		EXIT(GFN_STUFFED_WRITEPAGE);
		return -EIO;
	}

	ogfs_trans_add_bh(sdp, bh);

	kaddr = kmap(page);
	memcpy(BH_DATA(bh) + sizeof(ogfs_dinode_t) + from,
		   (char *) kaddr + from, to - from);
	kunmap(page);

	brelse(bh);

	SetPageUptodate(page);

	EXIT(GFN_STUFFED_WRITEPAGE);

	return 0;
}

/**
 * ogfs_writepage - Write complete page
 * @page: Page to write
 *
 */

static int
ogfs_writepage(struct page *page, struct writeback_control *wbc)
{
	struct inode *inode = page->mapping->host;
	ogfs_inode_t *ip = vi2ino(inode);
	ogfs_sbd_t *sdp = ip->i_sbd;
	ogfs_ipreserv_t *res = &ip->i_res;
	int data_blocks = PAGE_CACHE_SIZE >> sdp->sd_sb.sb_bsize_shift;
	int ind = 2 * sdp->sd_max_height;
	int error;

	ENTER(GFN_WRITEPAGE);

	lock_kernel();

	error = ogfs_glock_i(ip, 0);
	if (error)
		goto fail_glock;

	OGFS_ASSERT(!ogfs_is_stuffed(ip) ||
		   ip->i_di.di_size <=
		   sdp->sd_sb.sb_bsize - sizeof(ogfs_dinode_t),);

	res->ir_diblks = 0;
	res->ir_metablks = ind;
	res->ir_datablks = data_blocks;
	res->ir_target = ip->i_di.di_goal_rgrp;
	res->ir_trans = TRANS_WRITEPAGE;

	error = ogfs_inplace_reserv(ip, 0);
	if (error)
		goto fail;

	/*  Trans may require:
	   a dinode block, RG bitmaps to allocate from, and indirect blocks  */

	error =
	    ogfs_trans_begin(sdp, TRANS_WRITEPAGE,
			    1 + res->ir_rgd->rd_ri.ri_length + ind, 0);
	if (error)
		goto fail_ipres;

	if (ogfs_is_stuffed(ip)) {
		error = stuffed_writepage(inode, page, 0, ip->i_di.di_size);
		unlock_page(page);
	} else
		error = block_write_full_page(page, ogfs_get_block, NULL);

	if (error)
		goto fail_end_trans;

	ogfs_trans_end(sdp);

	ogfs_inplace_relse(ip);

	ogfs_gunlock_i(ip, 0);

	unlock_kernel();

	EXIT(GFN_WRITEPAGE);

	return 0;

fail_end_trans:
	ogfs_trans_mark(sdp, TRF_POISONED);
	ogfs_trans_end(sdp);

fail_ipres:
	ogfs_inplace_relse(ip);

fail:
	ClearPageUptodate(page);
	if (PageLocked(page))
		unlock_page(page);
	ogfs_gunlock_i(ip, 0);

fail_glock:
	unlock_kernel();
	EXIT(GFN_WRITEPAGE);
	return error;
}

/**
 * ogfs_prepare_write - Prepare to write to a file
 * @file: The file to write to
 * @page: The page which is to be prepared for writing
 * @from: From (byte range within page)
 * @to: To (byte range within page)
 *
 */

static int
ogfs_prepare_write(struct file *file, struct page *page, unsigned from,
		  unsigned to)
{
	struct inode *inode = file->f_dentry->d_inode;
	ogfs_inode_t *ip = vi2ino(inode);
	ogfs_sbd_t *sdp = ip->i_sbd;
	int error = 0;

	ENTER(GFN_PREPARE_WRITE);

	if (ogfs_is_stuffed(ip)) {
		if (((uint64) page->index << PAGE_CACHE_SHIFT) + to >
		    sdp->sd_sb.sb_bsize - sizeof(ogfs_dinode_t)) {
			error = ogfs_unstuff_dinode(ip);
			if (error)
				goto out;
			error =
			    block_prepare_write(page, from, to, ogfs_get_block);
#warning another fixme
//		} else if (!Page_Uptodate(page))
		} else if (!test_bit(PG_uptodate, &(page)->flags))
			error = stuffed_readpage(ip, page);
	} else
		error = block_prepare_write(page, from, to, ogfs_get_block);

out:
	EXIT(GFN_PREPARE_WRITE);

	return error;
}

/**
 * ogfs_commit_write - Commit write to a file
 * @file: The file to write to
 * @page: The page containing the data
 * @from: From (byte range within page)
 * @to: To (byte range within page)
 *
 */

static int
ogfs_commit_write(struct file *file, struct page *page, unsigned from,
		 unsigned to)
{
	struct inode *inode = file->f_dentry->d_inode;
	ogfs_inode_t *ip = vi2ino(inode);
	int error;

	ENTER(GFN_COMMIT_WRITE);

	if (ogfs_is_stuffed(ip)) {
		error = stuffed_writepage(inode, page, from, to);
		if (error)
			ClearPageUptodate(page);
	} else
		error = generic_commit_write(file, page, from, to);

	EXIT(GFN_COMMIT_WRITE);

	return error;
}

/**
 * ogfs_bmap - Block map function
 * @mapping: Address space info
 * @block: The block to map
 *
 */

static sector_t
ogfs_bmap(struct address_space *mapping, sector_t block)
{
	struct inode *inode = mapping->host;
	ogfs_inode_t *ip = vi2ino(inode);
	sector_t rv;

	ENTER(GFN_BMAP);

	lock_kernel();

	rv = ogfs_glock_i(ip, GL_SHARED);
	if (rv)
		goto out;

	if (ogfs_is_stuffed(ip))
		rv = 0;
	else
		rv = generic_block_bmap(inode->i_mapping, block, ogfs_get_block);

	ogfs_gunlock_i(ip, 0);

out:
	unlock_kernel();

	EXIT(GFN_BMAP);
	return rv;
}

/**
 * ogfs_permission - check permissions on a file
 * @ip: the file
 * @mask: the type of access we want
 * @cred: the structure defining our credentials
 *
 * Returns: 0 for Permission Granted, -EXXX on failure
 */

int
ogfs_permission(ogfs_inode_t * ip, int mask)
{
	int error;

	ENTER(GFN_PERMISSION);

	OGFS_ASSERT(ip->i_vino,);
	error = permission(ip->i_vino, mask, NULL);

	EXIT(GFN_PERMISSION);

	return error;
}

/**
 * shrink_icache_sb - toss unused inodes for a superblock
 * @sb: superblock
 *
 * Discard unused inodes for a given superblock being sure to bypass
 * inodes which are LOCKed.
 */
/*
int shrink_icache_sb(struct super_block * sb)
{
	LIST_HEAD(dispose);
	struct list_head *next;
	int count = 0;

        spin_lock(&inode_lock);

	next = inode_unused.next;
	for (;;) {
	struct list_head * tmp = next;
		struct inode * inode;

		next = next->next;
		if (tmp == &inode_unused)
			break;
		inode = list_entry(tmp, struct inode, i_list);

		if (inode->i_sb != sb)
			continue;

		if (inode->i_state & I_LOCK)
			continue;

		if (atomic_read(&inode->i_count))
			BUG();

		//list_del(&inode->i_hash);
		//INIT_LIST_HEAD(&inode->i_hash);
		list_del(&inode->i_list);
		list_add(&inode->i_list, &dispose);
		inode->i_state = I_FREEING;
		count++;
	}
	inodes_stat.nr_unused -= count;
	spin_unlock(&inode_lock);

	{
		struct inode *inode;
		struct list_head *tmp;

		for (tmp = dispose.next; tmp != &dispose; tmp = tmp->next) {
			inode = list_entry(tmp, struct inode, i_list);

			filemap_fdatasync(inode->i_mapping);
			fsync_inode_buffers(inode);
			fsync_inode_data_buffers(inode);
			filemap_fdatawait(inode->i_mapping);
		}
	}

	dispose_list(&dispose);
	return count;
}
*/

/**
 * ogfs_drop_excess_inodes - drop unused inodes
 * @sdp: the filesystem
 *
 */

void
ogfs_drop_excess_inodes(ogfs_sbd_t * sdp)
{
	ENTER(GFN_DROP_EXCESS_INODES);

	if (!sdp->sd_args.ar_localcaching) {
		shrink_dcache_sb(sdp->sd_vsb);
#warning need to figure out what to replace this with
		//shrink_icache_sb(sdp->sd_vsb);
		//shrink_icache_memory(); /* not this */
	}

	EXIT(GFN_DROP_EXCESS_INODES);
}

struct inode_operations ogfs_file_iops = {
	//revalidate:	ogfs_irevalidate,
	setattr:	ogfs_setattr
};

struct address_space_operations ogfs_file_aops = {
	writepage:	ogfs_writepage,
	readpage:	ogfs_readpage,
	sync_page:	block_sync_page,
	prepare_write:	ogfs_prepare_write,
	commit_write:	ogfs_commit_write,
	bmap:		ogfs_bmap,
};

struct inode_operations ogfs_dev_iops = {
	//revalidate:	ogfs_irevalidate,
	setattr:	ogfs_setattr
};

struct inode_operations ogfs_dir_iops = {
	create:		ogfs_create,
	lookup:		ogfs_lookup,
	link:		ogfs_link,
	unlink:		ogfs_unlink,
	symlink:	ogfs_symlink,
	mkdir:		ogfs_mkdir,
	rmdir:		ogfs_rmdir,
	mknod:		ogfs_mknod,
	rename:		ogfs_rename,
	//revalidate:	ogfs_irevalidate,
	setattr:	ogfs_setattr
};

struct inode_operations ogfs_symlink_iops = {
	readlink:	ogfs_readlink,
	follow_link:	ogfs_follow_link,
	//revalidate:	ogfs_irevalidate
};
