/*
 *
 *    Copyright 1995-1999 Regents of the University of Minnesota
 *    Portions Copyright 1999-2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 */

#include "ogfs_debug.h"
#include "ogfs.h"
#include "blklist.h"
#include "bmap.h"
#include "dio.h"
#include "file.h"
#include "glock.h"
#include "glops.h"
#include "inode.h"
#include "super.h"
#include "trans.h"
#include "blklist.h"

#ifdef DEBUG_TRACE
/**
 * ogfs_debug_change - Change debug level
 * @cmd: The command
 * @arg: Pointer to argument
 *
 * Returns: 0 on success or error code
 */

int
ogfs_debug_change(unsigned int cmd, void *arg)
{
	unsigned int flag;
	unsigned char *val;

	ENTER(GFN_DEBUG_CHANGE);

	if (copy_from_user(&flag, arg, sizeof(unsigned int))) {
		EXIT(GFN_DEBUG_CHANGE);
		return -EFAULT;
	}

	if (flag >= DEBUG_FLAGS) {
		EXIT(GFN_DEBUG_CHANGE);
		return -EINVAL;
	}

	val = &ogfs_debug_flags[flag / 8];

	if (cmd == OGFS_DEBUG_ON)
		*val |= 1 << (flag % 8);
	else
		*val &= ~(1 << (flag % 8));

	EXIT(GFN_DEBUG_CHANGE);

	return 0;
}
#endif				/*  DEBUG_TRACE  */

#ifdef DEBUG_PRINT
/**
 * ogfs_debug_dump - Dump debug information
 * @arg: Pointer to buffer
 *
 * Returns: Length of buffer used, or error code on failure
 */

int
ogfs_debug_dump(void *arg)
{
	char *p;
	int error = 0;

	ENTER(GFN_DEBUG_DUMP);

	/*  If they didn't give us a pointer,
	   tell them how big a buffer to pass in.  */

	if (!arg) {
		EXIT(GFN_DEBUG_DUMP);
		return OGFS_ICDB_BUF_LEN;
	}

	p = (char *) arg;

	down(&ogfs_icdb_lock);

	if (copy_to_user(p, ogfs_icdb_buf + ogfs_icdb_point,
	     OGFS_ICDB_BUF_LEN - ogfs_icdb_point))
		error = -EFAULT;
	if (!error && ogfs_icdb_point)
		 if (copy_to_user(p + (OGFS_ICDB_BUF_LEN - ogfs_icdb_point),
				  ogfs_icdb_buf, ogfs_icdb_point))
			 error = -EFAULT;

	up(&ogfs_icdb_lock);

	EXIT(GFN_DEBUG_DUMP);

	return error;
}

/**
 * ogfs_print_frag - Print information about block locations
 * @ip: The inode to print information about
 *
 */

void
ogfs_print_frag(ogfs_inode_t * ip)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	uint64 iblock, block, blocks;
	uint64 start = 0, daddr = 0, run = 0;
	int not_new = 0;
	int error;

	ENTER(GFN_PRINT_FRAG);

	db_printf("Extent list for inode %" FMT_64 "u/%" FMT_64 "u\n",
		  ip->i_num.no_formal_ino, ip->i_num.no_addr);

	error = ogfs_glock_i(ip, GL_SHARED);
	if (error) {
		EXIT(GFN_PRINT_FRAG);
		return;
	}

	if (ogfs_is_stuffed(ip)) {
		db_printf("  stuffed\n");
		ogfs_gunlock_i(ip, 0);
		EXIT(GFN_PRINT_FRAG);
		return;
	}

	db_printf("%15s%15s%15s\n", "logical block", "extent length",
		  "disk block");

	blocks = (ip->i_di.di_size + 1) >> sdp->sd_sb.sb_bsize_shift;
	for (iblock = 0; iblock < blocks; iblock++) {
		error = ogfs_inode_get_block(ip, iblock, &not_new, &block);
		if (error) {
			ogfs_gunlock_i(ip, 0);
			EXIT(GFN_PRINT_FRAG);
			return;
		}

		if (daddr && (daddr + run == block)) {
			run++;
		} else {
			if (daddr)
				db_printf("%15" FMT_64 "u%15" FMT_64 "u%15"
					  FMT_64 "u\n", start, run, daddr);

			start = iblock;
			daddr = block;
			run = 1;
		}
	}

	if (daddr)
		db_printf("%15" FMT_64 "u%15" FMT_64 "u%15" FMT_64 "u\n",
			  start, run, daddr);

	ogfs_gunlock_i(ip, 0);

	EXIT(GFN_PRINT_FRAG);
}
#endif				/*  DEBUG_PRINT  */

#if DEBUG_PROFILE
/**
 * ogfs_prof_dump - Profile dump
 * @arg: The buffer
 *
 * Returns: 0 on success, error code otherwise
 */

int
ogfs_prof_dump(void *arg)
{
	int x;
	int error = 0;

	ENTER(GFN_PROF_DUMP);

	/*  If they didn't give us a pointer,
	   tell them how big a buffer to pass in.  */

	if (!arg) {
		EXIT(GFN_PROF_DUMP);
		return DEBUG_FLAGS;
	}

	con_printf("ogfs_prof_neg_count = %" FMT_64 "u\n", ogfs_prof_neg_count);

	if (copy_to_user
	    (arg, ogfs_prof_stats, DEBUG_FLAGS * sizeof(ogfs_profile_t)))
		error = -EFAULT;
	else {
		memset(ogfs_prof_stats, 0,
			   DEBUG_FLAGS * sizeof(ogfs_profile_t));
		for (x = 0; x < DEBUG_FLAGS; x++)
			ogfs_prof_stats[x].min_micros = (uint64) - 1;
		ogfs_prof_neg_count = 0;
	}

	EXIT(GFN_PROF_DUMP);

	return error;
}
#endif				/*  DEBUG_PROFILE  */

/**
 * ogfs_get_super - Dump the superblock into a buffer
 * @sb: The superblock
 * @ptr: The buffer pointer
 *
 * Returns: 0 or error code
 */

int
ogfs_get_super(ogfs_sbd_t * sdp, void *arg)
{
	ogfs_glock_t *gl;
	struct buffer_head *bh;
	int error;

	ENTER(GFN_GET_SUPER);

	error =
	    ogfs_glock_num(sdp, OGFS_SB_LOCK, &ogfs_meta_glops, NULL, GL_SHARED,
			  &gl);
	if (error)
		goto out;

	bh = ogfs_dread(sdp, OGFS_SB_ADDR >> sdp->sd_fsb2bb_shift, gl,
		       DR_START | DR_WAIT);
	if (!bh) {
		error = -EIO;
		goto out_gunlock;
	}

	error = copy_to_user(arg, BH_DATA(bh), BH_SIZE(bh));

	brelse(bh);

out_gunlock:
	ogfs_gunlock(sdp, gl, 0);
	ogfs_put_glstruct(gl);

out:
	EXIT(GFN_GET_SUPER);
	return error;
}

/**
 * ogfs_jread_ioctl - Read from a journaled data file via ioctl
 * @ip: The OGFS inode
 * @arg: The argument from ioctl
 *
 * Returns: Amount of data copied or error
 */

int
ogfs_jread_ioctl(ogfs_inode_t * ip, void *arg)
{
	ogfs_jio_t jt;
	int ret;

	ENTER(GFN_JREAD_IOCTL);

	if (copy_from_user(&jt, arg, sizeof(ogfs_jio_t))) {
		ret = -EFAULT;
		goto out;
	}

	ret = ogfs_glock_i(ip, GL_SHARED);
	if (ret)
		goto out;

	ret =
	    ogfs_readi(ip, jt.jio_data, jt.jio_offset, jt.jio_size,
		      ogfs_copy2user);

	ogfs_gunlock_i(ip, 0);

out:
	EXIT(GFN_JREAD_IOCTL);
	return ret;
}

/**
 * ogfs_jwrite_ioctl - Write to a journaled file via ioctl
 * @ip: The OGFS inode
 * @arg: The argument from ioctl
 *
 * Returns: Amount of data copied or error
 */

int
ogfs_jwrite_ioctl(ogfs_inode_t * ip, void *arg)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	ogfs_ipreserv_t *res = &ip->i_res;
	ogfs_jio_t jt;
	unsigned int ind = 0, tmp, data_blocks, trans_reserv;
	int ret;

	ENTER(OGFS_JWRITE_IOCTL);

	if (copy_from_user(&jt, arg, sizeof(ogfs_jio_t))) {
		ret = -EFAULT;
		goto out;
	}

	/*  Compute maximum number of data blocks we need.
	   (We might need a block to unstuff the dinode)  */

	data_blocks = (jt.jio_size >> sdp->sd_sb.sb_bsize_shift) + 3;

	/*  Compute maximum number of indirect blocks we need  */

	for (tmp = data_blocks; tmp > sdp->sd_diptrs;) {
		tmp = DIV_RU(tmp, sdp->sd_inptrs) + 1;
		ind += tmp;
	}

	ind += 2 * (sdp->sd_max_height - 1);

	ret = ogfs_glock_i(ip, 0);
	if (ret)
		goto out;

	res->ir_diblks = 0;
	if (ogfs_is_jdata(ip)) {
		res->ir_metablks = ind + data_blocks;
		res->ir_datablks = 0;
	} else {
		res->ir_metablks = ind;
		res->ir_datablks = data_blocks;
	}
	res->ir_target = ip->i_di.di_goal_rgrp;
	res->ir_trans = TRANS_JWRITE;

	ret = ogfs_inplace_reserv(ip, 0);
	if (ret)
		goto out_gunlock;

	/*  Trans may require:
	   All blocks for a RG bitmap, all the "data" blocks, whatever indirect blocks we
	   need, and a modified dinode  */

	if (ogfs_is_jdata(ip))
		trans_reserv =
		    1 + data_blocks + res->ir_rgd->rd_ri.ri_length + ind;
	else
		trans_reserv = 1 + res->ir_rgd->rd_ri.ri_length + ind;

	ret = ogfs_trans_begin(sdp, TRANS_JWRITE, trans_reserv, 0);
	if (ret)
		goto out_relse;

	ret =
	    ogfs_writei(ip, jt.jio_data, jt.jio_offset, jt.jio_size,
		       ogfs_copy_from_user);

	ogfs_trans_end(sdp);

out_relse:
	ogfs_inplace_relse(ip);

out_gunlock:
	ip->i_gl->gl_vn++;
	ogfs_gunlock_i(ip, GL_SYNC);

out:
	EXIT(OGFS_JWRITE_IOCTL);

	return ret;
}

/**
 * ogfs_stat_ogfs_ioctl - Do an OGFS specific statfs
 * @sdp: the filesystem
 * @arg: the ogfs_usage_t structure
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_stat_ogfs_ioctl(ogfs_sbd_t * sdp, void *arg)
{
	ogfs_usage_t usage;
	int error = 0;

	ENTER(GFN_STAT_OGFS_IOCTL);

	error = ogfs_stat_ogfs(sdp, &usage, TRUE);
	if (error)
		goto out;

	if (copy_to_user(arg, &usage, sizeof(ogfs_usage_t)))
		error = -EFAULT;

out:
	EXIT(GFN_STAT_OGFS_IOCTL);

	return error;
}

/**
 * ogfs_stat_rgrp_ioctl - Print out statistics info of a rgrp
 * @sdp: the filesystem
 * @arg: the ogfs_usage_t structure
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_stat_rgrp_ioctl(ogfs_sbd_t * sdp, void *arg)
{
	ogfs_rgrpstat_t stat;
	int error = 0;

	if(copy_from_user(&stat, arg, sizeof(ogfs_rgrpstat_t))){
		error = -EFAULT;
		goto out;
	};
	
	memset(&(stat.usage), 0, sizeof(ogfs_usage_t));
	stat.usage.block_size = sdp->sd_sb.sb_bsize;

	error = ogfs_rindex_hold(sdp);
	if (error) {
		return error;
	}

	error = ogfs_stat_rgrp(sdp, stat.rnum, &(stat.usage));
	if (error)
		goto out;


	if (copy_to_user(arg, &stat, sizeof(ogfs_rgrpstat_t)))
		error = -EFAULT;

out:
	ogfs_rindex_release(sdp);
	return error;
}

/**
 * ogfs_reclaim_one_ioctl - ioctl function for reclaiming all free metadata 
 * 				block from one rgrp
 * @sdp: the filesystem
 * @arg: rgrp number
 * */
int ogfs_reclaim_one_ioctl(ogfs_sbd_t *sdp, void *arg)
{
	int error;
	unsigned int rnum = *(unsigned int *)arg;
	
	error = ogfs_rindex_hold(sdp);
	if (error) {
		return error;
	}

	error=ogfs_reclaim_one(sdp, rnum);

	ogfs_rindex_release(sdp);
	return error;
}

/**
 * ogfs_reclaim_all_ioctl - ioctl function for reclaiming all
 * 				free metadata block
 * @sdp: the filesystem
 * */
int ogfs_reclaim_all_ioctl(ogfs_sbd_t *sdp)
{
	int error;
	unsigned int rnum;
	
	error = ogfs_rindex_hold(sdp);
	if (error) {
		return error;
	}

	for ( rnum = 0 ; rnum < sdp->sd_rgcount; rnum++) {
		error=ogfs_reclaim_one(sdp, rnum);
		if (error)
			break;
	}		

	ogfs_rindex_release(sdp);
	return error;
}
