/*
 *
 *    Copyright 1995-1999 Regents of the University of Minnesota
 *    Portions Copyright 2000-2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 *    This file is maintained by:
 *      Kenneth W. Preslan <kpreslan@sistina.com>
 *
 */

#include "ogfs_debug.h"
#include "ogfs.h"

#define pv(struct, member, fmt) \
{ \
  if (console) \
  { \
    con_printf("  "#member" = "fmt"\n", struct->member); \
  } \
  else \
  { \
    db_printf("  "#member" = "fmt"\n", struct->member); \
  } \
}

#define CPIN_08(s1, s2, member, count) {ogfs_memcpy((s1->member), (s2->member), (count));}
#define CPOUT_08(s1, s2, member, count) {ogfs_memcpy((s2->member), (s1->member), (count));}
#define CPIN_16(s1, s2, member) {(s1->member) = ogfs16_to_cpu((s2->member));}
#define CPOUT_16(s1, s2, member) {(s2->member) = cpu_to_ogfs16((s1->member));}
#define CPIN_32(s1, s2, member) {(s1->member) = ogfs32_to_cpu((s2->member));}
#define CPOUT_32(s1, s2, member) {(s2->member) = cpu_to_ogfs32((s1->member));}
#define CPIN_64(s1, s2, member) {(s1->member) = ogfs64_to_cpu((s2->member));}
#define CPOUT_64(s1, s2, member) {(s2->member) = cpu_to_ogfs64((s1->member));}

/**
 * ogfs_rgrp_save_in - Read in rgrp data
 * @rs: the cpu-order structure
 * @buf: the disk-order buffer
 *
 */

void
ogfs_rgrp_save_in(ogfs_rgrp_save_t * rs, lm_lvb_t * lvb)
{
	ogfs_rgrp_save_t *str = (ogfs_rgrp_save_t *) lvb->lvb_data;

	ENTER(GFN_RGRP_SAVE_IN);

	OGFS_ASSERT(lvb->lvb_length >= sizeof(ogfs_rgrp_save_t),);

	CPIN_32(rs, str, rs_magic);
	CPIN_32(rs, str, rs_free);
	CPIN_32(rs, str, rs_useddi);
	CPIN_32(rs, str, rs_freedi);
	CPIN_32(rs, str, rs_usedmeta);
	CPIN_32(rs, str, rs_freemeta);
	CPIN_16(rs, str, rs_owner);

	EXIT(GFN_RGRP_SAVE_IN);
}

/**
 * ogfs_rgrp_save_out - Write out rgrp data
 * @rs: the cpu-order structure
 * @buf: the disk-order buffer
 *
 */

void
ogfs_rgrp_save_out(ogfs_rgrp_save_t * rs, lm_lvb_t * lvb)
{
	ogfs_rgrp_save_t *str = (ogfs_rgrp_save_t *) lvb->lvb_data;

	ENTER(GFN_RGRP_SAVE_OUT);

	OGFS_ASSERT(lvb->lvb_length >= sizeof(ogfs_rgrp_save_t),);

	CPOUT_32(rs, str, rs_magic);
	CPOUT_32(rs, str, rs_free);
	CPOUT_32(rs, str, rs_useddi);
	CPOUT_32(rs, str, rs_freedi);
	CPOUT_32(rs, str, rs_usedmeta);
	CPOUT_32(rs, str, rs_freemeta);
	CPOUT_16(rs, str, rs_owner);

	EXIT(GFN_RGRP_SAVE_OUT);
}

/**
 * ogfs_rgrp_save_print - Print out rgrp data
 * @rs: the cpu-order structure
 * @console - TRUE if this should be printed to the console,
 *            FALSE if it should be just printed to the incore debug
 *            buffer
 */

void
ogfs_rgrp_save_print(ogfs_rgrp_save_t * rs, int console)
{
	ENTER(GFN_RGRP_SAVE_PRINT);

	pv(rs, rs_magic, "%u");
	pv(rs, rs_free, "%u");
	pv(rs, rs_useddi, "%u");
	pv(rs, rs_freedi, "%u");
	pv(rs, rs_usedmeta, "%u");
	pv(rs, rs_freemeta, "%u");
	pv(rs, rs_owner, "%u");

	EXIT(GFN_RGRP_SAVE_PRINT);
}

/* plock funcs. */
void
ogfs_plock_elem_in(plock_element_t * pl, lm_lvb_t * lvb)
{
	plock_element_t *str = (plock_element_t *) lvb->lvb_data;
	ENTER(GFN_PLOCK_ELEM_IN);

	OGFS_ASSERT(lvb->lvb_length >= sizeof(plock_element_t),);

	CPIN_64(pl, str, begin);
	CPIN_64(pl, str, end);
	CPIN_32(pl, str, magic);
	CPIN_32(pl, str, pid);
	CPIN_16(pl, str, Next);
	pl->ltype = str->ltype;

	EXIT(GFN_PLOCK_ELEM_IN);
}

void
ogfs_plock_elem_out(plock_element_t * pl, lm_lvb_t * lvb)
{
	plock_element_t *str = (plock_element_t *) lvb->lvb_data;

	ENTER(GFN_PLOCK_ELEM_OUT);

	OGFS_ASSERT(lvb->lvb_length >= sizeof(plock_element_t),);

	CPOUT_64(pl, str, begin);
	CPOUT_64(pl, str, end);
	CPOUT_32(pl, str, magic);
	CPOUT_32(pl, str, pid);
	CPOUT_16(pl, str, Next);
	str->ltype = pl->ltype;

	EXIT(GFN_PLOCK_ELEM_OUT);
}

void
ogfs_plock_head_in(plock_head_t * pl, lm_lvb_t * lvb)
{
	plock_head_t *str = (plock_head_t *) lvb->lvb_data;

	ENTER(GFN_PLOCK_HEAD_IN);

	OGFS_ASSERT(lvb->lvb_length >= sizeof(plock_head_t),);

	CPIN_32(pl, str, magic);
	CPIN_32(pl, str, Jid_group);

	EXIT(GFN_PLOCK_HEAD_IN);
}

void
ogfs_plock_head_out(plock_head_t * pl, lm_lvb_t * lvb)
{
	plock_head_t *str = (plock_head_t *) lvb->lvb_data;

	ENTER(GFN_PLOCK_HEAD_OUT);

	OGFS_ASSERT(lvb->lvb_length >= sizeof(plock_head_t),);

	CPOUT_32(pl, str, magic);
	CPOUT_32(pl, str, Jid_group);

	EXIT(GFN_PLOCK_HEAD_OUT);
}

void
ogfs_plock_jid_head_in(plock_jid_head_t * pl, lm_lvb_t * lvb)
{
	plock_jid_head_t *str = (plock_jid_head_t *) lvb->lvb_data;

	ENTER(GFN_PLOCK_JID_HEAD_IN);

	CPIN_32(pl, str, magic);
	CPIN_32(pl, str, lock_group);
	CPIN_16(pl, str, lock_count);
	CPIN_16(pl, str, first);

	EXIT(GFN_PLOCK_JID_HEAD_IN);
}

void
ogfs_plock_jid_head_out(plock_jid_head_t * pl, lm_lvb_t * lvb)
{
	plock_jid_head_t *str = (plock_jid_head_t *) lvb->lvb_data;

	ENTER(GFN_PLOCK_JID_HEAD_OUT);

	CPOUT_32(pl, str, magic);
	CPOUT_32(pl, str, lock_group);
	CPOUT_16(pl, str, lock_count);
	CPOUT_16(pl, str, first);

	EXIT(GFN_PLOCK_JID_HEAD_OUT);
}
