/*
 *
 *    Copyright 2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 */

#include "ogfs_debug.h"
#include "ogfs.h"
#include "dio.h"

extern spinlock_t ogfs_bufferlist_lock;

/**
 * ogfs_glock_arch_init - Initialize the arch dependent part of an ogfs_glock_t
 * @sdp: the filesystem
 * @gl: the glock
 *
 */

void
ogfs_glock_arch_init(ogfs_sbd_t * sdp, ogfs_glock_t * gl)
{
	ENTER(GFN_GLOCK_ARCH_INIT);

	INIT_LIST_HEAD(&gl->gl_arch.gl_bufs);

	EXIT(GFN_GLOCK_ARCH_INIT);
}

/**
 * ogfs_trans_prep_for_ail - called before a trans is put into the AIL
 * @sdp: the filesystem
 * @tr: the transaction
 *
 */

void
ogfs_trans_prep_for_ail(ogfs_sbd_t * sdp, ogfs_trans_t * tr)
{
	ENTER(GFN_TRANS_PREP_FOR_AIL);

	INIT_LIST_HEAD(&tr->tr_arch.tr_bufs);

	EXIT(GFN_TRANS_PREP_FOR_AIL);
}

/**
 * ogfs_trans_check_empty - Check whether or not a trans in the AIL has been synced
 * @sdp: the filesystem
 * @tr: the transaction
 *
 */

int
ogfs_trans_check_empty(ogfs_sbd_t * sdp, ogfs_trans_t * tr)
{
	ogfs_bufdata_t *bd;
	struct buffer_head *bh;

	ENTER(GFN_TRANS_CHECK_EMPTY);

	spin_lock(&ogfs_bufferlist_lock);

	while (!list_empty(&tr->tr_arch.tr_bufs)) {
		bd = list_entry(tr->tr_arch.tr_bufs.next, ogfs_bufdata_t,
				bd_arch.bd_ail_list);
		bh = bd->bd_bh;

		atomic_inc(&bh->b_count);
		spin_unlock(&ogfs_bufferlist_lock);

		ogfs_lock_buffer(bh);

		if (!bd->bd_arch.bd_pinned && !buffer_dirty(bh)
		    && !buffer_locked(bh)) {
			spin_lock(&ogfs_bufferlist_lock);
			if (bd->bd_arch.bd_ail_tr) {
				bd->bd_arch.bd_ail_tr = NULL;
				list_del(&bd->bd_arch.bd_ail_list);
			}

			ogfs_unlock_buffer(bh);
			atomic_dec(&bh->b_count);
			continue;
		}

		ogfs_unlock_buffer(bh);

		atomic_dec(&bh->b_count);

		EXIT(GFN_TRANS_CHECK_EMPTY);
		return FALSE;
	}

	spin_unlock(&ogfs_bufferlist_lock);

	EXIT(GFN_TRANS_CHECK_EMPTY);

	return TRUE;
}

/*
 * ogfs_mark_vsb - set or clear superblock dirty flag
 * @sdp: the filesystem
 * @dirty: it TRUE set sb dirty, else mark it clean
 */

void
ogfs_mark_vsb(ogfs_sbd_t * sdp, int dirty)
{
	sdp->sd_vsb->s_dirt = dirty;
}
