/*
 *
 *    Copyright 1999-2001 Sistina Software, Inc.
 *    Portions Copyright 2001 The OpenGFS Project
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 */

#include "ogfs_debug.h"
#include "ogfs.h"

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>

#include "mount.h"


/**
 * ogfs_make_args - Parse mount arguments
 * @data:
 * @args:
 *
 * Return: 1 if data was valid, 0 if data is not valid
 */

int
ogfs_make_args(char *data, ogfs_args_t * args)
{
	char *this_token, *value;
	int ret = 1;
	unsigned int count = 0;

	ENTER(GFN_MAKE_ARGS);
	/*  If someone preloaded options, use those instead  */
	down(&ogfs_mount_args_lock);
	if (ogfs_mount_args) {
		data = ogfs_mount_args;
		count = ogfs_mount_args_size;

		db_printf("Overriding mount options with:  %s\n", data);

		ogfs_mount_args = NULL;
	}
	up(&ogfs_mount_args_lock);
	memset(args, 0, sizeof(ogfs_args_t));

	/*  Set some defaults  */
	args->ar_dir_alloc_method = OGFS_DAMETH_DEFAULT;
	args->ar_buf_readahead = 8;
	args->ar_atime_quantum = OGFS_ATIME_DEFAULT;

	/*  Split the options into tokens with the "," character and process
	   them  */
	if (data) {
		for (this_token = strsep(&data, ","); this_token;
		     this_token = strsep(&data, ",")) {
			/* This splits the token in half if there is an "=" in
			 * it */
			value = strchr(this_token, '=');
			if (value)
				*value++ = 0;
			/* This can be extended to only allow valid lock
			 * protocols */
			if (!strcmp(this_token, "lockproto") && value)
			{
				strncpy(
					args->ar_lockproto, value,
					sizeof(args->ar_lockproto));
				args->ar_lockproto[
					sizeof(args->ar_lockproto) - 1] = 0;
			}
			/* This can be extended to only allow valid lock table
			 * names */
			else if (!strcmp(this_token, "locktable") && value)
			{
				strncpy(
					args->ar_locktable, value,
					sizeof(args->ar_locktable));
				args->ar_locktable[
					sizeof(args->ar_locktable) - 1] = 0;
			}
			/* This can be extended to only allow valid host data
			 * fields */
			else if (!strcmp(this_token, "hostdata") && value)
			{
				strncpy(
					args->ar_hostdata, value,
					sizeof(args->ar_hostdata));
				args->ar_hostdata[
					sizeof(args->ar_hostdata) - 1] = 0;
			}
			else if (!strcmp(this_token, "diralloc") && value) {
				if (!strcmp(value, "single"))
					args->ar_dir_alloc_method =
					    OGFS_DAMETH_SINGLE;
				else if (!strcmp(value, "roundrobin"))
					args->ar_dir_alloc_method =
					    OGFS_DAMETH_RROBIN;
				else if (!strcmp(value, "random"))
					args->ar_dir_alloc_method =
					    OGFS_DAMETH_RANDOM;
				else
					ret = 0;
			}
			else if (!strcmp(this_token, "ignore_local_fs"))
				args->ar_ignore_local_fs = TRUE;
			else if (!strcmp(this_token, "localflocks"))
				args->ar_localflocks = TRUE;
			else if (!strcmp(this_token, "localcaching"))
				args->ar_localcaching = TRUE;
			else if (!strcmp(this_token, "buf_readahead")
				 && value)
				args->ar_buf_readahead =
					simple_strtoul(value, NULL, 10);
			else if (!strcmp(this_token, "atime_quantum")
				 && value)
				args->ar_atime_quantum =
					simple_strtoul(value, NULL, 10);
			else if (!strcmp(this_token, "upgrade"))
				args->ar_upgrade = TRUE;
			else
				ret = 0;

			if (!ret)
				break;
		}
	}
#ifdef CONFIG_SSI
	/* SSI: Always over-ride hostdata with ics_netaddr, since we
	 * we are maintaining cluster-wide mount consistency.  All nodes
	 * are mounting without the hostdata parameter being passed.
	 */
	{
		extern char ics_netaddr[];

		{
			strncpy(
				args->ar_hostdata, ics_netaddr,
				sizeof(args->ar_hostdata));
			args->ar_hostdata[
				sizeof(args->ar_hostdata) - 1] = 0;
		}
	}
#endif /* CONFIG_SSI */

	if (!ret)
		con_printf("OGFS:  invalid mount option(s)\n");
	else {
		db_printf("ar_lockproto = \"%s\"\n", args->ar_lockproto);
		db_printf("ar_locktable = \"%s\"\n", args->ar_locktable);
		db_printf("ar_hostdata = \"%s\"\n", args->ar_hostdata);
		db_printf("ar_dir_alloc_method = %u\n",
			  args->ar_dir_alloc_method);
		db_printf("ar_ignore_local_fs = %s\n",
			  (args->ar_ignore_local_fs) ? "TRUE" : "FALSE");
		db_printf("ar_localflocks = %s\n",
			  (args->ar_localflocks) ? "TRUE" : "FALSE");
		db_printf("ar_localcaching = %s\n",
			  (args->ar_localcaching) ? "TRUE" : "FALSE");
		db_printf("ar_buf_readahead = %u\n", args->ar_buf_readahead);
		db_printf("ar_atime_quantum = %u\n", args->ar_atime_quantum);
		db_printf("ar_upgrade = %u\n", args->ar_upgrade);
	}

	/*  Free the option space  */
	if (count)
		gfree(data, count);
	EXIT(GFN_MAKE_ARGS);

	return ret;
}

/**
 * ogfs_proc_write - Read in some mount options
 * @file: unused
 * @buffer: a buffer of mount options
 * @count: the length of the mount options
 * @data: unused
 *
 * Called when someone writes to /proc/fs/ogfs.
 * It allows you to specify mount options when you can't do it
 * from mount.  i.e. from a inital ramdisk
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_proc_write(struct file *file, const char *buffer,
	       unsigned long count, void *data)
{
	int error;
	char *p;

	ENTER(GFN_PROC_WRITE);
	//MOD_INC_USE_COUNT;
	down(&ogfs_mount_args_lock);
	if (ogfs_mount_args) {
		gfree(ogfs_mount_args, ogfs_mount_args_size);
		ogfs_mount_args = NULL;
	}
	if (!count) {
		error = 0;
		goto fail;
	}
	ogfs_mount_args = (char *) gmalloc(count + 1);
	if (!ogfs_mount_args) {
		error = -ENOMEM;
		goto fail;
	}
	ogfs_mount_args_size = count + 1;
	error = copy_from_user(ogfs_mount_args, buffer, count);
	if (error) {
		error = -EFAULT;
		goto fail_gfree;
	}
	ogfs_mount_args[count] = 0;
	/*  Get rid of extra newlines  */
	for (p = ogfs_mount_args; *p; p++)
		if (*p == '\n')
			*p = 0;
	up(&ogfs_mount_args_lock);
	//MOD_DEC_USE_COUNT;

	return count;

fail_gfree:
	gfree(ogfs_mount_args, ogfs_mount_args_size);
	ogfs_mount_args = NULL;

fail:
	up(&ogfs_mount_args_lock);
	//MOD_DEC_USE_COUNT;
	EXIT(GFN_PROC_WRITE);
	return error;
}
