/*
 *
 *    Copyright 2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 */

/*
 *
 * This be those wacky posix locks.  Also known as fcntl record locks.
 *
 * list of list plocking.
 * Plocks are stored in lvbs.  Current OGFS and lock recovery wipes all locks
 * that are only referenced by a dead node.  Keep this in mind.
 * So we cannot intermix a list from multinodes, a well placed crash will
 * break the list, and confuse others, and make recovery very difficult.
 *
 * So, each node (marked by its JID) gets its own list.  By being smart
 * about the numbers we use for the lock names, we can calculate the ids of
 * all these lvbs.  So if a node dies, all of its plocks die, and since
 * there's no list pointer like stuff, nothing is lost.  When another node
 * checks for plocks, it will just see that that node has none.
 *
 * This does of course make some things slower, and require that we apply
 * some careful checking in places.  But the advantages are well worth the
 * price.
 *
 * So what's this about being smart about the lock names?  well, there are
 * now three lock types, One for the plock head, one for plock jid heads,
 * and one for plock list elements.
 *
 * plock head's number is just the inode number of who's getting the locks.
 *
 * plock jid head's number is the high 16 bits are the jid, 32 bits for the
 * plock group, then 16 bits of zero.  (The plock group number has to be
 * intelligently picked, and is the first limit.)
 *
 * plock list elements' number: top 16 bits are the jid, next 32 are the
 * plock group, last 16 are the lock id. (This is the second limit.)
 *
 * With some fancy work, things could probably be compressed and or
 * stretched a bit.  future work if needed.
 *
 * Where do plock groups come from?  Well, we start with the lower 32 bits of
 * the inode.  Since this creates overlap, we need to make sure that noone
 * else is using this group.  If it is, increment the group id, and try
 * again.
 *
 * We keep track of which groups are in use by grabbing another lock.  The
 * first node grabs the lock, and others just get a hold. This way, when the
 * last node drops it, it is automagically freed.
 *
 * The name of this lock is high 16 bits turned on, next 32 bits as the
 * group ID, low 16 bits off.  This is also mapped into the plockelem lock
 * type. (This is now the third limit.)
 *
 * This induces some limits.
 * 1:  only 2**32 files can have plocks
 * 2:  each node can only put (2**16)-1 plocks on a single file.
 * 3:  There cannot be a journal that has the id of 0xffff
 *
 * learn to live with that.
 *
 * If a node has plocks, it holds the plockhead, and its list.
 * If a node has no plocks, it holds nothing.
 *
 * To do a plock action, a node must grab the plhead and all lists exl.
 * When it is done, it drops all other node's lists completely.
 * If it still has items, it move the plhead and its items to held and
 * unlocked.
 * If it has no items, it drops everything.
 *
 * TODO:
 *	Review error handling.  There are some places where the might be
 *	leaks or some other issues.
 */

#include "ogfs_debug.h"
#include "ogfs.h"
#include "fs.h"
#include "glock.h"
#include "glops.h"
#include "plock.h"
#include "lvb.h"
#include "super.h"

#ifndef MAX
#define MAX(a,b) ((a>b)?a:b)
#endif
#ifndef MIN
#define MIN(a,b) ((a<b)?a:b)
#endif

/* lvbs: Now in lvb.h */

/* incore: */
typedef struct plock_ic_s {
	struct list_head pl_list;
	ogfs_glock_t *gl;
	plock_element_t hdata;
	u16 lid;
} ogfs_pl_t;

typedef struct plock_jid_ic_s {
	struct list_head plj_list;

	ogfs_glock_t *gl;
	plock_jid_head_t hdata;

	u16 jid;
	struct list_head freeids;

	struct list_head locks;
} pljid_t;

typedef struct plock_head_ic_s {
	ogfs_glock_t *gl;
	plock_head_t hdata;

	ogfs_glock_t *pgrpl;	/* plock group marker. */

	/* maybe change this to a struct pointer instead of a list?? */
	struct list_head mlcks;	/* only has one item, a pljid_t with my plocks */

	struct list_head olcks;	/* a list of pljid_t with every other plock. */
} plhd_t;

/****************************************************************************/
/* TODO Look into ways of doing this without the need for memory
 * allocations. (if there is such a thing)
 */
typedef struct frl_s {
	struct list_head frl_list;
	uint16 first;
	uint16 last;
} frl_t;

/**
 * init_free_list - set up the default free list
 * @list: <> an itemless just initialized list_head
 *
 * Puts the first item onto the list which claims entire range is free to
 * use.  Later calls to mark_used and next_free will change the shape of
 * this range. (possible making multiple)
 *
 * Returns: errors
 */
int
init_free_list(struct list_head * list)
{
	frl_t *frl = NULL;
	ENTER(INIT_FREE_LIST);

	frl = (frl_t *) gmalloc(sizeof(frl_t));
	if (!frl) {
		EXIT(INIT_FREE_LIST);
		return -ENOMEM;
	}
	memset(frl, 0, sizeof(frl_t));
	INIT_LIST_HEAD(&frl->frl_list);
	frl->first = 1;		/* zero is always used */
	frl->last = GND16;	/* last possible id. */

	list_add(&frl->frl_list, list);

	EXIT(INIT_FREE_LIST);
	return 0;
}

/**
 * destroy_free_list - free all mem in list
 * @list: <> list of free ranges to destroy
 */
void
destroy_free_list(struct list_head * list)
{
	struct list_head *tmp;
	frl_t *frl = NULL;
	ENTER(DESTROY_FREE_LIST);
	tmp = list;
	for (;;) {
		tmp = tmp->next;
		if (tmp == list)
			break;
		frl = list_entry(tmp, frl_t, frl_list);
		list_del(tmp);
		gfree(frl, sizeof(frl));
	}
	EXIT(DESTROY_FREE_LIST);
}

/**
 * next_free_list - return the next available lockid
 * @list: <> list of possibles
 *
 * Just pick up the next available, and take it out of the list.
 *
 * Returns: new lockid or 0xffff (error)
 */
u16
next_free_list(struct list_head * list)
{
	struct list_head *tmp;
	frl_t *frl;
	u16 ret = GND16;

	ENTER(NEXT_FREE_LIST);
	tmp = list;
	tmp = tmp->next;
	if (tmp == list)
		goto out;	/* shouldn't really ever happen... */
	frl = list_entry(tmp, frl_t, frl_list);

	if (frl->first == GND16)
		goto out;

	ret = frl->first;
	if (frl->first == frl->last) {
		/* emptied this extent, remove it. */
		list_del(tmp);
		gfree(frl, sizeof(frl_t));
	} else
		frl->first++;	/* move first up */

 out:
	IFDB(NEXT_FREE_LIST, db_printf("Next free id in list is %u\n", ret);
	    );
	EXIT(NEXT_FREE_LIST);
	return ret;
}

/**
 * return_to_frl - return a number to the free lists
 * @list: <> list of possibles
 * @num: < value to return
 *
 * Puts a number `back' into the list of freely available ones.
 *
 * free range list had better be sorted.....
 * Returns: int
 */
int
return_to_frl(struct list_head * list, u16 num)
{
	struct list_head *tmp;
	frl_t *frl, *new = NULL;

	ENTER(RETURN_TO_FRL);
	IFDB(RETURN_TO_FRL, db_printf("freeing %u\n", num);
	    );

	tmp = list;
	for (;;) {
		tmp = tmp->next;
		if (tmp == list)
			break;
		frl = list_entry(tmp, frl_t, frl_list);

		if (num < frl->first) {
			new = (frl_t *) gmalloc(sizeof(frl_t));
			if (!new) {
			}
			INIT_LIST_HEAD(&new->frl_list);
			new->first = num;
			new->last = num;
			list_add_tail(&new->frl_list, &frl->frl_list);
		} else if (num + 1 == frl->first) {
			frl->first--;	/* expand to include num */
			break;
		} else if (num - 1 == frl->last) {
			frl->last++;	/* expand to include num */
			break;
		}
		/* Need some kind of add to end????? */
	}
	EXIT(RETURN_TO_FRL);
	return 0;
}

/**
 * print_free_list - print out list of open ranges.
 * @list: < list of possibles
 */
void
print_free_list(struct list_head * list)
{
	struct list_head *tmp;
	frl_t *frl;
	ENTER(PRINT_FREE_LIST);
	tmp = list;
	for (;;) {
		tmp = tmp->next;
		if (tmp == list)
			break;
		frl = list_entry(tmp, frl_t, frl_list);
		db_printf("  Unused range %u to %u\n", frl->first, frl->last);
	}
	EXIT(PRINT_FREE_LIST);
}

/****************************************************************************/
/**
 * pack_plem_bid - take the parts of a plock element id, and make a bid
 * @jid: < journal id
 * @plgrp: < plock group id
 * @lid: < lock id
 *
 * Returns: u64 bid
 */
u64
pack_plem_bid(u16 jid, u32 plgrp, u16 lid)
{
	u64 bid = 0;
	ENTER(PACK_PLEM_BID);
	bid = jid & 0xffff;
	bid <<= 32;
	bid |= plgrp & 0xffffffff;
	bid <<= 16;
	bid |= lid;
	EXIT(PACK_PLEM_BID);
	return bid;
}

/****************************************************************************/
/**
 * find_unused_plockgroup - find a vaule to use for the plock group
 * @sdp: < super block pointer
 * @plhd: plock header struct.
 * @guess: <> an intial guess on in, the found value on out
 *
 * Not really sure of the best way to do this.
 * I think I'll do a simple test..incr loop for starts.
 *
 * TODO I haven't actually verified that this works in the conflict
 * case, only in the non-conflict case.
 *
 * Returns: <>0: Error, =0: Ok
 */
int
find_unused_plockgroup(ogfs_sbd_t * sdp, plhd_t * plhd, u32 * guess)
{
	ogfs_glock_t *gl;
	u64 bid = 0;
	int err = 0;

	ENTER(FIND_UNUSED_PLOCKGROUP);

	plhd->pgrpl = NULL;
	for (;;) {
		bid = pack_plem_bid(GND16, (*guess), (u16) 0);

		IFDB(FIND_UNUSED_PLOCKGROUP,
		     db_printf("Trying to get groupid:%u (bid:0x%" FMT_64
			       "x)\n", *guess, bid);
		    );

		gl = ogfs_get_glstruct(sdp, bid, &ogfs_plock_glops, NULL, CREATE);
		if (!gl) {
			err = -1;
			goto exit;
		}

		err = ogfs_hold_lvb(sdp, gl);
		if (err) {
			ogfs_put_glstruct(gl);
			goto exit;
		}

		err = ogfs_glock(sdp, gl, GL_PERM);
		if (err) {
			ogfs_unhold_lvb(sdp, gl);
			ogfs_put_glstruct(gl);
			goto exit;
		}

		if (*((u8 *) (gl->gl_lvb->lvb_data)) == 0) {
			/*YEAH! it is ours.  just hang onto the lvb, this will keep the lock
			 * on the server.  We'll have everone else pick up a hold on this
			 * too, to keep it around, then when the last drops, it goes buhbye.
			 */
			*((u8 *) (gl->gl_lvb->lvb_data)) = 0xff;
			ogfs_sync_lvb(sdp, gl);
			ogfs_gunlock(sdp, gl, GL_NOCACHE);
			plhd->pgrpl = gl;
			break;
		}

		/* damn, someone else has it, incr, try again. */
		ogfs_gunlock(sdp, gl, GL_NOCACHE);
		ogfs_unhold_lvb(sdp, gl);
		ogfs_put_glstruct(gl);

		(*guess)++;
	}

exit:
	EXIT(FIND_UNUSED_PLOCKGROUP);
	return err;
}

/**
 * hold_plockgroup - puta a hold onto the plockgroup
 * @sdp: < super block pointer
 * @plhd: <> plock header
 * @pgrp: < plock group number
 *
 * Just another hold onto the plockgroup marker, so that it sticks around.
 * If the pgrpl glock is there, it won't grab it again.
 *
 * Returns: =0: Ok, <>0: Error
 */
int
hold_plockgroup(ogfs_sbd_t * sdp, plhd_t * plhd, u32 pgrp)
{
	u64 bid = 0;
	int err = 0;

	ENTER(HOLD_PLOCKGROUP);

	if (plhd->pgrpl == NULL) {
		/* if we've got it held, don't need to hold it again. */

		bid = pack_plem_bid(~((u16) 0), pgrp, (u16) 0);

		plhd->pgrpl =
		    ogfs_get_glstruct(sdp, bid, &ogfs_plock_glops, NULL, CREATE);
		if (!plhd->pgrpl) {
			err = -1;
			goto exit;
		}

		err = ogfs_hold_lvb(sdp, plhd->pgrpl);
		if (err) {
			ogfs_put_glstruct(plhd->pgrpl);
			goto exit;
		}

	}

	/* should i do some kind of verification here??? */
	/* k, we've made a hold on this, so if the original holder drops, its not
	   * lost.
	 */
exit:
	EXIT(HOLD_PLOCKGROUP);
	return err;
}

/**
 * drop_plockgroup - drop the plockgroup marker.
 * @sdp: < super block pointer
 * @plhd: < plock header
 *
 * We have no reason to keep this plock group reserved, so we'll let the
 * marker go.  When the last node lets this go, this plock group will again
 * be available.
 *
 */
void
drop_plockgroup(ogfs_sbd_t * sdp, plhd_t * plhd)
{
	ENTER(DROP_PLOCKGROUP);
	OGFS_ASSERT(plhd->pgrpl != NULL,);
	ogfs_unhold_lvb(sdp, plhd->pgrpl);
	ogfs_put_glstruct(plhd->pgrpl);
	plhd->pgrpl = NULL;
	EXIT(DROP_PLOCKGROUP);
}

/****************************************************************************/

/**
 * load_plock_head - get the plock head
 * @ip: <> inode pointer
 *
 * load the plock head struct on an inode.  Fills ing the info if it is the
 * first time called.
 *
 * Returns: plhd_t* or NULL
 */
plhd_t *
load_plock_head(ogfs_inode_t * ip)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	plhd_t *plhd = NULL;
	int err = 0;

	ENTER(GFN_LOAD_PLOCK_HEAD);

	down(&ip->i_plock_mux);

	if (ip->i_plock_gl == NULL) {

		atomic_inc(&ip->i_count);	/* grab a reference, jic */
		/* do this here so if it fails its easier to clean up */
		plhd = (plhd_t *) gmalloc(sizeof(plhd_t));
		if (!plhd) {
			atomic_dec(&ip->i_count);
			up(&ip->i_plock_mux);
			EXIT(GFN_LOAD_PLOCK_HEAD);
			return NULL;
		}

		ip->i_plock_gl = ogfs_get_glstruct(sdp, ip->i_num.no_formal_ino,
						  &ogfs_plock_head_glops, NULL,
						  CREATE);
		if (!ip->i_plock_gl) {
			gfree(plhd, sizeof(plhd_t));
			atomic_dec(&ip->i_count);
			up(&ip->i_plock_mux);
			EXIT(GFN_LOAD_PLOCK_HEAD);
			return NULL;
		}
		err = ogfs_hold_lvb(sdp, ip->i_plock_gl);
		if (err) {
			ogfs_put_glstruct(ip->i_plock_gl);
			gfree(plhd, sizeof(plhd_t));
			atomic_dec(&ip->i_count);
			up(&ip->i_plock_mux);
			EXIT(GFN_LOAD_PLOCK_HEAD);
			return NULL;
		}

		memset(plhd, 0, sizeof(plhd_t));
		INIT_LIST_HEAD(&plhd->mlcks);
		INIT_LIST_HEAD(&plhd->olcks);

		ip->i_plock_gl->gl_object = plhd;
		plhd->gl = ip->i_plock_gl;

	} else {
		plhd = (plhd_t *) ip->i_plock_gl->gl_object;
		OGFS_ASSERT(plhd != NULL,);
		OGFS_ASSERT(plhd->gl == ip->i_plock_gl,);
	}

	atomic_inc(&ip->i_plock_cnt);
	up(&ip->i_plock_mux);

	/* pids had better barrer here. */
	err = ogfs_glock(sdp, ip->i_plock_gl, GL_PERM);
	if (err) {
		/* probably no correct */
		ogfs_unhold_lvb(sdp, ip->i_plock_gl);
		ogfs_put_glstruct(ip->i_plock_gl);
		EXIT(GFN_LOAD_PLOCK_HEAD);
		return NULL;
	}

	OGFS_ASSERT(ip->i_plock_gl->gl_lvb != NULL,);	/* right here bebe. */
	OGFS_ASSERT(ip->i_plock_gl->gl_lvb->lvb_data != NULL,);

	/* ok, got here, got the head. */
	ogfs_plock_head_in(&plhd->hdata, ip->i_plock_gl->gl_lvb);

	if (plhd->hdata.magic == 0) {
		/* its new, init it. */
		plhd->hdata.magic = PLOCK_HEAD_MAGIC;

		plhd->hdata.Jid_group = ip->i_num.no_formal_ino & 0xffffffff;
		err = find_unused_plockgroup(sdp, plhd, &plhd->hdata.Jid_group);
		/* if error, undo everything, return null */
		if (err) {
			ogfs_gunlock(sdp, ip->i_plock_gl, GL_NOCACHE);
			ogfs_unhold_lvb(sdp, ip->i_plock_gl);
			ogfs_put_glstruct(ip->i_plock_gl);
			EXIT(GFN_LOAD_PLOCK_HEAD);
			return NULL;
		}

		ogfs_plock_head_out(&plhd->hdata, ip->i_plock_gl->gl_lvb);
		ogfs_sync_lvb(sdp, ip->i_plock_gl);
	} else {
		err = hold_plockgroup(sdp, plhd, plhd->hdata.Jid_group);
		/* if error, undo everything, return null */
		if (err) {
			ogfs_gunlock(sdp, ip->i_plock_gl, GL_NOCACHE);
			ogfs_unhold_lvb(sdp, ip->i_plock_gl);
			ogfs_put_glstruct(ip->i_plock_gl);
			EXIT(GFN_LOAD_PLOCK_HEAD);
			return NULL;
		}
	}

	EXIT(GFN_LOAD_PLOCK_HEAD);
	return plhd;
}

/**
 * sync_plock_head - syncs the plock head to lvb store
 * @ip: < inode pointer
 * @plhd: < plock header thingy
 *
 * this will dealloc the plhd if there are no local plocks.
 *
 */
void
sync_plock_head(ogfs_inode_t * ip, plhd_t * plhd)
{
	ogfs_sbd_t *sdp = ip->i_sbd;

	ENTER(GFN_SYNC_PLOCK_HEAD);
	/* at this point, all plock jidheads and elems should have been
	 * dropped/synced!!
	 */
	OGFS_ASSERT(plhd->gl == ip->i_plock_gl,);

	ogfs_plock_head_out(&plhd->hdata, plhd->gl->gl_lvb);

	OGFS_ASSERT(list_empty(&plhd->olcks),);

	if (list_empty(&plhd->mlcks)) {
		IFDB(GFN_SYNC_PLOCK_HEAD, db_printf("fully dropping plhd.\n");
		    );
		/* nothing here for me, do a full drop. */

		ogfs_gunlock(sdp, ip->i_plock_gl, GL_NOCACHE);

		down(&ip->i_plock_mux);
		if (atomic_dec_and_test(&ip->i_plock_cnt)) {
			/* if nooneelse just attached themselves. */
			drop_plockgroup(sdp, plhd);
			ogfs_unhold_lvb(sdp, ip->i_plock_gl);
			ip->i_plock_gl->gl_object = NULL;
			gfree(plhd, sizeof(plhd_t));

			ogfs_put_glstruct(ip->i_plock_gl);
			ip->i_plock_gl = NULL;

			atomic_dec(&ip->i_count);
			OGFS_ASSERT(atomic_read(&ip->i_plock_cnt) == 0,);
		}
		up(&ip->i_plock_mux);

	} else {
		IFDB(GFN_SYNC_PLOCK_HEAD, db_printf("just syncing plhd.\n");
		    );
		/* still have stuff; unlock, but maintain holds. */
		ogfs_sync_lvb(sdp, plhd->gl);
		ogfs_gunlock(sdp, plhd->gl, GL_NOCACHE);
		atomic_dec(&ip->i_plock_cnt);
	}

	EXIT(GFN_SYNC_PLOCK_HEAD);
}

/**
 * load_plock_jid_head -  load jid list head from lvb store
 * @ip: < inode pointer
 * @plhd: < plock header thingy
 * @which: < which jid head to load
 *
 *
 * Returns: pljid_t
 */
pljid_t *
load_plock_jid_head(ogfs_inode_t * ip, plhd_t * plhd, u16 which)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	pljid_t *pljid = NULL;
	u64 bid = 0;
	int err = 0;

	ENTER(LOAD_PLOCK_JID_HEAD);
	IFDB(LOAD_PLOCK_JID_HEAD,
	     db_printf("Loading jid head for jid %u\n", which);
	    );

	if (which == GND16) {
		EXIT(LOAD_PLOCK_JID_HEAD);
		return NULL;
	}

	bid = pack_plem_bid(which, plhd->hdata.Jid_group, (u16) 0);

	pljid = (pljid_t *) gmalloc(sizeof(pljid_t));
	if (!pljid) {
		EXIT(LOAD_PLOCK_JID_HEAD);
		return NULL;
	}
	memset(pljid, 0, sizeof(pljid_t));
	INIT_LIST_HEAD(&pljid->plj_list);
	INIT_LIST_HEAD(&pljid->locks);
	INIT_LIST_HEAD(&pljid->freeids);
	pljid->jid = which;

	pljid->gl = ogfs_get_glstruct(sdp, bid, &ogfs_plock_glops, NULL, CREATE);
	if (!pljid->gl) {
		gfree(pljid, sizeof(pljid_t));
		EXIT(LOAD_PLOCK_JID_HEAD);
		return NULL;
	}
	err = ogfs_hold_lvb(sdp, pljid->gl);
	if (err) {
		ogfs_put_glstruct(pljid->gl);
		gfree(pljid, sizeof(pljid_t));
		EXIT(LOAD_PLOCK_JID_HEAD);
		return NULL;
	}
	err = ogfs_glock(sdp, pljid->gl, GL_PERM);
	if (err) {
		ogfs_unhold_lvb(sdp, pljid->gl);
		ogfs_put_glstruct(pljid->gl);
		gfree(pljid, sizeof(pljid_t));
		EXIT(LOAD_PLOCK_JID_HEAD);
		return NULL;
	}

	pljid->gl->gl_object = pljid;

	ogfs_plock_jid_head_in(&pljid->hdata, pljid->gl->gl_lvb);

	if (pljid->hdata.magic == 0) {
		/* first time, init. */
		pljid->hdata.magic = PLOCK_JID_HEAD_MAGIC;
		pljid->hdata.lock_group = plhd->hdata.Jid_group;
		pljid->hdata.lock_count = 0;
		pljid->hdata.first = GND16;

		ogfs_plock_jid_head_out(&pljid->hdata, pljid->gl->gl_lvb);
	}

	EXIT(LOAD_PLOCK_JID_HEAD);
	return pljid;
}

/**
 * sync_plock_jid_head - sync the jid list head to the lvb store.
 * @ip: < inode pointer
 * @pljid: < jid list head pointer
 *
 */
void
sync_plock_jid_head(ogfs_inode_t * ip, pljid_t * pljid)
{
	ogfs_sbd_t *sdp = ip->i_sbd;

	ENTER(SYNC_PLOCK_JID_HEAD);

	ogfs_plock_jid_head_out(&pljid->hdata, pljid->gl->gl_lvb);

	ogfs_sync_lvb(sdp, pljid->gl);
	ogfs_gunlock(sdp, pljid->gl, GL_NOCACHE);

	EXIT(SYNC_PLOCK_JID_HEAD);
}

/**
 * drop_my_plock_jid_head - drop all refs to a jid head pointer
 * @ip: < inode pointer
 * @pljid: <> jid head pointer. (this get dealloced)
 *
 */
void
drop_plock_jid_head(ogfs_inode_t * ip, pljid_t * pljid)
{
	ogfs_sbd_t *sdp = ip->i_sbd;

	ENTER(DROP_PLOCK_JID_HEAD);

	ogfs_plock_jid_head_out(&pljid->hdata, pljid->gl->gl_lvb);

	OGFS_ASSERT(list_empty(&pljid->locks),);

	destroy_free_list(&pljid->freeids);	/* jic */
	ogfs_unhold_lvb(sdp, pljid->gl);
	ogfs_gunlock(sdp, pljid->gl, GL_NOCACHE);
	pljid->gl->gl_object = NULL;
	ogfs_put_glstruct(pljid->gl);

	list_del(&pljid->plj_list);
	gfree(pljid, sizeof(pljid_t));

	EXIT(DROP_PLOCK_JID_HEAD);
}

/**
 * load_other_plock_elems - get a hold on someone else's plock elements
 * @sdp: < superblock pointer
 * @pljid: <> jid list head.
 *
 * This load element routine is optimised for the loading of the other
 * plocks.  This means a few assumptions have been made.
 * 1: None of the items being loaded exist in our memory.  This is the
 *    first laod
 * 2: No modifications will be made on these.
 *
 * Returns: int
 */
int
load_other_plock_elems(ogfs_sbd_t * sdp, pljid_t * pljid)
{
	ogfs_pl_t *pl = NULL;
	struct list_head *pltmp, *pltop;
	int i, err = 0;
	u64 bid;
	u16 Next;

	ENTER(LOAD_OTHER_PLOCK_ELEMS);

	Next = pljid->hdata.first;

	for (i = 0; i < pljid->hdata.lock_count && Next != GND16; i++) {
		bid = pack_plem_bid(pljid->jid, pljid->hdata.lock_group, Next);

		pl = (ogfs_pl_t *) gmalloc(sizeof(ogfs_pl_t));
		if (!pl) {
			err = -ENOMEM;
			goto mega_error;
		}
		memset(pl, 0, sizeof(ogfs_pl_t));
		INIT_LIST_HEAD(&pl->pl_list);
		pl->lid = Next;

		pl->gl =
		    ogfs_get_glstruct(sdp, bid, &ogfs_plock_glops, NULL, CREATE);
		if (!pl->gl) {
			err = -ENOMEM;
			gfree(pl, sizeof(ogfs_pl_t));
			goto mega_error;
		}
		err = ogfs_hold_lvb(sdp, pl->gl);
		if (err) {
			ogfs_put_glstruct(pl->gl);
			gfree(pl, sizeof(ogfs_pl_t));
			goto mega_error;
		}
		err = ogfs_glock(sdp, pl->gl, GL_PERM);
		if (err) {
			ogfs_unhold_lvb(sdp, pl->gl);
			ogfs_put_glstruct(pl->gl);
			gfree(pl, sizeof(ogfs_pl_t));
			goto mega_error;
		}

		pl->gl->gl_object = pl;

		ogfs_plock_elem_in(&pl->hdata, pl->gl->gl_lvb);

		list_add(&pl->pl_list, &pljid->locks);

		Next = pl->hdata.Next;
	}

	EXIT(LOAD_OTHER_PLOCK_ELEMS);
	return err;
mega_error:
	/* gah, gotta drop everything we tried to load, moving old list back if
	 * there is one.
	 */
	pltmp = pltop = &pljid->locks;
	for (;;) {
		pltmp = pltmp->next;
		if (pltmp == pltop)
			break;
		pl = list_entry(pltmp, ogfs_pl_t, pl_list);

		ogfs_unhold_lvb(sdp, pl->gl);
		pl->gl->gl_object = NULL;
		ogfs_put_glstruct(pl->gl);

		list_del(pltmp);
		gfree(pl, sizeof(pl));
	}

	EXIT(LOAD_OTHER_PLOCK_ELEMS);
	return err;
}

/**
 * lock_my_plock_elems - grab the exl lock on my plock elements
 * @sdp: < superblock pointer
 * @pljid: <> jid list head.
 *
 * Grabs the exl lock on all of our plocks.
 * Backs off if needed.
 *
 * Returns: error or zero
 */
int
lock_my_plock_elems(ogfs_sbd_t * sdp, pljid_t * pljid)
{
	struct list_head *pltmp, *pltop;
	ogfs_pl_t *pl = NULL;
	int err = 0;

	ENTER(LOCK_MY_PLOCK_ELEMS);

	pltmp = pltop = &pljid->locks;
	for (;;) {
		pltmp = pltmp->next;
		if (pltmp == pltop)
			break;
		pl = list_entry(pltmp, ogfs_pl_t, pl_list);

		err = ogfs_glock(sdp, pl->gl, GL_NOCACHE);
		if (err) {
			/*back off. */
			for (;;) {	/* whips through everything before, unlocking. */
				pltmp = pltmp->prev;
				if (pltmp == pltop)
					break;
				pl = list_entry(pltmp, ogfs_pl_t, pl_list);
				ogfs_gunlock(sdp, pl->gl, GL_NOCACHE);
			}
			break;
		}
		ogfs_plock_elem_in(&pl->hdata, pl->gl->gl_lvb);

	}
	EXIT(LOCK_MY_PLOCK_ELEMS);
	return err;
}

/**
 * sync_plock_elems - sync list of plock elems to lvb store.
 * @sdp: < superblock pointer
 * @pljid: <> jid list head.
 *
 */
void
sync_plock_elems(ogfs_sbd_t * sdp, pljid_t * pljid)
{
	struct list_head *pltmp, *pltop;
	ogfs_pl_t *pl = NULL;

	ENTER(SYNC_PLOCK_ELEMS);

	pltmp = pltop = &pljid->locks;
	for (;;) {
		pltmp = pltmp->next;
		if (pltmp == pltop)
			break;
		pl = list_entry(pltmp, ogfs_pl_t, pl_list);

		ogfs_plock_elem_out(&pl->hdata, pl->gl->gl_lvb);

		ogfs_sync_lvb(sdp, pl->gl);
		ogfs_gunlock(sdp, pl->gl, GL_NOCACHE);

	}
	EXIT(SYNC_PLOCK_ELEMS);
}

/**
 * drop_plock_elems - drop all refs and memory to a list of plock elems
 * @sdp: < superblock pointer
 * @pljid: <> jid list head
 *
 */
void
drop_plock_elems(ogfs_sbd_t * sdp, pljid_t * pljid)
{
	struct list_head *pltmp, *pltop;
	ogfs_pl_t *pl = NULL;

	ENTER(DROP_PLOCK_ELEMS);

	pltmp = pltop = &pljid->locks;
	for (;;) {
		pltmp = pltmp->next;
		if (pltmp == pltop)
			break;
		pl = list_entry(pltmp, ogfs_pl_t, pl_list);

		ogfs_plock_elem_out(&pl->hdata, pl->gl->gl_lvb);

		ogfs_unhold_lvb(sdp, pl->gl);
		ogfs_gunlock(sdp, pl->gl, GL_NOCACHE);
		pl->gl->gl_object = NULL;
		ogfs_put_glstruct(pl->gl);

		list_del(pltmp);
		gfree(pl, sizeof(ogfs_pl_t));

	}
	EXIT(DROP_PLOCK_ELEMS);
}

/**
 * print_plock_list - print the list out to the debug console.
 * @pljid: < pointer to plock journal header struct
 *
 * if you can't figure this out from the code, leave.
 */
void
print_plock_list(pljid_t * pljid)
{
	struct list_head *pltmp, *pltop;
	ogfs_pl_t *pl = NULL;

	ENTER(PRINT_PLOCK_LIST);

	db_printf(" Jid head\n");
	db_printf("     magic      = %u\n", pljid->hdata.magic);
	db_printf("     lock_group = %u\n", pljid->hdata.lock_group);
	db_printf("     lock_count = %u\n", pljid->hdata.lock_count);
	db_printf("     first      = %u\n", pljid->hdata.first);

	pltmp = pltop = &pljid->locks;
	for (;;) {
		pltmp = pltmp->next;
		if (pltmp == pltop)
			break;
		pl = list_entry(pltmp, ogfs_pl_t, pl_list);

		db_printf("  Plock range\n");
		db_printf("        begin = %" FMT_64 "u\n", pl->hdata.begin);
		db_printf("        end   = %" FMT_64 "u\n", pl->hdata.end);
		db_printf("        magic = %u\n", pl->hdata.magic);
		db_printf("        pid   = %u\n", pl->hdata.pid);
		db_printf("        Next  = %u\n", pl->hdata.Next);
		db_printf("        ltype = %u\n", pl->hdata.ltype);

	}
	EXIT(PRINT_PLOCK_LIST);
}

/**
 * load_my_plocks - lock/load all pointers to my plocks
 * @ip: < inode pointer
 * @plhd: <> plock header struct thingy
 *
 *
 * Returns: int
 */
int
load_my_plocks(ogfs_inode_t * ip, plhd_t * plhd)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	pljid_t *pljid;
	int err = 0;

	ENTER(LOAD_MY_PLOCKS);

	if (list_empty(&plhd->mlcks)) {
		pljid =
		    load_plock_jid_head(ip, plhd, sdp->sd_lockstruct.ls_jid);

		ogfs_plock_head_out(&plhd->hdata, plhd->gl->gl_lvb);

		init_free_list(&pljid->freeids);	/* at creation, they're all free */

		if (!pljid) {
			err = -ENOMEM;
			goto exit_lmp;
		}
		list_add(&pljid->plj_list, &plhd->mlcks);
	} else {
		struct list_head *pljidtmp;

		pljidtmp = &plhd->mlcks;
		pljidtmp = pljidtmp->next;
		OGFS_ASSERT(pljidtmp != NULL,);
		pljid = list_entry(pljidtmp, pljid_t, plj_list);

		err = ogfs_glock(sdp, pljid->gl, GL_PERM);
		if (err)
			goto exit_lmp;

		ogfs_plock_jid_head_in(&pljid->hdata, pljid->gl->gl_lvb);

		OGFS_ASSERT(pljid->hdata.magic == PLOCK_JID_HEAD_MAGIC,);
	}

	err = lock_my_plock_elems(sdp, pljid);
	if (err) {
		/* bah, gotta undo pljid. */
		if (pljid->hdata.lock_count > 0)
			sync_plock_jid_head(ip, pljid);
		else
			drop_plock_jid_head(ip, pljid);
	}

exit_lmp:
	EXIT(LOAD_MY_PLOCKS);
	return err;
}

/**
 * sync_my_plocks - sync the plock lvbs for my plocks
 * @ip: < inode pointer
 * @plhd: <> plock header struct thingy.
 *
 * syncs out my plocks.  If I am not holding any, it frees up the jid
 * structure.
 * i will never call drop on my locks. I will remove them all.  That's why
 * this is a sync, and there is no drop_my_plocks. it should never happen.
 */
void
sync_my_plocks(ogfs_inode_t * ip, plhd_t * plhd)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	struct list_head *pljidtmp;
	pljid_t *pljid;

	ENTER(SYNC_MY_PLOCKS);

	if (list_empty(&plhd->mlcks))
		goto exit_sync_my_plocks;

	pljidtmp = &plhd->mlcks;
	pljidtmp = pljidtmp->next;
	OGFS_ASSERT(pljidtmp != NULL,);
	pljid = list_entry(pljidtmp, pljid_t, plj_list);

	if (list_empty(&pljid->locks)) {
		/* I have no locks, drop. */
		drop_plock_jid_head(ip, pljid);
	} else {
		/* I still have locks.
		 */
		sync_plock_elems(sdp, pljid);
		sync_plock_jid_head(ip, pljid);
	}
exit_sync_my_plocks:
	EXIT(SYNC_MY_PLOCKS);
}

/**
 * load_other_plocks - lock and load a ref to other node's plocks
 * @ip: < inode pointer
 * @plhd: <> plock head to all plocks.
 *
 * when ever this is called, the plhd->olcks list is empty.
 *
 * loops over all `live' jids. (how do I determin this?)
 * plhd->olcks, is a list of pljids which will have the locks.
 *
 * heh, that max jid stunt has some issues.  Now doing smarter thing.
 * Just check every possible jid that this system has. (look up number of
 * journals from superblockpointer.)
 *
 * Returns: int
 */
int
load_other_plocks(ogfs_inode_t * ip, plhd_t * plhd)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	pljid_t *pljid;
	unsigned int i;
	int err = 0;

	ENTER(LOAD_OTHER_PLOCKS);

	OGFS_ASSERT(list_empty(&plhd->olcks),);

	for (i = 0; i < sdp->sd_journals; i++) {
		if (i == sdp->sd_lockstruct.ls_jid)
			continue;	/* skip myself. */
		pljid = load_plock_jid_head(ip, plhd, i);
		/* null is an error, die. */
		if (!pljid) {
			err = -ENOMEM;
			goto exit_load_other_plocks;
		}
		list_add(&pljid->plj_list, &plhd->olcks);
		if (pljid->hdata.lock_count == 0) {
			/* no locks here, drop it. */
			drop_plock_jid_head(ip, pljid);
		} else {
			/* there be locks, get em. */
			err = load_other_plock_elems(sdp, pljid);
			if (err)
				goto exit_load_other_plocks;
		}
	}

	EXIT(LOAD_OTHER_PLOCKS);
	return 0;
exit_load_other_plocks:
	/* drop everything we just got. */
	{
		struct list_head *pjtmp, *pjtop;
		pjtmp = pjtop = &plhd->olcks;
		for (;;) {
			pjtmp = pjtmp->next;
			if (pjtmp == pjtop)
				break;
			pljid = list_entry(pjtmp, pljid_t, plj_list);

			drop_plock_elems(sdp, pljid);
			drop_plock_jid_head(ip, pljid);
		}
	}
	EXIT(LOAD_OTHER_PLOCKS);
	return err;
}

/**
 * drop_other_plocks - drops all refs to plocks from other nodes
 * @ip: < inode pointer
 * @plhd: <> plock head to all plocks.
 *
 * drops all plocks for others.  Since the only thing we ever do with these
 * is read them, this just drops everything. (we'll never hold these longer
 * than the conflict check.) You also notice that we're really not
 * bothering to put anything back into the lvb.  This is ok, nothing should
 * have changed.
 *
 * if I was really cute, the otherplocks would only get held shared and not
 * exculsive. but I'm not.
 */
void
drop_other_plocks(ogfs_inode_t * ip, plhd_t * plhd)
{
	struct list_head *pjtmp, *pjtop;
	pljid_t *pljid;

	ENTER(DROP_OTHER_PLOCKS);

	pjtmp = pjtop = &plhd->olcks;
	for (;;) {
		pjtmp = pjtmp->next;
		if (pjtmp == pjtop)
			break;
		pljid = list_entry(pjtmp, pljid_t, plj_list);

		drop_plock_elems(ip->i_sbd, pljid);
		drop_plock_jid_head(ip, pljid);
	}
	EXIT(DROP_OTHER_PLOCKS);
}

/**
 * add_plock - adds a plock to the jid's chain.
 * @ip: < inode pointer
 * @pljid: <> jid list header
 * @plold: <> plock that we want to be after.
 * @begin: < where this lock starts
 * @end: < where this lock ends
 * @type: < what kind of lock
 *
 * It is assumed that you've already made sure that this lock doesn't
 * conflict when you call this.
 *
 * Returns: <0: Error, =0: Ok
 */
int
add_plock_elem(ogfs_inode_t * ip, pljid_t * pljid, ogfs_pl_t * plold,
	       u64 begin, u64 end, char type)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	ogfs_pl_t *pl = NULL;
	u64 bid = 0;
	int err = 0;
	uint16 newlid = 0;

	ENTER(ADD_PLOCK_ELEM);

	newlid = next_free_list(&pljid->freeids);
	if (newlid == GND16) {
		err = -ENOLCK;
		goto out;
	}
	/*we're full. */
	bid = pack_plem_bid(pljid->jid, pljid->hdata.lock_group, newlid);

	pl = (ogfs_pl_t *) gmalloc(sizeof(ogfs_pl_t));
	if (!pl) {
		err = -ENOMEM;
		goto out;
	}
	memset(pl, 0, sizeof(ogfs_pl_t));
	INIT_LIST_HEAD(&pl->pl_list);
	pl->lid = newlid;

	pl->gl = ogfs_get_glstruct(sdp, bid, &ogfs_plock_glops, NULL, CREATE);
	if (!pl->gl) {
		gfree(pl, sizeof(plhd_t));
		err = -ENOMEM;
		goto out;
	}
	err = ogfs_hold_lvb(sdp, pl->gl);
	if (err) {
		ogfs_put_glstruct(pl->gl);
		gfree(pl, sizeof(plhd_t));
		goto out;
	}
	err = ogfs_glock(sdp, pl->gl, GL_PERM);
	if (err) {
		ogfs_unhold_lvb(sdp, pl->gl);
		ogfs_put_glstruct(pl->gl);
		gfree(pl, sizeof(plhd_t));
		goto out;
	}

	ogfs_plock_elem_in(&pl->hdata, pl->gl->gl_lvb);

	OGFS_ASSERT(pl->hdata.magic == 0,	/* bad things otherwise. */
		   db_printf("magic is:0x%x\n", pl->hdata.magic);
		   db_printf("jid:0x%x group:0x%x lid:0x%x Forms:0x%" FMT_64
			     "x\n", pljid->jid, pljid->hdata.lock_group, newlid,
			     bid);
	    );

	pl->hdata.begin = begin;
	pl->hdata.end = end;
	pl->hdata.magic = PLOCK_ELEM_MAGIC;
	pl->hdata.pid = osi_pid();
	pl->hdata.ltype = type;
	/* TODO this seems krufty.  see if you can think of a better way. */
	if (pljid->hdata.lock_count == 0) {	/* first lock added. */
		pl->hdata.Next = GND16;	/* nothing to follow. */
		pljid->hdata.first = newlid;

		list_add(&pl->pl_list, &pljid->locks);	/* add first item */
		IFDB(ADD_PLOCK_ELEM, db_printf(" Added first item.\n");
		    );
	} else if (plold == NULL) {
		/* adding to end of list. */
		ogfs_pl_t *plend;
		struct list_head *pllast;

		/* last item in list is prev from head. */
		pllast = &pljid->locks;
		pllast = pllast->prev;

		plend = list_entry(pllast, ogfs_pl_t, pl_list);

		pl->hdata.Next = GND16;
		plend->hdata.Next = newlid;

		list_add(&pl->pl_list, &plend->pl_list);

		ogfs_plock_elem_out(&plend->hdata, plend->gl->gl_lvb);
		IFDB(ADD_PLOCK_ELEM, db_printf(" Added to end.\n");
		    );
	} else {
		ogfs_pl_t *pltwo;
		struct list_head *plprev;

		plprev = &plold->pl_list;
		plprev = plprev->prev;
		if (plprev == &pljid->locks) {
			/* adding in front of the first item. */
			pl->hdata.Next = pljid->hdata.first;
			pljid->hdata.first = newlid;

			list_add(&pl->pl_list, &pljid->locks);	/*add to front */
			IFDB(ADD_PLOCK_ELEM,
			     db_printf
			     (" Added item right after head, folowing me b:%"
			      FMT_64 "u " "e:%" FMT_64 "u t:%d\n",
			      plold->hdata.begin, plold->hdata.end,
			      plold->hdata.ltype);
			    );
		} else {
			/* adding between two list elemsnts. */
			pltwo = list_entry(plprev, ogfs_pl_t, pl_list);

			pl->hdata.Next = pltwo->hdata.Next;
			pltwo->hdata.Next = newlid;

			list_add_tail(&pl->pl_list, &plold->pl_list);	/* add before... */
			ogfs_plock_elem_out(&pltwo->hdata, pltwo->gl->gl_lvb);
			IFDB(ADD_PLOCK_ELEM,
			     db_printf(" Added item between b:%" FMT_64 "u e:%"
				       FMT_64 "u t:%u and " "b:%" FMT_64 "u e:%"
				       FMT_64 "u t:%u\n", pltwo->hdata.begin,
				       pltwo->hdata.end, pltwo->hdata.ltype,
				       plold->hdata.begin, plold->hdata.end,
				       plold->hdata.ltype);
			    );
		}
	}

	pljid->hdata.lock_count++;

	ogfs_plock_elem_out(&pl->hdata, pl->gl->gl_lvb);
	ogfs_plock_jid_head_out(&pljid->hdata, pljid->gl->gl_lvb);

out:
	EXIT(ADD_PLOCK_ELEM);
	return err;
}

/**
 * remove_plock - remove a plock structure from server.
 * @sdp: < pointer to the super block.
 * @pljid: <> head of the loaded plock list
 * @pl: <> Items to be removed. (this get removed from memory!!!)
 *
 * WARNING!! There is the assumption being made that the [re]load call makes
 * the list in the same order as the lvb->Next list.  If this is not
 * true, chaos will follow!!!!
 *
 * heh, found chaos.....
 */
void
remove_plock(ogfs_sbd_t * sdp, pljid_t * pljid, ogfs_pl_t * pl)
{
	ENTER(REMOVE_PLOCK);
	IFDB(REMOVE_PLOCK,
	     db_printf("Removing plock element: b:%" FMT_64 "u e:%" FMT_64
		       "u t:%u " " lid:%u next:%u\n", pl->hdata.begin,
		       pl->hdata.end, pl->hdata.ltype, pl->lid, pl->hdata.Next);
	    );

	if (pljid->locks.next == &pl->pl_list) {
		/* first item in list. */
		OGFS_ASSERT(pljid->hdata.first == pl->lid,	/*make sure it is as we think. */
			   db_printf(" first:%u  lid:%u\n", pljid->hdata.first,
				     pl->lid);
		    );
		return_to_frl(&pljid->freeids, pl->lid);
		pljid->hdata.first = pl->hdata.Next;
	} else {
		/* someone else infront of us. */
		ogfs_pl_t *pltwo;
		struct list_head *plprev;

		OGFS_ASSERT(pljid->hdata.lock_count > 1,);

		plprev = &pl->pl_list;
		plprev = plprev->prev;
		if (plprev == &pljid->locks) {
			/*we're first. */
			/* I don't think it should ever get here..... */
			OGFS_ASSERT(pljid->hdata.first == pl->lid,	/*make sure it is as we think. */
				   db_printf(" first:%u  lid:%u\n",
					     pljid->hdata.first, pl->lid);
			    );
			return_to_frl(&pljid->freeids, pl->lid);
			pljid->hdata.first = pl->hdata.Next;
		} else {
			pltwo = list_entry(plprev, ogfs_pl_t, pl_list);

			OGFS_ASSERT(pltwo->hdata.Next == pl->lid,	/*make sure it is as we think. */
				   db_printf(" Next:%u  lid:%u\n",
					     pltwo->hdata.Next, pl->lid);
			    );
			return_to_frl(&pljid->freeids, pl->lid);
			pltwo->hdata.Next = pl->hdata.Next;
		}
	}

	pljid->hdata.lock_count--;

	ogfs_unhold_lvb(sdp, pl->gl);
	ogfs_gunlock(sdp, pl->gl, GL_NOCACHE);
	ogfs_put_glstruct(pl->gl);

	list_del(&pl->pl_list);
	gfree(pl, sizeof(ogfs_pl_t));

	EXIT(REMOVE_PLOCK);
}

/**
 * check_for_confilcts - see if the request conflicts with an existing lock
 * @plhd: < head of all locks
 * @begin: < start of the attempt
 * @end: < end of the attempt
 * @type: < type of the attempt
 * @pl: > plck we conflicted with
 *
 * Nice idea.  we'll see if it is worth the nrg......
 * this will need that deadlock detect code that may never get written.
 *
 * Returns: =0: No conflict, =-EAGAIN: Conflicts.
 */
int
check_for_confilcts(plhd_t * plhd, u64 begin, u64 end, char type,
		    ogfs_pl_t ** plout)
{
	struct list_head *pjtmp, *pjtop, *pltmp, *pltop;
	int error = 0;
	ogfs_pl_t *pl = NULL;
	pljid_t *pljid;

	/* check for local conflicts. */
	pjtmp = pjtop = &plhd->mlcks;
	pjtmp = pjtmp->next;
	if (pjtmp != pjtop) {
		pljid = list_entry(pjtmp, pljid_t, plj_list);

		pltmp = pltop = &pljid->locks;
		for (;;) {
			pltmp = pltmp->next;
			if (pltmp == pltop)
				break;
			pl = list_entry(pltmp, ogfs_pl_t, pl_list);

			/*same pid skip */
			if (pl->hdata.pid == osi_pid())
				continue;
			/* no overlap, no conflict. */
			if (!((pl->hdata.end >= begin) &&
			      (end >= pl->hdata.begin)))
				continue;
			/* lock types conflict? */
			if ((type == PLOCK_SH) && (pl->hdata.ltype == PLOCK_SH))
				continue;
			/* TODO deadlock detection. */
			error = -EAGAIN;
			goto out;
		}
	}

	/* check for remote conflicts. */
	pjtmp = pjtop = &plhd->olcks;
	for (;;) {
		pjtmp = pjtmp->next;
		if (pjtmp == pjtop)
			break;
		pljid = list_entry(pjtmp, pljid_t, plj_list);

		pltmp = pltop = &pljid->locks;
		for (;;) {
			pltmp = pltmp->next;
			if (pltmp == pltop)
				break;
			pl = list_entry(pltmp, ogfs_pl_t, pl_list);

			/* no overlap, no conflict. */
			if (!((pl->hdata.end >= begin) &&
			      (end >= pl->hdata.begin)))
				continue;
			/* lock types conflict? */
			if ((type == PLOCK_SH) && (pl->hdata.ltype == PLOCK_SH))
				continue;
			error = -EAGAIN;
			goto out;
		}
	}
out:
	if (plout != NULL)
		*plout = pl;
	return error;
}

/****************************************************************************/

/* this should be configurable via some interface.... */
/* maybe set this up to be a simple random of 1 or 2?
 * I say this because I'm seeing a lot of processes take turns looking at
 * the same set.  Not sure if it would help, but maybe.
 */
#define PLOCK_RETRY_SLEEP_TIME 1
/**
 * ogfs_plock_act - do a posix lock or unlock op on this inode
 * @ip:  <> inode pointer of file
 * @begin: < where this lock starts.
 * @end: < where it stops.
 * @type: < what kind is it.
 * @wait: < should we block?
 *
 * This is heavily based on the posix locking code in the linux kernel.
 * Since they did a seemingly good job, I figured why should I kill myself
 * writing my own version?  I should just cause myself pain porting
 * thiers....
 *
 * Returns: <0: Error, =0: Ok
 */
int
ogfs_plock_act(ogfs_inode_t * ip, u64 begin, u64 end, char type, char wait)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	plhd_t *plhd = NULL;
	pljid_t *pljid = NULL;
	ogfs_pl_t *pl = NULL, *right = NULL, *left = NULL;
	struct list_head *pltmp, *pltop, *pjtmp;
	int error = 0, added = 0, i;

	ENTER(GFN_PLOCK_ACT);
	IFDB(GFN_PLOCK_ACT,
	     db_printf("plock action on inode:%" FMT_64 "u, begin:%" FMT_64 "u "
		       "end:%" FMT_64 "u, type:%u, wait:%u\n",
		       ip->i_num.no_formal_ino, begin, end, type, wait);
	    );

	if (type != PLOCK_UN) {
		for (;;) {
			error = ogfs_jindex_hold(sdp, TRUE);
			if (error)
				goto error_out;
			plhd = load_plock_head(ip);
			ogfs_jindex_release(sdp);	/* gets release nomatter the result of load */
			if (!plhd) {
				error = -ENOMEM;
				goto error_out;
			}
			error = load_my_plocks(ip, plhd);
			if (error)
				goto error_head;
			error = load_other_plocks(ip, plhd);
			if (error)
				goto error_mine;

			error =
			    check_for_confilcts(plhd, begin, end, type, NULL);
			if (error == -EDEADLK)
				goto out;
			if (error == -EAGAIN && wait == FALSE)
				goto out;	/*don't wait for conflicts */

			if (osi_pending_signals()) {
				error = -ERESTARTSYS;
				goto out;
			}

			if (error == 0)
				break;	/* no conflicts! */

			/* we must block here.  which currently is a poll.
			 * which means, drop everything, wait, start over.
			 */
			IFDB(GFN_PLOCK_ACT,
			     db_printf
			     ("Conflicts, dropping all, sleeping, and retrying. %ld\n",
			      osi_pid());
			    );

			drop_other_plocks(ip, plhd);
			sync_my_plocks(ip, plhd);
			sync_plock_head(ip, plhd);
			osi_sleep(PLOCK_RETRY_SLEEP_TIME);
		}		/*for(;;) */

	} else {		/* still gotta load the lists. */
		error = ogfs_jindex_hold(sdp, TRUE);
		if (error)
			goto error_out;
		plhd = load_plock_head(ip);
		ogfs_jindex_release(sdp);	/* gets release nomatter the result of load */
		if (!plhd) {
			error = -ENOMEM;
			goto error_out;
		}
		error = load_my_plocks(ip, plhd);
		if (error)
			goto error_head;
	}

	/* No conflict if we made it this far.
	 */
	IFDB(GFN_PLOCK_ACT,
	     db_printf("Got out of plock blocks, now processing locks. %ld\n",
		       osi_pid());
	    );

	/* this is where the real magic begins. */
	pjtmp = &plhd->mlcks;
	pjtmp = pjtmp->next;
	pljid = list_entry(pjtmp, pljid_t, plj_list);
	pltmp = pltop = &pljid->locks;
	for (i = 0;; i++) {
		pltmp = pltmp->next;
		if (pltmp == pltop) {
			pl = NULL;
			break;
		}
		pl = list_entry(pltmp, ogfs_pl_t, pl_list);

		IFDB(GFN_PLOCK_ACT,
		     db_printf("  Looking at Plock range; b:%" FMT_64 "u e:%"
			       FMT_64 "u " "t:%u  icount:%u pid:%u (c:%ld)\n",
			       pl->hdata.begin, pl->hdata.end, pl->hdata.ltype,
			       i, pl->hdata.pid, osi_pid());
		    );

		if (osi_pid() != pl->hdata.pid)
			continue;

		if (type == pl->hdata.ltype) {
			if (pl->hdata.end < begin - 1)
				continue;
			if (pl->hdata.begin > end + 1)
				break;

			if (pl->hdata.begin > begin)
				pl->hdata.begin = begin;
			else
				begin = pl->hdata.begin;
			if (pl->hdata.end < end)
				pl->hdata.end = end;
			else
				end = pl->hdata.end;
			if (added) {
				remove_plock(sdp, pljid, pl);
				continue;
			}
			added = 1;

		} else {
			if (pl->hdata.end < begin)
				continue;
			if (pl->hdata.begin > end)
				break;
			if (type == PLOCK_UN)
				added = 1;
			if (pl->hdata.begin < begin)
				left = pl;
			if (pl->hdata.end > end) {
				right = pl;
				break;
			}
			if (pl->hdata.begin >= begin) {
				if (added) {
					remove_plock(sdp, pljid, pl);
					continue;
				}
				pl->hdata.begin = begin;
				pl->hdata.end = end;
				pl->hdata.ltype = type;

				added = 1;
			}
		}		/* if type == pl->ltype */

	}			/*for(;;) */

	error = 0;
	if (!added) {
		if (type == PLOCK_UN)
			goto out;
		error = add_plock_elem(ip, pljid, pl, begin, end, type);
	}
	if (right) {
		if (right == left) {
			error = add_plock_elem(ip, pljid, pl,
					       right->hdata.begin,
					       right->hdata.end,
					       right->hdata.ltype);
		}
		right->hdata.begin = end + 1;
	}
	if (left) {
		left->hdata.end = begin - 1;
	}
out:
	drop_other_plocks(ip, plhd);
error_mine:
	sync_my_plocks(ip, plhd);
error_head:
	sync_plock_head(ip, plhd);
error_out:
	IFDB(GFN_PLOCK_ACT, if (error == -ERESTARTSYS)
	     db_printf("Lock interruped by a signal.\n");) ;
	EXIT(GFN_PLOCK_ACT);
	return error;
}

/****************************************************************************/
/**
 * ogfs_ptestlock - check the status of a lock
 * @ip:  < inode
 * @begin: <> where the lock begins
 * @end: <> where it ends
 * @type: <> what kind of lock
 * @pid: > whose pid conflicts.
 *
 * TODO I think I still need to set the return value.....
 *
 * Returns: <0: Error =0: Ok.
 */
int
ogfs_ptestlock(ogfs_inode_t * ip, u64 * begin, u64 * end, char *type, u32 * pid)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	plhd_t *plhd = NULL;
	ogfs_pl_t *pl = NULL;
	int error = 0;

	ENTER(GFN_PTESTLOCK);
	IFDB(GFN_PTESTLOCK,
	     db_printf("ptesting inode:%" FMT_64 "u, begin:%" FMT_64 "u end:%"
		       FMT_64 "u, " "type:%u\n", ip->i_num.no_formal_ino,
		       *begin, *end, *type);
	    );

	/* get plock list. */
	error = ogfs_jindex_hold(sdp, TRUE);
	if (error)
		goto error_out;
	plhd = load_plock_head(ip);	/* calls load if listempty. */
	ogfs_jindex_release(sdp);	/* gets release no matter the result of load */
	if (!plhd) {
		error = -ENOMEM;
		goto error_out;
	}
	error = load_my_plocks(ip, plhd);
	if (error)
		goto error_head;
	error = load_other_plocks(ip, plhd);
	if (error)
		goto error_mine;

	error = check_for_confilcts(plhd, *begin, *end, *type, &pl);
	if (pl != NULL) {
		*begin = pl->hdata.begin;
		*end = pl->hdata.end;
		*type = pl->hdata.ltype;
		*pid = pl->hdata.pid;
	}

	drop_other_plocks(ip, plhd);
error_mine:
	sync_my_plocks(ip, plhd);
error_head:
	sync_plock_head(ip, plhd);
error_out:
	EXIT(GFN_PTESTLOCK);
	return error;
}
