/*
 *
 *    Copyright 2000-2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 */

#include "ogfs_debug.h"
#include "ogfs.h"

#include "dio.h"
#include "fs.h"
#include "glock.h"
#include "glops.h"
#include "inode.h"
#include "log.h"
#include "recovery.h"
#include "super.h"

#define bn2seg(bn) (((bn) - jdesc->ji_addr) / sdp->sd_sb.sb_seg_size)
#define seg2bn(seg) ((seg) * sdp->sd_sb.sb_seg_size + jdesc->ji_addr)

typedef int (desc_handler_t) (ogfs_sbd_t * sdp, ogfs_jindex_t * jdesc,
			      uint64 start, void *data);

struct replay_stats {
	unsigned int replays;
	unsigned int skips;
	unsigned int sames;
};
typedef struct replay_stats replay_stats_t;

/**
 * ogfs_log_header - read the log header for a given segment
 * @sdp: the filesystem
 * @jdesc: the journal
 * @seg: the segment to look at
 * @lh: the log header to return
 *
 * Read the log header for a given segement in a given journal.  Do a few
 * sanity checks on it.
 *
 * Returns: 0 on success, 1 if the header was invalid or incomplete and, -EXXX on error
 */

static int
get_log_header(ogfs_sbd_t * sdp, ogfs_jindex_t * jdesc, uint32 seg,
	       ogfs_log_header_t * lh)
{
	struct buffer_head *bh;
	ogfs_log_header_t lh2;
	ogfs_glock_t *gl;

	ENTER(GFN_GET_LOG_HEADER);

	gl = ogfs_get_glstruct(sdp, jdesc->ji_addr, &ogfs_meta_glops, NULL,
			      NO_CREATE);
	bh = ogfs_dread(sdp, seg2bn(seg), gl, DR_START | DR_WAIT);
	ogfs_put_glstruct(gl);

	if (!bh) {
		EXIT(GFN_GET_LOG_HEADER);
		return -EIO;
	}

	ogfs_log_header_in(lh, BH_DATA(bh));
	ogfs_log_header_in(&lh2,
			  BH_DATA(bh) + OGFS_BASIC_BLOCK -
			  sizeof(ogfs_log_header_t));

	brelse(bh);

	if (memcmp(lh, &lh2, sizeof(ogfs_log_header_t)) != 0 ||
	    lh->lh_header.mh_magic != OGFS_MAGIC ||
	    lh->lh_header.mh_type != OGFS_METATYPE_LH) {
		EXIT(GFN_GET_LOG_HEADER);
		return 1;
	}

	EXIT(GFN_GET_LOG_HEADER);

	return 0;
}

/**
 * find_good_lh - find a good log header
 * @sdp: the filesystem
 * @jdesc: the journal
 * @seg: the segment to start searching from (it's also filled in with a new value.)
 * @lh: the log header to fill in
 * @forward: if true search forward in the log, else search backward
 *
 * Call get_log_header() to get a log header for a segment, but if the
 * segment is bad, either scan forward or backward until we find a good one.
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
find_good_lh(ogfs_sbd_t * sdp, ogfs_jindex_t * jdesc,
	     uint32 * seg, ogfs_log_header_t * lh, int forward)
{
	int error;
	uint32 orig_seg = *seg;

	ENTER(GFN_FIND_GOOD_LH);

	for (;;) {
		error = get_log_header(sdp, jdesc, *seg, lh);
		if (error <= 0) {
			EXIT(GFN_FIND_GOOD_LH);
			return error;
		}

		if (forward) {
			if (++*seg == jdesc->ji_nsegment)
				*seg = 0;
		} else {
			if (*seg-- == 0)
				*seg = jdesc->ji_nsegment - 1;
		}

		/* FIXME: endless loop when compiled without asserts! */
		OGFS_ASSERT(*seg != orig_seg,);
	}
}

/**
 * verify_jhead - make sure we've found the head of the log
 * @sdp: the filesystem
 * @jdesc: the journal
 * @head: this is filled in with the log descriptor of the head
 *
 * At this point, seg and lh should be either the head of the log or just
 * before.  Scan forward until we find the head.
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
verify_jhead(ogfs_sbd_t * sdp, ogfs_jindex_t * jdesc, ogfs_log_header_t * head)
{
	ogfs_log_header_t lh;
	uint32 seg;
	int error;

	ENTER(GFN_VERIFY_JHEAD);

	seg = bn2seg(head->lh_first);

	for (;;) {
		if (++seg == jdesc->ji_nsegment)
			seg = 0;

		error = get_log_header(sdp, jdesc, seg, &lh);
		if (error < 0) {
			EXIT(GFN_VERIFY_JHEAD);
			return error;
		}

		/* FIXME: endless loop if log is broken in a certain way */
		if (error == 1)
			continue;
		if (lh.lh_sequence == head->lh_sequence)
			continue;

		/* FIXME: endless loop if log is broken in a certain way */
		/* FIXME: this breaks the log if an integer overflow occurs.
		 * since the sequence number is initialised with a random
		 * value, this *does* happen sometimes. */
		if (lh.lh_sequence < head->lh_sequence)
			break;

		memcpy(head, &lh, sizeof(ogfs_log_header_t));
	}

	EXIT(GFN_VERIFY_JHEAD);

	return 0;
}

/**
 * ogfs_find_jhead - find the head of a log
 * @sdp: the filesystem
 * @jdesc: the journal
 * @head: the log descriptor for the head of the log is returned here
 *
 * Do a binary search of a journal and find the valid log entry with the
 * highest sequence number.  (i.e. the log head)
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_find_jhead(ogfs_sbd_t * sdp, ogfs_jindex_t * jdesc, ogfs_log_header_t * head)
{
	ogfs_log_header_t lh1, lh_m;
	uint32 seg1, seg2, seg_m;
	int error;

	ENTER(GFN_FIND_JHEAD);

	seg1 = 0;
	seg2 = jdesc->ji_nsegment - 1;

	for (;;) {
		seg_m = (seg1 + seg2) / 2;

		error = find_good_lh(sdp, jdesc, &seg1, &lh1, TRUE);
		if (error)
			break;

		if (seg1 == seg_m) {
			error = verify_jhead(sdp, jdesc, &lh1);
			memcpy(head, &lh1, sizeof(ogfs_log_header_t));
			break;
		}

		error = find_good_lh(sdp, jdesc, &seg_m, &lh_m, FALSE);
		if (error)
			break;

		if (lh1.lh_sequence <= lh_m.lh_sequence)
			seg1 = seg_m;
		else
			seg2 = seg_m;
	}

	IFDB(GFN_FIND_JHEAD, if (error) {
	     db_printf("Can't find jhead (%d)\n", error);}
	     else
	     {
	     db_printf("Journal Head:\n"); ogfs_log_header_print(head, FALSE);}
	) ;

	EXIT(GFN_FIND_JHEAD);

	return error;
}

/**
 * increment_blkno - move to the next block in a journal
 * @sdp: the filesystem
 * @jdesc: the journal
 * @addr: the block number to increment
 * @skip_header: if this is TRUE, skip log headers
 *
 * Replace @addr with the location of the next block in the log.
 * Take care of journal wrap and skip of log header if necessary.
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
increment_blkno(ogfs_sbd_t * sdp, ogfs_jindex_t * jdesc, uint64 * addr,
		int skip_headers)
{
	ogfs_log_header_t header;
	int error;

	ENTER(GFN_INCREMENT_BLKNO);

	(*addr)++;

	/*  Handle journal wrap  */

	if (*addr == seg2bn(jdesc->ji_nsegment))
		*addr -= jdesc->ji_nsegment * sdp->sd_sb.sb_seg_size;

	/*  Handle landing on a header block  */

	if (skip_headers && (*addr % sdp->sd_sb.sb_seg_size) == 0) {
		error = get_log_header(sdp, jdesc, bn2seg(*addr), &header);
		if (error < 0) {
			EXIT(GFN_INCREMENT_BLKNO);
			return error;
		}

		OGFS_ASSERT(!error,);	/*  Corrupt headers here are bad  */
		OGFS_ASSERT(*addr != header.lh_first,);

		(*addr)++;
		/*  Can't wrap here  */
	}

	EXIT(GFN_INCREMENT_BLKNO);

	return 0;
}

/**
 * foreach_descriptor - go through the active part of the log
 * @sdp: the filesystem
 * @jdesc: the journal
 * @start: the first log header in the active region
 * @end: the last log header (don't process the contents of this entry))
 * @handler: the function to call for every log descriptor
 * @data: opaque data to pass to @handler
 *
 * Call a given function once for every log descriptor in the active
 * portion of the log.
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
foreach_descriptor(ogfs_sbd_t * sdp, ogfs_jindex_t * jdesc,
		   uint64 start, uint64 end, desc_handler_t handler, void *data)
{
	ogfs_log_header_t header;
	ogfs_log_descriptor_t desc;
	struct buffer_head *bh;
	ogfs_glock_t *gl;
	int error;

	ENTER(GFN_FOREACH_DESCRIPTOR);

	while (start != end) {
		OGFS_ASSERT(!(start % sdp->sd_sb.sb_seg_size),);

		error = get_log_header(sdp, jdesc, bn2seg(start), &header);
		if (error < 0)
			goto fail;

		OGFS_ASSERT(!error,);	/*  Corrupt headers are bad  */
		OGFS_ASSERT(header.lh_first == start,);

		IFDB(GFN_FOREACH_DESCRIPTOR,
		     db_printf("\nHeader at %" FMT_64 "u\n", start);
		     ogfs_log_header_print(&header, FALSE);
		    );

		start++;

		for (;;) {
			gl = ogfs_get_glstruct(sdp, jdesc->ji_addr,
					      &ogfs_meta_glops, NULL, NO_CREATE);
			bh = ogfs_dread(sdp, start, gl, DR_START | DR_WAIT);
			ogfs_put_glstruct(gl);

			if (!bh) {
				error = -EIO;
				goto fail;
			}

			ogfs_metatype_check(bh, OGFS_METATYPE_LD);
			ogfs_desc_in(&desc, BH_DATA(bh));

			brelse(bh);

			IFDB(GFN_FOREACH_DESCRIPTOR,
			     db_printf("Descriptor at %" FMT_64 "u\n", start);
			     ogfs_desc_print(&desc, FALSE);
			    );

			if (desc.ld_type != OGFS_LOG_DESC_LAST) {
				error = handler(sdp, jdesc, start, data);
				if (error)
					goto fail;

				while (desc.ld_length--) {
					error =
					    increment_blkno(sdp, jdesc, &start,
							    TRUE);
					if (error)
						goto fail;
				}
			} else {
				while (desc.ld_length--) {
					error =
					    increment_blkno(sdp, jdesc, &start,
							    (desc.
							     ld_length) ? TRUE :
							    FALSE);
					if (error)
						goto fail;
				}

				break;
			}
		}
	}

	EXIT(GFN_FOREACH_DESCRIPTOR);

	return 0;

fail:
	EXIT(GFN_FOREACH_DESCRIPTOR);
	return error;
}

/**
 * do_replay_local - replay a metadata block (when in local fs mode)
 * @sdp: the filesystem
 * @jdesc: the journal
 * @tag: the block tag describing the metadata block
 * @blkno: the incore copy of the block
 * @rstats: stats about the replay
 *
 * Returns: -EXXX on failure, 0 on success
 */

static int
do_replay_local(ogfs_sbd_t * sdp, ogfs_jindex_t * jdesc,
		ogfs_block_tag_t * tag, uint64 blkno, replay_stats_t * rstats)
{
	ogfs_glock_t *gl;
	struct buffer_head *inplace_bh, *log_bh;
	ogfs_meta_header_t inplace_mh, log_mh;
	int do_replay = TRUE;
	int error = 0;

	ENTER(GFN_DO_REPLAY_LOCAL);

	/*  The lock should be held, so we don't need to hold a count on
	   the structure.  We do need a pointer to it, though.  */

	gl = ogfs_get_glstruct(sdp, jdesc->ji_addr, &ogfs_meta_glops, NULL,
			      NO_CREATE);
	OGFS_ASSERT(gl,);
	ogfs_put_glstruct(gl);

	inplace_bh = ogfs_dread(sdp, tag->bt_blkno, NULL, DR_START | DR_WAIT);
	if (!inplace_bh) {
		error = -EIO;
		goto fail;
	}
	ogfs_meta_check(inplace_bh);
	ogfs_meta_header_in(&inplace_mh, BH_DATA(inplace_bh));

	log_bh = ogfs_dread(sdp, blkno, gl, DR_START | DR_WAIT);
	if (!log_bh) {
		error = -EIO;
		goto fail_ip_drelse;
	}
	ogfs_meta_check(log_bh);
	ogfs_meta_header_in(&log_mh, BH_DATA(log_bh));

	if (log_mh.mh_generation < inplace_mh.mh_generation) {
		IFDB(GFN_DO_REPLAY_LOCAL,
		     db_printf("skip %" FMT_64 "u\n", tag->bt_blkno);
		    );

		do_replay = FALSE;
		rstats->skips++;
	} else if (log_mh.mh_generation == inplace_mh.mh_generation) {
		if (memcmp
		    (BH_DATA(inplace_bh), BH_DATA(log_bh),
		     sdp->sd_sb.sb_bsize) == 0) {
			IFDB(GFN_DO_REPLAY_LOCAL,
			     db_printf("same %" FMT_64 "u\n", tag->bt_blkno);
			    );

			do_replay = FALSE;
			rstats->sames++;
		}
	}

	if (do_replay) {
		IFDB(GFN_DO_REPLAY_LOCAL,
		     db_printf("replay %" FMT_64 "u\n", tag->bt_blkno);
		    );

		memcpy(BH_DATA(inplace_bh), BH_DATA(log_bh),
			   sdp->sd_sb.sb_bsize);

		error = ogfs_dwrite(sdp, inplace_bh, DW_DIRTY);
		if (error)
			goto fail_log_drelse;

		rstats->replays++;
	}

	brelse(log_bh);
	brelse(inplace_bh);

	EXIT(GFN_DO_REPLAY_LOCAL);

	return 0;

fail_log_drelse:
	brelse(log_bh);

fail_ip_drelse:
	brelse(inplace_bh);

fail:
	EXIT(GFN_DO_REPLAY_LOCAL);
	return error;
}

/**
 * do_replay_multi - replay a metadata block (when in multihost mode)
 * @sdp: the filesystem
 * @jdesc: the journal
 * @tag: the block tag describing the metadata block
 * @blkno: the incore copy of the block
 * @rstats: stats about the replay
 *
 * Returns: -EXXX on failure, 0 on success
 */

static int
do_replay_multi(ogfs_sbd_t * sdp, ogfs_jindex_t * jdesc,
		ogfs_block_tag_t * tag, uint64 blkno, replay_stats_t * rstats)
{
	ogfs_glock_t *gl;
	struct buffer_head *inplace_bh, *log_bh;
	ogfs_meta_header_t inplace_mh, log_mh;
	char *data;
	int do_replay = TRUE;
	int error = 0;

	ENTER(GFN_DO_REPLAY_MULTI);

	/*  The lock should be held, so we don't need to hold a count on
	   the structure.  We do need a pointer to it, though.  */

	gl = ogfs_get_glstruct(sdp, jdesc->ji_addr, &ogfs_meta_glops, NULL,
			      NO_CREATE);
	OGFS_ASSERT(gl,);
	ogfs_put_glstruct(gl);

	data = gmalloc(sdp->sd_sb.sb_bsize);
	if (!data) {
		error = -ENOMEM;
		goto fail;
	}
	inplace_bh = ogfs_init_temp_bh(sdp, tag->bt_blkno, data, FALSE);
	if (!inplace_bh) {
		error = -ENOBUFS;
		goto fail_free;
	}
	error = ogfs_dreread(sdp, inplace_bh, DR_START | DR_WAIT);
	if (error)
		goto fail_dispose;
	ogfs_meta_check(inplace_bh);
	ogfs_meta_header_in(&inplace_mh, BH_DATA(inplace_bh));

	log_bh = ogfs_dread(sdp, blkno, gl, DR_START | DR_WAIT);
	if (!log_bh) {
		error = -EIO;
		goto fail_dispose;
	}
	ogfs_meta_check(log_bh);
	ogfs_meta_header_in(&log_mh, BH_DATA(log_bh));

	if (log_mh.mh_generation < inplace_mh.mh_generation) {
		IFDB(GFN_DO_REPLAY_MULTI,
		     db_printf("skip %" FMT_64 "u\n", tag->bt_blkno);
		    );

		do_replay = FALSE;
		rstats->skips++;
	} else if (log_mh.mh_generation == inplace_mh.mh_generation) {
		if (memcmp(BH_DATA(log_bh), data, sdp->sd_sb.sb_bsize) == 0) {
			IFDB(GFN_DO_REPLAY_MULTI,
			     db_printf("same %" FMT_64 "u\n", tag->bt_blkno);
			    );

			do_replay = FALSE;
			rstats->sames++;
		}
	}

	ogfs_dispose_temp_bh(inplace_bh);
	gfree(data, sdp->sd_sb.sb_bsize);

	if (do_replay) {
		IFDB(GFN_DO_REPLAY_MULTI,
		     db_printf("replay %" FMT_64 "u\n", tag->bt_blkno);
		    );

		inplace_bh =
		    ogfs_init_temp_bh(sdp, tag->bt_blkno, BH_DATA(log_bh), TRUE);
		if (!inplace_bh)
			error = -ENOBUFS;
		else {
			error = ogfs_dwrite(sdp, inplace_bh, DW_START | DW_WAIT);
			ogfs_dispose_temp_bh(inplace_bh);
		}

		if (error) {
			brelse(log_bh);
			goto fail;
		}

		rstats->replays++;
	}

	brelse(log_bh);

	EXIT(GFN_DO_REPLAY_MULTI);

	return 0;

fail_dispose:
	ogfs_dispose_temp_bh(inplace_bh);

fail_free:
	gfree(data, sdp->sd_sb.sb_bsize);

fail:
	EXIT(GFN_DO_REPLAY_MULTI);
	return error;
}

/**
 * replay_metadata - Do the actual journal replay for a descriptor
 * @sdp: the filesystem
 * @jdesc: the journal
 * @start: the block containing the descriptor to replay
 * @data: unused opaque data
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
replay_metadata(ogfs_sbd_t * sdp, ogfs_jindex_t * jdesc, uint64 cblk, void *data)
{
	replay_stats_t *rstats = (replay_stats_t *) data;
	ogfs_glock_t *gl;
	ogfs_log_descriptor_t desc;
	ogfs_block_tag_t tag;
	struct buffer_head *bh;

	uint64 blkno;
	uint32 tagn;
	unsigned int count_ctl, offset;
	int error;

	ENTER(GFN_REPLAY_METADATA);

	/*  The lock should be held, so we don't need to hold a count on
	   the structure.  We do need a pointer to it, though.  */

	gl = ogfs_get_glstruct(sdp, jdesc->ji_addr, &ogfs_meta_glops, NULL,
			      NO_CREATE);
	OGFS_ASSERT(gl,);
	ogfs_put_glstruct(gl);

	bh = ogfs_dread(sdp, cblk, gl, DR_START | DR_WAIT);
	if (!bh) {
		error = -EIO;
		goto fail;
	}

	ogfs_metatype_check(bh, OGFS_METATYPE_LD);
	ogfs_desc_in(&desc, BH_DATA(bh));

	brelse(bh);

	/*  If this isn't a metadata descriptor, skip it.  */

	if (desc.ld_type != OGFS_LOG_DESC_METADATA) {
		EXIT(GFN_REPLAY_METADATA);
		return 0;
	}

	/*  Set up the pointer to the control blocks  */

	offset = sizeof(ogfs_log_descriptor_t);

	/*  Set up the pointer to the metadata blocks.
	   (Note we have to take care to skip log headers.)  */

	blkno = cblk;

	count_ctl = ogfs_struct2blk(sdp, desc.ld_data1, sizeof(ogfs_block_tag_t));
	while (count_ctl--) {
		error = increment_blkno(sdp, jdesc, &blkno, TRUE);
		if (error)
			goto fail;
	}

	for (tagn = 0; tagn < desc.ld_data1; tagn++) {
		/*  Read in the tag  */

		bh = ogfs_dread(sdp, cblk, gl, DR_START | DR_WAIT);
		if (!bh) {
			error = -EIO;
			goto fail;
		}

		ogfs_tag_in(&tag, BH_DATA(bh) + offset);

		brelse(bh);

		IFDB(GFN_REPLAY_METADATA,
		     db_printf("Block tag for %" FMT_64 "u\n", blkno);
		     ogfs_tag_print(&tag, FALSE);
		    );

		if (sdp->sd_args.ar_localcaching)
			error =
			    do_replay_local(sdp, jdesc, &tag, blkno, rstats);
		else
			error =
			    do_replay_multi(sdp, jdesc, &tag, blkno, rstats);

		if (error)
			goto fail;

		/*  Advance to the next block tag  */

		offset += sizeof(ogfs_block_tag_t);

		if (offset + sizeof(ogfs_block_tag_t) > sdp->sd_sb.sb_bsize) {
			offset = 0;

			error = increment_blkno(sdp, jdesc, &cblk, TRUE);
			if (error)
				goto fail;
		}

		/*  Advance to the next metadata block  */

		error = increment_blkno(sdp, jdesc, &blkno, TRUE);
		if (error)
			goto fail;
	}

	EXIT(GFN_REPLAY_METADATA);

	return 0;

fail:
	EXIT(GFN_REPLAY_METADATA);
	return error;
}

/**
 * clean_journal - mark a dirty journal as being clean
 * @sdp: the filesystem
 * @jdesc: the journal
 * @head: the head journal to start from
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
clean_journal(ogfs_sbd_t * sdp, ogfs_jindex_t * jdesc, ogfs_log_header_t * head)
{
	ogfs_log_header_t lh;
	ogfs_log_descriptor_t desc;
	struct buffer_head *bh;
	ogfs_glock_t *gl;

	uint32 seg;
	uint64 blkno;
	int error;

	ENTER(GFN_CLEAN_JOURNAL);

	gl = ogfs_get_glstruct(sdp, jdesc->ji_addr, &ogfs_meta_glops, NULL,
			      NO_CREATE);
	OGFS_ASSERT(gl,);

	seg = bn2seg(head->lh_first);

	for (;;) {
		if (++seg == jdesc->ji_nsegment)
			seg = 0;

		error = get_log_header(sdp, jdesc, seg, &lh);
		if (error < 0)
			goto fail;

		/*  Rewrite corrupt header blocks  */

		if (error == 1) {
			bh = ogfs_dgetblk(sdp, seg2bn(seg), gl);
			if (!bh) {
				error = -EIO;
				goto fail;
			}
			ogfs_prep_new_buffer(bh);

			ogfs_buffer_clear(bh);
			ogfs_log_header_out(head, BH_DATA(bh));
			ogfs_log_header_out(head,
					   BH_DATA(bh) + OGFS_BASIC_BLOCK -
					   sizeof(ogfs_log_header_t));
			error =
			    ogfs_dwrite(sdp, bh, DW_DIRTY | DW_START | DW_WAIT);
			brelse(bh);

			if (error)
				goto fail;
		}

		/*  Stop when we get to the end of the log.  */

		if (lh.lh_sequence < head->lh_sequence)
			break;
	}

	/*  Build a "last" descriptor for the transaction we are
	   about to commit by writing the shutdown header.  */

	memset(&desc, 0, sizeof(ogfs_log_descriptor_t));

	desc.ld_header.mh_magic = OGFS_MAGIC;
	desc.ld_header.mh_type = OGFS_METATYPE_LD;
	desc.ld_header.mh_format = OGFS_FORMAT_LD;
	desc.ld_type = OGFS_LOG_DESC_LAST;
	desc.ld_length = 0;

	for (blkno = head->lh_first + 1; blkno != seg2bn(seg);) {
		if (blkno % sdp->sd_sb.sb_seg_size)
			desc.ld_length++;
		if (++blkno == seg2bn(jdesc->ji_nsegment))
			blkno -= jdesc->ji_nsegment * sdp->sd_sb.sb_seg_size;
	}

	/*  Write the descriptor  */

	bh = ogfs_dgetblk(sdp, head->lh_first + 1, gl);
	if (!bh) {
		error = -EIO;
		goto fail;
	}
	ogfs_prep_new_buffer(bh);

	ogfs_buffer_clear(bh);
	ogfs_desc_out(&desc, BH_DATA(bh));
	error = ogfs_dwrite(sdp, bh, DW_DIRTY | DW_START | DW_WAIT);
	brelse(bh);

	if (error)
		goto fail;

	/*  Build a log header that says the journal is clean  */

	memset(&lh, 0, sizeof(ogfs_log_header_t));
	lh.lh_header.mh_magic = OGFS_MAGIC;
	lh.lh_header.mh_type = OGFS_METATYPE_LH;
	lh.lh_header.mh_format = OGFS_FORMAT_LH;
	lh.lh_flags = OGFS_LOG_HEAD_UNMOUNT;
	lh.lh_first = seg2bn(seg);
	lh.lh_sequence = head->lh_sequence + 1;
	/*  Don't care about tail  */
	lh.lh_last_nopd = head->lh_last_nopd;

	/*  Write the header  */

	bh = ogfs_dgetblk(sdp, lh.lh_first, gl);
	if (!bh) {
		error = -EIO;
		goto fail;
	}
	ogfs_prep_new_buffer(bh);

	ogfs_buffer_clear(bh);
	ogfs_log_header_out(&lh, BH_DATA(bh));
	ogfs_log_header_out(&lh,
			   BH_DATA(bh) + OGFS_BASIC_BLOCK -
			   sizeof(ogfs_log_header_t));
	error = ogfs_dwrite(sdp, bh, DW_DIRTY | DW_START | DW_WAIT);
	brelse(bh);

	if (error)
		goto fail;

	ogfs_put_glstruct(gl);

	EXIT(GFN_CLEAN_JOURNAL);

	return 0;

fail:
	ogfs_put_glstruct(gl);
	EXIT(GFN_CLEAN_JOURNAL);
	return error;
}

/**
 * ogfs_recover_journal - recovery a given journal
 * @sdp: the filesystem
 * @jid: the number of the journal to recover
 *
 * Acquire a journals lock, check to see if the journal is clean, and
 * do recovery if necessary.
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_recover_journal(ogfs_sbd_t * sdp, unsigned int jid)
{
	ogfs_log_header_t head;
	ogfs_jindex_t *jdesc;
	ogfs_glock_t *gl;
	replay_stats_t rstats;
	uint64 tstart, tstop;
	int error;

	ENTER(GFN_RECOVER_JOURNAL);

	/*  Find the right ogfs_jindex_t.  */

	OGFS_ASSERT(jid < sdp->sd_journals,);
	jdesc = &sdp->sd_jindex[jid];

	con_printf("OGFS:  Trying to acquire journal lock %u...\n", jid);

	/*  Aquire the journal lock so we can do recovery  */

	error = ogfs_glock_num(sdp, jdesc->ji_addr, &ogfs_meta_glops, NULL,
			      LM_FLAG_NOEXP | ((sdp->sd_lockstruct.ls_first) ? 0
					       : LM_FLAG_TRY), &gl);
	switch (error) {
	case 0:
		break;

	case GLR_TRYFAILED:
		con_printf("OGFS:  Busy\n");
		error = 0;

	default:
		goto fail;
	};

	con_printf("OGFS:  Looking at journal %u...\n", jid);

	error = ogfs_find_jhead(sdp, jdesc, &head);
	if (error)
		goto fail_gunlock;

	if (!(head.lh_flags & OGFS_LOG_HEAD_UNMOUNT)) {
		memset(&rstats, 0, sizeof(replay_stats_t));

		con_printf("OGFS:  Replaying journal...\n");
		tstart = osi_gettimeofday();

		/*  Acquire an exclusive hold on the transaction lock  */

		error = ogfs_glock(sdp, sdp->sd_trans_gl, LM_FLAG_NOEXP);
		if (error)
			goto fail_gunlock;

		error = foreach_descriptor(sdp, jdesc,
					   head.lh_tail, head.lh_first,
					   replay_metadata, &rstats);
		if (error)
			goto fail_gunlock_tr;

		ogfs_gunlock(sdp, sdp->sd_trans_gl, GL_NOCACHE);

		tstop = osi_gettimeofday();
		con_printf
		    ("OGFS:  examined %u blocks (%u needed to be replayed) in %"
		     FMT_64 "u seconds\n",
		     rstats.replays + rstats.skips + rstats.sames,
		     rstats.replays, DIV_RU(tstop - tstart, 1000000));
		db_printf("OGFS:  replays = %u, skips = %u, sames = %u\n",
			  rstats.replays, rstats.skips, rstats.sames);

		error = clean_journal(sdp, jdesc, &head);
		if (error)
			goto fail_gunlock_tr;
	}

	DO_PROF(LM_RESET_EXP,
		sdp->sd_lockstruct.ls_ops->reset_exp(sdp->sd_lockstruct.
						     ls_lockspace, jid);
	    );

	clear_bit(jid, sdp->sd_dirty_j);

	ogfs_gunlock(sdp, gl, GL_NOCACHE);
	ogfs_put_glstruct(gl);

	con_printf("OGFS:  Done\n");

	EXIT(GFN_RECOVER_JOURNAL);

	return 0;

fail_gunlock_tr:
	ogfs_gunlock(sdp, sdp->sd_trans_gl, GL_NOCACHE);

fail_gunlock:
	ogfs_gunlock(sdp, gl, GL_NOCACHE);
	ogfs_put_glstruct(gl);

	con_printf("OGFS:  %s\n", (error) ? "Failed" : "Done");

fail:
	EXIT(GFN_RECOVER_JOURNAL);
	return error;
}

/**
 * ogfs_check_journals - Recover any dirty journals
 * @sdp: the filesystem
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_check_journals(ogfs_sbd_t * sdp)
{
	unsigned int j;
	int error = 0;

	ENTER(GFN_CHECK_JOURNALS);

	/*  Just try to refresh the journal index.
	   If we can't get the glock, go ahead and try to recover anyway.
	   After mount, we are the only one messing with the journal index,
	   so we don't have to worry about sd_jindex disappearing beneath
	   us.  */

	error = ogfs_jindex_hold(sdp, FALSE);
	switch (error) {
	case 0:
		ogfs_jindex_release(sdp);
		break;
	case GLR_TRYFAILED:
		error = 0;
		break;
	default:
		goto out;
	}

	for (j = 0; j < sdp->sd_journals; j++)
		if (test_and_clear_bit(j, sdp->sd_dirty_j)) {
			OGFS_ASSERT(j != sdp->sd_lockstruct.ls_jid,);

			error = ogfs_recover_journal(sdp, j);
			if (error)
				break;
		}

out:
	EXIT(GFN_CHECK_JOURNALS);

	return error;
}

/**
 * collect_nopen - called by foreach_descriptor() to collect nopen counts
 * @sdp: the filesystem
 * @jdesc: the journal
 * @start: the address of the log descriptor
 * @data: unused opaque data
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
collect_nopen(ogfs_sbd_t * sdp, ogfs_jindex_t * jdesc, uint64 start, void *data)
{
	ogfs_glock_t *gl;
	ogfs_log_descriptor_t desc;
	struct buffer_head *bh;
	ogfs_tr_nop_t *nop;
	int is_first = TRUE;
	int *dump = (int *) data;
	unsigned int index, offset;
	ogfs_inum_t inum;
	int x, error;

	ENTER(GFN_COLLECT_NOPEN);

	/*  We only can recover our own journal for now.  */

	OGFS_ASSERT(jdesc == &sdp->sd_my_jdesc,);

	/*  The lock should be held, so we don't need to hold a count on
	   the structure.  We do need a pointer to it, though.  */

	gl = ogfs_get_glstruct(sdp, jdesc->ji_addr, &ogfs_meta_glops, NULL,
			      NO_CREATE);
	OGFS_ASSERT(gl,);
	ogfs_put_glstruct(gl);

	bh = ogfs_dread(sdp, start, gl, DR_START | DR_WAIT);
	if (!bh) {
		error = -EIO;
		goto fail;
	}

	ogfs_metatype_check(bh, OGFS_METATYPE_LD);
	ogfs_desc_in(&desc, BH_DATA(bh));

	brelse(bh);

	switch (desc.ld_type) {
	case OGFS_LOG_DESC_NOPOPEN:
		if (*dump) {
			OGFS_ASSERT(desc.ld_data1,);
			*dump = FALSE;
		} else
			OGFS_ASSERT(!desc.ld_data1,);
		break;

	case OGFS_LOG_DESC_NOPCLOSE:
		OGFS_ASSERT(!*dump,);
		break;

	default:
		OGFS_ASSERT(!*dump,);
		goto out;
	}

	for (index = 0; index < desc.ld_length; index++) {
		bh = ogfs_dread(sdp, start, gl, DR_START | DR_WAIT);
		if (!bh) {
			error = -EIO;
			goto fail;
		}

		if (is_first) {
			offset = sizeof(ogfs_log_descriptor_t);
			is_first = FALSE;
		} else
			offset = 0;

		for (; offset + sizeof(ogfs_inum_t) <= sdp->sd_sb.sb_bsize;
		     offset += sizeof(ogfs_inum_t)) {
			ogfs_inum_in(&inum, BH_DATA(bh) + offset);

			if (inum.no_formal_ino) {
				for (x = 0; x < 2; x++) {
					nop =
					    (ogfs_tr_nop_t *)
					    gmalloc(sizeof(ogfs_tr_nop_t));
					if (!nop) {
						brelse(bh);
						error = -ENOMEM;
						goto fail;
					}
					memset(nop, 0,
						   sizeof(ogfs_tr_nop_t));

					nop->nop_type = desc.ld_type;
					nop->nop_inum = inum;

					if (x)
						ogfs_nopen_merge(sdp, nop,
								&sdp->
								sd_nopen_od_list,
								&sdp->
								sd_nopen_od_count);
					else
						ogfs_nopen_merge(sdp, nop,
								&sdp->
								sd_nopen_ic_list,
								&sdp->
								sd_nopen_ic_count);
				}
			}
		}

		brelse(bh);

		error = increment_blkno(sdp, jdesc, &start, TRUE);
		if (error)
			goto fail;
	}

out:
	EXIT(GFN_COLLECT_NOPEN);

	return 0;

fail:
	EXIT(GFN_COLLECT_NOPEN);
	return error;
}

/**
 * ogfs_recover_nopen - recover the nopen counts in this machine's journal
 * @sdp: the filesystem
 * @head: the log descriptor for the head of the log
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_recover_nopen(ogfs_sbd_t * sdp, ogfs_log_header_t * head)
{
	int dump = TRUE;
	int error;

	ENTER(GFN_RECOVER_NOPEN);

	con_printf("OGFS:  Looking for unlinked inodes...\n");

	/*  Collect the list of nopen counts  */

	OGFS_ASSERT(head->lh_last_nopd,);

	error = foreach_descriptor(sdp, &sdp->sd_my_jdesc,
				   head->lh_last_nopd, head->lh_first,
				   collect_nopen, &dump);
	if (error)
		goto fail;

	OGFS_ASSERT(!dump,);

	/*  We need to make sure if we crash during this nopen dump that
	   all intermediate headers in the transaction point to the last
	   nopen dump before the one we're making so we don't lose it.  */

	sdp->sd_last_nopd = head->lh_last_nopd;

	con_printf("OGFS:  Found %d inodes\n",
		   atomic_read(&sdp->sd_nopen_od_count));

	EXIT(GFN_RECOVER_NOPEN);

	return 0;

fail:
	con_printf("OGFS:  Failed\n");

	EXIT(GFN_RECOVER_NOPEN);
	return error;
}
