/*
 *
 *    Copyright 1995-2000 Regents of the University of Minnesota
 *    Portions Copyright 1999-2001 Sistina Software, Inc.
 *    Portions Copyright 2001-2003 The OpenGFS Project.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 *    This file was maintained by:
 *      Kenneth W. Preslan <kpreslan@sistina.com>
 *
 */

#include "file_linux.h"
#include "ogfs_debug.h"
#include "ogfs.h"
#include "blklist.h"
#include "bmap.h"
#include "daemon.h"
#include <linux/dcache.h>
#include <linux/namei.h>
#include "dcache.h"
#include "dio.h"
#include "ogfs_linux.h"
#include "glock.h"
#include "glops.h"
#include "fs.h"
#include "inode.h"
#include "inode_linux.h"
#include "locking.h"
#include "log.h"
#include "mount.h"
#include "recovery.h"
#include "super.h"
#include "trans.h"
#include "version.h"

#include <asm/statfs.h>
#include <linux/statfs.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/sched.h>
#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>
#include <linux/blkdev.h>
#ifdef CONFIG_SSI
#include <cluster/nsc.h>
#include <cluster/ssi/ogfs.h>
#endif /* CONFIG_SSI */
#ifdef DEBUG_SYMBOLS
#define EXPORT_SYMTAB
#endif				/*  DEBUG_SYMBOLS  */
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/kmod.h>
#include <linux/proc_fs.h>


#include "ogfs_debug.h"
#include "ogfs.h"
#include "blklist.h"
#include "bmap.h"
#include "dio.h"
#include "file.h"
#include "format.h"
#include "fs.h"
#include "glock.h"
#include "glops.h"
#include "inode.h"
#include "recovery.h"
#include "super.h"

/**
 * ogfs_check_sb - Check superblock
 * @sdp: the filesystem
 * @sb: The superblock
 * @silent: Don't print a message if the check fails
 *
 * Checks the version code of the FS is one that we understand how to
 * read and that the sizes of the various on-disk structures have not
 * changed.
 */

int
ogfs_check_sb(ogfs_sbd_t * sdp, ogfs_sb_t * sb, int silent)
{
	unsigned int x;
	int error = 0;

	ENTER(GFN_CHECK_SB);

	if (sb->sb_header.mh_magic != OGFS_MAGIC ||
	    sb->sb_header.mh_type != OGFS_METATYPE_SB) {
		if (!silent)
			con_printf("OGFS:  not an OGFS filesystem\n");
		error = -EINVAL;
		goto out;
	}

	/*  If format numbers match exactly, we're done.  */

	if (sb->sb_fs_format == OGFS_FORMAT_FS &&
	    sb->sb_multihost_format == OGFS_FORMAT_MULTI)
		goto out;

	if (sb->sb_fs_format != OGFS_FORMAT_FS) {
		for (x = 0; ogfs_old_fs_formats[x]; x++)
			if (ogfs_old_fs_formats[x] == sb->sb_fs_format)
				break;

		if (!ogfs_old_fs_formats[x]) {
			con_printf
			    ("OGFS:  code version (%u, %u) is incompatible with ondisk format (%u, %u)\n",
			     OGFS_FORMAT_FS, OGFS_FORMAT_MULTI, sb->sb_fs_format,
			     sb->sb_multihost_format);
			con_printf
			    ("OGFS:  I don't know how to upgrade this FS\n");
			error = -EINVAL;
			goto out;
		}
	}

	if (sb->sb_multihost_format != OGFS_FORMAT_MULTI) {
		for (x = 0; ogfs_old_multihost_formats[x]; x++)
			if (ogfs_old_multihost_formats[x] ==
			    sb->sb_multihost_format)
				break;

		if (!ogfs_old_multihost_formats[x]) {
			con_printf
			    ("OGFS:  code version (%u, %u) is incompatible with ondisk format (%u, %u)\n",
			     OGFS_FORMAT_FS, OGFS_FORMAT_MULTI, sb->sb_fs_format,
			     sb->sb_multihost_format);
			con_printf
			    ("OGFS:  I don't know how to upgrade this FS\n");
			error = -EINVAL;
			goto out;
		}
	}

	if (!sdp->sd_args.ar_upgrade) {
		con_printf
		    ("OGFS:  code version (%u, %u) is incompatible with ondisk format (%u, %u)\n",
		     OGFS_FORMAT_FS, OGFS_FORMAT_MULTI, sb->sb_fs_format,
		     sb->sb_multihost_format);
		con_printf
		    ("OGFS:  Use the \"upgrade\" mount option to upgrade the FS\n");
		con_printf("OGFS:  See the HOWTO for more details\n");
		error = -EINVAL;
		goto out;
	}

      out:
	EXIT(GFN_CHECK_SB);

	return error;
}

/**
 * ogfs_read_sb - Read super block
 * @sdp: The OGFS superblock
 * @gl: the glock for the superblock (assumed to be held)
 * @silent: Don't print message if mount fails
 *
 */

int
ogfs_read_sb(ogfs_sbd_t * sdp, ogfs_glock_t * gl, int silent)
{
	struct buffer_head *bh;
	uint64 block;

	ENTER(GFN_READ_SB);

	block = OGFS_SB_ADDR * OGFS_BASIC_BLOCK / sdp->sd_sb.sb_bsize;

	bh = ogfs_dread(sdp, block, gl, DR_START | DR_WAIT);
	if (!bh) {
		if (!silent)
			con_printf("OGFS:  can't read superblock\n");
		EXIT(GFN_READ_SB);
		return -EIO;
	}

	OGFS_ASSERT(sizeof(ogfs_sb_t) <= BH_SIZE(bh),);

	ogfs_sb_in(&sdp->sd_sb, BH_DATA(bh));

	brelse(bh);

	if (ogfs_check_sb(sdp, &sdp->sd_sb, silent)) {
		EXIT(GFN_READ_SB);
		return -EINVAL;
	}

	sdp->sd_fsb2bb_shift = ffs(sdp->sd_sb.sb_bsize / OGFS_BASIC_BLOCK) - 1;
	sdp->sd_diptrs =
	    (sdp->sd_sb.sb_bsize - sizeof(ogfs_dinode_t)) / sizeof(uint64);
	sdp->sd_inptrs =
	    (sdp->sd_sb.sb_bsize - sizeof(ogfs_indirect_t)) / sizeof(uint64);
	sdp->sd_jbsize = sdp->sd_sb.sb_bsize - sizeof(ogfs_jdata_t);
	sdp->sd_hash_bsize = sdp->sd_sb.sb_bsize / 2;
	sdp->sd_hash_ptrs = sdp->sd_hash_bsize / sizeof(uint64);

	EXIT(GFN_READ_SB);

	return 0;
}

/**
 * ogfs_do_upgrade - upgrade a filesystem
 * @sdp: The OGFS superblock
 *
 */

int
ogfs_do_upgrade(ogfs_sbd_t * sdp, ogfs_glock_t * gl_sb)
{
	ogfs_glock_t *gl_j;
	ogfs_log_header_t lh;
	struct buffer_head *bh;
	unsigned int x;
	uint64 block;
	int error = 0;

	ENTER(GFN_DO_UPGRADE);

	/*  If format numbers match exactly, we're done.  */

	if (sdp->sd_sb.sb_fs_format == OGFS_FORMAT_FS &&
	    sdp->sd_sb.sb_multihost_format == OGFS_FORMAT_MULTI) {
		con_printf("OGFS:  no upgrade necessary\n");
		sdp->sd_args.ar_upgrade = FALSE;
		goto out;
	}

	error = ogfs_jindex_hold(sdp, FALSE);
	switch (error) {
	case 0:
		break;

	case GLR_TRYFAILED:
		error = -EBUSY;

	default:
		goto fail;
	}

	for (x = 0; x < sdp->sd_journals; x++) {
		error =
		    ogfs_glock_num(sdp, sdp->sd_jindex[x].ji_addr,
				  &ogfs_meta_glops, NULL, LM_FLAG_TRY, &gl_j);
		switch (error) {
		case 0:
			break;

		case GLR_TRYFAILED:
			error = -EBUSY;

		default:
			goto fail_jindex_relse;
		}

		error = ogfs_find_jhead(sdp, &sdp->sd_jindex[x], &lh);

		ogfs_gunlock(sdp, gl_j, 0);
		ogfs_put_glstruct(gl_j);

		if (error)
			goto fail_jindex_relse;

		if (!(lh.lh_flags & OGFS_LOG_HEAD_UNMOUNT)) {
			error = -EBUSY;
			goto fail_jindex_relse;
		}
	}

	/*  We don't need to journal this change because we're changing
	   only one sector of one block.  We definitely don't want to have
	   the journaling code running at this point.  */

	block = OGFS_SB_ADDR * OGFS_BASIC_BLOCK / sdp->sd_sb.sb_bsize;
	bh = ogfs_dread(sdp, block, gl_sb, DR_START | DR_WAIT);
	if (!bh) {
		error = -EIO;
		goto fail_jindex_relse;
	}

	ogfs_sb_in(&sdp->sd_sb, BH_DATA(bh));

	error = ogfs_check_sb(sdp, &sdp->sd_sb, FALSE);
	OGFS_ASSERT(!error,);

	sdp->sd_sb.sb_fs_format = OGFS_FORMAT_FS;
	sdp->sd_sb.sb_multihost_format = OGFS_FORMAT_MULTI;

	ogfs_sb_out(&sdp->sd_sb, BH_DATA(bh));

	error = ogfs_dwrite(sdp, bh, DW_DIRTY | DW_START | DW_WAIT);

	brelse(bh);

	ogfs_jindex_release(sdp);

	if (!error) {
		con_printf("OGFS:  upgrade successful\n");
		sdp->sd_args.ar_upgrade = FALSE;
	}

      out:
	EXIT(GFN_DO_UPGRADE);

	return error;

      fail_jindex_relse:
	ogfs_jindex_release(sdp);

      fail:
	if (error == -EBUSY)
		con_printf
		    ("OGFS:  can't upgrade:  the FS is still busy or contains dirty journals\n");
	else
		con_printf("OGFS:  can't upgrade:  error = %d\n", error);
	EXIT(GFN_DO_UPGRADE);
	return error;
}

/**
 * ogfs_clear_journals - Clear all the journal index information
 * @sdp: The OGFS superblock
 *
 */

void
ogfs_clear_journals(ogfs_sbd_t * sdp)
{
	int j;
	ENTER(GFN_CLEAR_JOURNALS);

	if (sdp->sd_jindex) {
		/* if external, there is a dev path/name in the jindex */
		for (j = 0;j < sdp->sd_journals; j++) {
			if ((sdp->sd_jindex + j)->ji_device[0])
				ogfs_put_jdev_info(sdp, j);
		}

		gfree(sdp->sd_jdevs, sdp->sd_journals
				* sizeof(struct block_device *));
		sdp->sd_jdevs = NULL;
		gfree(sdp->sd_jindex, sdp->sd_journals
				* sizeof(ogfs_jindex_t));
		sdp->sd_jindex = NULL;
	}
	sdp->sd_journals = 0;

	EXIT(GFN_CLEAR_JOURNALS);
}

/**
 * ogfs_journal_update - Update the journal index information
 * @ip: The journal index inode
 *
 * Returns: 0 on success, error code otherwise
 */

int
ogfs_ji_update(ogfs_inode_t * ip)
{
	ogfs_sbd_t *sdp = ip->i_sbd;
	ogfs_jindex_t buf;
	int rv;
	unsigned int j;
	uint64_t base;

	ENTER(GFN_JI_UPDATE);

	if (ip->i_di.di_size == 0) {
		con_printf("OGFS:  jindex dinode contains no journals\n");
		rv = -EINVAL;
		goto out;
	}
	if (ip->i_di.di_size % sizeof(ogfs_jindex_t) != 0) {
		con_printf("OGFS:  jindex dinode has bad size\n");
		rv = -EINVAL;
		goto out;
	}

	ogfs_clear_journals(sdp);

	sdp->sd_jindex = (ogfs_jindex_t *) gmalloc(ip->i_di.di_size);
	if (!sdp->sd_jindex)
	{
		con_printf("OGFS:  can't alloc mem for jindex\n");
		rv = -ENOMEM;
		goto out;
	}
	memset(sdp->sd_jindex, 0, ip->i_di.di_size);

	for (j = 0;; j++) {
		rv = ogfs_internal_read(ip, (char *) &buf,
				       j * sizeof(ogfs_jindex_t),
				       sizeof(ogfs_jindex_t));
		if (!rv)
			break;
		if (rv != sizeof(ogfs_jindex_t)) {
			if (rv > 0) {
				con_printf("OGFS:  bad size for jindex\n");
				rv = -EIO;
			}
			goto out;
		}

		ogfs_jindex_in(sdp->sd_jindex + j, (char *) &buf);
	}

	if (j == 0) {
		con_printf("OGFS:  contains no journals\n");
		rv = -EINVAL;
		goto out;
	}
	if (j * sizeof(ogfs_jindex_t) != ip->i_di.di_size) {
		con_printf("OGFS:  jindex dinode size mismatch\n");
		rv = -EINVAL;
		goto out;
	}
	if (j > OGFS_CLIENTS) {
		con_printf("OGFS:  too many journals\n");
		rv = -EINVAL;
		goto out;
	}

	sdp->sd_journals = j;
	sdp->sd_jiinode_vn = ip->i_gl->gl_vn;

	/* external journal support */
	sdp->sd_jdevs = (struct block_device **)
			gmalloc(j * sizeof(struct block_device *));
	if (!sdp->sd_jdevs)
	{
		con_printf("OGFS:  can't alloc mem for journal block devs\n");
		rv = -ENOMEM;
		goto out;
	}
	memset(sdp->sd_jdevs, 0, j * sizeof(struct block_device *));
	base = 18446744073709551615ULL;  /* UINT64_MAX */

	/* if external, there is a dev path/name in the jindex */
	for (j = 0;j < sdp->sd_journals; j++) {
		if ((sdp->sd_jindex + j)->ji_device[0]) {
			rv = ogfs_get_jdev_info(sdp, j);
			if (rv) {
				con_printf("OGFS:  ogfs_get_jdev_info() "
					"failed, jnl ID %d\n", j);
				rv = -EINVAL;
				break;
			}
			/* determine base (block #) of external journals */
			if ((sdp->sd_jindex + j)->ji_addr < base)
				base = (sdp->sd_jindex + j)->ji_addr;
		}
	}
	db_printf("external journal base = %" FMT_64 "u\n", base);
	sdp->sd_jext_base = base;

      out:
	EXIT(GFN_JI_UPDATE);
	return rv;
}

/**
 * ogfs_jindex_hold - Grab a lock on the jindex
 * @sdp: The OGFS superblock
 * @wait: block
 *
 * This is very similar to the ogfs_rindex_hold() function, except that
 * in general we hold the jindex lock for longer periods of time and
 * we grab it far less frequently (in general) then the rgrp lock.
 *
 * Returns: 0 on success, error code otherwise
 */

int
ogfs_jindex_hold(ogfs_sbd_t * sdp, int wait)
{
	ogfs_inode_t *ip = sdp->sd_jiinode;
	ogfs_glock_t *gl = ip->i_gl;
	int error;

	ENTER(GFN_JINDEX_HOLD);

	error = ogfs_glock_i(ip, GL_SHARED | ((wait) ? 0 : LM_FLAG_TRY));
	if (error)
		goto out;

	if (sdp->sd_jiinode_vn != gl->gl_vn) {
		error = ogfs_ji_update(ip);
		if (error)
			ogfs_gunlock_i(ip, 0);
	}

      out:
	EXIT(GFN_JINDEX_HOLD);
	return error;
}

/**
 * ogfs_ji_release - Release hold over journal index
 * @sdp: The GFs superblock
 *
 * Drop the lock on the journal index, allows the number and size of
 * journals to vary once more
 *
 */

void
ogfs_jindex_release(ogfs_sbd_t * sdp)
{
	ENTER(GFN_JINDEX_RELEASE);
	ogfs_gunlock_i(sdp->sd_jiinode, 0);
	EXIT(GFN_JINDEX_RELEASE);
}

/**
 * ogfs_compute_reserv - Compute various maximum sizes
 * @sdp: The OGFS superblock
 *
 */

void
ogfs_compute_reserv(ogfs_sbd_t * sdp)
{
	ogfs_inode_t ip;
	uint32 hash_blocks, ind_blocks, leaf_blocks;
	uint32 tmp_blocks;

	ENTER(GFN_COMPUTE_RESERV);

	/*  Compute maximum reservation required to add a entry to a directory
	 */

	hash_blocks =
	    DIV_RU(sizeof(uint64) * (1 << OGFS_DIR_MAX_DEPTH), sdp->sd_jbsize);

	ind_blocks = 0;
	for (tmp_blocks = hash_blocks; tmp_blocks > sdp->sd_diptrs;) {
		tmp_blocks = DIV_RU(tmp_blocks, sdp->sd_inptrs);
		ind_blocks += tmp_blocks;
	}

	leaf_blocks = (2 + OGFS_DIR_MAX_DEPTH);

	sdp->sd_max_dirres = hash_blocks + ind_blocks + leaf_blocks;

	/*  Compute the maximum height of a file's metadata tree  */

	ip.i_sbd = sdp;		/*  Damn kludge!!!  */
	ip.i_di.di_flags = OGFS_DIF_JDATA;
	ip.i_di.di_size = (uint64) - 1;

	sdp->sd_max_height = ogfs_calc_tree_height(&ip, 0);

	OGFS_ASSERT(sdp->sd_max_height <= OGFS_MAX_META_HEIGHT,);

	EXIT(GFN_COMPUTE_RESERV);
}

/**
 * ogfs_get_riinode - Initialise the rindex inode for the superblock
 * @sdp: The OGFS superblock
 *
 * Returns: 0 on success, error code otherwise
 */

int
ogfs_get_riinode(ogfs_sbd_t * sdp)
{
	ogfs_glock_t *gl;
	int error;

	ENTER(GFN_GET_RIINODE);

	error =
	    ogfs_glock_num(sdp, sdp->sd_sb.sb_rindex_di.no_formal_ino,
			  &ogfs_inode_glops, NULL, GL_SHARED, &gl);
	if (error)
		goto out;

	sdp->sd_riinode =
	    ogfs_get_istruct(sdp, gl, &sdp->sd_sb.sb_rindex_di, CREATE);
	if (!sdp->sd_riinode) {
		ogfs_gunlock(sdp, gl, 0);
		ogfs_put_glstruct(gl);
		error = -ENOMEM;
		goto out;
	}

	sdp->sd_riinode_vn = gl->gl_vn - 1;
	set_bit(GLF_STICKY, &gl->gl_flags);

	ogfs_gunlock(sdp, gl, 0);
	ogfs_put_glstruct(gl);

      out:
	EXIT(GFN_GET_RIINODE);

	return error;
}

/**
 * ogfs_put_riinode - Put the rindex dinode
 * @sdp: The OGFS superblock
 *
 */

void
ogfs_put_riinode(ogfs_sbd_t * sdp)
{
	ENTER(GFN_PUT_RIINODE);
	ogfs_put_istruct(sdp->sd_riinode);
	sdp->sd_riinode = NULL;
	EXIT(GFN_PUT_RIINODE);
}

/**
 * ogfs_get_jiinode - Initialise the jindex inode for the superblock
 * @sdp: The OGFS superblock
 *
 * Returns: 0 on success, error code otherwise
 */

int
ogfs_get_jiinode(ogfs_sbd_t * sdp)
{
	ogfs_glock_t *gl;
	int error;

	ENTER(GFN_GET_JIINODE);

	error =
	    ogfs_glock_num(sdp, sdp->sd_sb.sb_jindex_di.no_formal_ino,
			  &ogfs_inode_glops, NULL, GL_SHARED, &gl);
	if (error)
		goto out;

	sdp->sd_jiinode =
	    ogfs_get_istruct(sdp, gl, &sdp->sd_sb.sb_jindex_di, CREATE);
	if (!sdp->sd_jiinode) {
		ogfs_gunlock(sdp, gl, 0);
		ogfs_put_glstruct(gl);
		error = -ENOMEM;
		goto out;
	}

	sdp->sd_jiinode_vn = gl->gl_vn - 1;
	set_bit(GLF_STICKY, &gl->gl_flags);

	ogfs_gunlock(sdp, gl, 0);
	ogfs_put_glstruct(gl);

      out:
	EXIT(GFN_GET_JIINODE);

	return error;
}

/**
 * ogfs_put_jiinode - Put the jindex dinode
 * @sdp: The OGFS superblock
 *
 */

void
ogfs_put_jiinode(ogfs_sbd_t * sdp)
{
	ENTER(GFN_PUT_JIINODE);
	ogfs_put_istruct(sdp->sd_jiinode);
	sdp->sd_jiinode = NULL;
	EXIT(GFN_PUT_JIINODE);
}

/**
 * ogfs_setup_dameth - Setup the directory allocation method
 * @sdp: the filesystem
 *
 * Returns:  0 on success, -EXXX on failure
 */

int
ogfs_setup_dameth(ogfs_sbd_t * sdp)
{
	ogfs_rgrpd_t *rgd;
	unsigned int rg;
	int error = 0;

	ENTER(GFN_SETUP_DAMETH);

	error = ogfs_rindex_hold(sdp);
	if (error)
		goto fail;

	switch (sdp->sd_args.ar_dir_alloc_method) {
	case OGFS_DAMETH_SINGLE:
		/*  This code could be improved when we get LVBs.
		   We could verify that two machines aren't
		   picking the same RG.  */

		if (sdp->sd_rgcount >= sdp->sd_journals)
			rg = sdp->sd_rgcount * sdp->sd_lockstruct.ls_jid /
			    sdp->sd_journals;
		else
			rg = sdp->sd_lockstruct.ls_jid % sdp->sd_rgcount;

		break;

	case OGFS_DAMETH_RROBIN:
	case OGFS_DAMETH_RANDOM:
		/*  Do the same thing for both of these:  just initialize
		   sd_dir_alloc_target to a random value.  */

		rg = ogfs_randrange(0, sdp->sd_rgcount);

		break;

	default:
		con_printf("OGFS:  weird directory allocation method (%u)\n",
			   sdp->sd_args.ar_dir_alloc_method);
		error = -EINVAL;
		goto fail_ri_relse;
	}

	rgd = ogfs_rgrpd_get_by_num(sdp, rg);
	if (!rgd) {
		con_printf("OGFS:  can't determine initial resource group\n");
		error = -ENOENT;
		goto fail_ri_relse;
	}
	sdp->sd_dir_alloc_target = rgd->rd_ri.ri_addr;
	ogfs_rgrpd_put(rgd);

	ogfs_rindex_release(sdp);

	EXIT(GFN_SETUP_DAMETH);

	return 0;

      fail_ri_relse:
	ogfs_rindex_release(sdp);

      fail:
	EXIT(GFN_SETUP_DAMETH);
	return error;
}

/**
 * ogfs_update_datarget - Update the directory allocation target
 * @sdp: the filesystem
 * @res: the last inplace reservation
 *
 * Returns:  0 on success, -EXXX on failure
 */

int
ogfs_update_datarget(ogfs_sbd_t * sdp, ogfs_ipreserv_t * res)
{
	int error = 0;

	ENTER(GFN_UPDATE_DATARGET);

	switch (sdp->sd_args.ar_dir_alloc_method) {
	case OGFS_DAMETH_SINGLE:
		sdp->sd_dir_alloc_target = res->ir_rgd->rd_ri.ri_addr;

		break;

	case OGFS_DAMETH_RROBIN:
		sdp->sd_dir_alloc_target = res->ir_rgd->rd_ri.ri_addr + 1;

		break;

	case OGFS_DAMETH_RANDOM:
		error = ogfs_setup_dameth(sdp);

		break;

	default:
		con_printf("OGFS:  weird directory allocation method (%u)\n",
			   sdp->sd_args.ar_dir_alloc_method);
		error = -EINVAL;
	}

	EXIT(GFN_UPDATE_DATARGET);

	return error;
}

/**
 * ogfs_stat_rgrp - fill rgrp statistics info
 * @sdp: the filesystem
 * @rnum: rgrp number
 * @usage: the usage structure
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_stat_rgrp(ogfs_sbd_t * sdp, unsigned int rnum, ogfs_usage_t * usage)
{
	ogfs_rgrpd_t *rgd = NULL;
	int error = 0;

	rgd = ogfs_rgrpd_get_by_num(sdp, rnum);
	if (!rgd)
		return -ENOENT;

	error = ogfs_glock_rg(rgd, GL_SHARED | GL_SKIP);
	if (error) {
		ogfs_rgrpd_put(rgd);
		return error;
	}
	OGFS_ASSERT(rgd->rd_rs.rs_magic == OGFS_MAGIC,);

	/* Either lvb has no owner or the current machine owns
	 * it.  Everything is fine. */
	if (!rgd->rd_rs.rs_owner
	    || rgd->rd_rs.rs_owner ==
	    sdp->sd_lockstruct.ls_jid + 1) {
		IFDB(GFN_STAT_OGFS,
		     ogfs_rindex_print(&rgd->rd_ri, FALSE);
		     ogfs_rgrp_save_print(&rgd->rd_rs, FALSE);
		    );

		usage->total_blocks += rgd->rd_ri.ri_data;
		usage->free += rgd->rd_rs.rs_free;
		usage->used_dinode += rgd->rd_rs.rs_useddi;
		usage->free_dinode += rgd->rd_rs.rs_freedi;
		usage->used_meta += rgd->rd_rs.rs_usedmeta;
		usage->free_meta += rgd->rd_rs.rs_freemeta;

		ogfs_gunlock_rg(rgd, GL_SKIP);

	} else {
		/* Another machine owns the lvb, i.e. it is
		 * crashed and the data is stal.  Reinitialise
		 * it. */
		ogfs_gunlock_rg(rgd, GL_SKIP);

		error = ogfs_rgrp_lvb_init(rgd, TRUE);
	}

	ogfs_rgrpd_put(rgd);
	
	return error;
}


/**
 * ogfs_stat_ogfs - Do a statfs
 * @sdp: the filesystem
 * @usage: the usage structure
 * @interruptible:  Stop if there is a signal pending
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_stat_ogfs(ogfs_sbd_t * sdp, ogfs_usage_t * usage, int interruptible)
{
	uint64 rnum;
	int error = 0;

	ENTER(GFN_STAT_OGFS);

	memset(usage, 0, sizeof(ogfs_usage_t));
	usage->block_size = sdp->sd_sb.sb_bsize;

	error = ogfs_rindex_hold(sdp);
	if (error)
		return error;

	for ( rnum = 0 ; rnum < sdp->sd_rgcount; rnum++) {
		error = ogfs_stat_rgrp(sdp, rnum, usage);
		
		if (error)
			break;
		
		if (interruptible && osi_pending_signals()) {
			error = -ERESTARTSYS;
			break;
		}
	}

	ogfs_rindex_release(sdp);

	EXIT(GFN_STAT_OGFS);

	return error;
}


static struct super_block *ogfs_get_sb(struct file_system_type *fs_type,
        int flags, const char *dev_name, void *data);

static struct file_system_type ogfs_fs_type = {
        .owner          = THIS_MODULE,
        .name           = "ogfs",
        .get_sb         = ogfs_get_sb,
        .kill_sb        = kill_block_super,
        .fs_flags       = FS_REQUIRES_DEV,
};

/*
extern int ogfs_fill_super(struct super_block *sb, void *data, int silent);

static struct super_block *ogfs_get_sb(struct file_system_type *fs_type,
        int flags, const char *dev_name, void *data)
{
        return get_sb_bdev(fs_type, flags, dev_name, data, ogfs_fill_super);
}

static struct file_system_type ogfs_fs_type = {
        .owner          = THIS_MODULE,
        .name           = "ogfs",
        .get_sb         = ogfs_get_sb,
        .kill_sb        = kill_block_super,
        .fs_flags       = FS_REQUIRES_DEV,
};
*/

char *ogfs_mount_args = NULL;
struct semaphore ogfs_mount_args_lock;
uint32 ogfs_random_number;
unsigned int ogfs_mount_args_size = 0;

struct proc_dir_entry *ogfs_proc_entry = NULL;

#ifndef DEBUG_MEM

static kmem_cache_t *ogfs_glock_cachep = NULL;
static kmem_cache_t *ogfs_inode_cachep = NULL;
static kmem_cache_t *ogfs_bufdata_cachep = NULL;
static kmem_cache_t *ogfs_rgrpd_cachep = NULL;

#endif

ogfs_glock_t *
ogfs_alloc_glock(void)
{
#ifdef DEBUG_MEM
	return gmalloc(sizeof(ogfs_glock_t));
#else
	return kmem_cache_alloc(ogfs_glock_cachep, GFP_NOFS);
#endif
}

void
ogfs_free_glock(ogfs_glock_t * gl)
{
#ifdef DEBUG_MEM
	gfree(gl, sizeof(ogfs_glock_t));
#else
	kmem_cache_free(ogfs_glock_cachep, gl);
#endif
}

ogfs_inode_t *
ogfs_alloc_inode(void)
{
#ifdef DEBUG_MEM
	return gmalloc(sizeof(ogfs_inode_t));
#else
	return kmem_cache_alloc(ogfs_inode_cachep, GFP_NOFS);
#endif
}

void
ogfs_free_inode(ogfs_inode_t * ip)
{
#ifdef DEBUG_MEM
	gfree(ip, sizeof(ogfs_inode_t));
#else
	kmem_cache_free(ogfs_inode_cachep, ip);
#endif
}

ogfs_bufdata_t *
ogfs_alloc_bufdata(void)
{
#ifdef DEBUG_MEM
	return gmalloc(sizeof(ogfs_bufdata_t));
#else
	return kmem_cache_alloc(ogfs_bufdata_cachep, GFP_NOFS);
#endif
}

void
ogfs_free_bufdata(ogfs_bufdata_t * bd)
{
#ifdef DEBUG_MEM
	gfree(bd, sizeof(ogfs_bufdata_t));
#else
	kmem_cache_free(ogfs_bufdata_cachep, bd);
#endif
}

ogfs_rgrpd_t *
ogfs_alloc_rgrpd(void)
{
#ifdef DEBUG_MEM
	return gmalloc(sizeof(ogfs_rgrpd_t));
#else
	return kmem_cache_alloc(ogfs_rgrpd_cachep, GFP_NOFS);
#endif
}

void
ogfs_free_rgrpd(ogfs_rgrpd_t * rgd)
{
#ifdef DEBUG_MEM
	gfree(rgd, sizeof(ogfs_rgrpd_t));
#else
	kmem_cache_free(ogfs_rgrpd_cachep, rgd);
#endif
}


struct statfs;

#ifdef CONFIG_SSI
struct vfs_operations ogfs_mnt_ops = {
	get_uniqueid: 	ogfs_get_uniqueid,
	cluster_mount:	ogfs_notify_all,
	cluster_umount:	ogfs_client_umount,
	get_mount:	ogfs_get_mount,
};
#endif /* CONFIG_SSI */

/**
 * init_ogfs_fs - Register OGFS as a filesystem
 *
 * Returns: 0 on success, error code on failure
 */

int __init
init_ogfs_fs(void)
{
	int error;

#ifndef DEBUG_MEM
	ogfs_glock_cachep =
	    kmem_find_general_cachep(sizeof(ogfs_glock_t), GFP_NOFS);
	ogfs_inode_cachep =
	    kmem_find_general_cachep(sizeof(ogfs_inode_t), GFP_NOFS);
	ogfs_bufdata_cachep =
	    kmem_find_general_cachep(sizeof(ogfs_bufdata_t), GFP_NOFS);
	ogfs_rgrpd_cachep =
	    kmem_find_general_cachep(sizeof(ogfs_rgrpd_t), GFP_NOFS);

	if (!ogfs_glock_cachep || !ogfs_inode_cachep ||
	    !ogfs_bufdata_cachep || !ogfs_rgrpd_cachep) {
		printk("OGFS: cannot create OGFS kmem caches\n");
		return -ENOMEM;
	}
#endif

	error = register_filesystem(&ogfs_fs_type);
	if (error)
		return error;

	printk(KERN_INFO "%s\n", OGFS_RELEASE_STRING);

	init_MUTEX(&ogfs_mount_args_lock);

	ogfs_proc_entry =
	    create_proc_read_entry("fs/ogfs", S_IFREG | 0200, NULL, NULL, NULL);
	ogfs_proc_entry->write_proc = ogfs_proc_write;

	/*  Initialize the random number generator  */

	ogfs_random_number = osi_gettimeofday() & 0xFFFFFFFF;

#ifdef DEBUG_MEM
	{
		int x;
		for (x = 0; x < GMALLOC_HASH_SIZE; x++)
			INIT_LIST_HEAD(&gmalloc_head[x]);
	}
	memset(&gm_stats, 0, sizeof(gmalloc_stats_t));
	spin_lock_init(&gmalloc_lock);
#endif

#ifdef DEBUG_BUF
	{
		int x;
		for (x = 0; x < BUF_HASH_SIZE; x++)
			INIT_LIST_HEAD(&buf_head[x]);
	}
	memset(&buf_stats, 0, sizeof(buf_stats_t));
	spin_lock_init(&buf_lock);
#endif

#ifdef DEBUG_TRACE
	/* init debug flags ... you can't change these using ogfs_tool, */
	/*   until after the fs is mounted.  Set them up manually here. */
#if 0
	memset(ogfs_debug_flags, 0xFF, DEBUG_FLAG_BYTES); /* all active */
#else
	memset(ogfs_debug_flags, 0, DEBUG_FLAG_BYTES);    /* all off */
#endif
#define dbon(x) ogfs_debug_flags[(x) / 8] |= (1 << ((x) % 8))
#define dboff(x) ogfs_debug_flags[(x) / 8] &= ~(1 << ((x) % 8))
	dboff(GFN_SCAN_CACHED_INODES);
	dboff(GFN_SCAN_HELD_GLOCKS);
	dbon(GFN_FIND_JHEAD);
	dbon(GFN_FOREACH_DESCRIPTOR);
	dbon(GFN_DO_REPLAY_LOCAL);
	dbon(GFN_DO_REPLAY_MULTI);
	dbon(GFN_REPLAY_METADATA);
	dbon(GFN_RECOVER_JOURNAL);
	dbon(GFN_RECOVER_NOPEN);
#endif				/*  DEBUG_TRACE  */

#ifdef DEBUG_PRINT
	memset(ogfs_icdb_buf, 0, OGFS_ICDB_BUF_LEN);
	init_MUTEX(&ogfs_icdb_lock);
#endif				/*  DEBUG_PRINT  */

#ifdef DEBUG_PROFILE
	{
		int x;
		memset(ogfs_prof_stats, 0,
			   DEBUG_FLAGS * sizeof(ogfs_profile_t));
		for (x = 0; x < DEBUG_FLAGS; x++)
			ogfs_prof_stats[x].min_micros = (uint64) - 1;
	}
#endif				/*  DEBUG_PROFILE  */

	db_printf("sizeof(ogfs_meta_header_t) = %d\n",
		  sizeof(ogfs_meta_header_t));
	db_printf("sizeof(ogfs_sb_t) = %d\n", sizeof(ogfs_sb_t));
	db_printf("sizeof(ogfs_jindex_t) = %d\n", sizeof(ogfs_jindex_t));
	db_printf("sizeof(ogfs_rindex_t) = %d\n", sizeof(ogfs_rindex_t));
	db_printf("sizeof(ogfs_rgrp_t) = %d\n", sizeof(ogfs_rgrp_t));
	db_printf("sizeof(ogfs_indirect_t) = %d\n", sizeof(ogfs_indirect_t));
	db_printf("sizeof(ogfs_dinode_t) = %d\n", sizeof(ogfs_dinode_t));
	db_printf("sizeof(ogfs_dirent_t) = %d\n", sizeof(ogfs_dirent_t));
	db_printf("sizeof(ogfs_leaf_t) = %d\n", sizeof(ogfs_leaf_t));
	db_printf("sizeof(ogfs_log_header_t) = %d\n", sizeof(ogfs_log_header_t));
	db_printf("sizeof(ogfs_log_descriptor_t) = %d\n",
		  sizeof(ogfs_log_descriptor_t));
	db_printf("sizeof(ogfs_block_tag_t) = %d\n",
		  sizeof(ogfs_block_tag_t));
	db_printf("\n");
	db_printf("sizeof(ogfs_bitmap_t) = %d\n", sizeof(ogfs_bitmap_t));
	db_printf("sizeof(ogfs_rgrpd_t) = %d\n", sizeof(ogfs_rgrpd_t));
	db_printf("sizeof(ogfs_glock_t) = %d\n", sizeof(ogfs_glock_t));
	db_printf("sizeof(ogfs_bufdata_t) = %d\n", sizeof(ogfs_bufdata_t));
	db_printf("sizeof(ogfs_inode_t) = %d\n", sizeof(ogfs_inode_t));
	db_printf("sizeof(ogfs_trans_t) = %d\n", sizeof(ogfs_trans_t));
	db_printf("sizeof(ogfs_sbd_t) = %d\n", sizeof(ogfs_sbd_t));
	db_printf("sizeof(ogfs_dcached_t) = %d\n", sizeof(ogfs_dcached_t));

	return 0;
}

/**
 * exit_ogfs_fs - Unregister the file system
 *
 */

void __exit
exit_ogfs_fs(void)
{
	/*  Free any unused mount options  */

	down(&ogfs_mount_args_lock);
	if (ogfs_mount_args) {
		gfree(ogfs_mount_args, ogfs_mount_args_size);
		ogfs_mount_args = NULL;
	}
	up(&ogfs_mount_args_lock);

	remove_proc_entry("fs/ogfs", NULL);

	unregister_filesystem(&ogfs_fs_type);

	buf_check();
	gmalloc_check();
}

MODULE_DESCRIPTION("Global File System");
MODULE_AUTHOR("The OpenGFS Project");
MODULE_LICENSE("GPL");

module_init(init_ogfs_fs);
module_exit(exit_ogfs_fs);


/**
 * ogfs_read_super - Read in superblock
 * @sb: The VFS superblock
 * @data: Mount options
 * @silent: Don't complain if it's not an OGFS filesystem
 *
 * Returns: The VFS superblock, or NULL on error
 */

static __inline__ struct super_block *
_ogfs_read_super(struct super_block *sb, void *data, int silent)
{
	ogfs_sbd_t *sdp;
	ogfs_glock_t *gl_sb, *gl_i;
	ogfs_log_header_t head;
	ogfs_bhlist_t *bl;
	ogfs_tr_nop_t *nop;
	int super = TRUE, jindex = TRUE;
	int x;
	int error;

	ENTER(GFN_READ_SUPER);

	/* this ends up as vmalloc */
	sdp = (ogfs_sbd_t *) gmalloc(sizeof(ogfs_sbd_t));
	if (!sdp) {
		con_printf("OGFS:  out of memory\n");
		goto fail;
	}
	memset(sdp, 0, sizeof(ogfs_sbd_t));

	//vsb2sbd(sb) = sdp;
	sb->s_fs_info = sdp;
	sdp->sd_vsb = sb;

	/*  Set up the buffer cache and fill in some fake values
	   to allow us to read in the superblock.  */

	sdp->sd_sb.sb_bsize = bdev_hardsect_size(sb->s_bdev);
	if (sdp->sd_sb.sb_bsize < OGFS_BASIC_BLOCK)
		sdp->sd_sb.sb_bsize = OGFS_BASIC_BLOCK;
	set_blocksize(sb->s_bdev, sdp->sd_sb.sb_bsize);
	if (sizeof(ogfs_sb_t) > sdp->sd_sb.sb_bsize) {
		con_printf("OGFS:  superblock larger than device sector size\n");
		goto fail_sbd_free;
	}
	sdp->sd_jext_base = (uint64)1 << 63; /* UINT64_MAX / 2 */

	/* Init rgrp variables */

	INIT_LIST_HEAD(&sdp->sd_rglist);
	rwlock_init(&sdp->sd_rglock);

	/*  Init Glock structures  */

	for (x = 0; x < OGFS_GHASH_SIZE; x++) {
		INIT_LIST_HEAD(&sdp->sd_ghash[x].hb_notheld);
		INIT_LIST_HEAD(&sdp->sd_ghash[x].hb_held);
		INIT_LIST_HEAD(&sdp->sd_ghash[x].hb_perm);

		rwlock_init(&sdp->sd_ghash[x].hb_lock);
	}

	sdp->sd_gl_heldtime = 300;	/*  Time in seconds  */

	if (!ogfs_make_args((char *) data, &sdp->sd_args)) {
		con_printf("OGFS: ogfs_make_args() failed\n");
		goto fail_sbd_free;
	}

	if (ogfs_mount_lockproto(sdp, silent)) {
		con_printf("OGFS: ogfs_mount_lockproto() failed\n");
		goto fail_sbd_free;
	}

	if (sdp->sd_lockstruct.ls_ops->local_fs
	    && !sdp->sd_args.ar_ignore_local_fs) {
		db_printf("local filesystem\n");

		/*  Force local [p|f]locks  */
		db_printf("forcing 'localflocks' option\n");
		sdp->sd_args.ar_localflocks = TRUE;

		/*  Force local read ahead and caching  */
		db_printf("forcing 'localcaching' option\n");
		sdp->sd_args.ar_localcaching = TRUE;
	}

	/*  Start up the glockd thread  */

	init_MUTEX_LOCKED(&sdp->sd_glockd_sema);
	atomic_set(&sdp->sd_glockd_ctrl, 5);  /* 5 second exec frequency */

	error = kernel_thread(ogfs_glockd, sdp, 0);
	if (error < 0) {
		con_printf("OGFS:  can't start glockd thread (%d)\n", error);
		goto fail_gla_free;
	}
	down(&sdp->sd_glockd_sema);

	error = ogfs_glock_num(sdp, OGFS_MOUNT_LOCK, &ogfs_nondisk_glops, NULL,
			      GL_PERM | LM_FLAG_NOEXP, NULL);
	if (error) {
		con_printf("OGFS:  can't acquire mount glock (%d)\n", error);
		goto fail_glockd;
	}

	error = ogfs_glock_num(sdp, OGFS_LIVE_LOCK, &ogfs_nondisk_glops, NULL,
			      GL_SHARED | GL_PERM | GL_DISOWN | LM_FLAG_NOEXP,
			      NULL);
	if (error) {
		con_printf("OGFS:  can't acquire live glock (%d)\n", error);
		goto fail_gunlock_mount;
	}

	error = ogfs_glock_num(sdp, OGFS_SB_LOCK, &ogfs_meta_glops, NULL,
			      (sdp->sd_args.ar_upgrade) ? 0 : GL_SHARED,
			      &gl_sb);
	if (error) {
		con_printf("OGFS:  can't acquire superblock glock (%d)\n",
			   error);
		goto fail_gunlock_live;
	}

	if (ogfs_read_sb(sdp, gl_sb, silent) < 0) {
		con_printf("OGFS: ogfs_read_sb() failed\n");
		goto fail_gunlock_sb;
	}

	/*  Set up the buffer cache and SB for real  */

	if (sdp->sd_sb.sb_bsize > PAGE_SIZE) {
		con_printf
		    ("OGFS:  the blocksize (%u) is too big for the page size (%lu)\n",
		     sdp->sd_sb.sb_bsize, PAGE_SIZE);
		goto fail_gunlock_sb;
	}

	set_blocksize(sb->s_bdev, sdp->sd_sb.sb_bsize);

	/*  Copy out mount flags  */

	sdp->sd_noatime = (sb->s_flags & (MS_NOATIME | MS_NODIRATIME));
	sdp->sd_sync_mount = FALSE;	/*  Let the VFS deal with it  */

	/*  Setup up Virtual Super Block  */

	sb->s_blocksize = sdp->sd_sb.sb_bsize;
	sb->s_blocksize_bits = sdp->sd_sb.sb_bsize_shift;
	sb->s_magic = OGFS_MAGIC;
	sb->s_op = &ogfs_sops;
#ifdef CONFIG_SSI
	sb->s_mnt_op = &ogfs_mnt_ops;
#endif /* CONFIG_SSI */
	sb->s_flags |= MS_NOATIME | MS_NODIRATIME;
	sb->s_maxbytes = ~0ULL;

	ogfs_compute_reserv(sdp);

	/*  Read in journal index inode  */

	error = ogfs_get_jiinode(sdp);
	if (error) {
		con_printf("OGFS:  can't get journal index inode (%d)\n", error);
		goto fail_gunlock_sb;
	}

	/*  Upgrade version numbers if we need to  */

	if (sdp->sd_args.ar_upgrade) {
		error = ogfs_do_upgrade(sdp, gl_sb);
		if (error) {
			con_printf("OGFS: ogfs_do_upgrade() failed\n");
			goto fail_ji_free;
		}
	}

	/*  Get a handle on the transaction glock  */

	sdp->sd_trans_gl =
	    ogfs_get_glstruct(sdp, OGFS_TRANS_LOCK, &ogfs_trans_glops, NULL,
			     CREATE);
	if (!sdp->sd_trans_gl) {
		error = -EIO;
		con_printf("OGFS: ogfs_get_glstruct() failed\n");
		goto fail_ji_free;
	}
	set_bit(GLF_STICKY, &sdp->sd_trans_gl->gl_flags);

	/*  Load in the journal indexes  */

	error = ogfs_jindex_hold(sdp, TRUE);
	if (error) {
		con_printf("OGFS: can't read journal index (%d)\n", error);
		goto fail_trans_gl;
	}

	/* journal info for this machine */
	if (sdp->sd_lockstruct.ls_jid >= sdp->sd_journals) {
		con_printf
		    ("OGFS:  there are %u journals, you're trying to mount #%u\n",
		     sdp->sd_journals, sdp->sd_lockstruct.ls_jid + 1);
		goto fail_gunlock_ji;
	}
	sdp->sd_my_jdesc = sdp->sd_jindex[sdp->sd_lockstruct.ls_jid];
	sdp->sd_my_jdev = sdp->sd_jdevs[sdp->sd_lockstruct.ls_jid];

	/*  Initialize some log stuff (some of which is needed for recovery). */

	sdp->sd_log_seg_free = sdp->sd_my_jdesc.ji_nsegment - 1;

	INIT_LIST_HEAD(&sdp->sd_log_ail);
	INIT_LIST_HEAD(&sdp->sd_log_incore);
	init_MUTEX(&sdp->sd_log_lock);
	INIT_LIST_HEAD(&sdp->sd_log_bhlist_cache);
	INIT_LIST_HEAD(&sdp->sd_nopen_od_list);
	INIT_LIST_HEAD(&sdp->sd_nopen_ic_list);

	error =
	    ogfs_glock_num(sdp, sdp->sd_my_jdesc.ji_addr, &ogfs_meta_glops, NULL,
			  GL_PERM | LM_FLAG_NOEXP, &sdp->sd_my_jnl_gl);
	if (error) {
		con_printf("OGFS:  can't acquire the journal glock (%d)\n",
			   error);
		goto fail_gunlock_ji;
	}

	if (sdp->sd_lockstruct.ls_first) {
		for (x = 0; x < sdp->sd_journals; x++) {
			error = ogfs_recover_journal(sdp, x);
			if (error) {
				con_printf
				    ("OGFS:  error recovering journal %u (%d)\n",
				     x, error);
				goto fail_gunlock_journal;
			}
		}

		DO_PROF(LM_OTHERS_MAY_MOUNT,
			sdp->sd_lockstruct.ls_ops->others_may_mount(sdp->
								    sd_lockstruct.
								    ls_lockspace););
		sdp->sd_lockstruct.ls_first = FALSE;
	} else {
		error = ogfs_recover_journal(sdp, sdp->sd_lockstruct.ls_jid);
		if (error) {
			con_printf("OGFS:  error recovering my journal (%d)\n",
				   error);
			goto fail_gunlock_journal;
		}
	}

	ogfs_jindex_release(sdp);
	jindex = FALSE;

	/*  Disown my Journal glock  */


	error = ogfs_glock(sdp, sdp->sd_my_jnl_gl, GL_DISOWN);
	if (error) {
		con_printf("OGFS:  failed to disown my jrnl lock(%d)\n", error);
		goto fail_gunlock_journal;
	}
	ogfs_gunlock(sdp, sdp->sd_my_jnl_gl, 0);

	error = ogfs_find_jhead(sdp, &sdp->sd_my_jdesc, &head);
	if (error) {
		con_printf("OGFS:  can't find head of journal (%d)\n", error);
		goto fail_gunlock_journal;
	}
	if (!(head.lh_flags & OGFS_LOG_HEAD_UNMOUNT)) {
		con_printf("OGFS:  something wrong after ogfs_find_jhead\n");
		goto fail_gunlock_journal;
	}

	/*  Initialize some head of the log stuff  */

	sdp->sd_sequence = head.lh_sequence;
	sdp->sd_log_head = head.lh_first + 1;

	/*  Start up the recover thread
	   Make sure we clear all dirty bits or we might panic
	   if the lock module reported us expired while we were
	   replaying our journal.  */

	memset(sdp->sd_dirty_j, 0, OGFS_CLIENT_BYTES);

	init_MUTEX_LOCKED(&sdp->sd_recoverd_sema);
	atomic_set(&sdp->sd_recoverd_ctrl, 60);

	error = kernel_thread(ogfs_recoverd, sdp, 0);
	if (error < 0) {
		con_printf("OGFS:  can't start recoverd thread (%d)\n", error);
		goto fail_gunlock_journal;
	}
	down(&sdp->sd_recoverd_sema);

	/*  Recover nopen counts  */

	if (head.lh_last_nopd) {
		error = ogfs_recover_nopen(sdp, &head);
		if (error) {
			con_printf("OGFS:  can't recover nopen counts (%d)\n",
				   error);
			goto fail_nopen_list;	/*  Clean up from a partial nopen read  */
		}
	}

	/*  Dump out the first list of open inodes.  */

	ogfs_nopen_dump(sdp, TRUE);

	/*  Get a handle on the rename lock  */

	sdp->sd_rename_gl =
	    ogfs_get_glstruct(sdp, OGFS_RENAME_LOCK, &ogfs_nondisk_glops, NULL,
			     CREATE);
	if (!sdp->sd_rename_gl) {
		error = -EIO;
		con_printf("OGFS: ogfs_get_glstruct() 2 failed\n");
		goto fail_nopen_list;
	}

	/*  Read in the resource index inode  */

	error = ogfs_get_riinode(sdp);
	if (error) {
		con_printf("OGFS:  can't get resource index inode (%d)\n",
			   error);
		goto fail_rename_gl;
	}

	/*  Get the root inode  */

	error =
	    ogfs_glock_num(sdp, sdp->sd_sb.sb_root_di.no_formal_ino,
			  &ogfs_inode_glops, NULL, GL_SHARED, &gl_i);
	if (error) {
		con_printf("OGFS:  can't acquire root glock (%d)\n", error);
		goto fail_ri_free;
	}

	sdp->sd_rooti =
	    ogfs_get_istruct(sdp, gl_i, &sdp->sd_sb.sb_root_di, CREATE);

	ogfs_gunlock(sdp, gl_i, 0);
	ogfs_put_glstruct(gl_i);

	if (!sdp->sd_rooti) {
		con_printf("OGFS:  can't read in root inode\n");
		goto fail_ri_free;
	}

	/*  We're through with the superblock lock  */

	ogfs_gunlock(sdp, gl_sb, 0);
	ogfs_put_glstruct(gl_sb);
	super = FALSE;

	/*  Get the inode/dentry  */

	sb->s_root = d_alloc_root(ogfs_iget(sb, sdp->sd_rooti));

	ogfs_put_istruct(sdp->sd_rooti);

	if (!sb->s_root) {
		con_printf("OGFS:  can't get root dentry\n");
		goto fail_ri_free;
	}

	/*  Start up the inoded thread  */

	init_MUTEX_LOCKED(&sdp->sd_inoded_sema);
	atomic_set(&sdp->sd_inoded_ctrl, 5);

	error = kernel_thread(ogfs_inoded, sdp, 0);
	if (error < 0) {
		con_printf("OGFS:  can't start inoded thread (%d)\n", error);
		goto fail_dput;
	}
	down(&sdp->sd_inoded_sema);

	/*  Set up the directory allocation method  */

	error = ogfs_setup_dameth(sdp);
	if (error) {
		con_printf
		    ("OGFS:  can't set directory allocation method (%d)\n",
		     error);
		goto fail_inoded;
	}

	ogfs_gunlock_num(sdp, OGFS_MOUNT_LOCK, &ogfs_nondisk_glops, NULL,
			GL_NOCACHE);

	EXIT(GFN_READ_SUPER);

	return sb;

fail_inoded:
	atomic_set(&sdp->sd_inoded_ctrl, 0);
	wake_up_process(sdp->sd_inoded_process);
	down(&sdp->sd_inoded_sema);

fail_dput:
	dput(sb->s_root);
	while (invalidate_inodes(sdp->sd_vsb))	/*  Do Nothing  */
		;

fail_ri_free:
	ogfs_put_riinode(sdp);
	ogfs_clear_rgrpd(sdp);

fail_rename_gl:
	ogfs_put_glstruct(sdp->sd_rename_gl);

fail_nopen_list:
	while (!list_empty(&sdp->sd_nopen_od_list)) {
		nop =
		    list_entry(sdp->sd_nopen_od_list.next, ogfs_tr_nop_t,
				   nop_list);
		list_del(&nop->nop_list);
		gfree(nop, sizeof(ogfs_tr_nop_t));
	}
	while (!list_empty(&sdp->sd_nopen_ic_list)) {
		nop =
		    list_entry(sdp->sd_nopen_ic_list.next, ogfs_tr_nop_t,
				   nop_list);
		list_del(&nop->nop_list);
		gfree(nop, sizeof(ogfs_tr_nop_t));
	}
	while (!list_empty(&sdp->sd_log_bhlist_cache)) {
		bl = list_entry(sdp->sd_log_bhlist_cache.next,
				    ogfs_bhlist_t, bl_list);
		list_del(&bl->bl_list);
		gfree(bl, sizeof(ogfs_bhlist_t));
	}

	atomic_set(&sdp->sd_recoverd_ctrl, 0);
	wake_up_process(sdp->sd_recoverd_process);
	down(&sdp->sd_recoverd_sema);

fail_gunlock_journal:
	ogfs_gunlock(sdp, sdp->sd_my_jnl_gl, 0);
	ogfs_put_glstruct(sdp->sd_my_jnl_gl);

fail_gunlock_ji:
	if (jindex)
		ogfs_jindex_release(sdp);

fail_trans_gl:
	ogfs_put_glstruct(sdp->sd_trans_gl);

fail_ji_free:
	ogfs_put_jiinode(sdp);
	ogfs_clear_journals(sdp);

      fail_gunlock_sb:
	if (super) {
		ogfs_gunlock(sdp, gl_sb, 0);
		ogfs_put_glstruct(gl_sb);
	}

fail_gunlock_live:
	ogfs_gunlock_num(sdp, OGFS_LIVE_LOCK, &ogfs_nondisk_glops, NULL, 0);

fail_gunlock_mount:
	ogfs_gunlock_num(sdp, OGFS_MOUNT_LOCK, &ogfs_nondisk_glops, NULL, 0);

fail_glockd:
	atomic_set(&sdp->sd_glockd_ctrl, 0);
	wake_up_process(sdp->sd_glockd_process);
	down(&sdp->sd_glockd_sema);

fail_gla_free:
	ogfs_clear_gla(sdp);
	ogfs_unmount_lockproto(sdp);

fail_sbd_free:
	gfree(sdp, sizeof(ogfs_sbd_t));

fail:
	//vsb2sbd(sb) = NULL;
	sb->s_fs_info = NULL
	//sb->s_bdev = NODEV;

	EXIT(GFN_READ_SUPER);
	return NULL;
}


/* #ifndef CONFIG_SSI
struct super_block *  */
//ogfs_read_super(struct super_block *sb, void *data, int silent)
static int
ogfs_fill_super(struct super_block *sb, void *data, int silent)
{
	//int error;

	//return _ogfs_read_super(sb, data, silent);
	if(!_ogfs_read_super(sb, data, silent))  {
		con_printf("OGFS: failed _ogfs_read_super\n");
		return -1;
	}
	return 0;

}

static struct super_block *ogfs_get_sb(struct file_system_type *fs_type,
        int flags, const char *dev_name, void *data)
{
        return get_sb_bdev(fs_type, flags, dev_name, data, ogfs_fill_super);
}

/* #else */
#if 0
struct super_block *
ogfs_read_super (struct super_block *sb, void *data, int silent)
{
	struct ogfs_mount_data *cluster_data = data;

	/* Don't do anything special if SSI isn't initialized yet */
	if (cluster_state != CLUSTER_STATE_POSTROOT && cluster_state !=
	    CLUSTER_STATE_POSTINIT)
		return _ogfs_read_super(sb, data, silent);

	/* User mode mount just pass data on to original read_super */
	if (cluster_data == NULL || cluster_data->magic != OGFS_MOUNT_MAGIC)
		return _ogfs_read_super(sb, data, silent);

	/* Make sure this isn't left in the super block,
	 * it is used only internally for ogfs_notify_all().
	 */
	sb->s_flags &= ~MS_REMOTE;
	switch (cluster_data->mode) {
	case OGFS_DISCOVER:
	case OGFS_NOTIFY:
		return _ogfs_read_super(sb, NULL, silent);
	}

	printk(KERN_ERR "ogfs_read_super: Unknown mode\n");
	BUG();
	/* NOTREACHED */
}
//#endif /* CONFIG_SSI */
#endif /* 0 */

/**
 * connect_inode_to_vnode - Connect an OGFS inode to a Linux vnode
 * @ip: the OGFS inode
 * @inode: the Linux vnode
 *
 */

static void
connect_inode_to_vnode(ogfs_inode_t * ip, struct inode *inode)
{
	ENTER(GFN_CONNECT_INODE_TO_VNODE);

	ogfs_hold_istruct(ip);

	/*  Horrible kludge due to race between inode creation and destruction.  */
	while (ip->i_vino) {
		//ogfs_yield();
		schedule_timeout(1);
	}

	ip->i_vino = inode;
	vi2ino(inode) = ip;

	ogfs_inode_attr_in(ip);

	if (S_ISREG(inode->i_mode)) {
		inode->i_op = &ogfs_file_iops;
		inode->i_fop = &ogfs_file_fops;
		inode->i_mapping->a_ops = &ogfs_file_aops;
	} else if (S_ISDIR(inode->i_mode)) {
		inode->i_op = &ogfs_dir_iops;
		inode->i_fop = &ogfs_dir_fops;
	} else if (S_ISLNK(inode->i_mode)) {
		inode->i_op = &ogfs_symlink_iops;
	} else {
		inode->i_op = &ogfs_dev_iops;
		init_special_inode(inode, inode->i_mode, inode->i_rdev);
	}

	EXIT(GFN_CONNECT_INODE_TO_VNODE);
}

/**
 * ogfs_iget_for_nfs - Get an inode based on its number
 * @sdp: the filesystem
 * @inum: the inode number
 * @inode: the linux inode
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
ogfs_iget_for_nfs(ogfs_sbd_t * sdp, ogfs_inum_t * inum, struct inode *inode)
{
	ogfs_inode_t *ip;
	ogfs_glock_t *gl;
	int error;

	ENTER(GFN_IGET_FOR_NFS);

	error =
	    ogfs_glock_num(sdp, inum->no_formal_ino, &ogfs_inode_glops, NULL,
			  GL_SHARED, &gl);
	if (error)
		goto fail;

	ip = ogfs_get_istruct(sdp, gl, inum, CREATE);
	if (!ip)
		goto fail_gunlock;

	if (ip->i_di.di_flags & OGFS_DIF_UNUSED) {
		error = -EINVAL;
		goto fail_idestroy;
	}

	connect_inode_to_vnode(ip, inode);

	ogfs_put_istruct(ip);

	ogfs_gunlock(sdp, gl, 0);
	ogfs_put_glstruct(gl);

	EXIT(GFN_IGET_FOR_NFS);

	return 0;

fail_idestroy:
	ogfs_put_istruct(ip);
	ogfs_destroy_istruct(ip);

fail_gunlock:
	ogfs_gunlock(sdp, gl, 0);
	ogfs_put_glstruct(gl);

fail:
	EXIT(GFN_IGET_FOR_NFS);
	return error;
}

/**
 * ogfs_panic_inode - Panic
 * @inode: A VFS inode
 *
 * This is just here to make KNFSD happy.  It's never called.
 */

/*  UNUSED -- DELETE ME
static void
ogfs_panic_inode(struct inode *inode)
{
	ENTER(GFN_PANIC_INODE);

	OGFS_ASSERT(0,);

	EXIT(GFN_PANIC_INODE);
}
*/

/**
 * ogfs_read_inode - Read an inode
 * @inode: A VFS inode
 * @opaque: The inum of the dinode to read in
 *
 * Pull in an inode out of the OGFS inode cache.
 *
 * This will change when we start making a distinction between
 * formal inode numbers and disk addresses.
 *
 */

#warning fixme *_read_inode changed in 2.6

static void
ogfs_read_inode(struct inode *inode, void *opaque)
{
	ENTER(GFN_READ_INODE);

	if (opaque) {
		ogfs_inode_t *ip;

		/*  This section runs when we're called from OGFS via ogfs_iget()/iget4().  */

		OGFS_ASSERT(opaque,);
		ip = (ogfs_inode_t *) opaque;

		connect_inode_to_vnode(ip, inode);
	} else {
		ogfs_inum_t inum;
		int error;

		/*  This section runs when we're called from KNFSD.
		   When we start using the 2.4 NFS patches, KNFSD will call some
		   other function and this function won't contain this section.  */

		inum.no_formal_ino = inode->i_ino;
		inum.no_addr = inode->i_ino;	/*  This won't always be true  */

		error = ogfs_iget_for_nfs(inode->i_sb->s_fs_info, &inum, inode);
		if (error)
			make_bad_inode(inode);
	}

	EXIT(GFN_READ_INODE);
}

/*
 * ogfs_vread_inode - wrapper for ogfs_read_inode for vfs ops
 *
 * @inode: A VFS inode
 */
static void
ogfs_vread_inode(struct inode *inode)
{
	void *opaque = NULL;
	ogfs_read_inode(inode, opaque);
}

/**
 * ogfs_statfs - Gather and return stats about the filesystem
 * @sb: The superblock
 * @statfsbuf: The buffer
 *
 * Returns: 0 on success or error code
 */

static int
ogfs_statfs(struct super_block *sb, struct kstatfs *sfs)
{
	//ogfs_sbd_t *sdp = vsb2sbd(sb);
	ogfs_sbd_t *sdp = sb->s_fs_info;
	ogfs_usage_t usage;
	int error;

	ENTER(GFN_STATFS);

	error = ogfs_stat_ogfs(sdp, &usage, TRUE);
	if (error) {
		EXIT(GFN_STATFS);
		return error;
	}

	memset(sfs, 0, sizeof(struct statfs));

	sfs->f_type = (long)OGFS_MAGIC;
	sfs->f_bsize = (__u32)usage.block_size;
	sfs->f_blocks = (__u32)usage.total_blocks -
	    (usage.used_dinode + usage.free_dinode) -
	    (usage.used_meta + usage.free_meta);
	sfs->f_bfree = (__u32)usage.free;
	sfs->f_bavail = (__u32)usage.free;
	sfs->f_files = (__u32)usage.used_dinode + usage.free_dinode;
	sfs->f_ffree = (__u32)usage.free_dinode;
	sfs->f_namelen = (__u32)OGFS_FNAMESIZE;

	EXIT(GFN_STATFS);

	return 0;
}

/**
 * ogfs_clear_inode - Deallocate an inode when VFS is done with it
 * @inode: The VFS inode
 *
 */

static void
ogfs_clear_inode(struct inode *inode)
{
	ogfs_inode_t *ip;
	ogfs_sbd_t *sdp;

	ENTER(GFN_CLEAR_INODE);

	ip = vi2ino(inode);
	if (ip) {
		sdp = ip->i_sbd;

		ip->i_vino = NULL;
		vi2ino(inode) = NULL;

		ogfs_put_istruct(ip);

		atomic_inc(&sdp->sd_free_icount);
		if (atomic_read(&sdp->sd_free_icount) > 50)
			wake_up_process(sdp->sd_glockd_process);
	}

	EXIT(GFN_CLEAR_INODE);
}

/**
 * ogfs_remount_fs - called when the FS is remounted
 * @sb:  the filesystem
 * @flags:  the remount flags
 * @data:  extra data passed in (not used right now)
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
ogfs_remount_fs(struct super_block *sb, int *flags, char *data)
{
	//ogfs_sbd_t *sdp = vsb2sbd(sb);
	ogfs_sbd_t *sdp = sb->s_fs_info;

	ENTER(GFN_REMOUNT_FS);

	sdp->sd_noatime = (*flags & (MS_NOATIME | MS_NODIRATIME));
	sdp->sd_sync_mount = FALSE;	/*  Let the VFS deal with it  */

	/*  Don't let the VFS update atimes.  */
	*flags |= MS_NOATIME | MS_NODIRATIME;

	EXIT(GFN_REMOUNT_FS);

	return 0;
}

/**
 * ogfs_write_super - disk commit all incore transactions
 * @sb: the filesystem
 *
 * This function is called every time sync(2) is called.
 * After this exits, all dirty buffers and synced.
 */

static void
ogfs_write_super(struct super_block *sb)
{
	//ogfs_sbd_t *sdp = vsb2sbd(sb);
	ogfs_sbd_t *sdp = sb->s_fs_info;

	ENTER(GFN_WRITE_SUPER);

	ogfs_log_flush(sdp, NULL);
	ogfs_pull_tail(sdp);

	EXIT(GFN_WRITE_SUPER);
}

/**
 * ogfs_put_super - Unmount the filesystem
 * @sb: The VFS superblock
 *
 */

static void
ogfs_put_super(struct super_block *sb)
{
	//ogfs_sbd_t *sdp = vsb2sbd(sb);
	ogfs_sbd_t *sdp = sb->s_fs_info;
	ogfs_tr_nop_t *nop;
	ogfs_bhlist_t *bl;

	ENTER(GFN_PUT_SUPER);

	/*  Get rid of any VFS inodes laying around  */
	while (invalidate_inodes(sdp->sd_vsb))	/*  Do Nothing  */
		;

	/*  Kill off the inode thread  */
	atomic_set(&sdp->sd_inoded_ctrl, 0);
	wake_up_process(sdp->sd_inoded_process);
	down(&sdp->sd_inoded_sema);

	ogfs_put_riinode(sdp);

	/*  Kill off the recoverd thread  */
	atomic_set(&sdp->sd_recoverd_ctrl, 0);
	wake_up_process(sdp->sd_recoverd_process);
	down(&sdp->sd_recoverd_sema);

	ogfs_put_jiinode(sdp);

	/*  Kill off the glock thread  */
	atomic_set(&sdp->sd_glockd_ctrl, 0);
	wake_up_process(sdp->sd_glockd_process);
	down(&sdp->sd_glockd_sema);

	/*  Flush out all incore transactions  */
	ogfs_log_flush(sdp, NULL);

	/*  Dump out what's left of the nopen list.  */
	ogfs_nopen_dump(sdp, TRUE);

	/*  Clear out the AIL  */
	ogfs_pull_tail(sdp);

	/*  Shutdown the log manager  */
	ogfs_shutdown_log(sdp);

	/*  At this point, we're through modifying the disk  */

	/*  Release the journal lock and the mount lock  */

	ogfs_put_glstruct(sdp->sd_trans_gl);
	ogfs_put_glstruct(sdp->sd_rename_gl);

	ogfs_gunlock(sdp, sdp->sd_my_jnl_gl, 0);
	ogfs_put_glstruct(sdp->sd_my_jnl_gl);

	ogfs_gunlock_num(sdp, OGFS_LIVE_LOCK, &ogfs_nondisk_glops, NULL, 0);

	/*  Get rid of rgrp bitmap structures  */
	ogfs_clear_rgrpd(sdp);
	ogfs_clear_journals(sdp);

	/*  Take apart glock structures and buffer lists  */
	ogfs_clear_gla(sdp);

	/*  Unmount the locking protocol  */
	ogfs_unmount_lockproto(sdp);

	/*  At this point, we're through participating in the lockspace  */

	/*  Get rid of any nopen count structures  */

	while (!list_empty(&sdp->sd_nopen_od_list)) {
		nop =
		    list_entry(sdp->sd_nopen_od_list.next, ogfs_tr_nop_t,
				   nop_list);
		list_del(&nop->nop_list);
		gfree(nop, sizeof(ogfs_tr_nop_t));
	}
	while (!list_empty(&sdp->sd_nopen_ic_list)) {
		nop =
		    list_entry(sdp->sd_nopen_ic_list.next, ogfs_tr_nop_t,
				   nop_list);
		list_del(&nop->nop_list);
		gfree(nop, sizeof(ogfs_tr_nop_t));
	}

	/*  Get rid of the ogfs_bhlist_t cache  */

	while (!list_empty(&sdp->sd_log_bhlist_cache)) {
		bl = list_entry(sdp->sd_log_bhlist_cache.next,
				    ogfs_bhlist_t, bl_list);
		list_del(&bl->bl_list);
		gfree(bl, sizeof(ogfs_bhlist_t));
	}

	gfree(sdp, sizeof(ogfs_sbd_t));

	//vsb2sbd(sb) = NULL;
	sb->s_fs_info = NULL;

	EXIT(GFN_PUT_SUPER);
}

struct super_operations ogfs_sops = {
	//.alloc_inode	= ogfs_alloc_inode,
	//.destroy_inode	= ogfs_destroy_inode,
	.read_inode	= ogfs_vread_inode,
	.put_super	= ogfs_put_super,
	.statfs		= ogfs_statfs,
	.clear_inode	= ogfs_clear_inode,
	.remount_fs	= ogfs_remount_fs,
	.write_super	= ogfs_write_super
};

