/*
 *
 *    Copyright 1999-2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 */

#include "ogfs_debug.h"
#include "ogfs.h"

#include "blklist.h"
#include "dio.h"
#include "glock.h"
#include "inode.h"
#include "log.h"
#include "trans.h"

/**
 * ogfs_trans_print - Print a transaction to the console
 * @sdp: the filesystem
 * @tr: The OGFS transaction
 * @where: Situation of transaction
 *
 */

void
ogfs_trans_print(ogfs_sbd_t * sdp, ogfs_trans_t * tr, int where)
{
	ogfs_glock_t *gl;
	ogfs_bufdata_t *bd;
	ogfs_tr_nop_t *nop;
	struct list_head *ntmp, *gtmp, *btmp, *head;
	char *type;
	uint32 meta_type;
	uint64 blk;

	ENTER(GFN_TRANS_PRINT);

	switch (tr->tr_type) {
	case TRANS_COMPOUND:
		type = "compound";
		break;
	case TRANS_CREATE:
		type = "create";
		break;
	case TRANS_LINK:
		type = "link";
		break;
	case TRANS_RENAME:
		type = "rename";
		break;
	case TRANS_WRITE:
		type = "write";
		break;
	case TRANS_BMAP:
		type = "bmap";
		break;
	case TRANS_TRUNC_GROW:
		type = "truncate grow";
		break;
	case TRANS_PREPWRITE:
		type = "prepare write";
		break;
	case TRANS_NOTE_CH:
		type = "notify change";
		break;
	case TRANS_SETATTR:
		type = "set attributes";
		break;
	case TRANS_DIR2FILE:
		type = "dir to file";
		break;
	case TRANS_UNLINK:
		type = "unlink";
		break;
	case TRANS_RMDIR:
		type = "rmdir";
		break;
	case TRANS_LEAF_DEALLOC:
		type = "leaf deallocate";
		break;
	case TRANS_FILE_DEALLOC:
		type = "file deallocate";
		break;
	case TRANS_I_DEALLOC:
		type = "inode deallocate";
		break;
	default:
		type = "unknown";
		break;
	}

	db_printf
	    ("Transaction:  %s  flags = 0x%.8X, res=%u segs (%u blocks orig)\n",
	     type, tr->tr_flags, tr->tr_seg_reserved, tr->tr_blks_asked);
	db_printf
	    ("  tr_num_buf = %u, tr_num_gl = %u, tr_num_nop_o = %u,"
	     " tr_num_nop_c = %u\n",
	     tr->tr_num_buf, tr->tr_num_gl, tr->tr_num_nop_o, tr->tr_num_nop_c);

	for (ntmp = tr->tr_nop_change.next; ntmp != &tr->tr_nop_change;
	     ntmp = ntmp->next) {
		nop = list_entry(ntmp, ogfs_tr_nop_t, nop_list);
		switch (nop->nop_type) {
		case OGFS_LOG_DESC_NOPOPEN:
			db_printf("  inode %" FMT_64 "u/%" FMT_64
				  "u is now open\n",
				  nop->nop_inum.no_formal_ino,
				  nop->nop_inum.no_addr);
			break;
		case OGFS_LOG_DESC_NOPCLOSE:
			db_printf("  inode %" FMT_64 "u/%" FMT_64
				  "u is now closed\n",
				  nop->nop_inum.no_formal_ino,
				  nop->nop_inum.no_addr);
			break;
		default:
			db_printf("  weird nopen value (%d)\n", nop->nop_type);
			break;
		}
	}

	for (gtmp = tr->tr_gl.next; gtmp != &tr->tr_gl; gtmp = gtmp->next) {
		if (where == TRANS_IS_NEW) {
			gl = list_entry(gtmp, ogfs_glock_t, gl_new_list);
			head = &gl->gl_new_bufs;
		} else if (where == TRANS_IS_INCORE) {
			gl = list_entry(gtmp, ogfs_glock_t, gl_incore_list);
			head = &gl->gl_incore_bufs;
		} else {
			OGFS_ASSERT(0, con_printf("where = %d\n", where);
		    );
		}
		blk = ogfs_lockname2blk(sdp, &gl->gl_name);
		db_printf("  Glock (%" FMT_64 "u, %u)\n", blk,
			  gl->gl_name.ln_type);
		for (btmp = head->next; btmp != head; btmp = btmp->next) {
			if (where == TRANS_IS_NEW) {
				bd = list_entry(btmp, ogfs_bufdata_t,
						    bd_new_list);
			}
			else {
				bd = list_entry(
					btmp, ogfs_bufdata_t, bd_incore_list);
			}
			meta_type =
			    ((ogfs_meta_header_t *)BH_DATA(bd->bd_bh))->mh_type;
			meta_type = ogfs32_to_cpu(meta_type);
			db_printf("    buffer %" FMT_64 "u %u\n",
				  BH_BLKNO(bd->bd_bh), meta_type);
		}
	}

	EXIT(GFN_TRANS_PRINT);
}

/**
 * new_trans - Make a new transaction structure
 * @sdp: The OGFS superblock
 *
 * Returns: the new transaction
 */

static ogfs_trans_t *
new_trans(ogfs_sbd_t * sdp)
{
	ogfs_trans_t *tr = NULL;

	ENTER(GFN_GET_TRANS);

	tr = (ogfs_trans_t *) gmalloc(sizeof(ogfs_trans_t));
	if (tr) {
		memset(tr, 0, sizeof(ogfs_trans_t));
		INIT_LIST_HEAD(&tr->tr_gl);
		INIT_LIST_HEAD(&tr->tr_nop_change);
		INIT_LIST_HEAD(&tr->tr_nop_change_ic);
		INIT_LIST_HEAD(&tr->tr_bhlist);
		OGFS_ASSERT(!get_transaction(),);
		set_transaction(tr);
	}

	EXIT(GFN_GET_TRANS);

	return tr;
}

/**
 * ogfs_trans_begin - Perpare to start a transaction
 * @sdp: The OGFS superblock
 * @type: The type of the transaction
 * @meta_blocks: Reserve this many metadata blocks in the log
 * @extra_blocks: Number of non-metadata blocks to reserve
 *
 * Allocate the ogfs_trans_t struct.  Do in-place and
 * log reservations.
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
ogfs_trans_begin(ogfs_sbd_t * sdp, int type, unsigned int meta_blocks,
		unsigned int extra_blocks)
{
	ogfs_trans_t *tr;
	unsigned int blocks;
	int error;

	ENTER(GFN_TRANS_BEGIN);

	error = ogfs_glock(sdp, sdp->sd_trans_gl, GL_SHARED);
	if (error)
		goto fail;
	/*  Actually get the transaction  */
	tr = new_trans(sdp);
	if (!tr) {
		error = -ENOMEM;
		goto fail_gunlock;
	}
	tr->tr_type = type;
	/*  Do log reservation  */
	tr->tr_blks_asked = meta_blocks;
	blocks =
	    ogfs_struct2blk(sdp, meta_blocks,
			   sizeof(ogfs_block_tag_t)) + meta_blocks +
	    extra_blocks + 1;
	tr->tr_seg_reserved = ogfs_blk2seg(sdp, blocks);
	error = ogfs_log_reserv(sdp, tr->tr_seg_reserved);
	if (error)
		goto fail_free_trans;
	IFDB(GFN_TRANS_PRINT, ogfs_trans_print(sdp, tr, TRANS_IS_NEW);
	    );

	EXIT(GFN_TRANS_BEGIN);

	return 0;

fail_free_trans:
	set_transaction(NULL);
	gfree(tr, sizeof(ogfs_trans_t));

fail_gunlock:
	ogfs_gunlock(sdp, sdp->sd_trans_gl, 0);

fail:
	EXIT(GFN_TRANS_BEGIN);
	return error;
}

/**
 * clean_transaction_nop_lists - free the elements of the tr_nop_change and
 * tr_nop_change_ic lists in a transaction structure
 * @tr: the transaction
 *
 */

static __inline void
clean_transaction_nop_lists(ogfs_trans_t * tr)
{
	ogfs_tr_nop_t *nop;

	ENTER(GFN_CLEAN_TRANSACTION_NOP_LISTS);

	while (!list_empty(&tr->tr_nop_change)) {
		nop =
			list_entry(tr->tr_nop_change.next, ogfs_tr_nop_t,
				   nop_list);
		list_del(&nop->nop_list);
		gfree(nop, sizeof(ogfs_tr_nop_t));
	}
	while (!list_empty(&tr->tr_nop_change_ic)) {
		nop =
			list_entry(tr->tr_nop_change_ic.next,
				   ogfs_tr_nop_t, nop_list);
		list_del(&nop->nop_list);
		gfree(nop, sizeof(ogfs_tr_nop_t));
	}

	EXIT(GFN_CLEAN_TRANSACTION_NOP_LISTS);
	return;
}

/**
 * do_poison - poison a transaction
 * @sdp: the filesystem
 * @tr: the transaction
 *
 */

static void
do_poison(ogfs_sbd_t * sdp, ogfs_trans_t * tr)
{
	struct list_head *gtmp, *btmp;
	ogfs_glock_t *gl;
	ogfs_bufdata_t *bd;

	ENTER(GFN_DO_POISON);

	for (gtmp = tr->tr_gl.next; gtmp != &tr->tr_gl; gtmp = gtmp->next) {
		gl = list_entry(gtmp, ogfs_glock_t, gl_new_list);

		set_bit(GLF_POISONED, &gl->gl_flags);

		for (btmp = gl->gl_new_bufs.next; btmp != &gl->gl_new_bufs;
		     btmp = btmp->next) {
			bd = list_entry(btmp, ogfs_bufdata_t, bd_new_list);
			memcpy(BH_DATA(bd->bd_bh), bd->bd_frozen,
				   sdp->sd_sb.sb_bsize);
		}
	}
	clean_transaction_nop_lists(tr);

	EXIT(GFN_DO_POISON);
}

/**
 * ogfs_trans_end - End a transaction
 * @sdp: The OGFS superblock
 *
 * If buffers were actually added to the transaction,
 * commit it.
 */

void
ogfs_trans_end(ogfs_sbd_t * sdp)
{
	ogfs_trans_t *tr;
	ogfs_glock_t *gl;
	struct list_head *gtmp;
	int null_trans = TRUE;

	ENTER(GFN_TRANS_END);

	tr = get_transaction();
	OGFS_ASSERT(tr,);
	set_transaction(NULL);
	/*  Verify that all the glocks were passed to this transaction and
	 * look for null transactions  */
	for (gtmp = tr->tr_gl.next; gtmp != &tr->tr_gl; gtmp = gtmp->next) {
		gl = list_entry(gtmp, ogfs_glock_t, gl_new_list);
		OGFS_ASSERT(atomic_read(&gl->gl_locked) &&
			   gl->gl_pid == osi_pid() &&
			   gl->gl_state == LM_ST_EXCLUSIVE,);
		ogfs_put_glstruct(gl);
		if (!list_empty(&gl->gl_new_bufs))
			null_trans = FALSE;
	}

	if (null_trans) {
		while (!list_empty(&tr->tr_gl)) {
			gl = list_entry(
				tr->tr_gl.next, ogfs_glock_t, gl_new_list);
			list_del_init(&gl->gl_new_list);
			gl->gl_new_tr = NULL;
		}
		clean_transaction_nop_lists(tr);
		ogfs_log_relse(sdp, tr->tr_seg_reserved);
		gfree(tr, sizeof(ogfs_trans_t));
		ogfs_gunlock(sdp, sdp->sd_trans_gl, 0);

		EXIT(GFN_TRANS_END);
		return;
	}
	OGFS_ASSERT(tr->tr_num_buf <= tr->tr_blks_asked,
		   ogfs_trans_print(sdp, tr, TRANS_IS_NEW);
	    );
	/*  Handle poisoning  */
	if (tr->tr_flags & TRF_POISONED)
		do_poison(sdp, tr);
	IFDB(GFN_TRANS_PRINT, ogfs_trans_print(sdp, tr, TRANS_IS_NEW);
	    );
	ogfs_log_commit(sdp, tr);
	ogfs_gunlock(sdp, sdp->sd_trans_gl, 0);

	EXIT(GFN_TRANS_END);

	return;
}

/**
 * ogfs_trans_add_gl - Add a glock to a transaction
 * @sdp: the filesystem
 * @gl: the glock
 *
 * Add the given glock to this process's transaction
 */

void
ogfs_trans_add_gl(ogfs_sbd_t * sdp, ogfs_glock_t * gl)
{
	ogfs_trans_t *tr;

	ENTER(GFN_TRANS_ADD_GL);

	if (!gl->gl_new_tr) {
		OGFS_ASSERT(atomic_read(&gl->gl_locked) &&
			   gl->gl_pid == osi_pid() &&
			   gl->gl_state == LM_ST_EXCLUSIVE,);
		OGFS_ASSERT(list_empty(&gl->gl_new_list),);
		OGFS_ASSERT(list_empty(&gl->gl_new_bufs),);
		tr = get_transaction();
		OGFS_ASSERT(tr,);
		list_add(&gl->gl_new_list, &tr->tr_gl);
		gl->gl_new_tr = tr;
		ogfs_hold_glstruct(gl);	/*  Released in ogfs_trans_end()  */
		tr->tr_num_gl++;
		set_bit(GLF_DIRTY, &gl->gl_flags);
	}

	EXIT(GFN_TRANS_ADD_GL);

	return;
}

/**
 * ogfs_trans_add_bh - Add a buffer to the current transaction
 * @sdp: The OGFS superblock
 * @bh: The buffer to add
 *
 * Add a buffer to the current transaction.  The glock for the buffer
 * should be held.  This pins the buffer as well.
 *
 * Call this as many times as you want during transaction formation.
 * It only does its work once.
 */

void
ogfs_trans_add_bh(ogfs_sbd_t * sdp, struct buffer_head * bh)
{
	ogfs_bufdata_t *bd;
	ogfs_glock_t *gl;
	char *data;

	ENTER(GFN_TRANS_ADD_BH);

	ogfs_meta_check(bh);
	bd = BH_BUFDATA(bh);
	if (!list_empty(&bd->bd_new_list))
		goto out;
	gl = bd->bd_gl;
	if (!gl->gl_new_tr)
		ogfs_trans_add_gl(sdp, gl);
	list_add(&bd->bd_new_list, &gl->gl_new_bufs);
	gl->gl_new_tr->tr_num_buf++;
	ogfs_dpin(bh);
	/*  FIXME  */
	for (;;) {
		data = (char *) gmalloc(sdp->sd_sb.sb_bsize);
		if (data)
			break;

		con_printf("OGFS: Failed to allocate mem \n");
		osi_sleep_unint(1);
	}
	ogfs_lock_buffer(bh);
	OGFS_ASSERT(!bd->bd_frozen,);
	bd->bd_frozen = data;
	memcpy(bd->bd_frozen, BH_DATA(bh), sdp->sd_sb.sb_bsize);
	ogfs_unlock_buffer(bh);

out:
	EXIT(GFN_TRANS_ADD_BH);
}

/**
 * ogfs_trans_mark - Mark a flag in a transaction
 * @sdp: the filesystem
 * @flag: the flag to mark
 *
 */

void
ogfs_trans_mark(ogfs_sbd_t * sdp, unsigned int flag)
{
	ogfs_trans_t *tr;

	ENTER(GFN_TRANS_MARK);
	tr = get_transaction();
	OGFS_ASSERT(tr,);
	tr->tr_flags |= flag;

	EXIT(GFN_TRANS_MARK);
}

/**
 * ogfs_trans_nopen_change - Add a nopen change to the current transaction
 * @sdp: the filesystem
 * @type: the type of nopen change (%OGFS_LOG_DESC_NOP_OPEN or %OGFS_LOG_DESC_NOP_CLOSE)
 * @inum: the inode number
 *
 * Returns:  0 on success, -EXXX on failure
 */

int
ogfs_trans_nopen_change(ogfs_sbd_t * sdp, unsigned int type, ogfs_inum_t * inum)
{
	ogfs_trans_t *tr;
	ogfs_tr_nop_t *nop;
	int x, error = 0;

	ENTER(GFN_TRANS_NOPEN_CHANGE);

	tr = get_transaction();
	OGFS_ASSERT(tr,);
	for (x = 0; x < 2; x++) {
		nop = gmalloc(sizeof(ogfs_tr_nop_t));
		if (!nop) {
			error = -ENOMEM;
			goto out;
		}
		memset(nop, 0, sizeof(ogfs_tr_nop_t));
		nop->nop_type = type;
		nop->nop_inum = *inum;
		if (x)
			list_add(&nop->nop_list, &tr->tr_nop_change);
		else
			list_add(&nop->nop_list, &tr->tr_nop_change_ic);
	}

	switch (type) {
	case OGFS_LOG_DESC_NOPOPEN:
		tr->tr_num_nop_o++;
		break;
	case OGFS_LOG_DESC_NOPCLOSE:
		tr->tr_num_nop_c++;
		break;
	default:
		OGFS_ASSERT(0, con_printf("type = %u\n", type);
		    );
	}

	OGFS_ASSERT(inum->no_formal_ino && inum->no_addr,);

out:
	EXIT(GFN_TRANS_NOPEN_CHANGE);

	return error;
}
