/*
 *
 *    Copyright 2001 The OpenGFS Project.
 *    Portions Copyright 2000-2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 * 
 */

#define EXPORT_SYMTAB

#include <linux/module.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/kmod.h>
#include <linux/string.h>
#include <asm/semaphore.h>

#include "global.h"
#include "lm_interface.h"


MODULE_DESCRIPTION("Lock Module Harness");
MODULE_AUTHOR("The OpenGFS Project");
MODULE_LICENSE("GPL");

static LIST_HEAD(lm_list);
static DECLARE_RWSEM(lhm_mutex);


/**
 * lm_find_proto - Search for a low-level locking protocol
 * @name: name of the protocol
 *
 * Returns the lm_wrapper struct of the locking protocol or
 * NULL if none was found.
 */
static struct lm_lockops *
lm_find_proto(char *name)
{
	struct lm_lockops *proto;
	struct list_head *tmp;

	list_for_each(tmp, &lm_list) {
		proto = list_entry(tmp, struct lm_lockops, list);

		if (strcmp(proto->proto_name, name) == 0)
			return proto;
	}

	return NULL;
}

/**
 * lm_register_proto - Register a low-level locking protocol
 * @proto: the protocol definition
 *
 * Returns: 0 on success, -EEXIST if a protocol with the same
 * name is already registered.
 */
int
lm_register_proto(struct lm_lockops *proto)
{
	int error = 0;

	down_write(&lhm_mutex);
	if (lm_find_proto(proto->proto_name)) {
		printk(KERN_ERR "harness: protocol %s already registered\n",
				proto->proto_name);
		error = -EEXIST;
	} else
		list_add(&proto->list, &lm_list);
	up_write(&lhm_mutex);

	return error;
}

/**
 * lm_unregister_proto - Unregister a low-level locking protocol
 * @proto: the protocol definition
 */
void
lm_unregister_proto(struct lm_lockops *proto)
{
	struct lm_lockops *p;

	down_write(&lhm_mutex);
	p = lm_find_proto(proto->proto_name);
	if (p)
		list_del(&p->list);
	up_write(&lhm_mutex);
}

/**
 * lm_mount - Mount a lock protocol
 * @proto_name - the name of the protocol
 * @table_name - the name of the lock space
 * @host_data - data specific to this host
 * @cb - the callback to the code using the lock module
 * @fsdata - data to pass back with the callback
 * @lockstruct - a structure returned describing the mount
 *
 * Returns: 0 on success, -EXXX on failure
 */
int
lm_mount(char *proto, char *table, char *data, lm_callback_t cb,
		lm_fsdata_t *fsdata, struct lm_lockstruct *lockstruct)
{
	struct lm_lockops *p;
	int error = -ENOENT;

	down_read(&lhm_mutex);
	p = lm_find_proto(proto);
	if (p)
		goto found;
	up_read(&lhm_mutex);

	request_module(proto);

	down_read(&lhm_mutex);
	p = lm_find_proto(proto);
	if (p)
		goto found;

	printk(KERN_ERR "harness: can't find protocol %s\n", proto);
	goto out;

found:
	printk(KERN_INFO "harness: found protocol %s\n", proto);
	lockstruct->ls_ops = p;
	lockstruct->ls_lockspace =
	    p->mount(table, data, cb, fsdata, &lockstruct->ls_jid,
			    &lockstruct->ls_first);

	if (lockstruct->ls_lockspace)
		error = 0;
	else
		error = -EINVAL;

out:
	up_read(&lhm_mutex);
	return error;
}

EXPORT_SYMBOL(lm_register_proto);
EXPORT_SYMBOL(lm_unregister_proto);
EXPORT_SYMBOL(lm_mount);
