/*
 *
 *    Copyright 1999-2001 Sistina Software, Inc.
 *    Portions Copyright 2003 Brian Jackson
 *    Portions Copyright 2003 The OpenGFS Project
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 *    This file was maintained by:
 *      Kenneth W. Preslan <kpreslan@sistina.com>
 *
 */
#ifndef _LM_INTERFACE_H
#define _LM_INTERFACE_H

/*
 * This file defines the interface to the locking modules
 *
 * Sooner or later, I need to put all the documentation back into this file.
 * In the mean time, here are some notes.
 *
 *  - The lock module is now responsible for STOMITHing the an expired
 *    client before calling the callback with type LM_CB_EXPIRED.
 *
 *  - If mount() operation returns first == TRUE, OGFS will check all the
 *    journals.  OGFS itself can't/shouldn't stomith the machines, so the
 *    lock module needs to make sure that there are no zombie machines on
 *    any of the journals.  (i.e. this should probably be on the first mount
 *    of the lock space where all mounts by other machines are blocked.)
 *    OGFS will call others_may_mount() when the filesystem is in a consistent
 *    state.
 *
 *  - OGFS can issue multiple simultaneous get_lock()s for the same lockname.
 *    The lock module needs to deal with it, either by
 *     1)  building a hash table to lookup the structures and keeping a
 *	   reference count so there is only on lm_lock_t for a given
 *	   lockname.
 *     2)  just dealing with multiple lm_lock_t structures for a given
 *         lockname.
 *
 */

#include "global.h"

typedef void lm_lockspace_t;
typedef void lm_lock_t;
typedef void lm_fsdata_t;
typedef void (*lm_callback_t) (lm_fsdata_t * fsdata, unsigned int type,
			       void *data);

/*  Lock types 
    LM_TYPE_RESERVED, LM_TYPE_CIDBUF, and LM_TYPE_MOUNT aren't ever used by OGFS.  
    MSB is used for LVB's, so valid types can be a maximum of 0x7F  */

#define LM_TYPE_RESERVED       (0x00)
#define LM_TYPE_CIDBUF         (0x01)
#define LM_TYPE_MOUNT          (0x02)
#define LM_TYPE_NONDISK        (0x03)
#define LM_TYPE_INODE          (0x04)
#define LM_TYPE_RGRP           (0x05)
#define LM_TYPE_META           (0x06)
#define LM_TYPE_NOPEN          (0x07)
#define LM_TYPE_FLOCK          (0x08)
#define LM_TYPE_PLOCK          (0x09)
#define LM_TYPE_PLOCK_HEAD     (0x0A)
#define LM_TYPE_LVB_MASK       (0x80)
#define LM_TYPE_DEADMAN        (0x0B)

/*  States passed to lock()  */

#define LM_ST_UNLOCKED         (0)
#define LM_ST_EXCLUSIVE        (1)
#define LM_ST_DEFERRED         (2)
#define LM_ST_SHARED           (3)

/*  Flags passed to lock()  */

#define LM_FLAG_TRY            (0x00000100)
#define LM_FLAG_NOEXP          (0x00000200)

/*  Flags returned by lock()  */

#define LM_OUT_ST_MASK         (0x00000003)
#define LM_OUT_CANCELED        (0x00000004)
#define LM_OUT_NEED_E          (0x00000008)
#define LM_OUT_NEED_D          (0x00000010)
#define LM_OUT_NEED_S          (0x00000020)

/*  Callback types  */

#define LM_CB_NEED_E           (257)
#define LM_CB_NEED_D           (258)
#define LM_CB_NEED_S           (259)
#define LM_CB_EXPIRED          (260)
#define LM_CB_DROPLOCKS        (261)

/*  Lock Value Block flags  */

#define LM_LVB_PERM            (0x00000001)

struct lm_lockname {
	uint64 ln_number;
	unsigned int ln_type;
};
typedef struct lm_lockname lm_lockname_t;

#define lm_name_equal(name1, name2) (((name1)->ln_number == (name2)->ln_number) && ((name1)->ln_type == (name2)->ln_type))

struct lm_lvb {
	int lvb_flags;
	unsigned int lvb_length;
	char *lvb_data;
};
typedef struct lm_lvb lm_lvb_t;

struct lm_lockops {
	char proto_name[256];
	unsigned int local_fs;
	struct list_head list;

	/*  Mount/Unmount  */

	lm_lockspace_t *(*mount) (char *table_name, char *host_data,
				  lm_callback_t cb, lm_fsdata_t * fsdata,
				  unsigned int *jid, unsigned int *first);
	void (*others_may_mount) (lm_lockspace_t * lockspace);
	void (*unmount) (lm_lockspace_t * lockspace);

	/*  Lock oriented operations  */

	lm_lock_t *(*get_lock) (lm_lockspace_t * lockspace,
				lm_lockname_t * name);
	void (*put_lock) (lm_lock_t * lock);

	int (*lock) (lm_lock_t * lock, unsigned int cur_state,
		     unsigned int req_state, unsigned int flags);
	int (*unlock) (lm_lock_t * lock, unsigned int cur_state);

	int (*reset) (lm_lock_t * lock);
	void (*cancel) (lm_lock_t * lock, int on);

	lm_lvb_t *(*hold_lvb) (lm_lock_t * lock);
	void (*unhold_lvb) (lm_lock_t * lock, lm_lvb_t * lvb);
	void (*sync_lvb) (lm_lock_t * lock, lm_lvb_t * lvb);

	/*  Client oriented operations  */

	void (*reset_exp) (lm_lockspace_t * lockspace, unsigned int jid);
};
typedef struct lm_lockops lm_lockops_t;

struct lm_lockstruct {
	unsigned int ls_jid;
	unsigned int ls_first;
	lm_lockspace_t *ls_lockspace;
	lm_lockops_t *ls_ops;
};
typedef struct lm_lockstruct lm_lockstruct_t;

/*  Bottom interface  */

int lm_register_proto(lm_lockops_t * proto);
void lm_unregister_proto(lm_lockops_t * proto);

/*  Top interface  */

int lm_mount(char *proto_name,
	     char *table_name, char *host_data,
	     lm_callback_t cb, lm_fsdata_t * fsdata,
	     lm_lockstruct_t * lockstruct);

#endif /* _LM_INTERFACE_H */
