/*
 *
 *    Copyright 1999-2001 Sistina Software, Inc.
 *    Portions Copyright 2003 Brian Jackson
 *    Portions Copyright 2003 The OpenGFS Project
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 *    This file was maintained by:
 *      David C. Teigland <teigland@sistina.com>
 *
 */

#include "config.h"

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/in.h>
#include <asm/uaccess.h>

#include "memexp.h"
#include "cluster_info.h"

struct message {
	uint32 type;
	uint32 ln_type;		/* adjust according to lm_lockname_t defn */
	uint64 ln_number;
};
typedef struct message message_t;

/**
 * message_in
 *
 * @mess:
 * @buff:
 *
 */

STATIC void
message_in(message_t * mess, unsigned char *buff)
{
	message_t *m = (message_t *) buff;

	mess->type = be32_to_cpup(&m->type);
	mess->ln_type = be32_to_cpup(&m->ln_type);
	mess->ln_number = be64_to_cpup(&m->ln_number);
}

/**
 * message_out
 *
 * @mess:
 * @buff:
 *
 */

STATIC void
message_out(message_t * mess, unsigned char *buff)
{
	message_t *m = (message_t *) buff;

	m->type = cpu_to_be32p(&mess->type);
	m->ln_type = cpu_to_be32p(&mess->ln_type);
	m->ln_number = cpu_to_be64p(&mess->ln_number);
}

/**
 * open_cb_socket
 *
 * @md:
 *
 */

int
open_cb_socket(memexp_t * md)
{
	struct sockaddr_in myaddr;

	int error;
	uint32 ip;

	error = sock_create(AF_INET, SOCK_DGRAM, 0, &md->cb_recv_sock);
	if (error < 0) {
		printk("%s:  can't open callback receive socket\n", PROTO_NAME);
		return (error);
	}

	myaddr.sin_family = AF_INET;
	myaddr.sin_port = cpu_to_be16(md->cb_port);

	error = cid2ipaddr(md, md->cid, &ip);
	if (error) {
		sock_release(md->cb_recv_sock);
		printk("%s:  can't match cid to ipaddr\n", PROTO_NAME);
		return (error);
	}

	myaddr.sin_addr.s_addr = cpu_to_be32(ip);

	error =
	    md->cb_recv_sock->ops->bind(md->cb_recv_sock,
					(struct sockaddr *) &myaddr,
					sizeof(struct sockaddr_in));
	if (error < 0) {
		sock_release(md->cb_recv_sock);
		printk("%s:  can't bind to address\n", PROTO_NAME);
		return (error);
	}

	error = sock_create(AF_INET, SOCK_DGRAM, 0, &md->cb_send_sock);
	if (error < 0) {
		sock_release(md->cb_recv_sock);
		printk("%s:  can't open callback send socket\n", PROTO_NAME);
		return (error);
	}

	init_MUTEX(&md->cb_send_lock);

	return (0);
}

/**
 * close_cb_socket
 *
 * @md:
 *
 */

void
close_cb_socket(memexp_t * md)
{
	sock_release(md->cb_send_sock);
	sock_release(md->cb_recv_sock);
}

/**
 * send_mess
 *
 * @md:
 * @to:
 * @ln_number:
 * @ln_type:
 * @type:
 *
 */

void
send_mess(memexp_t * md, uint32 to, uint64 ln_number, uint32 ln_type,
	  uint32 type, int retry)
{
	message_t mess;

	struct sockaddr_in youraddr;
	struct msghdr msg;
	struct iovec iov;

	unsigned long proc_flags;
	mm_segment_t oldfs;
	sigset_t oldset;

	char buff[sizeof(message_t)];
	int error;

again:
	mess.type = type;
	mess.ln_type = ln_type;
	mess.ln_number = ln_number;

	message_out(&mess, buff);

	youraddr.sin_family = AF_INET;
	youraddr.sin_addr.s_addr = cpu_to_be32(to);
	youraddr.sin_port = cpu_to_be16(md->cb_port);

	iov.iov_base = buff;
	iov.iov_len = sizeof(message_t);

	msg.msg_name = (struct sockaddr *) &youraddr;
	msg.msg_namelen = sizeof(struct sockaddr_in);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags = 0;

	down(&md->cb_send_lock);

	oldfs = get_fs();
	set_fs(get_ds());

	spin_lock_irqsave(&current->sighand->siglock, proc_flags);
	oldset = current->blocked;
	sigfillset(&current->blocked);
	recalc_sigpending();
	spin_unlock_irqrestore(&current->sighand->siglock, proc_flags);

	error = sock_sendmsg(md->cb_send_sock, &msg, iov.iov_len);

	spin_lock_irqsave(&current->sighand->siglock, proc_flags);
	current->blocked = oldset;
	recalc_sigpending();
	spin_unlock_irqrestore(&current->sighand->siglock, proc_flags);

	set_fs(oldfs);

	up(&md->cb_send_lock);

	if (error < 0) {
		printk("%s:  error sending callback to 0x%.8X (%d)\n",
		       PROTO_NAME, to, error);
		if (retry) {
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(2*HZ);
			goto again;
		}
	}
}

/**
 * memexp_callback
 *
 * @data:
 *
 */

int
memexp_callback(void *data)
{
	memexp_t *md = (memexp_t *) data;
	message_t mess;

	struct msghdr msg;
	struct iovec iov;

	unsigned long flags;
	mm_segment_t oldfs;

	struct sockaddr_in youraddr;

	unsigned char buff[sizeof(message_t)];
	int n;
	lm_lockname_t lockname;

	MEP_ASSERT(md,);

	/*  Set up the thread  */

	lock_kernel();

	daemonize("memexp: memexp_callback");

	current->session = 1;
	//current->pgrp = 1;
#ifdef SKIP
	current->tty = NULL;
#endif
	current->tty_old_pgrp = 0;

	spin_lock_irqsave(&current->sighand->siglock, flags);
	sigfillset(&current->blocked);
	recalc_sigpending();
	spin_unlock_irqrestore(&current->sighand->siglock, flags);

	sprintf(current->comm, "memexp_callback");

	/*  Wake up the process who started us  */

	up(&md->cb_recvd_sem);

	while (1) {
		iov.iov_base = buff;
		iov.iov_len = sizeof(message_t);

		msg.msg_name = (struct sockaddr *) &youraddr;
		msg.msg_namelen = sizeof(struct sockaddr_in);
		msg.msg_iov = &iov;
		msg.msg_iovlen = 1;
		msg.msg_control = NULL;
		msg.msg_controllen = 0;
		msg.msg_flags = 0;

		oldfs = get_fs();
		set_fs(get_ds());

		n = sock_recvmsg(md->cb_recv_sock, &msg, iov.iov_len, 0);

		set_fs(oldfs);

		if (md->cb_recvd_stop)
			break;

		if (n == sizeof(message_t)) {
			message_in(&mess, buff);

			if (mess.type == LM_CB_NEED_S
			    || mess.type == LM_CB_NEED_E
			    || mess.type == LM_CB_NEED_D
			    || mess.type == LM_CB_DROPLOCKS) {
				lockname.ln_type = mess.ln_type;
				lockname.ln_number = mess.ln_number;
				md->cb(md->fsdata, mess.type,
				       (void *) &lockname);
			} else
				printk("%s:  unrecognized callback type (%u)\n",
				       PROTO_NAME, mess.type);
		} else if (n < 0)
			printk("%s:  error doing receive (%d)\n", PROTO_NAME,
			       n);
		else
			printk("%s:  bad message length (%d)\n", PROTO_NAME, n);
	}

	up(&md->cb_recvd_sem);

	unlock_kernel();

	return (0);
}
