/*
 *
 *    Copyright 1999-2001 Sistina Software, Inc.
 *    Portions Copyright 2003 Brian Jackson
 *    Portions Copyright 2003 The OpenGFS Project
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 *    This file was maintained by:
 *      David C. Teigland <teigland@sistina.com>
 *
 */

/*
 * cidbuf states:
 *
 *  0. UNUSED - not owned
 *  1. USED - not owned
 *  2. STOMITH - owned (exclusive)
 *  3. STOMITH_DONE - not owned
 *  4. EXPIRED_MARKING - owned (exclusive)
 *  5. EXPIRED - not owned
 *  6. RESETTING - owned (exclusive)
 *
 * States 0 and 1 are the normal states of cidbufs for cid's which are
 * currently used or not used.  States 2 to 6 are recovery states.  Gaining
 * ownership of an exclusive recovery state (2, 4 or 6) is simply done by
 * being the first to store your cid in the cidbuf's owner (cleaner) field.
 * This owner field is set to OWNER_NONE (-1) in unowned (non-exclusive)
 * states.
 *
 * If you notice the cidbuf of a machine A is USED, but has not been
 * heartbeated in the required interval, you attempt to gain ownership
 * of cidbuf A in the STOMITH state (2).
 *
 * If you own machine A's cidbuf in the STOMITH state you do two things:
 * First you stomith A.  Second, you reset to OWNER_NONE any other cidbufs
 * for which A is the owner.  Step two allows machines which A may have
 * been recovering when it died to be recovered by someone else.
 * When you're done with these two steps you move cidbuf A into the
 * STOMITH_DONE state (3) with OWNER_NONE.  When any machine sees cidbuf A in
 * state 3 it will attempt to gain ownership of cidbuf A for recovery state 4.
 *
 * If you own machine A's cidbuf in the EXPIRED_MARKING state (4), you call
 * lsm_mark_expired().  When this is finished you move cidbuf A into state 5.
 *
 * Any machine which sees cidbuf A in state 5 will report it to OGFS.  OGFS
 * guarantee's that only one machine will do journal replay in this state.
 *
 * When OGFS is finished with journal replay, it calls reset_expired which
 * moves cidbuf A into the RESETTING state (6).  Since one machine had
 * exlusive access to do journal recovery, only that machine will call
 * reset_expired, so moving cidbuf A into state 6 will be attempted by
 * one machine only.  After moving cidbuf A into the RESETTING state, the
 * machine calls lsm_reset_expired().  When this is finished, cidbuf A is
 * moved into the UNUSED state (0).  (If a machine is recovering
 * its own journal, this last transition is into the USED state (1).)
 *
 * If a machine B notices a cidbuf C in any one of the exclusive recovery
 * states with an owner of OWNER_NONE, it means someone was doing recovery of
 * C but died before finishing.  In this case, B will try to gain ownership
 * of cidbuf C in its current state.  If B succeeds, it will start the
 * recovery process at the point appropriate for the current state.
 *
 */

#include "config.h"

#include <linux/module.h>
#include <linux/slab.h>

#include "memexp.h"
#include "ll_memexp.h"
#include "cluster_info.h"
#include "lock_state.h"
#include "stomith_ops.h"

struct client_data {
	cidbuf_t *clients;
	uint32 bufsize;
};
typedef struct client_data client_data_t;

/**
 * cidbuf_in - byte swap in place
 * @buf:
 *
 */

cidbuf_t *
cidbuf_in(unsigned char *buf)
{
	cidbuf_t *cb = (cidbuf_t *) buf;

	cb->counter = be64_to_cpu(cb->counter);
	cb->cleaner = be16_to_cpu(cb->cleaner);

	return cb;
}

/**
 * cidbuf_out - byte swap in place
 * @buf:
 *
 */

unsigned char *
cidbuf_out(unsigned char *buf)
{
	cidbuf_t *cb = (cidbuf_t *) buf;

	cb->counter = cpu_to_be64(cb->counter);
	cb->cleaner = cpu_to_be16(cb->cleaner);

	return buf;
}

/**
 * get_cid_state
 * @mep: the lock module's "superblock"
 * @cid: the client of interest
 * @state: output current state (inuse)
 * @owner: output current owner (cleaner)
 *
 * This is always called right after rereading client state.
 *
 */

STATIC int
get_cid_state(memexp_t * mep, uint16 cid, unsigned char *state, uint32 * owner)
{
	client_data_t *cd = (client_data_t *) mep->client_data;

	*state = cd->clients[cid].inuse;
	*owner = cd->clients[cid].cleaner;

	return 0;
}

/**
 * set_cid_state
 * @mep: the lock module's "superblock"
 * @cid: the client of interest
 * @state: set state to this (inuse)
 * @owner: set owner to this (cleaner)
 *
 * This is called after rereading client state.  In some state transitions
 * the state/owner could change between then and the first load here -
 * those are the races to claim an exclusive recovery state.  In other state
 * transitions the state/owner should never change between then and the
 * first load because the present state is exclusive - in those cases this
 * routine should never fail.
 *
 */

STATIC int
set_cid_state(memexp_t * mep, uint16 cid, unsigned char state, uint32 owner)
{
	client_data_t *cd = (client_data_t *) mep->client_data;
	ll_memexp_ops_t *dmep = mep->ll_memexp_ops;
	lm_lockname_t lockname;
	cidbuf_t *cidbuf;
	char buf[cd->bufsize];
	buf_load_t bld;
	buf_store_t bst;
	int error;

	lockname.ln_number = cid;
	lockname.ln_type = LM_TYPE_CIDBUF;

	error = dmep->load(mep, &lockname, &bld, buf, cd->bufsize);
	if (error)
		return error;

	MEP_ASSERT(bld.inuse, printk("cidbuf %u not inuse\n", cid);
	    );

	cidbuf = cidbuf_in(buf);

	if (cidbuf->inuse != cd->clients[cid].inuse ||
	    cidbuf->cleaner != cd->clients[cid].cleaner)
		return -1;

	cidbuf->inuse = state;
	cidbuf->cleaner = owner;

	cidbuf_out(buf);
	prepare_store(&bld, &bst);
	error = dmep->store(mep, &lockname, &bst, buf, cd->bufsize);
	if (error)
		return -1;

	cd->clients[cid].inuse = state;
	cd->clients[cid].cleaner = owner;

	return 0;
}

/**
 * reset_owner
 * @mep: the lock module's "superblock"
 * @cid: the client of interest
 *
 * Set owner (cleaner) to OWNER_NONE on cidbufs for which the owner
 * died and we acquired the first exclusive recovery state.
 *
 */

STATIC int
reset_owner(memexp_t * mep, uint16 cid)
{
	client_data_t *cd = (client_data_t *) mep->client_data;
	int error, i;

	for (i = 0; i < MAX_CLIENTS; i++) {
		if (cd->clients[i].cleaner == cid) {
			error =
			    set_cid_state(mep, i, cd->clients[i].inuse,
					  OWNER_NONE);

			MEP_ASSERT(!error,
				   printk("set_cid_state: %d, %u\n", i, cid);
			    );
		}
	}

	return 0;
}

/**
 * stomith_cid
 * @mep: the lock module's "superblock"
 * @excid: the expired client of interest
 *
 */

int
stomith_cid(memexp_t * mep, uint16 excid)
{
	int error, i;
	uint32 jid, nstomith = 0, ipaddr = 0;
	struct cluster_stomith victim, local;

	MEP_ASSERT(excid != mep->cid, printk("bug in memexp code\n");
	    );

	error = cid2jid(mep, excid, &jid);
	if (error) {
		printk("%s:  stomith_cid bad cid2jid map (%u)\n", PROTO_NAME,
		       excid);
		return -1;
	}

	error = cid2ipaddr(mep, excid, &ipaddr);
	if (error) {
		printk("%s:  stomith_cid bad cid2ipaddr map (%u)\n", PROTO_NAME,
		       excid);
		return -1;
	}

	mep->ll_memexp_ops->stomith(mep, ipaddr);

	jid2nstomith(mep, jid, &nstomith);

	for (i = 0; i < nstomith; i++) {
		error = get_stomith_desc(mep, jid, &victim, i);
		if (error)
			continue;

		get_stomith_desc(mep, mep->jid, &local, i);

		error = do_stomith(&victim, &local);
		if (!error)
			break;
	}

	/* This sleep is a quiet period to allow any outstanding I/O from the
	   stomithed machine to reach any devices.  Eight seconds is probably
	   longer than necessary. */
	set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout(8*HZ);

	return error;
}

/**
 * recover_expired_client
 * @mep: the lock module's "superblock"
 * @excid: the expired client of interest
 *
 * This checks the current status of the cidbuf's state and
 * owner to determine if it should try to pick up recovery
 * at some point.  The approach right now is for a machine
 * to finish one stage of exclusive recovery and then put
 * the cidbuf back into an unowned state so anyone could pick
 * pick up the next exclusive recovery stage.
 *
 * Returns:  1 if this routine could be called again immediately to
 *           continue with the next recovery step.
 *           0 if there is no more recovery to do
 *
 */

STATIC int
recover_expired_client(memexp_t * mep, uint16 excid)
{
	unsigned char cidstate;
	uint32 owner;
	uint32 jid;
	int error;

	/* there's a chance excid is ourself at mount time and we want to ignore it
	   if it happens */

	if (excid == mep->cid)
		return 0;

	get_cid_state(mep, excid, &cidstate, &owner);

	MEP_ASSERT(cidstate != CID_NOT_USED,
		   printk("expired and unused cidbuf %u\n", excid);
	    );

	/* case1: a live machine just died; case2: whoever was
	   stomithing excid died before finishing the job */

	if (cidstate == CID_USED ||
	    (cidstate == CID_STOMITH && owner == OWNER_NONE)) {
		error = set_cid_state(mep, excid, CID_STOMITH, mep->cid);
		if (error)
			return 0;

		printk("%s:  client recovery stomithing cid %u\n", PROTO_NAME,
		       excid);

		error = stomith_cid(mep, excid);
		if (error) {
			printk("%s:  client_state stomith failed\n",
			       PROTO_NAME);
			error =
			    set_cid_state(mep, excid, CID_STOMITH, OWNER_NONE);

			MEP_ASSERT(!error,
				   printk("set_cid_state cid %u\n", excid);
			    );
			return 0;
		}

		reset_owner(mep, excid);

		error = set_cid_state(mep, excid, CID_STOMITH_DONE, OWNER_NONE);

		MEP_ASSERT(!error, printk("set_cid_state cid %u\n", excid);
		    );
		return 1;
	}

	/* case1: someone finished stomithing excid and moved state to
	   STOMITH_DONE; case2: whoever was marking expired on excid
	   died before finishing */

	else if (cidstate == CID_STOMITH_DONE ||
		 (cidstate == CID_EXPIRED_MARKING && owner == OWNER_NONE)) {
		error =
		    set_cid_state(mep, excid, CID_EXPIRED_MARKING, mep->cid);
		if (error)
			return 0;

		printk("%s:  client recovery marking expired cid %u\n",
		       PROTO_NAME, excid);

		error = lsm_mark_expired(mep, excid);

		MEP_ASSERT(!error,
			   printk("mark_expired cid %u (%d)\n", excid, error);
		    );

		/* move state of dead machine's cid to CID_EXPIRED (with no owner).
		   after this everyone will start sending OGFS callbacks and OGFS
		   will arbitrate who does the journal replay (by acquiring journal lock) */

		error = set_cid_state(mep, excid, CID_EXPIRED, OWNER_NONE);

		MEP_ASSERT(!error, printk("set_cid_state cid %u\n", excid);
		    );
		return 1;
	}

	/* someone finished mark expired and moved state to EXPIRED,
	   now everyone will see this and send an OGFS callback */

	else if (cidstate == CID_EXPIRED) {
		error = cid2jid(mep, excid, &jid);
		if (error) {
			printk("%s:  bad cid2jid map (%u)\n", PROTO_NAME,
			       excid);
			return 0;
		}

		mep->cb(mep->fsdata, LM_CB_EXPIRED, (void *) &jid);

		return 0;
	}

	/* whoever was resetting excid's expired locks died before finishing,
	   we'll try and finish */

	else if (cidstate == CID_RESETING && owner == OWNER_NONE) {
		error = set_cid_state(mep, excid, CID_RESETING, mep->cid);
		if (error)
			return 0;

		printk("%s:  client recovery resetting expired cid %u\n",
		       PROTO_NAME, excid);

		lsm_reset_expired(mep, excid);

		/* move state of dead machine's cid to UNUSED, or if it's our own
		   to USED */

		if (excid == mep->cid) {
			error = set_cid_state(mep, excid, CID_USED, OWNER_NONE);

			MEP_ASSERT(!error,
				   printk("set_cid_state cid %u\n", excid);
			    );
		} else {
			error =
			    set_cid_state(mep, excid, CID_NOT_USED, OWNER_NONE);

			MEP_ASSERT(!error,
				   printk("set_cid_state cid %u\n", excid);
			    );
		}

		return 0;
	}

	else {
		/* Someone else is doing the recovery.  Cases include:
		   cidstate == CID_STOMITH && owner is valid
		   cidstate == CID_EXPIRED_MARKING && owner is valid
		   cidstate == CID_RESETING && owner is valid */
	}

	return 0;
}

/**
 * reread_client_state
 * @mep: the lock module's "superblock"
 *
 * This updates the cd->clients information from the cidbufs.
 * Along the lines of what used to be Report Expired, except this
 * doesn't mess with the cidbuf states at all.
 */

int
reread_client_state(memexp_t * mep)
{
	client_data_t *cd = (client_data_t *) mep->client_data;
	ll_memexp_ops_t *dmep = mep->ll_memexp_ops;
	lm_lockname_t lockname;
	unsigned char buf[cd->bufsize];
	buf_load_t bld;
	cidbuf_t *cidbuf;
	int i, error;

	for (i = 0; i < MAX_CLIENTS; i++) {
		lockname.ln_number = i;
		lockname.ln_type = LM_TYPE_CIDBUF;

		error = dmep->load(mep, &lockname, &bld, buf, cd->bufsize);
		if (error)
			return error;

		if (!bld.inuse) {
			cd->clients[i].inuse = CID_NOT_USED;
			cd->clients[i].counter = 0;
			cd->clients[i].cleaner = OWNER_NONE;
			continue;
		}

		cidbuf = cidbuf_in(buf);

		if (cidbuf->counter == cd->clients[i].counter)
			cd->clients[i].dead = TRUE;
		else
			cd->clients[i].dead = FALSE;

		cd->clients[i].counter = cidbuf->counter;
		cd->clients[i].cleaner = cidbuf->cleaner;
		cd->clients[i].inuse = cidbuf->inuse;
	}

	if (cd->clients[mep->cid].dead) {
		MEP_ASSERT(0,
			   printk("counter: %" FMT_64
				  "u\nlast hb: %u\nnow: %u\n" "timeout: %u\n",
				  cd->clients[mep->cid].counter,
				  (unsigned int)mep->heartbeat_time,
				  (unsigned int)jiffies , mep->node_timeout);
		    );
	}

	return 0;
}

/**
 * check_expired_clients
 * @mep: the lock module's "superblock"
 *
 * Called after reread_client_state() from node_monitor daemon to see if
 * any clients need to be recovered.
 */

int
check_expired_clients(memexp_t * mep)
{
	client_data_t *cd = (client_data_t *) mep->client_data;
	int i;

	for (i = 0; i < MAX_CLIENTS; i++) {
		/* someone just quit heartbeating their cidbuf */

		if (cd->clients[i].inuse == CID_USED && cd->clients[i].dead)
			while (recover_expired_client(mep, i)) ;

		/* this machine is in a recovery state, but if the recoverer died,
		   the owner will be none and we need to try and pick up the job */

		else if (cd->clients[i].inuse != CID_USED &&
			 cd->clients[i].inuse != CID_NOT_USED) {
			/* We can short circuit the process here if we know we're the one doing
			   the cleaning.  I think I'm alive... I think, therefore I am... ok  */

			if (cd->clients[i].cleaner != mep->cid)
				while (recover_expired_client(mep, i)) ;
		}
	}

	return 0;
}

/**
 * reset_expired_client
 * @mep: the lock module's "superblock"
 * @cid: the client to reset
 *
 * We were called here by OGFS which had an exclusive journal recovery lock.
 * Because there should be only one client which gets called here, there should
 * be no race to get the exclusive RESETING recovery state.
 */

STATIC int
reset_expired_client(memexp_t * mep, uint16 excid)
{
	client_data_t *cd = (client_data_t *) mep->client_data;
	int error;

	if (cd->clients[excid].inuse != CID_EXPIRED)
		return 0;

	printk("%s:  client recovery reset expired cid %u (called)\n",
	       PROTO_NAME, excid);

      retry:

	error = set_cid_state(mep, excid, CID_RESETING, mep->cid);
	if (error) {
		/* this may fail due to conflict with our own refresh thread */
		printk("%s:  set_cid_state failed cid %u, RESETING, %u\n",
		       PROTO_NAME, excid, mep->cid);

		if (excid == mep->cid)
			goto retry;

		MEP_ASSERT(0,);
	}

	lsm_reset_expired(mep, excid);

	if (excid == mep->cid) {
		/* this may fail due to conflict with our own refresh thread */
		while (set_cid_state(mep, excid, CID_USED, OWNER_NONE))
			printk
			    ("%s:  set_cid_state failed cid %u, USED, OWNER_NONE\n",
			     PROTO_NAME, excid);
	} else {
		error = set_cid_state(mep, excid, CID_NOT_USED, OWNER_NONE);

		MEP_ASSERT(!error, printk("set_cid_state cid %u\n", excid);
		    );
	}

	return 0;
}

/**
 * client_refresh
 * @mep: the lock module's "superblock"
 *
 * Heartbeat my own cidbuf.
 */

void
client_refresh(memexp_t * mep)
{
	client_data_t *cd = (client_data_t *) mep->client_data;
	ll_memexp_ops_t *dmep = mep->ll_memexp_ops;
	lm_lockname_t lockname;
	unsigned char buf[cd->bufsize];
	buf_load_t bld;
	buf_store_t bst;
	cidbuf_t *cidbuf;
	int error;

	lockname.ln_number = mep->cid;
	lockname.ln_type = LM_TYPE_CIDBUF;

      retry:

	error = dmep->load(mep, &lockname, &bld, buf, cd->bufsize);
	if (error)
		return;

	MEP_ASSERT(bld.inuse, printk("cidbuf %u buffer not inuse\n", mep->cid);
	    );

	/* when recovering ourself, we need to be able to heartbeat our own
	   expired cidbuf */

	cidbuf = cidbuf_in(buf);

	if (cidbuf->inuse != CID_USED && cidbuf->inuse != CID_EXPIRED &&
	    cidbuf->inuse != CID_RESETING)
		MEP_ASSERT(0,
			   printk("cidbuf %u with inuse %u\n", mep->cid,
				  cidbuf->inuse);
	    );

	cidbuf->counter++;
	cidbuf_out(buf);
	prepare_store(&bld, &bst);
	error = dmep->store(mep, &lockname, &bst, buf, cd->bufsize);
	if (error)
		goto retry;

	if ((jiffies - mep->heartbeat_time) >= (mep->node_timeout*HZ) &&
			mep->heartbeat_time) {
		printk("%s:  Warning - heartbeat delay greater than timeout\n",
		       PROTO_NAME);
		printk("  last: %u, now: %u, diff: %u, timeout: %u\n",
		       (unsigned int)mep->heartbeat_time,
		       (unsigned int)jiffies,
		       (unsigned int)(jiffies - mep->heartbeat_time),
		       mep->node_timeout);
	}

	mep->heartbeat_time = jiffies;
}

/**
 * client_state_init
 * @mep: the lock module's "superblock"
 *
 * This must be called after lsm_init().
 */

int
client_state_init(memexp_t * mep)
{
	lsm_t *lsm = (lsm_t *) mep->lsm_data;
	ll_memexp_ops_t *dmep = mep->ll_memexp_ops;
	client_data_t *cd;
	lm_lockname_t lockname;
	unsigned char buf[lsm->bufsize];
	buf_load_t bld;
	buf_store_t bst;
	cidbuf_t *cidbuf;
	int error;

	/*
	 * Initial values of the clients are set on the
	 * first reread_client_state.
	 */
	cd = kmalloc(sizeof(client_data_t), GFP_NOFS);
	if (!cd) {
		printk("%s:  out of memory\n", PROTO_NAME);
		return -ENOMEM;
	}

	cd->clients = kmalloc(MAX_CLIENTS * sizeof(cidbuf_t), GFP_NOFS);
	if (!cd->clients) {
		kfree(cd);
		printk("%s:  out of memory\n", PROTO_NAME);
		return -ENOMEM;
	}

	memset(cd->clients, 0, MAX_CLIENTS * sizeof(cidbuf_t));
	cd->bufsize = lsm->bufsize;
	mep->client_data = cd;

	lockname.ln_number = mep->cid;
	lockname.ln_type = LM_TYPE_CIDBUF;

	error = dmep->load(mep, &lockname, &bld, buf, cd->bufsize);
	if (error)
		goto fail;

	cidbuf = cidbuf_in(buf);

	/* the buffer itself may not be inuse if this is a new mount.  If it's
	   a remount, the buffer may be in use, but the cidbuf->inuse should
	   be UNUSED or EXPIRED.  It will be EXPIRED if we did a reset_cluster
	   in check_mount_state, ie we're recovering ourself */

	if (!bld.inuse || cidbuf->inuse == CID_NOT_USED) {
		if (!bld.inuse)
			cidbuf->counter = 1;
		else
			cidbuf->counter++;

		cidbuf->inuse = CID_USED;
		cidbuf->cleaner = OWNER_NONE;
		cidbuf_out(buf);
		prepare_store(&bld, &bst);
		error = dmep->store(mep, &lockname, &bst, buf, cd->bufsize);
		if (error) {
			printk("%s:  first store on cidbuf failed\n",
			       PROTO_NAME);
			goto fail;
		}
	} else if (cidbuf->inuse == CID_EXPIRED) {
		cidbuf->counter++;
		cidbuf->cleaner = OWNER_NONE;
		cidbuf_out(buf);
		prepare_store(&bld, &bst);
		error = dmep->store(mep, &lockname, &bst, buf, cd->bufsize);
		if (error) {
			printk("%s:  first store on cidbuf failed (expired)\n",
			       PROTO_NAME);
			goto fail;
		}
	} else {
		printk("%s:  client_state_init illegal cidbuf state (%u)\n",
		       PROTO_NAME, cidbuf->inuse);
		goto fail;
	}

	return 0;

fail:
	kfree(cd->clients);
	kfree(cd);
	mep->client_data = NULL;
	return -1;
}

/**
 * client_state_relse
 * @mep: the lock module's "superblock"
 *
 * This must be called before lsm_shutdown().  Release my cidbuf by
 * keeping it in use and setting the cidbuf->inuse to CID_NOT_USED.
 * Deallocating the cidbuf buffer causes problems with fast mount/umount/mount.
 *
 * This function should return an error code.
 */

void
client_state_relse(memexp_t * mep)
{
	client_data_t *cd = (client_data_t *) mep->client_data;
	ll_memexp_ops_t *dmep = mep->ll_memexp_ops;
	lm_lockname_t lockname;
	unsigned char buf[cd->bufsize];
	buf_load_t bld;
	buf_store_t bst;
	cidbuf_t *cidbuf;
	int error;

	lockname.ln_number = mep->cid;
	lockname.ln_type = LM_TYPE_CIDBUF;

      retry:

	error = dmep->load(mep, &lockname, &bld, buf, cd->bufsize);
	if (error)
		return;

	MEP_ASSERT(bld.inuse, printk("cidbuf %u is not inuse\n", mep->cid);
	    );

	cidbuf = cidbuf_in(buf);
	cidbuf->inuse = CID_NOT_USED;
	cidbuf_out(buf);
	prepare_store(&bld, &bst);
	error = dmep->store(mep, &lockname, &bst, buf, cd->bufsize);
	if (error)
		goto retry;

	kfree(cd->clients);
	kfree(cd);
	mep->client_data = NULL;
}

/**
 * client_alive
 * @mep: the lock module's "superblock"
 * @cid:
 *
 */

int
client_alive(memexp_t * mep, uint16 cid)
{
	client_data_t *cd = (client_data_t *) mep->client_data;

	if (cid >= MAX_CLIENTS) {
		printk("%s:  client_alive invalid cid %u\n", PROTO_NAME, cid);
		return FALSE;
	}

	if (cd->clients[cid].inuse == CID_USED && !cd->clients[cid].dead)
		return TRUE;

	return FALSE;
}

/**
 * memexp_reset_exp
 * @lockspace:
 * @jid:
 *
 */

void
memexp_reset_exp(lm_lockspace_t * lockspace, unsigned int jid)
{
	memexp_t *mep = (memexp_t *) lockspace;
	uint16 cid = 0;

	jid2cid(mep, jid, &cid);

	reset_expired_client(mep, cid);
}

/* vim:sw=8 ts=8
 */
