/*
 *
 *    Copyright 1999-2001 Sistina Software, Inc.
 *    Portions Copyright 2003 Brian Jackson
 *    Portions Copyright 2003 The OpenGFS Project
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 *    This file was maintained by:
 *      David C. Teigland <teigland@sistina.com>
 *
 */

#include "config.h"

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/buffer_head.h>
#include <linux/slab.h>

#include "memexp.h"
#include "cluster_config.h"
#include "util.h"
#include "stomith.h"

/* Cache cluster info in these structs */

struct node_info {
	uint16 jid;
	uint16 cid;
	uint32 ipaddr;
	uint32 nstomith;
	struct node_info *next;
};
typedef struct node_info node_info_t;

struct node_group {
	uint32 count;
	node_info_t *list;
};
typedef struct node_group node_group_t;

/**
 * init_cluster_info
 *
 * @md: the lock module's "superblock"
 * @ip: local IP address
 *
 */

int
init_cluster_info(memexp_t * md, uint32 ip)
{
	struct buffer_head *bh;
	cluster_global_t cg;
	cluster_node_t cn;
	node_group_t *group = NULL;
	node_info_t *ni, *ni2;
	int block = 0, found = 0;
	//dev_t dev;
	struct block_device *bdev;

	//dev = md->cfile->f_dentry->d_inode->i_rdev;
	bdev = md->cfile->f_dentry->d_inode->i_bdev;
	set_blocksize(bdev, CIDEV_BLOCKSIZE);

	bh = __bread(bdev, 0, CIDEV_BLOCKSIZE);
	if (!bh) {
		printk("%s:  can't read Cluster Information Device\n",
		       PROTO_NAME);
		return (-1);
	}

	cluster_global_in(&cg, bh->b_data);

	if (cg.cg_magic != OGFS_CLUSTER_GLOBAL) {
		//mark_buffer_uptodate(bh, FALSE);
		clear_bit(BH_Uptodate, &bh->b_state);
		brelse(bh);
		return (-1);
	}

	if ((((cg.cg_version >> 16) & 0xff) !=
	     ((OGFS_CLUSTER_VERSION >> 16) & 0xff))
	    || (((cg.cg_version >> 8) & 0xff) !=
		((OGFS_CLUSTER_VERSION >> 8) & 0xff))) {
		//mark_buffer_uptodate(bh, FALSE);
		clear_bit(BH_Uptodate, &bh->b_state);
		brelse(bh);
		printk
		    ("%s:  version mismatch: cidev (%u, %u), memexp (%u, %u)\n",
		     PROTO_NAME, (cg.cg_version >> 16) & 0xff,
		     (cg.cg_version >> 8) & 0xff,
		     (OGFS_CLUSTER_VERSION >> 16) & 0xff,
		     (OGFS_CLUSTER_VERSION >> 8) & 0xff);
		return (-1);
	}

	md->cb_port = (unsigned short) cg.cg_cbport;
	md->node_timeout = cg.cg_node_timeout;

	memcpy(md->lockdev_name, cg.cg_lockdev, OGFS_LOCKNAME_LEN);

	//mark_buffer_uptodate(bh, FALSE);
	clear_bit(BH_Uptodate, &bh->b_state);
	brelse(bh);

	block++;

	group = kmalloc(sizeof(node_group_t), GFP_NOFS);
	if (!group)
		return (-ENOMEM);

	memset(group, 0, sizeof(node_group_t));

	/*  Read cluster info  */

	while (1) {
		bh = __bread(bdev, block, CIDEV_BLOCKSIZE);

		if (!bh) {
			printk
			    ("%s:  error reading Cluster Information Device.\n",
			     PROTO_NAME);
			goto fail;
		}

		cluster_node_in(&cn, bh->b_data);

		if (cn.cn_magic == OGFS_CLUSTER_GLOBAL) {	/* end of the configured nodes */
			//mark_buffer_uptodate(bh, FALSE);
			clear_bit(BH_Uptodate, &bh->b_state);
			brelse(bh);
			break;
		}

		if (cn.cn_magic != OGFS_CLUSTER_NODE)	/* a node was removed from the config */
			goto next;

		if (cn.cn_ipaddr == ip) {	/* our own config block */
			md->jid = block - 1;
			md->cid = cn.cn_cid;
			found = TRUE;
		}

		ni = kmalloc(sizeof(node_info_t), GFP_NOFS);
		if (!ni)
			goto fail;

		ni->jid = block - 1;
		ni->cid = cn.cn_cid;
		ni->ipaddr = cn.cn_ipaddr;
		ni->nstomith = cn.cn_num_stomith;

		if (group->list) {
			ni->next = group->list;
			group->list = ni;
		} else {
			group->list = ni;
			ni->next = NULL;
		}

	      next:
		//mark_buffer_uptodate(bh, FALSE);
		clear_bit(BH_Uptodate, &bh->b_state);
		brelse(bh);
		block++;
	}

	if (!found) {
		char ipstr[16];
		inet_ntoa(ip, ipstr);
		printk("%s:  no cluster info for %s\n", PROTO_NAME, ipstr);
		goto fail_2;
	}

	group->count = block - 1;
	md->cluster_data = group;

	return (0);

      fail:

	if (bh) {
		//mark_buffer_uptodate(bh, FALSE);
		clear_bit(BH_Uptodate, &bh->b_state);
		brelse(bh);
	}

      fail_2:

	ni = group->list;

	while (ni) {
		ni2 = ni;
		ni = ni->next;
		kfree(ni2);
	}

	kfree(group);

	return (-1);
}

/**
 * release_cluster_info
 *
 * @md: the lock module's "superblock"
 *
 */

void
release_cluster_info(memexp_t * md)
{
	node_group_t *group = (node_group_t *) md->cluster_data;
	node_info_t *ni, *ni2;

	MEP_ASSERT(group,);

	ni = group->list;

	while (ni) {
		ni2 = ni;
		ni = ni->next;
		kfree(ni2);
	}

	kfree(group);

	md->cluster_data = NULL;
}

/**
 * reread_cluster_info
 *
 * @md: the lock module's "superblock"
 *
 */

int
reread_cluster_info(memexp_t * md)
{
	struct buffer_head *bh;
	cluster_node_t cn;
	node_group_t *group = (node_group_t *) md->cluster_data;
	node_info_t *ni;
	int block = 1;
	struct block_device *bdev;

	bdev = md->cfile->f_dentry->d_inode->i_bdev;
	while (1) {
		bh = __bread(bdev, block, CIDEV_BLOCKSIZE);
		if (!bh) {
			printk
			    ("%s:  error reading Cluster Information Device.\n",
			     PROTO_NAME);
			goto fail;
		}

		cluster_node_in(&cn, bh->b_data);

		/* end of the configured nodes */

		if (cn.cn_magic == OGFS_CLUSTER_GLOBAL) {
			//mark_buffer_uptodate(bh, FALSE);
			clear_bit(BH_Uptodate, &bh->b_state);
			brelse(bh);
			break;
		}

		/* a node was removed from the config */

		if (cn.cn_magic != OGFS_CLUSTER_NODE)
			goto next;

		/* skip whatever's up to date */

		ni = group->list;

		while (ni) {
			if (ni->cid == cn.cn_cid) {
				MEP_ASSERT(ni->jid == (block - 1),);
				goto next;
			}

			ni = ni->next;
		}

		/* new CID */

		MEP_ASSERT(!ni,);

		/* if an old jid is being reused by a new cid, update values */

		ni = group->list;

		while (ni) {
			if (ni->jid == (block - 1)) {
				ni->cid = cn.cn_cid;
				ni->ipaddr = cn.cn_ipaddr;
				ni->nstomith = cn.cn_num_stomith;

				printk
				    ("%s:  discovered new client reusing jid %u, cid %u, ipaddr %u, nstomith %u\n",
				     PROTO_NAME, ni->jid, ni->cid, ni->ipaddr,
				     ni->nstomith);

				goto next;
			}

			ni = ni->next;
		}

		/* new jid, make a new ni entry */
		ni = kmalloc(sizeof(node_info_t), GFP_NOFS);
		if (!ni)
			goto fail;

		ni->jid = block - 1;
		ni->cid = cn.cn_cid;
		ni->ipaddr = cn.cn_ipaddr;
		ni->nstomith = cn.cn_num_stomith;

		printk
		    ("%s:  discovered new client.  new jid %u,  cid %u,  ipaddr %u,  nstomith %u\n",
		     PROTO_NAME, ni->jid, ni->cid, ni->ipaddr, ni->nstomith);

		group->count++;

		if (group->list) {
			ni->next = group->list;
			group->list = ni;
		} else {
			group->list = ni;
			ni->next = NULL;
		}

	      next:
		//mark_buffer_uptodate(bh, FALSE);
		clear_bit(BH_Uptodate, &bh->b_state);
		brelse(bh);
		block++;
	}

	return (0);

      fail:

	if (bh) {
		//mark_buffer_uptodate(bh, FALSE);
		clear_bit(BH_Uptodate, &bh->b_state);
		brelse(bh);
	}

	return (-1);
}

/**
 * cid2jid
 *
 * @md: the lock module's "superblock"
 * @cid: input
 * @jid: output
 *
 */

int
cid2jid(memexp_t * md, uint16 cid, uint32 * jid)
{
	node_group_t *group = (node_group_t *) md->cluster_data;
	node_info_t *ni;
	int retried = 0;

	MEP_ASSERT(group,);

      retry:

	ni = group->list;

	while (ni) {
		if (ni->cid == cid) {
			*jid = ni->jid;
			return (0);
		}

		ni = ni->next;
	}

	if (retried)
		return (-1);

	else {
		reread_cluster_info(md);
		retried = TRUE;
		goto retry;
	}
}

/**
 * cid2ipaddr
 *
 * @md: the lock module's "superblock"
 * @cid: input
 * @ip: output
 *
 */

int
cid2ipaddr(memexp_t * md, uint16 cid, uint32 * ip)
{
	node_group_t *group = (node_group_t *) md->cluster_data;
	node_info_t *ni;
	int retried = 0;

	MEP_ASSERT(group,);

      retry:

	ni = group->list;

	while (ni) {
		if (ni->cid == cid) {
			*ip = ni->ipaddr;
			return (0);
		}

		ni = ni->next;
	}

	if (retried)
		return (-1);

	else {
		reread_cluster_info(md);
		retried = TRUE;
		goto retry;
	}
}

/**
 * jid2cid
 *
 * @md: the lock module's "superblock"
 * @jid: input
 * @cid: output
 *
 */

int
jid2cid(memexp_t * md, uint32 jid, uint16 * cid)
{
	node_group_t *group = (node_group_t *) md->cluster_data;
	node_info_t *ni;
	int retried = 0;

	MEP_ASSERT(group,);

      retry:

	ni = group->list;

	while (ni) {
		if (ni->jid == jid) {
			*cid = ni->cid;
			return (0);
		}

		ni = ni->next;
	}

	if (retried)
		return (-1);

	else {
		reread_cluster_info(md);
		retried = TRUE;
		goto retry;
	}
}

/**
 * jid2nstomith
 *
 * @md: the lock module's "superblock"
 * @jid: input
 * @nstomith: output
 *
 */

int
jid2nstomith(memexp_t * md, uint32 jid, uint32 * nstomith)
{
	node_group_t *group = (node_group_t *) md->cluster_data;
	node_info_t *ni;
	int retried = 0;

	MEP_ASSERT(group,);

	*nstomith = 0;

      retry:

	ni = group->list;

	while (ni) {
		if (ni->jid == jid) {
			*nstomith = ni->nstomith;
			return (0);
		}

		ni = ni->next;
	}

	if (retried)
		return (0);

	else {
		reread_cluster_info(md);
		retried = TRUE;
		goto retry;
	}
}

/**
 * get_stomith_desc
 *
 * @md: the lock module's "superblock"
 * @jid: input
 * @st: stomith descriptor output
 * @n: which stomith descriptor to select
 *
 */

int
get_stomith_desc(memexp_t * md, uint32 jid, struct cluster_stomith *st, int n)
{
	cluster_node_t cn;
	struct cluster_stomith cs;
	struct buffer_head *bh;
	int block = jid + 1;
	int x, i, offset;
	uint32 magic;
	struct block_device *bdev;

	bdev = md->cfile->f_dentry->d_inode->i_bdev;
	bh = __bread(bdev, block, CIDEV_BLOCKSIZE);
	if (!bh) {
		printk("%s:  error reading Cluster Information Device.\n",
		       PROTO_NAME);
		return (-1);
	}

	cluster_node_in(&cn, bh->b_data);

	if (cn.cn_magic != OGFS_CLUSTER_NODE)
		goto fail;

	offset = sizeof(cluster_node_t);

	for (i = 0; i <= n; i++) {
		if (offset + 4 > CIDEV_BLOCKSIZE)	/* 4 is size of magic */
			goto fail;

		magic = be32_to_cpup((uint32 *) (bh->b_data + offset));

		if (ondisk_stomith_size(magic) + offset > CIDEV_BLOCKSIZE)
			goto fail;

		cluster_stomith_in(&cs, bh->b_data + offset);

		offset += ondisk_stomith_size(cs.cs_magic);
	}

	//mark_buffer_uptodate(bh, FALSE);
	clear_bit(BH_Uptodate, &bh->b_state);
	brelse(bh);

#warning fixme
/*
	for (x = 0; stomith_names[x].name[0]; x++) {
		if (cs.cs_magic == stomith_names[x].number)
			break;
	}

	if (!stomith_names[x].name[0])
		return (-1);
*/
	*st = cs;

	return (0);

      fail:

	//mark_buffer_uptodate(bh, FALSE);
	clear_bit(BH_Uptodate, &bh->b_state);
	brelse(bh);
	return (-1);
}

/**
 * cluster_node_count
 *
 * @md: the lock module's "superblock"
 *
 */

int
cluster_node_count(memexp_t * md)
{
	node_group_t *group = (node_group_t *) md->cluster_data;

	MEP_ASSERT(group,);

	return (group->count);
}

/**
 * show_cluster_info
 *
 * @md: the lock module's "superblock"
 *
 */

void
show_cluster_info(memexp_t * md)
{
	node_group_t *group = (node_group_t *) md->cluster_data;
	node_info_t *ni;
	char ip[16];

	if (!group)
		return;

	ni = group->list;

	printk("%s:  Cluster Information Device summary\n", PROTO_NAME);
	while (ni) {
		inet_ntoa(ni->ipaddr, ip);

		printk("  jid: %u, cid: %u, ipaddr: %s, nstomith: %u\n",
		       ni->jid, ni->cid, ip, ni->nstomith);

		ni = ni->next;
	}
	printk("  Node timeout: %u\n", md->node_timeout);
}
