/*
 *
 *    Copyright 1999-2001 Sistina Software, Inc.
 *    Portions Copyright 2003 Brian Jackson
 *    Portions Copyright 2003 The OpenGFS Project
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 *    This file was maintained by:
 *      David C. Teigland <teigland@sistina.com>
 *
 */

#include "config.h"

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <asm/uaccess.h>

#include "memexp.h"
#include "client_state.h"

/**
 * lock_refresh - thread to periodically increment heartbeat counter for
 * this node
 *
 * @data: memexp "superblock" structure
 *
 * calls client_refresh() once every HEARTBEAT_INTERVAL seconds
 */

int
lock_refresh(void *data)
{
	memexp_t *mep = (memexp_t *) data;
	unsigned long begin, end, secs;
	unsigned long flags;

	MEP_ASSERT(mep,);

	/*  Set up the thread  */

#warning test me
	//lock_kernel();

	daemonize("memexp: lock_refresh");

	spin_lock_irqsave(&current->sighand->siglock, flags);
	sigfillset(&current->blocked);
	recalc_sigpending();
	spin_unlock_irqrestore(&current->sighand->siglock, flags);

	current->session = 1;
	//current->pgrp = 1;
#ifdef SKIP
	current->tty = NULL;
#endif
	current->tty_old_pgrp = 0;

	sprintf(current->comm, "memexp_heartbeat");

	mep->lock_refresh_ts = current;	/* save this for the wakeup */

	while (!mep->lock_refresh_stop) {
		begin = jiffies;
		client_refresh(mep);
		end = jiffies;

		secs = CT_TO_SECS(end - begin);
		if (secs > HEARTBEAT_INTERVAL) {
			printk("%s:  slow heartbeat time (%u, %u, %u)\n",
				   PROTO_NAME, (unsigned int) begin,
				   (unsigned int) end,
				   (unsigned int) (end - begin));
			continue;
		}
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout((HEARTBEAT_INTERVAL-secs) * HZ);
	}

	up(&mep->lock_refresh_sem);

	//unlock_kernel();
	return (0);
}

/**
 * node_monitor - thread to periodically check entire cluster for dead nodes
 *
 * @data: memexp "superblock" structure
 *
 * calls reread_client_state() and check_expired_clients()
 * once every mep->node_timeout seconds
 */

int
node_monitor(void *data)
{
	memexp_t *mep = (memexp_t *) data;
	unsigned long flags;

	MEP_ASSERT(mep,);

	/*  Set up the thread  */

	lock_kernel();

	daemonize("memexp: node_monitor");

	spin_lock_irqsave(&current->sighand->siglock, flags);
	sigfillset(&current->blocked);
	recalc_sigpending();
	spin_unlock_irqrestore(&current->sighand->siglock, flags);

	current->session = 1;
	//current->pgrp = 1;
#ifdef SKIP
	current->tty = NULL;
#endif
	current->tty_old_pgrp = 0;

	sprintf(current->comm, "memexp_monitor");

	mep->node_monitor_ts = current;	/* save this for the wakeup */

	while (!mep->node_monitor_stop) {
		reread_client_state(mep);
		check_expired_clients(mep);

		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(mep->node_timeout * HZ);
	}

	up(&mep->node_monitor_sem);

	unlock_kernel();
	return (0);
}
