/*
 *    Copyright Sistina Software
 *    Portions Copyright 2003 Brian Jackson
 *    Portions Copyright 2003 The OpenGFS Project
 */

#include "config.h"

#include <linux/module.h>
#include <linux/kernel.h>
#include <asm/uaccess.h>
#include <linux/slab.h>

#include "osi.h"
#include "memexp.h"
#include "ll_memexp.h"

#include "dmep_tcp.h"
#include "dmep_scsi.h"

#include "osi.h"
#include "list.h"

memexp_t *ip_mep;
memexp_t *scsi_mep;

static int
byte_cmp(void *this, void *that, int size)
{
	int i;

	for (i = 0; i < size; i++) {
		if (((u8 *) this)[i] < ((u8 *) that)[i]) {
			return -1;
		} else if (((u8 *) this)[i] > ((u8 *) that)[i]) {
			return 1;
		}
	}
	return 0;
}

STATIC int
debug_load(memexp_t * mep, lm_lockname_t * lck,
	   buf_load_t * rdf, void *buf, u32 buflen)
{
	int error = 0;
	lm_lockname_t ip_lck, scsi_lck;
	buf_load_t ip_rdf, scsi_rdf;
	u8 ip_buf[buflen], scsi_buf[buflen];

	memcpy(&ip_lck, lck, sizeof(lm_lockname_t));
	memcpy(&scsi_lck, lck, sizeof(lm_lockname_t));
	memcpy(&ip_rdf, rdf, sizeof(buf_load_t));
	memcpy(&scsi_rdf, rdf, sizeof(buf_load_t));
	memcpy(ip_buf, buf, buflen);
	memcpy(scsi_buf, buf, buflen);

	error =
	    ip_mep->ll_memexp_ops->load(ip_mep, &ip_lck, &ip_rdf, ip_buf,
					buflen);

	if (error < 0) {
		printk("IP DMEP error :: load unable to complete.\n");
		return error;
	}

	error =
	    scsi_mep->ll_memexp_ops->load(scsi_mep, &scsi_lck, &scsi_rdf,
					  scsi_buf, buflen);

	if (error < 0) {
		printk("SCSI DMEP error :: load unable to complete.\n");
		return error;
	}

	/* comparison code */
	if (byte_cmp(&ip_lck, &scsi_lck, sizeof(lm_lockname_t))) {
		printk("\ndebug_load operation failure.\n");
		printk("lm_lockname_t's differ!\n");
		printk("=======================\n");

		printk("ip_lck.ln_number  : %" FMT_64 "u\n",
			   ip_lck.ln_number);
		printk("ip_lck.ln_type    : %u\n", ip_lck.ln_type);

		printk("scsi_lck.ln_number: %" FMT_64 "u\n",
			   ip_lck.ln_number);
		printk("scsi_lck.ln_type  : %u\n", ip_lck.ln_type);
		error = -1;
	} else {
		memcpy(lck, &scsi_lck, sizeof(lm_lockname_t));
	}

	if (ip_rdf.inuse != scsi_rdf.inuse) {
		printk("\ndebug_load operation failure.\n");
		printk("buf_load_t's differ!\n");
		printk("====================\n");

		printk("ip_rdf.inuse: %u\n", ip_rdf.inuse);
		/* IPO = Informational Purposes Only */
		printk("ip_rdf.full: %u - IPO\n", ip_rdf.full);
		printk("ip_rdf.seq %" FMT_64 "u - IPO\n", ip_rdf.seq);
		printk("ip_rdf.pbn %" FMT_64 "u -IPO\n", ip_rdf.pbn);

		printk("scsi_rdf.inuse: %u\n", scsi_rdf.inuse);
		printk("scsi_rdf.full: %u - IPO\n", scsi_rdf.full);
		printk("scsi_rdf.seq %" FMT_64 "u - IPO\n", scsi_rdf.seq);
		printk("scsi_rdf.pbn %" FMT_64 "u - IPO\n", scsi_rdf.pbn);
		/* Attention -- this should return -1, not do memcpy */
		memcpy(rdf, &ip_rdf, sizeof(buf_load_t));
	} else {
		memcpy(rdf, &ip_rdf, sizeof(buf_load_t));
	}

	if (byte_cmp(ip_buf, scsi_buf, buflen)) {
		int i;
		printk("\ndebug_load operation failure.\n");
		printk("(void *)buffers differ!\n");
		printk("=======================\n");
		printk("ip_buf:\n");
		for (i = 0; i < buflen; i++) {
			if (!(i % 8))
				printk("\n");
			printk("%.2x ", ip_buf[i]);
		}
		printk("\n\n");
		printk("scsi_buf:\n");
		for (i = 0; i < buflen; i++) {
			if (!(i % 8))
				printk("\n");
			printk("%.2x ", scsi_buf[i]);
		}
		printk("\n\n");
		error = -1;
	} else {
		memcpy(buf, ip_buf, buflen);
	}

	return error;
}

STATIC int
debug_store(memexp_t * mep, lm_lockname_t * lck,
	    buf_store_t * rdf, void *buf, u32 buflen)
{

	int error = 0;
	lm_lockname_t ip_lck, scsi_lck;
	buf_store_t ip_rdf, scsi_rdf;
	u8 ip_buf[buflen], scsi_buf[buflen];

	memcpy(&ip_lck, lck, sizeof(lm_lockname_t));
	memcpy(&scsi_lck, lck, sizeof(lm_lockname_t));
	memcpy(&ip_rdf, rdf, sizeof(buf_load_t));
	memcpy(&scsi_rdf, rdf, sizeof(buf_load_t));
	memcpy(ip_buf, buf, buflen);
	memcpy(scsi_buf, buf, buflen);

	error = ip_mep->ll_memexp_ops->store(ip_mep, &ip_lck, &ip_rdf,
					     ip_buf, buflen);

	if (error < 0) {
		printk("IP DMEP error :: store unable to complete.\n");
		return error;
	}

	error = scsi_mep->ll_memexp_ops->store(scsi_mep, &scsi_lck,
					       &scsi_rdf, scsi_buf, buflen);

	if (error < 0) {
		printk("SCSI DMEP error :: store unable to complete.\n");
		return error;
	}

	/* comparison code */
	if (byte_cmp(&ip_lck, &scsi_lck, sizeof(lm_lockname_t))) {
		printk("\ndebug_store operation failure.\n");
		printk("lm_lockname_t's differ!\n");
		printk("=======================\n");

		printk("ip_lck.ln_number  : %" FMT_64 "u\n",
			   ip_lck.ln_number);
		printk("ip_lck.ln_type    : %u\n", ip_lck.ln_type);

		printk("scsi_lck.ln_number: %" FMT_64 "u\n",
			   ip_lck.ln_number);
		printk("scsi_lck.ln_type  : %u\n", ip_lck.ln_type);
		error = -1;
	} else {
		memcpy(lck, &scsi_lck, sizeof(lm_lockname_t));
	}

	if (ip_rdf.inuse != scsi_rdf.inuse) {
		printk("\ndebug_load operation failure.\n");
		printk("buf_store_t's differ!\n");
		printk("====================\n");

		printk("ip_rdf.inuse: %u\n", ip_rdf.inuse);
		/* IPO = Informational Purposes Only */
		printk("ip_rdf.seq %" FMT_64 "u - IPO\n", ip_rdf.seq);
		printk("ip_rdf.pbn %" FMT_64 "u -IPO\n", ip_rdf.pbn);

		printk("scsi_rdf.inuse: %u\n", scsi_rdf.inuse);
		printk("scsi_rdf.seq %" FMT_64 "u - IPO\n", scsi_rdf.seq);
		printk("scsi_rdf.pbn %" FMT_64 "u - IPO\n", scsi_rdf.pbn);
		error = -1;
	} else {
		memcpy(rdf, &scsi_rdf, sizeof(buf_load_t));
	}

	if (byte_cmp(ip_buf, scsi_buf, buflen)) {
		int i;
		printk("\ndebug_store operation failure.\n");
		printk("(void *)buffers differ!\n");
		printk("=======================\n");
		printk("ip_buf:\n");
		for (i = 0; i < buflen; i++) {
			if (!(i % 8))
				printk("\n");
			printk("%.2x ", ip_buf[i]);
		}
		printk("\n\n");
		printk("scsi_buf:\n");
		for (i = 0; i < buflen; i++) {
			if (!(i % 8))
				printk("\n");
			printk("%.2x ", scsi_buf[i]);
		}
		printk("\n\n");
		error = -1;
	} else {
		memcpy(buf, ip_buf, buflen);
	}

	return error;
}

STATIC int
debug_dump(memexp_t * mep, void **cookie, dump_status_t * heads,
	   u8 * bufs, u32 cnt, u32 sze, u32 * ret)
{
	int error = 0;

	error = ip_mep->ll_memexp_ops->dump(mep, cookie, heads,
					    bufs, cnt, sze, ret);

	if (error < 0) {
		printk("IP DMEP error :: dump unable to complete.\n");
		return error;
	}

	error = scsi_mep->ll_memexp_ops->dump(mep, cookie, heads,
					      bufs, cnt, sze, ret);

	if (error < 0) {
		printk("SCSI DMEP error :: dump unable to complete.\n");
		return error;
	}

	/* comparison code */
	/* copy code */

	return error;
}

STATIC int
debug_modepg(memexp_t * mep, u8 set, uint64 * numbufs, uint32 * datasize)
{
	int error = 0;
	uint64 ip_numbufs, scsi_numbufs;
	uint32 ip_datasize, scsi_datasize;

	ip_numbufs = scsi_numbufs = *numbufs;
	ip_datasize = scsi_datasize = *datasize;
	error = ip_mep->ll_memexp_ops->modepg(ip_mep, set,
					      &ip_numbufs, &ip_datasize);

	if (error < 0) {
		printk("IP DMEP error :: modepg unable to complete.\n");
		return error;
	}

	error = scsi_mep->ll_memexp_ops->modepg(scsi_mep, set,
						&scsi_numbufs, &scsi_datasize);

	if (error < 0) {
		printk("SCSI DMEP error :: modepg unable to complete.\n");
		return error;
	}

	/* comparison code */

	/* numbufs is not correctly reported from modepg functions */
	if ( /*(ip_numbufs == scsi_numbufs) && */
	    (ip_datasize == scsi_datasize)) {
		*numbufs = ip_numbufs;
		*datasize = ip_datasize;
	} else {
		printk("ip_numbufs   : %" FMT_64 "u\n"
			   "scsi_numbufs : %" FMT_64 "u\n"
			   "ip_datasize  : %u\n"
			   "scsi_datasize: %u\n",
			   ip_numbufs, scsi_numbufs, ip_datasize,
			   scsi_datasize);
		*numbufs = 0;
		*datasize = 0;
		error = -1;
	}

	return error;
}

STATIC int
debug_init(memexp_t * mep, char *args)
{
	int error = 0;
	char *divider;

	ip_mep = (memexp_t *) kmalloc(sizeof(memexp_t), GFP_KERNEL);
	if (!ip_mep) {
		printk("%s:  out of memory\n", PROTO_NAME);
		return -ENOMEM;
	}

	scsi_mep = (memexp_t *) kmalloc(sizeof(memexp_t), GFP_KERNEL);
	if (!scsi_mep) {
		kfree(ip_mep);
		printk("%s:  out of memory\n", PROTO_NAME);
		return -ENOMEM;
	}

	memcpy(ip_mep, mep, sizeof(memexp_t));
	memcpy(scsi_mep, mep, sizeof(memexp_t));
	memset(ip_mep->lockdev_name, '\0', 256);
	memset(scsi_mep->lockdev_name, '\0', 256);

	ip_mep->ll_memexp_ops = &ll_dmep_tcp_ops;
	//scsi_mep->ll_memexp_ops = &ll_dmep_scsi_ops;

	if ((divider = strstr(mep->lockdev_name, "::"))) {
		strcpy(scsi_mep->lockdev_name, (divider + 2));
		strncpy(ip_mep->lockdev_name, mep->lockdev_name,
			(strlen(mep->lockdev_name) - strlen(divider)));
	} else {
		kfree(ip_mep);
		kfree(scsi_mep);
		return -1;
	}

	printk("\nIP lockdev   :: %s\n", ip_mep->lockdev_name);
	printk("SCSI lockdev :: %s\n\n", scsi_mep->lockdev_name);

	error = ip_mep->ll_memexp_ops->init(ip_mep, args);

	if (error < 0) {
		printk("IP DMEP error :: init unable to complete.\n");
		kfree(ip_mep);
		kfree(scsi_mep);
		return error;
	}

	error = scsi_mep->ll_memexp_ops->init(scsi_mep, args);

	if (error < 0) {
		printk("SCSI DMEP error :: init unable to complete.\n");
		kfree(ip_mep);
		kfree(scsi_mep);
		return error;
	}

	/* comparison code */

	return error;
}

STATIC int
debug_release(memexp_t * mep)
{
	int error = 0;

	error = ip_mep->ll_memexp_ops->release(ip_mep);

	if (error < 0) {
		printk("IP DMEP error :: release unable to complete.\n");
		return error;
	}

	error = scsi_mep->ll_memexp_ops->release(scsi_mep);

	if (error < 0) {
		printk("SCSI DMEP error :: release unable to complete.\n");
		return error;
	}

	/* comparison code */

	mep->ll_memexp_data = NULL;
	return error;
}

STATIC int
debug_stomith(memexp_t * mep, uint32 ipaddr)
{
	int error = 0;

	error = ip_mep->ll_memexp_ops->stomith(ip_mep, ipaddr);

	if (error < 0) {
		printk("IP DMEP error :: stomith unable to complete.\n");
		return error;
	}

	error = scsi_mep->ll_memexp_ops->stomith(scsi_mep, ipaddr);

	if (error < 0) {
		printk("SCSI DMEP error :: stomith unable to complete.\n");
		return error;
	}

	/* comparison code */

	return error;
}

struct ll_memexp_ops ll_dmep_debug_ops = {
	load:debug_load,
	store:debug_store,
	dump:debug_dump,
	modepg:debug_modepg,
	init:debug_init,
	release:debug_release,
	stomith:debug_stomith
};
