/*
 *
 *    Copyright 1999-2001 Sistina Software, Inc.
 *    Portions Copyright 2001 The OpenGFS Project
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 *    This file was maintained by:
 *      David C. Teigland <teigland@sistina.com>
 *
 */

#include "config.h"

#include <linux/module.h>
#include <linux/kernel.h>
#include <asm/uaccess.h>
#include <linux/slab.h>

#include "osi.h"
#include "memexp.h"
#include "ll_memexp.h"
#include "pool_lock.h"
#include "pool_ondisk.h"
#include "pool_incore.h"

#define DMEP_SEGMENT_NUMBER 0

#define DMEP_OUT_OPCODE 0x89
#define DMEP_IN_OPCODE 0x85

struct dmep_scsi_data {
	struct nameidata nd;
	struct dentry *lockdev_dcache;
	struct inode *lockdev_inode;
};
typedef struct dmep_scsi_data dmep_scsi_data_t;

struct dump_info {
	kdev_t dev;
	uint64 pbn;
};
typedef struct dump_info dump_info_t;

extern pool_t pools[];

/**
 * isbdev - check wether a nameidata represents a a useable block device
 * @nd:		&struct nameidata to check
 *
 */
static int
isbdev(struct nameidata *nd)
{
#ifdef IS_NODEV
	if (IS_NODEV(nd->dentry->d_inode))
		return 0;
#else
	if (nd->mnt->mnt_flags & MNT_NODEV)
		return 0;
#endif
	return S_ISBLK(nd->dentry->d_inode->i_mode);
}

/**
 * next_pool_dev - return the next lock device in the pool
 * @pldev: the pool device
 * @in: the current lock device (NULL if first)
 * @out: the next lock device (after in)
 *
 * Returns: 0 if next device was found, -1 if no next device was found
 */

STATIC int
next_pool_dev(kdev_t * pldev, kdev_t * in, kdev_t * out)
{
	pool_t *pp;
	subpool_t *sp;
	dmepdev_t *dev;
	int next = 0, i, j;

	pp = &pools[MINOR(*pldev)];

	for (i = 0; i < pp->p_num_spools; i++) {
		sp = &pp->p_spools[i];

		for (j = 0; j < sp->sp_dmepdevs; j++) {
			dev = &sp->sp_dmepdevlist[j];

			if (!in) {
				*out = dev->dm_dev;
				return (0);
			}

			if (next) {
				*out = dev->dm_dev;
				return (0);
			}

			if (dev->dm_dev == *in)
				next = TRUE;
		}
	}

	return (-1);
}

/**
 * get_rdf_64 - extract 64 bit number from return data buffer
 * @src: buffer
 * @val: number
 *
 */

STATIC void
get_rdf_64(unsigned char *src, uint64 * val)
{
	uint64 number = 0;
	int i;

	for (i = 0; i < 8; i++)
		number += ((src[i] & 0xFF) << (8 * (7 - i)));

	*val = number;
}

/**
 * fill_cdb - pack DMEP command parameters into a cdb structure
 * @cdb: the cdb buffer
 * @lockname: contains number and type which form bid
 * @opcode: DMEP opcode for load or store
 * @action: which type of load or store
 * @segment: DMEP segment number
 * @len: the length
 *
 */

STATIC void
fill_cdb(unsigned char *cdb, lm_lockname_t * lockname, uint8 opcode,
	 uint8 action, uint8 segment, uint32 len)
{
	uint64 number;
	number = lockname->ln_number;

	cdb[0] = opcode;
	cdb[1] = action & 0x1F;
	cdb[2] = segment;
	cdb[3] = lockname->ln_type & 0xFF;

	cdb[4] = (number >> 56) & 0xFF;
	cdb[5] = (number >> 48) & 0xFF;
	cdb[6] = (number >> 40) & 0xFF;
	cdb[7] = (number >> 32) & 0xFF;
	cdb[8] = (number >> 24) & 0xFF;
	cdb[9] = (number >> 16) & 0xFF;
	cdb[10] = (number >> 8) & 0xFF;
	cdb[11] = (number >> 0) & 0xFF;

	cdb[12] = (len >> 16) & 0xFF;
	cdb[13] = (len >> 8) & 0xFF;
	cdb[14] = (len >> 0) & 0xFF;

	cdb[15] = 0;
}

/**
 * unfill_cdb - unpack DMEP command parameters from a cdb
 * @cdb: the cdb buffer
 * @lockname: contains number and type which form bid
 * @opcode: DMEP opcode for load or store
 * @action: which type of load or store
 * @segment: DMEP segment number
 * @len: the length
 *
 */

STATIC void
unfill_cdb(unsigned char *cdb, lm_lockname_t * lockname, uint8 * opcode,
	   uint8 * action, uint8 * segment, uint32 * len)
{
	*opcode = cdb[0] & 0xFF;
	*action = cdb[1] & 0xFF;
	*segment = cdb[2] & 0xFF;
	lockname->ln_type = cdb[3] & 0xFF;
	get_rdf_64(cdb + 4, &lockname->ln_number);
	*len = (cdb[12] & 0xFF) << 16;
	*len += (cdb[13] & 0xFF) << 8;
	*len += (cdb[14] & 0xFF);
}

/**
 * fill_store - pack PDF for store command
 * @out: the PDF buffer
 * @store: contains sequence number and pbn
 * @length:
 * @action:
 *
 */

STATIC void
fill_store(unsigned char *out, buf_store_t * store, uint32 length, uint8 action)
{
	out[0] = (length >> 16) & 0xFF;
	out[1] = (length >> 8) & 0xFF;
	out[2] = (length >> 0) & 0xFF;

	out[3] = action & 0x1F;

	out[4] = (store->inuse ? 0x80 : 0x00);

	out[5] = 0;

	out[8] = (store->seq >> 56) & 0xFF;
	out[9] = (store->seq >> 48) & 0xFF;
	out[10] = (store->seq >> 40) & 0xFF;
	out[11] = (store->seq >> 32) & 0xFF;
	out[12] = (store->seq >> 24) & 0xFF;
	out[13] = (store->seq >> 16) & 0xFF;
	out[14] = (store->seq >> 8) & 0xFF;
	out[15] = (store->seq >> 0) & 0xFF;

	out[16] = (store->pbn >> 56) & 0xFF;
	out[17] = (store->pbn >> 48) & 0xFF;
	out[18] = (store->pbn >> 40) & 0xFF;
	out[19] = (store->pbn >> 32) & 0xFF;
	out[20] = (store->pbn >> 24) & 0xFF;
	out[21] = (store->pbn >> 16) & 0xFF;
	out[22] = (store->pbn >> 8) & 0xFF;
	out[23] = (store->pbn >> 0) & 0xFF;
}

/**
 * unfill_store - unpack PDF for store command
 * @pdf: the PDF buffer
 * @store: contains sequence number and pbn
 * @length:
 * @action:
 *
 */

STATIC void
unfill_store(unsigned char *pdf, buf_store_t * store, uint32 * length,
	     uint8 * action)
{
	*length = (pdf[0] & 0xFF) << 16;
	*length += (pdf[1] & 0xFF) << 8;
	*length += (pdf[2] & 0xFF);

	*action = pdf[3] & 0xFF;

	store->inuse = (pdf[4] ? 1 : 0);

	get_rdf_64(pdf + 8, &store->seq);
	get_rdf_64(pdf + 16, &store->pbn);
}

/**
 * dmep_show_error
 * @cdb:
 * @ec:
 * @pdf:
 *
 */

STATIC void
dmep_show_error(unsigned char *cdb, error_code_t * ec, unsigned char *pdf)
{
	int i;
	unsigned int opcode = (cdb[0] & 0xFF);

	if ((opcode == DMEP_OUT_OPCODE) && (ec->sense_key == 0xe) &&
	    (ec->asc == 0x26) && (ec->ascq == 0xe)) {
		/* store failures are valid due to sequence number mismatches */
		return;
	}

	printk("%s:  DMEP command failure\n", PROTO_NAME);
	printk("%s:    SK/ASC/ASCQ = %.2x/%.2x/%.2x\n", PROTO_NAME,
		   ec->sense_key, ec->asc, ec->ascq);
	printk("%s:    SKS         = %.2x %.2x %.2x\n", PROTO_NAME,
		   ec->specific[0], ec->specific[1], ec->specific[2]);
	printk("%s:    CDB: ", PROTO_NAME);

	for (i = 0; i < 16; i++)
		printk("%02x ", 0xFF & cdb[i]);
	printk("\n");

	{
		lm_lockname_t lockname;
		uint8 opcode, action, segment;
		uint32 len;

		unfill_cdb(cdb, &lockname, &opcode, &action, &segment, &len);

		printk("%s:    lock_type = %u, lock_number = %" FMT_64
			   "u\n", PROTO_NAME, lockname.ln_type,
			   lockname.ln_number);
		printk
		    ("%s:    opcode = 0x%x, action = %u, segment = %u, len = %u\n",
		     PROTO_NAME, opcode, action, segment, len);
	}

	if (opcode == DMEP_OUT_OPCODE && pdf != NULL) {
		buf_store_t store;
		uint32 length;
		uint8 action;

		printk("%s:    PDF: ", PROTO_NAME);
		for (i = 0; i < 24; i++)
			printk("%02x ", 0xFF & pdf[i]);
		printk("\n");

		unfill_store(pdf, &store, &length, &action);

		printk("%s:    length = %u, action = %u, seq = %" FMT_64
			   "u, pbn = %" FMT_64 "u\n", PROTO_NAME, length,
			   action, store.seq, store.pbn);
	}
}

/**
 * dmep_scsi_init
 * @mep: the lock module "superblock"
 * @args:
 *
 */

STATIC int
dmep_scsi_init(memexp_t * mep, char *args)
{
	dmep_scsi_data_t *pdata;
	int error = 0;

	pdata = kmalloc(sizeof(dmep_scsi_data_t), GFP_NOFS);
	if (!pdata)
		return (-ENOMEM);

	if (path_init(mep->lockdev_name, 0, &(pdata->nd)))
		error = path_walk(mep->lockdev_name, &(pdata->nd));

	if (!pdata->nd.dentry || !pdata->nd.dentry->d_inode ||
	    !isbdev(&pdata->nd)) {
		path_release(&pdata->nd);
		printk("%s:  can't open lock dev %s\n", PROTO_NAME,
		       mep->lockdev_name);
		kfree(pdata);
		return (-ENXIO);
	}

	pdata->lockdev_inode = pdata->nd.dentry->d_inode;
	error = blkdev_get(pdata->lockdev_inode->i_bdev, FMODE_READ,
			   O_RDONLY, -1);

	if (error) {
		path_release(&pdata->nd);
		printk("%s:  can't open lock dev %s\n", PROTO_NAME,
		       mep->lockdev_name);
		kfree(pdata);
		return (-EIO);
	}

	pdata->lockdev_dcache = pdata->nd.dentry;

	mep->ll_memexp_data = pdata;
	return (error);
}

/**
 * dmep_scsi_release
 * @mep: the lock module "superblock"
 *
 */

STATIC int
dmep_scsi_release(memexp_t * mep)
{
	dmep_scsi_data_t *pdata = (dmep_scsi_data_t *) mep->ll_memexp_data;

	MEP_ASSERT(pdata,);

	blkdev_put(pdata->lockdev_inode->i_bdev, -1);

	dput(pdata->lockdev_dcache);

	kfree(pdata);

	mep->ll_memexp_data = NULL;

	return (0);
}

/**
 * dmep_scsi_load
 * @mep: the lock module "superblock"
 * @lockname: holds lock number and type
 * @ret: the return values
 * @data: return data buffer copied here
 * @datalen: length of data buffer
 *
 */

STATIC int
dmep_scsi_load(memexp_t * mep, lm_lockname_t * lockname,
	       buf_load_t * ret, void *data, uint32 datalen)
{
	dmep_scsi_data_t *pdata = (dmep_scsi_data_t *) mep->ll_memexp_data;
	unsigned char cdb[16];
	unsigned char *retbuf;
	uint32 retlen;
	error_code_t ec;
	int error;

	MEP_ASSERT(pdata,);
	memset(ret, 0, sizeof(buf_load_t));
	memset(data, 0, datalen);

	/* intermediate buffer includes RDF and buffer */

	retlen = 24 + datalen;
	retbuf = kmalloc(retlen, GFP_NOFS);
	if (!retbuf)
		return (-ENOMEM);
	memset(retbuf, 0, retlen);

	/* opcode: 0x85, action: 0, segment: n */

	fill_cdb(cdb, lockname, DMEP_IN_OPCODE, 0, DMEP_SEGMENT_NUMBER, retlen);

	/* pool only uses the 8 byte lock number to map to a lock device */

	error = pool_memexp(pdata->lockdev_inode->i_rdev, LOGICAL_DEVICE, cdb,
			    lockname->ln_number, retbuf, retlen, &ec);
	if (error) {
		dmep_show_error(cdb, &ec, NULL);
		goto out;
	}

	/* prepare info for returning */

	ret->inuse = retbuf[4] & 0x80;
	ret->full = retbuf[5];
	get_rdf_64(retbuf + 8, &ret->seq);
	get_rdf_64(retbuf + 16, &ret->pbn);
	memcpy(data, retbuf + 24, datalen);

      out:
	kfree(retbuf);
	return (0);
}

/**
 * dmep_scsi_store
 *
 * @mep: the lock module "superblock"
 * @lockname: holds lock number and type
 * @store: input store parameters, pbn and seq number
 * @data: buffer containing data to store
 * @datalen: length of data buffer
 *
 */

STATIC int
dmep_scsi_store(memexp_t * mep, lm_lockname_t * lockname,
		buf_store_t * store, void *data, uint32 datalen)
{
	dmep_scsi_data_t *pdata = (dmep_scsi_data_t *) mep->ll_memexp_data;
	unsigned char cdb[16];
	unsigned char *outbuf;
	int outlen;
	error_code_t ec;
	int error = 0;

	MEP_ASSERT(pdata,);

	/* intermediate buffer holds PDF and data buffer */

	outlen = 24 + datalen;
	outbuf = kmalloc(outlen, GFP_NOFS);
	if (!outbuf)
		return (-ENOMEM);

	/* opcode: 0x89, action: 0, segment: n */

	fill_cdb(cdb, lockname, DMEP_OUT_OPCODE, 0, DMEP_SEGMENT_NUMBER,
		 outlen);

	/* action: 0 */

	fill_store(outbuf, store, outlen, 0);

	/* opaque buffer is copied after the parameter data info */

	memcpy(outbuf + 24, data, datalen);

	/* pool only uses the 8 byte lock number to map to a lock device */

	error = pool_memexp(pdata->lockdev_inode->i_rdev, LOGICAL_DEVICE, cdb,
			    lockname->ln_number, outbuf, outlen, &ec);
	if (error)
		dmep_show_error(cdb, &ec, outbuf);

	kfree(outbuf);
	return (error);
}

/**
 * dmep_scsi_modepg - this is no longer really a mode page, but the name stuck
 * @mep: the lock module "superblock"
 * @set: choose sense or select action
 * @numbufs: not used
 * @size: size of buffers
 *
 */

STATIC int
dmep_scsi_modepg(memexp_t * mep, uint8 set, uint64 * numbufs, uint32 * size)
{
	dmep_scsi_data_t *pdata = (dmep_scsi_data_t *) mep->ll_memexp_data;
	unsigned char cdb[16];
	unsigned char *retbuf;
	uint32 retlen;
	error_code_t ec;
	int error, first = TRUE;
	lm_lockname_t lockname;
	kdev_t dev;
	uint32 bufsize;

	MEP_ASSERT(pdata,);

	*numbufs = 0;
	*size = 0;

	if (set)
		return (-1);

	/* RDF will be returned in this buffer */

	retlen = 20;
	retbuf = kmalloc(retlen, GFP_NOFS);
	if (!retbuf)
		return (-ENOMEM);

	/* this is ignored for sense config */

	lockname.ln_number = 0;
	lockname.ln_type = 0;

	/* opcode: 0x85, action: 2, segment: n */

	fill_cdb(cdb, &lockname, DMEP_IN_OPCODE, 2, DMEP_SEGMENT_NUMBER,
		 retlen);

	/* first real memexp device */

	error = next_pool_dev(&pdata->lockdev_inode->i_rdev, NULL, &dev);
	if (error < 0) {
		printk("%s:  no DMEP device\n", PROTO_NAME);
		goto fail;
	}

	/* sense on all real devices */

	while (1) {
		memset(retbuf, 0, retlen);

		error =
		    pool_memexp(dev, REAL_DEVICE, cdb, 0, retbuf, retlen, &ec);
		if (error) {
			printk("%s:  failed to get DMEP info from device\n",
				   PROTO_NAME);
			dmep_show_error(cdb, &ec, NULL);
			goto fail;
		}

		bufsize = 0;
		bufsize = ((retbuf[16] & 0xFF) << 16) +
		    ((retbuf[17] & 0xFF) << 8) + ((retbuf[18] & 0xFF) << 0);

		if (!bufsize) {
			printk("%s:  DMEP buffer size set to zero: %u,%u\n",
			       PROTO_NAME, MAJOR(dev), MINOR(dev));
			goto fail;
		} else {
			printk
			    ("%s:  detected DMEP buffer size %u on device %u,%u\n",
			     PROTO_NAME, bufsize, MAJOR(dev), MINOR(dev));
		}

		if (first) {
			*size = bufsize;
			first = FALSE;
		} else {
			if (*size != bufsize) {
				printk("%s:  inconsistent DMEP buffer sizes\n",
				       PROTO_NAME);
				goto fail;
			}
		}
		error =
		    next_pool_dev(&pdata->lockdev_inode->i_rdev, &dev, &dev);
		if (error < 0) {
			kfree(retbuf);
			return (0);
		}
	}

      fail:

	kfree(retbuf);
	return (-1);
}

/**
 * dmep_scsi_dump
 * @mep: the lock module "superblock"
 * @cookie: holds info this routine needs between calls
 * @heads: headers describing each buffer
 * @bufs: the buffers are packed in this space
 * @count: number of buffers which can be accepted
 * @size: the buffer size
 * @retcount: number of buffers actually returned
 *
 * Returns:  -EXXXX on error, 0 or 1 on success
 *            0: there are no more buffers to dump
 *            1: there are more buffers, call dump again
 */

STATIC int
dmep_scsi_dump(memexp_t * mep, void **cookie, dump_status_t * heads,
	       unsigned char *bufs, uint32 count, uint32 size,
	       uint32 * retcount)
{
	dmep_scsi_data_t *pdata = (dmep_scsi_data_t *) mep->ll_memexp_data;
	unsigned char cdb[16];
	unsigned char *retbuf;
	uint32 retlen;
	dump_info_t *di;
	kdev_t dev, new;
	uint64 pbn = 0, pbn2 = 0;
	uint32 bytes, realcount;
	unsigned char *src;
	int error, i;
	lm_lockname_t lockname;
	unsigned int more;
	error_code_t ec;

	*retcount = 0;

	/* figure out which real scsi device and real pbn to start with */

	if (*cookie) {
		di = (dump_info_t *) (*cookie);
		dev = di->dev;
		pbn = di->pbn;
	} else {
		pbn = 0;
		error =
		    next_pool_dev(&pdata->lockdev_inode->i_rdev, NULL, &dev);
		if (error < 0)
			return (-1);

		di = kmalloc(sizeof(dump_info_t), GFP_NOFS);
		if (!di)
			return (-ENOMEM);

		*cookie = di;
	}

	/* 8 byte header for dump command as a whole,
	 * 28 byte header for each opaque buffer returned,
	 * count opaque buffers returned,
	 * size is the length of each opaque buffer */

	retlen = 8 + (28 * count) + (size * count);
	retbuf = kmalloc(retlen, GFP_NOFS);
	if (!retbuf)
		return (-ENOMEM);
	memset(retbuf, 0, retlen);

	/* the type field is just "reserved" in this case */

	lockname.ln_number = pbn;
	lockname.ln_type = 0;

	/* action: 1, segment: n */

	fill_cdb(cdb, &lockname, DMEP_IN_OPCODE, 1, DMEP_SEGMENT_NUMBER,
		 retlen);

	/* the lock number field is unused with REAL_DEVICE */

	error = pool_memexp(dev, REAL_DEVICE, cdb, 0, retbuf, retlen, &ec);
	if (error) {
		dmep_show_error(cdb, &ec, NULL);
		kfree(di);
		return (-1);
	}

	/* pick off the 8 header bytes for the whole dump */

	bytes = 0;
	bytes = ((retbuf[0] & 0xFF) << 16) +
	    ((retbuf[1] & 0xFF) << 8) + ((retbuf[2] & 0xFF) << 0);

	MEP_ASSERT((retbuf[3] & 0x1F) == 1,);

	more = (retbuf[4] & 0x80);

	/* actual number of returned buffers may differ from count */

	realcount = (bytes - 8) / (28 + size);

	if ((bytes - 8) % (28 + size))
		printk
		    ("%s:  warning - something is wrong with buffer byte counts\n",
		     PROTO_NAME);

	/* split up the headers and the data
	   headers go into "heads" array and raw buffer contents go into "bufs" array */

	src = retbuf + 8;

	for (i = 0; i < realcount; i++) {
		src += 3;	/* reserved bytes */

		heads->lockname.ln_type = (unsigned int) (*src);
		get_rdf_64(src + 1, &heads->lockname.ln_number);

		src += 9;	/* bid (lock type + number) */

		get_rdf_64(src, &heads->seq);

		src += 8;	/* sequence number */

		get_rdf_64(src, &heads->pbn);
		get_rdf_64(src, &pbn2);

		src += 8;	/* pbn */

		memcpy(bufs, src, size);

		src += size;	/* opaque buffer */

		bufs += size;
		heads++;
	}

	/* return values */

	*retcount = realcount;

	if (!more) {
		error =
		    next_pool_dev(&pdata->lockdev_inode->i_rdev, &dev, &new);

		if (error < 0) {
			kfree(di);
			return (0);
		}

		di->dev = new;
		di->pbn = 0;
	} else {
		di->pbn = pbn2 + 1;
		di->dev = dev;
	}

	kfree(retbuf);
	return (1);
}

/**
 * dmep_scsi_stomith - not used
 * @mep: the lock module "superblock"
 * @ipaddr:
 *
 */

STATIC int
dmep_scsi_stomith(memexp_t * mep, uint32 ipaddr)
{
	return (0);
}

struct ll_memexp_ops ll_dmep_scsi_ops = {
	load:dmep_scsi_load,
	store:dmep_scsi_store,
	dump:dmep_scsi_dump,
	modepg:dmep_scsi_modepg,
	init:dmep_scsi_init,
	release:dmep_scsi_release,
	stomith:dmep_scsi_stomith
};
