/*
 *
 *    Copyright 2001 Sistina Software, Inc.
 *    Portions Copyright 2003 Brian Jackson
 *    Portions Copyright 2003 The OpenGFS Project
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 *    This file was maintained by:
 *      Mike C Tilstra <conrad@sistina.com>
 *
 */

/*
 * This is a first hack.  I've got to rework the server a bit before this
 * can handle multi-requests-pending.  So in the meantime this is
 * completely serialized.  But that should be good enough. (for now)
 * Probably easier to debug anyways.
 *
 *
 * To handle the need of reconnecting a lost connection.  All of the
 * communication parts of this have been placed within a thing I'm calling
 * an ATTEMPT block.  This Attempts to execute some code, where it calls
 * RETRY when there is an error.  RETRY jumps to the RESTART block.  In
 * RESTART we determin if we can recover from this error, and then do what
 * is needed to clean up before we try again.  The end of a RESTART block
 * jumps to the start of the ATTEMPT block.  The end of an ATTEMPT block
 * breaks out of the ATTEMPT.
 *
 * Loads are easy.  We ATTEMPT the load. If there is a recoverable error,
 * we re-issue the load.  All RESTART does here is reconnects to a server.
 *
 * Store is difficult.  We ATTEMPT to do a store.  On recoverable errors,
 * we reconnect and issue a Verify to the server.  If the verify returns
 * failure, we redo the store.  If the Verify returns success, we break out
 * stating that the store was a success.
 *
 * WARNING! The curent implementation of the server that makes the Verify
 * work relies on the fact that both this code here and the server are
 * serial. (One lock op at a time.)  This will be worked on in the future,
 * but for anyone else looking into things here, this is a very important
 * thing to keep in mind.
 *
 *
 * ----
 * We've been seeing some weird things with the schedulaling of threads.
 * If I make this support multi-requests-pending, try to do it without a
 * listener thread. This is doable.  Basically I think you have one of all
 * the pending doing the listening.  If it receives something for some
 * other request, put the data into a buff, and wake them so they can deal.
 * If it is for us, store it, pass off the task of listening to another
 * pending request, and deal with the data.
 *
 * Will have to fix up server first.  It must have the ablity to have
 * multiple pending requests. (well, it doesn't NEED that, but it is either
 * that, or some half-assed hack to get things closer.)  Basically, I have
 * to allow the possiblity of having to `active' store ops.  Currently the
 * store cannot do this.
 *
 * Its not actually to active stores, but we have to be able to handle a
 * store AND a verify from the same client.  And since verify is only
 * called from within store.....
 *
 */

#include "config.h"

#include <linux/module.h>

#include "osi.h"
#include "list.h"

#include <linux/net.h>
#include <linux/in.h>
#include <linux/socket.h>
#include <net/sock.h>

#include "dmep_tcp.h"
#include "util.h"
#include "wiredtypes.h"

#ifndef MAX
#define MAX(a,b) ((a>b)?a:b)
#endif
#include "timed.h"

/* The list elements for the Queued sleepers. */
typedef struct Qsleep_s {
	struct list_head Qsp_list;
	struct semaphore sleeper;
} Qsleep_t;
typedef struct QsleepHead_s {
	struct semaphore outer;
	struct semaphore Qmux;
	unsigned int queue_depth;
	struct list_head IOQ;
} QslpHead_t;

/* This here is a private structure to this interface.  It holds all the
 * nicities like sockets and mutexts.
 */
typedef struct sm_s {
	struct socket *sock;
	struct sockaddr_in adr;
	int retry_pause;

	QslpHead_t Qedmux;
} tcp_mep_t;

#define RETRY_PAUSE_DEFAULT 5

/* prototype and junk for internal functions. */
typedef enum { DXwrite = 0, DXread, DXpeek } DXtype;
static char *DXstrs[] = { "write", "read", "peek" };
static int data_xmit(DXtype send, struct socket *sock, void *buf, u32 size);

/*************************************************************************/
/* This is an abstraction.
 * It keeps thing clearer in my head, YMMV.
 */
#define ATTEMPT while(1){
#define RESTART break; \
                __restarthere:
#define ENDATTEMPT }
#ifdef DEBUG
#define RETRY {printk("===>RETRY at %d from err(%d)\n",__LINE__,err); \
   goto __restarthere;}
#else				/* DEBUG */
#define RETRY goto __restarthere
#endif /*DEBUG*/
/*************************************************************************/
 Timestats_Init(dmep_load_time);
Timestats_Init(dmep_store_time);
Timestats_Init(dmep_dump_time);
Timestats_Init(dmep_release_time);
Timestats_Init(dmep_modpage_time);
Timestats_Init(dmep_ban_time);
/*************************************************************************/

/**
 * Qdown - wait for your turn to the IO pipe
 * @tmt: < tcp_mep_t structure
 *
 * This function and its friend Qup for a fifo like ordering to the
 * proccesses that try to get by here.  This is to try and avoid some
 * starvation we had been seeing.
 *
 */
void
Qdown(QslpHead_t * qm)
{
	Qsleep_t tmp;

	if (down_trylock(&qm->outer) != 0) {
		/* ok, someone is working. put ourselves on the Q. */

		INIT_LIST_HEAD(&tmp.Qsp_list);
		init_MUTEX_LOCKED(&tmp.sleeper);

		down(&qm->Qmux);
		list_add_tail(&tmp.Qsp_list, &qm->IOQ);
		qm->queue_depth++;
		up(&qm->Qmux);

		/* now wait for this to awaken */
		down(&tmp.sleeper);

		/* remove myself from the list of waiters. */
		down(&qm->Qmux);
		list_del(&tmp.Qsp_list);
		qm->queue_depth--;
		up(&qm->Qmux);

		/* when it comes up we'll grab the outer */
		down(&qm->outer);
	}
}

/**
 * Qup - tell the next in line that you're done.
 * @tmt: < tcp_mep_t structure
 *
 * This function and its friend Qdown for a fifo like ordering to the
 * proccesses that try to get by here.  This is to try and avoid some
 * starvation we had been seeing.
 *
 */
void
Qup(QslpHead_t * qm)
{
	Qsleep_t *Qsp = NULL;
	struct list_head *lst = NULL;

	down(&qm->Qmux);
	if (!list_empty(&qm->IOQ)) {

		lst = &qm->IOQ;
		lst = lst->next;
		MEP_ASSERT(lst != &qm->IOQ,
			   printk
			   ("pid<%6u> Next wrapped head. shouldn't be here.\n",
			    current->pid););
		Qsp = list_entry(lst, Qsleep_t, Qsp_list);

		up(&Qsp->sleeper);	/* wake up the next guy */

	} else
		MEP_ASSERT(qm->queue_depth == 0,
			   printk
			   ("==u<%6u> no items in list, but queue_depth %u\n",
			    current->pid, qm->queue_depth););
	up(&qm->Qmux);

	up(&qm->outer);
}

void
init_Qmux(QslpHead_t * qm)
{
	init_MUTEX(&qm->outer);
	init_MUTEX(&qm->Qmux);
	INIT_LIST_HEAD(&qm->IOQ);
}

/*************************************************************************/
/**
 * show_error - prints out errors in sorta english.
 * @ec: < an errorcode.
 */
STATIC void
show_error(errorcode_t ec)
{
	printk(" Error received: ");
#define CaseError(x) case (x): printk("%s\n",#x); break
	switch (ec) {
		CaseError(Mel_Err_BadSize);
		CaseError(Mel_Err_BidZero);
		CaseError(Mel_Err_NoBid);
		CaseError(Mel_Err_BadSeq);
		CaseError(Mel_Err_BadPbn);
		CaseError(Mel_Pbn_Space_Full);
		CaseError(Mel_You_are_Banned);
		CaseError(Mel_Verify_Failed);
	default:
		printk("Unknown errorcode %d\n", ec);
	}
#undef CaseError
}

/**
 * printbid - prints out a bid in hex.
 * @bid: bid buffer to print.
 */
void
printbid(u8 bid[wreBIDSIZE])
{
	int i;
	printk("0x");
	for (i = 0; i < BIDSIZE; i++)
		printk("%02x", bid[i] & 0xFF);
}

/**
 * pack_bid - convert a lm_lockname into a bid.
 * @ln: < the lockname struct
 * @bid: > bid buffer.
 *
 */
STATIC void
pack_bid(lm_lockname_t * ln, u8 bid[wreBIDSIZE])
{
	bid[0] = ln->ln_type & 0xFF;

	bid[1] = (ln->ln_number >> 56) & 0xFF;
	bid[2] = (ln->ln_number >> 48) & 0xFF;
	bid[3] = (ln->ln_number >> 40) & 0xFF;
	bid[4] = (ln->ln_number >> 32) & 0xFF;
	bid[5] = (ln->ln_number >> 24) & 0xFF;
	bid[6] = (ln->ln_number >> 16) & 0xFF;
	bid[7] = (ln->ln_number >> 8) & 0xFF;
	bid[8] = (ln->ln_number >> 0) & 0xFF;
	bid[9] = 0;		/* oopsy */
}

/**
 * unpack_bid - convert a bid into a lm_lockname.
 * @ln: > the lockname struct
 * @bid: < bid buffer.
 *
 */
STATIC void
unpack_bid(lm_lockname_t * ln, u8 bid[wreBIDSIZE])
{
	ln->ln_type = bid[0];

	ln->ln_number = (u64) bid[1] << 56;
	ln->ln_number |= (u64) bid[2] << 48;
	ln->ln_number |= (u64) bid[3] << 40;
	ln->ln_number |= (u64) bid[4] << 32;
	ln->ln_number |= (u64) bid[5] << 24;
	ln->ln_number |= (u64) bid[6] << 16;
	ln->ln_number |= (u64) bid[7] << 8;
	ln->ln_number |= (u64) bid[8] << 0;
}

/**
 * restartable - is the given errorcode one we can recover from?
 * @err: < an error number.
 *
 * switch that case.  is this something we can do a reconnect and continue
 * from?
 *
 * I think I've the correct list here.
 *
 * Returns: TRUE or FALSE
 */
STATIC int
restartable(int err)
{
	switch (abs(err)) {
	case 0:		/* no error is always restartable. */
	case EPIPE:
	case EAGAIN:
	case ECOMM:
	case ESTRPIPE:
	case ENETRESET:
	case ECONNABORTED:
	case ECONNRESET:
	case ENOTCONN:
	case ESHUTDOWN:
	case ECONNREFUSED:
	case EHOSTUNREACH:
	case ERESTARTSYS:
		return TRUE;
	default:
		return FALSE;
	}
}

/**
 * dtcp_connect - connects us to the server.
 * @smep: <> The details of what we need to know.
 *
 * Pulled out the routines for connecting so we can attpemt reconnects.
 *
 * This now loops for ever...
 *
 * Returns: Error value or 0.
 */
STATIC int
dtcp_connect(tcp_mep_t * smep)
{
	int err = 0;
	unsigned int retry_count = 0;

	while (restartable(err)) {
		if (smep->sock) {
			/* kill off what might be there. */
			sock_release(smep->sock);
			smep->sock = NULL;
		}

		err = sock_create(AF_INET, SOCK_STREAM, 0, &smep->sock);
		if (err) {
			printe("%u: Failed to create socket (%d)\n",
			       retry_count, err);
		} else {
			err =
			    smep->sock->ops->connect(smep->sock,
						     (struct sockaddr *) &smep->
						     adr,
						     sizeof(struct sockaddr_in),
						     0);
			if (err) {
				if ((retry_count % 1000) == 0) {
					printe("%u: Failed to connect to"
					       " server (%d)\n", retry_count,
					       err);
				}
			} else {
				printi("%u: Connected to server\n",
				       retry_count);
				err = 0;
				/* if here, we drop out with err == 0. */
				break;
			}
		}

		retry_count++;

		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(smep->retry_pause*HZ);
	}

	if (err != 0)
		printe
		    ("%u:Unable to reconnect to server. Expect hell to break loose.\n",
		     retry_count);
	return err;
}

/**
 * dtcp_init - Starts up a connection to the dmep server.
 * @mep: <> The place that holds all we know.
 * @args: < string contating whatever we may need to get started.
 *
 * wonder if it is safe to munge that string....
 *
 * Returns: <0: Errorcode 0: Ok.
 */
STATIC int
dtcp_init(memexp_t * mep, char *args)
{
	tcp_mep_t *smep;
	char *c, *p;
	int err = 0;

	mep->ll_memexp_data = kmalloc(sizeof(tcp_mep_t), GFP_NOFS);
	if (!mep->ll_memexp_data) {
		err = -ENOMEM;
		goto exit_dtcp_init;
	}
	memset(mep->ll_memexp_data, 0, sizeof(tcp_mep_t));
	smep = mep->ll_memexp_data;

	/* get ip and port out of the arguments. */
	c = strstr(args, ":");
	if (!c) {
		printe("Bad argument, should be serverip:serverport\n");
		err = -1;
		goto exit_dtcp_init;
	}
	*c++ = '\0';

	/* look for optional retry_pause */
	p = strstr(c, ":");
	if (p) {
		*p++ = '\0';
		smep->retry_pause = atoi(p);
	} else {
		smep->retry_pause = RETRY_PAUSE_DEFAULT;
	}

	smep->adr.sin_family = AF_INET;
	smep->adr.sin_port = cpu_to_be16(atoi(c));
	if (!smep->adr.sin_port) {
		printe("Bad port number (%s) 0x%x\n", c, smep->adr.sin_port);
		err = -1;
		goto exit_dtcp_init;
	}
	if (inet_aton(args, &smep->adr.sin_addr.s_addr) < 0) {
		printe("Bad IP number. (%s) 0x%x\n", args,
		       smep->adr.sin_addr.s_addr);
		err = -1;
		goto exit_dtcp_init;
	}
	cpu_to_be32s(&smep->adr.sin_addr.s_addr);

	smep->sock = NULL;
#if 0
	err = dtcp_connect(smep);
	if (err)
		goto exit_dtcp_init;
#else
	/* Don't loop on mount.
	 * Copying this here is far simplier that overloading the connect
	 * function to handle crap that is only needed at mount time.
	 */
	err = sock_create(AF_INET, SOCK_STREAM, 0, &smep->sock);
	if (err) {
		printe("Failed to create socket (%d)\n", err);
		goto exit_dtcp_init;
	}
	err =
	    smep->sock->ops->connect(smep->sock, (struct sockaddr *) &smep->adr,
				     sizeof(struct sockaddr_in), 0);
	if (err) {
		printe("Failed to connect to server (%d)\n", err);
		goto exit_dtcp_init;
	}
#endif

	init_Qmux(&smep->Qedmux);

	/* Need to do an unban here. */
	{
		struct sock *tsk;
		char ob[sizeof(me_hdr_t) + sizeof(u32)];
		me_hdr_t *hdr = (me_hdr_t *) ob;
		u32 *IP = (u32 *) (ob + sizeof(me_hdr_t));

		hdr->Opcode = Mel_UnBanMe;
		hdr->Key = 4242;
		hdr->length = cpu_to_be16(sizeof(u32));
		tsk = smep->sock->sk;
		// *IP = tsk->rcv_saddr;
		*IP = smep->adr.sin_addr.s_addr;

#ifdef DEBUG
		printk("IP: %x\n", *IP);
#endif

		err =
		    data_xmit(DXwrite, smep->sock, ob,
			      sizeof(me_hdr_t) + sizeof(u32));
		if (err)
			goto exit_dtcp_init;

		err = data_xmit(DXread, smep->sock, ob, sizeof(me_hdr_t));
		if (err)
			goto exit_dtcp_init;
		if (hdr->Opcode == Mel_Error) {
			errorcode_t er;
			err =
			    data_xmit(DXread, smep->sock, &er,
				      sizeof(errorcode_t));
			if (err)
				goto exit_dtcp_init;
			show_error(er);
			err = -1;	/* need better value here */
		}
	}

	Timestats_reset(dmep_load_time);
	Timestats_reset(dmep_store_time);
	Timestats_reset(dmep_dump_time);
	Timestats_reset(dmep_release_time);
	Timestats_reset(dmep_modpage_time);
	Timestats_reset(dmep_ban_time);
	return err;

exit_dtcp_init:
	kfree(mep->ll_memexp_data);
	return err;
}

/**
 * dtcp_release - stop.
 * @mep: <> The place that holds all we know.
 *
 * Everything we need to shut ourselves off should be in this single
 * stucture.
 *
 * Returns: Errorvalue.
 */
STATIC int
dtcp_release(memexp_t * mep)
{
	tcp_mep_t *smep = mep->ll_memexp_data;
	int err = 0;

	time_start(dmep_release_time);
	if (!smep) {
		err = -EINVAL;
		goto exit_dtcp_release;
	}

	Qdown(&smep->Qedmux);

	sock_release(smep->sock);

	Qup(&smep->Qedmux);

	kfree(mep->ll_memexp_data);

	time_stop(dmep_release_time);

	time_print(dmep_load_time);
	time_print(dmep_store_time);
	time_print(dmep_dump_time);
	time_print(dmep_release_time);
	time_print(dmep_modpage_time);
	time_print(dmep_ban_time);
      exit_dtcp_release:
	return err;
}

/**
 * dtcp_load - load a memeory buffer
 * @mep: <> The place that holds all we know.
 * @lck: < The lock name that we want to work with.
 * @rdf: > The returned rdf.
 * @buf: > The contents of the memory buffer.
 * @buflen: < how much room we've got.
 *
 * They had better use the correct buf len....
 *
 * Returns: 0: Ok, <>0: Error.
 */
STATIC int
dtcp_load(memexp_t * mep, lm_lockname_t * lck, buf_load_t * rdf,
	  void *buf, u32 buflen)
{
	tcp_mep_t *smep = mep->ll_memexp_data;
	u8 ob[sizeof(me_hdr_t) + sizeof(me_cdb_t)];
	me_hdr_t *hdr = (me_hdr_t *) ob;
	me_cdb_t *mycdb = (me_cdb_t *) (ob + sizeof(me_hdr_t));
	me_hdr_t rhdr;
	me_rdf_t myrdf;
	int err = 0;
	unsigned long begin, end;

	memset(rdf, 0, sizeof(buf_load_t));
	memset(buf, 0, buflen);

	time_start(dmep_load_time);
	if (!smep) {
		err = -EINVAL;
		goto exit_dtcp_load;
	}

	hdr->Opcode = Mel_Load;
	hdr->Key = 4242;	/* usable key? */
	hdr->length = cpu_to_be16(sizeof(me_cdb_t));
	pack_bid(lck, mycdb->bid);
	mycdb->allocLen = cpu_to_be16(sizeof(me_rdf_t) + buflen);

	Qdown(&smep->Qedmux);

	ATTEMPT {		/* Read and Write buffers CANNOT over lap! */
		begin = jiffies;

		err = data_xmit(DXwrite, smep->sock, ob,
				sizeof(me_hdr_t) + sizeof(me_cdb_t));
		if (err)
			RETRY;

		err = data_xmit(DXread, smep->sock, &rhdr, sizeof(me_hdr_t));
		if (err)
			RETRY;

		end = jiffies;
		if ((end - begin) >= (15 * HZ))
			printk("dmep_tcp load delay warning: %u, %u, %u\n",
			       (unsigned int) begin, (unsigned int) end,
			       (unsigned int) (end - begin));

		if (rhdr.Key != 4242)
			printe("===> Header Key failure. %u\n", rhdr.Key);
		if (rhdr.Opcode == Mel_Error) {
			errorcode_t er;
			err =
			    data_xmit(DXread, smep->sock, &er,
				      sizeof(errorcode_t));
			if (err)
				RETRY;
			show_error(er);
			err = (int) er;
		} else {
			if (be16_to_cpu(rhdr.length) !=
			    sizeof(me_rdf_t) + buflen)
				printe
				    ("AAAAHHH!!!! load got %u, we wanted %u!!!\n",
				     be16_to_cpu(rhdr.length),
				     (unsigned int) sizeof(me_rdf_t) + buflen);
			err =
			    data_xmit(DXread, smep->sock, &myrdf,
				      sizeof(me_rdf_t));
			if (err)
				RETRY;
			err = data_xmit(DXread, smep->sock, buf, buflen);
			if (err)
				RETRY;

			rdf->inuse = myrdf.InUse;
			rdf->full = myrdf.fullness;
			rdf->seq = be64_to_cpu(myrdf.seq);
			rdf->pbn = be64_to_cpu(myrdf.pbn);
		}

	}
	RESTART {
		err = dtcp_connect(smep);
		if (err)
			break;	/* cannot recover from this, return errcode */
	}
	ENDATTEMPT;

	Qup(&smep->Qedmux);

      exit_dtcp_load:
	time_stop(dmep_load_time);

	return err;
}

/**
 * dtcp_verify - does the verify call.
 * @smep: <
 * @pbn: <
 *
 * THE ONLY PLACE THIS FUNCTION SHOULD EVER BE CALLED FROM IS THE RESTART
 * BLOCK IN DTCP_STORE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 *
 * Returns: 0: ok, <0: err, >0: Verify failure
 */
STATIC int
dtcp_verify(tcp_mep_t * smep, u64 pbn)
{
	char sob[sizeof(me_hdr_t) + sizeof(u64)];
	me_hdr_t *thdr = (me_hdr_t *) sob;
	u64 *pb = (u64 *) (sob + sizeof(me_hdr_t));
	int err = 0;

	thdr->Opcode = Mel_Verify;
	thdr->Key = 4242;
	thdr->length = cpu_to_be16(sizeof(u64));
	*pb = cpu_to_be64(pbn);

	err = data_xmit(DXwrite, smep->sock, sob,
			sizeof(me_hdr_t) + sizeof(u64));
	if (err)
		return err;

	err = data_xmit(DXread, smep->sock, thdr, sizeof(me_hdr_t));
	if (err)
		return err;
	if (thdr->Opcode == Mel_Error) {
		errorcode_t ec;
		err = data_xmit(DXread, smep->sock, &ec, sizeof(errorcode_t));
		if (err)
			return err;
#if 0
		printk("==> Verify Failed\n");
#endif
		show_error(ec);
		err = (int) ec;	/* random positive */
	} else {
		/* ok, means this store succeded. */
#if 0
		printk("==> Verify succeded.\n");
#endif
		err = 0;
	}
	return err;
}

/**
 * dtcp_store - stores a memory buffer.
 * @mep: <> The place that holds all we know.
 * @lck: < The lock name that we want to work with.
 * @rdf: < The seq and pbn for this store.
 * @buf: < The contents of the memory buffer.
 * @buflen: < how big is buf
 *
 * Returns: 0: Ok, <>0: Error
 */
STATIC int
dtcp_store(memexp_t * mep, lm_lockname_t * lck, buf_store_t * rdf,
	   void *buf, u32 buflen)
{
	tcp_mep_t *smep = mep->ll_memexp_data;
	int err = 0;
	unsigned int buffysize = 0;
	char *buffy = NULL, *data = NULL;
	me_hdr_t *hdr = NULL;
	me_hdr_t rhdr;
	me_cdb_t *mycdb = NULL;
	me_rdf_t *myrdf = NULL;
	unsigned long begin, end;

	time_start(dmep_store_time);
	if (!smep) {
		err = -EINVAL;
		goto exit_dtcp_store;
	}

	buffysize =
	    sizeof(me_hdr_t) + sizeof(me_cdb_t) + sizeof(me_rdf_t) + buflen;
	buffy = kmalloc(buffysize, GFP_NOFS);
	if (!buffy) {
		err = -ENOMEM;
		goto exit_dtcp_store;
	}

	hdr = (me_hdr_t *) buffy;
	mycdb = (me_cdb_t *) (buffy + sizeof(me_hdr_t));
	myrdf = (me_rdf_t *) (buffy + sizeof(me_hdr_t) + sizeof(me_cdb_t));
	data = buffy + sizeof(me_hdr_t) + sizeof(me_cdb_t) + sizeof(me_rdf_t);

	hdr->Opcode = Mel_Store;
	hdr->Key = 4242;
	hdr->length = cpu_to_be16(buffysize - sizeof(me_hdr_t));

	pack_bid(lck, mycdb->bid);
	mycdb->allocLen = cpu_to_be16(sizeof(me_rdf_t) + buflen);

	myrdf->InUse = rdf->inuse;
	myrdf->fullness = 0;
	myrdf->seq = cpu_to_be64(rdf->seq);
	myrdf->pbn = cpu_to_be64(rdf->pbn);

	memcpy(data, buf, buflen);

#if 0
	if (!rdf->inuse && lck->ln_type != LM_TYPE_NOPEN) {
		printk(" freeing buffer pbn:%" FMT_64 "u seq:%" FMT_64 "u ",
		       rdf->pbn, rdf->seq);
		printbid(mycdb->bid);
		printk("\n");
	}
#endif

	Qdown(&smep->Qedmux);

	ATTEMPT {
		err = 0;
#if 0
		if (crap)
			printk("Top of store attempt loop\n");
#endif
		begin = jiffies;

		err = data_xmit(DXwrite, smep->sock, buffy, buffysize);
		if (err)
			RETRY;

		err = data_xmit(DXread, smep->sock, &rhdr, sizeof(me_hdr_t));
		if (err)
			RETRY;

		end = jiffies;
		if ((end - begin) >= (15 * HZ))
			printk("dmep_tcp load delay warning: %u, %u, %u\n",
			       (unsigned int) begin, (unsigned int) end,
			       (unsigned int) (end - begin));

		if (rhdr.Opcode == Mel_Error) {
			errorcode_t er;
			err =
			    data_xmit(DXread, smep->sock, &er,
				      sizeof(errorcode_t));
#if 0
			if (err)
				RETRY;	/* who cares at this point? */
#endif
			/* If we cannot get the error byte, big  deal. Return the
			 * failure, and the next load will fix things.
			 * (close reopen has a nice way of fixing buffers.)
			 */
			if (!err && er != Mel_Err_BadSeq
			    && er != Mel_Err_BadPbn) {
				printe(" For pbn:%" FMT_64 "u seq:%" FMT_64
				       "u ", rdf->pbn, rdf->seq);
				show_error(er);
			}
			err = (int) er;	/* need better value here */
		} else {
			if (rhdr.length != 0)
				printe
				    ("AAAAHHH!!!! store got a %u, we wanted 0!!!!!!\n",
				     be16_to_cpu(rhdr.length));
			/* No data other than the Mel_Ok header is sent. (unless there was an
			 * error of course.)  So this should be all we need.
			 */
			err = 0;
		}
	}
	RESTART {
		err = dtcp_connect(smep);
		if (err)
			break;	/* cannot recover from this, return errcode */
#if 0
		printk("Gonna verify pbn: %" FMT_64 "u bid: ", rdf->pbn);
		printbid(mycdb->bid);
		printk("\n");
		printk(" InUse: %u\n", rdf->inuse);
#endif
		err = dtcp_verify(smep, rdf->pbn);
		if (err == 0) {
			/* Verify succeded. we're done */
#if 0
			printk("==>Good Verify.\n");
#endif
			break;
		} else if (err < 0) {
#if 0
			printk("==> Verify.\n");
#endif
			RETRY;
		} else {	/* otherwise verify failed. jump to top of ATTEMPT */
#if 0
			printk("==> Gotta do it again.\n");
#endif
		}
#if 0
		printk("going to resend store\n");
		crap = 1;
#endif
	}
	ENDATTEMPT;

	Qup(&smep->Qedmux);
      exit_dtcp_store:
#if 0
	if (err) {
		printk("=>Store returning %d on bid ", err);
		printbid(mycdb->bid);
		printk("\n");
	}
#endif
	kfree(buffy);
	time_stop(dmep_store_time);
	return err;
}

/**
 * dtcp_modepg - the modepage.
 * @mep: <> The place that holds all we know.
 * @set: > Value ignored.
 * @numbufs: <> The number of buffers you want.
 * @datasize: <> The size of each memory buffer.
 *
 * Get the size of each buffer, and the number available.
 *
 * Returns: Errorvalue
 */
STATIC int
dtcp_modepg(memexp_t * mep, u8 set, uint64 * numbufs, uint32 * datasize)
{
	tcp_mep_t *smep = mep->ll_memexp_data;
	char ob[sizeof(me_hdr_t) + sizeof(modepage_t)];
	char ib[sizeof(me_hdr_t) + sizeof(modepage_t)];
	me_hdr_t *hdr = (me_hdr_t *) ob;
	me_hdr_t *ihdr = (me_hdr_t *) ib;
	modepage_t *mp = (modepage_t *) (ib + sizeof(me_hdr_t));
	int err = 0;

	time_start(dmep_modpage_time);
	memset(ob, 0, sizeof(me_hdr_t) + sizeof(modepage_t));
#if 0
	hdr->Opcode = (set == 0) ? Mel_ModePgR : Mel_ModePgW;
#endif
	hdr->Opcode = Mel_ModePgR;	/* ogfs client will never do a write. */
	hdr->Key = 4242;
	hdr->length = cpu_to_be16(sizeof(modepage_t));

#if 0				/* not allowed to change this data. */
	mp->numbufs = cpu_to_be64(*numbufs);
	mp->datasize = cpu_to_be32(*datasize);
#endif

	Qdown(&smep->Qedmux);

	ATTEMPT {
		err = data_xmit(DXwrite, smep->sock, ob,
				sizeof(me_hdr_t) + sizeof(modepage_t));
		if (err)
			RETRY;
		err = data_xmit(DXread, smep->sock, ib,
				sizeof(me_hdr_t) + sizeof(modepage_t));
		if (err)
			RETRY;
	}
	RESTART {
		err = dtcp_connect(smep);
		if (err)
			break;	/* cannot recover from this, return errcode */
	}
	ENDATTEMPT;

	Qup(&smep->Qedmux);

	if (be16_to_cpu(ihdr->length) != sizeof(modepage_t))
		printe("AAAAHHH!!!! ModePg got %u, we wanted %u!!!!!!!\n",
		       be16_to_cpu(ihdr->length),
		       (unsigned int) sizeof(modepage_t));

	*numbufs = be64_to_cpu(mp->numbufs);
	*datasize = be32_to_cpu(mp->datasize);

	time_stop(dmep_modpage_time);
	return err;
}

/**
 * dtcp_dump - dump out many buffers.
 * @mep: <> The place that holds all we know.
 * @cookie: <> Thing so we can track where we're at.
 * @heads: <> Where the buffer headers are stuffed.
 * @bufs: <> Where the buffer contents are stuffed.
 * @cnt: > How many buffers the alloced memory can hold.
 * @sze: > The size of one of the buffers.
 * @ret: < How many buffers we found.
 *
 * heads and bufs need to be malloced by the caller.  They had better be
 * big enough too.  heads should be (sizeof(dmp_hdr_t) * cnt) and bufs
 * should be (sze * cnt)
 *
 * Returns: Errorvalue or 0 if no more buffers, or 1 if there are.
 * Returns: 0: no more buffers.
 *          1: more buffers.
 *         <0: Error
 *         >1: Error
 */
STATIC int
dtcp_dump(memexp_t * mep, void **cookie, dump_status_t * heads,
	  u8 * bufs, u32 cnt, u32 sze, u32 * ret)
{
	tcp_mep_t *smep = mep->ll_memexp_data;
	int err = 1;
	u64 *pbn = NULL;
	char ob[sizeof(me_hdr_t) + sizeof(dumpb_cdb_t)];
	me_hdr_t *hdr = (me_hdr_t *) ob;
	me_hdr_t rhdr;
	dumpb_cdb_t *cdb = (dumpb_cdb_t *) (ob + sizeof(me_hdr_t));

	time_start(dmep_dump_time);
	/* printk("Welcome to dtcp_dump (pid:%u)\n", current->pid); */
	if (*cookie != NULL) {
		pbn = (u64 *) * cookie;
#if 0
		printk(" Reusing cookie, pbn %" FMT_64 "u\n", *pbn);
#endif
	} else {
		*cookie = kmalloc(sizeof(u64), GFP_NOFS);
		if (!*cookie)
			return -ENOMEM;
		pbn = *cookie;
		*pbn = 0;
	}

	hdr->Opcode = Mel_Dump;
	hdr->Key = 4242;
	hdr->length = cpu_to_be16(sizeof(dumpb_cdb_t));

	cdb->pbn = cpu_to_be64(*pbn);
	cdb->want = cpu_to_be32(cnt);

	Qdown(&smep->Qedmux);

	ATTEMPT {
		err = data_xmit(DXwrite, smep->sock, ob,
				sizeof(me_hdr_t) + sizeof(dumpb_cdb_t));
		if (err)
			RETRY;

		err = data_xmit(DXread, smep->sock, &rhdr, sizeof(me_hdr_t));
		if (err)
			RETRY;
		if (rhdr.Opcode == Mel_Error) {
			errorcode_t er;
			err =
			    data_xmit(DXread, smep->sock, &er,
				      sizeof(errorcode_t));
			if (err)
				RETRY;
			show_error(er);
			err = (int) er;	/* fudge hell if Mel_Err_BadSize, oh well */
			*ret = 0;
		} else {
			u32 rcnt, i;
#if 0
			printk("Dump is receiving %u bytes\n",
			       be16_to_cpu(hdr->length));
#endif
			{
				char stop[9];
				err = data_xmit(DXread, smep->sock, stop, 8);
				if (strncmp("DUMPSTRT", stop, 8) != 0) {
					dump_buffer(stop, 8);
					panic("Bad magic in dump header!!\n");
				}
			}
			err = data_xmit(DXread, smep->sock, &rcnt, sizeof(u32));
			if (err)
				RETRY;
			be32_to_cpus(&rcnt);
#if 0
			printk(" Gonna read %u dump elements.\n", rcnt);
#endif
			for (i = 0; i < rcnt; i++) {
				dump_element_t delm;
				err =
				    data_xmit(DXread, smep->sock, &delm,
					      sizeof(dump_element_t));
				if (err)
					RETRY;
				unpack_bid(&heads[i].lockname, delm.bid);
				heads[i].seq = be64_to_cpu(delm.seq);
				heads[i].pbn = be64_to_cpu(delm.pbn);
#if 0
				printk("dump ele: pbn:%" FMT_64 "u seq:%" FMT_64
				       "u ln_num:%" FMT_64 "u" " ln_type:%u\n",
				       heads[i].pbn, heads[i].seq,
				       heads[i].lockname.ln_number,
				       heads[i].lockname.ln_type);
#endif

				err = data_xmit(DXread, smep->sock, bufs, sze);
				if (err)
					RETRY;
				bufs += sze;

				*pbn = heads[i].pbn;
			}
			*pbn = *pbn + 1;

			*ret = rcnt;	/* return how many there was. */

			if (rcnt < cnt) {
				/* we got fewer than what was asked for so there must be no more. */
				err = 0;
				kfree(*cookie);
#if 0
				printk("Freed cookie.\n");
#endif
				*cookie = NULL;
			} else
				err = 1;
		}
	}
	RESTART {
		err = dtcp_connect(smep);
		if (err)
			break;	/* cannot recover from this, return errcode */
	}
	ENDATTEMPT;

	Qup(&smep->Qedmux);

	time_stop(dmep_dump_time);
	return err;
}

/**
 * dtcp_stomith
 * @mep: <> The place that holds all we know.
 * @ipaddr: < ip address of client we want to ban
 *
 * GAH.  can't we call this a ban here?  stomith makes no sense at all with
 * what will be done.
 *
 * Returns: Errorvalue
 */

STATIC int
dtcp_stomith(memexp_t * mep, uint32 ipaddr)
{
	tcp_mep_t *smep = mep->ll_memexp_data;
	int err = 0;
	char ob[sizeof(me_hdr_t) + sizeof(u32)];
	me_hdr_t *hdr = (me_hdr_t *) ob;
	u32 *IP = (u32 *) (ob + sizeof(me_hdr_t));

	time_start(dmep_ban_time);
	hdr->Opcode = Mel_BanReq;
	hdr->Key = 4242;
	hdr->length = cpu_to_be16(sizeof(u32));

	*IP = cpu_to_be32(ipaddr);

#if 0
	printk("IPparam %x\n", ipaddr);
	printk("Banning %x\n", *IP);
#endif

	Qdown(&smep->Qedmux);

	/* Since the server will actually only ban an ip once, this can repeat
	 * itself.
	 */
	ATTEMPT {
		/* FIXME Ack! The recv needs to put the data somewhere other than
		 * into ob.  Since that over writes the stuff I'm trying to send.
		 */
		err =
		    data_xmit(DXwrite, smep->sock, ob,
			      sizeof(me_hdr_t) + sizeof(u32));
		if (err)
			RETRY;

		err = data_xmit(DXread, smep->sock, ob, sizeof(me_hdr_t));
		if (err)
			RETRY;
		if (hdr->Opcode == Mel_Error) {
			errorcode_t er;
			err =
			    data_xmit(DXread, smep->sock, &er,
				      sizeof(errorcode_t));
			if (err)
				RETRY;
			show_error(er);
			err = -1;	/* need better value here */
		}
	}
	RESTART {
		err = dtcp_connect(smep);
		if (err)
			break;	/* cannot recover from this, return errcode */
	}
	ENDATTEMPT;

	Qup(&smep->Qedmux);

	time_stop(dmep_ban_time);
	return err;
}

/**
 * dx_timeout - send a sig to the caller.
 * @ptr: < task struct pointer to the calling process
 *
 */
void
dx_timeout(unsigned long ptr)
{
	printi("Ack! data_xmit timed out.  Lets try to recover.\n");
	send_sig(SIGVTALRM, (struct task_struct *) ptr, 0);
}

/**
 * data_xmit - send or recieve data
 * @send: < What kind of xmit this is. DXread, DXpeek, DXwrite
 * @sock: < Pointer to the socket stucture that we're to xmit on.
 * @buf: <> Data pointer to send or receive.
 * @size: < How much data to move.
 *
 * Basic wrapper that moves and protects those bits of the kernel that we
 * don't want fsck up.
 *
 * Note: This now has a timeout.  Such that ops must take less than the
 * timeout value.  Currently this is hard coded to 20 seconds.  The only
 * time this should ever be an issue is when the server has failed in a
 * maner that the lower tcp/ip levels don't detect it.
 *
 * Returns: <0: Error code 0: ok
 */
static int
data_xmit(DXtype send, struct socket *sock, void *buf, u32 size)
{
	mm_segment_t oldfs;
	int result = -EIO;
	struct msghdr msg;
	struct iovec iov;
	unsigned long flags;
	sigset_t oldset;
	struct timer_list timer;
	unsigned long begin, end;

	/* set up timout */
	init_timer(&timer);
	timer.function = dx_timeout;
	timer.data = (unsigned long) current;
	timer.expires = jiffies + (20 * HZ);
	add_timer(&timer);

	/* move into transmit */
	oldfs = get_fs();
	set_fs(get_ds());

	spin_lock_irqsave(&current->sighand->siglock, flags);
	oldset = current->blocked;
#if 0
	sigfillset(&current->blocked);
#endif
	siginitsetinv(&current->blocked, sigmask(SIGVTALRM));
	recalc_sigpending();
	spin_unlock_irqrestore(&current->sighand->siglock, flags);

	do {
		iov.iov_base = buf;
		iov.iov_len = (u16) size;
		msg.msg_name = NULL;
		msg.msg_namelen = 0;
		msg.msg_iov = &iov;
		msg.msg_iovlen = 1;
		msg.msg_control = NULL;
		msg.msg_controllen = 0;
		msg.msg_namelen = 0;
		msg.msg_flags = 0;
		/* lookout. len param in socks is int.  But we're doing a long.
		 * could show problems here. (though would think the loop would
		 * handle it.)
		 */

		begin = jiffies;

		if (send == DXwrite)
			result = sock_sendmsg(sock, &msg, (u16) size);
		else if (send == DXread)
			result =
			    sock_recvmsg(sock, &msg, (u16) size, MSG_WAITALL);
		else if (send == DXpeek)
			result = sock_recvmsg(sock, &msg, (u16) size, MSG_PEEK);

		end = jiffies;
		if ((end - begin) >= (10 * HZ))
			printk
			    ("data_xmit send/recv delay warning: %u, %u, %u\n",
			     (unsigned int) begin, (unsigned int) end,
			     (unsigned int) (end - begin));

		if (result <= 0) {
			printe("%s - sock=%p at buf=%p, size=%u returned %d.\n",
			       DXstrs[send], sock, buf, size, result);
			break;
		}
		size -= (u32) result;
		buf += result;
		if (size != 0)
			printk
			    ("DMEPTCP WARN! packet fragment! tell conrad if you see this! %d\n",
			     size);
	} while (size > 0);

	spin_lock_irqsave(&current->sighand->siglock, flags);
	current->blocked = oldset;
	recalc_sigpending();
	if (result == -ERESTARTSYS)
		flush_signals(current);	/* otherwise we'll loop */
	spin_unlock_irqrestore(&current->sighand->siglock, flags);

	set_fs(oldfs);

	/* shutdown timer. */
	del_timer(&timer);

	/* less than 0: error, return error code
	 * equal to 0 : xmit failed, return EAGAIN
	 * greaterthan: success, return 0
	 */
	return (result < 0) ? result : ((result == 0) ? -EAGAIN : 0);
}

/* function lookup sturcture. */
struct ll_memexp_ops ll_dmep_tcp_ops = {
	load:dtcp_load,
	store:dtcp_store,
	dump:dtcp_dump,
	modepg:dtcp_modepg,
	init:dtcp_init,
	release:dtcp_release,
	stomith:dtcp_stomith
};

/* vim:ai cin et sw=3 ts=3
 */
