/*
 *
 *    Copyright 1999-2001 Sistina Software, Inc.
 *    Portions Copyright 2001-2003 The OpenGFS Project
 *    Portions Copyright 2003 Brian Jackson
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 *    This file was maintained by:
 *      David C. Teigland <teigland@sistina.com>
 *
 */

#include "config.h"

#include <linux/module.h>
#include <linux/slab.h>

#include "memexp.h"
#include "ll_memexp.h"
#include "lvbops.h"

/**
 * lsm_lock_in
 * @lkp:
 * @buf:
 *
 */
void
lsm_lock_in(lsm_lock_t * lkp, char *buf)
{
	lsm_lock_t *blkp = (lsm_lock_t *) buf;

	/* takes care of all the byte fields at once */
	memcpy(lkp, blkp, sizeof(lsm_lock_t));

	lkp->expired_holders = be16_to_cpu(blkp->expired_holders);
	lkp->live_holders = be16_to_cpu(blkp->live_holders);
	lkp->lvb_holders = be16_to_cpu(blkp->lvb_holders);
}

/**
 * lsm_lock_out
 * @lkp:
 * @buf:
 *
 */
void
lsm_lock_out(lsm_lock_t * lkp, char *buf)
{
	lsm_lock_t *blkp = (lsm_lock_t *) buf;

	/* takes care of all the byte fields at once */
	memcpy(blkp, lkp, sizeof(lsm_lock_t));

	blkp->expired_holders = cpu_to_be16(lkp->expired_holders);
	blkp->live_holders = cpu_to_be16(lkp->live_holders);
	blkp->lvb_holders = cpu_to_be16(lkp->lvb_holders);
}

/**
 * me_lock_print
 * @lkp:
 *
 */
static void
me_lock_print(me_lock_t * mlk)
{
	printk("lock number %" FMT_64 "u type %u\n", mlk->lockname.ln_number,
	       mlk->lockname.ln_type);
}

/**
 * lsm_lock_print
 * @lkp:
 *
 */
void
lsm_lock_print(lsm_lock_t * lkp)
{
	int i;

	printk("lock buffer\n");
	printk("  state:           %u\n", lkp->state);
	printk("  conv:            %u\n", lkp->conv);
	printk("  lvb_perm:        %u\n", lkp->lvb_perm);
	printk("  lvb_lock:        %u\n", lkp->lvb_lock);
	printk("  expired_holders: %u\n", lkp->expired_holders);
	printk("  live_holders:    %u\n", lkp->live_holders);
	printk("  lvb_holders:     %u\n", lkp->lvb_holders);

	printk("  holders: ");
	for (i = 0; i < MAX_CLIENTS; i++)
		if (test_holder(i, lkp))
			printk("%u ", i);
	printk("\n");

	printk("  lvb: ");
	for (i = 0; i < MAX_CLIENTS; i++)
		if (test_lvb(i, lkp))
			printk("%u ", i);
	printk("\n");

	printk("  expired: ");
	for (i = 0; i < MAX_CLIENTS; i++)
		if (test_expired(i, lkp))
			printk("%u ", i);
	printk("\n");

	printk("  conversion: ");
	for (i = 0; i < MAX_CLIENTS; i++)
		if (test_conversion(i, lkp))
			printk("%u ", i);
	printk("\n");
}

/**
 * verify_pre_state - verify that our current hold/in-use state on lock is
 *     valid for use as starting point for requested action
 * @mep:
 * @action:
 * @bld:
 * @llk:
 *
 */
static int
verify_pre_state(me_lock_t * mlk, int action, buf_load_t * bld,
		 lsm_lock_t * llk)
{
	memexp_t *mep = mlk->mep;
	uint32 cid = mep->cid;

	/* if the buffer is not in use I'd better not think I have the
	   lock in some held state */

	if (!bld->inuse) {
		MEP_ASSERT(action == ME_ACT_U2E || action == ME_ACT_U2D ||
			   action == ME_ACT_U2S, me_lock_print(mlk);
			   lsm_lock_print(llk);
		    );
		return 0;
	}

	switch (action) {
	case ME_ACT_U2E:
	case ME_ACT_U2D:
	case ME_ACT_U2S:
		MEP_ASSERT(!test_holder(cid, llk), me_lock_print(mlk);
			   lsm_lock_print(llk); printk("action: %d\n", action);
		    );
		break;

	case ME_ACT_E2D:
	case ME_ACT_E2S:
	case ME_ACT_E2U:
		MEP_ASSERT(test_holder(cid, llk), me_lock_print(mlk);
			   lsm_lock_print(llk); printk("action: %d\n", action);
		    );
		MEP_ASSERT(llk->state == ME_ST_EXCLUSIVE, me_lock_print(mlk);
			   lsm_lock_print(llk); printk("action: %d\n", action);
		    );
		break;

	case ME_ACT_D2E:
	case ME_ACT_D2S:
	case ME_ACT_D2U:
		MEP_ASSERT(test_holder(cid, llk), me_lock_print(mlk);
			   lsm_lock_print(llk); printk("action: %d\n", action);
		    );
		MEP_ASSERT(llk->state == ME_ST_DEFERRED, me_lock_print(mlk);
			   lsm_lock_print(llk); printk("action: %d\n", action);
		    );
		break;

	case ME_ACT_S2E:
	case ME_ACT_S2D:
	case ME_ACT_S2U:
		MEP_ASSERT(test_holder(cid, llk), me_lock_print(mlk);
			   lsm_lock_print(llk); printk("action: %d\n", action);
		    );
		MEP_ASSERT(llk->state == ME_ST_SHARED, me_lock_print(mlk);
			   lsm_lock_print(llk); printk("action: %d\n", action);
		    );
		break;

	case ME_ACT_X2U:
		MEP_ASSERT(test_holder(cid, llk), me_lock_print(mlk);
			   lsm_lock_print(llk); printk("action: %d\n", action);
		    );
		break;

	default:
		MEP_ASSERT(0, printk("invalid action %u\n", action);
		    );
	}

	return 0;
}

/**
 * check_compat - determine compatability of action with current lock state
 * @mep:
 * @action:
 * @bld:
 * @llk:
 * @force: acquire lock even if it has expired holders
 *
 * possible actions: U2E U2D U2S E2D E2S D2E D2S S2E S2D
 * possible lock states: buffer unused, U, E, D, S
 * no error checking because verify_pre_state should be called before this
 *
 * Returns: 1 (compatable), 0 (not-compat, try conv), -1 (not-compat)
 */
static int
check_compat(memexp_t * mep, int action, buf_load_t * bld,
	     lsm_lock_t * llk, int force)
{
	if (!bld->inuse)
		return 1;

	if (llk->expired_holders && !force)
		return -1;

	switch (action) {
	case ME_ACT_U2E:

		if ((llk->state == ME_ST_UNLOCKED) &&
		    (llk->conv == NO_CONV || test_conversion(mep->cid, llk)))
			return 1;
		goto try_conv;

	case ME_ACT_D2E:
	case ME_ACT_S2E:

		if (((llk->conv == NO_CONV) || test_conversion(mep->cid, llk))
		    && (llk->live_holders == 1))
			return 1;
		goto try_conv;

	case ME_ACT_U2D:

		if ((llk->state != ME_ST_EXCLUSIVE) &&
		    (llk->state != ME_ST_SHARED) &&
		    (llk->conv == NO_CONV || test_conversion(mep->cid, llk)))
			return 1;
		goto try_conv;

	case ME_ACT_S2D:
	case ME_ACT_D2S:

		if (((llk->conv == NO_CONV) || test_conversion(mep->cid, llk))
		    && (llk->live_holders == 1))
			return 1;
		goto try_conv;

	case ME_ACT_U2S:

		if ((llk->state != ME_ST_EXCLUSIVE) &&
		    (llk->state != ME_ST_DEFERRED) &&
		    (llk->conv == NO_CONV || test_conversion(mep->cid, llk)))
			return 1;
		goto try_conv;

	case ME_ACT_E2D:
	case ME_ACT_E2S:

		return 1;

	default:
		MEP_ASSERT(0, printk("invalid action %d\n", action);
		    );
	}

      try_conv:

	if (llk->conv == NO_CONV)
		return 0;

	return -1;
}

/**
 * new_me_state - the state to which action is moving
 * @action:
 *
 */
static int
new_me_state(int action)
{
	switch (action) {
	case ME_ACT_U2E:
	case ME_ACT_D2E:
	case ME_ACT_S2E:
		return ME_ST_EXCLUSIVE;

	case ME_ACT_U2D:
	case ME_ACT_E2D:
	case ME_ACT_S2D:
		return ME_ST_DEFERRED;

	case ME_ACT_U2S:
	case ME_ACT_E2S:
	case ME_ACT_D2S:
		return ME_ST_SHARED;

	default:
		MEP_ASSERT(0, printk("invalid action %d\n", action);
		    );
	}
	return -1;
}

/**
 * update_llk_state - make changes to lock state for the current transition
 * @mep:
 * @mlk:
 * @action:
 * @llk:
 *
 */
static void
update_llk_state(memexp_t * mep, me_lock_t * mlk, int action, lsm_lock_t * lkp)
{
	lkp->state = new_me_state(action);
	set_holder(mep->cid, lkp);

	if (lkp->conv && test_conversion(mep->cid, lkp)) {
		lkp->conv = NO_CONV;
		clear_conversion(mep->cid, lkp);
	}

	if ((action == ME_ACT_E2D || action == ME_ACT_E2S) &&
	    (mlk->lvb && (mlk->lvb->lvb_flags & LM_LVB_PERM)))
		lkp->lvb_perm = 1;
}

/**
 * __lsm_lock - handle lock operations which end in some locked state (or NOP)
 * @action: anything ending in E, D or S (or NOP)
 * @mlk: memexp lock struct
 * @status: results of lsm lock operation
 * @force: acquire lock even if it has expired holders
 *
 * loads lock from lock storage, changes state, stores lock
 *
 */
static int
__lsm_lock(int action, me_lock_t * mlk, lsm_status_t * status, int force)
{
	memexp_t *mep = mlk->mep;
	ll_memexp_ops_t *mops = mep->ll_memexp_ops;
	lsm_t *lsm = (lsm_t *) mep->lsm_data;
	lsm_lock_t llk;
	buf_load_t bld;
	buf_store_t bst;
	char buf[lsm->bufsize];
	int compat, success = 1, error;

	memset(status, 0, sizeof(lsm_status_t));

      retry:

	error = mops->load(mep, &mlk->lockname, &bld, buf, lsm->bufsize);
	if (error)
		return error;

	/* old stuff may be left in an unused buffer */
	if (!bld.inuse)
		memset(buf, 0, lsm->bufsize);

	lsm_lock_in(&llk, buf);

	if (action == ME_ACT_NOP)
		goto out;

	verify_pre_state(mlk, action, &bld, &llk);
	compat = check_compat(mep, action, &bld, &llk, force);

	/* lock state is compat with action - proceed with lock acquisition */

	if (compat > 0) {
		update_llk_state(mep, mlk, action, &llk);
		lsm_lock_out(&llk, buf);
		prepare_store(&bld, &bst);
		error =
		    mops->store(mep, &mlk->lockname, &bst, buf, lsm->bufsize);
		if (error)
			goto retry;
	}

	/* lock state is not compat with action - try to get conversion */

	else if (compat == 0) {
		llk.conv = YES_CONV;
		set_conversion(mep->cid, &llk);
		lsm_lock_out(&llk, buf);
		prepare_store(&bld, &bst);
		error =
		    mops->store(mep, &mlk->lockname, &bst, buf, lsm->bufsize);
		if (error)
			goto retry;
		success = 0;
	}

	/* lock state is not compat or it is and no conversion possible */

	else
		success = 0;

      out:

	status->success = success;
	status->state = llk.state;
	status->conv = llk.conv;
	status->hvconv = (llk.conv && test_conversion(mep->cid, &llk));
	status->full = bld.full;
	status->live_holders = llk.live_holders;
	status->expired_holders = llk.expired_holders;
	memcpy(&status->cid_list, &llk.holders, MAX_CLIENTS / 8);

	return 0;
}

/**
 * lsm_unlock - handle lock operations which end in unlocked state
 * @mep:
 *
 * actions may be: E2U D2U S2U X2U
 *
 * loads lock from lock storage, changes state, stores lock
 */
static int
lsm_unlock(int action, me_lock_t * mlk, lsm_status_t * status)
{
	memexp_t *mep = mlk->mep;
	ll_memexp_ops_t *mops = mep->ll_memexp_ops;
	lsm_t *lsm = (lsm_t *) mep->lsm_data;
	lsm_lock_t llk;
	buf_load_t bld;
	buf_store_t bst;
	char buf[lsm->bufsize];
	int freebuf, error;

	memset(status, 0, sizeof(lsm_status_t));

      retry:

	error = mops->load(mep, &mlk->lockname, &bld, buf, lsm->bufsize);
	if (error)
		return error;

	MEP_ASSERT(bld.inuse,);
	lsm_lock_in(&llk, buf);
	verify_pre_state(mlk, action, &bld, &llk);

	/* update lock state */

	if (action == ME_ACT_E2U &&
	    (mlk->lvb && (mlk->lvb->lvb_flags & LM_LVB_PERM)))
		llk.lvb_perm = 1;

	if (llk.conv && test_conversion(mep->cid, &llk)) {
		llk.conv = NO_CONV;
		clear_conversion(mep->cid, &llk);
	}

	freebuf = 0;
	if (llk.live_holders == 1) {
		llk.state = ME_ST_UNLOCKED;
		if (llk.conv == NO_CONV && llk.lvb_perm == 0 &&
		    llk.expired_holders == 0 && llk.lvb_holders == 0)
			freebuf = 1;
	}

	clear_holder(mep->cid, &llk);
	lsm_lock_out(&llk, buf);
	prepare_store(&bld, &bst);
	if (freebuf)
		bst.inuse = 0;

	error = mops->store(mep, &mlk->lockname, &bst, buf, lsm->bufsize);
	if (error)
		goto retry;

	status->success = 1;
	status->state = llk.state;
	status->conv = llk.conv;
	status->hvconv = (llk.conv && test_conversion(mep->cid, &llk));
	status->full = bld.full;
	status->live_holders = llk.live_holders;
	status->expired_holders = llk.expired_holders;
	memcpy(status->cid_list, llk.holders, (MAX_CLIENTS / 8));

	return 0;
}

/**
 * lsm_lock - general function supporting all lock and unlock operations
 * @action: ME_ACT_*
 * @mlk: memexp lock struct
 * @status: results of lsm lock operation
 * @force: acquire lock even if it's expired
 *
 * This function also handles LVB reading/writing; rules are:
 * 1) Any transition that starts in the exclusive state stores the LVB.
 * 2) Any other transition that ends in a non-unlocked state loads the LVB.
 */
int
lsm_lock(int action, me_lock_t * mlk, lsm_status_t * status, int force)
{
	int error;

	if (mlk->lvb) {
		if (action == ME_ACT_U2E ||
		    action == ME_ACT_U2D ||
		    action == ME_ACT_U2S ||
		    action == ME_ACT_D2E ||
		    action == ME_ACT_D2S ||
		    action == ME_ACT_S2E || action == ME_ACT_S2D) {
			error = __lsm_lock(action, mlk, status, force);
			if (error)
				return error;

			lvb_read(mlk);
			return 0;
		}

		if (action == ME_ACT_E2D || action == ME_ACT_E2S) {
			lvb_write(mlk);
			return __lsm_lock(action, mlk, status, force);
		}

		if (action == ME_ACT_E2U) {
			lvb_write(mlk);
			return lsm_unlock(action, mlk, status);
		}

		if (action == ME_ACT_D2U ||
		    action == ME_ACT_S2U || action == ME_ACT_X2U) {
			return lsm_unlock(action, mlk, status);
		}

		/* ME_ACT_NOP (or some bogus action) is all that's left */

		return __lsm_lock(action, mlk, status, force);
	}

	/* no lvb */

	if (action == ME_ACT_E2U ||
	    action == ME_ACT_D2U ||
	    action == ME_ACT_S2U || action == ME_ACT_X2U) {
		return lsm_unlock(action, mlk, status);
	}

	return __lsm_lock(action, mlk, status, force);
}

/**
 * lsm_drop_conversion
 * @mlk:
 * @status:
 *
 */
int
lsm_drop_conversion(me_lock_t * mlk, lsm_status_t * status)
{
	memexp_t *mep = mlk->mep;
	ll_memexp_ops_t *mops = mep->ll_memexp_ops;
	lsm_t *lsm = (lsm_t *) mep->lsm_data;
	lsm_lock_t llk;
	buf_load_t bld;
	buf_store_t bst;
	char buf[lsm->bufsize];
	int error;

	memset(status, 0, sizeof(lsm_status_t));

      retry:

	error = mops->load(mep, &mlk->lockname, &bld, buf, lsm->bufsize);
	if (error)
		return error;

	if (!bld.inuse)
		memset(buf, 0, lsm->bufsize);

	lsm_lock_in(&llk, buf);

	if (llk.conv && test_conversion(mep->cid, &llk)) {
		llk.conv = NO_CONV;
		clear_conversion(mep->cid, &llk);
		lsm_lock_out(&llk, buf);
		prepare_store(&bld, &bst);
		error =
		    mops->store(mep, &mlk->lockname, &bst, buf, lsm->bufsize);
		if (error)
			goto retry;
	}

	status->success = 1;
	status->state = llk.state;
	status->conv = llk.conv;
	status->hvconv = (llk.conv && test_conversion(mep->cid, &llk));
	status->full = bld.full;
	status->live_holders = llk.live_holders;
	status->expired_holders = llk.expired_holders;
	memcpy(&status->cid_list, &llk.holders, MAX_CLIENTS / 8);

	return 0;
}

/**
 * cleanup_lvb - decide what to do with an lvb during recovery
 * @mep:
 * @excid:
 *
 */
static int
cleanup_lvb(memexp_t * mep, lm_lockname_t * name, uint32 excid)
{
	ll_memexp_ops_t *mops = mep->ll_memexp_ops;
	lsm_t *lsm = (lsm_t *) mep->lsm_data;
	lsm_lock_t llk;
	lm_lockname_t lvbname;
	char buf[lsm->bufsize];
	buf_load_t bld;
	buf_store_t bst;
	int error;

      retry_lock:

	error = mops->load(mep, name, &bld, buf, lsm->bufsize);
	if (error)
		return error;

	MEP_ASSERT(bld.inuse,);

	lsm_lock_in(&llk, buf);

	if (!test_lvb(excid, &llk))
		return 0;

	if ((llk.lvb_perm == 1) || (llk.lvb_holders > 1)) {
		clear_lvb(excid, &llk);
		lsm_lock_out(&llk, buf);
		prepare_store(&bld, &bst);
		error = mops->store(mep, name, &bst, buf, lsm->bufsize);
		if (error)
			goto retry_lock;

		return 0;
	}

	/* if we get here, the excid has the last lvb reference meaning the lvb
	   must be freed.. freeing the lvb requires holding the lvb_lock */

	if (!llk.lvb_lock) {
		llk.lvb_lock = TRUE;
		lsm_lock_out(&llk, buf);
		prepare_store(&bld, &bst);
		error = mops->store(mep, name, &bst, buf, lsm->bufsize);
		if (error)
			goto retry_lock;
	}

	lock2lvb(name, &lvbname);

	error = mops->load(mep, &lvbname, &bld, buf, lsm->bufsize);
	if (error)
		return error;

	MEP_ASSERT(bld.inuse,);
	bst.inuse = 0;
	bst.seq = bld.seq;
	bst.pbn = bld.pbn;
	error = mops->store(mep, &lvbname, &bst, buf, lsm->bufsize);
	if (error)
		MEP_ASSERT(0,);

	/* we've freed lvb buffer, now clear the final lvb ref from lock.
	   cleanup_lock will check into freeing the lock */

      store_final:

	error = mops->load(mep, name, &bld, buf, lsm->bufsize);
	if (error)
		return error;

	lsm_lock_in(&llk, buf);

	MEP_ASSERT(bld.inuse,);
	MEP_ASSERT(test_lvb(excid, &llk),);
	MEP_ASSERT(llk.lvb_lock,);

	clear_lvb(excid, &llk);
	llk.lvb_lock = 0;

	lsm_lock_out(&llk, buf);
	prepare_store(&bld, &bst);
	error = mops->store(mep, name, &bst, buf, lsm->bufsize);
	if (error)
		goto store_final;

	return 0;
}

/**
 * cleanup_lock - change state of lock due to recovery
 * @mep:
 * @excid:
 *
 */
static int
cleanup_lock(memexp_t * mep, lm_lockname_t * name, uint32 excid)
{
	ll_memexp_ops_t *mops = mep->ll_memexp_ops;
	lsm_t *lsm = (lsm_t *) mep->lsm_data;
	lsm_lock_t llk;
	char buf[lsm->bufsize];
	buf_load_t bld;
	buf_store_t bst;
	int error;

      retry:

	error = mops->load(mep, name, &bld, buf, lsm->bufsize);
	if (error)
		return error;

	MEP_ASSERT(bld.inuse,);

	lsm_lock_in(&llk, buf);

	if (llk.state == ME_ST_EXCLUSIVE && test_holder(excid, &llk)) {
		MEP_ASSERT(llk.live_holders == 1,);
		set_expired(excid, &llk);
	}

	if (test_conversion(excid, &llk)) {
		llk.conv = 0;
		clear_conversion(excid, &llk);
	}

	clear_holder(excid, &llk);

	if (llk.live_holders == 0)
		llk.state = ME_ST_UNLOCKED;

	/* clearing all that may mean that the lock should now be freed */

	if (llk.live_holders == 0 &&
	    llk.expired_holders == 0 &&
	    llk.lvb_holders == 0 && llk.conv == NO_CONV && llk.lvb_perm == 0)
		bst.inuse = 0;
	else
		bst.inuse = 1;

	bst.seq = bld.seq;
	bst.pbn = bld.pbn;

	lsm_lock_out(&llk, buf);

	error = mops->store(mep, name, &bst, buf, lsm->bufsize);
	if (error)
		goto retry;

	return 0;
}

/**
 * lsm_mark_expired - set dead cid's exclusive locks as "expired",
 *    free LVB if possible, free lock if possible.
 * @mep:
 * @excid: expired cid
 *
 */
int
lsm_mark_expired(memexp_t * mep, uint16 excid)
{
	ll_memexp_ops_t *mops = mep->ll_memexp_ops;
	lsm_t *lsm = (lsm_t *) mep->lsm_data;
	dump_status_t *dhead;
	unsigned char *dbuf;
	lsm_lock_t llk;
	void *cookie = NULL;
	u32 retcount = 0;
	int more, error = -ENOMEM, i, type;

	dbuf = kmalloc(lsm->dump_length, GFP_NOFS);
	if (!dbuf)
		goto out;

	dhead = kmalloc(lsm->dump_count * sizeof(dump_status_t), GFP_NOFS);
	if (!dhead)
		goto out_free;

again:
	/*
	 * Get a chunk of buffers -
	 * 	dhead's give name etc of each buffer.
	 */
	more = mops->dump(mep, &cookie, dhead, dbuf, lsm->dump_count,
			lsm->bufsize, &retcount);

	for (i = 0; i < retcount; i++) {
		type = dhead[i].lockname.ln_type;

		/*
		 * Skip buffer types which memexp uses itself,
		 * ie. aren't used by OGFS.
		 */
		if (type == LM_TYPE_CIDBUF ||
		    type == LM_TYPE_MOUNT ||
		     (type & LM_TYPE_LVB_MASK))
			continue;

		lsm_lock_in(&llk, &dbuf[i * lsm->bufsize]);

		/*
		 * Excid had no reference on this lock.
		 */

		if (!test_holder(excid, &llk) &&
		    !test_conversion(excid, &llk) &&
		    !test_expired(excid, &llk) &&
		    !test_lvb(excid, &llk))
			continue;

		/*
		 * At this point we know that the expired cid had something to
		 * do with this lock, so we need to fix it up.
		 * First we may need to free the lock's lvb if excid had the
		 * last hold on the lvb.
		 */
		error = cleanup_lvb(mep, &dhead[i].lockname, excid);
		if (error)
			goto out_free2;

		cleanup_lock(mep, &dhead[i].lockname, excid);
		if (error)
			goto out_free2;
	}

	if (more)
		goto again;

	error = 0;

out_free2:
	kfree(dhead);
out_free:
	kfree(dbuf);
out:
	return error;
}

/**
 * lsm_reset_expired - dead cid has been recovered so clear all expired refs
 * @mep:
 * @excid:
 *
 */
int
lsm_reset_expired(memexp_t * mep, uint16 excid)
{
	ll_memexp_ops_t *mops = mep->ll_memexp_ops;
	lsm_t *lsm = (lsm_t *) mep->lsm_data;
	dump_status_t *dhead;
	unsigned char *dbuf;
	lsm_lock_t llk;
	char buf[lsm->bufsize];
	buf_load_t bld;
	buf_store_t bst;
	void *cookie = NULL;
	uint32 retcount = 0;
	uint64 tmpseq, tmppbn;
	int more, error = -ENOMEM, i, type;

	dbuf = kmalloc(lsm->dump_length, GFP_NOFS);
	if (!dbuf)
		goto out;

	dhead = kmalloc(lsm->dump_count * sizeof(dump_status_t), GFP_NOFS);
	if (!dhead)
		goto out_free;

again:
	/*
	 * Get a chunk of buffers, dhead's give name etc of each buffer.
	 */
	more = mops->dump(mep, &cookie, dhead, dbuf, lsm->dump_count,
			lsm->bufsize, &retcount);

	for (i = 0; i < retcount; i++) {
		type = dhead[i].lockname.ln_type;

		/*
		 * Skip buffers which aren't locks.
		 */

		if (type == LM_TYPE_CIDBUF ||
		    type == LM_TYPE_MOUNT ||
		     (type & LM_TYPE_LVB_MASK))
			continue;

		lsm_lock_in(&llk, &dbuf[i * lsm->bufsize]);

		/*
		 * Excid had no reference on this lock.
		 */
		if (!test_expired(excid, &llk))
			continue;

		tmpseq = dhead[i].seq;
		tmppbn = dhead[i].pbn;

retry:
		/*
		 * Clear all traces of excid from lock.
		 */
		clear_expired(excid, &llk);

		/*
		 * Clearing expired may mean that the lock should now
		 * be freed.
		 */
		if (llk.state == ME_ST_UNLOCKED &&
		    llk.live_holders == 0 &&
		    llk.expired_holders == 0 &&
		    llk.lvb_holders == 0 &&
		    llk.conv == NO_CONV && llk.lvb_perm == 0)
			bst.inuse = 0;
		else
			bst.inuse = 1;

		bst.seq = tmpseq;
		bst.pbn = tmppbn;

		lsm_lock_out(&llk, buf);

		error = mops->store(mep, &dhead[i].lockname, &bst, buf,
				lsm->bufsize);
		if (error) {
			error = mops->load(mep, &dhead[i].lockname, &bld,
					buf, lsm->bufsize);
			if (error)
				goto out_free2;

			if (!bld.inuse)	/* is this valid? */
				continue;

			lsm_lock_in(&llk, buf);
			tmpseq = bld.seq;
			tmppbn = bld.pbn;
			goto retry;
		}
	}
	if (more)
		goto again;

	error = 0;

out_free2:
	kfree(dbuf);
out_free:
	kfree(dhead);
out:
	return error;
}

/**
 * lsm_init
 * @mep:
 *
 * Returns: 0 if successful. -ENOMEM if out of memory
 */
int
lsm_init(memexp_t *mep)
{
	ll_memexp_ops_t *mops = mep->ll_memexp_ops;
	lsm_t *lsm;
	u64 numbufs;
	u32 datasize;
	int error;

	lsm = kmalloc(sizeof(*lsm), GFP_NOFS);
	if (!lsm)
		return -ENOMEM;
	mep->lsm_data = lsm;

	error = mops->modepg(mep, 0, &numbufs, &datasize);
	if (error < 0) {
		printk("%s:  failed sense on DMEP device/server\n", PROTO_NAME);
		goto fail_free_lsm;
	}

	if (datasize < sizeof(lsm_lock_t)) {
		printk("%s:  buffers are too small on DMEP device\n",
		       PROTO_NAME);
		error = -1;
		goto fail_free_lsm;
	}

	lsm->bufsize = datasize;
	lsm->dump_length = DBUFFER_LENGTH;
	lsm->dump_count = lsm->dump_length / datasize;

	return 0;

fail_free_lsm:
	kfree(lsm);
	return error;
}

/**
 * lsm_shutdown
 * @mep:
 *
 */
int
lsm_shutdown(memexp_t *mep)
{
	lsm_t *lsm = (lsm_t *)mep->lsm_data;

	kfree(lsm);
	return 0;
}

/**
 * lsm_check_state - not used (useful for debugging)
 *
 */
void
lsm_check_state(me_lock_t * mlk, int state, int flags)
{
	memexp_t *mep = mlk->mep;
	ll_memexp_ops_t *mops = mep->ll_memexp_ops;
	lsm_t *lsm = (lsm_t *) mep->lsm_data;
	lsm_lock_t llk;
	buf_load_t bld;
	char buf[lsm->bufsize];
	int error;

	if (flags & LM_FLAG_TRY)
		return;

	error = mops->load(mep, &mlk->lockname, &bld, buf, lsm->bufsize);
	if (error)
		return;

	if (!bld.inuse)
		memset(buf, 0, lsm->bufsize);

	lsm_lock_in(&llk, buf);

	if (state == -1) {
		if (test_holder(mep->cid, &llk))
			error = 1;
		if (test_conversion(mep->cid, &llk))
			error = 1;
	} else {
		if (state != llk.state)
			error = 1;
		if (!test_holder(mep->cid, &llk))
			error = 1;
		if (test_conversion(mep->cid, &llk))
			error = 1;
	}

	if (error) {
		me_lock_print(mlk);
		lsm_lock_print(&llk);
		printk("req state: %d\n", state);
	}
}
