/*
 *
 *    Copyright 1999-2001 Sistina Software, Inc.
 *    Portions Copyright 2003 Brian Jackson
 *    Portions Copyright 2003 The OpenGFS Project
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 *    This file was maintained by:
 *      David C. Teigland <teigland@sistina.com>
 *
 */

//#include "yield.h"

#include <linux/module.h>
#include <linux/slab.h>

#include "memexp.h"
#include "ll_memexp.h"
#include "moduleops.h"
#include "callbacks.h"
#include "cluster_info.h"
#include "client_state.h"
#include "lock_state.h"
#include "lvbops.h"

/* current state selects row, reqested selects column */

static int lock_state_change[16] = {
	-1, ME_ACT_U2E, ME_ACT_U2D, ME_ACT_U2S,
	-1, -1, ME_ACT_E2D, ME_ACT_E2S,
	-1, ME_ACT_D2E, -1, ME_ACT_D2S,
	-1, ME_ACT_S2E, ME_ACT_S2D, -1
};

/**
 * me2lm_state - return LM_ST_X equivalent of ME_ST_X
 * @mestate:
 *
 * The me/lm states are currently the same, so this isn't really needed.
 */

STATIC __inline__ int
me2lm_state(int mestate)
{
	switch (mestate) {
	case ME_ST_UNLOCKED:
		return LM_ST_UNLOCKED;
	case ME_ST_EXCLUSIVE:
		return LM_ST_EXCLUSIVE;
	case ME_ST_DEFERRED:
		return LM_ST_DEFERRED;
	case ME_ST_SHARED:
		return LM_ST_SHARED;
	default:
		return -1;
	}
}

/**
 * lm2me_state - return ME_ST_X equivalent of LM_ST_X
 * @lmstate:
 *
 */

STATIC __inline__ int
lm2me_state(int lmstate)
{
	switch (lmstate) {
	case LM_ST_UNLOCKED:
		return ME_ST_UNLOCKED;
	case LM_ST_EXCLUSIVE:
		return ME_ST_EXCLUSIVE;
	case LM_ST_DEFERRED:
		return ME_ST_DEFERRED;
	case LM_ST_SHARED:
		return ME_ST_SHARED;
	default:
		return -1;
	}
}

/**
 * calculate_action - determine action to get from cur to req state
 * @cur_lmstate: LM_ST_X
 * @req_lmstate: LM_ST_X
 *
 */

STATIC int
calculate_action(unsigned int cur_lmstate, unsigned int req_lmstate)
{
	int cur = lm2me_state(cur_lmstate);
	int req = lm2me_state(req_lmstate);
	int new = lock_state_change[4 * cur + req];

	MEP_ASSERT((new != -1), printk("invalid lock action: cur %d, req %d\n",
				       cur, req);
	    );
	return new;
}

/**
 * in_cid_list - check if cid is set in list
 * @cid:
 * @list:
 *
 */

int
in_cid_list(uint16 cid, char *list)
{
	unsigned char byte;
	unsigned char mask = (1 << (cid % 8));

	byte = list[cid / 8];

	if (byte & mask)
		return TRUE;

	return FALSE;
}

/**
 * local_lmstate - return the state in which this machine holds the lock
 * @mep:
 * @status:
 *
 */

STATIC int
local_lmstate(memexp_t * mep, lsm_status_t * status)
{
	int rs;

	if (in_cid_list(mep->cid, status->cid_list))
		rs = me2lm_state(status->state);
	else
		rs = LM_ST_UNLOCKED;

	return rs;
}

/**
 * test_conv_deadlock - check if the lock may be in conversion deadlock
 * @mep:
 * @status:
 *
 */

STATIC int
test_conv_deadlock(memexp_t * mep, lsm_status_t * status)
{
	if (status->conv &&
	    !status->hvconv && in_cid_list(mep->cid, status->cid_list))
		return TRUE;
	return FALSE;
}

/**
 * failure_allowed
 * @action:
 *
 */

STATIC int
failure_allowed(int action)
{
	if (action == ME_ACT_E2S || action == ME_ACT_E2D)
		return FALSE;
	return TRUE;
}

/**
 * unlock
 * @mep:
 * @cur_mestate:
 * @mlk:
 * @status:
 *
 */

STATIC int
unlock(memexp_t * mep, int cur_mestate, me_lock_t * mlk, lsm_status_t * status)
{
	int error, action;

	switch (cur_mestate) {
	case ME_ST_EXCLUSIVE:
		action = ME_ACT_E2U;
		break;
	case ME_ST_SHARED:
		action = ME_ACT_S2U;
		break;
	case ME_ST_DEFERRED:
		action = ME_ACT_D2U;
		break;
	default:
		action = ME_ACT_X2U;	/* cur_state may be -1 when it doesn't matter */
	}

	error = lsm_lock(action, mlk, status, 0);
	if (error)
		return error;

	MEP_ASSERT(status->success,);
	MEP_ASSERT(!status->hvconv,);
	MEP_ASSERT(!in_cid_list(mep->cid, status->cid_list),);

	return 0;
}

/**
 * drop_conversion
 * @mep:
 * @mlk:
 * @status:
 *
 * Modifies status
 */

STATIC int
drop_conversion(me_lock_t * mlk, lsm_status_t * status)
{
	int error;

	error = lsm_drop_conversion(mlk, status);
	if (error)
		return error;

	MEP_ASSERT(status->success,);
	MEP_ASSERT(!status->hvconv,);

	return 0;
}

/**
 * wait_expired
 * @mep:
 * @mlk:
 * @status:
 *
 * Modifies the values of status
 *
 * Returns:  -EXXXX on error,  1 if canceled, 0 on success
 */

STATIC int
wait_expired(memexp_t * mep, me_lock_t * mlk, lsm_status_t * status)
{
	int error = 0;

	/* poll lock until it's not expired */

	while (1) {
		if (mlk->cancel)
			return 1;

		error = lsm_lock(ME_ACT_NOP, mlk, status, 0);
		if (error)
			return error;

		MEP_ASSERT(status->success,);

		if (!status->expired_holders)
			break;

		printk("%s:  waiting on expired lock: %u, %" FMT_64 "u\n",
		       PROTO_NAME, mlk->lockname.ln_type,
		       mlk->lockname.ln_number);
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(HZ);
	}

	return 0;
}

/**
 * kick_holders - notify all holders of lock to release it
 * @mep:
 * @mlk:
 * @status:
 * @req_state:
 * @action:
 * @tries:
 *
 */

STATIC void
kick_holders(memexp_t * mep, me_lock_t * mlk, lsm_status_t * status,
	     int req_state, int action, int tries)
{
	uint16 cid;
	int error, i, cb = 0;
	uint32 ipaddr = 0;

	switch (req_state) {
	case ME_ST_EXCLUSIVE:
		cb = LM_CB_NEED_E;
		break;
	case ME_ST_DEFERRED:
		cb = LM_CB_NEED_D;
		break;
	case ME_ST_SHARED:
		cb = LM_CB_NEED_S;
		break;
	default:
		MEP_ASSERT(0,
			   printk("cb invalid request state %d\n", req_state);
		    );
	}

	for (cid = 0; cid < MAX_CLIENTS; cid++) {
		if (in_cid_list(cid, status->cid_list) && cid != mep->cid) {
			error = cid2ipaddr(mep, cid, &ipaddr);
			if (error) {
				printk("%s:  bad cid (%u) to ip addr mapping\n",
				       PROTO_NAME, cid);
				printk(" lock holder cids: ");
				for (i = 0; i < MAX_CLIENTS; i++)
					if (in_cid_list(i, status->cid_list))
						printk("%u ", i);
				printk("\n");

				continue;
			}

			send_mess(mep, ipaddr, mlk->lockname.ln_number,
				  mlk->lockname.ln_type, cb, FALSE);

			if (tries % 10000 == 9500)
				printk("%s:  asking for lock (%u, %" FMT_64
				       "u) action %d from %u\n", PROTO_NAME,
				       mlk->lockname.ln_type,
				       mlk->lockname.ln_number, action, cid);
		}
	}
}

/**
 * drop_locks - notify all nodes to release excess locks
 * @mep:
 *
 * Called when lock storage is getting rather full
 *
 */

STATIC void
drop_locks(memexp_t * mep)
{
	uint16 cid;
	uint32 ipaddr = 0;
	int error;

	/*
	 * Two seconds between each cb
	 */
	if ((jiffies-mep->cb_meter) < (2*HZ))
		return;
	mep->cb_meter = jiffies;

	for (cid = 0; cid < MAX_CLIENTS; cid++) {
		if (client_alive(mep, cid)) {
			error = cid2ipaddr(mep, cid, &ipaddr);
			if (error) {
				printk
				    ("%s:  bad cid (%u) to ip addr mapping for drop_locks\n",
				     PROTO_NAME, cid);
				continue;
			}

			printk("%s:  asking cid %u to drop locks\n", PROTO_NAME,
			       cid);

			send_mess(mep, ipaddr, 0, 0, LM_CB_DROPLOCKS, FALSE);
		}
	}
}

/**
 * memexp_get_lock - allocate and init memexp private data structure
 * @lockspace:
 * @name:
 *
 * The G-Lock layer attaches the allocated structure to a G-Lock
 */

STATIC lm_lock_t *
memexp_get_lock(lm_lockspace_t * lockspace, lm_lockname_t * name)
{
	me_lock_t *mlk;

	mlk = kmalloc(sizeof(me_lock_t), GFP_NOFS);
	if (!mlk)
		return mlk;

	memset(mlk, 0, sizeof(me_lock_t));

	mlk->mep = (memexp_t *) lockspace;
	mlk->lockname = *name;

	return ((lm_lock_t *) mlk);
}

/**
 * memexp_put_lock - deallocate memexp private data structure, and any
 *    attached LVB structure
 * @lk:
 *
 */

STATIC void
memexp_put_lock(lm_lock_t * lk)
{
	me_lock_t *mlk = (me_lock_t *) lk;

	MEP_ASSERT(mlk,);

	if (mlk->lvb)
		memexp_unhold_lvb(lk, mlk->lvb);

	kfree(mlk);
}

/**
 * process_results - interpret results of call to lsm_lock()
 * @mlk:
 * @status:
 * @req_state:
 * @ret_state:
 * @action:
 * @flags:
 * @tries:
 *
 * Returns: 0 (Success), 1 (Retry), other (error)
 */

STATIC int
process_results(me_lock_t * mlk, lsm_status_t * status,
		int req_lmstate, int *ret_state, int *action,
		unsigned int flags, int tries)
{
	memexp_t *mep = mlk->mep;
	int error, rs;

	if (status->full > DROPLOCK_THRESHOLD)
		drop_locks(mep);

	/* successful lsm lock operation */

	if (status->success) {
		rs = me2lm_state(status->state);
		MEP_ASSERT(rs == req_lmstate,);
		MEP_ASSERT(in_cid_list(mep->cid, status->cid_list),);
		MEP_ASSERT(!status->hvconv,);

		if (status->expired_holders)
			MEP_ASSERT(flags & LM_FLAG_NOEXP,);

		goto ret_done;
	}

	/* failed lsm lock operation */

	MEP_ASSERT(failure_allowed(*action),
		   printk("action %d failed\n", *action);
	    );

	if (flags & LM_FLAG_TRY) {
		if (test_conv_deadlock(mep, status)) {
			error = unlock(mep, -1, mlk, status);
			if (error)
				return error;

			rs = LM_ST_UNLOCKED;
			goto ret_done;
		}

		rs = local_lmstate(mep, status);
		goto ret_done;
	}

	if (status->expired_holders && !(flags & LM_FLAG_NOEXP))
		goto wait_expired;

	if (test_conv_deadlock(mep, status)) {
		error = unlock(mep, -1, mlk, status);
		if (error)
			return error;

		*action = ME_ACT_U2E;	/* only place retry action is changed */
		goto ret_retry;
	}

	if ((tries < 2) || (tries % 500 == 0))
		kick_holders(mep, mlk, status, req_lmstate, *action, tries);
	goto ret_retry;

      wait_expired:
	error = wait_expired(mep, mlk, status);
	if (error < 0)
		return error;

      ret_retry:
	if (mlk->cancel) {
		rs = local_lmstate(mep, status);
		rs |= LM_OUT_CANCELED;
		goto ret_done;
	}
	return 1;

      ret_done:
	error = 0;
	*ret_state = rs;

	if (status->hvconv)
		error = drop_conversion(mlk, status);

	return error;
}

/**
 * memexp_lock
 * @lk:
 * @cur_lmstate:
 * @req_lmstate:
 * @flags:
 *
 */

STATIC int
memexp_lock(lm_lock_t * lk, unsigned int cur_lmstate,
	    unsigned int req_lmstate, unsigned int flags)
{
	me_lock_t *mlk = (me_lock_t *) lk;
	memexp_t *mep = mlk->mep;
	lsm_status_t status;
	int action, ret_state = 0;
	int error, tries = 0;

	MEP_ASSERT(mep,);

	action = calculate_action(cur_lmstate, req_lmstate);

      retry:

	error = lsm_lock(action, mlk, &status,
			 (flags & LM_FLAG_NOEXP) ? TRUE : FALSE);
	if (error)
		return error;

	error = process_results(mlk, &status, req_lmstate, &ret_state,
				&action, flags, tries);
	if (error < 0)
		return error;
	if (error > 0) {
		tries++;
		//ogfs_yield();
		yield();
		goto retry;
	}

	return ret_state;
}

/**
 * memexp_unlock
 * @lk:
 * @cur_state:
 *
 */

STATIC int
memexp_unlock(lm_lock_t * lk, unsigned int cur_lmstate)
{
	me_lock_t *mlk = (me_lock_t *) lk;
	lsm_status_t status;

	MEP_ASSERT(mlk,);
	MEP_ASSERT(mlk->mep,);

	return unlock(mlk->mep, (int) lm2me_state(cur_lmstate), mlk, &status);
}

/**
 * memexp_cancel - set or reset lock's cancel flag, to stop or allow retries
 * @lk:
 *
 */

STATIC void
memexp_cancel(lm_lock_t * lk, int on)
{
	me_lock_t *mlk = (me_lock_t *) lk;

	MEP_ASSERT(mlk,);

	mlk->cancel = on ? 1 : 0;
}

/**
 * memexp_reset
 * @lk:
 *
 */

STATIC int
memexp_reset(lm_lock_t * lk)
{
	me_lock_t *mlk = (me_lock_t *) lk;
	memexp_t *mep;
	lsm_status_t status;
	int error;

	MEP_ASSERT(mlk,);
	MEP_ASSERT(mlk->mep,);

	mep = mlk->mep;

	error = lsm_lock(ME_ACT_X2U, mlk, &status, 0);
	if (error)
		return error;

	MEP_ASSERT(!in_cid_list(mep->cid, status.cid_list),);

	/* drop conversion if we have it */

	if (status.hvconv)
		error = drop_conversion(mlk, &status);

	return error;
}

struct lm_lockops memexp_ops = {
	proto_name:		PROTO_NAME,
	mount:			memexp_mount,
	others_may_mount:	memexp_others_may_mount,
	unmount:		memexp_unmount,
	get_lock:		memexp_get_lock,
	put_lock:		memexp_put_lock,
	lock:			memexp_lock,
	unlock:			memexp_unlock,
	reset:			memexp_reset,
	cancel:			memexp_cancel,
	hold_lvb:		memexp_hold_lvb,
	unhold_lvb:		memexp_unhold_lvb,
	sync_lvb:		memexp_sync_lvb,
	reset_exp:		memexp_reset_exp,
};
