/*
 *
 *    Copyright 1999-2001 Sistina Software, Inc.
 *    Portions Copyright 2003 Brian Jackson
 *    Portions Copyright 2003 The OpenGFS Project
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 *    This file was maintained by:
 *      David C. Teigland <teigland@sistina.com>
 *
 */

#include "config.h"

#include <linux/module.h>
#include <linux/slab.h>

#include "memexp.h"
#include "ll_memexp.h"
#include "lock_state.h"

/**
 * lock2lvb
 * @lockname:
 * @lvbname:
 *
 */

void
lock2lvb(lm_lockname_t * lockname, lm_lockname_t * lvbname)
{
	lvbname->ln_number = lockname->ln_number;
	lvbname->ln_type = lockname->ln_type | LM_TYPE_LVB_MASK;
}

/**
 * lvb_read
 * @lk:
 *
 */

int
lvb_read(me_lock_t * mlk)
{
	memexp_t *mep = mlk->mep;
	ll_memexp_ops_t *mops = mep->ll_memexp_ops;
	lm_lockname_t lvbname;
	lsm_t *lsm = (lsm_t *) mep->lsm_data;
	char buf[lsm->bufsize];
	buf_load_t bld;
	int error;

	MEP_ASSERT(mlk->lvb,);
	MEP_ASSERT(mlk->lvb->lvb_data,);
	MEP_ASSERT(mlk->lvb->lvb_length == lsm->bufsize,);

	lock2lvb(&mlk->lockname, &lvbname);

	error = mops->load(mep, &lvbname, &bld, buf, lsm->bufsize);
	if (error)
		return error;

	MEP_ASSERT(bld.inuse,);

	memcpy(mlk->lvb->lvb_data, buf, lsm->bufsize);

	return 0;
}

/**
 * lvb_write
 * @lk:
 *
 */

int
lvb_write(me_lock_t * mlk)
{
	memexp_t *mep = mlk->mep;
	ll_memexp_ops_t *mops = mep->ll_memexp_ops;
	lm_lockname_t lvbname;
	lsm_t *lsm = (lsm_t *) mep->lsm_data;
	char buf[lsm->bufsize];
	buf_load_t bld;
	buf_store_t bst;
	int error;

	MEP_ASSERT(mlk->lvb,);
	MEP_ASSERT(mlk->lvb->lvb_data,);
	MEP_ASSERT(mlk->lvb->lvb_length == lsm->bufsize,);

	lock2lvb(&mlk->lockname, &lvbname);

      retry:

	error = mops->load(mep, &lvbname, &bld, buf, lsm->bufsize);
	if (error)
		return error;

	MEP_ASSERT(bld.inuse,);

	memcpy(buf, mlk->lvb->lvb_data, lsm->bufsize);

	prepare_store(&bld, &bst);
	error = mops->store(mep, &lvbname, &bst, buf, lsm->bufsize);
	if (error)
		goto retry;

	return 0;
}

/**
 * memexp_hold_lvb
 * @lk:
 *
 */

lm_lvb_t *
memexp_hold_lvb(lm_lock_t * lk)
{
	me_lock_t *mlk = (me_lock_t *) lk;
	memexp_t *mep = mlk->mep;
	ll_memexp_ops_t *mops = mep->ll_memexp_ops;
	lm_lockname_t lvbname;
	lm_lvb_t *lvb;
	lsm_t *lsm = (lsm_t *) mep->lsm_data;
	char buf[lsm->bufsize];
	lsm_lock_t llk;
	buf_load_t bld;
	buf_store_t bst;
	int error;

	if (mlk->lvb)
		return (mlk->lvb);

	lvb = kmalloc(sizeof(lm_lvb_t), GFP_NOFS);
	if (!lvb)
		return NULL;
	memset(lvb, 0, sizeof(lm_lvb_t));

	lvb->lvb_length = lsm->bufsize;
	lvb->lvb_data = kmalloc(lsm->bufsize, GFP_NOFS);
	if (!lvb->lvb_data)
		goto Enomem;
	mlk->lvb = lvb;

retry_lock:
	error = mops->load(mep, &mlk->lockname, &bld, buf, lsm->bufsize);
	if (error)
		goto Eload;

	if (!bld.inuse)
		memset(buf, 0, lsm->bufsize);

	lsm_lock_in(&llk, buf);

	if (!bld.inuse)
		llk.state = ME_ST_UNLOCKED;

	if (llk.lvb_lock)
		goto retry_lock; /* someone's locked this while doing unhold */

	set_lvb(mep->cid, &llk);

	lsm_lock_out(&llk, buf);
	prepare_store(&bld, &bst);
	error = mops->store(mep, &mlk->lockname, &bst, buf, lsm->bufsize);
	if (error)
		goto retry_lock;

retry_lvb:
	lock2lvb(&mlk->lockname, &lvbname);

	error = mops->load(mep, &lvbname, &bld, buf, lsm->bufsize);
	if (error)
		goto Eload;

	if (!bld.inuse) {
		prepare_store(&bld, &bst);	/* sets inuse state of buffer */
		memset(buf, 0, lsm->bufsize);
		error = mops->store(mep, &lvbname, &bst, buf, lsm->bufsize);
		if (error)
			goto retry_lvb;
	}

	/* data not necessarily consistent, but why not copy it */
	memcpy(lvb->lvb_data, buf, lsm->bufsize);
	return lvb;

Eload:
	kfree(lvb->lvb_data);
Enomem:
	kfree(lvb);
	return NULL;
}

/**
 * memexp_unhold_lvb
 * @lk:
 * @lvb:
 *
 * This function should return an error code.
 */

void
memexp_unhold_lvb(lm_lock_t * lk, lm_lvb_t * lvb)
{
	me_lock_t *mlk = (me_lock_t *) lk;
	memexp_t *mep = mlk->mep;
	ll_memexp_ops_t *mops = mep->ll_memexp_ops;
	lsm_t *lsm = (lsm_t *) mep->lsm_data;
	lm_lockname_t lvbname;
	char buf[lsm->bufsize];
	lsm_lock_t llk;
	buf_load_t bld;
	buf_store_t bst;
	int error;

	MEP_ASSERT(mlk->lvb == lvb,);

      retry_lock:

	error = mops->load(mep, &mlk->lockname, &bld, buf, lsm->bufsize);
	if (error)
		return;

	lsm_lock_in(&llk, buf);

	MEP_ASSERT(bld.inuse,);
	MEP_ASSERT(test_lvb(mep->cid, &llk),);

	if (lvb->lvb_flags & LM_LVB_PERM)
		llk.lvb_perm = 1;

	if ((llk.lvb_perm == 1) || (llk.lvb_holders > 1)) {
		/* in this case simply remove our lvb reference */

		clear_lvb(mep->cid, &llk);
		lsm_lock_out(&llk, buf);
		prepare_store(&bld, &bst);
		error =
		    mops->store(mep, &mlk->lockname, &bst, buf, lsm->bufsize);
		if (error)
			goto retry_lock;

		kfree(lvb->lvb_data);
		kfree(lvb);
		mlk->lvb = NULL;
		return;
	}

	/* we have to free the LVB buffer before clearing the reference in lock.
	   the lvb_lock flag prevents hold_lvb's while we're freeing the lvb buffer */

	llk.lvb_lock = TRUE;
	lsm_lock_out(&llk, buf);
	prepare_store(&bld, &bst);
	error = mops->store(mep, &mlk->lockname, &bst, buf, lsm->bufsize);
	if (error)
		goto retry_lock;

	lock2lvb(&mlk->lockname, &lvbname);

	/* at this point, we hold the last reference on the lvb and hold the
	   lvb_lock flag, so no one else should be touching the lvb */

	error = mops->load(mep, &lvbname, &bld, buf, lsm->bufsize);
	if (error)
		return;

	MEP_ASSERT(bld.inuse,);
	bst.inuse = 0;
	bst.seq = bld.seq;
	bst.pbn = bld.pbn;
	error = mops->store(mep, &lvbname, &bst, buf, lsm->bufsize);
	if (error)
		MEP_ASSERT(0,);

	/* we've freed lvb buffer, now clear our final lvb reference in lock */

      store_final:

	error = mops->load(mep, &mlk->lockname, &bld, buf, lsm->bufsize);
	if (error)
		return;

	lsm_lock_in(&llk, buf);

	MEP_ASSERT(bld.inuse,);
	MEP_ASSERT(test_lvb(mep->cid, &llk),);
	MEP_ASSERT(llk.lvb_lock,);

	clear_lvb(mep->cid, &llk);
	llk.lvb_lock = 0;

	/* this final lvb reference may be the only thing keeping the lock itself
	   inuse, so we may need to free the lock */

	if ((llk.state == ME_ST_UNLOCKED) &&
	    (llk.lvb_holders == 0) &&
	    (llk.live_holders == 0) &&
	    (llk.expired_holders == 0) && (llk.conv == NO_CONV))
		bst.inuse = 0;
	else
		bst.inuse = 1;
	bst.seq = bld.seq;
	bst.pbn = bld.pbn;

	lsm_lock_out(&llk, buf);

	error = mops->store(mep, &mlk->lockname, &bst, buf, lsm->bufsize);
	if (error)
		goto store_final;

	kfree(lvb->lvb_data);
	kfree(lvb);
	mlk->lvb = NULL;
}

/**
 * memexp_sync_lvb
 * @lk:
 * @lvb:
 *
 * This function should return an error code.
 */

void
memexp_sync_lvb(lm_lock_t * lk, lm_lvb_t * lvb)
{
	me_lock_t *mlk = (me_lock_t *) lk;
	memexp_t *mep = mlk->mep;
	ll_memexp_ops_t *mops = mep->ll_memexp_ops;
	lsm_t *lsm = (lsm_t *) mep->lsm_data;
	lm_lockname_t lvbname;
	lsm_lock_t llk;
	char buf[lsm->bufsize];
	buf_load_t bld;
	buf_store_t bst;
	int error;

	MEP_ASSERT(mlk->lvb == lvb,);

      retry_lock:

	error = mops->load(mep, &mlk->lockname, &bld, buf, lsm->bufsize);
	if (error)
		return;

	lsm_lock_in(&llk, buf);

	MEP_ASSERT(bld.inuse,);
	MEP_ASSERT(test_lvb(mep->cid, &llk),);
	MEP_ASSERT(llk.state == ME_ST_EXCLUSIVE,);
	MEP_ASSERT(test_holder(mep->cid, &llk),);
	MEP_ASSERT(lvb->lvb_data,);
	MEP_ASSERT(lvb->lvb_length == lsm->bufsize,);

	/* write out the PERM flag if it's not yet set */

	if ((lvb->lvb_flags & LM_LVB_PERM) && (llk.lvb_perm == 0)) {
		llk.lvb_perm = 1;
		lsm_lock_out(&llk, buf);
		prepare_store(&bld, &bst);
		error =
		    mops->store(mep, &mlk->lockname, &bst, buf, lsm->bufsize);
		if (error)
			goto retry_lock;
	}

	/* do the lvb sync */

	lock2lvb(&mlk->lockname, &lvbname);

	error = mops->load(mep, &lvbname, &bld, buf, lsm->bufsize);
	if (error)
		return;

	MEP_ASSERT(bld.inuse,);

	memcpy(buf, lvb->lvb_data, lsm->bufsize);
	prepare_store(&bld, &bst);
	error = mops->store(mep, &lvbname, &bst, buf, lsm->bufsize);
	if (error)
		goto retry_lock;
}
