/*
 *
 *    Copyright 1999-2001 Sistina Software, Inc.
 *    Portions Copyright 2001-2003 The OpenGFS Project
 *    Portions Copyright 2003 Brian Jackson
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 *    This file was maintained by:
 *      David C. Teigland <teigland@sistina.com>
 *
 *    Fix potential buffer overrun 2001/09/01
 *                                       Jes Sorensen <jes@trained-monkey.org>
 */

#define __KERNEL_SYSCALLS__

#include "config.h"
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/file.h>
#include <linux/unistd.h>

#include "memexp.h"
#include "ll_memexp.h"
#include "util.h"
#include "callbacks.h"
#include "cluster_info.h"
#include "client_state.h"
#include "mount_state.h"
#include "lock_state.h"
#include "daemon.h"
#include "dmep_tcp.h"
#include "dmep_scsi.h"
#include "dmep_debug.h"
//#include "kcompat.h"

memexp_t *list = NULL;

/**
 * memexp_mount
 *
 * @table_name: name of the cidev
 * @host_data: IP address
 * @cb: OGFS callback function
 * @fsdata: opaque OGFS handle
 * @jid: output jid
 * @first: output flag TRUE if this is the first mount
 *
 * Returns: handle to lock space
 */

lm_lockspace_t *
memexp_mount(char *table_name, char *host_data,
	     lm_callback_t cb, lm_fsdata_t * fsdata,
	     unsigned int *jid, unsigned int *first)
{
	memexp_t *mep;
	char name[256], *port;
	uint32 ip;
	int error;

	//MOD_INC_USE_COUNT;

	name[255] = '\0';
	strncpy(name, table_name, 255);

	inet_aton(host_data, &ip);

	/*  Check to make sure this lock table isn't already being used  */

	for (mep = list; mep; mep = mep->next)
		if (!strcmp(mep->table_name, name)) {
			printk("%s:  \"%s\" is already in use\n", PROTO_NAME,
			       name);
			goto fail;
		}

	/*  Set up our main structure  */

	mep = (memexp_t *) kmalloc(sizeof(memexp_t), GFP_KERNEL);
	if (!mep) {
		printk("%s:  out of memory\n", PROTO_NAME);
		goto fail;
	}
	memset(mep, 0, sizeof(memexp_t));

	strcpy(mep->table_name, name);

	mep->cb = cb;
	mep->fsdata = fsdata;

	mep->cfile = filp_open(name, O_RDONLY, 0);
	if (IS_ERR(mep->cfile)) {
		printk("%s: open error %ld\n", PROTO_NAME, PTR_ERR(mep->cfile));
		mep->cfile = NULL;
	}
	if (!mep->cfile) {
		printk("%s:  can't open cidev %s\n", PROTO_NAME, name);
		goto fail_free_mep;
	}

	if (!S_ISBLK(mep->cfile->f_dentry->d_inode->i_mode)) {
		printk("%s: cidev %s is not a blockdevice\n", PROTO_NAME, name);
		goto fail_free_mep;
	}

	/*  Get information from cidev  */

	error = init_cluster_info(mep, ip);
	if (error) {
		printk("%s:  bad parameters from %s\n", PROTO_NAME, name);
		goto fail_dput;
	}
	show_cluster_info(mep);

	/*  Set up inter-machine callback socket  */

	if (open_cb_socket(mep) < 0)
		goto fail_rels_cluster;

	/*  Determine if lock source is device, server or debug */

	/* The debug lockdev_name string would look like, **
	   ** 10.0.3.1:56789::/dev/pool/brassow              **
	   ** So, we have 2 different lock devs              */
	if ((port = strstr(mep->lockdev_name, "::"))) {
		mep->ll_memexp_ops = &ll_dmep_debug_ops;
	} else {
		mep->ll_memexp_ops = &ll_dmep_tcp_ops;
	}
	/*  Set up the dmep device or server  */

	error = mep->ll_memexp_ops->init(mep, mep->lockdev_name);
	if (error < 0) {
		printk("%s:  DMEP device or server init failed (%d)\n",
		       PROTO_NAME, error);
		goto fail_closesock;
	}

	/*  Set up LSM  */

	error = lsm_init(mep);
	if (error < 0) {
		printk("%s:  LSM init failed (%d)\n", PROTO_NAME, error);
		goto fail_rels_dmep;
	}

	/*  Wait for possible mount state recovery  */

	error = check_mount_state(mep, first);
	if (error < 0) {
		printk("%s:  mount lock check failed (%d)\n", PROTO_NAME,
		       error);
		goto fail_lsm;
	}

	/*  Initialize client state  */

	error = client_state_init(mep);
	if (error < 0) {
		printk("%s:  client state init failed\n", PROTO_NAME);
		goto fail_lsm;
	}

	/*  Start lock callback thread  */

	init_MUTEX_LOCKED(&mep->cb_recvd_sem);

	error = kernel_thread(memexp_callback, mep, 0);
	if (error < 0) {
		printk("%s:  can't start callback thread (%d)\n", PROTO_NAME,
		       error);
		goto fail_client_state;
	}

	down(&mep->cb_recvd_sem);
	init_MUTEX_LOCKED(&mep->cb_recvd_sem);

	/*  Start refresh thread  */

	init_MUTEX_LOCKED(&mep->lock_refresh_sem);

	error = kernel_thread(lock_refresh, mep, 0);
	if (error < 0) {
		printk("%s:  can't start refresh thread (%d)\n", PROTO_NAME,
		       error);
		goto fail_callback;
	}

	/*  Start expiration monitor thread  */

	init_MUTEX_LOCKED(&mep->node_monitor_sem);

	error = kernel_thread(node_monitor, mep, 0);
	if (error < 0) {
		printk("%s:  can't start monitor thread (%d)\n", PROTO_NAME,
		       error);
		goto fail_refresh;
	}

	/*  Success  */

	mep->next = list;
	list = mep;

	*jid = mep->jid;

	return ((lm_lockspace_t *) mep);

fail_refresh:
	mep->lock_refresh_stop = TRUE;
	wake_up_process(mep->lock_refresh_ts);
	down(&mep->lock_refresh_sem);

fail_callback:
	mep->cb_recvd_stop = TRUE;
	cid2ipaddr(mep, mep->cid, &ip);
	send_mess(mep, ip, 0, 0, 0, TRUE);
	down(&mep->cb_recvd_sem);

fail_client_state:
	client_state_relse(mep);

fail_lsm:
	lsm_shutdown(mep);

fail_rels_dmep:
	mep->ll_memexp_ops->release(mep);

fail_closesock:
	close_cb_socket(mep);

fail_rels_cluster:
	release_cluster_info(mep);

fail_dput:
	filp_close(mep->cfile, NULL);

fail_free_mep:
	kfree(mep);

fail:
	//MOD_DEC_USE_COUNT;
	return (NULL);
}

/**
 * memexp_others_may_mount
 *
 * @lockspace: handle to specific lock space
 *
 * OGFS calls this function if it was the first mounter after it's done
 * checking all the journals.
 *
 */

void
memexp_others_may_mount(lm_lockspace_t * lockspace)
{
	memexp_t *mep = (memexp_t *) lockspace;

	others_may_mount(mep);
}

/**
 * memexp_umount
 *
 * @lockspace: handle to specific lock space
 *
 */

void
memexp_unmount(lm_lockspace_t * lockspace)
{
	memexp_t *tmp, *last = NULL, *mep = (memexp_t *) lockspace;
	int done = FALSE;
	uint32 ip;

	if (list == mep) {
		list = list->next;
		done = TRUE;
	} else
		for (tmp = list; tmp; last = tmp, tmp = tmp->next)
			if (tmp == mep) {
				last->next = tmp->next;
				done = TRUE;
			}

	if (!done)
		printk("%s:  given a weird lockspace pointer\n", PROTO_NAME);
	else {
		mep->cb_recvd_stop = TRUE;
		cid2ipaddr(mep, mep->cid, &ip);
		send_mess(mep, ip, 0, 0, 0, TRUE);
		down(&mep->cb_recvd_sem);

		mep->node_monitor_stop = TRUE;
		wake_up_process(mep->node_monitor_ts);
		down(&mep->node_monitor_sem);

		mep->lock_refresh_stop = TRUE;
		wake_up_process(mep->lock_refresh_ts);
		down(&mep->lock_refresh_sem);

		/* close and release stuff */

		client_state_relse(mep);

		lsm_shutdown(mep);

		mep->ll_memexp_ops->release(mep);

		release_cluster_info(mep);

		close_cb_socket(mep);

		filp_close(mep->cfile, NULL);

		kfree(mep);
	}

	//MOD_DEC_USE_COUNT;
}
