/*
 *
 *    Copyright 2001 Sistina Software, Inc.
 *    Portions Copyright 2003 Brian Jackson
 *    Portions Copyright 2003 The OpenGFS Project
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 *    This file was maintained by:
 *      David C. Teigland <teigland@sistina.com>
 *
 */

#include "config.h"
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/slab.h>

#include "memexp.h"
#include "ll_memexp.h"
#include "lock_state.h"
#include "client_state.h"

struct mlbuf_s {
	uint16 state;
	uint16 owner;
};
typedef struct mlbuf_s mlbuf_t;

/**
 * mlbuf_in
 * @buf:
 *
 */

STATIC mlbuf_t *
mlbuf_in(unsigned char *buf)
{
	mlbuf_t *ml = (mlbuf_t *) buf;

	ml->state = be16_to_cpu(ml->state);
	ml->owner = be16_to_cpu(ml->owner);

	return ml;
}

/**
 * mlbuf_out
 * @buf:
 *
 */

STATIC unsigned char *
mlbuf_out(unsigned char *buf)
{
	mlbuf_t *ml = (mlbuf_t *) buf;

	ml->state = cpu_to_be16(ml->state);
	ml->owner = cpu_to_be16(ml->owner);

	return buf;
}

/**
 * load_cid_state
 * @mep: the lock module's "superblock"
 * @cid: client
 * @state: the inuse cidbuf field
 * @counter: the counter which heartbeating increments
 *
 */

STATIC int
load_cid_state(memexp_t * mep, uint16 cid, unsigned char *state,
	       uint64 * counter)
{
	lsm_t *lsm = (lsm_t *) mep->lsm_data;
	ll_memexp_ops_t *dmep = mep->ll_memexp_ops;
	unsigned char buf[lsm->bufsize];
	buf_load_t bld;
	lm_lockname_t lockname;
	cidbuf_t *cidbuf;
	int error;

	lockname.ln_number = cid;
	lockname.ln_type = LM_TYPE_CIDBUF;

	error = dmep->load(mep, &lockname, &bld, buf, lsm->bufsize);
	if (error < 0) {
		printk("%s:  DMEP load failed on nr %" FMT_64 "u, type %u\n",
		       PROTO_NAME, lockname.ln_number, lockname.ln_type);
		return -1;
	}

	if (!bld.inuse) {
		*state = CID_NOT_USED;
		*counter = 0;
		return 0;
	}

	cidbuf = cidbuf_in(buf);

	*state = cidbuf->inuse;
	*counter = cidbuf->counter;

	return 0;
}

/**
 * check_live_cid
 * @mep: the lock module's "superblock"
 * @cid: client to monitor
 *
 * Returns: 0 if cid is dead, 1 if cid is heartbeating, -1 on error
 *
 * This routine is supposed to determine whether the mount lock holder is dead
 * or alive.  A client does not begin heartbeating until after check_mount_state
 * is completed, so we have to estimate the maximum amount of time it would
 * take a machine to get through the mount process and begin heartbeating.  In most
 * cases this comes down to how long reset_cluster() takes to complete.
 */

STATIC int
check_live_cid(memexp_t * mep, uint16 cid)
{
	uint64 start = 0, counter = 0;
	unsigned char state;
	int i, error, n = 0;

	/* see how many machines the ml holder has to recover, we'll multiply this
	   number by a constant to decide how long to wait */

	for (i = 0; i < MAX_CLIENTS; i++) {
		load_cid_state(mep, i, &state, &counter);

		if (state != CID_NOT_USED)
			n++;
	}

	error = load_cid_state(mep, cid, &state, &start);
	if (error) {
		printk("%s:  load_cid_state failed (%u)\n", PROTO_NAME, cid);
		return -1;
	}

	/* for each machine needing recovery we'll give the ml holder 1 minute */

	for (i = 0; i < n * 60; i++) {
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(HZ);

		error = load_cid_state(mep, cid, &state, &counter);
		if (error) {
			printk("%s:  load_cid_state failed (%u)\n", PROTO_NAME,
			       cid);
			return -1;
		}

		if (counter != start) {
			printk("%s:  check_live_cid: %u is alive\n", PROTO_NAME,
			       cid);
			return 1;
		}
	}

	return 0;
}

/**
 * check_live_cluster
 * @mep: the lock module's "superblock"
 *
 * Returns: 0 if all cids are dead, 1 if anyone is heartbeating, -1 on error
 */

STATIC int
check_live_cluster(memexp_t * mep)
{
	uint64 start[MAX_CLIENTS], counter;
	int i, t;
	unsigned char state;

	for (i = 0; i < MAX_CLIENTS; i++)
		load_cid_state(mep, i, &state, &start[i]);

	for (t = 0; t < mep->node_timeout; t++) {
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(HZ);

		for (i = 0; i < MAX_CLIENTS; i++) {
			load_cid_state(mep, i, &state, &counter);

			if (counter != start[i])
				return 1;
		}
	}

	printk
	    ("%s:  Cluster is dead.  Will STOMITH all previously mounted machines.\n",
	     PROTO_NAME);
	return 0;
}

/**
 * reset_cluster
 * @mep: the lock module's "superblock"
 * @prev_state: previous state of Mount Lock
 * @prev_owner: owner of prev_state
 *
 */

STATIC int
reset_cluster(memexp_t * mep, unsigned char prev_state, uint16 prev_owner)
{
	lsm_t *lsm = (lsm_t *) mep->lsm_data;
	ll_memexp_ops_t *dmep = mep->ll_memexp_ops;
	unsigned char buf[lsm->bufsize];
	buf_load_t bld;
	buf_store_t bst;
	lm_lockname_t lockname;
	cidbuf_t *cidbuf;
	unsigned char state;
	uint64 counter;
	int error, i;

	/* do required stomiths */

	if (prev_state == MOUNT_STATE_INIT)
		while (stomith_cid(mep, prev_owner)) ;

	else {
		for (i = 0; i < MAX_CLIENTS; i++) {
			if (i == mep->cid)
				continue;

			load_cid_state(mep, i, &state, &counter);

			if (state != CID_NOT_USED)
				while (stomith_cid(mep, i)) ;
		}
	}

	/* mark all buffers expired */

	for (i = 0; i < MAX_CLIENTS; i++) {
		load_cid_state(mep, i, &state, &counter);

		if (state == CID_NOT_USED)
			continue;

		error = lsm_mark_expired(mep, i);
		MEP_ASSERT(!error,
			   printk("mark_expired failed on cid %u (%d)\n", i,
				  error);
		    );
	}

	/* set all cidbufs to EXPIRED */

	for (i = 0; i < MAX_CLIENTS; i++) {
		lockname.ln_number = i;
		lockname.ln_type = LM_TYPE_CIDBUF;

		while (dmep->load(mep, &lockname, &bld, buf, lsm->bufsize)) ;

		if (!bld.inuse)
			continue;

		cidbuf = cidbuf_in(buf);

		if (cidbuf->inuse == CID_NOT_USED)
			continue;

		cidbuf->inuse = CID_EXPIRED;
		cidbuf->cleaner = OWNER_NONE;
		cidbuf_out(buf);
		prepare_store(&bld, &bst);
		error = dmep->store(mep, &lockname, &bst, buf, lsm->bufsize);
		MEP_ASSERT(!error, printk("DMEP store failed\n");
		    );
	}

	return 0;
}

/**
 * set_ml_state - set the state of the mount lock
 * @mep: the lock module's "superblock"
 *
 */

STATIC int
set_ml_state(memexp_t * mep, buf_load_t * bldp, unsigned char state,
	     uint32 owner)
{
	lsm_t *lsm = (lsm_t *) mep->lsm_data;
	ll_memexp_ops_t *dmep = mep->ll_memexp_ops;
	unsigned char buf[lsm->bufsize];
	lm_lockname_t lockname;
	buf_store_t bst;
	mlbuf_t *ml;
	int error;

	lockname.ln_number = 0;
	lockname.ln_type = LM_TYPE_MOUNT;

	ml = (mlbuf_t *) buf;
	ml->state = state;
	ml->owner = owner;

	mlbuf_out(buf);
	prepare_store(bldp, &bst);
	error = dmep->store(mep, &lockname, &bst, buf, lsm->bufsize);
	if (!error)
		return 0;

	printk("%s:  failed to set ML to state: %u, owner: %u\n", PROTO_NAME,
	       state, owner);
	return -1;
}

/**
 * check_mount_state
 * @mep: the lock module's "superblock"
 * @first: output, TRUE if this mount is the first one
 *
 * lsm_init must happen before this
 */

int
check_mount_state(memexp_t * mep, int *first)
{
	lsm_t *lsm = (lsm_t *) mep->lsm_data;
	ll_memexp_ops_t *dmep = mep->ll_memexp_ops;
	unsigned char buf[lsm->bufsize];
	unsigned char cidstate;
	uint16 owner = 0;
	buf_load_t bld;
	lm_lockname_t lockname;
	mlbuf_t *ml;
	uint64 counter;
	int error;

	while (1) {
		lockname.ln_number = 0;
		lockname.ln_type = LM_TYPE_MOUNT;

		error = dmep->load(mep, &lockname, &bld, buf, lsm->bufsize);
		if (error < 0) {
			printk("%s:  DMEP load failed\n", PROTO_NAME);
			return -1;
		}

		/* ML state is NULL */

		if (!bld.inuse) {
			error =
			    set_ml_state(mep, &bld, MOUNT_STATE_INIT, mep->cid);
			if (!error) {
				*first = TRUE;
				return 0;
			}
		}

		else {
			ml = mlbuf_in(buf);

			/* ML state is INIT */

			if (ml->state == MOUNT_STATE_INIT) {
				if (ml->owner != mep->cid
				    && check_live_cid(mep, ml->owner)) {
					set_current_state(TASK_INTERRUPTIBLE);
					schedule_timeout(2*HZ);
					continue;
				}

				if (ml->owner == mep->cid) {
					error =
					    set_ml_state(mep, &bld,
							 MOUNT_STATE_INIT,
							 mep->cid);
					if (!error) {
						*first = TRUE;
						return (0);
					}
					continue;
				}

				owner = ml->owner;

				error =
				    set_ml_state(mep, &bld, MOUNT_STATE_RESET,
						 mep->cid);
				if (!error) {
					reset_cluster(mep, MOUNT_STATE_INIT,
						      owner);
					*first = TRUE;
					return 0;
				}
			}

			/* ML state is RESET */

			else if (ml->state == MOUNT_STATE_RESET) {
				if (ml->owner != mep->cid
				    && check_live_cid(mep, ml->owner)) {
					set_current_state(TASK_INTERRUPTIBLE);
					schedule_timeout(2*HZ);
					continue;
				}

				error =
				    set_ml_state(mep, &bld, MOUNT_STATE_RESET,
						 mep->cid);
				if (!error) {
					reset_cluster(mep, MOUNT_STATE_RESET,
						      0);
					*first = TRUE;
					return 0;
				}
			}

			/* ML state is OK */

			else if (ml->state == MOUNT_STATE_OK) {
				error =
				    load_cid_state(mep, mep->cid, &cidstate,
						   &counter);
				if (error)
					continue;

				if (cidstate == CID_NOT_USED) {	/* also true if !bld.inuse on cidbuf */
					*first = FALSE;
					return 0;
				}

				if (check_live_cluster(mep)) {
					set_current_state(TASK_INTERRUPTIBLE);
					schedule_timeout(2*HZ);
					continue;
				}

				error =
				    set_ml_state(mep, &bld, MOUNT_STATE_RESET,
						 mep->cid);
				if (!error) {
					reset_cluster(mep, MOUNT_STATE_OK, 0);
					*first = TRUE;
					return 0;
				}
			}

			else {
				printk
				    ("%s:  Warning: unknown mount buffer state\n",
				     PROTO_NAME);
			}
		}
	}

	return 0;
}

/**
 * others_may_mount - Set the mount buffer state to OK so others can mount
 * @mep: the lock module's "superblock"
 *
 */

void
others_may_mount(memexp_t * mep)
{
	lsm_t *lsm = (lsm_t *) mep->lsm_data;
	ll_memexp_ops_t *dmep = mep->ll_memexp_ops;
	unsigned char buf[lsm->bufsize];
	lm_lockname_t lockname;
	buf_load_t bld;
	buf_store_t bst;
	mlbuf_t *ml;
	int error;

	lockname.ln_number = 0;
	lockname.ln_type = LM_TYPE_MOUNT;

	while (dmep->load(mep, &lockname, &bld, buf, lsm->bufsize)) ;

	MEP_ASSERT(bld.inuse, printk("others_may_mount mount lock not inuse\n");
	    );

	ml = mlbuf_in(buf);

	if (ml->state != MOUNT_STATE_INIT && ml->state != MOUNT_STATE_RESET)
		MEP_ASSERT(0,
			   printk("invalid mount lock state (%u)\n", ml->state);
	    );

	if (ml->owner != mep->cid)
		MEP_ASSERT(0,
			   printk("invalid mount lock owner (%u)\n", ml->owner);
	    );

	ml->state = MOUNT_STATE_OK;
	ml->owner = OWNER_NONE;

	mlbuf_out(buf);
	prepare_store(&bld, &bst);
	error = dmep->store(mep, &lockname, &bst, buf, lsm->bufsize);

	MEP_ASSERT(!error, printk("others_may_mount store failed\n");
	    );
}

/* vim:sw=8 ts=8
 */
