/*
 *
 *    Copyright 2000-2001 Sistina Software, Inc.
 *
 *    This is free software released under the GNU General Public License.
 *    There is no warranty for this software.  See the file COPYING for
 *    details.
 *
 *    See the file AUTHORS for a list of contributors.
 *
 *    This file was maintained by:
 *      Mike C Tilstra <conrad@sistina.com>
 *
 */

#ifndef _MEMEXP_WIREDTYPES_H
#define _MEMEXP_WIREDTYPES_H

/* Load Request:
 *   me_hdr_t: Mel_Load
 *   me_cdb_t
 * Load Reply:
 *   me_hdr_t: Mel_Ok
 *   me_rdf_t
 *   u8[]            Size is predetermined from modepage
 *
 * Store Request:
 *   me_hdr_t: Mel_Store
 *   me_cdb_t
 *   me_rdf_t
 *   u8[]            Size is predetermined from modepage
 * Store Reply:
 *   me_hdr_t: Mel_Ok
 *
 * Verify Request:
 *   me_hdr_t: Mel_Verify
 *   u64            PBN
 * Verify Reply:
 *   me_hdr_t: Mel_Ok
 *
 * ModePage Request:
 *   me_hdr_t: Mel_ModePgR or Mel_ModePgW
 *   me_modepage_t
 * ModePage Reply:
 *   me_hdr_t: Mel_Ok
 *   me_modepage_t
 *
 * Ban Request:
 *   me_hdr_t: Mel_BanReq
 *   u32            Ip in BigE
 * Ban Reply:
 *   me_hdr_t: Mel_Ok
 *
 * Unban Request:
 *   me_hdr_t: Mel_UnBanMe
 *   u32            Ip in BigE
 * Unban Reply:
 *   me_hdr_t: Mel_Ok
 *
 * Status Request:
 *   me_hdr_t: Mel_StatReq
 *   u8             Selector for which status structure we want returned.
 * Status Reply:
 *   me_hdr_t: Mel_Ok
 *   srv_status_t
 *
 * Dump Request:
 *   me_hdr_t: Mel_Dump
 *   dumpb_cdb_t
 * Dump Reply:
 *   me_hdr_t: Mel_Ok
 *   u8[8] = DUMPSTRT
 *   u32                 Size of following array
 *   {
 *     dump_element_t
 *     u8[]              This buffer's length is determined elsewhere
 *   }[]                 This group repeats
 *      This current dump reoutine wil not work for udp.  But I could tweek
 *      it a little.  Basically just send one of the dump_ele and u8[]
 *      pairs in each udp packet. (if they fit.)
 *      Could also pull the pack-skipping stunt that the server uses to
 *      write to disk on 512byte bounries.
 *
 *
 * NOTE!!!!
 * Any of the Replys can be replaced with the following in error
 * conditions.
 *   me_hdr_t: Mel_Error
 *   errorcode_t
 */

/* DMEP commands. */
#define Mel_Load    0x9B	/* follow header with me_cdb_t */
#define Mel_Store   0x9C	/* follow header with me_cdb_t and me_rdf_t */
#define Mel_Dump    0x9D	/* follow header with dumpb_cdb_t */
#define Mel_Verify  0x9E
#define Mel_ModePgR 0x31	/* follow header with modepage_t */
#define Mel_ModePgW 0x30	/* follow header with modepage_t */

/* Extensions to DMEP.  Not required, but make things nicer. */
#define Mel_StatReq 0x67	/* follow header with srv_status_t */
#define Mel_BanReq  0x44	/* follow header with u32 (IP in Be) */
#define Mel_UnBanMe 0x42	/* follow header with u32 (IP in Be) */
#define Mel_Option  0x54	/* Set a server option without restarting. */

/* Reply codes. */
#define Mel_Error   0x70	/* follow header with errorcode_t */
#define Mel_Ok      0x66

/* Possible errors to return: */
#define Mel_Err_BidZero 0x02
#define Mel_Err_NoBid   0x03
#define Mel_Err_BadSeq  0x04
#define Mel_Err_BadPbn  0x05
#define Mel_Pbn_Space_Full 0x06
#define Mel_You_are_Banned 0x07
#define Mel_Verify_Failed  0x08
#define Mel_Err_BadSize 0x09	/* changed to 9 to avoid a collision. */
#define Mel_Err_NotImpl 0xff

struct me_hdr_s {
	u8 Opcode;
	u32 Key;		/* not used yet... */
	u16 length;		/* how much data is following. */
} __attribute__ ((packed));
typedef struct me_hdr_s me_hdr_t;

#define wreBIDSIZE 10		/* since it might be different... */
struct me_cdb_s {
	u8 bid[wreBIDSIZE];
	u16 allocLen;
} __attribute__ ((packed));
typedef struct me_cdb_s me_cdb_t;

struct me_rdf_s {		/* this may get an Opcode field.  later */
	u8 InUse;		/* just the upper bit. */
	u8 fullness;
	u64 seq;
	u64 pbn;
	/* data follows */
} __attribute__ ((packed));
typedef struct me_rdf_s me_rdf_t;

struct dumpb_cdb_s {
	u64 pbn;
	u32 want;		/* how many buffers to send */
} __attribute__ ((packed));
typedef struct dumpb_cdb_s dumpb_cdb_t;

/* dump RDF is a u32 which is the number of buffers being returned, followed
 * by dump_element and some data. where the last two repeat as needed. */
#define dump_elm_magic 0x44554d50
struct dump_element_s {
	u8 InUse;		/* just upper bit *//* why is this still here? */
	u8 bid[wreBIDSIZE];
	u64 seq;
	u64 pbn;
	/* data follows */
} __attribute__ ((packed));
typedef struct dump_element_s dump_element_t;

struct modepage_s {
	u64 numbufs;
	u32 datasize;
} __attribute__ ((packed));
typedef struct modepage_s modepage_t;

typedef u8 errorcode_t;

/* Status packet.  only really useful to me, but hey...
 *
 * NOTE: the ogfs memexp clients in the kernel should NEVER access this
 * structure.
 *
 * I really don't need to do all these inuse checks.  I'll leave the code
 * there, but comment it out I think.  More useful information now will
 * involve statistics about the memblock usages.
 * So what new info??
 *  - Max locks held
 *  - current locks held
 *  - ?mean locks held?
 *  - locks per second
 *  -  max, min, mean, std for lps too.
 */
#define STAT_HSHCNT (1)
#define STAT_counts (2)
#define STAT_LPS    (3)
#define STAT_CLILST (4)		/* this is a bit of a hack. maybe kill if not DEBUG? */
#define STAT_LDUR   (5)

/* locks per second struct.  (it ended up more like seconds per lock op...)*/
#define LPSSTEP (10)
typedef struct lps_s {
	u64 step;
	u64 lasttime;
	u64 min;
	u64 max;
	u64 std;
	u64 total;
	u64 count;
} lps_t;

/* lock op durration.  Length of a sing lock op. */
typedef struct ldur_s {
	u64 lasttime;
	u64 min;
	u64 max;
	u64 std;
	u64 total;
	u64 count;
} ldur_t;

typedef struct counters_s {
	u32 loads;		/* how many loads have happened? */
	u32 stores;		/* how many stores have happened? */
	u32 sucstr;		/* how many stores that succeded */
	u32 dumps;		/* how many dumps have happened? */
	u32 bans;		/* how many bans have happened? */
	u32 unbans;		/* how many unbans have happened? */
	u32 errors;		/* how many error replies sent */
	u32 verify;		/* how many verifies have happened? */
} counters_t;

#define MAXdens 12
struct hshcnt_s {
	u16 bcnt[MAXdens];	/* how many hash buckets have 0 items? 1 item? ect. */
	u32 inuse;
	u32 numkeys;
	u32 maxelems;
} __attribute__ ((packed));
typedef struct hshcnt_s hshcnt_t;

#define SIZE_STATS MAX(MAX(sizeof(lps_t),sizeof(ldur_t)), \
      MAX(sizeof(counters_t),sizeof(hshcnt_t)))

#endif /* _MEMEXP_WIREDTYPES_H */
