/*
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * General Public License for more details.
 *
 */

/*
 * Copyright (c) 2000 VERITAS Software Corporation.
 *
 */
/****************************************************************************
 *  Header: remcom.c,v 1.34 91/03/09 12:29:49 glenne Exp $
 *
 *  Module name: remcom.c $
 *  Revision: 1.34 $
 *  Date: 91/03/09 12:29:49 $
 *  Contributor:     Lake Stevens Instrument Division$
 *
 *  Description:     low level support for gdb debugger. $
 *
 *  Considerations:  only works on target hardware $
 *
 *  Written by:	     Glenn Engel $
 *  Updated by:	     David Grothe <dave@gcom.com>
 *  Updated by:	     Robert Walsh <rjwalsh@durables.org>
 *  Updated by:	     wangdi <wangdi@clusterfs.com>
 *  ModuleState:     Experimental $
 *
 *  NOTES:	     See Below $
 *
 *  Modified for 386 by Jim Kingdon, Cygnus Support.
 *  Compatibility with 2.1.xx kernel by David Grothe <dave@gcom.com>
 *
 *  Changes to allow auto initilization.  All that is needed is that it
 *  be linked with the kernel and a break point (int 3) be executed.
 *  The header file <asm/kgdb.h> defines BREAKPOINT to allow one to do
 *  this. It should also be possible, once the interrupt system is up, to
 *  call putDebugChar("+").  Once this is done, the remote debugger should
 *  get our attention by sending a ^C in a packet. George Anzinger
 *  <george@mvista.com>
 *  Integrated into 2.2.5 kernel by Tigran Aivazian <tigran@sco.com>
 *  Added thread support, support for multiple processors,
 *	support for ia-32(x86) hardware debugging.
 *	Amit S. Kale ( akale@veritas.com )
 *
 *  Modified to support debugging over ethernet by Robert Walsh
 *  <rjwalsh@durables.org> and wangdi <wangdi@clusterfs.com>, based on
 *  code by San Mehat.
 *
 *
 *  To enable debugger support, two things need to happen.  One, a
 *  call to set_debug_traps() is necessary in order to allow any breakpoints
 *  or error conditions to be properly intercepted and reported to gdb.
 *  Two, a breakpoint needs to be generated to begin communication.  This
 *  is most easily accomplished by a call to breakpoint().  Breakpoint()
 *  simulates a breakpoint by executing an int 3.
 *
 *************
 *
 *    The following gdb commands are supported:
 *
 * command	    function				   Return value
 *
 *    g		    return the value of the CPU registers  hex data or ENN
 *    G		    set the value of the CPU registers	   OK or ENN
 *
 *    mAA..AA,LLLL  Read LLLL bytes at address AA..AA	   hex data or ENN
 *    MAA..AA,LLLL: Write LLLL bytes at address AA.AA	   OK or ENN
 *
 *    c		    Resume at current address		   SNN	 ( signal NN)
 *    cAA..AA	    Continue at address AA..AA		   SNN
 *
 *    s		    Step one instruction		   SNN
 *    sAA..AA	    Step one instruction from AA..AA	   SNN
 *
 *    k		    kill
 *
 *    ?		    What was the last sigval ?		   SNN	 (signal NN)
 *
 * All commands and responses are sent with a packet which includes a
 * checksum.  A packet consists of
 *
 * $<packet info>#<checksum>.
 *
 * where
 * <packet info> :: <characters representing the command or response>
 * <checksum>	 :: < two hex digits computed as modulo 256 sum of <packetinfo>>
 *
 * When a packet is received, it is first acknowledged with either '+' or '-'.
 * '+' indicates a successful transfer.	 '-' indicates a failed transfer.
 *
 * Example:
 *
 * Host:		  Reply:
 * $m0,10#2a		   +$00010203040506070809101112131415#42
 *
 ****************************************************************************/
#define KGDB_VERSION "<20030915.1651.33>"
#include <linux/config.h>
#include <linux/types.h>
#include <asm/string.h>		/* for strcpy */
#include <linux/kernel.h>
#include <linux/sched.h>
#include <asm/system.h>
#include <asm/ptrace.h>		/* for linux pt_regs struct */
#include <asm/kgdb_local.h>
#include <linux/list.h>
#include <asm/atomic.h>
#include <asm/processor.h>
#include <asm/unwind.h>
#include <asm/delay.h>
#include <asm/sections.h>
#include <linux/irq.h>
#include <linux/reboot.h>
#include <linux/inet.h>
#include <linux/netpoll.h>

/*
 * IA64 ToDo:
 *	1) more testing needs to be done when modifying registers.
 *	2) probably could use cleanup in register get/put from unw_info but gdb work
 *	   is on going (WIP) to reduce g-packet on wire for serial
 *	3) until gdb work is complete and accepted by gdb folks, the serial interface
 *	   isn't complete.  the current g-packet is over 10,000 bytes which is too
 *	   large for a serial line. without a g-packet a delay is used with large putpacket
	   requests.  however, this won't solve a gdb sent packet which is a g-packet.
 *	4) NMI for hung machine.  Well we don't have real NMI!
 */

/************************************************************************
 *
 * external low-level support routines
 */
typedef void (*Function) (void);	/* pointer to a function */

/* Thread reference */
typedef unsigned char threadref[8];

extern int tty_putDebugChar(int);     /* write a single character      */
extern int tty_getDebugChar(void);    /* read and return a single char */
extern void tty_flushDebugChar(void); /* flush pending characters      */
extern int eth_putDebugChar(int);     /* write a single character      */
extern int eth_getDebugChar(void);    /* read and return a single char */
extern void eth_flushDebugChar(void); /* flush pending characters      */

/************************************************************************/
/* BUFMAX defines the maximum number of characters in inbound/outbound buffers*/
/* at least NUMREGBYTES*2 are needed for register packets */
/* Longer buffer is needed to list all threads */
/* purloined from gdb ia64 config support */
#define NUM_REGS 590
#define REGISTER_BYTES (NUM_REGS*8+128*8)
#define REGISTER_BYTE(N) (((N) * 8) \
  + ((N) <= IA64_FR0_REGNUM ? 0 : 8 * (((N) > IA64_FR127_REGNUM) ? 128 : (N) - IA64_FR0_REGNUM)))
#define	REGISTER_SIZE(N) (((N) >= IA64_FR0_REGNUM && (N) <= IA64_FR127_REGNUM) ? 16 : 8)
#define IA64_GR0_REGNUM         0
#define IA64_FR0_REGNUM         128
#define IA64_FR127_REGNUM       (IA64_FR0_REGNUM+127)
#define IA64_PR0_REGNUM         256
#define IA64_BR0_REGNUM         320
#define IA64_VFP_REGNUM         328
#define IA64_PR_REGNUM          330
#define IA64_IP_REGNUM          331
#define IA64_PSR_REGNUM         332
#define IA64_CFM_REGNUM         333
#define IA64_AR0_REGNUM         334
#define IA64_NAT0_REGNUM        462
#define IA64_NAT31_REGNUM       (IA64_NAT0_REGNUM+31)
#define IA64_NAT32_REGNUM       (IA64_NAT0_REGNUM+32)
#define IA64_RSC_REGNUM		(IA64_AR0_REGNUM+16)
#define IA64_BSP_REGNUM		(IA64_AR0_REGNUM+17)
#define IA64_BSPSTORE_REGNUM	(IA64_AR0_REGNUM+18)
#define IA64_RNAT_REGNUM	(IA64_AR0_REGNUM+19)
#define IA64_FCR_REGNUM		(IA64_AR0_REGNUM+21)
#define IA64_EFLAG_REGNUM	(IA64_AR0_REGNUM+24)
#define IA64_CSD_REGNUM		(IA64_AR0_REGNUM+25)
#define IA64_SSD_REGNUM		(IA64_AR0_REGNUM+26)
#define IA64_CFLG_REGNUM	(IA64_AR0_REGNUM+27)
#define IA64_FSR_REGNUM		(IA64_AR0_REGNUM+28)
#define IA64_FIR_REGNUM		(IA64_AR0_REGNUM+29)
#define IA64_FDR_REGNUM		(IA64_AR0_REGNUM+30)
#define IA64_CCV_REGNUM		(IA64_AR0_REGNUM+32)
#define IA64_UNAT_REGNUM	(IA64_AR0_REGNUM+36)
#define IA64_FPSR_REGNUM	(IA64_AR0_REGNUM+40)
#define IA64_ITC_REGNUM		(IA64_AR0_REGNUM+44)
#define IA64_PFS_REGNUM		(IA64_AR0_REGNUM+64)
#define IA64_LC_REGNUM		(IA64_AR0_REGNUM+65)
#define IA64_EC_REGNUM		(IA64_AR0_REGNUM+66)

#define	REGISTER_INDEX(N)	(REGISTER_BYTE(N) / sizeof (unsigned long))
#define BUFMAX (REGISTER_BYTES*2+10)
static char remcomInBuffer[BUFMAX];
static char remcomOutBuffer[BUFMAX];
static short error;
/*
 * gdb 'qPart' packet support
 */

struct reg_offset {
	int 	regnum;
	int	offset;
} reg_offset[] = {
{IA64_GR0_REGNUM+1, 0},			/*	r1 */
{IA64_GR0_REGNUM+2, 8},
{IA64_GR0_REGNUM+3, 16},
{IA64_GR0_REGNUM+4, 24},
{IA64_GR0_REGNUM+5, 32},
{IA64_GR0_REGNUM+6, 40},
{IA64_GR0_REGNUM+7, 48},
{IA64_GR0_REGNUM+8, 56},
{IA64_GR0_REGNUM+9, 64},
{IA64_GR0_REGNUM+10, 72},
{IA64_GR0_REGNUM+11, 80},
{IA64_GR0_REGNUM+12, 88},
{IA64_GR0_REGNUM+13, 96},
{IA64_GR0_REGNUM+14, 104},
{IA64_GR0_REGNUM+15, 112},
{IA64_GR0_REGNUM+16, 120},
{IA64_GR0_REGNUM+17, 128},
{IA64_GR0_REGNUM+18, 136},
{IA64_GR0_REGNUM+19, 144},
{IA64_GR0_REGNUM+20, 152},
{IA64_GR0_REGNUM+21, 160},
{IA64_GR0_REGNUM+22, 168},
{IA64_GR0_REGNUM+23, 176},
{IA64_GR0_REGNUM+24, 184},
{IA64_GR0_REGNUM+25, 192},
{IA64_GR0_REGNUM+26, 200},
{IA64_GR0_REGNUM+27, 208},
{IA64_GR0_REGNUM+28, 216},
{IA64_GR0_REGNUM+29, 224},
{IA64_GR0_REGNUM+30, 232},
{IA64_GR0_REGNUM+31, 240},		/* r31 */
{IA64_FR0_REGNUM+6, 248},		/* fr6 */
{IA64_FR0_REGNUM+7, 264},
{IA64_FR0_REGNUM+8, 280},
{IA64_FR0_REGNUM+9, 296},
{IA64_FR0_REGNUM+10, 312},
{IA64_FR0_REGNUM+11, 328},		/* fr11 */
{IA64_PR_REGNUM, 344},			/* pr  */
{IA64_BR0_REGNUM, 352},			/* b0 */
{IA64_BR0_REGNUM+1, 360},
{IA64_BR0_REGNUM+2, 368},
{IA64_BR0_REGNUM+3, 376},
{IA64_BR0_REGNUM+4, 384},
{IA64_BR0_REGNUM+5, 392},
{IA64_BR0_REGNUM+6, 400},
{IA64_BR0_REGNUM+7, 408},		/* br7 */
{IA64_IP_REGNUM,  416},			/* ip */
{IA64_CFM_REGNUM, 424},			/* cfm */
{IA64_RSC_REGNUM, 432},			/* rsc */
{IA64_BSP_REGNUM, 440},			/* bp */
{IA64_BSPSTORE_REGNUM, 448},		/* bpstore */
{IA64_RNAT_REGNUM, 456},		/* rnat */
{IA64_UNAT_REGNUM, 464},		/* unat */
{IA64_FPSR_REGNUM, 472},		/* fpsr */
{IA64_PSR_REGNUM,  480},		/* psr */
{IA64_PFS_REGNUM,  488},		/* pfs */
};

static int rpacket_size;
int norpacket_udelay = 10000;
char * hex2mem(char *buf, char *mem, int count, int may_fault);
char * mem2hex(char *mem, char *buf, int count, int may_fault);

long
find_roffset(long regnum)
{
	int i;

	for (i = 0; i < (sizeof (reg_offset) / sizeof (reg_offset[0])); i++)
		if (reg_offset[i].regnum == regnum)
			return reg_offset[i].offset;

	return -1L;
}

static void
process_gpacket_request(void)
{
	int regnum, offset, size;

	if (!rpacket_size)
		rpacket_size = reg_offset[(sizeof (reg_offset) / sizeof (reg_offset[0])) - 1].
			offset + 8;
	hex2mem(&remcomInBuffer[20], (char *) &regnum, sizeof(regnum), 0);
	if (remcomInBuffer[20 + sizeof(regnum)*2] != ',')
		goto bad;
	hex2mem(&remcomInBuffer[20 + sizeof(regnum)*2 + 1], (char *) &size, sizeof(size), 0);
	if (size != sizeof (size))
		goto bad;
	offset = find_roffset(regnum);
	mem2hex((char *) &offset, &remcomOutBuffer[0], sizeof (offset), 0);
	remcomOutBuffer[sizeof(offset) * 2] = 0;
	return;
bad:
	remcomOutBuffer[0] = 'E';
	remcomOutBuffer[1] = 0;
	return;
}

void
rpacket_process(unsigned long *reg, char *buf, int direction)
{
	int i;

	for (i = 0; i < (sizeof (reg_offset) / sizeof (reg_offset[0])); i++)
		if (!direction)
			hex2mem(buf + reg_offset[i].offset * 2,
				(char *) &reg[REGISTER_INDEX(reg_offset[i].regnum)],
				REGISTER_SIZE(reg_offset[i].regnum), 0);
		else
			mem2hex((char *) &reg[REGISTER_INDEX(reg_offset[i].regnum)],
				buf + reg_offset[i].offset * 2,
				REGISTER_SIZE(reg_offset[i].regnum), 0);

	if (direction)
		buf[rpacket_size * 2] = 0;
}



char *kgdb_version = KGDB_VERSION;

/*  debug >  0 prints ill-formed commands in valid packets & checksum errors */
int debug_regs = 0;		/* set to non-zero to print registers */

/* filled in by an external module */
char *gdb_module_offsets;

static const char hexchars[] = "0123456789abcdef";

/* Number of bytes of registers.  */
#define NUMREGBYTES 	REGISTER_BYTES
#define	BREAKNUM	0x6666UL
#define	KGDBBREAKNUM	0x6665UL

static void inline
kgdb_pc(struct pt_regs *regs, unsigned long pc)
{
	regs->cr_iip = pc & ~0xf;
	ia64_psr(regs)->ri = pc & 0x3;
	return;
}

void
breakpoint(void)
{
	asm volatile ("break.m 0x6665");
}

/***************************  ASSEMBLY CODE MACROS *************************/
/*
 * Put the error code here just in case the user cares.
 * Likewise, the vector number here (since GDB only gets the signal
 * number through the usual means, and that's not very specific).
 * The called_from is the return address so he can tell how we entered kgdb.
 * This will allow him to seperate out the various possible entries.
 */
#define REMOTE_DEBUG 0		/* set != to turn on printing (also available in info) */

#define PID_MAX PID_MAX_DEFAULT

#ifdef CONFIG_SMP
void smp_send_nmi_allbutself(void);
#define IF_SMP(x) x
#undef MAX_NO_CPUS
#ifndef CONFIG_NO_KGDB_CPUS
#define CONFIG_NO_KGDB_CPUS 2
#endif
#if CONFIG_NO_KGDB_CPUS > NR_CPUS
#define MAX_NO_CPUS NR_CPUS
#else
#define MAX_NO_CPUS CONFIG_NO_KGDB_CPUS
#endif
#define hold_init hold_on_sstep: 1,
#define MAX_CPU_MASK (unsigned long)((1LL << MAX_NO_CPUS) - 1LL)
#define NUM_CPUS num_online_cpus()
#else
#define IF_SMP(x)
#define hold_init
#undef MAX_NO_CPUS
#define MAX_NO_CPUS 1
#define NUM_CPUS 1
#endif
#define NOCPU (struct task_struct *)0xbad1fbad
struct kgdb_state {
	int		exceptionVector;
	int		signo;
	unsigned long	err_code;
	struct pt_regs *regs;
	struct unw_frame_info
			*unw;
	int		ret;
};
/* *INDENT-OFF*	 */
struct kgdb_info {
	int used_malloc;
	void *called_from;
	long long entry_itc;
	int errcode;
	unsigned long vector;
	int print_debug_info;
	unsigned long ia64_regs[REGISTER_BYTES / sizeof(long)];
#ifdef CONFIG_SMP
	int hold_on_sstep;
	struct {
		volatile struct task_struct *task;
		int pid;
		int hold;
		struct pt_regs *regs;
		unsigned long ia64_regs[REGISTER_BYTES / sizeof(long)];
	} cpus_waiting[MAX_NO_CPUS];
#endif
} kgdb_info = {hold_init print_debug_info:REMOTE_DEBUG, vector:-1UL};

/* *INDENT-ON*	*/

#define used_m kgdb_info.used_malloc
/*
 * This is little area we set aside to contain the stack we
 * need to build to allow gdb to call functions.  We use one
 * per cpu to avoid locking issues.  We will do all this work
 * with interrupts off so that should take care of the protection
 * issues.
 */
#define	MAX_HW_BREAKPOINT	(20)
long hw_break_total_dbr, hw_break_total_ibr;
#define	HW_BREAKPOINT	(hw_break_total_dbr + hw_break_total_ibr)
#define	WATCH_INSTRUCTION	0x0
#define WATCH_WRITE		0x1
#define	WATCH_READ		0x2
#define	WATCH_ACCESS		0x3

#define	HWCAP_DBR	((1 << WATCH_WRITE) | (1 << WATCH_READ))
#define	HWCAP_IBR	(1 << WATCH_INSTRUCTION)
struct hw_breakpoint {
	unsigned enabled;
	unsigned long capable;
	unsigned long type;
	unsigned long mask;
	unsigned long addr;
} *breakinfo;

#define LOOKASIDE_SIZE (200 + (sizeof(struct hw_breakpoint) * MAX_HW_BREAKPOINT))
#define MALLOC_MAX   LOOKASIDE_SIZE	/* Max malloc size */
struct {
	unsigned int esp;
	int array[LOOKASIDE_SIZE];
} fn_call_lookaside[MAX_NO_CPUS];

#define IF_BIT 0x200
#define TF_BIT 0x100

#define MALLOC_ROUND 8-1

static char malloc_array[MALLOC_MAX];
IF_SMP(static void to_gdb(const char *mess));
void *
malloc(int size)
{

	if (size <= (MALLOC_MAX - used_m)) {
		int old_used = used_m;
		used_m += ((size + MALLOC_ROUND) & (~MALLOC_ROUND));
		return &malloc_array[old_used];
	} else {
		return NULL;
	}
}

/*
 * I/O dispatch functions...
 * Based upon kgdboe, either call the ethernet
 * handler or the serial one..
 */
void
putDebugChar(int c)
{
	if (!kgdboe) {
		tty_putDebugChar(c);
	} else {
		eth_putDebugChar(c);
	}
}

int
getDebugChar(void)
{
	if (!kgdboe) {
		char ch;
		while (1) {
			ch = tty_getDebugChar() & 0x7f;
			if (ch != 0x7f)
				return ch;
		}
	} else {
		return eth_getDebugChar();
	}
}

void
flushDebugChar(void)
{
	if (!kgdboe) {
		tty_flushDebugChar();
	} else {
		eth_flushDebugChar();
	}
}

/*
 * Gdb calls functions by pushing agruments, including a return address
 * on the stack and the adjusting EIP to point to the function.	 The
 * whole assumption in GDB is that we are on a different stack than the
 * one the "user" i.e. code that hit the break point, is on.  This, of
 * course is not true in the kernel.  Thus various dodges are needed to
 * do the call without directly messing with EIP (which we can not change
 * as it is just a location and not a register.	 To adjust it would then
 * require that we move every thing below EIP up or down as needed.  This
 * will not work as we may well have stack relative pointer on the stack
 * (such as the pointer to regs, for example).

 * So here is what we do:
 * We detect gdb attempting to store into the stack area and instead, store
 * into the fn_call_lookaside.array at the same relative location as if it
 * were the area ESP pointed at.  We also trap ESP modifications
 * and uses these to adjust fn_call_lookaside.esp.  On entry
 * fn_call_lookaside.esp will be set to point at the last entry in
 * fn_call_lookaside.array.  This allows us to check if it has changed, and
 * if so, on exit, we add the registers we will use to do the move and a
 * trap/ interrupt return exit sequence.  We then adjust the eflags in the
 * regs array (remember we now have a copy in the fn_call_lookaside.array) to
 * kill the interrupt bit, AND we change EIP to point at our set up stub.
 * As part of the register set up we preset the registers to point at the
 * begining and end of the fn_call_lookaside.array, so all the stub needs to
 * do is move words from the array to the stack until ESP= the desired value
 * then do the rti.  This will then transfer to the desired function with
 * all the correct registers.  Nifty huh?
 */
extern asmlinkage void fn_call_stub(void);
extern asmlinkage void fn_rtn_stub(void);
/*					   *INDENT-OFF*	 */
/*__asm__("fn_rtn_stub:\n\t"
	"movl %eax,%esp\n\t"
	"fn_call_stub:\n\t"
	"1:\n\t"
	"addl $-4,%ebx\n\t"
	"movl (%ebx), %eax\n\t"
	"pushl %eax\n\t"
	"cmpl %esp,%ecx\n\t"
	"jne  1b\n\t"
	"popl %eax\n\t"
	"popl %ebx\n\t"
	"popl %ecx\n\t"
	"iret \n\t");
*/
/*					     *INDENT-ON*  */
#define gdb_ia64vector	kgdb_info.vector
#define gdb_ia64errcode kgdb_info.errcode
#define waiting_cpus	kgdb_info.cpus_waiting
#define remote_debug	kgdb_info.print_debug_info
#define hold_cpu(cpu)	kgdb_info.cpus_waiting[cpu].hold
/* gdb locks */

#ifdef CONFIG_SMP
static int in_kgdb_called;
static spinlock_t waitlocks[MAX_NO_CPUS] =
    {[0 ... MAX_NO_CPUS - 1] = SPIN_LOCK_UNLOCKED };
/*
 * The following array has the thread pointer of each of the "other"
 * cpus.  We make it global so it can be seen by gdb.
 */
volatile int in_kgdb_entry_log[MAX_NO_CPUS];
volatile struct pt_regs *in_kgdb_here_log[MAX_NO_CPUS];
/*
static spinlock_t continuelocks[MAX_NO_CPUS];
*/
spinlock_t kgdb_spinlock = SPIN_LOCK_UNLOCKED;
/* waiters on our spinlock plus us */
static atomic_t spinlock_waiters = ATOMIC_INIT(1);
static int spinlock_count = 0;
static int spinlock_cpu = 0;
/*
 * Note we use nested spin locks to account for the case where a break
 * point is encountered when calling a function by user direction from
 * kgdb. Also there is the memory exception recursion to account for.
 * Well, yes, but this lets other cpus thru too.  Lets add a
 * cpu id to the lock.
 */
#define KGDB_SPIN_LOCK(x) if( spinlock_count == 0 || \
			      spinlock_cpu != smp_processor_id()){\
				      atomic_inc(&spinlock_waiters); \
				      while (! spin_trylock(x)) {\
					    in_kgdb(linux_regs, unw_info);\
				      }\
				      atomic_dec(&spinlock_waiters); \
				      spinlock_count = 1; \
				      spinlock_cpu = smp_processor_id(); \
			  }else{  \
				      spinlock_count++; \
			  }
#define KGDB_SPIN_UNLOCK(x) if( --spinlock_count == 0) spin_unlock(x)
#else
unsigned kgdb_spinlock = 0;
#define KGDB_SPIN_LOCK(x) --*x
#define KGDB_SPIN_UNLOCK(x) ++*x
#endif

int
hex(char ch)
{
	if ((ch >= 'a') && (ch <= 'f'))
		return (ch - 'a' + 10);
	if ((ch >= '0') && (ch <= '9'))
		return (ch - '0');
	if ((ch >= 'A') && (ch <= 'F'))
		return (ch - 'A' + 10);
	return (-1);
}

/* scan for the sequence $<data>#<checksum>	*/
void
getpacket(char *buffer)
{
	unsigned char checksum;
	unsigned char xmitcsum;
	int i;
	int count;
	char ch;

	do {
		/* wait around for the start character, ignore all other characters */
		while ((ch = (getDebugChar() & 0x7f)) != '$') ;
		checksum = 0;
		xmitcsum = -1;

		count = 0;

		/* now, read until a # or end of buffer is found */
		while (count < BUFMAX) {
			ch = getDebugChar() & 0x7f;
			if (ch == '#')
				break;
			checksum = checksum + ch;
			buffer[count] = ch;
			count = count + 1;
		}
		buffer[count] = 0;

		if (ch == '#') {
			xmitcsum = hex(getDebugChar() & 0x7f) << 4;
			xmitcsum += hex(getDebugChar() & 0x7f);
			if ((remote_debug) && (checksum != xmitcsum)) {
				printk
				    ("bad checksum.	My count = 0x%x, sent=0x%x. buf=%s\n",
				     checksum, xmitcsum, buffer);
			}

			if (checksum != xmitcsum) {
				putDebugChar('-');	/* failed checksum */
			}
			else {
				putDebugChar('+');	/* successful transfer */
				/* if a sequence char is present, reply the sequence ID */
				if (buffer[2] == ':') {
					putDebugChar(buffer[0]);
					putDebugChar(buffer[1]);
					/* remove sequence chars from buffer */
					count = strlen(buffer);
					for (i = 3; i <= count; i++)
						buffer[i - 3] = buffer[i];
				}
			}
		}
	} while (checksum != xmitcsum);

	if (remote_debug)
		printk("R:%s\n", buffer);
	flushDebugChar();
}

/* send the packet in buffer.  */

void
putpacket(char *buffer)
{
	unsigned char checksum;
	int count;
	char ch;

	/*  $<packet info>#<checksum>. */

	if (!kgdboe) {
		do {
			if (remote_debug)
				printk("T:%s\n", buffer);
			putDebugChar('$');
			checksum = 0;
			count = 0;

			while ((ch = buffer[count])) {
				putDebugChar(ch);
				checksum += ch;
				count += 1;
				if (!rpacket_size && ((count % 200) == 0))
					udelay(norpacket_udelay);
			}

			putDebugChar('#');
			putDebugChar(hexchars[checksum >> 4]);
			putDebugChar(hexchars[checksum % 16]);
			flushDebugChar();

		} while ((getDebugChar() & 0x7f) != '+');
	} else {
		/*
		 * For udp, we can not transfer too much bytes once.
		 * We only transfer MAX_SEND_COUNT size bytes each time
		 */

#define MAX_SEND_COUNT 30

		int send_count = 0, i = 0;
		char send_buf[MAX_SEND_COUNT];

		do {
			if (remote_debug)
				printk("T:%s\n", buffer);
			putDebugChar('$');
			checksum = 0;
			count = 0;
			send_count = 0;
			while ((ch = buffer[count])) {
				if (send_count >= MAX_SEND_COUNT) {
					for(i = 0; i < MAX_SEND_COUNT; i++) {
						putDebugChar(send_buf[i]);
					}
					flushDebugChar();
					send_count = 0;
				} else {
					send_buf[send_count] = ch;
					checksum += ch;
					count ++;
					send_count++;
				}
			}
			for(i = 0; i < send_count; i++)
				putDebugChar(send_buf[i]);
			putDebugChar('#');
			putDebugChar(hexchars[checksum >> 4]);
			putDebugChar(hexchars[checksum % 16]);
			flushDebugChar();
		} while ((getDebugChar() & 0x7f) != '+');
	}
}


void
debug_error(char *format, char *parm)
{
	if (remote_debug)
		printk(format, parm);
}

static void
print_regs(struct pt_regs *regs)
{
	printk("B0=0x%lx\n", regs->b0);
	printk("IPSR=0x%lx\n", regs->cr_ipsr);
	printk("IIP=0x%lx\n", regs->cr_iip);
	printk("\n");

}				/* print_regs */



static void
unw_ar_regs(struct unw_frame_info *unw, unsigned long *regs, int put)
{
	if (put)
		unw_access_ar(unw, UNW_AR_BSP, &regs[REGISTER_INDEX(IA64_BSP_REGNUM)], put);
	else
		unw_get_bsp(unw, &regs[REGISTER_INDEX(IA64_BSP_REGNUM)]);
	unw_access_ar(unw, UNW_AR_BSPSTORE, &regs[REGISTER_INDEX(IA64_BSPSTORE_REGNUM)], put);
	unw_access_ar(unw, UNW_AR_PFS, &regs[REGISTER_INDEX(IA64_PFS_REGNUM)], put);
	unw_access_ar(unw, UNW_AR_RNAT, &regs[REGISTER_INDEX(IA64_RNAT_REGNUM)], put);
	unw_access_ar(unw, UNW_AR_UNAT, &regs[REGISTER_INDEX(IA64_UNAT_REGNUM)], put);
	unw_access_ar(unw, UNW_AR_LC, &regs[REGISTER_INDEX(IA64_LC_REGNUM)], put);
	unw_access_ar(unw, UNW_AR_EC, &regs[REGISTER_INDEX(IA64_EC_REGNUM)], put);
	unw_access_ar(unw, UNW_AR_FPSR, &regs[REGISTER_INDEX(IA64_FPSR_REGNUM)], put);
	unw_access_ar(unw, UNW_AR_RSC, &regs[REGISTER_INDEX(IA64_RSC_REGNUM)], put);
	unw_access_ar(unw, UNW_AR_CCV, &regs[REGISTER_INDEX(IA64_CCV_REGNUM)], put);
	unw_access_ar(unw, UNW_AR_CSD, &regs[REGISTER_INDEX(IA64_CSD_REGNUM)], put);
	unw_access_ar(unw, UNW_AR_SSD, &regs[REGISTER_INDEX(IA64_SSD_REGNUM)], put);
	return;
}

static void
unw_get_regs(struct unw_frame_info *unw, unsigned long *regs, struct pt_regs *ptregs)
{
	int i, j;
	char nat;
	struct ia64_fpreg *fr, freg;

	memset(regs, 0, sizeof (kgdb_info.ia64_regs));

	for (i = 1; i < 32; i++) {
		unw_access_gr(unw, i, &regs[REGISTER_INDEX(IA64_GR0_REGNUM + i)], &nat, 0);
		regs[REGISTER_INDEX(IA64_NAT0_REGNUM + i)] = nat;
	}

	for (i = 1; i < 8; i++)
		unw_access_br(unw, i, &regs[REGISTER_INDEX(IA64_BR0_REGNUM + i)], 0);

	if (ptregs)
		regs[REGISTER_INDEX(IA64_BR0_REGNUM)] = ptregs->b0;
	else
		unw_access_br(unw, 0, &regs[REGISTER_INDEX(IA64_BR0_REGNUM)], 0);

	if (ptregs)
		fr = &ptregs->f6;
	else
		fr = &freg;

	for (i = 6; i < 12; i++) {
		if (!ptregs)
			unw_access_fr(unw, i, fr, 0);
		regs[REGISTER_INDEX(IA64_FR0_REGNUM + 6 + i)] = fr->u.bits[0];
		regs[REGISTER_INDEX(IA64_FR0_REGNUM + 6 + i + 1)] = fr->u.bits[1];
		if (ptregs)
			fr++;
	}

	unw_ar_regs(unw, regs, 0);

	unw_access_pr(unw, &regs[REGISTER_INDEX(IA64_PR_REGNUM)], 0);
	for (j = 1, i = 0; i < 64; i++, j <<= 1, i++)
		regs[REGISTER_INDEX(IA64_PR0_REGNUM + i)] =
			(regs[REGISTER_INDEX(IA64_PR_REGNUM)] & j) ? 1 : 0;


	if (!ptregs) {
		unw_get_ip(unw, &regs[REGISTER_INDEX(IA64_IP_REGNUM)]);
		unw_get_cfm(unw, &regs[REGISTER_INDEX(IA64_CFM_REGNUM)]);
	} else {
		regs[REGISTER_INDEX(IA64_IP_REGNUM)] = ptregs->cr_iip;
		regs[REGISTER_INDEX(IA64_PSR_REGNUM)] = ptregs->cr_ipsr;
		regs[REGISTER_INDEX(IA64_CFM_REGNUM)] = ptregs->cr_ifs;
	}

	return;
}

static void
kgdb_get_block_task(struct task_struct *p, unsigned long *ia64regs)
{
	struct unw_frame_info info;
	unsigned long ip;
	int count = 0;

	unw_init_from_blocked_task(&info, p);
	ip = 0UL;
	do {
		if (unw_unwind(&info) < 0)
			return;
		unw_get_ip(&info, &ip);
		if (!in_sched_functions(ip))
			break;
	} while (count++ < 16);

	if (!ip)
		return;

	unw_get_regs(&info, ia64regs, (struct pt_regs *) 0);

	return;
}

static void
regs_to_gdb_regs(struct kgdb_state *state)
{
	unw_get_regs(state->unw, kgdb_info.ia64_regs, state->regs);
	return;
}				/* regs_to_gdb_regs */

static void
gdb_regs_to_regs(struct kgdb_state *state)
{
	int i;
	char nat;
	unsigned long *regs;
	struct ia64_fpreg *fr;
	struct unw_frame_info *unw;

	unw = state->unw;
	regs = kgdb_info.ia64_regs;

	for (i = 1; i < 32; i++) {
		nat = (char) regs[REGISTER_INDEX(IA64_NAT0_REGNUM + i)];
		unw_access_gr(unw, i, &regs[REGISTER_INDEX(IA64_GR0_REGNUM + i)], &nat, 1);
	}

	for (i = 0; i < 8; i++)
		unw_access_br(unw, i, &regs[REGISTER_INDEX(IA64_BR0_REGNUM + i)], 1);

	for (fr = &state->regs->f6, i = 6; i < 12; i++, fr++) {
		fr->u.bits[0] = regs[REGISTER_INDEX(IA64_FR0_REGNUM + 6 + i)];
		fr->u.bits[1] = regs[REGISTER_INDEX(IA64_FR0_REGNUM + 6 + i + 1)];
	}

	unw_ar_regs(unw, regs, 1);

	unw_access_pr(unw, &regs[IA64_PR_REGNUM], 1);

	state->regs->cr_iip = regs[REGISTER_INDEX(IA64_IP_REGNUM)];
	state->regs->cr_ipsr = regs[REGISTER_INDEX(IA64_PSR_REGNUM)];
	state->regs->cr_ifs = regs[REGISTER_INDEX(IA64_CFM_REGNUM)];

	return;

}				/* gdb_regs_to_regs */

int thread_list = 0;

void
get_gdb_regs(struct task_struct *p, struct kgdb_state *state)
{
	IF_SMP(int i);
	if (!p || p == current) {
		regs_to_gdb_regs(state);
		return;
	}
#ifdef CONFIG_SMP
	for (i = 0; i < MAX_NO_CPUS; i++) {
		if (p == kgdb_info.cpus_waiting[i].task &&
			!user_mode(kgdb_info.cpus_waiting[i].regs)){
			memcpy(kgdb_info.ia64_regs, kgdb_info.cpus_waiting[i].ia64_regs,
				sizeof(kgdb_info.ia64_regs));
			return;
		}
	}
#endif
/*
 * This code is to give a more informative notion of where a process
 * is waiting.	It is used only when the user asks for a thread info
 * list.  If he then switches to the thread, s/he will find the task
 * is in schedule, but a back trace should show the same info we come
 * up with. Some of this code was purloined from get_wchan;
 */

	if (!thread_list)
		return;
	else if (p->state == TASK_RUNNING)
		return;

	kgdb_get_block_task(p, kgdb_info.ia64_regs);

	return;

}

/* Indicate to caller of mem2hex or hex2mem that there has been an
   error.  */
static volatile int mem_err = 0;
static volatile int mem_err_expected = 0;
static volatile int mem_err_cnt = 0;

int
get_char(char *addr)
{
	return *addr;
}

void
set_char(char *addr, int val, int may_fault)
{
	if (may_fault) {
		unsigned long valid;
		if ((unsigned long) addr < PAGE_SIZE)
			valid = 0;
		else
			asm volatile("probe.w %0 = %1, 0" : "=r" (valid) : "r" (addr) : "memory");
		if (!valid) {
			mem_err = 1;
			return;
		}

	}
	*addr = val;
}

/* convert the memory pointed to by mem into hex, placing result in buf */
/* return a pointer to the last char put in buf (null) */
/* If MAY_FAULT is non-zero, then we should set mem_err in response to
   a fault; if zero treat a fault like any other fault in the stub.  */
char *
mem2hex(char *mem, char *buf, int count, int may_fault)
{
	int i;
	unsigned char ch;

	if (may_fault) {
		mem_err_expected = 1;
		mem_err = 0;
	}
	for (i = 0; i < count; i++) {
		/* printk("%lx = ", mem) ; */

		if (may_fault) {
			unsigned long valid;
			if ((unsigned long) mem < PAGE_SIZE)
				valid = 0;
			else
				asm volatile("probe.r %0 = %1, 0" :
					"=r" (valid) : "r" (mem) : "memory");
			if (!valid)
				mem_err = 1;
			else
				ch = get_char(mem++);
		}
		else
			ch = get_char(mem++);

		/* printk("%02x\n", ch & 0xFF) ; */
		if (may_fault && mem_err) {
			if (remote_debug)
				printk("Mem fault fetching from addr %lx\n",
				       (long) (mem - 1));
			*buf = 0;	/* truncate buffer */
			return (buf);
		}
		*buf++ = hexchars[ch >> 4];
		*buf++ = hexchars[ch % 16];
	}
	*buf = 0;
	if (may_fault)
		mem_err_expected = 0;
	return (buf);
}

/* convert the hex array pointed to by buf into binary to be placed in mem */
/* return a pointer to the character AFTER the last byte written */
/* NOTE: We use the may fault flag to also indicate if the write is to
 * the registers (0) or "other" memory (!=0)
 */
char *
hex2mem(char *buf, char *mem, int count, int may_fault)
{
	int i;
	unsigned char ch;

	if (may_fault) {
		mem_err_expected = 1;
		mem_err = 0;
	}
	for (i = 0; i < count; i++) {
		ch = hex(*buf++) << 4;
		ch = ch + hex(*buf++);
		set_char(mem++, ch, may_fault);

		if (may_fault && mem_err) {
			if (remote_debug)
				printk("Mem fault storing to addr %lx\n",
				       (long) (mem - 1));
			return (mem);
		}
	}
	if (may_fault)
		mem_err_expected = 0;
	return (mem);
}

/**********************************************/
/* WHILE WE FIND NICE HEX CHARS, BUILD AN INT */
/* RETURN NUMBER OF CHARS PROCESSED	      */
/**********************************************/
int
hexToLong(char **ptr, unsigned long *intValue)
{
	int numChars = 0;
	int hexValue;

	*intValue = 0UL;

	while (**ptr) {
		hexValue = hex(**ptr);
		if (hexValue >= 0) {
			*intValue = (*intValue << 4) | hexValue;
			numChars++;
		} else
			break;

		(*ptr)++;
	}

	return (numChars);
}

int
hexToInt(char **ptr, int *intValue)
{
	int numChars = 0;
	int hexValue;

	*intValue = 0;

	while (**ptr) {
		hexValue = hex(**ptr);
		if (hexValue >= 0) {
			*intValue = (*intValue << 4) | hexValue;
			numChars++;
		} else
			break;

		(*ptr)++;
	}

	return (numChars);
}

#define stubhex(h) hex(h)
#ifdef old_thread_list

static int
stub_unpack_int(char *buff, int fieldlength)
{
	int nibble;
	int retval = 0;

	while (fieldlength) {
		nibble = stubhex(*buff++);
		retval |= nibble;
		fieldlength--;
		if (fieldlength)
			retval = retval << 4;
	}
	return retval;
}
#endif
static char *
pack_hex_byte(char *pkt, int byte)
{
	*pkt++ = hexchars[(byte >> 4) & 0xf];
	*pkt++ = hexchars[(byte & 0xf)];
	return pkt;
}

#define BUF_THREAD_ID_SIZE 16

static char *
pack_threadid(char *pkt, threadref * id)
{
	char *limit;
	unsigned char *altid;

	altid = (unsigned char *) id;
	limit = pkt + BUF_THREAD_ID_SIZE;
	while (pkt < limit)
		pkt = pack_hex_byte(pkt, *altid++);
	return pkt;
}

#ifdef old_thread_list
static char *
unpack_byte(char *buf, int *value)
{
	*value = stub_unpack_int(buf, 2);
	return buf + 2;
}

static char *
unpack_threadid(char *inbuf, threadref * id)
{
	char *altref;
	char *limit = inbuf + BUF_THREAD_ID_SIZE;
	int x, y;

	altref = (char *) id;

	while (inbuf < limit) {
		x = stubhex(*inbuf++);
		y = stubhex(*inbuf++);
		*altref++ = (x << 4) | y;
	}
	return inbuf;
}
#endif
void
int_to_threadref(threadref * id, int value)
{
	unsigned char *scan;

	scan = (unsigned char *) id;
	{
		int i = 4;
		while (i--)
			*scan++ = 0;
	}
	*scan++ = (value >> 24) & 0xff;
	*scan++ = (value >> 16) & 0xff;
	*scan++ = (value >> 8) & 0xff;
	*scan++ = (value & 0xff);
}
int
int_to_hex_v(unsigned char * id, int value)
{
	unsigned char *start = id;
	int shift;
	int ch;

	for (shift = 28; shift >= 0; shift -= 4) {
		if ((ch = (value >> shift) & 0xf) || (id != start)) {
			*id = hexchars[ch];
			id++;
		}
	}
	if (id == start)
		*id++ = '0';
	return id - start;
}
#ifdef old_thread_list

static int
threadref_to_int(threadref * ref)
{
	int i, value = 0;
	unsigned char *scan;

	scan = (char *) ref;
	scan += 4;
	i = 4;
	while (i-- > 0)
		value = (value << 8) | ((*scan++) & 0xff);
	return value;
}
#endif
static int
cmp_str(char *s1, char *s2, int count)
{
	while (count--) {
		if (*s1++ != *s2++)
			return 0;
	}
	return 1;
}

#if 1				/* this is a hold over from 2.4 where O(1) was "sometimes" */
extern struct task_struct *kgdb_get_idle(int cpu);
#define idle_task(cpu) kgdb_get_idle(cpu)
#else
#define idle_task(cpu) init_tasks[cpu]
#endif

extern int kgdb_pid_init_done;

#ifdef	CONFIG_KGDB_EARLY
struct task_struct kgdb_task = {.comm = "kgdb-dummy"};
#endif

struct task_struct *
getthread(int pid)
{
	struct task_struct *thread;
	if (pid >= PID_MAX && pid <= (PID_MAX + MAX_NO_CPUS)) {

		return idle_task(pid - PID_MAX);
	} else {
		/*
		 * find_task_by_pid is relatively safe all the time
		 * Other pid functions require lock downs which imply
		 * that we may be interrupting them (as we get here
		 * in the middle of most any lock down).
		 * Still we don't want to call until the table exists!
		 */
		if (kgdb_pid_init_done){
			thread = find_task_by_pid(pid);
			if (thread) {
				return thread;
			}
		}
	}
#ifdef	CONFIG_KGDB_EARLY
	if (!kgdb_pid_init_done)
		return &kgdb_task;
#endif
	return NULL;
}
/* *INDENT-OFF*	 */


unsigned long hw_breakpoint_status;

int hw_breakpoint_init;

void
do_init_hw_break(void)
{
	s64 status;
	int i;

	hw_breakpoint_init = 1;

#ifdef	CONFIG_IA64_HP_SIM
	hw_break_total_ibr = 8;
	hw_break_total_dbr = 8;
	status = 0;
#else
	status = ia64_pal_debug_info(&hw_break_total_ibr, &hw_break_total_dbr);
#endif

	if (status) {
		printk(KERN_INFO "do_init_hw_break: pal call failed %d\n", (int) status);
		return;
	}

	if (HW_BREAKPOINT > MAX_HW_BREAKPOINT) {
		printk(KERN_INFO "do_init_hw_break: %d exceeds max %d\n", (int) HW_BREAKPOINT,
			(int) MAX_HW_BREAKPOINT);

		while ((HW_BREAKPOINT > MAX_HW_BREAKPOINT) && hw_break_total_ibr != 1)
			hw_break_total_ibr--;
		while (HW_BREAKPOINT > MAX_HW_BREAKPOINT)
			hw_break_total_dbr--;
	}

	breakinfo = malloc(HW_BREAKPOINT * sizeof(struct hw_breakpoint));

	if (!breakinfo) {
		printk(KERN_INFO "Failed to allocate hardware break array\n");
		return;
	}

	memset(breakinfo, 0, HW_BREAKPOINT * sizeof(struct hw_breakpoint));

	for (i = 0; i < hw_break_total_dbr; i++)
		breakinfo[i].capable = HWCAP_DBR;

	for (; i < HW_BREAKPOINT; i++)
		breakinfo[i].capable = HWCAP_IBR;

	return;
}

void
correct_hw_break(void)
{
	int breakno;

	if (!breakinfo)
		return;

	for (breakno = 0; breakno < HW_BREAKPOINT; breakno++) {
		if (breakinfo[breakno].enabled) {
			if (breakinfo[breakno].capable & HWCAP_IBR) {
				int ibreakno = breakno - hw_break_total_dbr;
				ia64_set_ibr(ibreakno << 1, breakinfo[breakno].addr);
				ia64_set_ibr((ibreakno << 1) + 1,
					(~breakinfo[breakno].mask & ((1UL << 56UL) - 1)) |
					(1UL << 56UL) | (1UL << 63UL));
			}
			else {
				ia64_set_dbr(breakno << 1, breakinfo[breakno].addr);
				ia64_set_dbr((breakno << 1) + 1,
					(~breakinfo[breakno].mask & ((1UL << 56UL) - 1)) |
					(1UL << 56UL) | (breakinfo[breakno].type << 62UL));
			}
		}
		else  {
			if (breakinfo[breakno].capable & HWCAP_IBR)
				ia64_set_ibr(((breakno - hw_break_total_dbr) << 1) + 1, 0);
			else
				ia64_set_dbr((breakno << 1) + 1, 0);
		}
	}

	return;
}

int
hardware_breakpoint(unsigned long addr, int length, int type, int action)
{
	int breakno, found, watch;
	unsigned long mask;
	extern unsigned long _start[];

	if (!hw_breakpoint_init)
		do_init_hw_break();

	if (!breakinfo)
		return 0;
	else if (addr == (unsigned long) _start)
		return 1;

	if (type == WATCH_ACCESS)
		mask = HWCAP_DBR;
	else
		mask = 1UL << type;

	for (watch = 0, found = 0, breakno = 0; breakno < HW_BREAKPOINT; breakno++) {
		if (action) {
			if (breakinfo[breakno].enabled || !(breakinfo[breakno].capable & mask))
				continue;
			breakinfo[breakno].enabled = 1;
			breakinfo[breakno].type = type;
			breakinfo[breakno].mask = length - 1;
			breakinfo[breakno].addr = addr;
			watch = breakno;
		} else if (breakinfo[breakno].enabled &&
			((length < 0 && breakinfo[breakno].addr == addr) ||
			((breakinfo[breakno].capable & mask) &&
			(breakinfo[breakno].mask == (length - 1)) &&
			(breakinfo[breakno].addr == addr)))) {
			breakinfo[breakno].enabled = 0;
			breakinfo[breakno].type = 0UL;
		}
		else
			continue;
		found++;
		if (type != WATCH_ACCESS)
			break;
		else if (found == 2)
			break;
		else
			mask = HWCAP_IBR;
	}

	if (type == WATCH_ACCESS && found == 1) {
		breakinfo[watch].enabled = 0;
		found = 0;
	}

	return found;
}

#ifdef	oldbreak_protocol
int
remove_hw_break(unsigned breakno)
{
	if (!breakinfo[breakno].enabled)
		return -1;

	breakinfo[breakno].enabled = 0;

	return 0;
}

int
set_hw_break(unsigned breakno, unsigned type, unsigned len, unsigned addr)
{
	if (breakinfo[breakno].enabled)
		return -1;

	breakinfo[breakno].enabled = 1;
	breakinfo[breakno].type = type;
	breakinfo[breakno].mask = len - 1;
	breakinfo[breakno].addr = addr;
	return 0;
}
#endif

static void inline
normalize(struct unw_frame_info *running, struct pt_regs *regs)
{
	unsigned long sp;

	/*
	 * unwind to last frame before exception which will be an error
	 * and then fetch the bsp at exception time.
	 */

	do {
		unw_get_sp(running, &sp);
		if ((sp + 0x10) >= (unsigned long) regs)
			break;
	} while (unw_unwind(running) >= 0);

	return;
}

#ifdef CONFIG_SMP
static int in_kgdb_console = 0;


static void
snap_regs(struct unw_frame_info *unw_info, void *data)
{
	struct pt_regs *regs;
	int cpu;

	regs = data;
	normalize(unw_info, regs);
	cpu = smp_processor_id();
	unw_get_regs(unw_info, kgdb_info.cpus_waiting[cpu].ia64_regs, regs);

	return;
}

int
in_kgdb(struct pt_regs *regs, struct unw_frame_info *unw_info)
{
	unsigned flags;
	int cpu = smp_processor_id();
	in_kgdb_called = 1;

	preempt_disable();

	if (!spin_is_locked(&kgdb_spinlock)) {
		if (in_kgdb_here_log[cpu] ||	/* we are holding this cpu */
		    in_kgdb_console) {	/* or we are doing slow i/o */
			preempt_enable_no_resched();
			return 1;
		}
		preempt_enable_no_resched();
		return 0;
	}

	/* As I see it the only reason not to let all cpus spin on
	 * the same spin_lock is to allow selected ones to proceed.
	 * This would be a good thing, so we leave it this way.
	 * Maybe someday....  Done !

	 * in_kgdb() is called from an NMI so we don't pretend
	 * to have any resources, like printk() for example.
	 */

	kgdb_local_irq_save(flags);	/* only local here, to avoid hanging */
	/*
	 * log arival of this cpu
	 * The NMI keeps on ticking.  Protect against recurring more
	 * than once, and ignor the cpu that has the kgdb lock
	 */
	in_kgdb_entry_log[cpu]++;
	in_kgdb_here_log[cpu] = regs;
	if (cpu == spinlock_cpu || waiting_cpus[cpu].task)
		goto exit_in_kgdb;

	/*
	 * For protection of the initilization of the spin locks by kgdb
	 * it locks the kgdb spinlock before it gets the wait locks set
	 * up.	We wait here for the wait lock to be taken.  If the
	 * kgdb lock goes away first??	Well, it could be a slow exit
	 * sequence where the wait lock is removed prior to the kgdb lock
	 * so if kgdb gets unlocked, we just exit.
	 */

	while (spin_is_locked(&kgdb_spinlock) &&
	       !spin_is_locked(waitlocks + cpu)) ;
	if (!spin_is_locked(&kgdb_spinlock))
		goto exit_in_kgdb;

	if (unw_info)
		unw_get_regs(unw_info, kgdb_info.cpus_waiting[cpu].ia64_regs, regs);
	else if (!user_mode(regs)) {
		if (current->state == TASK_RUNNING)
			unw_init_running(snap_regs, regs);
		else
			kgdb_get_block_task(current, kgdb_info.cpus_waiting[cpu].ia64_regs);
	}
	waiting_cpus[cpu].task = current;
	waiting_cpus[cpu].pid = (current->pid) ? : (PID_MAX + cpu);
	waiting_cpus[cpu].regs = regs;

	spin_unlock_wait(waitlocks + cpu);


	/*
	 * log departure of this cpu
	 */
	waiting_cpus[cpu].task = 0;
	waiting_cpus[cpu].pid = 0;
	waiting_cpus[cpu].regs = 0;
	correct_hw_break();
      exit_in_kgdb:
	in_kgdb_here_log[cpu] = 0;
	kgdb_local_irq_restore(flags);
	preempt_enable_no_resched();
	return 1;
	/*
	   spin_unlock(continuelocks + smp_processor_id());
	 */
}

void
smp__in_kgdb(struct pt_regs regs)
{
	in_kgdb(&regs, NULL);
}
#else
int
in_kgdb(struct pt_regs *regs, struct unw_frame_info *unw)
{
	return (kgdb_spinlock);
}
#endif

void
printexceptioninfo(int exceptionNo, int errorcode, char *buffer)
{
	switch (exceptionNo) {
	case 1:		/* debug exception */
		break;
	case 3:		/* breakpoint */
		sprintf(buffer, "Software breakpoint");
		return;
	default:
		sprintf(buffer, "Details not available");
		return;
	}
	sprintf(buffer, "Unknown trap");
	return;
}

/*
 * This function does all command procesing for interfacing to gdb.
 *
 * NOTE:  The INT nn instruction leaves the state of the interrupt
 *	  enable flag UNCHANGED.  That means that when this routine
 *	  is entered via a breakpoint (INT 3) instruction from code
 *	  that has interrupts enabled, then interrupts will STILL BE
 *	  enabled when this routine is entered.	 The first thing that
 *	  we do here is disable interrupts so as to prevent recursive
 *	  entries and bothersome serial interrupts while we are
 *	  trying to run the serial port in polled mode.
 *
 * For kernel version 2.1.xx the kgdb_cli() actually gets a spin lock so
 * it is always necessary to do a restore_flags before returning
 * so as to let go of that lock.
 */

static void do_kgdb_handle_exception(struct unw_frame_info *, void *data);

int
kgdb_handle_exception(int exceptionVector, int signo, unsigned long err_code, struct pt_regs *linux_regs)
{
	struct kgdb_state info;

	/*
	 * If the entry is not from the kernel then return to the Linux
	 * trap handler and let it process the interrupt normally.
	 */
	if (user_mode(linux_regs)) {
		printk("ignoring non-kernel exception\n");
		print_regs(linux_regs);
		return (0);
	}

	preempt_disable();

	kgdb_info.called_from = __builtin_return_address(0);

	info.exceptionVector = exceptionVector;
	info.signo = signo;
	info.err_code = err_code;
	info.regs = linux_regs;
	info.unw = (void *) 0;
	unw_init_running(do_kgdb_handle_exception, &info);

	preempt_enable_no_resched();

	return info.ret;
}

static int kgdb_dbregs_enabled;

static void
do_kgdb_handle_exception(struct unw_frame_info *unw_info, void *data)
{
	int exceptionVector, signo, have_regs = 0;
	unsigned long err_code;
	struct pt_regs *linux_regs;
	struct kgdb_state *info;
	struct task_struct *usethread = NULL;
	struct task_struct *thread_list_start = 0, *thread = NULL;
	int length;
	unsigned long addr;
	char *ptr;
	int newPC;
	threadref thref;
	int threadid;
	int thread_min = PID_MAX + MAX_NO_CPUS;
#ifdef old_thread_list
	int maxthreads;
#endif
	int nothreads;
	unsigned long flags;
#define	gdb_regs	kgdb_info.ia64_regs
	IF_SMP(int entry_state = 0);	/* 0, ok, 1, no nmi, 2 sync failed */
#define NO_NMI 1
#define NO_SYNC 2
#define	ptregs	(*linux_regs)
#define NUMREGS NUM_REGS

	info = data;
	info->unw = unw_info;
	exceptionVector = info->exceptionVector;
	signo = info->signo;
	err_code = info->err_code;
	linux_regs = info->regs;

	normalize(unw_info, linux_regs);

	/*
	 * If we're using eth mode, set the 'mode' in the netdevice.
	 */
	if (kgdboe)
		netpoll_set_trap(1);

	kgdb_local_irq_save(flags);

	/* Get kgdb spinlock */

	KGDB_SPIN_LOCK(&kgdb_spinlock);
	kgdb_info.entry_itc = ia64_get_itc();
	/*
	 * We depend on this spinlock and the NMI watch dog to control the
	 * other cpus.	They will arrive at "in_kgdb()" as a result of the
	 * NMI and will wait there for the following spin locks to be
	 * released.
	 */
#ifdef CONFIG_SMP

#if 0
	if (cpu_callout_map & ~MAX_CPU_MASK) {
		printk("kgdb : too many cpus, possibly not mapped"
		       " in contiguous space, change MAX_NO_CPUS"
		       " in kgdb_stub and make new kernel.\n"
		       " cpu_callout_map is %lx\n", cpu_callout_map);
		goto exit_just_unlock;
	}
#endif
	if (spinlock_count == 1) {
		long time = 0, end_time;
		int i;
		int cpu_logged_in[MAX_NO_CPUS] = {[0 ... MAX_NO_CPUS - 1] = (0)
		};
		if (remote_debug) {
			printk("kgdb : cpu %d entry, syncing others\n",
			       smp_processor_id());
		}
		for (i = 0; i < MAX_NO_CPUS; i++) {
			/*
			 * Use trylock as we may already hold the lock if
			 * we are holding the cpu.  Net result is all
			 * locked.
			 */
			spin_trylock(&waitlocks[i]);
		}
		for (i = 0; i < MAX_NO_CPUS; i++)
			cpu_logged_in[i] = 0;
		/*
		 * Wait for their arrival.  We know the watch dog is active if
		 * in_kgdb() has ever been called, as it is always called on a
		 * watchdog tick.
		 */
		time = ia64_get_itc();
		end_time = time + 2;	/* Note: we use the High order bits! */
		i = 1;
		if (num_online_cpus() > 1) {
			int me_in_kgdb = in_kgdb_entry_log[smp_processor_id()];
			smp_send_nmi_allbutself();

			while (i < num_online_cpus() && time != end_time) {
				int j;
				for (j = 0; j < MAX_NO_CPUS; j++) {
					if (waiting_cpus[j].task &&
					    waiting_cpus[j].task != NOCPU &&
					    !cpu_logged_in[j]) {
						i++;
						cpu_logged_in[j] = 1;
						if (remote_debug) {
							printk
							    ("kgdb : cpu %d arrived at kgdb\n",
							     j);
						}
						break;
					} else if (!waiting_cpus[j].task &&
						   !cpu_online(j)) {
						waiting_cpus[j].task = NOCPU;
						cpu_logged_in[j] = 1;
						waiting_cpus[j].hold = 1;
						break;
					}
					if (!waiting_cpus[j].task && in_kgdb_here_log[j]) {
						int wait = 100000;
						while (wait-- && !waiting_cpus[j].task );
						if (!waiting_cpus[j].task  &&
							in_kgdb_here_log[j]) {
							printk
							    ("kgdb : cpu %d stall"
							     " in in_kgdb\n",
							     j);
							i++;
							cpu_logged_in[j] = 1;
							waiting_cpus[j].task =
								(struct task_struct *) 1;
						}
					}
				}

				if (in_kgdb_entry_log[smp_processor_id()] >
				    (me_in_kgdb + 10)) {
					break;
				}

				time = ia64_get_itc();
			}
			if (i < num_online_cpus()) {
				printk
				    ("kgdb : time out, proceeding without sync\n");
#if 0
				printk("kgdb : Waiting_cpus: 0 = %d, 1 = %d\n",
				       waiting_cpus[0].task != 0,
				       waiting_cpus[1].task != 0);
				printk("kgdb : Cpu_logged in: 0 = %d, 1 = %d\n",
				       cpu_logged_in[0], cpu_logged_in[1]);
				printk
				    ("kgdb : in_kgdb_here_log in: 0 = %d, 1 = %d\n",
				     in_kgdb_here_log[0] != 0,
				     in_kgdb_here_log[1] != 0);
#endif
				entry_state = NO_SYNC;
			} else {
#if 0
				int ent =
				    in_kgdb_entry_log[smp_processor_id()] -
				    me_in_kgdb;
				printk("kgdb : sync after %d entries\n", ent);
#endif
			}
		} else {
			if (remote_debug) {
				printk
				    ("kgdb : %ld cpus, but watchdog not active\n"
				     "proceeding without locking down other cpus\n",
				     num_online_cpus());
				entry_state = NO_NMI;
			}
		}
	}
#endif

	if (remote_debug) {
		unsigned long *lp = (unsigned long *) &linux_regs;

		printk("handle_exception(exceptionVector=%d, "
		       "signo=%d, err_code=%ld, linux_regs=%p)\n",
		       exceptionVector, signo, err_code, linux_regs);
		if (debug_regs) {
			print_regs(&ptregs);
			printk("Stk: %8lx %8lx %8lx %8lx"
			       "  %8lx %8lx %8lx %8lx\n",
			       lp[0], lp[1], lp[2], lp[3],
			       lp[4], lp[5], lp[6], lp[7]);
			printk("     %8lx %8lx %8lx %8lx"
			       "  %8lx %8lx %8lx %8lx\n",
			       lp[8], lp[9], lp[10], lp[11],
			       lp[12], lp[13], lp[14], lp[15]);
			printk("     %8lx %8lx %8lx %8lx  "
			       "%8lx %8lx %8lx %8lx\n",
			       lp[16], lp[17], lp[18], lp[19],
			       lp[20], lp[21], lp[22], lp[23]);
			printk("     %8lx %8lx %8lx %8lx  "
			       "%8lx %8lx %8lx %8lx\n",
			       lp[24], lp[25], lp[26], lp[27],
			       lp[28], lp[29], lp[30], lp[31]);
		}
	}

	/* Disable hardware debugging while we are in kgdb */
	/* Get the debug register status register */
	hw_breakpoint_status = ia64_getreg(_IA64_REG_PSR);
	if (hw_breakpoint_status & IA64_PSR_DB)
		ia64_setreg(_IA64_REG_PSR_L, hw_breakpoint_status ^ IA64_PSR_DB);


	switch (exceptionVector) {
	case -1:	/* general death */
	case 11:	/* break  fix signo */
		switch (err_code) { 	/* break_num */
		case BREAKNUM:
			signo = SIGTRAP;
			break;
		case KGDBBREAKNUM:
			signo = SIGTRAP;
			if (ia64_psr(linux_regs)->ri < 2)
				kgdb_pc(linux_regs, linux_regs->cr_iip +
					ia64_psr(linux_regs)->ri + 1);
			else
				kgdb_pc(linux_regs, linux_regs->cr_iip + 16);
			break;
		}
		break;
	case 29:	/*  hardware breakpoint ibr/dbr fault */
	case 36:	/* single step trap */
		signo = SIGTRAP;
		break;
	case 6:		/* ikey_miss */
	case 7:		/* dkey_miss */
	case 24:	/* general exception */
	case 31:	/*  unsupported data reference */
		signo = SIGSEGV;
		break;
	case 13:	/* reserved  */
	case 14:	/*  " ""     */
	case 15:	/*  " ""     */
	case 16:	/*  " ""     */
	case 17:	/*  " ""     */
	case 18:	/*  " ""     */
	case 19:	/*  " ""     */
	case 28:	/*  " ""     */
	case 25:	/* disable fp */
	case 32:	/* floating point */
	case 33:	/* floating point trap */
	case 34:	/* lower privilege fault */
	case 35:	/* taken branch trap */
	case 37:	/* reserved */
	case 38:	/* reserved */
	case 39:	/* reserved */
	case 40:	/* reserved */
	case 41:	/* reserved */
	case 42:	/* reserved */
	case 43:	/* reserved */
	case 44:	/* reserved */
	case 45:	/* ia32_exception */
	case 47:	/* ia32 interrupt  */
	default:	/* reserved and undefined */
		signo = SIGILL;
		break;
	case 5:		/* kernel page fault */
		if (mem_err_expected) {
			/*
			 * This fault occured because of the
			 * get_char or set_char routines.  These
			 * two routines use either eax of edx to
			 * indirectly reference the location in
			 * memory that they are working with.
			 * For a page fault, when we return the
			 * instruction will be retried, so we
			 * have to make sure that these
			 * registers point to valid memory.
			 */
			mem_err = 1;	/* set mem error flag */
			mem_err_expected = 0;
			mem_err_cnt++;	/* helps in debugging */
			if (ia64_psr(linux_regs)->ri < 2)
				kgdb_pc(linux_regs, linux_regs->cr_iip +
					ia64_psr(linux_regs)->ri + 1);
			else
				kgdb_pc(linux_regs, linux_regs->cr_iip + 16);
			if (remote_debug)
				printk("Return after memory error: "
				       "mem_err_cnt=%d\n", mem_err_cnt);
			if (debug_regs)
				print_regs(&ptregs);
			goto exit_kgdb;
		}
		break;
	}
	if (remote_debug)
		printk("kgdb : entered kgdb on cpu %d\n", smp_processor_id());

	gdb_ia64vector = exceptionVector;
	gdb_ia64errcode = err_code;
#ifdef CONFIG_SMP
	/*
	 * OK, we can now communicate, lets tell gdb about the sync.
	 * but only if we had a problem.
	 */
	switch (entry_state) {
	case NO_NMI:
		to_gdb("NMI not active, other cpus not stopped\n");
		break;
	case NO_SYNC:
		to_gdb("Some cpus not stopped, see 'kgdb_info' for details\n");
	default:;
	}

#endif
/*
 * Set up the gdb function call area.
 */

      IF_SMP(once_again:)
	    /* reply to host that an exception has occurred */
	remcomOutBuffer[0] = 'S';
	remcomOutBuffer[1] = hexchars[signo >> 4];
	remcomOutBuffer[2] = hexchars[signo % 16];
	remcomOutBuffer[3] = 0;

	putpacket(remcomOutBuffer);

	while (1 == 1) {
		error = 0;
		remcomOutBuffer[0] = 0;
		getpacket(remcomInBuffer);
		switch (remcomInBuffer[0]) {
		case '?':
			remcomOutBuffer[0] = 'S';
			remcomOutBuffer[1] = hexchars[signo >> 4];
			remcomOutBuffer[2] = hexchars[signo % 16];
			remcomOutBuffer[3] = 0;
			break;
		case 'd':
			remote_debug = !(remote_debug);	/* toggle debug flag */
			printk("Remote debug %s\n",
			       remote_debug ? "on" : "off");
			break;
		case 'p':	/* fetch register */
		{
			int regnum;

			if (!have_regs) {
				get_gdb_regs(usethread, info);
				have_regs = 1;
			}

			hex2mem(&remcomInBuffer[1], (char *) &regnum, sizeof(regnum), 0);
			if (regnum >= NUMREGS) {
				remcomOutBuffer[0] = 'E';
				remcomOutBuffer[1] = 0;
				break;
			}
			mem2hex((char *) &gdb_regs[REGISTER_INDEX(regnum)], remcomOutBuffer,
				REGISTER_SIZE(regnum), 0);
			remcomOutBuffer[REGISTER_SIZE(regnum) * 2] = 0;
			break;
		}
		case 'g':	/* return the value of the CPU registers */
			get_gdb_regs(usethread, info);
			if (rpacket_size)
				rpacket_process(gdb_regs, remcomOutBuffer, 1);
			else
				mem2hex((char *) gdb_regs,
					remcomOutBuffer, NUMREGBYTES, 0);
			break;
		case 'G':	/* set the value of the CPU registers - return OK */
			if (rpacket_size)
				rpacket_process(gdb_regs, &remcomInBuffer[1], 0);
			else
			  hex2mem(&remcomInBuffer[1],
				(char *) gdb_regs, NUMREGBYTES, 0);
			if (!usethread || usethread == current) {
				gdb_regs_to_regs(info);
				strcpy(remcomOutBuffer, "OK");
			} else {
				strcpy(remcomOutBuffer, "E00");
			}
			break;

		case 'P':{	/* set the value of a single CPU register -
				   return OK */
				/*
				 * For some reason, gdb wants to talk about psudo
				 * registers (greater than 15).	 These may have
				 * meaning for ptrace, but for us it is safe to
				 * ignor them.	We do this by dumping them into
				 * _GS which we also ignor, but do have memory for.
				 */
				int regno;

				ptr = &remcomInBuffer[1];
				regs_to_gdb_regs(info);
				if ((!usethread || usethread == current) &&
				    hexToInt(&ptr, &regno) &&
				    *ptr++ == '=' && (regno >= 0)) {
					regno =
					    (regno >= NUMREGS ? 0 : regno);
					hex2mem(ptr, (char *) &gdb_regs[REGISTER_INDEX(regno)],
						4, 0);
					gdb_regs_to_regs(info);
					strcpy(remcomOutBuffer, "OK");
					break;
				}
				strcpy(remcomOutBuffer, "E01");
				break;
			}

			/* mAA..AA,LLLL	 Read LLLL bytes at address AA..AA */
		case 'm':
			/* TRY TO READ %x,%x.  IF SUCCEED, SET PTR = 0 */
			ptr = &remcomInBuffer[1];
			if (hexToLong(&ptr, &addr) &&
			    (*(ptr++) == ',') && (hexToInt(&ptr, &length))) {
				ptr = 0;
				/*
				 * hex doubles the byte count
				 */
				if (length > (BUFMAX / 2))
					length = BUFMAX / 2;
				mem2hex((char *) addr,
					remcomOutBuffer, length, 1);
				if (mem_err) {
					strcpy(remcomOutBuffer, "E03");
					debug_error("memory fault\n", NULL);
				}
			}

			if (ptr) {
				strcpy(remcomOutBuffer, "E01");
				debug_error
				    ("malformed read memory command: %s\n",
				     remcomInBuffer);
			}
			break;

			/* MAA..AA,LLLL:
			   Write LLLL bytes at address AA.AA return OK */
		case 'M':
			/* TRY TO READ '%x,%x:'.  IF SUCCEED, SET PTR = 0 */
			ptr = &remcomInBuffer[1];
			if (hexToLong(&ptr, &addr) &&
			    (*(ptr++) == ',') &&
			    (hexToInt(&ptr, &length)) && (*(ptr++) == ':')) {
				extern unsigned long _start[];

				if (addr == (unsigned long) _start)
					strcpy(remcomOutBuffer, "OK");

				else {
					hex2mem(ptr, (char *) addr, length, 1);

					if (mem_err) {
						strcpy(remcomOutBuffer, "E03");
						debug_error("memory fault\n", NULL);
					} else {
						if (kernel_text_address(addr))
							flush_icache_range(addr, addr + length);
						strcpy(remcomOutBuffer, "OK");
					}
				}

				ptr = 0;
			}
			if (ptr) {
				strcpy(remcomOutBuffer, "E02");
				debug_error
				    ("malformed write memory command: %s\n",
				     remcomInBuffer);
			}
			break;
		case 'S':
			remcomInBuffer[0] = 's';
		case 'C':
			/* Csig;AA..AA where ;AA..AA is optional
			 * continue with signal
			 * Since signals are meaning less to us, delete that
			 * part and then fall into the 'c' code.
			 */
			ptr = &remcomInBuffer[1];
			length = 2;
			while (*ptr && *ptr != ';') {
				length++;
				ptr++;
			}
			if (*ptr) {
				do {
					ptr++;
					*(ptr - length++) = *ptr;
				} while (*ptr);
			} else {
				remcomInBuffer[1] = 0;
			}

			/* cAA..AA  Continue at address AA..AA(optional) */
			/* sAA..AA  Step one instruction from AA..AA(optional) */
			/* D	    detach, reply OK and then continue */
		case 'c':
		case 's':
		case 'D':

			/* try to read optional parameter,
			   pc unchanged if no parm */
			ptr = &remcomInBuffer[1];
			if (hexToLong(&ptr, &addr)) {
				if (remote_debug)
					printk("Changing EIP to 0x%lx\n", addr);

				ptregs.cr_iip = addr;
			}

			newPC = ptregs.cr_iip;

			/* clear the trace bit */
			ptregs.cr_ipsr &= ~IA64_PSR_SS;

			/* set the trace bit if we're stepping */
			if (remcomInBuffer[0] == 's')
				ptregs.cr_ipsr |= IA64_PSR_SS;

			/* detach is a friendly version of continue. Note that
			   debugging is still enabled (e.g hit control C)
			 */
			if (remcomInBuffer[0] == 'D') {
				strcpy(remcomOutBuffer, "OK");
				putpacket(remcomOutBuffer);
			}

			if (remote_debug) {
				printk("Resuming execution\n");
				print_regs(&ptregs);
			}

			if (kgdboe)
				netpoll_set_trap(0);

			correct_hw_break();
			ptregs.cr_ipsr |= IA64_PSR_DB;
			goto exit_kgdb;

			/* kill the program */
		case 'k':	/* do nothing */
			break;

			/* query */
		case 'q':
			nothreads = 0;
			switch (remcomInBuffer[1]) {
			case 'P':
				/* qPart:gpacket:read::<offset>,<size> */
				if (strncmp(&remcomInBuffer[1], "Part", 4) ||
					remcomInBuffer[5] != ':' ||
					strncmp(&remcomInBuffer[6], "gpacket", 7) ||
					remcomInBuffer[13] != ':' ||
					strncmp(&remcomInBuffer[14], "read", 4) ||
					remcomInBuffer[18] != ':' ||
					remcomInBuffer[19] != ':')
					break;
				process_gpacket_request();
				break;
			case 'f':
				threadid = 1;
				thread_list = 2;
				thread_list_start = (usethread ? : current);
			case 's':
				if (!cmp_str(&remcomInBuffer[2],
					     "ThreadInfo", 10))
					break;

				remcomOutBuffer[nothreads++] = 'm';
				for (; threadid < PID_MAX + MAX_NO_CPUS;
				     threadid++) {
					thread = getthread(threadid);
					if (thread) {
						nothreads += int_to_hex_v(
							&remcomOutBuffer[
								nothreads],
							threadid);
						if (thread_min > threadid)
							thread_min = threadid;
						remcomOutBuffer[
							nothreads] = ',';
						nothreads++;
						if (nothreads > BUFMAX - 10)
							break;
					}
				}
				if (remcomOutBuffer[nothreads - 1] == 'm') {
					remcomOutBuffer[nothreads - 1] = 'l';
				} else {
					nothreads--;
				}
				remcomOutBuffer[nothreads] = 0;
				break;

#ifdef old_thread_list /* Old thread info request */
			case 'L':
				/* List threads */
				thread_list = 2;
				thread_list_start = (usethread ? : current);
				unpack_byte(remcomInBuffer + 3, &maxthreads);
				unpack_threadid(remcomInBuffer + 5, &thref);
				do {
					int buf_thread_limit =
					    (BUFMAX - 22) / BUF_THREAD_ID_SIZE;
					if (maxthreads > buf_thread_limit) {
						maxthreads = buf_thread_limit;
					}
				} while (0);
				remcomOutBuffer[0] = 'q';
				remcomOutBuffer[1] = 'M';
				remcomOutBuffer[4] = '0';
				pack_threadid(remcomOutBuffer + 5, &thref);

				threadid = threadref_to_int(&thref);
				for (nothreads = 0;
				     nothreads < maxthreads &&
				     threadid < PID_MAX + MAX_NO_CPUS;
				     threadid++) {
					thread = getthread(threadid);
					if (thread) {
						int_to_threadref(&thref,
								 threadid);
						pack_threadid(remcomOutBuffer +
							      21 +
							      nothreads * 16,
							      &thref);
						nothreads++;
						if (thread_min > threadid)
							thread_min = threadid;
					}
				}

				if (threadid == PID_MAX + MAX_NO_CPUS) {
					remcomOutBuffer[4] = '1';
				}
				pack_hex_byte(remcomOutBuffer + 2, nothreads);
				remcomOutBuffer[21 + nothreads * 16] = '\0';
				break;
#endif
			case 'C':
				/* Current thread id */
				remcomOutBuffer[0] = 'Q';
				remcomOutBuffer[1] = 'C';
				threadid = current->pid;
				if (!threadid) {
					/*
					 * idle thread
					 */
					for (threadid = PID_MAX;
					     threadid < PID_MAX + MAX_NO_CPUS;
					     threadid++) {
						if (current ==
						    idle_task(threadid -
							      PID_MAX))
							break;
					}
				}
				int_to_threadref(&thref, threadid);
				pack_threadid(remcomOutBuffer + 2, &thref);
				remcomOutBuffer[18] = '\0';
				break;

			case 'E':
				/* Print exception info */
				printexceptioninfo(exceptionVector,
						   err_code, remcomOutBuffer);
				break;
			case 'T':{
				char * nptr;
				/* Thread extra info */
				if (!cmp_str(&remcomInBuffer[2],
					    "hreadExtraInfo,", 15)) {
					break;
				}
				ptr = &remcomInBuffer[17];
				hexToInt(&ptr, &threadid);
				thread = getthread(threadid);
				nptr = &thread->comm[0];
				length = 0;
				ptr = &remcomOutBuffer[0];
				do {
					length++;
					ptr = pack_hex_byte(ptr, *nptr++);
				 } while (*nptr && length < 16);
				/*
				 * would like that 16 to be the size of
				 * task_struct.comm but don't know the
				 * syntax..
				 */
				*ptr = 0;
			}
			}
			break;

			/* task related */
		case 'H':
			switch (remcomInBuffer[1]) {
			case 'g':
				ptr = &remcomInBuffer[2];
				hexToInt(&ptr, &threadid);
				thread = getthread(threadid);
				if (!thread) {
					remcomOutBuffer[0] = 'E';
					remcomOutBuffer[1] = '\0';
					break;
				}
				/*
				 * Just in case I forget what this is all about,
				 * the "thread info" command to gdb causes it
				 * to ask for a thread list.  It then switches
				 * to each thread and asks for the registers.
				 * For this (and only this) usage, we want to
				 * fudge the registers of tasks not on the run
				 * list (i.e. waiting) to show the routine that
				 * called schedule. Also, gdb, is a minimalist
				 * in that if the current thread is the last
				 * it will not re-read the info when done.
				 * This means that in this case we must show
				 * the real registers. So here is how we do it:
				 * Each entry we keep track of the min
				 * thread in the list (the last that gdb will)
				 * get info for.  We also keep track of the
				 * starting thread.
				 * "thread_list" is cleared when switching back
				 * to the min thread if it is was current, or
				 * if it was not current, thread_list is set
				 * to 1.  When the switch to current comes,
				 * if thread_list is 1, clear it, else do
				 * nothing.
				 */
				usethread = thread;
				if ((thread_list == 1) &&
				    (thread == thread_list_start)) {
					thread_list = 0;
				}
				if (thread_list && (threadid == thread_min)) {
					if (thread == thread_list_start) {
						thread_list = 0;
					} else {
						thread_list = 1;
					}
				}
				/* follow through */
			case 'c':
				remcomOutBuffer[0] = 'O';
				remcomOutBuffer[1] = 'K';
				remcomOutBuffer[2] = '\0';
				break;
			}
			break;

			/* Query thread status */
		case 'T':
			ptr = &remcomInBuffer[1];
			hexToInt(&ptr, &threadid);
			thread = getthread(threadid);
			if (thread) {
				remcomOutBuffer[0] = 'O';
				remcomOutBuffer[1] = 'K';
				remcomOutBuffer[2] = '\0';
				if (thread_min > threadid)
					thread_min = threadid;
			} else {
				remcomOutBuffer[0] = 'E';
				remcomOutBuffer[1] = '\0';
			}
			break;

#ifdef	oldbreak_protocol
		case 'Y': /* set up a hardware breakpoint */
			ptr = &remcomInBuffer[1];
			hexToInt(&ptr, &breakno);
			ptr++;
			hexToInt(&ptr, &breaktype);
			ptr++;
			hexToInt(&ptr, &length);
			ptr++;
			hexToLong(&ptr, &addr);
			if (set_hw_break(breakno & 0x3,
					 breaktype & 0x3,
					 length & 0x3, addr) == 0) {
				strcpy(remcomOutBuffer, "OK");
			} else {
				strcpy(remcomOutBuffer, "ERROR");
			}
			break;

			/* Remove hardware breakpoint */
		case 'y':
			ptr = &remcomInBuffer[1];
			hexToInt(&ptr, &breakno);
			if (remove_hw_break(breakno & 0x3) == 0) {
				strcpy(remcomOutBuffer, "OK");
			} else {
				strcpy(remcomOutBuffer, "ERROR");
			}
			break;
#endif
		/*
		 * Set/Remove watchpoint
		 *
		 */
		case 'Z':
		case 'z':
			if (!kgdb_dbregs_enabled)
				break;

			switch (remcomInBuffer[1]) {
			case '0':	/* insert hwd break */
			case '1':	/* hardware breakpoint */
			case '2':	/* write watchpoint */
			case '3':	/* read watchpoint */
			case '4':	/* access watchpoint */
				if (remcomInBuffer[2] != ',') {
					strcpy(remcomOutBuffer, "ERROR");
					break;
				}
				ptr = &remcomInBuffer[3];
				if (hexToLong(&ptr, &addr)) {
					ptr++;
					if (!hexToInt(&ptr, &length)) {
						strcpy(remcomOutBuffer, "ERROR");
						break;
					}
				}
				else {
					strcpy(remcomOutBuffer, "ERROR");
					break;
				}

				if (hardware_breakpoint(addr, length,
					(remcomInBuffer[1] != '0' ?
					remcomInBuffer[1] - '1' : WATCH_INSTRUCTION),
					remcomInBuffer[0] == 'Z'))
					strcpy(remcomOutBuffer, "OK");
				else
					strcpy(remcomOutBuffer, "ERROR");
				break;
			default:
				strcpy(remcomOutBuffer, "ERROR");
				break;
			}
			break;
		case 'R':	/* reboot */
			strcpy(remcomOutBuffer, "OK");
			putpacket(remcomOutBuffer);
			/*to_gdb("Rebooting\n"); */
			machine_restart(NULL);

		}		/* switch */

		/* reply to the request */
		putpacket(remcomOutBuffer);
	}			/* while(1==1) */
	/*
	 *  reached by goto only.
	 */
      exit_kgdb:
#ifdef CONFIG_SMP
	/*
	 * Release gdb wait locks
	 * Sanity check time.  Must have at least one cpu to run.  Also single
	 * step must not be done if the current cpu is on hold.
	 */
	if (spinlock_count == 1) {
		int ss_hold = (ptregs.cr_ipsr & IA64_PSR_SS) && kgdb_info.hold_on_sstep;
		int cpu_avail = 0;
		int i;

		for (i = 0; i < MAX_NO_CPUS; i++) {
			if (!cpu_online(i))
				break;
			if (!hold_cpu(i)) {
				cpu_avail = 1;
			}
		}
		/*
		 * Early in the bring up there will be NO cpus on line...
		 */
		if (!cpu_avail && !cpus_empty(cpu_online_map)) {
			to_gdb("No cpus unblocked, see 'kgdb_info.hold_cpu'\n");
			goto once_again;
		}
		if (hold_cpu(smp_processor_id()) && (ptregs.cr_ipsr & IA64_PSR_SS)) {
			to_gdb
			    ("Current cpu must be unblocked to single step\n");
			goto once_again;
		}

		if (!ss_hold) {
			int i;
			for (i = 0; i < MAX_NO_CPUS; i++) {
				if (!hold_cpu(i)) {
					spin_unlock(&waitlocks[i]);
				}
			}
		} else {
			spin_unlock(&waitlocks[smp_processor_id()]);
		}
		/* Release kgdb spinlock */
		KGDB_SPIN_UNLOCK(&kgdb_spinlock);
		/*
		 * If this cpu is on hold, this is where we
		 * do it.  Note, the NMI will pull us out of here,
		 * but will return as the above lock is not held.
		 * We will stay here till another cpu releases the lock for us.
		 */
		spin_unlock_wait(waitlocks + smp_processor_id());
		kgdb_local_irq_restore(flags);
		info->ret = 0;
		return;
	}
#if 0
exit_just_unlock:
#endif
#endif
	/* Release kgdb spinlock */
	KGDB_SPIN_UNLOCK(&kgdb_spinlock);
	kgdb_local_irq_restore(flags);
	info->ret = 0;
	return;
}

/* this function is used to set up exception handlers for tracing and
 * breakpoints.
 * This function is not needed as the above line does all that is needed.
 * We leave it for backward compatitability...
 */
void
set_debug_traps(void)
{
	/*
	 * linux_debug_hook is defined in traps.c.  We store a pointer
	 * to our own exception handler into it.

	 * But really folks, every hear of labeled common, an old Fortran
	 * concept.  Lots of folks can reference it and it is define if
	 * anyone does.	 Only one can initialize it at link time.  We do
	 * this with the hook.	See the statement above.  No need for any
	 * executable code and it is ready as soon as the kernel is
	 * loaded.  Very desirable in kernel debugging.

	 linux_debug_hook = handle_exception ;
	 */

	/* In case GDB is started before us, ack any packets (presumably
	   "$?#xx") sitting there.
	   putDebugChar ('+');

	   initialized = 1;
	 */
}

/* This function will generate a breakpoint exception.	It is used at the
   beginning of a program to sync up with a debugger and can be used
   otherwise as a quick means to stop program execution and "break" into
   the debugger. */
/* But really, just use the BREAKPOINT macro.  We will handle the int stuff
 */

#ifdef later
/*
 * possibly we should not go thru the traps.c code at all?  Someday.
 */
void
do_kgdb_int3(struct pt_regs *regs, long error_code)
{
	kgdb_handle_exception(3, 5, error_code, regs);
	return;
}
#endif
#undef regs
#ifdef CONFIG_TRAP_BAD_SYSCALL_EXITS
asmlinkage void
bad_sys_call_exit(int stuff)
{
	struct pt_regs *regs = (struct pt_regs *) &stuff;
	printk("Sys call %d return with %x preempt_count\n",
	       (int) regs->orig_eax, preempt_count());
}
#endif
#ifdef CONFIG_STACK_OVERFLOW_TEST
#include <asm/kgdb.h>
asmlinkage void
stack_overflow(void)
{
#ifdef BREAKPOINT
	BREAKPOINT;
#else
	printk("Kernel stack overflow, looping forever\n");
#endif
	while (1) {
	}
}
#endif

#if defined(CONFIG_SMP) || defined(CONFIG_KGDB_CONSOLE)
char gdbconbuf[BUFMAX];

static void
kgdb_gdb_message(const char *s, unsigned count)
{
	int i;
	int wcount;
	char *bufptr;
	/*
	 * This takes care of NMI while spining out chars to gdb
	 */
	IF_SMP(in_kgdb_console = 1);
	gdbconbuf[0] = 'O';
	bufptr = gdbconbuf + 1;
	while (count > 0) {
		if ((count << 1) > (BUFMAX - 2)) {
			wcount = (BUFMAX - 2) >> 1;
		} else {
			wcount = count;
		}
		count -= wcount;
		for (i = 0; i < wcount; i++) {
			bufptr = pack_hex_byte(bufptr, s[i]);
		}
		*bufptr = '\0';
		s += wcount;

		putpacket(gdbconbuf);

	}
	IF_SMP(in_kgdb_console = 0);
}
#endif
#ifdef CONFIG_SMP
static void
to_gdb(const char *s)
{
	int count = 0;
	while (s[count] && (count++ < BUFMAX)) ;
	kgdb_gdb_message(s, count);
}
#endif
#ifdef CONFIG_KGDB_CONSOLE
#include <linux/console.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>

void
kgdb_console_write(struct console *co, const char *s, unsigned count)
{

	if (gdb_ia64vector == -1) {
		/*
		 * We have not yet talked to gdb.  What to do...
		 * lets break, on continue we can do the write.
		 * But first tell him whats up. Uh, well no can do,
		 * as this IS the console.  Oh well...
		 * We do need to wait or the messages will be lost.
		 * Other option would be to tell the above code to
		 * ignore this breakpoint and do an auto return,
		 * but that might confuse gdb.	Also this happens
		 * early enough in boot up that we don't have the traps
		 * set up yet, so...
		 */
		breakpoint();
	}
	kgdb_gdb_message(s, count);
}

/*
 * ------------------------------------------------------------
 * Serial KGDB driver
 * ------------------------------------------------------------
 */

static struct console kgdbcons = {
	name:"kgdb",
	write:kgdb_console_write,
#ifdef CONFIG_KGDB_USER_CONSOLE
	device:kgdb_console_device,
#endif
	flags:CON_PRINTBUFFER | CON_ENABLED,
	index:-1,
};

/*
 * The trick here is that this file gets linked before printk.o
 * That means we get to peer at the console info in the command
 * line before it does.	 If we are up, we register, otherwise,
 * do nothing.	By returning 0, we allow printk to look also.
 */
static int kgdb_console_enabled;

int __init
kgdb_console_init(char *str)
{
	if ((strncmp(str, "kgdb", 4) == 0) || (strncmp(str, "gdb", 3) == 0)) {
		register_console(&kgdbcons);
		kgdb_console_enabled = 1;
	}
	return 0;		/* let others look at the string */
}

__setup("console=", kgdb_console_init);

int __init
kgdb_enable_dbregs(char *str)
{
	kgdb_dbregs_enabled = 1;
	return 1;
}

__setup("kgdb_debug_regs=", kgdb_enable_dbregs);

#ifdef CONFIG_KGDB_USER_CONSOLE
static kdev_t kgdb_console_device(struct console *c);
/* This stuff sort of works, but it knocks out telnet devices
 * we are leaving it here in case we (or you) find time to figure it out
 * better..
 */

/*
 * We need a real char device as well for when the console is opened for user
 * space activities.
 */

static int
kgdb_consdev_open(struct inode *inode, struct file *file)
{
	return 0;
}

static ssize_t
kgdb_consdev_write(struct file *file, const char *buf,
		   size_t count, loff_t * ppos)
{
	int size, ret = 0;
	static char kbuf[128];
	static DECLARE_MUTEX(sem);

	/* We are not reentrant... */
	if (down_interruptible(&sem))
		return -ERESTARTSYS;

	while (count > 0) {
		/* need to copy the data from user space */
		size = count;
		if (size > sizeof (kbuf))
			size = sizeof (kbuf);
		if (copy_from_user(kbuf, buf, size)) {
			ret = -EFAULT;
			break;;
		}
		kgdb_console_write(&kgdbcons, kbuf, size);
		count -= size;
		ret += size;
		buf += size;
	}

	up(&sem);

	return ret;
}

struct file_operations kgdb_consdev_fops = {
	open:kgdb_consdev_open,
	write:kgdb_consdev_write
};
static kdev_t
kgdb_console_device(struct console *c)
{
	return MKDEV(TTYAUX_MAJOR, 1);
}

/*
 * This routine gets called from the serial stub in the i386/lib
 * This is so it is done late in bring up (just before the console open).
 */
void
kgdb_console_finit(void)
{
	if (kgdb_console_enabled) {
		char *cptr = cdevname(MKDEV(TTYAUX_MAJOR, 1));
		char *cp = cptr;
		while (*cptr && *cptr != '(')
			cptr++;
		*cptr = 0;
		unregister_chrdev(TTYAUX_MAJOR, cp);
		register_chrdev(TTYAUX_MAJOR, "kgdb", &kgdb_consdev_fops);
	}
}
#endif
#endif
#ifdef CONFIG_KGDB_TS
#include <asm/msr.h>		/* time stamp code */
#include <asm/hardirq.h>	/* in_interrupt */
#ifdef CONFIG_KGDB_TS_64
#define DATA_POINTS 64
#endif
#ifdef CONFIG_KGDB_TS_128
#define DATA_POINTS 128
#endif
#ifdef CONFIG_KGDB_TS_256
#define DATA_POINTS 256
#endif
#ifdef CONFIG_KGDB_TS_512
#define DATA_POINTS 512
#endif
#ifdef CONFIG_KGDB_TS_1024
#define DATA_POINTS 1024
#endif
#ifndef DATA_POINTS
#define DATA_POINTS 128		/* must be a power of two */
#endif
#define INDEX_MASK (DATA_POINTS - 1)
#if (INDEX_MASK & DATA_POINTS)
#error "CONFIG_KGDB_TS_COUNT must be a power of 2"
#endif
struct kgdb_and_then_struct {
#ifdef CONFIG_SMP
	int on_cpu;
#endif
	struct task_struct *task;
	long long at_time;
	int from_ln;
	char *in_src;
	void *from;
	int *with_shpf;
	int data0;
	int data1;
};
struct kgdb_and_then_struct2 {
#ifdef CONFIG_SMP
	int on_cpu;
#endif
	struct task_struct *task;
	long long at_time;
	int from_ln;
	char *in_src;
	void *from;
	int *with_shpf;
	struct task_struct *t1;
	struct task_struct *t2;
};
struct kgdb_and_then_struct kgdb_data[DATA_POINTS];

struct kgdb_and_then_struct *kgdb_and_then = &kgdb_data[0];
int kgdb_and_then_count;

void
kgdb_tstamp(int line, char *source, int data0, int data1)
{
	static spinlock_t ts_spin = SPIN_LOCK_UNLOCKED;
	int flags;
	kgdb_local_irq_save(flags);
	spin_lock(&ts_spin);
	kgdb_and_then->at_time = ia64_get_itc();
#ifdef CONFIG_SMP
	kgdb_and_then->on_cpu = smp_processor_id();
#endif
	kgdb_and_then->task = current;
	kgdb_and_then->from_ln = line;
	kgdb_and_then->in_src = source;
	kgdb_and_then->from = __builtin_return_address(0);
	kgdb_and_then->with_shpf = (int *) (((flags & IF_BIT) >> 9) |
					    (preempt_count() << 8));
	kgdb_and_then->data0 = data0;
	kgdb_and_then->data1 = data1;
	kgdb_and_then = &kgdb_data[++kgdb_and_then_count & INDEX_MASK];
	spin_unlock(&ts_spin);
	kgdb_local_irq_restore(flags);
#ifdef CONFIG_PREEMPT

#endif
	return;
}
#endif
typedef int gdb_debug_hook(int exceptionVector,
			   int signo, int err_code, struct pt_regs *linux_regs);
gdb_debug_hook *linux_debug_hook = &kgdb_handle_exception;	/* histerical reasons... */

static int kgdb_need_breakpoint[NR_CPUS];

void kgdb_schedule_breakpoint(void)
{
	kgdb_need_breakpoint[smp_processor_id()] = 1;
}

void kgdb_process_breakpoint(void)
{
	/*
	 * Handle a breakpoint queued from inside network driver code
         * to avoid reentrancy issues
	 */
	if (kgdb_need_breakpoint[smp_processor_id()]) {
		kgdb_need_breakpoint[smp_processor_id()] = 0;
		BREAKPOINT;
	}
}

