/*
 *  acpi_wacom.c - Wacom Digitizer ACPI Driver ($Revision: 1.1 $)
 *
 *  Copyright (C) 2004 Jamey Hicks <jamey.hicks@hp.com>
 *  Copyright (C) 2001, 2002 Andy Grover <andrew.grover@intel.com>
 *  Copyright (C) 2001, 2002 Paul Diefenbaugh <paul.s.diefenbaugh@intel.com>
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <acpi/acpi.h>
#include <acpi/acpi_bus.h>
#include <acpi/acpi_drivers.h>

#define ACPI_WACOM_COMPONENT		0x80000000
#define ACPI_WACOM_CLASS			"wacom"
#define ACPI_WACOM_HID 			"WACF005"
#define ACPI_WACOM_DRIVER_NAME		"ACPI Wacom Driver"
#define ACPI_WACOM_DEVICE_NAME		"Wacom Digitizer"
#define ACPI_WACOM_FILE_STATE		"state"
#define ACPI_WACOM_NOTIFY_STATUS		0x80
#define ACPI_WACOM_STATUS_OFFLINE		0x0D
#define ACPI_WACOM_STATUS_ONLINE		0x0F
#define ACPI_WACOM_STATUS_UNKNOWN		0xFF

#define _COMPONENT		ACPI_WACOM_COMPONENT
ACPI_MODULE_NAME		("acpi_wacom")

MODULE_AUTHOR("Jamey Hicks");
MODULE_DESCRIPTION(ACPI_WACOM_DRIVER_NAME);
MODULE_LICENSE("GPL");

int acpi_wacom_add (struct acpi_device *device);
int acpi_wacom_remove (struct acpi_device *device, int type);
static int acpi_wacom_start (struct acpi_device *device);
static int acpi_wacom_stop (struct acpi_device *device, int type);
static int acpi_wacom_open_fs(struct inode *inode, struct file *file);

static struct acpi_driver acpi_wacom_driver = {
	.name =		ACPI_WACOM_DRIVER_NAME,
	.class =	ACPI_WACOM_CLASS,
	.ids =		ACPI_WACOM_HID,
	.ops =		{
				.add =		acpi_wacom_add,
				.remove =	acpi_wacom_remove,
				.start =	acpi_wacom_start,
				.stop =		acpi_wacom_stop,
			},
};

struct acpi_wacom {
	acpi_handle		handle;
	unsigned long		state;
};

static struct file_operations acpi_wacom_fops = {
	.open		= acpi_wacom_open_fs,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= single_release,
};

/* --------------------------------------------------------------------------
                               Wacom Management
   -------------------------------------------------------------------------- */

static int
acpi_wacom_get_state (
	struct acpi_wacom		*ac)
{
	acpi_status		status = AE_OK;

	ACPI_FUNCTION_TRACE("acpi_wacom_get_state");
	printk("acpi_wacom_get_state ac=%p\n", ac);

	if (!ac)
		return_VALUE(-EINVAL);

	status = acpi_evaluate_integer(ac->handle, "_STA", NULL, &ac->state);
	if (ACPI_FAILURE(status)) {
		ACPI_DEBUG_PRINT((ACPI_DB_ERROR,
			"Error reading Wacom state\n"));
		ac->state = ACPI_WACOM_STATUS_UNKNOWN;
		return_VALUE(-ENODEV);
	}
	
	return_VALUE(0);
}


/* --------------------------------------------------------------------------
                              FS Interface (/proc)
   -------------------------------------------------------------------------- */

struct proc_dir_entry		*acpi_wacom_dir;

int acpi_wacom_seq_show(struct seq_file *seq, void *offset)
{
	struct acpi_wacom		*ac = (struct acpi_wacom *) seq->private;

	ACPI_FUNCTION_TRACE("acpi_wacom_seq_show");

	if (!ac)
		return 0;

	if (acpi_wacom_get_state(ac)) {
		seq_puts(seq, "ERROR: Unable to read Wacom state\n");
		return 0;
	}

	seq_puts(seq, "state:                   ");
	switch (ac->state) {
	case ACPI_WACOM_STATUS_OFFLINE:
		seq_puts(seq, "off-line\n");
		break;
	case ACPI_WACOM_STATUS_ONLINE:
		seq_puts(seq, "on-line\n");
		break;
	default:
		seq_puts(seq, "unknown\n");
		break;
	}

	return 0;
}
	
static int acpi_wacom_open_fs(struct inode *inode, struct file *file)
{
	return single_open(file, acpi_wacom_seq_show, PDE(inode)->data);
}

static int
acpi_wacom_add_fs (
	struct acpi_device	*device)
{
	struct proc_dir_entry	*entry = NULL;

	ACPI_FUNCTION_TRACE("acpi_wacom_add_fs");

	if (!acpi_device_dir(device)) {
		acpi_device_dir(device) = proc_mkdir(acpi_device_bid(device),
						     acpi_wacom_dir);
		if (!acpi_device_dir(device))
			return_VALUE(-ENODEV);
	}

	/* 'state' [R] */
	entry = create_proc_entry(ACPI_WACOM_FILE_STATE,
				  S_IRUGO, acpi_device_dir(device));
	if (!entry)
		ACPI_DEBUG_PRINT((ACPI_DB_ERROR,
			"Unable to create '%s' fs entry\n",
			ACPI_WACOM_FILE_STATE));
	else {
		entry->proc_fops = &acpi_wacom_fops;
		entry->data = acpi_driver_data(device);
	}

	return_VALUE(0);
}


static int
acpi_wacom_remove_fs (
	struct acpi_device	*device)
{
	ACPI_FUNCTION_TRACE("acpi_wacom_remove_fs");

	if (acpi_device_dir(device)) {
		remove_proc_entry(ACPI_WACOM_FILE_STATE, acpi_device_dir(device));
		remove_proc_entry(acpi_device_bid(device), acpi_wacom_dir);
		acpi_device_dir(device) = NULL;
	}

	return_VALUE(0);
}


/* --------------------------------------------------------------------------
                                   Driver Model
   -------------------------------------------------------------------------- */

void
acpi_wacom_notify (
	acpi_handle		handle,
	u32			event,
	void			*data)
{
	struct acpi_wacom		*ac = (struct acpi_wacom *) data;
	struct acpi_device	*device = NULL;

	ACPI_FUNCTION_TRACE("acpi_wacom_notify");

	if (!ac)
		return;

	if (acpi_bus_get_device(ac->handle, &device))
		return_VOID;

	switch (event) {
	case ACPI_WACOM_NOTIFY_STATUS:
		acpi_wacom_get_state(ac);
		acpi_bus_generate_event(device, event, (u32) ac->state);
		break;
	default:
		ACPI_DEBUG_PRINT((ACPI_DB_INFO,
			"Unsupported event [0x%x]\n", event));
		break;
	}

	return_VOID;
}


int
acpi_wacom_add (
	struct acpi_device	*device)
{
	int			result = 0;
	acpi_status		status = AE_OK;
	struct acpi_wacom		*wacom = NULL;

	ACPI_FUNCTION_TRACE("acpi_wacom_add");
	printk("acpi_wacom_add device=%p\n", device);

	if (!device)
		return_VALUE(-EINVAL);

	wacom = kmalloc(sizeof(struct acpi_wacom), GFP_KERNEL);
	if (!wacom)
		return_VALUE(-ENOMEM);
	memset(wacom, 0, sizeof(struct acpi_wacom));

	wacom->handle = device->handle;
	strcpy(acpi_device_name(device), ACPI_WACOM_DEVICE_NAME);
	strcpy(acpi_device_class(device), ACPI_WACOM_CLASS);
	acpi_driver_data(device) = wacom;

	result = acpi_wacom_get_state(wacom);
	if (result)
		goto end;

	result = acpi_wacom_add_fs(device);
	if (result)
		goto end;

#if 0
	status = acpi_install_notify_handler(ac->handle,
		ACPI_DEVICE_NOTIFY, acpi_wacom_notify, ac);
	if (ACPI_FAILURE(status)) {
		ACPI_DEBUG_PRINT((ACPI_DB_ERROR,
			"Error installing notify handler\n"));
		result = -ENODEV;
		goto end;
	}
#endif

	printk(KERN_INFO PREFIX "%s [%s] (%s)\n", 
		acpi_device_name(device), acpi_device_bid(device), 
		wacom->state?"on-line":"off-line");

end:
	if (result) {
		acpi_wacom_remove_fs(device);
		kfree(wacom);
	}

	return_VALUE(result);
}


int
acpi_wacom_remove (
	struct acpi_device	*device,
	int			type)
{
	acpi_status		status = AE_OK;
	struct acpi_wacom		*ac = NULL;

	ACPI_FUNCTION_TRACE("acpi_wacom_remove");

	if (!device || !acpi_driver_data(device))
		return_VALUE(-EINVAL);

	ac = (struct acpi_wacom *) acpi_driver_data(device);

#if 0
	status = acpi_remove_notify_handler(ac->handle,
		ACPI_DEVICE_NOTIFY, acpi_wacom_notify);
	if (ACPI_FAILURE(status))
		ACPI_DEBUG_PRINT((ACPI_DB_ERROR,
			"Error removing notify handler\n"));
#endif

	acpi_wacom_remove_fs(device);

	kfree(ac);

	return_VALUE(0);
}

static acpi_status
acpi_wacom_io_ports (
	struct acpi_resource	*resource,
	void			*context)
{
	int i;
	ACPI_FUNCTION_TRACE("acpi_wacom_io_ports");
	printk("acpi_wacom_io_ports: type=%#x length=%d\n", resource->id, resource->length);
	switch (resource->id) {
	case ACPI_RSTYPE_IRQ: {
		struct acpi_resource_irq *ari = &resource->data.irq;
		printk("  edge_level=%x active_high=%x shared=%x num=%x int=%x\n",
		       ari->edge_level, ari->active_high_low, ari->shared_exclusive, ari->number_of_interrupts, ari->interrupts[0]);
		} break;
	case ACPI_RSTYPE_IO: {
		struct acpi_resource_io *ari = &resource->data.io;
		printk("  io_decode=%x min_base=%x max_base=%x alignment=%x range_length=%x\n",
		       ari->io_decode, ari->min_base_address, ari->max_base_address, ari->alignment, ari->range_length);
		} break;
	}
	return_VALUE(0);
}

static int
acpi_wacom_start (
	struct acpi_device	*device)
{
	acpi_status		status = AE_OK;
	struct acpi_wacom *wacom = NULL;
	struct acpi_buffer acpi_buffer;

	ACPI_FUNCTION_TRACE("acpi_wacom_start");

	if (!device)
		return_VALUE(-EINVAL);

	wacom = acpi_driver_data(device);

	if (!wacom)
		return_VALUE(-EINVAL);

	/*
	 * Get I/O port addresses. Convert to GAS format.
	 */
	printk("acpi_wacom: walking resources\n");
	status = acpi_walk_resources(device->handle, METHOD_NAME__CRS,
				     acpi_wacom_io_ports, wacom);
	if (ACPI_FAILURE(status)) {
		ACPI_DEBUG_PRINT((ACPI_DB_ERROR, "Error getting I/O port addresses"));
		return_VALUE(-ENODEV);
	}

#if 0
	/*
	 * Install GPE handler
	 */
	status = acpi_install_gpe_handler(NULL, ec->gpe_bit,
		ACPI_GPE_EDGE_TRIGGERED, &acpi_wacom_gpe_handler, ec);
	if (ACPI_FAILURE(status)) {
		return_VALUE(-ENODEV);
	}

	status = acpi_install_address_space_handler (ec->handle,
			ACPI_ADR_SPACE_EC, &acpi_wacom_space_handler,
			&acpi_wacom_space_setup, ec);
	if (ACPI_FAILURE(status)) {
		acpi_remove_gpe_handler(NULL, ec->gpe_bit, &acpi_wacom_gpe_handler);
		return_VALUE(-ENODEV);
	}
#endif

	printk("acpi_wacom: getting possible resources\n");
	acpi_get_possible_resources(device->handle, &acpi_buffer);
	printk("acpi_wacom: setting resources\n");
	acpi_set_current_resources(device->handle, &acpi_buffer);
	return_VALUE(AE_OK);
}


static int
acpi_wacom_stop (
	struct acpi_device	*device,
	int			type)
{
	acpi_status		status = AE_OK;
	struct acpi_wacom		*wacom = NULL;
	struct acpi_buffer ret;

	ACPI_FUNCTION_TRACE("acpi_wacom_stop");

	if (!device)
		return_VALUE(-EINVAL);

	wacom = acpi_driver_data(device);

	status = acpi_evaluate_object(device->handle, "_DIS", NULL, &ret);
	if (ACPI_FAILURE(status))
		return_VALUE(-ENODEV);

#if 0
	status = acpi_remove_address_space_handler(ec->handle,
		ACPI_ADR_SPACE_EC, &acpi_wacom_space_handler);
	if (ACPI_FAILURE(status))
		return_VALUE(-ENODEV);

	status = acpi_remove_gpe_handler(NULL, ec->gpe_bit, &acpi_wacom_gpe_handler);
	if (ACPI_FAILURE(status))
		return_VALUE(-ENODEV);
#endif

	return_VALUE(0);
}

int __init
acpi_wacom_init (void)
{
	int			result = 0;

	ACPI_FUNCTION_TRACE("acpi_wacom_init");

	acpi_wacom_dir = proc_mkdir(ACPI_WACOM_CLASS, acpi_root_dir);
	if (!acpi_wacom_dir)
		return_VALUE(-ENODEV);

	result = acpi_bus_register_driver(&acpi_wacom_driver);
	if (result < 0) {
		remove_proc_entry(ACPI_WACOM_CLASS, acpi_root_dir);
		return_VALUE(-ENODEV);
	}

	return_VALUE(0);
}


void __exit
acpi_wacom_exit (void)
{
	ACPI_FUNCTION_TRACE("acpi_wacom_exit");

	acpi_bus_unregister_driver(&acpi_wacom_driver);

	remove_proc_entry(ACPI_WACOM_CLASS, acpi_root_dir);

	return_VOID;
}


module_init(acpi_wacom_init);
module_exit(acpi_wacom_exit);
