/*
 *  gzloop.c: Read-only compressed loop blockdevice
 *
 *  clp file looks like:
 *  [32-bit uncompressed block size: network order]
 *  [32-bit number of blocks (n_blocks): network order]
 *  [64-bit file offsets of start of blocks: native host byte order]
 *    * (n_blocks + 1).
 * n_blocks of:
 *   [compressed block]
 *
 *  Inspired by compressed_loop.c by Paul `Rusty' Russell & Klaus Knopper.
 *
 * Copyright (C) 2003 MandrakeSoft (pixel at mandrakesoft dot com)
 * Redistribution of this file is permitted under the GNU Public License.
 *
 */


#include <linux/module.h>
#include <linux/init.h>
#include <linux/zutil.h>
#include <linux/loop.h>

#define LO_CRYPT_GZ	13

MODULE_LICENSE ("GPL");
MODULE_DESCRIPTION ("Read-only compressed loop blockdevice");
MODULE_AUTHOR("Pascal Rigaux <pixel@mandrakesoft.com>");


#define CLOOP_HEADROOM 128

struct cloop_head {
	char preamble[CLOOP_HEADROOM];
	u_int32_t block_size;
	u_int32_t num_blocks;
};

struct compressed_loop {
	int block_size;
	int num_blocks;

	/* An array of offsets of compressed blocks within the file */
	loff_t *offsets;

	int buffered_blocknum;
	char *buffer;
	char *compressed_buffer;

	z_stream zstream;
	struct semaphore mutex;
};

/*********************************************************************************/
/* taken from crypto_loop */
/*********************************************************************************/
static int file_read(request_queue_t *q, struct file *file, char *buf, int size, loff_t ppos)
{
	mm_segment_t fs;
	int x, y, z;

	y = 0;
	do {
		z = size - y;
		fs = get_fs();
		set_fs(get_ds());
		x = file->f_op->read(file, buf + y, z, &ppos);
		set_fs(fs);

		if (!x)
			return 1;
		
		if (x < 0) {
			if ((x == -EAGAIN) || (x == -ENOMEM) || (x == -ERESTART) || (x == -EINTR)) {
				blk_run_queue(q);
				set_current_state(TASK_INTERRUPTIBLE);
				schedule_timeout(HZ / 2);
				continue;
			}
			return 1;
		}
		y += x;
	} while (y < size);
	return 0;
}
/*********************************************************************************/
/* end of taken from crypto_loop */
/*********************************************************************************/

static int uncompress(struct compressed_loop *cloop, char *dest, unsigned long dest_size,
                      char *source, unsigned long source_size)
{
	/* Most of this code can be found in fs/cramfs/uncompress.c */
	int err;
	cloop->zstream.next_in = source;
	cloop->zstream.avail_in = source_size;
	cloop->zstream.next_out = dest;
	cloop->zstream.avail_out = dest_size;
	err = zlib_inflateReset(&cloop->zstream);
	if (err != Z_OK)
	{
		printk(KERN_ERR "zlib_inflateReset error %d\n", err);
		zlib_inflateEnd(&cloop->zstream); zlib_inflateInit(&cloop->zstream);
	}
	err = zlib_inflate(&cloop->zstream, Z_FINISH);
	if (err != Z_STREAM_END || cloop->zstream.total_out != dest_size) return err;
	return Z_OK;
}

/* This looks more complicated than it is */
static int cloop_load_block(struct loop_device *lo, struct compressed_loop *cloop, int blocknum)
{
	unsigned int buf_length;
	int ret;

	if (blocknum > cloop->num_blocks || blocknum < 0)
	{
		printk(KERN_WARNING "%s: Invalid block number %d requested.\n", lo->lo_file_name, blocknum);
		cloop->buffered_blocknum = -1;
		return -1;
	}

	if (blocknum == cloop->buffered_blocknum) return 0;

	buf_length = cloop->offsets[blocknum + 1] - cloop->offsets[blocknum];

        /* Load one compressed block from the file. */
	file_read(lo->lo_queue, lo->lo_backing_file, cloop->compressed_buffer, buf_length, cloop->offsets[blocknum]);

	/* Do the uncompression */
	ret = uncompress(cloop, cloop->buffer, cloop->block_size, cloop->compressed_buffer, buf_length);

	if (ret != 0)
	{
		printk(KERN_ERR "%s: error %i uncompressing block %u %u/%u "
		       "%Lu-%Lu\n", lo->lo_file_name, ret, blocknum,
		       cloop->block_size, buf_length,
		       cloop->offsets[blocknum], cloop->offsets[blocknum+1]);
		cloop->buffered_blocknum = -1;
		return -1;
	}

	cloop->buffered_blocknum = blocknum;
	return 0;
}


static int cloop_receive(struct loop_device *lo, struct bio_vec *bvec, int bsize, loff_t pos)
{
	int retval=0;
	struct compressed_loop *cloop = lo->key_data;
	char *dest = kmap(bvec->bv_page) + bvec->bv_offset;
	int size = bvec->bv_len;

	/* !! do_div is magical, the following does: */
	/* block_num = pos / cloop->block_size */
	/* offset_in_block = pos % cloop->block_size */
	loff_t block_num = pos;
	int offset_in_block = do_div(block_num, cloop->block_size);

	down_interruptible(&cloop->mutex);

	for (; size > 0; block_num++)
	{
		int used_buffer_size = min(size, cloop->block_size - offset_in_block);

		retval = cloop_load_block(lo, cloop, block_num);
		if (retval) break;

		memcpy(dest, &cloop->buffer[offset_in_block], used_buffer_size);
		offset_in_block = 0; /* not needed anymore */
		dest += used_buffer_size; 
		size -= used_buffer_size; 
	}
	kunmap(bvec->bv_page);
	up(&cloop->mutex);
	return retval;
}

#define cond_vfree(p) if (p) vfree(p); p = NULL

static void cloop_free(struct compressed_loop *cloop)
{
	if (cloop)
	{
		cond_vfree(cloop->zstream.workspace);
		cond_vfree(cloop->compressed_buffer);
		cond_vfree(cloop->buffer);
		cond_vfree(cloop->offsets);
		vfree(cloop);
	}
}

#define vmalloc_or_fail(size) ({ void *p = vmalloc(size); err = -ENOMEM; if (!p) goto cloop_init_failed; p; })

static int cloop_init(struct loop_device *lo, const struct loop_info64 *info)
{
	int err = -EINVAL;
	struct inode *inode = lo->lo_backing_file->f_dentry->d_inode;
	struct compressed_loop *cloop;

	if (!(lo->lo_flags & LO_FLAGS_READ_ONLY))
	{
		printk(KERN_ERR "gzloop can only be used on read-only loop\n");
		return -EINVAL;
	}

	lo->key_data = cloop = vmalloc(sizeof(struct compressed_loop));
	if (!cloop) return -ENOMEM;


	cloop->buffered_blocknum = -1;

	/* ensure cond_vfree works */
	cloop->zstream.workspace = cloop->offsets = NULL;
	cloop->buffer = cloop->compressed_buffer = NULL;

	{ /* read clp header */
		struct cloop_head header;
		if (file_read(lo->lo_queue, lo->lo_backing_file, (char *) &header, sizeof(struct cloop_head), 0) != 0)
			goto cloop_init_failed;
		cloop->block_size = ntohl(header.block_size);
		cloop->num_blocks = ntohl(header.num_blocks);

		if (cloop->block_size % 512 != 0)
		{
			printk(KERN_ERR "%s: blocksize %u not multiple of 512\n",
			       info->lo_file_name, cloop->block_size);
			goto cloop_init_failed;
		}
	}

	{ /* read offsets */
		int size_offsets = sizeof(loff_t) * (cloop->num_blocks + 1);

		if (!S_ISBLK(inode->i_mode) && sizeof(struct cloop_head) + size_offsets > inode->i_size)
		{
			printk(KERN_ERR "%s: file too small for %u blocks\n",
			       info->lo_file_name, cloop->num_blocks);
			goto cloop_init_failed;
		}

		cloop->offsets = vmalloc_or_fail(size_offsets);
		if (file_read(lo->lo_queue, lo->lo_backing_file, (char *) cloop->offsets, size_offsets, sizeof(struct cloop_head)) != 0)
			goto cloop_init_failed;

		if (!S_ISBLK(inode->i_mode) && cloop->offsets[cloop->num_blocks] != inode->i_size)
		{
			printk(KERN_ERR "%s: final offset wrong (%Lu not %Lu)\n",
			       info->lo_file_name, cloop->offsets[cloop->num_blocks], inode->i_size);
			goto cloop_init_failed;
		}
	}

	{ /* search for largest block */
		unsigned long largest_block = 0;
		int i;
		for(i = 0; i < cloop->num_blocks; i++)
			largest_block = max(largest_block, (unsigned long) (cloop->offsets[i+1] - cloop->offsets[i]));

		printk(KERN_INFO "%s: %u blocks, %u bytes/block, largest block is %lu bytes.\n",
		       info->lo_file_name, cloop->num_blocks, cloop->block_size, largest_block);
		
		cloop->compressed_buffer = vmalloc_or_fail(largest_block);
	}

	lo->lo_sizelimit = (loff_t) cloop->num_blocks * cloop->block_size;
	lo->lo_offset = 0;

	cloop->buffer = vmalloc_or_fail(cloop->block_size);
	cloop->zstream.workspace = vmalloc_or_fail(zlib_inflate_workspacesize());
	zlib_inflateInit(&cloop->zstream);

	init_MUTEX(&cloop->mutex);

	return 0;

 cloop_init_failed:
	cloop_free(cloop);
	return err;
}

static int cloop_release(struct loop_device *lo)
{
	struct compressed_loop *cloop = lo->key_data;
	if (cloop)
	{
		cloop_free(cloop);
		return 0;
	} else
	{
		printk(KERN_ERR "gzloop_release(): cloop == NULL ????\n");
		return -EINVAL;
	}
}


static struct loop_func_table gz_funcs = { 
	number: LO_CRYPT_GZ,
	do_receive: cloop_receive,
	init: cloop_init,
	release: cloop_release,
	owner: THIS_MODULE
}; 	


static int __init
init_gzloop(void)
{
	int rc;

	if ((rc = loop_register_transfer (&gz_funcs))) {
		printk (KERN_ERR "gzloop: register loop transfer function failed\n");
		return rc;
	}

	printk (KERN_INFO "gzloop: loaded\n");
	return 0;
}

static void __exit
cleanup_gzloop(void)
{
	if (loop_unregister_transfer (LO_CRYPT_GZ))
		printk (KERN_ERR "gzloop: unregistering transfer funcs failed\n");

	printk (KERN_INFO "gzloop: unloaded\n");
}

module_init(init_gzloop);
module_exit(cleanup_gzloop);
