/* Dazuko RSBAC. Allow RSBAC Linux file access control for 3rd-party applications.
   Copyright (c) 2004 H+BEDV Datentechnik GmbH
   Written by John Ogness <jogness@antivir.de>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "dazuko_rsbac.h"
#include "dazuko_xp.h"
#include "dazukoio.h"

#include <linux/init.h>
#include <linux/unistd.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <asm/uaccess.h>
#include <linux/random.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/vermagic.h>
#endif

#include <rsbac/types.h>
#include <rsbac/reg.h>
#include <rsbac/adf.h>
#include <rsbac/aci.h>
#include <rsbac/getname.h>
#include <rsbac/error.h>   
#include <rsbac/proc_fs.h>  
#include <rsbac/debug.h>  


#if defined(DEVFS_SUPPORT) || LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/devfs_fs_kernel.h>
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
int linux_dazuko_device_read(struct file *file, char *buffer, size_t length, loff_t *pos);
int linux_dazuko_device_write(struct file *file, const char *buffer, size_t length, loff_t *pos);
#else
ssize_t linux_dazuko_device_read(struct file *file, char *buffer, size_t length, loff_t *pos);
ssize_t linux_dazuko_device_write(struct file *file, const char *buffer, size_t length, loff_t *pos);
#endif
int linux_dazuko_device_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long param);
int linux_dazuko_device_open(struct inode *inode, struct file *file);
int linux_dazuko_device_release(struct inode *inode, struct file *file);

extern struct xp_atomic active;

static int			dev_major = -1;
static rsbac_reg_handle_t	handle = -1;

static struct file_operations	fops = {
					read: linux_dazuko_device_read,		/* read */
					write: linux_dazuko_device_write,	/* write */
					ioctl: linux_dazuko_device_ioctl,	/* ioctl */
					open: linux_dazuko_device_open,		/* open */
					release: linux_dazuko_device_release,	/* release */
				};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)

/* The following code is taken directly from Linux in the file:
   include/linux/sched.h */

#ifndef __wait_event_interruptible
#define __wait_event_interruptible(wq, condition, ret)		 \
do {								 \
	struct wait_queue __wait;				 \
								 \
	__wait.task = current;					 \
	add_wait_queue(&wq, &__wait);				 \
	for (;;) {						 \
		current->state = TASK_INTERRUPTIBLE;		 \
		mb();						 \
		if (condition)					 \
			break;					 \
		if (!signal_pending(current)) {			 \
			schedule();				 \
			continue;				 \
		}						 \
		ret = -ERESTARTSYS;				 \
		break;						 \
	}							 \
	current->state = TASK_RUNNING;				 \
	remove_wait_queue(&wq, &__wait);			 \
} while (0)
#endif

#ifndef wait_event_interruptible
#define wait_event_interruptible(wq, condition)			 \
({								 \
	int __ret = 0;						 \
	if (!(condition))					 \
		__wait_event_interruptible(wq, condition, __ret);\
	__ret;							 \
})
#define wait_event(wq, condition)				 \
({								 \
	int __ret = 0;						 \
	if (!(condition))					 \
		__wait_event_interruptible(wq, condition, __ret);\
	__ret;							 \
})
#endif

#endif


/* mutex */

inline int xp_init_mutex(struct xp_mutex *mutex)
{
	#ifdef init_MUTEX
		init_MUTEX(&(mutex->mutex));
	#else
		sema_init(&(mutex->mutex), 1);
	#endif

	return 0;
}

inline int xp_down(struct xp_mutex *mutex)
{
	down(&(mutex->mutex));
	return 0;
}

inline int xp_up(struct xp_mutex *mutex)
{
	up(&(mutex->mutex));
	return 0;
}

inline int xp_destroy_mutex(struct xp_mutex *mutex)
{
	return 0;
}


/* read-write lock */

inline int xp_init_rwlock(struct xp_rwlock *rwlock)
{
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
		rwlock_init(&(rwlock->rwlock));
	#else
		rwlock->rwlock = RW_LOCK_UNLOCKED;
	#endif
	return 0;
}

inline int xp_write_lock(struct xp_rwlock *rwlock)
{
	write_lock(&(rwlock->rwlock));
	return 0;
}

inline int xp_write_unlock(struct xp_rwlock *rwlock)
{
	write_unlock(&(rwlock->rwlock));
	return 0;
}

inline int xp_read_lock(struct xp_rwlock *rlock)
{
	read_lock(&(rlock->rwlock));
	return 0;
}

inline int xp_read_unlock(struct xp_rwlock *rlock)
{
	read_unlock(&(rlock->rwlock));
	return 0;
}

inline int xp_destroy_rwlock(struct xp_rwlock *rwlock)
{
	return 0;
}


/* wait-notify queue */

inline int xp_init_queue(struct xp_queue *queue)
{
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
		init_waitqueue_head(&(queue->queue));
	#else
		queue = NULL;
	#endif

	return 0;
}

inline int xp_wait_until_condition(struct xp_queue *queue, int (*cfunction)(void *), void *cparam, int allow_interrupt)
{
	/* wait until cfunction(cparam) != 0 (condition is true) */

	if (allow_interrupt)
	{
		return wait_event_interruptible(queue->queue, cfunction(cparam) != 0);
	}
	else
	{
		wait_event(queue->queue, cfunction(cparam) != 0);
	}

	return 0;
}

inline int xp_notify(struct xp_queue *queue)
{
	wake_up(&(queue->queue));
	return 0;
}

inline int xp_destroy_queue(struct xp_queue *queue)
{
	return 0;
}


/* memory */

inline void* xp_malloc(size_t size)
{
	return kmalloc(size, GFP_KERNEL);
}

inline int xp_free(void *ptr)
{
	kfree(ptr);
	return 0;
}

inline int xp_copyin(const void *user_src, void *kernel_dest, size_t size)
{
	return copy_from_user(kernel_dest, user_src, size);
}

inline int xp_copyout(const void *kernel_src, void *user_dest, size_t size)
{
	return copy_to_user(user_dest, kernel_src, size);
}

inline int xp_verify_user_writable(const void *user_ptr, size_t size)
{
	#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
		return verify_area(VERIFY_WRITE, user_ptr, size);
	#else
		return 0;
	#endif
}

inline int xp_verify_user_readable(const void *user_ptr, size_t size)
{
	#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
		return verify_area(VERIFY_READ, user_ptr, size);
	#else
		return 0;
	#endif
}


/* path attribute */

inline int xp_is_absolute_path(const char *path)
{
	return (path[0] == '/');
}


/* atomic */

inline int xp_atomic_set(struct xp_atomic *atomic, int value)
{
	atomic_set(&(atomic->atomic), value);
	return 0;
}

inline int xp_atomic_inc(struct xp_atomic *atomic)
{
	#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		#ifdef MODULE
			if (atomic == &active)
				MOD_INC_USE_COUNT;
		#endif
	#endif

	atomic_inc(&(atomic->atomic));
	return 0;
}

inline int xp_atomic_dec(struct xp_atomic *atomic)
{
	#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		#ifdef MODULE
			if (atomic == &active)
				MOD_DEC_USE_COUNT;
		#endif
	#endif

	atomic_dec(&(atomic->atomic));
	return 0;
}

inline int xp_atomic_read(struct xp_atomic *atomic)
{
	return atomic_read(&(atomic->atomic));
}


/* file descriptor */

inline int xp_copy_file(struct xp_file *dest, struct xp_file *src)
{
	return 0;
}

inline int xp_compare_file(struct xp_file *file1, struct xp_file *file2)
{
	return 0;
}

inline int xp_file_struct_check(struct dazuko_file_struct *dfs)
{
	int	length;

	/* make sure we have access to everything */

	if (dfs == NULL)
		return -1;

	if (dfs->extra_data == NULL)
		return -1;

	if (dfs->extra_data->dentry == NULL)
		return -1;

	if (dfs->extra_data->dentry->d_inode == NULL)
		return -1;

	/* ok, we have everything we need */

	length = rsbac_get_full_path_length(dfs->extra_data->dentry);
	if (length < 1)
		return -1;

	dfs->extra_data->full_filename = xp_malloc(length + 1);
	if (dfs->extra_data->full_filename == NULL)
		return -1;

	/* the full_filename will need to be deleted later */
	dfs->extra_data->free_full_filename = 1;

	if (rsbac_get_full_path(dfs->extra_data->dentry, dfs->extra_data->full_filename, length + 1) < 1)
		return -1;

	/* find the actual value of the length */
	dfs->extra_data->full_filename_length = dazuko_get_filename_length(dfs->extra_data->full_filename);

	/* reference copy of full path */
	dfs->filename = dfs->extra_data->full_filename;

	dfs->filename_length = dfs->extra_data->full_filename_length;

	dfs->file_p.size = dfs->extra_data->dentry->d_inode->i_size;
	dfs->file_p.set_size = 1;
	dfs->file_p.uid = dfs->extra_data->dentry->d_inode->i_uid;
	dfs->file_p.set_uid = 1;
	dfs->file_p.gid = dfs->extra_data->dentry->d_inode->i_gid;
	dfs->file_p.set_gid = 1;
	dfs->file_p.mode = dfs->extra_data->dentry->d_inode->i_mode;
	dfs->file_p.set_mode = 1;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	dfs->file_p.device_type = dfs->extra_data->dentry->d_inode->i_dev;
#else
	dfs->file_p.device_type = dfs->extra_data->dentry->d_inode->i_rdev;
#endif
	dfs->file_p.set_device_type = 1;

	return 0;
}

int xp_file_struct_check_cleanup(struct dazuko_file_struct *dfs)
{
	return 0;
}

static int dazuko_file_struct_cleanup(struct dazuko_file_struct **dfs)
{
	if (dfs == NULL)
		return 0;

	if (*dfs == NULL)
		return 0;

	xp_file_struct_check_cleanup(*dfs);

	if ((*dfs)->extra_data != NULL)
	{
		if ((*dfs)->extra_data->free_full_filename)
			xp_free((*dfs)->extra_data->full_filename);

		xp_free((*dfs)->extra_data);
	}

	xp_free(*dfs);

	*dfs = NULL;

	return 0;
}


/* daemon id */

int xp_id_compare(struct xp_daemon_id *id1, struct xp_daemon_id *id2)
{
	if (id1 == NULL || id2 == NULL)
		return -1;

	/* if file's are available and they match,
	 * then we say that the id's match */
	if (id1->file != NULL && id1->file == id2->file)
		return 0;

	if (id1->pid == id2->pid)
		return 0;

	return 1;
}

int xp_id_free(struct xp_daemon_id *id)
{
	xp_free(id);

	return 0;
}

struct xp_daemon_id* xp_id_copy(struct xp_daemon_id *id)
{
	struct xp_daemon_id	*ptr;

	if (id == NULL)
		return NULL;

	ptr = (struct xp_daemon_id *)xp_malloc(sizeof(struct xp_daemon_id));

	if (ptr != NULL)
	{
		ptr->pid = id->pid;
		ptr->file = id->file;
	}

	return ptr;
}


/* system calls */

static inline int dazuko_rsbac_process(enum rsbac_adf_request_t request,	/* open, close, exec, etc */
					enum rsbac_target_t target,		/* file, dir, etc */
					union rsbac_target_id_t tid,		/* dentry (for files) */
					int is_pre)
{
	struct dazuko_file_struct *dfs = NULL;
	struct xp_daemon_id xp_id;
	int error = 0;
	int check_error = 0;
	struct event_properties event_p;
	int event;
	int daemon_allowed;

	switch (request)
	{
		case R_DELETE:
			if (target == T_FILE)
			{
				event = DAZUKO_ON_UNLINK;
				daemon_allowed = 1;
			}
			else if (target == T_DIR)
			{
				event = DAZUKO_ON_RMDIR;
				daemon_allowed = 1;
			}
			else
			{
				goto dazuko_rsbac_process_quickout;
			}
			break;

		case R_CLOSE:
			if (target == T_FILE)
			{
				event = DAZUKO_ON_CLOSE;
				daemon_allowed = 1;
			}
			else
			{
				goto dazuko_rsbac_process_quickout;
			}
			break;

		case R_EXECUTE:
			if (target == T_FILE)
			{
				event = DAZUKO_ON_EXEC;
				daemon_allowed = 0;
			}
			else
			{
				goto dazuko_rsbac_process_quickout;
			}
			break;

		case R_APPEND_OPEN:
		case R_READ_WRITE_OPEN:
		case R_READ_OPEN:
		case R_WRITE_OPEN:
			if (target == T_FILE)
			{
				event = DAZUKO_ON_OPEN;
				daemon_allowed = 1;
			}
			else
			{
				goto dazuko_rsbac_process_quickout;
			}
			break;

		default:
			goto dazuko_rsbac_process_quickout;
	}

	xp_id.pid = current->pid;
	xp_id.file = NULL;

	check_error = dazuko_sys_check(event, daemon_allowed, &xp_id);

	if (!check_error)
	{
		dazuko_bzero(&event_p, sizeof(event_p));
/*
		event_p.flags = flags;
		event_p.set_flags = 1;
		event_p.mode = mode;
		event_p.set_mode = 1;
*/
		event_p.pid = current->pid;
		event_p.set_pid = 1;
		event_p.uid = current->uid;
		event_p.set_uid = 1;

		dfs = (struct dazuko_file_struct *)xp_malloc(sizeof(struct dazuko_file_struct));
		if (dfs != NULL)
		{
			dazuko_bzero(dfs, sizeof(struct dazuko_file_struct));

			dfs->extra_data = (struct xp_file_struct *)xp_malloc(sizeof(struct xp_file_struct));
			if (dfs->extra_data != NULL)
			{
				dazuko_bzero(dfs->extra_data, sizeof(struct xp_file_struct));

				dfs->extra_data->dentry = tid.file.dentry_p;

				if (is_pre)
					error = dazuko_sys_pre(event, dfs, &event_p);
				else
					dazuko_sys_post(event, dfs, NULL, &event_p);
			}
			else
			{
				xp_free(dfs);
				dfs = NULL;
			}

			dazuko_file_struct_cleanup(&dfs);
		}
	}

	if (!is_pre)
		return 0;

	if (error)
		return NOT_GRANTED;

	return GRANTED;

dazuko_rsbac_process_quickout:
	if (is_pre)
		return DO_NOT_CARE;

	return 0;
}

static inline int dazuko_rsbac_pre(enum rsbac_adf_request_t request,		/* open, close, exec, etc */
					rsbac_pid_t owner_pid,			/* some PID */
					enum rsbac_target_t target,		/* file, dir, etc */
					union rsbac_target_id_t tid,		/* dentry (for files) */
					enum rsbac_attribute_t attr,		/* keine ahnung */
					union rsbac_attribute_value_t attr_val,	/* tons of rsbac attribs */
					rsbac_uid_t owner)			/* some UID */
{
	return dazuko_rsbac_process(request, target, tid, 1);
}


static int dazuko_rsbac_post(enum rsbac_adf_request_t request,
				rsbac_pid_t owner_pid,
				enum rsbac_target_t target,
				union rsbac_target_id_t tid,
				enum  rsbac_target_t new_target,
				union rsbac_target_id_t new_tid,
				enum  rsbac_attribute_t attr,
				union rsbac_attribute_value_t attr_val,
				rsbac_uid_t owner)
{
	return dazuko_rsbac_process(request, target, tid, 0);
}


/* system hook */

inline int xp_sys_hook()
{
#define RSBAC_REG_REGISTER_RETRIES 1024
	/* Called insmod when inserting the module. */

	int retries;

	/* register with rsbac */

	struct rsbac_reg_entry_t entry;

	/* clearing registration entries */
	memset(&entry, 0, sizeof(entry));

	strcpy(entry.name, "Dazuko");
	entry.request_func = dazuko_rsbac_pre;
	entry.set_attr_func = dazuko_rsbac_post;
	entry.switch_on = TRUE;

	for (retries=0 ; retries<RSBAC_REG_REGISTER_RETRIES ; retries++)
	{
		/* get a random handle */
		get_random_bytes(&handle, sizeof(handle));

		/* the handle needs to be positive */
		if (handle < 0)
			handle *= -1;

		entry.handle = handle;

		handle = rsbac_reg_register(RSBAC_REG_VERSION, entry);

		if (handle != -RSBAC_EEXISTS)
		{
			/* there wasn't a problem with the handle value */
			break;
		}
	}

	if (handle < 0)
	{
		xp_print("dazuko: failed to register with RSBAC\n");
		return -ENOEXEC;
	}

	/* register the dazuko device */
#ifdef DEVFS_SUPPORT
	dev_major = devfs_register_chrdev(0, DEVICE_NAME, &fops);
	devfs_register(NULL, DEVICE_NAME, DEVFS_FL_DEFAULT,
		dev_major, 0, S_IFCHR | S_IRUSR | S_IWUSR,
		&fops, NULL);
#else
	dev_major = register_chrdev(0, DEVICE_NAME, &fops);
#endif
	if (dev_major < 0)
	{
		xp_print("dazuko: unable to register device chrdev, err=%d\n", dev_major);
		return dev_major;
	}

	/* initialization complete */

	return 0;
}

inline int xp_sys_unhook()
{
	/* Called by rmmod when removing the module. */

	int	error;

#ifdef DEVFS_SUPPORT
	error = devfs_unregister_chrdev(dev_major, DEVICE_NAME);
	devfs_unregister(devfs_find_handle(NULL, DEVICE_NAME, dev_major, 0, DEVFS_SPECIAL_CHR, 0));
#else
	error = unregister_chrdev(dev_major, DEVICE_NAME);
#endif
	if (error < 0)
	{
		xp_print("dazuko: error unregistering chrdev, err=%d\n", error);
	}

	if (rsbac_reg_unregister(handle) != 0)
	{
		xp_print("dazuko: failed to unregister with RSBAC (system may be left in an unstable state!)\n");
	}

	return 0;
}


/* output */

int xp_print(const char *fmt, ...)
{
	va_list args;
	char *p;
	size_t size = 1024;

	p = (char *)xp_malloc(size);
	if (!p)
		return -1;

	va_start(args, fmt);
	dazuko_vsnprintf(p, size-1, fmt, args);
	va_end(args);

	p[size-1] = 0;

	printk(p);
	rsbac_printk(p);

	xp_free(p);

	return 0;
}


/* ioctl's */

int linux_dazuko_device_open(struct inode *inode, struct file *file)
{
	DPRINT(("dazuko: linux_dazuko_device_open() [%d]\n", current->pid));

	return 0;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
int linux_dazuko_device_read(struct file *file, char *buffer, size_t length, loff_t *pos)
#else
ssize_t linux_dazuko_device_read(struct file *file, char *buffer, size_t length, loff_t *pos)
#endif
{
	/* Reading from the dazuko device simply
	 * returns the device number. This is to
	 * help out the daemon. */

	char	tmp[20];
	size_t	dev_major_len;

	DPRINT(("dazuko: linux_dazuko_device_read() [%d]\n", current->pid));

	/* non-root daemons are ignored */
	if (current->uid != 0)
		return 0;

	/* only one read is allowed */
	if (*pos != 0)
		return 0;

	if (dev_major < 0)
		return -ENODEV;

	/* print dev_major to a string
	 * and get length (with terminator) */
	dazuko_bzero(tmp, sizeof(tmp));

	dev_major_len = dazuko_snprintf(tmp, sizeof(tmp), "%d", dev_major) + 1;

	if (tmp[sizeof(tmp)-1] != 0)
	{
		xp_print("dazuko: failing device_read, device number overflow for dameon %d (dev_major=%d)\n", current->pid, dev_major);
		return -EFAULT;
	}

	if (length < dev_major_len)
		return -EINVAL;

	/* copy dev_major string to userspace */
	if (xp_copyout(tmp, buffer, dev_major_len) != 0)
		return -EFAULT;

	*pos = dev_major_len;

	return dev_major_len;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
int linux_dazuko_device_write(struct file *file, const char *buffer, size_t length, loff_t *pos)
#else
ssize_t linux_dazuko_device_write(struct file *file, const char *buffer, size_t length, loff_t *pos)
#endif
{
	struct dazuko_request	*u_request;
	struct xp_daemon_id	xp_id;
	char			tmpbuffer[32];
	char			*value;
	int			size;

	/* non-root daemons are ignored */
	if (current->uid != 0)
		return length;

	size = length;
	if (length >= sizeof(tmpbuffer))
		size = sizeof(tmpbuffer) -1;

	/* copy request pointer string to kernelspace */
	if (xp_copyin(buffer, tmpbuffer, size) != 0)
		return -EFAULT;

	tmpbuffer[size] = 0;

	if (dazuko_get_value("\nRA=", buffer, &value) != 0)
	{
		xp_print("dazuko: error: linux_dazuko_device_write.RA missing\n");
		return -EFAULT;
	}

	u_request = (struct dazuko_request *)simple_strtoul(value, NULL, 10);

	xp_free(value);

	xp_id.pid = current->pid;
	xp_id.file = file;

	if (dazuko_handle_user_request(u_request, &xp_id) == 0)
		return length;
	else
		return -EINTR;
}

int linux_dazuko_device_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long param)
{
	/* A daemon uses this function to interact with
	 * the kernel. A daemon can set scanning parameters,
	 * give scanning response, and get filenames to scan. */

	struct xp_daemon_id	xp_id;
	int			error = 0;

	/* non-root daemons are ignored */
	if (current->uid != 0)
		return 0;

	if (param == 0)
	{
		xp_print("dazuko: error: linux_dazuko_device_ioctl(..., 0)\n");
		return -EFAULT;
	}

	xp_id.pid = current->pid;
	xp_id.file = file;

	error = dazuko_handle_user_request_compat12((void *)param, _IOC_NR(cmd), &xp_id);

	if (error != 0)
	{
		/* general error occurred */

		return -EPERM;
	}

	return error;
}

int linux_dazuko_device_release(struct inode *inode, struct file *file)
{
	struct xp_daemon_id	xp_id;

	DPRINT(("dazuko: dazuko_device_release() [%d]\n", current->pid));

	/* non-root daemons are ignored */
	if (current->uid != 0)
		return 0;

	xp_id.pid = current->pid;
	xp_id.file = file;

	return dazuko_unregister_daemon(&xp_id);
}


/* init/exit */

int __init linux_dazuko_init(void)
{
	return dazuko_init();
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
void __exit linux_dazuko_exit(void)
#else
void linux_dazuko_exit(void)
#endif
{
	dazuko_exit();
}


#ifdef MODULE

int init_module(void)
{
	return linux_dazuko_init();
}

void cleanup_module(void)
{
	linux_dazuko_exit();
}
        
MODULE_AUTHOR("H+BEDV Datentechnik GmbH <linux_support@antivir.de>");
MODULE_DESCRIPTION("allow 3rd-party file access control");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#else
static const char __module_license[] __attribute__((section(".modinfo"))) = "license=GPL";
#endif   
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
MODULE_INFO(vermagic, VERMAGIC_STRING);
#endif
                
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
EXPORT_NO_SYMBOLS;
#endif

#else
         
module_init(linux_dazuko_init);
module_exit(linux_dazuko_exit);
/* module_init(int linux_dazuko_init(void)); */
/* module_exit(void linux_dazuko_exit(void)); */

#endif

