/*************************************************** */
/* Rule Set Based Access Control                     */
/* Implementation of the Access Control Decision     */
/* Facility (ADF) - JAIL module                      */
/* File: rsbac/adf/jail/syscalls.c                   */
/*                                                   */
/* Author and (c) 1999-2004: Amon Ott <ao@rsbac.org> */
/*                                                   */
/* Last modified: 27/Apr/2004                        */
/*************************************************** */

#include <linux/string.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/version.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/syscalls.h>
#include <linux/file.h>
#endif
#include <rsbac/types.h>
#include <rsbac/aci.h>
#include <rsbac/error.h>
#include <rsbac/rkmem.h>
#include <rsbac/debug.h>
#include <rsbac/helpers.h>
#include <rsbac/getname.h>
#include <rsbac/network.h>
#include <asm/uaccess.h>

/************************************************* */
/*           Global Variables                      */
/************************************************* */

static rsbac_jail_id_t next_id = 1;

/************************************************* */
/*          Internal Help functions                */
/************************************************* */

/************************************************* */
/*          Externally visible functions           */
/************************************************* */

extern long sys_chroot(const char * filename);
extern long sys_chdir(const char * filename);

/* Create a jail for current process */
/* Note: It is allowed to create jails within jails, but with restrictions */
int rsbac_jail_sys_jail(rsbac_version_t version,
                        char * path,
                        rsbac_jail_ip_t ip,
                        rsbac_jail_flags_t flags,
                        rsbac_cap_vector_t max_caps)
  {
    union rsbac_target_id_t i_tid;
    union rsbac_attribute_value_t i_attr_val1;
    int err = 0;
    int chk_addr_ret;

    if(version != RSBAC_JAIL_VERSION)
      return -RSBAC_EINVALIDVERSION;
    chk_addr_ret = inet_addr_type(ip);
    if (ip != INADDR_ANY &&
        chk_addr_ret != RTN_LOCAL &&
        chk_addr_ret != RTN_MULTICAST &&
        chk_addr_ret != RTN_BROADCAST)
      return -EADDRNOTAVAIL;

#if 0
    err = rsbac_get_owner(&i_tid.user);
    if(err)
      return err;
    if (rsbac_get_attr(JAIL,
                       T_USER,
                       i_tid,
                       A_jail_role,
                       &i_attr_val1,
                       TRUE))
      {
        rsbac_ds_get_error("rsbac_jail_sys_jail()", A_jail_role);
      }
    /* if user role, then deny */
    if (i_attr_val1.system_role == SR_user)
      return(-EPERM);
#endif

    /* Get jail_id for this process */
    i_tid.process = current->pid;
    if (rsbac_get_attr(JAIL,
                       T_PROCESS,
                       i_tid,
                       A_jail_id,
                       &i_attr_val1,
                       TRUE))
      {
        rsbac_ds_get_error("rsbac_jail_sys_jail()", A_jail_id);
        return(-RSBAC_EREADFAILED);
      }
    if(i_attr_val1.jail_id)
      { /* this process is already in a jail -> limit ip and flags */
        if (rsbac_get_attr(JAIL,
                           T_PROCESS,
                           i_tid,
                           A_jail_flags,
                           &i_attr_val1,
                           TRUE))
          {
            rsbac_ds_get_error("rsbac_jail_sys_jail()", A_jail_flags);
            return(-RSBAC_EREADFAILED);
          }
        flags &= i_attr_val1.jail_flags;
        if (rsbac_get_attr(JAIL,
                           T_PROCESS,
                           i_tid,
                           A_jail_ip,
                           &i_attr_val1,
                           TRUE))
          {
            rsbac_ds_get_error("rsbac_jail_sys_jail()", A_jail_ip);
            return(-RSBAC_EREADFAILED);
          }
        if(i_attr_val1.jail_ip)
          ip = i_attr_val1.jail_ip;
        if (rsbac_get_attr(JAIL,
                           T_PROCESS,
                           i_tid,
                           A_jail_max_caps,
                           &i_attr_val1,
                           TRUE))
          {
            rsbac_ds_get_error("rsbac_jail_sys_jail()", A_jail_max_caps);
            return(-RSBAC_EREADFAILED);
          }
        max_caps &= i_attr_val1.jail_max_caps;
      }

    if(path)
      {
        mm_segment_t                    oldfs;
	struct file * file;
	struct files_struct *files = current->files;
	int fd;

        err = sys_chroot(path);
        if(err)
          return err;
        /* Set current user space to kernel space, because sys_chdir() takes name */
        /* from user space */
        oldfs = get_fs();
        set_fs(KERNEL_DS);
        err = sys_chdir("/");
        /* Set current user space back to user space */
        set_fs(oldfs);

restart:
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	spin_lock(&files->file_lock);
#else
	read_lock(&files->file_lock);
#endif
	for(fd=0; fd < files->max_fds; fd++)
	  {
	    file = files->fd[fd];
	    if(   file
	       && file->f_dentry
	       && file->f_dentry->d_inode
	       && S_ISDIR(file->f_dentry->d_inode->i_mode)
	      )
	      {
                char * filename;

                #ifdef CONFIG_RSBAC_LOG_FULL_PATH
                filename = rsbac_kmalloc(CONFIG_RSBAC_MAX_PATH_LEN + 4);
                if(filename)
                  rsbac_get_full_path(file->f_dentry, filename, CONFIG_RSBAC_MAX_PATH_LEN);
                #else
                filename = rsbac_kmalloc(RSBAC_MAXNAMELEN + 4);
                if(filename)
                  rsbac_get_full_path(file->f_dentry, filename, RSBAC_MAXNAMELEN);
                #endif

#ifdef CONFIG_RSBAC_RMSG
                rsbac_printk(KERN_INFO
                             "rsbac_jail_sys_jail(): avoid possible chroot breakout by closing open dir fd %u, inode %u, device %02u:%02u, path %s\n",
                             fd,
                             file->f_dentry->d_inode->i_ino,
                             MAJOR(file->f_dentry->d_sb->s_dev),
                             MINOR(file->f_dentry->d_sb->s_dev),
                             filename);
#endif
#ifdef CONFIG_RSBAC_RMSG_NOSYSLOG
                if (!rsbac_nosyslog)
#endif
                printk(KERN_INFO
                       "rsbac_jail_sys_jail(): avoid possible chroot breakout by closing open dir fd %u, inode %lu, device %02u:%02u, path %s\n",
                       fd,
                       file->f_dentry->d_inode->i_ino,
                       MAJOR(file->f_dentry->d_sb->s_dev),
                       MINOR(file->f_dentry->d_sb->s_dev),
                       filename);
                if(filename)
                  rsbac_kfree(filename);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
                spin_unlock(&files->file_lock);
#else
                read_unlock(&files->file_lock);
#endif
	        sys_close(fd);
	        goto restart;
	      }
	  }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	spin_unlock(&files->file_lock);
#else
	read_unlock(&files->file_lock);
#endif
      }

    /* Set jail_id for this process */
    i_attr_val1.jail_id = next_id++;
    if (rsbac_set_attr(JAIL,
                       T_PROCESS,
                       i_tid,
                       A_jail_id,
                       i_attr_val1))
      {
        rsbac_ds_set_error("rsbac_jail_sys_jail()", A_jail_id);
        return(-RSBAC_EWRITEFAILED);
      }
    /* Set jail_ip for this process */
    i_attr_val1.jail_ip = ip;
    if (rsbac_set_attr(JAIL,
                       T_PROCESS,
                       i_tid,
                       A_jail_ip,
                       i_attr_val1))
      {
        rsbac_ds_set_error("rsbac_jail_sys_jail()", A_jail_ip);
        return(-RSBAC_EWRITEFAILED);
      }
    /* Set jail_flags for this process */
    i_attr_val1.jail_flags = flags;
    if (rsbac_set_attr(JAIL,
                       T_PROCESS,
                       i_tid,
                       A_jail_flags,
                       i_attr_val1))
      {
        rsbac_ds_set_error("rsbac_jail_sys_jail()", A_jail_flags);
        return(-RSBAC_EWRITEFAILED);
      }
    /* Set jail_max_caps for this process */
    i_attr_val1.jail_max_caps = max_caps;
    if (rsbac_set_attr(JAIL,
                       T_PROCESS,
                       i_tid,
                       A_jail_max_caps,
                       i_attr_val1))
      {
        rsbac_ds_set_error("rsbac_jail_sys_jail()", A_jail_flags);
        return(-RSBAC_EWRITEFAILED);
      }
    return err;
  }

/* end of rsbac/adf/jail/syscalls.c */
